      program p3dr
C     Parallel 3D Cartesian Reconstruction program
C                                                     MAIN PROGRAM: P3DR
C **********************************************************************
C *                                                                    *
C * Main routine: P3DR in file aamain.f                                *
C * Author: Robert E. Lynch (rel@cs.purdue.edu) and                    *
C *         Hong Lin (linh@cs.purdue.edu)                              *
C *         Yongchang Ji (yji@cs.ucf.edu)                              *
C * Date:   1997 -- 2002                                               *
C * Function: Main program for cartesian reconstruction                *
C * Algorithm: P3DR uses a set of pixel frames of projections of       *
C *   an object to reconstruct the 3D density map of the object.       *
C *   Firstly, a 2D Fourier transform is carried out on each           *
C *   pixel frame; then the 2D DFT is used in estimating 3D DFT        *
C *   of the object; finally, a reverse Fourier transform is           *
C *   carried out on the 3D DFT to get the density values of the       *
C *   object.                                                          *
C *      P3DR reads in orientations and pixel frames from two          *
C *      files with extension .dat and .pif and carries out            *
C *      reconstruction based on these data.                           *
C * Modification history:                                              *
C *   - Robert E. Lynch, Version 0.0, 1997                           *
C *   - Hong Lin, Version 0.1, Apr.20, 1998                          *
C *   - Hong Lin, Version 0.6, Jul.6, 1998                           *
C *   - Hong Lin, Version 1.0, Aug.18, 1998                          *
C *   - Hong Lin, Version 2.0, Sep.29, 1999.                         *
C *   - Hong Lin, Version 2.9, June 17, 2000                           *
C *   - R.E. Lynch, Aug 2000                                           *
C *        Program name set to P3DR                                    *
C *        Moved print of time and memory-use to subroutine MEM_TIME   *
C *        from main routine --- THIS DOESN"T WORK !!!!!!              *
C *        The setting of the value of  diameter  was removed --- the  *
C *          number of angstroms per pixels is not known until the     *
C *          header of the orientation file is read.                   *
C *        Used new variable  interpolants_size  to store the size     *
C *          of the array  interpolants  which is passed to several    *
C *          routines.  Simplified the routine  intlz_interp           *
C *                                                                    *
C *   - Yongchang Ji, Final Version, Aug.2000 -- Jan. 2002             *
C *                                                                    *
C **********************************************************************

C confine_3D -- When confine_3D is .true., the edge-length of 3D
C   DFT is set to the same as the edge-length of pixel frames. 
C 
      implicit none

      character(len=256) :: infile

#ifdef MPI
      include 'mpif.h'
#endif

      include '../include/allprogs.inc'
      include '../include/info.inc'
      include '../include/com_01.inc'
      include '../include/mem_time.inc'
      include '../include/pifhead.inc'

      type(pifhead_t) imghead

      logical :: confine_3D = .false.

      integer :: mcw, inunit, err_code, alloc_stat
      integer first_iq, last_iq, max_iq_slab,
     +     first_z,last_z, first_hk,last_hk, max_hk_slab, num_hkl,
     +     den_min_x, den_max_x, den_min_y, den_max_y, 
     +     den_min_z, den_max_z, mem_size, 
     +     min_xyz, max_xyz, hkl_max, max_z_slab,
     +     first_rr_out, last_rr_out, max_rr_out_slab,
     +     first_u, last_u, max_u_slab,
     +     first_y, last_y, max_y_slab,
     +     interpolants_size
      integer i, j, id, nptles, ptle_count(256)
      integer nrffts_1_row, nrffts_1_col, nrffts_1_omega,
     $     nrffts_2_row, nrffts_2_col, nrffts_2_omega

C     Variables used for noise suppression
      integer nfsc, locate
      real convert_to_invang
      real, allocatable :: fsc_freq(:), fsc_value(:)

      double precision elapsed_time
      
      integer*8 clock1, clock2

      character(len=18) :: subname = 'P3DR main routine'
      
      real map_min, map_max, map_min_local, map_max_local
      real enlargefact, imgprjbig, imgprjstd

      integer, allocatable   :: indx(:)
      real, allocatable      :: pixels(:), x(:), y(:),mag(:)
      real, allocatable      :: theta(:), phi(:), omega(:)
      real, allocatable      :: delf_maj_ptle(:), delf_min_ptle(:)
      real, allocatable      :: ang_maj_ptle(:)
      real, allocatable      :: f_vru(:), f_vru_tmp(:), F_XK_dZ(:)
      real, allocatable      :: rho(:,:,:), density_map(:,:,:)
      real, allocatable      :: ctf_val(:,:), abs_ctf(:,:)
      complex, allocatable   :: f_hkl(:)
      double precision, allocatable :: interpolants(:), f_hkl_ref(:)

      call system_clock(clock1)

#ifdef MPI
      mcw = mpi_comm_world
      call MPI_INIT (mpi_err)
      if (mpi_err.ne.0) call error_stop(progname, subname, 
     +     'Error in MPI_INIT')
      call MPI_COMM_RANK (MCW, iam, mpi_err)
      if (mpi_err.ne.0) call error_stop(progname, subname, 
     +     'Error in MPI_COMM_RANK')
      call MPI_COMM_SIZE (MCW, num_nodes, mpi_err)
      if (mpi_err.ne.0) call error_stop(progname, subname, 
     +     'Error in MPI_COMM_SIZE')
      if (num_nodes > max_numnodes) then
         write(*, 9001) num_nodes, max_numnodes
 9001    format(//' Requested number of nodes = NUM_NODES    = ',i10/
     +        ' Available number of nodes = MAX_NUMNODES = ',i10/)
         call error_stop(progname, subname,  
     +        'NUM_NODES requested exceeds MAX_NUMNODES')
      endif
#else
      iam = 0
      num_nodes = 1
#endif
      
      progname = 'P3DR'
      
      if( iam == 0 ) then
         write(*,*) 'P3DR: Parallel 3D reconstruction'
         write(*,*) 'Parallel version run on ', num_nodes , ' CPUs'
         write(*,*) 'For doc see: http://cryoem.ucsd.edu/programs.shtm'
         write(*,*) 
         
         call parse_cmd_line(infile)
         
         if(len_trim(infile) == 0) then
            inunit = 5
            call info(inunit)
         else
            inunit=17
            open(file=infile, unit=inunit)
            call info(inunit)
            close(inunit)
         endif
      endif
      
#ifdef MPI
C     Broadcast parameters from root to all other processes
      call bcast_parameters 
      call mpi_bcast(oddeven, 1, mpi_integer, 0, mcw, mpi_err)
#endif
      
C     Process FSC file
      if(iam == 0) then
         nfsc = 0
         if(fsc_file_name(1:1) == ' ') then
            write(*,*)
            write(*,*) 'No FSC file provided - no noise suppression'
            write(*,*)
         else
            write(*,*)
            write(*,*) 'Noise suppression w/ FSC file = ',
     $           trim(fsc_file_name)
            write(*,*)
            open(unit=21, file=fsc_file_name)
            read(21,*) nfsc
         endif
      endif

#ifdef MPI
      call mpi_bcast(nfsc, 1, mpi_integer, 0, mcw, mpi_err)
#endif      

      allocate(fsc_freq(nfsc), fsc_value(nfsc))      
      if(iam == 0 .and. nfsc > 0) then
         do i=1,nfsc
            read(21,*) fsc_freq(i), fsc_value(i)
         enddo
         close(21)
         write(*,*) 'Number of FSC values = ', nfsc
         write(*,*) 'i   spatial freq      FSC'
         do i=1,nfsc
            write(*,'(i3,4x,f7.4,6x,f7.4)') 
     $           i, fsc_freq(i), fsc_value(i)
         enddo
         write(*,*)
      endif
      
#ifdef MPI      
      call mpi_bcast(fsc_freq,  nfsc, mpi_real, 0, mcw, mpi_err)
      call mpi_bcast(fsc_value, nfsc, mpi_real, 0, mcw, mpi_err)
#endif

      call intlz_params(hkl_max, num_hkl, min_xyz, max_xyz, 
     +     nrffta_row, nrffta_col, nrffta_omega, nfc_2d_row, 
     +     nfc_2d_col, nrffts_1_row, nrffts_1_col, 
     +     nrffts_1_omega, nrffts_2_row, nrffts_2_col, 
     +     nrffts_2_omega, den_min_x, den_max_x, den_min_y, 
     +     den_max_y, den_min_z, den_max_z,
     +     confine_3D)
      
      if( iam == 0 ) then
         print *, ' zero_fill after  call of INTLZ_PARAMS =',zero_fill
         print *, ' nrffta_row after call of INTLZ_PARAMS =',nrffta_row
      endif
      
      call intlz_arrays(hkl_max, min_xyz, max_xyz, min_z, max_z,
     +     first_hk, last_hk, max_hk_slab, first_iq, last_iq,
     +     max_iq_slab, first_rr_out, last_rr_out, max_rr_out_slab,
     +     first_u, last_u, max_u_slab,
     +     first_y, last_y, max_y_slab, first_z, last_z, max_z_slab,
     +     npix_row)      
      
C     **********************************************************************
C     *             2D DFT and estimating 3D DFT from 2D DFT               *
C     **********************************************************************
      
      interpolants_size = (hkl_max +1)*((2*hkl_max + 1)**2) * 3
      
      allocate(ctf_val(0:nfc_2d_row, -nfc_2d_col:nfc_2d_col), 
     +     abs_ctf(0:nfc_2d_row, -nfc_2d_col:nfc_2d_col),
     +     interpolants(interpolants_size),
     +     stat = alloc_stat)
      
      if(alloc_stat > 0) then
         call error_stop (progname, subname,  
     +        'Memory allocation ctf_val, abs_ctf, interpolants failed')
      endif 
      
      mem_interpolants = dble( 8 * interpolants_size ) 
      
      mem_size = (nfc_2d_row +1) * (nfc_2d_col * 2 +1)
      mem_ctf_val = dble(8 * mem_size)
      
      mem_total = mem_total + mem_ctf_val + mem_interpolants 
      mem_peak  = max(mem_peak, mem_total)
      
C     Zero interpolants array
      interpolants = 0.0        ! Fortran 90 array syntax
      
#ifdef MPI
      call mpi_bcast(param_fnames, filename_len*pfilesmx, mpi_character,
     $     0, mcw, mpi_err)
      call mpi_bcast(img_fnames, filename_len*pfilesmx, mpi_character,
     $     0, mcw, mpi_err)
#endif
      
C     Determine numbers of particles in parameter files
      if(iam == 0) then
         ptle_count = 0
         call get_ptle_counts(nfiles, filename_len, param_fnames,
     $        oddeven, ptle_count)
         nptles = sum(ptle_count)
         
         if(nptles == 0) then
            call error_stop('P3DR', 'main program', 
     +           'All particle parameter files are empty')
         else
            write(*,*) 'Total number of images: ', nptles
            do i=1,nfiles
               write(*,*) 'file #', i, ': ', ptle_count(i), 
     $              trim(param_fnames(i))
            enddo
         endif
      endif
      
C     Broadcast numbers of particles in parameter files
#ifdef MPI
      call mpi_bcast(nfiles,     1,      mpi_integer, 0, mcw, mpi_err)
      call mpi_bcast(nptles,     1,      mpi_integer, 0, mcw, mpi_err)
      call mpi_bcast(ptle_count, nfiles, mpi_integer, 0, mcw, mpi_err)
#endif
      
C     Allocate global arrays for particle parameters
      allocate(indx(nptles), theta(nptles), phi(nptles), omega(nptles),
     $     x(nptles), y(nptles), mag(nptles))
      
      if(per_ptle_ctf == 1) then
         allocate(delf_maj_ptle(nptles), delf_min_ptle(nptles),
     $        ang_maj_ptle(nptles))
      endif

C     Get particle parameters
      if(iam == 0) then
         if(per_ptle_ctf == 1) then
            call get_ptle_ioomctf(nfiles, filename_len, param_fnames,
     $           oddeven, nptles, nbox, indx, theta, phi, omega,
     $           x, y, mag, delf_maj_ptle, delf_min_ptle, ang_maj_ptle)
         else                  
            call get_ptle_ioom(nfiles, filename_len, param_fnames,
     $           oddeven, nptles, nbox, indx, theta, phi, omega,
     $           x, y, mag)
         endif
      endif

C     Distrubute particle information to all other processes
#ifdef MPI
      call mpi_bcast(indx,  nptles, mpi_integer, 0, mcw, mpi_err)
      call mpi_bcast(theta, nptles, mpi_real,    0, mcw, mpi_err)
      call mpi_bcast(phi,   nptles, mpi_real,    0, mcw, mpi_err)
      call mpi_bcast(omega, nptles, mpi_real,    0, mcw, mpi_err)
      call mpi_bcast(x,     nptles, mpi_real,    0, mcw, mpi_err)
      call mpi_bcast(y,     nptles, mpi_real,    0, mcw, mpi_err)
      call mpi_bcast(mag,   nptles, mpi_real,    0, mcw, mpi_err)
      if(per_ptle_ctf == 1) then
         call mpi_bcast(delf_maj_ptle, nptles, mpi_real, 0,mcw,mpi_err)
         call mpi_bcast(delf_min_ptle, nptles, mpi_real, 0,mcw,mpi_err)
         call mpi_bcast(ang_maj_ptle,  nptles, mpi_real, 0,mcw,mpi_err)
      endif
#endif
      
      id = 0
      do i = 1, nfiles
         npix_row         =  nbox(i)
         npix_col         =  nbox(i)
         angstrom_per_pix =  pixsizs(i)
         
         allocate(pixels(npix_col*npix_row))
         
         call pif_open(img_fnames(i), pix_file_unit, 'r')
         call get_1st_ortid(j, param_fnames(i))
         call read_1_pif(j,npix_row,npix_col, pix_file_unit, imgprjbig)
         call pif_close(pix_file_unit)
         if(i == 1) then
            imgprjstd   = imgprjbig
            enlargefact = 1. 
         else 
            enlargefact = imgprjstd / imgprjbig
         endif
         
C     Initialize CTF arrays
         ctf_val = 1.0          ! Fortran 90 array syntax
         abs_ctf = 1.0          ! Fortran 90 array syntax
         
C     Calculate CTF corrections
         if(ctfmode > 0 .and. ctfmode <= 4)  then
            if(per_ptle_ctf == 0) then
               call ctf_para(ctf_val, abs_ctf, 
     $              0, nfc_2d_row, -nfc_2d_col, nfc_2d_col, .true.,
     $              ctfmode, filter, tempfac, ctf_ff1, ctf_ff2,
     $              pixsizs(i), units, volts(i), 
     $              delf_majs(i), delf_mins(i), 
     $              ang_majs(i), ampfacs(i),
     $              cs_coeffs(i), enlargefact, res_max, res_min)
            endif
         endif
         
         do j=1,ptle_count(i)
            id = id + 1
            
            if(iam == mod(id-1,num_nodes)) then
               call pif_open(img_fnames(i), pix_file_unit, 'r')
               imghead%lunum = pix_file_unit
               call pif_read_gh(imghead)
               imghead%imgnum = indx(id)
               call pif_read_dh(imghead)
               imghead%imgnum = indx(id)
               call read_pixel_frame(imghead, npix_row*npix_col, 
     $              pixels, err_code)
               call pif_close(pix_file_unit)

               if(per_ptle_ctf == 1) then 
                  call ctf_para(ctf_val, abs_ctf, 
     $                 0, nfc_2d_row, -nfc_2d_col, nfc_2d_col, .true., 
     $                 ctfmode, filter, tempfac, ctf_ff1, ctf_ff2, 
     $                 pixsizs(i), units, volts(i), 
     $                 delf_maj_ptle(id), delf_min_ptle(id), 
     $                 ang_maj_ptle(id), ampfacs(i), 
     $                 cs_coeffs(i), enlargefact, res_max, res_min) 
               endif            
               
               call cmpt_intrps(hkl_max, interpolants,
     $              interpolants_size, pixels, npix_col*npix_row,
     $              ctf_val, abs_ctf, theta(id), phi(id), omega(id),
     $              x(id), y(id), mag(id))
            endif
         enddo
         
         deallocate(pixels)
      enddo
      
      deallocate(indx, theta, phi, omega, x, y, mag) 

      if(per_ptle_ctf == 1) then
         deallocate(delf_maj_ptle, delf_min_ptle, ang_maj_ptle)
      endif


C     Apply 32 symmetry operations to interpolants array
      if(symm_code == 532) then
         call apply32(interpolants, hkl_max)
      endif
      
      deallocate(ctf_val, abs_ctf)
      mem_total = mem_total - mem_ctf_val

      convert_to_invang = (mapmagfactor * nfc_2d_row) /
     $     (nrffta_row * angstrom_per_pix * hkl_max)

crss      write(*,*) 'preparing for noise suppression'
crss      write(*,*) 'nrffta_row        = ', nrffta_row
crss      write(*,*) 'angstrom_per_pix  = ', angstrom_per_pix
crss      write(*,*) 'hkl_max           = ', hkl_max
crss      write(*,*) 'mapmagfactor      = ', mapmagfactor
crss      write(*,*) 'nfc_2d_row        = ', nfc_2d_row
crss      write(*,*) 'convert_to_invang = ', convert_to_invang
crss      write(*,*)

C     Apply noise suppression algorithm
C     Rosenthal and Henderson, JMB 333 721-745 (2003)
      if (nfsc > 0) then
         call noise_suppression(interpolants, hkl_max, 
     $        convert_to_invang, nfsc, fsc_freq, fsc_value)
      endif



C     Noise suppression code - need to figure out where to apply RSS

      
C     **********************************************************************
C     *        Combining estimates of 3D DFT along three directions        *
C     *        arrange the data before 3D DFT                              * 
C     **********************************************************************
      
      mem_size    = (num_hkl**2) * (-first_iq+last_iq+1) * 3
      if(mem_size < 0) mem_size = 0
      allocate(f_hkl_ref(mem_size), stat = alloc_stat)
      if (alloc_stat > 0 ) 
     +     call error_stop(progname, subname,
     +     'Error in allocating memory for F_HKL_ref')
      mem_f_hkl_ref = dble(8 * mem_size)
      mem_total = mem_total + mem_f_hkl_ref
      mem_peak  = max(mem_peak, mem_total)
      call exch_intp(hkl_max, interpolants, f_hkl_ref)
      deallocate (interpolants)
      mem_total = mem_total - mem_interpolants
      mem_size    = (num_hkl**2) * (-first_iq+last_iq+1)
      if(mem_size < 0) mem_size = 0
      allocate(f_hkl(mem_size), stat = alloc_stat)
      if (alloc_stat > 0 )
     +     call error_stop(progname, subname,
     +     'Error in allocating memory for F_HKL')
      mem_f_hkl = dble(8 * mem_size)
      mem_total = mem_total + mem_f_hkl
      mem_peak  = max(mem_peak, mem_total)
      call realtocomplx(hkl_max, first_iq, last_iq, 
     +     f_hkl_ref, f_hkl)
      deallocate (f_hkl_ref)
      mem_total = mem_total - mem_f_hkl_ref
      
      
C     ----- This is the end of combining 3D DFT along 3 directions -----
      
C     ********************************************************************
C     *                    2D Fourier synthesis                          *
C     ********************************************************************
      
      
C     Allocate memory
      mem_size = ((-min_xyz+max_xyz+1)**2) *
     +     (-first_rr_out+last_rr_out+1)
      if(mem_size < 0) mem_size = 0
      allocate(f_vru(mem_size), stat = alloc_stat)
      if (alloc_stat > 0) call error_stop(progname, subname, 
     +     'Error in allocating memory for F_VRU')
      mem_f_vru = 4 * mem_size
      mem_total = mem_total + mem_f_vru
      mem_peak  = max(mem_peak, mem_total)
      
C     FFTSYNTH_1
C     **********
      call fftsynth_1_m_slab(hkl_max,
     +     first_iq, last_iq, first_rr_out, last_rr_out,
     +     f_hkl, nrffts_1_row, nrffts_1_col, nrffts_1_omega, 
     +     min_x, max_x, min_z, max_z, 
     +     min_xyz, max_xyz, f_vru, mem_fftsynth_1_m )
      
      deallocate(f_hkl)
      mem_peak = max(mem_peak, mem_total+mem_fftsynth_1_m)
      mem_total = mem_total - mem_f_hkl
      
C     ------------ This is the end of 2D Fourier synthesis -----------------
      
C     **********************************************************************
C     *           Exchanging data for 1D Fourier synthesis                 *
C     **********************************************************************
      
C     Allocate storage for subroutine EXCHANGE_2
      mem_size = (-min_xyz+max_xyz+1) * (-min_xyz+max_xyz+1) *
     +     (-frst_lst_U(1,iam)+frst_lst_U(2,iam)+1)
      allocate(F_VRU_TMP(mem_size), stat = alloc_stat)
      if (alloc_stat > 0) call error_stop(progname,
     +     'exchange_2_m', 'Error in allocating memory for F_VRU_TMP!')
      mem_F_VRU_TMP = dble(4 * mem_size)
      mem_total = mem_total + mem_F_VRU_TMP
      mem_peak  = max(mem_peak, mem_total)
      
      call exchange_2_slab( min_xyz, max_xyz, F_VRU,
     +     nrffts_1_row, F_VRU_TMP )
      
      deallocate(F_VRU)
      mem_total = mem_total - mem_F_VRU
      
      mem_size = nrffts_1_row*(-frst_lst_u(1,iam)+frst_lst_u(2,iam)+1)
      mem_F_XK_dZ = dble(4 * (-min_x+max_x+1) * mem_size)
      allocate(F_XK_dZ((-min_x+max_x+1) * mem_size), stat = alloc_stat)
      if(alloc_stat > 0 ) then
         call error_stop (progname, subname,  
     +        ' Memory allocation failed')
      endif
      mem_total = mem_total + mem_F_XK_dZ 
      mem_peak  = max(mem_peak, mem_total)
      
      call rearrange_2_slab(min_xyz, max_xyz, nrffts_1_row,
     +     F_VRU_TMP, min_x, max_x, F_XK_dZ )
      
C     Free memory used in 2D Fourier synthesis.
      deallocate(F_VRU_TMP)
      mem_total = mem_total - mem_F_VRU_TMP
      
C     ----- This is the end of exchanging data for 1D Fourier synthesis ----
C     ------ Before move onto the next step, synchronize all the nodes -----
      
C     **********************************************************************
C     *                        1D Fourier synthesis                        *
C     **********************************************************************
      
      
C     Allocate storage for 1D Fourier synthesis.
      mem_size = ((max_x-min_x+1)**2) * (-first_z+last_z+1)
      if(mem_size < 0) mem_size = 0
      allocate(rho(min_x:max_x, min_y:max_y, first_z:last_z),
     +     stat = alloc_stat)
      if (alloc_stat > 0) call error_stop(progname, subname, 
     +     'Error in allocating memory for RHO')
      mem_rho = dble(4 * mem_size)
      mem_total  = mem_total + mem_rho 
      mem_peak   = max(mem_peak, mem_total)
      
C     CALL FFTSYNTH_2
C     ***************
      call fftsynth_2_m_slab( nrffts_1_row, min_x, max_x, min_y, 
     +     max_y, min_z, max_z, F_XK_dZ, nrffts_2_row, 
     +     nrffts_2_col, nrffts_2_omega, rho, 
     +     mem_fftsynth_2_m )
      mem_peak = max(mem_peak, mem_total+mem_fftsynth_2_m)
      
C     Free memory used in 2D Fourier synthesis.
      deallocate(F_XK_dZ)
      mem_total = mem_total - mem_F_XK_dZ
      
C     -------------- This is the end of 1D Fourier synthesis --------------
      
C     *********************************************************************
C     clear the density map and output density map
C     *********************************************************************
      
      mem_size = (-den_min_x+den_max_x+1)*
     +     (-den_min_y+den_max_y+1)* max_z_slab
C     according to the set_indice, node 1 alway has the most largest slab,
      
      allocate(density_map(den_min_x:den_max_x, den_min_y:den_max_y,
     $     max_z_slab), stat = alloc_stat)
      mem_density_map = dble(4*mem_size)
      mem_total = mem_total + mem_density_map
      mem_peak  = max(mem_peak, mem_total)
      if (alloc_stat > 0) call error_stop(progname,
     +     subname,  'Error in allocating density_map')
      
      call density_clear(rho, min_x, max_x, min_y, max_y,
     +     den_min_x, den_max_x,
     +     frst_lst_z(1,iam), frst_lst_z(2,iam),
     +     density_map, map_min_local, map_max_local, apo_border)
      
C     Get the global maximum and minimum density over entire map
      
#ifdef MPI
      call mpi_reduce(map_min_local, map_min, 1, MPI_REAL, MPI_MIN,
     +     0, mcw, mpi_err)
      call mpi_reduce(map_max_local, map_max, 1, MPI_REAL, MPI_MAX,
     +     0, mcw, mpi_err)
      if( mpi_err .ne. 0 ) call error_stop(progname,'density_clear',
     +     'Error in calling MPI_BCAST background')
#else
      map_min = map_min_local
      map_max = map_max_local
#endif   
            
      deallocate(rho)
      mem_total = mem_total - mem_rho
      
Cji   Write density map into MAP file.

      call output_density( density_map, den_min_x,
     +     den_max_x, max_z_slab, MAP_MIN, MAP_MAX )
      
C     Free memory for storing the density values.
      
      deallocate(density_map)
      deallocate(fsc_freq, fsc_value)
      mem_total = mem_total - mem_density_map

C ------- This the end of writing density values into MAP file --------
C ----- Before reporting the timing the memory allocating mesage, -----
C ------------------- synchronize all the nodes. ----------------------

#ifdef MPI
      call mpi_barrier(mcw, mpi_err)
#endif
      
      call system_clock(clock2)
      time = elapsed_time(clock1, clock2)
      
      if ( iam == 0 ) then
         write(*,*) 'Total run time = ', time
      endif

#ifdef MPI
      call MPI_FINALIZE ( mpi_err )
#endif

      end program p3dr


      function locate(xx,n,x)
      implicit none
      integer, intent(in) :: n
      real, intent(in) :: xx(n)
      real, intent(in) :: x
      integer :: locate

C     Given an array xx(1:n), and given a value x, returns a value j
C     such that x is between xx(j) and xx(j+1). xx must be monotonic,
C     either increasing or decreasing. j=0 or j=n is returned to
C     indicate that x is out of range. (Adapted from Numerical Recipes
C     in Fortran 90

      integer :: jl, jm, ju
      logical :: ascnd
      ascnd = (xx(n) >= xx(1))
      jl=0
      ju=n+1
      do
         if (ju-jl <= 1) exit
         jm=(ju+jl)/2
         if (ascnd .eqv. (x >= xx(jm))) then
            jl=jm
         else
            ju=jm
         endif
      enddo
      if (x == xx(1)) then
         locate=1
      else if (x == xx(n)) then
         locate=n-1
      else
         locate=jl
      endif
      end function locate
