

/*******************************************************************
 * ROBEM - Rob's Electron Microscopy imaging software.
 * 
 * Program to take Purdue portable PIF image files, "box" them
 * and write out Purdue portable PIF packed format images for use
 * in reconstruction.  
 *
 * Also reads MAP files and displays with 3D or 2D rendering.
 *
 * 
 * Rob Ashmore
 * Biology Department
 * Purdue University
 * February 1996
 *
 *******************************************************************/

#define VERSION "2.00_200307015"
#define IS_MAIN_MODULE 1
#define IS_ROBEM_MODULE 1

#include <limits.h>
#include <string.h>
#include <stdio.h>
#include <math.h>
#include <ctype.h>
#include <stdlib.h>
#include <time.h>
#include <sys/types.h>
#include <unistd.h>

#include <Xm/Xm.h>
#include <Xm/GadgetP.h>
#include <Xm/PrimitiveP.h>
#include <Xm/MessageB.h>
#include <Xm/ScrollBar.h>
#include <Xm/ToggleB.h>
#include <Xm/Label.h>
#include <Xm/DialogS.h>
#include <Xm/Text.h>
#include <Xm/List.h>
#include <Mrm/MrmAppl.h>
#include <X11/cursorfont.h>
#include <Xm/FileSB.h>

#ifdef NEEDTHIS
#include <cda_def.h>
#endif

#include "../include/libPIF.h"
#include "../libR/libUtil.h"
#include "../libR/appodize.h"
#include "lists.h"
#include "robem.h"
#include "3dgraphics.h"

#include "bitmaps/zoom.bit"
#include "bitmaps/zoommask.bit"
#include "bitmaps/lantern.bit"
#include "bitmaps/lanternmask.bit"
#include "bitmaps/color.xbm"
#include "bitmaps/uncolor.xbm"
#include "bitmaps/section.xbm"
#include "bitmaps/unsection.xbm"
#include "bitmaps/locked.xbm"
#include "bitmaps/unlocked.xbm"
#include "bitmaps/play.xbm"
#include "bitmaps/stop.xbm"
#include "bitmaps/playing.xbm"
#include "bitmaps/stopped.xbm"
#include "bitmaps/doflashlight.xbm"
#include "bitmaps/noflashlight.xbm"
#include "bitmaps/dorock.xbm"
#include "bitmaps/norock.xbm"
#include "bitmaps/infini.xbm"
#include "bitmaps/noinfini.xbm"
#include "bitmaps/go.xbm"
#include "bitmaps/nogo.xbm"
#include "bitmaps/help.xbm"
#include "bitmaps/nohelp.xbm"
#include "bitmaps/exit.xbm"
#include "bitmaps/noexit.xbm"
#include "bitmaps/reset.xbm"
#include "bitmaps/noreset.xbm"
#include "bitmaps/slow.xbm"
#include "bitmaps/noslow.xbm"
#include "bitmaps/movie.xbm"
#include "bitmaps/unmovie.xbm"



/************************************************************
 * malloc_verify() - Debugging routine for checking to see
 * if malloc/free are happy.
*/
void malloc_verify()
{/* MALLOC_VERIFY */
  char *inStr;

  CK(inStr = (char *)malloc(sizeof(char)*5));
  free(inStr);

}/* MALLOC_VERIFY */



/************************************************************************
 * OS transfer point.  The main routine does all the one-time setup and *
 * then calls XtAppMainLoop.                                            *
 ************************************************************************/
int main(int argc, char **argv)
{/* MAIN */
  int i,*dispThis=0;
  char *s1;
  char robemStr[256];
  String fallbacks[] =
    { "*fontList:-dec-terminal-bold-r-normal--14-140-75-75-c-80-iso8859-1=tag",
	NULL
	};
  XmString simpleString;
  FILE *fIn;

  /* Show them our version number */
  sprintf(versionStr,"%s",banner("RobEM"));

  /* Initialize our C code */
  init_globals(argc, argv);

  /* Initialize MRM */
  MrmInitialize();                 

  XtToolkitInitialize();
  app_context = XtCreateApplicationContext();

  XtAppSetFallbackResources(app_context, fallbacks);

  /* Finally, open the display */
  dsply = XtOpenDisplay(app_context, NULL,
			"robem",
			"robem", NULL,
			0, &argc, argv);
  
  /* Add our remapping of keys to widgets */
  XtAppAddActions(app_context,rwa_action_table, 1);
  XtAppAddActions(app_context,box_action_table, 1);
  XtAppAddActions(app_context,editBoxText_action_table, 1);
  XtAppAddActions(app_context,map_abang_action_table, 1);
  XtAppAddActions(app_context,aoverb_action_table, 1);
  XtAppAddActions(app_context,fftImgNum_action_table, 1);
  XtAppAddActions(app_context,fftImgNumArrow_action_table, 1);
  XtAppAddActions(app_context,surfDens_action_table, 1);
  XtAppAddActions(app_context,zoom_action_table, 1);
  XtAppAddActions(app_context,zoomArrow_action_table, 1);
  XtAppAddActions(app_context,imgNum_action_table, 1);
  XtAppAddActions(app_context,radius_action_table, 1);
  XtAppAddActions(app_context,boxInit_action_table, 1);
  XtAppAddActions(app_context,fftStep_action_table, 1);
  XtAppAddActions(app_context,boxNum_action_table, 1);
  XtAppAddActions(app_context,paintSize_action_table, 1);
  XtAppAddActions(app_context,icosa_action_table, 1);
  XtAppAddActions(app_context,ta_action_table, 1);
  XtAppAddActions(app_context,ta2_action_table, 1);
  XtAppAddActions(app_context,gbs_action_table, 1);
  XtAppAddActions(app_context,cmScroll_action_table, 1);
  XtAppAddActions(app_context,version_action_table, 1);
  XtAppAddActions(app_context,paramFile_action_table, 1);
  XtAppAddActions(app_context,paramMapFile_action_table, 1);
  XtAppAddActions(app_context,paramScanFile_action_table, 1);
  XtAppAddActions(app_context,projContour_action_table, 1);
  XtAppAddActions(app_context,projasu_action_table, 1);
  XtAppAddActions(app_context,ppin_action_table, 1);
  XtAppAddActions(app_context,stepPP_action_table, 1);
  XtAppAddActions(app_context,stepPPtext_action_table, 1);
  XtAppAddActions(app_context,ppintext_action_table, 1);
  XtAppAddActions(app_context,binPP_action_table, 1);
  XtAppAddActions(app_context,binPPtext_action_table, 1);
  XtAppAddActions(app_context,strtxpp_action_table, 1);
  XtAppAddActions(app_context,ppMinMax_action_table, 1);
  XtAppAddActions(app_context,ctfValue_action_table, 1);
  XtAppAddActions(app_context,ctfCurve_action_table, 1);
  XtAppAddActions(app_context,latText_action_table, 1);
  XtAppAddActions(app_context,ctfText_action_table, 1);
  XtAppAddActions(app_context,rotflt_action_table, 1);
  XtAppAddActions(app_context,cDisp_zArrow_act_table, 1);
  XtAppAddActions(app_context,subText_action_table, 1);
  XtAppAddActions(app_context,diffSec_action_table, 1);
  XtAppAddActions(app_context,smMMax_action_table, 1);
  XtAppAddActions(app_context,phase_action_table, 1);
  XtAppAddActions(app_context,lopText_action_table, 1);
  XtAppAddActions(app_context,hipText_action_table, 1);
  XtAppAddActions(app_context,coravg_action_table, 1);
  XtAppAddActions(app_context,lohiRadiusText_action_table, 1);

  /* Explicitly call this class "robem" */
  toplevel_widget = XtVaAppCreateShell(
				       argv[0],
				       "robem",
				       topLevelShellWidgetClass, 
				       /* XmTopLevelShell or
					  whatever */
				       dsply,
				       NULL);

  /* Now create our colormap */
#define NORMAL_WAY 1      
#ifdef NORMAL_WAY
  visual= DefaultVisual(dsply, DefaultScreen(dsply));
  visual_depth = (unsigned int)DefaultDepth(dsply, DefaultScreen(dsply));
  if(visual_depth!=8)
    {
      /* OLDWAY 
	 create24Colormap(); 
	 createDefaultColormap();
      */
      createDefaultColormap();
    }
#else
  create24Colormap(); 
#endif



  /***********************************************************************
   * Open the UID files (the output of the UIL compiler) in the hierarchy*
   ***********************************************************************/
  /* We're calling the robem.uid file now new_robem.uid so the robem_1.07 can
   * use the older robem.uid */
  CK(db_filename_vec[0] = (char *)malloc(sizeof(char)*512));

  /* Kludge this for now so the production user gets the right .UID */
  s1 = strrchr(argv[0],'/');
  if(s1)
    sprintf(robemStr,"%s",s1+1);
  else
    sprintf(robemStr,"%s",argv[0]);
  if(EQU(argv[0],"./robem"))
    sprintf(db_filename_vec[0],"./new_robem.uid"); 
  else if(EQU(argv[0],"/bio/baker9/rwa/devel64/robem/robem"))
    sprintf(db_filename_vec[0],"/bio/baker9/rwa/devel64/robem/new_robem.uid"); 
  else if(EQU(argv[0],"/home/rwa/devel/robem/robem"))
    sprintf(db_filename_vec[0],"/home/rwa/devel/robem/new_robem.uid"); 
  else if(EQU(argv[0],"/baker9/rwa/devel/robem/robem"))
    sprintf(db_filename_vec[0],"/baker9/rwa/devel/robem/new_robem.uid"); 
  else if(EQU(argv[0],"/usr/baker9/rwa/devel/robem/robem"))
    sprintf(db_filename_vec[0],"/usr/baker9/rwa/devel/robem/new_robem.uid"); 
  else 
    sprintf(db_filename_vec[0],"%s/new_%s.uid",getenv("EMDIR"),robemStr); 
  /* Give feedabck for easy debugging of EMDIR env var defn */
  printf("[%s]\n",db_filename_vec[0]);

  fIn = fopen(db_filename_vec[0],"r");
  if(!fIn)
    {/* Error */
      printf("\nI cannot find [%s]\n\n",db_filename_vec[0]);
      printf("Make sure you've got the environmental variable EMDIR\n\
  defined properly as:\n\n\
       setenv EMDIR /usr/local/baker\n\n\
  or as appropriate.\n\n");
      exit(0);
    }/* Error */
  fclose(fIn);

  if (MrmOpenHierarchy(db_filename_num, /* Number of files. */
		       db_filename_vec,       /* Array of file names.  */
		       NULL,                  /* Default OS extenstion. */
		       &s_MrmHierarchy)       /* Pointer to returned MRM ID */
      !=MrmSUCCESS)
    s_error("can't open hierarchy");
  
  
  /* Register the items MRM needs to bind for us. */
  MrmRegisterNames(reglist, reglist_num);
  
  /* Go get the main part of the application. */
  if (MrmFetchWidget(s_MrmHierarchy, "S_MAIN_WINDOW", toplevel_widget,
		     &main_window_widget, &dummy_class) != MrmSUCCESS)
    s_error("can't fetch main window");

  /* Go get the fftFeedback Screen part of the application. */
  if (MrmFetchWidget(s_MrmHierarchy, "fftFeedbackScreen", toplevel_widget,
		     &fftFeedbackScreen, &dummy_class) != MrmSUCCESS)
    s_error("can't fetch fftFeedback window");

  /* Go get the dio Screen part of the application. */
  if (MrmFetchWidget(s_MrmHierarchy, "dioScreen", toplevel_widget,
		     &dioScreen, &dummy_class) != MrmSUCCESS)
    s_error("can't fetch dio window");

  /* Go get the batchjob Screen part of the application. */
  if (MrmFetchWidget(s_MrmHierarchy, "batchjobscreen", toplevel_widget,
		     &batchjobscreen, &dummy_class) != MrmSUCCESS)
    s_error("can't fetch batchjob window");

  /* Go get the subtractMap Screen part of the application. */
  if (MrmFetchWidget(s_MrmHierarchy, "subtractMapScreen", toplevel_widget,
		     &subtractMapScreen, &dummy_class) != MrmSUCCESS)
    s_error("can't fetch subtratcMap window");

  /* Go get the fourierFilt Screen part of the application. */
  if (MrmFetchWidget(s_MrmHierarchy, "fourierFiltScreen", toplevel_widget,
		     &fourierFiltScreen, &dummy_class) != MrmSUCCESS)
    s_error("can't fetch fourierFilt window");

  /* Go get the rescale Screen part of the application. */
  if (MrmFetchWidget(s_MrmHierarchy, "rescalescreen", toplevel_widget,
		     &rescalescreen, &dummy_class) != MrmSUCCESS)
    s_error("can't fetch rescale window");

  /* Go get the invert Screen part of the application. */
  if (MrmFetchWidget(s_MrmHierarchy, "invertscreen", toplevel_widget,
		     &invertscreen, &dummy_class) != MrmSUCCESS)
    s_error("can't fetch invert window");

  /* Go get the movement Screen part of the application. */
  if (MrmFetchWidget(s_MrmHierarchy, "movementscreen", toplevel_widget,
		     &movementscreen, &dummy_class) != MrmSUCCESS)
    s_error("can't fetch movement window");

  /* Go get the partproc Screen part of the application. */
  if (MrmFetchWidget(s_MrmHierarchy, "partprocscreen", toplevel_widget,
		     &partprocscreen, &dummy_class) != MrmSUCCESS)
    s_error("can't fetch partproc window");

  /* Go get the model Screen part of the application. */
  if (MrmFetchWidget(s_MrmHierarchy, "modelscreen", toplevel_widget,
		     &modelScreenWidget, &dummy_class) != MrmSUCCESS)
    s_error("can't fetch model window");

  /* Go get the binning Screen part of the application. */
  if (MrmFetchWidget(s_MrmHierarchy, "binningscreen", toplevel_widget,
		     &binningScreenWidget, &dummy_class) != MrmSUCCESS)
    s_error("can't fetch binning window");

  /* Go get the interp Screen part of the application. */
  if (MrmFetchWidget(s_MrmHierarchy, "interpscreen", toplevel_widget,
		     &interpScreenWidget, &dummy_class) != MrmSUCCESS)
    s_error("can't fetch interp window");

  /* Go get the version Screen part of the application. */
  if (MrmFetchWidget(s_MrmHierarchy, "versionscreen", toplevel_widget,
		     &versionScreenWidget, &dummy_class) != MrmSUCCESS)
    s_error("can't fetch version window");

  /* Go get the status Screen part of the application. */
  if (MrmFetchWidget(s_MrmHierarchy, "statusscreen", toplevel_widget,
		     &statusScreenWidget, &dummy_class) != MrmSUCCESS)
    s_error("can't fetch status window");

  /* Go get the colormapScreen part of the application. */
  if (MrmFetchWidget(s_MrmHierarchy, "colormapscreen", toplevel_widget,
		     &colorScreenWidget, &dummy_class) != MrmSUCCESS)
    s_error("can't fetch colormap window");

  /* Go get the helpScreen part of the application. */
  if (MrmFetchWidget(s_MrmHierarchy, "helpscreen", toplevel_widget,
		     &helpScreenWidget, &dummy_class) != MrmSUCCESS)
    s_error("can't fetch help window");

  /* Go get the editBoxScreen part of the application. */
  if (MrmFetchWidget(s_MrmHierarchy, "editBoxscreen", toplevel_widget,
		     &editBoxScreenWidget, &dummy_class) != MrmSUCCESS)
    s_error("can't fetch editBox window");

  /* Go get the optionScreen part of the application. */
  if (MrmFetchWidget(s_MrmHierarchy, "optionscreen", toplevel_widget,
		     &optionScreenWidget, &dummy_class) != MrmSUCCESS)
    s_error("can't fetch option window");

  /* Go get the mapShadeScreen part of the application. */
  if (MrmFetchWidget(s_MrmHierarchy, "mapShadescreen", toplevel_widget,
		     &mapShadeScreenWidget, &dummy_class) != MrmSUCCESS)
    s_error("can't fetch mapShade window");

  /* Go get the statscreen part of the application. */
  if (MrmFetchWidget(s_MrmHierarchy, "statscreen", toplevel_widget,
		     &statscreen, &dummy_class) != MrmSUCCESS)
    s_error("can't fetch stat window");

  /* Go get the refscreen part of the application. */
  if (MrmFetchWidget(s_MrmHierarchy, "refscreen", toplevel_widget,
		     &refscreen, &dummy_class) != MrmSUCCESS)
    s_error("can't fetch ref window");

  /* Go get the movie_contour_screen part of the application. */
  if (MrmFetchWidget(s_MrmHierarchy, "movie_contour_screen", toplevel_widget,
		     &movie_contour_screen, &dummy_class) != MrmSUCCESS)
    s_error("can't fetch movie_contour window");

  /* Go get the pointscreen part of the application. */
  if (MrmFetchWidget(s_MrmHierarchy, "pointscreen", toplevel_widget,
		     &pointscreen, &dummy_class) != MrmSUCCESS)
    s_error("can't fetch point window");

  /* Go get the truncaddscreen part of the application. */
  if (MrmFetchWidget(s_MrmHierarchy, "truncaddscreen", toplevel_widget,
		     &truncaddscreen, &dummy_class) != MrmSUCCESS)
    s_error("can't fetch truncadd window");

  /* Go get the rotatescreen part of the application. */
  if (MrmFetchWidget(s_MrmHierarchy, "rotatescreen", toplevel_widget,
		     &rotatescreen, &dummy_class) != MrmSUCCESS)
    s_error("can't fetch rotate window");

  /* Go get the fix2Dscreen part of the application. */
  if (MrmFetchWidget(s_MrmHierarchy, "fix2Dscreen", toplevel_widget,
		     &fix2Dscreen, &dummy_class) != MrmSUCCESS)
    s_error("can't fetch fix2D window");

  /* Go get the fixscreen part of the application. */
  if (MrmFetchWidget(s_MrmHierarchy, "fixscreen", toplevel_widget,
		     &fixscreen, &dummy_class) != MrmSUCCESS)
    s_error("can't fetch fix window");

  /* Go get the fixitscreen part of the application. */
  if (MrmFetchWidget(s_MrmHierarchy, "fixitscreen", toplevel_widget,
		     &fixitscreen, &dummy_class) != MrmSUCCESS)
    s_error("can't fetch fixit window");

  /* Go get the imgconvscreen part of the application. */
  if (MrmFetchWidget(s_MrmHierarchy, "imgconvscreen", toplevel_widget,
		     &imgconvscreen, &dummy_class) != MrmSUCCESS)
    s_error("can't fetch imgconv window");

  /* Go get the emcoravgscreen part of the application. */
  if (MrmFetchWidget(s_MrmHierarchy, "emcoravgscreen", toplevel_widget,
		     &emcoravgscreen, &dummy_class) != MrmSUCCESS)
    s_error("can't fetch emcoravg window");

  /* Go get the fourAverScreen part of the application. */
  if (MrmFetchWidget(s_MrmHierarchy, "fourAverScreen", toplevel_widget,
		     &fourAverScreen, &dummy_class) != MrmSUCCESS)
    s_error("can't fetch fourAver window");

  /* Go get the crysDispscreen part of the application. */
  if (MrmFetchWidget(s_MrmHierarchy, "crysDispscreen", toplevel_widget,
		     &crysDispscreen, &dummy_class) != MrmSUCCESS)
    s_error("can't fetch crysDisp window");

  /* Go get the rotfltscreen part of the application. */
  if (MrmFetchWidget(s_MrmHierarchy, "rotfltscreen", toplevel_widget,
		     &rotfltscreen, &dummy_class) != MrmSUCCESS)
    s_error("can't fetch rotflt window");

  /* Go get the fftscreen part of the application. */
  if (MrmFetchWidget(s_MrmHierarchy, "fftscreen", toplevel_widget,
		     &fftscreen, &dummy_class) != MrmSUCCESS)
    s_error("can't fetch fft window");

  /* Go get the circAvgscreen part of the application. */
  if (MrmFetchWidget(s_MrmHierarchy, "circAvgscreen", toplevel_widget,
		     &circAvgScreenWidget, &dummy_class) != MrmSUCCESS)
    s_error("can't fetch circAvg window");

  /* Go get the auto2DDRscreen part of the application. */
  if (MrmFetchWidget(s_MrmHierarchy, "auto2DDRScreen", toplevel_widget,
		     &auto2DDRScreenWidget, &dummy_class) != MrmSUCCESS)
    s_error("can't fetch auto2DDR window");

  /* Go get the autoDRscreen part of the application. */
  if (MrmFetchWidget(s_MrmHierarchy, "autoDRScreen", toplevel_widget,
		     &autoDRScreenWidget, &dummy_class) != MrmSUCCESS)
    s_error("can't fetch autoDR window");

  /* Go get the appodizeScreen part of the application. */
  if (MrmFetchWidget(s_MrmHierarchy, "appodizescreen", toplevel_widget,
		     &appodizeScreenWidget, &dummy_class) != MrmSUCCESS)
    s_error("can't fetch appodize window");

  /* Go get the boxScreen part of the application. */
  if (MrmFetchWidget(s_MrmHierarchy, "boxscreen", toplevel_widget,
		     &boxScreenWidget, &dummy_class) != MrmSUCCESS)
    s_error("can't fetch box window");

  /* Go get the infoScreen part of the application. */
  if (MrmFetchWidget(s_MrmHierarchy, "infoMessageBox", toplevel_widget,
		     &infoBox, &dummy_class) != MrmSUCCESS)
    s_error("can't fetch info window");
  XtUnmanageChild(XmMessageBoxGetChild(infoBox,XmDIALOG_HELP_BUTTON));

  /* Go get the infoAnswerscreen part of the application. */
  if (MrmFetchWidget(s_MrmHierarchy, "infoAnswerscreen", toplevel_widget,
		     &infoAnswerscreen, &dummy_class) != MrmSUCCESS)
    s_error("can't fetch infoAnswer window");

  /* Go get the normalScreen part of the application. */
  if (MrmFetchWidget(s_MrmHierarchy, "normalScreen", toplevel_widget,
		     &normalScreen, &dummy_class) != MrmSUCCESS)
    s_error("can't fetch normal window");

  /* Go get the cavgPPScreen part of the application. */
  if (MrmFetchWidget(s_MrmHierarchy, "cavgPPScreen", toplevel_widget,
		     &cavgPPScreen, &dummy_class) != MrmSUCCESS)
    s_error("can't fetch cavgPP window");

  /* Go get the manipPPScreen part of the application. */
  if (MrmFetchWidget(s_MrmHierarchy, "manipPPScreen", toplevel_widget,
		     &manipPPScreen, &dummy_class) != MrmSUCCESS)
    s_error("can't fetch manipPP window");

  /* Go get the delatOmegaPPScreen part of the application. */
  if (MrmFetchWidget(s_MrmHierarchy, "deltaOmegaPPScreen", toplevel_widget,
		     &deltaOmegaPPScreen, &dummy_class) != MrmSUCCESS)
    s_error("can't fetch infoAnswer window");

  /* Go get the delatXYpartprocScreen part of the application. */
  if (MrmFetchWidget(s_MrmHierarchy, "deltaXYpartprocScreen", toplevel_widget,
		     &deltaXYpartprocScreen, &dummy_class) != MrmSUCCESS)
    s_error("can't fetch infoAnswer window");


  /* Now go looking for fonts */
  if(robemFontList != 0)
    {/* They've specified an override font */
      XtSetArg(arglst[0], XmNfontList,  robemFontList);
      XtSetArg(arglst[1], XmNlabelFontList,  robemFontList);
      XtSetArg(arglst[2], XmNbuttonFontList,  robemFontList);
      XtSetValues(optionScreenWidget, arglst, 3);
      XtSetValues(editBoxWidget, arglst, 3);
      XtSetValues(colorScreenWidget, arglst, 3);
      XtSetValues(main_window_widget, arglst, 3);
      XtSetValues(imgNumTextWidget, arglst, 3);
      XtSetValues(pointLabelWidget[1], arglst, 3);
      XtSetValues(pointLabelWidget[2], arglst, 3);
      XtSetValues(pointLabelWidget[3], arglst, 3);
      XtSetValues(pointLabelWidget[4], arglst, 3);
      XtSetValues(toplevel_widget, arglst, 3);
      XtSetValues(infoBox, arglst, 3);
      XtSetValues(imgNumButtonW, arglst, 3);
      XtSetValues(XtParent(main_window_widget), arglst, 3);
      XtSetValues(toplevel_widget, arglst, 3);
      XFlush(dsply);
    }/* They've specified an override font */
  
  /* Manage the main part and realize everything.  The interface comes up
   * on the display now.  Manage all of the screens now so we can alter
   * what we need to in XLibInitialize() - Unmanage the unneeded ones
   * thereafter.
   */
  inited= FALSE;
  XtManageChild(partprocscreen);
  XtManageChild(editBoxScreenWidget);
  XtManageChild(mapShadeScreenWidget);
  XtManageChild(colorScreenWidget);
  XtManageChild(optionScreenWidget);
  XtManageChild(fftscreen);  
  XtManageChild(rotfltscreen); 
  XtManageChild(circAvgScreenWidget);
  XtManageChild(emcoravgscreen);
  XtManageChild(imgconvscreen);
  XtManageChild(statscreen);
/*  XtManageChild(refscreen);
  XtManageChild(appodizeScreenWidget);
  XtManageChild(boxScreenWidget);
  XtManageChild(infoBox); */
  XtManageChild(infoAnswerscreen); 
  XtManageChild(normalScreen); 
  XtManageChild(subtractMapScreen); 
  XtManageChild(main_window_widget);
  XFlush(dsply);

  XlibInitialize();

  /* Let's make sure we're sized right */
  XtSetArg(arglst[0], XmNwidth,  500);
  XtSetArg(arglst[1], XmNheight, 700);
  XtSetArg(arglst[2], XmNunitType, XmPIXELS);
  XtSetValues(optionScreenWidget, arglst, 3);
  XtResizeWindow(optionScreenWidget);
  XFlush(dsply);
  XtSetArg(arglst[0], XmNwidth,  550);
  XtSetArg(arglst[1], XmNheight, 550);
  XtSetValues(editBoxWidget, arglst, 3);
  XtResizeWindow(editBoxWidget);
  XFlush(dsply);

  if(!debugSwitch || debugLevel!=4)
    {
      XtUnmanageChild(showDebugToggleW); 
      XtUnmanageChild(showBoxingStatsToggleW); 
      XtUnmanageChild(boxGrid2W); 
      showDebug = FALSE;
      showBoxingSwitch = FALSE;
    }

  XtUnmanageChild(subtractMapScreen); 
  XtUnmanageChild(normalScreen); 
  XtUnmanageChild(infoBox);
  XtUnmanageChild(appodizeScreenWidget); 
  XtUnmanageChild(boxScreenWidget); 
  XtUnmanageChild(refscreen); 
  XtUnmanageChild(statscreen);
  XtUnmanageChild(emcoravgscreen);
  XtUnmanageChild(imgconvscreen);
  XtUnmanageChild(circAvgScreenWidget);
  XtUnmanageChild(fftscreen); 
  XtUnmanageChild(rotfltscreen); 
  XtUnmanageChild(colorScreenWidget);
  XtUnmanageChild(mapShadeScreenWidget);
  XtUnmanageChild(editBoxScreenWidget);
  XtUnmanageChild(optionScreenWidget);
  XtUnmanageChild(icosRadProjToggleW);
  XtUnmanageChild(radProjToggleW);
  XtUnmanageChild(radProjCSToggleW);
  XtUnmanageChild(partprocscreen);
  XtUnmanageChild(parmTogsW);
  XtUnmanageChild(projasuW);
  XtUnmanageChild(infoAnswerscreen); 
  XFlush(dsply);
  XSync(dsply,0);
  XFlush(dsply);


  /* Create our cursors */
  watch_cursor= XCreateFontCursor(dsply, XC_watch);
  cross_cursor= XCreateFontCursor(dsply, XC_cross);
  bogosity_cursor= XCreateFontCursor(dsply, XC_hand1);
  if(hiColor)
    XRecolorCursor(dsply, cross_cursor,
		   &color_defs[YELLOW],
		   &color_defs[BLACK]);
  else
    XRecolorCursor(dsply, cross_cursor,
		   &color_defs[pixels_return[YELLOW]],
		   &color_defs[pixels_return[BLACK]]);
  XFlush(dsply);
  inited= TRUE;
  ref_box_flag = FALSE;
  
  /* masterProcess only needed for 8 bit displays */
  if(!masterProcess)
    {/* DOn't let the child-ren play with this... */
      XtSetArg(arglst[0], XmNsensitive, FALSE);
      XtSetValues(colorButtonW, arglst, 1);  
      XtSetValues(colorToolWidget, arglst, 1);  
    }/* DOn't let the child-ren play with this... */

  if(!liteSwitch)
    {/* Don't let them choose display redraws */
      XtUnmanageChild(drawLiteButtonW);
      XtUnmanageChild(drawFFTLiteButtonW);
      XtUnmanageChild(drawFFT2LiteButtonW);
    }/* Don't let them choose display redraws */
  /* *dispThis = image_window1;   */

  /* See if they want to configure things for themselves */
  readResourceFile();

  /* Put up version number */
  sprintf(outStr,"Robem: [Version %s]",versionStr);
  createSimpleString(&simpleString,outStr);
  XtSetArg(arglst[0], XmNdialogTitle, simpleString );
  XtSetArg(arglst[1], XmNrecomputeSize, FALSE );
  XtSetValues(main_window_widget, arglst, 2); 
  XFlush(dsply);
  XmStringFree(simpleString); 


  /* See if they want to do some automatic operations */
  checkAutomatic();


  /* Sit around forever waiting to process X-events.  We never leave
   * XtAppMainLoop. From here on, we only execute our callback routines. */
  XtAppMainLoop(app_context);

  return(0);

}/* MAIN */



/*-----------------------------------------------------------------------*/
/*-----------------------------------------------------------------------*/
/*-----------------------------------------------------------------------*/
/*************************************************************************
 * NON-Callback Routines - These rotuines are of the more generic C      *
 * language variety and are not called directly from X.  The callback    *
 * routines that are called by X do make heavy use of these, however.    *
 *************************************************************************/
/*-----------------------------------------------------------------------*/
/*-----------------------------------------------------------------------*/
/*-----------------------------------------------------------------------*/

/*************************************************************************
 * checkAutomatic() - Check to see if they've requested anything from
 * the command line to be done automatically by robem.
 *
 * A NOTE ABOUT USING XVFB.  If we are running in batch mode we want
 * to have Xvfb running as a virtual XServer by doing this sometime on the
 * machine (probably at startup):
 *
 *     % Xvfb :1 -screen 1 1600x1200x16
 *
 * Then for the Xterm that's running robem in batch mode:
 *
 *     % setenv DISPLAY :1.1
 *
 * Then:
 *
 *     % robem -pre a.pif -noise 2.0 SHOT
 *
 * etc ...
 *
 * All XEvents then get served by the Xvfb process and the user can
 * go on with their other processing needs as this runs in background.
 *
 *
*/
void checkAutomatic()
{/* CHECKAUTOMATIC */
  int tag,outWidth,outHeight;
  char *s1;

  if(forceAvgBkgSwitch )
    {/* setup toggle */
      XtSetArg(arglst[0], XmNset, TRUE );
      XtSetValues(forceAvgBkgToggleW, arglst, 1); 
      forceAvgBkg = TRUE;
    }/* setup toggle */

  /* Set the zoom stuff first */
  if(zoomItSwitch)
    {/* zoom */
      sizeFactor = zoomItFactor;
      sprintf(outStr,"%.2f",sizeFactor);
      XmTextSetString(zoomTextW, outStr);   /* Feedback */
      XFlush(dsply);

      /* I need a log base 2 routine "log2()" that will allow
       * me to convert the input gotten above into meaningful
       * zoomFactors. 
       * e.g.
       *
       *      zoomFactor = (float)log2(sizeFactor);
       *
       */
      zoomFactor = (float)rwalog2(sizeFactor);
      zoomFactor = MAX(-4.0,MIN(zoomFactor,4.0));

      zoomItSwitch = FALSE;
    }/* zoom */

  /* See if they want to preload something */
  if(preloadSwitch)
    {/* Get it */
      readWrite= READING;
      gotFilename(0,0,0);
    }/* Get it */

  /* See if they want normalize and linear gradient */
  if(dumpNormStats)
    {/* normalize stats */
      tag = 1;

      normalize_activate(0, &tag, 0);
      exit(0);
    }/* normalize stats */


  /* See if they want spherical average */
  if(dumpSpherAvg)
    {/* spherical average */
      FILE *fDump;
      float *thisRint;
      fDump = fopen(dumpSpherAvgFile,"w");
      thisRint   = circAvg[0].rint;
      if(fDump)
	{/* ok */
	  for(i=0;i<(numRows/2)+1;i++)
	    fprintf(fDump,"%d %f\n",
		    i,
		    thisRint[i]
		    );
	  fclose(fDump);
	}/* ok */

      exit(0);
    }/* spherical average */


  /* Do lingrad */
  if(doLinAlso)	
    {/* linear gradient */
      printf("Linear Gradient \n");
      linGrad_activate(0,0,0); 
    }/* linear gradient */

  /* See if they want to fix values */
  if(doFix)
    {/* fix */

      printf("Remove blemishes\n");
      /* WHY ???
	 tag = 1;
	 fix_activate(0, &tag, 0); */

      tag = FIX2D;
      removeBlemish_activate(0,0,0);

      fix2D_do_activate(0, &tag, 0);

      if(!normLinSwitch)
	{/* write it in normlin area */
	  s1 = strrchr(fileName.addr,'.');
	  if(s1)
	    *s1 = '\0';
	  if(doOutFile)
	    sprintf(outStr,"%s",outFname);
	  else
	    sprintf(outStr,"%s_fix.pif",fileName.addr);
	  sprintf(fileName.addr,"%s",outStr);
	  /* Calc our dimensions */
	  if(packRadius <= 0)
	    outHeight= outWidth = purdueHeader.numCols;
	  else outHeight= outWidth = (packRadius*2)+1;	
	  if(purdueHeader.numCols != purdueHeader.numRows)
	    {outHeight = purdueHeader.numRows;
	    outWidth  = purdueHeader.numCols;}
	  write_pif(outWidth, outHeight, fileName.addr, PACKEDPORTABLE);
	  exit(0);
	}/* write it in normlin area */
    }/* fix */

  /* See if they want normalize and linear gradient */
  if(normLinSwitch)
    {/* normalize and linear gradient */

      printf("Normalize\n");
      tag = 1;

      normalize_activate(0, &tag, 0);
      normalize_activate_2(0, &tag, 0);

      if(!reBoxSwitch)
	{/* write it in rebox area */
	  s1 = strrchr(fileName.addr,'.');
	  if(s1)
	    *s1 = '\0';
	  if(doOutFile)
	    sprintf(outStr,"%s",outFname);
	  else
	    sprintf(outStr,"%s_norm.pif",fileName.addr);
	  sprintf(fileName.addr,"%s",outStr);
	  /* Calc our dimensions */
	  if(packRadius <= 0)
	    outHeight= outWidth = purdueHeader.numCols;
	  else outHeight= outWidth = (packRadius*2)+1;	
	  if(purdueHeader.numCols != purdueHeader.numRows)
	    {outHeight = purdueHeader.numRows;
	    outWidth  = purdueHeader.numCols;}
	  write_pif(outWidth, outHeight, fileName.addr, PACKEDPORTABLE);
	  exit(0);
	}/* write it in rebox area */
    }/* normalize and linear gradient */


  /* See if they want to rebox an exisitng boxed PIF file  */
  if(reBoxSwitch)
    {/* rebox it */

      printf("Rebox\n");

      /* Actually do the reboxing */
      boxRadius = autoBoxBoxSize;
      featherRadius = autoBoxFeatherSize;

      s1 = strrchr(fileName.addr,'.');
      if(s1)
	*s1 = '\0';
      if(doOutFile)
	sprintf(outStr,"%s",outFname);
      else
	sprintf(outStr,"%s_rebox.pif",fileName.addr);
      sprintf(fileName.addr,"%s",outStr);
      /* Calc our dimensions */
      if(packRadius <= 0)
	outHeight= outWidth = purdueHeader.numCols;
      else outHeight= outWidth = (packRadius*2)+1;	
      if(purdueHeader.numCols != purdueHeader.numRows)
	{outHeight = purdueHeader.numRows;
	 outWidth  = purdueHeader.numCols;}

      write_pif(outWidth, outHeight, fileName.addr, PACKEDPORTABLE);
      exit(0);
    }/* rebox it */

  /* See if they want autoBoxing */
  if(autoBoxSwitch)
    {/* autobox */
      char thisFileName[512];

      emmaRadius = autoBoxSize;
      emptyShellWidth = 5;
      boxRadius  = autoBoxBoxSize;
      featherRadius  = autoBoxFeatherSize;

      /* First autobox */
      tag = AUTOBOX;
      autoBox_activate(0, &tag, 0);
      /* Then correlate */
      for(i=0; i<numBoxes; i++)
	box[i].reference = 1; /* Turn them all ON */
      tag = CORRELATE;
      autoBox_activate(0, &tag, 0);

      /* Let's write the box coords */
      sprintf(thisFileName,"%s",fileName.addr);
      s1= strrchr(thisFileName,'.');
      if(s1)
	sprintf(s1,".bcrd");
      else sprintf(thisFileName,"%s.bcrd",fileName.addr);
      writeBoxCoords(thisFileName, BOXCOORDS);

      raiseIt(statusScreenWidget);
      XFlush(dsply);
      XSync(dsply,0);

      /* Let's write the file */
      s1 = strrchr(fileName.addr,'.');
      if(s1)
	*s1 = '\0';
      if(doOutFile)
	sprintf(outStr,"%s",outFname);
      else
	sprintf(outStr,"%s_box.pif",fileName.addr);
      sprintf(fileName.addr,"%s",outStr);

      outHeight= outWidth = (boxRadius*2)+1;
      /* Shouldn't we use the featherRadius here to determine
       * outWidth and outHeight for write_pif() ??? */
      outHeight= outWidth = (featherRadius*2)+1;
      write_pif(outWidth, outHeight, fileName.addr, CIRCLEPACKED);
      exit(0);
    }/* autobox */


  /* See if they want autoBoxing */
  if(boxitSwitch)
    {/* boxit */

      /* Read box coords */
      readBoxCoords(boxitFname);

      /* Override with what we've got from command line */
      emmaRadius = boxitSize;
      emptyShellWidth = 5;
      boxRadius  = boxitBoxSize;
      featherRadius  = boxitFeatherSize;

      /* Let's write the file */
      s1 = strrchr(fileName.addr,'.');
      if(s1)
	*s1 = '\0';
      if(doOutFile)
	sprintf(outStr,"%s",outFname);
      else
	sprintf(outStr,"%s_box.pif",fileName.addr);
      sprintf(fileName.addr,"%s",outStr);

      raiseIt(statusScreenWidget);
      XFlush(dsply);
      XSync(dsply,0);

      outHeight= outWidth = (boxRadius*2)+1;
      /* Shouldn't we use the featherRadius here to determine
       * outWidth and outHeight for write_pif() ??? */
      outHeight= outWidth = (featherRadius*2)+1;
      write_pif(outWidth, outHeight, fileName.addr, CIRCLEPACKED);
      exit(0);
    }/* boxit */


  /* See if they want to add noise */
  if(doNoise)
    {/* noise */

      tag = FIX2D;

      fix_do_activate(0, &tag, 0);

      s1 = strrchr(fileName.addr,'.');
      if(s1)
	*s1 = '\0';
      if(doOutFile)
	sprintf(outStr,"%s",outFname);
      else
	{
	  if(noiseType == GAUSSNOISE)
	    sprintf(outStr,"%s_%.1fgausnoise.pif",fileName.addr,noiseStd);
	  else
	    sprintf(outStr,"%s_%.1fshotnoise.pif",fileName.addr,noiseStd);
	}
      sprintf(fileName.addr,"%s",outStr);
      /* Calc our dimensions */
      if(packRadius <= 0)
	outHeight= outWidth = purdueHeader.numCols;
      else outHeight= outWidth = (packRadius*2)+1;	
      if(purdueHeader.numCols != purdueHeader.numRows)
	{outHeight = purdueHeader.numRows;
	 outWidth  = purdueHeader.numCols;}
      if(pifMode==20)
	write_pif(outWidth, outHeight, fileName.addr, MAP);
      else
	write_pif(outWidth, outHeight, fileName.addr, PACKEDPORTABLE);
      exit(0);
    }/* noise */


  /* See if they want to apply CTF */
  if(doApplyCtf)
    {/* applyCTF */

      idim1= idim2= 1024;   /* MAX fft size */
      CK(fftImage_gold= (float *)malloc(sizeof(float)*(idim1*(idim2*2))));
      writeFFTType = APPLYCTFR4 ;
      ctfFilter = applyCtfFilter;

      tag = 0;
      fft_activate(0, &tag, 0);
      ctf_all_activate();

      exit(0);
    }/* applyCTF */


  /* See if they want to trunc */
  if(doTrunc)
    {/* trunc */

      if(fileType!=MAP)
	{/* error */
	  printf("\n************************************************\n");
	  printf("* Cannot TRUNC this file type [%d]\n",fileType);
	  printf("************************************************\n");
	  exit(0);
	}/* error */

      truncadd_do_activate();
      
      s1 = strrchr(fileName.addr,'.');
      if(s1)
	*s1 = '\0';
      if(doOutFile)
	sprintf(outStr,"%s",outFname);
      else
	sprintf(outStr,"%s_trunc.pif",fileName.addr);
      sprintf(fileName.addr,"%s",outStr);
      outHeight = purdueHeader.numRows;
      outWidth  = purdueHeader.numCols;
      write_pif(outWidth, outHeight, fileName.addr, MAP);
      exit(0);
    }/* trunc */



  /* See if they want to interp */
  if(doInterp)
    {/* interp */
      if(fileType!=EMPORTABLE &&
	 fileType!=BYTEPORTABLE &&
	 fileType!=PORTABLE)
	{/* error */
	  printf("\n************************************************\n");
	  printf("* Cannot interp this file type [%d]\n",fileType);
	  printf("************************************************\n");
	  exit(0);
	}/* error */

      interpolate_activate();
      
      s1 = strrchr(fileName.addr,'.');
      if(s1)
	*s1 = '\0';
      if(doOutFile)
	sprintf(outStr,"%s",outFname);
      else
	sprintf(outStr,"%s_interp.pif",fileName.addr);
      sprintf(fileName.addr,"%s",outStr);
      outHeight = purdueHeader.numRows;
      outWidth  = purdueHeader.numCols;
      write_pif(outWidth, outHeight, fileName.addr, EMPORTABLE);
      exit(0);
    }/* interp */



}/* CHECKAUTOMATIC */




/******************************************************
 * ie3tvax() - Convert IEEE floats in Ultrix/C to 
 * VAX format float.  Never could get this to work right.
*/
void ie3tvax(int *inFloat)
{/* IE3TVAX */
  int tmpFloat;

  tmpFloat= (int)*inFloat;

  *inFloat= 0;
  *inFloat |= ((tmpFloat & 0177777) << 16);
  *inFloat |= ((tmpFloat & 037777600000) >> 16);
  *inFloat |= 0400;

}/* IE3TVAX */




/*******************************************************
 * vaxtie3() - Convert VAX format floats to IEEE floats 
 * in Ultrix/C. Never could get this to work right.
*/
void vaxtie3(unsigned int *inFloat)
{/* VAXTIE3 */
  unsigned int tmpFloat;

  tmpFloat= (unsigned int)*inFloat;

  *inFloat= 0;
  *inFloat |= ((tmpFloat & 037777600000) >> 16);
  *inFloat |= ((tmpFloat & 0177777) << 16);
  *inFloat &= 037677777777;   /* Turn this bit OFF */


/*  Here's some fortran code I got off the net to do this conversion.
    Implement it when I understand it.
c
c convert single precision floating point numbers from vax representation
c to ieee while running on rs6000. originally vax version of ie3fos from
c cernlib.
c modifications indicated by    ! <<<
c
 
      subroutine ie3fvax (ms,mt,nwdo,jbad)
c
c cern proglib# m220    ie3fos          .version kernvax  2.20  861204
 
c-    convert single precision for output with copy
c-    from source in native to target in ieee data format
 
      integer      nwdo                 ! <<<
c      dimension    ms(99), mt(99)      ! <<<
      dimension    ms(nwdo), mt(nwdo)   ! <<<
 
      double precision   thdb
      dimension    this(2)
      equivalence  (thdb,this)
      equivalence  (itha,tha,this(1)), (ithb,thb,this(2))
 
      parameter    (mskb16 = '00008000'x)
      parameter    (ibadcs = '7f80fe00'x)
      parameter    (ibadcd = '7ff01fc0'x)
      parameter    (iovpcs = '7f800000'x)
      parameter    (iovpcd = '7ff00000'x)
      parameter    (iovncs = 'ff800000'x)
      parameter    (iovncd = 'fff00000'x)
 
 
      jbad = 0
      jms  = 0
      jmt  = 0
 
  301 do 349  jl=1,nwdo
      itha = ms(jms+1)
      if (itha.eq.0)         go to 347
c     jsign = itha .and. mskb16                ! <<<
      jsign = and(itha,mskb16)                 ! <<<
c     jexp  = jibits (itha,7,8)                ! <<<
      jexp  = ibits (itha,7,8)                 ! <<<
      if (jexp.le.2)         go to 331
      if (jexp.eq.255)       go to 333
c     jmant = jishft(jibits(itha,0,7), 16) .or. jibits(itha,16,16)    ! <<<
      jmant = or(ishft(ibits(itha,0,7), 16),ibits(itha,16,16))        ! <<<
c     itha  = jishft(jsign,16) .or. jishft(jexp-2,23) .or. jmant      ! <<<
      itha  = or( or(ishft(jsign,16),ishft(jexp-2,23)), jmant)        ! <<<
      go to 347
 
c--      zero / nan
  331 if (jsign.eq.0)        go to 341
      if (jexp.ne.0)         go to 341
      ithb = ibadcs
      go to 336
 
c--      overflow
  333 if (jsign.eq.0)    then
          ithb = iovpcs
        else
          ithb = iovncs
        endif
  336 jbad = jl
      itha = ithb
      go to 347
 
  341 itha = 0
  347 mt(jmt+1) = itha
      jmt = jmt + 1
  349 jms = jms + 1
 
      return
      end
*/

}/* VAXTIE3 */




/*************************************************************
 * calcScaleFactor() - Calculate the factor needed to convert
 * out floatIntegers into floats when we read this file later.
 * Just scan all of our values and pick a value that maintains 
 * our dynamic rarenge.
 * Future goal - need to clean up all of the different versions
 * of calculating scale factors into one single rotine.
*/
void calcScaleFactor(float *floatScaleFactor, 
		     float *recipFloatScaleFactor, 
		     float *int4scaleF, 
		     float img_min, float img_max,
		     float img_avg, float img_std,
		     int fileType)
{/* CALCSCALEFACTOR */
  float minVal,maxVal;

  minVal = -360.0;  /* To account for alp,bet,gam etc... */
  maxVal = +360.0;

  minVal= MIN(minVal,(float)img_min);
  minVal= MIN(minVal,(float)img_max);
  minVal= MIN(minVal,img_avg);
  minVal= MIN(minVal,img_std);

  maxVal= MAX(maxVal,(float)img_min);
  maxVal= MAX(maxVal,(float)img_max);
  maxVal= MAX(maxVal,img_avg);
  maxVal= MAX(maxVal,img_std);

  /* Get greatest magnitude */
  if(minVal < 0.0)
    minVal = 0.0 - minVal;
  maxVal = MAX(minVal,maxVal);
  /* Now let's give it a little extra so we're guaranteed not
   * to run over the top */
  maxVal *= 1.02;

  /*********
    printf("maxshort [%d,%f]    maxint[%d]\n",MAXSHORT,(float)MAXSHORT,MAXINT);
    printf("ft,map [%d,%d] \n",fileType,MAP);
    printf("maxVal=[%f]\n",maxVal);
  *********/

  /* We're doing MAPs as shorts so make it smaller */
  if(fileType==MAP)
    *floatScaleFactor = (double)((double) MAXSHORT / (double) maxVal);
  else
    *floatScaleFactor = (double)((double) MAXINT / (double) maxVal);

  *int4scaleF = (double)((double) MAXINT / (double) maxVal);

  *recipFloatScaleFactor = (double)((double)1.0 / (double)*floatScaleFactor); 

}/* CALCSCALEFACTOR */



/************************************************************
 * today() - Return todays date and current time in STRING 
 * (static) form.
*/
char *today()
{/* TODAY */
  static char s1[132];
  char *ampm;
  int seconds,
      minutes,
      hours,
      hyear,
      hmonth,
      hday;
  struct tm *tim;
  time_t tSec;

  tSec= time(NULL); 
  tim= localtime((time_t *)&tSec); 

  seconds= tim->tm_sec;
  minutes= tim->tm_min;
  hours=   tim->tm_hour;
  hday=    tim->tm_mday;
  hmonth=  tim->tm_mon + 1;
  hyear=   tim->tm_year + 1900;

  if(hours > 12)
    {/* PM */
      hours -= 12;
      ampm= "PM";
    }/* PM */
  else if(hours==12)
    ampm= "PM";
  else ampm= "AM";

  sprintf(s1,"%02d/%02d/%04d,  %02d:%02d:%02d%s",
         hmonth, hday, hyear, hours, minutes, seconds, ampm);

  return(s1);
}/* TODAY */





/****************************************************************
 * calcShortStats() - Calculates statistics from the short image
 * from the disk.
 * Future goal - combine all calc stats type routines into one
 * routine.  Be cure to take into account what type of stats 
 * are needed e.g. boxing particles has a different method than
 * for say MAP data etc...
*/
void calcShortStats(unsigned short *inbuffer, int *img_bkg, int *img_min,
		    int *img_max, float *img_avg, float *img_std,
		    int width, int height, int boxWidth)
{/* CALCSHORTSTATS */
  int i,j;
  int pixelValue,tot;
  float totValues;
  double var,sumsq;

  /* Init all */
  *img_bkg = 0;
  *img_min = 9999;
  *img_max = -9999;
  *img_min = INT_MAX;  /* dgm INT_MAX instead of 9999 */
  *img_max = INT_MIN;  /* dgm INT_MIN instead of -9999 */
  *img_avg = 0;
  *img_std = 0;

  /* Accumulate intermediate values */
  totValues = tot = 0;
  var = sumsq = 0.0;
  for(i=0;i<height;i++)
    {/* height */
      for(j=0;j<width;j++)
	{/* Look at all pixels */
	  pixelValue = (int)inbuffer[((i*boxWidth)+j)];
	  if(pixelValue > 0)
	    {/* Only use non-zeros */
	      *img_min = MIN(*img_min,pixelValue);
	      *img_max = MAX(*img_max,pixelValue);
	      sumsq += (pixelValue * pixelValue);
	      totValues +=  (float)pixelValue;
	      tot++;
	    }/* Only use non-zeros */
	}/* Look at all pixels */
/* printf("[%d][%d] - %d\n",i,22,(int)inbuffer[((i*boxWidth)+22)]); */
    }/* height */
  
  /* Calc the background */
  *img_bkg= (short)calcShortBackground(inbuffer, width, height, boxWidth );

  /* Calc stats */
  *img_avg = (float)((float)totValues / (float)(width * height));
  *img_std = 0.0;
  var = (sumsq / (float)(width*height)) - (*img_avg * *img_avg);
  if(var > 0.0)
    *img_std = (float)sqrt(var);

}/* CALCSHORTSTATS */




/******************************************************
 * calcRectMedian() - Calculates median from the 
 * selected rectangle.
 * Future goal - combine all calc stats type routines into one
 * routine.  Be cure to take into account what type of stats 
 * are needed e.g. boxing particles has a different method than
 * for say MAP data etc...
*/
void calcRectMedian(float *inbuffer, int *median,
		    int width, int height)
{/* CALCRECTMEDIAN */
  int i,j;
  int tot,totSoFar;
  float pixelValue;
  float totValues;
  int histVals[257];

  /* Init all */
  *median = 0.0;
  for(i=0;i<257;i++)
    histVals[i]= 0;

  /* Accumulate intermediate values */
  totValues = 0.0;
  tot = 0;
  for(i=0;i<height;i++)
    {/* height */
      for(j=0;j<width;j++)
	{/* Look at all pixels */
	  pixelValue = (float)inbuffer[((i*width)+j)];
	  histVals[(int)pixelValue]++;
	  totValues +=  (float)pixelValue;
	  tot++;
	}/* Look at all pixels */
    }/* height */
  
  totSoFar = 0;
  for(i=0;i<257;i++)
    {/* Look for the median value */
      totSoFar += histVals[i];
      if(totSoFar >= (tot/2))
	{/* We've reached the half-way point */
	  *median = i;
	  break;
	}/* We've reached the half-way point */
    }/* Look for the median value */


}/* CALCRECTMEDIAN */



/******************************************************
 * calcStats() - Calculates statistics from the image
 * Future goal - combine all calc stats type routines into one
 * routine.  Be cure to take into account what type of stats 
 * are needed e.g. boxing particles has a different method than
 * for say MAP data etc...
*/
void calcStats(float *inbuffer, float *img_bkg, float *img_min,
	       float *img_max, float *img_avg, float *img_std,
	       int width, int height, int boxWidth)
{/* CALCSTATS */
  int i,j,tot,totBkg;
  float pixelValue;
  float totValues,totBkgValues;
  double var,sumsq;

  /* Init all */
  *img_bkg = 0;
  *img_min = 9999;
  *img_max = -9999;
  *img_min = INT_MAX;  /* dgm INT_MAX instead of 9999 */
  *img_max = INT_MIN;  /* dgm INT_MIN instead of -9999 */
  *img_avg = 0;
  *img_std = 0;

  /* Accumulate intermediate values */
  totValues = 0.0;
  tot = 0;
  totBkgValues = 0.0;
  totBkg = 0;
  var = sumsq = 0.0;
  for(i=0;i<height;i++)
    {/* height */
      for(j=0;j<width;j++)
	{/* Look at all pixels */
	  pixelValue = (float)inbuffer[((i*width)+j)];
	  *img_min = MIN(*img_min,pixelValue);
	  *img_max = MAX(*img_max,pixelValue);
	  if(pixelValue > 0)
	    {/* Only use non-zeros */
	      sumsq += (pixelValue * pixelValue);
	      totValues +=  (float)pixelValue;
	      tot++;
	    }/* Only use non-zeros */
	  if(i==0 || j==0 || (i==(height-1)) || (j==(width-1)))
	     {/* Accum for RECT background */
	       totBkgValues += pixelValue;
	       totBkg++;
	     }/* Accum for RECT background */
	}/* Look at all pixels */
    }/* height */
  
  /* Calc the background */
  if(fileType==BYTEPORTABLE || fileType==EMPORTABLE || fileType==REBOX)
    *img_bkg= (float)calcFloatBackground(inbuffer, width, height, boxWidth,
				       fileType,
				       bkgrndMethod,
				       boxMethod,
				       packRadius,
				       boxRadius,
				       featherRadius,
				       recipSizeFactor,
				       tryColors );
  else if(fileType==RECTPACKED && totBkg > 0)
    {
      *img_bkg = (int)(totBkgValues / (float)totBkg);
    }
  else *img_bkg = inbuffer[0]; /* e.g. Don't recalc Packed boxes or MAPS!! */

  /* Calc stats */
  *img_avg = (float)((float)totValues / (float)(width * height));
  *img_std = 0.0;
  var = (sumsq / (float)(width*height)) - (*img_avg * *img_avg);
  if(var > 0.0)
    *img_std = (float)sqrt(var);
  else *img_std = (float)sqrt(fabs(var));

}/* CALCSTATS */



/******************************************************
 * calcByteStats() - Calculates statistics from the image
 * Future goal - combine all calc stats type routines into one
 * routine.  Be cure to take into account what type of stats 
 * are needed e.g. boxing particles has a different method than
 * for say MAP data etc...
*/
void calcByteStats(BYTE *inbuffer, float *img_bkg, float *img_min,
	       float *img_max, float *img_avg, float *img_std,
	       int width, int height, int boxWidth)
{/* CALCBYTESTATS */
  int i,j;
  int pixelValue,tot,totBkg;
  float totValues,totBkgValues;
  double var,sumsq;

  /* Init all */
  *img_bkg = 0;
  *img_min = 9999;
  *img_max = -9999;
  *img_min = INT_MAX;  /* dgm INT_MAX instead of 9999 */
  *img_max = INT_MIN;  /* dgm INT_MIN instead of -9999 */
  *img_avg = 0;
  *img_std = 0;

  /* Accumulate intermediate values */
  totValues = 0.0;
  tot = 0;
  totBkgValues = 0.0;
  totBkg = 0;
  var = sumsq = 0.0;
  for(i=0;i<height;i++)
    {/* height */
      for(j=0;j<width;j++)
	{/* Look at all pixels */
	  pixelValue = (int)inbuffer[((i*width)+j)];
	  *img_min = MIN(*img_min,pixelValue);
	  *img_max = MAX(*img_max,pixelValue);
	  if(pixelValue > 0)
	    {/* Only use non-zeros */
	      sumsq += (pixelValue * pixelValue);
	      totValues +=  (float)pixelValue;
	      tot++;
	    }/* Only use non-zeros */
	  if(i==0 || j==0 || (i==(height-1)) || (j==(width-1)))
	     {/* Accum for RECT background */
	       totBkgValues += pixelValue;
	       totBkg++;
	     }/* Accum for RECT background */
	}/* Look at all pixels */
    }/* height */
  
  /* Calc the background */
  if(fileType==BYTEPORTABLE || fileType==EMPORTABLE || fileType==REBOX)
    *img_bkg= (float)calcBackground(inbuffer, width, height, boxWidth,
				     fileType,
				     bkgrndMethod,
				     boxMethod,
				     packRadius,
				     boxRadius,
				     featherRadius,
				     recipSizeFactor,
				     tryColors );
  else if(fileType==RECTPACKED && totBkg > 0)
    {
      *img_bkg = (float)(totBkgValues / (float)totBkg);
    }
  else *img_bkg = inbuffer[0]; /* e.g. Don't recalc Packed boxes or MAPS!! */

  /* Calc stats */
  *img_avg = (float)((float)totValues / (float)(width * height));
  *img_std = 0.0;
  var = (sumsq / (float)(width*height)) - (*img_avg * *img_avg);
  if(var > 0.0)
    *img_std = (float)sqrt(var);
  else *img_std = (float)sqrt(fabs(var));

}/* CALCBYTESTATS */



/*************************************************************
 * particleCalcStats() - Calculates statistics from the 
 * particle X values.
*/
void particleCalcStats(int *part_min, int *part_max, 
		       float *part_avg, float *part_std)
{/* PARTICLECALCSTATS */
  P_INFO *thisInfo;
  int i,j;
  int pixelValue,tot;
  float totValues;
  double var,sumsq;

  /* Init all */
  *part_min = 9999;
  *part_max = -9999;
  *part_min = INT_MAX;  /* dgm INT_MAX instead of 9999 */
  *part_max = INT_MIN;  /* dgm INT_MIN instead of -9999 */
  *part_avg = 0;
  *part_std = 0;

  /* Accumulate intermediate values */
  totValues = 0.0;
  tot = 0;
  var = sumsq = 0.0;
  thisInfo = info.info;
  while(thisInfo != NULL)
    {/* Do all of them */

      pixelValue = (int)thisInfo->x;
      *part_min = MIN(*part_min,pixelValue);
      *part_max = MAX(*part_max,pixelValue);
      sumsq += (pixelValue * pixelValue);
      totValues +=  (float)pixelValue;
      tot++;

      thisInfo = thisInfo->next;
      
    }/* Do all of them */
  
  /* Calc stats */
  *part_avg = (float)((float)totValues / (float)tot);
  *part_std = 0.0;
  var = (sumsq / (float)(tot)) - (*part_avg * *part_avg);
  if(var > 0.0)
    *part_std = (float)sqrt(var);
  else *part_std = (float)sqrt(fabs(var));

}/* PARTICLECALCSTATS */



/************************************************************
 * calcScanStats() - Calculates statistics from the image
 * but only from the scan inside the image.
 * Future goal - combine all calc stats type routines into one
 * routine.  Be cure to take into account what type of stats 
 * are needed e.g. boxing particles has a different method than
 * for say MAP data etc...
*/
void calcScanStats(float *inbuffer, 
		   float *img_bkg, float *img_min,
		   float *img_max, float *img_avg, float *img_std,
		   int width, int height, int boxWidth)
{/* CALCSCANSTATS */
  int i,j,tot;
  float pixelValue;
  float totValues,distnce;
  double var,sumsq;

  /* Init all */
  /* DON't DO THIS WE MAY NEED THE VALUE THAT'S THERE!!! *img_bkg = 0; */
  *img_min = 9999;
  *img_max = -9999;
  *img_min = INT_MAX;  /* dgm INT_MAX instead of 9999 */
  *img_max = INT_MIN;  /* dgm INT_MIN instead of -9999 */
  *img_avg = 0;
  *img_std = 0;

  /* Accumulate intermediate values */
  totValues = 0.0;
  tot = 0;
  var = sumsq = 0.0;
  for(i=0;i<height;i++)
    {/* height */
      for(j=0;j<width;j++)
	{/* Look at all pixels */
	  pixelValue = (float)inbuffer[((i*width)+j)];
	  *img_min = MIN(*img_min,pixelValue);
	  *img_max = MAX(*img_max,pixelValue);
	  sumsq += (pixelValue * pixelValue);
	  totValues +=  (float)pixelValue;
	  tot++;
	}/* Look at all pixels */
    }/* height */
  
  /* Calc stats */
  *img_avg = (float)((float)totValues / (float)(width * height));
  *img_std = 0.0;
  var = (sumsq / (float)(width*height)) - (*img_avg * *img_avg);
  if(var > 0.0)
    *img_std = (float)sqrt(var);
  else *img_std = (float)sqrt(fabs(var));

  *img_bkg = (int)*img_avg;

}/* CALCSCANSTATS */



/************************************************************
 * calcSubScanStats() - Calculates statistics from the image
 * but only from the subregion inside the image.
 * Future goal - combine all calc stats type routines into one
 * routine.  Be cure to take into account what type of stats 
 * are needed e.g. boxing particles has a different method than
 * for say MAP data etc...
*/
void calcSubScanStats(float *inbuffer, 
		      float *img_bkg, float *img_min,
		      float *img_max, float *img_avg, float *img_std,
		      int width, int height, int boxWidth, 
		      int offsetX, int offsetY)
{/* CALCSUBSCANSTATS */
  int i,j,tot;
  float pixelValue;
  float totValues,distnce;
  double var,sumsq;

  /* Init all */
  /* DON't DO THIS WE MAY NEED THE VALUE THAT'S THERE!!! *img_bkg = 0; */
  *img_min = 9999;
  *img_max = -9999;
  *img_min = INT_MAX;  /* dgm INT_MAX instead of 9999 */
  *img_max = INT_MIN;  /* dgm INT_MINinstead of -9999 */
  *img_avg = 0.0;
  *img_std = 0.0;

  /* Accumulate intermediate values */
  totValues = 0.0;
  tot = 0;
  var = sumsq = 0.0;
  for(i=0;i<height;i++)
    {/* height */
      for(j=0;j<width;j++)
	{/* Look at all pixels */
	  pixelValue = (float)inbuffer[(((i+offsetY)*boxWidth)+j+offsetX)];
	  *img_min = MIN(*img_min,pixelValue);
	  *img_max = MAX(*img_max,pixelValue);
	  sumsq += (pixelValue * pixelValue);
	  totValues +=  (float)pixelValue;
	  tot++;
	}/* Look at all pixels */
    }/* height */
  
  /* Calc stats */
  *img_avg = (float)((float)totValues / (float)(tot));
  *img_std = 0.0;
  var = (sumsq / (float)(width*height)) - (*img_avg * *img_avg);
  if(var > 0.0)
    *img_std = (float)sqrt(var);
  else *img_std = (float)sqrt(fabs(var));

  *img_bkg = (int)*img_avg;

}/* CALCSUBSCANSTATS */



/************************************************************
 * calcCircleStats() - Calculates statistics from the image
 * but only from the circle inside the image.
 * Future goal - combine all calc stats type routines into one
 * routine.  Be cure to take into account what type of stats 
 * are needed e.g. boxing particles has a different method than
 * for say MAP data etc...
*/
void calcCircleStats(float *inbuffer, 
		     float *img_bkg, float *img_min,
		     float *img_max, float *img_avg, float *img_std,
		     int width, int height, int boxWidth)
{/* CALCCIRCLESTATS */
  int i,j,tot;
  float pixelValue;
  float totValues,distnce;
  double var,sumsq;
  int xcen,ycen,iMycen,jMxcen,thisRadius,xSq,ySq;

  /* Init all */
  /* DON't DO THIS WE MAY NEED THE VALUE THAT'S THERE!!! *img_bkg = 0; */
  *img_min = 9999;
  *img_max = -9999;
  *img_min = INT_MAX;  /* dgm INT_MAX instead of 9999 */
  *img_max = INT_MIN;  /* dgm INT_MIN instead of -9999 */
  *img_avg = 0;
  *img_std = 0;

  if(fileType==PACKEDPORTABLE || fileType==REBOX)
    {
      if(packRadius <= 0)
	thisRadius= height+width;
      else thisRadius= packRadius-borderWidth;
    }
    thisRadius= boxRadius;

  /* Accumulate intermediate values */
  totValues = tot = 0;
  var = sumsq = 0.0;
  xcen = width / 2;
  ycen = height / 2;
  //  printf("xcen = ycen = [%d,%d]\n",xcen,ycen);
  //  printf("height=[%d]  width = [%d]\n",height,width);
  //  printf("thisRadius = [%d]\n",thisRadius);

  for(i=0;i<height;i++)
    {/* height */
      iMycen = (i)-ycen;
      xSq    = (iMycen)*(iMycen);
      for(j=0;j<width;j++)
	{/* Look at all pixels */
	  pixelValue = (float)inbuffer[((i*boxWidth)+j)];
	  jMxcen = (j)-xcen;
	  ySq= (jMxcen)*(jMxcen);
	  distnce = (float)sqrt((double)(xSq+ySq)); 
	  if(distnce<= thisRadius)
	    {/* we've found one in the circle */
	      *img_min = MIN(*img_min,pixelValue);
	      *img_max = MAX(*img_max,pixelValue);
	      sumsq += (pixelValue * pixelValue);
	      totValues +=  (float)pixelValue;
	      tot++;
	    }/* we've found one in the circle */
	}/* Look at all pixels */
    }/* height */
  // printf("tot=[%d] totValues = [%f]  sumsq = [%f]\n",tot,totValues,sumsq);

  /* Calc the background */
  if(fileType==BYTEPORTABLE || fileType==EMPORTABLE  || fileType==PORTABLE 
     || fileType==REBOX)
    {/* They are boxing from a scan */
      if(backgroundMode==USECIRCLE)
	*img_bkg= (float)calcFloatBackground(inbuffer, width, height, boxWidth,
					     fileType,
					     bkgrndMethod,
					     boxMethod,
					     packRadius,
					     boxRadius,
					     featherRadius,
					     recipSizeFactor,
					     tryColors);
      else if(backgroundMode==USERECT)
	{
	  printf("calcRectBackground not yet implemented.\n");
	  /* *img_bkg= (short)calcRectBackground(
	   inbuffer, width, 
	   height, boxWidth );  */
	}
      else if(backgroundMode==PICK)
	*img_bkg = inbuffer[0]; 
      else
	printf("Unknown background method [%d]\n",backgroundMode);
    }/* They are boxing from a scan */
  else if(packRadius <= 0)
    {/* e.g. Don't recalc Packed boxes that have no radius!!! */
      *img_bkg = purdueHeader.img_bkg;
    }/* e.g. Don't recalc Packed boxes that have no radius!!! */
  else *img_bkg = inbuffer[0]; /* e.g. Don't recalc Packed boxes or MAPS!! */

  /* Calc stats */
  /* Only so many are used for tot */
  *img_avg = (float)((float)totValues / (float)(tot));
  *img_std = 0.0;
  var = (sumsq / (float)(tot)) - (*img_avg * *img_avg);
  // New MISHA way ??? var = (sumsq  - (*img_avg * *img_avg))  / (float)tot;
  // printf("var[%f]  avg[%f]\n",var,*img_avg);

  if(var > 0.0)
    *img_std = (float)sqrt(var);
  else if(var == 0.0)
    *img_std = (float)0.0;
  else *img_std = (float)sqrt(fabs(var));
  // printf("std[%f]\n",*img_std);

}/* CALCCIRCLESTATS */





/***************************************************************
 * calcRealStats() - Calculates statistics from the real image.
 * Future goal - combine all calc stats type routines into one
 * routine.  Be cure to take into account what type of stats 
 * are needed e.g. boxing particles has a different method than
 * for say MAP data etc...
*/
void calcRealStats(float *inbuffer, 
		   float *img_bkg, float *img_min,
		   float *img_max, float *img_avg, float *img_std,
		   int width, int height, int boxWidth, int nimg)
{/* CALCREALSTATS */
  int i,j,k;
  int tot;
  float pixelValue;
  float totValues,distnce;
  double var,sumsq;
  int xcen,ycen,iMycen,jMxcen,thisRadius,xSq,ySq;

  /* Init all */
  /* DON't DO THIS WE MAY NEED THE VALUE THAT'S THERE!!! *img_bkg = 0; */
  *img_min = 9999.0;
  *img_max = -9999.0;
  *img_min = HUGE;  /* dgm HUGE instead of 9999.0 */
  *img_max = -HUGE; /* dgm -HUGE instead of -9999.0 */
  *img_avg = 0.0;
  *img_std = 0.0;

  if(fileType==PACKEDPORTABLE || fileType==REBOX)
    {
      if(packRadius <= 0)
	thisRadius= height+width;
      else thisRadius= packRadius;
    }
  else thisRadius= boxRadius;

  /* Accumulate intermediate values */
  totValues = tot = 0;
  var = sumsq = 0.0;
  xcen = ycen = thisRadius;
  for(k=0;k<nimg;k++)
    {/* Do all imgs */
      for(i=0;i<height;i++)
	{/* height */
	  iMycen = (i*recipSizeFactor)-ycen;
	  xSq    = (iMycen)*(iMycen);
	  for(j=0;j<width;j++)
	    {/* Look at all pixels */
	      pixelValue = (float)inbuffer[((i*width)+j)];
	      
	      jMxcen = (j*recipSizeFactor)-xcen;
	      ySq= (jMxcen)*(jMxcen);
	      distnce = (float)sqrt((double)(xSq+ySq)); 
/* FOrce this to always happen - e.g. does radius mean anything here ??? */
distnce = 0.0;
	      if(distnce<= thisRadius)
		{/* we've found one in the circle */
		  *img_min = MIN(*img_min,pixelValue);
		  *img_max = MAX(*img_max,pixelValue);
/*	      if(pixelValue > 0.0)
    {/+ Only use non-zeros */
		  sumsq += (pixelValue * pixelValue);
		  totValues +=  (float)pixelValue;
		  tot++;
/*		}/+ Only use non-zeros */
		}/* we've found one in the circle */
	    }/* Look at all pixels */
	}/* height */
    }/* Do all imgs */
  printf("totValues[%f] to[%d]\n",totValues,tot);
      
  /* Calc stats */
  *img_avg = (float)((float)totValues / (float)(width * height));
  *img_std = 0.0;
  var = (sumsq / (float)(width*height)) - (*img_avg * *img_avg);
  if(var > 0.0)
    *img_std = (float)sqrt(var);
  else *img_std = (float)sqrt(fabs(var));
  
  /* Accumulate intermediate values for background */
  totValues = tot = 0;
  var = sumsq = 0.0;
  xcen = ycen = thisRadius;
  for(i=0;i<height;i++)
    {/* height */
      /* Left hand side */
      pixelValue = (float)inbuffer[((i*width))];
      totValues +=  (float)pixelValue;
      tot++;
      /* Right hand side */
      pixelValue = (float)inbuffer[((i*width)+(width-1))];
      totValues +=  (float)pixelValue;
      tot++;
    }/* height */
  for(i=0;i<width;i++)
    {/* width */
      /* Left hand side */
      pixelValue = (float)inbuffer[i];
      totValues +=  (float)pixelValue;
      tot++;
      /* Right hand side */
      pixelValue = (float)inbuffer[(((height-1)*width)+i)];
      totValues +=  (float)pixelValue;
      tot++;
    }/* width */
  
  /* Calc the background - just average all values around the perimeter
   * of the square */
  if(tot > 0)
    *img_bkg = (float)((float)totValues / (float)tot);

}/* CALCREALSTATS */


/**************************************************************
 * raiseIt() - Do everything necessary to raise a window to the
 * foreground on the display.
*/
void raiseIt(Widget w)
{/* RAISEIT */

  XtManageChild(w);
  XtPopup(XtParent(w), XtGrabNone);

  XmProcessTraversal(w, XmTRAVERSE_CURRENT); 
  XtRealizeWidget(w);  
  XRaiseWindow(dsply, XtWindow(XtParent(w)));
  XFlush(dsply);


}/* RAISEIT */




/**************************************************************
 * createSimpleString() - Create a simple XmString string .
*/
void createSimpleString(XmString *simpStr, char * outStr)
{/* CREATESIMPLESTRING */

  *simpStr= XmStringCreateLocalized(outStr);

}/* CREATESIMPLESTRING */




/*************************************************************
 * doLabel () -  DSet the label widget with the input string.
*/
void doLabel(char *str, Widget wid)
{/* doLabel */
  XmString simpleString;

  createSimpleString(&simpleString,str);
  XtSetArg(arglst[0], XmNlabelString, simpleString );
  XtSetArg(arglst[1], XmNrecomputeSize, FALSE );
  XtSetValues(wid, arglst, 2); 
  XFlush(dsply);
  XmStringFree(simpleString); 

}/* doLabel */





/**************************************************************************
 * createDefaultColormap() - DO whatever is necessary to get us the desired 
 * colormap.
*/
void createDefaultColormap()
{/*  CREATEDEFAULTCOLORMAP */
  FILE *fColor;
  int num_visuals,screen;
  XVisualInfo * visual_info;
  XVisualInfo * vi;
  XVisualInfo vinfo;
  int class;
  Bool              contig;
  unsigned int      nplanes;
  int numColormaps;
  int success;
  int depth, status;
  float scaleFactor;
  XColor screen_def_rtn, exact_def_rtn;


  printf("Default colormap.\n");
/*  CREATEDefaultCOLORMAP */
  visual= DefaultVisual(dsply, DefaultScreen(dsply));
  visual_depth = (unsigned int)DefaultDepth(dsply, DefaultScreen(dsply));
  
/*  CREATEDefaultCOLORMAP */

  /* Pull the switch */
  hiColor = TRUE;

  
/*  CREATEDefaultCOLORMAP */


  /* Allocate the color cells */
  CK(pixels_return= (unsigned long *)malloc(sizeof(unsigned long)*256));
  CK(dummy_pixels_return= (unsigned long *)malloc(sizeof(unsigned long)*256));
  /* Allocate the color definition structs */
  if (color_defs == NULL) 
    CK(color_defs = (XColor *) malloc (1025 * sizeof (XColor) ));
  if (mirror_color_defs == NULL) 
    CK(mirror_color_defs = (XColor *) malloc (1025 * sizeof (XColor) ));
  if (gen_color_defs == NULL) 
    CK(gen_color_defs = (XColor *) malloc (1025 * sizeof (XColor) ));
      
  if(masterProcess)
    {/* Create the colormap here is MASTER */


      printf("visual==[%d]\n",visual);
      
      color_map = XCreateColormap(dsply, DefaultRootWindow(dsply),
				  visual, AllocNone);
      printf("DEC colormap = [%d]\n",color_map);


/*  CREATEDefaultCOLORMAP */

      if (color_map == NULL) {
	fprintf (stderr, "Could not create color map for visual\n");
	exit (1);
      }
      
      /* Now let's query the DISPLAY until we get the maximum number of
       * contiguous colorcell entries in the colormap that we can .
       */
      success= FALSE;
      tryColors= 256;
      nplanes = 1;
      contig = True;
      plane_masks_return = (unsigned long *)malloc(sizeof(unsigned long)*
						   nplanes);
      dummy_plane_masks_return = (unsigned long *)malloc(sizeof(unsigned long)*
						   nplanes);
      nplanes = 0;   

/*  CREATEDefaultCOLORMAP */

      /* Init the color defs to point to each entry */
      for (i = 0; i <256 ; i++) 
	{/* Make the color map gray-scale */
	  color_defs[i].pixel= i;
	  gen_color_defs[i].pixel= i;
	  pixels_return[i] = i;
	}/* Make the color map gray-scale */

      /* Find out what's already there in the color table */
      XQueryColors(dsply, color_map, color_defs, 256);
      XQueryColors(dsply, color_map, gen_color_defs, 256);

      sprintf(outStr,"%s/cmap.robem",getenv("HOME"));      
      fColor = fopen(outStr,"w");

      if(fColor)
	{/* Write it */
	  fwrite(&tryColors,sizeof(int),1,fColor);
	  for(i=0;i<tryColors;i++)
	    {/* Write it all out */
	      fwrite(&pixels_return[i],sizeof(unsigned long),1,fColor);
	    }/* Write it all out */
	}/* Write it */
      else
	{/* Trouble */
	  printf("ERROR - Cannot write colormap file.\n");
	}/* Trouble */
      fclose(fColor);

    }/* Create the colormap here is MASTER */
  else
    {/* Read from file */

      sprintf(outStr,"%s/cmap.robem",getenv("HOME"));
      fColor = fopen(outStr,"r");

      if(fColor)
	{/* Read it */
	  fread(&tryColors,sizeof(int),1,fColor);
	  for(i=0;i<tryColors;i++)
	    {/* Read it all out */
	      fread(&pixels_return[i],sizeof(unsigned long),1,fColor);
	    }/* Read it all out */
	}/* Read it */
      else
	{/* SOmethings wrong */
	  printf("ERROR - Cannot read colormap file.\n");
	}/* SOmethings wrong */
      fclose(fColor);
      
    }/* Read from file */

  
/*  CREATEDefaultCOLORMAP */

  /* Now change the colormap for drawing in pixels_return[GREEN,RED,BLUE] 
   * First calculate some fudge factors.
   */
  pixelColor= (65535.0/256.0);
  scaleFactor = 255.0 / ((float)tryColors - 7.0);
  
  /* Make these entries in the colormap GRAYSCALE */
  for(i=6;i<(tryColors);i++)
    {/* Do the colors we have been assigned */
      gen_color_defs[pixels_return[i]].pixel = pixels_return[i];
      gen_color_defs[pixels_return[i]].red   = 
	(((i-6.0) * scaleFactor) * pixelColor);
      gen_color_defs[pixels_return[i]].green =
	(((i-6.0) * scaleFactor) * pixelColor);
      gen_color_defs[pixels_return[i]].blue  = 
	(((i-6.0) * scaleFactor) * pixelColor);
      gen_color_defs[pixels_return[i]].flags = DoRed | DoGreen | DoBlue;
      color_defs[pixels_return[i]].pixel = pixels_return[i];
      color_defs[pixels_return[i]].red   = (unsigned short)
	(((i-6.0) * scaleFactor) * pixelColor);
      color_defs[pixels_return[i]].green = (unsigned short)
	(((i-6.0) * scaleFactor) * pixelColor);
      color_defs[pixels_return[i]].blue  = (unsigned short)
	(((i-6.0) * scaleFactor) * pixelColor);
      color_defs[pixels_return[i]].flags = DoRed | DoGreen | DoBlue;
      color_defs[pixels_return[i]].pad = 0;
      gen_color_defs[pixels_return[i]].pad = 0;
      if(masterProcess)
	success = XAllocColor (dsply, color_map, 
			       &color_defs[i]);
      pixels_return[i] = color_defs[i].pixel;
      if(hiColor)
	gen_color_defs[i].pixel =  i;
    }/* Do the colors we have been assigned */
  
  
/*  CREATEDefaultCOLORMAP */

  /* Now define our robem colors */
  color_defs[WHITE].pixel = WHITE;
  color_defs[WHITE].red   = 255 * pixelColor;
  color_defs[WHITE].green = 255 * pixelColor;
  color_defs[WHITE].blue  = 255 * pixelColor;
  color_defs[WHITE].flags = DoRed | DoGreen | DoBlue;
  if(masterProcess)
    {XAllocColor (dsply, color_map, &color_defs[WHITE]);
     pixels_return[WHITE] = color_defs[WHITE].pixel;}

  color_defs[YELLOW].pixel = YELLOW;
  color_defs[YELLOW].red   = 255 * pixelColor;
  color_defs[YELLOW].green = 255 * pixelColor;
  color_defs[YELLOW].blue  = 0 * pixelColor;
  color_defs[YELLOW].flags = DoRed | DoGreen | DoBlue;
  if(masterProcess)
    {XAllocColor (dsply, color_map, &color_defs[YELLOW]);
     pixels_return[YELLOW] = color_defs[YELLOW].pixel ;}

  color_defs[GREEN].pixel = GREEN;
  color_defs[GREEN].red   = 0;
  color_defs[GREEN].green = 255 * pixelColor;
  color_defs[GREEN].blue  = 0;
  color_defs[GREEN].flags = DoRed | DoGreen | DoBlue;
  if(masterProcess)
    {XAllocColor (dsply, color_map, &color_defs[GREEN]);
     pixels_return[GREEN] = color_defs[GREEN].pixel;}

  color_defs[pixels_return[RED]].pixel = RED;
  color_defs[pixels_return[RED]].red   = 255 * pixelColor;
  color_defs[pixels_return[RED]].green = 0;
  color_defs[pixels_return[RED]].blue  = 0;
  color_defs[pixels_return[RED]].flags = DoRed | DoGreen | DoBlue;
  if(masterProcess)
    {XAllocColor (dsply, color_map, &color_defs[pixels_return[RED]]);
     pixels_return[RED] = color_defs[RED].pixel;}

  color_defs[BLUE].pixel = BLUE;
  color_defs[BLUE].red   = 0;
  color_defs[BLUE].green = 0;
  color_defs[BLUE].blue  = 255 * pixelColor;
  color_defs[BLUE].flags = DoRed | DoGreen | DoBlue;
  if(masterProcess)
    {XAllocColor (dsply, color_map, &color_defs[BLUE]);
     pixels_return[BLUE] = color_defs[BLUE].pixel;}

  color_defs[PLACEHOLDER].pixel = PLACEHOLDER;
  color_defs[PLACEHOLDER].red   = 0 * pixelColor;
  color_defs[PLACEHOLDER].green = 255 * pixelColor;
  color_defs[PLACEHOLDER].blue  = 0;
  color_defs[PLACEHOLDER].flags = DoRed | DoGreen | DoBlue;
  if(masterProcess)
    {XAllocColor (dsply, color_map, &color_defs[PLACEHOLDER]);
     pixels_return[PLACEHOLDER] = color_defs[PLACEHOLDER].pixel;}

/*  CREATEDefaultCOLORMAP */

  color_defs[BLACK].pixel = BLACK;
  color_defs[BLACK].red   = 0 * pixelColor;
  color_defs[BLACK].green = 0 * pixelColor;
  color_defs[BLACK].blue  = 0 * pixelColor;
  color_defs[BLACK].flags = DoRed | DoGreen | DoBlue;
  if(masterProcess)
    {XAllocColor (dsply, color_map, &color_defs[BLACK]);
     pixels_return[BLACK] = color_defs[BLACK].pixel;}
  
 noHiColor:;

}/*  CREATEDEFAULTCOLORMAP */






/**************************************************************************
 * create24Colormap() - DO whatever is necessary to get us the desired 
 * colormap.
 * Experiment - For 24 bit display with robem try this
 * line in your .Xdefaults file:
 * 
 *     robem*useColorObj:       false
 *     robem*background:	MidnightBlue
 *     robem*foreground:	lightBlue
 * 
 * This turns off the use of entries in the default colormap to color
 * widgets which leads to bogus values being used when we create
 * our own 24 bit colormap. 
 *
*/
void create24Colormap()
{/*  CREATE24COLORMAP */
  FILE *fColor;
  int num_visuals,screen;
  XVisualInfo * visual_info;
  XVisualInfo * vi;
  XVisualInfo vinfo;
  int class;
  XVisualInfo * found_vi;
  Bool              contig;
  unsigned int      nplanes;
  int numColormaps;
  int success;
  int depth, status;
  float scaleFactor;
  XColor screen_def_rtn, exact_def_rtn;



  /* [1.2] Find Appropriate Visual */
  visual_info = XGetVisualInfo (dsply, VisualNoMask, NULL, &num_visuals);
  if (visual_info == NULL || num_visuals == 0) 
    {
      fprintf (stderr, "XGetVisualInfo failed\n");
      exit (1);
    }	    
  found_vi = NULL;
  
/*  CREATE24COLORMAP */

  /* [1.3] Find visual such that:
   * desired_class = 2;                    ==2 for PseudoColor 
   * desired_depth *= 3;			needed if separate colors
   * desired_depth_delta = 3;
   *
   *	1. desired_depth is as large as possible (up to true desired depth)
   *	2. visual depth is the smallest of those supported >= desired_depth
   *	3. visual class is the `most desired' for the image type
   */
  vi = visual_info;
  for(i=0;i<num_visuals;i++)
    {/* Look for one we want */
      printf("Found depth[%d] class[%d]\n",vi->depth,vi->class);
      if (vi->depth >= 16 && (vi->class == 4))  /* PseudoColor==3 
						  Direct==5 
						  TrueColor ==4 class */
	found_vi = vi;
      if(found_vi)
	break;
      vi++;
    }/* Look for one we want */
  
  
/*  CREATE24COLORMAP */

  if (found_vi == NULL) 
    {/* no go */
      printf("Could not find > 8 bit visual type.\n");
      hiColor = FALSE;
      goto noHiColor;
    }/* no go */
  else
    {/* Get screen */
      Screen *scr;

      /* Pull the switch */
      hiColor = TRUE;

      /* Use this for our 24 bit screens */
      screen = found_vi->screen;
      root_window = RootWindow (dsply, screen); 

      scr = ScreenOfDisplay(dsply, screen);
      printf("Min Cmaps[%d]\n",XMinCmapsOfScreen(scr));
      printf("Max Cmaps[%d]\n",maxCmaps= XMaxCmapsOfScreen(scr));

    }/* Get screen */

  
/*  CREATE24COLORMAP */


  /* Get visual for default 8 bits screens */
  visual= DefaultVisual(dsply, DefaultScreen(dsply));
  visual_depth = (unsigned int)DefaultDepth(dsply, DefaultScreen(dsply));
  
  /* Allocate the color cells */
  CK(pixels_return= (unsigned long *)malloc(sizeof(unsigned long)*256));
  CK(dummy_pixels_return= (unsigned long *)malloc(sizeof(unsigned long)*256));
  /* Allocate the color definition structs */
  if (color_defs == NULL) 
    CK(color_defs = (XColor *) malloc (1025 * sizeof (XColor) ));
  if (mirror_color_defs == NULL) 
    CK(mirror_color_defs = (XColor *) malloc (1025 * sizeof (XColor) ));
  if (gen_color_defs == NULL) 
    CK(gen_color_defs = (XColor *) malloc (1025 * sizeof (XColor) ));
      
printf("masterProcess=[%d]\n",masterProcess);
printf("PseudoColor[%d]  DirectColor[%d] TrueColor[%d]\n",
       PseudoColor,DirectColor,TrueColor);
      
  if(masterProcess)
    {/* Create the colormap here is MASTER */


      class = found_vi->class;
      depth = found_vi->depth;

      status = XMatchVisualInfo(dsply, screen, depth, class, &vinfo);
      if( status == 0) 
	{
	  printf("No Matching Visuals\n");
	  exit(1);
	}
      printf("Visual ID: %x\n", vinfo.visualid);
      visual       = vinfo.visual;
      visual_depth = depth;
  
      printf("visual==[%d]\n",visual);
      
      color_map = XCreateColormap(dsply, DefaultRootWindow(dsply),
				  visual, AllocNone);
      printf("DEC colormap = [%d]\n",color_map);

      i = 0;
      XtSetArg(args[i], XtNcolormap, color_map); i++;
      XtSetArg(args[i], XmNvisual, visual); i++;
      XtSetArg(args[i], XmNdepth, visual_depth); i++;
      XtSetValues(toplevel_widget,args,i);
      XFlush(dsply);


/*  CREATE24COLORMAP */

      if (color_map == NULL) {
	fprintf (stderr, "Could not create color map for visual\n");
	exit (1);
      }
      
      /* Now let's query the DISPLAY until we get the maximum number of
       * contiguous colorcell entries in the colormap that we can .
       */
      success= FALSE;
      tryColors= 256;
      nplanes = 1;
      contig = True;
      plane_masks_return = (unsigned long *)malloc(sizeof(unsigned long)*
						   nplanes);
      dummy_plane_masks_return = (unsigned long *)malloc(sizeof(unsigned long)*
						   nplanes);
      nplanes = 0;   

/*  CREATE24COLORMAP */

      /* Init the color defs to point to each entry */
      for (i = 0; i <256 ; i++) 
	{/* Make the color map gray-scale */
	  color_defs[i].pixel= i;
	  gen_color_defs[i].pixel= i;
	  pixels_return[i] = i;
	}/* Make the color map gray-scale */

      /* Find out what's already there in the color table */
      XQueryColors(dsply, color_map, color_defs, 256);
      XQueryColors(dsply, color_map, gen_color_defs, 256);

      sprintf(outStr,"%s/cmap.robem",getenv("HOME"));
      fColor = fopen(outStr,"w");

      if(fColor)
	{/* Write it */
	  fwrite(&tryColors,sizeof(int),1,fColor);
	  for(i=0;i<tryColors;i++)
	    {/* Write it all out */
	      fwrite(&pixels_return[i],sizeof(unsigned long),1,fColor);
	    }/* Write it all out */
	}/* Write it */
      else
	{/* Trouble */
	  printf("ERROR - Cannot write colormap file.\n");
	}/* Trouble */
      fclose(fColor);

    }/* Create the colormap here is MASTER */
  else
    {/* Read from file */

      sprintf(outStr,"%s/cmap.robem",getenv("HOME"));
      fColor = fopen(outStr,"r");

      if(fColor)
	{/* Read it */
	  fread(&tryColors,sizeof(int),1,fColor);
	  for(i=0;i<tryColors;i++)
	    {/* Read it all out */
	      fread(&pixels_return[i],sizeof(unsigned long),1,fColor);
	    }/* Read it all out */
	}/* Read it */
      else
	{/* SOmethings wrong */
	  printf("ERROR - Cannot read colormap file.\n");
	}/* SOmethings wrong */
      fclose(fColor);
      
    }/* Read from file */

  
/*  CREATE24COLORMAP */

  /* Now change the colormap for drawing in pixels_return[GREEN,RED,BLUE] 
   * First calculate some fudge factors.
   */
  pixelColor= (65535.0/256.0);
  scaleFactor = 255.0 / ((float)tryColors - 7.0);
  
  /* Make these entries in the colormap GRAYSCALE */
  for(i=6;i<(tryColors);i++)
    {/* Do the colors we have been assigned */
      gen_color_defs[pixels_return[i]].pixel = i;
      gen_color_defs[pixels_return[i]].red   = 
	(((i-6.0) * scaleFactor) * pixelColor);
      gen_color_defs[pixels_return[i]].green =
	(((i-6.0) * scaleFactor) * pixelColor);
      gen_color_defs[pixels_return[i]].blue  = 
	(((i-6.0) * scaleFactor) * pixelColor);
      gen_color_defs[pixels_return[i]].flags = DoRed | DoGreen | DoBlue;
      color_defs[pixels_return[i]].pixel = i;
      color_defs[pixels_return[i]].red   = (unsigned short)
	(((i-6.0) * scaleFactor) * pixelColor);
      color_defs[pixels_return[i]].green = (unsigned short)
	(((i-6.0) * scaleFactor) * pixelColor);
      color_defs[pixels_return[i]].blue  = (unsigned short)
	(((i-6.0) * scaleFactor) * pixelColor);
      color_defs[pixels_return[i]].flags = DoRed | DoGreen | DoBlue;
      color_defs[pixels_return[i]].pad = 0;
      gen_color_defs[pixels_return[i]].pad = 0;
      if(masterProcess)
	success = XAllocColor (dsply, color_map, 
			       &color_defs[i]);
      pixels_return[i] = color_defs[i].pixel;

    }/* Do the colors we have been assigned */
  
  
/*  CREATE24COLORMAP */

  /* Now define our robem colors */
  color_defs[WHITE].pixel = WHITE;
  color_defs[WHITE].red   = 255 * pixelColor;
  color_defs[WHITE].green = 255 * pixelColor;
  color_defs[WHITE].blue  = 255 * pixelColor;
  color_defs[WHITE].flags = DoRed | DoGreen | DoBlue;
  if(masterProcess)
    {XAllocColor (dsply, color_map, &color_defs[WHITE]);
     pixels_return[WHITE] = color_defs[WHITE].pixel;}

  color_defs[YELLOW].pixel = YELLOW;
  color_defs[YELLOW].red   = 255 * pixelColor;
  color_defs[YELLOW].green = 255 * pixelColor;
  color_defs[YELLOW].blue  = 0 * pixelColor;
  color_defs[YELLOW].flags = DoRed | DoGreen | DoBlue;
  if(masterProcess)
    {XAllocColor (dsply, color_map, &color_defs[YELLOW]);
     pixels_return[YELLOW] = color_defs[YELLOW].pixel ;}

  color_defs[GREEN].pixel = GREEN;
  color_defs[GREEN].red   = 0;
  color_defs[GREEN].green = 255 * pixelColor;
  color_defs[GREEN].blue  = 0;
  color_defs[GREEN].flags = DoRed | DoGreen | DoBlue;
  if(masterProcess)
    {XAllocColor (dsply, color_map, &color_defs[GREEN]);
     pixels_return[GREEN] = color_defs[GREEN].pixel;}

  color_defs[pixels_return[RED]].pixel = RED;
  color_defs[pixels_return[RED]].red   = 255 * pixelColor;
  color_defs[pixels_return[RED]].green = 0;
  color_defs[pixels_return[RED]].blue  = 0;
  color_defs[pixels_return[RED]].flags = DoRed | DoGreen | DoBlue;
  if(masterProcess)
    {XAllocColor (dsply, color_map, &color_defs[pixels_return[RED]]);
     pixels_return[RED] = color_defs[RED].pixel;}

  color_defs[BLUE].pixel = BLUE;
  color_defs[BLUE].red   = 0;
  color_defs[BLUE].green = 0;
  color_defs[BLUE].blue  = 255 * pixelColor;
  color_defs[BLUE].flags = DoRed | DoGreen | DoBlue;
  if(masterProcess)
    {XAllocColor (dsply, color_map, &color_defs[BLUE]);
     pixels_return[BLUE] = color_defs[BLUE].pixel;}

  color_defs[PLACEHOLDER].pixel = PLACEHOLDER;
  color_defs[PLACEHOLDER].red   = 0 * pixelColor;
  color_defs[PLACEHOLDER].green = 255 * pixelColor;
  color_defs[PLACEHOLDER].blue  = 0;
  color_defs[PLACEHOLDER].flags = DoRed | DoGreen | DoBlue;
  if(masterProcess)
    {XAllocColor (dsply, color_map, &color_defs[PLACEHOLDER]);
     pixels_return[PLACEHOLDER] = color_defs[PLACEHOLDER].pixel;}

/*  CREATE24COLORMAP */

  color_defs[BLACK].pixel = BLACK;
  color_defs[BLACK].red   = 0 * pixelColor;
  color_defs[BLACK].green = 0 * pixelColor;
  color_defs[BLACK].blue  = 0 * pixelColor;
  color_defs[BLACK].flags = DoRed | DoGreen | DoBlue;
  if(masterProcess)
    {XAllocColor (dsply, color_map, &color_defs[BLACK]);
     pixels_return[BLACK] = color_defs[BLACK].pixel;}
  
#define NEEDTHISgf 1
#ifdef NEEDTHISgf
  /* Make this colormap is the one that's used! */
  XListInstalledColormaps(dsply, root_window, &numColormaps);
  printf("numColormaps[%d]\n",numColormaps);
  if(masterProcess)
    XInstallColormap (dsply, color_map);/* Install the color map */

  XListInstalledColormaps(dsply, root_window, &numColormaps);
  printf("After numColormaps[%d]\n",numColormaps);
#endif  /* ifdef NEEDTHISgf */

 noHiColor:;

}/*  CREATE24COLORMAP */





/**************************************************************
 * mainRobem_sub_UNIX() - Mimic FORTRAN call to read .img files.
 * Used for development in UNIX.
 * This rtn is very historic - it does some of the input
 * read for robem.  read_pif() is newer and does the MAP files.
 * Future goal - merge this rtn into read_pif().
*/
void mainRobem_sub_UNIX(unsigned short **int_buffer, 
			fString_t *fName,
			int *i, int *j)
{/* MAINROBEM_SUB_UNIX */
  int k,l,m,writeIt,logicalRecordLength;
  int inWidth,inHeight,imgOffset;
  unsigned short junk,junk1,junk2;
  int dummyBuf[32];    /* junk2 keeps getting overflowed into fIn/fOut
			* (depending on the whim of the compiler/linker)
			* so keep this puppy here for protection */
  char floatScaleStr[22],versionStr[64];
  char junkChar[1024];
  char descriptionStr[180];
  char timeStr[132];
  char micrographStr[116];
  char scanNumberStr[18];
  FILE *fIn,*fOut;
  float inBuf[MAXFILESIZE];
  float i4Buf[MAXFILESIZE];
  XmString simpleString;
  forPifHead_t thisHead;
  fString_t inFile,rWrite;
  int luNum, firstTime;

  inFile.addr = (char *)malloc(sizeof(char)*264);
  rWrite.addr = (char *)malloc(sizeof(char)*64);

  /* Init */
  fOut= fIn= 0;

  /******************************
   *   READING                  *
   ******************************/
  /* We are reading some type of file here either .IMG (Purdue
   * INTEGER*2 format) or .PCK (Purdue packed) or .PPI
   * (Purdue Portable Image format) or .PIF (EM portable).
   */
  
  XtSetArg(arglst[0], XmNsensitive, TRUE);
  XtSetValues(truncaddButtonW, arglst, 1); 
  XtSetValues(removeBlemishButtonW, arglst, 1); 
  XtSetValues(editBoxButtonW, arglst, 1); 
  XtSetValues(pcxButtonW, arglst, 1); 
  XtSetValues(correlateButtonW, arglst, 1);  
  XtSetValues(correlateButton2W, arglst, 1);  
  XtSetValues(autoBox_widget, arglst, 1);  
  XtSetValues(linGradWidget, arglst, 1); 
  /*      XtSetValues(autoBox2_widget, arglst, 1);   */
  
  XtSetArg(arglst[0], XmNsensitive, FALSE);
  XtSetValues(paintMapButtonW, arglst, 1); 
  XtSetValues(mapMovieButtonW, arglst, 1); 
  XtSetValues(mapShadeButtonW, arglst, 1); 
  /*      XtSetValues(fixButtonW, arglst, 1);   */
  XtSetValues(invertButtonW, arglst, 1); 
  XtSetValues(rescaleButtonW, arglst, 1); 
  XtSetValues(modelWidget, arglst, 1); 
  XtSetValues(subtractMapButtonW, arglst, 1); 
  XtSetValues(calcSFButtonW, arglst, 1); 
  /* XtSetValues(RcalcSFButtonW, arglst, 1);  */
  XtSetValues(hardRotateButtonW, arglst, 1); 
  XtSetValues(rotateButtonW, arglst, 1); 
  XtSetValues(fourierFiltWidget, arglst, 1); 
  XtSetValues(addAverageWidget, arglst, 1); 
  XtSetValues(checkBkgWidget, arglst, 1); 
  XtSetValues(normalizeWidget, arglst, 1); 
  XtSetValues(normalizeBkgWidget, arglst, 1); 
  XFlush(dsply);
  
  *i= *j= 0;
  
  fIn = fopen((*fName).addr,"r");
  if(fIn == NULL)
    {/*  Ooops can't read it!!! */
      sprintf(outStr,
	      "Can't read [%s] ",(*fName).addr);
      askThem(outStr,TRUE,XmDIALOG_INFORMATION);
      goto bugOut;
    }/*  Ooops can't read it!!! */
  
  
  if(fIn)
    {/* Read the file */
      
      fclose(fIn);
      
      thisHead.luNum = 7;
      sprintf(inFile.addr,"%s",(*fName).addr);
      inFile.length = strlen((*fName).addr);
      sprintf(rWrite.addr,"r");

      pif_open_(inFile.addr, &(thisHead.luNum), rWrite.addr);

      pif_read_gh(&thisHead);

      thisHead.imgNum = 1;
      pif_read_dh(&thisHead);

      /* Init featherMode */
      if(thisHead.borderType == 1)
	{featherMode=LINEAR;
	XtSetArg(arglst[0], XmNset, TRUE);
	XtSetValues(apLinearW, arglst, 1);
	XtSetArg(arglst[0], XmNset, FALSE);
	XtSetValues(apOldW, arglst, 1);
	XtSetValues(apExponentialW, arglst, 1);
	XtSetValues(apGaussianW, arglst, 1);
	}
      else if(thisHead.borderType == 2)
	{featherMode=GAUSSIAN;
	XtSetArg(arglst[0], XmNset, TRUE);
	XtSetValues(apGaussianW, arglst, 1);
	XtSetArg(arglst[0], XmNset, FALSE);
	XtSetValues(apOldW, arglst, 1);
	XtSetValues(apExponentialW, arglst, 1);
	XtSetValues(apLinearW, arglst, 1);
	}
      else if(thisHead.borderType == 3)
	{featherMode=EXPONENTIAL;
	XtSetArg(arglst[0], XmNset, TRUE);
	XtSetValues(apExponentialW, arglst, 1);
	XtSetArg(arglst[0], XmNset, FALSE);
	XtSetValues(apOldW, arglst, 1);
	XtSetValues(apGaussianW, arglst, 1);
	XtSetValues(apLinearW, arglst, 1);
	}
      else 
	{featherMode=NONE;
	XtSetArg(arglst[0], XmNset, TRUE);
	XtSetValues(apOldW, arglst, 1);
	XtSetArg(arglst[0], XmNset, FALSE);
	XtSetValues(apExponentialW, arglst, 1);
	XtSetValues(apGaussianW, arglst, 1);
	XtSetValues(apLinearW, arglst, 1);
	}

      purdueHeader.numCols= thisHead.numCols;
      purdueHeader.numRows= thisHead.numRows;
      purdueHeader.nimg=  thisHead.nimg;
      
      fileType = readType;
      
      if(thisHead.htype==1)
	{/* All the same */
	  purdueHeader.numCols = thisHead.numCols;
	  purdueHeader.numRows = thisHead.numRows;
	  purdueHeader.numSecs = thisHead.numSecs;
	}/* All the same */
      else
	{/* can\'t goon */
	  char infoStr[64];
	  sprintf(infoStr,"Don't understand this type [%d]\n",
		  thisHead.htype);
	  askThem(infoStr,FALSE,XmDIALOG_INFORMATION);
	  fileType = -99;
	  goto bugOut;
	}/* can\'t goon */
      
      pifMode = thisHead.mode;
      if(pifMode==20 || pifMode==21 || pifMode==22)
	{/* MAP data */
	  purdueHeader.numCols= 
	    purdueHeader.numRows= 
	      *i = *j = 0;
	  askThem("Use Read/MAP to read this file.",FALSE,
		  XmDIALOG_INFORMATION);
	  goto bugOut;
	}/* MAP data */
      if((pifMode==6)||(pifMode==46)||(pifMode==7))
	{/* byte/short datya */
	  /* This is for CIRCLE boxes and also for rectangle
	   * boxes.  
	   */
	  fileType=PACKEDPORTABLE;
	  XtSetArg(arglst[0], XmNsensitive, FALSE);
	  XtSetValues(vertScroll, arglst, 1); 
	  XtSetValues(horiScroll, arglst, 1); 
	}/* byte/short datya */
      else if(pifMode==0)
	{/* byte datya */
	  /* Let's just force them to have one image for now */
	  purdueHeader.nimg = 1;
	  XtSetArg(arglst[0], XmNsensitive, TRUE);
	  XtSetValues(vertScroll, arglst, 1); 
	  XtSetValues(horiScroll, arglst, 1); 
	  fileType=BYTEPORTABLE;
	  /* get dynamic memory */
	  if(*int_buffer != NULL)
	    free(*int_buffer);
	  CK(*int_buffer = (unsigned short *)malloc(
						    ((defDisplayBufferSize*
						      defDisplayBufferSize)*
						 sizeof(unsigned short))+1));
	  
	}/* byte datya */
      else if(pifMode==1)
	{/* short data */
	  XtSetArg(arglst[0], XmNsensitive, TRUE);
	  XtSetValues(vertScroll, arglst, 1); 
	  XtSetValues(horiScroll, arglst, 1); 
	  fileType=PORTABLE;
	  
	  /* get dynamic memory */
	  if(*int_buffer != NULL)
	    free(*int_buffer);
	  CK(*int_buffer = (unsigned short *)malloc(
						    ((defDisplayBufferSize*
						      defDisplayBufferSize)*
						sizeof(unsigned short))+1));
	  
	}/* short data */
      else if(pifMode==2)
	{/* float data */
	  XtSetArg(arglst[0], XmNsensitive, TRUE);
	  XtSetValues(vertScroll, arglst, 1); 
	  XtSetValues(horiScroll, arglst, 1); 
	  fileType=PORTABLE;
	  
	  /* get dynamic memory */
	  if(*int_buffer != NULL)
	    free(*int_buffer);
	  CK(*int_buffer = (unsigned short *)malloc(
						    ((defDisplayBufferSize*
						      defDisplayBufferSize)*
						sizeof(unsigned short))+1));
	}/* float data */
      
      maxPackPixel= -999999999.0;
      minPackPixel= 999999999.0;
      
      firstTime = TRUE;
      for(m=0;m<purdueHeader.nimg;m++)
	{/* Do each image */
	  
	  sprintf(outStr,"Reading image[%d] of [%d] total.",
		  (m+1),purdueHeader.nimg);
	  if(checkStatus(outStr))
	    {/* nimg=m*/
	      goto bugOut;
	    }/* nimg=m*/
	  
	  thisHead.imgNum = m+1;

	  /* [2] Read header for each image */
	  pif_read_dh(&thisHead);

	  purdueHeader.img_bkg = thisHead.img_bkg;
	  if(fileType==PACKEDPORTABLE  && firstTime)
	    {/* Setup params */
	      firstTime = FALSE;
	      borderType = thisHead.borderType;
	      featherRadius= packRadius = thisHead.packRadius;
	      if(thisHead.borderWidth < thisHead.numCols)
		boxRadius = featherRadius - thisHead.borderWidth;
	      else boxRadius = featherRadius;
	      sprintf(outStr,"%d",((2*featherRadius)+1));
	      XmTextSetString(featherRadiusTextWidget, outStr); 
	      sprintf(outStr,"%d",featherRadius-boxRadius);
	      XmTextSetString(boxRadiusTextWidget, outStr); 
	    }/* Setup params */
	  purdueHeader.img_max = thisHead.datai4_max;
	  purdueHeader.img_min = thisHead.datai4_min;
	  purdueHeader.img_avg = thisHead.datar4_avg;
	  purdueHeader.img_std = thisHead.datar4_std;

	  if(fileType==PACKEDPORTABLE)
	    {/* packed portable */
	      if(packImg[m]!=NULL)
		{free(packImg[m]);
		 packImg[m]= NULL;}
	      CK(packImg[m]= (float *)malloc(
					   sizeof(float) *
					   purdueHeader.numCols*
					   purdueHeader.numRows));
	      
	    }/* packed portable */
	  logicalRecordLength= inWidth  = purdueHeader.numCols;
	  inHeight = purdueHeader.numRows;
	  
	  /* Remember where the header ends so we can get the
	   * image buffer in virtual_read_activate()
	   */
	  headOff= 1024;
	  
	  /* Clip this guy to 1k x 1k and let them look at 
	   * 1k x 1k pages at a time */
	  if(fileType!=PACKEDPORTABLE)
	    {/* not packed */
	      inWidth = MAX(0,MIN(defDisplayBufferSize,
				  purdueHeader.numCols));
	      inHeight = MAX(0,MIN(defDisplayBufferSize,
				   purdueHeader.numRows));
	    }/* not packed */
	  else
	    {/* is packed */
	      inWidth  = purdueHeader.numCols;
	      inHeight = purdueHeader.numRows;
	    }/* is packed */
	  
	  /* [3] Now read the image info */
	  if(fileType==PACKEDPORTABLE)
	    {/* packed */
	      int z,view;
	      BYTE *byte_buffer;
	      unsigned short *short_buffer;

	      imgOffset  = thisHead.imgOffset;

	      if(thisHead.mode==6)
		{/* byte */
		  byte_buffer = (BYTE *)malloc(sizeof(BYTE) *
					       (inWidth*inHeight));
		  pif_read_byte_image(&(thisHead.luNum),
				      byte_buffer,
				      &imgOffset,
				      &(thisHead.numCols),
				      &(thisHead.numRows),
				      &(thisHead.numSecs));
		  for(z=0;z<(thisHead.numCols*thisHead.numRows);z++)
		    (packImg[m])[z] = (float)byte_buffer[z];
		  free(byte_buffer);
		}/* byte */
	      else if(thisHead.mode==46)
		{/* float */
		  pif_read_boxi4(&thisHead, packImg[m]);
		}/* float */
	      else if(thisHead.mode==7)
		{/* short */
		  short_buffer = (unsigned short *)malloc(sizeof(unsigned 
								 short) *
					       (inWidth*inHeight));
		  pif_read_short_image(&(thisHead.luNum),
				      short_buffer,
				      &imgOffset,
				      &(thisHead.numCols),
				      &(thisHead.numRows),
				      &(thisHead.numSecs));

		  for(z=0;z<(thisHead.numCols*thisHead.numRows);z++)
		    (packImg[m])[z] = (float)short_buffer[z];
		  free(short_buffer);
		}/* short */
			
	      for(k=0;k<=(inHeight/2);k++)
		{/* Switch it a row at a time */
		  int yOut,yIn;
		  /* We keep these packed images on the disk in Tim's
		   * x,y orientation (e.g. drawn from bottom up so 
		   * if we ever convert to .PCK format it's already 
		   * in the right space.
		   */
		  yOut = (((purdueHeader.numRows-1)-k)*inWidth);
		  yIn  = (k*inWidth);

		  for(l=0;l<logicalRecordLength;l++)
		    {/* COnvert char to int */
		      inBuf[0] = (float)(packImg[m])[(yOut+l)];
		      inBuf[1] = (float)(packImg[m])[(yIn+l)];

		      (packImg[m])[(yOut+l)]= (float)inBuf[1];
		      (packImg[m])[(yIn+l)] = (float)inBuf[0];
		      
		      /* Now find min/max */
		      maxPackPixel= (MAX(maxPackPixel,inBuf[0]));
		      minPackPixel= (MIN(minPackPixel,inBuf[0]));
		      maxPackPixel= (MAX(maxPackPixel,inBuf[1]));
		      minPackPixel= (MIN(minPackPixel,inBuf[1]));

		    }/* COnvert char to int */
		  /* NO pesky EndOfRecord short here... */
		}/* Switch it a row at a time */
	      
	      deletePackedImage[m]= FALSE;
	      
	    }/* packed */
	  else if(fileType==PORTABLE || fileType==BYTEPORTABLE)
	    {/* PORTABLE */
	      
	      /* Pass back to the routine that called us the dimensions
	       * of this image */
	      *i= purdueHeader.numCols;
	      *j= purdueHeader.numRows;
	      file_width  = *i;
	      file_height = *j;
	      
	      /* Setup Scroll bars and display_wdith/height */
	      formResize(0,0,0);   
	      
	      if((file_width > display_width) ||
		 (file_height > display_height) )
		fullScreen= TRUE;
	      else fullScreen= FALSE;
	      
	      /* This changes whenever a new image  */
	      /* is read                            */
	      image_width = defDisplayBufferSize;   
	      image_height = defDisplayBufferSize;  
	      
	      /* If we can just read the whole thing into core
	       * memory then virtual_read_activate() won't have to
	       * work as hard. */
	      if((fileType==BYTEPORTABLE)||(fileType==PORTABLE))
		{/* Let's try and optimize this */
		  BYTE  *byte_buffer;
		  unsigned short *short_buffer;
		  /* Don't let's CK() this guy - we'll just keep
		   * it our secret... :-) */
		  if(fileType==BYTEPORTABLE)
		    CK(byte_buffer= (BYTE *)malloc(sizeof(BYTE)*
						   (file_width)*
						   (file_height)));
		  else if(fileType==PORTABLE && pifMode==1)
		    CK(short_buffer= (unsigned short *)malloc(sizeof(unsigned short)*
						   (file_width)*
						   (file_height)));
		  CK(inCore_buffer= (float *)malloc(sizeof(float)*
						   (file_width)*
						   (file_height)));

		  if(inCore_buffer)
		    imgInCore= TRUE;
		  else 
		    {/* Trouble */
		      imgInCore= FALSE;
		      printf("Trouble allocating inCore_buffer!!!\n");
		      exit(0);
		    }/* Trouble */
		  
		  sprintf(outStr,"Reading image[%d] of [%d] total.\n\
Trying to read entire image into buffer...",
			  (m+1),purdueHeader.nimg);
		  if(checkStatus(outStr))
		    goto bugOut;
		  
		  if(imgInCore)
		    {/* OK to go */
		      int z, status, recordLength;
		      recordLength = (file_width*file_height);
		      imgOffset  = 1024; 
		      if(fileType==BYTEPORTABLE)
			pif_read_byte_image(&(thisHead.luNum),
					    byte_buffer,
					    &imgOffset,
					    &(thisHead.numCols),
					    &(thisHead.numRows),
					    &(thisHead.numSecs));
		      else if(fileType==PORTABLE && pifMode==1)
			pif_read_short_image(&(thisHead.luNum),
					    short_buffer,
					    &imgOffset,
					    &(thisHead.numCols),
					    &(thisHead.numRows),
					    &(thisHead.numSecs));
		      else if(fileType==PORTABLE && pifMode==2)
			pif_read_wmapi4_(&thisHead,
					 inCore_buffer,
					 &(thisHead.numCols),
					 &(thisHead.numRows),
					 &(thisHead.numSecs));
		      if(pifMode != 2)
			{/* dont do for float */
			  for(z=0;z<(file_width*file_height);z++)
			    {
			      if(fileType==BYTEPORTABLE)
				inCore_buffer[z] = (float)(byte_buffer[z]);
			      else if(fileType==PORTABLE)
				inCore_buffer[z] = (float)(short_buffer[z]);
			    }
			}/* dont do for float */
		      status= imgOffset;
		      if(status==0)
			askThem("Trouble reading file [IMGINCORE].",
				FALSE,
				XmDIALOG_INFORMATION);
		      else
			if(purdueHeader.img_max == 0)
			  {/* Find it */
			    int z;
			    float maxP;
			    maxP = 0;
			    for(z=0;z<(file_width*file_height);z++)
			      maxP= MAX(maxP,inCore_buffer[z]);
			    purdueHeader.img_max = maxP;
			  }/* Find it */

		    }/* OK to go */
		  
		  if(fileType==BYTEPORTABLE)
		    free(byte_buffer);
		  else if(fileType==PORTABLE)
		    free(short_buffer);
		  
		}/* Let's try and optimize this */
	      
	      /* DOne below ... 
		 pif_close(&(thisHead.luNum));
	      */

	      sprintf(outStr,"Reading image[%d] of [%d] total.\n\
Writing to display now ...",
		      (m+1),purdueHeader.nimg);
	      if(checkStatus(outStr))
		goto bugOut;
	      virtual_read_activate(-99,-99);

	      XtSetArg(arglst[0], XmNsensitive, TRUE);
	      XtSetValues(interpolateWidget, arglst, 1); 
	      XtSetValues(binningWidget, arglst, 1); 

	    }/* PORTABLE */
	  
	}/* Do each image */
      

      if(minMaxColor)
	{/* Use option screen defaults */
	  char *tourStr;
	  printf("Using option defaults...\n");
	  tourStr= XmTextGetString(minMapTextW);
	  minPackPixel = (float)strtod(tourStr,0);
	  XtFree(tourStr);
	  tourStr= XmTextGetString(maxMapTextW);
	  maxPackPixel = (float)strtod(tourStr,0);
	  XtFree(tourStr);
	}/* Use option screen defaults */


      /* Now that we've read all of the packed iamges we 
       * can display them */
      if(fileType==PACKEDPORTABLE)
	{/* packed */
	  
#define TRY_SCALING_PACKED 1
#ifdef TRY_SCALING_PACKED
	  /* Now clip this guy to our colormap */ 
	  if(minPackPixel < 0.0)
	    scalePackFactor = (((float)tryColors - 7.0) / 
			       (maxPackPixel-minPackPixel));
	  else scalePackFactor = (((float)tryColors - 7.0) / 
				  maxPackPixel);
#else
	  /* Now clip this guy to our colormap */ 
          if(purdueHeader.mode==6)
	    scalePackFactor = (((float)tryColors - 7.0) / 256.0);
          else
	    scalePackFactor = (((float)tryColors - 7.0) / 32000.0);
#endif
	  printf("minPackPixel [%f]  max[%f]     scale[%f]\n",
		 minPackPixel,
		 maxPackPixel,
		 scalePackFactor);
	  
	  stuffPackedImage(purdueHeader.numCols,
			   purdueHeader.numRows,
			   packImg,purdueHeader.nimg,
			   1);  /* Start at zero'th(1) image */
	  
	  *i= defDisplayBufferSize;
	  *j= defDisplayBufferSize;
	  
	  createSimpleString(&simpleString, "ImgNum");
	  XtSetArg(arglst[0], XmNlabelString, simpleString);
	  XtSetArg(arglst[1], XmNrecomputeSize, FALSE );
	  XtSetValues(imgNumButtonW, arglst, 2); 
	  XFlush(dsply);
	  XmStringFree(simpleString); 
	  
	  XtSetArg(arglst[0], XmNsensitive, TRUE);
	  XtSetValues(addAverageWidget, arglst, 1); 
	  XtSetValues(fourierFiltWidget, arglst, 1); 
	  XtSetValues(checkBkgWidget, arglst, 1); 
	  XtSetValues(normalizeWidget, arglst, 1); 
	  XtSetValues(normalizeBkgWidget, arglst, 1); 
	  XtSetValues(linGradWidget, arglst, 1); 
	  XFlush(dsply);
	  
	  sprintf(outStr,"1");
	  XmTextSetString(imgNumTextWidget,outStr);
	  
	  /* We've got some PIF files that don't have the proper
	   * packRadius in them for some reason.  The bug was
	   * it either put 0 in there for packRadius or it put
	   * numCols.  Let's just default to numCols/2 below if we 
	   * encounter the bug.  */
	  /* Always start the numbering of these particles from 0 */
	  lastNumBox = 0;
	  sprintf(outStr,"%d",lastNumBox+1);
	  XmTextSetString(boxInitWidget, outStr);
	  XFlush(dsply);
	  
	}/* packed */
      
      
    }/* EM portable */
  
  /* Not needed - done below...
     pif_close(&(thisHead.luNum));
  */
  borderWidth = thisHead.borderWidth;
  
  /* We've got an historic bug wqhereby either 0 or numCols was
   * placed in packRadius field.
   * We'll default packradius (See note above )  */
  if((packRadius==0 || packRadius==purdueHeader.numCols) &&
     fileType==PACKEDPORTABLE)
    {
      featherRadius = packRadius = purdueHeader.numCols/2;
      if(thisHead.borderWidth < purdueHeader.numCols)
	boxRadius = featherRadius - thisHead.borderWidth;
      else boxRadius = featherRadius;
      sprintf(outStr,"%d",((2*featherRadius)+1));
      XmTextSetString(featherRadiusTextWidget, outStr); 
      sprintf(outStr,"%d",featherRadius-boxRadius);
      XmTextSetString(boxRadiusTextWidget, outStr); 
    }
  /*
     Messes up writing packed boxes where we've added noise e.g.   ??
     */

  if(map_abang == 0.0)
    map_abang = 90.0;
  if(aoverb == 0.0)
    aoverb = 1.0;
  
  sprintf(outStr,"Calculating stats ...");
  if(checkStatus(outStr))
    goto bugOut;
  
  if(fileType==PACKEDPORTABLE)
    {/* Stuff the first one */
      
      /* Just believe the file header */
      calcCircleStats(packImg[0],
		      &purdueHeader.img_bkg, 
		      &purdueHeader.img_min,
		      &purdueHeader.img_max, 
		      &purdueHeader.img_avg, 
		      &purdueHeader.img_std,
		      purdueHeader.numCols, purdueHeader.numRows, 
		      purdueHeader.numCols);
      
      /* Write stats */
      sprintf(outStr,"\nStatistics for file [%s] Image [%d]\n\
\n\
File size   [%d]x[%d]  nimg[%d]\n\
\n\
img_bkg     [%f]\n\
MIN,MAX     [%f],[%f]\n\
AVG,STD     [%f],[%f]\n\
\n\
\n",
	      (*fName).addr, (0+1),
	      purdueHeader.numCols, purdueHeader.numRows, 
	      purdueHeader.nimg,
	      purdueHeader.img_bkg,
	      purdueHeader.img_min, purdueHeader.img_max, 
	      purdueHeader.img_avg, purdueHeader.img_std
	      );
      XmTextSetString(statTextWidget,outStr);
    }/* Stuff the first one */
  else
    {/* Try you rbest */
      /* Write stats */
      sprintf(outStr,"\nStatistics for file [%s] Image[%d]\n\
\n\
File size   [%d]x[%d]\n\
Background  [%f]\n\
MIN,MAX     [%f],[%f]\n\
AVG,STD     [%f],[%f]\n\
",
	      (*fName).addr,1,
	      purdueHeader.numCols,	  purdueHeader.numRows,
	      purdueHeader.img_bkg,
	      purdueHeader.img_min,purdueHeader.img_max,
	      purdueHeader.img_avg,purdueHeader.img_std
	      );
      XmTextSetString(statTextWidget,outStr);
    }/* Try you rbest */
  
  if(fileType==PACKEDPORTABLE)
    {/* Remember packed portable */
      lastPackedNumImgs = purdueHeader.nimg;
      XtSetArg(arglst[0], XmNsensitive, FALSE );
      XtSetValues(useContrastW, arglst, 1); 
      /* XtSetValues(histEqualW, arglst, 1);  We can do this now */
    }/* Remember packed portable */
  else 
    {/* Scanned images */
      lastPackedNumImgs = 0;
      XtSetArg(arglst[0], XmNsensitive, TRUE );
      XtSetValues(useContrastW, arglst, 1); 
      XtSetValues(histEqualW, arglst, 1); 
    }/* Scanned images */

  /* Let's get a baseline of info */
  pif_copy_head_(&thisHead, &pifHeader);

  pifHeader.numCols = purdueHeader.numCols;
  pifHeader.numRows = purdueHeader.numRows;
  pifHeader.nimg    = purdueHeader.nimg;
  pifHeader.img_bkg = purdueHeader.img_bkg;;
  pifHeader.datai4_min = purdueHeader.img_min;
  pifHeader.datai4_max = purdueHeader.img_max;
  pifHeader.datar4_avg = purdueHeader.img_avg;
  pifHeader.datar4_std = purdueHeader.img_std;
  sprintf(pifHeader.descriptionStr,"%s",descriptionStr);

 bugOut:;
  pif_close(&(thisHead.luNum));

  if(fileType != MAP && fileType != PACKEDPORTABLE)
    {XtSetArg(arglst[0], XmNsensitive, FALSE);
    XtSetValues(truncaddButtonW, arglst, 1); }

  free(inFile.addr);
  free(rWrite.addr);

  fflush(NULL);

}/* MAINROBEM_SUB_UNIX */





/********************************************************
 * encput() - used by writePCX() rtn.
*/
encput(byt, cnt, fid)
unsigned char byt, cnt;
FILE *fid;
{
  if (cnt) {
        if ((cnt == 1) && (0xC0 != (0xC0 & byt)))
                {
                if (EOF == putc((int )byt, fid))
                        return(0);     /* disk write error (probably full) */
                return(1);
                }
        else
                {
                if (EOF == putc((int )0xC0 | cnt, fid))
                        return (0);      /* disk write error */
                if (EOF == putc((int )byt, fid))
                        return (0);      /* disk write error */
                return (2);
                }
        }
   return (0);
}

/********************************************************
 * encLine() - used by writePCX() rtn.
 * This subroutine encodes one scanline and writes it to a file.
 * It returns number of bytes written into outBuff, 0 if failed. 
*/
encLine(inBuff, inLen, fp)
unsigned char *inBuff;    /* pointer to scanline data */
int inLen;                        /* length of raw scanline in bytes */
FILE *fp;                        /* file to be written to */
{
unsigned char this, last;
int srcIndex, i;
register int total;
register unsigned char runCount;     /* max single runlength is 63 */
  total = 0;
  runCount = 1;
  last = *(inBuff);

/* Find the pixel dimensions of the image by calculating 
[XSIZE = Xmax - Xmin + 1] and [YSIZE = Ymax - Ymin + 1].  
Then calculate how many bytes are in a "run" */

  for (srcIndex = 1; srcIndex < inLen; srcIndex++)
        {
        this = *(++inBuff);
        if (this == last)     /* There is a "run" in the data, encode it */
                {
                runCount++;
                if (runCount == 63)
                        {
                        if (! (i = encput(last, runCount, fp)))
                                return (0);
                        total += i;
                        runCount = 0;
                        }
                }
        else                /* No "run"  -  this != last */
                {
                if (runCount)
                        {
                        if (! (i = encput(last, runCount, fp)))
                                return(0);
                        total += i;
                        }
                last = this;
                runCount = 1;
                }
        }        /* endloop */
  if (runCount)        /* finish up */
        {
        if (! (i = encput(last, runCount, fp)))
                return (0);
        return (total + i);
        }
  return (total);
}



/*********************************************************
 * writePCX() - writes a pcx file given the proper header
 * info and the image data stored in the cuedMap[].shad
 * BYTE buffer. It writes nimg files named by filename
*/
void writePCX(char *filename,int nimg, int startImg)
{ /* writePCX */
  FILE *fp;
  int x,y,z,thisNimg,thisX;
  int count1,count2; /* These are just counters used in mapping the 
			image to PCX */
  BYTE bytecount,thisByte;
  char name[80];
  BYTE identifier=10;
  BYTE temp;
  short stemp;
  float min,max,scaleFactor,minF,maxF;
  float * q;
  BYTE *PCX_Buffer;
  float rmin,rmax,gmin,gmax,bmin,bmax,rsf,bsf,gsf;
  int ncol,nrow;

  if(!lowRes && !dispSections)
    {/* Account for hires images */
      ncol = largestDim;
      nrow = largestDim;
    }/* Account for hires images */
  else
    {/* lowres */
      ncol = numCols;
      nrow = numRows;

      /* Trap for assymetric images */
      if(ncol != nrow)
	{
	  if(ncol < nrow)
	    ncol= nrow;
	  else nrow = ncol;
	}
    }/* lowres */

  if(!dispSections)
    {/* Find global min/max for radially queued images */
      q = (float *)(map_image_buffer);
      minF=  maxF= *q;
      for (x=startImg;x<nimg;x++) 
	{ /* For all images */
	  q = (float *)(map_image_buffer+(x*ncol*nrow));
	  for (y=0;y<ncol*nrow;y++) 
	    { /* For each pixel */
	      if (*q < minF)
		minF = *q;
	      if (*q > maxF)
		maxF = *q;
	      q++;
	    } /* For each pixel */
	} /* For all images */
    }/* Find global min/max for radially queued images */

  if(rockIt)
    thisNimg = (nimg*2)-1;
  else thisNimg = nimg;

  for (x=startImg;x<thisNimg;x++) { /* For all images */
    sprintf(name,"%s_%03d.pcx",filename,x-startImg);
    fp = fopen(name,"w");
    if (!fp) {
      printf ("Error: Opening file %s\n",name);
      exit(0);
    }

    /* Write the header info */
    fwrite(&identifier,sizeof(BYTE),1,fp);
    temp = 5; /* Version 5 for PC Paintbrush for Win */
    fwrite(&temp,sizeof(BYTE),1,fp);
    temp=1;  /* Encoding = 1 for RLE encoded */
    fwrite(&temp,sizeof(BYTE),1,fp);
    temp=8;  /* Bits Per Pixel = 8 for 256 color */
    fwrite(&temp,sizeof(BYTE),1,fp);
    stemp=0;  /* Upper left corner is at 0,0 */
    fwrite(&stemp,sizeof(short),1,fp);
    fwrite(&stemp,sizeof(short),1,fp);

    stemp = ncol-1;
    fwrite(&stemp,sizeof(short),1,fp);
    stemp = nrow-1;
    fwrite(&stemp,sizeof(short),1,fp);
    stemp = ncol;
    fwrite(&stemp,sizeof(short),1,fp);
    stemp = nrow;
    fwrite(&stemp,sizeof(short),1,fp);

    temp=0;
    for (z=0;z<48;z++)
      fwrite(&temp,sizeof(BYTE),1,fp);
    fwrite(&temp,sizeof(BYTE),1,fp);
    temp=1; /* Number of bit planes = 1 */
    fwrite(&temp,sizeof(BYTE),1,fp);
    if (ncol%2==0)
      stemp = ncol;
    else
      stemp = ncol-1;
    fwrite(&stemp,sizeof(short),1,fp);
    stemp=1; /* PaletteType = 1 for color image */
    fwrite(&stemp,sizeof(short),1,fp);
    temp=0; /* Pad the rest with 0 */
    for (z=0;z<58;z++)
      fwrite(&temp,sizeof(BYTE),1,fp);


    if(x >= nimg)
      thisX = thisNimg - x - 1;
    else thisX = x;

  if (!dispSections)  {  /* If a 3D representation */

    /* Now lets do some stuff to prepare for the image data */
    /* Find the scaleFactor */
    min=cuedMap[thisX].shad[0];
    max=cuedMap[thisX].shad[0];
    for (y=0;y<ncol*nrow;y++) { /* For each pixel */
      if (cuedMap[thisX].shad[y]<min)
        min = cuedMap[thisX].shad[y];
      if (cuedMap[thisX].shad[y]>max)
        max = cuedMap[thisX].shad[y];
    } /* For each pixel */

    CK(PCX_Buffer = (BYTE *)malloc(sizeof(BYTE)*ncol*nrow));

    if(!radialMapShade)
      {/* Not radial */
        scaleFactor = (float)(tryColors-8)/(float)(max-min);
        for (y=0;y<ncol*nrow;y++)
          {
	    if(hiColor)
	      PCX_Buffer[y] = (BYTE)((cuedMap[thisX].shad[y]-min)*scaleFactor)+
	      8;
	    else
	      PCX_Buffer[y] = (BYTE)((cuedMap[thisX].shad[y]-min)*scaleFactor)+
		pixels_return[8];
	 }
      }/* Not radial */
    else
      {/* Can't scale radially cued */
        int thisPixel;
        scaleFactor = (float)(tryColors-8)/(float)(maxF-minF);
        for (y=0;y<ncol*nrow;y++)
          {/* Do all */
	    if(hiColor)
	      PCX_Buffer[y] = (BYTE)((cuedMap[thisX].shad[y]-minF)*
				     scaleFactor) + 
				       8 ;
	    else
	      PCX_Buffer[y] = (BYTE)((cuedMap[thisX].shad[y]-minF)*
				     scaleFactor) + 
		pixels_return[8] ;
          }/* Do all */
      }/* Can't scale radially cued */

  } /* If 3D map */
  else { /* If a section representation */
    CK(PCX_Buffer = (BYTE *)malloc(sizeof(BYTE)*(ncol)*
				(nrow)));
    stuffDisplayMapBuffer(thisX,1,0);
    for (count1=0,count2=0;count1<(nrow)*BOX_WIDTH;
		count1+=BOX_WIDTH,count2+=ncol)
      memcpy(PCX_Buffer+count2,editBox_display_buffer+count1,
	     (ncol));
  } /* If a section representation */
    for(y=0;y<nrow;y++) { /* Row */
      if (ncol%2==0)
        encLine(&PCX_Buffer[y*ncol],ncol,fp);
      else
        /* WHY ncol-1 ??? encLine(&PCX_Buffer[y*ncol],ncol-1,fp); */
        encLine(&PCX_Buffer[y*ncol],ncol-1,fp);
    } /* For each pixel */

    /* Write the palette */
    temp=12;  /* 12 precedes the palette */
    fwrite(&temp,sizeof(BYTE),1,fp);

    for (z=0;z<256;z++) { /* For all the palette */
      temp = (BYTE)((unsigned short)color_defs[z].red/256);
      fwrite(&temp,sizeof(BYTE),1,fp);
      temp = (BYTE)((unsigned short)color_defs[z].green/256);
      fwrite(&temp,sizeof(BYTE),1,fp);
      temp = (BYTE)((unsigned short)color_defs[z].blue/256);
      fwrite(&temp,sizeof(BYTE),1,fp);
    } /* For all the palette */

    if (PCX_Buffer)
      free(PCX_Buffer);
    fclose(fp);
  } /* For all images */

} /* writePCX */





/*********************************************************
 * writePCXbyte() - writes a pcx file given the proper header
 * info and the image data stored in the either the
 * packImg[] or inCore_buffer BYTE buffers. It writes nimg 
 * files named by filename
 * RWA Oct 9, 2002 - Need to rewrite this one so it can handle 
 * > 8 bit data for inCore_buffer and packImg[].
*/
void writePCXbyte(char *filename,int nimg)
{ /* writePCXbyte */
  FILE *fp;
  int x,y,z;
  int count1,count2; /* These are just counters used in mapping the 
			image to PCX */
  BYTE bytecount,thisByte;
  char name[80];
  BYTE identifier=10;
  BYTE temp;
  short stemp;
  float min,max,scaleFactor;
  BYTE *PCX_Buffer;
  float *thisBuf;
  float rmin,rmax,gmin,gmax,bmin,bmax,rsf,bsf,gsf;
  int ncol,nrow;

  ncol = purdueHeader.numCols;
  nrow = purdueHeader.numRows;

  sprintf(name,"%s",filename);
  fp = fopen(name,"w");
  if (!fp) {
    printf ("Error: Opening file %s\n",name);
    exit(0);
  }
  
  /* Write the header info */
  fwrite(&identifier,sizeof(BYTE),1,fp);
  temp = 5; /* Version 5 for PC Paintbrush for Win */
  fwrite(&temp,sizeof(BYTE),1,fp);
  temp=1;  /* Encoding = 1 for RLE encoded */
  fwrite(&temp,sizeof(BYTE),1,fp);
  temp=8;  /* Bits Per Pixel = 8 for 256 color */
  fwrite(&temp,sizeof(BYTE),1,fp);
  stemp=0;  /* Upper left corner is at 0,0 */
  fwrite(&stemp,sizeof(short),1,fp);
  fwrite(&stemp,sizeof(short),1,fp);
  
  stemp = ncol-1;
  fwrite(&stemp,sizeof(short),1,fp);
  stemp = nrow-1;
  fwrite(&stemp,sizeof(short),1,fp);
  stemp = ncol;
  fwrite(&stemp,sizeof(short),1,fp);
  stemp = nrow;
  fwrite(&stemp,sizeof(short),1,fp);
  
  temp=0;
  for (z=0;z<48;z++)
    fwrite(&temp,sizeof(BYTE),1,fp);
  fwrite(&temp,sizeof(BYTE),1,fp);
  temp=1; /* Number of bit planes = 1 */
  fwrite(&temp,sizeof(BYTE),1,fp);
  if (ncol%2==0)
    stemp = ncol;
  else
    stemp = ncol-1;
  fwrite(&stemp,sizeof(short),1,fp);
  stemp=1; /* PaletteType = 1 for color image */
  fwrite(&stemp,sizeof(short),1,fp);
  temp=0; /* Pad the rest with 0 */
  for (z=0;z<58;z++)
    fwrite(&temp,sizeof(BYTE),1,fp);
  
  if(fileType==PACKEDPORTABLE)
    thisBuf = packImg[(nimg-1)];
  else thisBuf = inCore_buffer;
  
  /* Now lets do some stuff to prepare for the image data */
  /* Find the scaleFactor */
  x = imgNum - 1;
  min=thisBuf[0];
  max=thisBuf[0];
  for (y=0;y<ncol*nrow;y++) { /* For each pixel */
    if (thisBuf[y]<min)
      min = thisBuf[y];
    if (thisBuf[y]>max)
      max = thisBuf[y];
  } /* For each pixel */
  
  CK(PCX_Buffer = (BYTE *)malloc(sizeof(BYTE)*ncol*nrow));
  for (y=0;y<ncol*nrow;y++)
    PCX_Buffer[y] = (BYTE)(thisBuf[y]);
  
  for(y=0;y<nrow;y++) { /* Row */
    if (ncol%2==0)
      encLine(&PCX_Buffer[y*ncol],ncol,fp);
    else
      encLine(&PCX_Buffer[y*ncol],ncol-1,fp);
  } /* For each pixel */
  
  /* Write the palette */
  temp=12;  /* 12 precedes the palette */
  fwrite(&temp,sizeof(BYTE),1,fp);
  
  for (z=0;z<256;z++) { /* For all the palette */
    temp = (BYTE)((unsigned short)color_defs[z].red/256);
    fwrite(&temp,sizeof(BYTE),1,fp);
    temp = (BYTE)((unsigned short)color_defs[z].green/256);
    fwrite(&temp,sizeof(BYTE),1,fp);
    temp = (BYTE)((unsigned short)color_defs[z].blue/256);
    fwrite(&temp,sizeof(BYTE),1,fp);
  } /* For all the palette */

  if (PCX_Buffer)
    free(PCX_Buffer);
  fclose(fp);

  sprintf(outStr,"PCX file [%s] sucessfully written",name);
  askThem(outStr,FALSE,XmDIALOG_INFORMATION);

} /* writePCXbyte */





/************************************************************
 * read_pif() - Read image(s) in pif format.
 & Future goal - get all input to robem read here.
*/
void read_pif(int *ncol, int *nrow, int *nsec,
	      fString_t *inFileName)
{/* READ_PIF */
  int i,j;
  fString_t rWrite;
  int htype,imgOffset,convTsb;
  char versionStr[256], descriptionStr[100];
  XmString simpleString;
  fString_t descStr;
  float alp, bet, gam;
  int binCol,binRow,binSec;
  int strtCol,strtRow,strtSec;

  i= 4;
  pifHeader.luNum = i;
  rWrite.addr = malloc(sizeof(char)*4);
  rWrite.length = 1;
  descStr.addr = malloc(sizeof(char)*64);
  descStr.length = 32;
  sprintf(rWrite.addr,"r");

  pif_open_(inFileName->addr, &i, rWrite.addr);

  pif_read_gh_(&pifHeader);
  nimg                  = pifHeader.nimg;
  recipFloatScaleFactor = pifHeader.recipFloatScaleFactor;
  htype                 = pifHeader.htype;
  numCols               = pifHeader.numCols;
  numRows               = pifHeader.numRows;
  numSecs               = pifHeader.numSecs;
  pifMode               = pifHeader.mode;
  fileMrcEven           = pifHeader.even;
  fileMrcX              = pifHeader.mrcX;
  fileMrcY              = pifHeader.mrcY;
  fileMrcZ              = pifHeader.mrcZ;


  if(pifMode!=20 && pifMode!=40 && pifMode!=21 && 
     pifMode!=22 && pifMode != 46)
    {/* NOt a map!! */
      *nsec = -99;
      goto noRead;
    }/* NOt a map!! */

  if(pifMode==20 || pifMode==40 || pifMode==21 || pifMode==22 || pifMode==46)
    {/* MAP file */
      int thisImg;
      j= 0;
      thisImg = j + 1;  /* For fortran */
      
      pifHeader.imgNum = thisImg;
      pif_read_dh_(&pifHeader);
      
      recipFloatScaleFactor = pifHeader.recipFloatScaleFactor;
      purdueHeader.img_bkg  = pifHeader.img_bkg;
      purdueHeader.img_min  = (int)pifHeader.datai4_min;
      purdueHeader.img_max  = (int)pifHeader.datai4_max;
      purdueHeader.map_min  = pifHeader.datar4_min;
      purdueHeader.map_max  = pifHeader.datar4_max;
      purdueHeader.img_avg  = pifHeader.datar4_avg;
      purdueHeader.img_std  = pifHeader.datar4_std;
      numCols    = pifHeader.numCols;
      numRows    = pifHeader.numRows;
      numSecs    = pifHeader.numSecs;
      imgOffset  = pifHeader.imgOffset; 
      pifMode    = pifHeader.mode; 
      htype      = pifHeader.htype; 
      sprintf(title,"%s",pifHeader.descriptionStr);
      aoverb     = pifHeader.aoverb; 
      map_abang  = pifHeader.map_abang;
      a_cell     = pifHeader.a_cell;
      b_cell     = pifHeader.b_cell;
      c_cell     = pifHeader.c_cell; 
      alp        = pifHeader.alp; 
      bet        = pifHeader.bet; 
      gam        = pifHeader.gam; 
      dela       = pifHeader.dela; 
      delb       = pifHeader.delb; 
      delc       = pifHeader.delc;
      binFactor  = pifHeader.binFactor;
      dthe       = pifHeader.dthe; 
      dphi_90    = pifHeader.dphi_90; 
      symmetry   = pifHeader.symmetry; 
      
      mrc_alpha = (int) alp;
      mrc_beta = (int) bet;
      mrc_gamma = (int) gam;
      fileBinSecs= binSec  = numSecs;
      fileBinRows= binRow  = numRows;
      fileBinCols= binCol  = numCols;
      fileBinFactor = binFactor;


      if(pifMode==22)
	{/* SPecial case for PFTS data - nrot*4 */
	  binRow  = numRows/binFactor;
	  /* We store it nrot/4 now ... */
	  binCol  = numCols/binFactor;
	}/* SPecial case for PFTS data - nrot*4 */
      else
	{/* Normal */
	  if(binFactor==2)
	    {/* 1/2 */
	      if(pifMode==20)
		binSec  = numSecs/2+1;  
	      binRow  = numRows/2+1;
	      binCol  = numCols/2+1;
	      if((binSec%2)==0)
		binSec--;
	      if((binRow%2)==0)
		binRow--;
	      if((binCol%2)==0)
		binCol--;
	    }/* 1/2 */
	}/* Normal */
      
      if((float_image_buffer1) &&
	 (float_image_buffer1 != lowRes_buffer1))
	free(float_image_buffer1);
      if(lowRes_buffer1)
	free(lowRes_buffer1);

      float_image_buffer1= NULL;
      lowRes_buffer1 = NULL;
      
      if(pifMode == 46)
	binSec = pifHeader.nimg;

      CK(float_image_buffer1 = (float *)malloc(sizeof(float)*
					       binCol*binRow*binSec));
      
      
      maxMapPixel= -99999999.9;
      minMapPixel= 99999999.9;
      maxMapPixel= -HUGE; /* dgm -HUGE instead of -99999.9 */
      minMapPixel= HUGE;  /* dgm HUGE instead of 9999999.9 */
      convTsb = TRUE;
      pif_tsbworld(convTsb);
      
      for(j=0;j<nimg;j++)
	{/* Read all iamges */
	  if(pifMode==20)
	    {/* read it */
	      binFactor = 1;
	      strtCol = strtRow = strtSec = 1;
	      pif_read_map_short_image_(&i, float_image_buffer1,
					&imgOffset,
					&binCol, &binRow, &binSec,
					&recipFloatScaleFactor,
					&minMapPixel,&maxMapPixel, &binFactor,
					&strtCol, &strtRow, &strtSec,
					&binCol, &binRow, &binSec);
	    }/* read it */
	  else if(pifMode==40)
	    {/* read it */
	      int l;
	      binFactor = 1;
	      strtCol = strtRow = strtSec = 1;
	      pif_read_wmapi4_(&pifHeader,
			       float_image_buffer1,
			       &binCol, &binRow, &binSec);
	      pifMode = 20;  /* Kludge for now YJI */
	      minMapPixel = maxMapPixel = float_image_buffer1[0];
	      for(l=0;l<(binCol*binRow*binSec);l++)
		{/* find */
		  maxMapPixel = MAX(maxMapPixel,float_image_buffer1[l]);
		  minMapPixel = MIN(minMapPixel,float_image_buffer1[l]);
		}/* find */
	    }/* read it */
	  else if (pifMode==21 || pifMode == 22)
	    {/* Read from disk */
	      binFactor = 1;
	      pif_read_map_int_image_(&i, float_image_buffer1,
				      &imgOffset,
				      &binCol, &binRow, &binSec,
				      &recipFloatScaleFactor,
				      &minMapPixel,&maxMapPixel, &binFactor, 
				      &pifMode);
	    }/* Read from disk */
	  else if (pifMode == 46)
	    {/* Read from disk */
	      int ll;
	      int localBinSec;  /* We've only one section per img */
	      thisImg = j + 1;  /* For fortran */
	      pifHeader.imgNum = thisImg;
	      pif_read_dh_(&pifHeader);
	      imgOffset  = 
		pifHeader.imgOffset= 512 +
		(512 * (j+1)) +
		(j * binCol * binRow *4);
	      imgOffset  = pifHeader.imgOffset; 
	      binFactor = 1;
	      localBinSec = 1;
	      pif_read_map_int_image_(&i, 
				      (float_image_buffer1+(j*binCol*binRow)),
				      &imgOffset,
				      &binCol, &binRow, &localBinSec,
				      &recipFloatScaleFactor,
				      &minMapPixel,&maxMapPixel, &binFactor, 
				      &pifMode);

	    }/* Read from disk */
	  else
	    {/* Error */
	      askThem("Bad mode for MAP.",FALSE,XmDIALOG_INFORMATION);
	    }/* Error */
	  
	}/* Read all iamges */
    }/* MAP file */
  
  pif_close_(&i);
  
  numCols = binCol;
  numRows = binRow;
  numSecs = binSec;
  origNumSecs= numSecs;

  if(pifMode == 20 && minMapPixel > -1.0 && maxMapPixel < 1.0)
    {/* voxels have very small values */
      float  scaler;
      int done,i;

      done = FALSE;
      scaler = 1.0;
      while(!done)
	{/* find scaler */
	  if((scaler * minMapPixel < -10.0) &&
	     (scaler * maxMapPixel >  10.0))
	    done = TRUE;
	  else scaler *= 10.0;
	  if(scaler > 100000000.0)
	    done = TRUE;
	}/* find scaler */

      printf("Multiplying voxel densities by [%f].\n" ,scaler);

      for(i=0;i<(numCols*numRows*numSecs);i++)
	float_image_buffer1[i] *= scaler;
 
      minMapPixel *= scaler;
      maxMapPixel *= scaler;

    }/* voxels have very small values */


  if(pifMode == 46)
    {/* Let's try converting mode==46 to PACKEDPORTABLE */
      int i,j,k,maxI,minI;
      float *thisPackImg;

      fileType = PACKEDPORTABLE;

      printf("MAX[%f]MIN[%f]   \n",
	     maxMapPixel,minMapPixel);

      maxI = minI = (float_image_buffer1[0]);
      for(k=0;k<binSec;k++)
	{/* DO all imgs */
	  if(packImg[k]!=NULL)
	    {free(packImg[k]);
	    packImg[k]= NULL;}
	  CK(packImg[k]= (float *)malloc(
					 sizeof(float) *
					 binCol*
					 binRow));
	  thisPackImg = packImg[k];
	  for(i=0;i<binRow;i++)
	    for(j=0;j<binCol;j++)
	      {
		thisPackImg[((i*binCol)+j)] = 
		  float_image_buffer1[(k*binCol*binRow)+((i*binCol)+j)] ;
		minI = MIN(minI,thisPackImg[((i*binCol)+j)]);
		maxI = MAX(maxI,thisPackImg[((i*binCol)+j)]);
	      }
	  
	}/* DO all imgs */

      minPackPixel = minI;
      maxPackPixel = maxI;

      if(minMaxColor)
	{/* Use option screen defaults */
	  char *tourStr;
	  printf("Using option defaults \n");
	  tourStr= XmTextGetString(minMapTextW);
	  minPackPixel = (float)strtod(tourStr,0);
	  XtFree(tourStr);
	  tourStr= XmTextGetString(maxMapTextW);
	  maxPackPixel = (float)strtod(tourStr,0);
	  XtFree(tourStr);
	}/* Use option screen defaults */

      /* Now clip this guy to our colormap */ 
      if(minPackPixel < 0.0)
	scalePackFactor = (((float)tryColors - 7.0) / 
			   (maxPackPixel-minPackPixel));
      else scalePackFactor = (((float)tryColors - 7.0) / 
			      maxPackPixel);
      printf("minPackPixel [%f]  max[%f]     scale[%f]\n",
	     minPackPixel,
	     maxPackPixel,
	     scalePackFactor);
      
      purdueHeader.numCols = binCol;
      purdueHeader.numRows = binRow;
      purdueHeader.nimg    = binSec;
      packRadius = purdueHeader.numCols/2;
      stuffPackedImage(purdueHeader.numCols,
		       purdueHeader.numRows,
		       packImg,purdueHeader.nimg,
		       1);  /* Start at zero'th(1) image */
      
      
    }/* Let's try converting mode==46 to PACKEDPORTABLE */

  /* Global default */
  sprintf(outStr,"1");
  XmTextSetString(imgNumTextWidget,outStr);
  
  if(pifMode==20)
    {/* We've got a map here */

      imgNum = numCols/2; /* Let's default to the middle section */
      sprintf(outStr,"%d",imgNum+1);
      purdueHeader.nimg = numSecs;  /* Fool it to take our value */
      XmTextSetString(imgNumTextWidget,outStr);
      
      calcSF_work(calcSFs);

      /* Now remember these forever */
      lowRes_buffer1 = float_image_buffer1;
      lowResCols = numCols;
      lowResRows = numRows;
      lowResSecs = numSecs;

      origWidthFromFile = numCols;

    }/* We've got a map here */

  purdueHeader.numCols = *ncol = numCols;
  purdueHeader.numRows = *nrow = numRows;
  purdueHeader.numSecs = *nsec = origNumSecs;

  xCen = (double)(numCols/2);
  yCen = (double)(numRows/2);
  zCen = (double)(origNumSecs/2);

  largestDim = numCols;
  largestDim = MAX(largestDim,numRows);
  largestDim = MAX(largestDim,origNumSecs);

  /* We've got to make this larger for asymmetric maps 
   * and all maps that have information "outside" of the
   * normal sphere of our normal maps,
   * or else we just keep cliping the box into a sphere */
  if(numCols==largestDim && numRows==largestDim
     && (origNumSecs)==largestDim)
    {/* Symmetric */
      aSymmetric = FALSE;
      origGlobMaxRadius= globMaxRadius = (double)((double)origNumSecs/2)+5.0;
      globMinRadius = (double)0.0;
    }/* Symmetric */
  else
    {/* asymmetric */
      aSymmetric = TRUE;
      origGlobMaxRadius=  globMaxRadius = (double)((double)origNumSecs);
      globMinRadius = (double)0.0;
    }/* asymmetric */
  
  free(rWrite.addr);

  /* Set scrolls on depthcue screen */
  /* Give feedback */
  sprintf(outStr,"%.0f",globMinRadius);
  createSimpleString(&simpleString,outStr);
  XtSetArg(arglst[0], XmNvalue, outStr );
  XtSetArg(arglst[1], XmNrecomputeSize, FALSE );
  XtSetValues(minRadiusValueWidget, arglst, 2); 
  XFlush(dsply);
  XmStringFree(simpleString); 
  
  sprintf(outStr,"%.0f",globMaxRadius);
  createSimpleString(&simpleString,outStr);
  XtSetArg(arglst[0], XmNvalue, outStr );
  XtSetArg(arglst[1], XmNrecomputeSize, FALSE );
  XtSetValues(maxRadiusValueWidget, arglst, 2); 
  XFlush(dsply);
  XmStringFree(simpleString); 
  
  
  /* Radius is origNumSecs/2 */
  XtVaSetValues(maxRadiusScrollWidget,XmNmaximum, 
		(int)((globMaxRadius*2)+5),NULL);
  XtVaSetValues(maxRadiusScrollWidget,XmNsliderSize, (int)(5),NULL);
  XtVaSetValues(maxRadiusScrollWidget,XmNvalue, (int)(globMaxRadius),NULL);
		
  XtVaSetValues(minRadiusScrollWidget,XmNmaximum, 
		(int)((globMaxRadius*2)+5),NULL);
  XtVaSetValues(minRadiusScrollWidget,XmNsliderSize, (int)(5),NULL);
  XtVaSetValues(minRadiusScrollWidget,XmNvalue, (int)(globMinRadius),NULL);
  
  printf("MAX[%f]  MIN[%f]\n",maxMapPixel,minMapPixel);
  secMaxPixel = maxMapPixel;
  secMinPixel = minMapPixel;

  minSurf = (int)MAX(minMapPixel,0.0);
  maxSurf = MAX(maxMapPixel,(0.0-minMapPixel));
  /* We're doing the surfaceDensity in tenths now */
  if (pifMode == 46)
    {
      XtVaSetValues(surfDensScrollWidget,XmNmaximum, 
		    (int)((maxSurf*10.0)+50.0),NULL);
      XtVaSetValues(surfDensScrollWidget,XmNminimum, (int)(minSurf*10.0),NULL);
      XtVaSetValues(surfDensScrollWidget,XmNsliderSize, (int)(500),NULL);
      XtVaSetValues(surfDensScrollWidget,XmNpageIncrement, (int)500,NULL);
      XtVaSetValues(surfDensScrollWidget,XmNincrement, (int)100,	NULL);
      XtVaSetValues(surfDensScrollWidget,XmNvalue, 
		    (int)(purdueHeader.img_bkg),NULL);
		  
    }
  else if(maxSurf > 5000.0)
    {
      XtVaSetValues(surfDensScrollWidget,XmNmaximum, 
		    (int)((maxSurf*10.0)+50.0),NULL);
      XtVaSetValues(surfDensScrollWidget,XmNminimum, (int)(minSurf*10.0),NULL);
      XtVaSetValues(surfDensScrollWidget,XmNsliderSize, 
		    (int)(maxSurf*0.01),NULL);
      XtVaSetValues(surfDensScrollWidget,XmNpageIncrement, 
		    (int)(maxSurf*0.01),NULL);
      XtVaSetValues(surfDensScrollWidget,XmNincrement, 
		    (int)(maxSurf*0.002),NULL);
      XtVaSetValues(surfDensScrollWidget,XmNvalue, (int)(10),NULL);
		  
    }
  else if(maxSurf > 500.0)
    {
    XtVaSetValues(surfDensScrollWidget,XmNmaximum, 
		  (int)((maxSurf*10.0)+50.0),NULL);
    XtVaSetValues(surfDensScrollWidget,XmNminimum, (int)(minSurf*10.0),NULL);
    XtVaSetValues(surfDensScrollWidget,XmNsliderSize, (int)(500),NULL);
    XtVaSetValues(surfDensScrollWidget,XmNpageIncrement, (int)500,NULL);
    XtVaSetValues(surfDensScrollWidget,XmNincrement, (int)100,	NULL);
    XtVaSetValues(surfDensScrollWidget,XmNvalue, (int)(10),NULL);
		  
    }
  else
    {
      XtVaSetValues(surfDensScrollWidget,XmNmaximum, 
		    (int)((maxSurf*10.0)+50.0),NULL);
      XtVaSetValues(surfDensScrollWidget,XmNminimum, (int)(minSurf*10.0),NULL);
      XtVaSetValues(surfDensScrollWidget,XmNsliderSize, (int)(50),NULL);
      XtVaSetValues(surfDensScrollWidget,XmNpageIncrement, (int)50,NULL);
      XtVaSetValues(surfDensScrollWidget,XmNincrement, (int)10,	NULL);
      XtVaSetValues(surfDensScrollWidget,XmNvalue, (int)(10),NULL);
		  
    }
  surfaceDensity= 1.0;  /* 1 tenth of 10 */
  
  /* Give feedback */
  sprintf(outStr,"%d",-1);
  createSimpleString(&simpleString,outStr);
  XtSetArg(arglst[0], XmNvalue, outStr );
  XtSetArg(arglst[1], XmNrecomputeSize, FALSE );
  XtSetValues(surfDensValueWidget, arglst, 2); 
  XFlush(dsply);
  XmStringFree(simpleString); 

  createSimpleString(&simpleString,"SecNum");
  XtSetArg(arglst[0], XmNlabelString, simpleString);
  XtSetArg(arglst[1], XmNrecomputeSize, FALSE );
  XtSetValues(imgNumButtonW, arglst, 2); 
  XFlush(dsply);
  XmStringFree(simpleString); 
  

  /* To make this work for now let's stuff the nimg field with the
   * number of sections.  Really we should change any and all references
   * to nimg to be origNumSecs. We really only work on one image at a time
   * so this may never have any affect.
   */
  purdueHeader.nimg = origNumSecs;

  /* Write stats */
  if (pifMode != 46)
    calcMapStats(float_image_buffer1,
		 &purdueHeader.img_bkg, 
		 &purdueHeader.map_min,
		 &purdueHeader.map_max, 
		 &purdueHeader.img_avg, 
		 &purdueHeader.img_std,
		 &numCols, &numRows, &origNumSecs);

  /* Write stats */
  sprintf(outStr,"\nStatistics for file [%s]\n\
\n\
Title [%s]\n\
\n\
File size   [%d]x[%d]x[%d]\n\
Background  [%f]\n\
MIN,MAX     [%f],[%f]\n\
AVG,STD     [%f],[%f]\n\
\n",
	  inFileName->addr,
	  title,
	  numCols,	  numRows, origNumSecs,
	  purdueHeader.img_bkg, 
	  minMapPixel, maxMapPixel,
	  purdueHeader.img_avg,	  purdueHeader.img_std
	  );
  XmTextSetString(statTextWidget,outStr);
  XFlush(dsply);

 noRead:;


}/* READ_PIF */





/*********************************************************
 * tsbworld_float() - Turn this float buffer upside down 
 * to conform with the way Tim stores them on disk.
 * Tim's old  Fortran code stored them in reverse row
 * order.
*/
void tsbworld_float(float *in_buffer,
		    float *out_buffer,
		    int thisNcols, int thisNrows, int thisNsecs)
{/* TSBWORLD_FLOAT */
  int   rowCol,tsbRow,yOut,yOut2,i,j,k;

  rowCol = thisNrows * thisNcols;
  for(i=0;i<thisNsecs;i++)
    {/* DO all secs */
      for(j=0;j<thisNrows;j++)
	{/* Do rows in TSB order */
	  /* Do rows in TSB order */
	  tsbRow = thisNrows - j - 1;
	  yOut = (i*thisNcols*thisNrows)+(tsbRow*thisNcols);
	  yOut2= (i*thisNcols*thisNrows)+(j*thisNcols);
	  for(k=0;k<thisNcols;k++)
	    {/* Write it a short at a time */
	      out_buffer[yOut2+k]= (float)in_buffer[yOut + k];
	    }/* Write it a short at a time */
	}/* Do rows in TSB order */
    }/* DO all secs */
  
}/* TSBWORLD_FLOAT */





/*********************************************************
 * tsbworld_float_box() - Turn this float buffer upside down 
 * to conform with the way Tim stores them on disk.
 * Tim's old  Fortran code stored them in reverse row
 * order.
*/
void tsbworld_float_box(float *in_buffer,
			float *out_buffer,
			int boxWidth, int ncol, int nrow)
{/* TSBWORLD_FLOAT_BOX */
  int i,j;
  float tB1,tB2;

  for(i=0;i<=(nrow/2);i++)
    {/* DO all rows */
      for(j=0;j<ncol;j++)
	{/* All cols */
	  tB1 = in_buffer[(i*boxWidth)+j];
	  tB2 = in_buffer[((nrow-i-1)*boxWidth)+j];
	  out_buffer[(i*ncol)+j] = tB2;
	  out_buffer[((nrow-i-1)*ncol)+j] = tB1;
	}/* All cols */
    }/* DO all rows */
  
}/* TSBWORLD_FLOAT_BOX */





/*********************************************************
 * tsbworld_int() - Turn this int buffer upside down 
 * to conform with the way Tim stores them on disk.
 * Tim's old  Fortran code stored them in reverse row
 * order.
*/
void tsbworld_int(int *in_buffer,
		   int *out_buffer,
		   int box_width, int ncol, int nrow)
{/* TSBWORLD_INT */
  int i,j;
  int tB1,tB2;

  for(i=0;i<=(nrow/2);i++)
    {/* DO all rows */
      for(j=0;j<ncol;j++)
	{/* All cols */
	  tB1 = in_buffer[(i*box_width)+j];
	  tB2 = in_buffer[((nrow-i-1)*box_width)+j];
	  out_buffer[(i*ncol)+j] = tB2;
	  out_buffer[((nrow-i-1)*ncol)+j] = tB1;
	}/* All cols */
    }/* DO all rows */

}/* TSBWORLD_INT */





/*********************************************************
 * tsbworld_byte() - Turn this byte buffer upside down 
 * to conform with the way Tim stores them on disk.
 * Tim's old  Fortran code stored them in reverse row
 * order.
*/
void tsbworld_byte(BYTE *in_buffer,
		   BYTE *out_buffer,
		   int box_width, int ncol, int nrow)
{/* TSBWORLD_BYTE */
  int i,j;
  BYTE tB1,tB2;

  for(i=0;i<=(nrow/2);i++)
    {/* DO all rows */
      for(j=0;j<ncol;j++)
	{/* All cols */
	  tB1 = in_buffer[(i*box_width)+j];
	  tB2 = in_buffer[((nrow-i-1)*box_width)+j];
	  out_buffer[(i*ncol)+j] = tB2;
	  out_buffer[((nrow-i-1)*ncol)+j] = tB1;
	}/* All cols */
    }/* DO all rows */

}/* TSBWORLD_BYTE */




/****************************************************************
 * checkRemoveBoxes() - Remove any boxes that require pixels
 * outside the dimensions of the scanned file.
 *
 */
void checkRemoveBoxes()
{/* CHECKREMOVEBOXES */
  int beginX,endX,beginY,endY;
  int i,j,k,numRemoved;

  numRemoved = i = 0;
  while(i<numBoxes)
    {/* Do all boxes */
      beginY = box[i].y - featherRadius;
      beginX = box[i].x - featherRadius;
      endY   = box[i].y + featherRadius;
      endX   = box[i].x + featherRadius;
      if(beginY <= 0 ||
	 beginX <= 0 ||
	 endX >= file_width ||
	 endY >= file_height)
	{/* Box is outside of file */
	  removeBox(box[i].x, box[i].y, TRUE);
	  sprintf(outStr,"Removing box [%d] - outside file limits.\n",
		  i+1+numRemoved);
	  askThem(outStr,TRUE,XmDIALOG_INFORMATION);
	  numRemoved++;
	}/* Box is outside of file */
      else
	{/* OK to look at next slot */
	  i++;
	}/* OK to look at next slot */

    }/* Do all boxes */

}/* CHECKREMOVEBOXES */




/************************************************************
 * write_pif() - Write image(s) in pif format.  All PIF output
 * from robem should be written here.
*/
void write_pif(int ncol, int nrow, char *outFileName, int thisFileType)
{/* WRITE_PIF */
  int file_id,htype,mode,i,j,k,m,n,o,p,firstImg;
  int writeIt,noRectYet, afterDeletingNumImages,numImages;
  float *out_image_buffer;
  char junkChar[1024];
  char  floatScaleStr[17];
  char versionStr[33];
  char descriptionStr[285];
  char timeStr[36];
  char micrographStr[200];
  int  saveNimg,appendAndCurrentImgs;
  char scanNumberStr[210];
  FILE *fOut;
  int aoverbI,map_abangI,delaI,delbI,delcI,dtheI,dphi_90I;
  fString_t fName,rWrite;
  int imgOffset, thisNcol, thisNrow, thisNimages;
  forPifHead_t thisHead;

  if(thisFileType==CIRCLEPACKED)
    {/* Check to see if we've got outliers */
      checkRemoveBoxes();
    }/* Check to see if we've got outliers */

  firstImg = 0;
  appendAndCurrentImgs = purdueHeader.nimg;
  if(pifMode == 20)
    appendAndCurrentImgs = 1;
  else if(pifMode == 46)
    appendAndCurrentImgs = purdueHeader.nimg;
  else if(thisFileType==CIRCLEPACKED)
    appendAndCurrentImgs = numBoxes;

  thisHead.luNum = 4;
  rWrite.addr = malloc(sizeof(char)*4);
  rWrite.length = 4;
  fName.addr = malloc(sizeof(char)*264);
  fName.length = 264;
  sprintf(rWrite.addr,"r+");
  sprintf(fName.addr,outFileName);

  noRectYet = TRUE;
  writeIt = TRUE;
  thisNimages = 0;

  if(thisFileType==PACKEDPORTABLE)
    {/* Find out how many we've got to write */
      afterDeletingNumImages = purdueHeader.nimg;
      for(i=0;i<purdueHeader.nimg;i++)
	if(deletePackedImage[i])
	  afterDeletingNumImages--;
    }/* Find out how many we've got to write */

  if(thisFileType==PACKEDPORTABLE || thisFileType==CIRCLEPACKED ||
     thisFileType==RECTPACKED || thisFileType==MAP || 
     thisFileType==EMPORTABLE)
    {/* PACKEDPORTABLE/emportable or MAP */

      if((writeType==APPENDCIRCLEPACKED) ||
	 (writeType==APPENDRECT))
	{/* Open for append */

	  pif_open_(fName.addr, &(thisHead.luNum), rWrite.addr);
	  
	  pif_read_gh_(&thisHead);
	  thisHead.imgNum = 1;
	  pif_read_dh_(&thisHead);
	  
	  thisNcol    = thisHead.numCols;
	  thisNrow    = thisHead.numRows;
	  thisNimages = thisHead.nimg;
	  appendAndCurrentImgs += thisNimages;
	  mode        = thisHead.mode;
	  firstImg    = thisHead.nimg;
	  
	  /* Check for inconsistencies */
	  if((thisNcol != ncol) ||
	     (thisNrow != nrow))
	    {/* Error */
	      sprintf(outStr,
		      "Image sizes differ [%d][%d] in [%s] [%d][%d] other,",
		      numCols, numRows,
		      outFileName,thisNcol,thisNrow);
	      askThem(outStr,TRUE,XmDIALOG_INFORMATION);
	      writeIt= FALSE;
	    }/* Error */
	  
	  
	  /* Now update the header to show new data at the end */
	  if(thisFileType==CIRCLEPACKED)
	    {/* Boxes */
	      if(boxMethod==USECIRCLE)
		purdueHeader.nimg = numBoxes;
	      else if (boxMethod==USERECT || boxMethod==USESQUARE)
		purdueHeader.nimg = 1;
	    }/* Boxes */
	  else if(thisFileType==RECTPACKED)
	    {/* Boxes */
	      purdueHeader.nimg = 1;
	    }/* Boxes */
	  else if(thisFileType != PACKEDPORTABLE)
	    purdueHeader.nimg = 1;
	  
	  if(fileType==PACKEDPORTABLE)
	    numImages = thisNimages + afterDeletingNumImages;
	  else numImages = thisNimages + purdueHeader.nimg;

	  thisHead.nimg    = numImages;
	  /* thisHead.numSecs = numImages; */

	  if(mode != 46)
	    {/* Can't do */
	      printf("Can't append mode != 46 files!!\n");
	      exit(0);
	    }/* Can't do */

	  if((mode==46) &&
	     (float_image_buffer1 == NULL))
	    float_image_buffer1 = (float *)malloc(
						  sizeof(float)*
						  appendAndCurrentImgs*
						  thisHead.numCols*
						  thisHead.numRows
						  );

	  for(j=0;j<firstImg;j++)
	    {/* Read all */
	      int binFactor,binCol,binRow,imgOffset,thisImg,pifMode;
	      int localBinSec;  /* We've only one section per img */
	      recipFloatScaleFactor =  thisHead.recipFloatScaleFactor;
	      thisImg = j + 1;  /* For fortran */
	      thisHead.imgNum = thisImg;
	      binCol = thisHead.numCols;
	      binRow = thisHead.numRows;
	      pifMode = 46;
	      pif_read_dh_(&thisHead);
	      imgOffset  = 
		thisHead.imgOffset= 512 +
		(512 * (j+1)) +
		(j * binCol * binRow *4);
	      imgOffset  = thisHead.imgOffset; 
	      binFactor = 1;
	      localBinSec = 1;
	      pif_read_map_int_image_(&thisHead.luNum, 
				      (float_image_buffer1+(j*binCol*binRow)),
				      &imgOffset,
				      &binCol, &binRow, &localBinSec,
				      &recipFloatScaleFactor,
				      &minMapPixel,&maxMapPixel, &binFactor, 
				      &pifMode);
	    }/* Read all */

	  writeType = CIRCLEPACKED; /* Just choose somethign else so
				     * these data can be written as normal */
	  /* OLD way	  pif_write_gh(&thisHead); */
	  pif_close(&(thisHead.luNum));

	}/* Open for append */
      else if(!normLinSwitch && !autoBoxSwitch && !reBoxSwitch && 
	      !boxitSwitch && !doFix && !doInterp && !doNoise && !doTrunc)  
	                       /* We always have the right filename for 
				* normLinSwitch boxitSwitch doFix and 
				* autoBoxSwitch and doInterp */
	{/* Check to make sure we're not overwriting it */
	  fOut= NULL;
	  fOut = fopen(outFileName,"r");
	  if(fOut != NULL)
	    {/*  Ooops we may be overwriting it!!! */
	      fclose(fOut);
	      sprintf(outStr,
		      "Do you really want to overwrite [%s]?",outFileName);
	      writeIt= askThem(outStr,TRUE,XmDIALOG_QUESTION);
	    }/*  Ooops we may be overwriting it!!! */
	  fOut= NULL;
	}/* Check to make sure we're not overwriting it */
      
      if(writeIt)
	{/* OK to proceed */
	  
	  if((writeType==APPENDCIRCLEPACKED)  ||
	     (writeType==APPENDRECT))
	    {
	      sprintf(rWrite.addr,"r+");     /* Append it */
	    }
	  else
	    {
	      sprintf(rWrite.addr,"w");     /* Write it */
	      pif_init_head(&thisHead);
	      thisHead.luNum = 4;
	    }

	  pif_open_(fName.addr, &(thisHead.luNum), rWrite.addr);
	  
	  if((writeType==APPENDCIRCLEPACKED) ||
	     (writeType==APPENDRECT))
	    {/* Check to make sure we've got the same image dimensions */
	      /* Already done above */
	    }/* Check to make sure we've got the same image dimensions */
	  else
	    {/* Write new header */
	      

	      /*******************************************
	       * 
	       *  WRIITING NEW GLOBAL PIF HEADER
	       *
	       *******************************************/
	      /* [1] Write global file header */
	      
	      /* Pick appropriate statistics calculation based
	       * on data type we have */
	      if(thisFileType==CIRCLEPACKED)
		{/* packed */
		  if(boxMethod==USECIRCLE)
		    {/* Circles */
		      stuffDisplayBoxBuffer(0, image_fileName.addr, 1);
		      calcCircleStats(editBox_mirror_buffer,
				      &purdueHeader.img_bkg, 
				      &purdueHeader.img_min,
				      &purdueHeader.img_max, 
				      &purdueHeader.img_avg, 
				      &purdueHeader.img_std,
				      ncol, nrow, BOX_WIDTH);

		    }/* Circles */
		  else if(boxMethod==USERECT || boxMethod==USESQUARE)
		    {/* Rects */
		      CK(out_image_buffer = (float *)malloc(
						     sizeof(float)*
								 (ncol*nrow)));
		      stuffRectangleBuffer(out_image_buffer, 
					   image_fileName.addr);
		      
		      calcStats(out_image_buffer,
				&purdueHeader.img_bkg, 
				&purdueHeader.img_min,
				&purdueHeader.img_max, 
				&purdueHeader.img_avg, 
				&purdueHeader.img_std,
				ncol, nrow, ncol);
		      noRectYet = FALSE;
		    }/* Rects */
		}/* packed */
	      else if(thisFileType==RECTPACKED)
		{/* packed */
		  int saveFileType,i,j;
		  CK(out_image_buffer = (float *)malloc(
						      sizeof(float)*
								(ncol*nrow)));
		  stuffRectangleBuffer(out_image_buffer, 
				       image_fileName.addr);

		  saveFileType = fileType;
		  fileType = RECTPACKED;
		  calcStats(out_image_buffer,
			    &purdueHeader.img_bkg, 
			    &purdueHeader.img_min,
			    &purdueHeader.img_max, 
			    &purdueHeader.img_avg, 
			    &purdueHeader.img_std,
			    ncol, nrow, ncol);
		  fileType = saveFileType;
		  noRectYet = FALSE;
		}/* packed */
	      else if(thisFileType==PACKEDPORTABLE)
		{/* Packed Image file input  */
		  int i,thisNsec;
		  float *thisPackImg;
		  float img_bkg, img_min, img_max, img_avg, img_std;
		  /* We've got to find the maxima so we can calc the scale
		     factors correctly */
		  thisNsec = 1;
		  for(i=0;i<purdueHeader.nimg;i++)
		    {/* Do all */
		      sprintf(outStr,"Statistics calculating for image [%d].",
			      i+1);
		      checkStatus(outStr);
		      thisPackImg = packImg[i];
		      calcMapStats(thisPackImg,
				   &img_bkg, 
				   &img_min,
				   &img_max, 
				   &img_avg, 
				   &img_std,
				   &ncol, &nrow, &thisNsec);
		      purdueHeader.img_bkg = MAX(purdueHeader.img_bkg,img_bkg);
		      purdueHeader.img_min = MIN(purdueHeader.img_min,img_min);
		      purdueHeader.img_max = MAX(purdueHeader.img_max,img_max);
		      purdueHeader.img_avg = MAX(purdueHeader.img_avg,img_avg);
		      purdueHeader.img_std = MAX(purdueHeader.img_std,img_std);
		    }/* Do all */
		  stuffDisplayPackBoxBuffer(0, 1, TRUE);
		}/* Packed Image file input  */
	      else if(thisFileType==MAP)
		{
		  if(paintMapSwitch)
		    calcMapStats(paintMap,
				 &purdueHeader.img_bkg, 
				 &purdueHeader.map_min,
				 &purdueHeader.map_max, 
				 &purdueHeader.img_avg, 
				 &purdueHeader.img_std,
				 &ncol, &nrow, &origNumSecs);
		  else
		    calcMapStats(float_image_buffer1,
				 &purdueHeader.img_bkg, 
				 &purdueHeader.map_min,
				 &purdueHeader.map_max, 
				 &purdueHeader.img_avg, 
				 &purdueHeader.img_std,
				 &ncol, &nrow, &origNumSecs);
		}
	      else if(thisFileType==EMPORTABLE)
		{
		  if(fileType==PORTABLE && pifMode==2)
		    {
		      origNumSecs = 1;
		      calcMapStats(inCore_buffer,
				   &purdueHeader.img_bkg, 
				   &purdueHeader.map_min,
				   &purdueHeader.map_max, 
				   &purdueHeader.img_avg, 
				   &purdueHeader.img_std,
				   &ncol, &nrow, &origNumSecs);
		    }
		  else
		    calcStats(inCore_buffer,
			      &purdueHeader.img_bkg, 
			      &purdueHeader.img_min,
			      &purdueHeader.img_max, 
			      &purdueHeader.img_avg, 
			      &purdueHeader.img_std,
			      ncol, nrow, ncol);
		}	      
		     
	      if(thisFileType==MAP)
		calcScaleFactor(&floatScaleFactor,		     
				&recipFloatScaleFactor, 
				&int4scaleF, 
				(float)(purdueHeader.map_min-1.0),
				(float)(purdueHeader.map_max+1.0),
				purdueHeader.img_avg,purdueHeader.img_std,
				thisFileType);
	      else if(fileType==PORTABLE && pifMode==2)
		calcScaleFactor(&floatScaleFactor,		     
				&recipFloatScaleFactor, 
				&int4scaleF, 
				(float)(purdueHeader.map_min-1.0),
				(float)(purdueHeader.map_max+1.0),
				purdueHeader.img_avg,purdueHeader.img_std,
				fileType);
	      else
		calcScaleFactor(&floatScaleFactor,		     
				&recipFloatScaleFactor, 
				&int4scaleF, 
				purdueHeader.img_min,purdueHeader.img_max,
				purdueHeader.img_avg,purdueHeader.img_std,
				thisFileType);

	      thisHead.recipFloatScaleFactor = recipFloatScaleFactor;
	      thisHead.floatScaleFactor = floatScaleFactor;
	      thisHead.int4scaleF = int4scaleF;

	      /* Let's not forget the pixelSize */
	      thisHead.pixelSize   = pifHeader.pixelSize;	      
	      thisHead.pixSizUnits = pifHeader.pixSizUnits;	      

	      /* purdueHeader.nimg already knows how many we've got for
	       * PACKEDPORTABLE except if some have been deleted... */
	      if(thisFileType==CIRCLEPACKED)
		{/* Boxes */
		  if(boxMethod==USECIRCLE)
		    purdueHeader.nimg = numBoxes;
		  else if (boxMethod==USERECT || boxMethod==USESQUARE)
		    purdueHeader.nimg = 1;
		}/* Boxes */
	      else if(thisFileType==RECTPACKED)
		{/* Boxes */
		  purdueHeader.nimg = 1;
		}/* Boxes */
	      else if(thisFileType == MAP)
		purdueHeader.nimg = 1;
	      else if(thisFileType == EMPORTABLE)
		purdueHeader.nimg = 1;
	      else if(thisFileType != PACKEDPORTABLE)
		purdueHeader.nimg = 1;
	      
	      if(thisFileType == PACKEDPORTABLE)
		saveNimg = afterDeletingNumImages;
	      else saveNimg = purdueHeader.nimg;
	      thisHead.nimg = saveNimg + firstImg;
	      
	      thisHead.htype= 1;

	      if(fileType==MAP)
		{/* Special */
		  mrc_nx = fileBinCols;
		  mrc_ny = fileBinRows;
		}/* Special */
	      else
		{/* Normal */
		  mrc_nx = ncol;
		  mrc_ny = nrow;
		}/* Normal */
	      if(fileType==PACKEDPORTABLE)
		mrc_nz = afterDeletingNumImages + firstImg;
	      else mrc_nz = purdueHeader.nimg + firstImg;
	      /* I propose that unless the image is threeD then we just have
	       * mrc_nz= 1 here. */
	      mrc_nz= 1;
	      if(fileType==MAP)
		mrc_nz= fileBinSecs;
	      if(fileType==EMPORTABLE)
		mrc_nz= 1;

	      thisHead.numCols = mrc_nx;
	      thisHead.numRows = mrc_ny;
	      thisHead.numSecs = mrc_nz;

	      thisHead.even = fileMrcEven;
	      thisHead.mrcX = fileMrcX;
	      thisHead.mrcY = fileMrcY;
	      thisHead.mrcZ = fileMrcZ;

	      if(thisFileType==PACKEDPORTABLE || thisFileType==CIRCLEPACKED)
		{
		  if(fileType==BYTEPORTABLE)
		    mode = 46;
		  else if(pifMode==6)
		    mode = 46;/* ALWAYS force it to be float output */
		  else if(pifMode==7)
		    mode = 46;/* ALWAYS force it to be float output */
		  else if(pifMode==46)
		    mode = 46;
		  else if(pifMode==1)
		    mode = 46;
		  else if(pifMode==2)
		    mode = 46;
		  else
		    mode= 7;
		  if(forceByte)
		    mode = 6;
		}
	      else if(thisFileType==RECTPACKED)
		mode = 6;
	      else if(thisFileType==EMPORTABLE)
		{
		  if(fileType==PORTABLE && pifMode==2)
		    mode = 2;
		  else if(pifMode==1)
		    mode = 1;
		  else 
		    mode = 0;
		}
	      else if(thisFileType==MAP)
		mode = 20;
	      
	      thisHead.mode = mode;

	      pif_write_gh(&thisHead);
	      
	    }/* Write new header */
	  

	  /*********************************************
	   * 
	   *  WRITE EACH Individually
	   * 
	   *********************************************/
	  if((mode==46) &&
	     (float_image_buffer1 == NULL))
	    CK(float_image_buffer1 = (float *)malloc(
						  sizeof(float)*
						  appendAndCurrentImgs*
						  thisHead.numCols*
						  thisHead.numRows
						  ));


	  saveNimg = -1;
	  /* OLDWAY	  
	  for(m=firstImg;m<(firstImg+purdueHeader.nimg);m++)
	  */
	  printf("first [%d] nimg[%d]  tot[%d]\n",
		 firstImg,purdueHeader.nimg,thisHead.nimg);

	  for(m=firstImg;m<appendAndCurrentImgs;m++) 
	    {/* Do each image */
	      
	      sprintf(outStr,"Processing image [%d] of [%d] total.",
		      m+1,thisHead.nimg);
	      checkStatus(outStr);

	      if((fileType==PACKEDPORTABLE && !deletePackedImage[m-firstImg]) 
		 ||
		 fileType!=PACKEDPORTABLE)
		{/* OK to write */
		  saveNimg++;
		  if(thisFileType==CIRCLEPACKED)
		    {/* packed */
		      if(boxMethod==USECIRCLE)
			{/* Circles */
			  stuffDisplayBoxBuffer(m-firstImg, 
						image_fileName.addr, 1);
			  calcCircleStats(editBox_mirror_buffer,
					  &purdueHeader.img_bkg, 
					  &purdueHeader.img_min,
					  &purdueHeader.img_max, 
					  &purdueHeader.img_avg, 
					  &purdueHeader.img_std,
					  ncol, nrow, BOX_WIDTH);
			  /* DEBUG  if(m==0)
			    printf("111 bkg[%f]min[%f]max[%f]avg[%f]std[%f]\n",
				   purdueHeader.img_bkg,
				   purdueHeader.img_min,
				   purdueHeader.img_max,
				   purdueHeader.img_avg,
				   purdueHeader.img_std
				   );
			  */
			}/* Circles */
		      else if(boxMethod==USERECT || boxMethod==USESQUARE)
			{/* Rects */
			  if(noRectYet)
			    {/* Dont' do is done above */
			      CK(out_image_buffer = 
				 (float *)malloc(sizeof(float)*
							 (ncol*nrow)));
			      stuffRectangleBuffer(out_image_buffer, 
						   image_fileName.addr);
			      
			      calcStats(out_image_buffer,
					&purdueHeader.img_bkg, 
					&purdueHeader.img_min,
					&purdueHeader.img_max, 
					&purdueHeader.img_avg, 
					&purdueHeader.img_std,
					ncol, nrow, ncol);
			    }/* Dont' do is done above */
			}/* Rects */
		    }/* packed */
		  else if(thisFileType==RECTPACKED)
		    {/* packed */
		      if(noRectYet)
			{/* Dont' do is done above */
			  CK(out_image_buffer = 
			     (float *)malloc(sizeof(float)*
						     (ncol*nrow)));
			  stuffRectangleBuffer(out_image_buffer, 
					       image_fileName.addr);
			  
			  calcStats(out_image_buffer,
				    &purdueHeader.img_bkg, 
				    &purdueHeader.img_min,
				    &purdueHeader.img_max, 
				    &purdueHeader.img_avg, 
				    &purdueHeader.img_std,
				    ncol, nrow, ncol);
			}/* Dont' do is done above */
		    }/* packed */
		  else if(thisFileType==PACKEDPORTABLE)
		    {/* Packed file input */
		      stuffDisplayPackBoxBuffer(m-firstImg, 1, TRUE);
		      calcCircleStats(editBox_mirror_buffer,
				      &purdueHeader.img_bkg, 
				      &purdueHeader.img_min,
				      &purdueHeader.img_max, 
				      &purdueHeader.img_avg, 
				      &purdueHeader.img_std,
				      ncol, nrow, BOX_WIDTH);
		      /* DEBUG if(m==0)  
			printf("222 bkg[%f]min[%f]max[%f]avg[%f]std[%f]\n",
			       purdueHeader.img_bkg,
			       purdueHeader.img_min,
			       purdueHeader.img_max,
			       purdueHeader.img_avg,
			       purdueHeader.img_std
			       );
			*/
		    }/* Packed file input */
		  else if(thisFileType==MAP)
		    {/* MAP file input */
		      if(paintMapSwitch)
			calcMapStats(paintMap,
				     &purdueHeader.img_bkg, 
				     &purdueHeader.map_min,
				     &purdueHeader.map_max, 
				     &purdueHeader.img_avg, 
				     &purdueHeader.img_std,
				     &ncol, &nrow, &origNumSecs);
		      else
			calcMapStats(float_image_buffer1,
				     &purdueHeader.img_bkg, 
				     &purdueHeader.map_min,
				     &purdueHeader.map_max, 
				     &purdueHeader.img_avg, 
				     &purdueHeader.img_std,
				     &ncol, &nrow, &origNumSecs);
		    }/* MAP file input */
		  else if(thisFileType==EMPORTABLE)
		    {/* IMG file input */
		      calcStats(inCore_buffer,
				&purdueHeader.img_bkg, 
				&purdueHeader.img_min,
				&purdueHeader.img_max, 
				&purdueHeader.img_avg, 
				&purdueHeader.img_std,
				ncol, nrow, ncol);
		    }/* IMG file input */
		  
		  /* [2] Write header for each image */
		  if(fileType==MAP)
		    {/* Special */
		      mrc_nx = fileBinCols;
		      mrc_ny = fileBinRows;
		    }/* Special */
		  else
		    {/* Normal */
		      mrc_nx = ncol;
		      mrc_ny = nrow;
		    }/* Normal */
		  if(fileType==PACKEDPORTABLE)
		    mrc_nz = 1;
		  else if (fileType==MAP)
		    mrc_nz = fileBinSecs;
		  else if (fileType==EMPORTABLE)
		    mrc_nz = 1;
		  else 
		    mrc_nz = 1;

		  thisHead.numCols = mrc_nx;
		  thisHead.numRows = mrc_ny;
		  thisHead.numSecs = mrc_nz;
		  thisHead.mode    = mode;

		  mrc_nxstart= mrc_nystart= mrc_nzstart= 0;
		  if(thisFileType==CIRCLEPACKED)
		    {/* PACKED files */
		      mrc_nxstart = purdueHeader.img_bkg;
		      mrc_nystart = featherRadius;
		    }/* PACKED files */
		  else if(thisFileType==RECTPACKED)
		    {/* PACKED files */
		      mrc_nxstart = purdueHeader.img_bkg;
		      mrc_nystart = boxRadius;
		      mrc_nystart = -99; /* Make it so we can write out new */
		    }/* PACKED files */
		  else if(thisFileType==PACKEDPORTABLE)
		    {/* PACKED files */
		      mrc_nxstart = purdueHeader.img_bkg;
		      mrc_nystart = packRadius;
		    }/* PACKED files */
		  
		  thisHead.img_bkg    = mrc_nxstart;
		  thisHead.packRadius = mrc_nystart;
		  
		  thisHead.datai4_max= purdueHeader.img_max;
		  thisHead.datar4_max= (float)purdueHeader.img_max;
		  thisHead.datai4_min= purdueHeader.img_min;
		  thisHead.datar4_min= (float)purdueHeader.img_min;
		  thisHead.datar4_avg= (float)purdueHeader.img_avg;
		  thisHead.datar4_std= (float)purdueHeader.img_std;

		  thisHead.aoverb     = aoverb;
		  thisHead.map_abang  = map_abang;
		  thisHead.dela       = dela;
		  thisHead.delb       = delb;
		  thisHead.delc       = delc;
		  thisHead.dthe       = dthe;
		  thisHead.dphi_90    = dphi_90;
		  thisHead.symmetry   = symmetry;
		  if(featherMode==LINEAR)
		    thisHead.borderType = 1;
		  else if(featherMode==GAUSSIAN)
		    thisHead.borderType = 2;
		  else if(featherMode==EXPONENTIAL)
		    thisHead.borderType = 3;
		  else 
		    thisHead.borderType = 0;
		  /* Carry through the original borderType */
		  if(fileType==PACKEDPORTABLE && thisHead.borderType == 0)
		    thisHead.borderType = borderType;
		  thisHead.borderWidth = featherRadius-boxRadius;
		  /* We need to capture how far in the boxRadius is 
		   * from the edge */
		  thisHead.borderWidth = thisHead.packRadius-boxRadius;

		  if(fileType==MAP)
		    thisHead.binFactor     = fileBinFactor;
		  else
		    thisHead.binFactor     = binFactor;

		  /* If we're appending we need to offset that here
		   * with thisNimages - also we may not write ALL
		   * of the images because they've "deleted" them
		   * - use saveNimg for that. */
		  thisHead.imgNum = thisNimages+saveNimg+1;

		  thisHead.transX = pifHeader.transX;
		  thisHead.transY = pifHeader.transY;
		  thisHead.transZ = pifHeader.transZ;
		  thisHead.isTrunc = pifHeader.isTrunc;
		  thisHead.origX = pifHeader.origX;
		  thisHead.origY = pifHeader.origY;
		  thisHead.origZ = pifHeader.origZ;

		  pif_write_dh(&thisHead);
		  thisHead.imgOffset += 512;  /* Just wrote the header */

		  /* Now call pif_write_?????_image() after we 
		   * rework the buffers
		   * so it's in Tim's row order. */

		  /* [3] Now write the image info */
		  if(thisFileType==CIRCLEPACKED)
		    {/* packed */
		      
		      /* We keep these packed images on the disk in Tim's
		       * x,y orientation (e.g. drawn from bottom up so 
		       * if we ever convert to .PCK format it's already 
		       * in the right space.
		       */
		      
		      if(boxMethod==USECIRCLE)
			{/* Circles */
			  BYTE *tmpByte_buffer;
			  unsigned short *tmpShort_buffer;
			  float *tmpInt_buffer;
			  int n;
			  float min,max,scaleF;
			  CK(tmpByte_buffer=(BYTE *)malloc(sizeof(BYTE)*
							  ncol*nrow));
			  CK(tmpShort_buffer=(unsigned short *)
			     malloc(sizeof(unsigned short)*
				    ncol*nrow));
			  CK(tmpInt_buffer=(float *)malloc(sizeof(float)*
							   ncol*nrow));
			  tsbworld_float_box(editBox_mirror_buffer,
					     tmpInt_buffer,
					     BOX_WIDTH,
					     ncol,nrow);
			  max= min = tmpInt_buffer[0];
			  for(p=0;p<nrow;p++)
			    for(n=0;n<ncol;n++)
			      {
				min = MIN(min,
					  (float)tmpInt_buffer[((p*ncol)+n)]);
				max = MAX(max,
					  (float)tmpInt_buffer[((p*ncol)+n)]);
			      }
			  if(mode==6 || mode==46)
			    scaleF = 1.0;
			  else
			    scaleF = 32000.0 / (max+2.0);
			  /* Try this we got the raw data from the file
			   * so there's no need to scale it ... */
			  scaleF = 1.0;
			  max = 0.0;
			  for(p=0;p<nrow;p++)
			    for(n=0;n<ncol;n++)
			      {
				if(mode==6)
				  tmpByte_buffer[((p*ncol)+n)] = (BYTE)
				    ((float)
				     tmpInt_buffer[((p*ncol)+n)]*
				     scaleF
				     );
				else
				  tmpShort_buffer[((p*ncol)+n)] =
				    (unsigned short)((float)
						   tmpInt_buffer[((p*ncol)+n)]*
						   scaleF
						   );
			      }
			  if(mode==6)
			    pif_write_byte_image_(&(thisHead.luNum),
						  tmpByte_buffer,
						  &(thisHead.imgOffset),
						  &(thisHead.numCols),
						  &(thisHead.numRows),
						  &(thisHead.numSecs));
			  else
			    {/* SAve for later */
			      /* OLDWAY pif_write_boxi4_(&thisHead,
				 tmpInt_buffer);  */
			      int kk,ll;
			      for(ll=0;ll<thisHead.numRows;ll++)
				for(kk=0;kk<thisHead.numCols;kk++)
				  {
				    float_image_buffer1[
							((m) * 
							 thisHead.numCols * 
							 thisHead.numRows) +
							((ll*thisHead.numCols)+
							 kk)
				    ] = 
				      tmpInt_buffer[((ll*thisHead.numCols)+kk)];
				  }
			      
			    }/* SAve for later */
			  
			  free(tmpByte_buffer);
			  free(tmpShort_buffer);
			  free(tmpInt_buffer);
			}/* Circles */
		      else if (boxMethod==USERECT || boxMethod==USESQUARE)
			{/* rects */
			  BYTE *tmpByte_buffer;
			  float min,max,scaleF;
			  CK(tmpByte_buffer=(BYTE *)malloc(sizeof(BYTE)*
							  ncol*nrow));
			  /* Do nothing */
			  max= min = out_image_buffer[0];
			  for(p=0;p<nrow;p++)
			    for(n=0;n<ncol;n++)
			      {
				min = MIN(min,
					  out_image_buffer[((p*ncol)+n)]);
				max = MAX(max,
					  out_image_buffer[((p*ncol)+n)]);
			      }
			  scaleF = 1.0;
			  for(p=0;p<nrow;p++)
			    for(n=0;n<ncol;n++)
			      {
				tmpByte_buffer[((p*ncol)+n)] =
				  (BYTE)((float)
					out_image_buffer[((p*ncol)+n)]*
					 scaleF
					);
			      }
			  pif_write_byte_image_(&(thisHead.luNum),
						tmpByte_buffer,
						&(thisHead.imgOffset),
						&(thisHead.numCols),
						&(thisHead.numRows),
						&(thisHead.numSecs));
		  
			  free(out_image_buffer);
			  free(tmpByte_buffer);
			}/* rects */
		    }/* packed */
		  else if(thisFileType==RECTPACKED)
		    {/* packed */
		      BYTE *tmpByte_buffer;
		      unsigned short *tmpShort_buffer;
		      float *tmpInt_buffer;
		      float min,max,scaleF;
		      CK(tmpByte_buffer=(BYTE *)malloc(sizeof(BYTE)*
						       ncol*nrow));
		      CK(tmpShort_buffer=(unsigned short *)
			 malloc(sizeof(unsigned short)*
						       ncol*nrow));
		      CK(tmpInt_buffer=(float *)malloc(sizeof(float)*
						     ncol*nrow));
		      tsbworld_float_box(out_image_buffer,
					 tmpInt_buffer,
					 ncol,
					 ncol,nrow);
		      max= min = tmpInt_buffer[0];
		      for(p=0;p<nrow;p++)
			for(n=0;n<ncol;n++)
			  {
			    min = MIN(min,
				      tmpInt_buffer[((p*ncol)+n)]);
			    max = MAX(max,
				      tmpInt_buffer[((p*ncol)+n)]);
			  }
		      scaleF = 1.0;
		      for(p=0;p<nrow;p++)
			for(n=0;n<ncol;n++)
			  {
			    if(mode==6)
			      tmpByte_buffer[((p*ncol)+n)] = (BYTE)
				((float)
				 tmpInt_buffer[((p*ncol)+n)]*
				 scaleF
				 );
			    else
			      tmpShort_buffer[((p*ncol)+n)] =
				(unsigned short)((float)
						 tmpInt_buffer[((p*ncol)+n)]*
						 scaleF
						 );
			  }
		      if(mode==6)
			pif_write_byte_image_(&(thisHead.luNum),
					      tmpByte_buffer,
					      &(thisHead.imgOffset),
					      &(thisHead.numCols),
					      &(thisHead.numRows),
					      &(thisHead.numSecs));
		      else
			pif_write_short_image_(&(thisHead.luNum),
					       tmpShort_buffer,
					       &(thisHead.imgOffset),
					       &(thisHead.numCols),
					       &(thisHead.numRows),
					       &(thisHead.numSecs));
		      free(tmpByte_buffer);
		      free(tmpShort_buffer);
		      free(tmpInt_buffer);
		      free(out_image_buffer);
		    }/* packed */
		  else if(thisFileType==PACKEDPORTABLE)
		    {/* packed file input */
		      BYTE *tmpByte_buffer;
		      unsigned short *tmpShort_buffer;
		      float *tmpFloat_buffer;
		      float *tmpInt_buffer;
		      float min,max,scaleF;
		      CK(tmpShort_buffer=(unsigned short *)malloc(sizeof(
						       unsigned short)*
						       ncol*nrow));
		      CK(tmpByte_buffer=(BYTE *)malloc(sizeof(BYTE)*
						       ncol*nrow));
		      CK(tmpInt_buffer=(float *)malloc(sizeof(float)*
						     ncol*nrow));
		      CK(tmpFloat_buffer=(float *)malloc(sizeof(float)*
						     ncol*nrow));

		      tsbworld_float_box(editBox_mirror_buffer,
					 tmpInt_buffer,
					 BOX_WIDTH,
					 ncol,nrow);

		      max= min = tmpInt_buffer[0];
		      for(p=0;p<nrow;p++)
			for(n=0;n<ncol;n++)
			  {
			    min = MIN(min,
				      tmpInt_buffer[((p*ncol)+n)]);
			    max = MAX(max,
				      tmpInt_buffer[((p*ncol)+n)]);
			  }

		      if(mode==6)
			scaleF = 1.0;
		      else if(mode==46)
			scaleF = 1.0 ;
		      else
			scaleF = 32000.0 / (max);

		      for(p=0;p<nrow;p++)
			for(n=0;n<ncol;n++)
			  {
			    if(mode==6)
			      tmpByte_buffer[((p*ncol)+n)] = (BYTE)
				((float)
				 tmpInt_buffer[((p*ncol)+n)]*
				 scaleF
				 );
			    else if(mode==46)
			      tmpFloat_buffer[((p*ncol)+n)] = (float)
				((float)
				 tmpInt_buffer[((p*ncol)+n)]
				 );
			    else
			      tmpShort_buffer[((p*ncol)+n)] =
				(unsigned short)((float)
						 tmpInt_buffer[((p*ncol)+n)]*
						 scaleF
						 );
			  }
		      if(mode==6)
			pif_write_byte_image_(&(thisHead.luNum),
					      tmpByte_buffer,
					      &(thisHead.imgOffset),
					      &(thisHead.numCols),
					      &(thisHead.numRows),
					      &(thisHead.numSecs));
		      else if(mode==46)
			{
			  int kk,ll;
			  for(ll=0;ll<thisHead.numRows;ll++)
			    for(kk=0;kk<thisHead.numCols;kk++)
			      {/* Do all */
	
				float_image_buffer1[
					     ((m) * 
					      thisHead.numCols * 
					      thisHead.numRows) +
					     ((ll*thisHead.numCols)+kk)
				] = 
				  tmpFloat_buffer[((ll*thisHead.numCols)+kk)];
			      }/* Do all */
			}
		      else
			pif_write_short_image_(&(thisHead.luNum),
					      tmpShort_buffer,
					      &(thisHead.imgOffset),
					      &(thisHead.numCols),
					      &(thisHead.numRows),
					      &(thisHead.numSecs));
		      free(tmpByte_buffer);
		      free(tmpShort_buffer);
		      free(tmpInt_buffer);
		      free(tmpFloat_buffer);
		    }/* packed file input */
		  else if(thisFileType==MAP)
		    {/* Do nothing for now */
		      float *tmp_float_buffer;
		      int saveBinFactor,saveNC,saveNR,saveNS;
		      CK(tmp_float_buffer=(float *)malloc(sizeof(float)*
							  numCols*numRows*
							  origNumSecs));
		      if(paintMapSwitch)
			tsbworld_float(paintMap,
				       tmp_float_buffer,
				       numCols,numRows,origNumSecs);
		      else
			tsbworld_float(float_image_buffer1,
				       tmp_float_buffer,
				       numCols,numRows,origNumSecs);
		      /* Suppose we've got a bin==2 map here we've read in.
		       * The header contains the original rehydrated 
		       * numcols/rows/secs.  SO just write it like it's
		       * bin==1 with appropriate numcols/rows/secs */
		      saveBinFactor = thisHead.binFactor;
		      saveNC = thisHead.numCols;
		      saveNR = thisHead.numRows;
		      saveNS = thisHead.numSecs;
		      thisHead.numCols = numCols;
		      thisHead.numRows = numRows;
		      thisHead.numSecs = origNumSecs;
		      thisHead.binFactor = 1;
		      pif_write_wmapi2(&thisHead,
				       tmp_float_buffer,
				       &numCols,
				       &numRows,
				       &origNumSecs);
		      thisHead.binFactor = saveBinFactor;
		      thisHead.numCols = saveNC;
		      thisHead.numRows = saveNR;
		      thisHead.numSecs = saveNS;
		      free(tmp_float_buffer);

		    }/* Do nothing for now */
		  else if(thisFileType==EMPORTABLE)
		    {/* Do nothing for now */
		      BYTE *byte_buffer;
		      unsigned short *short_buffer;
		      float min,max,scaleF;
		      int m,n,ncol,nrow;
		      ncol = thisHead.numCols;
		      nrow = thisHead.numRows;
		      if(fileType==PORTABLE && pifMode!=2)
			{/* short */
			  short_buffer = (unsigned short *)
			    malloc(sizeof(unsigned short)*
							 ncol*nrow );
			}/* short */
		      else if(fileType==BYTEPORTABLE)
			{/* byte */
			  byte_buffer = (BYTE *)malloc(sizeof(BYTE)*
							 ncol*nrow );
			}/* byte */
		      min=max=inCore_buffer[0];
		      for(m=0;m<nrow;m++)
			for(n=0;n<ncol;n++)
			  {
			    min = MIN(min,inCore_buffer[((m*ncol)+ n)]);
			    max = MAX(max,inCore_buffer[((m*ncol)+ n)]);
			  }
		      if(fileType==PORTABLE)
			{
			  if(max > 32000.0)
			    scaleF = 32000.0 / max;
			  else
			    scaleF = 1.0;
			}
		      else
			scaleF = 255.0 / max;
		      /****************************
		       * weve got to stuff the header appropriately for
		       * the case where filetype==PORTABLE and pifmode==1
		       * ????????
		       * RWA Jan 2007 
		       thisHead.floatScaleFactor = scaleF;
		       thisHead.recipFloatScaleFactor = 1.0/scaleF;
		       pif_write_gh(&thisHead);
		       */
		      for(m=0;m<nrow;m++)
			for(n=0;n<ncol;n++)
			  {/* Rescale it */
			    if(fileType==PORTABLE && pifMode != 2)
			      {/* short */
				short_buffer[((m*ncol)+ n)] = (unsigned short)
				  ((float)inCore_buffer[((m*ncol)+ n)]*scaleF);
			      }/* short */
			    else if(fileType==BYTEPORTABLE)
			      {/* byte */
				byte_buffer[((m*ncol)+ n)] = (BYTE)
				  ((float)inCore_buffer[((m*ncol)+ n)]*scaleF);
			      }/* byte */
			  }/* Rescale it */
		      if(fileType==PORTABLE && pifMode == 2)
			{/* floatInt*4 */
			  pif_write_wmapi4_(&(thisHead),
					    inCore_buffer,
					    &(thisHead.numCols),
					    &(thisHead.numRows),
					    &(thisHead.numSecs));
			}/* floatInt*4 */
		      else if(fileType==PORTABLE)
			{/* short */
			  pif_write_short_image_(&(thisHead.luNum),
						 short_buffer,
						 &(thisHead.imgOffset),
						 &(thisHead.numCols),
						 &(thisHead.numRows),
						 &(thisHead.numSecs));
			  free(short_buffer);
			}/* short */
		      else if(fileType==BYTEPORTABLE)
			{/* byte */
			  pif_write_byte_image_(&(thisHead.luNum),
						byte_buffer,
						&(thisHead.imgOffset),
						&(thisHead.numCols),
						&(thisHead.numRows),
						&(thisHead.numSecs));
			  free(byte_buffer);
			}/* byte */
		    }/* Do nothing for now */
		  
		}/* OK to write */

	    }/* Do each image */
	  
	}/* OK to proceed */
      
      if(mode==46)
	{/* Rewrite the header for mode 46 to get floatScaelFactor correct */
	  int i,j,k,totWritten;
	  int localInt;
	  char cipStr[32];
	  float backFloat;

	  if(origNumSecs == 0)
	    origNumSecs = thisHead.nimg;

#define REDO_SCALING 1
#ifdef REDO_SCALING
	  /* We calculated the scalefactors above  - we may have 
	   * underestimated some outliers here ???  */
	  calcMapStats(float_image_buffer1,
		       &purdueHeader.img_bkg, 
		       &purdueHeader.map_min,
		       &purdueHeader.map_max, 
		       &purdueHeader.img_avg, 
		       &purdueHeader.img_std,
		       &ncol, &nrow, &(thisHead.nimg));

	  calcScaleFactor(&floatScaleFactor,		     
			  &recipFloatScaleFactor, 
			  &int4scaleF, 
			  (purdueHeader.map_min-1.0),
			  (purdueHeader.map_max+1.0),
			  purdueHeader.img_avg,purdueHeader.img_std,
			  fileType);
	  thisHead.floatScaleFactor = floatScaleFactor;
	  thisHead.recipFloatScaleFactor = recipFloatScaleFactor;
	  thisHead.int4scaleF = int4scaleF;
#endif /* REDO_SCALING */

	  pif_write_gh(&thisHead);
	  
	  thisHead.datai4_max= purdueHeader.img_max;
	  thisHead.datar4_max= (float)purdueHeader.img_max;
	  thisHead.datai4_min= purdueHeader.img_min;
	  thisHead.datar4_min= (float)purdueHeader.img_min;
	  thisHead.datar4_avg= (float)purdueHeader.img_avg;
	  thisHead.datar4_std= (float)purdueHeader.img_std;
	  thisHead.imgNum = 1;
	  
	  totWritten = 0;
	  /* OLDWAY 	  for(i=0;i<purdueHeader.nimg;i++) 
	     for(i=0;i<thisHead.nimg;i++)*/
	  for(i=0;i<appendAndCurrentImgs;i++)
	    {/* Just blow out the data - we've already stuffed the headers */
	      if(i < firstImg ||
		 (!deletePackedImage[i-firstImg]))
		{/* Only do ones we want */
		  sprintf(outStr,"Writing image [%d] of [%d] total.",
			  totWritten+1,thisHead.nimg);
		  checkStatus(outStr);
		  // thisHead.imgNum = thisNimages + totWritten + 1;
		  thisHead.imgNum = totWritten + 1;
		  calcCircleStats((float_image_buffer1+
				   (i*(thisHead.numCols*
						thisHead.numRows))),
				  &purdueHeader.img_bkg, 
				  &purdueHeader.img_min,
				  &purdueHeader.img_max, 
				  &purdueHeader.img_avg, 
				  &purdueHeader.img_std,
				  ncol, nrow, ncol);
		  /*DEBUG   if(i==0)
		    printf("333 bkg[%f]min[%f]max[%f]avg[%f]std[%f]\n",
		    purdueHeader.img_bkg,
		    purdueHeader.img_min,
		    purdueHeader.img_max,
		    purdueHeader.img_avg,
		    purdueHeader.img_std
		    );
		    printf("Before writing we found these:\n\
		    \tmin[%f]\n\
		    \tmax[%f]\n\
		    \tbkg[%f]\n\
		    \tavg[%f]\n\
		    \tstd[%f]\n\n",
		    purdueHeader.img_min,
		    purdueHeader.img_max,
		    purdueHeader.img_bkg,
		    purdueHeader.img_avg,
		    purdueHeader.img_std);
		  */
		  
		  thisHead.img_bkg   = purdueHeader.img_bkg;
		  thisHead.datai4_max= purdueHeader.img_max;
		  thisHead.datar4_max= (float)purdueHeader.img_max;
		  thisHead.datai4_min= purdueHeader.img_min;
		  thisHead.datar4_min= (float)purdueHeader.img_min;
		  thisHead.datar4_avg= (float)purdueHeader.img_avg;
		  thisHead.datar4_std= (float)purdueHeader.img_std;
		  pif_write_dh(&thisHead);
		  pif_write_boxi4_(&thisHead,
				   (float_image_buffer1+
				    (i*thisHead.numCols*
				     thisHead.numRows)));
		  totWritten ++ ;
		}/* Only do ones we want */
	    }/* Just blow out the data - we've already stuffed the headers */
	  
	}/* Rewrite the header for mode 46 to get floatScaleFactor correct */
      
      pif_close(&(thisHead.luNum));
      
      if((mode==46) &&
	 (float_image_buffer1 != NULL))
	{free(float_image_buffer1);
	float_image_buffer1 = NULL;  /* For next time */}

    }/* PACKEDPORTABLE/emportable or MAP */
  
  /****************************************
   *   Depth Cued images                  *
   ****************************************/
  else if(thisFileType==DEPTHCUED && shadedMaps)
    {/* DEPTHCUED */

      fOut= NULL;
      fOut = fopen(outFileName,"r");
      if(fOut != NULL)
	{/*  Ooops we may be overwriting it!!! */
	  fclose(fOut);
	  sprintf(outStr,
		  "Do you really want to overwrite [%s]?",outFileName);
	  writeIt= askThem(outStr,TRUE,XmDIALOG_QUESTION);
	}/*  Ooops we may be overwriting it!!! */
      fOut= NULL;
      
      if(writeIt)
	{/* OK to proceed */
	  

	  sprintf(rWrite.addr,"w");     /* Write it */
	  pif_init_head(&thisHead);
	  thisHead.luNum = 5;

	  pif_open_(fName.addr, &(thisHead.luNum), rWrite.addr);
	  
	  /* [1] Write global file header */
	  stuffDisplayMapBuffer(0,1,0);
	  calcStats(editBox_int_buffer,
		    &purdueHeader.img_bkg, 
		    &purdueHeader.img_min,
		    &purdueHeader.img_max, 
		    &purdueHeader.img_avg, 
		    &purdueHeader.img_std,
		    ncol, nrow, BOX_WIDTH);
	  
	  calcScaleFactor(&floatScaleFactor,		     
			  &recipFloatScaleFactor, 
			  &int4scaleF, 
			  purdueHeader.img_min,purdueHeader.img_max,
			  purdueHeader.img_avg,purdueHeader.img_std,
			  thisFileType);
	  thisHead.recipFloatScaleFactor = recipFloatScaleFactor;
	  thisHead.int4scaleF = int4scaleF;
	  
	  thisHead.nimg = purdueHeader.nimg = mapNumSecs;
	      
	  thisHead.htype = htype= 1;
	  thisHead.numCols = mrc_nx = ncol;
	  thisHead.numRows = mrc_ny = nrow;
	  thisHead.numSecs = mrc_nz = purdueHeader.nimg;
	  thisHead.mode = mode = DEPTHCUED;
	  
	  pif_write_gh_(&thisHead);

	  for(m=0;m<purdueHeader.nimg;m++)
	    {/* Do each image */
	      
	      stuffDisplayMapBuffer(m,1,0);
	      calcStats(editBox_int_buffer,
			&purdueHeader.img_bkg, 
			&purdueHeader.img_min,
			&purdueHeader.img_max, 
			&purdueHeader.img_avg, 
			&purdueHeader.img_std,
			ncol, nrow, BOX_WIDTH);
	      
	      /* [2] Write header for each image */
	      thisHead.numCols = mrc_nx = ncol;
	      thisHead.numRows = mrc_ny = nrow;
	      thisHead.numSecs = mrc_nz = purdueHeader.nimg;

	      thisHead.img_bkg = purdueHeader.img_bkg;
	      thisHead.datai4_min = mrc_min= purdueHeader.img_min;
	      thisHead.datai4_max = mrc_max= purdueHeader.img_max;
	      thisHead.datar4_min = (float) purdueHeader.img_min;
	      thisHead.datar4_max = (float) purdueHeader.img_max;
	      thisHead.datar4_avg = purdueHeader.img_avg;
	      thisHead.datar4_std = purdueHeader.img_std;
	      if(featherMode==LINEAR)
		thisHead.borderType = 1;
	      else if(featherMode==GAUSSIAN)
		thisHead.borderType = 2;
	      else if(featherMode==EXPONENTIAL)
		thisHead.borderType = 3;
	      else 
		thisHead.borderType = 0;
	      thisHead.borderWidth = featherRadius-boxRadius;
	      
	      thisHead.imgNum = 1;
	      pif_write_dh_(&thisHead);

	      /* [3] Now write the image info */
	      pif_write_byte_image(&(thisHead.luNum),
				   editBox_display_buffer,
				   &(thisHead.imgOffset),
				   &(thisHead.numCols),
				   &(thisHead.numRows),
				   &(thisHead.numSecs));
	      
	    }/* Do each image */
	  
	noWriteCued:;
	  pif_close(&(thisHead.luNum));
	}/* OK to proceed */
      
    }/* DEPTHCUED */
  
  if(isDatabase)
    {/* Save this file to protected database area */
      char sysCall [1024];
      int noSuccess;
      sprintf(sysCall,"~leis/db/savedata.py %s %s ",
	      databaseRootStr,
	      databaseProjectID
	      );
      printf("call[%s]\n",sysCall);
      noSuccess = system(sysCall);
      printf("noSuccess[%d]\n",noSuccess);
      if(noSuccess)
	{/* Can't do it */
	  sprintf(outStr,"Problems saving database file[%s/%s]",
		  databaseProjectID,
		  databaseStr);
	  askThem(outStr,TRUE,XmDIALOG_QUESTION);
	}/* Can't do it */

    }/* Save this file to protected database area */



}/* WRITE_PIF */




/**************************************************************
 * calcShortBackground() - Calculate statistics about the image.
 * Used by the stuffDisplayBoxBuffer() routine(s) to mask the
 * background of the boxed images.  Note tha this rtn throws out 
 * data > 2.0 STDev .
 * Future goal - make one rtn that does all stats.
*/
int calcShortBackground(unsigned short *inBuf, int width, int height, 
	int boxWidth )
{/* CALCSHORTBACKGROUND */
  int i,j,totPnts,xcen,ycen,iMycen,jMxcen,thisRadius;
  float absTmp,back,std,avg,var,sumSq,distnce,cutOff,tmp,xSq,ySq;
  unsigned short b1;

  /* Init the background value */
  back = 0.0;
  if(fileType==PACKEDPORTABLE)
    thisRadius= packRadius;
  else thisRadius= boxRadius;

  /* Which method do they want? */
  sumSq = 0.0;
  if(bkgrndMethod==USEAVGPIXEL)
    {/* Average it */
      if(fileType != PORTABLE &&
	 (boxMethod==USECIRCLE && fileType!=EMPORTABLE) &&
	 (width==height))
	{/* Calc for circle boxes */
	  xcen = ycen = thisRadius;
	  totPnts = 0;
	  for(i=0;i<height;i++)
	    for(j=0;j<width;j++)
	      {/* Look at all pixels */

		b1= inBuf[((i*boxWidth)+j)];
		iMycen = (i*recipSizeFactor)-ycen;
		jMxcen = (j*recipSizeFactor)-xcen;
		xSq= (iMycen)*(iMycen);
		ySq= (jMxcen)*(jMxcen);
		distnce = (float)sqrt((double)(xSq+ySq)); 
		if(distnce>=thisRadius && ((distnce-1.0)<= thisRadius))
		  {/* we've found one on the thisRadius */
		    tmp  =  (float)b1;
		    back += tmp;
		    totPnts++;
		    sumSq += (tmp*tmp);
		}/* we've found one on the thisRadius */
	      }/* Look at all pixels */

	  std     = 0.0;
	  avg     = back/totPnts;
	  var     = (sumSq/totPnts) - (avg*avg);
	  if(var >= 0.0) 
	    std = sqrt(var);

	/*'****************************************************'
	 *'*  Throw out data points more than 2*STD from AVG  *'
	 *'****************************************************'*/
	  back   = 0.0;
	  totPnts = 0;
	  sumSq  = 0.0;
	  cutOff= avg + (2.0*std);
	  for(i=0;i<height;i++)
	    for(j=0;j<width;j++)
	      {/* Look at all pixels */
		b1 = inBuf[((i*boxWidth)+j)];
		tmp= (float)b1;
		iMycen = (i*recipSizeFactor)-ycen;
		jMxcen = (j*recipSizeFactor)-xcen;
		xSq= (iMycen)*(iMycen);
		ySq= (jMxcen)*(jMxcen);
		distnce = (float)sqrt((double)(xSq+ySq)); 
		if(distnce>=thisRadius && ((distnce-1.0)<= thisRadius))
		  {/* we've found one on the thisRadius */
		    if(tmp < avg)
		      absTmp = avg + (avg - (float)tmp);
		    else absTmp = (float)tmp;
		    if(absTmp <= cutOff) 
		      {/* then       */
			back   += tmp;
			totPnts++;
			sumSq  += (tmp*tmp);
		      }/* then       */
		}/* we've found one on the thisRadius */
	      }/* Look at all pixels */

	  std = 0.0;
	  avg = back/totPnts;
	  var = (sumSq/totPnts) - (avg*avg);
	  if(var >= 0.0) 
	    std = sqrt(var);
	  
	  back = (int)(avg);
	  
	}/* Calc for circle boxes */
      else
	{/* rectangles */

	  back = 0.0;
	  for(i=0; i<width;i++)
	    {/* Do all cols */
	      back += (float)inBuf[(i)];	      /* bottom row */
	      back += (float)inBuf[(((height-1)*boxWidth)+i)]; /* top row */
	    }/* Do all cols */
	  for(i=1; i<(height-2);i++)
	    {/* Do all cols */
	      back += (float)inBuf[(i*boxWidth)];      /* leftColumn */
	      back += (float)inBuf[((i*boxWidth)+(height-1))]; /*rightColumn*/
	    }/* Do all cols */
	  /* Compute the average background pixel */
	  back    = back/((2*width)+(2*height)-2);

	  /* We're challenged here with calculating the background for 
	   * a boxed particle saved as an "IMG" file.  Kludge for now
	   * until we can get convert to pass the background from the 
	   * header to us. */

	}/* rectangles */
    }/* Average it */
  else if(bkgrndMethod==USERANDPIXEL)
    {/* Pass back a random one */
      float factor;
      /* Get our fraction of the random number */
      factor= 1.0 / DENOMINATOR;
      
      /* Just get me a random number */
      back= ((tryColors-7) * (random() * factor));
      
    }/* Pass back a random one */
  
  return((int)back);
			    
}/* CALCSHORTBACKGROUND */


/**************************************************************
 * calcBoxSizeFactor() - Calculate how much we have to correct
 * for based on how zoomed we are in the image.
 * Used for display purposes.
*/
void calcBoxSizeFactor(int thisZoomFactor)
{/* CALCBoxSizeFACTOR */
  int k;

  /* Get ZOOMiness factor */
  sizeFactor= 1.0;
  if(thisZoomFactor >=1)
    {/* ZoomIn */
      for(k=1;k<thisZoomFactor;k++)
	sizeFactor *= 2.0;
    }/* ZoomIn */
  else
    {/* ZoomOut */
      for(k=thisZoomFactor;k<1;k++)
	sizeFactor *= 0.5;
    }/* ZoomOut */

  recipSizeFactor = 1.0 / sizeFactor ;

}/* CALCBoxSizeFACTOR */




/**********************************************************
 * calcSizeFactor() - Calculate how much we have to correct
 * for based on how zoomed we are in the image.
 * Used for display purposes.
*/
void calcSizeFactor(float thisZoomFactor)
{/* CALCSIZEFACTOR */

  if(!fullScreen)
    {/* normal */
      /* Get ZOOMiness factor */
      if(thisZoomFactor==0.0)
	sizeFactor = 1.0;
      else
	sizeFactor = pow(2.0,thisZoomFactor);
    }/* normal */
  else
    {/* Size it for full zoom */
      float scaleX,scaleY;
      if(file_width > 0 && file_height > 0)
	{/* OK to do */
	  scaleX = (float)display_width / (float)file_width;
	  scaleY = (float)display_height / (float)file_height;
	}/* OK to do */
      else
	{/* Jut try something */
	  scaleX = scaleY = 1.0;
	}/* Jut try something */
      if(scaleY > scaleX)
	sizeFactor = scaleX;
      else sizeFactor = scaleY;
      zoomFactor = (float)rwalog2(sizeFactor);
    }/* Size it for full zoom */

  recipSizeFactor = 1.0 / sizeFactor ;

}/* CALCSIZEFACTOR */



/***************************************************************************
 * evalSwitches() - Evaluate any switches they may have thrown from the
 * command line.  A lot of these are used by Xdy for his auto processing
 * rtns.
 */
void evalSwitches(int argc, char **argv)
{/* EVALSWITCHES */
  int i,m;

  /* Analyze switches */
  isDatabase    = FALSE;
  scaleItSwitch = FALSE;
  liteSwitch    = FALSE;
  masterProcess = TRUE;
  debugSwitch   = FALSE;
  debugLevel    = 0;
  shiftFFT      = FALSE;
  normLinSwitch = FALSE;
  reBoxSwitch   = FALSE;
  autoBoxSwitch = FALSE;
  boxitSwitch   = FALSE;
  doLinAlso     = FALSE;
  doFix         = FALSE;
  doNoise       = FALSE;
  doInterp      = FALSE;
  doTrunc       = FALSE;
  doOutFile     = FALSE;
  dumpNormStats = FALSE;
  dumpSpherAvg  = FALSE;
  forceAvgBkgSwitch = FALSE;

  for(i=0;i<argc;i++)
    {/* Look at them all */
      if(*(argv[i])=='-')
	{/* Could be our switch */
	  if((strstr(argv[i],"-scaleit")!=0) ||
		  (strstr(argv[i],"-SCALEIT")!=0))
	    {scaleItSwitch= TRUE;}
	  else if((strstr(argv[i],"-debug")!=0) ||
		  (strstr(argv[i],"-DEBUG")!=0))
	    {/* debug */
	      char *s1=0;
	      debugSwitch= TRUE;
	      s1 = strchr(argv[i],':');
	      if(s1)
		debugLevel = (int)strtod((s1+1),0);
	      printf("debugLevel [%d]\n",debugLevel);
	    }/* debug */
	  else if((strstr(argv[i],"-child")!=0) ||
		  (strstr(argv[i],"-CHILD")!=0))
	    {/* child */
	      masterProcess= FALSE;
	      if(hiColor)
		{/* error */
		  printf(
    "-CHILD switch not allowed for > 8 bit display.  Resetting ... \n");
		  masterProcess = TRUE;
		}/* error */

	    }/* child */
	  else if((strstr(argv[i],"-lite")!=0) ||
		  (strstr(argv[i],"-LITE")!=0))
	    {/* lite */
	      liteSwitch = TRUE;
	    }/* lite */
	  else if((strstr(argv[i],"-forceavgbkg")!=0) ||
		  (strstr(argv[i],"-forceAvgBkg")!=0) ||
		  (strstr(argv[i],"-FORCEAVGBKG")!=0))
	    {/* forceAvgBkg */
	      /* Toggle the force avg bkg options toggle */
	      forceAvgBkgSwitch = TRUE;
	    }/* forceAvgBkg */
	  else if((strstr(argv[i],"-rmblem")!=0) ||
		  (strstr(argv[i],"-RMBLEM")!=0))
	    {/* fix */
	      doFix = TRUE;
	      batchFixStd = 0.0;
	      if((i+1)<argc)
		{
		  sprintf(outStr,"%s",argv[i+1]);
		  batchFixStd = (float)strtod(outStr,0);
		}
	    }/* fix */
	  else if((strstr(argv[i],"-noise")!=0) ||
		  (strstr(argv[i],"-NOISE")!=0))
	    {/* noise */
	      doNoise = TRUE;
	      noiseStd = 0.0;
	      noiseType = GAUSSNOISE;
	      if((i+1)<argc)
		{
		  sprintf(outStr,"%s",argv[i+1]);
		  noiseStd = (float)strtod(outStr,0);
		}
	      if((i+2)<argc)
		{
		  sprintf(outStr,"%s",argv[i+2]);
		  if(EQU(outStr,"SHOT") || EQU(outStr,"shot"))
		    noiseType = NOISE;
		  if(EQU(outStr,"GAUSSIAN") || EQU(outStr,"gaussian"))
		    noiseType = GAUSSNOISE;
		}
	    }/* noise */
	  else if((strstr(argv[i],"-applyctf")!=0) ||
		  (strstr(argv[i],"-applyCtf")!=0) ||
		  (strstr(argv[i],"-applyCTF")!=0) ||
		  (strstr(argv[i],"-APPLYCTF")!=0))
	    {/* applyCTF */
	      FILE *fIn;
	      int units;
	      doApplyCtf = TRUE;
	      applyCtfFilter = 1;
	      applyCtfWiener=
		applyCtfTFactor=
		applyCtfAngMajor=
		applyCtfPixelSize=
		applyCtfDefocus_in=
		applyCtfDefocus_minor_in=
		applyCtfAmpContrast_in=
		applyCtfSphericalAbbr = 0.0;
	      applyCtfMode = 1;
	      applyCtfVolts= 1;

	      /* paramFname */
	      if((i+1)<argc)
		{/* read paramFname */
		  sprintf(outStr,"%s",argv[i+1]);
		  fIn = fopen(outStr,"r");
		  if(fIn)
		    {/* read */
		      get_str(fIn, '\n', outStr);
		      if(!preloadSwitch)
			{/* Just get the fname from the param file */
			  preloadSwitch = TRUE;
			  sprintf(preLoadStr,"%s",outStr);
			}/* Just get the fname from the param file */
		      get_str(fIn, ',', outStr);
		      applyCtfPixelSize = (float)strtod(outStr,0);
		      get_str(fIn, ',', outStr);
		      units = (int)strtod(outStr,0);
		      if(units != 2)
			applyCtfPixelSize  *= 0.1;  /* always in nm here */
		      get_str(fIn, ',', outStr);
		      applyCtfVolts = (int)strtod(outStr,0);
		      get_str(fIn, ',', outStr);
		      applyCtfAmpContrast_in = (float)strtod(outStr,0);
		      get_str(fIn, ',', outStr);
		      applyCtfDefocus_in = (float)strtod(outStr,0);
		      get_str(fIn, ',', outStr);
		      applyCtfDefocus_minor_in = (float)strtod(outStr,0);
		      get_str(fIn, ',', outStr);
		      applyCtfAngMajor = (float)strtod(outStr,0);
		      get_str(fIn, '\n', outStr);
		      applyCtfSphericalAbbr = (float)strtod(outStr,0);

		      fclose(fIn);
		    }/* read */
		}/* read paramFname */
	      /* filter */
	      if((i+2)<argc)
		{sprintf(outStr,"%s",argv[i+2]);
		  applyCtfFilter = (int)strtod(outStr,0);}
	      /* wiener */
	      if((i+3)<argc)
		{sprintf(outStr,"%s",argv[i+3]);
		  applyCtfWiener = (float)strtod(outStr,0);}
	      /* tFactor */
	      if((i+4)<argc)
		{sprintf(outStr,"%s",argv[i+4]);
		  applyCtfTFactor = (float)strtod(outStr,0);}
	      /* mode */
	      if((i+5)<argc)
		{sprintf(outStr,"%s",argv[i+5]);
		  applyCtfMode = (int)strtod(outStr,0);}
	    }/* applyCTF */
	  else if((strstr(argv[i],"-interp")!=0) ||
		  (strstr(argv[i],"-INTERP")!=0))
	    {/* interp */
	      doInterp = TRUE;
	      doInterpOrigSize=
		doInterpOutputSize = 0.0;
	      if((i+1)<argc)
		{
		  sprintf(outStr,"%s",argv[i+1]);
		  doInterpOrigSize = (float)strtod(outStr,0);
		}
	      if((i+2)<argc)
		{
		  sprintf(outStr,"%s",argv[i+2]);
		  doInterpOutputSize = (float)strtod(outStr,0);
		}
	    }/* interp */
	  else if((strstr(argv[i],"-trunc")!=0) ||
		  (strstr(argv[i],"-TRUNC")!=0))
	    {/* trunc */
	      doTrunc = TRUE;
	      truncVal = 0.0;
	      if((i+1)<argc)
		{ sprintf(outStr,"%s",argv[i+1]);
		  truncBegX = (int)strtod(outStr,0);}
	      if((i+2)<argc)
		{ sprintf(outStr,"%s",argv[i+2]);
		  truncEndX = (int)strtod(outStr,0);}
	      if((i+3)<argc)
		{ sprintf(outStr,"%s",argv[i+3]);
		  truncBegY = (int)strtod(outStr,0);}
	      if((i+4)<argc)
		{ sprintf(outStr,"%s",argv[i+4]);
		  truncEndY = (int)strtod(outStr,0);}
	      if((i+5)<argc)
		{ sprintf(outStr,"%s",argv[i+5]);
		  truncBegZ = (int)strtod(outStr,0);}
	      if((i+6)<argc)
		{ sprintf(outStr,"%s",argv[i+6]);
		  truncEndZ = (int)strtod(outStr,0);}
	      if((i+7)<argc)
		{ sprintf(outStr,"%s",argv[i+7]);
		  truncVal = (float)strtod(outStr,0);}

	    }/* trunc */
	  else if((strstr(argv[i],"-lingrad")!=0) ||
		  (strstr(argv[i],"-LINGRAD")!=0))
	    {/* normLinGrad */
	      doLinAlso = TRUE;
	    }/* normLinGrad */
	  else if((strstr(argv[i],"-outfile")!=0) ||
		  (strstr(argv[i],"-outFile")!=0) ||
		  (strstr(argv[i],"-OUTFILE")!=0))
	    {/* outFname */
	      doOutFile = TRUE;
	      if((i+1)<argc)
		sprintf(outFname,"%s",argv[i+1]);
	    }/* outFname */
	  else if((strstr(argv[i],"-box")!=0) ||
		  (strstr(argv[i],"-BOX")!=0))
	    {/* boxit */
	      boxitSwitch = TRUE;
	      boxitSize = 50;
	      boxitBoxSize = 55;
	      boxitFeatherSize = 55;
	      if((i+1)<argc)
		sprintf(boxitFname,"%s",argv[i+1]);
	      else
		sprintf(boxitFname,"UNK");
	      if((i+2)<argc)
		{
		  if(EQU(argv[i+2],"GAUSSIAN"))
		    featherMode = GAUSSIAN;
		  else if(EQU(argv[i+2],"gaussian"))
		     featherMode = GAUSSIAN;
		  else if(EQU(argv[i+2],"EXPONENTIAL"))
		     featherMode = EXPONENTIAL;
		  else if(EQU(argv[i+2],"exponential"))
		     featherMode = EXPONENTIAL;
		  else if(EQU(argv[i+2],"LINEAR"))
		     featherMode = LINEAR;
		  else if(EQU(argv[i+2],"linear"))
		     featherMode = LINEAR;
		  else if(EQU(argv[i+2],"NONE"))
		     featherMode = OLD;
		  else if(EQU(argv[i+2],"none"))
		     featherMode = OLD;
		 }
	      if((i+3)<argc)
		{
		  sprintf(outStr,"%s",argv[i+3]);
		  boxitSize = (int)strtod(outStr,0);
		}
	      if((i+4)<argc)
		{
		  sprintf(outStr,"%s",argv[i+4]);
		  boxitBoxSize = (int)strtod(outStr,0);
		}
	      if((i+5)<argc)
		{
		  sprintf(outStr,"%s",argv[i+5]);
		  boxitFeatherSize = (int)strtod(outStr,0);
		}
	      if(boxitBoxSize <= boxitSize)
		boxitBoxSize = boxitSize + 1;
	      if(boxitFeatherSize < boxitBoxSize)
		boxitFeatherSize = boxitBoxSize ;
	    }/* boxit */
	  else if((strstr(argv[i],"-rebox")!=0) ||
		  (strstr(argv[i],"-reBox")!=0) ||
		  (strstr(argv[i],"-REBOX")!=0))
	    {/* reBox */
	      reBoxSwitch = TRUE;
	      autoBoxSize = 50;
	      autoBoxBoxSize = 55;
	      autoBoxFeatherSize = 55;
	      if((i+1)<argc)
		{
		  if(EQU(argv[i+1],"GAUSSIAN"))
		    featherMode = GAUSSIAN;
		  else if(EQU(argv[i+1],"gaussian"))
		     featherMode = GAUSSIAN;
		  else if(EQU(argv[i+1],"EXPONENTIAL"))
		     featherMode = EXPONENTIAL;
		  else if(EQU(argv[i+1],"exponential"))
		     featherMode = EXPONENTIAL;
		  else if(EQU(argv[i+1],"LINEAR"))
		     featherMode = LINEAR;
		  else if(EQU(argv[i+1],"linear"))
		     featherMode = LINEAR;
		  else if(EQU(argv[i+1],"NONE"))
		     featherMode = OLD;
		  else if(EQU(argv[i+1],"none"))
		     featherMode = OLD;
		 }
	      if((i+2)<argc)
		{
		  sprintf(outStr,"%s",argv[i+2]);
		  autoBoxBoxSize = (int)strtod(outStr,0);
		}
	      if((i+3)<argc)
		{
		  sprintf(outStr,"%s",argv[i+3]);
		  autoBoxFeatherSize = (int)strtod(outStr,0);
		}
	      if(autoBoxFeatherSize < autoBoxBoxSize)
		autoBoxFeatherSize = autoBoxBoxSize ;

	    }/* reBox */
	  else if((strstr(argv[i],"-autobox")!=0) ||
		  (strstr(argv[i],"-AUTOBOX")!=0))
	    {/* autoBox */
	      autoBoxSwitch = TRUE;
	      autoBoxSize = 50;
	      autoBoxBoxSize = 55;
	      autoBoxFeatherSize = 55;
	      autoBoxThin = 1;
	      autoBoxDilates = 0;
	      autoBoxIter = 1;
	      if((i+1)<argc)
		{
		  if(EQU(argv[i+1],"GAUSSIAN"))
		    featherMode = GAUSSIAN;
		  else if(EQU(argv[i+1],"gaussian"))
		     featherMode = GAUSSIAN;
		  else if(EQU(argv[i+1],"EXPONENTIAL"))
		     featherMode = EXPONENTIAL;
		  else if(EQU(argv[i+1],"exponential"))
		     featherMode = EXPONENTIAL;
		  else if(EQU(argv[i+1],"LINEAR"))
		     featherMode = LINEAR;
		  else if(EQU(argv[i+1],"linear"))
		     featherMode = LINEAR;
		  else if(EQU(argv[i+1],"NONE"))
		     featherMode = OLD;
		  else if(EQU(argv[i+1],"none"))
		     featherMode = OLD;
		 }
	      if((i+2)<argc)
		{
		  sprintf(outStr,"%s",argv[i+2]);
		  autoBoxSize = (int)strtod(outStr,0);
		}
	      if((i+3)<argc)
		{
		  sprintf(outStr,"%s",argv[i+3]);
		  autoBoxBoxSize = (int)strtod(outStr,0);
		}
	      if((i+4)<argc)
		{
		  sprintf(outStr,"%s",argv[i+4]);
		  autoBoxFeatherSize = (int)strtod(outStr,0);
		}
	      if((i+5)<argc)
		{
		  sprintf(outStr,"%s",argv[i+5]);
		  autoBoxThin = (int)strtod(outStr,0);
		}
	      if((i+6)<argc)
		{
		  sprintf(outStr,"%s",argv[i+6]);
		  autoBoxIter = (int)strtod(outStr,0);
		}
	      if((i+7)<argc)
		{
		  sprintf(outStr,"%s",argv[i+7]);
		  autoBoxDilates = (int)strtod(outStr,0);
		}
	      if(autoBoxBoxSize <= autoBoxSize)
		autoBoxBoxSize = autoBoxSize + 1;
	      if(autoBoxFeatherSize < autoBoxBoxSize)
		autoBoxFeatherSize = autoBoxBoxSize ;
	    }/* autoBox */
	  else if((strstr(argv[i],"-dumpnorm")!=0) ||
		  (strstr(argv[i],"-dumpNorm")!=0) ||
		  (strstr(argv[i],"-DUMPNORM")!=0))
	    {/* dumpnorm */
	      dumpNormStats = TRUE;
	      if((i+1)<argc)
		{
		  sprintf(dumpNormFile,"%s",argv[i+1]);
		}
	    }/* dumpnorm */
	  else if((strstr(argv[i],"-dumpspheravg")!=0) ||
		  (strstr(argv[i],"-dumpSpherAvg")!=0) ||
		  (strstr(argv[i],"-DUMPSPHERAVG")!=0))
	    {/* dumpspheravg */
	      dumpSpherAvg = TRUE;
	      sprintf(dumpSpherAvgFile,"dump_spheravg.dat");
	      if((i+1)<argc)
		{
		  sprintf(dumpSpherAvgFile,"%s",argv[i+1]);
		}
	    }/* dumpspheravg */
	  else if((strstr(argv[i],"-norm")!=0) ||
		  (strstr(argv[i],"-NORM")!=0))
	    {/* normLinGrad */
	      normLinSwitch = TRUE;
	      normLinMean = normLinStd = 0.0;
	      if((i+1)<argc)
		{
		  sprintf(outStr,"%s",argv[i+1]);
		  normLinMean = (float)strtod(outStr,0);
		}
	      if((i+2)<argc)
		{
		  sprintf(outStr,"%s",argv[i+2]);
		  normLinStd = (float)strtod(outStr,0);
		}
	      printf("Setting norm mean/std [%f,%f]\n",
		     normLinMean,normLinStd);
	    }/* normLinGrad */
	  else if((strstr(argv[i],"-pre")!=0) ||
		  (strstr(argv[i],"-PRE")!=0))
	    {/* preload */
	      preloadSwitch = TRUE;
	      sprintf(preLoadStr,"%s",argv[i+1]);
	    }/* preload */
	  else if((strstr(argv[i],"-dat")!=0) ||
		  (strstr(argv[i],"-DAT")!=0))
	    {/* preload */
	      isDatabase = TRUE;
	      databaseIter = 1;
	      if((i+1)<argc)
		sprintf(databaseProjectID,"%s",argv[i+1]);
	      else
		{printf("No database projectID specified.\n");
		exit(0);}
	    }/* preload */
	  else if((strstr(argv[i],"-river")!=0) ||
		  (strstr(argv[i],"-RIVER")!=0))
	    {/* river */
	      riverSwitch = TRUE;
	    }/* river */
	  else if((strstr(argv[i],"-zoom")!=0) ||
		  (strstr(argv[i],"-ZOOM")!=0))
	    {/* zoom */
	      zoomItSwitch = TRUE;
	      if((i+1) < argc)
		{/* OK */
		  zoomItFactor = (float)strtod(argv[i+1],0);
		}/* OK */
	      else
		{/* error */
		  printf("ZOOM switch needs a zoomFactor argument like:\n\n\
     -zoom 2\n\n");
		  exit(0);
		}/* error */
	    }/* zoom */
	  else if((strstr(argv[i],"-shade")!=0) ||
		  (strstr(argv[i],"-SHADE")!=0))
	    {/* shade */
	      shadeItSwitch = TRUE;
	      if((i+1) < argc)
		{/* OK */
		  shadeItDensity = (float)strtod(argv[i+1],0);
		}/* OK */
	      else
		{/* error */
		  printf("SHADE switch needs a density argument like:\n\n\
     -shade 16\n\n");
		  exit(0);
		}/* error */
	    }/* shade */
	  else if((strstr(argv[i],"-shiftFFT")!=0) ||
		  (strstr(argv[i],"-shiftfft")!=0) ||
		  (strstr(argv[i],"-SHIFTFFT")!=0))
	    {/* shiftFFT */
	      shiftFFT = TRUE;
	    }/* shiftFFT */
	  else if((strstr(argv[i],"-roblite")!=0) ||
		  (strstr(argv[i],"-ROBLITE")!=0))
	    {/* roblite */
	      liteSwitch = TRUE;
	      drawLite = FALSE;
	    }/* roblite */
	  else if((strstr(argv[i],"-usage")!=0) ||
		  (strstr(argv[i],"-USAGE")!=0))
	    {/* Usage */
	      printf(
        "robem \n\
     -APPLYCTF  paramFname filter wiener tFactor mode\n\
     -AUTOBOX  featherMode(GAUSSIAN,EXPONENTIAL,LINEAR,NONE)\n\
               emmaRadius boxRadius featherRadius numThin numIter numDilates\n\
     -BOX  boxCoordsFilename featherMode(GAUSSIAN,EXPONENTIAL,LINEAR,NONE)\n\
               emmaRadius boxRadius featherRadius\n\
     -DATABASE projectID\n\
     -DEBUG:n  (n is the debug level)\n\
     -DUMPNORM dumpNormFile\n\
     -DUMPSPHERAVG dumpSpericalAverageFile\n\
     -FORCEAVGBKG\n\
     -INTERP origSize outputSize\n\
     -LINGRAD  (Used with the -NORM switch)\n\
     -NOISE noiseStd noiseType(SHOT,GAUSSIAN)\n\
     -NORM normMean normStd (if normMean==0.0 && normStd==0.0 then use what\n\
                             robEM calculates for those values)\n\
                             **Use -LINGRAD to do linear gradient also**\n\
     -OUTFILE fname\n\
     -PRELOAD fname\n\
     -REBOX  featherMode(GAUSSIAN,EXPONENTIAL,LINEAR,NONE)\n\
               boxRadius featherRadius\n\
     -RMBLEM stdDev  (if stDev==0.0 then use program default for stDev)\n\
     -SCALEIT\n\
     -SHADE density\n\
     -TRUNC begX endX begY endY begZ endZ fillValue\n\
     -USAGE\n\
     -ZOOM zoomFactor\n\
 \n");
	      exit(0);
	    }/* Usage */
	}/* Could be our switch */
    }/* Look at them all */


  if(isDatabase)
    {/* get root */
      char *s1;
      char tmpStr[512];
      sprintf(tmpStr,preLoadStr);
      s1 = strrchr(tmpStr,'.');
      if(s1)
	*s1 = '\0';
      sprintf(databaseRootStr,"%s",tmpStr);
    }/* get root */


}/* EVALSWITCHES */




/**********************************************************************
 * readResourceFile()-  This needs to be developed further.  Currently
 * users can put certain configuration defaults in a ~/.robemrc file
 * that will change the config of robem at startup.  For example,
 * some folks like to have the immediateUpdate function on the 3D
 * rendering screen always be defaulted to ON whereas others don't 
 * like that so much.
*/
void readResourceFile()
{/* READRESOURCEFILE */
  FILE *fIn;
  char inStr[512];
  char key[256],state[256];
  XmString simpleString;

  sprintf(outStr,"%s/.robemrc",getenv("HOME"));
  fIn = fopen(outStr,"r");
  
  if(fIn)
    {/* There is one to read */

      while(!feof(fIn))
	{/* Do all input */
	  get_str(fIn, '\n', inStr);
	  
	  sscanf(inStr,"%s",key);
	  if(EQU(key,"IMMEDIATEUPDATE"))
	    {/* Immediate Update */
	      sscanf((inStr+(strlen(key))), "%s",state);
	      if(EQU(state,"AUTO"))
		{/* Set it up for automatic */
		  /* Not yet - when lowRes YES 
		     immediateMapShade = TRUE; 
		     XtSetArg(arglst[0], XmNset, TRUE);
		     XtSetValues(immediateUpdateW, arglst, 1); 
		     */
		  immediateResource = FALSE;          
		}/* Set it up for automatic */
	      else if(EQU(state,"MANUAL"))
		{/* Manual */
		  immediateResource = TRUE;          
		  immediateMapShade = FALSE;
		  XtSetArg(arglst[0], XmNset, FALSE);
		  XtSetValues(immediateUpdateW, arglst, 1); 
		}/* Manual */
	      else
		printf("I don't understand [%s][%s] from resource file.\n",
		       key,state);
	    }/* Immediate Update */
	  else if(EQU(key,"TRUEPHI"))
	    {/*  */
	      sscanf((inStr+(strlen(key))), "%s",state);
	      if(EQU(state,"ON"))
		{/* Set it ON */
		  truePhi = TRUE;
		  XtSetArg(arglst[0], XmNset, TRUE);
		  XtSetValues(truePhiW, arglst, 1); 
		}/* Set it ON */
	      else if(EQU(state,"OFF"))
		{/* off */
		  truePhi = FALSE;
		  XtSetArg(arglst[0], XmNset, FALSE);
		  XtSetValues(truePhiW, arglst, 1); 
		}/* off */
	      else
		printf("I don't understand [%s][%s] from resource file.\n",
		       key,state);
	    }/*  */
	  else if(EQU(key,"FORCEAVGBKG"))
	    {/*  */
	      sscanf((inStr+(strlen(key))), "%s",state);
	      if(EQU(state,"ON"))
		{/* Set it ON */
		  forceAvgBkg = TRUE;
		  XtSetArg(arglst[0], XmNset, TRUE);
		  XtSetValues(forceAvgBkgToggleW, arglst, 1); 
		}/* Set it ON */
	      else if(EQU(state,"OFF"))
		{/* off */
		  forceAvgBkg = FALSE;
		  XtSetArg(arglst[0], XmNset, FALSE);
		  XtSetValues(forceAvgBkgToggleW, arglst, 1); 
		}/* off */
	      else
		printf("I don't understand [%s][%s] from resource file.\n",
		       key,state);
	    }/*  */
	  else if(EQU(key,"CALCSFS"))
	    {/*  */
	      sscanf((inStr+(strlen(key))), "%s",state);
	      if(EQU(state,"ON"))
		{/* Set it ON */
		  calcSFs = TRUE;
		  XtSetArg(arglst[0], XmNset, TRUE);
		  XtSetValues(calcSfsW, arglst, 1); 
		}/* Set it ON */
	      else if(EQU(state,"OFF"))
		{/* off */
		  calcSFs = FALSE;
		  XtSetArg(arglst[0], XmNset, FALSE);
		  XtSetValues(calcSfsW, arglst, 1); 
		}/* off */
	      else
		printf("I don't understand [%s][%s] from resource file.\n",
		       key,state);
	    }/*  */
	  else if(EQU(key,"CALCBIN"))
	    {/*  */
	      sscanf((inStr+(strlen(key))), "%s",state);
	      if(EQU(state,"ON"))
		{/* Set it ON */
		  calcBin = FALSE;
		  XtSetArg(arglst[0], XmNset, FALSE);
		  XtSetValues(calcBinW, arglst, 1); 
		}/* Set it ON */
	      else if(EQU(state,"OFF"))
		{/* off */
		  calcBin = FALSE;
		  XtSetArg(arglst[0], XmNset, FALSE);
		  XtSetValues(calcBinW, arglst, 1); 
		}/* off */
	      else
		printf("I don't understand [%s][%s] from resource file.\n",
		       key,state);
	    }/*  */
	  else if(EQU(key,"MAPSHADE"))
	    {/*  */
	      sscanf((inStr+(strlen(key))), "%s",state);
	      if(EQU(state,"ON"))
		{/* Set it ON */
		  secondMapShade = TRUE;
		  XtSetArg(arglst[0], XmNset, TRUE);
		  XtSetValues(secondShadeWidget, arglst, 1); 
		}/* Set it ON */
	      else if(EQU(state,"OFF"))
		{/* off */
		  secondMapShade = FALSE;
		  XtSetArg(arglst[0], XmNset, FALSE);
		  XtSetValues(secondShadeWidget, arglst, 1); 
		}/* off */
	      else
		printf("I don't understand [%s][%s] from resource file.\n",
		       key,state);

	      if(secondMapShade)
		{/* Give feedback */
		  sprintf(outStr,"Shade ON");
		  XtSetArg(arglst[0], XmNsensitive, TRUE );
		  XtSetValues(fractScrollWidget, arglst, 1); 
		  XtSetValues(angxScrollWidget, arglst, 1); 
		  XtSetValues(angyScrollWidget, arglst, 1); 
		  XtSetValues(fractLabelWidget, arglst, 1); 
		  XtSetValues(angxLabelWidget, arglst, 1); 
		  XtSetValues(angyLabelWidget, arglst, 1); 
		  XtSetValues(fractValueWidget, arglst, 1); 
		  XtSetValues(angxValueWidget, arglst, 1); 
		  XtSetValues(angyValueWidget, arglst, 1); 
		  XtSetValues(fractToggleWidget, arglst, 1); 
		  XtSetValues(angxToggleWidget, arglst, 1); 
		  XtSetValues(angyToggleWidget, arglst, 1); 
		}/* Give feedback */
	      else 
		{/* Give feedback */
		  sprintf(outStr,"Shade OFF");
		  XtSetArg(arglst[0], XmNsensitive, FALSE );
		  XtSetValues(fractScrollWidget, arglst, 1); 
		  XtSetValues(angxScrollWidget, arglst, 1); 
		  XtSetValues(angyScrollWidget, arglst, 1); 
		  XtSetValues(fractLabelWidget, arglst, 1); 
		  XtSetValues(angxLabelWidget, arglst, 1); 
		  XtSetValues(angyLabelWidget, arglst, 1); 
		  XtSetValues(fractValueWidget, arglst, 1); 
		  XtSetValues(angxValueWidget, arglst, 1); 
		  XtSetValues(angyValueWidget, arglst, 1); 
		  XtSetValues(fractToggleWidget, arglst, 1); 
		  XtSetValues(angxToggleWidget, arglst, 1); 
		  XtSetValues(angyToggleWidget, arglst, 1); 
		}/* Give feedback */
	      createSimpleString(&simpleString,outStr);
	      XtSetArg(arglst[0], XmNlabelString, simpleString );
	      XtSetArg(arglst[1], XmNrecomputeSize, FALSE );
	      XtSetValues(secondShadeWidget, arglst, 2); 
	      XFlush(dsply);
	      XmStringFree(simpleString); 

	      /* Turn radial Projections OFF */
	      radialProjections = FALSE;
	      sprintf(outStr,"Radial/Icos Proj OFF");
	      createSimpleString(&simpleString,outStr);
	      XtSetArg(arglst[0], XmNlabelString, simpleString );
	      XtSetArg(arglst[1], XmNrecomputeSize, FALSE );
	      XtSetValues(radProjShadeWidget, arglst, 2); 
	      XFlush(dsply);
	      XmStringFree(simpleString);
	      /* Now reset toggles */
	      XtSetArg(arglst[0], XmNset, FALSE);
	      XtSetValues(radProjShadeWidget, arglst, 1);
	      XtUnmanageChild(icosRadProjToggleW);
	      XtUnmanageChild(radProjToggleW);
	      XtUnmanageChild(radProjCSToggleW);
	      XtManageChild(radialCSToggleW);
	      XtManageChild(radialTextWidget);
	      XtManageChild(radialLabelWidget);
	      XFlush(dsply);
	      
	      /* Now activate the surface density widget */
	      XtSetArg(arglst[0], XmNsensitive, TRUE );
	      XtSetValues(surfDensScrollWidget, arglst, 1); 
	      XtSetValues(surfDensLabelWidget, arglst, 1); 
	      XtSetValues(surfDensValueWidget, arglst, 1); 
	      XtSetValues(surfDensToggleWidget, arglst, 1); 
	      
	    }/*  */
	  else if(*key=='!')
	    {/*  */
	      /* OK is a comment */
	    }/*  */
	  else if(EQU(key,""))
	    {/*  */
	      /* OK is a blank line */
	    }/*  */
	  else 
	    {/*  */
	      printf("I don't understand [%s] from resource file.\n",
		     key);
	    }/*  */
	}/* Do all input */
	  
      fclose(fIn);
    }/* There is one to read */
  else
    {/* */
      printf("No resource file.\n");
    }/* */

}/* READRESOURCEFILE */






/******************************************************************
 * init_globals() - Init out anything that needs it before we
 * start executing.  See groundZero() for anything that needs
 * initing later on in the execution.  e.g. init_globals() is
 * done once at prg startup and groundZero() is done whenever
 * a new file is being read.
*/
void init_globals(int argc, char **argv)
{/* INIT_GLOBALS */
  int i,m;
  unsigned int seedVal;

  hiColor = FALSE;

  drawLite = TRUE;
  riverSwitch = FALSE;
  preloadSwitch = FALSE;
  shadeItSwitch = FALSE;
  zoomItSwitch = FALSE;
  featherMode = EXPONENTIAL;
  evalSwitches(argc,argv);

  if(liteSwitch)
    defDisplayBufferSize = LITE_DEF_DISPLAY_BUFFER_SIZE;
  else
    defDisplayBufferSize = DEF_DISPLAY_BUFFER_SIZE;

  /* Colors */
  redLevel =  greenLevel = blueLevel = 0;
  brightLevel = 255.0;

  /* Generic counters etc... */
  r4_min= -1.0e+38;
  r4_max=  1.0e+38;
  fullScreen= FALSE;
  borderType = 0;
  boxRadius  = 50;
  featherRadius  = 50;
  emmaRadius = 45;
  emptyShellWidth = 5;
  boxRadiusDisp = TRUE;
  emptyShellDisp = FALSE;
  emmaRadiusDisp = TRUE;
  featherRadiusDisp = TRUE;
  refineDisp = FALSE;
  clusterDisp = FALSE;
  referenceDisp = FALSE;
  imgNum= 1;
  curBox= numBoxes= numRects= numPoly= 0;
  findBox = -99;
  makeSquareBox = FALSE;
  boxZoom = USEPOINT;
  boxMethod = USECIRCLE;
  backgroundMode = USECIRCLE;
  bkgrndMethod= USEAVGPIXEL;
  dispMethod = CIRCLEDISP;
  histMethod= FALSE;
  histoEqualize = FALSE;
  writeType = EMPORTABLE;
  colorType = USEGENERIC;
  zoomFactor = 0.0;
  oldZoomFactor = -99.0;
  oldSizeFactor = -99.0;
  editBoxZoom= 1;
  stepCM = 256;
  virXoffset= virYoffset= xoffset= yoffset= 0;
  rectBox.drawable= FALSE;
  rectBox.width = rectBox.height= 0;
  buffer_width= buffer_height= 0;
  locked= FALSE;
  largeImage= FALSE;
  inited= FALSE;
  noCallbackScroll= FALSE;
  makingMovie= FALSE;
  memoryMiser = TRUE;
  aSymmetric = FALSE;
  imgInCore = FALSE;
  buildingCmap = FALSE;

  /* Seed the random number generator */
  seedVal = (unsigned int)time(NULL);
  srandom((unsigned int)seedVal);
      
  /* Electronic ruler */
  rulerMag= rulerPixelWidth= 0.0;

  /* Filename descriptor for FORTRAN */
  image_fileName.addr = (char *)calloc(2048, sizeof(char));  
  image_fileName.length = 2048;
  fileName.addr = (char *)calloc(2048, sizeof(char));  
  fileName.length = 2048;
  image_fileName.type = 14;
  image_fileName.class = 1;
  fileName.type = 14;
  fileName.class = 1;

  for(i=0;i<MAXBOX;i++)
    {/* Init thes out */
      packImg[i]= NULL;
      fix2DpackImg_buffer[i] = NULL;
      fix2DpackImg_diff[i] = NULL;
      box[i].editBox_buffer= 
	box[i].edit_mirror_buffer= NULL;
      box[i].inMemory= FALSE;
    }/* Init thes out */
  lastPackedNumImgs= 0;
  lastDeleted= 0;

  /* Map shading stuff */
  secondMapShade = FALSE;
  immediateMapShade = FALSE;
  immediateResource = FALSE;          
  radialMapShade = FALSE;
  radialProjections = FALSE;
  stereoMapShade = FALSE;
  lowRes = TRUE;
  globFract= 0.75;
  globAngx= 10.0;
  globAngy= -10.0;   /* Why negative??? */
  globOmega =
    globTheta = 
      globPhi = 
	globMaxRadius= 
	  origGlobMaxRadius= 
	    globMinRadius= 0.0;
  globPixelSize = 1.0;
  rotateXYZ = FALSE;
  whichStep = NONE;
  shadedMaps = FALSE;
  for(m=0;m<MAXSECTIONS;m++)
    cuedMap[m].depth= cuedMap[m].radial= cuedMap[m].shad= NULL;
  tmpRotMap = rotMap = NULL;
  int_image_buffer1= NULL;
  float_image_buffer1= NULL;
  rwa_tmp2d_map= rwa_tmp_map= NULL;
  /* Start these guys out small */
  CK(rwa_tmp2d_map = (double *)malloc((2) * sizeof(double)));
  CK(rwa_tmp_map   = (double *)malloc((2) * sizeof(double)));
  CK(map_image_buffer   = (float *)malloc((2) * sizeof(float)));
  
  /* Boxing stuff */
  initLastNumBox = FALSE;
  lastNumBox = 0;
  dispThickCircles = FALSE;
  showCC = FALSE;
  showDebug = TRUE;
  showBoxingSwitch = FALSE;  /* Not yet */
  darkCircleRef = FALSE;
  cycleRef = FALSE;
  refReadFile = FALSE;

  autoZoom = TRUE;
  rockIt = FALSE;
  flashLightIt = FALSE;

  fullScreen_buffer.dispx = fullScreen_buffer.dispy = 0;

  irev= -1;
  rampSwitch = FALSE;
  rainbowSwitch = FALSE;

  /* CircAvg stuff */
  caInvertContrast= FALSE;  
  currentCircAvg= 0;
  selectedCircAvg= -99;
  circAvgSize= 1;
  caOffsetX = 0;

  maxRadiusSwitch = TRUE;
  movieSpeed= 0;
  flickSpeed = 0;
  totContours = 0;

  rotateAxis = Y;
  expandStep = 8;

  /* FFT stuff */
  fftSize = 512;
  idim1= idim2= 1024;          /* Allocate the buffer to 1024 - MAX size */
  maxFFTvalue= 100000000.00;
  minFFTvalue= 0.00;
  tempFactor = 0;
  for(i=0;i<1024;i++)
    for(j=0;j<2048;j++)
      {/* Init */
	ctfValues[i][j] = 1.0;
	tempFactors[i][j] = 1.0;
      }/* Init */
  fftMode= FALSE;
  /* fftType= LINEARFFT; */
  fftType= LOGFFT;
/* OLD way - let's do it all the time now 
  backTransform = FALSE;
*/
  backTransform = FALSE;
  pickLattice = 1;
  pickCTF     = -99;
  pickedLatPts = TRUE;
  latPt[0].x = 1;
  latPt[0].y = 0;
  latPt[1].x = 0;
  latPt[1].y = 1;
  latPt[0].measuredX = 290;
  latPt[0].measuredY = 256;
  latPt[1].measuredX = 256;
  latPt[1].measuredY = 222;
  calcLattice(TRUE);

  contour= TRUE;
  contourToggle= FALSE;
  lopHipSwitch = FALSE;
  filterToggle= FALSE;
  drawLatToggle= FALSE;
  drawHiLoToggle= TRUE;
  layerLines= FALSE;
  drawCTF = FALSE;
  ctfEnhance = TRUE;
  bumpContrast = FALSE;
  dispCenters = TRUE;
  measureFFT = TRUE;
  filtSizeValue = 5.0;
  icut= 0;                      /* Start at Hard Cutoff */
  for(i=0;i<5;i++)
    latScrollValue[i]= 0.0;
  pixSize = 1.0;                /* FFT screen Pixel Size */
  fftAngsPerPixel = ((float)pixSize*10.0);
  lowResoFilter = 320000;       /* Ridiculously high */
  hiResoFilter  = 0;            /* Ridiculously low */
  latLineOrder = 1;             /* Start with something */
  latValChanged = FALSE;
  averageFFT = FALSE;
  fftZoom = FALSE;
  fftZoomFactor = 1;
  rasterStep = 1.00;
  ctfCurveNum = 0;
  maxCtfCurve = 1;
  applyCTF = FALSE;
  for(i=0;i<MAXCTFS;i++)
    {/* Init them all */
      ctfPts[i].latX0= 0;
      ctfPts[i].latY0= 0;
      ctfPts[i].latX1= 0;
      ctfPts[i].latY1= 0;
      ctfPts[i].pickedPts= FALSE;
    }/* Init them all */
  for(i=0;i<3;i++)
    { fftMeasures[i].x = 0; fftMeasures[i].y = 0;}
  fftMeasures[0].x = 256; fftMeasures[0].y = 256; /* Start in center */
  fftImage_gold = NULL;
  fftImage_fourAver= NULL;
  fftImage_buffer2= NULL;
  fftMotionFeedback = FALSE;
  pixSize = 0.357;
  fftAngsPerPixel = ((float)pixSize*10.0);
  scanStep = 14.0;               /* um */
  magnification = 39200.0;       /* X */
  wavelength = 0.0037;             /* nm-1 */
  spAberration = 2.0;            /* mm times 10^6 */
  numTimesApplyCTF= 0;

  /* ROFLT stuff */
  rotfltSwitch = FALSE;
  yFactor = 4.0;
  xFactor = 1.0;
  saveLinYFactor = 4;
  saveLogYFactor = 1;
  nrot = 5;
  lg_end = 100;
  nannuli = 30;
  logROTFLTplot = FALSE;
  fft_step = 1.0;
  phi0= omega= 0.0;
  jzero= 0;                     /* 0 means to include it - believe it */
  sump1= sump2= sump3= 0.0;
  backTransROTFLT= FALSE;
  stretchRotflt = TRUE;
  minRotfltScaleF= 999999999999.0;
  minRotfltScaleF= HUGE;  /* dgm HUGE instead of 999999999999.0 */
  
  /* crysDisp stuff */
  ncellx= 1;
  ncelly= 1;
  updateCrys = FALSE;
  crysDisping = FALSE;

  /* fourAver stuff */
  backSize = 2;
  fourAverRadius= 1.0;
  fourAverCut= 2.0;
  fourierAveragedCrysDisp = FALSE;
  refinePhaseOrigin= imposeSymmetry = 0;
  deltaPhaseShift = FALSE;
  
  /* unblob */
  unblobSwitch = FALSE;
  reverseSections= FALSE;
  negUniverseCut= FALSE;     /* For cut */
  cleanUnblob= FALSE;    /* For unblob */
  revRadProj = FALSE;
  icosRadProj = FALSE;
  radProjCS = FALSE;
  radialCS = FALSE;

  deltaCont = 0.0;
  numCont = 0;
  useCont = FALSE;
  breakCont = 0.0;

  contrastStretch = FALSE;
  fftMinMaxLock = FALSE;
  disp1024fft = FALSE;
  fftDispType = INTENSITY;       
  fftMeasLineNum = 11;

  /* Fix values stuff */
  fixMode= FIXOVER;
  fixIcosFill = FALSE;

  calcFFT1DcircAvg= FALSE;
  disp1Dca = FALSE;
  ctf1dCalcIt = TRUE;
  ctf1dSaveFftAvg = NULL;
  ctf2dCalcIt = TRUE;
  ctf2dSaveFftasum = NULL;
  ctf2dSaveFftasumsq = NULL;

  /* Autobox stuff */
  nThins = 1;
  nDilates = 0;
  prag = 1.5;
  refCutoff = 1.15 ;       /* Ratio cutoff to throw out particles */

  /* Param File stuff */
  dispCenters = TRUE;
  haveParamFile = FALSE;
  symmetry = 532;
  icoDispMode = UNITCELL;

  /* Alway set these to ON */
  drawIcos    = FALSE;
  useOmega    = FALSE;
  findNearest = FALSE;
  icosaMotioning = TRUE;
  drawIcos = TRUE;
  useOmega = TRUE;
  findNearest = TRUE;
  icosaMotioning = FALSE;

  icosdist = 20;
  startbox = 0.000;
  endbox = 1.000;
  yscale = 480;
  linelength = 480.0;
  step = (endbox-startbox)/480.0;
  pixelstep = (step*linelength)/(endbox-startbox);
  ico_toolsScrollValue = 0;
  newXY = FALSE;
  idtoggle= thetatoggle= phitoggle= omegatoggle= xtoggle= ytoggle=
    magtoggle= cc1toggle= cc2toggle= cc3toggle= distancetoggle=
      deltaXtoggle = deltaYtoggle= absDeltaXtoggle= absDeltaYtoggle= FALSE;
  mapProj = mapProjContour = FALSE;
  incSort = TRUE;
  volumeDisp = FALSE;
  taArrowX= taArrowY= taArrowZ= TRUE;
  deltaStdX = deltaStdY= 0.0;

  info.info = NULL;

  fortranBuffer = FALSE;   /* Have we allocated with fortran - we need
			    * to know this for VMS so we can use the
			    * free() in vms_only.f */

  pixelSizeFactor = 1.0;
  calcSFs   = FALSE;
  calcBin   = FALSE;
  truePhi   = TRUE;
  minMaxColor = forceAvgBkg =  forceByte = FALSE;
  dispIcosa = FALSE;
  pixelSizeSave = cutMin0Save = cutMax0Save = cutMin1Save =
      cutMax1Save = cutMin2Save = cutMax2Save = -99.0;
  icosaScaleF = 1.0;

  subtractingMap = FALSE;
  readEmcoravg = FALSE;
  readImgconv  = FALSE;

  doBinning = 0;
  doManipPP = 0;
  doCavgPP = 0;
  writeCavg = FALSE;
  cavgPPcenter = 0;
  totCavgPPrint = 0;

  subtract_image_buffer1 = subtract_image_buffer2 = 0;
  /* We've always got to have something allocated for this */
  CK(subtract_map_image_buffer   = (float *)malloc((2) * sizeof(float)));
  CK(subtract_map_image_buffer2  = (float *)malloc((2) * sizeof(float)));
  subMapYmxb_buffer = diffSection_buffer = 0;

  diffSection = TRUE;
  subCanDrawWindow = TRUE;
  findMagHist_toggle= FALSE;
  findRadHist_toggle= FALSE;
  drawingFindRad = FALSE;
  appodizing = FALSE;
  writeFFTType = FALSE;
  rdensit= wrdensit= ccoefit= FALSE;
  wccoefit= TRUE;
  s_wccoefit= FALSE;
  sprintf(RDoutStr,"");
  sprintf(WRDoutStr,"");
  sprintf(CCoutStr,"");
  sprintf(WCCoutStr,"");
  sprintf(S_WCCoutStr,"");
  flickerIt = 12;
  threshScrollValue = -2000.0;
  magSecScrollValue = 1.0;
  diffSecScrollValue = 1.0;
  diffSec2ScrollValue= 0.0;
  pointColor = 0;
  totParamFiles= 0;
  for(i=0;i<MAXPACKEDFILES;i++)
    batchListStr[i]= 0;
  readPackedFile = TRUE;
  subMap2d3d= 2;
  saveSecondMap = FALSE;
  icosaQuarter = FALSE;
  showRadii = TRUE;
  threshDoit = TRUE;
  threshAbs = FALSE;
  submapFindRad  = FALSE;
  whichIcos = ICOSNORMAL;
  destroyIcos = FALSE;
  writeFourAverSF = FALSE;

  /* Auto Defocus refine */
  doingAutoDefocusRefine=  adrCircAvgDraw=  adrSDraw=  adrSminusBDraw=
    adrCtfsqDraw=  adrDDraw=   adrBDraw=  adrBlankDraw= FALSE;
  ca1d = ca1d_std = NULL;
  adr_s= adr_sminusb= adr_ctfsq= adr_d= adr_b= NULL;
  ca1d_cnt = NULL;
  doingAuto2DDefocusRefine = FALSE;

  paintMapSwitch= FALSE;
  paintMap= NULL;

  calcLowResMap= FALSE;
  globAccum = (int *)NULL;

  fix2DinCore_buffer= NULL;
  fix2DinCore_diff= NULL;
  autoMinMaxThreshold = AMMCORRCOEF;
  refCircAvg  = FALSE;
  refAutomate = FALSE;
  newOrigin = TRUE;
  rotfltCrosshair = FALSE;
  rotfltResiduals = FALSE;
  CK(sigma= (float *)malloc(sizeof(float)*(22)*(22)));
  coravgRadius = 10.0;
  lastPlotClickX = lastPlotClickY = 50;  /* Just give it something */
  coravgLowLimValue = 0.0;
  useAvgCorAvg = FALSE;
  showLowLim = FALSE;
  drawFix2Dgrid = FALSE;
  drawFix2Dstats = FALSE;
  flickerFix2D = FALSE;
  firstTimeLoHiRes = TRUE;
  blindWriting = FALSE;
  normalExcise = TRUE;
  ctfFilter = 1;

  numExp= 0;                                    /* Debug exposures */

  drawAuto2Dctf = FALSE;
  drawAuto2Dscurve = FALSE;
  drawAuto2Dctflines = FALSE;
  auto2DdrawBuf = (float *)malloc(sizeof(float)*256*512);
  auto2DdataBuf = (float *)malloc(sizeof(float)*256*512);
  emptyBox = FALSE;
  a2drSDraw=a2drSminusBDraw=a2drCtfsqDraw=a2drDDraw=a2drBDraw = FALSE;
  a2dr_s= a2dr_sminusb= a2dr_ctfsq= a2dr_d= a2dr_b= NULL;
  fftasum = poly_paras = NULL;

  modelType = SQUARE;
  model_undo_buffer = 0;
  mbImmediate = FALSE;
  isIcosRepl  = FALSE;

  countIt = 0;
  firstAMMSTDEVtime = TRUE;
  showBoxingLevel = AMMCORRCOEF;
  threshLo= 0.0;
  threshHi= 99999999.9;


}/* INIT_GLOBALS */




/******************************************************************
 * groundZero() - Initialize these things when you do a major 
 * start over in the code.  For example, when you read a new file 
 * or etc...
*/
void groundZero()
{/* GROUNDZERO */
  int i;
  XmString simpleString;

  if(projASU)
    {/* But *do* get rid of any unwanted memory */
      for(i=0;i<purdueHeader.nimg;i++)
	{
	  if(packImg[i])
	    {free(packImg[i]);
	     packImg[i]= NULL;}
	}
      projASU = FALSE;
      purdueHeader.nimg = 0;
      lastPackedNumImgs= 0;
    }/* But *do* get rid of any unwanted memory */
  
  if(haveParamFile)
    {/* Get rid of any parameter file we've got */
      haveParamFile = FALSE;
      sprintf(inParamFilename,"");
      sprintf(paramBoxesFilename,"");
      XmTextSetString(paramFileTextW,inParamFilename); 
      purdueHeader.nimg= 0;
      lastPackedNumImgs= 0;
      Free_Particle_Info(&info);
    }/* Get rid of any parameter file we've got */

  /* Delete our last packed images */
  for(i=0;i<lastPackedNumImgs;i++)
    {/* Init these out */
      free(packImg[i]);
      packImg[i]= NULL;
    }/* Init thes out */
  lastPackedNumImgs= 0;
  for(i=0;i<MAXBOX;i++)
    {/* Init these out */
      packImg[i]= NULL;
    }/* Init thes out */

  for(i=0;i<numBoxes;i++)
    if(box[i].inMemory==TRUE)
      {/* Free it */
	free(box[i].editBox_buffer);
	box[i].editBox_buffer= NULL;
	box[i].inMemory= FALSE;
      }/* Free it */
  
  numBoxes= numRects= numPoly= 0;
  xoffset= yoffset= 0;
  rectBox.drawable= FALSE; 
  rectBox.width = rectBox.height= 0;
  fullScreen= FALSE;
  histZoomFactor = 20;
  noCallbackScroll= FALSE;
/*  Let's try not re-initing these ...
  zoomFactor = 0.0;
  oldZoomFactor = -99.0;
  XmTextSetString(zoomTextW, "1.00");   /+ We're back to normal zoom +/
*/
  oldSizeFactor = -99.0;
  calcSizeFactor(zoomFactor); 
  makingMovie= FALSE;
  curBox= 0;
  packRadius= 0;
  averageFFT = FALSE;
  numTimesApplyCTF= 0;
  phaseOrigX = 0.0;
  phaseOrigY = 0.0;

  /* Reset the FFT widgets */
  XtSetArg(arglst[0], XmNset, FALSE);
  XtSetValues(a128FFTWidget, arglst, 1); 
  XtSetValues(a256FFTWidget, arglst, 1); 
  XtSetValues(a1024FFTWidget, arglst, 1); 
  XtSetArg(arglst[0], XmNset, TRUE);
  XtSetValues(a512FFTWidget, arglst, 1); 
  fftSize = 512;

  if(imgInCore)
    free(inCore_buffer);
  imgInCore = FALSE;

  if(initLastNumBox)
    {/* Init it */
      lastNumBox = 0;
      sprintf(outStr,"%d",lastNumBox+1);
      XmTextSetString(boxInitWidget, outStr);
      XFlush(dsply);
    }/* Init it */

  /* Free up MAP stuff */
  for(i=0;i<MAXSECTIONS;i++)
    {/* Do all section */
      if(cuedMap[i].shad)
	free(cuedMap[i].shad);
      if(cuedMap[i].depth)
	free(cuedMap[i].depth);
      if(cuedMap[i].radial)
	free(cuedMap[i].radial);
      cuedMap[i].shad= cuedMap[i].depth= cuedMap[i].radial= NULL;
    }/* Do all section */

  XtUnmanageChild(mapShadeScreenWidget);
  XtUnmanageChild(circAvgScreenWidget);
  XtUnmanageChild(boxScreenWidget);

  if(rotMap)
    free(rotMap);
  tmpRotMap = rotMap = NULL;
  shadedMaps = FALSE;

  if(int_image_buffer1)
    free(int_image_buffer1);
  int_image_buffer1= NULL;

  if((float_image_buffer1) &&
     (float_image_buffer1 != lowRes_buffer1))
    free(float_image_buffer1);
  if(lowRes_buffer1)
    free(lowRes_buffer1);

  float_image_buffer1= NULL;
  lowRes_buffer1 = NULL;

/* DONT DO THIS!!!  We do this when we reallocate!!!
  free(  rwa_tmp2d_map );
  free(  rwa_tmp_map   );
  rwa_tmp2d_map = NULL;
  rwa_tmp_map   = NULL;
*/

  if(subtract_image_buffer2)
    free(subtract_image_buffer2);
  subtract_image_buffer2 = 0;
  if(subtract_image_buffer1)
    free(subtract_image_buffer1);
  subtract_image_buffer1 = 0;
  if(diffSection_buffer)
    free(diffSection_buffer);
  diffSection_buffer = 0;
  if(subMapYmxb_buffer)
    free(subMapYmxb_buffer);
  subMapYmxb_buffer = 0;

  /* Default menu choices */
  XtSetArg(arglst[0], XmNsensitive, TRUE);
  XtSetValues(truncaddButtonW, arglst, 1); 
  XtSetValues(removeBlemishButtonW, arglst, 1); 
  XtSetValues(editBoxButtonW, arglst, 1); 
  XtSetValues(pcxButtonW, arglst, 1); 
  XtSetValues(correlateButtonW, arglst, 1);  
  XtSetValues(correlateButton2W, arglst, 1);  
  XtSetValues(autoBox_widget, arglst, 1);  
  XtSetValues(rawButtonW, arglst, 1);   
  /*  XtSetValues(autoBox2_widget, arglst, 1);   */
  
  XtSetArg(arglst[0], XmNsensitive, FALSE);
  XtSetValues(paintMapButtonW, arglst, 1); 
  XtSetValues(mapMovieButtonW, arglst, 1); 
  /*  XtSetValues(circAvgButtonW, arglst, 1);  */
  XtSetValues(mapShadeButtonW, arglst, 1); 
  /*  XtSetValues(fixButtonW, arglst, 1);  */
  XtSetValues(invertButtonW, arglst, 1); 
  XtSetValues(rescaleButtonW, arglst, 1); 
  XtSetValues(calcSFButtonW, arglst, 1); 
  XtSetValues(modelWidget, arglst, 1); 
  XtSetValues(subtractMapButtonW, arglst, 1); 
  /* XtSetValues(RcalcSFButtonW, arglst, 1);  */
  XtSetValues(hardRotateButtonW, arglst, 1); 
  XtSetValues(rotateButtonW, arglst, 1); 
  
  fullScreen_buffer.dispx = fullScreen_buffer.dispy = 0;
  
  /* Sections or cued maps */
  dispSections= TRUE;
  /* Start without it */
  XtSetArg(arglst[0], XmNsensitive, FALSE);
  XtSetValues(sectionWidget, arglst, 1); 
  
  if(fftImage_gold)
    free(fftImage_gold);
  if(fftImage_fourAver)
    free(fftImage_fourAver);
  if(fftImage_buffer2)
    free(fftImage_buffer2);
  fftImage_gold = NULL;
  fftImage_fourAver= NULL;
  fftImage_buffer2= NULL;
  fft_screen_deactivate(0,0,0);
  rotflt_screen_deactivate(0,0,0);
  crysDisp_screen_deactivate(0,0,0);
  fourAver_deactivate(0,0,0);
  emcoravg_screen_deactivate(0,0,0);
  imgconv_screen_deactivate(0,0,0);
  mapShadeScreen_deactivate(0,0,0);
  editBox_screen_deactivate(0,0,0);
  autoDR_deactivate(0,0,0);
  auto2DDR_deactivate(0,0,0);
/* OBSOLETE   unblob_screen_deactivate(0,0,0); */

  XtSetArg(arglst[0], XmNsensitive, FALSE);
  XtSetValues(interpolateWidget, arglst, 1); 
  XtSetValues(binningWidget, arglst, 1); 
  XtSetValues(addAverageWidget, arglst, 1); 
  XtSetValues(fourierFiltWidget, arglst, 1); 
  XtSetValues(checkBkgWidget, arglst, 1); 
  XtSetValues(normalizeWidget, arglst, 1); 
  XtSetValues(normalizeBkgWidget, arglst, 1); 
  XtSetValues(linGradWidget, arglst, 1); 

  XmTextSetString(frameTextWidget, ""); 

  XFlush(dsply);

  /* Init the labels for autoboxing */
  sprintf(outStr,"     ");
  createSimpleString(&simpleString,outStr);
  XtSetArg(arglst[0], XmNlabelString, simpleString );
  XtSetArg(arglst[1], XmNrecomputeSize, FALSE );
  XtSetValues(autoBoxLabelW, arglst, 2); 
  XtSetValues(refCorrLabelW, arglst, 2); 
  XtSetValues(autoCorrLabelW, arglst, 2); 
  XFlush(dsply);
  XmStringFree(simpleString); 
  XtSetArg(arglst[0], XmNsensitive, FALSE);
  XtSetArg(arglst[1], XmNset, FALSE);
/*  XtSetValues(showccToggleW, arglst, 2);   */
  XFlush(dsply);
  showCC = FALSE;
  showDebug = TRUE;

  haveParamFile = FALSE;
  projASU       = FALSE;
  Free_Particle_Info(&info);
  /* Make them read the map file second */
  XmTextSetString(paramMapFileTextW,""); 
  XmTextSetString(paramScanFileTextW,""); 

  XFlush(dsply);

  /* Make them recalc any Hires buffers */
  pixelSizeSave = cutMin0Save = cutMax0Save = cutMin1Save =
      cutMax1Save = cutMin2Save = cutMax2Save = -99.0;

  subCanDrawWindow = TRUE;
  drawingFindRad = FALSE;

  doingAuto2DDefocusRefine= doingAutoDefocusRefine= FALSE;

  paintMapSwitch= FALSE;
  if(paintMap)
    free(paintMap);
  paintMap = NULL;

  ctf1dCalcIt = TRUE;
  if(ctf1dSaveFftAvg != NULL)
    free(ctf1dSaveFftAvg);
  ctf1dSaveFftAvg = NULL;
  ctf2dCalcIt = TRUE;
  if(ctf2dSaveFftasum != NULL)
    free(ctf2dSaveFftasum);
  ctf2dSaveFftasum = NULL;
  if(ctf2dSaveFftasumsq != NULL)
    free(ctf2dSaveFftasumsq);
  ctf2dSaveFftasumsq = NULL;

  firstGridding = TRUE;  /* Make a new grid first time empty boxing */

  /* reset these in case we did empties last time */
  refAutomate = FALSE;
  darkCircleRef =  FALSE;
  
  XtSetArg(arglst[0], XmNset, FALSE);
  XtSetValues(darkRefToggleW, arglst, 1);
  XtSetArg(arglst[0], XmNset, FALSE);
  XtSetValues(refAutomateW, arglst, 1);
  
}/* GROUNDZERO */




/* @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
 * XlibInitialize() - Initialize for Xlib calls.  Setup
 * the colormap we'll use and graphics contexts.  Create 
 * all of the Images and allocate memory for all of our
 * XmDrawingAreas.  Add event handlers for motion events
 * where needed and setup our cursors.
 */
void XlibInitialize()
{/* XLIBINITIALIZE */
  Window tWin;
  XWindowAttributes winAttrib;
  XSetWindowAttributes setWinAttrib;
  static const char dotted[8]= {4,4,4,4};

  /* Enable save unders and backing store */
  if((XDoesSaveUnders(XScreenOfDisplay(dsply, DefaultScreen(dsply)))) &&
     (XDoesBackingStore(XScreenOfDisplay(dsply, DefaultScreen(dsply)))))
    {
      attrs.save_under = TRUE;
      attrs.backing_store = WhenMapped;
      printf("Enabling 'Backing Store' and 'SaveUnder' features\n");
      XChangeWindowAttributes(dsply, XtWindow(drawingWidget), 
			      (CWSaveUnder | CWBackingStore), &attrs);
      XChangeWindowAttributes(dsply, XtWindow(fftDrawingWidget), 
			      (CWSaveUnder | CWBackingStore), &attrs);
      XChangeWindowAttributes(dsply, XtWindow(editBoxWidget), 
			      (CWSaveUnder | CWBackingStore), &attrs);
      XChangeWindowAttributes(dsply, XtWindow(icosaWidget), 
			      (CWSaveUnder | CWBackingStore), &attrs);
      XFlush(dsply);
      /* Other candidates for this:
	 image_window1= XtWindow(drawingWidget);
	 rotflt_window1= XtWindow(rotfltDrawWidget); 
	 fft_window1= XtWindow(fftDrawingWidget);
	 fftContour_window1= XtWindow(contourWidget);
	 circAvg_window1= XtWindow(circAvgWidget);
	 color_window1= XtWindow(colormapWidget);
	 test_color_window1= XtWindow(testcolormapWidget);
	 hist_color_window1= XtWindow(histWidget);
	 editBox_window1= XtWindow(editBoxWidget);
	 icosa_window1= XtWindow(icosaWidget);
      */
    }

  /* Init dimensions */
  buffer_width= image_width = 
    buffer_height= image_height = defDisplayBufferSize;
  image_colors = 3;


  /* Create Image buffers for our drawingAreas */
  /* MAIN drawing area on main screen */
  image_image1 = XCreateImage (dsply, visual, visual_depth, 
			       ZPixmap, 0, NULL, image_width, 
			       image_height, 32, 0);
  if (image_image1 == NULL) {
    fprintf (stderr, "XCreateImage (image) failed\n");
    exit(1);
  }
  
  /* Histogram area on colormap screen */
  hist_color_image1 = XCreateImage (dsply, visual, visual_depth, 
			       ZPixmap, 0, NULL, 32, 
			       256, 32, 0);
  if (hist_color_image1 == NULL) {
    fprintf (stderr, "XCreateImage (image) failed\n");
    exit(1);
  }
  
  /* Test color area on colormap screen */
  test_color_image1 = XCreateImage (dsply, visual, visual_depth, 
			       ZPixmap, 0, NULL, 32, 
			       32, 32, 0);
  if (test_color_image1 == NULL) {
    fprintf (stderr, "XCreateImage (image) failed\n");
    exit(1);
  }
  
  /* Colormap display area on colormap screen */
  color_image1 = XCreateImage (dsply, visual, visual_depth, 
			       ZPixmap, 0, NULL, 32, 
			       256, 32, 0);
  if (color_image1 == NULL) {
    fprintf (stderr, "XCreateImage (image) failed\n");
    exit(1);
  }
  
  /* ROTFLT display area  */
  rotflt_image1 = XCreateImage (dsply, visual, visual_depth, 
			       ZPixmap, 0, NULL, 512, 
			       512, 32, 0);
  if (rotflt_image1 == NULL) {
    fprintf (stderr, "XCreateImage (image) failed\n");
    exit(1);
  }

  /* FFT display area  */
  fft_image1 = XCreateImage (dsply, visual, visual_depth, 
			       ZPixmap, 0, NULL, 512, 
			       512, 32, 0);
  if (fft_image1 == NULL) {
    fprintf (stderr, "XCreateImage (image) failed\n");
    exit(1);
  }
  fftContour_image1 = XCreateImage (dsply, visual, visual_depth, 
			       ZPixmap, 0, NULL, 128, 
			       128, 32, 0);
  if (fftContour_image1 == NULL) {
    fprintf (stderr, "XCreateImage (image) failed\n");
    exit(1);
  }
  
  /* EMCORAVG display area  */
  emcoravg_image1 = XCreateImage (dsply, visual, visual_depth, 
				  ZPixmap, 0, NULL, 256, 
				  256, 32, 0);
  if (emcoravg_image1 == NULL) {
    fprintf (stderr, "XCreateImage (image) failed\n");
    exit(1);
  }
  
  /* Circular Average display area  */
  circAvg_image1 = XCreateImage (dsply, visual, visual_depth, 
			       ZPixmap, 0, NULL, 256, 
			       256, 32, 0);
  if (circAvg_image1 == NULL) {
    fprintf (stderr, "XCreateImage (image) failed\n");
    exit(1);
  }
  
  /* Area on the edit box screen for viewing the boxes */
  editBox_image1 = XCreateImage (dsply, visual, visual_depth, 
			       ZPixmap, 0, NULL, BOX_WIDTH, 
			       BOX_HEIGHT, 32, 0);
  if (editBox_image1 == NULL) {
    fprintf (stderr, "XCreateImage (image) failed\n");
    exit(1);
  }
  
  /* ICOSA display area  */
  icosa_image1 = XCreateImage (dsply, visual, visual_depth, 
			       ZPixmap, 0, NULL, 512, 
			       512, 32, 0);
  if (icosa_image1 == NULL) {
    fprintf (stderr, "XCreateImage (image) failed\n");
    exit(1);
  }

  /* Get Window pointers from MOTIF */
  image_window1= XtWindow(drawingWidget);
  rotflt_window1= XtWindow(rotfltDrawWidget); 
  fft_window1= XtWindow(fftDrawingWidget);
  fftContour_window1= XtWindow(contourWidget);
  emcoravg_window1= XtWindow(emcoravgDrawWidget);
  circAvg_window1= XtWindow(circAvgWidget);
  color_window1= XtWindow(colormapWidget);
  test_color_window1= XtWindow(testcolormapWidget);
  hist_color_window1= XtWindow(histWidget);
  editBox_window1= XtWindow(editBoxWidget);
  icosa_window1= XtWindow(icosaWidget);

  
  /* Now tell MOTIF what kindof events you'll listen to in this
   * window.
   */
   event_mask  = ButtonPressMask;
   event_mask |= ButtonReleaseMask   | KeyPressMask | KeyReleaseMask;
   event_mask |= PointerMotionMask   | LeaveWindowMask | ExposureMask;
   XSelectInput(dsply, image_window1, event_mask);
   XSelectInput(dsply, emcoravg_window1, event_mask);
   XSelectInput(dsply, circAvg_window1, event_mask);
   XSelectInput(dsply, rotflt_window1, event_mask); 
   XSelectInput(dsply, fft_window1, event_mask);
   XSelectInput(dsply, color_window1, event_mask);
   XSelectInput(dsply, test_color_window1, event_mask);
   XSelectInput(dsply, hist_color_window1, event_mask);
   XSelectInput(dsply, editBox_window1, event_mask);
   XSelectInput(dsply, icosa_window1, event_mask);
     
  /* BYTE mirroring buffers */
  CK(mirror_image_buffer1 = (BYTE *)malloc ((sizeof(BYTE))));
                                       /* Just give it something for now */
  CK(editBox_undo_buffer = 
    (float *)malloc ((sizeof(float)*(BOX_WIDTH*BOX_HEIGHT))));
  
  CK(editBox_mirror_buffer = 
    (float *)malloc ((sizeof(float)*(BOX_WIDTH*BOX_HEIGHT))));

  for(i=0;i<(BOX_WIDTH*BOX_HEIGHT);i++)
    {
      editBox_undo_buffer[i] = 
	editBox_mirror_buffer[i] = 0.0;
    }

  /* Allocate for our image storage for XWindows buffers
   * Use *4 to account for 32 bit depth if needed.
   */
  if(hiColor)
    CK(image_buffer1_disp = (BYTE *)malloc ((buffer_width*buffer_height)*4*
				       (sizeof(BYTE))));
  CK(image_buffer1 = (BYTE *)malloc ((buffer_width*buffer_height)*4*
				     (sizeof(BYTE))));
  CK(emcoravg_buffer1 = (BYTE *)malloc ((sizeof(BYTE)*4*(256*256))));
  CK(circAvg_buffer1 = (BYTE *)malloc ((sizeof(BYTE)*4*(256*256))));
  CK(rotflt_buffer1 = (BYTE *)malloc ((sizeof(BYTE)*4*(512*512))));
  CK(fft_display_buffer1 = (BYTE *)malloc ((sizeof(BYTE)*4*(512*512))));
  CK(fft_buffer1 = (BYTE *)malloc ((sizeof(BYTE)*4*(512*512))));
  CK(fft_buffer2 = (BYTE *)malloc ((sizeof(BYTE)*4*(512*512))));
  CK(fft_gold_buffer1 = (BYTE *)malloc ((sizeof(BYTE)*4*(512*512))));
  CK(fft_contour_buffer = (BYTE *)malloc ((sizeof(BYTE)*4*(128*128))));
  CK(fft_contour_buffer2 = (BYTE *)malloc ((sizeof(BYTE)*4*(128*128))));
  CK(color_buffer1 = (BYTE *)malloc ((sizeof(BYTE)*4*(32*256))));
  CK(test_color_buffer1 = (BYTE *)malloc ((sizeof(BYTE)*4*(32*32))));
  CK(hist_color_buffer1 = (BYTE *)malloc ((sizeof(BYTE)*4*(32*256))));
  CK(editBox_int_buffer = 
    (float *)malloc ((sizeof(float)*4*(BOX_WIDTH*BOX_HEIGHT))));
  CK(editBox_display_buffer = 
    (BYTE *)malloc ((sizeof(BYTE)*4*(BOX_WIDTH*BOX_HEIGHT))));
  CK(editBox_display_buffer2 = 
    (BYTE *)malloc ((sizeof(BYTE)*4*(BOX_WIDTH*BOX_HEIGHT))));
  CK(icosa_buffer1 = (BYTE *)malloc ((sizeof(BYTE)*4*(512*512))));


  /* Couldn't get aholt of enough storage? */
  if (image_buffer1 == NULL) {
    perror ("malloc image_buffer");
    exit(1);
  }
  if (color_buffer1 == NULL) {
    perror ("malloc color_buffer");
    exit(1);
  }
  
  /* OK tell the image where to find it's data */
  image_image1->data = (char *)image_buffer1;
  rotflt_image1->data = (char *)rotflt_buffer1;
  if(hiColor)
    fft_image1->data = (char *)fft_display_buffer1;
  else
    fft_image1->data = (char *)fft_buffer1;
  if(hiColor)
    fftContour_image1->data = (char *)fft_contour_buffer2;
  else
    fftContour_image1->data = (char *)fft_contour_buffer;
  emcoravg_image1->data = (char *)emcoravg_buffer1;
  circAvg_image1->data = (char *)circAvg_buffer1;
  color_image1->data = (char *)color_buffer1;
  hist_color_image1->data = (char *)hist_color_buffer1;
  test_color_image1->data = (char *)test_color_buffer1;
  if(hiColor)
    editBox_image1->data = (char *)editBox_display_buffer2;
  else
    editBox_image1->data = (char *)editBox_display_buffer;
  icosa_image1->data = (char *)icosa_buffer1;
  
  /* Init up the main drawing Area to BLACK.  */
  for(i=0;i<buffer_height;i++)
    for(j=0;j<buffer_width;j++)
      {
	if(hiColor)
	  XPutPixel(image_image1, j, i, pixels_return[BLACK]); 
	else
	  image_buffer1[((i*buffer_width)+j)] = pixels_return[BLACK]; 
      }

  /* Historic Christmas image used as default image of robem - may
   * be a good place to implement a logo if we ever get that far.
   */
  christmasTime = FALSE;

  /* This is an optimization thing - just blow this out to the screen
   * numRows times whenever you need to make sure there's no unwanted
   * pixels showing.
   */
  for(j=0;j<NUMCOLS;j++)
    stufBuf[j]= pixels_return[BLACK];
  for(j=0;j<NUMCOLS;j++)  
    stufIntBuf[j]= pixels_return[BLACK];  
                                                 /* These are shorts */
  for(j=0;j<NUMCOLS;j++)
    stufBLKBuf[j]= 0;
  for(j=0;j<NUMCOLS;j++)  
    stufBLKIntBuf[j]= (unsigned short)0;  
                                                 /* These are shorts */

  /* Init up the colormap display drawingArea */
  for(i=0;i<256;i++)
    for(j=0;j<32;j++)
      color_buffer1[((i*32)+j)] = (255-i);

  /* Init up the TEST color drawing Area - we just show them this
   * placeholder value that's "mixed" by the scrolls beside it (RGB)
   */
  for(i=0;i<32;i++)
    for(j=0;j<32;j++)
      test_color_buffer1[((i*32)+j)] = pixels_return[PLACEHOLDER];
  
  /* Init up the rotflt drawing Area to BLACK.  */
  for(i=0;i<512;i++)
    for(j=0;j<512;j++)
      rotflt_buffer1[((i*512)+j)] = pixels_return[BLACK];

  /* Init up the fft drawing Area to BLACK.  */
  for(i=0;i<512;i++)
    for(j=0;j<512;j++)
      {
	if(hiColor)
	  fft_gold_buffer1[((i*512)+j)] = 
	    fft_buffer1[((i*512)+j)] = BLACK;
	else
	  fft_gold_buffer1[((i*512)+j)] = 
	    fft_buffer1[((i*512)+j)] = pixels_return[BLACK];
      }
  for(i=0;i<128;i++)
    for(j=0;j<128;j++)
      {
	if(hiColor)
	  fft_contour_buffer[((i*128)+j)] = BLACK;
	else
	  fft_contour_buffer[((i*128)+j)] = pixels_return[BLACK];
      }
  
  /* Init up the emcoravg drawing Area to BLACK.  */
  for(i=0;i<256;i++)
    for(j=0;j<256;j++)
      emcoravg_buffer1[((i*256)+j)] = pixels_return[BLACK];
  
  /* Init up the circAvg drawing Area to BLACK.  */
  for(i=0;i<256;i++)
    for(j=0;j<256;j++)
      circAvg_buffer1[((i*256)+j)] = pixels_return[BLACK];
  
  /* Init up the icosa drawing Area to GREY.  */
  for(i=0;i<512;i++)
    for(j=0;j<512;j++)
      icosa_buffer1[((i*512)+j)] = pixels_return[BLACK];

  /* Setup Graphics contexts */
  if ((fontstruct = XLoadQueryFont(dsply, "9x15bold")) == NULL) 
    {/* Try generic fixed */
      fprintf(stderr, "%s: display %s doesn't know font %s\n",
	      "robem", "DisplayString(dsply)", "9x15bold");
      if ((fontstruct = XLoadQueryFont(dsply, FONT)) == NULL) {
	fprintf(stderr, "%s: display %s doesn't know font %s\n",
		"robem", "DisplayString(dsply)", FONT);
	exit(1);
      }
    }/* Try generic fixed */
  gc_mask = 0;
  gc_values.function = GXcopy;		gc_mask |= GCFunction;
  gc_values.plane_mask = AllPlanes;	gc_mask |= GCPlaneMask;
  /*gc_values.foreground = white_pixel;	gc_mask |= GCForeground;*/
  /*gc_values.background = black_pixel;	gc_mask |= GCBackground;*/
  gc_values.foreground = 1;	        gc_mask |= GCForeground;
  gc_values.background = black_pixel;	gc_mask |= GCBackground;
  gc_values.graphics_exposures = False;	gc_mask |= GCGraphicsExposures;
  /* gc_values.font = fontstruct->fid;     gc_mask |= GCFont;  */

  image_gc = XCreateGC (dsply, image_window1, gc_mask, &gc_values);
  color_gc = XCreateGC (dsply, color_window1, gc_mask, &gc_values);
  rotflt_gc = XCreateGC (dsply, rotflt_window1, gc_mask, &gc_values);
  fft_gc = XCreateGC (dsply, fft_window1, gc_mask, &gc_values);
  fftContour_gc = XCreateGC (dsply, fftContour_window1, gc_mask, &gc_values);
  emcoravg_gc = XCreateGC (dsply, emcoravg_window1, gc_mask, &gc_values);
  circAvg_gc = XCreateGC (dsply, circAvg_window1, gc_mask, &gc_values);
  hist_color_gc = XCreateGC (dsply, hist_color_window1, gc_mask, &gc_values);
  test_color_gc = XCreateGC (dsply, test_color_window1, gc_mask, &gc_values);
  editBox_gc = XCreateGC (dsply, editBox_window1, gc_mask, &gc_values);
  icosa_gc = XCreateGC (dsply, icosa_window1, gc_mask, &gc_values);
  
  gc_values.foreground = pixels_return[WHITE];  gc_mask |= GCForeground;
  image_gcWHITE = XCreateGC (dsply, image_window1, gc_mask, &gc_values);
  
  gc_values.foreground = pixels_return[RED];  gc_mask |= GCForeground;
  image_gcRED = XCreateGC (dsply, image_window1, gc_mask, &gc_values);
  image_rotfltgcRED = XCreateGC (dsply, rotflt_window1, gc_mask, &gc_values);
  
  gc_values.foreground = pixels_return[YELLOW];  gc_mask |= GCForeground;
  image_gcYELLOW = XCreateGC (dsply, image_window1, gc_mask, &gc_values);
  
  gc_values.foreground = pixels_return[GREEN];  gc_mask |= GCForeground;
  image_gcGREEN = XCreateGC (dsply, image_window1, gc_mask, &gc_values);
  
  gc_values.foreground = pixels_return[BLUE];  gc_mask |= GCForeground;
  image_gcBLUE = XCreateGC (dsply, image_window1, gc_mask, &gc_values);
  
  gc_values.foreground = pixels_return[BLACK];  gc_mask |= GCForeground;
  image_gcBLACK = XCreateGC (dsply, image_window1, gc_mask, &gc_values);
  
  gc_values.foreground = pixels_return[PLACEHOLDER];  gc_mask |= GCForeground;
  image_gcPLACEHOLDER = XCreateGC (dsply, image_window1, gc_mask, &gc_values);
  
  /* Now setup for dashed lines */
  gc_values.foreground = pixels_return[RED];	gc_mask |= GCForeground;
  gc_values.line_style = LineOnOffDash; gc_mask |= GCLineStyle;
  image_gcDOTred = XCreateGC (dsply, image_window1, gc_mask, &gc_values);
  XSetDashes(dsply, image_gcDOTred, 0, dotted, 4);
  gc_values.foreground = pixels_return[GREEN];	gc_mask |= GCForeground;
  gc_values.line_style = LineOnOffDash; gc_mask |= GCLineStyle;
  image_gcDOTgreen = XCreateGC (dsply, image_window1, gc_mask, &gc_values);
  XSetDashes(dsply, image_gcDOTgreen, 0, dotted, 4);
  gc_values.foreground = pixels_return[BLUE];	gc_mask |= GCForeground;
  gc_values.line_style = LineOnOffDash; gc_mask |= GCLineStyle;
  image_gcDOT = XCreateGC (dsply, image_window1, gc_mask, &gc_values);
  XSetDashes(dsply, image_gcDOT, 0, dotted, 4);

  for(i=0;i<256;i++)
    {/* Make all of our GREY colors */
      gc_mask = 0;
      gc_values.function = GXcopy;		gc_mask |= GCFunction;
      gc_values.plane_mask = AllPlanes;	gc_mask |= GCPlaneMask;
      if(i<tryColors)
	gc_values.foreground = pixels_return[i];     
      else
	gc_values.foreground = pixels_return[tryColors-1]; 
      gc_mask |= GCForeground;
      gc_values.background = black_pixel;	gc_mask |= GCBackground;
      gc_values.graphics_exposures = False;	gc_mask |= GCGraphicsExposures;
      gc_values.font = fontstruct->fid;     gc_mask |= GCFont; 

      image_gcGREY[i] = XCreateGC (dsply, fft_window1, gc_mask, &gc_values);
      
    }/* Make all of our GREY colors */
  image_gcLAT = image_gcGREY[100];

  /* Display the image in the window created above */
  image_width = DEF_DISPLAY_SIZE;   /* This changes whenever a new image  */
  image_height = DEF_DISPLAY_SIZE;  /* is read                            */

  tWin = (Window)XtWindow(drawingWidget);
  /* Get new window dimensions */
  if(tWin > 0)
    {/* Valid window */
      XGetWindowAttributes(dsply,tWin,&winAttrib);
      display_width = (int)(winAttrib.width);
      display_height = (int)(winAttrib.height);
    }/* Valid window */

  /* Let's also look for motion events in these windows */
  XtAddEventHandler(rotfltDrawWidget, 
		    /* ButtonMotionMask,  */
		    PointerMotionMask, 
		    FALSE, (XtEventHandler)rotfltMotion, NULL);
  XtAddEventHandler(fftDrawingWidget, 
		    /* ButtonMotionMask,  */
		    PointerMotionMask, 
		    FALSE, (XtEventHandler)fftMotion, NULL);
  XtAddEventHandler(drawingWidget, 
		    /* ButtonMotionMask,  */
		    PointerMotionMask, 
		    FALSE, (XtEventHandler)plotMotion, NULL);
  XtAddEventHandler(circAvgWidget, 
		    /* ButtonMotionMask,  */
		    LeaveWindowMask, 
		    FALSE, (XtEventHandler)circAvgMotion, NULL);
  XtAddEventHandler(colormapWidget, 
		    /* ButtonMotionMask,  */
		    LeaveWindowMask, 
		    FALSE, (XtEventHandler)colormapMotion, NULL);
  XtAddEventHandler(histWidget, 
		    /* ButtonMotionMask,  */
		    LeaveWindowMask, 
		    FALSE, (XtEventHandler)colormapMotion, NULL);
  XtAddEventHandler(circAvgWidget, 
		    /* ButtonMotionMask,  */
		    PointerMotionMask, 
		    FALSE, (XtEventHandler)circAvgMotion, NULL);
  XtAddEventHandler(colormapWidget, 
		    /* ButtonMotionMask,  */
		    PointerMotionMask, 
		    FALSE, (XtEventHandler)colormapMotion, NULL);
  XtAddEventHandler(histWidget, 
		    /* ButtonMotionMask,  */
		    PointerMotionMask, 
		    FALSE, (XtEventHandler)colormapMotion, NULL);
  XtAddEventHandler(editBoxWidget, 
		    /* ButtonMotionMask,  */
		    PointerMotionMask, 
		    FALSE, (XtEventHandler)editBoxMotion, NULL);
  XtAddEventHandler(icosaWidget, 
		    /* ButtonMotionMask,  */
		    PointerMotionMask, 
		    FALSE, (XtEventHandler)icosaMotion, NULL);
  XFlush (dsply);
  
  
  /* ****************************************************************
   * OK let's setup for our customized icon images for some of our
   * pushbutton widgets 
   * ****************************************************************/

  /* LANTERN cursors and pixmaps */
  shape_pixmap= XCreateBitmapFromData(dsply,
				      image_window1,
				      lantern_bits,
				      lantern_width,
				      lantern_height);
  mask_pixmap = XCreateBitmapFromData(dsply,
				      image_window1,
				      lanternmask_bits,
				      lanternmask_width,
				      lanternmask_height);
  lanternPixmap = XCreatePixmapFromBitmapData(dsply,
					  image_window1,
					  lantern_bits,
					  lantern_width,
					  lantern_height,
					  pixels_return[BLACK],
					  pixels_return[WHITE],
					  visual_depth);
  if(hiColor)
    lanternCursor   = XCreatePixmapCursor(dsply,
					  shape_pixmap,
					  mask_pixmap,
					  &color_defs[BLACK],
					  &color_defs[WHITE],
					  lantern_x_hot,
					  lantern_y_hot);
  else
    lanternCursor   = XCreatePixmapCursor(dsply,
					  shape_pixmap,
					  mask_pixmap,
					  &color_defs[pixels_return[BLACK]],
					  &color_defs[pixels_return[WHITE]],
					  lantern_x_hot,
					  lantern_y_hot);

  
  /* ZOOM cursors and pixmaps */
  shape_pixmap= XCreateBitmapFromData(dsply,
				      image_window1,
				      zoom_bits,
				      zoom_width,
				      zoom_height);
  mask_pixmap = XCreateBitmapFromData(dsply,
				      image_window1,
				      zoommask_bits,
				      zoommask_width,
				      zoommask_height);
  zoomPixmap = XCreatePixmapFromBitmapData(dsply,
					  image_window1,
					  zoom_bits,
					  zoom_width,
					  zoom_height,
					  pixels_return[BLACK],
					  pixels_return[RED],
					  visual_depth);
  if(hiColor)
    zoomCursor   = XCreatePixmapCursor(dsply,
				       shape_pixmap,
				       mask_pixmap,
				       &color_defs[BLACK],
				       &color_defs[RED],
				       zoom_x_hot,
				       zoom_y_hot);
  else
    zoomCursor   = XCreatePixmapCursor(dsply,
				       shape_pixmap,
				       mask_pixmap,
				       &color_defs[pixels_return[BLACK]],
				       &color_defs[pixels_return[RED]],
				       zoom_x_hot,
				       zoom_y_hot);

  
  /* Unlocked/locked pixmaps */
  lockedPixmap = XCreatePixmapFromBitmapData(dsply,
					     XtWindow(XtParent(lockWidget)),
					     locked_bits,
					     locked_width,
					     locked_height,
					     pixels_return[BLACK],
					     pixels_return[WHITE],
					     visual_depth);
  unlockedPixmap = XCreatePixmapFromBitmapData(dsply,
					       XtWindow(XtParent(lockWidget)),
					       unlocked_bits,
					       unlocked_width,
					       unlocked_height,
					       pixels_return[BLACK],
					       pixels_return[WHITE],
					       visual_depth);
  XtSetArg(arglst[0], XmNlabelPixmap, unlockedPixmap);
  XtSetArg(arglst[1], XmNselectPixmap, lockedPixmap);
  XtSetArg(arglst[2], XmNrecomputeSize, FALSE);
  XtSetValues(lockWidget, arglst, 3); 
  

  /* Play/Stop pixmaps */
  playPixmap = XCreatePixmapFromBitmapData(dsply,
				     XtWindow(XtParent(playFormWidget)),
					     play_bits,
					     play_width,
					     play_height,
					     pixels_return[BLACK],
					     pixels_return[WHITE],
					     visual_depth);
  playingPixmap = XCreatePixmapFromBitmapData(dsply,
				     XtWindow(XtParent(playFormWidget)),
					     playing_bits,
					     playing_width,
					     playing_height,
					     pixels_return[BLACK],
					     pixels_return[WHITE],
					     visual_depth);
  infiniPixmap = XCreatePixmapFromBitmapData(dsply,
				       XtWindow(XtParent(playFormWidget)),
					       infini_bits,
					       infini_width,
					       infini_height,
					       pixels_return[BLACK],
					       pixels_return[WHITE],
					       visual_depth);
  noinfiniPixmap = XCreatePixmapFromBitmapData(dsply,
				       XtWindow(XtParent(playFormWidget)),
					       noinfini_bits,
					       noinfini_width,
					       noinfini_height,
					       pixels_return[BLACK],
					       pixels_return[WHITE],
					       visual_depth);
  flashLightPixmap = XCreatePixmapFromBitmapData(dsply,
				       XtWindow(XtParent(playFormWidget)),
					       doflashlight_bits,
					       doflashlight_width,
					       doflashlight_height,
					       pixels_return[BLACK],
					       pixels_return[WHITE],
					       visual_depth);
  noflashLightPixmap = XCreatePixmapFromBitmapData(dsply,
				       XtWindow(XtParent(playFormWidget)),
					       noflashlight_bits,
					       noflashlight_width,
					       noflashlight_height,
					       pixels_return[BLACK],
					       pixels_return[WHITE],
					       visual_depth);
  rockPixmap = XCreatePixmapFromBitmapData(dsply,
				       XtWindow(XtParent(playFormWidget)),
					       dorock_bits,
					       dorock_width,
					       dorock_height,
					       pixels_return[BLACK],
					       pixels_return[WHITE],
					       visual_depth);
  norockPixmap = XCreatePixmapFromBitmapData(dsply,
				       XtWindow(XtParent(playFormWidget)),
					       norock_bits,
					       norock_width,
					       norock_height,
					       pixels_return[BLACK],
					       pixels_return[WHITE],
					       visual_depth);
  stopPixmap = XCreatePixmapFromBitmapData(dsply,
				       XtWindow(XtParent(playFormWidget)),
					       stop_bits,
					       stop_width,
					       stop_height,
					       pixels_return[RED],
					       pixels_return[WHITE],
					       visual_depth);
  stoppedPixmap = XCreatePixmapFromBitmapData(dsply,
				       XtWindow(XtParent(playFormWidget)),
					       stopped_bits,
					       stopped_width,
					       stopped_height,
					       pixels_return[RED],
					       pixels_return[WHITE],
					       visual_depth);
  goPixmap = XCreatePixmapFromBitmapData(dsply,
				       XtWindow(XtParent(playFormWidget)),
					       go_bits,
					       go_width,
					       go_height,
					       pixels_return[GREEN],
					       pixels_return[BLACK],
					       visual_depth);
  nogoPixmap = XCreatePixmapFromBitmapData(dsply,
				       XtWindow(XtParent(playFormWidget)),
					       nogo_bits,
					       nogo_width,
					       nogo_height,
					       pixels_return[GREEN],
					       pixels_return[BLACK],
					       visual_depth);
  helpPixmap = XCreatePixmapFromBitmapData(dsply,
				       XtWindow(XtParent(playFormWidget)),
					       help_bits,
					       help_width,
					       help_height,
					       pixels_return[BLACK],
					       pixels_return[WHITE],
					       visual_depth);
  nohelpPixmap = XCreatePixmapFromBitmapData(dsply,
				       XtWindow(XtParent(playFormWidget)),
					       nohelp_bits,
					       nohelp_width,
					       nohelp_height,
					       pixels_return[BLACK],
					       pixels_return[WHITE],
					       visual_depth);
  exitPixmap = XCreatePixmapFromBitmapData(dsply,
				       XtWindow(XtParent(playFormWidget)),
					       exit_bits,
					       exit_width,
					       exit_height,
					       pixels_return[GREEN],
					       pixels_return[BLACK],
					       visual_depth);
  noexitPixmap = XCreatePixmapFromBitmapData(dsply,
				       XtWindow(XtParent(playFormWidget)),
					       noexit_bits,
					       noexit_width,
					       noexit_height,
					       pixels_return[GREEN],
					       pixels_return[BLACK],
					       visual_depth);
  moviePixmap = XCreatePixmapFromBitmapData(dsply,
				       XtWindow(XtParent(surfDensToggleWidget)),
					       movie_bits,
					       movie_width,
					       movie_height,
					       pixels_return[BLACK],
					       pixels_return[WHITE],
					       visual_depth);
  unmoviePixmap = XCreatePixmapFromBitmapData(dsply,
				       XtWindow(XtParent(surfDensToggleWidget)),
					       unmovie_bits,
					       unmovie_width,
					       unmovie_height,
					       pixels_return[BLACK],
					       pixels_return[WHITE],
					       visual_depth);
  slowPixmap = XCreatePixmapFromBitmapData(dsply,
				       XtWindow(XtParent(playFormWidget)),
					       slow_bits,
					       slow_width,
					       slow_height,
					       pixels_return[RED],
					       pixels_return[BLACK],
					       visual_depth);
  noslowPixmap = XCreatePixmapFromBitmapData(dsply,
				       XtWindow(XtParent(playFormWidget)),
					       noslow_bits,
					       noslow_width,
					       noslow_height,
					       pixels_return[RED],
					       pixels_return[BLACK],
					       visual_depth);

  /* Unreset/reset pixmaps */
  resetPixmap = XCreatePixmapFromBitmapData(dsply,
					     XtWindow(XtParent(latResetButW)),
					     reset_bits,
					     reset_width,
					     reset_height,
					     pixels_return[BLACK],
					     pixels_return[WHITE],
					     visual_depth);
  noresetPixmap = XCreatePixmapFromBitmapData(dsply,
					       XtWindow(
						    XtParent(latResetButW)),
					       noreset_bits,
					       noreset_width,
					       noreset_height,
					       pixels_return[BLACK],
					       pixels_return[WHITE],
					       visual_depth);

  /* Unsection/section pixmaps */
  sectionPixmap = XCreatePixmapFromBitmapData(dsply,
					     XtWindow(XtParent(sectionWidget)),
					     section_bits,
					     section_width,
					     section_height,
					     pixels_return[BLACK],
					     pixels_return[WHITE],
					     visual_depth);
  unsectionPixmap = XCreatePixmapFromBitmapData(dsply,
					       XtWindow(
						    XtParent(sectionWidget)),
					       unsection_bits,
					       unsection_width,
					       unsection_height,
					       pixels_return[BLACK],
					       pixels_return[WHITE],
					       visual_depth);

  /* Uncolor/color pixmaps */
  colorPixmap = XCreatePixmapFromBitmapData(dsply,
					     XtWindow(
					       XtParent(colorToolWidget)),
					     color_bits,
					     color_width,
					     color_height,
					     pixels_return[BLACK],
					     pixels_return[WHITE],
					     visual_depth);
  uncolorPixmap = XCreatePixmapFromBitmapData(dsply,
					       XtWindow(
						XtParent(colorToolWidget)),
					       uncolor_bits,
					       uncolor_width,
					       uncolor_height,
					       pixels_return[BLACK],
					       pixels_return[WHITE],
					       visual_depth);

  XtSetArg(arglst[0], XmNlabelPixmap, colorPixmap);
  XtSetArg(arglst[1], XmNselectPixmap, uncolorPixmap);
  XtSetArg(arglst[2], XmNrecomputeSize, FALSE);
  XtSetValues(colorToolWidget, arglst, 3); 
  XtSetArg(arglst[0], XmNlabelPixmap, sectionPixmap);
  XtSetArg(arglst[1], XmNselectPixmap, unsectionPixmap);
  XtSetArg(arglst[2], XmNrecomputeSize, FALSE);
  XtSetValues(sectionWidget, arglst, 3); 
  XtSetArg(arglst[0], XmNlabelPixmap, playPixmap);
  XtSetArg(arglst[1], XmNselectPixmap, playingPixmap);
  XtSetArg(arglst[2], XmNrecomputeSize, FALSE);
  XtSetValues(playWidget, arglst, 3); 
  XtSetArg(arglst[0], XmNlabelPixmap, stopPixmap);
  XtSetArg(arglst[1], XmNselectPixmap, stoppedPixmap);
  XtSetArg(arglst[2], XmNrecomputeSize, FALSE);
  XtSetValues(stopWidget, arglst, 3); 
  XtSetArg(arglst[0], XmNlabelPixmap, stopPixmap);
  XtSetArg(arglst[1], XmNselectPixmap, stoppedPixmap);
  XtSetArg(arglst[2], XmNrecomputeSize, FALSE);
  XtSetValues(mapShadeStopWidget, arglst, 3); 
  XtSetArg(arglst[0], XmNlabelPixmap, stopPixmap);
  XtSetArg(arglst[1], XmNselectPixmap, stoppedPixmap);
  XtSetArg(arglst[2], XmNrecomputeSize, FALSE);
  XtSetValues(boxStopWidget, arglst, 3); 
  XtSetValues(versionStopW, arglst, 3); 
  XtSetValues(statusStopW, arglst, 3); 
  XtSetValues(fftStopW, arglst, 3); 

  XtSetArg(arglst[0], XmNlabelPixmap, goPixmap);
  XtSetArg(arglst[1], XmNselectPixmap, nogoPixmap);
  XtSetArg(arglst[2], XmNrecomputeSize, FALSE);
  XtSetValues(flickArrowUpWidget, arglst, 3); 
  XtSetValues(movieArrowUpWidget, arglst, 3); 
  XtSetArg(arglst[0], XmNlabelPixmap, slowPixmap);
  XtSetArg(arglst[1], XmNselectPixmap, noslowPixmap);
  XtSetArg(arglst[2], XmNrecomputeSize, FALSE);
  XtSetValues(flickArrowDownWidget, arglst, 3); 
  XtSetValues(movieArrowDownWidget, arglst, 3); 

  XtSetArg(arglst[0], XmNlabelPixmap, exitPixmap);
  XtSetArg(arglst[1], XmNselectPixmap, noexitPixmap);
  XtSetArg(arglst[2], XmNrecomputeSize, FALSE);
  XtSetValues(dioExitW, arglst, 3); 
  XtSetValues(fftFeedbackExitW, arglst, 3); 
  XtSetValues(modelExitW, arglst, 3); 
  XtSetValues(binningExitW, arglst, 3); 
  XtSetValues(interpExitW, arglst, 3); 
  XtSetValues(autoDRExitW, arglst, 3); 
  XtSetValues(auto2DDRExitW, arglst, 3); 
  XtSetValues(normalExitW, arglst, 3); 
  XtSetValues(fourAverExitW, arglst, 3); 
  XtSetValues(truncaddExitW, arglst, 3); 
  XtSetValues(rotateExitW, arglst, 3); 
  XtSetValues(fixExitW, arglst, 3); 
  XtSetValues(fix2DExitW, arglst, 3); 
  XtSetValues(pointExitW, arglst, 3); 
  XtSetValues(boxExitW, arglst, 3); 
  XtSetValues(optionsExitW, arglst, 3); 
  XtSetValues(fftExitW, arglst, 3); 
  XtSetValues(circAvgExitW, arglst, 3); 
  XtSetValues(crysDispExitW, arglst, 3); 
  XtSetValues(emcoravgExitW, arglst, 3); 
  XtSetValues(imgconvExitW, arglst, 3); 
  XtSetValues(rotfltExitW, arglst, 3); 
  XtSetValues(statExitW, arglst, 3); 
  XtSetValues(contourExitW, arglst, 3); 
  XtSetValues(colorExitW, arglst, 3); 
  XtSetValues(mapShadeExitW, arglst, 3); 
  XtSetValues(movieDoneWidget, arglst, 3); 
  XtSetValues(versionExitW, arglst, 3); 
  XtSetValues(partprocExitW, arglst, 3); 
  XtSetValues(moveExitW, arglst, 3); 
  XtSetValues(batchjobExitW, arglst, 3); 
  XtSetValues(invertExitW, arglst, 3); 
  XtSetValues(rescaleExitW, arglst, 3); 
  XtSetValues(fourierFiltExitW, arglst, 3); 
  XtSetValues(helpExitW, arglst, 3); 
  XtSetValues(subtractExitW, arglst, 3); 
  XtSetValues(deltaXYpartprocExitW, arglst, 3); 
  XtSetValues(deltaOmegaPPExitW, arglst, 3); 
  XtSetValues(manipPPExitW, arglst, 3); 
  XtSetValues(cavgPPExitW, arglst, 3); 
  XtSetValues(appodizeExitW, arglst, 3); 


  XtSetArg(arglst[0], XmNlabelPixmap, resetPixmap);
  XtSetArg(arglst[1], XmNselectPixmap, noresetPixmap);
  XtSetArg(arglst[2], XmNrecomputeSize, FALSE);
  XtSetValues(refineResetButtonW, arglst, 3); 
  XtSetValues(apOrigResetW, arglst, 3); 
  XtSetValues(adrabResetW, arglst, 3); 
  XtSetValues(latResetButW, arglst, 3); 
  XtSetValues(crysDispResetW, arglst, 3); 
  XtSetValues(fourAverResetW, arglst, 3); 
  XtSetValues(smMinMaxResetW, arglst, 3); 
  XtSetValues(diffSecResetW, arglst, 3); 
  XtSetValues(magSecResetW, arglst, 3); 
  XtSetValues(findMagResetW, arglst, 3); 
  XtSetValues(findRadResetW, arglst, 3); 
  XtSetValues(findMag2ResetW, arglst, 3); 
  XtSetValues(findMag3ResetW, arglst, 3); 
  XtSetValues(unblobResetW, arglst, 3); 

  XtSetArg(arglst[0], XmNlabelPixmap, helpPixmap);
  XtSetArg(arglst[1], XmNselectPixmap, nohelpPixmap);
  XtSetArg(arglst[2], XmNrecomputeSize, FALSE);
  XtSetValues(imgconvHelpW, arglst, 3); 
  XtSetValues(dioHelpW, arglst, 3); 
  XtSetValues(rotfltHelpW, arglst, 3); 
  XtSetValues(modelHelpW, arglst, 3); 
  XtSetValues(binningHelpW, arglst, 3); 
  XtSetValues(interpHelpW, arglst, 3); 
  XtSetValues(autoDRHelpW, arglst, 3); 
  XtSetValues(auto2DDRHelpW, arglst, 3); 
  XtSetValues(normalHelpW, arglst, 3); 
  XtSetValues(fourAverHelpW, arglst, 3); 
  XtSetValues(boxHelpW, arglst, 3); 
  XtSetValues(appodizeHelpW, arglst, 3); 
  XtSetValues(fourierFiltHelpW, arglst, 3); 
  XtSetValues(fftHelpW, arglst, 3); 
  XtSetValues(mapShadeHelpW, arglst, 3); 
  XtSetValues(helpToolW, arglst, 3);  
  XtSetValues(helpFixW, arglst, 3); 
  XtSetValues(helpFix2DW, arglst, 3); 
  XtSetValues(helpRotateW, arglst, 3); 
  XtSetValues(helpTruncAddW, arglst, 3); 
  XtSetValues(subtractHelpW, arglst, 3); 
  XtSetValues(deltaXYpartprocHelpW, arglst, 3); 
  XtSetValues(deltaOmegaPPHelpW, arglst, 3); 
  XtSetValues(manipPPHelpW, arglst, 3); 
  XtSetValues(cavgPPHelpW, arglst, 3); 
  XtSetArg(arglst[2], XmNrecomputeSize, TRUE);
  XtSetValues(ppHelpW, arglst, 3); 


  XtSetArg(arglst[0], XmNlabelPixmap, moviePixmap);
  XtSetArg(arglst[1], XmNselectPixmap, unmoviePixmap);
  XtSetArg(arglst[2], XmNrecomputeSize, FALSE);
  XtSetValues(fractToggleWidget, arglst, 3); 
  XtSetValues(angxToggleWidget, arglst, 3); 
  XtSetValues(angyToggleWidget, arglst, 3); 
  XtSetValues(thetaToggleWidget, arglst, 3); 
  XtSetValues(phiToggleWidget, arglst, 3); 
  XtSetValues(omegaToggleWidget, arglst, 3); 
  XtSetValues(maxRadiusToggleWidget, arglst, 3); 
  XtSetValues(pixelSizeToggleWidget, arglst, 3); 
  XtSetValues(unblobsizeToggleW, arglst, 3); 
  XtSetValues(unblobradiusToggleW, arglst, 3); 
  XtSetValues(maxxToggleW, arglst, 3); 
  XtSetValues(maxyToggleW, arglst, 3); 
  XtSetValues(maxzToggleW, arglst, 3); 
  XtSetValues(surfDensToggleWidget, arglst, 3);  
  /* OLDWAY  XtSetArg(arglst[0], XmNlabelPixmap, unmoviePixmap);
  XtSetArg(arglst[1], XmNselectPixmap, unmoviePixmap);
  XtSetArg(arglst[2], XmNrecomputeSize, FALSE);
  XtSetValues(surfDensToggleWidget, arglst, 3);  */


  XtSetArg(arglst[0], XmNlabelPixmap, infiniPixmap);
  XtSetArg(arglst[1], XmNselectPixmap, noinfiniPixmap);
  XtSetArg(arglst[2], XmNrecomputeSize, FALSE);
  XtSetValues(infiniWidget, arglst, 3); 
/*  XtSetValues(versionToolWidget, arglst, 3);  */
  XtSetValues(versionScrollW, arglst, 3); 
  XtSetArg(arglst[0], XmNlabelPixmap, noflashLightPixmap);
  XtSetArg(arglst[1], XmNselectPixmap, flashLightPixmap);
  XtSetArg(arglst[2], XmNrecomputeSize, FALSE);
  XtSetValues(flashLightWidget, arglst, 3); 
  XtSetArg(arglst[0], XmNlabelPixmap, norockPixmap);
  XtSetArg(arglst[1], XmNselectPixmap, rockPixmap);
  XtSetArg(arglst[2], XmNrecomputeSize, FALSE);
  XtSetValues(rockWidget, arglst, 3); 
  

  /* Start without it */
  XtSetArg(arglst[0], XmNsensitive, FALSE);
  XtSetValues(sectionWidget, arglst, 1); 
  XtSetArg(arglst[0], XmNsensitive, FALSE);
  XtSetValues(stepToggleWidget, arglst, 1); 
  XtSetValues(stepToggle2Widget, arglst, 1); 
  XtSetValues(frameTextWidget, arglst, 1); 
  XtSetValues(stepTextWidget, arglst, 1); 
  XtSetValues(stepValueLabelWidget, arglst, 1); 
  XtSetValues(numFramesLabelW, arglst, 1); 
  XtSetValues(movieFormW, arglst, 1); 

  boxing_GC = image_gcGREEN;


}/* XLIBINITIALIZE */




/****************************************************************
 * pickAutoZoom() - Try and guess which zoom factor they want.
*/
void pickAutoZoom(int width,int height)
{/* PICKAUTOZOOM*/
  int guess, guessFound,done, zoomedWidth,zoomedHeight;

}/* PICKAUTOZOOM*/




/***************************************************************
 * checkStatus() - Check to see if they want to stop this or not.
 * Very important if they have asked for a long computation as
 * in a 3D Render movie and find out half way through it they 
 * either don't want to wait that long or they setup the wrong
 * params.  This gives them a way to abort by hitting the STOP
 * pushbutton.
*/
int checkStatus(char *thisStr)
{/* CHECKSTATUS */

  if(drawLite)
    {/* OK to do */
      XmTextSetString(statusTextW, thisStr);
      
      /* Frame the location of the stop button */
      XDrawLine(dsply,XtWindow(statusStopW),image_gcRED,
		12,12,20,20);
      XDrawLine(dsply,XtWindow(statusStopW),image_gcRED,
		20,12,12,20);
      XFlush(dsply);
      
      /* See if they want to bugout or not */
      if(XCheckTypedWindowEvent(dsply,
				XtWindow(statusStopW),
				ButtonPress,
				&event))
	{/* STOP!!! */
	  XFlush(dsply);
	  return(TRUE);
	}/* STOP!!! */
      else
	{/* OK to continue ... */
	  return(FALSE);
	}/* OK to continue ... */
    }/* OK to do */
  else
    return(FALSE);

}/* CHECKSTATUS */




/***************************************************************
 * s_error() - All errors are fatal.
 */
void s_error(char *problem_string)
{/* S_ERROR */
  printf("%s\n", problem_string);
  exit(0);
}/* S_ERROR */



/****************************************************************
 * calcHistLength()- Use the histZoomFactor to make sure we
 * don't draw in the histogram window more than 32 pixels.
*/
int calcHistLength(int histVal)
{/* CALCHISTLENGTH */
  int rtnVal;

  rtnVal = (int)((float)histogram[(256-histVal)] * histScaleFactor);
  
  rtnVal = MIN(rtnVal,32);
  
  return(rtnVal);

}/* CALCHISTLENGTH */






/***************************************************************
 * stuffIt() - Stuff the presumably shorter intShort buffer into
 * the char buffer buff.  Used when writing VMS file compatible
 * files from our C code here.  It's easy to write the file
 * in a memory buffer here and then just blow it out into the
 * VMS structured file one "chunk" of memory at a time.
*/
void stuffIt(unsigned char *buff, int len, unsigned char *intShort)
{/* STUFFIT */
  int i;

  for(i=0;i<len;i++)
    *(buff+i) = (unsigned char)*(intShort + i);

}/* STUFFIT */





/************************************************************
 * stuffImage()- Take the integer buffer we've read from the
 * .IMG disk file and stuff it into our BYTE buffer for 
 * display.  Make the integer buffer disk data "fit" into our
 * colormap by scaling it.  e.g. You may have data from 0 to
 * 255 on the disk (per pixel) but we've only got 237 colors
 * available to us for our color table.
 *
 * We maintain a "mirror" buffer here for speed in rezooming
 * the image.  For example if we need to zoom the image it
 * is easier to read the data already converted to our colormap
 * doman and in character form than trying to reread the integer 
 * buffer and reconverting it.  Keep that integer buffer around,
 * though, to stuff the box buffers etc...  We need to always
 * write out the original data NOT the colromap converted stuff
 * when we write.
*/
void stuffImage(unsigned short *thisBuffer, int i, int j, int scaleIt,
		int fullScreenFirst)
{/* STUFFIMAGE */
  int k,l,x,y,thisOffset,yOff,yMirror;
  int min,max,lastCol,lastRow;
  float scaleFactor;
  BYTE *p,*z;  
  unsigned short *q;

  /* Calculate zoominess */
  calcSizeFactor(zoomFactor);

  if(thisBuffer==NULL)
    {/* Can't go on */
      goto destructor;
    }/* Can't go on */

  /* Let's trap here for images less than image_width x image_height */
  lastCol = display_width;
  lastRow = display_height;
  if(fullScreen)
    {/* CHekc these */
      if((numCols*sizeFactor) < display_width)
	lastCol = (purdueHeader.numCols*sizeFactor);
      if((numRows*sizeFactor) < display_height)
	lastRow = (purdueHeader.numRows*sizeFactor);
      if(disp1024fft && fftSize == 1024 && XtIsManaged(fftscreen))
	{/* special */
	  lastCol = (1024*sizeFactor);
	  lastRow = (1024*sizeFactor);
	}/* special */
    }/* CHekc these */

  /* Reset offsets */
  xoffset= yoffset= 0;
  
  /* Just to make sure let's init this */
  for(l=0;l<(buffer_height*3);l++)
    {/* Init them */
      memcpy((&image_buffer1[(l*buffer_width)]),stufBuf,buffer_width);
    }/* Init them */

  /* Now clip this guy to our colormap */ 
  scaleFactor = (((float)tryColors - 7.0) / 256.0);

  /* Init histogram and point to our integer and display 
   * buffers */
  for(y=0;y<257;y++)
    histogram[y]= 0;
  if(hiColor)
    p= (BYTE *) image_buffer1_disp;
  else
    p= (BYTE *) image_buffer1;
  q= (unsigned short *)  thisBuffer; 

  free(mirror_image_buffer1);
  CK(mirror_image_buffer1 = (BYTE *)malloc ((i*j)*(sizeof(BYTE))));
  z= (BYTE *)  mirror_image_buffer1; 

  scaleItSwitch = scaleIt;
  if(scaleItSwitch)
    {/* Try scaling it */
      min= 99999999;
      max = 0;
      min = INT_MAX; /* dgm INT_MAX instead of 99999999 */
      max = INT_MIN ;/* dgm INT_MIN instead of 0 */
      for(y=0;y<lastRow;y++)    
	{/* Tim draws from the bottom up - We are opposite */
	  yMirror = y*i;
	  for(x=0;x<lastCol;x++)
	    {/* Clip the entire image */
	      
	      q = (thisBuffer + (yMirror+x));
	      if(flickerFix2D && *q==30000)
		{/* DO nothing */}
	      else
		{
		  min = MIN((*q),min);
		  max = MAX((*q),max);
		}
	    }/* Clip the entire image */
	  
	}/* Tim draws from the bottom up - We are opposite */

      if(fileType==PORTABLE)
	scaleFactor = scaleFactor * (65500.0 / max);
      else
	scaleFactor = scaleFactor * (255.0 / max);

    }/* Try scaling it */
  else 
    {/* No scaling */
      if(fileType==BYTEPORTABLE)
	{/* byte */
	  max=255;min=0;
	  /* What about if you want to maintian the values on the
	   * disk - e.g. suppose you max BYTE value is 127 why
	   * should you have to "dummy" it down??? */
	  if(purdueHeader.img_max < tryColors)
	    {/* Do this */
	      max = (tryColors-7);
	      min = 0;
	    }/* Do this */
	}/* byte */
      else
	{/* short */
	  max = min = thisBuffer[0] ;
	  for(k=0;k<image_height;k++)
	    for(l=0;l<image_width;l++)
	      {
		min = MIN(thisBuffer[((k*image_width)+l)],min);
		max = MAX(thisBuffer[((k*image_width)+l)],max);
	      }
	}/* short */
    }/* No scaling */

  /* Now clip this guy to our colormap - 8 bit colormap used to
  * reserve the first 7 spaces for defined colors like red,green,blue etc.*/ 
  scaleFactor = (
		 ((float)tryColors  - 7.0)
		 / 
		 (max - min)
		 );

  /* Convert integer to char within our colormap and put it
   * in the display buffer(s) */
  for(y=0;y<lastRow;y++)    
    {/* Tim draws from the bottom up - We are opposite */
      yOff= y*buffer_width;
      yMirror = y*i;
      for(x=0;x<lastCol;x++)
	{/* Clip the entire image */
	  
	  p = (image_buffer1 + (yOff+x));
	  q = (thisBuffer + (yMirror+x));
	  z = (mirror_image_buffer1 + (yMirror+x));
	  
	  /* We've got to scale these for now because we may get one
	   * that is larger than we've got colormap entries for.
	   */
	  thisOffset = (int)( (  ( ((float)*(q)- (float)min) * 
				       scaleFactor )   ) + 6.0);

	  if((thisOffset >=6) && (thisOffset<tryColors))
	    {/* OK to do */
	      if(hiColor)
		*(z)= thisOffset;
	      else
		*(z)= *(p)= (BYTE)pixels_return[thisOffset];
	      XPutPixel(image_image1, x, y, pixels_return[thisOffset]);
	      
	      if(hiColor)
		histogram[(thisOffset)]+= 1;
	      else
		histogram[(pixels_return[thisOffset])]+= 1;
	    }/* OK to do */
	  else 
	    /* printf ("x[%d]y[%d] q[%d] tOff[%d]\n",x,y,*(q),thisOffset); */
	    ;

	  if(flickerFix2D)
	    {/* Do flicker case */
	      if(*q==30000)
		{
		  if(hiColor)
		    {
		      *(z)= *(p)= RED;		
		      XPutPixel(image_image1, x, y, pixels_return[RED]);
		    }
		  else
		    *(z)= *(p)= (BYTE)pixels_return[RED];		
		}
	    }/* Do flicker case */

	}/* Clip the entire image */
    }/* Tim draws from the bottom up - We are opposite */
  

  if(histoEqualize && !fullScreenFirst)
    {/* Histogram equalize */
      int thisWidth,thisHeight;
      /* Init histogram and point to our integer and display 
       * buffers */
      for(y=0;y<257;y++)
	histogram[y]= 0;
      
      if(file_height < display_height)
	thisHeight = file_height;
      else thisHeight = display_height;
      if(file_width < display_width)
	thisWidth = file_width;
      else thisWidth = display_width;
      if(hiColor)
	histEqualize(thisWidth,thisHeight, image_buffer1, 
		     7,245,int_image_buffer1,
		     min,max,buffer_width);
      else
	histEqualize(thisWidth,thisHeight, image_buffer1, 
		     pixels_return[7],(tryColors-7),image_buffer1,
		     min,max,buffer_width);
    }/* Histogram equalize */

 destructor:;

}/* STUFFIMAGE */




/************************************************************
 * stuffMapImage()- Take the integer buffer we've read from the
 * .MAP disk file and stuff it into our BYTE buffer for 
 * display.  Make the integer buffer disk data "fit" into our
 * colormap by scaling it.  e.g. You may have data from 0 to
 * 255 on the disk (per pixel) but we've only got 237 colors
 * available to us for our color table.  For this MAP data let's
 * just start out with the first section.
 *
 * We maintain a "mirror" buffer here for speed in rezooming
 * the image.  For example if we need to zoom the image it
 * is easier to read the data already converted to our colormap
 * doman and in character form than trying to reread the integer 
 * buffer and reconverting it.  Keep that integer buffer around,
 * though, to stuff the box buffers etc...  We need to always
 * write out the original data NOT the colromap converted stuff
 * when we write.
*/
void stuffMapImage(float *map_img_buf, 
		   float *thisBuffer, 
		   int ncols, 
		   int nrows, 
		   int nsecs,
		   int startImg,
		   int passedOffset)
{/* stuffMapImage */
  int k,l,x,y,thisOffset,yOff,yMirror,firstTime;
  int sectionsPerRow,rowsPerImage,thisNumSecs, thisX;
  float thisPixel;
  BYTE *p,*z;  
  float *q;

  firstTime = TRUE;
  /* Force it to always display the depth cued buffer if we've got it 
   * or want it (e.g. from the section toggle etc...)*/
  if(shadedMaps && !dispSections)
    {/* Shaded maps */
      thisBuffer= map_img_buf;
      ncols = largestDim;
      nrows = largestDim;
    }/* Shaded maps */

  /* Computer numbers from 0 not 1 */
  startImg--;

  /* Calculate zoominess */
  calcSizeFactor(zoomFactor);

  if(passedOffset==0)
    {/* Just to make sure let's init this */
      for(l=0;l<(buffer_height*3);l++)
	{/* Init them */
	  memcpy((&image_buffer1[(l*buffer_width)]),stufBuf,buffer_width);
	}/* Init them */
    }/* Just to make sure let's init this */

  if((minMapPixel==maxMapPixel))
    {/* Can't do it */
      sprintf(outStr,"MAP file has bogus data min[%.2f]max[%.2f]",
	      minMapPixel,maxMapPixel);
      askThem(outStr,FALSE,XmDIALOG_QUESTION);
      goto destructor;
    }/* Can't do it */

  /* Now clip this guy to our colormap */ 
  if(minMapPixel < 0.0)
    scaleMapFactor = (((float)tryColors - 7.0) / (maxMapPixel-minMapPixel));
  else scaleMapFactor = (((float)tryColors - 7.0) / maxMapPixel);

  if(shadedMaps && !dispSections)
    {/* scale it */

      /* We REALLY need to have this here for the radial shading
       * to work properly.  dph_rescale() sets it up properly. */
      if(radialMapShade)
	scaleMapFactor= 1.0;  /* Already done in dph_rescale_rob() */

    }/* scale it */
  
  /* Now find out how we're going to put these on the display */
  if(buffer_width < display_width)
    sectionsPerRow= (int)(buffer_width / (int)(ncols*sizeFactor));
  else sectionsPerRow= (int)(display_width / (int)(ncols*sizeFactor));
  sectionsPerRow= MAX(1,sectionsPerRow);
  if(buffer_height < display_height)
    rowsPerImage = buffer_height / (int)(nrows*sizeFactor);
  else rowsPerImage = display_height / (int)(nrows*sizeFactor);
  rowsPerImage = MAX(1,rowsPerImage);
  thisNumSecs = (sectionsPerRow * rowsPerImage) + startImg;
  thisNumSecs = MAX((startImg+1),MIN(thisNumSecs,nsecs));

  /* Stuff globals */
  mapSectionsPerRow = sectionsPerRow;
  mapStartImage = startImg;

  /* Init histogram and point to our integer and display 
   * buffers */
  for(y=0;y<257;y++)
    histogram[y]= 0;
  p= (BYTE *) image_buffer1;
  q= (float *)  thisBuffer; 

  /* Convert integer to char within our colormap and put it
   * in the display buffer(s) */
  lastImg= thisNumSecs;
  for(l=startImg;l<thisNumSecs;l++)
    {/* Do all sections */
      int rowOffset,colOffset;

      rowOffset = (int)((l-startImg) / sectionsPerRow)*(nrows*sizeFactor);
      colOffset = (int)((l-startImg) % sectionsPerRow)*(ncols*sizeFactor);

      colOffset += passedOffset;

      imgNums[l].x= colOffset;
      imgNums[l].y= rowOffset;

      for(y=0;y<(nrows*sizeFactor);y++)    
	{/* Tim draws from the bottom up - We are opposite */
	  yOff= ((rowOffset+y)*buffer_width + colOffset);
	  yMirror = ((int)(y*recipSizeFactor))*ncols;
	  for(x=0;x<(ncols*sizeFactor);x++)
	    {/* Clip the entire image */
	      
	      p = (image_buffer1 + (yOff+x));
	      q = (thisBuffer + ((l*nrows*ncols)+
				 yMirror+
				 (int)(x*recipSizeFactor))); 

	      if(*q == -1234567.0)
		{/* Special case for subtract map thresholding */
		  thisOffset = GREEN;
		}/* Special case for subtract map thresholding */
	      else if(*q == -1234568.0)
		{/* Special case for subtract map thresholding */
		  thisOffset = RED;
		}/* Special case for subtract map thresholding */
	      else
		{/* normal */

		  /* We've got to scale these for now because we may get one
		   * that is larger than we've got colormap entries for.
		   */
		  if(minMapPixel < 0.0)
		    thisPixel= (float)(*(q) + (0.0 - minMapPixel));
		  else thisPixel= (float)*(q);
		  
		  thisOffset = ((int) ( (int) ( ((float)thisPixel) * 
					       scaleMapFactor )   ) + 6);
		}/* normal */
	      
	      if((thisOffset >=0) && (thisOffset<256))
		{/* OK to do */

		  if(colOffset+x > display_width)
		    thisX = display_width - colOffset-1;
		  else
		    thisX = x;
		  XPutPixel(image_image1, colOffset+thisX, rowOffset+y,
			    pixels_return[thisOffset]);

		  if(hiColor)
		    histogram[((int)thisOffset)]+= 1;
		  else
		    histogram[((int)thisOffset + pixels_return[0])]+= 1;
		  contrastMax = MAX(contrastMax,*(p));
		  contrastMin = MIN(contrastMin,*(p));
		}/* OK to do */
	      else
		{
		  if(firstTime)
		    printf ("colormap problems 1\n\
x[%d]y[%d] q[%d]\n",x,y,(int)*(q));
		  firstTime = FALSE;
		}

	    }/* Clip the entire image */
	}/* Tim draws from the bottom up - We are opposite */

    }/* Do all sections */

  image_width= ncols;
  image_height= nrows;
  
 destructor:;

}/* stuffMapImage */




/*************************************************************
 * stuffPackedImage()- For memory conservation reasons
 * we've got to get these off the disk.
 * Currently we assume all of the data is in memory and 
 * nothing is needed from the disk.
*/
void stuffPackedImage(int width,int height,
		      float **pkImg, int nimg,
		      int startImg)
{/* STUFFPACKEDIMAGE */
  int i,m,l,k,outWidth,yOff;
  int yMirror,thisOffset,firstTime;
  int sectionsPerRow,rowsPerImage,thisNumSecs;
  float hTsf,wTsf,thisFloat;
  int  x,y;
  int  hTsfInt,wTsfInt,numRowsDone;
  float *thisBuffer,*thisBuffer2,thisPixel,*p,*q, *w;
  BYTE *s;
  P_INFO *thisInfo,*foundID;

  firstTime = TRUE;
  if(fileType==PACKEDPORTABLE)
    {/* Stuff the first one */
      
      calcCircleStats(pkImg[(startImg-1)],
		      &purdueHeader.img_bkg, 
		      &purdueHeader.img_min,
		      &purdueHeader.img_max, 
		      &purdueHeader.img_avg, 
		      &purdueHeader.img_std,
		      purdueHeader.numCols, purdueHeader.numRows, 
		      purdueHeader.numCols);
      
      /* Write stats */
      sprintf(outStr,"\nStatistics for file [%s] Image [%d]\n\
\n\
File size   [%d]x[%d]x[%d]\n\
Background  [%f]\n\
MIN,MAX     [%f],[%f]\n\
AVG,STD     [%f],[%f]\n\
\n\
\n",
	      fileName.addr, startImg,
	      purdueHeader.numCols,	  purdueHeader.numRows, 
	      purdueHeader.nimg,
	      purdueHeader.img_bkg,
	      purdueHeader.img_min, purdueHeader.img_max, 
	      purdueHeader.img_avg, purdueHeader.img_std
	      );
      XmTextSetString(statTextWidget,outStr);
    }/* Stuff the first one */
  
  /* Computer numbers from 0 not 1 */
  startImg--;
  if(startImg > nimg)
    startImg = nimg-1;

  /* Calculate zoominess */
  calcSizeFactor(zoomFactor);

  /* Just to make sure let's init this */
  for(l=0;l<(buffer_height*3);l++)
    {/* Init them */
      memcpy((&image_buffer1[(l*buffer_width)]),stufBuf,buffer_width);
    }/* Init them */


  /* Now find out how we're going to put these on the display */
  if(buffer_width < display_width)
    sectionsPerRow= (int)(buffer_width / (int)(width*sizeFactor));
  else sectionsPerRow= (int)(display_width / (int)(width*sizeFactor));
  sectionsPerRow= MAX(1,sectionsPerRow);
  if(buffer_height < display_height)
    rowsPerImage = buffer_height / (int)(height*sizeFactor);
  else rowsPerImage = display_height / (int)(height*sizeFactor);
  rowsPerImage = MAX(1,rowsPerImage);
  thisNumSecs = (sectionsPerRow * rowsPerImage) + startImg;
  thisNumSecs = MAX((startImg),MIN(thisNumSecs,(nimg)));

  /* Stuff globals */
  packSectionsPerRow = sectionsPerRow;
  packStartImage = startImg;

  /* Init histogram and point to our integer and display 
   * buffers */
  for(y=0;y<257;y++)
    histogram[y]= 0;
  p = (float *) inCore_buffer;
  s = (BYTE *)image_buffer1_disp;

  /* Convert integer to char within our colormap and put it
   * in the display buffer(s) */
  lastImg= thisNumSecs;
  if(haveParamFile && mapProj)
    lastImg = thisNumSecs = startImg +1;
  if(projASU)
    {/* Setup for projected ASU */
      int thisRow;
      thisRow = (int)(startImg / maxPhi);
      numRowsDone = thisRow-1;
      startImg = 0;  /* We're just using the loop to cycle through them all */
      /* We have less to do if we start later  ??? */
      thisNumSecs -= (thisRow * maxPhi);
    }/* Setup for projected ASU */

  for(l=startImg;l<thisNumSecs;l++)
    {/* Do all sections */
      int rowOffset,colOffset,thisL;

      if(projASU)
	{/* They have some projections in ASU form */
	  if((l % sectionsPerRow) >= maxPhi)
	    {/* They've zoomed so small they have more than maxPhi perline */
	      thisL = (maxPhi-1);/* Just do a blank one */
	      thisNumSecs++;     /* And make them do another one as
				  * this wasn't accounted for above */
	    }/* They've zoomed so small they have more than maxPhi perline */
	  else if((l % sectionsPerRow)==0)
	    {/* Go to next line */
	      numRowsDone++;
	      thisL = (numRowsDone * maxPhi);
	    }/* Go to next line */
	  else
	    {/* Just do this line */
	      thisL = (numRowsDone * maxPhi) + (l % sectionsPerRow);
	    }/* Just do this line */
	  if(thisL >= nimg)
	    {/* Don't even bother to do this one */
	      thisL = maxPhi-1;
	      goto nextL;
	    }/* Don't even bother to do this one */
	  thisL = MIN(thisL,nimg-1);
	}/* They have some projections in ASU form */
      else if(haveParamFile)
	{/* They may have sorted the boxes */
	  foundID = NULL;
	  thisInfo = info.info;
	  while(thisInfo != NULL)
	    {
	      if(thisInfo->rank==(l+1))
		{/* Found it */
		  foundID = thisInfo;
		  thisL = thisInfo->particle_number-1;
		  break;
		}/* Found it */
	      else thisInfo = thisInfo->next;
	    }

	}/* They may have sorted the boxes */
      else
	{/* normal */
	  thisL = l;
	}/* normal */

      q= thisBuffer= (float *)pkImg[thisL]; 

      rowOffset = (int)((l-startImg) / sectionsPerRow)*(height*sizeFactor);
      colOffset = (int)((l-startImg) % sectionsPerRow)*(width*sizeFactor);

      imgNums[l].x= colOffset;
      imgNums[l].y= rowOffset;
      if(thisInfo==NULL)
	imgNums[l].particle_number = l+1; /* Couldn't find it */
      else imgNums[l].particle_number = thisL+1;

      hTsf = height*sizeFactor;
      wTsf = width *sizeFactor;
      for(y=0;y<(hTsf);y++)    
	{/* Tim draws from the bottom up - We are opposite */
	  yOff= ((rowOffset+y)*buffer_width + colOffset);
	  yMirror = ((int)(y*recipSizeFactor))*width;
	  for(x=0;x<(wTsf);x++)
	    {/* Clip the entire image */
	      
	      p = (inCore_buffer + (yOff+x));
	      s = (image_buffer1_disp + (yOff+x));
	      q = (thisBuffer + (yMirror + (int)(x*recipSizeFactor)));
	      
	      /* We've got to scale these for now because we may get one
	       * that is larger than we've got colormap entries for.
	       */
	      if(minPackPixel < 0.0)
		thisPixel= *(q) + (0.0 - minPackPixel);
	      else thisPixel= *(q);
	      
	      thisOffset = ((int) ( (int) ( ((float)thisPixel) * 
					   scalePackFactor )   ) + 6);
	      
	      /* This is done from the part proc screen
	       * so we can see something next to the 
	       * projected map image.  This should
	       * clip all but the darkest images at the
	       * top but it's better than nothing ...  */
	      if(bumpContrast)
		{/* Let's just double it */
		  thisOffset *= 2;
		  thisOffset = MAX(MIN(255,thisOffset),0);
		}/* Let's just double it */
		  
	      if((thisOffset >=0) && (thisOffset<256))
		{/* OK to do */
		  *(s)= thisOffset;
		  XPutPixel(image_image1, x+colOffset, y+rowOffset,
			    pixels_return[thisOffset]);
		  if(hiColor)
		    histogram[(thisOffset)]+= 1;
		  else
		    histogram[(thisOffset + pixels_return[0])]+= 1;
		}/* OK to do */
	      else 
		{
		  if(firstTime)
		    printf ("colormap problems 2\n\
x[%d]y[%d] q[%d]\n",x,y,(int)*(q));
		  firstTime = FALSE;
		}

	      if(flickerFix2D)
		{/* DO for special case flickering */
		  thisBuffer2= (float *)fix2DpackImg_diff[thisL]; 
		  w = (thisBuffer2 + (yMirror + (int)(x*recipSizeFactor)));
		  if((*w))
		    {
		      if(hiColor)
			{
			  /* DONT DO IT !!  *(p) = pixels_return[RED]; */
			  XPutPixel(image_image1, x+colOffset, y+rowOffset,
				    pixels_return[RED]);
			}
		      else
			*(p) = (BYTE)pixels_return[RED];
		    }
		}/* DO for special case flickering */

	    }/* Clip the entire image */
	}/* Tim draws from the bottom up - We are opposite */

      if(projASU)
	{/* They have some projections in ASU form */
	}/* They have some projections in ASU form */
      else if(haveParamFile && dispCenters)
	{/* Let's draw the centers */
	  int xZoom,yZoom;
	  
	  if(foundID != NULL)
	    {/* OK to draw */
	      xZoom = (int)(foundID->x*sizeFactor);
	      yZoom = (int)((purdueHeader.numRows-foundID->y)*sizeFactor);
	      xZoom -= (int)(4);
	      yZoom -= (int)(4);

	      /* Draw one-half of the "X" */
	      x = xZoom;
	      for(y=yZoom;y<=(yZoom+8);y++)    
		{/* Tim draws from the bottom up - We are opposite */
		  if(hiColor)
		    {/* hi */
		      if(!foundID->toggle)
			XPutPixel(image_image1, x+colOffset, y+rowOffset,
				  pixels_return[RED]);
		      else
			XPutPixel(image_image1, x+colOffset, y+rowOffset,
				  pixels_return[BLUE]);
		    }/* hi */
		  else
		    {/* lo */
		      BYTE *p;
		      yOff= ((rowOffset+y)*buffer_width + colOffset);
		      p = (image_buffer1 + (yOff+x));
		      if(!foundID->toggle)
			*(p)= (BYTE)pixels_return[RED];
		      else *(p)= (BYTE)pixels_return[BLUE];
		    }/* lo */

		  x++;
		}/* Tim draws from the bottom up - We are opposite */
	      /* Draw other-half of the "X" */
	      y = yZoom;
	      for(x=(xZoom+8);x>=(xZoom);x--)    
		{/* Tim draws from the bottom up - We are opposite */
		  if(hiColor)
		    {/* hi */
		      if(!foundID->toggle)
			XPutPixel(image_image1, x+colOffset, y+rowOffset,
				  pixels_return[RED]);
		      else
			XPutPixel(image_image1, x+colOffset, y+rowOffset,
				  pixels_return[BLUE]);
		    }/* hi */
		  else
		    {/* lo */
		      BYTE *p;
		      yOff= ((rowOffset+y)*buffer_width + colOffset);
		      p = (image_buffer1 + (yOff+x));
		      if(!foundID->toggle)
			*(p)= (BYTE)pixels_return[RED];
		      else *(p)= (BYTE)pixels_return[BLUE];
		    }/* lo */
		  y++;
		}/* Tim draws from the bottom up - We are opposite */

	    }/* OK to draw */
	  else
	    {/* Can't find it */
	      /* printf("Can't find particle [%d]\n",l); */
	    }/* Can't find it */

	}/* Let's draw the centers */

      if(projASU)
	{/* They have some projections in ASU form */
	}/* They have some projections in ASU form */
      else if(haveParamFile && mapProj)
	{/* Display a map projection */
	  int i,j,k,lun,yOut1,yOut2;
	  float thisOmega,thisTheta,thisPhi,thisSF;
	  float maxF,minF,z1,z2,thisFloat;
	  float *projn,*map2d,*thisQ;
	  /*
	   **************************************************************
	   *    NOTE USE OF OMEGA = 90.0 TO GET CORRECT ORIENTATIONS    *
	   **************************************************************
	   */
	  if(foundID->projection==NULL)
	    {/* Then we've got to create it */

	      raiseIt(statusScreenWidget);
	      XFlush(dsply);
	      sprintf(outStr,"Creating MAP projection ... ");
	      if(checkStatus(outStr))
		{/* Do nothing for now ... */}

	      /* ORig way that looks 180 degrees off */
	      thisOmega = foundID->omega + 90.0;
	      thisOmega = foundID->omega - 90.0;  /* WORKS!! */
	      thisTheta = foundID->theta;
	      thisPhi   = foundID->phi;
	      z1 = 1.0;
	      z2 = (float)mapSecs;
	      CK(projn = (float *)malloc(sizeof(float)*mapCols*(mapRows+2)));
	      CK(map2d = (float *)malloc(sizeof(float)*mapCols*(mapRows+2)));

	      map_prj_ (&lun, 
			&mapCols, &mapRows, &mapSecs, 
			projn, map2d, 
			float_image_buffer1,
			&thisTheta, &thisPhi, &thisOmega, 
			&z1, &z2,
			&map_alp, &map_bet, &map_gam);

	      free(projn);
	      /* We've got to flip the projection in the Y direction.
	       * Tim draws from the bottom up -we're opposite
	       */
	      for(y=0;y<(mapRows/2);y++)
		{/* Rows */
		  yOut1 = (mapRows - y - 1)*mapCols;
		  yOut2 = y*mapCols;
		  for(x=0;x<mapCols;x++)
		    {/* Cols */
		      thisFloat        = map2d[(yOut2+x)];
		      map2d[(yOut2+x)] = map2d[(yOut1+x)];
		      map2d[(yOut1+x)] = thisFloat;
		    }/* Cols */
		}/* Rows */

	      foundID->projection = map2d;
	      XtUnmanageChild(statusScreenWidget);
	      XFlush(dsply);
	    }/* Then we've got to create it */

	  map2d = foundID->projection;
	  maxF = minF = map2d[0];
	  for(i=0;i<mapRows;i++)
	    {/* rows */
	      for(j=0;j<mapCols;j++)
		{/* cols */
		  minF = MIN(map2d[(i*mapCols)+j],minF);
		  maxF = MAX(map2d[(i*mapCols)+j],maxF);
		}/* cols */
	    }/* rows */
	  thisSF = (float)(tryColors-7.0)/(maxF-minF);

	  /* Now place this projection next to the original */
	  colOffset += (int)(width*sizeFactor);
	
	  /* Suppose we've got a binned MAP and a normal sized
	   * particle file.  Let's account for that with sizeFactor
	   */
	  sizeFactor *= (float)((float)(purdueHeader.numCols+1) / 
				(float)(mapCols+1));
	  recipSizeFactor = 1.0 / sizeFactor;
	  height = mapRows;
	  width  = mapCols;
	  for(y=0;y<(height*sizeFactor);y++)    
	    {/* Tim draws from the bottom up - We are opposite */
	      yOff= ((rowOffset+y)*buffer_width + colOffset);
	      yMirror = ((int)(y*recipSizeFactor))*width;
	      for(x=0;x<(width*sizeFactor);x++)
		{/* Clip the entire image */
		  
		  p = (inCore_buffer + (yOff+x));
		  thisQ = (map2d + (yMirror + (int)(x*recipSizeFactor)));
		  
		  thisFloat= *(thisQ);
		  
		  thisOffset = ((int) ( (int) ( ((float)thisFloat-minF) * 
					       thisSF )   ) + 6);
		  
		  if((thisOffset >=0) && (thisOffset<256))
		    {/* OK to do */
		      if(hiColor)
			XPutPixel(image_image1, x+colOffset, y+rowOffset,
				  pixels_return[thisOffset]);
		      else
			{
			XPutPixel(image_image1, x+colOffset, y+rowOffset,
				  pixels_return[thisOffset]);
			}
		    }/* OK to do */
		  else 
		  {
		    if(firstTime)
		      printf ("colormap problems 3\n\
x[%d]y[%d] q[%d]\n",x,y,(int)*(q));
		    firstTime = FALSE;
		  }
		  
		}/* Clip the entire image */
	    }/* Tim draws from the bottom up - We are opposite */
	  
	  if(mapProjContour)
	    {/* Now draw the contour in the main window */
	      int saveEditBoxZoom,deltaX,deltaY;
	      float thisTour;
	      char *tourStr;
	      /* drawContour uses editBoxZoom so let's stuff it
	       * temporarily with the main window zoom factor
	       * adjusted for any binning discrepancy between
	       * particles and map */
	      saveEditBoxZoom = editBoxZoom;
	      editBoxZoom = (zoomFactor + 1);
	      if(purdueHeader.numCols > mapCols)
		editBoxZoom += 1;   /* Map is binned */
	      calcSizeFactor(zoomFactor);
	      deltaX = (int)((foundID->x - (float)(purdueHeader.numCols/2))
			     * sizeFactor);
	      deltaY = (int)((foundID->y - (float)(purdueHeader.numRows/2))
			     * sizeFactor);
	      /* drawContour depends on these globals */
	      numCols = mapCols;
	      numRows = mapRows;
	      tourStr = XmTextGetString(projContourTextW);
	      thisTour = (float)strtod(tourStr,0);
	      XtFree(tourStr);
	      drawContour(thisTour,
			  TRUE,
			  0,
			  drawingWidget,
			  map2d,
			  deltaX,deltaY,
			  image_buffer1);
	      editBoxZoom = saveEditBoxZoom;
	    }/* Now draw the contour in the main window */

	}/* Display a map projection */

      nextL:;

    }/* Do all sections */

}/* stuffPackedImage */




/**************************************************************
 * convertDisp2Image() - Convert the coordinates gotten from the
 * mouse on the screen to original unzoomed image coordinates.  
 * Take into account the zoom and x,y offsets.
*/
int convertDisp2Image(int *inVal, int inOffset, int virOffset)
{/* CONVERTDISP2IMAGE */

  /* Also changed recipsizeFactor for sizeFactor below. */
  calcSizeFactor(zoomFactor);

  /* Map this to original image coords */
  *inVal= (int)(((*inVal)*recipSizeFactor)+(inOffset*recipSizeFactor));
  *inVal += virOffset;

  return(*inVal);

}/* CONVERTDISP2IMAGE */




/**************************************************************
 * convertImage2Disp() - Convert the original unzoomed coordinates 
 * into coordinates for the image on the screen that is displayed.
 * Take into account the zoom and x,y offsets.
*/
int convertImage2Disp(int *inVal, int inOffset, int virOffset)
{/* CONVERTImage2Disp */

  calcSizeFactor(zoomFactor);

  /* Map this to display coords */
  *inVal= (int)(((*inVal)*sizeFactor)-inOffset);
  *inVal -= virOffset;

  return(*inVal);

}/* CONVERTImage2Disp */





/*******************************************************
 * pushBox() - We've got a new box.  Record the x,y coords
 * as the coords of the image on the disk (we may not have
 * all of it displayed right now).  Also allocate the memory
 * for the box image and get it from what we've got in
 * memory right now.  If we don't get the image from memory
 * now when they've selected the point we'll have to read
 * it of the disk later.
 * 
 * Added 6-24-96:  Check to see if the coords of this box are
 * +- 10 pixels from another box - if so then just update the coords 
 * for that box - they are "tuning" the coords for that box.
*/
void pushBox(int x, int y, int findThisBox, int *thisBox)
{/* PUSHBOX */
  int k,foundIt,thisRadius;

  /* We also changed all of the references to sizeFactor below
   * to recipSizeFactor */
  calcSizeFactor(zoomFactor);
  
  /* Map this to original image coords */
  box[numBoxes].x = (x*recipSizeFactor)+(xoffset*recipSizeFactor)+virXoffset;
  box[numBoxes].y = (y*recipSizeFactor)+(yoffset*recipSizeFactor)+virYoffset;
  box[numBoxes].inMemory= FALSE;  /* They've got to "paint" it into memory */
  /* Add reference particle marker , Aug.20, 97, Lixin Zhu*/ 
  box[numBoxes].reference = ref_box_flag; 
  
  /* Check to see if this one's been done before */
  foundIt = FALSE;
  thisRadius = boxRadius;
  thisRadius = MAX(20,thisRadius);
  if(findThisBox)
    {/* FInd the box pointed to */
      *thisBox = -99;
      for(k=0;k<numBoxes;k++)
	{/* Check all */
	  if(((box[k].x >= (box[numBoxes].x-thisRadius))&&
	      (box[k].x <= (box[numBoxes].x+thisRadius))) &&
	     ((box[k].y >= (box[numBoxes].y-thisRadius))&&
	      (box[k].y <= (box[numBoxes].y+thisRadius))) )
	    {/* This is it */
	      if(ref_box_flag)
		{/* We are trying to find an existing box */
		  /* Do nothing */
		}/* We are trying to find an existing box */
	      else
		{/* New box */
		  *thisBox = k;
		}/* New box */
	    }/* This is it */
	}/* Check all */
    }/* FInd the box pointed to */
  else
    {/* Try to add it */
      if(*thisBox >= 0)
	{/* Found it last time */
	  k = *thisBox;
	  foundIt = TRUE;
	  if(ref_box_flag)
	    {/* We are trying to find an existing box */
	      /* Do nothing */
	    }/* We are trying to find an existing box */
	  else
	    {/* New box */
	      box[k].x = box[numBoxes].x;
	      box[k].y = box[numBoxes].y;
	    }/* New box */
	  box[k].reference = ref_box_flag;
	}/* Found it last time */
      else
	{/* Look for it */
	  for(k=0;k<numBoxes;k++)
	    {/* Check all */
	      if(((box[k].x >= (box[numBoxes].x-thisRadius))&&
		  (box[k].x <= (box[numBoxes].x+thisRadius))) &&
		 ((box[k].y >= (box[numBoxes].y-thisRadius))&&
		  (box[k].y <= (box[numBoxes].y+thisRadius))) )
		{/* This is it */
		  foundIt = TRUE;
		  if(ref_box_flag)
		    {/* We are trying to find an existing box */
		      /* Do nothing */
		    }/* We are trying to find an existing box */
		  else
		    {/* New box */
		      box[k].x = box[numBoxes].x;
		      box[k].y = box[numBoxes].y;
		    }/* New box */
		  box[k].reference = ref_box_flag;
		}/* This is it */
	    }/* Check all */
	}/* Look for it */
      /* Check to see if we can add it to the end */
      if(!foundIt)
	{/* Add to end? */
	  if( ((box[numBoxes].x-boxRadius) < 0) ||
	     ((box[numBoxes].y-boxRadius) < 0) ||
	     ((box[numBoxes].x+boxRadius) > (file_width-1)) ||
	     ((box[numBoxes].y+boxRadius) > (file_height-1)))
	    {/* Can't add to list - it's out of the image */
	    }/* Can't add to list - it's out of the image */
	  else
	    {/* OK add it */
	      
	      numBoxes++;
	      
	    }/* OK add it */
	}/* Add to end? */
    }/* Try to add it */
  

}/* PUSHBOX */




/*******************************************************
 * popBox() - Take the top box off the box stack.
*/
void popBox()
{/* POPBOX */

  numBoxes--;
  if(numBoxes < 0)
    numBoxes = 0;

  if(box[numBoxes].inMemory)
    {/* Get rid of memory */
      if(box[numBoxes].edit_mirror_buffer)
	free(box[numBoxes].edit_mirror_buffer);
    }/* Get rid of memory */
/* malloc_verify(); */

}/* POPBOX */



/***************************************************************
 * removeBox() - Remove the specified box from the box stack.
*/
void removeBox(int x, int y, int diskCoords)
{/* REMOVEBOX */
  int i,j,foundIt,findSize;

  /* We also changed all of the references to sizeFactor below
   * to recipSizeFactor */
  calcSizeFactor(zoomFactor);

  /* How close do they have to click to the center of the circle
   * before we say they've found this one?  Even if they've reduced
   * the image way down still give them boxRadius pixels (in orig
   * unzoomed image space).
   */
  findSize = (boxRadius/2);

  if(!diskCoords)
    {/* Map this to original image coords */
      x = (x*recipSizeFactor)+(xoffset*recipSizeFactor)+virXoffset;
      y = (y*recipSizeFactor)+(yoffset*recipSizeFactor)+virYoffset;
    }/* Map this to original image coords */

  /* Find the box that fits these coords */
  foundIt= -99;
  for(i=0;i<numBoxes;i++)
    {/* Look for a hit */
      if(((x+findSize) > box[i].x && box[i].x > (x-findSize)) &&
	 ((y+findSize) > box[i].y && box[i].y > (y-findSize)) )
	{/* Found it */
	  foundIt= i;
	}/* Found it */

    }/* Look for a hit */

  if(box[foundIt].inMemory)
    {/* Get rid of memory */
      if(box[foundIt].edit_mirror_buffer)
	free(box[foundIt].edit_mirror_buffer);
    }/* Get rid of memory */

  /* Move other boxes in its place */
  if(foundIt > -99)
    {/* found one */
      for(j=foundIt;j<numBoxes;j++)
	{/* Move them up */
	  box[j].x = box[j+1].x;
	  box[j].y = box[j+1].y;
	  box[j].reference= box[j+1].reference;
	  box[j].edit_mirror_buffer = box[j+1].edit_mirror_buffer;
	  box[j].inMemory = box[j+1].inMemory;
	  box[j].correlation = box[j+1].correlation;
	  box[j].avgdensity  = box[j+1].avgdensity;
	  box[j].stdev       = box[j+1].stdev;
	}/* Move them up */
      numBoxes--;
      if(numBoxes < 0)
	numBoxes = 0;
    }/* found one */

}/* REMOVEBOX */




/********************************************************
 * pushRectBox() - Whenever they are trying to make a
 * rectangle (or square) box call here.  Check to see
 * which button they've pressed and build the box 
 * accordingly.
*/
void pushRectBox(int button, int x, int y)
{/* PUSHERECTBOX */
  int k,origX,origY;

  /* Changed sizeFactor to recipSizeFactor below also. */
  calcSizeFactor(zoomFactor);

  /* Map this to original image coords */
  origX = (x*recipSizeFactor)+(xoffset*recipSizeFactor)+virXoffset;
  origY = (y*recipSizeFactor)+(yoffset*recipSizeFactor)+virYoffset;

  /* Branch based on which button was pressed */
  if((button == Button1) && !locked)
    {/* Upper leftahnd corner  */
      if(rectBox.drawable==FALSE)
	{/* Initthese */
	  rectBox.x= origX;
	  rectBox.y= origY;
	  rectBox.lowx = origX+1;
	  rectBox.lowy = origY+1;
	}/* Initthese */
      else
	{/* Check to make sure we've got it right */

	  if(origY > rectBox.lowy)
	    rectBox.lowy= origY;
	  else
	    rectBox.y= origY;
	  if(origX > rectBox.lowx)
	    rectBox.lowx= origX;
	  else
	    rectBox.x= origX;

	  /* Now check for the SQUARE subset of rectangle */
	  if(boxMethod==USESQUARE)
	    {/* square it up */
	      int xDiff,yDiff;
	      xDiff = (rectBox.lowx-rectBox.x);
	      yDiff = (rectBox.lowy-rectBox.y);
	      if( yDiff > xDiff )
		rectBox.lowy = rectBox.y + xDiff;
	      else if( xDiff > yDiff )
		rectBox.lowx = rectBox.x + yDiff;
	      if(makeSquareBox)
		{/* They want it a certain size */
		  rectBox.lowy = rectBox.y + makeSquareBox;
		  rectBox.lowx = rectBox.x + makeSquareBox;
		}/* They want it a certain size */
	    }/* square it up */

	}/* Check to make sure we've got it right */

      /* Now we have a valid starting point to pick a Lower
       * right from */
      rectBox.drawable= TRUE;
      rectBox.width  = rectBox.lowx - rectBox.x;
      rectBox.height = rectBox.lowy - rectBox.y;
      
    }/* Upper leftyhand corner */
  else if((button == Button2) && !locked)
    {/* Got the lower righthand or such */

      if(rectBox.drawable)
	{/* Make them start with leftmost button */
	  /* Force so that our concept of lower right 
	   * really is lower right */
	  if(origY > rectBox.y)
	    rectBox.lowy= origY;
	  else
	    {/* Reverse them */
	      rectBox.lowy= rectBox.y;
	      rectBox.y= origY;
	    }/* Reverse them */
	  if(origX > rectBox.x)
	    rectBox.lowx= origX;
	  else
	    {/* Reverse them */
	      rectBox.lowx= rectBox.x;
	      rectBox.x= origX;
	    }/* Reverse them */

	  /* Now check for the SQUARE subset of rectangle */
	  if(boxMethod==USESQUARE)
	    {/* square it up */
	      int xDiff,yDiff;
	      xDiff = (rectBox.lowx-rectBox.x);
	      yDiff = (rectBox.lowy-rectBox.y);
	      if( yDiff > xDiff )
		rectBox.y = rectBox.lowy - xDiff;
	      else if( xDiff > yDiff )
		rectBox.x = rectBox.lowx - yDiff;
	      if(makeSquareBox)
		{/* They want it a certain size */
		  rectBox.lowy = rectBox.y + makeSquareBox;
		  rectBox.lowx = rectBox.x + makeSquareBox;
		}/* They want it a certain size */
	    }/* square it up */
	}/* Make them start with leftmost button */
      
    }/* Got the lower righthand or such */
  else if((button == Button3))
    {/* Move it */
      if(rectBox.drawable)
	{/* Move it to the new location */
	  /* Make the lower right hand corner the landmark */
	  rectBox.y = origY - (rectBox.lowy - rectBox.y);
	  rectBox.x = origX - (rectBox.lowx - rectBox.x);
	  rectBox.lowy = origY;
	  rectBox.lowx = origX;
	}/* Move it to the new location */

    }/* Move it */


}/* PUSHRECTBOX */





/*******************************************************
 * pushPolyBox() - Place the given coords on the polygon 
 * stack.
 * Future goal - Try and get this polyBox stuff debugged.
 * It's my understanding that noone has used it so far
 * but it might come in handy later...???
*/
void pushPolyBox(int x, int y)
{/* PUSHPOLYBOX */
  int k;

  /* Also changed sizeFactor to recipSizeFactor below */
  calcSizeFactor(zoomFactor);

  /* Map this to original image coords */
  poly[numPoly].x = (x*recipSizeFactor)+(xoffset*recipSizeFactor)+virXoffset;
  poly[numPoly].y = (y*recipSizeFactor)+(yoffset*recipSizeFactor)+virYoffset;

  numPoly++;

}/* PUSHPOLYBOX */



/*******************************************************
 * popPolyBox() - Pop the last one off the poly stack.
*/
void popPolyBox()
{/* POPPOLYBOX */

  numPoly--;
  if(numPoly < 0)
    numPoly = 0;

}/* POPPOLYBOX */



/*******************************************************
 * removePolyBox() - Remove the given point from the
 * poly stack.
*/
void removePolyBox(int x, int y)
{/* REMOVEPOLYBOX */
  int i,j,foundIt;
  int k;

  /* Also changed sizeFactor to recipSizeFactor below */
  calcSizeFactor(zoomFactor);

  /* Map this to original image coords */
  x = (x*recipSizeFactor)+(xoffset*recipSizeFactor)+virXoffset;
  y = (y*recipSizeFactor)+(yoffset*recipSizeFactor)+virYoffset;

  /* Look for a match for these coords */
  foundIt= -99;
  for(i=0;i<numPoly;i++)
    {/* Look for a hit */
      if(((x+4) > poly[i].x && poly[i].x > (x-4)) &&
	 ((y+4) > poly[i].y && poly[i].y > (y-4)) )
	{/* Found it */
	  foundIt= i;
	}/* Found it */
    }/* Look for a hit */

  /* Found one so move others in its place */
  if(foundIt > -99)
    {/* found one */
      for(j=foundIt;j<numPoly;j++)
	{/* Move them up */
	  poly[j].x = poly[j+1].x;
	  poly[j].y = poly[j+1].y;
	}/* Move them up */
      numPoly--;
      if(numPoly < 0)
	numPoly = 0;
    }/* found one */

}/* REMOVEPOLYBOX */





/*****************************************************
 * filterFiles() - Set the file selection widget to
 * filter for the fileStr files.
*/
void filterFiles(char *filesStr, int tag)
{/* FILTERFILES */
  XmString tcs;
  XmString xmDirMask;
  char *dirMask;
  char *s1;
  static int lastFilter;

  /* We need to filter every time to make sure we 
   * show them the latest version of the file for 
   * reading OR writing.  e.g. If we wrote box coords
   * file rob.bcrd;3 before and now we're reading box 
   * coords we want to filter the selection box to show
   * rob.bcrd;3 to be read instead of the rob.bcrd;2 which
   * was the last filtering (e.g. ;2 has older data than
   * ;3).
   */
  XtSetArg(arglst[0],XmNdirMask,&xmDirMask);
  XtGetValues(fileSel_widget,arglst,1);
  
  XmStringGetLtoR(xmDirMask,XmSTRING_DEFAULT_CHARSET,&dirMask);
  sprintf(outStr,"%s",dirMask);

  if(tag != lastFilter)
    {/* Only do if we're changing file types */
      
      s1 = strrchr(outStr,'/');
      
      if(tag==ARCHIVE)
	sprintf(outStr,filesStr);
      else if(tag==DEMO)
	sprintf(outStr,filesStr);
      else if(tag==TESTARCH)
	sprintf(outStr,filesStr);
      else if(s1)
	sprintf((s1+1),filesStr);
      else sprintf(outStr,filesStr);
      
    }/* Only do if we're changing file types */
  
  createSimpleString(&tcs,outStr);  
  XmFileSelectionDoSearch(fileSel_widget, tcs);
  XFlush(dsply);
  XmStringFree(tcs); 
  XmStringFree(xmDirMask); 
  XtFree(dirMask);
  
  XFlush(dsply);

  lastFilter = tag;

}/* FILTERFILES */




/***********************************************************
 * writeColormap ()- Write the current colormap out
 * to an ASCII disk file.  NOTE:  We get a new set of
 * colormap entries everytime we run this program based
 * on what other clients are using the display.  If we
 * should run again and get fewer colormap entries than 
 * we have now we may have trouble with the colormap we've 
 * just written to disk.
 * This mechanism was used in our 8 bit display environment 
 * where we wanted multiple instances of robem running and 
 * displaying to the same display.  We just allocated the 
 * colormap in the masterProcess and then just shared the 
 * values of that colormap between robem instances.
*/
void writeColormap(char *fName)
{/* WRITECOLORMAP */
  int i,writeIt;
  FILE *fOut;
  
  /* Make sure we're not overwriting a good one */
  fOut= 0;
  writeIt= TRUE;
  fOut= fopen(fName,"r");
  if(fOut)
    {/* Protect against overwriting it */
      char infoStr[512];
      sprintf(infoStr,"[%s] Already exists - OVERWRITE it?",fName);
      writeIt= askThem(infoStr,TRUE,XmDIALOG_QUESTION);
      fclose(fOut);
    }/* Protect against overwriting it */

  /* OK let's write */
  if(writeIt)
    {/* OK to write */
      fOut= 0;
      fOut= fopen(fName,"w");
      if(fOut)
	{/* Write it */
	  for(i=0;i<256;i++)
	    {/* Write the current colormap */
	      fprintf(fOut,"%d %d %d %d %d\n",
		     (color_defs[i].pixel),
		     (color_defs[i].red),
		     (color_defs[i].green),
		     (color_defs[i].blue),
		     (color_defs[i].flags));
	    }/* Write the current colormap */
	}/* Write it */
      else askThem("Can't open file.",TRUE,XmDIALOG_INFORMATION);
    }/* OK to write */
  fclose(fOut);

}/* WRITECOLORMAP */





/************************************************************
 * readColormap ()- Read a range colormap from disk.
 * Set colors to appropriate rgb values based on radial range.
 * The disk file has format:
 *
 *        radius red green blue
 *
 * Here's an example file:
12  0     0      0
45  0     0      65000
75  0     35000  65000
111 0     65000  0
135 65000 0      0
175 35000 65000  35000
555 65000 0      35000
 *
 * In the above example up to radius 12 the colormap is set to BLACK.
 * Up to radius 45 the color is blue.  Up to 75 color greenish-blue.
 * Up to radius 111 then green etc...
 * 65535 is the MAX any color can be set to.
 *
*/
void readColormap(char *fName)
{/* READCOLORMAP */
  int i,tot,foundIt;
  int radius[512], red[512],green[512],blue[512],flags[512];
  char inStr[512];
  FILE *fIn;
  
  /* Read it */
  fIn= 0;
  fIn= fopen(fName,"r");
  if(fIn)
    {/* Read it */
      tot = 0;
      while(!feof(fIn))
	{/* Read the current colormap */
	  get_str(fIn,' ',inStr);
 	  radius[tot] = (int)strtod(inStr,0);
	  get_str(fIn,' ',inStr);
 	  red[tot] = (int)strtod(inStr,0);
	  get_str(fIn,' ',inStr);
 	  green[tot] = (int)strtod(inStr,0);
	  get_str(fIn,'\n',inStr);
 	  blue[tot] = (int)strtod(inStr,0);
	  flags[tot] = 7;
	  if(radius[tot] > 0)
	    tot++;
	}/* Read the current colormap */

      /* Now update colormap */
      for(i=0;i<256;i++)
	{/* Read the current colormap */
	  foundIt = -99;
	  for(j=0;j<tot;j++)
	    {/* search all */
	      if(i < radius[j])
		{/* Do this one */
		  foundIt = j;
		  break;
		}/* Do this one */
	    }/* search all */
	  if(foundIt >= 0)
	    {/* ok */
	      color_defs[i].pixel = i;
	      color_defs[i].red = red[foundIt];
	      color_defs[i].green = green[foundIt];
	      color_defs[i].blue = blue[foundIt];
	      color_defs[i].flags = flags[foundIt];
	    }/* ok */
	}/* Read the current colormap */



      /* Set the current colormap on the display */
      if(masterProcess)
	for(i=6;i<tryColors;i++)
	  {
	    if(hiColor)
	      {XAllocColor (dsply, color_map, &color_defs[i]);
	       pixels_return[i] = color_defs[i].pixel;}
	    else
	      XStoreColor (dsply, color_map, &color_defs[pixels_return[i]]);
	  }
      /* Update display */
      if(fileType==PACKEDPORTABLE)
	stuffPackedImage(purdueHeader.numCols,
			 purdueHeader.numRows,
			 packImg,purdueHeader.nimg,
			 imgNum);  /* Start at zero'th image */
      else if(fileType==MAP)
	stuffMapImage((float *)map_image_buffer, 
		      (float *)float_image_buffer1, 
		      file_width,file_height,
		      nsecs,
		      imgNum, 0);  /* Start at zero'th image */
      
      DrawWindow(); 
    }/* Read it */
  else askThem("Can't open file.",TRUE,XmDIALOG_INFORMATION);
  fclose(fIn);

}/* READCOLORMAP */




/******************************************************
 * pushColormap ()- Save the current colormap in case 
 * we've got to restore it later.
*/
void pushColormap()
{/* PUSHCOLORMAP */
  int i;
  
  for(i=0;i<256;i++)
    {/* Remember the current colormap */
      mirror_color_defs[i].pixel = color_defs[i].pixel;
      mirror_color_defs[i].red   = color_defs[i].red;
      mirror_color_defs[i].green = color_defs[i].green;
      mirror_color_defs[i].blue  = color_defs[i].blue;
      mirror_color_defs[i].flags = color_defs[i].flags;
    }/* Remember the current colormap */

}/* PUSHCOLORMAP */




/******************************************************
* popColormap ()- Restore saved colormap.
*/
void popColormap()
{/* POPCOLORMAP */
  int i;

  for(i=0;i<256;i++)
    {/* Remember the old current colormap */
      color_defs[i].pixel = mirror_color_defs[i].pixel;
      color_defs[i].red   = mirror_color_defs[i].red;
      color_defs[i].green = mirror_color_defs[i].green;
      color_defs[i].blue  = mirror_color_defs[i].blue;
      color_defs[i].flags = mirror_color_defs[i].flags;
    }/* Remember the old current colormap */

  if(masterProcess)
    for(i=6;i<tryColors;i++)
      {/* Stuff the display with it */
	if(hiColor)
	  {XAllocColor (dsply, color_map, &color_defs[i]);
	   pixels_return[i] = color_defs[i].pixel;}
	else
	  XStoreColor (dsply, color_map, &color_defs[pixels_return[i]]);
      }/* Stuff the display with it */

  /* Now redraw the histogram */
  histogram_activate(&minP, &maxP);

}/* POPCOLORMAP */





/* @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
 * pointerLabelDraw() - When they navigate the cursor over
 * the image displayed this rtn gives them feedback as to
 * what the values for the underlying pixel are.
*/
void pointerLabelDraw(int x, int y)
{/* POINTERLABELDRAW */
  static int lastY= -999;
  int i, j, k;
  float scaleFactor;
  int thisX,thisY,rectX,rectY;
  float grayLevel;
  XmString simpleString;
  int thisImg,sectY,sectX;

  if(inited)
    {/* Don't do this until we're up and running */
      /* Show them where they've started */
      calcSizeFactor(zoomFactor);

      /* Get these for this virtual page only */
      thisX = ((x*recipSizeFactor)+(xoffset*recipSizeFactor)) + virXoffset;
      thisY = ((y*recipSizeFactor)+(yoffset*recipSizeFactor)) + virYoffset;
      
      grayLevel = (float)findRealValue(x,y,&thisImg,&sectY,&sectX);

      if(rectBox.drawable)
	{/* OK we've got one */

	  if(rectBox.x > rectBox.lowx)
	    rectX = rectBox.x - rectBox.lowx;
	  else rectX = rectBox.lowx - rectBox.x;
	  if(rectBox.y > rectBox.lowy)
	    rectY = rectBox.y - rectBox.lowy;
	  else rectY = rectBox.lowy - rectBox.y;

	  sprintf(outStr,"R[%dx%d]  %d,%d",
		  rectBox.width,
		  rectBox.height,
		  sectX,sectY);
	}/* OK we've got one */
      else sprintf(outStr,"x,y,z[%3d,%3d,%3d] [%f]",
		   sectX,sectY,thisImg,grayLevel);
      
      if(fileType!=PACKEDPORTABLE)
	drawPointScreen(outStr,1);

      if(fileType==PACKEDPORTABLE)
	{/* Do for packed boxes */
	  sprintf(outStr,"Pixel Value [%f]",grayLevel);
	  drawPointScreen(outStr,1);

	  sprintf(outStr,"Image[%d] col,row[%d,%d]",
		  thisImg,sectX,sectY);
	  drawPointScreen(outStr,4);
	}/* Do for packed boxes */


      if(fftSize==1024 && disp1024fft &&XtIsManaged(fftscreen))
	{/* give 1024 fft feedback */
	  int hereX,hereY;
	  float thisRadius,apart,bpart,amp,phase;

	  hereX = thisX - 512.0;
	  hereY = 0.0 - (thisY - 512.0);

	  if(fftImage_gold)
	    {/* OK to do */
	      fft_point_(fftImage_gold,
			 &fftSize,&fftSize,
			 &hereX,&hereY,&thisRadius,&apart,&bpart);
	    }/* OK to do */
	  if(x != 512)
	    sprintf(outStr,"x[%d]y[%d] radius[%.0f] Angs[%.2f]",
		    hereX,hereY,thisRadius,
		    ((1024.0/((float)(x-512.0)/(float)zoomFactor))*
		     fftAngsPerPixel));
	      else
		sprintf(outStr,"x[%d]y[%d] radius[%.0f] Angs[undef]",
			hereX,hereY,thisRadius);
	  drawPointScreen(outStr,2);

	  sprintf(outStr,"a[%.2E] b[%.2E]",
		  apart,bpart);
	  drawPointScreen(outStr,3);

	  convert_abpart_(&apart,&bpart,&amp,&phase);
	  sprintf(outStr,"amp[%.2E] phase[%.2E]",
		  amp,phase);
	  drawPointScreen(outStr,4);

	}/* give 1024 fft feedback */

    }/* Don't do this until we're up and running */
  
}/* POINTERLABELDRAW */





/*************************************************************
 * stuffDisplayBoxBuffer()- For memory conservation reasons
 * we've got to get these off the disk.
 * We assume now that all data are in memory already so no
 * reading of the disk required.
*/
void stuffDisplayBoxBuffer(int thisBox, char *filename, 
			   int thisZoomFactor)
{/* stuffDisplayBoxBUFFER */
  int l,k,outHeight,outWidth,logicalRecordLength,yOff;
  int yOut,yIn,xIn,beginY,beginX,min,max,endX,endY,thisBoxRadius;
  int thisPixel;
  FILE *fIn;

  /* OK let's read our box and calc the backgorund */
  if(numBoxes > 0)
    {/* OK to continue */

      thisBoxRadius = featherRadius;

      /* Init histogram and point to our integer and display 
       * buffers */
      for(y=0;y<257;y++)
	histogram[y]= 0;
      
      min= 99999999;
      max = 0;
      min = INT_MAX;  /* dgm INT_MAX instead of 999999 */
      max = INT_MIN;  /* dgm INT_MAX instead of 0 */
      scaleBoxFactor = (((float)tryColors - 7.0) / 256.0);

      calcBoxSizeFactor(thisZoomFactor);

      /* Init display buffer */
      if(hiColor)
	{/* hi color */
	  for(k=0;k<(BOX_HEIGHT);k++)
	    for(l=0;l<BOX_WIDTH;l++)
	      XPutPixel(editBox_image1, l, k, 
			pixels_return[BLACK]
			);

	}/* hi color */
      else
	{/* normal */
	for(i=0;i<(BOX_HEIGHT*4);i++)
	  {/* Init them */
	    memcpy((&editBox_display_buffer[(i*BOX_WIDTH)]),
		   stufBuf,BOX_WIDTH);
	  }/* Init them */
	}/* normal */
      
      /* Get dimensions */
      outHeight= outWidth = ((thisBoxRadius*2)+1)*sizeFactor;
      if((outHeight > BOX_HEIGHT) ||
	 (outWidth > BOX_WIDTH))
	{/* Too large */
	  askThem(
		  "[STUFFDISPLAYBOXBUFFER] Error - box too large for buffer.",
		  FALSE,XmDIALOG_INFORMATION);
	  goto noStuff;
	}/* Too large */
      
      outHeight = MAX(0,MIN(outHeight,BOX_HEIGHT));
      outWidth = MAX(0,MIN(outWidth,BOX_WIDTH));
      
      beginY = box[thisBox].y - thisBoxRadius;
      beginX = box[thisBox].x - thisBoxRadius;
      endY   = box[thisBox].y + thisBoxRadius;
      endX   = box[thisBox].x + thisBoxRadius;
      if(beginY <= 0 ||
	 beginX <= 0 ||
	 endX >= file_width ||
	 endY >= file_height)
	{/* Box is outside of file */
	  if(hiColor)
	    {
	      for(k=0;k<BOX_HEIGHT;k++)
		for(l=0;l<BOX_WIDTH;l++)
		  editBox_mirror_buffer[((k*BOX_WIDTH)+l)]= 
		    editBox_undo_buffer[((k*BOX_WIDTH)+l)]= 
		    (float)BLACK;
	      /* Mark bad box with and X */
	      l = 0;
	      for(k=0;k<outHeight;k++)
		{
		  editBox_mirror_buffer[((k*BOX_WIDTH)+l)]= 
		    editBox_undo_buffer[((k*BOX_WIDTH)+l)]= 
		    128.0;
		  l++;
		}
	      l = outWidth - 1;
	      for(k=0;k<outHeight;k++)
		{
		  editBox_mirror_buffer[((k*BOX_WIDTH)+l)]= 
		    editBox_undo_buffer[((k*BOX_WIDTH)+l)]= 
		    128.0;
		  l--;
		}
	    }
	  else
	    {
	      for(k=0;k<BOX_HEIGHT;k++)
		for(l=0;l<BOX_WIDTH;l++)
		  editBox_mirror_buffer[((k*BOX_WIDTH)+l)]= 
		    editBox_undo_buffer[((k*BOX_WIDTH)+l)]= 
		      (float)pixels_return[BLACK];
	    }
	  sprintf(outStr,"Box number [%d] is outside of file ",thisBox+1);
	  askThem(outStr,FALSE,XmDIALOG_INFORMATION);
	  goto noStuff;
	}/* Box is outside of file */

      if(!box[thisBox].inMemory)
	{/* Read it afresh from the disk */
	  /* Open the disk file */
	  fIn= NULL;
	  fIn= fopen(filename,"r");
	  if(fIn==NULL)
	    {/* Error */
	      askThem("[STUFFDISPLAYBOXBUFFER] Error reading file.",
		      FALSE,XmDIALOG_INFORMATION);
	      goto noStuff;
	    }/* Error */
	  
	  /* Read off the disk */
	  logicalRecordLength = BOX_WIDTH;
	  beginY = box[thisBox].y - thisBoxRadius;
	  beginX = box[thisBox].x - thisBoxRadius;
	  for(k=0;k<outHeight;k++)
	    {/* Do the y direction */
	      
	      /* Use the +k because of EOR marker * 2.0 because it's 
	       * shorts not chars */
	      if(fileType == PORTABLE)
		{/* Don't use the EOR markers */
		  yOff = headOff + 
		    (2.0 * (
			    ((beginY + (int)(k*recipSizeFactor)) * 
			     file_width) +  
			    /* Y direction*/
			    beginX                            /* X */
			    ));
		}/* Don't use the EOR markers */
	      else if(fileType==BYTEPORTABLE)
		{/* Don't use the EOR markers */
		  yOff = headOff + 
		    (1.0 * (
			    ((beginY + (int)(k*recipSizeFactor)) * 
			     file_width) +  
			    /* Y direction*/
			    beginX                            /* X */
			    ));
		}/* Don't use the EOR markers */
	      else
		{/* DO use the EOR markers */
		  yOff = headOff + 
		    (2.0 * (
			    ((beginY  + (int)(k*recipSizeFactor)) * 
			     file_width) +
			    /* Y direction*/
			    beginX +                            /* X */
			    (beginY + (int)(k*recipSizeFactor)) 
			                   /* EOR markers */
			    ));
		}/* DO use the EOR markers */
	      /* SEEK it */
	      sysError= fseek(fIn,yOff,0);
	      if(sysError)
		{/* Can't go on */
		  int k2,l2;
		  askThem("[SDB seek] Error reading file.",
			  TRUE,XmDIALOG_INFORMATION);
		  /* Problems */
		  for(k2=0;k2<outHeight;k2++)
		    {/* Do the y direction */
		      int yOut;
		      yOut = (k2 * BOX_WIDTH);
		      /* Stuff all > radius in background character */
		      for(l2=0;l2<outWidth;l2++)
			{/* Stuff the outbuffer */
			  editBox_int_buffer[(yOut+(l2))]= 
			    editBox_mirror_buffer[(yOut+(l2))]= 
			      editBox_undo_buffer[(yOut+(l2))]= 
				editBox_display_buffer[(yOut+(l2))]=  0.0;
			}/* Stuff the outbuffer */
		    }/* Do the y direction */
		  fclose(fIn);
		  goto noStuff;
		}/* Can't go on */
	      
	      if((fileType==BYTEPORTABLE)||(fileType==PORTABLE))
		{/* Read byte data */
		  int thisI;
		  int thisYoff;
		  thisYoff = 
		    (1.0 * (
			    ((beginY + (int)(k*recipSizeFactor)) * 
			     file_width) +  
			    /* Y direction*/
			    beginX                            /* X */
			    ));
		  for(thisI=0;thisI<logicalRecordLength;thisI++)
		    tempBox_int_buffer[thisI] =
		      inCore_buffer[(thisYoff)+thisI];
		  
		  /* Topmost point of box */
		  yOut = (k * BOX_WIDTH);
		  /* Stuff display and mirror buffers */
		  for(l=0;l<outWidth;l++)
		    {/* Stuff the outbuffer */
		      /* Leftmost point in box */
		      xIn = (l * recipSizeFactor);
		      editBox_int_buffer[(yOut+(l))]= 
			editBox_undo_buffer[(yOut+l)]= 
			editBox_mirror_buffer[(yOut+l)]= 
			(float)tempBox_int_buffer[xIn];
		      thisPixel= (int)tempBox_int_buffer[xIn];
		      min = MIN(thisPixel,min);
		      max = MAX(thisPixel,max);
		    }/* Stuff the outbuffer */
		}/* Read byte data */
	      else
		{/* error */
		  askThem("[SDB fileType] I don't understand this type.",
			  FALSE,XmDIALOG_INFORMATION);
		  fclose(fIn);
		  goto noStuff;
		}/* error */
	      
	    }/* Do the y direction */
	  
	  fclose(fIn);
	}/* Read it afresh from the disk */
      else
	{/* We should already have it because it's not deleted yet */
	  float *thisBoxBuf;
	  int yIn,xIn,thisHeight,thisWidth;

	  thisBoxBuf = box[thisBox].edit_mirror_buffer;

	  /* Get dimensions */
	  thisWidth = (thisBoxRadius*2)+1;
	  outHeight= outWidth = ((thisBoxRadius*2)+1)*sizeFactor;
	  outHeight = MAX(0,MIN(outHeight,BOX_HEIGHT));
	  outWidth = MAX(0,MIN(outWidth,BOX_WIDTH));

	  for(i=0;i<outHeight;i++)
	    {/* Rows */
	      yIn = i * recipSizeFactor;
	      yOut = (i * BOX_WIDTH);
	      for(j=0;j<outWidth;j++)
		{/* Cols */
		  xIn = j * recipSizeFactor;
		  editBox_mirror_buffer[(yOut+j)]= 
		    editBox_undo_buffer[(yOut+j)]= 
		    editBox_int_buffer[(yOut+j)] = 
		    (float)thisBoxBuf[((yIn*thisWidth)+xIn)];
		  thisPixel= (int)thisBoxBuf[((yIn*thisWidth)+xIn)];
		  min = MIN(thisPixel,min);
		  max = MAX(thisPixel,max);
		}/* Cols */
	    }/* Rows */
	}/* We should already have it because it's not deleted yet */
      

      calcCircleStats(editBox_mirror_buffer,
		      &purdueHeader.img_bkg, 
		      &purdueHeader.img_min,
		      &purdueHeader.img_max, 
		      &purdueHeader.img_avg, 
		      &purdueHeader.img_std,
		      outWidth, outHeight, 
		      BOX_WIDTH);

      backChar = (float)purdueHeader.img_bkg;
      appodize_it(editBox_int_buffer, 
		  outWidth, 
		  outHeight, BOX_WIDTH,
		  fileType,
		  bkgrndMethod,
		  boxMethod,
		  packRadius,
		  boxRadius,
		  featherRadius,
		  recipSizeFactor,
		  sizeFactor,
		  tryColors,
		  featherMode,
		  &backChar,
		  thisBoxRadius,
		  forceAvgBkg,
		  purdueHeader.img_avg
		 );

      /* Stuff appropriate buffers */
      for(k=0;k<outHeight;k++)
	{/* Do the y direction */
	  int yOut,yOut2;
	  /* Where is the center? */
	  yOut  = (k * BOX_WIDTH);
	  yOut2 = (k * outWidth);
	  for(l=0;l<outWidth;l++)
	    {/* Stuff the outbuffer */
	      editBox_mirror_buffer[(yOut+(l))]= 
		editBox_undo_buffer[(yOut+(l))]= 
		  (float)editBox_int_buffer[(yOut+(l))];
	    }/* Stuff the outbuffer */
	}/* Do the y direction */


      /* OK let's now fit this image to our colormap */
      scaleBoxFactor = scaleBoxFactor * (255.0 / max);
      printf("[%d]  max==[%f]  bkg[%f]\n", thisBox, (float)max, backChar);
      for(i=0;i<outHeight;i++)
	{/* Rows */
	  yIn = i * recipSizeFactor;
	  yOut = (i * BOX_WIDTH);
	  for(j=0;j<outWidth;j++)
	    {/* Cols */
	      xIn = j * recipSizeFactor;
	      if(hiColor)
		{
		  editBox_display_buffer[(yOut+j)] = 
		    (BYTE)(((float)editBox_int_buffer[(
						       yOut+j)] * 
			    scaleBoxFactor) + 6.0);
		  XPutPixel(editBox_image1, j, i, 
			    pixels_return[(int)(((float)
				       editBox_int_buffer[(yOut+j)] * 
				scaleBoxFactor) + 6.0)]
			  );
	      }
	      else
		{
		  editBox_display_buffer[(yOut+j)] = 
		    pixels_return[(int)(((float)editBox_int_buffer[(
							    yOut+j)] * 
				scaleBoxFactor) + 6.0)];

		  XPutPixel(editBox_image1, l, k, 
			    editBox_display_buffer[(yOut+j)] ); /* Why l,k ? */
		}

	      histogram[(editBox_display_buffer[(yOut+j)])]+= 1;
	    }/* Cols */
	}/* Rows */

      /* Now let's make sure our backDispChar is right */
      if(hiColor)
	backDispChar = (int)(((float)backChar * 
					    scaleBoxFactor) + 6.0);
      else
	backDispChar = pixels_return[(int)(((float)backChar * 
					    scaleBoxFactor) + 6.0)];

      histogram_activate(&minP, &maxP);

    }/* OK to continue */
  
 noStuff:;
  
}/* stuffDisplayBoxBUFFER */





/*************************************************************
 * stuffRectangleBuffer()- For memory conservation reasons
 * we've got to get these off the disk.
 * Same - assume memory holds all data now - no disk read.
*/
void stuffRectangleBuffer(float *outBuf, char *filename)
{/* STUFFRECTANGLEBUFFER */
  int l,k,outHeight,outWidth,logicalRecordLength,yOff;
  int yOut,yIn,xIn,beginY,beginX;
  float sizeFactor,recipSizeFactor;
  FILE *fIn;

  if(blindWriting)
    {/* Just get from memory */
      float *thisImg;
      thisImg = packImg[(imgNum-1)];
      outWidth  = purdueHeader.numCols;
      outHeight = purdueHeader.numRows;
      for(k=0;k<outHeight;k++)
	{/* rows */
	  yOut = (k * outWidth);
	  for(l=0;l<outWidth;l++)
	    {/* Stuff the outbuffer */
	      /* Leftmost point in box */
	      outBuf[(yOut+l)]= 
		(float)thisImg[yOut+l];
	    }/* Stuff the outbuffer */
	}/* rows */
      goto blindWrite;
    }/* Just get from memory */

  /* NO zoom */
  recipSizeFactor= 1.0;
  sizeFactor= 1.0;;

  outHeight= rectBox.lowy - rectBox.y;
  outWidth = rectBox.lowx - rectBox.x;
  rectBox.width  = rectBox.lowx - rectBox.x;
  rectBox.height = rectBox.lowy - rectBox.y;

  if(outWidth > MAXFILESIZE)
    {/* Too large!! */
      askThem("[STUFFRECTANGLEBUFFER] Box too large.",
	      FALSE,XmDIALOG_INFORMATION);
      goto noStuff;
    }/* Too large!! */


  /* Read off the disk */
  logicalRecordLength = outWidth;
  beginY = rectBox.y;
  beginX = rectBox.x;
  for(k=0;k<outHeight;k++)
    {/* Do the y direction */
      
      /* Use the +k because of EOR marker * 2.0 because it's 
       * shorts not chars */
      if(fileType == PORTABLE)
	{/* Don't use the EOR markers */
	  /* Use yOff to find int offset in inCore_buffer */
	  yOff = 
	    (1.0 * (
		    ((beginY + (int)(k*recipSizeFactor)) * file_width) +  
		    /* Y direction*/
		    beginX                            /* X */
		    ));
	}/* Don't use the EOR markers */
      else if(fileType == BYTEPORTABLE)
	{/* Don't use the EOR markers */
	  /* headOff = 0;  Use yOff to find int offset in inCore_buffer */
	  yOff = 
	    (1.0 * (
		    ((beginY + (int)(k*recipSizeFactor)) * file_width) +  
		    /* Y direction*/
		    beginX                            /* X */
		    ));
	}/* Don't use the EOR markers */
      else if(crysDisping)
	{/* Just get the disply buffer - we're crysDisping a packed box */
	  yOff = 0;
	}/* Just get the disply buffer - we're crysDisping a packed box */
      else
	{/* DO use the EOR markers */
	  /* headOff = 0;  Use yOff to find int offset in inCore_buffer */
	  yOff = 
	    /* ORIGway   (2.0 * ( */
	    (1.0 * (
		    ((beginY  + (int)(k*recipSizeFactor)) * file_width) +
		    /* Y direction*/
		    beginX +                            /* X */
		    (beginY + (int)(k*recipSizeFactor)) /* EOR markers */
		    ));
	}/* DO use the EOR markers */

      if(crysDisping)
	{/* Just get the disply buffer - we're crysDisping a packed box */
	}/* Just get the disply buffer - we're crysDisping a packed box */
      else
	{/* SEEK it */
	}/* SEEK it */

      if((fileType==BYTEPORTABLE)||(fileType==PORTABLE))
	{/* Read shorts from inCore_buffer */
	  /* Topmost point of box */
	  yOut = (k * outWidth);
	  /* Stuff display and mirrro buffers */
	  for(l=0;l<outWidth;l++)
	    {/* Stuff the outbuffer */
	      /* Leftmost point in box */
	      xIn = (l * recipSizeFactor);
	      outBuf[(yOut+l)]= 
		inCore_buffer[yOff+l];
	    }/* Stuff the outbuffer */
	}/* Read shorts from inCore_buffer */
      else if(crysDisping)
	{/* Just get the disply buffer - we're crysDisping a packed box */

	  for(l=0;l<logicalRecordLength;l++)
	    tempBox_buffer[l] = (float)image_buffer1[((k+rectBox.y)*
						      buffer_width)+
						     (rectBox.x) + l];
	  /* Topmost point of box */
	  yOut = (k * outWidth);
	  /* Stuff display and mirrro buffers */
	  for(l=0;l<outWidth;l++)
	    {/* Stuff the outbuffer */
	      /* Leftmost point in box */
	      xIn = (l);
	      outBuf[(yOut+l)]= 
		(float)tempBox_buffer[xIn];
	    }/* Stuff the outbuffer */
	}/* Just get the disply buffer - we're crysDisping a packed box */
      else
	{/* error */
	  askThem("[SRB fileType] I don't understand this type.",
		  FALSE,XmDIALOG_INFORMATION);
	  goto noStuff;
	}/* error */
      
    }/* Do the y direction */
  
 noStuff:;

 blindWrite:;

}/* STUFFRECTANGLEBUFFER */




/*************************************************************
 * stuffDisplayMapBuffer()- Put the MAP section number passed
 * into editBox_display_buffer which will be put on the screen.
*/
void stuffDisplayMapBuffer(int thisSec, int thisZoomFactor, int boxOffset)
{/* stuffDisplayMapBuffer */
  int i,j,l,k,thisNumCols,thisNumRows;
  int y,x,lastRow,lastCol,thisOffset,saveBoxOffset;
  float *yMirror,*yMirror2;
  float thisPixel,offset;
  float *q;
  float thisPixelF;
  float *q2,*thisMap,*yMirrorF;

  calcBoxSizeFactor(thisZoomFactor);

  if(liteSwitch)
    {saveBoxOffset = boxOffset;
      boxOffset = 1;}

  if(boxOffset==0)
    {/* OK to init */
      if(hiColor)
	{/* hi color */
	  for(k=0;k<(BOX_HEIGHT);k++)
	    for(l=0;l<BOX_WIDTH;l++)
	      XPutPixel(editBox_image1, l, k, 
			pixels_return[BLACK]
			);

	}/* hi color */
      else
	{/* normal */
	for(i=0;i<(BOX_HEIGHT*4);i++)
	  {/* Init them */
	    memcpy((&editBox_display_buffer[(i*BOX_WIDTH)]),
		   stufBuf,BOX_WIDTH);
	  }/* Init them */
	}/* normal */
    }/* OK to init */

  if(liteSwitch)
    boxOffset = saveBoxOffset;

  q= (float *)float_image_buffer1; 

  /* Convert float to char within our colormap and put it
   * in the display buffer(s) */
  /* Do only this section */
  l= thisSec;
  if(minMapPixel < 0.0)
    offset= (0.0 - minMapPixel);
  else offset= 0.0; 

  /* Find the zoomed col/row values */
  thisNumCols = numCols;
  thisNumRows = numRows;
  if(shadedMaps && !dispSections)
    {/* May be an asymmetrical map */
      thisNumCols = largestDim;
      thisNumRows = largestDim;
    }/* May be an asymmetrical map */
  lastRow= (thisNumRows * sizeFactor);
  lastCol= (thisNumCols * sizeFactor);
  boxOffset *= sizeFactor;  /* Got to get the offset into zoomed space */
  lastRow= MIN(lastRow,BOX_HEIGHT);
  lastCol= MIN(lastCol,BOX_HEIGHT);
  
  /* Init histogram */
  for(y=0;y<257;y++)
    histogram[y]= 0;

  if(shadedMaps && !dispSections)
    {/* We've got depth cued shaded data */

      /* Point to shaded buffer */
      if(!radialProjections &&
	 !secondMapShade && 
	 !radialMapShade
	 )
	thisMap = cuedMap[l].depth;
      else
	thisMap = cuedMap[l].shad;
      
      for(y=0;y<lastRow;y++)    
	{/* Tim draws from the bottom up - We are opposite */
	  yMirrorF = (float *)(thisMap+ 
			       (((int)(y*recipSizeFactor))*thisNumCols));
	  for(x=0;x<lastCol;x++)
	    {/* Clip the entire image */

	      q2 = (float *)(yMirrorF+((int)(x*recipSizeFactor)));

	      /* We've got to scale these for now because we may get one
	       * that is larger than we've got colormap entries for.
	       */
	      if(*q2==0.0)
		thisPixelF= 0.0;
	      else 
		thisPixelF= (float)(*(q2) + offset);

	      thisOffset = ((int) ( (int) ( ((float)thisPixelF) * 
					   scaleMapFactor )   ) + 6);
	      
	      if((thisOffset >=0) && (thisOffset<256))
		{/* OK to do */
		  if(hiColor)
		    histogram[((int)thisOffset)]+= 1;
		  else
		    histogram[((int)thisOffset + pixels_return[0])]+= 1;

		  if(hiColor)
		    editBox_display_buffer[((y*BOX_WIDTH)+x+boxOffset)]=
		      (BYTE)thisOffset;
		  else
		    editBox_display_buffer[((y*BOX_WIDTH)+x+boxOffset)]=
		      (BYTE)pixels_return[thisOffset];
		    XPutPixel(editBox_image1, 
			      x+boxOffset, 
			      y,
			      pixels_return[thisOffset]);
		}/* OK to do */

	    }/* Clip the entire image */
	}/* Tim draws from the bottom up - We are opposite */

      /* Now draw the new histogram */
      histogram_activate(&minP, &maxP);

    }/* We've got depth cued shaded data */
  else
    {/* We've gto section data */
      
      for(y=0;y<lastRow;y++)    
	{/* Tim draws from the bottom up - We are opposite */
	  yMirror = (float *)(float_image_buffer1 + 
				       (l*thisNumRows*thisNumCols)+
				       (((int)(y*recipSizeFactor))*
					thisNumCols));
	  if(paintMapSwitch)
	    yMirror2 = (float *)(paintMap + 
				 (l*thisNumRows*thisNumCols)+
				 (((int)(y*recipSizeFactor))*
				  thisNumCols));
	  for(x=0;x<lastCol;x++)
	    {/* Clip the entire image */
	      
	      q = yMirror+((int)(x*recipSizeFactor));
	      
	      /* We've got to scale these for now because we may get one
	       * that is larger than we've got colormap entries for.
	       */
	      thisPixel= *(q) + offset;
	      
	      thisOffset = ((int) ( (int) ( ((float)thisPixel) * 
					   scaleMapFactor )   ) + 6);
	      
	      if((thisOffset >=0) && (thisOffset<256))
		{/* OK to do */
		  if(hiColor)
		    histogram[((int)thisOffset)]+= 1;
		  else
		    histogram[((int)thisOffset + pixels_return[0])]+= 1;

		  if(hiColor)
		    XPutPixel(editBox_image1, 
			      x+boxOffset, 
			      y,
			      pixels_return[thisOffset]);
		  editBox_display_buffer[((y*BOX_WIDTH)+x+boxOffset)]=
		    (BYTE)thisOffset;
		  
		}/* OK to do */

	      if(paintMapSwitch)
		{/* special */
		  q2 = yMirror2+((int)(x*recipSizeFactor));
		      if(normalExcise)
			{/* normal */
			  if((*q2)!= 0.0)
			    {
			      if(hiColor)
				XPutPixel(editBox_image1, 
					  x+boxOffset, 
					  y,
					  pixels_return[RED]);
			      editBox_display_buffer[((y*BOX_WIDTH)+
						      x+boxOffset)]=
						(BYTE)RED;
			    }
			}/* normal */
		      else
			{/* reverse */
			  if((*q2)== 0.0)
			    {
			      if(hiColor)
				XPutPixel(editBox_image1, 
					  x+boxOffset, 
					  y,
					  pixels_return[RED]);
				editBox_display_buffer[((y*BOX_WIDTH)+
							x+boxOffset)]=
				  (BYTE)RED;
			    }
			}/* reverse */
		}/* special */

	    }/* Clip the entire image */
	}/* Tim draws from the bottom up - We are opposite */
    }/* We've gto section data */
  
 noStuff:;

}/* stuffDisplayMapBuffer */




/*************************************************************
 * stuffDisplayPackBoxBuffer()- Put the packed box passed
 * in on the screen.
*/
void stuffDisplayPackBoxBuffer(int thisSec, int thisZoomFactor,
			       int computeAppodize)
{/* STUFFDISPLAYPACKBOXBUFFER */
  int i,j,l,k,thisOffset;
  int y,x,lastRow,lastCol,outWidth,outHeight;
  float *yMirror,*thisPackImg,*tmpPackImg;
  float thisPixel,offset;
  float *q;

  /* Put appodize stuff here ??? */
  outWidth  = purdueHeader.numCols;
  outHeight = purdueHeader.numRows;
  thisPackImg = (float *)packImg[thisSec];
  tmpPackImg = (float *)malloc(sizeof(float)*outWidth*outHeight);
  for(k=0;k<outHeight;k++)
    {/* Do the y direction */
      int yOut;
      yOut  = (k * outWidth);
      for(l=0;l<outWidth;l++)
	{/* Stuff the outbuffer */
	  tmpPackImg[(yOut+(l))] = 
	    thisPackImg[(yOut+(l))];  
	}/* Stuff the outbuffer */
    }/* Do the y direction */

  if(computeAppodize)
    {/* OK to appodize */
      int thisBoxRadius;
      thisBoxRadius = featherRadius;
      calcCircleStats(tmpPackImg,
		      &purdueHeader.img_bkg, 
		      &purdueHeader.img_min,
		      &purdueHeader.img_max, 
		      &purdueHeader.img_avg, 
		      &purdueHeader.img_std,
		      outWidth, outHeight, 
		      outWidth);
      printf("Pack Appodizing [%d][%d]  [%d][%d][%d]\n",
	     outWidth, outHeight,
	     packRadius,boxRadius,featherRadius);
      printf("featherMode==[%d]\n",featherMode);
      backChar = (float)purdueHeader.img_bkg;
      appodize_it(tmpPackImg, 
		  outWidth, 
		  outHeight, outWidth,
		  fileType,
		  bkgrndMethod,
		  boxMethod,
		  packRadius,
		  boxRadius,
		  featherRadius,
		  recipSizeFactor,
		  sizeFactor,
		  tryColors,
		  featherMode,
		  &backChar,
		  thisBoxRadius,
		  forceAvgBkg,
		  purdueHeader.img_avg
		  );
    }/* OK to appodize */
  
  calcBoxSizeFactor(thisZoomFactor);

  if(hiColor)
    {/* hi color */
      for(k=0;k<(BOX_HEIGHT);k++)
	for(l=0;l<BOX_WIDTH;l++)
	  XPutPixel(editBox_image1, l, k, 
		    pixels_return[BLACK]
		    );
      
    }/* hi color */
  else
    {/* normal */
      for(i=0;i<BOX_HEIGHT*4;i++)
	{/* Init them */
	  memcpy((&editBox_display_buffer[(i*BOX_WIDTH)]),stufBuf,BOX_WIDTH);
	}/* Init them */
    }/* normal */

  /* Convert float to char within our colormap and put it
   * in the display buffer(s) */
  /* Do only this section */
  l= thisSec;
  if(minPackPixel < 0.0)
    offset= (0.0 - minPackPixel);
  else offset= 0.0; 

  /* Find the zoomed col/row values */
  lastRow= (purdueHeader.numRows * sizeFactor);
  lastCol= (purdueHeader.numCols * sizeFactor);
  lastRow= MIN(lastRow,BOX_HEIGHT);
  lastCol= MIN(lastCol,BOX_HEIGHT);
  
  for(y=0;y<257;y++)
    histogram[y]= 0;
  for(y=0;y<lastRow;y++)    
    {/* Tim draws from the bottom up - We are opposite */

      yMirror = (float *)(tmpPackImg + 
		 ((int)(y*recipSizeFactor)*purdueHeader.numCols));
      for(x=0;x<lastCol;x++)
	{/* Clip the entire image */
	  
	  q = yMirror+((int)(x*recipSizeFactor));
	  
	  /* We've got to scale these for now because we may get one
	   * that is larger than we've got colormap entries for.
	   */
	  thisPixel= *(q) + offset;
	  
	  thisOffset = ((int) ( (int) ( ((float)thisPixel) * 
				       scalePackFactor )   ) + 6);
	  
	  thisOffset = MAX(MIN(thisOffset,tryColors-1),0);
	  if(hiColor)
	    {/* hi */
	      XPutPixel(editBox_image1, x, y, pixels_return[thisOffset]);
	      editBox_display_buffer[((y*BOX_WIDTH)+x)]=
		(BYTE)thisOffset;
	    }/* hi */
	  else
	    editBox_display_buffer[((y*BOX_WIDTH)+x)]=
	      (BYTE)pixels_return[thisOffset];
	  editBox_mirror_buffer[((y*BOX_WIDTH)+x)]=
	    editBox_undo_buffer[((y*BOX_WIDTH)+x)]=
	      (float)*q;
		
	  if(hiColor)
	    histogram[((int)thisOffset)]+= 1;
	  else
	    histogram[((int)thisOffset + pixels_return[0])]+= 1;

	}/* Clip the entire image */
    }/* Tim draws from the bottom up - We are opposite */

  backChar= (float)editBox_undo_buffer[((2*BOX_WIDTH)+2)];
      
  if(hiColor)
    backDispChar = (float)(((float)backChar * 
			  scalePackFactor) + 6.0);
  else
    backDispChar = (float)pixels_return[(int)(((float)backChar * 
					scalePackFactor) + 6.0)];

  histogram_activate(&minP, &maxP);

 noStuff:;
  free(tmpPackImg);

}/* STUFFDISPLAYPACKBOXBUFFER */





/* @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
 * handle_cr() -  This isn't really a callback
 * but it is called by a callback.  This routine changes
 * the translation table interpretations for the simple 
 * text widgets so it knows what to do when it sees a 
 * carriage return.
*/
void handle_cr(Widget widget)
{/* HANDLE_CR */
  XtTranslations parsed_t_table;

  if(widget==imgNumTextWidget)
    {/* Do translations for study text widget */

      /* Modify this widgets translation table */
      parsed_t_table = XtParseTranslationTable(rwa_translation_table);
      XtOverrideTranslations(widget, parsed_t_table);
    }/* Do translations for study text widget */
  else if(widget==boxInitWidget)
    {/* Do translations for boxInit widget */

      /* Modify this widgets translation table */
      parsed_t_table = XtParseTranslationTable(box_translation_table);
      XtOverrideTranslations(widget, parsed_t_table);
    }/* Do translations for boxInit widget */
  else if(widget==editBoxTextWidget)
    {/* Do translations for boxInit widget */

      /* Modify this widgets translation table */
      parsed_t_table = XtParseTranslationTable(editBoxText_translation_table);
      XtOverrideTranslations(widget, parsed_t_table);
    }/* Do translations for boxInit widget */
  else if(widget==aoverbTextW ||
	  widget==scaleXvalueW ||
	  widget==adimTextW ||
	  widget==bdimTextW ||
	  widget==gammaXtalTextW 
	  )
    {/* Do translations for aoverb widget */

      /* Modify this widgets translation table */
      parsed_t_table = XtParseTranslationTable(aoverb_translation_table);
      XtOverrideTranslations(widget, parsed_t_table);
    }/* Do translations for aoverb widget */
  else if(widget==map_abangTextW)
    {/* Do translations for map_abang widget */

      /* Modify this widgets translation table */
      parsed_t_table = XtParseTranslationTable(map_abang_translation_table);
      XtOverrideTranslations(widget, parsed_t_table);
    }/* Do translations for map_abang widget */
  else if(widget==fftImgNumW)
    {/* Do translations for fftImgNum widget */

      /* Modify this widgets translation table */
      parsed_t_table = XtParseTranslationTable(fftImgNum_translation_table);
      XtOverrideTranslations(widget, parsed_t_table);
    }/* Do translations for fftImgNum widget */
  else if(widget==rotfltImgNumW)
    {/* Do translations for rotfltImgNum widget */

      /* Modify this widgets translation table */
      parsed_t_table = XtParseTranslationTable(fftImgNum_translation_table);
      XtOverrideTranslations(widget, parsed_t_table);
    }/* Do translations for rotfltImgNum widget */
  else if(widget==surfDensScrollWidget)
    {/* Do translations for depthCue/radial text widget */

      /* Modify this widgets translation table */
      parsed_t_table = XtParseTranslationTable(surfDens_translation_table);
      XtOverrideTranslations(widget, parsed_t_table);
    }/* Do translations for depthCue/radial text widget */
  else if(widget==fractScrollWidget)
    {/* Do translations for depthCue/radial text widget */

      /* Modify this widgets translation table */
      parsed_t_table = XtParseTranslationTable(surfDens_translation_table);
      XtOverrideTranslations(widget, parsed_t_table);
    }/* Do translations for depthCue/radial text widget */
  else if(widget==angxScrollWidget)
    {/* Do translations for depthCue/radial text widget */

      /* Modify this widgets translation table */
      parsed_t_table = XtParseTranslationTable(surfDens_translation_table);
      XtOverrideTranslations(widget, parsed_t_table);
    }/* Do translations for depthCue/radial text widget */
  else if(widget==angyScrollWidget)
    {/* Do translations for depthCue/radial text widget */

      /* Modify this widgets translation table */
      parsed_t_table = XtParseTranslationTable(surfDens_translation_table);
      XtOverrideTranslations(widget, parsed_t_table);
    }/* Do translations for depthCue/radial text widget */
  else if(widget==phiScrollWidget)
    {/* Do translations for depthCue/radial text widget */

      /* Modify this widgets translation table */
      parsed_t_table = XtParseTranslationTable(surfDens_translation_table);
      XtOverrideTranslations(widget, parsed_t_table);
    }/* Do translations for depthCue/radial text widget */
  else if(widget==omegaScrollWidget)
    {/* Do translations for depthCue/radial text widget */

      /* Modify this widgets translation table */
      parsed_t_table = XtParseTranslationTable(surfDens_translation_table);
      XtOverrideTranslations(widget, parsed_t_table);
    }/* Do translations for depthCue/radial text widget */
  else if(widget==thetaScrollWidget)
    {/* Do translations for depthCue/radial text widget */

      /* Modify this widgets translation table */
      parsed_t_table = XtParseTranslationTable(surfDens_translation_table);
      XtOverrideTranslations(widget, parsed_t_table);
    }/* Do translations for depthCue/radial text widget */
  else if(widget==minRadiusScrollWidget)
    {/* Do translations for depthCue/radial text widget */

      /* Modify this widgets translation table */
      parsed_t_table = XtParseTranslationTable(surfDens_translation_table);
      XtOverrideTranslations(widget, parsed_t_table);
    }/* Do translations for depthCue/radial text widget */
  else if(widget==maxRadiusScrollWidget)
    {/* Do translations for depthCue/radial text widget */

      /* Modify this widgets translation table */
      parsed_t_table = XtParseTranslationTable(surfDens_translation_table);
      XtOverrideTranslations(widget, parsed_t_table);
    }/* Do translations for depthCue/radial text widget */

  else if(widget==pixelSizeScrollWidget)
    {/* Do translations for depthCue/radial text widget */

      /* Modify this widgets translation table */
      parsed_t_table = XtParseTranslationTable(surfDens_translation_table);
      XtOverrideTranslations(widget, parsed_t_table);
    }/* Do translations for depthCue/radial text widget */

  else if(widget==surfDensValueWidget)
    {/* Do translations for depthCue/radial text widget */

      /* Modify this widgets translation table */
      parsed_t_table = XtParseTranslationTable(surfDens_translation_table);
      XtOverrideTranslations(widget, parsed_t_table);
    }/* Do translations for depthCue/radial text widget */
  else if(widget==fractValueWidget)
    {/* Do translations for depthCue/radial text widget */

      /* Modify this widgets translation table */
      parsed_t_table = XtParseTranslationTable(surfDens_translation_table);
      XtOverrideTranslations(widget, parsed_t_table);
    }/* Do translations for depthCue/radial text widget */
  else if(widget==angxValueWidget)
    {/* Do translations for depthCue/radial text widget */

      /* Modify this widgets translation table */
      parsed_t_table = XtParseTranslationTable(surfDens_translation_table);
      XtOverrideTranslations(widget, parsed_t_table);
    }/* Do translations for depthCue/radial text widget */
  else if(widget==angyValueWidget)
    {/* Do translations for depthCue/radial text widget */

      /* Modify this widgets translation table */
      parsed_t_table = XtParseTranslationTable(surfDens_translation_table);
      XtOverrideTranslations(widget, parsed_t_table);
    }/* Do translations for depthCue/radial text widget */
  else if(widget==phiValueWidget || widget==omegaValueWidget)
    {/* Do translations for depthCue/radial text widget */

      /* Modify this widgets translation table */
      parsed_t_table = XtParseTranslationTable(surfDens_translation_table);
      XtOverrideTranslations(widget, parsed_t_table);
    }/* Do translations for depthCue/radial text widget */
  else if(widget==thetaValueWidget)
    {/* Do translations for depthCue/radial text widget */

      /* Modify this widgets translation table */
      parsed_t_table = XtParseTranslationTable(surfDens_translation_table);
      XtOverrideTranslations(widget, parsed_t_table);
    }/* Do translations for depthCue/radial text widget */
  else if(widget==minRadiusValueWidget)
    {/* Do translations for depthCue/radial text widget */

      /* Modify this widgets translation table */
      parsed_t_table = XtParseTranslationTable(surfDens_translation_table);
      XtOverrideTranslations(widget, parsed_t_table);
    }/* Do translations for depthCue/radial text widget */
  else if(widget==maxRadiusValueWidget)
    {/* Do translations for depthCue/radial text widget */

      /* Modify this widgets translation table */
      parsed_t_table = XtParseTranslationTable(surfDens_translation_table);
      XtOverrideTranslations(widget, parsed_t_table);
    }/* Do translations for depthCue/radial text widget */
  else if(widget==pixelSizeValueWidget)
    {/* Do translations for depthCue/radial text widget */

      /* Modify this widgets translation table */
      parsed_t_table = XtParseTranslationTable(surfDens_translation_table);
      XtOverrideTranslations(widget, parsed_t_table);
    }/* Do translations for depthCue/radial text widget */
  else if(widget==zoomTextW)
    {/* Do translations for zoom text widget */
      /* Modify this widgets translation table */
      parsed_t_table = XtParseTranslationTable(zoom_translation_table);
      XtOverrideTranslations(widget, parsed_t_table);
    }/* Do translations for zoom text widget */
  else if((widget==zoomArrowUpWidget) ||
	  (widget==zoomArrowDownWidget)	  )
    {/* Do translations for zoom arrow widget */
      /* Modify this widgets translation table */
      parsed_t_table = XtParseTranslationTable(zoomArrow_translation_table);
      XtOverrideTranslations(widget, parsed_t_table);
    }/* Do translations for zoom arrow widget */
  else if((widget==cDispZoomArrowUpW) ||
	  (widget==cDispZoomArrowDownW)	  )
    {/* Do translations for zoom arrow widget */
      /* Modify this widgets translation table */
      parsed_t_table = XtParseTranslationTable(
			       cDisp_zArrow_trans_table);
      XtOverrideTranslations(widget, parsed_t_table);
    }/* Do translations for zoom arrow widget */
  else if((widget==boxUpWidget) ||
	  (widget==boxDownWidget)	  )
    {/* Do translations for zoom arrow widget */
      /* Modify this widgets translation table */
      parsed_t_table = XtParseTranslationTable(boxNum_translation_table);
      XtOverrideTranslations(widget, parsed_t_table);
    }/* Do translations for zoom arrow widget */
  else if((widget==icosaUpWidget) ||
	  (widget==icosaDownWidget)	  )
    {/* Do translations for icosa arrow widget */
      /* Modify this widgets translation table */
      parsed_t_table = XtParseTranslationTable(icosa_translation_table);
      XtOverrideTranslations(widget, parsed_t_table);
    }/* Do translations for icosa arrow widget */
  else if((widget==paintUpWidget) ||
	  (widget==paintDownWidget)	  )
    {/* Do translations for zoom arrow widget */
      /* Modify this widgets translation table */
      parsed_t_table = XtParseTranslationTable(paintSize_translation_table);
      XtOverrideTranslations(widget, parsed_t_table);
    }/* Do translations for zoom arrow widget */
  else if(widget==versionTextWidget)
    {/* Do translations for version widget */
      /* Modify this widgets translation table */
      parsed_t_table = XtParseTranslationTable(version_translation_table);
      XtOverrideTranslations(widget, parsed_t_table);
    }/* Do translations for version widget */
  else if(widget==paramFileTextW)
    {/* Do translations for paramFile widget */
      /* Modify this widgets translation table */
      parsed_t_table = XtParseTranslationTable(paramFile_translation_table);
      XtOverrideTranslations(widget, parsed_t_table);
    }/* Do translations for paramFile widget */
  else if(widget==paramMapFileTextW)
    {/* Do translations for paramMapFile widget */
      /* Modify this widgets translation table */
      parsed_t_table = XtParseTranslationTable(paramMapFile_translation_table);
      XtOverrideTranslations(widget, parsed_t_table);
    }/* Do translations for paramMapFile widget */
  else if(widget==projasuTextW)
    {/* Do translations for projasu widget */
      /* Modify this widgets translation table */
      parsed_t_table = XtParseTranslationTable(projasu_translation_table);
      XtOverrideTranslations(widget, parsed_t_table);
    }/* Do translations for projasu widget */
  else if(widget==projContourTextW)
    {/* Do translations for projContour widget */
      /* Modify this widgets translation table */
      parsed_t_table = XtParseTranslationTable(projContour_translation_table);
      XtOverrideTranslations(widget, parsed_t_table);
    }/* Do translations for projContour widget */
  else if(widget==strtXPPW || widget==endXPPW || widget==maxYPPW)
    {/* Do translations for strt end X PP widget */
      /* Modify this widgets translation table */
      parsed_t_table = XtParseTranslationTable(strtxpp_translation_table);
      XtOverrideTranslations(widget, parsed_t_table);
    }/* Do translations for strt end X PP widget */
  else if(widget==ctfFocusMinvalueW || widget==ctfAngMajorValueW 
	  || widget==ctffocusMajValueW )
    {/* Do translations for ctf Value widget */
      /* Modify this widgets translation table */
      parsed_t_table = XtParseTranslationTable(ctfValue_translation_table);
      XtOverrideTranslations(widget, parsed_t_table);
    }/* Do translations for ctf Value widget */
  else if(widget==latticeLabel2W[0] ||
	  widget==latticeLabel2W[1] ||
	  widget==latticeLabel2W[2] ||
	  widget==latticeLabel2W[3]
	  )
    {/* Do translations for lat TEXT widget */
      /* Modify this widgets translation table */
      parsed_t_table = XtParseTranslationTable(latText_translation_table);
      XtOverrideTranslations(widget, parsed_t_table);
    }/* Do translations for lat TEXT widget */
  else if(widget==ctfPixSizeW ||
	  widget==ctfMagW ||
	  widget==ctfScanStepW
	  )
    {/* Do translations for ctf TEXT widget */
      /* Modify this widgets translation table */
      parsed_t_table = XtParseTranslationTable(ctfText_translation_table);
      XtOverrideTranslations(widget, parsed_t_table);
    }/* Do translations for ctf TEXT widget */
  else if(widget==hipTextW)
    {/* Do translations for hip TEXT widget */
      /* Modify this widgets translation table */
      parsed_t_table = XtParseTranslationTable(hipText_translation_table);
      XtOverrideTranslations(widget, parsed_t_table);
    }/* Do translations for hip TEXT widget */
  else if((widget==lopRadiusTextW) || (widget==hipRadiusTextW))
    {/* Do translations for lop TEXT widget */
      /* Modify this widgets translation table */
      parsed_t_table = XtParseTranslationTable(lohiRadiusText_trans_table);
      XtOverrideTranslations(widget, parsed_t_table);
    }/* Do translations for lop TEXT widget */
  else if(widget==lopTextW)
    {/* Do translations for lop TEXT widget */
      /* Modify this widgets translation table */
      parsed_t_table = XtParseTranslationTable(lopText_translation_table);
      XtOverrideTranslations(widget, parsed_t_table);
    }/* Do translations for lop TEXT widget */
  else if(widget==subtractFile2TextW)
    {/* Do translations for sub TEXT widget */
      /* Modify this widgets translation table */
      parsed_t_table = XtParseTranslationTable(subText_translation_table);
      XtOverrideTranslations(widget, parsed_t_table);
    }/* Do translations for sub TEXT widget */
  else if(widget==diffSecTextWidget || 
	  widget==diffSec2TextWidget ||
	  widget==magSecTextWidget ||
	  widget==subMagStartW ||
	  widget==threshTextWidget)
    {/* Do translations for diffsec TEXT widget */
      /* Modify this widgets translation table */
      parsed_t_table = XtParseTranslationTable(diffSec_translation_table);
      XtOverrideTranslations(widget, parsed_t_table);
    }/* Do translations for diffsec TEXT widget */
  else if(widget==ppMinTextW || widget==ppMaxTextW)
    {/* Do translations for pp min max widget */
      /* Modify this widgets translation table */
      parsed_t_table = XtParseTranslationTable(ppMinMax_translation_table);
      XtOverrideTranslations(widget, parsed_t_table);
    }/* Do translations for pp min max widget */
  else if(widget==binPPUpW || widget==binPPDownW)
    {/* Do translations for binPP UP/DOWN widget */
      /* Modify this widgets translation table */
      parsed_t_table = XtParseTranslationTable(binPP_translation_table);
      XtOverrideTranslations(widget, parsed_t_table);
    }/* Do translations for binPP UP/DOWN widget */
  else if(widget==stepPPUpW || widget==stepPPDownW)
    {/* Do translations for stepPP UP/DOWN widget */
      /* Modify this widgets translation table */
      parsed_t_table = XtParseTranslationTable(stepPP_translation_table);
      XtOverrideTranslations(widget, parsed_t_table);
    }/* Do translations for stepPP UP/DOWN widget */
  else if(widget==ppinUpW || widget==ppinDownW)
    {/* Do translations for ppin UP/DOWN widget */
      /* Modify this widgets translation table */
      parsed_t_table = XtParseTranslationTable(ppin_translation_table);
      XtOverrideTranslations(widget, parsed_t_table);
    }/* Do translations for ppin UP/DOWN widget */
  else if(widget==paramScanFileTextW)
    {/* Do translations for paramScanFile widget */
      /* Modify this widgets translation table */
      parsed_t_table = XtParseTranslationTable(paramScanFile_translation_table);
      XtOverrideTranslations(widget, parsed_t_table);
    }/* Do translations for paramScanFile widget */
  else if((widget==colormapScrollWidget) ||
	  (widget==colormapScrollDownWidget)	  )
    {/* Do translations for zoom arrow widget */
      /* Modify this widgets translation table */
      parsed_t_table = XtParseTranslationTable(cmScroll_translation_table);
      XtOverrideTranslations(widget, parsed_t_table);
    }/* Do translations for zoom arrow widget */
  else if((widget==gbsUpArrowW) ||
	  (widget==gbsDownArrowW)	  )
    {/* Do translations for gbs arrow widget */
      /* Modify this widgets translation table */
      parsed_t_table = XtParseTranslationTable(gbs_translation_table);
      XtOverrideTranslations(widget, parsed_t_table);
    }/* Do translations for gbs arrow widget */
  else if((widget==taUpArrowW) ||
	  (widget==taDownArrowW)	  )
    {/* Do translations for ta arrow widget */
      /* Modify this widgets translation table */
      parsed_t_table = XtParseTranslationTable(ta_translation_table);
      XtOverrideTranslations(widget, parsed_t_table);
    }/* Do translations for ta arrow widget */
  else if((widget==ta2UpArrowW) ||
	  (widget==ta2DownArrowW)	  )
    {/* Do translations for ta arrow widget */
      /* Modify this widgets translation table */
      parsed_t_table = XtParseTranslationTable(ta2_translation_table);
      XtOverrideTranslations(widget, parsed_t_table);
    }/* Do translations for ta arrow widget */
  else if((widget==boxinitArrowUpWidget) ||
	  (widget==boxinitArrowDownWidget)	  )
    {/* Do translations for zoom arrow widget */
      /* Modify this widgets translation table */
      parsed_t_table = XtParseTranslationTable(boxInit_translation_table);
      XtOverrideTranslations(widget, parsed_t_table);
    }/* Do translations for zoom arrow widget */
  else if((widget==xorigUpArrowW) ||
	  (widget==xorigDownArrowW) ||
	  (widget==yorigUpArrowW) ||
	  (widget==yorigDownArrowW)	  )
    {/* Do translations for zoom arrow widget */
      /* Modify this widgets translation table */
      parsed_t_table = XtParseTranslationTable(rotflt_translation_table);
      XtOverrideTranslations(widget, parsed_t_table);
    }/* Do translations for zoom arrow widget */
  else if(widget==phasXupW || widget==phasXdownW ||
	  widget==phasYupW || widget==phasYdownW)
    {/* Do translations for smMMax UP/DOWN widget */
      /* Modify this widgets translation table */
      parsed_t_table = XtParseTranslationTable(phase_translation_table);
      XtOverrideTranslations(widget, parsed_t_table);
    }/* Do translations for smMMax UP/DOWN widget */
  else if(widget==coravgUpArrowW || widget==coravgDownArrowW)
    {/* Do translations for coravg UP/DOWN widget */
      /* Modify this widgets translation table */
      parsed_t_table = XtParseTranslationTable(coravg_translation_table);
      XtOverrideTranslations(widget, parsed_t_table);
    }/* Do translations for coravg UP/DOWN widget */
  else if(widget==smMinRadUpW || widget==smMinRadDownW ||
	  widget==smMaxRadUpW || widget==smMaxRadDownW)
    {/* Do translations for smMMax UP/DOWN widget */
      /* Modify this widgets translation table */
      parsed_t_table = XtParseTranslationTable(smMMax_translation_table);
      XtOverrideTranslations(widget, parsed_t_table);
    }/* Do translations for smMMax UP/DOWN widget */
  else if((widget==imgnumArrowUpWidget) ||
	  (widget==imgnumArrowDownWidget)
	  )
    {/* Do translations for zoom arrow widget */
      /* Modify this widgets translation table */
      parsed_t_table = XtParseTranslationTable(imgNum_translation_table);
      XtOverrideTranslations(widget, parsed_t_table);
    }/* Do translations for zoom arrow widget */
  else if((widget==fftImgnumArrowUpW) ||
	  (widget==fftImgnumArrowDownW) ||
	  (widget==rotfltImgnumArrowUpW) ||
	  (widget==rotfltImgnumArrowDownW)
	  )
    {/* Do translations for zoom arrow widget */
      /* Modify this widgets translation table */
      parsed_t_table = XtParseTranslationTable(fftImgNumArrow_translation_table);
      XtOverrideTranslations(widget, parsed_t_table);
    }/* Do translations for zoom arrow widget */
  else if(widget==ctfCurveTextW)
    {/* Do translations for ctfCurve widget */

      /* Modify this widgets translation table */
      parsed_t_table = XtParseTranslationTable(ctfCurve_translation_table);
      XtOverrideTranslations(widget, parsed_t_table);
    }/* Do translations for ctfCurve widget */
  else if(widget==ppinTextW)
    {/* Do translations for PP Imgnum widget */

      /* Modify this widgets translation table */
      parsed_t_table = XtParseTranslationTable(ppintext_translation_table);
      XtOverrideTranslations(widget, parsed_t_table);
    }/* Do translations for PP Imgnum widget */
  else if(widget==binTextW)
    {/* Do translations for PP Bin widget */

      /* Modify this widgets translation table */
      parsed_t_table = XtParseTranslationTable(binPPtext_translation_table);
      XtOverrideTranslations(widget, parsed_t_table);
    }/* Do translations for PP Bin widget */
  else if(widget==stepTextW)
    {/* Do translations for PP Step widget */

      /* Modify this widgets translation table */
      parsed_t_table = XtParseTranslationTable(stepPPtext_translation_table);
      XtOverrideTranslations(widget, parsed_t_table);
    }/* Do translations for PP Step widget */
  else if(widget==fftStepTextW ||
	  widget==xorigTextW ||
	  widget==yorigTextW 
	  )
    {/* Do translations for fftStep widget */

      /* Modify this widgets translation table */
      parsed_t_table = XtParseTranslationTable(fftStep_translation_table);
      XtOverrideTranslations(widget, parsed_t_table);
    }/* Do translations for fftStep widget */
  else if((widget==boxRadiusArrowUpWidget) ||
	  (widget==boxRadiusArrowDownWidget) ||	  
	  (widget==emptyShellArrowUpWidget) ||
	  (widget==emptyShellArrowDownWidget) ||	
	  (widget==emmaRadiusArrowUpWidget) ||
	  (widget==emmaRadiusArrowDownWidget) ||	
	  (widget==featherRadiusArrowUpWidget) ||
	  (widget==featherRadiusArrowDownWidget) 
	  )
    {/* Do translations for zoom arrow widget */
      /* Modify this widgets translation table */
      parsed_t_table = XtParseTranslationTable(radius_translation_table);
      XtOverrideTranslations(widget, parsed_t_table);
    }/* Do translations for zoom arrow widget */
  else if(
	  widget==    unblobMinScrollW ||
	  widget==    unblobMaxScrollW ||
	  widget==    unblobRadiusScrollW ||
	  widget==    unblobSizeScrollW ||
	  widget==    cutMinScrollW[0] ||
	  widget==    cutMinScrollW[1] ||
	  widget==    cutMinScrollW[2] ||
	  widget==    cutMaxScrollW[0] ||
	  widget==    cutMaxScrollW[1] ||
	  widget==    cutMaxScrollW[2] ||
	  widget==    mapShadeScreenWindow ||
	  widget==    unblobFormW ||
	  widget==    cutFormW
	  )
    {/* Do translations for depthCue/radial text widget */

      /* Modify this widgets translation table */
      parsed_t_table = XtParseTranslationTable(surfDens_translation_table);
      XtOverrideTranslations(widget, parsed_t_table);
    }/* Do translations for depthCue/radial text widget */

}/* HANDLE_CR */





/******************************************************
 * writeBoxCoords () - Write the current box coords to
 * an ASCII disk file.
*/
void writeBoxCoords(char *fName, int append)
{/* WRITEBOXCOORDS */
  int i,writeIt, newVersion,newLastBox,newNumBox,newRadius,tot,newFeather;
  char inStr[256];
  char firstChar;
  int oldX[MAXBOX],oldY[MAXBOX],status;
  FILE *fOut;
  
  if(append==APPENDCIRCLEPACKED)
    {/* Try and append it */
      fOut= 0;
      writeIt= TRUE;
      fOut= fopen(fName,"r");
      if(fOut)
	{/* Just append it  */

	  newVersion = FALSE;
	  firstChar = (char)fgetc(fOut);
	  if((firstChar == 'B'))
	    newVersion = TRUE;
	  fclose(fOut);
	  
	  fOut= fopen(fName,"r");
	  fscanf(fOut,"%d",&newLastBox);
	  fscanf(fOut,"%d",&newNumBox);
	  fscanf(fOut,"%d",&newRadius);
	  /* WHY ?? newRadius  += (int)strtod(inStr,0); */
	  if(newVersion)
	    fscanf(fOut,"%d",&newFeather);
	  if(newRadius != boxRadius)
	    {/* problems */
	      sprintf(inStr,
	   "Radius mismatch [%d][%d] - append aborted.",
		      boxRadius,newRadius);
	      status= askThem(inStr,FALSE,XmDIALOG_INFORMATION);
	      if(status)
		goto noWrite;
	    }/* problems */

	  for(i=0;i<newNumBox;i++)
	    {/* Read the current BoxCoords */
	      fscanf(fOut,"%d %d",
		     &oldX[i],
		     &oldY[i]);
	    }/* Read the current BoxCoords */
	  fclose(fOut);

	  fOut= fopen(fName,"w");
	  fprintf(fOut,"BCRD V2.00\n");
	  fprintf(fOut,"%d\n",newLastBox);
	  fprintf(fOut,"%d\n",newNumBox+numBoxes);
	  fprintf(fOut,"%d\n",boxRadius);
	  fprintf(fOut,"%d\n",featherRadius);
	  
	    for(i=0;i<newNumBox;i++)
	      {/* Read the current BoxCoords */
		fprintf(fOut,"%d %d\n",
			 oldX[i],
			 oldY[i]);
	      }/* Read the current BoxCoords */
	  fclose(fOut);

	}/* Just append it  */
    }/* Try and append it */
  else if (append==PARAMSCAN)
    {/* Just rewrite it */

      sprintf(outStr,"cp %s boxcoords.save",fName);
      system(outStr);

      writeIt = TRUE;
    }/* Just rewrite it */
  else
    {/* Don't overwrite a good one */
      fOut= 0;
      writeIt= TRUE;
      fOut= fopen(fName,"r");
      if(fOut && !autoBoxSwitch && !reBoxSwitch) 
	    /* We always can overwrite with auto(re)BoxSwitch */
	{/* Protect against overwriting it */
	  char infoStr[512];
	  sprintf(infoStr,"[%s] Already exists - OVERWRITE it?",fName);
	  writeIt= askThem(infoStr,TRUE,XmDIALOG_QUESTION);
	  fclose(fOut);
	}/* Protect against overwriting it */
    }/* Don't overwrite a good one */
  
  /* Write it */
  if(writeIt)
    {/* OK to write */
      char *radiusText;
      fOut= 0;
      if(append==APPENDCIRCLEPACKED)
	fOut= fopen(fName,"a");
      else fOut= fopen(fName,"w");
      if(fOut)
	{/* Write it */
	  if(append==APPENDCIRCLEPACKED)
	    {/* already done above */}
	  else
	    {/* Write new header */
	      fprintf(fOut,"BCRD V2.00\n");
	      fprintf(fOut,"%d\n",lastNumBox);
	      fprintf(fOut,"%d\n",numBoxes);
	      fprintf(fOut,"%d\n",boxRadius);
	      fprintf(fOut,"%d\n",featherRadius);
	    }/* Write new header */
	  
	  for(i=0;i<numBoxes;i++)
	    {/* Write the current BoxCoords */
	      fprintf(fOut,"%d %d\n",
		     box[i].x,
		     box[i].y);
	    }/* Write the current BoxCoords */
	}/* Write it */
      else askThem("Can't open file.",TRUE,XmDIALOG_INFORMATION);
    }/* OK to write */
	   
 noWrite:;
  fclose(fOut);

}/* WRITEBOXCOORDS */






/******************************************************
 * stripParens ()- Strip of unwanted parens and commas.
*/
void stripParens(char *inStr)
{/* stripParens */
  int  i,j;
  char *s1;

  s1= strrchr(inStr,')');
  if(s1)
    *s1= ' ';
  s1= strrchr(inStr,'(');
  if(s1)
    *s1= ' ';
  s1= strrchr(inStr,',');
  if(s1)
    *s1= ' ';

  s1= strchr(inStr,')');
  if(s1)
    *s1= ' ';
  s1= strchr(inStr,'(');
  if(s1)
    *s1= ' ';
  s1= strchr(inStr,',');
  if(s1)
    *s1= ' ';



}/* stripParens */



/******************************************************
 * readBoxCoords ()- Read box coords from a disk file.
*/
void readBoxCoords(char *fName)
{/* READBOXCOORDS */
  int i,emmaSwitch,newVersion;
  char firstChar;
  char inStr[320];
  FILE *fIn;
  
  /* Read it */
  fIn= 0;
  fIn= fopen(fName,"r");
  newVersion = emmaSwitch = FALSE;
  if(fIn)
    {/* Read first char */
      firstChar = (char)fgetc(fIn);
      if((firstChar == 'E'))
	emmaSwitch = TRUE;
      if((firstChar == 'B'))
	newVersion = TRUE;
      fclose(fIn);
    }/* Read first char */

  fIn= 0;
  fIn= fopen(fName,"r");
  if(fIn)
    {/* Read it */
      if(!emmaSwitch)
	{/* normal */
	  if(newVersion)
	    get_str(fIn,'\n',inStr); 	  
	  fscanf(fIn,"%d",&lastNumBox);
	  fscanf(fIn,"%d",&numBoxes);
	  fscanf(fIn,"%d",&boxRadius);
	  if(newVersion)
	    fscanf(fIn,"%d",&featherRadius);
	  for(i=0;i<numBoxes;i++)
	    {/* Read the current BoxCoords */
	      fscanf(fIn,"%d %d",
		     &box[i].x,
		     &box[i].y);
	    }/* Read the current BoxCoords */
	}/* normal */
      else
	{/* wing it */
	  /* Scan off first "E" line */
	  fscanf(fIn,"%s",inStr);
	  lastNumBox = 0;
	  numBoxes= 0;
	  boxRadius = 50;
	  while(!feof(fIn))
	    {/* Read all */

	      fscanf(fIn,"%s",inStr);
	      stripParens(inStr);
	      box[numBoxes].x = (int)strtod(inStr,0);
	      fscanf(fIn,"%s",inStr);
	      stripParens(inStr);
	      box[numBoxes].y = (int)strtod(inStr,0);
	      if(box[numBoxes].x > 0 ||box[numBoxes].y > 0 )
		numBoxes++;

	    }/* Read all */
	  numBoxes--;   /* LAst one is bogus */
	}/* wing it */

      emmaRadius    = MIN(emmaRadius,(boxRadius-1));
      if(!newVersion)
	featherRadius = MIN(featherRadius,(boxRadius));
      printf("emma[%d]box[%d]fea[%d]\n",emmaRadius,boxRadius,featherRadius);

      sprintf(outStr,"%d",((2*featherRadius)+1));
      XmTextSetString(featherRadiusTextWidget,outStr);
      sprintf(outStr,"%d",featherRadius-boxRadius);
      XmTextSetString(boxRadiusTextWidget,outStr);
      sprintf(outStr,"%d",lastNumBox+1);
      XmTextSetString(boxInitWidget, outStr);
      XFlush(dsply);
      fclose(fIn);

      raiseIt(boxScreenWidget);
      DrawWindow();

    }/* Read it */
  else askThem("Can't open file.",TRUE,XmDIALOG_INFORMATION);

}/* READBOXCOORDS */




/******************************************************
 * readState ()- Read the options state for the program
 * from an ASCII disk file.  Se the program "state"
 * based on the values read.
 * This was the thinking early on that we would be able 
 * to store and restore that state of all the program
 * variables and display widgets or at least the important 
 * ones.  Not sure if it's needed or not..
*/
void readState(char *fName)
{/* READSTATE */
  int i;
  char version[64], outStr[64];
  FILE *fIn;
  
  /* Open and read it */
  fIn= 0;
  fIn= fopen(fName,"r");
  if(fIn)
    {/* Read it */
      fscanf(fIn,"%s",version);
      fscanf(fIn,"%d",&locked);
      fscanf(fIn,"%d",&boxRadius);
    }/* Read it */
  else askThem("Can't open file.",TRUE,XmDIALOG_INFORMATION);
  fclose(fIn);

  /* Now change the "state" of the program */
  /* Set the lock toggle */
  XtVaSetValues(lockWidget,
		XmNset, locked,
		NULL);
  /* Set the boxRadiusTextWidget to (un)editable */
  XtVaSetValues(boxRadiusTextWidget,
		XmNeditable, !(locked),
		NULL);

  /* New radius value */
  sprintf(outStr,"%d",featherRadius-boxRadius);
  XmTextSetString(boxRadiusTextWidget,outStr);


}/* READSTATE */



/******************************************************
 * readSF ()- Read the structure factors file and
 * convert to a MAP on the fly.
*/
void readSF(char *fName)
{/* READSTATE */
  int i;
  char version[64], outStr[64];
  FILE *fIn;
  
  askThem("SF Not implemented",FALSE,XmDIALOG_INFORMATION);

  /* Open and read it */
  fIn= 0;
  fIn= fopen(fName,"r");
  if(fIn)
    {/* Read it */



    }/* Read it */
  else askThem("Can't open file.",TRUE,XmDIALOG_INFORMATION);
  fclose(fIn);

}/* READSF */




/******************************************************
 * writeRAWcontour ()-  Write the current image in RAW
 * format.  Used to get images to the Mac e.g.
*/
void writeRAWcontour(char *fName, BYTE *byte_buffer, int ncol, int nrow)
{/* writeRAWcontour */
  int i,j,k;
  BYTE outChar;
  BYTE *outBuffer;
  int outInt,writeIt,cmapColors,offset;
  int minByte,maxByte;
  float scaleFactor;
  short outShort;
  FILE *fOut;
  BYTE *tmpByte_buffer;
  fString_t forName,rWrite;
  int imgOffset;
  forPifHead_t thisHead;
  
  writeIt= TRUE;
  fOut= NULL;
  fOut = fopen(fName,"r");
  if(fOut != NULL)
    {/*  Ooops we may be overwriting it!!! */
      sprintf(outStr,
	      "Do you really want to overwrite [%s]?",fName);
      writeIt= askThem(outStr,TRUE,XmDIALOG_QUESTION);
    }/*  Ooops we may be overwriting it!!! */
  
  if(writeIt)
    {/* OK to proceed */
      
      if(fOut!=NULL)
	fclose(fOut);
      fOut= 0;
      fOut= fopen(fName,"w");

      if(fOut)
	{/* Write it */
	  
	  CK(outBuffer= (BYTE *)malloc(sizeof(BYTE)*nrow*ncol));
	  
	  /* Stuff outbuffer with each image */
	  for(i=0;i<nrow;i++)
	    for(k=0;k<ncol;k++)
	      outBuffer[((i*ncol)+k)]=
		byte_buffer[(i*BOX_WIDTH)+k];

	  
	  /* Now find the min and max */
	  minByte = 255;
	  maxByte = 0;
	  for(j=0;j<nrow;j++)
	    for(k=0;k<ncol;k++)
	      {/* Check all bytes */
		outChar = outBuffer[((j*ncol)+k)];
		minByte = MIN(minByte,(int)outChar);
		maxByte = MAX(maxByte,(int)outChar);
	      }/* Check all bytes */
	  
	  /* Now find scale factor */
	  minByte -= 5;      /* We'll put our colors in 0-4 */
	  scaleFactor = 255.0 / (float)(maxByte-minByte);
	  
	  /* Now scale the data */
	  for(j=0;j<nrow;j++)
	    for(k=0;k<ncol;k++)
	      {/* Check all bytes */
		outChar = outBuffer[((j*ncol)+k)];
		if(hiColor)
		  {/* do for hi color */
		    if(outChar == RED)
		      outBuffer[((j*ncol)+k)]= (BYTE)2;
		    else if(outChar == GREEN)
		      outBuffer[((j*ncol)+k)]= (BYTE)3;
		    else
		    outBuffer[((j*ncol)+k)]= outChar;
		  }/* do for hi color */
		else
		  {/* do for normal */
		    if(outChar == pixels_return[RED])
		      outBuffer[((j*ncol)+k)]= (BYTE)2;
		    if(outChar == pixels_return[GREEN])
		      outBuffer[((j*ncol)+k)]= (BYTE)3;
		    else
		    outBuffer[((j*ncol)+k)]= outChar;
		  }/* do for normal */
	      }/* Check all bytes */
	  
	  
	  /* Now find the min and max */
	  minByte = 255;
	  maxByte = 0;
	  for(j=0;j<nrow;j++)
	    for(k=0;k<ncol;k++)
	      {/* Check all bytes */
		outChar = outBuffer[((j*ncol)+k)];
		minByte = MIN(minByte,(int)outChar);
		maxByte = MAX(maxByte,(int)outChar);
	      }/* Check all bytes */
	  
	  fwrite(outBuffer,sizeof(BYTE),(ncol*nrow),fOut);
	  
	  sprintf(outStr,"RAW file stats:\n\
\n\
     Filename = [%s]\n\
\n\
     Width = [%d]\n\
     Height= [%d]\n\
\n\
     8 bit byte pixels\n\
     NO header\n\
\n\
     Min[%d]\n\
     Max[%d]\n\
\n\
     RED pixel   = 2\n\
     GREEN pixel = 3\n",
		  fName,
		  (ncol),nrow,minByte,maxByte);
	  askThem(outStr,FALSE,XmDIALOG_INFORMATION);
	}/* Write it */
      else askThem("Can't open file.",TRUE,XmDIALOG_INFORMATION);
      
    }/* OK to proceed */
  
  fclose(fOut);
  
  /* Now write PIF file of the above */
  pif_init_head(&thisHead);
  thisHead.luNum = 4;
  rWrite.addr = malloc(sizeof(char)*4);
  rWrite.length = 4;
  forName.addr = malloc(sizeof(char)*264);
  forName.length = 264;
  sprintf(rWrite.addr,"w");
  sprintf(forName.addr,"rawFile.pif");

  pif_open_(forName.addr, &(thisHead.luNum), rWrite.addr);
	  
  /*******************************************
   * 
   *  WRIITING NEW GLOBAL PIF HEADER
   *
   *******************************************/
  /* [1] Write global file header */
  
  /* Pick appropriate statistics calculation based
   * on data type we have */
  thisHead.datai4_min = 0;
  thisHead.datai4_max = 255;
  thisHead.numCols = ncol;
  thisHead.numRows = nrow;

  thisHead.mode = 6;

  pif_write_gh(&thisHead);
	      
  thisHead.img_bkg    = 0;
  thisHead.packRadius = (int)(ncol/2);
  thisHead.datai4_max= 255;
  thisHead.datar4_max= (float)255.0;
  thisHead.datai4_min= 0;
  thisHead.datar4_min= (float)0.0;
  thisHead.borderWidth = 0;
  thisHead.binFactor     = 1;
  thisHead.imgNum = 1;

  pif_write_dh(&thisHead);
  thisHead.imgOffset += 512;  /* Just wrote the header */

  CK(tmpByte_buffer=(BYTE *)malloc(sizeof(BYTE)*
				   ncol*nrow));
  tsbworld_byte(outBuffer,
		tmpByte_buffer,
		ncol,
		ncol,nrow);
  pif_write_byte_image_(&(thisHead.luNum),
			tmpByte_buffer,
			&(thisHead.imgOffset),
			&(thisHead.numCols),
			&(thisHead.numRows),
			&(thisHead.numSecs));
  free(tmpByte_buffer);
  free(outBuffer);
  free(forName.addr);
  free(rWrite.addr);
  pif_close(&(thisHead.luNum));

}/* writeRAWcontour */





/******************************************************
 * writeRAW ()-  Write the current image in RAW
 * format.  Used to get images to the Mac e.g.
*/
void writeRAW(char *fName)
{/* writeRAW */
  int i,j,k,colorMapOffset,clrMapIFDoffset,imageOffset,imageIFDoffset;
  BYTE outChar;
  BYTE colr_Map[4096];
  BYTE *outBuffer;
  int outInt,writeIt,cmapColors,offset;
  int minByte,maxByte,numColsTsizeFactor,numRowsTsizeFactor;
  float scaleFactor;
  short outShort;
  tiffHead_t   tiffHead;
  tiffIFD_t    tiffIFD;
  tiffTag_t    tiffTag;
  FILE *fOut;
  int numChoices,numWanted,thisNsecs;
  float minF,maxF,scaleF;
  XmString simpleString;
  BYTE *thisImg;

  sprintf(outStr,"Write current ImgNum");
  createSimpleString(&simpleString, outStr);
  XmListAddItem(infoListW, simpleString, 1);
  XFlush(dsply);
  XmStringFree(simpleString); 
  sprintf(outStr,"Write what's visible on screen");
  createSimpleString(&simpleString, outStr);
  XmListAddItem(infoListW, simpleString, 2);
  XFlush(dsply);
  XmStringFree(simpleString); 
  sprintf(outStr,"Write entire data set");
  createSimpleString(&simpleString, outStr);
  XmListAddItem(infoListW, simpleString, 3);
  XFlush(dsply);
  XmStringFree(simpleString); 
  
  sprintf(outStr,"Choose how much to write below:");
  numChoices = 3;
  numWanted= askThemAnswer(outStr,numChoices);
  
  if(shadedMaps && !dispSections)
    {/* Write cued map */

      writeIt= TRUE;
      fOut= NULL;
      fOut = fopen(fName,"r");
      if(fOut != NULL)
	{/*  Ooops we may be overwriting it!!! */
	  sprintf(outStr,
		  "Do you really want to overwrite [%s]?",fName);
	  writeIt= askThem(outStr,TRUE,XmDIALOG_QUESTION);
	}/*  Ooops we may be overwriting it!!! */
      
      if(writeIt)
	{/* OK to proceed */
	  
	  if(fOut!=NULL)
	    fclose(fOut);
	  fOut= 0;
	  fOut= fopen(fName,"w");
	  if(fOut)
	    {/* Write it */
	      
	      calcSizeFactor(zoomFactor);

	      if(numWanted==1)
		{/* Just do a screen dump */
		  
		  numColsTsizeFactor = display_width;
		  numRowsTsizeFactor = display_height;
		  CK(outBuffer= (BYTE *)malloc(sizeof(BYTE)*
					       (numRowsTsizeFactor)*
					       (numColsTsizeFactor)));
		  minF = maxF = image_buffer1_disp[0];
		  for(i=0;i<(numRowsTsizeFactor);i++)
		    for(k=0;k<(numColsTsizeFactor);k++)
		      {
			maxF = MAX(maxF,
				   image_buffer1_disp[(i*buffer_width)+k]);
			minF = MIN(minF,
				   image_buffer1_disp[(i*buffer_width)+k]);
		      }
		  scaleF = 254.0 / (maxF-minF);
		  for(i=0;i<(numRowsTsizeFactor);i++)
		    for(k=0;k<(numColsTsizeFactor);k++)
		      outBuffer[(i*numColsTsizeFactor)+k]=
			(BYTE)(((float)image_buffer1_disp[(i*
					   buffer_width)+k]-minF)*
			       scaleF);
		  
		  /* Now find the min and max */
		  minByte = 9999;
		  maxByte = -9999;
		  minByte = INT_MAX;  /* dgm INT_MAX instead of 9999 */
		  maxByte = INT_MIN;  /* dgm INT_MIN instead of -9999 */
		  for(j=0;j<(numRowsTsizeFactor);j++)
		    for(k=0;k<numColsTsizeFactor;k++)
		      {/* Check all bytes */
			outChar = outBuffer[((j*numColsTsizeFactor)+k)];
			minByte = MIN(minByte,(int)outChar);
			maxByte = MAX(maxByte,(int)outChar);
		      }/* Check all bytes */
		  
		  /* Now find scale factor */
		  scaleFactor = 255.0 / (float)(maxByte-minByte);
		  
		  /* Now scale the data */
		  for(j=0;j<(numRowsTsizeFactor);j++)
		    for(k=0;k<numColsTsizeFactor;k++)
		      {/* Check all bytes */
			outChar = outBuffer[((j*numColsTsizeFactor)+k)];
			outBuffer[((j*numColsTsizeFactor)+k)]=
			  ((float)(outChar-minByte)*scaleFactor);
		      }/* Check all bytes */
		  
		  /* Now find the min and max */
		  minByte = 9999;
		  maxByte = -9999;
		  minByte = INT_MAX;  /* dgm INT_MAX instead of 9999 */
		  maxByte = INT_MIN;  /* dgm INT_MIN instead of -9999 */
		  for(j=0;j<numRowsTsizeFactor;j++)
		    for(k=0;k<numColsTsizeFactor;k++)
		      {/* Check all bytes */
			outChar = outBuffer[((j*numColsTsizeFactor)+k)];
			minByte = MIN(minByte,(int)outChar);
			maxByte = MAX(maxByte,(int)outChar);
		      }/* Check all bytes */
		  
		  fwrite(outBuffer,
			 sizeof(BYTE),(numRowsTsizeFactor*
				       numColsTsizeFactor),fOut);
		  
		  sprintf(outStr,"RAW file stats:\n\
\n\
     Frames= [%d]\n\
     Width = [%d]\n\
     Height= [%d]\n\
\n\
     8 bit byte pixels\n\
     NO header\n\
\n\
     Min[%d]\n\
     Max[%d]\n",1,(numColsTsizeFactor),numRowsTsizeFactor,
			  minByte,maxByte);
		  askThem(outStr,FALSE,XmDIALOG_INFORMATION);
		  
		}/* Just do a screen dump */
	      else
		{/* Do as many as wanted */
		  int firstJ,lastJ;

		  if(numWanted == 0)
		    {
		      firstJ = imgNum - 1;
		      lastJ  = firstJ + 1;
		      thisNsecs = 1;
		    }
		  else
		    {
		      firstJ= 0;
		      lastJ = nsecs;
		      thisNsecs = nsecs;
		    }

		  numColsTsizeFactor = largestDim * sizeFactor;
		  numRowsTsizeFactor = largestDim * sizeFactor;
		  CK(outBuffer= (BYTE *)malloc(sizeof(BYTE)*
					       (numRowsTsizeFactor)*
					       (numColsTsizeFactor)*
					       thisNsecs));
		  
		  for(j=firstJ;j<lastJ;j++)
		    {/* Stuff outbuffer with each image */
		      /* EditBox Zoom is off by 1.0 */
		      stuffDisplayMapBuffer(j,(int)(zoomFactor+1.0),0);
		      for(i=0;i<(numRowsTsizeFactor);i++)
			for(k=0;k<(numColsTsizeFactor);k++)
			  outBuffer[(
				     ((j-firstJ)*numColsTsizeFactor)
				     +
				     (i*numColsTsizeFactor*thisNsecs)
				     +k)]=
				       editBox_display_buffer[(i*BOX_WIDTH)+k];
		    }/* Stuff outbuffer with each image */
		  
		  /* Now find the min and max */
		  minByte = 9999;
		  maxByte = -9999;
		  minByte = INT_MAX;  /* dgm INT_MAX instead of 9999 */
		  maxByte = INT_MIN;  /* dgm INT_MIN instead of -9999 */
		  for(i=0;i<thisNsecs;i++)
		    for(j=0;j<(numRowsTsizeFactor);j++)
		      for(k=0;k<numColsTsizeFactor;k++)
			{/* Check all bytes */
			  outChar = outBuffer[((j*numColsTsizeFactor)+
					       (i*numColsTsizeFactor*
						(numRowsTsizeFactor))+k)];
			  minByte = MIN(minByte,(float)outChar);
			  maxByte = MAX(maxByte,(float)outChar);
			}/* Check all bytes */
		  
		  /* Now find scale factor */
		  scaleFactor = 254.0 / (float)(maxByte-minByte);
		  
		  /* Now scale the data */
		  for(i=0;i<thisNsecs;i++)
		    for(j=0;j<(numRowsTsizeFactor);j++)
		      for(k=0;k<numColsTsizeFactor;k++)
			{/* Check all bytes */
			  outChar = outBuffer[((j*numColsTsizeFactor)+
					       (i*numColsTsizeFactor*
						numRowsTsizeFactor)+k)];
			  outBuffer[((j*numColsTsizeFactor)+
				     (i*numColsTsizeFactor*
				      numRowsTsizeFactor)+k)]=
				       ((float)(outChar-minByte)*scaleFactor);
			}/* Check all bytes */
		  
		  
		  /* Now find the min and max */
		  minByte = 9999;
		  maxByte = -9999;
		  minByte = INT_MAX;  /* dgm INT_MAX instead of 9999 */
		  maxByte = INT_MIN;  /* dgm INT_MIN instead of -9999 */
		  for(i=0;i<thisNsecs;i++)
		    for(j=0;j<numRowsTsizeFactor;j++)
		      for(k=0;k<numColsTsizeFactor;k++)
			{/* Check all bytes */
			  outChar = outBuffer[((j*numColsTsizeFactor)+
					       (i*numColsTsizeFactor*
						numRowsTsizeFactor)+k)];
			  minByte = MIN(minByte,(int)outChar);
			  maxByte = MAX(maxByte,(int)outChar);
			}/* Check all bytes */
		  
		  fwrite(outBuffer,
			 sizeof(BYTE),(numRowsTsizeFactor*
				       numColsTsizeFactor*thisNsecs),fOut);
		  
		  sprintf(outStr,"RAW file stats:\n\
\n\
     Frames= [%d]\n\
     Width = [%d]\n\
     Height= [%d]\n\
\n\
     8 bit byte pixels\n\
     NO header\n\
\n\
     Min[%d]\n\
     Max[%d]\n",thisNsecs,(thisNsecs*numColsTsizeFactor),numRowsTsizeFactor,
			  minByte,maxByte);
		  askThem(outStr,FALSE,XmDIALOG_INFORMATION);
		  
		  free(outBuffer);
		}/* Do as many as wanted */
	      
	    }/* Write it */
	  else askThem("Can't open file.",TRUE,XmDIALOG_INFORMATION);
	  
	}/* OK to proceed */
      
      fclose(fOut);
      
    }/* Write cued map */
  else if(projASU || fileType==PACKEDPORTABLE)
    {/* PROJASU */

      writeIt= TRUE;
      fOut= NULL;
      fOut = fopen(fName,"r");
      if(fOut != NULL)
	{/*  Ooops we may be overwriting it!!! */
	  sprintf(outStr,
		  "Do you really want to overwrite [%s]?",fName);
	  writeIt= askThem(outStr,TRUE,XmDIALOG_QUESTION);
	}/*  Ooops we may be overwriting it!!! */
      
      if(writeIt)
	{/* OK to proceed */
	  
	  if(fOut!=NULL)
	    fclose(fOut);
	  fOut= 0;
	  fOut= fopen(fName,"w");
	  if(fOut)
	    {/* Write it */
	      
	      calcSizeFactor(zoomFactor);

	      if(fileType==PACKEDPORTABLE)
		{/* Init */
		  mapCols = purdueHeader.numCols;
		  mapRows = purdueHeader.numRows;
		}/* Init */

	      if(numWanted==0)
		{/* Just write the current imgNum */
		  stuffDisplayPackBoxBuffer(0, (int)(zoomFactor+1.0),FALSE);
		  numColsTsizeFactor = mapCols * sizeFactor;
		  numRowsTsizeFactor = mapRows * sizeFactor;
		  CK(outBuffer= (BYTE *)malloc(sizeof(BYTE)*
					       (numRowsTsizeFactor)*
					    (numColsTsizeFactor)));
		  
		  for(i=0;i<(numRowsTsizeFactor);i++)
		    for(k=0;k<(numColsTsizeFactor);k++)
		      outBuffer[(i*numColsTsizeFactor)+k]=
			editBox_display_buffer[(i*BOX_WIDTH)+k];

		  /* Now find the min and max */
		  minByte = 9999;
		  maxByte = -9999;
		  minByte = INT_MAX;  /* dgm INT_MAX instead of 9999 */
		  maxByte = INT_MIN;  /* dgm INT_MIN instead of -9999 */
		  for(j=0;j<(numRowsTsizeFactor);j++)
		    for(k=0;k<numColsTsizeFactor;k++)
		      {/* Check all bytes */
			outChar = outBuffer[((j*numColsTsizeFactor)+k)];
			minByte = MIN(minByte,(int)outChar);
			maxByte = MAX(maxByte,(int)outChar);
		      }/* Check all bytes */
		  
		  /* Now find scale factor */
		  scaleFactor = 255.0 / (float)(maxByte-minByte);
		  
		  /* Now scale the data */
		  for(j=0;j<(numRowsTsizeFactor);j++)
		    for(k=0;k<numColsTsizeFactor;k++)
		      {/* Check all bytes */
			outChar = outBuffer[((j*numColsTsizeFactor)+k)];
			outBuffer[((j*numColsTsizeFactor)+k)]=
			  ((float)(outChar-minByte)*scaleFactor);
		      }/* Check all bytes */
		  
		  /* Now find the min and max */
		  minByte = 9999;
		  maxByte = -9999;
		  minByte = INT_MAX;  /* dgm INT_MAX instead of 9999 */
		  maxByte = INT_MIN;  /* dgm INT_MIN instead of -9999 */
		  for(j=0;j<numRowsTsizeFactor;j++)
		    for(k=0;k<numColsTsizeFactor;k++)
		      {/* Check all bytes */
			outChar = outBuffer[((j*numColsTsizeFactor)+k)];
			minByte = MIN(minByte,(int)outChar);
			maxByte = MAX(maxByte,(int)outChar);
		      }/* Check all bytes */
		  
		  fwrite(outBuffer,
			 sizeof(BYTE),(numRowsTsizeFactor*
				       numColsTsizeFactor),fOut);
		  
		  sprintf(outStr,"RAW file stats:\n\
\n\
     Frames= [%d]\n\
     Width = [%d]\n\
     Height= [%d]\n\
\n\
     8 bit byte pixels\n\
     NO header\n\
\n\
     Min[%d]\n\
     Max[%d]\n",1,(numColsTsizeFactor),numRowsTsizeFactor,
			  minByte,maxByte);
		  askThem(outStr,FALSE,XmDIALOG_INFORMATION);
		  
		  
		}/* Just write the current imgNum */
	      else if(numWanted==1)
		{/* Just do a screen dump */

		  numColsTsizeFactor = display_width;
		  numRowsTsizeFactor = display_height;
		  CK(outBuffer= (BYTE *)malloc(sizeof(BYTE)*
					       (numRowsTsizeFactor)*
					    (numColsTsizeFactor)));
		  
		  minF = maxF = image_buffer1_disp[0];
		  for(i=0;i<(numRowsTsizeFactor);i++)
		    for(k=0;k<(numColsTsizeFactor);k++)
		      {
			maxF = MAX(maxF,
				   image_buffer1_disp[(i*buffer_width)+k]);
			minF = MIN(minF,
				   image_buffer1_disp[(i*buffer_width)+k]);
		      }
		  scaleF = 254.0 / (maxF-minF);
		  for(i=0;i<(numRowsTsizeFactor);i++)
		    for(k=0;k<(numColsTsizeFactor);k++)
		      outBuffer[(i*numColsTsizeFactor)+k]=
			(BYTE)(((float)image_buffer1_disp[(i*buffer_width)+k]-minF)
			       *scaleF);

		  /* Now find the min and max */
		  minByte = 9999;
		  maxByte = -9999;
		  minByte = INT_MAX;  /* dgm INT_MAX instead of 9999 */
		  maxByte = INT_MIN;  /* dgm INT_MIN instead of -9999 */
		  for(j=0;j<(numRowsTsizeFactor);j++)
		    for(k=0;k<numColsTsizeFactor;k++)
		      {/* Check all bytes */
			outChar = outBuffer[((j*numColsTsizeFactor)+k)];
			minByte = MIN(minByte,(int)outChar);
			maxByte = MAX(maxByte,(int)outChar);
		      }/* Check all bytes */
		  
		  /* Now find scale factor */
		  scaleFactor = 255.0 / (float)(maxByte-minByte);
		  
		  /* Now scale the data */
		  for(j=0;j<(numRowsTsizeFactor);j++)
		    for(k=0;k<numColsTsizeFactor;k++)
		      {/* Check all bytes */
			outChar = outBuffer[((j*numColsTsizeFactor)+k)];
			outBuffer[((j*numColsTsizeFactor)+k)]=
			  ((float)(outChar-minByte)*scaleFactor);
		      }/* Check all bytes */
		  
		  /* Now find the min and max */
		  minByte = 9999;
		  maxByte = -9999;
		  minByte = INT_MAX;  /* dgm INT_MAX instead of 9999 */
		  maxByte = INT_MIN;  /* dgm INT_MIN instead of -9999 */
		  for(j=0;j<numRowsTsizeFactor;j++)
		    for(k=0;k<numColsTsizeFactor;k++)
		      {/* Check all bytes */
			outChar = outBuffer[((j*numColsTsizeFactor)+k)];
			minByte = MIN(minByte,(int)outChar);
			maxByte = MAX(maxByte,(int)outChar);
		      }/* Check all bytes */
		  
		  fwrite(outBuffer,
			 sizeof(BYTE),(numRowsTsizeFactor*
				       numColsTsizeFactor),fOut);
		  
		  sprintf(outStr,"RAW file stats:\n\
\n\
     Frames= [%d]\n\
     Width = [%d]\n\
     Height= [%d]\n\
\n\
     8 bit byte pixels\n\
     NO header\n\
\n\
     Min[%d]\n\
     Max[%d]\n",1,(numColsTsizeFactor),numRowsTsizeFactor,
			  minByte,maxByte);
		  askThem(outStr,FALSE,XmDIALOG_INFORMATION);
		  
		}/* Just do a screen dump */
	      else
		{/* Dump entire set */
		  
		  if(projASU)
		    {/* Calculated projections */
		      numColsTsizeFactor = mapCols * sizeFactor;
		      numRowsTsizeFactor = mapRows * sizeFactor;
		    }/* Calculated projections */
		  else
		    {/* Just boxed data */
		      numColsTsizeFactor = purdueHeader.numCols * sizeFactor;
		      numRowsTsizeFactor = purdueHeader.numRows * sizeFactor;
		    }/* Just boxed data */
		  CK(outBuffer= (BYTE *)malloc(sizeof(BYTE)*
					       (numRowsTsizeFactor)*
					    (numColsTsizeFactor)*
					    purdueHeader.nimg));
		  
		  for(i=0;i<purdueHeader.nimg;i++)
		    {/* Do all boxes */
		      int outRow,outCol;
		      stuffDisplayPackBoxBuffer(i, (int)(zoomFactor+1.0),
						FALSE);
		      if(projASU)
			{/* Calculated projections */
			  outRow = (int)(i / maxPhi)*(numColsTsizeFactor*
						      maxPhi)
			                       *numRowsTsizeFactor;
			  outCol = (int)(i % maxPhi)*numColsTsizeFactor;
			}/* Calculated projections */
		      else
			{/* Just boxed data */
			  outRow = (int)0;
			  outCol = (int)(i)*numColsTsizeFactor;
			}/* Just boxed data */
		      for(j=0;j<(numRowsTsizeFactor);j++)
			{/* Rows */
			  for(k=0;k<(numColsTsizeFactor);k++)
			    {/* Cols */
			      if(projASU)
				outBuffer[(outRow+outCol)+
					(j*numColsTsizeFactor*maxPhi)+k]=
					  editBox_display_buffer[(j*BOX_WIDTH)
								 +k];
			      else
				outBuffer[(outRow+outCol)+
					(j*numColsTsizeFactor*
					 purdueHeader.nimg)+k]=
					  editBox_display_buffer[(j*BOX_WIDTH)
								 +k];
			    }/* Cols */
			}/* Rows */
		    }/* Do all boxes */

		  /* Now find the min and max */
		  minByte = 9999;
		  maxByte = -9999;
		  minByte = INT_MAX;  /* dgm INT_MAX instead of 9999 */
		  maxByte = INT_MIN;  /* dgm INT_MIN instead of -9999 */
		  for(i=0;i<purdueHeader.nimg;i++)
		    {/* Do all boxes */
		      for(j=0;j<(numRowsTsizeFactor);j++)
			for(k=0;k<numColsTsizeFactor;k++)
			  {/* Check all bytes */
			    outChar = outBuffer[((i*numColsTsizeFactor*
						  numRowsTsizeFactor)+
						 (j*numColsTsizeFactor)+k)];
			    minByte = MIN(minByte,(int)outChar);
			    maxByte = MAX(maxByte,(int)outChar);
			  }/* Check all bytes */
		    }/* Do all boxes */
		  
		  /* Now find scale factor */
		  scaleFactor = 255.0 / (float)(maxByte-minByte);
		  
		  /* Now scale the data */
		  for(i=0;i<purdueHeader.nimg;i++)
		    {/* Do all boxes */
		      for(j=0;j<(numRowsTsizeFactor);j++)
			for(k=0;k<numColsTsizeFactor;k++)
			  {/* Check all bytes */
			    outChar = outBuffer[((i*numColsTsizeFactor*
						  numRowsTsizeFactor)+
						 (j*numColsTsizeFactor)+k)];
			    outBuffer[((i*numColsTsizeFactor*
					numRowsTsizeFactor)+
				       (j*numColsTsizeFactor)+k)]=
					 ((float)(outChar-minByte)*
					  scaleFactor);
			  }/* Check all bytes */
		    }/* Do all boxes */
		  
		  /* Now find the min and max */
		  minByte = 9999;
		  maxByte = -9999;
		  minByte = INT_MAX;  /* dgm INT_MAX instead of 9999 */
		  maxByte = INT_MIN;  /* dgm INT_MIN instead of -9999 */
		  for(i=0;i<purdueHeader.nimg;i++)
		    {/* Do all boxes */
		      for(j=0;j<numRowsTsizeFactor;j++)
			for(k=0;k<numColsTsizeFactor;k++)
			  {/* Check all bytes */
			    outChar = outBuffer[((i*numColsTsizeFactor*
						  numRowsTsizeFactor)+
						 (j*numColsTsizeFactor)+k)];
			    minByte = MIN(minByte,(int)outChar);
			    maxByte = MAX(maxByte,(int)outChar);
			  }/* Check all bytes */
		    }/* Do all boxes */
		      
		  fwrite(outBuffer,
			 sizeof(BYTE),(purdueHeader.nimg*numRowsTsizeFactor*
				       numColsTsizeFactor),fOut);
		  
		  sprintf(outStr,"RAW file stats:\n\
\n\
     Frames= [%d]\n\
     Width = [%d]\n\
     Height= [%d]\n\
OR\n\
     1 Frame\n\
     Width = [%d]\n\
     Height= [%d]\n\
\n\
     8 bit byte pixels\n\
     NO header\n\
\n\
     Min[%d]\n\
     Max[%d]\n",purdueHeader.nimg,(numColsTsizeFactor),numRowsTsizeFactor,
			  (projASU)?(maxPhi * numColsTsizeFactor)
			           :(purdueHeader.nimg * numColsTsizeFactor),
			  (projASU)?((purdueHeader.nimg/maxPhi) * 
				     numRowsTsizeFactor)
			           :(numRowsTsizeFactor),
			  minByte,maxByte);
		  askThem(outStr,FALSE,XmDIALOG_INFORMATION);
		  
		}/* Dump entire set */
	      
	      free(outBuffer);

	    }/* Write it */
	  else askThem("Can't open file.",TRUE,XmDIALOG_INFORMATION);

	}/* OK to proceed */
      
      fclose(fOut);
      
    }/* PROJASU */
  else if(readType==MAP || readType==FLOATMAP || readType==PIFMAP)
    {/* Try and write something */

      if(numWanted==1)
	{/* Just do a screen dump */
	  
	  writeIt= TRUE;
	  fOut= NULL;
	  fOut = fopen(fName,"r");
	  if(fOut != NULL)
	    {/*  Ooops we may be overwriting it!!! */
	      sprintf(outStr,
		      "Do you really want to overwrite [%s]?",fName);
	      writeIt= askThem(outStr,TRUE,XmDIALOG_QUESTION);
	    }/*  Ooops we may be overwriting it!!! */
	  
	  if(writeIt)
	    {/* OK to proceed */
	      
	      if(fOut!=NULL)
		fclose(fOut);
	      fOut= 0;
	      fOut= fopen(fName,"w");
	      if(fOut)
		{/* Write it */
		  
		  calcSizeFactor(zoomFactor);
		  
		  
		  numColsTsizeFactor = display_width;
		  numRowsTsizeFactor = display_height;
		  CK(outBuffer= (BYTE *)malloc(sizeof(BYTE)*
					       (numRowsTsizeFactor)*
					       (numColsTsizeFactor)));
		  
		  minF = maxF = image_buffer1_disp[0];
		  for(i=0;i<(numRowsTsizeFactor);i++)
		    for(k=0;k<(numColsTsizeFactor);k++)
		      {
			maxF = MAX(maxF,
				   image_buffer1_disp[(i*buffer_width)+k]);
			minF = MIN(minF,
				   image_buffer1_disp[(i*buffer_width)+k]);
		      }
		  scaleF = 254.0 / (maxF-minF);
		  for(i=0;i<(numRowsTsizeFactor);i++)
		    for(k=0;k<(numColsTsizeFactor);k++)
		      outBuffer[(i*numColsTsizeFactor)+k]=
			(BYTE)(((float)image_buffer1_disp[(i*
					buffer_width)+k]-minF)*
			       scaleF);
		  
		  /* Now find the min and max */
		  minByte = 9999;
		  maxByte = -9999;
		  minByte = INT_MAX;  /* dgm INT_MAX instead of 9999 */
		  maxByte = INT_MIN;  /* dgm INT_MIN instead of -9999 */
		  for(j=0;j<(numRowsTsizeFactor);j++)
		    for(k=0;k<numColsTsizeFactor;k++)
		      {/* Check all bytes */
			outChar = outBuffer[((j*numColsTsizeFactor)+k)];
			minByte = MIN(minByte,(int)outChar);
			maxByte = MAX(maxByte,(int)outChar);
		      }/* Check all bytes */
		  
		  /* Now find scale factor */
		  scaleFactor = 255.0 / (float)(maxByte-minByte);
		  
		  /* Now scale the data */
		  for(j=0;j<(numRowsTsizeFactor);j++)
		    for(k=0;k<numColsTsizeFactor;k++)
		      {/* Check all bytes */
			outChar = outBuffer[((j*numColsTsizeFactor)+k)];
			outBuffer[((j*numColsTsizeFactor)+k)]=
			  ((float)(outChar-minByte)*scaleFactor);
		      }/* Check all bytes */
		  
		  /* Now find the min and max */
		  minByte = 9999;
		  maxByte = -9999;
		  minByte = INT_MAX;  /* dgm INT_MAX instead of 9999 */
		  maxByte = INT_MIN;  /* dgm INT_MIN instead of -9999 */
		  for(j=0;j<numRowsTsizeFactor;j++)
		    for(k=0;k<numColsTsizeFactor;k++)
		      {/* Check all bytes */
			outChar = outBuffer[((j*numColsTsizeFactor)+k)];
			minByte = MIN(minByte,(int)outChar);
			maxByte = MAX(maxByte,(int)outChar);
		      }/* Check all bytes */
		  
		  fwrite(outBuffer,
			 sizeof(BYTE),(numRowsTsizeFactor*
				       numColsTsizeFactor),fOut);
		  
		  sprintf(outStr,"RAW file stats:\n\
\n\
     Frames= [%d]\n\
     Width = [%d]\n\
     Height= [%d]\n\
\n\
     8 bit byte pixels\n\
     NO header\n\
\n\
     Min[%d]\n\
     Max[%d]\n",1,(numColsTsizeFactor),numRowsTsizeFactor,
			  minByte,maxByte);
		  askThem(outStr,FALSE,XmDIALOG_INFORMATION);
		}/* Write it */
		  
	    }/* OK to proceed */
	  
	  fclose(fOut);

	}/* Just do a screen dump */
      else
	{/* DO this */
	  /*      stuffDisplayMapBuffer(curBox,editBoxZoom,0); */
	  stuffDisplayMapBuffer((imgNum-1),1,0);
	  /* calcBoxSizeFactor(editBoxZoom); */
	  writeRAWcontour(fName,editBox_display_buffer,
			  (int)(numCols),
			  (int)(numRows));
	}/* DO this */
      
    }/* Try and write something */
  else if((fileType==EMPORTABLE ||
	   fileType==BYTEPORTABLE ||
	   fileType==PORTABLE) && imgInCore)
    {/* Try and write something */

      writeIt= TRUE;
      fOut= NULL;
      fOut = fopen(fName,"r");
      if(fOut != NULL)
	{/*  Ooops we may be overwriting it!!! */
	  sprintf(outStr,
		  "Do you really want to overwrite [%s]?",fName);
	  writeIt= askThem(outStr,TRUE,XmDIALOG_QUESTION);

	}/*  Ooops we may be overwriting it!!! */

      if(writeIt)
	{/* OK to proceed */
	  
	  if(fOut!=NULL)
	    fclose(fOut);
	  fOut= 0;
	  fOut= fopen(fName,"w");
	  if(fOut)
	    {/* Write it */
	      
	      fwrite(inCore_buffer,
		     sizeof(float),(file_width*file_height),fOut);

	      sprintf(outStr,"RAW file stats:\n\
\n\
     Width = [%d]\n\
     Height= [%d]\n\
\n\
     32 bit float byte pixels\n\
     NO header\n\
\n",
		      file_width,file_height);
/*		      minByte,maxByte); */
	      askThem(outStr,FALSE,XmDIALOG_INFORMATION);
	    }/* Write it */
	  else askThem("Can't open file.",TRUE,XmDIALOG_INFORMATION);
	  
	}/* OK to proceed */
      
      fclose(fOut);
      
    }/* Try and write something */

  
}/* writeRAW */





/******************************************************
 * writeState ()-  Write the current option state of the
 * program to an ASCII disk file.
*/
void writeState(char *fName)
{/* WRITESTATE */
  int i;
  char version[64], outStr[64];
  FILE *fOut;
  
  /* Write it */
  fOut= 0;
  fOut= fopen(fName,"w");
  if(fOut)
    {/* Write it */
      fprintf(fOut,"%s\n",VERSION);
      fprintf(fOut,"%d\n",locked);
      fprintf(fOut,"%d\n",boxRadius);
    }/* Write it */
  else askThem("Can't open file.",TRUE,XmDIALOG_INFORMATION);
  fclose(fOut);

}/* WRITESTATE */





/* ************************************************************
 * virtual_read_activate() - For all images regardless of size
 * read a new pages everytime they move through the image with
 * the scrollbars.  Pay attention to how we're zoomed as if we're
 * zoomed smaller we can read more info off the disk if we're 
 * zoomed larger we need to scroll more slowly etc...
 * We implemented this way because we want to be able to look at
 * an infinitely large image.  If we "page" it we can look at it
 * even if we don't have enough internal memory to hold it all.
 */
void virtual_read_activate(int newX, int newY)
{/* virtual_READ_actIVATE */
  int y,x,l,m,n,yOut,yOff,logicalRecordLength,image_depth;
  int scaledVirYOffset,
      scaledVirXOffset;
  int thisFileType;
  float *temp_int_image_buffer1;
  BYTE *temp_byte_image_buffer1;
  unsigned short *thisFSbuf;
  BYTE thisByte;
  int thisInt, sizeFactorInt, yTsizeFactor;
  static float thisOldZoomFactor = 0.0;
  FILE *fIn;

  thisFileType = fileType;
  if(fftSize==1024 && disp1024fft && XtIsManaged(fftscreen))
    {/* Special display of 1024 fft */
      thisFileType = PORTABLE;

    }/* Special display of 1024 fft */

  /* This may take awhile so give them the watch cursor */
  attrs.cursor= watch_cursor;
  XChangeWindowAttributes(dsply, XtWindow(drawingWidget), 
			  CWCursor, &attrs);
  XFlush(dsply);

  if(rotfltSwitch)
    {/* We are roflt'ing */
      file_width = orig_file_width;
    }/* We are roflt'ing */

  if(thisFileType==PACKEDPORTABLE)
    {/* Just redraw for zoom */
      if(fullScreen==99)
	{/* Calc how much ZOOM we used to be in the file */
	  calcSizeFactor(thisOldZoomFactor);
	  /* Reset from full screen after we've calculated the coords
	   * we were in before (in fullscreen mode) */
	  fullScreen= FALSE;
	}/* Calc how much ZOOM we used to be in the file */

      stuffPackedImage(purdueHeader.numCols,
		       purdueHeader.numRows,
		       packImg,purdueHeader.nimg,
		       imgNum);  /* Start at zero'th image */
      goto noImg;
    }/* Just redraw for zoom */
  else if(thisFileType==MAP || thisFileType==FLOATMAP)
    {/* Just redraw for zoom */
      if(fullScreen==99)
	{/* Calc how much ZOOM we used to be in the file */
	  calcSizeFactor(thisOldZoomFactor);
	  /* Reset from full screen after we've calculated the coords
	   * we were in before (in fullscreen mode) */
	  fullScreen= FALSE;
	}/* Calc how much ZOOM we used to be in the file */

      stuffMapImage((float *)map_image_buffer, 
		    (float *)float_image_buffer1, 
		    file_width,file_height,
		    nsecs,
		    imgNum, 0);  /* Start at zero'th image */
      goto noImg;
    }/* Just redraw for zoom */

  if((fullScreen==TRUE) && display_width==fullScreen_buffer.dispx &&
     display_height==fullScreen_buffer.dispy)
    {/* Just use the one we calculated last time */
      thisFSbuf = fullScreen_buffer.fullImage;
      for(i=0;i<image_height;i++)
	{/* rows */
	for(j=0;j<image_width;j++)
	  {/* DO all */
	  int_image_buffer1[((i*image_width)+j)]=  (unsigned short)
	    thisFSbuf[((i*image_width)+j)];
	  }/* DO all */
	}/* rows */
      goto noImg;
    }/* Just use the one we calculated last time */


  if(!imgInCore)
    {/* Open disk file */
      fIn= NULL;
      fIn = fopen(image_fileName.addr,"r");
      if(fIn == NULL)
	{/*  Ooops can't read it!!! */
	  if(!EQU(fileName.addr,""))
	    {sprintf(outStr,
		  "Can't read [%s] ",fileName.addr);
	    askThem(outStr,TRUE,XmDIALOG_INFORMATION);}
	  goto noRead;
	}/*  Ooops can't read it!!! */
    }/* Open disk file */
  
  /* Calc zoom */
  calcSizeFactor(zoomFactor);

  /* Init out the image buffer */
  for(i=0;i<buffer_height;i++)
    {/* Init them */
      memcpy((&image_buffer1[(i*buffer_width)]),stufBLKBuf,buffer_width);
    }/* Init them */

  for(i=0;i<image_height;i++)
    {/* Init them */
      memcpy((&int_image_buffer1[(i*image_width)]),stufBLKIntBuf,
	     (image_width*2));
    }/* Init them */

  /* I think xoffset and yoffset will always be <=0 after this 
   * because we're clicking in the display_widht/height space.
   */
  xoffset = ((newX * recipSizeFactor)+(xoffset*recipSizeFactor)) - 
    (display_width / 2.0);
  yoffset = ((newY * recipSizeFactor)+(yoffset*recipSizeFactor)) - 
    (display_height / 2.0);
  xoffset = MIN((MAXSIZE-display_width),MAX(xoffset,0));
  yoffset = MIN((MAXSIZE-display_height),MAX(yoffset,0));


  if(newY > 0 && newY > 0)
    if(fullScreen==TRUE)
      {/* We are Normal */
	virXoffset = virYoffset = 0;
      }/* We are Normal */
    else 
      {/* We are zooming */
	int fileSpaceX,fileSpaceY,
	    dispXfileSpace,dispYfileSpace,
	    halfDW,halfDH;

	/* Calc how much ZOOM we used to be in the file */
	calcSizeFactor(thisOldZoomFactor);
	/* Reset from full screen after we've calculated the coords
	 * we were in before (in fullscreen mode) */
	if(fullScreen==99)
	  fullScreen= FALSE;
	fileSpaceX = virXoffset + ((int)(newX*recipSizeFactor));
	fileSpaceY = virYoffset + ((int)(newY*recipSizeFactor));

	/* Calc zoom for what we are now */
	calcSizeFactor(zoomFactor);
	halfDW =           ((int)(display_width / 2.0));
	halfDH =           ((int)(display_height / 2.0));
	dispXfileSpace = ((int)(halfDW * recipSizeFactor));
	dispYfileSpace = ((int)(halfDH * recipSizeFactor));

	/* Now then virtual offsets are just what we are in the file
	 * minus what we can display */
	virXoffset = fileSpaceX - dispXfileSpace;
	virYoffset = fileSpaceY - dispYfileSpace;

      }/* We are zooming */

  scaledVirYOffset = (int)(virYoffset);
  scaledVirXOffset = (int)(virXoffset);

  /* Larger images can have more readIn at lower zooms */
  if(sizeFactor < 1.0)
    {/* Try to get more off the disk */
      if(buffer_width < display_width)
	logicalRecordLength = buffer_width * recipSizeFactor;
      else logicalRecordLength = display_width * recipSizeFactor;
      logicalRecordLength = MAX(defDisplayBufferSize,
				MIN(file_width,logicalRecordLength));
      if(buffer_height < display_height)
	image_depth = buffer_height * recipSizeFactor;
      else image_depth = display_height * recipSizeFactor;
      image_depth = MIN(file_height,image_depth);

      scaledVirYOffset = MAX(0,MIN(scaledVirYOffset,
				   (file_height-image_depth)));
      scaledVirXOffset = MAX(0,MIN(scaledVirXOffset,
				   (file_width-logicalRecordLength)));
      virYoffset=   scaledVirYOffset;
      virXoffset=   scaledVirXOffset;
    }/* Try to get more off the disk */
  else 
    {/* MAX it out at 1 k */
      if(buffer_width < display_width)
	logicalRecordLength = buffer_width * recipSizeFactor;
      else logicalRecordLength = display_width * recipSizeFactor;
      logicalRecordLength = MAX(defDisplayBufferSize,
				MIN(file_width,logicalRecordLength));
      if(buffer_height < display_height)
	{/* Buffer smaller */
	  image_depth = buffer_height * recipSizeFactor;
	  scaledVirYOffset = MAX(0,MIN(scaledVirYOffset,
				       (MAX(file_height,buffer_height) -
					((int)(buffer_height*
					       recipSizeFactor)))));
	  scaledVirXOffset = MAX(0,MIN(scaledVirXOffset,
				       (MAX(file_width,buffer_width) -
					((int)(buffer_width*
					       recipSizeFactor)))));
	}/* Buffer smaller */
      else 
	{/* normal */
	  image_depth = display_height * recipSizeFactor;
	  scaledVirYOffset = MAX(0,MIN(scaledVirYOffset,
				       (MAX(file_height,display_height) -
					((int)(display_height*
					       recipSizeFactor)))));
	  scaledVirXOffset = MAX(0,MIN(scaledVirXOffset,
				       (MAX(file_width,display_width) -
					((int)(display_width*
					       recipSizeFactor)))));
	}/* normal */

      virYoffset=   scaledVirYOffset;
      virXoffset=   scaledVirXOffset;

    }/* MAX it out at 1 k */

  image_depth = MAX(0,MIN((file_height-((int)virYoffset)),
			  image_depth)); 
  logicalRecordLength = MAX(0,MIN((file_width-((int)virXoffset)),
				  logicalRecordLength));
  
  if(fullScreen || ((newY>0 && newX>0)))
    {/* Setup the scroll bars */
      int maxVal,thisPageSize;

      /* Don't call us back on this just yet!!! */
      noCallbackScroll= TRUE;

      /* First map coords to image coords */
      calcSizeFactor(zoomFactor);
      
      maxVal = (file_height - ((int)(display_height*recipSizeFactor)));
      maxVal = MAX(((int)(display_height*recipSizeFactor)),
		   MIN(file_height,maxVal));
      /* I *think* this is where the slider size for the main screen 
       * is defined.  Make sure that if you increase sliderSize you also
       * increase XmNmaximum by slidereSize as in:
       *
       *     XmNmaximum = maxVal + sliderSize;
       *
       * Also make sure that if you change these you can still use the scrolls
       * to get to any part of the screen you want.
       * 
       * Also here is where you'd probably want to fix the paging 
       * problem Paul has.  e.g. XmNpageIncrement now is calulated to be
       * display_height pages (dsply_height is the number of pixels 
       * you currently see on the scren) but maybe something like:
       * 
       *         XmNpageIncrement *= 0.90;  /+ 90.0%  +/
       * 
       * or such is what he wants???
       */
      maxVal = MAX(maxVal,(virYoffset+1));
      thisPageSize = (int)((int)((display_height*recipSizeFactor)*0.9));
      XtVaSetValues(vertScroll,XmNmaximum, maxVal+thisPageSize,NULL);
      XtVaSetValues(vertScroll,XmNminimum, 0,NULL);
      XtVaSetValues(vertScroll,XmNvalue, (virYoffset),NULL);
      XtVaSetValues(vertScroll,XmNsliderSize, thisPageSize,NULL);
      XtVaSetValues(vertScroll,XmNpageIncrement, ((int)thisPageSize),NULL);
		    
      if(buffer_width < display_width)
	{/* Buffer is samller than display */
	  maxVal = (file_width - ((int)(buffer_width*recipSizeFactor)));
	  maxVal = MAX(((int)(buffer_width*recipSizeFactor)),
		       MIN(file_width,maxVal));
	  maxVal = MAX(maxVal,(virXoffset+1));
	  thisPageSize = (int)((int)((buffer_width*recipSizeFactor)*0.9));
	}/* Buffer is samller than display */
      else
	{/* Normal */
	  maxVal = (file_width - ((int)(display_width*recipSizeFactor)));
	  maxVal = MAX(((int)(display_width*recipSizeFactor)),
		       MIN(file_width,maxVal));
	  maxVal = MAX(maxVal,(virXoffset+1));
	  thisPageSize = (int)((int)((display_width*recipSizeFactor)*0.9));
	}/* Normal */
      XtVaSetValues(horiScroll,XmNmaximum, maxVal+thisPageSize,NULL);
      XtVaSetValues(horiScroll,XmNminimum, 0,NULL);
      XtVaSetValues(horiScroll,XmNvalue, (virXoffset),NULL);
      XtVaSetValues(horiScroll,XmNsliderSize, thisPageSize,NULL);
      XtVaSetValues(horiScroll,XmNpageIncrement, ((int)thisPageSize),NULL);
      XFlush(dsply);

    }/* Setup the scroll bars */

  sprintf(outStr,"%.2f",sizeFactor);
  XmTextSetString(zoomTextW, outStr);   /* Feedback */
  XFlush(dsply);

  /* Figure out how we're going to do fractional zooming when making
   * the image larger.  Add 1.0 in case we're doing fractional zooms */
  if(sizeFactor > (float)((int)sizeFactor))
    sizeFactorInt = (int)sizeFactor + 1;

  /* We should always calc sizefactorint??? even if we're not doing 
   * fractional zooms ??? */
  sizeFactorInt = (int)sizeFactor + 1;

  /* Allocate our temp buffer and read */
    CK(temp_int_image_buffer1= 
       (float *)malloc(sizeof(float)*file_width));
  for(y=0;y<image_depth;y++)    
    {/* Do all rows */
      /* Use the +k because of EOR marker * 2.0 because it's 
       * shorts not chars */
      if(thisFileType == PORTABLE)  /* We're just using the inCore_buffer now 
				 *  so 1.0 not 2.0 for shorts like we'll need*/
	{/* Don't use the EOR markers */
	  yOff = headOff + 
	    (1.0 * (
		    ((y+scaledVirYOffset) 
		                          * file_width) +   /* Y direction*/
		    scaledVirXOffset                        /* X */
		    ));
	}/* Don't use the EOR markers */
      else if(thisFileType == BYTEPORTABLE)
	{/* Don't use the EOR markers */
	  yOff = headOff + 
	    (1.0 * (
		    ((y+scaledVirYOffset) 
		                          * file_width) +   /* Y direction*/
		    scaledVirXOffset                        /* X */
		    ));
	}/* Don't use the EOR markers */
      else
	{/* DO use the EOR markers */
	  yOff = headOff + 
	    (2.0 * (
		    ((y+scaledVirYOffset) * file_width) +   /* Y direction*/
		    scaledVirXOffset +                      /* X */
		    (scaledVirYOffset + y)                /* EOR markers */
		    ));
	}/* DO use the EOR markers */

      if(!imgInCore)
	{/* SEEK it */
	  sysError= fseek(fIn,yOff,0);
	  if(sysError)
	    {/* Can't go on */
	      askThem("[VRA] Error reading file.",TRUE,XmDIALOG_INFORMATION);
	      fclose(fIn);
	      goto noRead;
	    }/* Can't go on */
	}/* SEEK it */

      yTsizeFactor = (int)(y * sizeFactor);

      if((thisFileType==BYTEPORTABLE) || (thisFileType==PORTABLE))
	{/* Read bytes */
	  if(!imgInCore)
	    {/* From disk */
	    }/* From disk */
	  else
	    {/* Get from core memory */
	      int z;
	      for(z=0;z<logicalRecordLength;z++)
		temp_int_image_buffer1[z] = 
		  inCore_buffer[(yOff-headOff)+z];
	    }/* Get from core memory */

	  if(sizeFactor > 1.0)
	    {/* We are zooming out = making larger than normal */
	      for(m=0;m<(int)(sizeFactorInt);m++)
		{/* Do all zoomed lines */
		  yOff = (image_width * (yTsizeFactor + m)) ;
		  for(l=0;l<logicalRecordLength;l++)
		    {/* Do all cols */
		      thisInt= (int)temp_int_image_buffer1[l];
		      yOut = yOff + ((int)(l*sizeFactor));
		      /* Let's do one more incase we're fractionally
		       * zooming smaller -> e.g. for zoomFactor = 1.123 e.g.
		       */
		      for(n=0;n<(int)(sizeFactorInt);n++)
			{/* expand */
			  *(int_image_buffer1+yOut+n)= 
			                    (unsigned short)thisInt;
			}/* expand */
		      if(flickerFix2D)
			{/* SPecail case */
			  if(fix2DinCore_diff[
					      ((y+scaledVirYOffset) 
					       * file_width) +/* Y direction*/
					      scaledVirXOffset + /* X */
					      l
					      ])
			    {
			      for(n=0;n<(int)(sizeFactorInt);n++)
				{/* expand */
				  *(int_image_buffer1+yOut+n)= 
				    (unsigned short)30000;
				}
			    }
			}/* SPecail case */

		    }/* Do all cols */
		}/* Do all zoomed lines */
	    }/* We are zooming out */
	  else
	    {/* We are zooming IN = making smaller than normal*/
	      yOff = (image_width * (yTsizeFactor)) ;
	      for(l=0;l<logicalRecordLength;l+=(int)recipSizeFactor)
		{/* Do all cols */
		  yOut = yOff + ((int)(l*sizeFactor));
		  thisInt= 0;
		  for(n=0;n<(int)recipSizeFactor;n++)
		    thisInt += (int)temp_int_image_buffer1[l+n];
		  *(int_image_buffer1+yOut)= (unsigned short)(thisInt/
							 (int)recipSizeFactor);

		  if(flickerFix2D)
		    {/* SPecail case */
		      if(fix2DinCore_diff[
					  ((y+scaledVirYOffset) 
					   * file_width) +/* Y direction*/
					  scaledVirXOffset + /* X */
					  l
					  ])
			{
			  *(int_image_buffer1+yOut)= 
			    (unsigned short)30000;
			}
		    }/* SPecail case */
		  
		}/* Do all cols */
	    }/* We are zooming IN */
	}/* Read bytes */
      else
	{/* Read shorts */
	  askThem("[VRA] I don't understand this type.",
		  FALSE,XmDIALOG_INFORMATION);
	  fclose(fIn);
	  goto noRead;
	}/* Read shorts */
    }/* Do all rows */
    free(temp_int_image_buffer1);

  if(fullScreen && display_width!=fullScreen_buffer.dispx &&
     display_height!=fullScreen_buffer.dispy)
    {/* Remember this for next time */

      /* Stuff our display buffer */
      stuffImage(int_image_buffer1, image_width, image_height, 
		 contrastStretch,TRUE);
      
      thisFSbuf = fullScreen_buffer.fullImage;
      if(thisFSbuf!=NULL)
	free(thisFSbuf);
      CK(thisFSbuf=(unsigned short *)malloc(sizeof(unsigned short)*
					    image_width*image_height));
      for(i=0;i<image_height;i++)
	for(j=0;j<image_width;j++)
	  thisFSbuf[((i*image_width)+j)]=
	    (unsigned short)int_image_buffer1[((i*image_width)+j)];
      fullScreen_buffer.dispx = display_width;
      fullScreen_buffer.dispy = display_height; 
      fullScreen_buffer.fullImage = thisFSbuf;
   }/* Remember this for next time */

  thisOldZoomFactor = zoomFactor;

  if(!imgInCore)
    fclose(fIn);

 noImg:;
  /* Stuff our display buffer */
  if(thisFileType != PACKEDPORTABLE &&
     thisFileType != MAP && thisFileType!=FLOATMAP)
     {/* OK to stuff it */
       if(thisFileType==PORTABLE)
	 stuffImage(int_image_buffer1, image_width, image_height, 
		    contrastStretch, FALSE);  /*  WHY??? */
       else
	 stuffImage(int_image_buffer1, image_width, image_height, 
		    contrastStretch, FALSE);
     }/* OK to stuff it */

  DrawWindow();

  /* Redraw the histogram */
  histogram_activate(&minP, &maxP);
  
 noRead:;

  /* Restore the old cursor whatever it was */
  attrs.cursor= old_cursor;
  XChangeWindowAttributes(dsply, XtWindow(drawingWidget), 
			  CWCursor, &attrs);
 noDisplay:;
  XFlush(dsply);
  noCallbackScroll= FALSE;

}/* virtual_READ_actIVATE */




/*-----------------------------------*/
/*************************************
 * INFO screen callbacks             *
 *************************************/
/*-----------------------------------*/
/**************************************************************
 * askThem()- This is the generic display a message and get
 * them to respond screen.  We get a return value from globAsk
 * which is set by infoOK() or infoCancel() (see below).
*/
int askThem(char *infoStr, int systemModal, int dialogType)
{/* ASKTHEM */
  XmString simpleString;

  /* Init and try to position the window */
  globAsk = -99;
  XFlush(dsply);
  createSimpleString(&simpleString,infoStr);
  XtSetArg(arglst[0], XmNmessageString, simpleString);
  XtSetArg(arglst[1], XmNdialogType, dialogType);
  XtSetArg(arglst[2], XmNx, 150);
  XtSetArg(arglst[3], XmNy, 150);
  XtSetValues(infoBox, arglst, 4);
  XFlush(dsply);
  XmStringFree(simpleString); 
 
  /* Init it and pop it up sit here and wait until we get 
   * something back */
  gFinished= FALSE;
  XtManageChild(infoBox);
  XtPopup(XtParent(infoBox), XtGrabNone);

  XmProcessTraversal(infoBox, XmTRAVERSE_CURRENT); 
  XtRealizeWidget(infoBox);  
  XFlush(dsply);

  /* gFinished is set in the infoOK() or infoCancel() callback */
  while(!gFinished)
    XtAppProcessEvent(app_context, XtIMAll);

  /* Pop it down and return their response */
  XtUnmanageChild(infoBox);
  XtPopdown(XtParent(infoBox));
  XtSetArg(arglst[0], XmNdialogStyle, XmDIALOG_MODELESS);
  XtSetValues(XtParent(infoBox), arglst, 1);
  XFlush(dsply);

  return(globAsk);

}/* ASKTHEM */



/**************************************************************
 * askThemAnswer()- This is the generic display a message and get
 * them to respond screen.  We get a return value from globAsk
 * which is set by the text widget on the screen.
*/
int askThemAnswer(char *infoStr, int maxAnswer)
{/* ASKTHEMANSWER */

  /* Init and try to position the window */
  XmTextSetString(infoPromptTextW,infoStr); 
  globAskAnswer = -99;
  XFlush(dsply);

  /* Init it and pop it up sit here and wait until we get 
   * something back */
  gFinished= FALSE;
  XtManageChild(infoAnswerscreen);
  XtPopup(XtParent(infoAnswerscreen), XtGrabNone);

  XmProcessTraversal(infoAnswerscreen, XmTRAVERSE_CURRENT); 
  XtRealizeWidget(infoAnswerscreen);  
  XFlush(dsply);

  /* gFinished is set in the infoOK() or infoCancel() callback */
  while(!gFinished)
    XtAppProcessEvent(app_context, XtIMAll);

  /* Pop it down and return their response */
  XtUnmanageChild(infoAnswerscreen);
  XtPopdown(XtParent(infoAnswerscreen));
  XtSetArg(arglst[0], XmNdialogStyle, XmDIALOG_MODELESS);
  XtSetValues(XtParent(infoAnswerscreen), arglst, 1);
  XFlush(dsply);

  globAskAnswer = MAX(MIN(globAskAnswer,maxAnswer-1),0);

  for(i=0;i<maxAnswer;i++)
    {/* Deleteing */
      XmListDeletePos(infoListW,1);
    }/* Deleteing */

  return(globAskAnswer);

}/* ASKTHEMANSWER */




/*---------------------------------------------------------------------*/
/*---------------------------------------------------------------------*/
/*---------------------------------------------------------------------*/
/***********************************************************************
 * CALLBACK routines - Rotuines called directly from X whenever the    *
 * appropriate event occurs.                                           *
 ***********************************************************************/
/*---------------------------------------------------------------------*/
/*---------------------------------------------------------------------*/
/*---------------------------------------------------------------------*/

/************************************************************
 * infoOK() - If the OK button is pressed from the infoscreen
 * return here.  Set gFinished to cause the value of globAsk
 * be return to the starting point (see askThem() above).
*/
void infoOK(Widget widget, int *tag, XmAnyCallbackStruct *callback_data)
{/* infoOK */

  globAsk = TRUE;
  XtUnmanageChild(infoBox);
  gFinished= TRUE;

}/* infoOK */


/************************************************************
 * infoCancel() - If the CNACEL button is pressed from the 
 * infoscreen return here.  Set gFinished to cause the value 
 * of globAsk be return to the starting point (see askThem() 
 * above).
*/
void infoCancel(Widget widget, int *tag, XmAnyCallbackStruct *callback_data)
{/* infoCancel */

  globAsk = FALSE;
  XtUnmanageChild(infoBox);
  gFinished= TRUE;

}/* infoCancel */




/*-----------------------------------*/
/*************************************
 * MAIN screen Generic I/O           *
 *************************************/
/*-----------------------------------*/
/* ************************************************************
 * write_activate() - They selected the "Write" submenu.  Branch
 * on the appropriate.
 */
void write_activate(Widget widget, int *tag, 
		    XmAnyCallbackStruct *callback_data)
{/* WRITE_ACTIVATE */

  writeType = *tag;

  /* Try and setup the file browsing widget with the
   * appropriate default file typ etc..
   */
  if(writeType == CIRCLEPACKED)
    filterFiles("*.pif",CIRCLEPACKED); 
  else if(writeType == APPENDCIRCLEPACKED)
    filterFiles("*.pif",APPENDCIRCLEPACKED); 
  else if(writeType == APPENDRECT)
    filterFiles("*.pif",APPENDRECT); 
  else if(writeType == RECTPACKED)
    filterFiles("*.pif",RECTPACKED); 
  else if(writeType == BOXCOORDS)
    filterFiles("*.b*",BOXCOORDS); 
  else if(writeType == STATE)
    filterFiles("*.sta",STATE); 
  else if(writeType == ARCHIVE)
    filterFiles("bragg4:[arch]*.*",ARCHIVE); 
  else if(writeType == TIFF)
    filterFiles("*.raw",TIFF); 
  else if(writeType == PCX)
    filterFiles("*.pcx",PCX); 
  else if(writeType == IMG)
    filterFiles("*.pif",IMG); 
  else if(writeType == MAP)
    {/* map */
      filterFiles("*.pmp",PIFMAP);  
    }/* map */

  readWrite= WRITING;

  /* Future work with a databse might use the isDatabase flag
   * to get info from database */
  if(isDatabase)
    {/* Already know the filename */
      gotFilename(0,0,0);
    }/* Already know the filename */
  else
    {/* Ask them for a filename */
      raiseIt(fileSel_widget);
    }/* Ask them for a filename */

 noWriting:;

}/* WRITE_ACTIVATE */



/* ************************************************************
 * read_activate() - They selected the "Read" submenu.  Branch
 * on the appropriate.
 */
void read_activate(Widget widget, int *tag, 
		   XmAnyCallbackStruct *callback_data)
{/* READ_ACTIVATE */
  float q, qSq,xCen,yCen,zCen,mapRadius,thisPixel;
  float xSq,ySq,zSq;
  int l,m,k;
  XmString simpleString;

  if(*tag==ALL)

  /* Try and setup the file browsing widget with the
   * appropriate default file type etc..
   */
  filterFiles("*",ALL); 

  else if(*tag==MAP)
   filterFiles("*.map",MAP); 
  else if(*tag==MRC)
   filterFiles("*.*",MRC); 
  else if(*tag==PIFMAP)
   filterFiles("*.pif",PIFMAP); 
  else if(*tag==EMPORTABLE)
   filterFiles("*.pif",EMPORTABLE); 
  else if(*tag==PACKEDPORTABLE)
   filterFiles("*.pif",PACKEDPORTABLE); 
  else if(*tag==SF)
   filterFiles("*.sf",SF); 
  else if(*tag==BOXCOORDS)
   filterFiles("*.b*",BOXCOORDS); 
  else if(*tag==STATE)
   filterFiles("*.sta",STATE); 
  else if(*tag==ARCHIVE)
    {
      filterFiles("/bio/baker9d5/archive/*",ARCHIVE); 
    }
  else if(*tag==DEMO)
   filterFiles("bragg4:[arch.demo]*.*",DEMO); 
  else if(*tag==TESTARCH)
   filterFiles("/bio/baker9d7/test/*",TESTARCH); 

  /* Let gotFileName() find it out */
  if(*tag==BOXCOORDS || *tag==SF || *tag==STATE || *tag==MRC)
    readType= *tag;
  else readType = 0;

  readWrite= READING;
  XtUnmanageChild(partprocscreen);   /* Make them read from partproc screen 
				      * if they want particle processing */
  raiseIt(fileSel_widget);
  XFlush(dsply);

}/* READ_ACTIVATE */



/* ************************************************************
 * exit_activate() - They said they wanted to exit the program.
 */
void exit_activate(Widget widget, int *tag, 
		   XmAnyCallbackStruct *callback_data)
{/* exit_ACTIVATE */
  
    exit(0);
  
}/* exit_ACTIVATE */





/*-----------------------------------*/
/*************************************
 * MAIN screen form Resizing         *
 *************************************/
/*-----------------------------------*/
/* ************************************************************
 * formResize() - When the main screen formDialog is resized 
 * call here.  Get the new window size and setup our globals.
 */
void formResize(Widget widget, int *tag, 
		XmDrawingAreaCallbackStruct *callback_data)
{/* FORMRESIZE */
  Window tWin;
  int oldW,oldH,width,height,thisWidth,thisHeight,pageIncW,pageIncH; 
  XWindowAttributes winAttrib;
  XSetWindowAttributes setWinAttrib;

  if(inited)
    {/* Don't do unless we've got all the goodies we need */
      tWin = (Window)XtWindow(drawingWidget);
      
      /* Get new window dimensions */
      if(tWin > 0)
	{/* Valid window */
	  XGetWindowAttributes(dsply,tWin,&winAttrib);
	  oldW= width = (int)(winAttrib.width);
	  oldH= height = (int)(winAttrib.height);
	}/* Valid window */

      /* Can't have the window larger than the image data we've got!!! */
      width  = MIN(width,defDisplayBufferSize);
      height = MIN(height,defDisplayBufferSize);

      /* First map coords to image coords */
      calcSizeFactor(zoomFactor);
      
      /* Remember the dimensions of this */
      display_width = width;
      display_height= height;
      
      /* Stuff the width according to whether this is a 
       * large image or not
       */
      thisWidth = file_width;
      thisHeight = file_height;
      if(!largeImage)
	thisHeight= thisWidth= buffer_width;

      if(!largeImage)
	{/* Setup for larger images */
	  scrollWidth =  (NUMCOLS - DEF_DISPLAY_SIZE)+12;
	  pageIncW= ((int)(display_height*sizeFactor));
	}/* Setup for larger images */
      else 
	{/* Setup for smaller images */
	  scrollWidth = (thisWidth - width);
	  pageIncW= ((int)(display_height));
	}/* Setup for smaller images */
      if(!largeImage)
	{/* Setup for larger images */
	  scrollHeight =  (NUMROWS - DEF_DISPLAY_SIZE)+12;
	  pageIncH= ((int)(display_height*sizeFactor));
	}/* Setup for larger images */
      else 
	{/* Setup for smaller images */
	  scrollHeight = (thisHeight - height);
	  pageIncH= ((int)(display_height));
	}/* Setup for smaller images */
      
      xoffset= yoffset = 0;

      if(file_height>0)
	{/* Now set the scrollbars */
	  XtVaSetValues(vertScroll,XmNmaximum, 
			scrollHeight+(int)(pageIncH*0.9),NULL);
	  XtVaSetValues(vertScroll,XmNminimum, 0,NULL);
	  XtVaSetValues(vertScroll,XmNvalue, (virYoffset + yoffset),NULL);
	  XtVaSetValues(vertScroll,XmNsliderSize, (int)(pageIncH*0.9),NULL);
	  XtVaSetValues(vertScroll,XmNpageIncrement, (int)(pageIncH*0.9),NULL);
			
	  XtVaSetValues(horiScroll,XmNmaximum, 
			scrollWidth+(int)(pageIncW*0.9),NULL);
	  XtVaSetValues(horiScroll,XmNminimum, 0,NULL);
	  XtVaSetValues(horiScroll,XmNvalue, (virXoffset+xoffset),NULL);
	  XtVaSetValues(horiScroll,XmNsliderSize, (int)(pageIncW*0.9),NULL);
	  XtVaSetValues(horiScroll,XmNpageIncrement, (int)(pageIncW*0.9),NULL);
			
	}/* Now set the scrollbars */

      
    }/* Don't do unless we've got all the goodies we need */
  
  /* This changes whenever a new image  */
  /* is read                            */
  image_width  = defDisplayBufferSize;   
  image_height = defDisplayBufferSize;   
  
  if(inited && widget != 0)
    {/* Resized from interface */
      int newX,newY;
      newX = (display_width / 2);
      newY = (display_height / 2);
      virtual_read_activate(newX,newY);
    }/* Resized from interface */
  
}/* FORMRESIZE */




/*-----------------------------------*/
/*************************************
 * MAIN screen TOOLS                 *
 *************************************/
/*-----------------------------------*/
/**********************************************************
 * rwalog2() - Compute the log base 2 for this number.
*/
double rwalog2(float thisSizeFactor)
{/* RWALOG2 */

  return((double)(log10(thisSizeFactor) / log10(2.0)));

}/* RWALOG2 */




/* *******************************************************
 * zoomText_change() - They've changed ZOOM on the main
 * screen.  
 */
void zoomText_change(Widget widget, int *tag, 
		     XmAnyCallbackStruct *callback_data)
{/* ZOOMTEXT_CHANGE */
  char *zoomText;
  float thisSizeFactor;
  int newX,newY;

  if(inited)
    {/* Only do if it exists */
      zoomText = XmTextGetString(zoomTextW);
      thisSizeFactor = (float)strtod(zoomText,0);

      /* I need a log base 2 routine "log2()" that will allow
       * me to convert the input gotten above into meaningful
       * zoomFactors. 
       * e.g.
       *
       *      zoomFactor = (float)log2(thisSizeFactor);
       *
       */
      zoomFactor = (float)rwalog2(thisSizeFactor);

      zoomFactor = MAX(-4.0,MIN(zoomFactor,4.0));

      XtFree(zoomText);

      /* Now redraw with new zoominess */
      fullScreen= FALSE;
      newX = (display_width / 2);
      newY = (display_height / 2);
      virtual_read_activate(newX,newY);

    }/* Only do if it exists */

}/* ZOOMTEXT_CHANGE */




/* *******************************************************
 * toolMode_change() - They've changed modes on the main
 * screen.  Remember the cursor we're using and set
 * the globals.
 */
void toolMode_change(Widget widget, int *tag, 
		     XmAnyCallbackStruct *callback_data)
{/* TOOLMODE_CHANGE */
  XEvent event;
  int oldBoxZoom;
  
  boxZoom = *tag;
  if(boxZoom==USEZOOM)
    {/* Zooming */
      old_cursor = zoomCursor;
      attrs.cursor= zoomCursor;
      XChangeWindowAttributes(dsply, XtWindow(drawingWidget), 
			      CWCursor, &attrs);
      /*      XtUnmanageChild(pointscreen); */
    }/* Zooming */
  else if(boxZoom==USEBOX)
    {/* Boxing */
      old_cursor = cross_cursor;
      attrs.cursor= cross_cursor;
      XChangeWindowAttributes(dsply, XtWindow(drawingWidget), 
			      CWCursor, &attrs);
      /*      XtUnmanageChild(pointscreen); */
      if(callback_data->event!=0)
	raiseIt(boxScreenWidget);
    }/* Boxing */
  else if(boxZoom==USEPOINT)
    {/* Pointing */
      old_cursor = None;
      attrs.cursor= None;
      XChangeWindowAttributes(dsply, XtWindow(drawingWidget), 
			      CWCursor, &attrs);
      if(callback_data->event!=0)
	raiseIt(pointscreen);
    }/* Pointing */
  
}/* TOOLMODE_CHANGE */




/*-----------------------------------*/
/*************************************
 * MAIN screen DRAWING AREA          *
 *************************************/
/*-----------------------------------*/
/* ************************************************************
 * plotWindow_create() - Remember the drawing area window
 * pointer for use later.  Setup the file selection widget
 * here.
 */
void plotWindow_create(Widget widget, int *tag, 
		       XmAnyCallbackStruct *callback_data)
{/* PLOTWINDOW_CREATE */
  Arg arglst[32];
  XtCallbackRec callBck[5];    /* Our callbacks */
  
  /* Remember the widget number for later */
  drawingWidget= widget;
  
  /* Setup the file selection widget */
  fileSel_widget = (Widget)XmCreateFileSelectionDialog(main_window_widget,
						"Get file",NULL,0);
  createSimpleString(&globalTcs,"*.pif");

  XtSetArg(arglst[0], XmNx, 10);
  XtSetArg(arglst[1], XmNy, 10 );
  XtSetArg(arglst[2], XmNdirMask, globalTcs);  
  XtSetArg(arglst[3], XmNdefaultPosition, FALSE); 
  XtSetArg(arglst[4], XmNfontList, robemFontList); 
  XtSetArg(arglst[5], XmNlabelFontList,  robemFontList);
  XtSetArg(arglst[6], XmNbuttonFontList,  robemFontList);
  XtSetValues(fileSel_widget, arglst, 4);
  /* OLDWAY XtSetValues(fileSel_widget, arglst, 7); */

  /* Add callbacks to file selection widget */
  callBck[0].callback= (XtCallbackProc)gotFilename;
  callBck[0].closure= 0;
  callBck[1].callback= 0;
  callBck[1].closure= 0;
  XtAddCallbacks(fileSel_widget, XmNokCallback, 
		 callBck);  
  callBck[0].callback= (XtCallbackProc)cancel_fileSel;
  callBck[0].closure= 0;
  callBck[1].callback= 0;
  callBck[1].closure= 0;
  XtAddCallbacks(fileSel_widget, XmNcancelCallback, 
		 callBck);  
  XFlush(dsply);
/* malloc_verify(); */


}/* PLOTWINDOW_CREATE */




/*****************************************************************
 * rwaXDrawArc() - Draw circles on screen representing each
 * box.  Draw the circle thrice if they want a thicker display.
*/
void rwaXDrawArc(Display *disp, Window drawWin, GC thisgc,
		 int xPos, 
		 int yPos,
		 int wid1, int wid2,
		 int ang1, int ang2)
{/* RWAXDRAWARC */
  int i;
  
  if(dispThickCircles)
    {/* Do them thick */
      for(i=-1;i<2;i++)
	XDrawArc(disp, drawWin, image_gcWHITE,
		 xPos-i, 
		 yPos-i,
		 wid1+(i*2), wid2+(i*2),
		 ang1, ang2);
    }/* Do them thick */
  else
    {/* Do them thin */
      i = 0;
      XDrawArc(disp, drawWin, thisgc,
	       xPos-i, 
	       yPos-i,
	       wid1+(i*2), wid2+(i*2),
	       ang1, ang2);
    }/* Do them thin */

  
}/* RWAXDRAWARC */




/****************************************************************
 * drawFindRad() - Draw feedback info.
*/
void drawFindRad()
{/* DRAWFINDRAD */
  float scaleF,radlo,radhi;
  int   maxI,minI,myTurn,isAneg,binStep,allIt,thisY;
  GC    thisGC;
  float firstRad,firstCorr,*scaled_rdens;
  float minF,maxF;
  char *s1;

  if(ccoef==0)
    {/* Haven't done this yet */
      goto noDrawFindRad;
    }/* Haven't done this yet */

  drawingFindRad = TRUE;

  XtUnmanageChild(statusScreenWidget);
  XFlush(dsply);
  
  /* XClearWindow(dsply, XtWindow(drawingWidget)); */
  XFillRectangle(dsply, XtWindow(drawingWidget),
		 image_gcGREY[(int)(110)],0,0,
		 display_width,display_height);
  XFlush(dsply);
  
  if(nbins > 0)
    {
      thickStepFR = 800/nbins;  /* default */
      thinStepFR  = 200/nbins;
      thinStepFR  = 3;   /* Hard code for now */
    }
  else
    {
      thickStepFR = 800/10;  /* default */
      thinStepFR  = 200/10;
      thinStepFR  = 3;   /* Hard code for now */
    }

  thisGC = image_gcWHITE;
  XDrawLine(dsply, XtWindow(drawingWidget), thisGC,
	    83,display_height-100,
	    83,display_height-700);
  XDrawLine(dsply, XtWindow(drawingWidget), thisGC,
	    83+(0*thickStepFR),display_height-100,
	    83+(nbins*thickStepFR),display_height-100);
  sprintf(outStr,"Radius");
  XDrawString(dsply, XtWindow(drawingWidget), image_gcWHITE,
	      83+(thickStepFR*((nbins/2)-1)), display_height-20, 
	      outStr, strlen(outStr));
  sprintf(outStr,"Correlation and Rel Dens");
  XDrawString(dsply, XtWindow(drawingWidget), image_gcWHITE,
	      10, display_height-720, 
	      outStr, strlen(outStr));

  /* Don't use rdens to compute scale unlest it's the only one */
  allIt = rdensit+wrdensit+ccoefit+wccoefit+s_wccoefit;

  CK(scaled_rdens = (float *)malloc(sizeof(float)*nbins));
  if(rdensit)
    {/* Only do if rdensit */
      minF = 99999.9;
      maxF = -99999.9;
      minF = HUGE;   /* dgm HUGE instead of 99999.9 */
      maxF = -HUGE;  /* dgm -HUGE instead of -99999.9 */
      for(i=0;i<nbins;i++)
	{/* Check all rdens */
	  if(rdens[i] < minF)
	    minF = rdens[i];
	  if(rdens[i] > maxF)
	    maxF = rdens[i];
	}/* Check all rdens */
      if(maxF == minF)
	scaleF = 1.0;
      else
	{
	  if(minF > 0.0)
	    scaleF = (1.0 / (float)maxF);
	  else
	    {
	      if(maxF > (0.0 - minF))
		scaleF = (1.0 / (float)maxF);
	      else
		scaleF = (1.0 / (float)(0.0 - (float)minF));
	    }
	}
      for(i=0;i<nbins;i++)
	scaled_rdens[i] = rdens[i] * scaleF;
    }/* Only do if rdensit */

  /* Now let's scale everything */
  minF = 99999.9;
  maxF = -99999.9;
  minF = HUGE;   /* dgm HUGE instead of 99999.9 */
  maxF = -HUGE;  /* dgm -HUGE instead of -99999.9 */
  for(i=0;i<nbins;i++)
    {/* Search all data */
      if(rdensit && scaled_rdens[i] > maxF)
	maxF = scaled_rdens[i];
      if(wrdensit && wrdens[i] > maxF)
	maxF = wrdens[i];
      if(ccoefit && ccoef[i] > maxF)
	maxF = ccoef[i];
      if(wccoefit && wccoef[i] > maxF)
	maxF = wccoef[i];
      if(s_wccoefit && s_wccoef[i] > maxF)
	maxF = s_wccoef[i];

      if(rdensit && scaled_rdens[i] < minF)
	minF = scaled_rdens[i];
      if(wrdensit && wrdens[i] < minF)
	minF = wrdens[i];
      if(ccoefit && ccoef[i] < minF)
	minF = ccoef[i];
      if(wccoefit && wccoef[i] < minF)
	minF = wccoef[i];
      if(s_wccoefit && s_wccoef[i] < minF)
	minF = s_wccoef[i];
    }/* Search all data */
  if(minF < 0.0)
    minF = -1.0;
  else minF = 0.0;
  maxF = 1.0;
  if(maxF == minF)
    scaleF = 1.0;
  else
    scaleF = (600.0 / (float)((float)maxF - (float)minF));

  /* Write X axis labels */
  firstRad = 0.0;
  binStep = nbins / 10;
  if(binStep <= 0)
    binStep = 1;
  for(i=0;i<nbins;i+=binStep)
    {/* Do all */
      firstRad += radstep;

      sprintf(outStr,"%d",i);
      XDrawString(dsply, XtWindow(drawingWidget), image_gcWHITE,
		  83+(thickStepFR*i), display_height-65, 
		  outStr, strlen(outStr));
      sprintf(outStr,"%.1f",firstRad);
      XFlush(dsply);

    }/* Do all */
  
  /* Draw Zero */
  isAneg = FALSE;
  if(minF < 0.0)
    {/* Show them where zero is */
      isAneg = TRUE;
      XDrawLine(dsply, XtWindow(drawingWidget), image_gcWHITE,
		83+(0*thickStepFR),
		display_height-((int)((float)(0.0-minF)*
				      (float)scaleF)+
				100),
		83+(nbins*thickStepFR),
		display_height-((int)((float)(0.0-minF)*
				      (float)scaleF)+
				100)
		);
    }/* Show them where zero is */

  /* Now draw all curves */
  myTurn = FALSE;
  for(i=0;i<nbins;i++)
    {/* Do all */
      GC localThisGC;
      
      /* Draw tick mark */
      for(j=0;j<(thinStepFR*4);j++)
	XDrawLine(dsply, XtWindow(drawingWidget), image_gcWHITE,
		  83+((int)(((float)i)*thickStepFR))+j,display_height-99,
		  83+((int)(((float)i)*thickStepFR))+j,display_height-97);
      
      for(j=0;j<thinStepFR;j++)
	{/* Draw hsitogram */
	  if(isAneg)
	    {/* We do have negative values */
	      if(rdensit)
		{/* Do for rdens */
		  thisGC = image_gcWHITE;
		  thisY =  (int)((float)(scaled_rdens[i]-minF)*
				 (float)scaleF);
		  thisY = MIN(MAX(thisY,0),600);
		  XDrawLine(dsply, XtWindow(drawingWidget), 
			    thisGC,
			    83+(i*thickStepFR)+j,
			    display_height-((int)((float)(0.0-minF)*
						  (float)scaleF)+
					    100),
			    83+(i*thickStepFR)+j,
			    display_height-(thisY+100)
			    );
		}/* Do for rdens */
	      if(wrdensit)
		{/* Do for rdens */
		  thisGC = image_gcRED;
		  XDrawLine(dsply, XtWindow(drawingWidget), 
			    thisGC,
			    83+(i*thickStepFR)+j+3,
			    display_height-((int)((float)(0.0-minF)*
						  (float)scaleF)+
					    100),
			    83+(i*thickStepFR)+j+3,
			    display_height-
			    ((int)((float)(wrdens[i]-minF)*
				   (float)scaleF)+
			     100));
		}/* Do for rdens */
	      if(ccoefit)
		{/* Do for rdens */
		  thisGC = image_gcGREEN;
		  XDrawLine(dsply, XtWindow(drawingWidget), 
			    thisGC,
			    83+(i*thickStepFR)+j+6,
			    display_height-((int)((float)(0.0-minF)*
						  (float)scaleF)+
					    100),
			    83+(i*thickStepFR)+j+6,
			    display_height-
			    ((int)((float)(ccoef[i]-minF)*
				   (float)scaleF)+
			     100));
		}/* Do for rdens */
	      if(wccoefit)
		{/* Do for rdens */
		  thisGC = image_gcBLUE;
		  XDrawLine(dsply, XtWindow(drawingWidget), 
			    thisGC,
			    83+(i*thickStepFR)+j+9,
			    display_height-((int)((float)(0.0-minF)*
						  (float)scaleF)+
					    100),
			    83+(i*thickStepFR)+j+9,
			    display_height-
			    ((int)((float)(wccoef[i]-minF)*
				   (float)scaleF)+
			     100));
		}/* Do for rdens */
	      if(s_wccoefit)
		{/* Do for s_wccoef */
		  thisGC = image_gcBLACK;
		  if(i < (nbins-1))
		    XDrawLine(dsply, XtWindow(drawingWidget), 
			      thisGC,
			      83+(i*thickStepFR)+j+4,
			      display_height-((int)((float)(s_wccoef[i]-minF)*
						    (float)scaleF)+
					      100),
			      83+((i+1)*thickStepFR)+j+4,
			      display_height-
			      ((int)((float)(s_wccoef[i+1]-minF)*
				     (float)scaleF)+
			       100));
		}/* Do for s_wccoef */
	    }/* We do have negative values */
	  else
	    {/* Positive value */
	      if(rdensit)
		{
		  thisGC = image_gcWHITE;
		  thisY =  (int)((float)(scaled_rdens[i]-minF)*
				 (float)scaleF);
		  thisY = MIN(MAX(thisY,0),600);
		  XDrawLine(dsply, XtWindow(drawingWidget), 
			    thisGC,
			    83+(i*thickStepFR)+j,
			    display_height-100,
			    83+(i*thickStepFR)+j,
			    display_height-(thisY+100)
			    );
		}
	      if(wrdensit)
		{
		  thisGC = image_gcRED;
		  XDrawLine(dsply, XtWindow(drawingWidget), 
			    thisGC,
			    83+(i*thickStepFR)+j+3,
			    display_height-100,
			    83+(i*thickStepFR)+j+3,
			    display_height-((int)((float)(wrdens[i]-minF)*
						  (float)scaleF)+
					    100));
		}
	      if(ccoefit)
		{
		  thisGC = image_gcGREEN;
		  XDrawLine(dsply, XtWindow(drawingWidget), 
			    thisGC,
			    83+(i*thickStepFR)+j+6,
			    display_height-100,
			    83+(i*thickStepFR)+j+6,
			    display_height-((int)((float)(ccoef[i]-minF)*
						  (float)scaleF)+
					    100));
		}
	      if(wccoefit)
		{
		  thisGC = image_gcBLUE;
		  XDrawLine(dsply, XtWindow(drawingWidget), 
			    thisGC,
			    83+(i*thickStepFR)+j+9,
			    display_height-100,
			    83+(i*thickStepFR)+j+9,
			    display_height-((int)((float)(wccoef[i]-minF)*
						  (float)scaleF)+
					    100));
		}
	      if(s_wccoefit)
		{/* Do for s_wccoef */
		  thisGC = image_gcBLACK;
		  if(i < (nbins-1))
		    XDrawLine(dsply, XtWindow(drawingWidget), 
			      thisGC,
			      83+(i*thickStepFR)+j+4,
			      display_height-((int)((float)(s_wccoef[i]-minF)*
						    (float)scaleF)+
					      100),
			      83+((i+1)*thickStepFR)+j+4,
			      display_height-
			      ((int)((float)(s_wccoef[i+1]-minF)*
				     (float)scaleF)+
			       100));
		}/* Do for s_wccoef */
	    }/* Positive value */
	  XFlush(dsply);
	}/* Draw hsitogram */
    }/* Do all */
  
  /* Now do Y axis */
  myTurn = FALSE;
  firstCorr = minF;
  while(firstCorr < maxF)
    {/* Do all */
      sprintf(outStr,"%.2f",firstCorr);
      if(firstCorr < 0.0)
	XDrawString(dsply, XtWindow(drawingWidget), image_gcWHITE,
		    10,
		    display_height-
		    ((int)((float)(firstCorr-minF)*
			   (float)scaleF)+100),
		    outStr, strlen(outStr));
      else
	XDrawString(dsply, XtWindow(drawingWidget), image_gcWHITE,
		    10,
		    display_height-
		    ((int)((float)(firstCorr-minF)*
			   (float)scaleF)+100),
		    outStr, strlen(outStr));
      firstCorr += (float)((maxF-minF)/10.0);
      XFlush(dsply);
    }/* Do all */
  /* Do the last one */
  sprintf(outStr,"%.2f",firstCorr);
  if(firstCorr < 0.0)
    XDrawString(dsply, XtWindow(drawingWidget), image_gcWHITE,
		10,
		display_height-
		((int)((float)(firstCorr-minF)*
		       (float)scaleF)+100),
		outStr, strlen(outStr));
  else
    XDrawString(dsply, XtWindow(drawingWidget), image_gcWHITE,
		10,
		display_height-
		((int)((float)(firstCorr-minF)*
		       (float)scaleF)+100),
		outStr, strlen(outStr));

  free(scaled_rdens);

 noDrawFindRad:;

  XFlush(dsply);
  XSync(dsply,0);
  
}/* DRAWFINDRAD */





/****************************************************************
 * drawMag() - Draw feedback info .
*/
void drawMag()
{/* DRAWMAG */
  float scaleF;
  int   thickStep,thinStep,maxI,minI,myTurn,isAneg;
  GC    thisGC;
  
  if(coefs==0)
    {/* Haven't done this yet */
      goto noDrawMag;
    }/* Haven't done this yet */

  drawingFindRad = FALSE;

  /*
  XClearWindow(dsply, XtWindow(drawingWidget));

  XDrawRectangle(dsply, XtWindow(drawingWidget),
    image_gcBLACK,0,0,display_width,display_height); 
    */
  XFillRectangle(dsply, XtWindow(drawingWidget),
		 image_gcGREY[(int)(110)],0,0,
		 display_width,display_height);

  XFlush(dsply);
  
  if(magchk > 0)
    {
      thickStep = 800/magchk;  /* default */
      thinStep  = 400/magchk;
    }
  else
    {
      thickStep = 800/10;  /* default */
      thinStep  = 400/10;
    }
  
  maxI = minI = 0;
  for(i=0;i<magchk;i++)
    {/* Look for min MAX */
      if(coefs[i] > coefs[maxI])
	maxI = i;
      if(coefs[i] < coefs[minI])
	minI = i;
    }/* Look for min MAX */
  
  if(coefs[maxI] == coefs[minI])
    scaleF = 1.0;
  else
    scaleF = (600.0 / (float)((float)coefs[maxI] - (float)coefs[minI]));
  
  XtUnmanageChild(statusScreenWidget);
  XFlush(dsply);
  
  thisGC = image_gcYELLOW;
  XDrawLine(dsply, XtWindow(drawingWidget), thisGC,
	    10,display_height-100,
	    10,display_height-700);
  XDrawLine(dsply, XtWindow(drawingWidget), thisGC,
	    10+(0*thickStep),display_height-100,
	    10+(magchk*thickStep),display_height-100);
  sprintf(outStr,"Magnification");
  XDrawString(dsply, XtWindow(drawingWidget), image_gcYELLOW,
	      10+(thickStep*((magchk/2)-1)), display_height-20, 
	      outStr, strlen(outStr));
  sprintf(outStr,"Correlation");
  XDrawString(dsply, XtWindow(drawingWidget), image_gcBLUE,
	      10, display_height-720, 
	      outStr, strlen(outStr));
  isAneg = FALSE;
  if(coefs[minI] < 0.0)
    {/* Show them where zero is */
      isAneg = TRUE;
      XDrawLine(dsply, XtWindow(drawingWidget), image_gcBLUE,
		10+(0*thickStep),
		display_height-((int)((float)(0.0-coefs[minI])*
				      (float)scaleF)+
				100),
		10+(magchk*thickStep),
		display_height-((int)((float)(0.0-coefs[minI])*
				      (float)scaleF)+
				100)
		);
    }/* Show them where zero is */
  
  myTurn = FALSE;
  for(i=0;i<magchk;i++)
    {/* Do all */
      
      thisGC = image_gcRED;
      
      for(j=0;j<thinStep;j++)
	{/* Draw hsitogram */
	  if(isAneg)
	    {/* We do have negative values */
	      GC localThisGC;
	      if(coefs[i] < 0.0)
		localThisGC = image_gcGREEN;
	      else localThisGC = thisGC;
	      XDrawLine(dsply, XtWindow(drawingWidget), 
			localThisGC,
			10+(i*thickStep)+j,
			display_height-((int)((float)(0.0-coefs[minI])*
					      (float)scaleF)+
					100),
			10+(i*thickStep)+j,
			display_height-
			((int)((float)(coefs[i]-coefs[minI])*
			       (float)scaleF)+
			 100));
	    }/* We do have negative values */
	  else
	    XDrawLine(dsply, XtWindow(drawingWidget), 
		      thisGC,
		      10+(i*thickStep)+j,
		      display_height-100,
		      10+(i*thickStep)+j,
		      display_height-((int)((float)(coefs[i]-coefs[minI])*
					    (float)scaleF)+
				      100));
	}/* Draw hsitogram */
    }/* Do all */
  
  myTurn = FALSE;
  for(i=0;i<magchk;i++)
    {/* Do all */
      sprintf(outStr,"%.4f",mags[i]);
      /* Offset the mag print-out so you can read it */
      if(myTurn)
	XDrawString(dsply, XtWindow(drawingWidget), image_gcYELLOW,
		    10+(thickStep*i), display_height-50, 
		    outStr, strlen(outStr));
      else
	XDrawString(dsply, XtWindow(drawingWidget), image_gcYELLOW,
		    10+(thickStep*i), display_height-65, 
		    outStr, strlen(outStr));
      if(magchk > 10)
	myTurn = !myTurn; /* Only cycle for large numbers */
      sprintf(outStr,"%.4f",coefs[i]);
      if(coefs[i] < 0.0)
	XDrawString(dsply, XtWindow(drawingWidget), image_gcBLUE,
		    10+(thickStep*i), 
		    display_height-
		    ((int)((float)(coefs[i]-coefs[minI])*
			   (float)scaleF)+90),
		    outStr, strlen(outStr));
      else
	XDrawString(dsply, XtWindow(drawingWidget), image_gcBLUE,
		    10+(thickStep*i), 
		    display_height-
		    ((int)((float)(coefs[i]-coefs[minI])*
			   (float)scaleF)+110),
		    outStr, strlen(outStr));
      XFlush(dsply);
    }/* Do all */
  
  noDrawMag:;

  XFlush(dsply);
  XSync(dsply,0);

}/* DRAWMAG */
  


/***************************************************************
 * showBoxingHistogram() - Draw the histogram on screen showing
 * histogram of different boxing stats.
 */
void showBoxingHistogram(int tag)
{/* SHOWBOXINGHISTOGRAM */
  int i,j,k,maxHits;
  float minVal,maxVal;
  float scaleF,stepF,iter;
  bhist_t hist[512];

  for(i=0;i<256;i++)
    XDrawLine(dsply, XtWindow(drawingWidget), image_gcBLACK,
	      i,0,i,256);
  /* Now the axes */
  XDrawLine(dsply, XtWindow(drawingWidget), image_gcYELLOW,
	      0,220,256,220);
  XFlush(dsply);

  if(showBoxingLevel== AMMCORRCOEF)
    {/* DO corr coef */
      /* First find MIN and MAX */
      maxVal= minVal = box[0].correlation;
      for(i=0;i<numBoxes;i++)
	{/* Check all */
	  minVal = MIN(box[i].correlation,minVal);
	  maxVal = MAX(box[i].correlation,maxVal);
	}/* Check all */
    }/* DO corr coef */
  else if (showBoxingLevel==AMMAVGDENS)
    {/* DO avg dens */
      maxVal= minVal = box[0].avgdensity;
      for(i=0;i<numBoxes;i++)
	{/* Check all */
	  minVal = MIN(box[i].avgdensity,minVal);
	  maxVal = MAX(box[i].avgdensity,maxVal);
	}/* Check all */
    }/* DO avg dens */
  else if (showBoxingLevel==AMMSTDEV)
    {/* DO stdev */
      maxVal= minVal = box[0].stdev;
      for(i=0;i<numBoxes;i++)
	{/* Check all */
	  minVal = MIN(box[i].stdev,minVal);
	  maxVal = MAX(box[i].stdev,maxVal);
	}/* Check all */
    }/* DO stdev */
  stepF  = (maxVal-minVal) / 80.0;

  j = 0;
  for(iter=minVal;iter<maxVal;iter+=stepF)
    {/* setup Hist */
      /* Assign ranges */
      hist[j].rangeVal = iter;
      hist[j].hits = 0;
      j++;
    }/* setup Hist */
  for(i=0;i<numBoxes;i++)
    {/* Check all */
      for(j=0;j<79;j++)
	{/* Look at all hist cells */
	  if(showBoxingLevel== AMMCORRCOEF)
	    {/* DO corr coef */
	      if(box[i].correlation >= hist[j].rangeVal  &&
		 box[i].correlation < hist[j+1].rangeVal 
		 )
		{
		  hist[j].hits++;
		  goto foundIt;
		}
	    }/* DO corr coef */
	  else if (showBoxingLevel==AMMAVGDENS)
	    {/* DO avg dens */
	      if(box[i].avgdensity >= hist[j].rangeVal  &&
		 box[i].avgdensity < hist[j+1].rangeVal 
		 )
		{
		  hist[j].hits++;
		  goto foundIt;
		}
	    }/* DO avg dens */
	  else if (showBoxingLevel==AMMSTDEV)
	    {/* DO stdev */
	      if(box[i].stdev >= hist[j].rangeVal  &&
		 box[i].stdev < hist[j+1].rangeVal 
		 )
		{
		  hist[j].hits++;
		  goto foundIt;
		}
	    }/* DO stdev */
	}/* Look at all hist cells */
      /* If we got here than it's the last one */
      hist[79].hits++;
    foundIt:;
    }/* Check all */

  /* Now let's plot the histogram */
  maxHits = hist[0].hits;
  for(j=0;j<80;j++)
    if(hist[j].hits > maxHits)
      maxHits = hist[j].hits;
  scaleF = 200.0 / (float)maxHits;
  for(j=0;j<80;j++)
    {/* plot it */
      GC thisGC;
      thisGC = image_gcBLUE;
      if(threshLo >= hist[j].rangeVal &&
	 threshLo <= hist[j+1].rangeVal)
	thisGC = image_gcRED;
      if(threshHi >= hist[j].rangeVal &&
	 threshHi <= hist[j+1].rangeVal)
	thisGC = image_gcGREEN;
      if(j==79 && threshHi >= hist[79].rangeVal)
	thisGC = image_gcGREEN;
      for(i=0;i<3;i++)
	XDrawLine(dsply, XtWindow(drawingWidget), thisGC,
		  (j*3)+5+i,218,
		  (j*3)+5+i,219 - (int)((float)hist[j].hits*scaleF));

    }/* plot it */





}/* SHOWBOXINGHISTOGRAM */




/************************************************************
 * DrawWindow() - Redraw window whenever it's prudent.  Pay 
 * attention to different program states and draw the 
 * appropriate when needed.
 */
void DrawWindow()
{/* DRAWWINDOW */
  int i,j,k,l,tag;
  int arcRadius,thisFileType;
  char numStr[16];
  int draw_width, draw_height;
  int drawX,drawY,drawDestX,drawDestY,width,height;
  
  if(!drawLite)
    goto bugOut;

  thisFileType = fileType;
  if(fftSize==1024 && disp1024fft && XtIsManaged(fftscreen))
    {/* Special display of 1024 fft */
      thisFileType = PORTABLE;
    }/* Special display of 1024 fft */

  if(!subCanDrawWindow)
    {/* DO this for subtracting maps */

      /* Need to decide here if we are displaying MAG stuff or
       * radius stuff  */
      if(submapFindRad)
	{/* Draw the find radius stuff */
	  drawFindRad();
	}/* Draw the find radius stuff */
      else
	{/* Draw the mag stuff */
	  drawMag();
	}/* Draw the mag stuff */

      goto bugOut;
    }/* DO this for subtracting maps */

  if(globAccum)
    {/* Only do this for boxing */
      /* Stuff our display buffer  - we may have clusters displayed
       * from before... */
      stuffImage(int_image_buffer1, image_width, image_height, 
		 contrastStretch,TRUE);
    }/* Only do this for boxing */

  /* Show the clusters */
  if(clusterDisp && globAccum)
    {/* Display the clustered image */
      float scaleFactor;
      int dispX,dispY,pntX,pntY,paint_factor;
      int autoboxHeight, autoboxWidth;
      scaleFactor = (float)sizeFactor / (1.0 / (float)compress_factor);
      autoboxHeight = (int)file_height/compress_factor;
      autoboxWidth = (int)file_width/compress_factor;
      for(j=1;j<autoboxHeight-1;j++)
	{
	  dispY = (int)((float)j*scaleFactor);
	  dispY -= (int)((float)virYoffset*sizeFactor);
	  for(i=1;i<autoboxWidth-1;i++)
	    {/* cols */
	      dispX = (int)((float)i*scaleFactor);
	      dispX -= (int)((float)virXoffset * sizeFactor);
	      if(dispY > 0 && dispX > 0 &&
		 dispY < image_height &&
		 dispX < image_width )
		{/* OK to draw */
		  if(globAccum[j*autoboxWidth+i])
		    {/* Paint it blue */
		      paint_factor = (int)(scaleFactor * 
					   (float)compress_factor);
		      if(paint_factor < 1)
			paint_factor = 1; /* Got to at least show it once */
		      for(pntY=0;pntY<paint_factor;pntY++)
			for(pntX=0;pntX<paint_factor;pntX++)
			  {
			    if(hiColor)
			      {
				XPutPixel(image_image1, 
					  (dispX+pntX),
					  (dispY+pntY),
					  pixels_return[BLUE]);
			      }
			    else
			      image_buffer1[(((dispY+pntY)*image_width)+
					     dispX+pntX)]=
					       (BYTE)pixels_return[BLUE];
			  }
		    }/* Paint it blue */
		}/* OK to draw */
	    }/* cols */
	}
    }/* Display the clustered image */

  /* Draw all of image */
  draw_width = display_width;
  draw_height = display_height;
  if(XtIsManaged(main_window_widget))
    {/* Is managed */
      XPutImage(dsply, image_window1, image_gc, image_image1, 
		xoffset, yoffset,
		0, 0,  draw_width,  draw_height);
    }/* Is managed */

  /* Get ZOOMiness factor */
  calcSizeFactor(zoomFactor);

  if(haveParamFile)
    {/* We're trying to do the particle processing thing */
    }/* We're trying to do the particle processing thing */
  else
    {/* DO normal single file processing */

      /* Display any BOXes we may have */
      if(boxMethod == USECIRCLE)
	{/* Draw the circles */
	  /* Account for this being a large image and the coords 
	   * being in another "page" with virOffset
	   */
	  for(i=0;i<numBoxes;i++)
	    {/* Draw the circles */
	      drawX = (((box[i].x-virXoffset)*sizeFactor)-(xoffset));
	      drawY = (((box[i].y-virYoffset)*sizeFactor)-(yoffset));
	      if(boxRadiusDisp)
		{/* Draw box */
		  /* Size of the circle */
		  arcRadius = (int)(sizeFactor * boxRadius);
		  if(i==findBox || box[i].reference==1)
		    rwaXDrawArc(dsply, XtWindow(drawingWidget), image_gcRED,
				(drawX-arcRadius), 
				(drawY-arcRadius),
				(arcRadius*2),(arcRadius*2),
				0, (360*64));
		  else
		    rwaXDrawArc(dsply, XtWindow(drawingWidget), 
				image_gcRED,
				(drawX-arcRadius), 
				(drawY-arcRadius),
				(arcRadius*2),(arcRadius*2),
				0, (360*64));
		}/* Draw box */
	      if(featherRadiusDisp)
		{/* Draw feather */
		  /* Size of the circle */
		  arcRadius = (int)(sizeFactor * featherRadius);
		  if(i==findBox || box[i].reference==1)
		    rwaXDrawArc(dsply, XtWindow(drawingWidget), image_gcRED,
				(drawX-arcRadius), 
				(drawY-arcRadius),
				(arcRadius*2),(arcRadius*2),
				0, (360*64));
		  else
		    rwaXDrawArc(dsply, XtWindow(drawingWidget), 
				image_gcBLUE,
				(drawX-arcRadius), 
				(drawY-arcRadius),
				(arcRadius*2),(arcRadius*2),
				0, (360*64));
		  /* top */
		  XDrawLine(dsply, XtWindow(drawingWidget), 
			    image_gcBLUE,
			    (drawX-arcRadius), 
			    (drawY-arcRadius),
			    (drawX+arcRadius), 
			    (drawY-arcRadius)
			    );
		  /* bottom */
		  XDrawLine(dsply, XtWindow(drawingWidget), 
			    image_gcBLUE,
			    (drawX-arcRadius), 
			    (drawY+arcRadius),
			    (drawX+arcRadius), 
			    (drawY+arcRadius)
			    );
		  /* left */
		  XDrawLine(dsply, XtWindow(drawingWidget), 
			    image_gcBLUE,
			    (drawX-arcRadius), 
			    (drawY-arcRadius),
			    (drawX-arcRadius), 
			    (drawY+arcRadius)
			    );
		  /* right */
		  XDrawLine(dsply, XtWindow(drawingWidget), 
			    image_gcBLUE,
			    (drawX+arcRadius), 
			    (drawY-arcRadius),
			    (drawX+arcRadius), 
			    (drawY+arcRadius)
			    );
		}/* Draw feather */
	      if(emptyShellDisp)
		{/* Draw empty width */
		  /* Size of the circle */
		  arcRadius = (int)(sizeFactor * (emmaRadius-emptyShellWidth));
		  if(i==findBox || box[i].reference==1)
		    rwaXDrawArc(dsply, XtWindow(drawingWidget), image_gcWHITE,
				(drawX-arcRadius), 
				(drawY-arcRadius),
				(arcRadius*2),(arcRadius*2),
				0, (360*64));
		  else
		    rwaXDrawArc(dsply, XtWindow(drawingWidget), 
				image_gcWHITE,
				(drawX-arcRadius), 
				(drawY-arcRadius),
				(arcRadius*2),(arcRadius*2),
				0, (360*64));
		}/* Draw empty width */
	      if(emmaRadiusDisp)
		{/* Draw emma */
		  /* Size of the circle */
		  arcRadius = (int)(sizeFactor * emmaRadius);
		  if(i==findBox || box[i].reference==1)
		    rwaXDrawArc(dsply, XtWindow(drawingWidget), image_gcRED,
				(drawX-arcRadius), 
				(drawY-arcRadius),
				(arcRadius*2),(arcRadius*2),
				0, (360*64));
		  else
		    rwaXDrawArc(dsply, XtWindow(drawingWidget), 
				image_gcGREEN,
				(drawX-arcRadius), 
				(drawY-arcRadius),
				(arcRadius*2),(arcRadius*2),
				0, (360*64));
		}/* Draw emma */
	      /* Size of the circle */
	      arcRadius = (int)(sizeFactor * boxRadius);
	      sprintf(numStr,"%d",(lastNumBox+(i+1)));
	      if(dispMethod == CIRCLEDISP)
		{/* Draw circle numbers */
		  if(i==findBox || box[i].reference==1)
		    XDrawString(dsply, XtWindow(drawingWidget), image_gcRED,
				drawX, drawY, numStr, strlen(numStr));
		  else
		    XDrawString(dsply, XtWindow(drawingWidget), 
				boxing_GC,
				drawX, drawY, numStr, strlen(numStr));
		}/* Draw circle numbers */
	      if(debugSwitch && debugLevel==4)
		{/* Try giving this feedback */
		  if(showDebug)
		    {/* OK to do it */
		      sprintf(numStr,"Cor %.3f",box[i].correlation);
		      XDrawString(dsply, XtWindow(drawingWidget), 
				  image_gcBLUE,
				  drawX-arcRadius+5, drawY-5, 
				  numStr, strlen(numStr));
		      sprintf(numStr,"Avg %.3f",box[i].avgdensity);
		      XDrawString(dsply, XtWindow(drawingWidget), 
				  image_gcBLUE,
				  drawX-arcRadius+5, drawY+5, 
				  numStr, strlen(numStr));
		    }/* OK to do it */
		}/* Try giving this feedback */
	    }/* Draw the circles */
	}/* Draw the circles */
      else if((boxMethod==USERECT || boxMethod==USESQUARE) &&
	      rectBox.drawable)
	{/* RECT or SQUARE */
	  /* We use vir?offset so that we take into account that this 
	   * guy might be on another virtual page 
	   */
	  width = rectBox.lowx - rectBox.x;
	  height = rectBox.lowy - rectBox.y;
	  width = ((width*sizeFactor));
	  height = ((height*sizeFactor));
	  drawX = (((rectBox.x-virXoffset)*sizeFactor)-(xoffset));
	  drawY = (((rectBox.y-virYoffset)*sizeFactor)-(yoffset));
	  XDrawRectangle(dsply, XtWindow(drawingWidget), image_gcGREEN,
			 drawX, 
			 drawY,
			 width,
			 height);
	  rectBox.width  = rectBox.lowx - rectBox.x;
	  rectBox.height = rectBox.lowy - rectBox.y;
	  /* Draw rect dimensions */
	  sprintf(numStr,"[%d]x[%d]",
		  (int)(rectBox.width),
		  (int)(rectBox.height)
		  );
	  XDrawString(dsply, XtWindow(drawingWidget), 
		      image_gcRED,
		      drawX+15, drawY+15, numStr, strlen(numStr));
	}/* RECT or SQUARE */
      else if(boxMethod==USEPOLY)
	{/* POLY */
	  for(i=0;i<(numPoly-1);i++)
	    {/* Draw the lines */
	      /* We use vir?offset so that we take into account that this 
	       * guy might be on another virtual page 
	       */
	      drawX= (((int)((poly[i].x-virXoffset)*sizeFactor))-(xoffset));
	      drawY= (((int)((poly[i].y-virYoffset)*sizeFactor))-(yoffset));
	      
	      drawDestX= (((int)((poly[i+1].x-virXoffset)*sizeFactor))-
			  (xoffset));
	      drawDestY= (((int)((poly[i+1].y-virYoffset)*sizeFactor))-
			  (yoffset));
	      
	      XDrawLine(dsply, XtWindow(drawingWidget), image_gcGREEN,
			drawX, drawY,
			drawDestX, drawDestY);
	    }/* Draw the lines */
	}/* POLY */
    }/* DO normal single file processing */
  
  /* Check to see if we should be drawing numbers for our packed
   * images */
  if((thisFileType==PACKEDPORTABLE || thisFileType==MAP) &&
     (dispMethod==CIRCLEDISP) &&
     !projASU)
    {/* Draw the image numbers */
      for(i=(imgNum-1);i<lastImg;i++)
	{/* Do all visible */
	  if(thisFileType==PACKEDPORTABLE)
	    /* e.g. we may have sorted from the partproc screen */
	    sprintf(numStr,"%d",imgNums[i].particle_number);
	  else sprintf(numStr,"%d",i+1);
	  if(dispMethod == CIRCLEDISP)
	    {/* Draw circle numbers */
	      XDrawString(dsply, XtWindow(drawingWidget), image_gcPLACEHOLDER,
			  (imgNums[i].x+5),	  
			  (imgNums[i].y+15), 
			  numStr, strlen(numStr));
	    }/* Draw circle numbers */
	}/* Do all visible */
    }/* Draw the image numbers */
  
  if(thisFileType==PACKEDPORTABLE)
    {/* PAcked file images may be deleted */
      for(l=(imgNum-1);l<lastImg;l++)
	{/* Look at all */
	  if(deletePackedImage[l])
	    {/* Mark it so in the display */
	      int drawDestX,drawDestY,colOffset,rowOffset;
	      colOffset= imgNums[l].x;
	      rowOffset= imgNums[l].y;
	      drawDestX= colOffset+(int)((2*packRadius)+1)*sizeFactor;
	      drawDestY= rowOffset+(int)((2*packRadius)+1)*sizeFactor;
	      XDrawLine(dsply, XtWindow(drawingWidget), image_gcRED,
			colOffset, rowOffset,
			drawDestX, drawDestY);
	    }/* Mark it so in the display */
	}/* Look at all */
    }/* PAcked file images may be deleted */
  
  /* Now check to see if we've got to draw the radius for correlation
   * averaging screen */
  if(XtIsManaged(emcoravgscreen))
    {/* Draw radius */
      arcRadius = (int)(sizeFactor * coravgRadius);
      rwaXDrawArc(dsply, XtWindow(drawingWidget), image_gcRED,
		  (lastPlotClickX-arcRadius), 
		  (lastPlotClickY-arcRadius),
		  (arcRadius*2),(arcRadius*2),
		  0, (360*64));
    }/* Draw radius */

  /* Now check to see if we've got to draw the regions for fix2d */
  if(XtIsManaged(fix2Dscreen))
    {/* OK to do */
      int  i,j,stepX,stepY,thisX,thisY,afW,afH;
      char *fixText;
      int fixRegions;
      
      if(drawFix2Dgrid)
	{/* draw it */
	  fixText = XmTextGetString(fix2DRegionsTextW); 
	  fixRegions = (int)strtod(fixText,0);
	  fixRegions = MAX(MIN(fixRegions,MAXREGIONS),1);
	  XtFree(fixText);
	  
	  stepX = (int)(file_width/fixRegions);
	  stepY = (int)(file_height/fixRegions);
	  
	  afW = (int)(file_width*sizeFactor);
	  afH = (int)(file_height*sizeFactor);
	  afW = MIN(afW,(int)(display_width));
	  afH = MIN(afH,(int)(display_height));

	  for(i=0;i<=file_width;i+=stepX)
	    for(j=0;j<=file_height;j+=stepY)
	      {/* DO all */
		thisX = (int)((i - virXoffset) * sizeFactor);
		thisY = (int)((j - virYoffset) * sizeFactor);
		if((thisX >= 0) &&
		   (thisY >= 0) &&
		   ((thisX <= display_width) ||
		    (thisY <= display_height))
		   )
		  {/* OK to draw */
		    if(thisY < display_height)
		      XDrawLine(dsply, XtWindow(drawingWidget), image_gcGREEN,
				0, (int)thisY,
				(int)afW, (int)thisY);
		    if(thisX < display_width)
		      XDrawLine(dsply, XtWindow(drawingWidget), image_gcGREEN,
				(int)thisX, 0,
				(int)thisX, (int)afH);
		  }/* OK to draw */
		XFlush (dsply);

		if(drawFix2Dstats)
		  {/* draw it */
		    int drawX,drawY;
		    drawX = (int)(i/stepX);
		    drawY = (int)(j/stepY);
		    if((drawX < fixRegions) &&
		       (drawY < fixRegions) &&
		       (thisX >= 0) &&
		       (thisY >= 0) 
		       )
		      {/* Only do valid */
			sprintf(numStr,"Mean[%.2f]",
				fix2Davg[(drawX)][(drawY)]);
			XDrawString(dsply, XtWindow(drawingWidget), 
				    image_gcWHITE,
				    (int)(thisX+5),	  
				    (int)(thisY+20), 
				    numStr, strlen(numStr));
			sprintf(numStr,"Std[%.2f]",
				fix2Dstd[(drawX)][(drawY)]);
			XDrawString(dsply, XtWindow(drawingWidget), 
				    image_gcWHITE,
				    (int)(thisX+5),	  
				    (int)(thisY+40), 
				    numStr, strlen(numStr));
			XFlush (dsply);
		      }/* Only do valid */
		  }/* draw it */

	      }/* DO all */
	}/* draw it */

    }/* OK to do */

  if(showBoxingSwitch)
    showBoxingHistogram(tag);

  XFlush (dsply);
 bugOut:;

}/* DRAWWINDOW */




/************************************************************
 * DrawWindowExpose() - Redraw window e.g. on expose events.
 */
void DrawWindowExpose(Widget  widget,
		      int     *tag,
		      XmAnyCallbackStruct *callback_data)
{/* DRAWWINDOWEXPOSE */
  
  DrawWindow();

}/* DRAWWINDOWEXPOSE */




/* ************************************************************
 * drawingWindow_create() - Remember the drawing area window
 * number for use later. Also remember other created widgets
 * Future goal - try and get a more consistent rtn for the
 * creation of all widgets - e.g. labels/toggles/pushbuttons.
 */
void drawingWindow_create(Widget widget, int *tag, 
			  XmAnyCallbackStruct *callback_data)
{/* DRAWINGWINDOW_CREATE */
  
  /* Remember the widget number for later */
  if(*tag==COLOR)
    colormapWidget= widget;
  else if(*tag==TEST)
    testcolormapWidget = widget;
  else if(*tag==COLOREXIT)
    colorExitW = widget;
  else if(*tag==HIST)
    histWidget= widget;
  else if(*tag==EDITBOX)
    editBoxWidget= widget;
  else if(*tag==CIRCAVG)
    circAvgWidget= widget;
  else if(*tag==EMCORAVG)
    emcoravgDrawWidget= widget;
  else if(*tag==FFT)
    fftDrawingWidget= widget;
  else if(*tag==COLORMAPSCROLL)
    {/* Up */
      colormapScrollWidget= widget;
      handle_cr(widget);
    }/* Up */
  else if(*tag==COLORMAPSCROLLDOWN)
    {/* Down */
      colormapScrollDownWidget= widget;
      handle_cr(widget);
    }/* Down */
  
}/* DRAWINGWINDOW_CREATE */



/* @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
 * plotClick() - Every time the mouse button is clicked call
 * here.  This routine handles the selecting/deselecting of
 * box points and ruler points on the screen.
 *            Button1 is the Left mouse button
 *            Button2 is the Middle mouse button
 *            Button3 is the Right mouse button
 * Do the appropriate based on the current state of the program
 * parameters selected from the interface.
 */
void plotClick(Widget widget, int *tag, 
	       XmDrawingAreaCallbackStruct *draw_data)
{/* PLOTCLICK */
  XEvent event;
  int x, y, k, reDraw;
  char ans;
  char infoStr[1024];
  static int thisBox;
  int thisImg,sectY,sectX;
  GC thisGC;

  /* Do appropriate depending on event type */
  if(draw_data != 0)
    {/* Only look at events meant for us */

      event= *(draw_data->event);

      if(pointColor == 0)
	thisGC = image_gcRED;
      else if(pointColor == 1)
	thisGC = image_gcGREEN;
      else if(pointColor == 2)
	thisGC = image_gcBLUE;

      switch (event.type) 
	{/* Find out which event we have */
	case ButtonRelease:
	  x = (int)(event.xbutton.x);      /* Get x,y copords of point pos */
	  y = (int)(event.xbutton.y);
	  reDraw = TRUE;

	  if(boxZoom == USEBOX && isBoxing)
	    {/* Draw the appropriate box */	
	      if(boxMethod==USECIRCLE)
		{/* CIRCLE */
		  if((event.xbutton.button == Button1))
		    {/* PUSH it */
		      pushBox(x,y,FALSE,&thisBox);
		    }/* PUSH it */
		  if((event.xbutton.button == Button2))
		    {/* Remove it */
		      removeBox(x, y, FALSE);
		    }/* Remove it */
		  if((event.xbutton.button == Button3))
		    {/* Pop last one on the list */
		      /* DON't DO THIS ANYMORE popBox(); */
		    }/* Pop last one on the list */
		}/* CIRCLE */
	      else if(boxMethod==USERECT || boxMethod==USESQUARE)
		{/* RECT or SQUARE */
		  pushRectBox(event.xbutton.button, x, y);
		}/* RECT or SQUARE */
	      else if(boxMethod==USEPOLY)
		{/* POLY */
		  if((event.xbutton.button == Button1))
		    {/* PUSH it */
		      pushPolyBox(x,y);
		    }/* PUSH it */
		  if((event.xbutton.button == Button2))
		    {/* Pop top of stack */
		      popPolyBox();
		    }/* Pop top of stack */
		  if((event.xbutton.button == Button3))
		    {/* Push last one on the list */
		      /* Write last leg of polygon */
		      poly[numPoly].x = poly[0].x;
		      poly[numPoly].y = poly[0].y;
		      numPoly++;
		      
		    }/* Push last one on the list */
		}/* POLY */
	      
	      isBoxing = FALSE;

	      if(reDraw)
		DrawWindow();
	    }/* Draw the appropriate box */	
	  
	  break;
	  
	case ButtonPress:
	  /* They are trying to find points */
	  x = (int)(event.xbutton.x);      /* Get x,y copords of point pos */
	  y = (int)(event.xbutton.y);
	  reDraw = TRUE;

	  lastPlotClickX = x;  /* Rememebr these for things like emcoravg */
	  lastPlotClickY = y;  
	  if(XtIsManaged(emcoravgscreen))
	    DrawWindow(); /* Show new location */

	  if(christmasTime)
	    {/* Only blink lights for christmas */
	      int i,j;
	      for(i=0;i<buffer_height;i++)
		for(j=0;j<buffer_width;j++)
		  {
		    if(image_buffer1[(i*buffer_width)+j]== 
		       pixels_return[GREEN])
		      image_buffer1[(i*buffer_width)+j] = pixels_return[22];
		    else if(image_buffer1[(i*buffer_width)+j]== 
		       pixels_return[BLUE])
		      image_buffer1[(i*buffer_width)+j] = pixels_return[32];
		    else if(image_buffer1[(i*buffer_width)+j]== 
		       pixels_return[PLACEHOLDER])
		      image_buffer1[(i*buffer_width)+j] = pixels_return[33];
		    else if(image_buffer1[(i*buffer_width)+j]== 
		       pixels_return[22])
		      image_buffer1[(i*buffer_width)+j] = pixels_return[GREEN];
		    else if(image_buffer1[(i*buffer_width)+j]== 
		       pixels_return[32])
		      image_buffer1[(i*buffer_width)+j] = pixels_return[BLUE];
		    else if(image_buffer1[(i*buffer_width)+j]== 
		       pixels_return[33])
		      image_buffer1[(i*buffer_width)+j] = 
			pixels_return[PLACEHOLDER];
		  }
	    }/* Only blink lights for christmas */

	  if(haveParamFile && newXY)
	    {/* Find a new x,y center for this */
	      P_INFO *thisInfo;
	      float newX,newY;
	      int thisX,thisY;

	      newX = (float)((float)x * recipSizeFactor);
	      newY = (float)((float)y * recipSizeFactor);
	      newY = purdueHeader.numRows - newY;

	      if(foundPart != NULL &&
		 (foundPart->rank==imgNum))
		{/* OK to go  */
		  thisInfo = foundPart;
		}/* OK to go  */
	      else
		{/* Find it */
		  thisInfo = info.info;
		  while(thisInfo != NULL)
		    {/* Find this one */
		      if(thisInfo->rank==imgNum)
			{/* Found it */
			  break;
			}/* Found it */
		      thisInfo = thisInfo->next;
		    }/* Find this one */
		}/* Find it */
	      
	      if((event.xbutton.button == Button1))
		{/* They want to change the position */
		  if(newX > purdueHeader.numCols ||
		     newY > purdueHeader.numRows)
		    {/* Error */
		      sprintf(outStr,
			      "[%d][%d] is outside the valid range [%d][%d]\n",
			      (int)newX,(int)newY,
			      purdueHeader.numCols,
			      purdueHeader.numRows);
		      askThem(outStr,FALSE,XmDIALOG_INFORMATION);
		    }/* Error */
		  else
		    {/* OK to continue */
		      thisInfo->x = newX;
		      thisInfo->y = newY;
		    }/* OK to continue */
		}/* They want to change the position */
	      else if((event.xbutton.button == Button2))
		{/* They want to toggle it */
		  /* OLD way - we toggled the state
		   *  thisInfo->toggle = !thisInfo->toggle;
		  */
		  thisInfo->x = (float)(((float)purdueHeader.numCols / 
							2.0)-0.5);
		  thisInfo->y = (float)(((float)purdueHeader.numRows / 
							2.0)-0.5);
		}/* They want to toggle it */
	      else if((event.xbutton.button == Button3))
		{/* Restore original */
		  thisInfo->x = thisInfo->origX;
		  thisInfo->y = thisInfo->origY;
		}/* Restore original */
	      
	      /* Recalc these in case they've changed */
	      thisInfo->deltaX = thisInfo->x - (float)(((float)
							purdueHeader.numCols / 
							2.0)-0.5);
	      thisInfo->deltaY = thisInfo->y - (float)(((float)
							purdueHeader.numRows / 
							2.0)-0.5);
	      /* Display this on the parm screen */
	      convertXY(&thisX, &thisY, 
			thisInfo->phi, thisInfo->theta);
	      findNearestParticle(&thisX,&thisY,&foundPart,FALSE,
				  thisInfo->rank, TRUE,0);
	    }/* Find a new x,y center for this */
	  else
	    {/* Normal */
	      if(boxZoom == USEPOINT)
		{/* Draw the pointing info */	
		  int imageX,imageY;
		  float rulerLength;
		  
		  /* Left button is just give me the coords and
		   * pixel value */
		  if((event.xbutton.button == Button1))
		    {/* Draw the coords and pixel value */
		      /* Added by Brant C. French 7-3-97 */
		      /* This is used only if someone presses button 1 */
		      /* when the histogram option is selected */
		      
		      if (histMethod) 
			{ /* If histogram is registered */
			  int start_x=0,start_y=display_height-50;
			  
			  /* Draw a line across the screen showing what 
			   * histogram they are seeing */
			  XDrawLine(dsply,XtWindow(drawingWidget),
				    thisGC,
				    0,y,buffer_width,y);
			  
			  /* Lets draw the histogram now across the bottom 
			   * of the screen */
			  for (k=0;k<buffer_width;k++) {
			    XDrawLine(dsply, XtWindow(drawingWidget), 
				      thisGC,
				      start_x,start_y,k,
				      display_height-50-
				      (2*(int)image_buffer1[(((y+yoffset)*
					      buffer_width)+(k+xoffset))]));
			    start_x=k; 
			    start_y=display_height-50-
			      (2*(int)image_buffer1[(((y+yoffset)*
					       buffer_width)+(k+xoffset))]);
			  } /* Go across the image */
			} /* If histogram is registered */
		      else 
			{ /* If not histogram */
			  globX = x;
			  globY = y;
			  /* First map coords to image coords */
			  /* Also changed sizeFactor to recipSize Factor 
			   * below. */
			  calcSizeFactor(zoomFactor);
			  
			  /* Give coords in this virtual page only */
			  imageX = ((x*recipSizeFactor)+
				    (xoffset*recipSizeFactor));
			  imageY = ((y*recipSizeFactor)+
				    (yoffset*recipSizeFactor));
			  oldGlobX= imageX;
			  oldGlobY= imageY;
			    /* If they're measuring in their depth cued map
			     * they may be zoomed past image_width e.g. */
			    if(TRUE)
			      {/* Now print the info */
				findRealValue(x,y,&thisImg,
					      &sectY,&sectX);
				sprintf(infoStr,"[%d,%d][%f]",
					sectX,
					sectY,
					findRealValue(x,y,&thisImg,
						      &sectY,&sectX));
				/*
				   XDrawString(dsply, XtWindow(drawingWidget), 
				   image_gcRED,
				   x, y, infoStr, strlen(infoStr));
				   */
				/* Draw an 'X' */
				XDrawLine(dsply, XtWindow(drawingWidget), 
					  thisGC,
					  x-1,y-1,x+1,y+1);
				XDrawLine(dsply, XtWindow(drawingWidget), 
					  thisGC,
					  x-1,y+1,x+1,y-1);
				drawPointScreen(infoStr,2);
				/* Zero out others */
				sprintf(infoStr,"");
				drawPointScreen(infoStr,3);
				if(showCC)
				  {/* Let's do this in the fourth one */
				    float saveCorrelation;
				    int saveX,saveY,saveRef;
				    saveX = box[0].x;
				    saveY = box[0].y;
				    saveCorrelation = box[0].correlation;
				    saveRef = box[0].reference;
				    box[0].y = virYoffset+imageY;
				    box[0].x = virXoffset+imageX;
				    printf("Looking at [%d,%d]\n",
					   box[0].x,
					   box[0].y);
				    ref_correlate_activate();
				    sprintf(infoStr,"CorrCoef[%f]",
					    box[0].correlation);
				    box[0].x = saveX;
				    box[0].y = saveY;
				    box[0].correlation = saveCorrelation;
				    box[0].reference   = saveRef;
				  }/* Let's do this in the fourth one */
				drawPointScreen(infoStr,4);
			      }/* Now print the info */
			} /* If not histogram */
		    }/* Draw the coords and pixel value */
		  else if((event.xbutton.button == Button2))
		    {/* Draw ruler line */
		      
		      /* Added by Brant C. French 7-3-97 */
		      /* This is used only if someone presses button 2 */
		      /* when the histogram option is selected */
		      
		      if (histMethod) 
			{ /* If histogram is registered */
			  int start_x=0,start_y=display_height-50;
			  
			  /* Draw a line down the screen to show where the 
			   * histogram */
			  /* was taken from */
			  XDrawLine(dsply,XtWindow(drawingWidget),
				    image_gcGREEN,
				    x,0,x,buffer_height);
			  
			  /* Lets draw the histogram now across the bottom 
			   * of the screen */
			  for (k=0;k<buffer_height;k++) {
			    XDrawLine(dsply, XtWindow(drawingWidget), 
				      image_gcGREEN,
				      start_x,start_y,k,
				      display_height-50-
				      (2*(int)image_buffer1[(((k+yoffset)*
					       buffer_width)+(x+xoffset))]));
			    start_x=k; 
			    start_y=display_height-50-
			      (2*(int)image_buffer1[(((k+yoffset)*
						buffer_width)+(x+xoffset))]);
			  } /* Go across the image */
			} /* If histogram is registered */
		      else 
			{ /* No histogram */
			  /* Also changed sizeFactor to recipSizeFactor
			   * below. */
			  calcSizeFactor(zoomFactor);
			  
			  /* Get these for this virtual page only */
			  imageX = ((x*recipSizeFactor)+
				    (xoffset*recipSizeFactor));
			  imageY = ((y*recipSizeFactor)+
				    (yoffset*recipSizeFactor));
			  rulerLength = (double)sqrt((double)
						     ((imageX-oldGlobX)*
						      (imageX-oldGlobX)) + 
						     ((imageY-oldGlobY)*
						      (imageY-oldGlobY)));
			  
			    /* If they're measuring in their depth cued map
			     * they may be zoomed past image_width e.g. */
			    if(TRUE)
			      {/* Now print the info */
				int dispX,dispY;
				findRealValue(x,y,&thisImg,
					      &sectY,&sectX);
				sprintf(infoStr,"[%d,%d][%f]",
					sectX,
					sectY,
					findRealValue(x,y,&thisImg,
						      &sectY,&sectX));
				drawPointScreen(infoStr,3);
				/* 
				   XDrawString(dsply, XtWindow(drawingWidget), 
				   thisGC,
				   x, y, infoStr, strlen(infoStr));
				   */
				XDrawLine(dsply, XtWindow(drawingWidget), 
					  thisGC,
					  x,y,globX,globY);
				/* Display the length */
				/* First which one is higher? - 
				 * make glob higher */
				if(y > globY)
				  {/* Switch them */
				    int temp;
				    temp = globY;
				    globY= y;
				    y= temp;
				  }/* Switch them */
				if(x > globX)
				  {/* Switch them */
				    int temp;
				    temp = globX;
				    globX= x;
				    x= temp;
				  }/* Switch them */
				dispX = (x + ((globX - x)/2)) + 15;
				dispY = (y + ((globY - y)/2)) + 15;
				
				if(rulerMag > 0.0 && rulerPixelWidth > 0.0)
				  sprintf(infoStr,"Length= [%f] Angs\n\
    [%f] pixels",
					  (rulerLength* (rulerPixelWidth)),
					  rulerLength);
				else sprintf(infoStr,"Length= [%f] pixels",
					     rulerLength);
				drawPointScreen(infoStr,4);
				/*
				   XDrawString(dsply, XtWindow(drawingWidget), 
				   image_gcGREEN,
				   dispX, dispY, infoStr, strlen(infoStr));
				   */
			      }/* Now print the info */
			} /* No histogram */
		    }/* Draw ruler line */
		  else if((event.xbutton.button == Button3))
		    {/* reDraw window */
		      DrawWindow();
		      sprintf(infoStr,"");
		      drawPointScreen(infoStr,2);
		      drawPointScreen(infoStr,3);
		      drawPointScreen(infoStr,4);
		    }/* reDraw window */
		}/* Draw the pointing info */	
	      
	      else if(boxZoom == USEBOX)
		{/* Draw the appropriate box */	

		  if(fileType==PACKEDPORTABLE && !crysDisping)
		    {/* Do the packed delete thing */
		      int width,height,sectionsPerRow,rowsPerImage,thisSec;
		      
		      if((event.xbutton.button == Button1))
			{/* (De)selecting one at a time */
			  
			  width  = purdueHeader.numCols;
			  height = purdueHeader.numRows;
			  /* Now find out how we're going to put these on 
			   * the display */
			  if(buffer_width < display_width)
			    sectionsPerRow= (int)(buffer_width / 
						  (int)(width*sizeFactor));
			  else
			    sectionsPerRow= (int)(display_width / 
						  (int)(width*sizeFactor));
			  sectionsPerRow= MAX(1,sectionsPerRow);
			  if(buffer_height < display_height)
			    rowsPerImage = buffer_height / 
			      (int)(height*sizeFactor);
			  else
			    rowsPerImage = display_height / 
			      (int)(height*sizeFactor);
			  rowsPerImage = MAX(1,rowsPerImage);
			  thisSec = ((int)((int)x / (int)(width*sizeFactor))) 
			    + 
			    (sectionsPerRow * (int)((int)y / 
						    (int)(height*sizeFactor)));
			  /* they may be displaying images 42-78 so take that
			   * into account here. */
			  thisSec += (imgNum-1);
			  thisSec = MAX((imgNum-1),
					MIN(thisSec,(purdueHeader.nimg)));
			  
			  deletePackedImage[thisSec]= 
			           !deletePackedImage[thisSec];
			  stuffPackedImage(purdueHeader.numCols,
					   purdueHeader.numRows,
					   packImg,purdueHeader.nimg,
					   imgNum);/* Start at radius image*/
			  lastDeleted= thisSec;  /* Remember for button 2*/
			  
			}/* (De)selecting one at a time */
		      if((event.xbutton.button == Button2))
			{/* (De)selecting a range */
			  
			  width  = purdueHeader.numCols;
			  height = purdueHeader.numRows;
			  /* Now find out how we're going to put these on 
			   * the display */
			  if(buffer_width < display_width)
			    sectionsPerRow= (int)(buffer_width / 
						  (int)(width*sizeFactor));
			  else
			    sectionsPerRow= (int)(display_width / 
						  (int)(width*sizeFactor));
			  sectionsPerRow= MAX(1,sectionsPerRow);
			  if(buffer_height < display_height)
			    rowsPerImage = buffer_height / 
			      (int)(height*sizeFactor);
			  else
			    rowsPerImage = display_height / 
			      (int)(height*sizeFactor);
			  rowsPerImage = MAX(1,rowsPerImage);
			  thisSec = ((int)((int)x / (int)(width*sizeFactor))) 
			    + 
			      (sectionsPerRow * (int)((int)y / 
						  (int)(height*sizeFactor)));
			  /* they may be displaying images 42-78 so take that
			   * into account here. */
			  thisSec += (imgNum-1);
			  thisSec = MAX((imgNum-1),
					MIN(thisSec,(purdueHeader.nimg)));
			  
			  for(i=(lastDeleted+1);i<=thisSec;i++)
			    deletePackedImage[i]= !deletePackedImage[i];
			  stuffPackedImage(purdueHeader.numCols,
					   purdueHeader.numRows,
					   packImg,purdueHeader.nimg,
					   imgNum); /* Start at radius image */
			}/* (De)selecting a range */
		      else if((event.xbutton.button == Button3))
			{/* (De)selecting all of them */
			  int i;
			  for(i=0;i<purdueHeader.nimg;i++)
			    deletePackedImage[i]= !deletePackedImage[i];
			}/* (De)selecting all of them  */
		      DrawWindow();
		    }/* Do the packed delete thing */
		  else
		    {/* Normal boxing */
		      isBoxing = TRUE;
		      lastBoxY= lastBoxX= -999;
		      pushBox(x,y,TRUE, &thisBox);
		    }/* Normal boxing */
		  
		}/* Draw the appropriate box */
	      else if(boxZoom == USEZOOM)
		{/* Zoom this puppy */
		  if(oldSizeFactor==-99.0)
		    {/* Init this guy */
		      oldSizeFactor= sizeFactor;  
		      oldZoomFactor= zoomFactor;
		    }/* Init this guy */
		  if((event.xbutton.button == Button1))
		    {/* Zoom in */
		      if(fullScreen)
			{/* CAlc coords */
			  x *= recipSizeFactor;  /* Put into orig coords */
			  y *= recipSizeFactor;
			  x *= oldSizeFactor;    /* Put into display coords */
			  y *= oldSizeFactor;
			}/* CAlc coords */
		      zoomFactor= (float)((int)(zoomFactor+1.0));
		      if(zoomFactor > 4.0)
			zoomFactor= 4.0;
		      fullScreen= FALSE;
		    }/* Zoom in */
		  else if((event.xbutton.button == Button2))
		    {/* Zoom out */
		      if(fullScreen)
			{/* CAlc coords */
			  x *= recipSizeFactor;  /* Put into orig coords */
			  y *= recipSizeFactor;
			  x *= oldSizeFactor;    /* Put into display coords */
			  y *= oldSizeFactor;
			}/* CAlc coords */
		      zoomFactor= (float)((int)(zoomFactor-1.0));
		      if(zoomFactor < -4.0)
			zoomFactor= -4.0;
		      fullScreen= FALSE;
		    }/* Zoom out */
		  else if((event.xbutton.button == Button3))
		    {/* Place Holder */
		      if(fullScreen)
			{/* Was fs */
			  float saveZF;
			  /* Suppose we do this - stuff zoomFactor with
			   * oldZoomFactor and then toggle fullScreen?? */
			  saveZF = zoomFactor;
			  zoomFactor = oldZoomFactor;
			  x *= recipSizeFactor;  /* Put into orig coords */
			  y *= recipSizeFactor;
			  x *= oldSizeFactor;    /* Put into display coords */
			  y *= oldSizeFactor;
			  fullScreen= FALSE;
			}/* Was fs */
		      else 
			{/* Make it FS */
			  virXoffset= virYoffset=xoffset= yoffset= 0;
			  x=y=0;
			  /* zoomFactor = 0.0; May not have to set this ... */
			  fullScreen= TRUE;
			}/* Make it FS */
		    }/* Place Holder */
		  
		  /* Remember for next time */
		  oldZoomFactor = zoomFactor;
		  oldSizeFactor = sizeFactor;
		  
		  calcSizeFactor(zoomFactor);   
		  sprintf(outStr,"%.2f",  sizeFactor);
		  XmTextSetString(zoomTextW, outStr); 
		  
		  virtual_read_activate(x,y);
		  
		}/* Zoom this puppy */
	    }/* Normal */
	  
	  break;
	  
	case KeyPress:
	  ans= XLookupKeysym((XKeyEvent *)&event, 0);
	  
	  /* Move box based on what key they've pressed from the
	   * keynoard */
	  if((boxZoom == USEBOX) && (rectBox.drawable) && (!locked) &&
	     (boxMethod==USERECT || boxMethod==USESQUARE))
	    {/* Boxing square or rect */
	      if(ans=='S')
		{/* Right */
		  if(boxMethod==USESQUARE)
		    {rectBox.lowx++;
		     rectBox.lowy++;}
		  else rectBox.lowx++;
		}/* Right */
	      else if(ans=='R')
		{/* Up */
		  if(boxMethod==USESQUARE)
		    {rectBox.lowx++;
		     rectBox.lowy++;}
		  else rectBox.lowy++;
		}/* Up */
	      else if(ans=='Q')
		{/* Left */
		  if(boxMethod==USESQUARE)
		    {rectBox.lowx--;
		     rectBox.lowy--;}
		  else rectBox.lowx--;
		}/* Left */
	      else if(ans=='T')
		{/* Down */
		  if(boxMethod==USESQUARE)
		    {rectBox.lowx--;
		     rectBox.lowy--;}
		  else rectBox.lowy--;
		}/* Down */
	      /* Trap for end points */
	      if(rectBox.lowx <= rectBox.x)
		rectBox.lowx= rectBox.x + 1;
	      if(rectBox.lowy <= rectBox.y)
		rectBox.lowy= rectBox.y + 1;
	      DrawWindow();
	      calcSizeFactor(zoomFactor);
	      pointerLabelDraw(0,0);
	    }/* Boxing square or rect */

	  break;
	  
	default:
	  break;
	}/* Find out which event we have */
    }/* Only look at events meant for us */
  
  XFlush(dsply);            /* Make changes current */

  
}/* PLOTCLICK */



/* @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
 * plotMotion() - Whenever the pointer moves across our 
 * window call here.  This is where we redraw the circles
 * if they're moving a box or print the coords of the
 * pointer as they move.
 * This rtn is called a LOT so be very frugal in what you
 * put in here because it will be executed over and over
 * again as the cursor travels over the image(s).
*/
void plotMotion(Widget widget, caddr_t unused, 
		register XMotionEvent *pEvent)
{/* plotMotion */
  int x, y,lastDraw;
  int i, j, k;
  char infoStr[1024];

  /* Make all changes current */
  XFlush(dsply);               

  switch (pEvent->type) 
    {/* Find out which event we have */
	  
    case MotionNotify:
	      
      /* Get coords of pointer */
      x = (int)(pEvent->x);     
      y = (int)(pEvent->y);

      if(!subCanDrawWindow && drawingFindRad)
	{/* Draw for subtract map */
	  int thisBin,thisBck;
	  if(thickStepFR > 0)
	    {/* OK to do */
	      XFillRectangle(dsply, XtWindow(drawingWidget),
			     image_gcGREY[(int)(110)],
			     125,0,
			     400,80);
	      XFlush(dsply);
	      thisBin = ((x - 83) / thickStepFR);
	      thisBin = MIN(MAX(thisBin,0),nbins);
	      sprintf(RDoutStr,  "Radial Density          [%f][%d]",
		      rdens[thisBin],thisBin);
	      sprintf(WRDoutStr, "Weighted RadialDensity  [%f][%d]",
		      wrdens[thisBin],thisBin);
	      sprintf(CCoutStr,  "Correlation Coefficient [%f][%d]",
		      ccoef[thisBin],thisBin);
	      sprintf(WCCoutStr, "Weighted CCoeffficient  [%f][%d]",
		      wccoef[thisBin],thisBin);
	      sprintf(S_WCCoutStr, "Smoothed Wghted CCoef   [%f][%d]",
		      s_wccoef[thisBin],thisBin);
	      if(rdensit)
		XDrawString(dsply, XtWindow(drawingWidget), 
			    image_gcWHITE,
			    125, 10,
			    RDoutStr, strlen(RDoutStr));
	      if(wrdensit)
		XDrawString(dsply, XtWindow(drawingWidget), 
			    image_gcRED,
			    125, 25,
			    WRDoutStr, strlen(WRDoutStr));
	      if(ccoefit)
		XDrawString(dsply, XtWindow(drawingWidget), 
			    image_gcGREEN,
			    125, 40,
			    CCoutStr, strlen(CCoutStr));
	      if(wccoefit)
		XDrawString(dsply, XtWindow(drawingWidget), 
			    image_gcBLUE,
			    125, 55,
			    WCCoutStr, strlen(WCCoutStr));
	      if(s_wccoefit)
		XDrawString(dsply, XtWindow(drawingWidget), 
			    image_gcBLACK,
			    125, 70,
			    S_WCCoutStr, strlen(S_WCCoutStr));
	    }/* OK to do */

	}/* Draw for subtract map */

      pointerLabelDraw(x, y);

      if(isBoxing && (boxZoom == USEBOX) &&  (boxMethod == USECIRCLE))
	{/* Draw the appropriate box */	
	  int arcRadius,origX,origY,outOfBounds;

	  if((pEvent->state & Button2Mask))
	    {/* Remove it */
	      removeBox(x, y, FALSE);
	    }/* Remove it */

	  /* Already done for us in pointerlabeldraw and done the 
	   * way we want... 
	   * calcSizeFactor(zoomFactor); */

	  /* Map this to original image coords */
	  origX = (x*recipSizeFactor)+(xoffset*recipSizeFactor)+virXoffset;
	  origY = (y*recipSizeFactor)+(yoffset*recipSizeFactor)+virYoffset;

	  outOfBounds= FALSE;
	  if( ((origX-boxRadius) <= recipSizeFactor) ||
	     ((origY-boxRadius) <= recipSizeFactor) ||
	     ((origX+boxRadius) > (file_width-1)) ||
	     ((origY+boxRadius) > (file_height-1)))
	    {/* Can't add to list - it's out of the image */
	      outOfBounds= TRUE;
	    }/* Can't add to list - it's out of the image */

	  arcRadius = (int)(sizeFactor * featherRadius);
	  if(lastBoxX>0 && lastBoxY>0)
	    {/* Redraw the image */
	      XPutImage(dsply, image_window1, image_gc, image_image1, 
			lastBoxX, lastBoxY,
			lastBoxX, lastBoxY,  
			(2*(arcRadius+3)),  
			(2*(arcRadius+3)));
	    }/* Redraw the image */

	  if(featherRadiusDisp && (!outOfBounds))
	    {/* Draw feather */
	      /* Size of the circle */
	      arcRadius = (int)(sizeFactor * featherRadius);
	      rwaXDrawArc(dsply, XtWindow(drawingWidget), 
			  image_gcPLACEHOLDER,
			  (x-arcRadius), 
			  (y-arcRadius),
			  (arcRadius*2),(arcRadius*2),
			  0, (360*64));
	      /* top */
	      XDrawLine(dsply, XtWindow(drawingWidget), 
			image_gcYELLOW,
			    (x-arcRadius), 
			    (y-arcRadius),
			    (x+arcRadius), 
			    (y-arcRadius)
			);
	      /* bottom */
	      XDrawLine(dsply, XtWindow(drawingWidget), 
			image_gcYELLOW,
			    (x-arcRadius), 
			    (y+arcRadius),
			    (x+arcRadius), 
			    (y+arcRadius)
			);
	      /* left */
	      XDrawLine(dsply, XtWindow(drawingWidget), 
			image_gcYELLOW,
			    (x-arcRadius), 
			    (y-arcRadius),
			    (x-arcRadius), 
			    (y+arcRadius)
			);
	      /* right */
	      XDrawLine(dsply, XtWindow(drawingWidget), 
			image_gcYELLOW,
			    (x+arcRadius), 
			    (y-arcRadius),
			    (x+arcRadius), 
			    (y+arcRadius)
			);
	    }/* Draw feather */
	  if(emptyShellDisp)
	    {/* Draw  */
	      /* Size of the circle */
	      arcRadius = (int)(sizeFactor * (emmaRadius-emptyShellWidth));
	      if(!outOfBounds)
		rwaXDrawArc(dsply, XtWindow(drawingWidget), image_gcWHITE,
			    (x-arcRadius), 
			    (y-arcRadius),
			    (arcRadius*2),(arcRadius*2),
			    0, (360*64));
	    }/* Draw  */
	  if(emmaRadiusDisp)
	    {/* Draw  */
	      /* Size of the circle */
	      arcRadius = (int)(sizeFactor * emmaRadius);
	      if(!outOfBounds)
		rwaXDrawArc(dsply, XtWindow(drawingWidget), image_gcYELLOW,
			    (x-arcRadius), 
			    (y-arcRadius),
			    (arcRadius*2),(arcRadius*2),
			    0, (360*64));
	    }/* Draw  */
	  if(boxRadiusDisp)
	    {/* Draw  */
	      arcRadius = (int)(sizeFactor * boxRadius);
	      if(!outOfBounds)
		rwaXDrawArc(dsply, XtWindow(drawingWidget), 
			    image_gcPLACEHOLDER,
			    (x-arcRadius), 
			    (y-arcRadius),
			    (arcRadius*2),(arcRadius*2),
			    0, (360*64));
	    }/* Draw  */
	  /* Do this in feather radius */
	  arcRadius = (int)(sizeFactor * featherRadius);
	  lastBoxX = x - (arcRadius) - 1;
	  lastBoxY = y - (arcRadius) - 1;
	  if(buffer_width < display_width)
	    lastBoxX = MIN((buffer_width),MAX(lastBoxX,0));
	  else lastBoxX = MIN((display_width),MAX(lastBoxX,0));
	  if(buffer_height < display_height)
	    lastBoxY = MIN((buffer_height),MAX(lastBoxY,0));
	  else lastBoxY = MIN((display_height),MAX(lastBoxY,0));
	}/* Draw the appropriate box */	
      else if(isBoxing && 
	      (boxZoom == USEBOX) && (rectBox.drawable) &&
	      (boxMethod==USERECT || boxMethod==USESQUARE))
	{/* Draw the appropriate RECT */	
	  int width,height,drawX,drawY;

	  /* Now check for the SQUARE subset of rectangle */
	  if(boxMethod==USESQUARE)
	    {/* square it up */
	      int xDiff,yDiff;
	      if(pEvent->state & Button1Mask)
		{/* x,y */
		  drawX = (((rectBox.lowx-virXoffset)*sizeFactor)-
			   (xoffset));
		  drawY = (((rectBox.lowy-virYoffset)*sizeFactor)-
			   (yoffset));
		  xDiff = (drawX-x);
		  yDiff = (drawY-y);
		  if( yDiff > xDiff )
		    y = drawY - xDiff;
		  else if( xDiff > yDiff )
		    x = drawX - yDiff;
		}/* x,y */
	      else if(pEvent->state & Button2Mask)
		{/* x,y */
		  drawX = (((rectBox.x-virXoffset)*sizeFactor)-
			   (xoffset));
		  drawY = (((rectBox.y-virYoffset)*sizeFactor)-
			   (yoffset));
		  xDiff = (x-drawX);
		  yDiff = (y-drawY);
		  if( yDiff > xDiff )
		    y = drawY + xDiff;
		  else if( xDiff > yDiff )
		    x = drawX + yDiff;
		}/* x,y */
	      if(buffer_width < display_width)
		x= MIN((buffer_width),MAX(x,0));
	      else x= MIN((display_width),MAX(x,0));
	      if(buffer_width < display_width)
		y= MIN((buffer_width),MAX(y,0));
	      else y= MIN((display_width),MAX(y,0));
	    }/* square it up */

	  /* Redraw from last time */
	  width = rectBox.lowx - rectBox.x;
	  height = rectBox.lowy - rectBox.y;
	  width = ((width*sizeFactor));
	  height = ((height*sizeFactor));

	  if(lastBoxX>0 && lastBoxY>0)
	    {/* Redraw the image */
	      XPutImage(dsply, image_window1, image_gc, image_image1, 
			lastBoxX-1, lastBoxY-1,
			lastBoxX-1, lastBoxY-1,  
			width+3,  
			height+3);
	      XFlush(dsply);
	    }/* Redraw the image */

	  if(pEvent->state & Button3Mask)
	    pushRectBox(Button3,x,y);
	  else if(pEvent->state & Button2Mask)
	    pushRectBox(Button2,x,y);
	  else if(pEvent->state & Button1Mask)
	    pushRectBox(Button1,x,y);

	  /* We use viroffset so that we take into account that this 
	   * guy might be on another virtual page 
	   */
	  width = rectBox.lowx - rectBox.x;
	  height = rectBox.lowy - rectBox.y;
	  width = ((width*sizeFactor));
	  height = ((height*sizeFactor));
	  drawX = (((rectBox.x-virXoffset)*sizeFactor)-(xoffset));
	  drawY = (((rectBox.y-virYoffset)*sizeFactor)-(yoffset));
	  /* Draw rectangle */
	  XDrawRectangle(dsply, XtWindow(drawingWidget), image_gcGREEN,
			 drawX, 
			 drawY,
			 width,
			 height);
	  rectBox.width  = rectBox.lowx - rectBox.x;
	  rectBox.height = rectBox.lowy - rectBox.y;
	  /* Draw rect dimensions */
	  sprintf(infoStr,"[%d]x[%d]",
		  (int)(rectBox.width),
		  (int)(rectBox.height)
		  );
	  XDrawString(dsply, XtWindow(drawingWidget), 
		      image_gcRED,
		      drawX+15, drawY+15, infoStr, strlen(infoStr));
	  XFlush(dsply);
	  

	  if(pEvent->state & Button1Mask)
	    {/* button 1 */
	      lastBoxX = x;
	      lastBoxY = y;
	    }/* button 1 */
	  else 
	    {/* button 2 or 3 */
	      lastBoxX = x - width;
	      lastBoxY = y - height;
	    }/* button 2 or 3 */
	  if(buffer_width < display_width)
	    lastBoxX = MIN((display_width),MAX(lastBoxX,0));
	  else lastBoxX = MIN((display_width),MAX(lastBoxX,0));
	  if(buffer_height < display_height)
	    lastBoxY = MIN((buffer_height),MAX(lastBoxY,0));
	  else lastBoxY = MIN((display_height),MAX(lastBoxY,0));
	}/* Draw the appropriate RECT */	
      
      break;
	  
    default:
      break;
    }/* Find out which event we have */
  
  XFlush(dsply);              /* Make changes current */


}/* plotMotion */






/*-----------------------------------*/
/*************************************
 * MAIN screen scrollbars            *
 *************************************/
/*-----------------------------------*/
/* ************************************************************
 * plotVertChange() - When the vertical scrollbar on the main
 * screen changes value call here.  This is only used for
 * traversing scanned images when boxing.
 */
void plotVertChange(Widget widget, int *tag, 
		    XmScrollBarCallbackStruct *callback_data)
{/* PLOTVERTCHANGE */
  int thisVal,thisWidth,thisHeight;

  if(!noCallbackScroll)
    {/* Don't do this if we're farbling with it */
      /* yoffset is just the offset into the image buffer 2k x 2k
       * we're seeing on the screen.  virYoffset is the offset
       * into the file image that we want to read from the disk.
       */
      fullScreen= FALSE;    /* We are scrolling not zooming */

      if(!fftMode)
	{/* Read this new one from disk */
	  virYoffset = callback_data->value;
	  yoffset= 0;

	  if(buffer_width < display_width)
	    thisWidth = (int)(buffer_width * 0.5);
	  else thisWidth = (int)(display_width * 0.5);
	  if(buffer_height < display_height)
	    thisHeight = (int)(buffer_height * 0.5);
	  else thisHeight = (int)(display_height * 0.5);

	  virtual_read_activate(((int)(thisWidth)),
				((int)(thisHeight)));

	}/* Read this new one from disk */
      else  yoffset = callback_data->value;
      
      DrawWindow();
    }/* Don't do this if we're farbling with it */
  
}/* PLOTVERTCHANGE */



/* ************************************************************
 * plotHorizChange() - When the horizontal scrollbar on the main
 * screen changes value call here.
 * Also only for scanned image boxing traversing.
 */
void plotHorizChange(Widget widget, int *tag, 
		     XmScrollBarCallbackStruct *callback_data)
{/* PLOTHORIZCHANGE */
  int thisVal,thisWidth,thisHeight;

  if(!noCallbackScroll)
    {/* Don't do this if we're farbling with it */
      /* xoffset is just the offset into the image buffer 2k x 2k
       * we're seeing on the screen.  virXoffset is the offset
       * into the file image that we want to read from the disk.
       */
      fullScreen= FALSE;    /* We are scrolling not zooming */

      if(!fftMode)
	{/* Read this new one from disk */
	  virXoffset = callback_data->value;
	  xoffset = 0;

	  if(buffer_width < display_width)
	    thisWidth = (int)(buffer_width * 0.5);
	  else thisWidth = (int)(display_width * 0.5);
	  if(buffer_height < display_height)
	    thisHeight = (int)(buffer_height * 0.5);
	  else thisHeight = (int)(display_height * 0.5);

	  virtual_read_activate(((int)(thisWidth)),
				((int)(thisHeight)));

	}/* Read this new one from disk */
      else xoffset = callback_data->value;
      
      DrawWindow();
    }/* Don't do this if we're farbling with it */
  
}/* PLOTHORIZCHANGE */



/* ************************************************************
 * plotScrollCreate() - Remeber the toggle widget number for
 * setting the scrollbar values later.  See note above about
 * getting all widget creation done in s single rtn ???
 */
void plotScrollCreate(Widget widget, int *tag, 
		      XmAnyCallbackStruct *callback_data)
{/* PLOTSCROLLCREATE */

  if(*tag==VERT)
    {/* height */
      vertScroll= widget;
    }/* height */
  else if(*tag==HORI)
    {/* width */
      horiScroll= widget;
    }/* width */

}/* PLOTSCROLLCREATE */





/*-----------------------------------*/
/*************************************
 * FILE SELCTION WIDGET callbacks    *
 *************************************/
/*-----------------------------------*/
/* @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
 * cancel_fileSel() -  They want to abort the selection
 * of a file name.
*/
void cancel_fileSel(Widget widget, int *tag, 
		    XmFileSelectionBoxCallbackStruct *callback_data)
{/* CANCEL_FILESEL */

  XtUnmanageChild(fileSel_widget);

}/* CANCEL_FILESEL */




/***********************************************************
 * whichFileType() - Determine which file type we've got so
 * we can read it properly.  Only used for PIF files as
 * of this writing.
*/
whichFileType(char *fName, int *type)
{/* WHICHFILETYPE */
  int i,j,inInt;
  short junk1,junk2;
  int file_id1,file_id2,mode,seekPos,endianNess;
  FILE *fIn;
  fString_t inFile,rWrite;
  forPifHead_t thisHead;

  if(*type==BOXCOORDS || *type==SF || *type==STATE)
    goto bugOut;

  fIn= NULL;
  fIn = fopen(fName,"r");
  if(fIn == NULL)
    {/*  Ooops can't read it!!! */
      sprintf(outStr,
	      "Can't read [%s] ",fName);
      askThem(outStr,TRUE,XmDIALOG_INFORMATION);
      goto bugOut;
    }/*  Ooops can't read it!!! */

  if(fIn)
    {/* Read the file */

      fread(&inInt, sizeof(int),1,fIn);
      fclose(fIn);

	  i= 4;
	  thisHead.luNum = i;
	  rWrite.addr = malloc(sizeof(char)*4);
	  rWrite.length = 4;
	  inFile.addr = malloc(sizeof(char)*264);
	  inFile.length = 264;
	  sprintf(rWrite.addr,"r");
	  sprintf(inFile.addr,fName);

	  pif_open_(inFile.addr, &i, rWrite.addr);
	  
	  if(!pif_read_gh_(&thisHead))
	    {/* Don't know this one */
	      askThem("I don't understand this file type.",
		      FALSE,XmDIALOG_INFORMATION);	
	      *type = -99;
	      goto bugOut;
	    }/* Don't know this one */
	  
	  mode               = thisHead.mode;
	  
	  if(mode==20)
	    *type= PIFMAP;
	  else if(mode==40)
	    *type= PIFMAP;
	  else if(mode==21)
	    *type= PIFMAP;
	  else if(mode==22)
	    *type= PIFMAP;  /* Special PFTS floatInt*4 case */
	  else if (mode==0)
	    *type= EMPORTABLE;
	  else if (mode==1)
	    *type= EMPORTABLEI2;  /* short Zeiss data */
	  else if (mode==2)
	    *type= EMPORTABLEI2;  /* float Zeiss data */
	  else if (mode==1)
	    *type= EMPORTABLE;
	  else if (mode==6)
	    *type= EMPORTABLE;
	  else if (mode==7)
	    *type= EMPORTABLE;
	  else if (mode==46)
	    *type= EMPORTABLEI4;
	  else
	    askThem("I don't understand this mode",
		    FALSE,XmDIALOG_INFORMATION);
	  
	  free(rWrite.addr);
	  free(inFile.addr);

    }/* Read the file */
  
  pif_close(&(thisHead.luNum));

 bugOut:;

}/* WHICHFILETYPE */




/**************************************************************
 * askThemForMode() - Ask them how to interpret this MRC file.
 * Future goal - try and get this info from the MRC header.
*/
void askThemForMode(int *mode)
{/* ASKTHEMFORMODE */
  int numChoices,numWanted;
  XmString simpleString;
  
  sprintf(outStr,"2D scanned image");
  createSimpleString(&simpleString, outStr);
  XmListAddItem(infoListW, simpleString, 1);
  XFlush(dsply);
  XmStringFree(simpleString); 
  sprintf(outStr,"Set of 2D boxes");
  createSimpleString(&simpleString, outStr);
  XmListAddItem(infoListW, simpleString, 2);
  XFlush(dsply);
  XmStringFree(simpleString); 
  sprintf(outStr,"3D Map");
  createSimpleString(&simpleString, outStr);
  XmListAddItem(infoListW, simpleString, 3);
  XFlush(dsply);
  XmStringFree(simpleString); 
  
  sprintf(outStr,"How do you want me to interpret this file?: ");
  numChoices = 3;
  numWanted= askThemAnswer(outStr,numChoices);
  
  if(numWanted==0)
    *mode = 2;
  else if(numWanted==1)
    *mode = 46;
  else if(numWanted==2)
    *mode = 20;
  
}/* ASKTHEMFORMODE */





/* @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
 * gotFilename() - Get the filename they specified in file
 * selection widget.  Branch according to what operation they
 * want performed on this filename.
 * This rtn is called whenever a filename is selected from
 * the filename browsing widget.
*/
void gotFilename (Widget w, char *client_data, 
		  XmFileSelectionBoxCallbackStruct *callback_data)
{/* GOTFILENAME */
  int i,j,k,l,m,yOff; 
  char oldFilename[1024];
  char thisFileName[1024];
  char *s1,*s2,*s3;
  XmString simpleString;
  float xcen, ycen, zcen, max_r;
  int interp, nterp;
  int *irnum;
  float *sum_rint, *sumsq_rint, *stdev, *mrcBuf;
  short outWidth,outHeight,junkShort;


  christmasTime = FALSE;
      
  /* Pop down the file selection widget */
  XtUnmanageChild(fileSel_widget);
  XFlush(dsply);
      
  /* This may take awhile so give them the watch cursor */
  attrs.cursor= watch_cursor;
  XChangeWindowAttributes(dsply, XtWindow(drawingWidget), 
			  CWCursor, &attrs);
  XFlush(dsply);
  
  if(readWrite == READPARAM)
    {/* Just get the filename and exit */
      char *thisStr;

      if(!XmStringGetLtoR(callback_data->value, 
			  XmSTRING_DEFAULT_CHARSET, 
			  &thisStr))
	{printf("Error extracting filename.\n");}
      
      sprintf(gottenParamFname,"%s",thisStr);
      XtFree(thisStr);

      /* Just read it in */
      paramFileRead_work();

      goto destructor;

    }/* Just get the filename and exit */

  if(subtractingMap)
    {/* Just get the filename and exit */
      char *thisStr;

      if(!XmStringGetLtoR(callback_data->value, 
			  XmSTRING_DEFAULT_CHARSET, 
			  &thisStr))
	{printf("Error extracting filename.\n");}
      
      sprintf(subtractFilename,"%s",thisStr);
      XtFree(thisStr);
      sprintf(outStr,"%s",subtractFilename);
      XmTextSetString(subtractFile2TextW,outStr);
      
      XtSetArg(arglst[0], XmNcursorPosition, strlen(outStr) );
      XtSetValues(subtractFile2TextW, arglst, 1); 
      XFlush(dsply);

      /* Just read it in */
      subtractMap_read();

      goto destructor;

    }/* Just get the filename and exit */

  if(readEmcoravg)
    {/* Just get the filename and exit */
      char *thisStr;

      if(!XmStringGetLtoR(callback_data->value, 
			  XmSTRING_DEFAULT_CHARSET, 
			  &thisStr))
	{printf("Error extracting filename.\n");}
      

      /* Now read it and stuff the display bufffer */
      sprintf(emcoravgFilename,thisStr);
      read_emcoravg(thisStr);
      XtFree(thisStr);

      goto destructor;

    }/* Just get the filename and exit */

  if(refReadFile && readType==REFREADFILE)
    {/* Just get the filename and exit */
      char *thisStr;

      if(!XmStringGetLtoR(callback_data->value, 
			  XmSTRING_DEFAULT_CHARSET, 
			  &thisStr))
	{printf("Error extracting filename.\n");}
      

      /* Now read it and stuff the display bufffer */
      sprintf(refReadFileName,thisStr);

      sprintf(outStr,"%s",refReadFileName);
      createSimpleString(&simpleString,outStr);
      XtSetArg(arglst[0], XmNlabelString, simpleString );
      XtSetArg(arglst[1], XmNrecomputeSize, TRUE );
      XtSetValues(refReadFileLabelW, arglst, 2); 
      XFlush(dsply);
      XmStringFree(simpleString); 

      goto destructor;

    }/* Just get the filename and exit */

  if(readImgconv)
    {/* Just get the filename and exit */
      char *thisStr;

      if(!XmStringGetLtoR(callback_data->value, 
			  XmSTRING_DEFAULT_CHARSET, 
			  &thisStr))
	{printf("Error extracting filename.\n");}
      

      /* Now read it and stuff the display bufffer */
      sprintf(imgconvFilename,thisStr);
      read_imgconv(thisStr);
      XtFree(thisStr);

      goto destructor;

    }/* Just get the filename and exit */

  /* Get the filename */
  sprintf(oldFilename,"%s",fileName.addr);

  if(preloadSwitch)
    {/* Preload it */
      sprintf(fileName.addr,"%s",preLoadStr); 
      preloadSwitch = FALSE;  /* Done with it */
    }/* Preload it */
  else if(isDatabase)
    {/* database */
      int done;
      FILE *fTry;
      done = FALSE;
      while(!done)
	{/* Get a valid filename */
	  sprintf(databaseBcrdStr,"%s_box_%03d.bcrd",
		  databaseRootStr,
		  databaseIter);
	  sprintf(databaseStr,"%s_box_%03d.pif",
		  databaseRootStr,
		  databaseIter);
	  sprintf(fileName.addr,"%s",databaseStr);
	  fTry = fopen(databaseStr,"r");
	  if(fTry)
	    {/* Already there */
	      sprintf(outStr,"[%s] already exists - Overwrite(Yes) or Iterate filenumber(Cancel)?",databaseStr);
	      if(askThem(outStr,FALSE,XmDIALOG_QUESTION))
		{/* OK to go */
		  done = TRUE;
		}/* OK to go */
	      else
		databaseIter++;
	      fclose(fTry);
	    }/* Already there */
	  else
	    done = TRUE;
	}/* Get a valid filename */
    }/* database */
  else
    {/* normal */
      if(!XmStringGetLtoR(callback_data->value, 
			  XmSTRING_DEFAULT_CHARSET, 
			  &(fileName.addr)))
	{printf("Error extracting filename.\n");}
    }/* normal */

  raiseIt(statusScreenWidget);
  XFlush(dsply);
  XSync(dsply,0);

  /* Branch according to who callled us */
  if(readWrite==READING)
    {/* REad from Input file */

      sprintf(outStr,"Reading [%s]",fileName.addr);
      if(checkStatus(outStr))
	goto destructor;

      if(readType==MRC)
	{/* read mrc file */
	  char fname[256], two_d[2];
	  int m,n,o,p,thisNumSecs;

	  sprintf(two_d,"Y");
	  sprintf(fname,fileName.addr);

	  read_mrc_header_ (fname, &pifHeader, two_d);

	  recipFloatScaleFactor = pifHeader.recipFloatScaleFactor;
	  purdueHeader.img_bkg  = pifHeader.img_bkg;
	  purdueHeader.img_min  = (int)pifHeader.datai4_min;
	  purdueHeader.img_max  = (int)pifHeader.datai4_max;
	  purdueHeader.map_min  = pifHeader.datar4_min;
	  purdueHeader.map_max  = pifHeader.datar4_max;
	  purdueHeader.img_avg  = pifHeader.datar4_avg;
	  purdueHeader.img_std  = pifHeader.datar4_std;
	  purdueHeader.numCols    = pifHeader.numCols;
	  purdueHeader.numRows    = pifHeader.numRows;
	  purdueHeader.numSecs    = pifHeader.numSecs;
	  purdueHeader.nimg       = pifHeader.nimg;
	  numCols    = pifHeader.numCols;
	  numRows    = pifHeader.numRows;
	  origNumSecs = numSecs    = pifHeader.numSecs;
	  pifMode    = pifHeader.mode; 
	  htype      = pifHeader.htype; 
	  sprintf(title,"%s",pifHeader.descriptionStr);
	  aoverb     = pifHeader.aoverb; 
	  map_abang  = pifHeader.map_abang;
	  a_cell     = pifHeader.a_cell;
	  b_cell     = pifHeader.b_cell;
	  c_cell     = pifHeader.c_cell; 
	  dela       = pifHeader.dela; 
	  delb       = pifHeader.delb; 
	  delc       = pifHeader.delc;
	  binFactor  = pifHeader.binFactor;
	  dthe       = pifHeader.dthe; 
	  dphi_90    = pifHeader.dphi_90; 
	  symmetry   = pifHeader.symmetry; 
	  
	  mrc_alpha = (int) pifHeader.alp;
	  mrc_beta = (int) pifHeader.bet;
	  mrc_gamma = (int) pifHeader.gam;
	  fileBinSecs= numSecs;
	  fileBinRows= numRows;
	  fileBinCols= numCols;
	  fileBinFactor = binFactor;
	  
	  askThemForMode(&pifHeader.mode);
	  pifMode    = pifHeader.mode; 
	  if(pifMode==20)
	    numSecs = pifHeader.nimg;

	  if(pifHeader.mode==46)
	    thisNumSecs = pifHeader.nimg;
	  else if(pifHeader.mode==2)
	    thisNumSecs = 1;
	  else
	    {/* mode == 20 */
	      thisNumSecs = numSecs;
	      pifHeader.nimg = numSecs;
	      pifHeader.numSecs = numSecs;
	      origNumSecs = numSecs;
	      nsecs = numSecs;
	    }/* mode == 20 */
	  /* +2 in secs for even numbered sections in mrc */
	  CK(mrcBuf = (float *)malloc(sizeof(float)
				      *numCols*numRows*(thisNumSecs+2)));
	  read_mrc_file_ (fname, mrcBuf, two_d, 
			  &numCols,&numRows,&thisNumSecs);


	  if(pifHeader.mode==46)
	    {/* DO for packed boxes */
	      float *thisF,*thisP;
	      readType = MRCPACKED;
	      fileType = PACKEDPORTABLE;
	      minPackPixel = maxPackPixel = mrcBuf[123];
	      for(m=0;m<pifHeader.nimg;m++)
		{/* DO all */
		  CK(thisP= packImg[m]= (float *)malloc(
							sizeof(float) *
							numCols*
							numRows));
		  thisF = (float *)(mrcBuf+(m*numCols*numRows));
		  for(n=0;n<(numRows);n++)
		    {
		      for(o=0;o<(numCols);o++)
			{
			  thisP[((n*numCols)+o)] = thisF[(((numRows-n-1)*
							   numCols)+o)];
			  minPackPixel = MIN(minPackPixel,
					     thisP[((n*numCols)+o)]);
			  maxPackPixel = MAX(maxPackPixel,
					     thisP[((n*numCols)+o)]);
			}
		    }
		}/* DO all */
	      /* Now clip this guy to our colormap */ 
	      if(minPackPixel < 0.0)
		scalePackFactor = (((float)tryColors - 7.0) / 
				   (maxPackPixel-minPackPixel));
	      else scalePackFactor = (((float)tryColors - 7.0) / 
				      maxPackPixel);
	      packRadius = purdueHeader.numCols/2;
	      stuffPackedImage(purdueHeader.numCols,
			       purdueHeader.numRows,
			       packImg,purdueHeader.nimg,
			       1);  /* Start at zero'th(1) image */
      
	      free(mrcBuf);
	    }/* DO for packed boxes */
	  else if(pifHeader.mode==20)
	    {/* DO for maps */
	      int l,m,n,o;
	      readType = MRCMAP;
	      fileType = MAP;
	      if((float_image_buffer1) &&
		 (float_image_buffer1 != lowRes_buffer1))
		free(float_image_buffer1);
	      if(lowRes_buffer1)
		free(lowRes_buffer1);
	      float_image_buffer1 = mrcBuf;

	      minMapPixel = maxMapPixel = float_image_buffer1[0];
	      for(l=0;l<(numCols*numRows*numSecs);l++)
		{/* find */
		  maxMapPixel = MAX(maxMapPixel,float_image_buffer1[l]);
		  minMapPixel = MIN(minMapPixel,float_image_buffer1[l]);
		}/* find */

	      if(pifMode==20)
		{/* We've got a map here */
		  
		  imgNum = numCols/2; /* Let's default to the middle section */
		  sprintf(outStr,"%d",imgNum+1);
		  purdueHeader.nimg = numSecs;  /* Fool it to take our value */
		  XmTextSetString(imgNumTextWidget,outStr);
		  
		  calcSF_work(calcSFs);
		  
		  /* Now remember these forever */
		  lowRes_buffer1 = float_image_buffer1;
		  lowResCols = numCols;
		  lowResRows = numRows;
		  lowResSecs = numSecs;
		  
		  origWidthFromFile = numCols;
		  
		}/* We've got a map here */
	      
	      
	      xCen = (double)(numCols/2);
	      yCen = (double)(numRows/2);
	      zCen = (double)(origNumSecs/2);
	      
	      largestDim = numCols;
	      largestDim = MAX(largestDim,numRows);
	      largestDim = MAX(largestDim,origNumSecs);
	      
	      /* We've got to make this larger for asymmetric maps 
	       * and all maps that have information "outside" of the
	       * normal sphere of our normal maps,
	       * or else we just keep cliping the box into a sphere */
	      if(numCols==largestDim && numRows==largestDim
		 && (origNumSecs)==largestDim)
		{/* Symmetric */
		  aSymmetric = FALSE;
		  origGlobMaxRadius= globMaxRadius = 
		    (double)((double)origNumSecs/2)+5.0;
		  globMinRadius = (double)0.0;
		}/* Symmetric */
	      else
		{/* asymmetric */
		  aSymmetric = TRUE;
		  origGlobMaxRadius=  globMaxRadius = 
		    (double)((double)origNumSecs);
		  globMinRadius = (double)0.0;
		}/* asymmetric */
	      
	      /* Set scrolls on depthcue screen */
	      /* Give feedback */
	      sprintf(outStr,"%.0f",globMinRadius);
	      createSimpleString(&simpleString,outStr);
	      XtSetArg(arglst[0], XmNvalue, outStr );
	      XtSetArg(arglst[1], XmNrecomputeSize, FALSE );
	      XtSetValues(minRadiusValueWidget, arglst, 2); 
	      XFlush(dsply);
	      XmStringFree(simpleString); 
	      
	      sprintf(outStr,"%.0f",globMaxRadius);
	      createSimpleString(&simpleString,outStr);
	      XtSetArg(arglst[0], XmNvalue, outStr );
	      XtSetArg(arglst[1], XmNrecomputeSize, FALSE );
	      XtSetValues(maxRadiusValueWidget, arglst, 2); 
	      XFlush(dsply);
	      XmStringFree(simpleString); 
	      
	      
	      /* Radius is origNumSecs/2 */
	      XtVaSetValues(maxRadiusScrollWidget,XmNmaximum, 
			    (int)((globMaxRadius*2)+5),NULL);
	      XtVaSetValues(maxRadiusScrollWidget,XmNsliderSize, 
			    (int)(5),NULL);
	      XtVaSetValues(maxRadiusScrollWidget,XmNvalue, 
			    (int)(globMaxRadius),NULL);
	      
	      XtVaSetValues(minRadiusScrollWidget,XmNmaximum, 
			    (int)((globMaxRadius*2)+5),NULL);
	      XtVaSetValues(minRadiusScrollWidget,XmNsliderSize, 
			    (int)(5),NULL);
	      XtVaSetValues(minRadiusScrollWidget,XmNvalue, (int)
			    (globMinRadius),NULL);
	      
	      printf("MAX[%f]  MIN[%f]\n",maxMapPixel,minMapPixel);
	      secMaxPixel = maxMapPixel;
	      secMinPixel = minMapPixel;
	      
	      minSurf = (int)MAX(minMapPixel,0.0);
	      maxSurf = MAX(maxMapPixel,(0.0-minMapPixel));
	      /* We're doing the surfaceDensity in tenths now */
	      if (pifMode == 46)
		{
		  XtVaSetValues(surfDensScrollWidget,XmNmaximum, 
				(int)((maxSurf*10.0)+50.0),NULL);
		  XtVaSetValues(surfDensScrollWidget,XmNminimum, 
				(int)(minSurf*10.0),NULL);
		  XtVaSetValues(surfDensScrollWidget,XmNsliderSize, 
				(int)(500),NULL);
		  XtVaSetValues(surfDensScrollWidget,XmNpageIncrement, 
				(int)500,NULL);
		  XtVaSetValues(surfDensScrollWidget,XmNincrement, 
				(int)100,	NULL);
		  XtVaSetValues(surfDensScrollWidget,XmNvalue, 
				(int)(purdueHeader.img_bkg),NULL);
		  
		}
	      else if(maxSurf > 5000.0)
		{
		  XtVaSetValues(surfDensScrollWidget,XmNmaximum, 
				(int)((maxSurf*10.0)+50.0),NULL);
		  XtVaSetValues(surfDensScrollWidget,XmNminimum, 
				(int)(minSurf*10.0),NULL);
		  XtVaSetValues(surfDensScrollWidget,XmNsliderSize, 
				(int)(maxSurf*0.01),NULL);
		  XtVaSetValues(surfDensScrollWidget,XmNpageIncrement, 
				(int)(maxSurf*0.01),NULL);
		  XtVaSetValues(surfDensScrollWidget,XmNincrement, 
				(int)(maxSurf*0.002),NULL);
		  XtVaSetValues(surfDensScrollWidget,XmNvalue, 
				(int)(10),NULL);
		  
		}
	      else if(maxSurf > 500.0)
		{
		  XtVaSetValues(surfDensScrollWidget,XmNmaximum, 
				(int)((maxSurf*10.0)+50.0),NULL);
		  XtVaSetValues(surfDensScrollWidget,XmNminimum, 
				(int)(minSurf*10.0),NULL);
		  XtVaSetValues(surfDensScrollWidget,XmNsliderSize, 
				(int)(500),NULL);
		  XtVaSetValues(surfDensScrollWidget,XmNpageIncrement, 
				(int)500,NULL);
		  XtVaSetValues(surfDensScrollWidget,XmNincrement, 
				(int)100,	NULL);
		  XtVaSetValues(surfDensScrollWidget,XmNvalue, 
				(int)(10),NULL);
		  
		}
	      else
		{
		  XtVaSetValues(surfDensScrollWidget,XmNmaximum, 
				(int)((maxSurf*10.0)+50.0),NULL);
		  XtVaSetValues(surfDensScrollWidget,XmNminimum, 
				(int)(minSurf*10.0),NULL);
		  XtVaSetValues(surfDensScrollWidget,XmNsliderSize, 
				(int)(50),NULL);
		  XtVaSetValues(surfDensScrollWidget,XmNpageIncrement, 
				(int)50,NULL);
		  XtVaSetValues(surfDensScrollWidget,XmNincrement, 
				(int)10,	NULL);
		  XtVaSetValues(surfDensScrollWidget,XmNvalue, 
				(int)(10),NULL);
		  
		}
	      surfaceDensity= 1.0;  /* 1 tenth of 10 */
	      
	      /* Give feedback */
	      sprintf(outStr,"%d",-1);
	      createSimpleString(&simpleString,outStr);
	      XtSetArg(arglst[0], XmNvalue, outStr );
	      XtSetArg(arglst[1], XmNrecomputeSize, FALSE );
	      XtSetValues(surfDensValueWidget, arglst, 2); 
	      XFlush(dsply);
	      XmStringFree(simpleString); 
	      
	      createSimpleString(&simpleString,"SecNum");
	      XtSetArg(arglst[0], XmNlabelString, simpleString);
	      XtSetArg(arglst[1], XmNrecomputeSize, FALSE );
	      XtSetValues(imgNumButtonW, arglst, 2); 
	      XFlush(dsply);
	      XmStringFree(simpleString); 
	      
	      
	      /* To make this work for now let's stuff the nimg field with the
	       * number of sections.  Really we should change any and all 
	       * references
	       * to nimg to be origNumSecs. We really only work on one 
	       * image at a time
	       * so this may never have any affect.
	       */
	      purdueHeader.nimg = origNumSecs;
	      
	      /* Write stats */
	      calcMapStats(float_image_buffer1,
			   &purdueHeader.img_bkg, 
			   &purdueHeader.map_min,
			   &purdueHeader.map_max, 
			   &purdueHeader.img_avg, 
			   &purdueHeader.img_std,
			   &numCols, &numRows, &origNumSecs);
	      
	      /* Write stats */
	      sprintf(outStr,"\nStatistics for file [%s]\n\
\n\
Title [%s]\n\
\n\
File size   [%d]x[%d]x[%d]\n\
Background  [%f]\n\
MIN,MAX     [%f],[%f]\n\
AVG,STD     [%f],[%f]\n\
\n",
		      fileName.addr,
		      title,
		      numCols,	  numRows, origNumSecs,
		      purdueHeader.img_bkg, 
		      minMapPixel, maxMapPixel,
		      purdueHeader.img_avg,	  purdueHeader.img_std
		      );
	      XmTextSetString(statTextWidget,outStr);
	      XFlush(dsply);
	      
	    }/* DO for maps */
	  else if(pifHeader.mode==2)
	    {/* DO for scanned img */
	      file_width = numCols;
	      file_height = numRows;
	      inCore_buffer= (float *)mrcBuf;
	      if(inCore_buffer)
		imgInCore= TRUE;
	      else 
		{/* Trouble */
		  imgInCore= FALSE;
		  printf("Trouble allocating inCore_buffer!!!\n");
		  exit(0);
		}/* Trouble */
	      if(int_image_buffer1 != NULL)
		free(int_image_buffer1);
	      CK(int_image_buffer1 = (unsigned short *)malloc(
					   ((defDisplayBufferSize*
					     defDisplayBufferSize)*
					    sizeof(unsigned short))+1));
	      readType = MRCSCAN;
	      fileType = PORTABLE;
	      virtual_read_activate(-99,-99);
	    }/* DO for scanned img */

	}/* read mrc file */
      else  if(readType != PARAMFILE && readType != PARAMMAP && 
	       readType != BATCHJOB  && readType != PARAMSCAN)
	whichFileType(fileName.addr, &readType);
      
      if((readType==EMPORTABLE) || (readType==EMPORTABLEI2)
	 || readType==EMPORTABLEI4 || readType==MRCPACKED)
	{/* Read an image file*/
	  
	  /* Remember this for the next time */
	  if(numBoxes > 0)
	    lastNumBox = (lastNumBox + numBoxes);

	  /* Init for this new image */
	  if(readType != MRCPACKED)
	    groundZero();

	  /* Tell them what we're doing */
	  sprintf(outStr,"%d",lastNumBox+1);
	  XmTextSetString(boxInitWidget, outStr);
	  XFlush(dsply);

	  /* Remember what our filename was - especially if we had to 
	   * change it to portable or etc...
	   */
	  sprintf(image_fileName.addr,"%s",fileName.addr);


	  /* Let's see if we can't just use our own */
	  if(readType != MRCPACKED)
	    {/* We already read it */
	      mainRobem_sub_UNIX(&int_image_buffer1, 
				 &fileName, &i, &j);
	      
	      /* Ooops no image get out of here */
	      if(i==0 || j==0)
		goto destructor;
	    }/* We already read it */

	  /* Give feedback on main window title bar */
	  if(!masterProcess)
	    sprintf(outStr,"Robem: [%s] - %d images total. [%s]",
		    fileName.addr,
		    purdueHeader.nimg,"CHILD");
	  else
	    sprintf(outStr,"Robem: [%s] - %d images total.",fileName.addr,
		    purdueHeader.nimg);
	  createSimpleString(&simpleString,outStr);
	  XtSetArg(arglst[0], XmNdialogTitle, simpleString );
	  XtSetArg(arglst[1], XmNrecomputeSize, FALSE );
	  XtSetValues(main_window_widget, arglst, 2); 
	  XFlush(dsply);
	  XmStringFree(simpleString); 
	  XFlush(dsply);
/* malloc_verify(); */

	  /* Remember how big we are on disk */
	  file_width = i;
	  file_height = j;

	  pickAutoZoom(i,j);

	  /* Now clip this and only get 1k x 1k maximum at a time */
	  i= defDisplayBufferSize;
	  j= defDisplayBufferSize;

	  /* Point to where the display data is */
	  image_image1->data = (char *)image_buffer1;
	  
	}/* Read an image file*/

      else if(readType==MAP || readType==FLOATMAP || readType==PIFMAP
	      || readType==MRCMAP)
	{/* Read a MAP file*/
	  short thisPixel;

	  fileType= MAP;

	  /* Init for this new image */
	  if(readType != MRCMAP)
	    groundZero();

	  createSimpleString(&simpleString, "SecNum");
	  XtSetArg(arglst[0], XmNlabelString, simpleString);
	  XtSetArg(arglst[1], XmNrecomputeSize, FALSE );
	  XtSetValues(imgNumButtonW, arglst, 2); 
	  XFlush(dsply);
	  XmStringFree(simpleString); 
	  
	  sprintf(outStr,"1");
	  XmTextSetString(imgNumTextWidget,outStr);

	  XtSetArg(arglst[0], XmNsensitive, FALSE);
	  XtSetValues(removeBlemishButtonW, arglst, 1); 
	  XtSetValues(editBoxButtonW, arglst, 1); 
	  XtSetValues(pcxButtonW, arglst, 1); 
	  XtSetValues(correlateButtonW, arglst, 1);  
	  XtSetValues(autoBox_widget, arglst, 1);  
	  XtSetValues(correlateButton2W, arglst, 1);  
/*	  XtSetValues(autoBox2_widget, arglst, 1);   */
	  XtSetArg(arglst[0], XmNsensitive, TRUE);
	  XtSetValues(paintMapButtonW, arglst, 1); 
	  XtSetValues(mapMovieButtonW, arglst, 1); 
	  XtSetValues(mapShadeButtonW, arglst, 1); 
/*	  XtSetValues(fixButtonW, arglst, 1);  */
	  XtSetValues(truncaddButtonW, arglst, 1); 
	  XtSetValues(rescaleButtonW, arglst, 1); 
	  XtSetValues(modelWidget, arglst, 1); 
	  XtSetValues(subtractMapButtonW, arglst, 1); 
	  XtSetValues(calcSFButtonW, arglst, 1); 
	  /* XtSetValues(RcalcSFButtonW, arglst, 1);  */
	  XtSetValues(hardRotateButtonW, arglst, 1); 
	  XtSetValues(invertButtonW, arglst, 1); 
	  XtSetValues(rotateButtonW, arglst, 1); 
	  XtSetValues(rawButtonW, arglst, 1);   

	  /* Read the MAP file */
	  if(readType==PIFMAP || readType==EMPORTABLEI4)
	    {/* PIF MAP file */
	      read_pif(&i, &j, &nsecs, &fileName);
	      if(nsecs==-99)
		{/* Error */
		  sprintf(outStr,"Error reading [%s] as MAP file.",
			  fileName.addr);
		  askThem(outStr,FALSE,XmDIALOG_INFORMATION);
		  goto destructor;
		}/* Error */
	      fileType= MAP;
	      if(readType==EMPORTABLEI4)
		fileType = PACKEDPORTABLE;
	    }/* PIF MAP file */
	  else if(readType==MRCMAP)
	    {/* MRC MAP */
	      fileType= MAP;
	      nsecs = pifHeader.numSecs;
	      i = numCols;
	      j = numRows;
	    }/* MRC MAP */
	  else
	    {/* Old TSB format MAP */
	      sprintf(outStr,"[%s] is not a PIF file.",
		      fileName.addr);
	      askThem(outStr,FALSE,XmDIALOG_INFORMATION);
	      goto destructor;
	    }/* Old TSB format MAP */

	  origNumSecs= numSecs= nsecs;  /* stuff our global */

	  /* Remember what our filename was - especially if we had to 
	   * change it to portable or etc...
	   */
	  sprintf(image_fileName.addr,"%s",fileName.addr);

	  /* Ooops no image get out of here */
	  if(i==0 || j==0)
	    goto destructor;

	  /* Give feedback on main window title bar */
	  if(!masterProcess)
	    sprintf(outStr,"Robem: [%s] - [%d,%d,%d] [%s]",fileName.addr,
		    i,j,nsecs, "CHILD");
	  else
	    sprintf(outStr,"Robem: [%s] - [%d,%d,%d]",fileName.addr,
		    i,j,nsecs);
	  createSimpleString(&simpleString,outStr);
	  XtSetArg(arglst[0], XmNdialogTitle, simpleString );
	  XtSetArg(arglst[1], XmNrecomputeSize, FALSE );
	  XtSetValues(main_window_widget, arglst, 2); 
	  XFlush(dsply);
	  XmStringFree(simpleString); 
	  XFlush(dsply);

	  /* Let them toggle between sections or cued maps */
	  XtSetArg(arglst[0], XmNsensitive, TRUE);
	  XtSetValues(sectionWidget, arglst, 1); 
	  XFlush(dsply);

	  /* Remember how big we are on disk */
	  numCols= file_width = i;
	  numRows= file_height = j;

	  pickAutoZoom(i,j);

	  /* Point to where the display data is */
	  image_image1->data = (char *)image_buffer1;

	  if(fileType==FLOATMAP)
	    {/* DO crsytalDisplay */
	      int tag;
	      tag = 0;
	      crysDisp_screen_activate(0,&tag,0);
	    }/* DO crsytalDisplay */
	  else
	    {/* regular MAP disp */
	 
	      if(!calcSFs)
		{/* We stuff this in the calcSF_work() rtn */
		  /* Now stuff the display buffer from what was read off the
		   * disk - strt at zero'th(1) section */
		  
		  printf("mapi[%d] fi[%d] [%d] [%d] [%d] [%d]\n",
			 map_image_buffer,
			 float_image_buffer1, 
			 file_width,file_height,
			 nsecs,imgNum);


		  stuffMapImage((float *)map_image_buffer, 
				(float *)float_image_buffer1, 
				file_width,file_height,
				nsecs,imgNum, 0);
		  
		  /* Display the image in the window created above */
		  XPutImage (dsply, image_window1, image_gc, image_image1, 
			     0, 0,
			     0, 0, image_width, image_height);
		  XFlush (dsply);
		}/* We stuff this in the calcSF_work() rtn */
	      
	    }/* regular MAP disp */
	
	  if((maxMapPixel == minMapPixel))
	    {/* Can't do it */
	      goto destructor;
	    }/* Can't do it */

	  printf("Here circavg \n");

	  /* ************************************* *
	   * OK let's compute the Circular Average *
	   * ************************************* */
	  if(pifMode==20 && currentCircAvg < MAXBOX-2)
	    {/* OK - we've got room */
	      s1= fileName.addr;
	      s3= strrchr(s1,'/');

	      circAvg[currentCircAvg].mapName = 
		(char *)malloc(sizeof(char)*512);

	      if(s3)
		sprintf(circAvg[currentCircAvg].mapName,"%s",(s3+1));
	      else sprintf(circAvg[currentCircAvg].mapName,"%s",fileName.addr);
	      s1= circAvg[currentCircAvg].mapName;

	      createSimpleString(&simpleString, s1);
	      XmListAddItem(circAvgListW, simpleString, (currentCircAvg+1));  
	      XFlush(dsply);
	      XmStringFree(simpleString); 
	      printf("completed...\n");

	      xcen = (float)((float)numCols/2.0);
	      ycen = (float)((float)numRows/2.0);
	      zcen = (float)((float)numSecs/2.0);
	      interp= 2;  /* Original way */
	      interp= 1;  /* Check circAvgMotion() to make sure they agree */
	      nterp = interp*(int)(
				   (int)((MAX(numCols,
					      MAX(numRows,numSecs))))
				   / (int)2 
				   ) 
		+ 1;
	      max_r= (float)((float)(MAX(numCols,MAX(numRows,numSecs)))/2.0) ;
	      /* Let's go out to the corners */
	      nterp *= 2.5;
	      max_r *= 2.5;
	      nterp = numCols;
	      max_r = numCols;
	      CK(circAvg[currentCircAvg].rint= (float*)malloc(sizeof(float)*
							      1024));
	      CK(circAvg[currentCircAvg].stDev= (float*)malloc(sizeof(float)*
							      1024));

	      CK(irnum= (int *)malloc(sizeof(int)*numCols*interp));
	      CK(sum_rint= (float *)malloc(sizeof(float)*numCols*interp));
	      CK(sumsq_rint= (float *)malloc(sizeof(float)*numCols*interp));
	      circAvg[currentCircAvg].nterp = nterp;
	      map_1d_rob(float_image_buffer1,numCols,numRows,numSecs,
			 xcen,ycen,zcen,
			 interp,nterp,max_r,
			 circAvg[currentCircAvg].rint,
			 irnum,sum_rint,
			 sumsq_rint,
			 circAvg[currentCircAvg].stDev);
	      free(irnum);
	      free(sum_rint);
	      free(sumsq_rint);
	      currentCircAvg++;
	    }/* OK - we've got room */

	}/* Read a MAP file*/
      else if(readType==BOXCOORDS)
	{/* Read a box Coords */
	  if(fileType==PACKEDPORTABLE)
	    {/* Not allowed */
	      askThem("Boxing of boxed (mode==6) PIF files not allowed.",
		      FALSE,XmDIALOG_INFORMATION);
	    }/* Not allowed */
	  else
	    {readBoxCoords(fileName.addr); }
	}/* Read a box Coords */
      else if(readType==STATE)
	{/* Read a state file */
	  readState(fileName.addr);
	}/* Read a state file */
      else if(readType==SF)
	{/* Read a structFactor file */
	  readSF(fileName.addr);
	}/* Read a strucFactor file */
      else if(readType==PARAMFILE)
	{/* Read a parameter file */
	  read_param(fileName.addr, TRUE);
	}/* Read a parameter file */
      else if(readType==PARAMMAP)
	{/* Read a parameter file */
	  read_param_map(fileName.addr);
	}/* Read a parameter file */
      else if(readType==PARAMSCAN)
	{/* Read a parameter file */
	  read_param_scan(fileName.addr);
	}/* Read a parameter file */
      else if(readType==BATCHJOB)
	{/* Read a batch job file */
	  read_batchjob(fileName.addr);
	}/* Read a batch job file */

      /* Let's setup what YELLOW means */
      pixelColor= (65535.0/256.0);
      if(hiColor)
	{/* Setup for hi res color */
	  if((fileType==EMPORTABLE || fileType==PORTABLE || 
	      fileType==BYTEPORTABLE))
	    color_defs[YELLOW].red   = 255 * pixelColor;
	  else color_defs[YELLOW].red   = 0 * pixelColor;
	  color_defs[YELLOW].pixel = YELLOW;
	  color_defs[YELLOW].green = 255 * pixelColor;
	  color_defs[YELLOW].blue  = 0 * pixelColor;
	  color_defs[YELLOW].flags = DoRed | DoGreen | DoBlue;
	}/* Setup for hi res color */
      else
	{/* Setup for lo res color */
	  if((fileType==EMPORTABLE || fileType==PORTABLE || 
	      fileType==BYTEPORTABLE))
	    color_defs[pixels_return[YELLOW]].red   = 255 * pixelColor;
	  else color_defs[pixels_return[YELLOW]].red   = 0 * pixelColor;
	  color_defs[pixels_return[YELLOW]].pixel = pixels_return[YELLOW];
	  color_defs[pixels_return[YELLOW]].green = 255 * pixelColor;
	  color_defs[pixels_return[YELLOW]].blue  = 0 * pixelColor;
	  color_defs[pixels_return[YELLOW]].flags = DoRed | DoGreen | DoBlue;
	}/* Setup for lo res color */
      if(masterProcess)
	{
	  if(hiColor)
	    {XAllocColor (dsply, color_map, &color_defs[YELLOW]);
	     pixels_return[YELLOW] = color_defs[YELLOW].pixel;}
	  else
	    XStoreColor (dsply, color_map, &color_defs[pixels_return[YELLOW]]);
	}
      
      /* Update display */
      DrawWindow(); 

    }/* REad from Input file */
  else if(readWrite==WRITING)
    {/* Write to output file */
      int maxWidth,maxHeight;

      sprintf(outStr,"Writing [%s]",fileName.addr);
      if(checkStatus(outStr))
	goto destructor;

      if((writeType==CIRCLEPACKED) || 
	 (writeType==APPENDCIRCLEPACKED))
	{/* write the CIRCLE boxes */
	  short outWidth,outHeight;

	  if(fileType==PACKEDPORTABLE)  
	    {/* We read a packed portable file!!! */

	      /* Get all appodizing params before writing to write_pif() */
	      appodizing = PACKEDPORTABLE;
	      /* This is border width now */
	      sprintf(outStr,"%d",featherRadius-boxRadius);
	      XmTextSetString(appodBoxRadW,outStr);
	      sprintf(outStr,"%d",featherRadius);
	      XmTextSetString(appodFeaRadW,outStr);
	      raiseIt(appodizeScreenWidget);

	    }/* We read a packed portable file!!! */
	  else if(fileType==MAP)  
	    {/* Don't know hwo to do */
	      sprintf(outStr,
		      "I don't know how to box this type of PIF file [%d].",
		      pifMode);
	      askThem(outStr,FALSE,XmDIALOG_INFORMATION);
	    }/* Don't know hwo to do */
	  else
	    {/* We boxed a scanned image */
	      /* Write the Box coords by default - what if we're appending
	       * here???  We need to write a bcrd file FOR THE ORIGINAL
	       * image (stored in image_fileName.addr) also...*/
	      sprintf(thisFileName,"%s",image_fileName.addr);
	      s1= strrchr(thisFileName,'.');
	      if(isDatabase)
		sprintf(thisFileName,"%s",databaseBcrdStr);
	      else if(s1)
		sprintf(s1,".bcrd");
	      else sprintf(thisFileName,"%s.bcrd",image_fileName.addr);
	      writeBoxCoords(thisFileName, BOXCOORDS);
	      
	      /* Get all appodizing params before writing to write_pif() */
	      appodizing = CIRCLEPACKED;
	      /* This is border width now */
	      sprintf(outStr,"%d",featherRadius-boxRadius);
	      XmTextSetString(appodBoxRadW,outStr);
	      sprintf(outStr,"%d",featherRadius);
	      XmTextSetString(appodFeaRadW,outStr);
	      raiseIt(appodizeScreenWidget);

	    }/* We boxed a scanned image */
	  
	  /* All done - put back old cursor ... */
	  attrs.cursor= old_cursor;
	  XChangeWindowAttributes(dsply, XtWindow(drawingWidget), 
				  CWCursor, &attrs);
	  XFlush(dsply);

	}/* write the CIRCLE boxes */
      
      else if((writeType==RECTPACKED) ||
	      (writeType==APPENDRECT) )
	{/* write the RECT boxes */

	  /* Draw the rectangle */
	  if(fileType==PACKEDPORTABLE)  
	    {/* We read a packed portable file!!! */

	      /* Forget appodizing */
	      outWidth = purdueHeader.numCols;
	      outHeight= purdueHeader.numRows;
	      blindWriting = TRUE;
	      write_pif(outWidth, outHeight, fileName.addr, RECTPACKED);
	      blindWriting = FALSE;

	    }/* We read a packed portable file!!! */
	  else if(rectBox.drawable)
	    {/* Have they drawn one yet */

	      /* Write out Rectangle */
	      outWidth = rectBox.lowx - rectBox.x;
	      outHeight= rectBox.lowy - rectBox.y;
	      
	      write_pif(outWidth, outHeight, fileName.addr, RECTPACKED);
	      
	    }/* Have they drawn one yet */
	  else
	    askThem("No rectangle to write.",FALSE,XmDIALOG_INFORMATION);
	  
	}/* write the RECT boxes */

      else if(writeType==MAP)
	{/* Write a MAP */
	  outHeight= numRows;
	  outWidth = numCols;
	  write_pif(outWidth, outHeight, fileName.addr, MAP);
	}/* Write a MAP */

      else if(writeType==IMG)
	{/* Write a scanned Img */
	  if(fileType==PACKEDPORTABLE)
	    {/* Fudge it so it works */
	      float *thisPackImg;
	      outHeight= numRows;
	      outWidth = numCols;
	      if(imgInCore)
		free(inCore_buffer);
	      inCore_buffer = (float *)malloc(numCols*numRows);
	      thisPackImg = packImg[(imgNum-1)];
	      for(i=0;i<numRows;i++)
		for(j=0;j<numCols;j++)
		  {
		    inCore_buffer[((i*numCols)+j)]= 
		      thisPackImg[((i*numCols)+j)];
		  }
	      write_pif(outWidth, outHeight, fileName.addr, EMPORTABLE);
	      free(inCore_buffer);
	      imgInCore = FALSE;
	    }/* Fudge it so it works */
	  else
	    {/* Normal */
	      outHeight= file_height;
	      outWidth = file_width;
	      write_pif(outWidth, outHeight, fileName.addr, EMPORTABLE);
	    }/* Normal */
	}/* Write a scannedImg */

      else if(writeType==BOXCOORDS)
	{/* Write a box Coords */
	  writeBoxCoords(fileName.addr, BOXCOORDS);
	}/* Write a box Coords */

      else if(writeType==STATE)
	{/* Write a state file */
	  writeState(fileName.addr);
	}/* Write a state file */

      else if(writeType==PCX)
	{/* Write a pcx file */
	  writePCXbyte(fileName.addr,imgNum);
	}/* Write a pcx file */

      else if(writeType==TIFF)
	{/* Write a tiff/jpeg file */
	  char fName[256], *s2,*s3;
	  int lngth;
	  sprintf(fName,"%s",fileName.addr);
	  s2 = strrchr(fName,'.');
	  if(s2)
	    {
	      /* Trust them to have agood filename 
		 sprintf((s2),".raw");
		 */
	    }
	  else
	    {/* Cat it on */
	      lngth= strlen(fileName.addr);
	      sprintf((fName+lngth),".raw");
	    }/* Cat it on */

	  writeRAW(fName);  
	}/* Write a tiff/jpeg file */

      else if(writeType==PARAMFILE)
	{/* Write a MAP */
	  sprintf(outParamFilename,"%s",fileName.addr);
	  /* ROBEM still thinks there's a packed boxes file there... */
	  sprintf(fileName.addr,"%s",oldFilename);
	  write_param(outParamFilename);
	}/* Write a MAP */

      else askThem("[Error 583] Unknown write type",TRUE,XmDIALOG_ERROR);

    }/* Write to output file */
  else if(readWrite==READCM)
    {/* Read a colormap */
      sprintf(outStr,"Reading [%s]",fileName.addr);
      if(checkStatus(outStr))
	goto destructor;
      /* Save the old colormap */
      pushColormap();
      readColormap(fileName.addr);
      histogram_activate(&minP, &maxP);
    }/* Read a colormap */
  else if(readWrite==WRITECM)
    {/* Write a colormap */
      sprintf(outStr,"Writing [%s]",fileName.addr);
      if(checkStatus(outStr))
	goto destructor;
      writeColormap(fileName.addr);
    }/* Write a colormap */


 destructor:;
  /* Restore the old cursor whatever it was */
  attrs.cursor= old_cursor;
  XChangeWindowAttributes(dsply, XtWindow(drawingWidget), 
			  CWCursor, &attrs);

  XtUnmanageChild(statusScreenWidget);


  if(shadeItSwitch)
    {/* Automatically shade it */
      int i;

      /* Now shade them */
      mapShadeScreen_activate(0,0,0);

      /* Give feedback on what the surface density *really* is... */
      /* Make sure the user sees it with the right sign */
      surfaceDensity = shadeItDensity;
      sprintf(outStr,"%.2f",irev*surfaceDensity);
      XmTextSetString(surfDensValueWidget,outStr); 

      i = SURFDENS; /* DO we need this??? */
      mapShade_activate(0,&i,0);   

      shadeItSwitch = FALSE;  /* Don't do again next time */

    }/* Automatically shade it */


}/* GOTFILENAME */





/*-----------------------------------*/
/*************************************
 * IMGNUM callbacks                  *
 *************************************/
/*-----------------------------------*/
/* ************************************************************
 * imgNumText_create() - Remember the imgNum text widget number
 * for use later.  Init the widget to a value of 1.
 * A lot fo created widgets here could be put in the generic
 * widget creation rtn when that is built.
 */
void imgNumText_create(Widget widget, int *tag, 
		       XmAnyCallbackStruct *callback_data)
{/* IMGNUMTEXT_CREATE */
  
  if(*tag==IMGNUMBUTTON)
    {/* imgNum */
      /* Remember the widget number for later */
      imgNumTextWidget= widget;
      XmTextSetString(imgNumTextWidget, "1"); 
      handle_cr(imgNumTextWidget);
    }/* imgNum */
  else if (*tag==USEBOX)
    {/* boxInit */
      /* Remember the widget number for later */
      boxInitWidget= widget;
      XmTextSetString(boxInitWidget, "1"); 
      handle_cr(boxInitWidget);
    }/* boxInit */
  else if (*tag==USEBOXLABL)
    {/* boxInit */
      /* Remember the widget number for later */
      boxInitLabelW= widget;
    }/* boxInit */
  else if (*tag==USEZOOM)
    {/* zoom text */
      /* Remember the widget number for later */
      zoomTextW= widget;
      XmTextSetString(zoomTextW, "1.00");  
      handle_cr(zoomTextW);
    }/* zoom text */
  else if (*tag==IMGNUMDOWN)
    {/* imgnum DOWN */
      /* Remember the widget number for later */
      imgnumArrowDownWidget= widget;
      handle_cr(imgnumArrowDownWidget);
    }/* imgnum DOWN */
  else if (*tag==IMGNUMUP)
    {/* imgnum UP */
      /* Remember the widget number for later */
      imgnumArrowUpWidget= widget;
      handle_cr(imgnumArrowUpWidget);
    }/* imgnum UP */
  else if (*tag==ZOOMDOWN)
    {/* zoom DOWN */
      /* Remember the widget number for later */
      zoomArrowDownWidget= widget;
      handle_cr(zoomArrowDownWidget);
    }/* zoom DOWN */
  else if (*tag==ZOOMUP)
    {/* zoom UP */
      /* Remember the widget number for later */
      zoomArrowUpWidget= widget;
      handle_cr(zoomArrowUpWidget);
    }/* zoom UP */
  else if (*tag==BOXINITDOWN)
    {/* boxinit DOWN */
      /* Remember the widget number for later */
      boxinitArrowDownWidget= widget;
      handle_cr(boxinitArrowDownWidget);
    }/* boxinit DOWN */
  else if (*tag==BOXINITUP)
    {/* boxinit UP */
      /* Remember the widget number for later */
      boxinitArrowUpWidget= widget;
      handle_cr(boxinitArrowUpWidget);
    }/* boxinit UP */
  else if (*tag==XORIGUP)
    {/* xorig UP */
      /* Remember the widget number for later */
      xorigUpArrowW= widget;
      handle_cr(widget);
    }/* xorig UP */
  else if (*tag==XORIGDOWN)
    {/* xorig DOWN */
      /* Remember the widget number for later */
      xorigDownArrowW= widget;
      handle_cr(widget);
    }/* xorig DOWN */
  else if (*tag==YORIGUP)
    {/* yorig UP */
      /* Remember the widget number for later */
      yorigUpArrowW= widget;
      handle_cr(widget);
    }/* yorig UP */
  else if (*tag==YORIGDOWN)
    {/* yorig DOWN */
      /* Remember the widget number for later */
      yorigDownArrowW= widget;
      handle_cr(widget);
    }/* yorig DOWN */
  else if (*tag==DELTAX)
    {/*  */
      /* Remember the widget number for later */
      deltaXpartprocW= widget;
    }/*  */
  else if (*tag==DELTAY)
    {/*  */
      /* Remember the widget number for later */
      deltaYpartprocW= widget;
    }/*  */

  
}/* IMGNUMTEXT_CREATE */



/* ************************************************************
 * imgNumText_changed() - When they change the imgNumText widget
 * they need to press the "imgNum" button next to it to get
 * the change instantiated.  
 */
void imgNumText_changed(Widget widget, int *tag, 
			XmAnyCallbackStruct *callback_data)
{/* IMGNUMTEXT_CHANGED */
  char *imgNumText;

  if(!locked)
    {/* Can't do if we're locked out */
      if(*tag==USEBOX)
	{/* boxInit */
	  imgNumText = XmTextGetString(boxInitWidget); 
	  
	  lastNumBox =  (int)strtod(imgNumText, 0);
	  lastNumBox--;
	  if(lastNumBox < 0)
	    lastNumBox = 0;
	  
	  XtFree(imgNumText);

	}/* boxInit */
      else if(*tag==IMGNUMBUTTON)
	{/* imgNum */
	  imgNumText = XmTextGetString(imgNumTextWidget); 
	  
	  imgNum = (int)strtod(imgNumText, 0);
	  if(imgNum < 1)
	    imgNum = 1;
	  if(fileType==PACKEDPORTABLE || fileType==MAP)
	    imgNum = MIN(imgNum,purdueHeader.nimg);
	  if(imgNum < 1)
	    imgNum = 1;
	  if(shadedMaps && !dispSections)
	    imgNum = MIN(imgNum,mapNumSecs);

	  XtFree(imgNumText);
	  XFlush(dsply);

	}/* imgNum */

    }/* Can't do if we're locked out */
  else
    {/* Maybe they're trying to find a certain box */
      if(*tag==USEBOX &&
	 (fileType==EMPORTABLE || fileType==PORTABLE || 
	  fileType==BYTEPORTABLE))
	{/* boxInit */
	  imgNumText = XmTextGetString(boxInitWidget); 
	  
	  findBox =  (int)strtod(imgNumText, 0);
	  findBox -= (lastNumBox+1);  /* Humans number from 1 not 0 */

	  findBox = MIN(MAX(findBox,0),(numBoxes-1));

	  XtFree(imgNumText);

	}/* boxInit */
	 
    }/* Maybe they're trying to find a certain box */
  

}/* IMGNUMTEXT_CHANGED */




/* ************************************************************
 * doTheImgNum() - Called from imgNumArrowActivate and 
 * imgNumArrowActivateButton().  Do the work required to change
 * the imgNum concept.  speedFactor determines how "fast"
 * to change the value of imgNum
 */
void doTheImgNum(Widget widget, float speedFactor)
{/* DOTHEIMGNUM */
  char newVal[22];

  if(
     (widget==imgnumArrowUpWidget) ||
     (widget==fftImgnumArrowUpW) ||
     (widget==rotfltImgnumArrowUpW)
     )
    {/* Up */
      imgNum += (int)speedFactor;
    }/* Up */
  else if(
	  (widget==imgnumArrowDownWidget) ||
	  (widget==fftImgnumArrowDownW) ||
	  (widget==rotfltImgnumArrowDownW)
	  )
    {/* Dopwn */
      imgNum -= (int)speedFactor;
    }/* Down */
  else printf("ImgNum unchanged.\n");
  
  if(fileType==PACKEDPORTABLE || fileType==MAP)
    imgNum = MIN(imgNum,purdueHeader.nimg);
  if(imgNum < 1)
    imgNum = 1;
  if(shadedMaps && !dispSections)
    imgNum = MIN(imgNum,mapNumSecs);
  
  sprintf(newVal,"%d",imgNum);
  XmTextSetString(imgNumTextWidget, newVal);
  
  if(haveParamFile)
    {/* Make the param screen consistent */
      P_INFO *thisInfo;
      int thisX,thisY;

      thisInfo = info.info;
      while(thisInfo != NULL)
	{/* Find this one */
	  if(thisInfo->rank==imgNum)
	    {/* Found it */
	      break;
	    }/* Found it */
	  thisInfo = thisInfo->next;
	}/* Find this one */

      /* Display this on the parm screen */
      convertXY(&thisX, &thisY, thisInfo->phi, thisInfo->theta);
      findNearestParticle(&thisX,&thisY,&foundPart,FALSE,imgNum,TRUE,0);

      sprintf(outStr,"%d",thisInfo->rank);
      XmTextSetString(ppinTextW,outStr);

    }/* Make the param screen consistent */
  else
    {/* Try these */
      if(fileType==PACKEDPORTABLE)
	stuffPackedImage(purdueHeader.numCols,
			 purdueHeader.numRows,
			 packImg,purdueHeader.nimg,
			 imgNum);  /* Start at zero'th image */
      else if(fileType==MAP)
	stuffMapImage((float *)map_image_buffer, 
		      (float *)float_image_buffer1, 
		      file_width,file_height,
		      nsecs,
		      imgNum, 0);  /* Start at zero'th image */
      
      DrawWindow(); 
    }/* Try these */
  
  if(XtIsManaged(subtractMapScreen) && diffSection)
    diffSec_really_doit();

}/* DOTHEIMGNUM */




/* ************************************************************
 * imgNumArrowActivate() - Instead of making them type the 
 * new imgNum just let them clikc the arrow.  Change the imgNum
 * accordingly.  This rtn is for the first mouse button only.
 */
void imgNumArrowActivate(Widget widget, int *tag, 
			 XmPushButtonCallbackStruct *callback_data)
{/* IMGNUMARROWACTIVATE */
  float speedFactor;

  if(!locked)
    {/* Can't do if we're locked out */
      
      speedFactor = 1.0;

      doTheImgNum(widget, speedFactor);

    }/* Can't do if we're locked out */

}/* IMGNUMARROWACTIVATE */




/* ************************************************************
 * imgNumArrowActivateButton() - Instead of making them type the 
 * new imgNum just let them clikc the arrow.  Change the imgNum
 * accordingly.  This rtn is for the second and third mouse buttons.
 */
void imgNumArrowActivateButton(Widget widget, XEvent *event)
{/* IMGNUMARROWACTIVATEBUTTON */
  XEvent thisEvent;
  float speedFactor;

  if(!locked)
    {/* Can't do if we're locked out */
      
      thisEvent= *(event);

      /* Get which button */
      if((thisEvent.xbutton.button == Button1))
	speedFactor = 1.0;;
      if((thisEvent.xbutton.button == Button2))
	speedFactor = 2.0;
      if((thisEvent.xbutton.button == Button3))
	speedFactor = 5.0;

      doTheImgNum(widget, speedFactor);

    }/* Can't do if we're locked out */

}/* IMGNUMARROWACTIVATEBUTTON */




/* ************************************************************
 * boxInitActivate() - When the "BoxInit" button is pressed 
 * redisplay the window using the new boxInit.
 */
void boxInitActivate(Widget  widget,
		     int     *tag,
		     XmAnyCallbackStruct *callback_data)
{/* BOXINITACTIVATE */
  XmString simpleString;

  if(!locked)
    {/* Only do while not locked */

      sprintf(outStr,"%d",lastNumBox+1);
      XmTextSetString(boxInitWidget, outStr);
      XFlush(dsply);

      DrawWindow(); 

    }/* Only do while not locked */
  else
    {/* Draw the findBox stuff if appropriate */

      DrawWindow(); 

    }/* Draw the findBox stuff if appropriate */

}/* BOXINITACTIVATE */




/* ************************************************************
 * doTheBoxInit() - Called from boxInitArrowActivate and 
 * boxInitArrowActivateButton().  Do the work required to change
 * the first/find box concept. speedFactor determines how
 * much to increment or decrement the value.
 */
void doTheBoxInit(Widget widget, float speedFactor)
{/* DOTHEBOXINIT */

  if(!locked)
    {/* Can't do if we're locked out */
      
      lastNumBox +=  (int)speedFactor;
      if(lastNumBox < 0)
	lastNumBox = 0;
      
      sprintf(outStr,"%d",lastNumBox+1);
      XmTextSetString(boxInitWidget, outStr);
      XFlush(dsply);
      
      DrawWindow(); 
      
    }/* Can't do if we're locked out */
  else
    {/* Maybe they're trying to find a certain box */
      if((fileType==EMPORTABLE || fileType==PORTABLE || 
	  fileType==BYTEPORTABLE))
	{/* boxInit */
	  
	  findBox +=  (int)speedFactor;
	  findBox = MIN(MAX(findBox,0),(numBoxes-1));

	  sprintf(outStr,"%d",findBox+lastNumBox+1);
	  XmTextSetString(boxInitWidget, outStr);
	  XFlush(dsply);

	  DrawWindow(); 
	}/* boxInit */
	 
    }/* Maybe they're trying to find a certain box */
  
}/* DOTHEBOXINIT */




/* ************************************************************
 * boxInitArrowActivateButton() - Instead of making them type the 
 * new first/find box just let them clikc the arrow.  Change the first/
 * find box variable
 * accordingly.  This rtn is for the second and third mouse buttons.
 */
void boxInitArrowActivateButton(Widget widget, XEvent *event)
{/* BOXINITARROWACTIVATEBUTTON */
  XEvent thisEvent;
  float speedFactor;

  thisEvent= *(event);
  
  /* Get which button */
  if((thisEvent.xbutton.button == Button1))
    speedFactor = 1.0;;
  if((thisEvent.xbutton.button == Button2))
    speedFactor = 2.0;
  if((thisEvent.xbutton.button == Button3))
    speedFactor = 5.0;
  
  if(widget==boxinitArrowDownWidget)
    speedFactor *= -1.0;
  
  doTheBoxInit(widget, speedFactor);
  
}/* BOXINITARROWACTIVATEBUTTON */




/* ************************************************************
 * doTheBoxNum() - Called from boxNumArrowActivate and 
 * boxNumArrowActivateButton().  Do the work required to change
 * the box/section number on boxedit/movie screen.
 */
void doTheBoxNum(Widget widget, float speedFactor)
{/* DOTHEBOXNUM */
  char curStr[12];

  curBox+= (int)speedFactor;

  if(makingMovie)
    {/* Put the proper movie frame up there */
      /* Set the text widget to agree */
      if(shadedMaps && stereoMapShade)
	curBox = MAX(0,MIN(curBox,(numSecs-2)));
      else curBox = MAX(0,MIN(curBox,(numSecs-1)));
      /* Human numbers from one not zero */
      sprintf(curStr,"%d",(curBox+lastNumBox+1));
      XmTextSetString(editBoxTextWidget, curStr);
      XFlush(dsply);
      /* Display the new box */
      stuffDisplayMapBuffer(curBox,editBoxZoom,0);
      if(shadedMaps && stereoMapShade)
	stuffDisplayMapBuffer(curBox+1,editBoxZoom,numCols);
      DrawEditBoxWindow(TRUE);
    }/* Put the proper movie frame up there */
  else
    {/* Put box up */
      /* Set the text widget to agree */
      if(fileType==PACKEDPORTABLE)
	curBox = MAX(0,MIN(curBox,(purdueHeader.nimg-1)));
      else curBox = MAX(0,MIN(curBox,((numBoxes)-1)));
      /* Human numbers from one not zero */
      sprintf(curStr,"%d",(curBox+lastNumBox+1));
      XmTextSetString(editBoxTextWidget, curStr);
      XFlush(dsply);
      /* Display the new box */
      if(fileType==PACKEDPORTABLE)
	stuffDisplayPackBoxBuffer(curBox, editBoxZoom, FALSE);
      else stuffDisplayBoxBuffer(curBox,image_fileName.addr,editBoxZoom);
      DrawEditBoxWindow(FALSE);
    }/* Put box up */

}/* DOTHEBOXNUM */




/* ************************************************************
 * boxNumArrowActivateButton() - Instead of making them type the 
 * new box/section number to edit on the editBox screen, just 
 * let them click the arrow.  
 */
void boxNumArrowActivateButton(Widget widget, XEvent *event)
{/* boxNumARROWACTIVATEBUTTON */
  XEvent thisEvent;
  float speedFactor;

  thisEvent= *(event);
  
  /* Get which button */
  if((thisEvent.xbutton.button == Button1))
    speedFactor = 1.0;;
  if((thisEvent.xbutton.button == Button2))
    speedFactor = 2.0;
  if((thisEvent.xbutton.button == Button3))
    speedFactor = 5.0;
  
  if(widget==boxDownWidget)
    speedFactor *= -1.0;
  
  doTheBoxNum(widget, speedFactor);
  
}/* boxNumARROWACTIVATEBUTTON */




/* ************************************************************
 * doThePaintSize() - Called from paintSizeArrowActivate and 
 * paintSizeArrowActivateButton().  Do the work required to change
 * the paint brush size.
 */
void doThePaintSize(Widget widget, float speedFactor)
{/* DOTHEPAINTSIZE */

  paintSize+= (int)speedFactor;

  paintSize = MAX(1,MIN(paintSize, 50));

  sprintf(outStr,"%d",paintSize);
  XmTextSetString(editBoxPaintTextWidget,outStr); 
  
}/* DOTHEPAINTSIZE */





/* ************************************************************
 * paintSizeArrowActivateButton() - Instead of making them type the 
 * new paint brush size just let them click the arrow.  
 */
void paintSizeArrowActivateButton(Widget widget, XEvent *event)
{/* PAINTSIZEARROWACTIVATEBUTTON */
  XEvent thisEvent;
  float speedFactor;

  thisEvent= *(event);
  
  /* Get which button */
  if((thisEvent.xbutton.button == Button1))
    speedFactor = 1.0;;
  if((thisEvent.xbutton.button == Button2))
    speedFactor = 2.0;
  if((thisEvent.xbutton.button == Button3))
    speedFactor = 5.0;
  
  if(widget==paintDownWidget)
    speedFactor *= -1.0;
  
  doThePaintSize(widget, speedFactor);
  
}/* PAINTSIZEARROWACTIVATEBUTTON */





/* ************************************************************
 * doTheCmScroll() - Called from cmScrollArrowActivate and 
 * cmScrollArrowActivateButton().  Do the work required to scroll
 * the colormap.
 */
void doTheCmScroll(Widget widget, float speedFactor)
{/* DOTHECMSCROLL */
  int i,nextPixel;
  XColor tempColor[7];

  if(speedFactor > 0.0)
    {/* UP */

      for(i=0;i<6;i++)
	{/* Let's store all of these even though we mayn't need them all */
	  nextPixel = (tryColors-1) - i;
	  if(hiColor)
	    {/* Do for hi color */
	      tempColor[i].pixel = nextPixel;
	      tempColor[i].red   = color_defs[nextPixel].red;
	      tempColor[i].green = color_defs[nextPixel].green;
	      tempColor[i].blue  = color_defs[nextPixel].blue;
	      tempColor[i].flags = color_defs[nextPixel].flags;
	    }/* Do for hi color */
	  else
	    {/* Do for lo color */
	      tempColor[i].pixel = pixels_return[(nextPixel)];
	      tempColor[i].red   = color_defs[pixels_return[nextPixel]].red;
	      tempColor[i].green = color_defs[pixels_return[nextPixel]].green;
	      tempColor[i].blue  = color_defs[pixels_return[nextPixel]].blue;
	      tempColor[i].flags = color_defs[pixels_return[nextPixel]].flags;
	    }/* Do for lo color */
	}/* Let's store all of these even though we mayn't need them all */
      
      for(i=(tryColors-1);i>=(7+(int)speedFactor);i--)
	{/* Do the colors we have been assigned */
	  nextPixel = i - (int)speedFactor;
	  if(hiColor)
	    {/* hi */
	      color_defs[i].pixel = i;
	      color_defs[i].red   = 
		color_defs[nextPixel].red;
	      color_defs[i].green   = 
		color_defs[nextPixel].green;
	      color_defs[i].blue   = 
		color_defs[nextPixel].blue;
	      color_defs[i].flags   = 
		color_defs[nextPixel].flags;
	    }/* hi */
	  else
	    {/* lo */
	      color_defs[pixels_return[i]].pixel = pixels_return[i];
	      color_defs[pixels_return[i]].red   = 
		color_defs[pixels_return[nextPixel]].red;
	      color_defs[pixels_return[i]].green   = 
		color_defs[pixels_return[nextPixel]].green;
	      color_defs[pixels_return[i]].blue   = 
		color_defs[pixels_return[nextPixel]].blue;
	      color_defs[pixels_return[i]].flags   = 
		color_defs[pixels_return[nextPixel]].flags;
	    }/* lo */
	  if(masterProcess)
	    {
	      if(hiColor)
		{XAllocColor (dsply, color_map, &color_defs[i]);
		 pixels_return[i] = color_defs[i].pixel;}
	      else 
		XStoreColor (dsply, color_map, 
			     &color_defs[pixels_return[i]]);
	    }
	}/* Do the colors we have been assigned */

      for(i=(7);i<(7+(int)speedFactor);i++)
	{/* Do the rest */
	  nextPixel = ((7+(int)speedFactor-1)-i);
	  if(hiColor)
	    {/* hi */
	      color_defs[i].pixel = i;
	      color_defs[i].red   = tempColor[nextPixel].red;
	      color_defs[i].green = tempColor[nextPixel].green;
	      color_defs[i].blue  = tempColor[nextPixel].blue;
	      color_defs[i].flags = tempColor[nextPixel].flags;
	    }/* hi */
	  else
	    {/* lo */
	      color_defs[pixels_return[i]].pixel = pixels_return[i];
	      color_defs[pixels_return[i]].red   = tempColor[nextPixel].red;
	      color_defs[pixels_return[i]].green = tempColor[nextPixel].green;
	      color_defs[pixels_return[i]].blue  = tempColor[nextPixel].blue;
	      color_defs[pixels_return[i]].flags = tempColor[nextPixel].flags;
	    }/* lo */
	  if(masterProcess)
	    {
	      if(hiColor)
		{XAllocColor (dsply, color_map, &color_defs[i]);
		 pixels_return[i] = color_defs[i].pixel;}
	      else
		XStoreColor (dsply, color_map, 
			     &color_defs[pixels_return[i]]);
	    }
	}/* Do the rest */

    }/* UP */
  else if((int)speedFactor < 0)
    {/* DOWN */

      for(i=0;i<6;i++)
	{/* Let's store all of these even though we mayn't need them all */
	  nextPixel = (7+i);
	  if(hiColor)
	    {/* hi */
	      tempColor[i].pixel = nextPixel;
	      tempColor[i].red = color_defs[nextPixel].red;
	      tempColor[i].green = color_defs[nextPixel].green;
	      tempColor[i].blue = color_defs[nextPixel].blue;
	      tempColor[i].flags = color_defs[nextPixel].flags;
	    }/* hi */
	  else
	    {/* lo */
	      tempColor[i].pixel = pixels_return[(nextPixel)];
	      tempColor[i].red = color_defs[pixels_return[nextPixel]].red;
	      tempColor[i].green = color_defs[pixels_return[nextPixel]].green;
	      tempColor[i].blue = color_defs[pixels_return[nextPixel]].blue;
	      tempColor[i].flags = color_defs[pixels_return[nextPixel]].flags;
	    }/* lo */
	}/* Let's store all of these even though we mayn't need them all */
      
      for(i=(7);i<((tryColors)+(int)speedFactor);i++)
	{/* Do the colors we have been assigned */
	  nextPixel = i - (int)speedFactor;
	  if(hiColor)
	    {/* hi */
	      color_defs[i].pixel = i;
	      color_defs[i].red   = color_defs[nextPixel].red;
	      color_defs[i].green = color_defs[nextPixel].green;
	      color_defs[i].blue  = color_defs[nextPixel].blue;
	      color_defs[i].flags = color_defs[nextPixel].flags;
	    }/* hi */
	  else
	    {/* lo */
	      color_defs[pixels_return[i]].pixel = pixels_return[i];
	      color_defs[pixels_return[i]].red   = 
		color_defs[pixels_return[nextPixel]].red;
	      color_defs[pixels_return[i]].green   = 
		color_defs[pixels_return[nextPixel]].green;
	      color_defs[pixels_return[i]].blue   = 
		color_defs[pixels_return[nextPixel]].blue;
	      color_defs[pixels_return[i]].flags   = 
		color_defs[pixels_return[nextPixel]].flags;
	    }/* lo */
	  if(masterProcess)
	    {
	      if(hiColor)
		{XAllocColor (dsply, color_map, &color_defs[i]);
		 pixels_return[i] = color_defs[i].pixel;}
	      else
		XStoreColor (dsply, color_map, 
			     &color_defs[pixels_return[i]]);
	    }
	}/* Do the colors we have been assigned */

      for(i=((tryColors)+(int)speedFactor);i<(tryColors);i++)
	{/* Do the rest */
	  nextPixel = (i-((tryColors-1)+(int)speedFactor));
	  if(hiColor)
	    {/* hi */
	      color_defs[i].pixel = i;
	      color_defs[i].red   = tempColor[nextPixel].red;
	      color_defs[i].green = tempColor[nextPixel].green;
	      color_defs[i].blue  = tempColor[nextPixel].blue;
	      color_defs[i].flags = tempColor[nextPixel].flags;
	    }/* hi */
	  else
	    {/* lo */
	      color_defs[pixels_return[i]].pixel = pixels_return[i];
	      color_defs[pixels_return[i]].red   = tempColor[nextPixel].red;
	      color_defs[pixels_return[i]].green = tempColor[nextPixel].green;
	      color_defs[pixels_return[i]].blue  = tempColor[nextPixel].blue;
	      color_defs[pixels_return[i]].flags = tempColor[nextPixel].flags;
	    }/* lo */
	  if(masterProcess)
	    {
	      if(hiColor)
		{XAllocColor (dsply, color_map, &color_defs[i]);
		 pixels_return[i] = color_defs[i].pixel;}
	      else
		XStoreColor (dsply, color_map, 
			     &color_defs[pixels_return[i]]);
	    }
	}/* Do the rest */

    }/* DOWN */

}/* DOTHECMSCROLL */




/* ************************************************************
 * cmScrollArrowActivateButton() - Scroll the colormap.
 */
void cmScrollArrowActivateButton(Widget widget, XEvent *event)
{/* CMSCROLLARROWACTIVATEBUTTON */
  XEvent thisEvent;
  float speedFactor;

  thisEvent= *(event);
  
  /* Get which button */
  if((thisEvent.xbutton.button == Button1))
    speedFactor = 1.0;;
  if((thisEvent.xbutton.button == Button2))
    speedFactor = 2.0;
  if((thisEvent.xbutton.button == Button3))
    speedFactor = 5.0;
  
  if(widget==colormapScrollDownWidget)
    speedFactor *= -1.0;
  
  doTheCmScroll(widget, speedFactor);
  
}/* CMSCROLLARROWACTIVATEBUTTON */





/* ************************************************************
 * imgNumActivate() - When the "ImgNum" button is pressed 
 * redisplay the window using the new imgNum. From main screen.
 */
void imgNumActivate(Widget widget, int *tag, 
		    XmAnyCallbackStruct *callback_data)
{/* IMGNUMACTIVATE */

  if(!locked)
    {/* Only do while not locked */
      if(haveParamFile)
	{/* Make the param screen consistent */
	  P_INFO *thisInfo;
	  int thisX,thisY;
	  
	  thisInfo = info.info;
	  while(thisInfo != NULL)
	    {/* Find this one */
	      if(thisInfo->particle_number==imgNum)
		{/* Found it */
		  break;
		}/* Found it */
	      thisInfo = thisInfo->next;
	    }/* Find this one */
	  
	  /* Display this on the parm screen */
	  convertXY(&thisX, &thisY, thisInfo->phi, thisInfo->theta);
	  findNearestParticle(&thisX,&thisY,&foundPart,FALSE,imgNum,TRUE,0);
	  
	}/* Make the param screen consistent */
      else
	{/* Try these */
	  if(fileType==PACKEDPORTABLE)
	    stuffPackedImage(purdueHeader.numCols,
			     purdueHeader.numRows,
			     packImg,purdueHeader.nimg,
			     imgNum);  /* Start at zero'th image */
	  else if(fileType==MAP)
	    stuffMapImage((float *)map_image_buffer, 
			  (float *)float_image_buffer1, 
			  file_width,file_height,
			  nsecs,
			  imgNum, 0);  /* Start at zero'th image */
	  
	  DrawWindow();
	}/* Try these */
    }/* Only do while not locked */
  
}/* IMGNUMACTIVATE */




/*-----------------------------------*/
/*************************************
 * COLORMAP callbacks                *
 *************************************/
/*-----------------------------------*/
/* ************************************************************
 * colormap_ramp_change() - From colormap screen
 */
void colormap_ramp_change(Widget widget, int *tag, 
			  XmAnyCallbackStruct *callback_data)
{/* COLORMAP_RAMP_CHANGE */

  if(*tag==RAINBOW)
    rainbowSwitch = !rainbowSwitch;
  else if(*tag==TEST)
    rampSwitch = !rampSwitch;

}/* COLORMAP_RAMP_CHANGE */


/* ************************************************************
 * colormap_ramp_activate() - They are modifying the colormap
 * from the interface.
 */
void colormap_ramp_activate(Widget widget, int *tag, 
			    XmAnyCallbackStruct *callback_data)
{/* colormap_ramp_ACTIVATE */
  float percStep;
  int i,j,k;

  buildingCmap = TRUE;
  if(colorType==USEBUILD && (firstBuildY > 6) && (lastBuildY > 6))
    {/* Only do if we're building */
      
      if(!rampSwitch && !rainbowSwitch)
	{/* Set the colormap to this color - NO RAMPING */
	  pixelColor= (65535.0/256.0);
	  for(i=firstBuildY;i<=lastBuildY;i++)
	    {/* Do the colors we have been assigned */
	      if(hiColor)
		{/* hi */
		  color_defs[i].pixel = i;
		  color_defs[i].red   = 
		    ((int)(redLevel) * pixelColor);
		  color_defs[i].green =
		    ((int)(greenLevel) * pixelColor);
		  color_defs[i].blue  = 
		    ((int)(blueLevel) * pixelColor);
		  color_defs[i].flags = 
		    DoRed | DoGreen | DoBlue;
		}/* hi */
	      else
		{/* lo */
		  color_defs[pixels_return[i]].pixel = pixels_return[i];
		  color_defs[pixels_return[i]].red   = 
		    ((int)(redLevel) * pixelColor);
		  color_defs[pixels_return[i]].green =
		    ((int)(greenLevel) * pixelColor);
		  color_defs[pixels_return[i]].blue  = 
		    ((int)(blueLevel) * pixelColor);
		  color_defs[pixels_return[i]].flags = 
		    DoRed | DoGreen | DoBlue;
		}/* lo */
	      if(masterProcess)
		{
		  if(hiColor)
		    {XAllocColor (dsply, color_map, &color_defs[i]);
		     pixels_return[i] = color_defs[i].pixel;}
		  else
		    XStoreColor (dsply, color_map, 
				 &color_defs[pixels_return[i]]);
		}
	    }/* Do the colors we have been assigned */
	}/* Set the colormap to this color - NO RAMPING */
      else if(rampSwitch && !rainbowSwitch)
	{/* RAMPING */

	  if(lastBuildY > firstBuildY)
	    percStep = (brightLevel/256.0) / (float)(lastBuildY - firstBuildY);
	  else percStep = 0.0;
	  
	  /* Set the colormap to this color */
	  pixelColor= (65535.0/256.0);
	  for(i=firstBuildY;i<=lastBuildY;i++)
	    {/* Do the colors we have been assigned */
	      int thisRedLevel,thisGreenLevel,thisBlueLevel;
	      thisRedLevel = redLevel + (
					 255.0 *
					 ((i-firstBuildY)*percStep)
					 );
	      thisRedLevel = MIN(255,thisRedLevel);
	      thisGreenLevel = greenLevel + (
					     255.0 *
					     ((i-firstBuildY)*percStep)
					     );
	      thisGreenLevel = MIN(255,thisGreenLevel);
	      thisBlueLevel = blueLevel + (
					   255.0 *
					   ((i-firstBuildY)*percStep)
					   );
	      thisBlueLevel = MIN(255,thisBlueLevel);
	      if(hiColor)
		{/* hi */
		  color_defs[i].pixel = i;
		  color_defs[i].red   = 
		    ((int)(thisRedLevel) * pixelColor);
		  color_defs[i].green =
		    ((int)(thisGreenLevel) * pixelColor);
		  color_defs[i].blue  = 
		    ((int)(thisBlueLevel) * pixelColor);
		  color_defs[i].flags = 
		    DoRed | DoGreen | DoBlue;
		}/* hi */
	      else
		{/* lo */
		  color_defs[pixels_return[i]].pixel = pixels_return[i];
		  color_defs[pixels_return[i]].red   = 
		    ((int)(thisRedLevel) * pixelColor);
		  color_defs[pixels_return[i]].green =
		    ((int)(thisGreenLevel) * pixelColor);
		  color_defs[pixels_return[i]].blue  = 
		    ((int)(thisBlueLevel) * pixelColor);
		  color_defs[pixels_return[i]].flags = 
		    DoRed | DoGreen | DoBlue;
		}/* lo */
	      if(masterProcess)
		{
		  if(hiColor)
		    {XAllocColor (dsply, color_map, &color_defs[i]);
		     pixels_return[i] = color_defs[i].pixel;}
		  else
		    XStoreColor (dsply, color_map, 
				 &color_defs[pixels_return[i]]);
		}
	    }/* Do the colors we have been assigned */
	  
	  /* Now "darken" any of the pixels of a radially cued image */
	  if(secondMapShade && radialMapShade)
	    {/* darken appropriate */
	      int darkVal;
	      darkVal = (int)(65535/(expandStep+2));
	      darkVal *= globFract;  /* Use the fraction scroll to soften
				      * the shading */

	      for(i=(6+1+(int)radialStepSize);i<(tryColors-expandStep);i+= 
		                                                  expandStep)
		{/* Shade every four */
		  if(i>=firstBuildY && i <= lastBuildY)
		    {/* Only do marked */
		      for(j=(expandStep-2);j>=0;j--)
			{/* Ramp all */
			  if(hiColor)
			    {/* hi */
			      color_defs[i+j].pixel = i+j;
			      if((int)
				 ((int)color_defs[(i+(expandStep-1))].red 
				  - 
				  (int)(darkVal*(((expandStep-2)-j)+1))) < 0)
				color_defs[i+j].red   = 0;
			      else color_defs[i+j].red   = 
				color_defs[(i+(expandStep-1))].red - 
				      (int)((darkVal*(((expandStep-2)-j)+1)));
			      if((int)
				 ((int)color_defs[(i+(expandStep-1))].green 
				  - 
				  (int)(darkVal*(((expandStep-2)-j)+1))) < 0)
				color_defs[i+j].green = 0;
			      else color_defs[i+j].green = 
				color_defs[(i+(expandStep-1))].green - 
				     (int)((darkVal*(((expandStep-2)-j)+1)));
			      if((int)
				 ((int)color_defs[(i+(expandStep-1))].blue 
				  - 
				  (int)(darkVal*(((expandStep-2)-j)+1))) < 0)
				color_defs[i+j].blue = 0;
			      else color_defs[i+j].blue = 
				color_defs[(i+(expandStep-1))].blue 
				  - 
			      (int)((darkVal*(((expandStep-2)-j)+1)));
			      color_defs[i+j].flags = 
				DoRed | DoGreen | DoBlue;
			    }/* hi */
			  else
			    {/* lo */
			      color_defs[pixels_return[i+j]].pixel = 
				pixels_return[i+j];
			      if((int)
				 ((int)color_defs[
				      pixels_return[(i+(expandStep-1))]].red 
				  - 
				  (int)(darkVal*(((expandStep-2)-j)+1))) < 0)
				color_defs[pixels_return[i+j]].red   = 0;
			      else color_defs[pixels_return[i+j]].red   = 
				color_defs[
				    pixels_return[(i+(expandStep-1))]].red - 
				     (int)((darkVal*(((expandStep-2)-j)+1)));
			      if((int)
				 ((int)color_defs[
				    pixels_return[(i+(expandStep-1))]].green 
				  - 
				  (int)(darkVal*(((expandStep-2)-j)+1))) < 0)
				color_defs[pixels_return[i+j]].green = 0;
			      else color_defs[pixels_return[i+j]].green = 
				color_defs[
				  pixels_return[(i+(expandStep-1))]].green - 
				     (int)((darkVal*(((expandStep-2)-j)+1)));
			      if((int)
				 ((int)color_defs[
				     pixels_return[(i+(expandStep-1))]].blue 
				  - 
				  (int)(darkVal*(((expandStep-2)-j)+1))) < 0)
				color_defs[pixels_return[i+j]].blue = 0;
			      else color_defs[pixels_return[i+j]].blue = 
			   color_defs[pixels_return[(i+(expandStep-1))]].blue 
				  - 
			      (int)((darkVal*(((expandStep-2)-j)+1)));
			      color_defs[pixels_return[i+j]].flags = 
				DoRed | DoGreen | DoBlue;
			    }/* lo */
			  if(masterProcess)
			    {
			      if(hiColor)
				{XAllocColor (dsply,color_map,
					      &color_defs[i+j]);
				 pixels_return[i+j] = color_defs[i+j].pixel;}
			      else
				XStoreColor (dsply, color_map, 
					     &color_defs[pixels_return[i+j]]);
			    }
			}/* Ramp all */
		      
		      /* Do the BLACK special case */
		      if(hiColor)
			{/* hi */
			  color_defs[i].pixel = i;
			  color_defs[i].red   = 0;
			  color_defs[i].green = 0;
			  color_defs[i].blue = 0;
			}/* hi */
		      else
			{/* lo */
			  color_defs[pixels_return[i]].pixel = 
			    pixels_return[i];
			  color_defs[pixels_return[i]].red   = 0;
			  color_defs[pixels_return[i]].green = 0;
			  color_defs[pixels_return[i]].blue = 0;
			  color_defs[pixels_return[i]].flags = 
			    DoRed | DoGreen | DoBlue;
			}/* lo */

		      if(masterProcess)
			{
			  if(hiColor)
			    {XAllocColor (dsply, color_map, &color_defs[i]);
			     pixels_return[i] = color_defs[i].pixel;}
			  else
			    XStoreColor (dsply, color_map, 
					 &color_defs[pixels_return[i]]);
			}
		      
		    }/* Only do marked */
		  
		}/* Shade every four */
	      
	    }/* darken appropriate */
	  
	}/* RAMPING */
      else if(!rampSwitch && rainbowSwitch)
	{/* RAinbowING */

	  if(lastBuildY > firstBuildY)
	    percStep = (brightLevel/256.0) / (float)(lastBuildY - firstBuildY);
	  else percStep = 0.0;
	  
	  /* Set the colormap to this color */
	  pixelColor= (65535.0/256.0);
	  for(i=firstBuildY;i<=lastBuildY;i++)
	    {/* Do the colors we have been assigned */
	      int thisRedLevel,thisGreenLevel,thisBlueLevel;
	      thisRedLevel = redLevel + (
					 (382.0 *
					 ((i-firstBuildY)*percStep))
					 );
	      thisRedLevel = MAX(MIN(255,thisRedLevel),0);
	      thisGreenLevel = greenLevel + (
					     760.0 *
					     ((i-firstBuildY)*percStep)
					     );
	      if(thisGreenLevel > 511)
		thisGreenLevel = (thisGreenLevel-511);
	      else if(thisGreenLevel > 255)
		thisGreenLevel = 255-(thisGreenLevel-255);
	      thisGreenLevel = MAX(MIN(255,thisGreenLevel),0);
	      thisBlueLevel = blueLevel + (255.0 -
					   (382.0 *
					   ((i-firstBuildY)*percStep))
					   );
	      if(thisBlueLevel < 0)
		thisBlueLevel = 0 - (2*thisBlueLevel);
	      thisBlueLevel = MIN(255,thisBlueLevel);

	      if(hiColor)
		{/* hi */
		  color_defs[i].pixel = i;
		  color_defs[i].red   = 
		    ((int)(thisRedLevel) * pixelColor);
		  color_defs[i].green =
		    ((int)(thisGreenLevel) * pixelColor);
		  color_defs[i].blue  = 
		    ((int)(thisBlueLevel) * pixelColor);
		  color_defs[i].flags = 
		    DoRed | DoGreen | DoBlue;
		}/* hi */
	      else
		{/* lo */
		  color_defs[pixels_return[i]].pixel = pixels_return[i];
		  color_defs[pixels_return[i]].red   = 
		    ((int)(thisRedLevel) * pixelColor);
		  color_defs[pixels_return[i]].green =
		    ((int)(thisGreenLevel) * pixelColor);
		  color_defs[pixels_return[i]].blue  = 
		    ((int)(thisBlueLevel) * pixelColor);
		  color_defs[pixels_return[i]].flags = 
		    DoRed | DoGreen | DoBlue;
		}/* lo */
	      if(masterProcess)
		{
		  if(hiColor)
		    {XAllocColor (dsply, color_map, &color_defs[i]);
		     pixels_return[i] = color_defs[i].pixel;}
		  else
		    XStoreColor (dsply, color_map, 
				 &color_defs[pixels_return[i]]);
		}
	    }/* Do the colors we have been assigned */
	}/* RAinbowING */
      else
	{/* error */
	  askThem("Can't Ramp and Rainbow at the same time.",
		  FALSE,XmDIALOG_INFORMATION);
	  goto noRamp;
	}/* error */
     
      /* Redraw the histogram */
      histogram_activate(&minP, &maxP);

    }/* Only do if we're building */

 noRamp:;

}/* colormap_ramp_ACTIVATE */




/* ************************************************************
 * colormap_scroll_activate() - They want to "scroll" the 
 * colormap replacing the highest entry with the next to the
 * highest all the way down and then filling in the lowest entry 
 * with the one taken from the highest position.
 */
void colormap_scroll_activate(Widget widget, int *tag, 
			      XmAnyCallbackStruct *callback_data)
{/* COLORMAP_SCROLL_ACTIVATE */

  if(*tag==XmARROW_DOWN)
    doTheCmScroll(widget, -1);
  else doTheCmScroll(widget, 1);

}/* COLORMAP_SCROLL_ACTIVATE */




/* ************************************************************
 * paramFileGenerate_activate() - Put up the interface for doing 
 * paramFileGeneration.  This gives a default param file they
 * can use for bootstrapping their reconstruction.
 */
void paramFileGenerate_activate(Widget widget, int *tag, 
				XmAnyCallbackStruct *callback_data)
{/* paramFileGenerate_ACTIVATE */
  char *s1,*thisText;
  char outParamFile[1024],tmpParamFile[1024];
  int saveHPF, units, volts;
  float pixelSize,ampFac,delfMaj,delfMin,angMaj,csCoef;

  sprintf(tmpParamFile,"%s",fileName.addr);
  s1 = strrchr(tmpParamFile,'.');
  if(!s1)
    s1 = tmpParamFile+ (strlen(fileName.addr)-1);
  *s1 = '\0';

  sprintf(outParamFile,"%s.dat_001",tmpParamFile);
  s1 = strchr(fileName.addr,'/');
  if(s1)
    sprintf(paramBoxesFilename,"%s",fileName.addr);
  else
    sprintf(paramBoxesFilename,"%s/%s",getenv("PWD"),fileName.addr);


  thisText = XmTextGetString(ctfPixSizeW); 
  pixelSize = (float)strtod(thisText,0);
  XtFree(thisText);
  units = 1;  /* Angstroms */
  thisText = XmTextGetString(ctfKvW); 
  volts = (int)strtod(thisText,0)*1000;
  XtFree(thisText);
  thisText = XmTextGetString(ctfAmpFactorW); 
  ampFac = (float)strtod(thisText,0);
  XtFree(thisText);
  thisText = XmTextGetString(ctfMajorDefocusW); 
  delfMaj = (float)strtod(thisText,0);
  XtFree(thisText);
  thisText = XmTextGetString(ctfMinorDefocusW); 
  delfMin = (float)strtod(thisText,0);
  XtFree(thisText);
  thisText = XmTextGetString(ctfAngMajorW); 
  angMaj = (float)strtod(thisText,0);
  XtFree(thisText);
  thisText = XmTextGetString(ctfSpAberW); 
  csCoef = (float)strtod(thisText,0);
  XtFree(thisText);

  /* Stuff appropriate widgets */
  sprintf(outStr,"%f",pixelSize*10.0); /* Let's write it in Angstroms */
  XmTextSetString(pixsizeW,outStr);
  sprintf(outStr,"%d",units);
  XmTextSetString(unitsW,outStr);
  sprintf(outStr,"%f",(float)volts);
  XmTextSetString(voltsW,outStr);
  sprintf(outStr,"%f",ampFac);
  XmTextSetString(ampfacsW,outStr);
  sprintf(outStr,"%f",delfMaj);
  XmTextSetString(delfMajW,outStr);
  sprintf(outStr,"%f",delfMin);
  XmTextSetString(delfMinW,outStr);
  sprintf(outStr,"%f",angMaj);
  XmTextSetString(angMajW,outStr);
  sprintf(outStr,"%f",csCoef);
  XmTextSetString(csCoefW,outStr);

  saveHPF = haveParamFile;
  haveParamFile = TRUE;
  write_param(outParamFile);
  haveParamFile = saveHPF;


}/* paramFileGenerate_ACTIVATE */





/* ************************************************************
 * paramFileRead_activate() - Read a param file
 */
void paramFileRead_activate(Widget widget, int *tag, 
			    XmAnyCallbackStruct *callback_data)
{/* PARAMFILEREAD_ACTIVATE */

  readWrite= READPARAM;
  filterFiles("*.dat*",READPARAM); 
  raiseIt(fileSel_widget);

}/* PARAMFILEREAD_ACTIVATE */




/* ************************************************************
 * paramFileRead_work() - They've chosen a param file now read
 * it from disk.
 */
void paramFileRead_work()
{/* PARAMFILEREAD_WORK */
  char *s1,*thisText;
  char inStr[1024],tmpParamFile[1024];
  int saveHPF, units, volts;
  float pixelSize,ampFac,delfMaj,delfMin,angMaj,csCoef;
  FILE *fIn;

  sprintf(tmpParamFile,gottenParamFname);
  fIn = fopen(tmpParamFile, "r");

  if(fIn)
    {/* read */
      get_str(fIn, '\n', inStr); /* first line */

      get_str(fIn, ',', inStr); 
      XmTextSetString(ctfPixSizeW,inStr); 
      XmTextSetString(pixsizeW,inStr);

      get_str(fIn, ',', inStr); 
      XmTextSetString(unitsW,inStr);

      get_str(fIn, ',', inStr); 
      XmTextSetString(ctfKvW,inStr); 
      XmTextSetString(voltsW,inStr);

      get_str(fIn, ',', inStr); 
      XmTextSetString(ctfAmpFactorW,inStr); 
      XmTextSetString(ampfacsW,inStr);

      get_str(fIn, ',', inStr); 
      XmTextSetString(ctfMajorDefocusW,inStr); 
      XmTextSetString(delfMajW,inStr);

      get_str(fIn, ',', inStr); 
      XmTextSetString(ctfMinorDefocusW,inStr); 
      XmTextSetString(delfMinW,inStr);

      get_str(fIn, ',', inStr); 
      XmTextSetString(ctfAngMajorW,inStr); 
      XmTextSetString(angMajW,inStr);

      get_str(fIn, '\n', inStr); 
      XmTextSetString(ctfSpAberW,inStr); 
      XmTextSetString(csCoefW,inStr);
      
      fclose(fIn);
    }/* read */
  else
    {/* error */
      sprintf("Cannot find file [%s].",tmpParamFile);
      askThem(outStr,FALSE,XmDIALOG_INFORMATION);
    }/* error */

  XFlush(dsply);
  DrawFFTWindow();


}/* PARAMFILEREAD_WORK */





/* ************************************************************
 * auto2DDefocusRefine_activate() - Put up the interface for doing 
 * auto2DDefocusRefines.
 */
void auto2DDefocusRefine_activate(Widget widget, int *tag, 
				  XmAnyCallbackStruct *callback_data)
{/* auto2DDefocusRefine_ACTIVATE */
  float thisHiRes,pixelSize;
  char *ctfText;

  if(ctfMajorDefocusValue != 0.0 &&
     ctfMinorDefocusValue != 0.0 )
    {/* OK to proceed */

      sprintf(outStr,"%f",ctfMajorDefocusValue);
      XmTextSetString(a2ddrfjW,outStr);
      sprintf(outStr,"%f",ctfMinorDefocusValue);
      XmTextSetString(a2ddrfnW,outStr);
      sprintf(outStr,"%f",angMajor);
      XmTextSetString(a2ddragW,outStr);

      ctfText = XmTextGetString(ctfPixSizeW); 
      pixelSize   = (float)strtod(ctfText, 0);
      XtFree(ctfText);
      thisHiRes = pixelSize * 2.0 ;
      thisHiRes *= 10.0;   /* Put it in Angstroms */
      sprintf(outStr,"%f",thisHiRes);
      XmTextSetString(a2ddrtfW,outStr);

      if(fftasum)
	free(fftasum);
      if(poly_paras)
	free(poly_paras);
      if(a2dr_s)
	free(a2dr_s);
      if(a2dr_sminusb)
	free(a2dr_sminusb);
      if(a2dr_ctfsq)
	free(a2dr_ctfsq);
      if(a2dr_d)
	free(a2dr_d);
      if(a2dr_b)
	free(a2dr_b);
      fftasum      = (double *)malloc(sizeof(double)*(1024*2048));
      poly_paras   = (double *)malloc(sizeof(double)*(1024));
      a2dr_s       = (float *)malloc(sizeof(float)*(1024));
      a2dr_sminusb = (float *)malloc(sizeof(float)*(1024));
      a2dr_ctfsq   = (float *)malloc(sizeof(float)*(1024));
      a2dr_d       = (float *)malloc(sizeof(float)*(1024));
      a2dr_b       = (float *)malloc(sizeof(float)*(1024));

      raiseIt(auto2DDRScreenWidget); 
      doingAuto2DDefocusRefine= TRUE;
    }/* OK to proceed */
  else
    {/* Get them to estimate it first */
      askThem("You must estimate the defocus first.",
	      FALSE,XmDIALOG_INFORMATION);
      doingAuto2DDefocusRefine= FALSE;
    }/* Get them to estimate it first */

}/* auto2DDefocusRefine_ACTIVATE */



/*************************************************************
 * auto2DDR_work() -  Do what's nedded for auto 2d Defocus 
 * refinement.
*/
void auto2DDR_work()
{/* AUTO2DDR_WORK */
  int ncol,nrow,nimg,i,j,k,x,y;
  float pixelSize, volts, tFac, defMajor, defMinor, angle,
    ampContrast, Cs, res_hi, chisq;
  char *ctfText;
  XmString simpleString;
  float minF, maxF, scaleF;

  /* Res_hi  */
  ctfText = XmTextGetString(a2ddrtfW); 
  res_hi = (float)strtod(ctfText, 0);
  XtFree(ctfText);

  if(res_hi <= 0.0)
    {/* nogo */
      askThem("You must enter a value for Highest Resolution.",
	      FALSE,XmDIALOG_INFORMATION);
      goto noCanDo;
    }/* nogo */


  ncol = purdueHeader.numCols;
  nrow = purdueHeader.numRows;
  nimg = purdueHeader.nimg;

  ctfText = XmTextGetString(ctfPixSizeW); 
  pixelSize   = (float)strtod(ctfText, 0);
  XtFree(ctfText);
  pixelSize *= 10.0;     /* Convert to Angs */
  printf("pixelSize = [%f]\n",pixelSize);
  ctfText = XmTextGetString(ctfKvW); 
  volts   = (int)strtod(ctfText, 0);
  volts *= 1000.0;    /* Convert from kiloVolts to volts */
  XtFree(ctfText);
  ctfText = XmTextGetString(ctfSpAberW); 
  Cs   = (float)strtod(ctfText, 0);
  XtFree(ctfText);
  /* DONT DO THIS??? Cs *= 10000000.0;  /+ Put into Angstroms */
  ctfText = XmTextGetString(ctfAmpFactorW); 
  ampContrast = (float)strtod(ctfText, 0);
  XtFree(ctfText);

  ctfText = XmTextGetString(ctfAngMajorW); 
  angle = (float)strtod(ctfText, 0);
  /* We need to change the sign of angMajor here because our
   * XWindows based buffer is upside down.  When we write
   * to disk we need to use the sign just as given from
   * the interface */
  angle = 0.0 - angle ;
  XtFree(ctfText);
  printf("OLD angle [%f]\n",angle);

  ctfText = XmTextGetString(a2ddrfjW);
  defMajor = (float)strtod(ctfText,0);
  XtFree(ctfText);
  ctfText = XmTextGetString(a2ddrfnW);
  defMinor = (float)strtod(ctfText,0);
  XtFree(ctfText);
  ctfText = XmTextGetString(a2ddragW);
  angle = (float)strtod(ctfText,0);
  XtFree(ctfText);
  /* We need to change the sign of angMajor here because our
   * XWindows based buffer is upside down.  When we write
   * to disk we need to use the sign just as given from
   * the interface 
   DO WE REALLY NEED TO DO THIS ????
   angle = 0.0 - angle ;
  */
  printf("defmaj[%f] old[%f]\n",defMajor,ctfMajorDefocusValue);
  printf("defmin[%f] old[%f]\n",defMinor,ctfMinorDefocusValue);
  printf("NEW angle [%f]\n",angle);

  sprintf(outStr,"Calculating ...");
  XmTextSetString(a2ddrfjW,outStr);
  sprintf(outStr,"Calculating ...");
  XmTextSetString(a2ddrfnW,outStr);
  sprintf(outStr,"Calculating ...");
  XmTextSetString(a2ddragW,outStr);
  sprintf(outStr,"Calculating ...");
  XmTextSetString(a2ddrcsqW,outStr);
  XFlush(dsply);


  /* Yongchang floats all the data after it's read in */
  CK(allPackImg = (float *)malloc(sizeof(float)*nimg*ncol*nrow));
  for(i=0;i<nimg;i++)
    {
      float *thisP,backG;
      int reverseIt;
      thisP = packImg[i];
      backG = *thisP;
      for(j=0;j<(nrow);j++)
	{
	  /* I'm suspicious that Yongchang also doesn't do the TSB "flip"
	     in the Y direction ... ??? */
	  reverseIt = j;
	  reverseIt = nrow - j -1;
	  for(k=0;k<(ncol);k++)
	    {
	      allPackImg[((i*ncol*nrow)+(j*ncol)+k)] = 
		(thisP[(reverseIt*ncol)+k] - backG);
	    }
	}
    }

  /* For debugging ******************
  pixelSize = 2.95;
  volts = 300000.0;
  tFac = 0.0;
  defMajor = 2.0;
  defMinor = 2.0;
  angle = 0.0;
  ampContrast = 0.05;
  Cs = 2.0;
  res_hi = 20.0;
  *****************************/

  printf ("pixelSize= [%f]\n",pixelSize);
  printf ("volts= [%f]\n",volts);
  printf ("tFac= [%f]\n",tFac);
  printf ("defMajor= [%f]\n",defMajor);
  printf ("defMinor= [%f]\n",defMinor);
  printf ("angle= [%f]\n",angle);
  printf ("ampContrast= [%f]\n",ampContrast);
  printf ("Cs= [%f]\n",Cs);
  printf ("res_hi= [%f]\n",res_hi);

  if(ctf2dSaveFftasum == NULL)
    CK(ctf2dSaveFftasum = (float *)malloc(sizeof(float)*fftSize*fftSize*2));
  if(ctf2dSaveFftasumsq == NULL)
    CK(ctf2dSaveFftasumsq = (float *)malloc(sizeof(float)*fftSize*fftSize*2));

  yji_ctf2dfit_(poly_paras, &ncol, &nrow, &nimg, &fftSize, &pixelSize,
		&volts, &res_hi, &defMajor, &defMinor, &angle,
		&ampContrast, &Cs,
		allPackImg, &chisq, auto2DdrawBuf, auto2DdataBuf,
		fftasum, &hkl_max, &order_b, &order_d,
		&ctf2dCalcIt,
		ctf2dSaveFftasum, ctf2dSaveFftasumsq
		);
  ctf2dCalcIt = FALSE;  /* We just did */

  free(allPackImg);
  allPackImg = NULL;

  /* Scale buffer for display */
  minF = maxF = auto2DdrawBuf[0];
  for(i=0;i<(256*512);i++)
    {/* stats */
      maxF = MAX(auto2DdrawBuf[i],maxF);
      minF = MIN(auto2DdrawBuf[i],minF);
    }/* stats */
  scaleF = 200.0 - (maxF-minF);
  for(y = 0 ; y < 512 ;y ++)
    {/* Do all y */
      for(x = 0 ; x < 256 ;x ++)
	{/* Do all x */
	  auto2DdrawBuf[((y*256)+x)] = 10.0 +
	    (float)((auto2DdrawBuf[((y*256)+x)] - minF) * scaleF);
	}/* Do all x */
    }/* Do all y */

  printf("New vals - defmajor[%f] defminor[%f] angle[%f] chisq[%f] \n",
	 defMajor, defMinor, angle, chisq);

  /* Put it in the first 180 degrees */
  if(angle > 180.0)
    angle = angle - 180.0;

  sprintf(outStr,"%f",defMajor);
  XmTextSetString(a2ddrfjW,outStr);
  sprintf(outStr,"%f",defMinor);
  XmTextSetString(a2ddrfnW,outStr);
  sprintf(outStr,"%f",angle);
  XmTextSetString(a2ddragW,outStr);
  /* This is an input now ...
  sprintf(outStr,"%f",res_hi);
  XmTextSetString(a2ddrtfW,outStr);
  */
  sprintf(outStr,"%f",chisq);
  XmTextSetString(a2ddrcsqW,outStr);
  XFlush(dsply);

  auto2Dangle    = angle;
  auto2DdefMajor = defMajor;
  auto2DdefMinor = defMinor;


 noCanDo:;

}/* AUTO2DDR_WORK */


/* ************************************************************
 * auto2DDR_activate() - They want to do 2d defocus estimate.
 */
void auto2DDR_activate(Widget widget, int *tag, 
		       XmAnyCallbackStruct *callback_data)
{/* auto2DDR_ACTIVATE */

  auto2DDR_work();

}/* auto2DDR_ACTIVATE */



/* ************************************************************
 * auto2DDR_deactivate() - Free up memory and whatever else
 * is needed after copmpleting 2d defocus estimate.
 */
void auto2DDR_deactivate(Widget widget, int *tag, 
			 XmAnyCallbackStruct *callback_data)
{/* AUTO2DDR_DEACTIVATE */
  
  if(fftasum)
    {free(fftasum);
    fftasum = NULL;}
  if(poly_paras)
    {free(poly_paras);
    poly_paras = NULL;}
  if(a2dr_s)
    {free(a2dr_s);
    a2dr_s = NULL;}
  if(a2dr_sminusb)
    {free(a2dr_sminusb);
    a2dr_sminusb = NULL;}
  if(a2dr_ctfsq)
    {free(a2dr_ctfsq);
    a2dr_ctfsq = NULL;}
  if(a2dr_d)
    {free(a2dr_d);
    a2dr_d = NULL;}
  if(a2dr_b)
    {free(a2dr_b);
    a2dr_b = NULL;}
  doingAuto2DDefocusRefine= FALSE;
  XtUnmanageChild(auto2DDRScreenWidget); 

}/* AUTO2DDR_DEACTIVATE */



/* ************************************************************
 * auto2DDR_curves_activate() - What curve(s) do they want drawn
 * as feedback.
 */
void auto2DDR_curves_activate(Widget widget, int *tag, 
			      XmToggleButtonCallbackStruct *callback_data)
{/* AUTO2DDR_CURVES_ACTIVATE */

  if(*tag==A2DR_S)
    a2drSDraw = !a2drSDraw ;
  else if(*tag==A2DR_SMINUSB)
    a2drSminusBDraw = !a2drSminusBDraw ;
  else if(*tag==A2DR_CTFSQ)
    a2drCtfsqDraw = !a2drCtfsqDraw ;
  else if(*tag==A2DR_D)
    a2drDDraw = !a2drDDraw ;
  else if(*tag==A2DR_B)
    a2drBDraw = !a2drBDraw ;


  DrawFFTWindow();

}/* AUTO2DDR_CURVES_ACTIVATE */




/* ************************************************************
 * auto2DDR_ctflines_activate() - Toggle drawing the CTF lines.
 */
void auto2DDR_ctflines_activate(Widget widget, int *tag, 
				XmToggleButtonCallbackStruct *callback_data)
{/* AUTO2DDR_CTFLINES_ACTIVATE */

  drawAuto2Dctflines = !drawAuto2Dctflines;

  if(drawAuto2Dctflines)
    drawCTF = TRUE; 
  else
    drawCTF = FALSE; 

  DrawFFTWindow();

}/* AUTO2DDR_CTFLINES_ACTIVATE */





/* ************************************************************
 * auto2DDR_scurve_activate() - toggle drawing scurve
 */
void auto2DDR_scurve_activate(Widget widget, int *tag, 
			      XmToggleButtonCallbackStruct *callback_data)
{/* AUTO2DDR_SCURVE_ACTIVATE */

  drawAuto2Dscurve = !drawAuto2Dscurve;

  if(drawAuto2Dscurve)
    {
      XtSetArg(arglst[0], XmNset, FALSE);
      XtSetValues(pickCTFtoggleW, arglst, 1); 
      pickLattice = -99;  
      drawCTF = FALSE; 
    }

  DrawFFTWindow();

}/* AUTO2DDR_SCURVE_ACTIVATE */






/* ************************************************************
 * auto2DDR_drawmodified_activate() - toggle drawing modified 
 * display.
 */
void auto2DDR_drawmodified_activate(Widget widget, int *tag, 
				   XmToggleButtonCallbackStruct *callback_data)
{/* AUTO2DDR_DRAWMODIFIED_ACTIVATE */

  drawAuto2Dctf = !drawAuto2Dctf;

  DrawFFTWindow();

}/* AUTO2DDR_DRAWMODIFIED_ACTIVATE */






/* ************************************************************
 * auto2DDR_update_activate() - Update the display text widgets
 * and draw FFT screen.
 */
void auto2DDR_update_activate(Widget widget, int *tag, 
			      XmAnyCallbackStruct *callback_data)
{/* AUTO2DDR_UPDATE_ACTIVATE */
  char *thisText;

  thisText = XmTextGetString(a2ddrfjW);
  XmTextSetString(ctffocusMajValueW, thisText);
  XtFree(thisText);
  thisText = XmTextGetString(a2ddrfnW);
  XmTextSetString(ctfFocusMinvalueW, thisText);
  XtFree(thisText);

  thisText = XmTextGetString(a2ddragW);
  XmTextSetString(ctfAngMajorValueW, thisText);
  XmTextSetString(ctfAngMajorW, thisText); 
  XtFree(thisText);
  thisText = XmTextGetString(a2ddrtfW);
  XmTextSetString(ctfTFactorW, thisText);
  XtFree(thisText);

  XFlush(dsply);

  DrawFFTWindow();

}/* AUTO2DDR_UPDATE_ACTIVATE */





/* ************************************************************
 * autoDefocusRefine_activate() - Put up the interface for doing 
 * autoDefocusRefines.
 */
void autoDefocusRefine_activate(Widget widget, int *tag, 
				XmAnyCallbackStruct *callback_data)
{/* autoDefocusRefine_ACTIVATE */
  float thisHiRes,pixelSize;
  char *ctfText;

  if(ctfMajorDefocusValue != 0.0 &&
     ctfMinorDefocusValue != 0.0 )
    {/* OK to proceed */
      int ncol,nrow,nimg;

      ncol = purdueHeader.numCols;
      nrow = purdueHeader.numRows;
      nimg = purdueHeader.nimg;
      CK(allPackImg = (float *)malloc(sizeof(float)*nimg*ncol*nrow));

      ctfText = XmTextGetString(ctfPixSizeW); 
      pixelSize   = (float)strtod(ctfText, 0);
      XtFree(ctfText);
      thisHiRes = pixelSize * 2.0 ;
      thisHiRes *= 10.0;   /* Put it in Angstroms */
      sprintf(outStr,"%f",thisHiRes);
      XmTextSetString(adrHiResW,outStr);
      raiseIt(autoDRScreenWidget); 
      doingAutoDefocusRefine= TRUE;
    }/* OK to proceed */
  else
    {/* Get them to estimate it first */
      askThem("You must estimate the defocus first.",
	      FALSE,XmDIALOG_INFORMATION);
      doingAutoDefocusRefine= FALSE;
    }/* Get them to estimate it first */

}/* autoDefocusRefine_ACTIVATE */




/*************************************************************
 * auto1DDR_work() -  Do what's necessary to setup for 1D
 * defocus estimation .
*/
void auto1DDR_work(int orderBkg, int orderSig,
		   float *calcDefocus, float *calcChisq)
{/* AUTO1DDR_WORK */
  int ncol,nrow,nimg,i,j,k,x,y;
  float pixelSize, volts, tFac, defMajor,
    ampContrast, Cs, res_hi, chisq;
  char *ctfText;
  float *a;
  XmString simpleString;
  float minF, maxF, scaleF;
  int bkg_poly_order, decay_poly_order, ma;
  char *thisStr;

  /* Res_hi  */
  ctfText = XmTextGetString(a2ddrtfW); 
  res_hi = (float)strtod(ctfText, 0);
  XtFree(ctfText);

  if(res_hi <= 0.0)
    {/* nogo */
      askThem("You must enter a value for Highest Resolution.",
	      FALSE,XmDIALOG_INFORMATION);
      goto noCanDo;
    }/* nogo */


  ncol = purdueHeader.numCols;
  nrow = purdueHeader.numRows;
  nimg = purdueHeader.nimg;

  ctfText = XmTextGetString(adrHiResW);
  res_hi = (float)strtod(ctfText,0);
  XtFree(ctfText);

  ctfText = XmTextGetString(ctfPixSizeW); 
  pixelSize   = (float)strtod(ctfText, 0);
  XtFree(ctfText);
  pixelSize *= 10.0;     /* Convert to Angs */
  printf("pixelSize = [%f]\n",pixelSize);
  ctfText = XmTextGetString(ctfKvW); 
  volts   = (int)strtod(ctfText, 0);
  volts *= 1000.0;    /* Convert from kiloVolts to volts */
  XtFree(ctfText);
  ctfText = XmTextGetString(ctfSpAberW); 
  Cs   = (float)strtod(ctfText, 0);
  XtFree(ctfText);
  /* DONT DO THIS??? Cs *= 10000000.0;  /+ Put into Angstroms */
  ctfText = XmTextGetString(ctfAmpFactorW); 
  ampContrast = (float)strtod(ctfText, 0);
  XtFree(ctfText);

   
  bkg_poly_order   = orderBkg;
  decay_poly_order = orderSig;

  bkg_poly_order++; decay_poly_order++;	
  ma = bkg_poly_order+decay_poly_order+1;
  printf("bpo = %d  dpo = %d  ma = %d\n",
	 bkg_poly_order  ,
	 decay_poly_order ,
	 ma
	 );
  a = (float *)malloc(sizeof(float)*ma+5);
  for (i=1; i<=(bkg_poly_order+decay_poly_order); i++)  
    *(a+i) = 0.0;
  *(a) = 1.0;
  *(a+bkg_poly_order)=1.0;
  
  /* Let's get these from interface */
  /* a */
  thisStr = XmTextGetString(adra0W);
  a[1]    = (float)strtod(thisStr,0);
  XtFree(thisStr);
  thisStr = XmTextGetString(adra1W);
  a[2]    = (float)strtod(thisStr,0);
  XtFree(thisStr);
  thisStr = XmTextGetString(adra2W);
  a[3]    = (float)strtod(thisStr,0);
  XtFree(thisStr);
  thisStr = XmTextGetString(adra3W);
  a[4]    = (float)strtod(thisStr,0);
  XtFree(thisStr);
  thisStr = XmTextGetString(adra4W);
  a[5]    = (float)strtod(thisStr,0);
  XtFree(thisStr);
  thisStr = XmTextGetString(adra5W);
  a[6]    = (float)strtod(thisStr,0);
  XtFree(thisStr);
  /* b */
  thisStr = XmTextGetString(adrb0W);
  a[8]    = (float)strtod(thisStr,0);
  XtFree(thisStr);
  thisStr = XmTextGetString(adrb1W);
  a[9]    = (float)strtod(thisStr,0);
  XtFree(thisStr);
  thisStr = XmTextGetString(adrb2W);
  a[10]    = (float)strtod(thisStr,0);
  XtFree(thisStr);
  thisStr = XmTextGetString(adrb3W);
  a[11]    = (float)strtod(thisStr,0);
  XtFree(thisStr);
  thisStr = XmTextGetString(adrb4W);
  a[12]    = (float)strtod(thisStr,0);
  XtFree(thisStr);
  thisStr = XmTextGetString(adrb5W);
  a[13]    = (float)strtod(thisStr,0);
  XtFree(thisStr);
  
  /* Yongchang floats all the data after it's read in */
  for(i=0;i<nimg;i++)
    {
      float *thisP,backG;
      int reverseIt;
      thisP = packImg[i];
      backG = *thisP;
      for(j=0;j<(nrow);j++)
	{
	  /* I'm suspicious that Yongchang also doesn't do the TSB "flip"
	     in the Y direction ... ??? */
	  reverseIt = j;
	  reverseIt = nrow - j -1;
	  for(k=0;k<(ncol);k++)
	    {
	      allPackImg[((i*ncol*nrow)+(j*ncol)+k)] = 
		(thisP[(reverseIt*ncol)+k] - backG);
	    }
	}
    }

  /* Put in proper units */
  Cs *= 10000000.0; 
  defMajor =   ctfMajorDefocusValue * 10000.0;  /* Put it in Angstroms */

  /* For debugging ******************
  pixelSize = 3.57;
  volts = 200000.0;
  defMajor = 18832.0;
  ampContrast = 0.07;
  Cs = 20000000.0;
  res_hi = 7.14;
  *****************************/

  printf ("pixelSize= [%f]\n",pixelSize);
  printf ("volts= [%f]\n",volts);
  printf ("defMajor= [%f]\n",defMajor);
  printf ("ampContrast= [%f]\n",ampContrast);
  printf ("Cs= [%f]\n",Cs);
  printf ("res_hi= [%f]\n",res_hi);

  /* Init out all arrays */
  for(i=0;i<256;i++)
    {
      adr_s[i]       = 0.0;
      adr_sminusb[i] = 0.0;
      adr_ctfsq[i]   = 0.0;
      adr_d[i]       = 0.0;
      adr_b[i]       = 0.0;
    }

  ca1d_save = (float *)malloc(sizeof(float)*1024);
  if(ctf1dSaveFftAvg == NULL)
    CK(ctf1dSaveFftAvg = (float *)malloc(sizeof(float)*fftSize*fftSize*2));
  yji_ctf1dfit_(&ncol, &nrow, &nimg, &fftSize, &pixelSize,
		&volts, &res_hi, &defMajor, 
		&ampContrast, &Cs, allPackImg, &chisq, 
		a,
		adr_s,
		adr_sminusb,
		adr_ctfsq,
		adr_d,
		adr_b,
		ca1d_save,
		&ctf1dCalcIt,
		ctf1dSaveFftAvg
		);
  ctf1dCalcIt = FALSE;  /* We just did */

  for (i=0; i<256; i++) 
    ca1d[i] = ca1d_save[i];
  free(ca1d_save);

  

  printf("New 1D vals - defmajor[%f] chisq[%f] \n", defMajor, chisq);
  *calcDefocus = defMajor*10000.0; /* We divide it later ... */
  *calcChisq   = chisq;

  /* Now put new array values on interface */
  /* a */
  sprintf(outStr,"%12E",a[0]);
  XmTextSetString(adra0W,outStr);
  sprintf(outStr,"%12E",a[1]);
  XmTextSetString(adra1W,outStr);
  sprintf(outStr,"%12E",a[2]);
  XmTextSetString(adra2W,outStr);
  sprintf(outStr,"%12E",a[3]);
  XmTextSetString(adra3W,outStr);
  sprintf(outStr,"%12E",a[4]);
  XmTextSetString(adra4W,outStr);
  sprintf(outStr,"%12E",a[5]);
  XmTextSetString(adra5W,outStr);
  /* b */
  sprintf(outStr,"%12E",a[6]);
  XmTextSetString(adrb0W,outStr);
  sprintf(outStr,"%12E",a[7]);
  XmTextSetString(adrb1W,outStr);
  sprintf(outStr,"%12E",a[8]);
  XmTextSetString(adrb2W,outStr);
  sprintf(outStr,"%12E",a[9]);
  XmTextSetString(adrb3W,outStr);
  sprintf(outStr,"%12E",a[10]);
  XmTextSetString(adrb4W,outStr);
  sprintf(outStr,"%12E",a[11]);
  XmTextSetString(adrb5W,outStr);


 noCanDo:;

}/* AUTO1DDR_WORK */




/*************************************************************
 * autoDR_work() -  The following rtn is modified from a program
 * called ctffit_poly.c written by Wei Zhang:
 ***********************************************************************
 * File ctffit_poly.c: Fit polynomial and CTF functions to the circular 
 *	averaged 1D curve of a micrograph's power spectrum (CA1D): 
 *
 * 	S = B + CTF*CTF * D
 *	where, S is the circular averaged intensity of a micrograph's
 *		power spectrum;
 *	       B is the background noise;
 *	       D is the signal strengh decay;
 *	       B and D are modeled by polynomial functions;
 *
 * The input file: It is the CA1D ascii format file given by the program
 *	babe2:[wz.files.for.misc]empiffftavg.exe
 *	from the second line, the first collumn is the index number of 
 *	annulii, the second collumn is the scaled averaged intensity,
 *	the third collumn is the scaled standard division when CA1D
 *	averaging is calculated. The first line of the file is highest
 *	annulii number, ie. half of the FFT size.
 *
 * This program fits the model to the CA1D curve at the range between the
 *	first CTF zero to a resolution that a user defines. This Levenberg
 *	-Marquard method vary smoothly between the inverse-Hessian method
 *	and the steepest descent method (Numerical Recipes fo C, P683).
 *	It does least square fitting for non-linear functions. 
 *
 * The following subroutines are modified from Numerical Recipes:
 *	gaussj.c  mrqcof.c	mrqmin.c	nrutil.c	covsrt.c
 *
 * Wei Zhang, 2-2-2000, last modified at 3-24-2000 
 *    modified by W. Zhang 9-25-00 output file empft.dat_000
 ************************************************************************
 * Future goal - see if this rtn is worhtwhile for implementation.
*/
void autoDR_work(int orderBkg, int orderSig,
		 float *calcDefocus, float *calcChisq)
{/* AUTODR_WORK */
#define DEBUG1 0
#define DEBUG2 0
//  long tmp1, LOOP;
//  int ndata, *ia, ma, i, j, k, array_size, tmp2;
//  int bkg_poly_order, decay_poly_order, annlow,  annhigh;
//  float   tmp3, tmp4, big, small, wavelength, A, Cs, ctf, ctf2 ;
//  float volts;
//  double  *x, *y, *sig, *a, **covar, **alpha, *chisq, *alamda;
//  double  *x_original, *y_original, *std_original, tmp5, tmp6;
//  double  reslow, reshigh, pixel_size, fw, fw2, theta, pie, B, D;
//  double  chisqold1, chisqold2, defcusold, fftavgvalue, defocus;
//  char *InFilename, *yes_no, *Outfile1, *Outfile2, *Outfile3; 
//  char *Outfile4, *Imgfile, *thisText;
//  FILE *fp1;
//  char *thisStr;
//  
//  /* Constant variables */	
//  
//  big = 100000000.0;
//  small = 0.000000001;
//  pie = 3.14159265358979323;
//  wavelength = 0.0250921;	/* Electron wavelength for 200keV    */
//  A = 0.05;			/* Persontage of Amplitude Contrast */
//  Cs = 20000000.0;		/* Spherical aberration constant */
//	
//
//  thisText = XmTextGetString(ctfKvW); 
//  volts   = (int)strtod(thisText, 0);
//  volts *= 1000.0;    /* Convert from kiloVolts to volts */
//  XtFree(thisText);
//  wavelength = wavel_(&volts);
//  wavelength *= 10.0;  /* Put into Angstroms */
//
//printf ("wave= [%f]\n",wavelength);
//
//  thisText = XmTextGetString(ctfSpAberW); 
//  Cs   = (float)strtod(thisText, 0);
//  XtFree(thisText);
//  Cs *= 10000000.0;  /* Put into Angstroms */
//printf ("Cs= [%f]\n",Cs);
//
//  thisText = XmTextGetString(ctfAmpFactorW); 
//  A = (float)strtod(thisText, 0);
//  XtFree(thisText);
//printf ("AmpFac= [%f]\n",A);
//
//  if(riverSwitch)
//    fp1= fopen("a_river.txt","w"); 
//  
//  array_size = 256;
//  tmp3 = ca1d_avg;
//  /* fscanf(fp1, "%d %f", &array_size, &tmp3); 
//  fscanf(fp1, "%d %f", &array_size, &tmp3); 
//  fprintf(fp1, "%d %f\n", array_size, tmp3);   */
//  fftavgvalue = (double)tmp3;
//  
//  x = (double *)dvector(1,array_size );
//  y = (double *)dvector(1,array_size );
//  sig = (double *)dvector(1, array_size);
//  
//  x_original = dvector(1,array_size );
//  y_original = dvector(1,array_size );
//  std_original = dvector(1, array_size);
//  
//  for(i=1;i<=array_size; i++) 
//    { 
//      tmp2 = i;
//      tmp3 = ca1d[i-1] / ca1d_avg;
//      tmp4 = ca1d_std[i-1]  / ca1d_avg;
//      /* fscanf(fp1, "%d %f %f ", &tmp2,&tmp3, &tmp4); 
//      fscanf(fp1, "%d %f %f ", &tmp2,&tmp3, &tmp4); 
//      fprintf(fp1, "%d %f %f\n", tmp2,tmp3, tmp4); */
//      *(x_original+i) = (double)tmp2;
//      *(y_original+i) =(double)tmp3;
//      *(std_original+i) = (double)tmp4;	   
//
//      /* print  tmp3 and tmp4 for river */
//      if(riverSwitch)
//	fprintf(fp1,"%d %E %E\n",tmp2,tmp3,tmp4);
//
//
//    }
//
//  if(riverSwitch)
//    {fclose(fp1); 
//     printf("\n\nValues written to a_river.txt file.\n\n");
//   }
//
//  if(DEBUG1)      
//    for(i=1;i<=array_size; i++)
//      printf ("x,y,sig: %e, %e, %e\n",
//	      *(x_original+i),*(y_original+i),*(std_original+i) );
//  
//  
//  /* get initial values for parameters of B and D components
//  printf("\n");
//  
//  printf("\n************************************************");
//  printf("\n* The fitting model is:                        *");
//  printf("\n*     S = B + CTF*CTF *D                       *");
//  printf("\n*     where B is the background noise;         *");
//  printf("\n*           D is the signal strengh decay;     *");
//  printf("\n* Enter the order of polynomial functions to   *");
//  printf("\n*   the background noise and signal strength   *");
//  printf("\n*   decay: 3-4 are generally good for most     *");
//  printf("\n*   micrographs.                               *");
//  printf("\n* If 3 is used for both:                       *");
//  printf("\n*        B = a0 + a1*x + a2*x^2 +a3*x^3        *");
//  printf("\n*        D = b0 + b1*x + b2*x^2 +b3*x^3        *");
//  printf("\n*        where x is the annulii                *");
//  printf("\n************************************************");
//  
//  printf("\n  The order of polynomial to fit the background B: ");
//  scanf("%d",&bkg_poly_order);
//  printf("\n  The order of polynomial to fit signal strengh decay D: ");
//  scanf("%d",&decay_poly_order);
//  printf("\n");
//   */
//  bkg_poly_order   = orderBkg;
//  decay_poly_order = orderSig;
//
//  bkg_poly_order++; decay_poly_order++;	
//  ma = bkg_poly_order+decay_poly_order+1;
//  printf("bpo = %d  dpo = %d  ma = %d\n",
//	 bkg_poly_order  ,
//	 decay_poly_order ,
//	 ma
//	 );
// 
//  tmp1 = ma;
//  a = dvector(1,ma);
//  ia = ivector(1,ma);
//  covar = dmatrix(1,tmp1,1,tmp1);
//  alpha = dmatrix(1,tmp1,1,tmp1); 
//  
//  /*  Don't reinit these every time!!! a0,a1,b0,b1,b2 etc...
//  printf("\n");
//  printf("\n************************************************");
//  printf("\n* Initial Parameters for polynomial functions  *");
//  printf("\n*    are the starting values for iteration.    *");
//  printf("\n*    If you use default values,                *");
//  printf("\n*       a0 = b0 = 1.0                          *");
//  printf("\n*       a1 = a2 = a3 = ... = 0.0               *");
//  printf("\n*       b1 = b2 = b3 = ... = 0.0               *");
//  printf("\n*    Otherwise, you can give your own inputs.  *");
//  printf("\n************************************************");
//  printf("\n  Do you want to use the default values (y/n)? ");
//  scanf("%s",yes_no);
//  if(!(strncmp(yes_no,"y",1))) 
//  */
//  if(TRUE)
//    {
//      for (i=2; i<=(bkg_poly_order+decay_poly_order); i++)  
//	*(a+i) = 0.0;
//      *(a+1) = 1.0;
//      *(a+bkg_poly_order+1)=1.0;
//
//      /* Let's get these from interface */
//      /* a */
//      thisStr = XmTextGetString(adra0W);
//      a[1]    = (float)strtod(thisStr,0);
//      XtFree(thisStr);
//      thisStr = XmTextGetString(adra1W);
//      a[2]    = (float)strtod(thisStr,0);
//      XtFree(thisStr);
//      thisStr = XmTextGetString(adra2W);
//      a[3]    = (float)strtod(thisStr,0);
//      XtFree(thisStr);
//      thisStr = XmTextGetString(adra3W);
//      a[4]    = (float)strtod(thisStr,0);
//      XtFree(thisStr);
//      thisStr = XmTextGetString(adra4W);
//      a[5]    = (float)strtod(thisStr,0);
//      XtFree(thisStr);
//      thisStr = XmTextGetString(adra5W);
//      a[6]    = (float)strtod(thisStr,0);
//      XtFree(thisStr);
//      /* b */
//      thisStr = XmTextGetString(adrb0W);
//      a[8]    = (float)strtod(thisStr,0);
//      XtFree(thisStr);
//      thisStr = XmTextGetString(adrb1W);
//      a[9]    = (float)strtod(thisStr,0);
//      XtFree(thisStr);
//      thisStr = XmTextGetString(adrb2W);
//      a[10]    = (float)strtod(thisStr,0);
//      XtFree(thisStr);
//      thisStr = XmTextGetString(adrb3W);
//      a[11]    = (float)strtod(thisStr,0);
//      XtFree(thisStr);
//      thisStr = XmTextGetString(adrb4W);
//      a[12]    = (float)strtod(thisStr,0);
//      XtFree(thisStr);
//      thisStr = XmTextGetString(adrb5W);
//      a[13]    = (float)strtod(thisStr,0);
//      XtFree(thisStr);
//
//    }        
//  else 
//    {
//      /* Do something else */
//    }
//  
//  for (i=1; i<=(bkg_poly_order+decay_poly_order); i++) 
//    *(ia+i) = 1;
//  
//  /* get information of the defocus and the first CTF zero
//  printf("\n\nEnter the guessed defocus value (float, ): ");
//  scanf("%f",&tmp3);
//  */
//  tmp3 =   ctfMajorDefocusValue * 10000.0;  /* Put it in Angstroms */
//printf("Defocus[%f]\n",tmp3);
//
//  *(a+ma)= tmp3;
//  *(ia+ma) = 1;
//  
//  /*  	printf("\nEnter the resolution of the first CTF zero (float, ): ");
//	scanf("%f",&tmp3);
//	reslow = (double) tmp3;*/
//  /*
//  printf("\nEnter the highest resolution you want to fit (float, ): ");
//  scanf("%f",&tmp3);
//  */
//  thisText = XmTextGetString(adrHiResW);
//  tmp3 = (float)strtod(thisText,0);
//  XtFree(thisText);
//  reshigh = (double) tmp3;
//printf("ResHigh = [%f]\n",reshigh);
//
//  /*
//  printf("\nEnter the pixel_size (float, ): ");
//  scanf("%f",&tmp3);
//  */
//  thisText = XmTextGetString(ctfPixSizeW); 
//  tmp3   = (float)strtod(thisText, 0);
//  XtFree(thisText);
//  pixel_size= (double) tmp3 * 10.0;  /* Put it into Angstroms */
//printf("pixsiz = [%f]\n",pixel_size);
//  
//  defocus = *(a+ma);
//  fw2 = (defocus-sqrt(defocus*defocus-Cs*wavelength*2.0))/Cs;
//  fw = sqrt(fw2);
//  annlow = (int)(fw*(pixel_size*2*array_size)/wavelength+0.5);
//  reslow= 2.0*array_size*pixel_size/annlow;
//  
//  /*	annlow = (int) (2*array_size *pixel_size / reslow + 0.5); */
//  annhigh = (int) (2*array_size *pixel_size / reshigh + 0.5);
//  
//  if(annlow < 1.0) annlow = 1.0;
//  if(annhigh > (float)array_size) annhigh = (float)array_size;
//  
//  if(DEBUG2) {
//    printf ("reslow: %f \n", (float)reslow);
//    printf ("reshigh: %f \n", (float)reshigh);
//    printf ("pixel_size: %f \n", (float)pixel_size);
//    printf ("annlow: %d\n", annlow);
//    printf ("annhigh: %d\n", annhigh);
//  }
//
//  
//  if(DEBUG1) 
//    {
//      printf("\nInitial parameters");
//      printf("\n==================");
//      
//      for (i=1; i<=bkg_poly_order; i++)
//	printf("\n   a[%d]= %e, %d\n",i, *(a+i),*(ia+i));
//      for (i=(bkg_poly_order+1); i<=(bkg_poly_order+decay_poly_order); i++) 
//	printf("\n   b[%d]= %e, %d\n",i, *(a+i),*(ia+i));
//      printf("\n   defocus level: %e   .\n ",*(a+ma) );
//    }
//  
//  /* preparing data for curve fitting: x, y, sig and ndata */
//  x = x_original + annlow - 1;
//  y = y_original + annlow- 1;
//  sig = std_original + annlow - 1;
//  ndata = annhigh + 1 - annlow; 
//
//  if(DEBUG1) {
//    for(i=1;i<=ndata; i++)
//      printf ("\nx,y,sig: %f, %f, %f",*(x +i),*(y +i),*(sig +i) );
//  } 
//  
//  /* allocate memory for chisq and alamda */  	
//  chisq = (double *) malloc (sizeof(double ));
//  alamda = (double *) malloc (sizeof(double ));
//  
//  /**********************************************************        	
//   * In this iteration session, all parameters are adjusted *
//   **********************************************************/
//  *alamda = -1.0;
//  for (;;) 
//    {
//      mrqmin (x, y, sig, ndata, a, ia, ma, covar, alpha, chisq,fpoly_ctf,
//	      alamda, bkg_poly_order, decay_poly_order, wavelength, 
//	      array_size, A, Cs, pixel_size);
//      if(DEBUG1)
//	printf ("\n alamda, chisq, defc.: %e, %e, %e", 
//		*alamda, *chisq,a[ma] );   
//      if(*alamda>big || *alamda<small) break;
//    }
//  *alamda = 0.0;
//  mrqmin (x, y, sig, ndata, a, ia, ma, covar, alpha, chisq,fpoly_ctf,
//	  alamda,bkg_poly_order, decay_poly_order,wavelength, array_size,
//	  A, Cs, pixel_size);
//  
//  if (DEBUG2) {
//    for (i=1; i<=bkg_poly_order; i++)  
//      printf("\na[%d]= %e",i, a[i]);
//    for (i=1+bkg_poly_order; i<=bkg_poly_order+decay_poly_order; i++)  
//      printf("\nb[%d]= %e",i, a[i]);
//    printf ("\ndefocus: %e \n",  a[ma]);
//    printf ("chisq: %e \n\n",  *chisq); 
//  } 
//  /************************************************************************ 
//   * In this iteration session, parameters for B, D and defocus value are 
//   * adjusted seperated 
//   *************************************************************************/
//  chisqold1 = *chisq;
//  for(;;)  
//    {
//      /******************************/
//      /**** Refine defocus value ****/	 
//      /******************************/
//      tmp5 = a[ma]; 
//      chisqold2=*chisq;
//      defcusold=a[ma];
//      for (i=1; i<ma; i++)  
//	ia[i]=0;
//      ia[ma]=1;	
//      for (j=-20; j<=20;j++)
//	{
//	  *alamda = -1.0;
//	  a[ma] = tmp5+j*10.;
//	  //	     printf ("\n%d, defocus: %e ", j, a[ma]);
//	  for(;;) 
//	    {
//	      mrqmin (x, y, sig, ndata, a, ia, ma, covar, alpha, chisq,
//		      fpoly_ctf, alamda,bkg_poly_order, decay_poly_order,
//		      wavelength, array_size,  A, Cs, pixel_size);
//	      if(*alamda>big || *alamda<small) break;
//	    } 
//	  *alamda =0.0; 
//	  
//	  mrqmin (x, y, sig, ndata, a, ia, ma, covar, alpha, chisq,fpoly_ctf,
//		  alamda,bkg_poly_order, decay_poly_order,wavelength, 
//		  array_size, A, Cs, pixel_size);
//	  if(*chisq<=chisqold2) 
//	    {
//	      chisqold2 = *chisq;
//	      defcusold = a[ma];
//	    } 
//	  /* printf ("\n j, chisq, defc.:%d  %e, %e", j, *chisq,a[ma] );    */
//	}
//      /* printf ("\n\n   chisq, defc.:  %e, %e\n",  chisqold2, defcusold );*/
//      a[ma]=defcusold;  
//      /**********************************/
//      /**** Refine back ground curve ****/	 
//      /**********************************/
//      *alamda =-1.0; 
//      for (i=1; i<=bkg_poly_order; i++)  
//	ia[i]=1;
//      for (i=bkg_poly_order+1; i<=bkg_poly_order+decay_poly_order; i++)  
//	ia[i]=0;
//      ia[ma]=0;
//      for(;;) 
//	{
//	  mrqmin (x, y, sig, ndata, a, ia, ma, covar, alpha, chisq,fpoly_ctf,
//		  alamda,bkg_poly_order, decay_poly_order,wavelength, 
//		  array_size, A, Cs, pixel_size);
//	  if(*alamda>big || *alamda<small) break;
//	  
//	  /*****
//	    for (j=1; j<=ma; j++)  
//	    printf(" a[%d]= %e",j, a[j]);
//	    printf ("i: %d, chisq: %e \n\n", i,  *chisq);  	
//	    ****/
//	} 
//      *alamda =0.0; 
//      
//      mrqmin (x, y, sig, ndata, a, ia, ma, covar, alpha, chisq,fpoly_ctf,
//	      alamda,bkg_poly_order, decay_poly_order,wavelength, array_size,
//	      A, Cs, pixel_size);
//      
//      /*	for (i=1; i<ma; i++)  
//		printf("\na[%d]= %e",i, a[i]);
//		
//		printf ("\ndefocus: %e \n",  a[ma]);
//		
//		printf ("chisq: %e \n\n",  *chisq); */
//      /********************************************/
//      /**** Refine signal strength decay curve ****/	 
//      /********************************************/
//	*alamda =-1.0; 
//      for (i=1; i<=bkg_poly_order; i++)  
//	ia[i]=0;
//      for (i=1+bkg_poly_order; i<=bkg_poly_order+decay_poly_order; i++)  
//	ia[i]=1; 
//      ia[ma]=0;
//      for(;;) 
//	{
//	  mrqmin (x, y, sig, ndata, a, ia, ma, covar, alpha, chisq,fpoly_ctf,
//		  alamda,bkg_poly_order, decay_poly_order,wavelength, 
//		  array_size, A, Cs, pixel_size);
//	  if(*alamda>big || *alamda<small) break;
//	  /***
//	    for (j=1; j<=ma; j++)  
//	    printf(" a[%d]= %e",j, a[j]);
//	    printf ("i: %d, chisq: %e \n\n", i,  *chisq);  	
//	    ***/
//	} 
//      *alamda =0.0; 
//      
//      mrqmin (x, y, sig, ndata, a, ia, ma, covar, alpha, chisq,fpoly_ctf,
//	      alamda,bkg_poly_order, decay_poly_order,wavelength, array_size,
//	      A, Cs, pixel_size);
//      if((*chisq-chisqold1)<small) break;
//      chisqold1 = *chisq;	 
//    }
//  for (i=1; i<=bkg_poly_order; i++)  
//    printf("\na%d= %e",i-1, a[i]);
//  for (i=bkg_poly_order+1; i<=bkg_poly_order+decay_poly_order; i++)  
//    printf("\nb%d= %e",i-bkg_poly_order-1, a[i]);
//  printf ("\ndefocus: %e \n",  a[ma]);
//  printf ("chisq: %e \n\n",  *chisq); 	 
//  
//
//  *calcDefocus = a[ma];
//  *calcChisq   = *chisq;
//
//
//  /* Now stuff our plotting arrays with appropriate */
//  tmp6 = 0.0;
//  /* fprintf(fp1, "X,Y,B+CTF^2*D,Y-B,B,CTF*CTF,D\n"); */
//  for (i=1; i<=array_size; i++) 
//    {/* all array_size */
//	
//      tmp5 = x_original[i];
//      tmp6 = a[1]; 
//      for (j=2; j<=bkg_poly_order; j++) 
//	{
//	  tmp6 +=a[j]*tmp5;
//	  tmp5= tmp5*x_original[i];
//	}		
//      B = tmp6;
//      /*	     y_original[i]=y_original[i]-tmp6;*/
//	
//	
//      tmp5 = x_original[i];
//      tmp6 = a[1+bkg_poly_order]; 
//      for (j=2+bkg_poly_order; j<=bkg_poly_order+decay_poly_order; j++) 
//	{
//	  tmp6 +=a[j]*tmp5;
//	  tmp5= tmp5*x_original[i];
//	}		
//	
//      
//      fw =wavelength*x_original[i]/(array_size*2.0* pixel_size );
//      fw2= fw*fw;
//      theta = pie*2.0*(Cs*fw2*fw2/4.0-a[ma]*fw2/2.0)/wavelength;
//      ctf = sqrt(1.0-A*A)*sin(theta)-A*cos(theta);
//      ctf2 = ctf*ctf;
//      /* fprintf(fp1, "%f, %f, %f, %f, %f, %f, %f\n",  
//	      x_original[i], y_original[i], 
//	      B+ctf2*tmp6, y_original[i]-B,B, ctf2,tmp6);	*/
//      /* fprintf(fp1, "X,Y,B+CTF^2*D,Y-B,B,CTF*CTF,D\n"); */
//      adr_s[i-1]       = (B+ctf2*tmp6);
//      adr_sminusb[i-1] = adr_s[i-1] - B;
//      adr_ctfsq[i-1]   = ctf2;
//      adr_d[i-1]       = tmp6;
//      adr_b[i-1]       = B;
//
//      adr_s[i-1] *= ca1d_avg;
//      adr_b[i-1] *= ca1d_avg;
//
//    }/* all array_size */
//
//
//
//  /*********************************************/
//  
//  /*********************************************/
//  /* write out a file: bkg_poly_order, decay_poly_order
//     a0, a1, ...
//     b0, b1, ...
//     defocus
//     chisq */
//  
//#ifdef WRITE_OUTFILES
//  if ((fp1 = fopen (Outfile1, "w"))==NULL) {
//    printf("%s: can't open.", Outfile1  );
//    exit(-1);
//  }
//  else  
//    {
//      fprintf(fp1, "Fitting FFT CA1D curve: S = B + CTF^2 * D\n");
//      fprintf(fp1, "=========================================\n");
//      
//      fprintf(fp1, "filename: %s\n",InFilename);
//      fprintf(fp1, "Resolution range for fitting: %f - %f ()\n", 
//      (float)reslow, (float)reshigh);
//      fprintf(fp1, "FFT size: %d\n",array_size*2);
//      fprintf(fp1, "Pixel size: %f \n",(float)pixel_size);
//      
//      fprintf(fp1, "bkg_poly_order: %d\n",bkg_poly_order-1);
//      fprintf(fp1, "decay_poly_order: %d\n",decay_poly_order-1);
//      for (i=1; i<=bkg_poly_order; i++)
//	fprintf(fp1, "B%d= %e\n",i-1,a[i]);
//      for (i=bkg_poly_order+1; i<=bkg_poly_order+decay_poly_order; i++)  
//	fprintf(fp1,"D%d= %e\n",i-bkg_poly_order-1, a[i]);
//      
//      fprintf(fp1, "defocus(): %e\n",a[ma]);
//      fprintf(fp1, "chisq: %e\n",*chisq);
//      fprintf(fp1, "fftavgvalue: %e\n",fftavgvalue);
//      
//    }
//  fclose (fp1);
//  if ((fp1 = fopen (Outfile3, "w"))==NULL) {
//    printf("%s: can't open.", Outfile3  );
//    exit(-1);
//  }
//  else  
//    {
//      fprintf(fp1, "%d\n",array_size*2);
//      fprintf(fp1, "%f\n",(float)pixel_size);
//      
//      fprintf(fp1, "%d\n",bkg_poly_order-1);
//      fprintf(fp1, "%d\n",decay_poly_order-1);
//      for (i=1; i<=bkg_poly_order; i++)
//	fprintf(fp1, "%e\n",a[i]);
//      for (i=bkg_poly_order+1; i<=bkg_poly_order+decay_poly_order; i++)  
//	fprintf(fp1, "%e\n",a[i]);
//      
//      fprintf(fp1, "%e\n",a[ma]);
//      fprintf(fp1, "%e\n",fftavgvalue);
//      
//    }
//  fclose (fp1);
//
//  /* write out a file: this_x, (this_x - B), ctf*ctf, D */
//
//  if ((fp1 = fopen (Outfile2, "w"))==NULL) {
//    printf("%s: can't open.", Outfile2  );
//    exit(-1);
//  }
//  else  
//    {
//      tmp6 = 0.0;
//      fprintf(fp1, "X,Y,B+CTF^2*D,Y-B,B,CTF*CTF,D\n");
//      for (i=1; i<=array_size; i++) {
//	
//	tmp5 = x_original[i];
//	tmp6 = a[1]; 
//	for (j=2; j<=bkg_poly_order; j++) {
//	  tmp6 +=a[j]*tmp5;
//	  tmp5= tmp5*x_original[i];
//	}		
//	B = tmp6;
//	/*	     y_original[i]=y_original[i]-tmp6;*/
//	
//	
//	tmp5 = x_original[i];
//	tmp6 = a[1+bkg_poly_order]; 
//	for (j=2+bkg_poly_order; j<=bkg_poly_order+decay_poly_order; j++) {
//	       tmp6 +=a[j]*tmp5;
//	       tmp5= tmp5*x_original[i];
//	     }		
//	
//	
//	fw =wavelength*x_original[i]/(array_size*2.0* pixel_size );
//	fw2= fw*fw;
//	theta = pie*2.0*(Cs*fw2*fw2/4.0-a[ma]*fw2/2.0)/wavelength;
//	ctf = sqrt(1.0-A*A)*sin(theta)-A*cos(theta);
//	ctf2 = ctf*ctf;
//	fprintf(fp1, "%f, %f, %f, %f, %f, %f, %f\n",  
//		x_original[i], y_original[i], 
//	    	B+ctf2*tmp6, y_original[i]-B,B, ctf2,tmp6);	
//      }
//    }
//  
//  fclose (fp1);
//  
//  /* write out empft.dat type file */ 
//  if ((fp1 = fopen (Outfile4, "w"))==NULL) {
//    printf("%s: can't open.", Outfile2  );
//    exit(-1);
//  }
//  else {
//    fprintf(fp1, "%s\n", Imgfile);
//    fprintf(fp1, "%f, %d, %f, %f, %f, %f, %f, %f\n", pixel_size,1,
//	    200000.0,0.05,a[ma]/10000.,a[ma]/10000.,0.0,2.0); 
//  }
//  fclose (fp1);
//#endif /* ifdef WRITE_OUTFILES */
//  
//  /* free pointers */
//
//  /* Now put new array values on interface */
//  /* a */
//  sprintf(outStr,"%12E",a[1]);
//  XmTextSetString(adra0W,outStr);
//  sprintf(outStr,"%12E",a[2]);
//  XmTextSetString(adra1W,outStr);
//  sprintf(outStr,"%12E",a[3]);
//  XmTextSetString(adra2W,outStr);
//  sprintf(outStr,"%12E",a[4]);
//  XmTextSetString(adra3W,outStr);
//  sprintf(outStr,"%12E",a[5]);
//  XmTextSetString(adra4W,outStr);
//  sprintf(outStr,"%12E",a[6]);
//  XmTextSetString(adra5W,outStr);
//  /* b */
//  sprintf(outStr,"%12E",a[7]);
//  XmTextSetString(adrb0W,outStr);
//  sprintf(outStr,"%12E",a[8]);
//  XmTextSetString(adrb1W,outStr);
//  sprintf(outStr,"%12E",a[9]);
//  XmTextSetString(adrb2W,outStr);
//  sprintf(outStr,"%12E",a[10]);
//  XmTextSetString(adrb3W,outStr);
//  sprintf(outStr,"%12E",a[11]);
//  XmTextSetString(adrb4W,outStr);
//  sprintf(outStr,"%12E",a[12]);
//  XmTextSetString(adrb5W,outStr);
//
//  free (chisq);
//  free (alamda);
//  free_dvector (a, 1, ma);
//  free_ivector (ia, 1, ma);
//  free_dmatrix (covar, 1, ma, 1, ma);
//  free_dmatrix (alpha, 1, ma, 1, ma);
//
//
}/* AUTODR_WORK */




/* ************************************************************
 * autoDR_activate() - Setup and do auto defocus refine.
 */
void autoDR_activate(Widget widget, int *tag, 
		     XmAnyCallbackStruct *callback_data)
{/* autoDR_ACTIVATE */
  int iter;
  char *thisText;
  int  orderBkg,orderSig,idim_min,thisIdim1,thisIdim2;
  int arg1,arg2,arg3,arg4,hiPixel,loPixel,linear;
  float map_min,map_max,map_avg,map_std;
  float calcDefocus,calcChisq;
  float *tmpBuffer;

  for(iter=0;iter<*tag;iter++)
    {/* DO the calc */

      /* Now let's get the circAvg */
      thisIdim1= thisIdim2 = 512;
      idim_min = MIN(thisIdim1,thisIdim2);
      if(ca1d)
	free(ca1d);
      if(ca1d_std)
	free(ca1d_std);
      if(ca1d_cnt)
	free(ca1d_cnt);
      ca1d     = (float *)malloc(sizeof(float)*(idim_min/2)+2);
      ca1d_std = (float *)malloc(sizeof(float)*(idim_min/2)+2);
      ca1d_cnt = (float *)malloc(sizeof(float)*(idim_min/2)+2);

      if(adr_s)
	free(adr_s);
      if(adr_sminusb)
	free(adr_sminusb);
      if(adr_ctfsq)
	free(adr_ctfsq);
      if(adr_d)
	free(adr_d);
      if(adr_b)
	free(adr_b);
      adr_s       = (float *)malloc(sizeof(float)*(1024));
      adr_sminusb = (float *)malloc(sizeof(float)*(1024));
      adr_ctfsq   = (float *)malloc(sizeof(float)*(1024));
      adr_d       = (float *)malloc(sizeof(float)*(1024));
      adr_b       = (float *)malloc(sizeof(float)*(1024));
      
      arg1 = fftSize;
      arg2 = fftSize;
      arg3 = 512;
      arg4 = 512;
      if(hiColor)
	{
	  loPixel = (int)8;
	  hiPixel = (int)(tryColors-2);
	}
      else
	{/* 8 bit */
	  loPixel = (int)pixels_return[(8)];
	  hiPixel = (int)pixels_return[(tryColors-2)];
	}/* 8 bit */
      if(fftType==LINEARFFT)
	linear = -1;
      else linear = 0;     /* Log */
      fft_dsp_c(fftImage_buffer2, 
		fft_buffer2,
		arg1, arg2, arg3, arg4,
		globMaxFFTvalue,
		globMinFFTvalue,
		hiPixel,
		loPixel,
		linear,
		(fftDispType==AMPLITUDE)?TRUE:FALSE);

      tmpBuffer = (float *)malloc(sizeof(float)*thisIdim1*thisIdim2);
      for(i=0;i<thisIdim1;i++)
	for(j=0;j<thisIdim2;j++)
	  tmpBuffer[(i*thisIdim1)+j]= (float)fft_buffer2[(i*thisIdim1)+j];
      
      map2d_stats_(tmpBuffer, &thisIdim1, &thisIdim2,
		   &map_min, &map_max,
		   &map_avg, &map_std,
		   ca1d,ca1d_std,ca1d_cnt,
		   &idim_min
		   );

      ca1d_avg = map_avg;
      
      free(tmpBuffer);
      
      
      /* Now let's calc the defocus */
      /* Let's just force it to be 4 for now *
	 thisText = XmTextGetString(orderBkgW); 
	 orderBkg = (int)strtod(thisText,0);
	 XtFree(thisText);
	 thisText = XmTextGetString(orderSigW); 
	 orderSig = (int)strtod(thisText,0);
	 XtFree(thisText);
	 ****************************************/
      /* OLD way 
	 orderBkg = orderSig = 4;
      */
      orderBkg = orderSig = 5;
      
      auto1DDR_work(orderBkg,orderSig,
		  &calcDefocus, &calcChisq);
      sprintf(outStr,"%f",calcDefocus/10000.0); /* /10000.0 to convert to um */
      XmTextSetString(adrDefocusW,outStr);
      sprintf(outStr,"%f",calcChisq);
      XmTextSetString(adrChisqW,outStr);
      
      XFlush(dsply);
      
    }/* DO the calc */
  

}/* autoDR_ACTIVATE */




/* ************************************************************
 * autoDR_update_activate() - Update text widgets with defocus
 * estimate.
 */
void autoDR_update_activate(Widget widget, int *tag, 
			    XmAnyCallbackStruct *callback_data)
{/* AUTODR_UPDATE_ACTIVATE */
  char *thisText;

  thisText = XmTextGetString(adrDefocusW);
  XmTextSetString(ctffocusMajValueW, thisText);
  XmTextSetString(ctfFocusMinvalueW, thisText);

  XtFree(thisText);
  XFlush(dsply);

  DrawFFTWindow();

}/* AUTODR_UPDATE_ACTIVATE */




/* ************************************************************
 * autoDR_deactivate() - Free memory after the defocus refinement.
 */
void autoDR_deactivate(Widget widget, int *tag, 
		       XmAnyCallbackStruct *callback_data)
{/* AUTODR_DEACTIVATE */
  
  if(allPackImg)
    {free(allPackImg);
    allPackImg = NULL;}

  if(ca1d)
    {free(ca1d);
    ca1d = NULL;}
  if(ca1d_std)
    {free(ca1d_std);
    ca1d_std = NULL;}
  if(ca1d_cnt)
    {free(ca1d_cnt);
    ca1d_cnt = NULL;}
  if(adr_s)
    {free(adr_s);
    adr_s = NULL;}
  if(adr_sminusb)
    {free(adr_sminusb);
    adr_sminusb = NULL;}
  if(adr_ctfsq)
    {free(adr_ctfsq);
    adr_ctfsq = NULL;}
  if(adr_d)
    {free(adr_d);
    adr_d = NULL;}
  if(adr_b)
    {free(adr_b);
    adr_b = NULL;}
  doingAutoDefocusRefine= FALSE;
  XtUnmanageChild(autoDRScreenWidget); 

}/* AUTODR_DEACTIVATE */


/*****************************************************************
 * adrab_reset_activate() - Reset param text widgets for auto defocus
 * refine so a new refeinement can be started.
*/
void adrab_reset_activate(Widget widget, int *tag, 
			  XmAnyCallbackStruct *callback_data)
{/* ADRAB_RESET_ACTIVATE */

  sprintf(outStr,"1");
  XmTextSetString(adra0W,outStr);
  XmTextSetString(adrb0W,outStr);

  sprintf(outStr,"0");
  XmTextSetString(adra1W,outStr);
  XmTextSetString(adra2W,outStr);
  XmTextSetString(adra3W,outStr);
  XmTextSetString(adra4W,outStr);
  XmTextSetString(adra5W,outStr);
  XmTextSetString(adrb1W,outStr);
  XmTextSetString(adrb2W,outStr);
  XmTextSetString(adrb3W,outStr);
  XmTextSetString(adrb4W,outStr);
  XmTextSetString(adrb5W,outStr);



}/* ADRAB_RESET_ACTIVATE */




/* ************************************************************
 * binning_deactivate() - Close binning window
 */
void binning_deactivate(Widget widget, int *tag, 
			XmAnyCallbackStruct *callback_data)
{/* BINNING_DEACTIVATE */
  
  XtUnmanageChild(binningScreenWidget); 

}/* BINNING_DEACTIVATE */




/* ************************************************************
 * interp_deactivate() - Close down interp window.
 */
void interp_deactivate(Widget widget, int *tag, 
		       XmAnyCallbackStruct *callback_data)
{/* INTERP_DEACTIVATE */
  
  XtUnmanageChild(interpScreenWidget); 

}/* INTERP_DEACTIVATE */




/* *******************************************************
 * adr_toggle_change() - Whenever the toggle state is changed
 * call here.  For auto defocus refine
 */
void adr_toggle_change(Widget widget, int *tag, 
		       XmToggleButtonCallbackStruct *callback_data)
{/* ADR_TOGGLE_CHANGE */


  if(*tag==ADRCIRCAVG)
    {/*  */
      adrCircAvgDraw= !adrCircAvgDraw;
    }/*  */
  else if(*tag==ADRS)
    {/*  */
      adrSDraw= !adrSDraw;
    }/*  */
  else if(*tag==ADRSMINUSB)
    {/*  */
      adrSminusBDraw= !adrSminusBDraw;
    }/*  */
  else if(*tag==ADRCTFSQ)
    {/*  */
      adrCtfsqDraw= !adrCtfsqDraw;
    }/*  */
  else if(*tag==ADRD)
    {/*  */
      adrDDraw=  !adrDDraw;
    }/*  */
  else if(*tag==ADRB)
    adrBDraw=  !adrBDraw;
  else if(*tag==ADRBLANK)
    adrBlankDraw=  !adrBlankDraw;

  DrawFFTWindow();

}/* ADR_TOGGLE_CHANGE */



/* ************************************************************
 * version_scroll_activate() - Scroll the version number widget.
 * Useful on VMS but not sure if applicable to UNIX - keep 
 * in case a need can be found.
 */
void version_scroll_activate(Widget widget, int *tag, 
			     XmAnyCallbackStruct *callback_data)
{/* VERSION_SCROLL_ACTIVATE */
  int thisVersion,imgOffset;
  char *versionText,*s1,*s2;
  char thisFileName[512],rawFileName[512];
  XEvent event;
  forPifHead_t thisHead;
  FILE *fIn;
  XmString simpleString;
  fString_t rWrite,inFile;

  rWrite.addr = malloc(sizeof(char)*4);
  rWrite.length = 1;
  sprintf(rWrite.addr,"r");
  inFile.addr = malloc(sizeof(char)*64);
  inFile.length = 1;

  if(inited)
    {/* Don't do until we've got everything setup */

      if(*tag==SCROLL)
	{/* Scroll them puppies */

	  sprintf(rawFileName,"%s",fileName.addr);
	  if(s1= strrchr(rawFileName,';'))
	    *s1= '\0';
	  thisVersion = 1;
	  while(TRUE)
	    {/* Do until stopped */
	      
	      fIn = 0;
	      while(!fIn)
		{/* Try this */
		  sprintf(thisFileName,"%s;%d",rawFileName,thisVersion);
		  fIn = fopen(thisFileName, "r");
		  if(!fIn && thisVersion==1)
		    {/* Dont know what to do */
		      sprintf(outStr,"Cannot find this file [%s]",
			      thisFileName);
		      askThem(outStr,FALSE,XmDIALOG_INFORMATION);
		      goto noMoreScroll;
		    }/* Dont know what to do */
		  if(!fIn)
		    thisVersion = 1;
		  else thisVersion++;
		}/* Try this */
	      fclose(fIn);

	      /* Tell them what we're doing */
	      sprintf(outStr,"%d",thisVersion-1);
	      XmTextSetString(versionTextWidget, outStr);
	      XFlush(dsply);

	      sprintf(outStr,"%d",(thisVersion-1));
	      createSimpleString(&simpleString,outStr);
	      XtSetArg(arglst[0], XmNlabelString, simpleString );
	      XtSetArg(arglst[1], XmNrecomputeSize, FALSE );
	      XtSetValues(versionTextWidget, arglst, 2); 
	      XFlush(dsply);
	      XmStringFree(simpleString); 
	      
	      thisHead.luNum = 7;
	      sprintf(inFile.addr,"%s",thisFileName);
	      inFile.length = strlen(thisFileName);

	      pif_open_(inFile.addr, &(thisHead.luNum), rWrite.addr);

	      pif_read_gh(&thisHead);
	      thisHead.imgNum = 1;
	      pif_read_dh(&thisHead);

	      purdueHeader.numCols= thisHead.numCols;
	      purdueHeader.numRows= thisHead.numRows;
	      purdueHeader.nimg=  thisHead.nimg;

	      /* If we can just read the whole thing into core
	       * memory then virtual_read_activate() won't have to
	       * work as hard. */
	      if(imgInCore)
		free(inCore_buffer);
	      CK(inCore_buffer= (float *)malloc(sizeof(float)*
					    (thisHead.numCols)*
					    (thisHead.numRows)));
	      if(inCore_buffer)
		imgInCore= TRUE;
	      else imgInCore= FALSE;
	      
	      if(imgInCore)
		{/* OK to go */
		  imgOffset = 1024;  /* one image for now */
		  if(thisHead.mode==1)
		    {/* short */
		      unsigned short *short_buffer;
		      int z;
		      short_buffer = (unsigned short *)
			malloc(sizeof(unsigned short)*
					    (thisHead.numCols)*
					    (thisHead.numRows));
		      pif_read_short_image(&(thisHead.luNum),
					  short_buffer,
					  &imgOffset,
					  &(thisHead.numCols),
					  &(thisHead.numRows),
					  &(thisHead.numSecs));
		      for(z=0;z<(thisHead.numCols*thisHead.numRows);z++)
			inCore_buffer[z] = (float)short_buffer[z];
		      free(short_buffer);
		    }/* short */
		  else if(thisHead.mode==0)
		    {/* byte */
		      BYTE *byte_buffer;
		      int z;
		      byte_buffer = (BYTE *)malloc(sizeof(BYTE)*
					    (thisHead.numCols)*
					    (thisHead.numRows));
		      pif_read_byte_image(&(thisHead.luNum),
					  byte_buffer,
					  &imgOffset,
					  &(thisHead.numCols),
					  &(thisHead.numRows),
					  &(thisHead.numSecs));
		      for(z=0;z<(thisHead.numCols*thisHead.numRows);z++)
			inCore_buffer[z] = byte_buffer[z];
		      free(byte_buffer);
		    }/* byte */
		}/* OK to go */
	      else
		{/* error */
		  sprintf(outStr,"Cannot allocate space for [%s]",
			  thisFileName);
		  askThem(outStr,FALSE,XmDIALOG_INFORMATION);
		  goto noMoreScroll;
		}/* error */

	      pif_close(&(thisHead.luNum));
	      
	      virtual_read_activate(-99,-99);
	      
	      /* Give feedback on main window title bar */
	      if(!masterProcess)
		sprintf(outStr,"Robem: [%s] - %d images total. [%s]",
			inFile.addr,
			thisHead.nimg,"CHILD");
	      else
		sprintf(outStr,"Robem: [%s] - %d images total.",
			inFile.addr,
			thisHead.nimg);
	      createSimpleString(&simpleString,outStr);
	      XtSetArg(arglst[0], XmNdialogTitle, simpleString );
	      XtSetArg(arglst[1], XmNrecomputeSize, FALSE );
	      XtSetValues(main_window_widget, arglst, 2); 
	      XFlush(dsply);
	      XmStringFree(simpleString); 
	      XFlush(dsply);
	      
		/* Frame the location of the stop button */
		XDrawLine(dsply,XtWindow(versionStopW),image_gcRED,
			  12,12,20,20);
		XDrawLine(dsply,XtWindow(versionStopW),image_gcRED,
			  20,12,12,20);
		XFlush(dsply);

	      /* See if they want to bugout or not */
	      if(XCheckTypedWindowEvent(dsply,
					XtWindow(versionStopW),
					ButtonPress,
					&event))
		{/* STOP!!! */
		  XFlush(dsply);
		  goto noMoreScroll;
		}/* STOP!!! */
	      
	    }/* Do until stopped */
	}/* Scroll them puppies */
      else
	{/* Just do this version number */
	  versionText = XmTextGetString(versionTextWidget); 
	  
	  if(!EQU(versionText,""))
	    {/* Don't do in the interim state */
	      sprintf(rawFileName,"%s",fileName.addr);
	      if(s1= strrchr(rawFileName,';'))
		*s1= '\0';
	      sprintf(thisFileName,"%s;%s",rawFileName,versionText);

	      fIn = fopen(thisFileName, "r");
	      if(!fIn)
		{/* Dont know what to do */
		  sprintf(outStr,"Cannot find this file [%s]",
			  thisFileName);
		  askThem(outStr,FALSE,XmDIALOG_INFORMATION);
		  goto noMoreScroll;
		}/* Dont know what to do */
	      fclose(fIn);

	      thisHead.luNum = 7;
	      sprintf(inFile.addr,"%s",thisFileName);
	      inFile.length = strlen(thisFileName);

	      pif_open_(inFile.addr, &(thisHead.luNum), rWrite.addr);

	      pif_read_gh(&thisHead);
	      thisHead.imgNum = 1;
	      pif_read_dh(&thisHead);

	      purdueHeader.numCols= thisHead.numCols;
	      purdueHeader.numRows= thisHead.numRows;
	      purdueHeader.nimg=  thisHead.nimg;

	      /* If we can just read the whole thing into core
	       * memory then virtual_read_activate() won't have to
	       * work as hard. */
	      if(imgInCore)
		free(inCore_buffer);
	      CK(inCore_buffer= (float *)malloc(sizeof(float)*
					    (thisHead.numCols)*
					    (thisHead.numRows)));
	      if(inCore_buffer)
		imgInCore= TRUE;
	      else imgInCore= FALSE;
	      
	      if(imgInCore)
		{/* OK to go */
		  imgOffset = 1024;  /* one image for now */
		  if(thisHead.mode==1)
		    {/* short */
		      unsigned short *short_buffer;
		      int z;
		      short_buffer = (unsigned short *)
			malloc(sizeof(unsigned short)*
					    (thisHead.numCols)*
					    (thisHead.numRows));
		      pif_read_short_image(&(thisHead.luNum),
					  short_buffer,
					  &imgOffset,
					  &(thisHead.numCols),
					  &(thisHead.numRows),
					  &(thisHead.numSecs));
		      for(z=0;z<(thisHead.numCols*thisHead.numRows);z++)
			inCore_buffer[z] = (float)short_buffer[z];
		      free(short_buffer);
		    }/* short */
		  else if(thisHead.mode==0)
		    {/* byte */
		      BYTE *byte_buffer;
		      int z;
		      byte_buffer = (BYTE *)malloc(sizeof(BYTE)*
					    (thisHead.numCols)*
					    (thisHead.numRows));
		      pif_read_byte_image(&(thisHead.luNum),
					  byte_buffer,
					  &imgOffset,
					  &(thisHead.numCols),
					  &(thisHead.numRows),
					  &(thisHead.numSecs));
		      for(z=0;z<(thisHead.numCols*thisHead.numRows);z++)
			inCore_buffer[z] = byte_buffer[z];
		      free(byte_buffer);
		    }/* byte */
		}/* OK to go */
	      else
		{/* error */
		  sprintf(outStr,"Cannot allocate space for [%s]",
			  thisFileName);
		  askThem(outStr,FALSE,XmDIALOG_INFORMATION);
		  goto noMoreScroll;
		}/* error */

	      pif_close(&(thisHead.luNum));
	      
	      virtual_read_activate(-99,-99);

	      /* Give feedback on main window title bar */
	      if(!masterProcess)
		sprintf(outStr,"Robem: [%s] - %d images total. [%s]",
			inFile.addr,
			thisHead.nimg,"CHILD");
	      else
		sprintf(outStr,"Robem: [%s] - %d images total.",
			inFile.addr,
			thisHead.nimg);

	      createSimpleString(&simpleString,outStr);
	      XtSetArg(arglst[0], XmNdialogTitle, simpleString );
	      XtSetArg(arglst[1], XmNrecomputeSize, FALSE );
	      XtSetValues(main_window_widget, arglst, 2); 
	      XFlush(dsply);
	      XmStringFree(simpleString); 
	      XFlush(dsply);
	      
	      
	    }/* Don't do in the interim state */
	  XtFree(versionText);
	  
	}/* Just do this version number */

    }/* Don't do until we've got everything setup */

 noMoreScroll:;
  free(rWrite.addr);
  free(inFile.addr);

}/* VERSION_SCROLL_ACTIVATE */




/* ************************************************************
 * version_screen_deactivate() - Put away the interface for reading 
 * multiple file versions.
 */
void version_screen_deactivate(Widget widget, int *tag, 
			       XmAnyCallbackStruct *callback_data)
{/* VERSION_SCREEN_DEACTIVATE */
  
  XtUnmanageChild(versionScreenWidget);

}/* VERSION_SCREEN_DEACTIVATE */



/* ************************************************************
 * version_activate() - Put up the interface for reading 
 * multiple file versions.
 */
void version_activate(Widget widget, int *tag, 
		      XmAnyCallbackStruct *callback_data)
{/* version_ACTIVATE */
  
  raiseIt(versionScreenWidget);

}/* version_ACTIVATE */


/* ************************************************************
 * versionArrowActivate() - Arrow widget changes concept of
 * version number.
*/
void versionArrowActivate(Widget widget, int *tag, 
			  XmAnyCallbackStruct *callback_data)
{/* VERSIONARROWACTIVATE */
  char *versionText;
  int thisVersion;

  versionText = XmTextGetString(versionTextWidget); 
  thisVersion = (int)strtod(versionText,0);
  XtFree(versionText);
  /* is it up or down ? */
  if(*tag==VERSIONUP)
    {/* Up */
      thisVersion++;
    }/* Up */
  else if(*tag==VERSIONDOWN)
    {/* Dopwn */
      thisVersion--;
    }/* Down */
  thisVersion = MAX(thisVersion,1);

  sprintf(outStr,"%d",thisVersion);
  XmTextSetString(versionTextWidget, outStr); 

  version_scroll_activate(widget,tag,0);

}/* VERSIONARROWACTIVATE */





/* ************************************************************
 * colormap_activate() - Build the image histogram and throw the
 * screen up on the display.
 */
void colormap_activate(Widget widget, int *tag, 
		       XmAnyCallbackStruct *callback_data)
{/* colormap_ACTIVATE */
  
  histogram_activate(&minP, &maxP);
  raiseIt(colorScreenWidget);

}/* colormap_ACTIVATE */



/* ************************************************************
 * colormap_screen_deactivate() - Get the colormap screen
 * off the display.
 */
void colormap_screen_deactivate(Widget widget, int *tag, 
				XmAnyCallbackStruct *callback_data)
{/* colormap_screen_deactIVATE */
  
  XtUnmanageChild(colorScreenWidget);
  
}/* colormap_screen_deactIVATE */



/* ************************************************************
 * DrawColormapWindow() - Redraw window e.g. on expose events
 * or whenever it's prudent.
 */
void DrawColormapWindow()
{/* DrawColormapWindow */
  int draw_width, draw_height;
  
  draw_width = 32;
  draw_height = 256;
  if(inited && XtIsManaged(colorScreenWidget))
    XPutImage(dsply, color_window1, color_gc, color_image1, 0, 0,
	      0, 0,  draw_width,  draw_height);
  XFlush (dsply);
  
}/* DrawColormapWindow */



/* ************************************************************
 * DrawTestColormapWindow() - Redraw window e.g. on expose events
 * or whenever it's prudent.  Draws either the histogram area
 * of the test color area on the colormap screen.
 */
void DrawTestColormapWindow(Widget widget, int *tag, 
			    XmAnyCallbackStruct *callback_data)
{/* DrawTestColormapWindow */

  if(inited && XtIsManaged(colorScreenWidget))
    {/* Only do if managed */
      if(*tag==HIST)
	XPutImage(dsply, hist_color_window1, hist_color_gc, 
		  hist_color_image1, 
		  0, 0,
		  0, 0,  32,  256); 
      else if(*tag==TEST)
	{
	  XPutImage(dsply, test_color_window1, test_color_gc, 
		    test_color_image1, 
		    0, 0,
		    0, 0,  32,  32);
	}

      XFlush (dsply);
    }/* Only do if managed */
  
}/* DrawTestColormapWindow */



/* ************************************************************
 * colorScrollChange() - Whenever they've changed the value
 * of one of the scroll bars on the Alter Colormap screen
 * call here.  Set the level based on the position of the
 * scrollbar.  Set the PLACEHOLDER entry of our colormap
 * to the value of the RGB scrolls.  This will cause the 
 * TEST drawing area of the Alter Colormap screen to change to
 * this new color as it is comprised of pixels with 
 * PLACEHOLDER value.
 */
void colorScrollChange(Widget widget, int *tag, 
		       XmScrollBarCallbackStruct *callback_data)
{/* colorScrollCHANGE */
  XmString simpleString;

  if(*tag==2)
    {/* RED */
      redLevel = callback_data->value;
    }/* RED */
  else if(*tag==3)
    {/* green */
      greenLevel = callback_data->value;
    }/* green */
  else if(*tag==4)
    {/* blue */
      blueLevel = callback_data->value;
    }/* blue */
  else if(*tag==5)
    {/* blue */
      brightLevel = callback_data->value;
    }/* blue */
  else if(*tag==STEPCM)
    {/* stepCM */
      int thisStep,thisSection, numSections;
      float colorPerSection;

      stepCM = (int)callback_data->value;
      sprintf(outStr,"%3d",stepCM);
      createSimpleString(&simpleString,outStr);
      XtSetArg(arglst[0], XmNlabelString, simpleString );
      XtSetArg(arglst[1], XmNrecomputeSize, FALSE );
      XtSetValues(stepCMValueW, arglst, 2); 
      XFlush(dsply);
      XmStringFree(simpleString); 
      XFlush(dsply);

      /* Now step the colormap */
      for(i=7;i<(tryColors);i++)
	{/* Do the colors we have been assigned */
	  /* Divide the entire colormap in STEP sections stepCM is always
	   * >= 1 */
	  numSections = (int)((float)((float)tryColors-7.0)*
				(float)((float)stepCM/256.0))+1;
	  thisSection = (int)((float)((float)i-7.0) * 
			      (float)((float)numSections / 
				      (float)((float)tryColors-7.0)));
	  
	  /* Start the colorPerSection out as GRAYscale and override 
	   * if desired */
	  if(numSections > 1)
	    numSections -= 1;      /* To make color brighter */
	  colorPerSection = (float)(255.0 / (float)numSections);

	  if(hiColor)
	    color_defs[i].pixel = i;
	  else color_defs[pixels_return[i]].pixel = pixels_return[i];

	  if(colorType==USEBUILD)
	    colorPerSection = (float)((float)redLevel / (float)numSections);
	  thisStep = (int)thisSection * colorPerSection;

	  if(hiColor)
	    color_defs[i].red   = ((thisStep) * (pixelColor));
	  else
	    color_defs[pixels_return[i]].red   = ((thisStep) * (pixelColor));

	  if(colorType==USEBUILD)
	    colorPerSection = (float)((float)greenLevel / (float)numSections);
	  thisStep = (int)thisSection * colorPerSection;
	  if(hiColor)
	    color_defs[i].green = ((thisStep) * (pixelColor));
	  else
	    color_defs[pixels_return[i]].green = ((thisStep) * (pixelColor));

	  if(colorType==USEBUILD)
	    colorPerSection = (float)((float)blueLevel / (float)numSections);
	  thisStep = (int)thisSection * colorPerSection;
	  if(hiColor)
	    color_defs[i].blue  = ((thisStep) * (pixelColor));
	  else
	    color_defs[pixels_return[i]].blue  = ((thisStep) * (pixelColor));
	  
	  if(hiColor)
	    color_defs[i].flags =  DoRed | DoGreen | DoBlue;
	  else
	    color_defs[pixels_return[i]].flags =  DoRed | DoGreen | DoBlue;

	  if(masterProcess)
	    {
	      if(hiColor)
		{XAllocColor (dsply, color_map, &color_defs[i]);
		 pixels_return[i] = color_defs[i].pixel;}
	      else
		XStoreColor (dsply, color_map, &color_defs[pixels_return[i]]);
	    }
	  XFlush(dsply);
	}/* Do the colors we have been assigned */

    }/* stepCM */
  
  if(*tag!=STEPCM)
    {/* Give feedback */
      sprintf(outStr,"%3d",callback_data->value);
      createSimpleString(&simpleString,outStr);
      XtSetArg(arglst[0], XmNlabelString, simpleString );
      XtSetArg(arglst[1], XmNrecomputeSize, FALSE );
      XtSetValues(colormapValueWidget[*tag], arglst, 2); 
      XFlush(dsply);
      XmStringFree(simpleString); 
      
      /* Stuff the PLACEHOLDER entry of our colormap with this new value */
      pixelColor= (65535.0/256.0);
      if(hiColor)
	{/* hi */
	  int thisI;
	  color_defs[PLACEHOLDER].pixel =  PLACEHOLDER;
	  color_defs[PLACEHOLDER].red   =  ((int)(redLevel) * pixelColor);
	  color_defs[PLACEHOLDER].green =  ((int)(greenLevel) * pixelColor);
	  color_defs[PLACEHOLDER].blue  =  ((int)(blueLevel) * pixelColor);
	  color_defs[PLACEHOLDER].flags = DoRed | DoGreen | DoBlue;
	}/* hi */
      else
	{/* lo */
	  color_defs[pixels_return[PLACEHOLDER]].pixel = 
	    pixels_return[PLACEHOLDER];
	  color_defs[pixels_return[PLACEHOLDER]].red   = 
	    ((int)(redLevel) * pixelColor);
	  color_defs[pixels_return[PLACEHOLDER]].green =
	    ((int)(greenLevel) * pixelColor);
	  color_defs[pixels_return[PLACEHOLDER]].blue  = 
	    ((int)(blueLevel) * pixelColor);
	  color_defs[pixels_return[PLACEHOLDER]].flags = 
	    DoRed | DoGreen | DoBlue;
	}/* lo */
      if(masterProcess)
	{
	  if(hiColor)
	    {
	      int thisI;
	      XAllocColor (dsply,color_map, &color_defs[PLACEHOLDER]);
	      pixels_return[PLACEHOLDER] = color_defs[PLACEHOLDER].pixel;
	      XSetForeground(dsply, image_gcPLACEHOLDER, 
			     pixels_return[PLACEHOLDER]);
	      /* DO this ??? */
	      for(thisI = 0; thisI < 32;thisI++)
		XDrawLine(dsply, test_color_window1, image_gcPLACEHOLDER,
			  0,thisI,
			  32,thisI);
	    }
	  else
	    XStoreColor (dsply,color_map,
			 &color_defs[pixels_return[PLACEHOLDER]]);
	}

      if(colorType==USEBUILD)
	{/* Update the colormap based on this new color */
	  colormap_ramp_activate(0,0,0);
	}/* Update the colormap based on this new color */
    }/* Give feedback */

}/* colorScrollCHANGE */




/* @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
 * colorClick() - Every time the mouse button is clicked call
 * here.  This routine handles the selecting/deselecting of
 * points on the screen.
 */
void colorClick(Widget widget, int *tag, 
		XmDrawingAreaCallbackStruct *draw_data)
{/* COLORCLICK */
  XEvent event;
  int x, y, i, j;
  char ans;
  static int firstX, firstY, lastY;

  if(colorType==USEBUILD)
    {/* Only do if we're building */
      /* Do appropriate depending on event type */
      if(draw_data != 0)
	{/* Only look at events meant for us */
	  event= *(draw_data->event);
	  
	  switch (event.type) 
	    {/* Find out which event we have */
	    case ButtonRelease:
	      
	      
	      x = (int)(event.xbutton.x);  /* Get x,y copords of point pos */
	      y = (int)(event.xbutton.y);
	      
	      /* Released Button1 means they are done picking
	       * points of the range */
	      if((event.xbutton.button == Button1))
		{/* Left most button */
		  /* This y is the last one they want */
		  lastBuildY= y;
		  
		  /* Now convert the Y's into colormap space */
		  firstBuildY = ((tryColors) - firstBuildY);
		                         /* We display from bottom up */
		  lastBuildY = ((tryColors) - lastBuildY);      
		  firstBuildY = MIN((tryColors-1),MAX(firstBuildY,7));
		  lastBuildY = MIN((tryColors-1),MAX(lastBuildY,7));
		  if(firstBuildY > lastBuildY)
		    {/* switch them */
		      int temp;
		      temp = lastBuildY;
		      lastBuildY= firstBuildY;
		      firstBuildY= temp;
		    }/* switch them */

		  colormap_ramp_activate(0,0,0);
		  if(hiColor)
		    {
		      if(fileType==PACKEDPORTABLE)
			stuffPackedImage(purdueHeader.numCols,
					 purdueHeader.numRows,
					 packImg,purdueHeader.nimg,
					 imgNum);  /* Start at zero'th image */
		      else if(fileType==MAP)
			stuffMapImage((float *)map_image_buffer, 
				      (float *)float_image_buffer1, 
				      file_width,file_height,
				      nsecs,
				      imgNum, 0);  /* Start at zero'th image */
		      else
			virtual_read_activate(-99,-99);
		      DrawWindow();
		      histogram_activate(&minP, &maxP);
		    }
		  XFlush (dsply);
		}/* Left most button */
	      
	      break;
	      
	    case ButtonPress:
	      
	      x = (int)(event.xbutton.x);  /* Get x,y coords of point pos */
	      y = (int)(event.xbutton.y);
	      
	      /* They are picking the firsrt point in the range */
	      if((event.xbutton.button == Button1))
		{/* Button 1 */
		  firstX= x;
		  firstBuildY= y;
		  firstX = MIN(MAX(firstX,0),255);
		  firstBuildY = MIN(MAX(firstBuildY,0),255);
		  
		  /* Save the old colormap */
		  pushColormap();
		  /* Restore histogram */
		  histogram_activate(&minP, &maxP);
		  
		  /* Show them where they've started */
		  i= MIN((tryColors-1),MAX((tryColors-firstBuildY),7));
		  if(hiColor)
		    {/* hi */
		      color_defs[i].pixel = i;
		      color_defs[i].red   = 
			((int)(redLevel) * pixelColor);
		      color_defs[i].green =
			((int)(greenLevel) * pixelColor);
		      color_defs[i].blue  = 
			((int)(blueLevel) * pixelColor);
		      color_defs[i].flags = 
			DoRed | DoGreen | DoBlue;
		    }/* hi */
		  else
		    {/* lo */
		      color_defs[pixels_return[i]].pixel = pixels_return[i];
		      color_defs[pixels_return[i]].red   = 
			((int)(redLevel) * pixelColor);
		      color_defs[pixels_return[i]].green =
			((int)(greenLevel) * pixelColor);
		      color_defs[pixels_return[i]].blue  = 
			((int)(blueLevel) * pixelColor);
		      color_defs[pixels_return[i]].flags = 
			DoRed | DoGreen | DoBlue;
		    }/* lo */
		  if(masterProcess)
		    {
		      if(hiColor)
			{XAllocColor (dsply, color_map, &color_defs[i]);
			 pixels_return[i] = color_defs[i].pixel;}
		      else
			XStoreColor (dsply, color_map, 
				     &color_defs[pixels_return[i]]);
		    }
		  
		  /* Make changes show up in histgram window */
		  /* Now draw in image - the mirror image of normal */
		  i=(firstBuildY-1);
		  for(j=0;j<32;j++)
		    {
		      if(hiColor)
			XPutPixel(hist_color_image1,
				  j,i,
				  pixels_return[RED]);
		      else
			hist_color_buffer1[((i*32)+j)]= 
			  pixels_return[RED];
		    }
		  for(j=0;j<calcHistLength(i);j++)
		    {
		      if(hiColor)
			XPutPixel(hist_color_image1,
				  j,i,
				  pixels_return[WHITE]);
		      else
			hist_color_buffer1[((i*32)+j)]= 
			  pixels_return[WHITE];
		    }
		  
		  if(XtIsManaged(colorScreenWidget))
		    XPutImage(dsply, hist_color_window1, hist_color_gc, 
			      hist_color_image1, 0, 0,
			      0, 0,  32,  256);
		  XFlush (dsply);
		  
		}/* Button 1 */
	      
	      else if((event.xbutton.button == Button2))
		{/* Pop this puppy */
		  /* They want to undo */
		  popColormap();
		}/* Pop this puppy */
	      

	      break;
	      
	    case KeyPress:
	      ans= XLookupKeysym((XKeyEvent *)&event, 0);
	      printf ("Keyed[%c]\n",ans);
	      
	      break;
	      
	    default:
	      break;
	    }/* Find out which event we have */
	}/* Only look at events meant for us */
      
    }/* Only do if we're building */
  else if(colorType==USEFITTO)
    {/* Try and make a better contrast */
      /* Do appropriate depending on event type */
      if(draw_data != 0)
	{/* Only look at events meant for us */
	  event= *(draw_data->event);
	  
	  switch (event.type) 
	    {/* Find out which event we have */
	    case ButtonRelease:
	      
	      
	      x = (int)(event.xbutton.x);  /* Get x,y copords of point pos */
	      y = (int)(event.xbutton.y);
	      
	      /* They are done picking points they want
	       * in the range for the new contrast colormap */
	      if((event.xbutton.button == Button1))
		{/* Left most button */
		  int min,max;
		  float scaleFactor;

		  /* This y is the last in the range */
		  lastY= y;
		  
		  /* Map this to colormap space */
		  firstY = ((255) - firstY);
		                         /* We display from bottom up */
		  lastY = ((255) - lastY);      
		  firstY = MIN((255-1),MAX(firstY,7));
		  lastY = MIN((255-1),MAX(lastY,7));
		  if(firstY > lastY)
		    {/* switch them */
		      int temp;
		      temp = lastY;
		      lastY= firstY;
		      firstY= temp;
		    }/* switch them */
		  
		  
		  /* Now that we know the range let's expand our 
		   * colormap accordingly */
		  pixelColor= (65535.0/256.0);
		  
		  if(hiColor)
		    {
		      min = MAX(firstY,7);
		      max = MIN(lastY,255);
		    }
		  else
		    {
		      min = MAX(firstY,(7+(pixels_return[0])));
		      max = MIN(lastY,(255));
		    }
		  if(min==max)
		    max++;
		  scaleFactor = 255.0 / ((float)(max-min));  
		  /*      scaleFactor = ((float)(max-min))  /  256.0 ; */
		  
		  /* Make these entries in the colormap GRAYSCALE */
		  for(i=min;i<=(max);i++)
		    {/* Do the colors we have been assigned */
		      color_defs[i].pixel = i;
		      color_defs[i].red   = 
			((i-min) * (scaleFactor * pixelColor));
		      color_defs[i].green =
			((i-min) * (scaleFactor * pixelColor));
		      color_defs[i].blue  = 
			((i-min) * (scaleFactor * pixelColor));
		      color_defs[i].flags = 
			DoRed | DoGreen | DoBlue;
		      if(masterProcess)
			{
			  if(hiColor)
			    {XAllocColor (dsply, color_map, &color_defs[i]);
			     pixels_return[i] = color_defs[i].pixel;}
			  else
			    XStoreColor (dsply, color_map, &color_defs[i]);
			}
		    }/* Do the colors we have been assigned */
		  
		  if(hiColor)
		    {
		      if(fileType==PACKEDPORTABLE)
			stuffPackedImage(purdueHeader.numCols,
					 purdueHeader.numRows,
					 packImg,purdueHeader.nimg,
					 imgNum);  /* Start at zero'th image */
		      else if(fileType==MAP)
			stuffMapImage((float *)map_image_buffer, 
				      (float *)float_image_buffer1, 
				      file_width,file_height,
				      nsecs,
				      imgNum, 0);  /* Start at zero'th image */
		      else
			virtual_read_activate(-99,-99);
		      DrawWindow();
		      histogram_activate(&minP, &maxP);
		    }
		  XFlush (dsply);
		}/* Left most button */
	      
	      break;
	      
	    case ButtonPress:
	      
	      x = (int)(event.xbutton.x);  /* Get x,y coords of point pos */
	      y = (int)(event.xbutton.y);
	      
	      if((event.xbutton.button == Button1))
		{/* Button 1 */
		  /* This is the first point in the range for
		   * the new colormap */
		  firstX= x;
		  firstY= y;
		  firstX = MIN(MAX(firstX,0),255);
		  firstY = MIN(MAX(firstY,0),255);
		  /* Save the old colormap */
		  pushColormap();
		  /* Restore histogram */
		  histogram_activate(&minP, &maxP);

		}/* Button 1 */
	      
	      else if((event.xbutton.button == Button2))
		{/* Pop this puppy */
		  /* They've changed their minds */
		  popColormap();
		}/* Pop this puppy */
	      

	      break;
	      
	    case KeyPress:
	      ans= XLookupKeysym((XKeyEvent *)&event, 0);
	      printf ("Keyed[%c]\n",ans);
	      
	      break;
	      
	    default:
	      break;
	    }/* Find out which event we have */
	}/* Only look at events meant for us */
      
    }/* Try and make a better contrast */

  XFlush(dsply);            /* Make changes current */
  
}/* COLORCLICK */



/* ************************************************************
 * colormapValue_create() - Remember this widget for use later.
 * This widget is either the RED GREEN or BLUE scrollbar 
 * value indicator on the Alter Colromap screen.
 * We need to remember the widget number so we can update
 * it when the value of the scrollbar changes.
*/
void colormapValue_create(Widget widget, int *tag, 
			  XmAnyCallbackStruct *callback_data)
{/* COLORMAPVALUE_CREATE */
  
  /* Remember the widget number for later */
  if(*tag==RADIUS)
    radialValueWidget= widget;
  else if(*tag==STEPCM)
    stepCMValueW= widget;
  else if(*tag==NOCONTRAST)
    noContrastW= widget;
  else if(*tag==USECONTRAST)
    useContrastW= widget;
  else if(*tag==USEHISTEQUAL)
    histEqualW= widget;
  else
    colormapValueWidget[*tag]= widget;
  
}/* COLORMAPVALUE_CREATE */




/* @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
 * colormapMotion() - Whenever the pointer moves across our 
 * Alter Colormap drawing area(s) call here.  This rotuine
 * basically takes care of drawing that little black line 
 * you see when the cursor is dragged across the colormap
 * areas.
*/
void colormapMotion(Widget widget, caddr_t unused, 
		    register XMotionEvent *pEvent)
{/* colormapMotion */
  XmString simpleString;
  int x, y,lastDraw;
  static int lastY= -999;
  static int lastI= -999;
  static short lastRed,lastGreen,lastBlue;
  int i, j, thisI;

  /* Make all changes current */
  XFlush(dsply);               

  switch (pEvent->type) 
    {/* Find out which event we have */
	  
    case LeaveNotify:

      /* They've been here before redraw the old position
       * in the "normal" way so that we only see the current
       * position exhibited on the screen. */
      if(lastY > 0)
	{/* Let's redisplay the OLD position */
	  /* Make changes show up in histgram window */
	  thisI=(lastY);
	  /* Redraw the histogram window */
	  for(j=0;j<32;j++)
	    {
	      if(hiColor)
		XPutPixel(hist_color_image1,
			  j,thisI,
			  pixels_return[WHITE]);
	      else
		hist_color_buffer1[((thisI*32)+j)]= 
		  pixels_return[WHITE];
	    }
	  lastDraw = calcHistLength(thisI);
	  for(j=0;j<lastDraw;j++)
	    {
	      if(hiColor)
		XPutPixel(hist_color_image1,
			  j,thisI,
			  pixels_return[RED]);
	      else
		hist_color_buffer1[((thisI*32)+j)]= 
		  pixels_return[RED];
	    }
	  
	  /* Redraw the colromap window */
	  for(j=0;j<32;j++)
	    {
	      if(hiColor)
		XPutPixel(color_image1,
			  j,
			  thisI,
			  pixels_return[(256-thisI)]);
	      else
		color_buffer1[((thisI*32)+j)]=  (256-thisI);
	    }

	  /* Put them on the screen */
	  XPutImage(dsply, hist_color_window1, hist_color_gc, 
		    hist_color_image1, 0, lastY,
		    0, lastY,  32,  1);
	  XPutImage(dsply, color_window1, color_gc, 
		    color_image1, 0, lastY,
		    0, lastY,  32,  1);
	}/* Let's redisplay the OLD position */
      if(lastI > 0 && !buildingCmap)
	{/* Redo last ones */

	  color_defs[lastI].red= lastRed;
	  color_defs[lastI].green= lastGreen;
	  color_defs[lastI].blue= lastBlue;
	  if(masterProcess)
	    {
	      if(hiColor)
		{XAllocColor (dsply, color_map, &color_defs[lastI]);
		 pixels_return[lastI] = color_defs[lastI].pixel;}
	      else
		XStoreColor (dsply, color_map, &color_defs[lastI]);
	    }
	}/* Redo last ones */
      lastY= -999;
      lastI= -999;
      
      break;

    case MotionNotify:
	      
      /* Get coords of pointer */
      x = (int)(pEvent->x);     
      y = (int)(pEvent->y);
      
      /* Show them where they've started */
      y = MAX(y,0);
      i= MIN(tryColors,MAX((tryColors-y),7));
      if(hiColor)
	{/* DO nothing? */  
	  i= MIN(255,MAX((255-y),7));
	}/* DO nothing? */  
      else
	i += pixels_return[0]; 

      /* Tell them what position in the colormap we are */
      sprintf(outStr,"%3d",i);
      createSimpleString(&simpleString,outStr);
      XtSetArg(arglst[0], XmNlabelString, simpleString);
      XtSetArg(arglst[1], XmNrecomputeSize, FALSE );
      XtSetValues(colormapValueWidget[ENTRY], arglst, 2); 
      XFlush(dsply);
      XmStringFree(simpleString); 

      if(radialMapShade)
	{/* Tell them what radius we are */
	  if(i<=pixels_return[7])
	    sprintf(outStr,"%3d",0);
	  else sprintf(outStr,"%.1f",radialValue[i]);
	  createSimpleString(&simpleString,outStr);
	  XtSetArg(arglst[0], XmNlabelString, simpleString);
	  XtSetArg(arglst[1], XmNrecomputeSize, FALSE );
	  XtSetValues(radialValueWidget, arglst, 2); 
	  XFlush(dsply);
	  XmStringFree(simpleString); 
	}/* Tell them what radius we are */

      
      /* They've been here before redraw the old position
       * in the "normal" way so that we only see the current
       * position exhibited on the screen. */
      if(lastY > 0)
	{/* Let's redisplay the OLD position */
	  /* Make changes show up in histgram window */
	  thisI=(lastY);
	  /* Redraw the histogram window */
	  for(j=0;j<32;j++)
	    {
	      if(hiColor)
		XPutPixel(hist_color_image1,
			  j, thisI,
			  pixels_return[WHITE]);
	      else
		hist_color_buffer1[((thisI*32)+j)]= 
		  pixels_return[WHITE];
	    }
	  lastDraw = calcHistLength(thisI);
	  for(j=0;j<lastDraw;j++)
	    {
	      if(hiColor)
		XPutPixel(hist_color_image1,
			  j, thisI,
			  pixels_return[RED]);
	      else
		hist_color_buffer1[((thisI*32)+j)]= 
		  pixels_return[RED];
	    }
	  /* Redraw the colromap window */
	  for(j=0;j<32;j++)
	    {
	      if(hiColor)
		XPutPixel(color_image1,
			  j, thisI,
			  pixels_return[(256-thisI)]);
	      else
		color_buffer1[((thisI*32)+j)]=  (256-thisI);
	    }
	  /* Put them on the screen */
	  XPutImage(dsply, hist_color_window1, hist_color_gc, 
		    hist_color_image1, 0, lastY,
		    0, lastY,  32,  1);
	  XPutImage(dsply, color_window1, color_gc, 
		    color_image1, 0, lastY,
		    0, lastY,  32,  1);
	}/* Let's redisplay the OLD position */

      if(y > 0)
	{/* Let's display the NEW position */

	  /* Make changes show up in histgram window */
	  thisI=(y * 32);
	  for(j=0;j<32;j++)
	    {/* SHo them where it's at */
	      if(hiColor)
		{
		  XPutPixel(hist_color_image1,
			    j, y,
			    pixels_return[BLACK]);
		  XPutPixel(color_image1,
			    j, y,
			    pixels_return[BLACK]);
		}
	      else
		{
		  hist_color_buffer1[((thisI)+j)]= 
		    pixels_return[BLACK];
		  color_buffer1[((thisI)+j)]= 
		    pixels_return[BLACK];
		}
	    }/* SHo them where it's at */
	  
	  /* Put changes on the screen */
	  XPutImage(dsply, hist_color_window1, hist_color_gc, 
		    hist_color_image1, 0, y,
		    0, y,  32,  1);
	  XPutImage(dsply, color_window1, color_gc, 
		    color_image1, 0, y,
		    0, y,  32,  1);

	  /* Remeber what we've done for next time */
	  lastY= y;

	  if(lastI > 0 && !buildingCmap)
	    {/* Redo last ones */
	      color_defs[lastI].red= lastRed;
	      color_defs[lastI].green= lastGreen;
	      color_defs[lastI].blue= lastBlue;
	      if(masterProcess)
		{
		  if(hiColor)
		    {XAllocColor (dsply, color_map, &color_defs[lastI]);
		     pixels_return[lastI] = color_defs[lastI].pixel;}
		  else
		    XStoreColor (dsply, color_map, &color_defs[lastI]);
		}
	    }/* Redo last ones */
	  lastI = i;
	  lastRed = color_defs[i].red;
	  lastGreen = color_defs[i].green;
	  lastBlue = color_defs[i].blue;
	  color_defs[i].red    = 65000;
	  color_defs[i].green  = 0;
	  color_defs[i].blue   = 0;
	  if(masterProcess)
	    {
	      if(hiColor)
		{XAllocColor (dsply, color_map, &color_defs[i]);
		 pixels_return[i] = color_defs[i].pixel;}
	      else
		XStoreColor (dsply, color_map, &color_defs[i]);
	    }
	  buildingCmap = FALSE;  /* Done by now .. */
	}/* Let's display the NEW position */

      break;
	  
    default:
      break;
    }/* Find out which event we have */
  
  XFlush(dsply);              /* Make changes current */
  
}/* colormapMotion */





/**********************************************************
 * contrast_enhance() - Try and stretch the colormap so that
 * we are looking at the maximum contrast across the range of
 * pixels of the image.
*/
void contrast_enhance(int min, int max)
{/* CONTRAST_ENHANCE */
  int i,j,k,l,y,x;
  float scaleFactor;

  /* Now that we know the range let's expand our 
   * colormap accordingly */
  pixelColor= (65535.0/256.0);
  
  if(hiColor)
    {
      min = MAX(min,7);
      max = MIN(max,255);
    }
  else
    {
      min = MAX(min,(7+(pixels_return[0])));
      max = MIN(max,(255));
    }
  if(min==max)
    max= min+1;
  scaleFactor = 255.0 / ((float)(max-min));  
  /*      scaleFactor = ((float)(max-min))  /  256.0 ; */
  
  /* Make these entries in the colormap GRAYSCALE */
  for(i=min;i<=(max);i++)
    {/* Do the colors we have been assigned */
      color_defs[i].pixel = i;
      color_defs[i].red   = 
	((i-min) * (scaleFactor * pixelColor));
      color_defs[i].green =
	((i-min) * (scaleFactor * pixelColor));
      color_defs[i].blue  = 
	((i-min) * (scaleFactor * pixelColor));
      color_defs[i].flags = 
	DoRed | DoGreen | DoBlue;
      if(masterProcess)
	{
	  if(hiColor)
	    {XAllocColor (dsply, color_map, &color_defs[i]);
	     pixels_return[i] = color_defs[i].pixel;}
	  else
	    XStoreColor (dsply, color_map, &color_defs[i]);
	}
    }/* Do the colors we have been assigned */
  
}/* CONTRAST_ENHANCE */





/*-----------------------------------*/
/*************************************
 * OPTION screen callbacks           *
 *************************************/
/*-----------------------------------*/
/* ************************************************************
 * createTog() - Remeber the toggle widget number for
 * setting the toggle later.
 */
void createTog(Widget widget, int *tag, XmAnyCallbackStruct *callback_data)
{/* CREATETog */
  Widget tempW;

  if(*tag==USEGENERIC)
    tempW= widget;
  else if(*tag==USEFITTO)
    tempW= widget;
  else if(*tag==USERANDOM)
    tempW= widget;
  else if(*tag==USEBUILD)
    tempW= widget;
  else if(*tag==USEZOOM)
    tempW= widget;
  else if(*tag==USEBOX)
    tempW= widget;
  else if(*tag==USEPOINT)
    tempW= widget;
  else if(*tag==READCM)
    tempW= widget;
  else if(*tag==WRITECM)
    tempW= widget;
  else if(*tag==STEPCM)
    tempW= widget;
  else if(*tag==USEREVERSE)
    tempW= widget;
  else if(*tag==IMGNUMBUTTON)
    imgNumButtonW= widget;
  else if(*tag==DRAWLITEBUTTON)
    drawLiteButtonW= widget;
  else if(*tag==DRAWFFTLITEBUTTON)
    drawFFTLiteButtonW= widget;
  else if(*tag==DRAWFFT2LITEBUTTON)
    drawFFT2LiteButtonW= widget;

}/* CREATETog */


/* *******************************************************
 * toggle_change() - Whenever the toggle state is changed
 * call here.  Callback for the colormap mode changes
 * and also the lock on the main screen.
 */
void toggle_change(Widget widget, int *tag, 
		   XmToggleButtonCallbackStruct *callback_data)
{/* toggle_change */
  int i,y,x;
  int p,min,max;
  float scaleFactor,randNum,factor;
  char valStr[32];
  XmString simpleString;
  colorType = *tag;

  if(*tag==USEGENERIC)
    {/* They want to use the GENERIC colormap model */
      i= *tag;
      for(i=7;i<(tryColors);i++)
	{/* Do the colors we have been assigned */
	  if(hiColor)
	    {/* hi */
	      color_defs[i].pixel = 
		gen_color_defs[i].pixel;
	      color_defs[i].red   = 
		gen_color_defs[i].red;
	      color_defs[i].green = 
		gen_color_defs[i].green;
	      color_defs[i].blue  = 
		gen_color_defs[i].blue;
	      color_defs[i].flags = 
		gen_color_defs[i].flags;
	    }/* hi */
	  else
	    {/* lo */
	      color_defs[pixels_return[i]].pixel = 
		gen_color_defs[pixels_return[i]].pixel;
	      color_defs[pixels_return[i]].red   = 
		gen_color_defs[pixels_return[i]].red;
	      color_defs[pixels_return[i]].green = 
		gen_color_defs[pixels_return[i]].green;
	      color_defs[pixels_return[i]].blue  = 
		gen_color_defs[pixels_return[i]].blue;
	      color_defs[pixels_return[i]].flags = 
		gen_color_defs[pixels_return[i]].flags;
	    }/* lo */
	  if(masterProcess)
	    {
	      if(hiColor)
		{XAllocColor (dsply, color_map, &color_defs[i]);
		 pixels_return[i] = color_defs[i].pixel;}
	      else
		XStoreColor (dsply, color_map, &color_defs[pixels_return[i]]);
	    }
	}/* Do the colors we have been assigned */
    }/* They want to use the GENERIC colormap model  */
  else if(*tag==USEREVERSE)
    {/* They want to use the REVERSE Contrast colormap model */
      i= *tag;
      for(i=7;i<(tryColors);i++)
	{/* Do the colors we have been assigned */
	  if(hiColor)
	    {/* hi */
	      color_defs[i].pixel = i;
	      color_defs[i].red   = 
		gen_color_defs[((tryColors-i)+7)].red;
	      color_defs[i].green = 
		gen_color_defs[((tryColors-i)+7)].green;
	      color_defs[i].blue  = 
		gen_color_defs[((tryColors-i)+7)].blue;
	      color_defs[i].flags = 
		gen_color_defs[((tryColors-i)+7)].flags;
	    }/* hi */
	  else
	    {/* lo */
	      color_defs[pixels_return[i]].pixel = 
		gen_color_defs[pixels_return[i]].pixel;
	      color_defs[pixels_return[i]].red   = 
		gen_color_defs[pixels_return[((tryColors-i)+5)]].red;
	      color_defs[pixels_return[i]].green = 
		gen_color_defs[pixels_return[((tryColors-i)+5)]].green;
	      color_defs[pixels_return[i]].blue  = 
		gen_color_defs[pixels_return[((tryColors-i)+5)]].blue;
	      color_defs[pixels_return[i]].flags = 
		gen_color_defs[pixels_return[((tryColors-i)+5)]].flags;
	    }/* lo */
	  if(masterProcess)
	    {
	      if(hiColor)
		{XAllocColor (dsply, color_map, &color_defs[i]);
		 pixels_return[i] = color_defs[i].pixel;}
	      else
		XStoreColor (dsply, color_map, &color_defs[pixels_return[i]]);
	    }
	}/* Do the colors we have been assigned */
    }/* They want to use the REVERSE Contrast colormap model  */
  else if(*tag==USEFITTO)
    {/* They want to fit colormap to image */
      i= *tag;
      /* First calculate min and max */
      min= 65500;
      max= 0;
      for(y=0;y<image_height;y++)
	{
	  for(x=0;x<image_width;x++)
	    {/* Visit each pixel */
	      p= image_buffer1[((y*buffer_width)+x)];
	      if(p < min)
		min=p;
	      if(p > max)
		max=p;
	    }/* Visit each pixel */
	}
      contrast_enhance(min, max);

    }/* They want to fit colromap to image */
  else if(*tag==USERANDOM)
    {/* They want a randomized colormap */
      pixelColor= (65535.0/256.0);
      
      /* Get our fraction of the random number */
      factor= 1.0 / DENOMINATOR;
      
      /* Make these entries in the colormap GRAYSCALE */
      for(i=7;i<(tryColors);i++)
	{/* Do the colors we have been assigned */
	  if(hiColor)
	    {/* hi */
	      color_defs[i].pixel = i;

	      randNum= random() * factor;

	      color_defs[i].red   = 
		((int)(randNum * 255) * pixelColor);

	      randNum= random() * factor;

	      color_defs[i].green =
		((int)(randNum * 255) * pixelColor);

	      randNum= random() * factor;

	      color_defs[i].blue  = 
		((int)(randNum * 255) * pixelColor);
	      color_defs[i].flags = DoRed | DoGreen | DoBlue;
	    }/* hi */
	  else
	    {/* lo */
	      color_defs[pixels_return[i]].pixel = pixels_return[i];

	      randNum= random() * factor;

	      color_defs[pixels_return[i]].red   = 
		((int)(randNum * 255) * pixelColor);

	      randNum= random() * factor;

	      color_defs[pixels_return[i]].green =
		((int)(randNum * 255) * pixelColor);

	      randNum= random() * factor;

	      color_defs[pixels_return[i]].blue  = 
		((int)(randNum * 255) * pixelColor);
	      color_defs[pixels_return[i]].flags = DoRed | DoGreen | DoBlue;
	    }/* lo */
	  if(masterProcess)
	    {
	      if(hiColor)
		{XAllocColor (dsply, color_map, &color_defs[i]);
		 pixels_return[i] = color_defs[i].pixel;}
	      else
		XStoreColor (dsply, color_map, &color_defs[pixels_return[i]]);
	    }
	}/* Do the colors we have been assigned */
      
    }/* They want a randomized colormap */
  else if(*tag==USEBUILD)
    {/* They want to build their own colormap */
      pushColormap();
    }/* They want to build their own colormap */
  else if(*tag==READCM)
    {/* They want to read their own colormap */
      readWrite= READCM;
      filterFiles("*.cmp",READCM); 
      raiseIt(fileSel_widget);
    }/* They want to read their own colormap */
  else if(*tag==WRITECM)
    {/* They want to write their own colormap */
      readWrite= WRITECM;
      filterFiles("*.cmp",WRITECM); 
      raiseIt(fileSel_widget);
    }/* They want to write their own colormap */
  else if(*tag==STEPCM)
    {/* They want to step the colormap */
    }/* They want to step the colormap */
  else if(*tag==USELOCK)
    {/* They want to lock their Box size  */
      locked = !locked;
      /* Set the imgNumTextWidget to (un)editable */
      XtVaSetValues(imgNumTextWidget,
		    XmNeditable, !(locked),
			NULL);

      /* Must reset label of the toggle to reflect whether we're
       * specifying the first box or in find box mode.  Also reset the
       * text widget for either the firstBox value or the find Box value
       */
      if(locked)
	{createSimpleString(&simpleString,"Find Box:");}
      else 
	{createSimpleString(&simpleString,"FirstBox:");
	 findBox= -99; /* Start again */}
      
      XtVaSetValues(boxInitLabelW,
		    XmNlabelString, simpleString,
		    NULL);
      XFlush(dsply);
      XmStringFree(simpleString); 
      sprintf(valStr,"%d",lastNumBox+1);
      XmTextSetString(boxInitWidget,valStr);
      
      DrawWindow();
    }/* They want to lock their box size  */
  else if(*tag==FINDNEAREST)
    {/* Find nearest particle */
      findNearest = !findNearest;
      if(findNearest)
	icosaMotioning = FALSE;
      else icosaMotioning = TRUE;
    }/* Find nearest particle */
  else if(*tag==DRAWICOS)
    {/* Draw icosahedron */
      drawIcos = !drawIcos;
    }/* Draw icosahedron */
  else if(*tag==USEOMEGA)
    {/* use Omega */
      useOmega = !useOmega;
    }/* use Omega */
  else if(*tag==NEWXY)
    {/* find new x,y center */
      newXY = !newXY;
      if(newXY)
	{/* Set new xy cursor */
	  if(boxZoom==USEZOOM)
	    old_newXY_cursor = zoomCursor;
	  else if(boxZoom==USEBOX)
	    old_newXY_cursor = cross_cursor;
	  else if(boxZoom==USEPOINT)
	    old_newXY_cursor = None;

	  /* Put up newXY cursor */
	  old_cursor = bogosity_cursor;
	  attrs.cursor= bogosity_cursor;
	  XChangeWindowAttributes(dsply, XtWindow(drawingWidget), 
				  CWCursor, &attrs);

	}/* Set new xy cursor */
      else
	{/* Restore old cursor */
	  /* Restore the old cursor whatever it was */
	  attrs.cursor= old_newXY_cursor;
	  XChangeWindowAttributes(dsply, XtWindow(drawingWidget), 
				  CWCursor, &attrs);
	}/* Restore old cursor */
    }/* find new x,y center */
  else if(*tag==PARAMPROJ)
    {/* They want map projections */
      mapProj= !mapProj;
      if(mapProj && EQU(paramMapFilename,""))
	{/* Can't go on */
	  mapProj= FALSE;
	  XtSetArg(arglst[0], XmNset, FALSE);
	  XtSetValues(mapProjW, arglst, 1);
	  askThem("No MAP open can't do projection.",
		  FALSE,XmDIALOG_INFORMATION);
	}/* Can't go on */
      else 
	{/* Do it */

	  if(mapProj)
	    {
	    /*	    askThem(
		    "MAP projections may be 90 degrees rotated in omega.\n\
		    We're working on getting this resolved.",
		    FALSE,XmDIALOG_INFORMATION); */

	      if(haveParamFile)
		stuffPackedImage(purdueHeader.numCols,
				 purdueHeader.numRows,
				 packImg,purdueHeader.nimg,
				 imgNum);
	      else
		askThem("Must read a param file first.",
			FALSE,XmDIALOG_INFORMATION); 
	    }
	  DrawWindow(); 
	}/* Do it */
    }/* They want map projections */
  else if(*tag==PROJCONTOUR)
    {/* They want map projections contour */
      mapProjContour= !mapProjContour;

      if(EQU(paramMapFilename,""))
	{/* Can't go on */
	  askThem("No MAP open can't contour the projection.",
		  FALSE,XmDIALOG_INFORMATION);
	}/* Can't go on */
      else
	{/* Do it */
	  stuffPackedImage(purdueHeader.numCols,
			   purdueHeader.numRows,
			   packImg,purdueHeader.nimg,
			   imgNum);
	  DrawWindow(); 
	}/* Do it */
    }/* They want map projections contour */
  else if(*tag==BUMPCONTRAST)
    {/* Do our bump up the contrast kludge */
      bumpContrast = !bumpContrast; 
      stuffPackedImage(purdueHeader.numCols,
		       purdueHeader.numRows,
		       packImg,purdueHeader.nimg,
		       imgNum);
      DrawWindow(); 
    }/* Do our bump up the contrast kludge */
  else if(*tag==TABEGX)
    taArrowX= !taArrowX;
  else if(*tag==TABEGY)
    taArrowY= !taArrowY;
  else if(*tag==TABEGZ)
    taArrowZ= !taArrowZ;
  else if(*tag==DRAWFIX2DGRID)
    {drawFix2Dgrid = !drawFix2Dgrid;
     DrawWindow();}
  else if(*tag==DRAWFIX2DSTATS)
    {drawFix2Dstats = !drawFix2Dstats;
     DrawWindow();}
  else if(*tag==FIXICOS)
    {
      fixIcosFill= !fixIcosFill;
      askThem("Icosahedral Fill not implemented yet.",
	      FALSE,XmDIALOG_INFORMATION);
      fixIcosFill = FALSE;
    }
  else if(*tag==SAVESECOND)
    saveSecondMap = !saveSecondMap;
  else if(*tag==FOURAVERSF)
    {writeFourAverSF = !writeFourAverSF;
    }
  else if(*tag==ICOSAQUARTER)
    icosaQuarter = !icosaQuarter;
  else if(*tag==REFAUTOMATE)
    refAutomate = !refAutomate;
  else if(*tag==REFCIRCAVG)
    {
      refCircAvg = !refCircAvg;
      if(refCircAvg)
	{/* FOrce dark cirlce off */
	  darkCircleRef = FALSE;
	  XtSetArg(arglst[0], XmNset, FALSE);
	  XtSetValues(darkRefToggleW, arglst, 1);
	  XtSetArg(arglst[0], XmNsensitive, TRUE);
	  XtSetValues(refAllW, arglst, 1);  
	  XFlush(dsply);
	}/* FOrce dark cirlce off */
    }			       
  else if(*tag==SHOWLOWLIM)
    {/* show lowlim? ? */
      showLowLim = !showLowLim;
    }/* showlowlim ? */
  else if(*tag==REFREADFILE)
    {/* ref boxing read file */
      refReadFile = !refReadFile;
      if(refReadFile)
	{/* get fname */
	  readType= REFREADFILE;
	  readWrite= READING;
	  raiseIt(fileSel_widget);
	}/* get fname */
      else
	{/* no file */
	  sprintf(outStr,"No file.");
	  createSimpleString(&simpleString,outStr);
	  XtSetArg(arglst[0], XmNlabelString, simpleString );
	  XtSetArg(arglst[1], XmNrecomputeSize, TRUE );
	  XtSetValues(refReadFileLabelW, arglst, 2); 
	  XFlush(dsply);
	  XmStringFree(simpleString); 
	}/* no file */
    }/* ref boxing read file */
  else if(*tag==USEAVGCORAVG)
    {/* use the avg for coravg ? */
      int i,j,ncolmp,nrowmp;
      useAvgCorAvg = !useAvgCorAvg;
      ncolmp = emcoravgNcol;
      nrowmp = emcoravgNrow;
      if(useAvgCorAvg)
	for(i=0;i<nrowmp;i++)
	  for(j=0;j<ncolmp;j++)
	    emcoravg_image_buffer1[(i*ncolmp)+j] = 
	      (BYTE)emcoravg_image_buffer_avg[(i*ncolmp)+j];
      else
	for(i=0;i<nrowmp;i++)
	  for(j=0;j<ncolmp;j++)
	    emcoravg_image_buffer1[(i*ncolmp)+j] = 
	      (BYTE)emcoravg_image_buffer_disk[(i*ncolmp)+j];
      stuffEmcoravgImage();
    }/* use the avg for coravg ? */
  else if(*tag==SHOWRADII)
    {
      showRadii = !showRadii;
      if(diffSection)
	diffSec_really_doit();
    }
  else if(*tag==THRESHDOIT)
    {
      threshDoit = !threshDoit;
      if(diffSection)
	diffSec_really_doit();
    }
  else if(*tag==THRESHABS)
    {
      threshAbs = !threshAbs;
      if(diffSection)
	diffSec_really_doit();
    }
  else if(*tag==SHIFTORIGX)
    {
      deltaPhaseShift = !deltaPhaseShift;
      if(deltaPhaseShift)
	{/* Only allow delta */
	  XtSetArg(arglst[0], XmNsensitive, TRUE );
	  XtSetValues(shiftOrigXW, arglst, 1); 
	  XtSetValues(shiftOrigYW, arglst, 1); 
	  XtSetValues(shiftOrigXlabelW, arglst, 1); 
	  XtSetValues(shiftOrigYlabelW, arglst, 1); 
	  XtSetArg(arglst[0], XmNsensitive, FALSE );
	  XtSetValues(moveOrigXW, arglst, 1); 
	  XtSetValues(moveOrigYW, arglst, 1); 
	  XtSetValues(moveOrigXlabelW, arglst, 1); 
	  XtSetValues(moveOrigYlabelW, arglst, 1); 
	  XtSetValues(phasYdownW, arglst, 1); 
	  XtSetValues(phasYupW, arglst, 1); 
	  XtSetValues(phasXdownW, arglst, 1); 
	  XtSetValues(phasXupW, arglst, 1); 
	}/* Only allow delta */
      else
	{/* Only allow change of real value */
	  XtSetArg(arglst[0], XmNsensitive, FALSE );
	  XtSetValues(shiftOrigXW, arglst, 1); 
	  XtSetValues(shiftOrigYW, arglst, 1); 
	  XtSetValues(shiftOrigXlabelW, arglst, 1); 
	  XtSetValues(shiftOrigYlabelW, arglst, 1); 
	  XtSetArg(arglst[0], XmNsensitive, TRUE );
	  XtSetValues(moveOrigXW, arglst, 1); 
	  XtSetValues(moveOrigYW, arglst, 1); 
	  XtSetValues(moveOrigXlabelW, arglst, 1); 
	  XtSetValues(moveOrigYlabelW, arglst, 1); 
	  XtSetValues(phasYdownW, arglst, 1); 
	  XtSetValues(phasYupW, arglst, 1); 
	  XtSetValues(phasXdownW, arglst, 1); 
	  XtSetValues(phasXupW, arglst, 1); 
	}/* Only allow change of real value */

    }

  if(hiColor &&
     ((*tag==USEGENERIC) ||
     (*tag==USEREVERSE) ||
     (*tag==USEFITTO) ||
     (*tag==USERANDOM) ||
     (*tag==USEBUILD))
     )
    {
      if(fileType==PACKEDPORTABLE)
	stuffPackedImage(purdueHeader.numCols,
			 purdueHeader.numRows,
			 packImg,purdueHeader.nimg,
			 imgNum);  /* Start at zero'th image */
      else if(fileType==MAP)
	stuffMapImage((float *)map_image_buffer, 
		      (float *)float_image_buffer1, 
		      file_width,file_height,
		      nsecs,
		      imgNum, 0);  /* Start at zero'th image */
      else
	virtual_read_activate(-99,-99);
      DrawWindow();
      histogram_activate(&minP, &maxP);
    }
  
  XFlush (dsply);
  
 }/* toggle_change */



/* ************************************************************
 * options_activate() - Put up the options screen.
 */
void options_activate(Widget widget, int *tag, 
		      XmAnyCallbackStruct *callback_data)
{/* options_ACTIVATE */

  sprintf(outStr,"%f",minMapPixel);
  XmTextSetString(minMapTextW,outStr);
  sprintf(outStr,"%f",maxMapPixel);
  XmTextSetString(maxMapTextW,outStr);
  raiseIt(optionScreenWidget);

}/* options_ACTIVATE */



/* ************************************************************
 * option_screen_deactivate() - Get rid of the options screen.
 */
void option_screen_deactivate(Widget widget, int *tag, 
			      XmAnyCallbackStruct *callback_data)
{/* option_screen_deactIVATE */
  char *thisStr;

  thisStr = XmTextGetString(minMapTextW);
  minMapPixel = (float)strtod(thisStr,0);
  XtFree(thisStr);
  thisStr = XmTextGetString(maxMapTextW);
  maxMapPixel = (float)strtod(thisStr,0);
  XtFree(thisStr);

  XtUnmanageChild(optionScreenWidget);
  
}/* option_screen_deactIVATE */



/* *******************************************************
 * backgroundMode_change() - They want to change the backgrounding
 * mode.  
 */
void backgroundMode_change(Widget widget, int *tag, 
			   XmAnyCallbackStruct *callback_data)
{/* backgroundMODE_CHANGE */

  backgroundMode = *tag;

}/* backgroundMODE_CHANGE */



/* *******************************************************
 * featherMode_change() - They want to change the feathering
 * mode.  
 */
void featherMode_change(Widget widget, int *tag, 
			XmAnyCallbackStruct *callback_data)
{/* featherMODE_CHANGE */

  featherMode = *tag;

}/* featherMODE_CHANGE */


/*************************************************************
 * reset_xyz_scrolls() - Reset the x,y,z min and max scrolls
 * on the 3D rendering screen.
*/
void reset_xyz_scrolls()
{/* RESET_XYZ_SCROLLS */
  int i;
  int thisMin,thisMax;
  XmString simpleString;

  for(i=0;i<3;i++)
    {/* Do cut scrolls */
      int thisDimension;
      
      if(FALSE)
	{/* Setup for high res images */
	  thisDimension = 100;
	}/* Setup for high res images */
      else
	{/* Normal */
	  if(i==0)
	    thisDimension = numCols;
	  else if (i==1)
	    thisDimension = numRows;
	  else 
	    thisDimension = origNumSecs;
	}/* Normal */

      thisMin = 0;
      thisMax = thisDimension;
      if(lowRes)
	{/* setup in case they've got other values */
	  if(i==2 && cutMin2Save>=0.0)
	    thisMin = (int)cutMin2Save;
	  if(i==2 && cutMax2Save>=0.0)
	    thisMax = (int)cutMax2Save;
	  if(i==1 && cutMin1Save>=0.0)
	    thisMin = (int)cutMin1Save;
	  if(i==1 && cutMax1Save>=0.0)
	    thisMax = (int)cutMax1Save;
	  if(i==0 && cutMin0Save>=0.0)
	    thisMin = (int)cutMin0Save;
	  if(i==0 && cutMax0Save>=0.0)
	    thisMax = (int)cutMax0Save;
	}/* setup in case they've got other values */

      XtSetArg(arglst[0], XmNmaximum, (int)((thisDimension)+4));
      XtSetArg(arglst[1], XmNminimum, (int)(0));
      XtSetArg(arglst[2], XmNsliderSize, (int)(5));
      XtSetArg(arglst[3], XmNvalue, (int)(thisMin));
      XtSetArg(arglst[4], XmNpageIncrement, (int)(10));
      XtSetValues(cutMinScrollW[i], arglst, 5);
      XFlush(dsply);
      sprintf(outStr,"%d",thisMin+1);
      createSimpleString(&simpleString,outStr);
      XtSetArg(arglst[0], XmNlabelString, simpleString);
      XtSetValues(cutMinValueW[i], arglst, 1); 
      XFlush(dsply);
      XmStringFree(simpleString); 
      XFlush(dsply);
      XtSetArg(arglst[0], XmNmaximum, (int)((thisDimension)+6));
      XtSetArg(arglst[1], XmNminimum, (int)(0));
      XtSetArg(arglst[2], XmNsliderSize, (int)(5));
      XtSetArg(arglst[3], XmNvalue, (int)(thisMax));
      XtSetArg(arglst[4], XmNpageIncrement, (int)(10));
      XtSetValues(cutMaxScrollW[i], arglst, 5);
      XFlush(dsply);
      sprintf(outStr,"%d",thisMax);
      createSimpleString(&simpleString,outStr);
      XtSetArg(arglst[0], XmNlabelString, simpleString);
      XtSetValues(cutMaxValueW[i], arglst, 1); 
      XFlush(dsply);
      XmStringFree(simpleString); 
      XFlush(dsply);
      
      cutMinValues[i]= (float)thisMin;
      cutMaxValues[i]= (float)thisMax;
      perMinValues[i]= (float)(((float)thisMin/(float)thisDimension));
      perMaxValues[i]= (float)(((float)thisMax/(float)thisDimension));
      
      /* Now do percent label */
      sprintf(outStr,"%.1f%% to %.1f%%",      
	      perMinValues[i]*100.0,
	      perMaxValues[i]*100.0);
      createSimpleString(&simpleString,outStr);
      XtSetArg(arglst[0], XmNlabelString, simpleString);
      XtSetValues(perValueW[i], arglst, 1); 
      XFlush(dsply);
      XmStringFree(simpleString); 
      XFlush(dsply);


    }/* Do cut scrolls */
  

}/* RESET_XYZ_SCROLLS */




/* *******************************************************
 * boxMode_change() - They want to change the boxing
 * mode.  Setup for circle/rectangle/polygon or the
 * way they want to have the background done.
 */
void boxMode_change(Widget widget, int *tag, 
		    XmAnyCallbackStruct *callback_data)
{/* boxMODE_CHANGE */
  XmString simpleString;
  int i;

  if(*tag==USESQUARE)
    {/* They want to use the square box option model */
      boxMethod = *tag;
    }/* They want to use the square box option model */
  else if(*tag==USERECT)
    {/* They want to use circular boxing model */
      boxMethod = *tag;
    }/* They want to use circular boxing model */
  else if(*tag==USECIRCLE)
    {/* They want to use circular boxing model */
      boxMethod = *tag;
    }/* They want to use circular boxing model */
  else if(*tag==USEPOLY)
    {/* They want to use circular boxing model */
      boxMethod = *tag;
    }/* They want to use circular boxing model */
  else if(*tag==USEMAPSHADE)
    {/* They want to use circular boxing model */
      secondMapShade = !secondMapShade;
      if(secondMapShade)
	{/* Give feedback */
	  sprintf(outStr,"Shade ON");
	  XtSetArg(arglst[0], XmNsensitive, TRUE );
	  XtSetValues(fractScrollWidget, arglst, 1); 
	  XtSetValues(angxScrollWidget, arglst, 1); 
	  XtSetValues(angyScrollWidget, arglst, 1); 
	  XtSetValues(fractLabelWidget, arglst, 1); 
	  XtSetValues(angxLabelWidget, arglst, 1); 
	  XtSetValues(angyLabelWidget, arglst, 1); 
	  XtSetValues(fractValueWidget, arglst, 1); 
	  XtSetValues(angxValueWidget, arglst, 1); 
	  XtSetValues(angyValueWidget, arglst, 1); 
	  XtSetValues(fractToggleWidget, arglst, 1); 
	  XtSetValues(angxToggleWidget, arglst, 1); 
	  XtSetValues(angyToggleWidget, arglst, 1); 
	}/* Give feedback */
      else 
	{/* Give feedback */
	  sprintf(outStr,"Shade OFF");
	  XtSetArg(arglst[0], XmNsensitive, FALSE );
	  XtSetValues(fractScrollWidget, arglst, 1); 
	  XtSetValues(angxScrollWidget, arglst, 1); 
	  XtSetValues(angyScrollWidget, arglst, 1); 
	  XtSetValues(fractLabelWidget, arglst, 1); 
	  XtSetValues(angxLabelWidget, arglst, 1); 
	  XtSetValues(angyLabelWidget, arglst, 1); 
	  XtSetValues(fractValueWidget, arglst, 1); 
	  XtSetValues(angxValueWidget, arglst, 1); 
	  XtSetValues(angyValueWidget, arglst, 1); 
	  XtSetValues(fractToggleWidget, arglst, 1); 
	  XtSetValues(angxToggleWidget, arglst, 1); 
	  XtSetValues(angyToggleWidget, arglst, 1); 
	}/* Give feedback */
      createSimpleString(&simpleString,outStr);
      XtSetArg(arglst[0], XmNlabelString, simpleString );
      XtSetArg(arglst[1], XmNrecomputeSize, FALSE );
      XtSetValues(secondShadeWidget, arglst, 2); 
      XFlush(dsply);
      XmStringFree(simpleString); 

      /* Turn radial Projections OFF */
      radialProjections = FALSE;
      sprintf(outStr,"Radial/Icos Proj OFF");
      createSimpleString(&simpleString,outStr);
      XtSetArg(arglst[0], XmNlabelString, simpleString );
      XtSetArg(arglst[1], XmNrecomputeSize, FALSE );
      XtSetValues(radProjShadeWidget, arglst, 2); 
      XFlush(dsply);
      XmStringFree(simpleString);
      /* Now reset toggles */
      XtSetArg(arglst[0], XmNset, FALSE);
      XtSetValues(radProjShadeWidget, arglst, 1);
      XtUnmanageChild(icosRadProjToggleW);
      XtUnmanageChild(radProjToggleW);
      XtUnmanageChild(radProjCSToggleW);
      XtManageChild(radialCSToggleW);
      XtManageChild(radialTextWidget);
      XtManageChild(radialLabelWidget);
      XFlush(dsply);

      /* Now activate the surface density widget */
      XtSetArg(arglst[0], XmNsensitive, TRUE );
      XtSetValues(surfDensScrollWidget, arglst, 1); 
      XtSetValues(surfDensLabelWidget, arglst, 1); 
      XtSetValues(surfDensValueWidget, arglst, 1); 
      XtSetValues(surfDensToggleWidget, arglst, 1); 

    }/* They want to use circular boxing model */
  else if(*tag==USEIMMEDIATE)
    {/* They want to use circular boxing model */
      immediateMapShade = !immediateMapShade;
    }/* They want to use circular boxing model */
  else if(*tag==MBUSEIMMEDIATE)
    {/* They want immediate model */
      mbImmediate = !mbImmediate;
    }/* They want immediate model */
  else if(*tag==ICOSREPL)
    {/* They want icosrepl */
      isIcosRepl = !isIcosRepl;
      if(isIcosRepl)
	{/* they want these replacted on the icosah\edral 5-fold */
	  sprintf(outStr,"Radius:");
	  createSimpleString(&simpleString,outStr);
	  XtSetArg(arglst[0], XmNlabelString, simpleString );
	  XtSetArg(arglst[1], XmNrecomputeSize, FALSE );
	  XtSetValues(icosReplRadiusLabelXW, arglst, 2); 
	  XFlush(dsply);
	  XtUnmanageChild(icosReplRadiusLabelYW);
	  XtUnmanageChild(icosReplRadiusLabelZW);
	  XtUnmanageChild(mbCenteryScrollW);
	  XtUnmanageChild(mbCenteryTextW);
	  XtUnmanageChild(mbCenterzScrollW);
	  XtUnmanageChild(mbCenterzTextW);
	  XtUnmanageChild(mbRotatetScrollW);
	  XtUnmanageChild(mbRotatetTextW);
	  XtUnmanageChild(mbRotatepScrollW);
	  XtUnmanageChild(mbRotatepTextW);
	  XtUnmanageChild(mbRotateoScrollW);
	  XtUnmanageChild(mbRotateoTextW);
	}/* they want these replacted on the icosah\edral 5-fold */
      else
	{/* normal */
	  sprintf(outStr,"x:        ");
	  createSimpleString(&simpleString,outStr);
	  XtSetArg(arglst[0], XmNlabelString, simpleString );
	  XtSetArg(arglst[1], XmNrecomputeSize, FALSE );
	  XtSetValues(icosReplRadiusLabelXW, arglst, 2); 
	  XFlush(dsply);
	  XtManageChild(icosReplRadiusLabelYW);
	  XtManageChild(icosReplRadiusLabelZW);
	  XtManageChild(mbCenteryScrollW);
	  XtManageChild(mbCenteryTextW);
	  XtManageChild(mbCenterzScrollW);
	  XtManageChild(mbCenterzTextW);
	  XtManageChild(mbRotatetScrollW);
	  XtManageChild(mbRotatetTextW);
	  XtManageChild(mbRotatepScrollW);
	  XtManageChild(mbRotatepTextW);
	  XtManageChild(mbRotateoScrollW);
	  XtManageChild(mbRotateoTextW);
	}/* normal */
      XFlush(dsply);
    }/* They want icosrepl */
  else if(*tag==USESTEREO)
    {/* They want to use stereo mapping */
      stereoMapShade = !stereoMapShade;
      if(stereoMapShade)
	{/* Give feedback */
	  sprintf(outStr,"Stereo ON");
	}/* Give feedback */
      else 
	{/* Give feedback */
	  sprintf(outStr,"Stereo OFF");
	}/* Give feedback */
      createSimpleString(&simpleString,outStr);
      XtSetArg(arglst[0], XmNlabelString, simpleString );
      XtSetArg(arglst[1], XmNrecomputeSize, FALSE );
      XtSetValues(stereoShadeWidget, arglst, 2); 
      XFlush(dsply);
      XmStringFree(simpleString); 
    }/* They want to use stereo mapping */
  else if(*tag==LOHIRES)
    {/* They want to use loHiResoltuion */
      int i;
      float thisPixSize;

      if(firstTimeLoHiRes)
	{/* init */
	  calcSF_activate(widget, tag, callback_data);
	  firstTimeLoHiRes = FALSE;
	}/* init */

      lowRes = !lowRes;

      /* Always reset the sampling in case they haven't */
      sprintf(outStr,"1");
      XmTextSetString(sampleRenderWidget,outStr); 
      if(globFract < 0.75)
	{/* set shading fraction */
	  globFract = 0.75;
	  sprintf(outStr,"%.2f",globFract);
	  XmTextSetString(fractValueWidget,outStr); 
	  XtSetArg(arglst[0], XmNvalue, (int)(globFract*1000.0));
	  XtSetValues(fractScrollWidget, arglst, 1);
	  XFlush(dsply);
	}/* set shading fraction */

      if(lowRes)
	{/* Give feedback */

	  if(!immediateResource)
	    {/* Turn immediate update ON */
	      XtSetArg(arglst[0], XmNset, TRUE);
	      XtSetValues(immediateUpdateW, arglst, 1); 
	      immediateMapShade = TRUE;
	      XFlush(dsply);
	    }/* Turn immediate update ON */
	  
	  numCols = lowResCols;
	  numRows = lowResRows;
	  origNumSecs= numSecs = lowResSecs;
	  
	  fileBinSecs= numSecs;
	  fileBinRows= numRows;
	  fileBinCols= numCols;
	  
	  xCen = (double)(numCols/2);
	  yCen = (double)(numRows/2);
	  zCen = (double)(origNumSecs/2);
	  
	  largestDim = numCols;
	  largestDim = MAX(largestDim,numRows);
	  largestDim = MAX(largestDim,origNumSecs);
	  
#define WENEEDTHISlr 1
#ifdef WENEEDTHISlr
	  /* We've got to make this larger for asymmetric maps 
	   * and all maps that have information "outside" of the
	   * normal sphere of our normal maps,
	   * or else we just keep cliping the box into a sphere */
	  if(numCols==numRows && numRows==origNumSecs)
	    {/* Symmetric */
	      aSymmetric = FALSE;
	      origGlobMaxRadius= globMaxRadius = (double)
		((double)origNumSecs/2)+5.0;
	      globMinRadius = (double)0.0;
	    }/* Symmetric */
	  else
	    {/* asymmetric */
	      aSymmetric = TRUE;
	      origGlobMaxRadius=  globMaxRadius = (double)
		((double)origNumSecs);
	      globMinRadius = (double)0.0;
	    }/* asymmetric */
	  
	  /* Set scrolls on depthcue screen */
	  /* Give feedback */
	  sprintf(outStr,"%.0f",globMinRadius);
	  createSimpleString(&simpleString,outStr);
	  XtSetArg(arglst[0], XmNvalue, outStr );
	  XtSetArg(arglst[1], XmNrecomputeSize, FALSE );
	  XtSetValues(minRadiusValueWidget, arglst, 2); 
	  XFlush(dsply);
	  XmStringFree(simpleString); 
	  
	  sprintf(outStr,"%.0f",globMaxRadius);
	  createSimpleString(&simpleString,outStr);
	  XtSetArg(arglst[0], XmNvalue, outStr );
	  XtSetArg(arglst[1], XmNrecomputeSize, FALSE );
	  XtSetValues(maxRadiusValueWidget, arglst, 2); 
	  XFlush(dsply);
	  XmStringFree(simpleString); 
	  
	  
	  /* Radius is origNumSecs/2 */
	  XtVaSetValues(maxRadiusScrollWidget,XmNmaximum, 
			(int)((globMaxRadius*2)+5),NULL);
	  XtVaSetValues(maxRadiusScrollWidget,XmNsliderSize, (int)(5),NULL);
	  XtVaSetValues(maxRadiusScrollWidget,XmNvalue, 
			(int)(globMaxRadius),NULL);
			
	  XtVaSetValues(minRadiusScrollWidget,XmNmaximum, 
			(int)((globMaxRadius*2)+5),NULL);
	  XtVaSetValues(minRadiusScrollWidget,XmNsliderSize, (int)(5),NULL);
	  XtVaSetValues(minRadiusScrollWidget,XmNvalue, 
			(int)(globMinRadius),NULL);
			
#endif /* WE NEEDTHIS */
	  
	  if(whichStep==PIXELSIZE)
	    {/* We can't allow this for lowres */
	      XtSetArg(arglst[0], XmNset, FALSE );
	      whichStep = NONE;
	      XmTextSetString(frameTextWidget,""); 
	      sprintf(outStr,"NO Step:");
	      createSimpleString(&simpleString,outStr);
	      XtSetArg(arglst[0], XmNlabelString, simpleString );
	      XtSetArg(arglst[1], XmNrecomputeSize, FALSE );
	      XtSetValues(stepValueLabelWidget, arglst, 2); 
	      XFlush(dsply);
	      XmStringFree(simpleString); 
	    }/* We can't allow this for lowres */
	  sprintf(outStr,"Low Resolution ON");
	  XtSetArg(arglst[0], XmNsensitive, FALSE );
	  XtSetValues(pixelSizeScrollWidget, arglst, 1); 
	  XtSetValues(pixelSizeValueWidget, arglst, 1); 
	  
	  reset_xyz_scrolls();

	}/* Give feedback */
      else 
	{/* NOT lowres Give feedback */

	  /* Make them recalc any new hires buffer */
	  pixelSizeSave = -99.0;

	  /* Save where they cut for use when generating the HiRes buffer */
	  cutMin0Save  = (float)cutMinValues[0];
	  cutMax0Save  = (float)cutMaxValues[0];
	  cutMin1Save  = (float)cutMinValues[1];
	  cutMax1Save  = (float)cutMaxValues[1];
	  cutMin2Save  = (float)cutMinValues[2];
	  cutMax2Save  = (float)cutMaxValues[2];

	  /* Turn immediate update OFF */
	  XtSetArg(arglst[0], XmNset, FALSE);
	  XtSetValues(immediateUpdateW, arglst, 1); 
	  immediateMapShade = FALSE;
	  XFlush(dsply);

#define WENEEDTHISHighRes 1
#ifdef WENEEDTHISHighRes
	  
	  /* Set the radius ridiculously high so they can
	   * see all of it and then adjust as they need */
	  sprintf(outStr,"%.0f",(float)lowResSecs * 3.0);
	  createSimpleString(&simpleString,outStr);
	  XtSetArg(arglst[0], XmNvalue, outStr );
	  XtSetArg(arglst[1], XmNrecomputeSize, FALSE );
	  XtSetValues(maxRadiusValueWidget, arglst, 2); 
	  XFlush(dsply);
	  XmStringFree(simpleString); 

	  globMaxRadius= lowResSecs * 3.0;
	  
	  XtVaSetValues(maxRadiusScrollWidget,XmNmaximum, 
			(int)((lowResSecs*4)+5),NULL);
	  XtVaSetValues(maxRadiusScrollWidget,XmNsliderSize, (int)(5),NULL);
	  XtVaSetValues(maxRadiusScrollWidget,XmNvalue, 
			(int)(lowResSecs*3),NULL);
			
#endif /* WE NEEDTHIS */
	  sprintf(outStr,"High Resolution ON");
	  XtSetArg(arglst[0], XmNsensitive, TRUE );
	  XtSetValues(pixelSizeScrollWidget, arglst, 1); 
	  XtSetValues(pixelSizeValueWidget, arglst, 1); 
	}/* NOT lores Give feedback */

      /* WHy are we doing this - it's for the movie???
	 XtSetValues(pixelSizeToggleWidget, arglst, 1); 
      */
      
      /* Now setup the pixelSize */
      if(!lowRes)
	{/* Setup for High Res */
	  thisPixSize= origWidthFromFile/lowResCols;
	  XtSetArg(arglst[0], XmNvalue, (int)(thisPixSize*1000.0));
	  XtSetValues(pixelSizeScrollWidget, arglst, 1);

	  sprintf(outStr,"%.2f",thisPixSize);
	  XmTextSetString(pixelSizeValueWidget, outStr); 
	  XFlush(dsply);
	}/* Setup for High Res */

    }/* They want to use loHi resolution */
  else if(*tag==USECONT)
    {/* They want to use contours */
      useCont  = !useCont;
      if(useCont)
	{/* Give feedback */
	  sprintf(outStr,"Apply to movie.");
	}/* Give feedback */
      else 
	{/* Give feedback */
	  sprintf(outStr,"Don't apply to movie.");
	}/* Give feedback */
      createSimpleString(&simpleString,outStr);
      XtSetArg(arglst[0], XmNlabelString, simpleString );
      XtSetArg(arglst[1], XmNrecomputeSize, FALSE );
      XtSetValues(useContWidget, arglst, 2); 
      XFlush(dsply);
      XmStringFree(simpleString); 
    }/* They want to use contours */
  else if(*tag==USERADIAL)
    {/* They want to use radial mapping */
      radialMapShade = !radialMapShade;
      if(radialMapShade)
	{/* Give feedback */
	  sprintf(outStr,"Radial Cueing ON");
	}/* Give feedback */
      else 
	{/* Give feedback */
	  sprintf(outStr,"Radial Cueing OFF");
	}/* Give feedback */
      createSimpleString(&simpleString,outStr);
      XtSetArg(arglst[0], XmNlabelString, simpleString );
      XtSetArg(arglst[1], XmNrecomputeSize, FALSE );
      XtSetValues(radialShadeWidget, arglst, 2); 
      XFlush(dsply);
      XmStringFree(simpleString); 

      /* Turn radial Projections OFF */
      radialProjections = FALSE;
      sprintf(outStr,"Radial/Icos Proj OFF");
      createSimpleString(&simpleString,outStr);
      XtSetArg(arglst[0], XmNlabelString, simpleString );
      XtSetArg(arglst[1], XmNrecomputeSize, FALSE );
      XtSetValues(radProjShadeWidget, arglst, 2); 
      XFlush(dsply);
      XmStringFree(simpleString);
      /* Now reset toggles */
      XtSetArg(arglst[0], XmNset, FALSE);
      XtSetValues(radProjShadeWidget, arglst, 1);
      XtUnmanageChild(icosRadProjToggleW);
      XtUnmanageChild(radProjToggleW);
      XtUnmanageChild(radProjCSToggleW);
      XtManageChild(radialCSToggleW);
      XtManageChild(radialTextWidget);
      XtManageChild(radialLabelWidget);
      XFlush(dsply);

      /* Now activate the surface density widget */
      XtSetArg(arglst[0], XmNsensitive, TRUE );
      XtSetValues(surfDensScrollWidget, arglst, 1); 
      XtSetValues(surfDensLabelWidget, arglst, 1); 
      XtSetValues(surfDensValueWidget, arglst, 1); 
      XtSetValues(surfDensToggleWidget, arglst, 1); 

    }/* They want to use radial mapping */
  else if(*tag==USERADPROJ)
    {/* They want to use radial projections */
      radialProjections = !radialProjections;
      if(radialProjections)
	{/* Give feedback */

	  /* Turn radial and shading off */
	  radialMapShade = FALSE;
	  sprintf(outStr,"Radial Cueing OFF");
	  createSimpleString(&simpleString,outStr);
	  XtSetArg(arglst[0], XmNlabelString, simpleString );
	  XtSetArg(arglst[1], XmNrecomputeSize, FALSE );
	  XtSetValues(radialShadeWidget, arglst, 2); 
	  XFlush(dsply);
	  XmStringFree(simpleString);

	  secondMapShade = FALSE;
	  sprintf(outStr,"Shade OFF");
	  XtSetArg(arglst[0], XmNsensitive, FALSE );
	  XtSetValues(fractScrollWidget, arglst, 1); 
	  XtSetValues(angxScrollWidget, arglst, 1); 
	  XtSetValues(angyScrollWidget, arglst, 1); 
	  XtSetValues(fractLabelWidget, arglst, 1); 
	  XtSetValues(angxLabelWidget, arglst, 1); 
	  XtSetValues(angyLabelWidget, arglst, 1); 
	  XtSetValues(fractValueWidget, arglst, 1); 
	  XtSetValues(angxValueWidget, arglst, 1); 
	  XtSetValues(angyValueWidget, arglst, 1); 
	  XtSetValues(fractToggleWidget, arglst, 1); 
	  XtSetValues(angxToggleWidget, arglst, 1); 
	  XtSetValues(angyToggleWidget, arglst, 1); 
	  createSimpleString(&simpleString,outStr);
	  XtSetArg(arglst[0], XmNlabelString, simpleString );
	  XtSetArg(arglst[1], XmNrecomputeSize, FALSE );
	  XtSetValues(secondShadeWidget, arglst, 2); 
	  XFlush(dsply);
	  XmStringFree(simpleString); 

	  /* Now reset toggles */
	  XtSetArg(arglst[0], XmNset, FALSE);
	  XtSetValues(secondShadeWidget, arglst, 1);
	  XtSetValues(radialShadeWidget, arglst, 1);

	  /* Now setup cut scrolls and values widgets */
	  reset_xyz_scrolls();
	  
	  /* Now for radproj feedback */
	  sprintf(outStr,"Radial/Icos Proj ON");
	  XtManageChild(icosRadProjToggleW);
	  XtManageChild(radProjToggleW);
	  XtManageChild(radProjCSToggleW);
	  XtUnmanageChild(radialCSToggleW);
	  XtUnmanageChild(radialTextWidget);
	  XtUnmanageChild(radialLabelWidget);
	  XFlush(dsply);

	  /* Now DEactivate the surface density widget */
	  XtSetArg(arglst[0], XmNsensitive, FALSE );
	  XtSetValues(surfDensScrollWidget, arglst, 1); 
	  XtSetValues(surfDensLabelWidget, arglst, 1); 
	  XtSetValues(surfDensValueWidget, arglst, 1); 
	  XtSetValues(surfDensToggleWidget, arglst, 1); 
	  XFlush(dsply);

	}/* Give feedback */
      else 
	{/* Give feedback */
	  sprintf(outStr,"Radial/Icos Proj OFF");
	  XtUnmanageChild(icosRadProjToggleW);
	  XtUnmanageChild(radProjToggleW);
	  XtUnmanageChild(radProjCSToggleW);
	  XtManageChild(radialCSToggleW);
	  XtManageChild(radialTextWidget);
	  XtManageChild(radialLabelWidget);
	  XFlush(dsply);
	  /* Now activate the surface density widget */
	  XtSetArg(arglst[0], XmNsensitive, TRUE );
	  XtSetValues(surfDensScrollWidget, arglst, 1); 
	  XtSetValues(surfDensLabelWidget, arglst, 1); 
	  XtSetValues(surfDensValueWidget, arglst, 1); 
	  XtSetValues(surfDensToggleWidget, arglst, 1); 

	}/* Give feedback */
      createSimpleString(&simpleString,outStr);
      XtSetArg(arglst[0], XmNlabelString, simpleString );
      XtSetArg(arglst[1], XmNrecomputeSize, FALSE );
      XtSetValues(radProjShadeWidget, arglst, 2); 
      XFlush(dsply);
      XmStringFree(simpleString); 

    }/* They want to use radial projections */
  else if(*tag==USEAVGPIXEL)
    {/* They want to use average pixel bkgrnd */

      bkgrndMethod = *tag;

      /* Redraw this puppy */
      if(fileType==PACKEDPORTABLE)
	stuffDisplayPackBoxBuffer(curBox, editBoxZoom, FALSE);
      else stuffDisplayBoxBuffer(curBox,image_fileName.addr,editBoxZoom);
      DrawEditBoxWindow(FALSE);

    }/* They want to use average pixel bkgrnd */
  else if(*tag==USERANDPIXEL)
    {/* They want to use random pixel bkgrnd */

      bkgrndMethod = *tag;

      /* Redraw this puppy */
      if(fileType==PACKEDPORTABLE)
	stuffDisplayPackBoxBuffer(curBox, editBoxZoom, FALSE);
      else stuffDisplayBoxBuffer(curBox,image_fileName.addr,editBoxZoom);
      DrawEditBoxWindow(FALSE);

    }/* They want to use random pixel bkgrnd */
  else if(*tag==VOLUME)
    {/* They want to see what volume they are at */
      volumeDisp = !volumeDisp;
    }/* They want to see what volume they are at */
  

}/* boxMODE_CHANGE */




/* *******************************************************
 * histMode_change() - They want to toggle whether they
 * see a histogram or not 
 */
void histMode_change(Widget widget,int *tag,
			XmAnyCallbackStruct *callback_data)
{ /* histMode_change */
  static int oldTag= 0;

  if(*tag==USEHIST)
    {/* They want to see histogram */
      histMethod = !histMethod;
    }/* They want to see histogram */
  else if(*tag != oldTag)
    {/* DOn't do this twice */
      if(*tag==NOCONTRAST)
	{/* They want to contrast stretch */
	  contrastStretch = FALSE;
	  histoEqualize = FALSE;
	  if(fileType != PACKEDPORTABLE &&
	     fileType != MAP && fileType!=FLOATMAP)
	    stuffImage(int_image_buffer1, image_width, image_height, 
		       contrastStretch,FALSE);
	  DrawWindow();
	  histogram_activate(&minP, &maxP);
	}/* They want to contrast stretch */
      else if(*tag==USECONTRAST)
	{/* They want to contrast stretch */
	  contrastStretch = TRUE;
	  histoEqualize = FALSE;
	  if(fileType != PACKEDPORTABLE &&
	     fileType != MAP && fileType!=FLOATMAP)
	    stuffImage(int_image_buffer1, image_width, image_height, 
		       contrastStretch,FALSE);
	  DrawWindow();
	  histogram_activate(&minP, &maxP);
	}/* They want to contrast stretch */
      else if(*tag==USEHISTEQUAL)
	{/* They want to use histogram equalization */
	  histoEqualize = TRUE;
	  contrastStretch = FALSE;
	  if(fileType != PACKEDPORTABLE &&
	     fileType != MAP && fileType!=FLOATMAP)
	    stuffImage(int_image_buffer1, image_width, image_height, 
		       contrastStretch,FALSE);
	  DrawWindow();
	  histogram_activate(&minP, &maxP);
	}/* They want to use histogram equalization */
    }/* DOn't do this twice */

  oldTag = *tag;

} /* histMode_change */




/* *******************************************************
 * dispMode_change() - They want to toggle whether they
 * see the circle number displayed in the middle of their
 * circles or not.
*/
void dispMode_change(Widget widget, int *tag, 
		     XmAnyCallbackStruct *callback_data)
{/* dispMODE_CHANGE */
  
  if(*tag==CIRCLEDISP)
    {/* They want to use the square disp option model */
      if(dispMethod==CIRCLEDISP)
	{dispMethod = FALSE;       /* Toggle this guy */
	 XtSetArg(arglst[0], XmNset, FALSE);
       }
      else 
	{dispMethod= CIRCLEDISP;
	 XtSetArg(arglst[0], XmNset, TRUE);
       }
      /* We may want to set the toggle in the options screen */
      XtSetValues(circleDispWidget, arglst, 1);
      DrawWindow();
    }/* They want to use the square disp option model */
  else if(*tag==BOXCOUNTER)
    {/* They want to init the box counter */
      initLastNumBox = !initLastNumBox;
    }/* They want to init the box counter */
  else if(*tag==AUTOZOOM)
    {/* They want to init the box counter */
      autoZoom = !autoZoom;
    }/* They want to init the box counter */
  else if(*tag==FFTMOTIONFEED)
    fftMotionFeedback = !fftMotionFeedback;
  else if(*tag==CALCSFS)
    calcSFs = !calcSFs;
  else if(*tag==CALCBIN)
    calcBin = !calcBin;
  else if(*tag==TRUEPHI)
    truePhi = !truePhi;
  else if(*tag==FORCEBYTE)
    forceByte = !forceByte;
  else if(*tag==FORCEAVGBKG)
    {
      forceAvgBkg = !forceAvgBkg;
    }
  else if(*tag==MINMAXCOLOR)
    minMaxColor = !minMaxColor;
  else if(*tag==NORMALEXCISE)
    normalExcise = !normalExcise;
  else if(*tag==EMPTYBOX)
    emptyBox = !emptyBox;

}/* dispMODE_CHANGE */




/*-----------------------------------*/
/*************************************
 * HISTOGRAM callbacks               *
 *************************************/
/*-----------------------------------*/
/* ************************************************************
 * histogram_activate() - Plot the histogram on the colormap 
 * screen.
 */
void histogram_activate(int *minP, int *maxP)
{/* HISTOGRAM_ACTIVATE */
  int i,j,k,max,min;
  int p;
  float thisZoomFactor;
  
  /* Find min and max */
  min = 99999;
  max = -999;
  min = INT_MAX;  /* dgm INT_MAX instead of 99999 */
  max = INT_MIN;  /* dgm INT_MIN instead of -999 */
  for (i=0;i<256;i++)
    {/* find max and min */
      p = (int) histogram[i];
      if((p) < min)
	min = (p);
      if((p) > max)
	max = (p);
    }/* find max and min */
  *minP = min;
  *maxP = max;
  
  /* Zoom this guy appropriately */
  thisZoomFactor= 1.0;
  if(histZoomFactor >= 20)
    {/* ZoomIn */
      for(k=20;k<histZoomFactor;k++)
	thisZoomFactor *= 2.0;
    }/* ZoomIn */
  else
    {/* ZoomOut */
      for(k=histZoomFactor;k<20;k++)
	thisZoomFactor /= 2.0;
    }/* ZoomOut */

  /* Make sure we don't over step our bounds */
  if(max > 0 && thisZoomFactor > 0)
    histScaleFactor = 32.0 / (float) (max / thisZoomFactor);
  else histScaleFactor = 0;
  
  /* Now draw in image */
  for(i=0;i<=255;i++)
    {/* GO the y direction */
      for(j=0;j<32;j++)
	{
	  if(hiColor)
	    XPutPixel(hist_color_image1,
		      j, i,
		      pixels_return[WHITE]);
	  else
	    hist_color_buffer1[((i*32)+j)]= pixels_return[WHITE];
	}
    }/* GO the y direction */
  for(i=255;i>=0;i--)
    {/* GO the y direction */
      for(j=0;j<calcHistLength(i);j++)
	{
	  if(hiColor)
	    XPutPixel(hist_color_image1,
		      j, i,
		      pixels_return[RED]);
	  else
	    hist_color_buffer1[((i*32)+j)]= pixels_return[RED];
	}
    }/* GO the y direction */

  /* Redraw the colromap window */
  for(i=255;i>=0;i--)
    for(j=0;j<32;j++)
      {
	if(hiColor)
	  XPutPixel(color_image1,
		    j, i,
		    pixels_return[(256-i)]);
	else
	  color_buffer1[((i*32)+j)]=  (256-i);
      }
  
  if(XtIsManaged(colorScreenWidget))
    {/* Only do if managed */
      XPutImage(dsply, hist_color_window1, hist_color_gc, hist_color_image1, 
		0, 0,
		0, 0,  32,  256);
      XPutImage(dsply, color_window1, color_gc, color_image1, 
		0, 0,
		0, 0,  32,  256);
      XFlush (dsply);
    }/* Only do if managed */
  
  /* Draw the colormap window for good measure */
  DrawColormapWindow();

  
}/* HISTOGRAM_ACTIVATE */



/* ************************************************************
 * histArrowActivate() - They want to expand the histogram
 * on the Alter Colormap screen.
*/
void histArrowActivate(Widget widget, int *tag, 
		       XmAnyCallbackStruct *callback_data)
{/* HISTArrowActIVATE */

  /* is it up or down ? */
  if(*tag==XmARROW_UP)
    {/* Up */
     histZoomFactor++;
    }/* Up */
  else if(*tag==XmARROW_DOWN)
    {/* Dopwn */
      histZoomFactor--;
    }/* Down */
  else printf("HistZoom unchanged.\n");

  /* Clip it accordingly */
  if(histZoomFactor > 39)
    histZoomFactor= 39;
  if(histZoomFactor < 1)
    histZoomFactor= 1;

  /* Redraw the puppy */
  histogram_activate(&minP, &maxP);

}/* HISTArrowActIVATE */




/*-----------------------------------*/
/*************************************
 * MAPMOVIE callbacks                 *
 *************************************/
/*-----------------------------------*/
/* ************************************************************
 * mapMovie_activate() - They want to see their map sections
 * displayed as a "movie".
 */
void mapMovie_activate(Widget widget, int *tag, 
		       XmAnyCallbackStruct *callback_data)
{/* mapMovie_activate */
  XmString simpleString;

  /* Put up the proper widgets in the Form */
  makingMovie= TRUE;
  XtManageChild(playFormWidget);
  XtManageChild(movieSaveWidget);
  XtManageChild(movieMinWidget);
  XtManageChild(movieMaxWidget);
  XtManageChild(maxLabelWidget);
  XtManageChild(minLabelWidget);
  XtUnmanageChild(undoWidget);
  XtUnmanageChild(savePaintWidget);
  XtUnmanageChild(paintFormWidget);
  XtSetArg(arglst[0], XmNsensitive, TRUE);
  XtSetValues(playWidget, arglst, 1); 
  XtSetValues(stopWidget, arglst, 1); 
  XtSetValues(infiniWidget, arglst, 1); 
  XtSetValues(rockWidget, arglst, 1); 
  XtSetValues(flashLightWidget, arglst, 1); 
  createSimpleString(&simpleString,"Movie Display");
  XtSetArg(arglst[0], XmNdialogTitle, simpleString);
  XtSetValues(editBoxScreenWidget, arglst, 1); 
  XFlush(dsply);
  XmStringFree(simpleString); 

  /* Now setup the movie min/max scroll bars */
  startMovie = 0;
  if(dispSections)
    endMovie = numSecs-1;
  else endMovie = mapNumSecs - 1;
  sprintf(outStr,"%d",startMovie+1);
  createSimpleString(&simpleString,outStr);
  XtSetArg(arglst[0], XmNlabelString, simpleString);
  XtSetValues(minLabelWidget, arglst, 1); 
  XFlush(dsply);
  XmStringFree(simpleString); 

  sprintf(outStr,"%d",endMovie+1);
  createSimpleString(&simpleString,outStr);
  XtSetArg(arglst[0], XmNlabelString, simpleString);
  XtSetValues(maxLabelWidget, arglst, 1); 
  XFlush(dsply);
  XmStringFree(simpleString); 

  XtVaSetValues(movieMaxWidget,XmNmaximum, numSecs,NULL);
  XtVaSetValues(movieMaxWidget,XmNminimum, 0,NULL);
  XtVaSetValues(movieMaxWidget,XmNvalue, numSecs-1,NULL);
  XtVaSetValues(movieMaxWidget,XmNsliderSize, 1,NULL);
  XtVaSetValues(movieMaxWidget,XmNpageIncrement, 5,NULL);
  
  XtVaSetValues(movieMinWidget,XmNmaximum, numSecs,NULL);
  XtVaSetValues(movieMinWidget,XmNminimum, 0,NULL);
  XtVaSetValues(movieMinWidget,XmNvalue, 0,NULL);
  XtVaSetValues(movieMinWidget,XmNsliderSize, 1,NULL);
  XtVaSetValues(movieMinWidget,XmNpageIncrement, 5,NULL);
  
  raiseIt(editBoxScreenWidget);
  stuffDisplayMapBuffer(0,editBoxZoom,0);
  if(shadedMaps && stereoMapShade)
    stuffDisplayMapBuffer(1,editBoxZoom,numCols);
  DrawEditBoxWindow(TRUE);
  XFlush(dsply);
  
  if(*tag==PAINTMAP)
    {/* They want to paint the map */
      int i,j,k;
      paintMapSwitch = TRUE;
      XtUnmanageChild(movieSaveWidget);   /* Doesn't make sense */
      if(!paintMap)
	{/* Make a new one */
	  paintMap = (float *)malloc(sizeof(float)*numCols*
				     numRows*origNumSecs);
	  for(i=0;i<origNumSecs;i++)
	    for(j=0;j<numRows;j++)
	      for(k=0;k<numCols;k++)
		{
		  if(normalExcise)
		    paintMap[((i*numCols*numRows)+
			      (j*numCols)+k)]= 0.0;
		  else
		    paintMap[((i*numCols*numRows)+
			      (j*numCols)+k)]= 
				float_image_buffer1[((i*numCols*numRows)+
						     (j*numCols)+k)];
		}
	}/* Make a new one */
	    
      askThem("Mouse modes for Excise Cuts:\n\nLeft=Small Cut  \nMiddle=Small UNDO  \nRight=LARGE Cut  \n\nUse File/Write/MAP to write cutout to disk.",
	      FALSE,XmDIALOG_INFORMATION);

    }/* They want to paint the map */

}/* mapMovie_activate */




/* ************************************************************
 * subMapShadeScreen_activate() - They want to do 3D rendering
 * of their subtracted map.
 * 
 */
void subMapShadeScreen_activate(Widget widget, int *tag, 
				XmAnyCallbackStruct *callback_data)
{/* subMapShadeScreen_activate */
  int i,j,k,rowOut,secOut;
  float *tmpSubtract_image_buffer2;

  /* Compute third buffer */
  CK(tmpSubtract_image_buffer2 = (float *)malloc(sizeof(float)*
				      sub2binCol*sub2binRow*sub2binSec));
  for(i=0;i<sub2binSec;i++)
    {
      secOut = (i * sub2binRow*sub2binCol);
      for(j=0;j<sub2binRow;j++)
	{
	  rowOut = (j * sub2binCol);
	  for(k=0;k<sub2binCol;k++)
	    {
	      subtract_image_buffer2[secOut+rowOut+k] = 
		float_image_buffer1[secOut+rowOut+k]; 
	    }
	}
    }

  raiseIt(statusScreenWidget);
  XFlush(dsply);

  subtractMap_doit(subtract_image_buffer2,tmpSubtract_image_buffer2);
  free(tmpSubtract_image_buffer2);

  XtUnmanageChild(statusScreenWidget);

  /* Now shade them */
  mapShadeScreen_activate(0,0,0);

}/* subMapShadeScreen_activate */





/* ************************************************************
 * mapShadeScreen_activate() - They want to see their map sections
 * displayed as a depth-cue shade 3D surface.
 */
void mapShadeScreen_activate(Widget widget, int *tag, 
			     XmAnyCallbackStruct *callback_data)
{/* mapShadeScreen_activate */
  int thisDimension;
  XmString simpleString;

  minUnblob= (int)secMinPixel;
  maxUnblob= (int)0.0 - surfaceDensity;

  /* Divide by 100 for percentage */
  unblobSize= (int)((numCols*numRows*origNumSecs) * 
		    (float)((float)0.000/100.0));

  /* Now setup the scrolls accordingly */
  XtSetArg(arglst[0], XmNmaximum, (int)((numCols/2)+5));
  XtSetArg(arglst[1], XmNminimum, (int)(0));
  XtSetArg(arglst[2], XmNsliderSize, (int)(5));
  XtSetArg(arglst[3], XmNvalue, (int)(0));
  XtSetArg(arglst[4], XmNpageIncrement, (int)(10));
  XtSetValues(unblobRadiusScrollW, arglst, 5);
  XFlush(dsply);
  XtSetArg(arglst[0], XmNvalue, (int)(0));
  XtSetValues(unblobSizeScrollW, arglst, 1);
  XFlush(dsply);
  /* And the value labels */
  sprintf(outStr,"%d",(0));
  createSimpleString(&simpleString,outStr);
  XtSetArg(arglst[0], XmNlabelString, simpleString);
  XtSetValues(unblobRadiusValueW, arglst, 1); 
  XFlush(dsply);
  XmStringFree(simpleString); 
  XFlush(dsply);
  sprintf(outStr,"%.3f%%",(0.000));
  XmTextSetString(unblobSizeValueW,outStr); 
  XFlush(dsply);
  
  /* Now setup cut scrolls and values widgets */
  reset_xyz_scrolls();


  XtSetArg(arglst[0], XmNsensitive, TRUE);
  XtSetValues(paintMapButtonW, arglst, 1); 
  XtSetValues(mapMovieButtonW, arglst, 1); 
  XFlush(dsply);

  firstTimeLoHiRes = TRUE;

  if(!lowRes)
    {/* Setup for High Res */
      pixelSizeFactor= numCols/100.0;
      globPixelSize = pixelSizeFactor;
      
      XtSetArg(arglst[0], XmNvalue, (int)(globPixelSize*1000.0));
      XtSetValues(pixelSizeScrollWidget, arglst, 1);

      sprintf(outStr,"%.2f",globPixelSize);
      XmTextSetString(pixelSizeValueWidget, outStr); 
      XFlush(dsply);

    }/* Setup for High Res */


  XtManageChild(mapShadeScreenWidget);
  raiseIt(mapShadeScreenWidget);

}/* mapShadeScreen_activate */




/* ************************************************************
 * angleText_create() - Remember appropriate widgets for use
 * later.
 */
void angleText_create(Widget widget, int *tag, 
		      XmAnyCallbackStruct *callback_data)
{/* angleText_create */
  char outStr[16];

  if(*tag==FRAME)
    frameTextWidget= widget;
  else if(*tag==USERADIAL)
    {/* Radial step text widget */
      radialTextWidget= widget;
    }/* Radial step text widget */
  else if(*tag==USERADIALLABEL)
    {/* Radial step text widget */
      radialLabelWidget= widget;
    }/* Radial step text widget */
  else if(*tag==STEP)
    stepTextWidget= widget;
  else if(*tag==STEPTOGGLE)
    stepToggleWidget= widget;
  else if(*tag==STEPTOGGLE2)
    stepToggle2Widget= widget;
  else if(*tag==STEPVAL)
    stepValueLabelWidget= widget;
  else if(*tag==SAMPLERENDER)
    sampleRenderWidget= widget;

}/* angleText_create */



/*****************************************************************
 * updateNewVolume() - Update the volume text widget based on
 * what our new surface density is.
*/
void updateNewVolume()
{/* UPDATENEWVOLUME */
  float volume,thisSD,minRad,maxRad;
  int l,foundIt,printIt;
  int i,j,maxI,thickStep,modStep,foundZero;
  float scaleF;
  GC thisGC;
  float *this_float_image_buffer1;

  if(volumeDisp)
    {/* Draw it on the main screen */

      if(unblobSwitch)
	this_float_image_buffer1= unblobBuffer;
      else this_float_image_buffer1= float_image_buffer1;

      /*******************************************/
      /* And now let's compute our volume arrays */
      /*******************************************/
      nbins_lim = nbins = 200;
      /* nbins_lim = nbins = ((int)(maxMapPixel+0.49999)-
			   (int)(minMapPixel-0.49999)); */
      minRad = (float)globMinRadius;
      maxRad = (float)globMaxRadius;


      nbins_lim = nbins = 100;
      nbins_lim = nbins = ((int)(secMaxPixel+0.49999)-
			   (int)(secMinPixel-0.49999));
      if(nbins <= 0)
	{/* Impossible */
	  printf("Bad nbins - no volume possible.\n");
	  goto noVolume;
	}/* Impossible */
      while(nbins > 200)
	{nbins *= 0.5; nbins_lim *= 0.5;}

      map_vol_rwa_ (this_float_image_buffer1, 
		    &numCols, &numRows, &origNumSecs, 
		    &secMinPixel, &secMaxPixel,
		    &nbins_lim, 
		    volumeFraction, volumeBin, 
		    &nbins, volumeCutoff,
		    cutMinValues,
		    cutMaxValues,
		    &minRad,
		    &maxRad);

      
      thisSD = 0.0 - surfaceDensity;
      thisSD = (float)irev * surfaceDensity;
      foundIt = -99;
      for(i=0;i<(nbins-1);i++)
	{/* Look for this cutoff */
	  if((thisSD >= volumeCutoff[i]) &&
	     (thisSD < volumeCutoff[i+1]))
	    {/* This is it */
	  foundIt = i;
	  break;
	}/* This is it */
	}/* Look for this cutoff */
      
      if(foundIt >= 0)
	{/* OK */
	  if(irev==-1)
	    volume = 100.0 - volumeFraction[i];
	  else 
	    volume = volumeFraction[i];
	  sprintf(outStr,"%.2f %% volume\n%.1E A**3",
		  volume,
		  (numCols*numRows*origNumSecs*pifHeader.pixelSize)*
		  (volume/100.0)
		  );
	}/* OK */
      else 
	{/* not OK */
	  volume = -99.9;
	  sprintf(outStr,"");
	}/* not OK */
      
      XmTextSetString(volumeValueW, outStr); 
      XFlush(dsply);
      
      thickStep = 7;  /* default */
      thickStep = (int)(display_width-50) / nbins;
      foundZero = 0;
      modStep   = 20;

      maxI=volumeBin[0];
      for(i=0;i<nbins;i++)
	{/* Search all */
	  if(!((0.0 > volumeCutoff[i]) &&
	       (0.0 <= volumeCutoff[i+1])))
	    maxI = MAX(maxI,volumeBin[i]);
	  else
	    foundZero = i;
	}/* Search all */
      scaleF = (400.0 / (float)maxI);
      
      thisGC = image_gcYELLOW;
      sprintf(outStr,"%.0f",volumeCutoff[0]);
      XDrawString(dsply, XtWindow(drawingWidget), 
		  thisGC,
		  10, display_height-50, outStr, strlen(outStr));
      sprintf(outStr,"%.0f",volumeCutoff[(nbins-1)]);
      XDrawString(dsply, XtWindow(drawingWidget), 
		  thisGC,
		  10+(thickStep*(nbins-1)), display_height-50, 
		  outStr, strlen(outStr));

      for(j=0;j<thickStep;j++)
	{/* DO large */
	  XDrawLine(dsply, XtWindow(drawingWidget), thisGC,
		    10+(0*thickStep)+j,display_height-100,
		    10+(0*thickStep)+j,display_height-90);
	  XDrawLine(dsply, XtWindow(drawingWidget), thisGC,
		    10+((nbins-1)*thickStep)+j,display_height-100,
		    10+((nbins-1)*thickStep)+j,display_height-90);
	}/* DO large */

      for(i=0;i<nbins;i++)
	{/* Do all */
	  if(irev==-1)
	    {/* This way */
	      if(i < foundIt)
		thisGC = image_gcRED;
	      else if(i==foundIt)
		thisGC = image_gcBLUE;
	      else thisGC = image_gcYELLOW;
	    }/* This way */
	  else
	    {/* That way */
	      if(i > foundIt)
		thisGC = image_gcRED;
	      else if(i==foundIt)
		thisGC = image_gcBLUE;
	      else thisGC = image_gcYELLOW;
	    }/* That way */

	  for(j=0;j<thickStep;j++)
	    {/* Draw hsitogram */
	      XDrawLine(dsply, XtWindow(drawingWidget), 
			thisGC,
			10+(i*thickStep)+j,
			display_height-100,
			10+(i*thickStep)+j,
			display_height-((int)(volumeBin[i]*scaleF)+100));
	      /* DO ticks every modStep */
	      if((((int)fabs((float)i-(float)foundZero)%modStep)==0)   &&
		 (i > modStep) && (i < (nbins-modStep))&&
		 !((i <= foundZero+1) && (i >= foundZero-1)))
		{/* Tick it */
		  XDrawLine(dsply, XtWindow(drawingWidget), image_gcYELLOW,
			    10+(i*thickStep)+j,display_height-100,
			    10+(i*thickStep)+j,display_height-90);
		  sprintf(outStr,"%.0f",volumeCutoff[i]);
		  XDrawString(dsply, XtWindow(drawingWidget), image_gcYELLOW,
			      10+(thickStep*i), display_height-50, 
			      outStr, strlen(outStr));
		}/* Tick it */
	    }/* Draw hsitogram */

	  if((0.0 > volumeCutoff[i]) &&
	     (0.0 <= volumeCutoff[i+1]))
	    {/* Draw the zero */
	      thisGC = image_gcYELLOW;
	      sprintf(outStr,"%.0f",0.0);
	      XDrawString(dsply, XtWindow(drawingWidget), 
			  thisGC,
			  10+(thickStep*i), display_height-50, 
			  outStr, strlen(outStr));
	      for(j=0;j<thickStep;j++)
		XDrawLine(dsply, XtWindow(drawingWidget),  thisGC,
			  10+(i*thickStep)+j, display_height-100,
			  10+(i*thickStep)+j, display_height-90);
	    }/* Draw the zero */
	  
	  XFlush(dsply);
	}/* Do all */
    }/* Draw it on the main screen */

 noVolume:;

}/* UPDATENEWVOLUME */




/* ************************************************************
 * surfDensValueChange() - If they change the value in any of
 * thesw XmTextWidgets this remembers the value.
 */
void surfDensValueChange(Widget widget, int *tag,
			 XmAnyCallbackStruct *callback_data)
{ /* surfDensValueChange */
  char *surfDensText,*stepValText;
  XmString simpleString;
  int nFrames;
  float steps;

  if(*tag==SURFDENS)
    {/* Surface density */

      surfDensText = XmTextGetString(surfDensValueWidget); 

      surfaceDensity = strtod(surfDensText, 0);

      /* BRANT - make sure we don't have any bogus values here.
       * e.g. surfaceDensity <= 1.0 is undefined... ??? */
      /* RWA July 29, 1998
       * OK - irev determines the sign on the surfaceDensity
       * sooo we'll always force it to be positive or 0.0.
       * Display in the text widget based on the irev so we can
       * show them what they are actually looking at when shaded.
       */
      if(surfaceDensity < 0.0)
	surfaceDensity = 0.0 - surfaceDensity;

      /* BRANT - Now set the scroll bar to the value we just got.
       * This DOES NOT cause the scrollbar callback to be called.
       * We need to handle the immediateUpdate and I've done that
       * in surfDens_CR() below.  (Also XtAddActions() above)
       */
      /* surfDens scroll done in tenths so times 10 */
      XtSetArg(arglst[0], XmNvalue, (int)(surfaceDensity*10.0));
      XtSetValues(surfDensScrollWidget, arglst, 1);
      XFlush(dsply);

      maxUnblob = 0.0 - surfaceDensity;

      updateNewVolume();

    }/* Surface density */
  else if(*tag==FRACT)
    {/* Fract */

      surfDensText = XmTextGetString(fractValueWidget); 

      globFract = strtod(surfDensText, 0);

      /* BRANT - Now set the scroll bar to the value we just got.
       * e.g. It has to be done for all widgets... */
      /* This is times 1000 */
      XtSetArg(arglst[0], XmNvalue, (int)(globFract*1000.0));
      XtSetValues(fractScrollWidget, arglst, 1);
      XFlush(dsply);

    }/* Fract */
  else if(*tag==ANGX)
    {/* Angle X */
      surfDensText = XmTextGetString(angxValueWidget); 

      globAngx = strtod(surfDensText, 0);

      /* BRANT - Now set the scroll bar to the value we just got.
       * e.g. It has to be done for all widgets... */
      XtSetArg(arglst[0], XmNvalue, (int)(globAngx));
      XtSetValues(angxScrollWidget, arglst, 1);
      XFlush(dsply);

    }/* Angle X */
  else if(*tag==ANGY)
    {/* Angle Y */
      surfDensText = XmTextGetString(angyValueWidget); 

      globAngy = strtod(surfDensText, 0);

      XtSetArg(arglst[0], XmNvalue, (int)(globAngy));
      XtSetValues(angyScrollWidget, arglst, 1);
      XFlush(dsply);

      globAngy = 0.0 - globAngy;

    }/* Angle Y */
  else if(*tag==OMEGA)
    {/* OMEGA */
      surfDensText = XmTextGetString(omegaValueWidget); 

      globOmega = strtod(surfDensText, 0);

      /* We've got to decide here if what we've got is XYZ or TPO
       * check the rotateXYZ logical and then call XDY's rtns as appropriate
       * XYZ-TPO */
      /* Add 360.0 because we start at -360.0 (which is zero for scroll*/
      XtSetArg(arglst[0], XmNvalue, (int)(globOmega+360.0));
      XtSetValues(omegaScrollWidget, arglst, 1);
      XFlush(dsply);

    }/* OMEGA */
  else if(*tag==PHI)
    {/* PHI */
      surfDensText = XmTextGetString(phiValueWidget); 

      globPhi = strtod(surfDensText, 0);

      /* We've got to decide here if what we've got is XYZ or TPO
       * check the rotateXYZ logical and then call XDY's rtns as appropriate
       * XYZ-TPO */
      /* Add 360.0 because we start at -360.0 (which is zero for scroll*/
      XtSetArg(arglst[0], XmNvalue, (int)(globPhi+360.0));
      XtSetValues(phiScrollWidget, arglst, 1);
      XFlush(dsply);

    }/* PHI */
  else if(*tag==THETA)
    {/* Theta Y */
      surfDensText = XmTextGetString(thetaValueWidget); 

      globTheta = strtod(surfDensText, 0);

      /* We've got to decide here if what we've got is XYZ or TPO
       * check the rotateXYZ logical and then call XDY's rtns as appropriate
       * XYZ-TPO */
      /* Add 360.0 because we start at -360.0 (which is zero for scroll*/
      XtSetArg(arglst[0], XmNvalue, (int)(globTheta+360.0));
      XtSetValues(thetaScrollWidget, arglst, 1);
      XFlush(dsply);


    }/* Theta Y */
  else if(*tag==MINRADIUS)
    {/* Min Radius */
      surfDensText = XmTextGetString(minRadiusValueWidget); 

      globMinRadius = strtod(surfDensText, 0);

      /* BRANT - Now set the scroll bar to the value we just got.
       * e.g. It has to be done for all widgets... */
      XtSetArg(arglst[0], XmNvalue, (int)(globMinRadius));
      XtSetValues(minRadiusScrollWidget, arglst, 1);
      XFlush(dsply);


    }/* Min Radius */
  else if(*tag==MAXRADIUS)
    {/* Max Radius */
      surfDensText = XmTextGetString(maxRadiusValueWidget); 

      globMaxRadius = strtod(surfDensText, 0);

      /* BRANT - Now set the scroll bar to the value we just got.
       * e.g. It has to be done for all widgets... */
      XtSetArg(arglst[0], XmNvalue, (int)(globMaxRadius));
      XtSetValues(maxRadiusScrollWidget, arglst, 1);
      XFlush(dsply);


    }/* Max Radius */
  else if(*tag==PIXELSIZE)
    {/* pixelsize */
      surfDensText = XmTextGetString(pixelSizeValueWidget); 

      globPixelSize = strtod(surfDensText, 0);

      /* BRANT - Now set the scroll bar to the value we just got.
       * e.g. It has to be done for all widgets... */
      XtSetArg(arglst[0], XmNvalue, (int)(globPixelSize*1000.0));
      XtSetValues(pixelSizeScrollWidget, arglst, 1);
      XFlush(dsply);


    }/* pixelsize */
  else if(*tag==STEPTOGGLE)
    {/* Setup for step size etc... */

      if(whichStep==MAXRADIUS || whichStep==MINRADIUS)
	steps = 5.0;
      else if(whichStep==PIXELSIZE)
	steps = 1.0;
      else if(whichStep==ANGY || whichStep == ANGX)
	steps = 5.0;
      else if(whichStep==FRACT)
	steps = 0.05;
      else if(whichStep==SURFDENS)
	{
	  /* We're doing the surfaceDensity in tenths now */
	  if(maxSurf > 5000.0)
	    steps = (maxSurf*0.002)*.01;
	  else if(maxSurf > 500.0)
	    steps = 10.0;
	  else
	    steps = 1.0;
	}
      else if(whichStep==UNBLOBSIZE)
	steps = 0.001;
      else if(whichStep==UNBLOBRADIUS)
	steps = 5.0;
      else if(
	      (whichStep==MAXX) ||
	      (whichStep==MAXY) ||
	      (whichStep==MAXZ)
	      )
	steps = 5.0;
      else if(
	      (whichStep==THETA) ||
	      (whichStep==PHI) ||
	      (whichStep==OMEGA)
	      )
	steps = 5.0;
      else 
	steps = 1.0;

      /* Put the step value */
      sprintf(outStr,"%f",steps);
      XmTextSetString(stepTextWidget,outStr); 


    }/* Setup for step size etc... */
  else if(*tag==STEPTOGGLE2)
    {/* Setup for number of frames etc... */
      int thisFrames;

      /* Get the step value */
      stepValText = XmTextGetString(stepTextWidget); 
      steps       = (float)strtod(stepValText,0);
      XtFree(stepValText);

      /* Get the frames value */
      stepValText = XmTextGetString(frameTextWidget); 
      thisFrames  = (float)strtod(stepValText,0);
      XtFree(stepValText);

      if(steps ==0.0)
	{/* Init it */
	  if(whichStep==MAXRADIUS || whichStep==MINRADIUS)
	    steps = 5.0;
	  else if(whichStep==PIXELSIZE)
	    steps = 1.0;
	  else if(whichStep==ANGY || whichStep == ANGX)
	    steps = 5.0;
	  else if(whichStep==FRACT)
	    steps = 0.05;
	  else if(whichStep==SURFDENS)
	    {
	      /* We're doing the surfaceDensity in tenths now */
	      if(maxSurf > 5000.0)
		steps = (maxSurf*0.002)*.01;
	      else if(maxSurf > 500.0)
		steps = 10.0;
	      else
		steps = 1.0;
	    }
	  else if(whichStep==UNBLOBSIZE)
	    steps = 0.001;
	  else if(whichStep==UNBLOBRADIUS)
	    steps = 5.0;
	  else if(
		  (whichStep==MAXX) ||
		  (whichStep==MAXY) ||
		  (whichStep==MAXZ)
		  )
	    steps = 5.0;
	  else if(
		  (whichStep==THETA) ||
		  (whichStep==PHI) ||
		  (whichStep==OMEGA)
		  )
	    steps = 5.0;
	  else 
	    steps = 1.0;
	  /* Put the step value */
	  sprintf(outStr,"%f",steps);
	  XmTextSetString(stepTextWidget,outStr); 
	}/* Init it */

      if(steps!=0.0)
	{/* OK */
	  if(whichStep==MAXRADIUS || whichStep==MINRADIUS)
	    {/* Calculate how many frames you need */
	      nFrames = (int)((float)(globMaxRadius - globMinRadius)/ steps);
	      nFrames++;
	    }/* Calculate how many frames you need */
	  else if(whichStep==PIXELSIZE)
	    {/* Calculate how many frames you need */
	      if(thisFrames == 0)
		nFrames = 1;
	      else
		nFrames = thisFrames;
	    }/* Calculate how many frames you need */
	  else if(whichStep==ANGY || whichStep == ANGX)
	    {/* Calculate how many frames you need */
	      if(thisFrames == 0)
		nFrames = 1;
	      else
		nFrames = thisFrames;
	    }/* Calculate how many frames you need */
	  else if(whichStep==FRACT)
	    {/* Calculate how many frames you need */
	      if(steps > 0)
		nFrames = (float)(1.00000001 - globFract) / (float)steps;
	      else nFrames = (float)(globFract) / (float)steps;
	    }/* Calculate how many frames you need */
	  else if(whichStep==SURFDENS)
	    {/* Calculate how many frames you need */
	      if(thisFrames == 0)
		nFrames = 1;
	      else
		nFrames = thisFrames;
	    }/* Calculate how many frames you need */
	  else if(whichStep==UNBLOBSIZE)
	    {/* unblobSize changed */
	      if(thisFrames == 0)
		nFrames = 1;
	      else
		nFrames = thisFrames;
	    }/* unblobSize changed */
	  else if(whichStep==UNBLOBRADIUS)
	    {/* unblobRadius changed */
	      if(thisFrames == 0)
		nFrames = 1;
	      else
		nFrames = thisFrames;
	    }/* unblobRadius changed */
	  else if(
		  (whichStep==MAXX) ||
		  (whichStep==MAXY) ||
		  (whichStep==MAXZ)
		  )
	    {/* maxx changed */
	      if(whichStep==MAXX)
		{
		  nFrames = (int)((float)(cutMaxValues[0] - 
				   cutMinValues[0])/steps);
		}
	      if(whichStep==MAXY)
		{
		  nFrames = (int)((float)(cutMaxValues[1] - 
				   cutMinValues[1])/steps);
		}
	      if(whichStep==MAXZ)
		{
		  nFrames = (int)((float)(cutMaxValues[2] - 
				   cutMinValues[2])/steps);
		}
	      if(steps > 0)
		nFrames--;  /* Always one less */
	      else nFrames++;  /* Always one less */
	    }/* maxx changed */
	  else if(
		  (whichStep==THETA) ||
		  (whichStep==PHI) ||
		  (whichStep==OMEGA)
		  )
	    {/* Just assume it's angles */
	      if(globPhi==0.0)
		nFrames = (int)((180.0)/ steps);     /* 180 degrees */
	      else nFrames = (int)((360.0)/ steps);  /* 360 degrees */
	    }/* Just assume it's angles */
	  else 
	    {/* default */
	      nFrames = 1;
	    }/* default */
	}/* OK */
      else nFrames = 1;
      
      nFrames = (int)fabs((float)nFrames);

      sprintf(outStr,"%d",nFrames);
      XmTextSetString(frameTextWidget, outStr); 
      XFlush(dsply);

    }/* Setup for number of frames etc... */

} /* surfDensValueChange */




/*****************************************************************
 * surfDens_CR() - When a carriage return happens in any of
 * the text widgets on the depth/radial cueing screen call here.
*/
void surfDens_CR(Widget widget, int *tag, 
		 XmAnyCallbackStruct *callback_data)
{/* surfDens_CR */
  int i;

  mapShade_activate(0,&i,0);   /* Don't do this if we're just dragging */

}/* surfDens_CR */



/* ************************************************************
 * surfDensScrollChange() - Based on the change in the scroll 
 * bars below adjust the internal program variable accordingly.
 */
void surfDensScrollChange(Widget widget, int *tag, 
			  XmScrollBarCallbackStruct *callback_data)
{/* surfDensScrollChange */
  int newValue,i;
  XmString simpleString;

  if(*tag==SURFDENS)
    {/* Surface density */

      surfaceDensity= (float)callback_data->value;
      surfaceDensity *= 0.1;  /* Do it in tenths */
      maxUnblob = 0.0 - surfaceDensity;

      /* RWA July 29, 1998
       * OK - irev determines the sign on the surfaceDensity
       * sooo we'll always force it to be positive or 0.0.
       * Display in the text widget based on the irev so we can
       * show them what they are actually looking at when shaded.
       * Actually surfaceDenity should always be positive here as 
       * we're getting it from the scroll bar.
       */
      if(surfaceDensity < 0.0)
	surfaceDensity = 0.0 - surfaceDensity;
      sprintf(outStr,"%.2f",irev*surfaceDensity);
      XmTextSetString(surfDensValueWidget,outStr); 

      updateNewVolume();

    }/* Surface density */
  else if(*tag==FRACT)
    {/* fract */

      globFract= (float)callback_data->value;
      globFract *= 0.001;  /* This is in 1000ths */
      
      /* Give feedback */
      sprintf(outStr,"%.2f",globFract);
      createSimpleString(&simpleString,outStr);
      XtSetArg(arglst[0], XmNvalue, outStr );
      XtSetArg(arglst[1], XmNrecomputeSize, FALSE );
      XtSetValues(fractValueWidget, arglst, 2); 
      XFlush(dsply);
      XmStringFree(simpleString); 

    }/* fract */
  else if(*tag==ANGX)
    {/* angx */
      globAngx= (double)callback_data->value;
      /* Give feedback */
      sprintf(outStr,"%3d",callback_data->value);
      createSimpleString(&simpleString,outStr);
      XtSetArg(arglst[0], XmNvalue, outStr );
      XtSetArg(arglst[1], XmNrecomputeSize, FALSE );
      XtSetValues(angxValueWidget, arglst, 2); 
      XFlush(dsply);
      XmStringFree(simpleString); 

    }/* angx */
  else if(*tag==ANGY)
    {/* angy */
      globAngy= 0.0 - (double)callback_data->value;
      /* Give feedback */
      sprintf(outStr,"%3d",callback_data->value);
      createSimpleString(&simpleString,outStr);
      XtSetArg(arglst[0], XmNvalue, outStr );
      XtSetArg(arglst[1], XmNrecomputeSize, FALSE );
      XtSetValues(angyValueWidget, arglst, 2); 
      XFlush(dsply);
      XmStringFree(simpleString); 

    }/* angy */
  else if(*tag==THETA)
    {/* theta */
      globTheta= (double)callback_data->value;

      /* We've got to decide here if what we've got is XYZ or TPO
       * check the rotateXYZ logical and then call XDY's rtns as appropriate
       * XYZ-TPO */

      globTheta -= 360.0;
      if(globTheta != 0)
	globTilt = TRUE;
      else globTilt = FALSE;

      /* Give feedback */
      sprintf(outStr,"%.0f",globTheta);
      createSimpleString(&simpleString, outStr);
      XtSetArg(arglst[0], XmNvalue, outStr );
      XtSetArg(arglst[1], XmNrecomputeSize, FALSE );
      XtSetValues(thetaValueWidget, arglst, 2); 
      XFlush(dsply);
      XmStringFree(simpleString); 

    }/* theta */
  else if(*tag==OMEGA)
    {/* omega */
      globOmega= (double)callback_data->value;

      /* We've got to decide here if what we've got is XYZ or TPO
       * check the rotateXYZ logical and then call XDY's rtns as appropriate
       * XYZ-TPO */

      globOmega -= 360.0;
      if(globOmega != 0)
	globTilt = TRUE;
      else globTilt = FALSE;

      /* Give feedback */
      sprintf(outStr,"%.0f",globOmega);
      createSimpleString(&simpleString, outStr);
      XtSetArg(arglst[0], XmNvalue, outStr );
      XtSetArg(arglst[1], XmNrecomputeSize, FALSE );
      XtSetValues(omegaValueWidget, arglst, 2); 
      XFlush(dsply);
      XmStringFree(simpleString); 

    }/* omega */
  else if(*tag==PHI)
    {/* phi */
      globPhi= (double)callback_data->value;

      /* We've got to decide here if what we've got is XYZ or TPO
       * check the rotateXYZ logical and then call XDY's rtns as appropriate
       * XYZ-TPO */

      globPhi -= 360.0;
      if(globPhi != 0)
	globTilt = TRUE;
      else globTilt = FALSE;

      /* Give feedback */
      sprintf(outStr,"%.0f",globPhi);
      createSimpleString(&simpleString, outStr);
      XtSetArg(arglst[0], XmNvalue, outStr );
      XtSetArg(arglst[1], XmNrecomputeSize, FALSE );
      XtSetValues(phiValueWidget, arglst, 2); 
      XFlush(dsply);
      XmStringFree(simpleString); 

    }/* phi */
  else if(*tag==MAXRADIUS)
    {/* theta */
      globMaxRadius= (double)callback_data->value;

      /* Give feedback */
      sprintf(outStr,"%.0f",globMaxRadius);
      createSimpleString(&simpleString,outStr);
      XtSetArg(arglst[0], XmNvalue, outStr );
      XtSetArg(arglst[1], XmNrecomputeSize, FALSE );
      XtSetValues(maxRadiusValueWidget, arglst, 2); 
      XFlush(dsply);
      XmStringFree(simpleString); 

    }/* theta */
  else if(*tag==MINRADIUS)
    {/* minRadius */
      globMinRadius= (double)callback_data->value;

      /* Give feedback */
      sprintf(outStr,"%.0f",globMinRadius);
      createSimpleString(&simpleString,outStr);
      XtSetArg(arglst[0], XmNvalue, outStr );
      XtSetArg(arglst[1], XmNrecomputeSize, FALSE );
      XtSetValues(minRadiusValueWidget, arglst, 2); 
      XFlush(dsply);
      XmStringFree(simpleString); 

    }/* minRadius */
  else if(*tag==PIXELSIZE)
    {/* pixelSize */
      globPixelSize= (double)callback_data->value;
      globPixelSize *= 0.001;  /* Let's do these in thousands */

      /* Give feedback */
      sprintf(outStr,"%.2f",globPixelSize);
      createSimpleString(&simpleString,outStr);
      XtSetArg(arglst[0], XmNvalue, outStr );
      XtSetArg(arglst[1], XmNrecomputeSize, FALSE );
      XtSetValues(pixelSizeValueWidget, arglst, 2); 
      XFlush(dsply);
      XmStringFree(simpleString); 

    }/* pixelSize */

  if(immediateMapShade)
    mapShade_activate(0,&i,0);   /* Don't do this if we're just dragging */

}/* surfDensScrollChange */



/* ************************************************************
 * mapShadeScreen_deactivate() - They're done shading.
 */
void mapShadeScreen_deactivate(Widget widget, int *tag, 
			       XmAnyCallbackStruct *callback_data)
{/* mapShadeScreen_deactivate */
  int m;

  XtUnmanageChild(mapShadeScreenWidget);
  XtUnmanageChild(editBoxScreenWidget);
  XFlush(dsply);
  rotMap = NULL;
  tmpRotMap = NULL;

  for(m=0;m<MAXSECTIONS;m++)
    {/* Try different angles */
      if(cuedMap[m].radial != NULL)
	free(cuedMap[m].radial);
      cuedMap[m].radial= NULL;
      if(cuedMap[m].depth != NULL)
	free(cuedMap[m].depth);
      cuedMap[m].depth= NULL;
    }/* Try different angles */
  /* Already done in mapShade_activate()  free(unblobBuffer); */
  unblobBuffer = NULL;
  unblobSwitch = FALSE;

  /* It's important to remember how many sections we've got in the
   * original map file especially since the depthcue and shading 
   * routines rely on it!!!
   */
  numSecs= origNumSecs;

  if(XtIsManaged(subtractMapScreen))
    {/* Set it back to doing sections */
      maxMapPixel = purdueHeader.img_max;
      minMapPixel = purdueHeader.img_min;
      dispSections = TRUE;
      diffSection  = TRUE;
      imgNum = (purdueHeader.numCols / 2)+1;
      diffSec_really_doit();
      /*
      stuffMapImage((float *)map_image_buffer, 
		      (float *)float_image_buffer1, 
		      file_width,file_height,
		      1,imgNum, 0);
      stuffMapImage((float *)subMapYmxb_buffer, 
		    (float *)subMapYmxb_buffer, 
		    file_width,file_height,
		    1,1, numCols*sizeFactor);
      */
    }/* Set it back to doing sections */

}/* mapShadeScreen_deactivate */



/*************************************************************
 * label_create() - Remember widget numbers!
 */
void label_create(Widget widget, int *tag, XmAnyCallbackStruct *callback_data)
{/* LABEL_CREATE */

  if(*tag==FRACT)
    fractLabelWidget = widget;
  else if(*tag==SURFDENS)
    surfDensLabelWidget = widget;
  else if(*tag==ANGX)
    angxLabelWidget = widget;
  else if(*tag==ANGY)
    angyLabelWidget = widget;
  else if(*tag==TIMELEFT)
    timeLeftLabelWidget = widget;
  else if(*tag==TIMELEFTBOX)
    timeLeftBoxW = widget;
  else if (*tag==MOVIEMAX)
    maxLabelWidget = widget;
  else if (*tag==MOVIEMIN)
    minLabelWidget = widget;
  else if (*tag==LIGHTSOURCE)
    lightSourceLabelWidget = widget;
  else if (*tag==CONTOUR)
    movieContourLabelWidget = widget;
  else if (*tag==DELTA)
    movieDeltaLabelWidget = widget;
  else if (*tag==NUMCONT)
    movieNumContLabelWidget = widget;
  else if (*tag==BREAKPOINT)
    movieBreakLabelWidget = widget;
  else if (*tag==CONTOUREXIT)
    contourExitW = widget;
  else if (*tag==STATEXIT)
    statExitW = widget;
  else if (*tag==ROTFLTEXIT)
    rotfltExitW = widget;
  else if (*tag==CRYSDISPEXIT)
    crysDispExitW = widget;
  else if (*tag==EMCORAVGEXIT)
    emcoravgExitW = widget;
  else if (*tag==IMGCONVEXIT)
    imgconvExitW = widget;
  else if (*tag==IMGCONVHELP)
    imgconvHelpW = widget;
  else if (*tag==CIRCAVG)
    circAvgExitW = widget;
  else if (*tag==OPTIONSEXIT)
    optionsExitW = widget;
  else if (*tag==FFTEXIT)
    fftExitW = widget;
  else if (*tag==POINTEXIT)
    pointExitW = widget;
  else if (*tag==POINTCOLOR)
    pointColorW = widget;
  else if (*tag==BOXEXIT)
    boxExitW = widget;
  else if (*tag==AUTOBOX)
    minBoxScrollWidget = widget;
  else if (*tag==CORRELATE)
    maxBoxScrollWidget = widget;
  else if (*tag==AUTOBOXLABEL)
    autoBoxLabelW = widget;
  else if (*tag==REFCORRLABEL)
    refCorrLabelW = widget;
  else if (*tag==AUTOCORRLABEL)
    autoCorrLabelW = widget;
  else if (*tag==DILATE)
    dilateScrollW = widget;
  else if (*tag==DILATELABEL)
    dilateLabelW = widget;
  else if (*tag==THINNING)
    thinningScrollW = widget;
  else if (*tag==THINLABEL)
    thinLabelW = widget;
  else if (*tag==PRAG)
    pragScrollW = widget;
  else if (*tag==PRAGLABEL)
    pragLabelW = widget;
  else if (*tag==REFCUTOFF)
    refCutoffScrollW = widget;
  else if (*tag==REFCUTOFFLABEL)
    refCutoffLabelW = widget;
  else if (*tag==STOP)
    versionStopW = widget;
  else if (*tag==SCROLL)
    versionScrollW = widget;
  else if (*tag==VERSIONEXIT)
    versionExitW = widget;
  else if (*tag==STATUSSTOP)
    statusStopW = widget;
  else if (*tag==STATUSW)
    statusTextW = widget;
  else if (*tag==PARTPROCEXIT)
    partprocExitW = widget;
  else if (*tag==POINTTEXT)
    partPointTextW = widget;
  else if (*tag==MOVELABEL)
    movementLabelWidget = widget;
  else if (*tag==MOVEEXIT)
    moveExitW = widget;
  else if (*tag==RESCALEEXIT)
    rescaleExitW = widget;
  else if (*tag==HELPEXIT)
    helpExitW = widget;
  else if (*tag==FOURIERFILTEXIT)
    fourierFiltExitW = widget;
  else if (*tag==ROTFLTHELP)
    rotfltHelpW = widget;
  else if (*tag==BOXHELP)
    boxHelpW = widget;
  else if (*tag==APPODIZEHELP)
    appodizeHelpW = widget;
  else if (*tag==FOURIERFILTHELP)
    fourierFiltHelpW = widget;
  else if (*tag==FFTHELP)
    fftHelpW = widget;
  else if (*tag==FOURIERFILT)
    fourierFiltLabelW = widget;
  else if (*tag==INVERTEXIT)
    invertExitW = widget;
  else if (*tag==BATCHJOB)
    batchjobExitW = widget;
  else if (*tag==BATCHFILELABEL)
    batchjobFileLabelW = widget;
  else if (*tag==BINPPUP)
    {binPPUpW = widget;
     handle_cr(binPPUpW);}
  else if (*tag==BINPPDOWN)
    {binPPDownW = widget;
     handle_cr(binPPDownW);}
  else if (*tag==STEPPPUP)
    {stepPPUpW = widget;
     handle_cr(stepPPUpW);}
  else if (*tag==STEPPPDOWN)
    {stepPPDownW = widget;
     handle_cr(stepPPDownW);}
  else if (*tag==PPINUP)
    {ppinUpW = widget;
     handle_cr(ppinUpW);}
  else if (*tag==PPINDOWN)
    {ppinDownW = widget;
     handle_cr(ppinDownW);}
  else if (*tag==PROJASU)
    projasuW = widget;
  else if (*tag==TAUP)
    {taUpArrowW = widget;
     handle_cr(taUpArrowW);}
  else if (*tag==TADOWN)
    {taDownArrowW = widget;
     handle_cr(taDownArrowW);}
  else if (*tag==GBSUP)
    {gbsUpArrowW = widget;
     handle_cr(gbsUpArrowW);}
  else if (*tag==GBSDOWN)
    {gbsDownArrowW = widget;
     handle_cr(gbsDownArrowW);}
  else if (*tag==TAUP)
    {taUpArrowW = widget;
     handle_cr(taUpArrowW);}
  else if (*tag==TADOWN)
    {taDownArrowW = widget;
     handle_cr(taDownArrowW);}
  else if (*tag==TA2UP)
    {ta2UpArrowW = widget;
     handle_cr(ta2UpArrowW);}
  else if (*tag==TA2DOWN)
    {ta2DownArrowW = widget;
     handle_cr(ta2DownArrowW);}
  else if (*tag==OLDMINDENS)
    oldMinDensLabelW = widget;
  else if (*tag==OLDMAXDENS)
    oldMaxDensLabelW = widget;
  else if(*tag==BATCHHELPLABEL)
    batchHelpLabelW = widget;
  else if(*tag==PPMIN)
    ppMinScrollW = widget;
  else if(*tag==PPMAX)
    ppMaxScrollW = widget;
  else if(*tag==PPMINMAX)
    ppMinMaxFormW = widget;
  else if(*tag==USEIMMEDIATE)
    immediateUpdateW = widget;
  else if(*tag==MBUSEIMMEDIATE)
    mbImmediateUpdateW = widget;
  else if(*tag==ICOSREPLX)
    icosReplRadiusLabelXW = widget;
  else if(*tag==ICOSREPLY)
    icosReplRadiusLabelYW = widget;
  else if(*tag==ICOSREPLZ)
    icosReplRadiusLabelZW = widget;
  else if(*tag==TRUEPHI)
    truePhiW = widget;
  else if(*tag==CALCBIN)
    calcBinW = widget;
  else if(*tag==CALCSFS)
    calcSfsW = widget;
  else if(*tag==USEMAPSHADE)
    mapShadeHelpW = widget;
  else if(*tag==HELPTOOLW)
    helpToolW = widget;
  else if(*tag==FIXADD)
    helpFixW = widget;
  else if(*tag==FIX2D)
    helpFix2DW = widget;
  else if(*tag==THETAT)
    helpRotateW = widget;
  else if(*tag==TABEGX)
    helpTruncAddW = widget;
  else if(*tag==SUBTRACTMAPEXIT)
    subtractExitW = widget;
  else if(*tag==SUBTRACTMAPHELP)
    subtractHelpW = widget;
  else if(*tag==UNITCELL)
    ppHelpW = widget;
  else if(*tag==FOURAVERHELP)
    fourAverHelpW = widget;
  else if(*tag==FOURAVERRESET)
    fourAverResetW = widget;
  else if(*tag==FOURAVEREXIT)
    fourAverExitW = widget;
  else if(*tag==BACKSIZE)
    backSizeW = widget;
  else if(*tag==FAREFINE)
    refineLabelW = widget;
  else if(*tag==IMPOSE)
    imposeLabelW = widget;
  else if(*tag==MANIPPPHELP)
    manipPPHelpW = widget;
  else if(*tag==CAVGPPHELP)
    cavgPPHelpW = widget;
  else if(*tag==DELTAOMEGAPPHELP)
    deltaOmegaPPHelpW = widget;
  else if(*tag==MANIPPPEXIT)
    manipPPExitW = widget;
  else if(*tag==CAVGPPEXIT)
    cavgPPExitW = widget;
  else if(*tag==DELTAOMEGAPPEXIT)
    deltaOmegaPPExitW = widget;
  else if(*tag==DELTAXYPARTPROCHELP)
    deltaXYpartprocHelpW = widget;
  else if(*tag==DELTAXYPARTPROCEXIT)
    deltaXYpartprocExitW = widget;
  else if(*tag==NORMALHELP)
    normalHelpW = widget;
  else if(*tag==NORMALEXIT)
    normalExitW = widget;
  else if(*tag==AUTODREXIT)
    autoDRExitW = widget;
  else if(*tag==AUTODRHELP)
    autoDRHelpW = widget;
  else if(*tag==AUTO2DDREXIT)
    auto2DDRExitW = widget;
  else if(*tag==AUTO2DDRHELP)
    auto2DDRHelpW = widget;
  else if(*tag==THETA)
    thetaButtonW = widget;
  else if(*tag==PHI)
    phiButtonW = widget;
  else if(*tag==OMEGA)
    omegaButtonW = widget;
  else if(*tag==STOPAVGFFT)
    fftStopW = widget;
  else if(*tag==APPODIZEEXIT)
    appodizeExitW = widget;
  else if(*tag==MAGSECSCROLL)
    magSecScrollWidget = widget;
  else if(*tag==THRESHSCROLL)
    threshScrollWidget = widget;
  else if(*tag==DIFFSECSCROLL)
    diffSecScrollWidget = widget;
  else if(*tag==DIFFSEC2SCROLL)
    diffSec2ScrollWidget = widget;
  else if(*tag==SUBEQUATION)
    subEquationW = widget;
  else if(*tag==SMMIN)
    smMinLabelW = widget;
  else if(*tag==SMMAX)
    smMaxLabelW = widget;
  else if(*tag==SMMINMAXRESET)
    smMinMaxResetW = widget;
  else if(*tag==FLICKERTYPES)
    flickerTypeWidget = widget;
  else if(*tag==DIFFSECRESET)
    diffSecResetW = widget;
  else if(*tag==MAGSECRESET)
    magSecResetW = widget;
  else if(*tag==COMPUTE_SCALING)
    computeScalingW= widget;
  else if(*tag==TADIMX)
    taDimXW= widget;
  else if(*tag==TADIMY)
    taDimYW= widget;
  else if(*tag==TADIMZ)
    taDimZW= widget;
  else if(*tag==SUBTRACTMAP2D)
    subMap2dW= widget;
  else if(*tag==SUBTRACTMAP3D)
    subMap3dW= widget;
  else if(*tag==MAXRADUP)
    {smMaxRadUpW= widget;
     handle_cr(widget);}
  else if(*tag==MAXRADDOWN)
    {smMaxRadDownW= widget;
     handle_cr(widget);}
  else if(*tag==MINRADUP)
    {smMinRadUpW= widget;
     handle_cr(widget);}
  else if(*tag==MINRADDOWN)
    {smMinRadDownW= widget;
     handle_cr(widget);}
  else if(*tag==FINDMAGRESET)
    {findMagResetW = widget;
   }
  else if(*tag==FINDRADRESET)
    {findRadResetW = widget;
   }
  else if(*tag==FINDMAGRESET2)
    {findMag2ResetW = widget;
   }
  else if(*tag==FINDMAGRESET3)
    {findMag3ResetW = widget;
   }
  else if(*tag==FINDRADDISP)
    {findRadDispW = widget;
   }
  else if(*tag==FINDMAGDISP)
    {findMagDispW = widget;
   }
  else if(*tag==PHASYUP)
    {phasYupW= widget;
     handle_cr(widget);}
  else if(*tag==PHASYDOWN)
    {phasYdownW= widget;
     handle_cr(widget);}
  else if(*tag==PHASXUP)
    {phasXupW= widget;
     handle_cr(widget);}
  else if(*tag==PHASXDOWN)
    {phasXdownW= widget;
     handle_cr(widget);}
  else if(*tag==SHIFTORIGX)
    shiftOrigXlabelW= widget;
  else if(*tag==SHIFTORIGY)
    shiftOrigYlabelW= widget;
  else if(*tag==MOVEORIGX)
    moveOrigXlabelW= widget;
  else if(*tag==MOVEORIGY)
    moveOrigYlabelW= widget;
  else if(*tag==MODEL)
    modelLabelW= widget;
  else if(*tag==MODELHELP)
    modelHelpW= widget;
  else if(*tag==MODELEXIT)
    modelExitW= widget;
  else if(*tag==BINNINGHELP)
    binningHelpW= widget;
  else if(*tag==BINNINGEXIT)
    binningExitW= widget;
  else if(*tag==INTERPHELP)
    interpHelpW= widget;
  else if(*tag==INTERPEXIT)
    interpExitW= widget;
  else if(*tag==FFTFEEDEXIT)
    fftFeedbackExitW= widget;
  else if(*tag==DIOEXIT)
    dioExitW= widget;
  else if(*tag==DIOHELP)
    dioHelpW= widget;
  else if(*tag==IMGCONVHELP)
    imgconvHelpW= widget;
  else if(*tag==CIRCLEDISP)
    circleDispWidget= widget;
  else if(*tag==FFTMOTIONFEED)
    fftMotionFeedbackWidget= widget;
  else if(*tag==ADRABRESET)
    adrabResetW= widget;
  else if(*tag==WRITECAVG)
    {writeCavgToggleW= widget;
     XtUnmanageChild(writeCavgToggleW);}
  else if(*tag==NFOLDTEXT)
    {nFoldLabelW = widget;
     XtUnmanageChild(nFoldLabelW);}
  else if(*tag==BOXCOLOR)
    boxcolorPushW = widget;
  else if(*tag==FIXGAUSSSTD)
    {
      fixGaussStdLabelW = widget;
      XtUnmanageChild(widget);
    }
  else if(*tag==APRESET)
    apOrigResetW = widget;
  else if(*tag==REFINERESET)
    refineResetLabelW = widget;
  else if(*tag==FRAME)
    numFramesLabelW = widget;
  else if(*tag==IMGCONVLABEL)
    imgconvLabelW = widget;
  else if(*tag==FORCEAVGBKG)
    forceAvgBkgToggleW = widget;
  else if(*tag==REFREADFILE)
    refReadFileLabelW = widget;
  else if(*tag==A2DDRWARN)
    a2ddrWarnW = widget;
  else if(*tag==INFFLICKER)
    infFlickerWidget = widget;
  else if(*tag==ROTATETSCROLL)
    mbRotatetScrollW = widget;
  else if(*tag==ROTATEPSCROLL)
    mbRotatepScrollW = widget;
  else if(*tag==ROTATEOSCROLL)
    mbRotateoScrollW = widget;
  else if(*tag==DIMENSIONXSCROLL)
    mbDimensionxScrollW = widget;
  else if(*tag==DIMENSIONYSCROLL)
    mbDimensionyScrollW = widget;
  else if(*tag==DIMENSIONZSCROLL)
    mbDimensionzScrollW = widget;
  else if(*tag==CENTERXSCROLL)
    mbCenterxScrollW = widget;
  else if(*tag==CENTERYSCROLL)
    mbCenteryScrollW = widget;
  else if(*tag==CENTERZSCROLL)
    mbCenterzScrollW = widget;
  else if(*tag==OLD)
    apOldW = widget;
  else if(*tag==EXPONENTIAL)
    apExponentialW = widget;
  else if(*tag==GAUSSIAN)
    apGaussianW = widget;
  else if(*tag==LINEAR)
    apLinearW = widget;
  else if(*tag==BOXGRID2)
    boxGrid2W = widget;
  

}/* LABEL_CREATE */




/* ************************************************************
 * surfDensScrollCreate() - Remember widget numbers!
 */
void surfDensScrollCreate(Widget  widget, int *tag,
			  XmAnyCallbackStruct *callback_data)
{/* surfDensScrollCreate */

  if(*tag==SURFDENS)
    surfDensScrollWidget = widget;
  else if(*tag==FRACT)
    fractScrollWidget = widget;
  else if(*tag==ANGX)
    angxScrollWidget = widget;
  else if(*tag==ANGY)
    angyScrollWidget = widget;
  else if(*tag==THETA)
    thetaScrollWidget = widget;
  else if(*tag==PHI)
    phiScrollWidget = widget;
  else if(*tag==OMEGA)
    omegaScrollWidget = widget;
  else if(*tag==MAXRADIUS)
    maxRadiusScrollWidget = widget;
  else if(*tag==PIXELSIZE)
    pixelSizeScrollWidget = widget;
  else if(*tag==MINRADIUS)
    minRadiusScrollWidget = widget;
  else if(*tag==USEMAPSHADE)
    secondShadeWidget = widget;
  else if(*tag==USERADIAL)
    radialShadeWidget = widget;
  else if(*tag==USERADPROJ)
    radProjShadeWidget = widget;
  else if(*tag==USESTEREO)
    stereoShadeWidget = widget;
  else if(*tag==LOHIRES)
    loHiResWidget = widget;
  else if(*tag==MAPSHADESTOP)
    mapShadeStopWidget= widget;
  else if(*tag==USECONT)
    useContWidget= widget;

  /* A noop if widget isn't inside handle_cr()... */
  handle_cr(widget);

}/* surfDensScrollCreate */


/* ************************************************************
 * surfDensValue_create() - Remember widget numbers!
 */
void surfDensValue_create(Widget widget, int *tag, 
			  XmAnyCallbackStruct *callback_data)
{/* surfDensValue_create */

  if(*tag==SURFDENS)
    surfDensValueWidget = widget;
  else if(*tag==FRACT)
    fractValueWidget = widget;
  else if(*tag==ANGX)
    angxValueWidget = widget;
  else if(*tag==ANGY)
    angyValueWidget = widget;
  else if(*tag==THETA)
    thetaValueWidget = widget;
  else if(*tag==PHI)
    phiValueWidget = widget;
  else if(*tag==OMEGA)
    omegaValueWidget = widget;
  else if(*tag==MAXRADIUS)
    maxRadiusValueWidget = widget;
  else if(*tag==MINRADIUS)
    minRadiusValueWidget = widget;
  else if(*tag==VOLUME)
    volumeValueW = widget;
  else if(*tag==PIXELSIZE)
    pixelSizeValueWidget = widget;

  handle_cr(widget);

}/* surfDensValue_create */



/********************************************************
 * clearMap() - Clear the memory just allocated.
*/
void clearMap(float *thisMap,int nCols,int nRows, int nSecs)
{/* CLEARMAP */
  int i,j,k,thisSec,thisRow;

  for(i=0;i<nSecs;i++)
    {/* Sec */
      thisSec = (i*nCols*nRows);
      for(j=0;j<nRows;j++)
	{/* row */
	  thisRow = thisSec + (j*nCols);
	  for(k=0;k<nCols;k++)
	    thisMap[(thisRow + k)]= 0.0;
	}/* row */	    
    }/* Sec */

}/* CLEARMAP */





/*************************************************************
 * mapShade_activate() - They want to see their map sections
 * displayed as a depth-cue shade 3D surface.
 */
void mapShade_activate(Widget widget, int *tag, 
		       XmAnyCallbackStruct *callback_data)
{/* mapShade_activate */

  if(paintMapSwitch)
    {/* Let's just do the paintMap */
      mapShade_activate_work(widget, 
			     tag, 
			     callback_data,
			     &paintMap,
			     &map_image_buffer,
			     FALSE);
      
    }/* Let's just do the paintMap */
  else
    {/* Normal */
      mapShade_activate_work(widget, 
			     tag, 
			     callback_data,
			     &float_image_buffer1,
			     &map_image_buffer,
			     FALSE);
      
      if(subtract_image_buffer2)
	{/* Let's surface shade the second one also */
	  mapShade_activate_work(widget, 
				 tag, 
				 callback_data,
				 &subtract_image_buffer2,
				 &subtract_map_image_buffer2,
				 TRUE);
	}/* Let's surface shade the second one also */
      
      if(subtract_image_buffer1)
	{/* Let's surface shade the second one also */
	  mapShade_activate_work(widget, 
				 tag, 
				 callback_data,
				 &subtract_image_buffer1,
				 &subtract_map_image_buffer,
				 TRUE);
	}/* Let's surface shade the second one also */
      
      if(subtract_image_buffer2 || subtract_image_buffer1)
	diffSec_really_doit();
    }/* Normal */

}/* mapShade_activate */





/*************************************************************
 * mapShade_activate_work() - They want to see their map sections
 * displayed as a depth-cue shade 3D surface.
 * this rtmn does all of the real work in setting up for
 * 3D rendering
 */
void mapShade_activate_work(Widget widget, int *tag, 
			    XmAnyCallbackStruct *callback_data,
			    float **float_img_buf1,
			    float **map_img_buf,
			    int calcMapPixel)
{/* MAPSHADE_ACTIVATE_WORK */
  XmString simpleString;
  float tmp;
  char ikey;
  int i,j,k,l,m,o,numPasses;
  int col1,col2,row1,row2,sec1,sec2;
  int lun,numFrames,secLeft,minLeft;
  int thisSec,thisSecRow;
  float thisPixel;
  float *thisMap, *thisRadialMap, *thisShadMap; 
  char *numFramesText;
  float stepVal;
  char *stepValText;
  char *expValText;
  time_t tloc,oldtloc;
  struct tm *thisTm;
  char timeStr[256];
  float saveSurfaceDensity, saveGlobFract, 
        saveGlobAngx, saveGlobAngy, saveUnblobSize; 
  double saveGlobTheta, saveGlobPhi, saveGlobOmega, saveGlobMaxRadius,
         saveGlobMinRadius, saveGlobPixelSize;
  int thisExpStep,numFramesMm;
  float radialXcen, radialYcen, radialZcen;
  float thisMinMapPixel,thisMaxMapPixel;
  int   topRadialSec;
  float saveCutMax0Value,saveCutMax1Value,saveCutMax2Value;
  int samRenStep;
  char *thisStr;
	

  /* Give feedback on what the surface density *really* is... */
  /* Make sure the user sees it with the right sign */
  sprintf(outStr,"%.2f",irev*surfaceDensity);
  XmTextSetString(surfDensValueWidget,outStr); 

  saveCutMax0Value = cutMaxValues[0];
  saveCutMax1Value = cutMaxValues[1];
  saveCutMax2Value = cutMaxValues[2];

  if(*tag==MAPSHADESTOP)
    {/* Do nothing */
      XtSetArg(arglst[0], XmNlabelPixmap, stopPixmap);
      XtSetArg(arglst[1], XmNselectPixmap, stoppedPixmap);
      XtSetArg(arglst[2], XmNrecomputeSize, FALSE);
      XtSetValues(mapShadeStopWidget, arglst, 3); 
      XFlush(dsply);
    }/* Do nothing */
  else
    if(fileType==MAP)
      {/* OK we've got map data */

	/* Get the sample rendering step */
	thisStr = XmTextGetString(sampleRenderWidget); 
	samRenStep = (int)strtod(thisStr,0);
	XtFree(thisStr);
	
	firstSec= 0;
	lastSec = origNumSecs;

	/* Init the Unblob stuff */
	minUnblob= (int)secMinPixel;
	maxUnblob= (int)0.0 - surfaceDensity;
	
	/* Ok they're gonna have a 3D image so we can write it as tiff 
	 * if they want. */
	XtSetArg(arglst[0], XmNsensitive, TRUE);
	XtSetValues(rawButtonW, arglst, 1); 
	XFlush(dsply);
	
	/* Get the step value */
	stepValText = XmTextGetString(stepTextWidget); 
	stepVal     = (float)strtod(stepValText,0);
	XtFree(stepValText);
	if(whichStep==UNBLOBSIZE)
	  {/* This is done is percentages */
	    /* Divide by 100 for percentage */
	    stepVal= (float)((numCols*numRows*origNumSecs) * (stepVal/100.0));
	  }/* This is done is percentages */
	if(stepVal==0.0)
	  stepVal = 1.0;
	
	expValText = XmTextGetString(radialTextWidget); 
	thisExpStep     = (int)strtod(expValText,0);
	XtFree(expValText);

	if(thisExpStep > 0)
	  expandStep = MIN(MAX(4,thisExpStep),12);
	sprintf(outStr,"%d",expandStep);
	XmTextSetString(radialTextWidget, outStr);
	XFlush(dsply);
	
	if(whichStep==MAXRADIUS)
	  {/* Calculate how many frames you need */
	    numFrames = (int)((globMaxRadius - globMinRadius)/ stepVal);
	    /* NO???  	    numFrames++; */
	    sprintf(outStr,"%d",numFrames);
	    XmTextSetString(frameTextWidget, outStr); 
	    XFlush(dsply);
	  }/* Calculate how many frames you need */
	if(whichStep==MAXX || whichStep==MAXY || whichStep==MAXZ
	   )
	  {/* Calculate how many frames you need */
	    int maxFrames;

	    /* First see what they think they want */
	    numFramesText = XmTextGetString(frameTextWidget); 
	    numFrames = (int)strtod(numFramesText,0);
	    numFrames = MAX(numFrames,0);
	    XtFree(numFramesText);

	    if(whichStep==MAXX)
	      maxFrames =  (int)((cutMaxValues[0] - 
				  cutMinValues[0])/fabs(stepVal));
	    else if(whichStep==MAXY)
	      maxFrames =  (int)((cutMaxValues[1] - 
				  cutMinValues[1])/fabs(stepVal));
	    else if(whichStep==MAXZ)
	      maxFrames =  (int)((cutMaxValues[2] - 
				  cutMinValues[2])/fabs(stepVal));

	    if(numFrames <= 0 || numFrames > maxFrames)
	      {/* Then calc it ourselves */
		if(whichStep==MAXX)
		  {
		    numFrames = (int)((float)(cutMaxValues[0] - 
				       cutMinValues[0])/fabs(stepVal));
		  }
		if(whichStep==MAXY)
		  {
		    numFrames = (int)((float)(cutMaxValues[1] - 
				       cutMinValues[1])/fabs(stepVal));
		  }
		if(whichStep==MAXZ)
		  {
		    numFrames = (int)((float)(cutMaxValues[2] - 
				       cutMinValues[2])/fabs(stepVal));
		  }
		numFrames--;  /* Always one less */
		sprintf(outStr,"%d",numFrames);
		XmTextSetString(frameTextWidget, outStr); 
		XFlush(dsply);
	      }/* Then calc it ourselves */

	  }/* Calculate how many frames you need */
	else
	  {/* Get how many movie frames they want */
	    numFrames= 1;
	    numFramesText = XmTextGetString(frameTextWidget); 
	    numFrames = (int)strtod(numFramesText,0);
	    numFrames = MAX(numFrames,1);
	    XtFree(numFramesText);
	  }/* Get how many movie frames they want */


	/* Stuff rotate text widgets appropriatly */
	/* We've got to decide here if what we've got is XYZ or TPO
	 * check the rotateXYZ logical and then call XDY's rtns as appropriate
	 * XYZ-TPO */
	sprintf(outStr,"%f",globTheta);
	XmTextSetString(rotateThetaW, outStr); 
	sprintf(outStr,"%f",globPhi);
	XmTextSetString(rotatePhiW,outStr); 
	sprintf(outStr,"%f",globOmega);
	XmTextSetString(rotateOmegaW, outStr); 


	/* ADjust numFrames if stereo */
	if(stereoMapShade)
          {/* Offset it for stereo */
	    numFrames *= 2;
	    globTheta += 3;  /* Offset it for stereo */
          }/* Offset it for stereo */

	/* Init */
	map_min    = r4_max;
	map_max    = r4_min;

	/* First let's see if they want to do high resolution or not */
	if(!lowRes)
	  {/* Calc the high res one */

	    if((globPixelSize + ((float)stepVal*(float)numFrames)) <= 0.0)
	      {/* Trouble */
		askThem("I don't know how to do pixelSizes <= 0.0",
			FALSE,XmDIALOG_INFORMATION);
		goto allDone;
	      }/* Trouble */

	    numFramesMm = numFrames;
	    resolution_tryit_activate(0,&numFramesMm,0);
	    /* They may have changed the size of the buffer in
	     * resolution_tryit() so account for that here */
	    saveCutMax0Value = cutMaxValues[0];
	    saveCutMax1Value = cutMaxValues[1];
	    saveCutMax2Value = cutMaxValues[2];

	    if(whichStep==PIXELSIZE)
	      {/* Calc largestDim */
		float tempValue;
		/* pixelSizeFactor is numCols / 100.0 so let's
		 * calc the largestDim as:
		 *
		 *   (100.0 * pixelSizeFactor) / pixelSize
		 */
		largestDim = (int)((float)((100.0 * pixelSizeFactor) / 
					   globPixelSize)
				   * cutFactor);
		tempValue =  ((float)(100.0 * pixelSizeFactor) / 
			      (float)(globPixelSize+(float)((float)stepVal*
					      (float)numFrames)));
		/* Now account for if it's cut down */
		tempValue *= cutFactor;
		largestDim = MAX(largestDim,(int)tempValue);
	      }/* Calc largestDim */
	    else
	      {/* Just use this one */
		largestDim = numCols;
		largestDim = MAX(largestDim,numRows);
		largestDim = MAX(largestDim,origNumSecs);
	      }/* Just use this one */
	    largestDim *= 1.1;  /* Let's see if this helps */
	  }/* Calc the high res one */
	else
	  {/* Then let's just use the gold std */
	    if((*float_img_buf1)==float_image_buffer1)
	      {/* This is the kahuna */
		if(((*float_img_buf1)) &&
		   ((*float_img_buf1) != lowRes_buffer1))
		  free((*float_img_buf1));
		(*float_img_buf1) = lowRes_buffer1;
		numCols = lowResCols;
		numRows = lowResRows;
		origNumSecs= numSecs = lowResSecs;
	      }/* This is the kahuna */
	    else
	      {/* Maybe the subtract buffer */
	      }/* Maybe the subtract buffer */
		
	    file_width  =  purdueHeader.numCols = numCols;
	    file_height =  purdueHeader.numRows = numRows;
	    nsecs = numSecs;
	    
	    fileBinSecs= numSecs;
	    fileBinRows= numRows;
	    fileBinCols= numCols;
	
	    xCen = (double)(numCols/2);
	    yCen = (double)(numRows/2);
	    zCen = (double)(origNumSecs/2);

	    largestDim = numCols;
	    largestDim = MAX(largestDim,numRows);
	    largestDim = MAX(largestDim,origNumSecs);
	  }/* Then let's just use the gold std */

	/* Do unblob if necessary */
	unblobSwitch = FALSE;
	if(whichStep == MAXX)
	  {/* DO x */
	    if(stepVal < 0.0)
	      {/* start at cutMaxValues[0] */}
	    else cutMaxValues[0] = cutMinValues[0]+1;
	  }/* DO x */
	if(whichStep == MAXY)
	  {/* DO y */
	    if(stepVal < 0.0)
	      {/* start at cutMaxValues[1] */}
	    else cutMaxValues[1] = cutMinValues[1]+1;
	  }/* DO y */
	if(whichStep == MAXZ)
	  {/* DO z */
	    if(stepVal < 0.0)
	      {/* start at cutMaxValues[2] */}
	    else cutMaxValues[2] = cutMinValues[2]+1;
	  }/* DO z */
	if(whichStep == UNBLOBSIZE ||
	   whichStep == UNBLOBRADIUS ||
	   unblobSize != 0.0 || 
	   reverseSections 
	   )
	  {/* OK unblob/cut it */
	    int thisSecRow;
	    unblobSwitch = TRUE;
	    unblobBuffer = NULL;
	    unblobBuffer=(float *)malloc(sizeof(float)*
					    numCols*numRows*origNumSecs);
	    if(!unblobBuffer)
	      {/* Can't allocate it */
		if(askThem("Insufficient memory - do you want to unblob the\n\
Gold Standard memory buffer?  (This is a \n\
destructive operation - you'll have to re-read\n\
from disk to undo this.)",FALSE,XmDIALOG_QUESTION))
		  unblobBuffer = (*float_img_buf1);
		else
		  unblobSwitch = FALSE;
	      }/* Can't allocate it */
	    if(unblobSwitch)
	      unblob_tryit_activate(numFrames,0,0);
	  }/* OK unblob/cut it */
	else
	  unblobSwitch = FALSE;

	/* Allocate main window buffer */
	/* We always keep something allocated for this */
	free((*map_img_buf));
	(*map_img_buf)= NULL;
	CK((*map_img_buf)= (float *)malloc(sizeof(float)*
					     largestDim*largestDim
					     *numFrames));

	if((*map_img_buf)==NULL)
	  {/* Can't get memory */
	    askThem("Cannot allocate image buffer.",
		    FALSE,XmDIALOG_INFORMATION);
	  }/* Can't get memory */
	else
	  {/* Go ahead */
	    
	    /* Remember where we started */
	    saveSurfaceDensity = surfaceDensity; 
	    saveGlobFract = globFract; 
	    saveGlobAngx = globAngx; 
	    saveGlobAngy = globAngy; 
	    saveGlobTheta = globTheta; 
	    saveGlobPhi = globPhi; 
	    saveGlobOmega = globOmega; 
	    saveGlobMaxRadius = globMaxRadius; 
	    saveGlobMinRadius = globMinRadius; 
	    saveGlobPixelSize = globPixelSize; 
	    saveUnblobSize  = unblobSize;

	    ikey = 'M';

	    /* RWA July 29, 1998  We're trying to make this
	     * consistent with the rotate_do_activate() routine.
	     * globTheta  = 0.0 - globTheta;
	     */

	    /* Don't do this for X it's OK */
	    /* globPhi = 0.0 - globPhi; */

	    if(whichStep==MAXRADIUS)
	      {/* Let's start off at the minRadius */
		/* WHY ???  globMaxRadius = globMinRadius+1; */
		globMaxRadius = globMinRadius + stepVal;
	      }/* Let's start off at the minRadius */

	    shadedMaps = TRUE;
	    XtSetArg(arglst[0], XmNsensitive, TRUE);
	    XtSetValues(paintMapButtonW, arglst, 1); 
	    XtSetValues(mapMovieButtonW, arglst, 1); 
	    XFlush(dsply);
	    
	    /* What time is it? */
	    time(&oldtloc);
	    
	    /* Allocate these for the shading routine emmap_rob_shd1() */
	    free(rwa_tmp2d_map);
	    free(rwa_tmp_map);
	    CK(rwa_tmp2d_map = (double *)malloc(((largestDim * largestDim) + 
						 22) *
					       sizeof(double)));
	    CK(rwa_tmp_map   = (double *)malloc((largestDim + 22) * 
					       sizeof(double)));

	    for(m=0;m<numFrames;m++)
	      {/* Alloc memory */
		/* We may have changed the dimensions of these by
		 * going to a high resolution display (and back) etc. 
		 * realloc these */
		if(cuedMap[m].depth != NULL)
		  free(cuedMap[m].depth);
		if(cuedMap[m].shad != NULL)
		  free(cuedMap[m].shad);
		if(cuedMap[m].radial != NULL)
		  free(cuedMap[m].radial);

		CK(cuedMap[m].depth= (float *)malloc(sizeof(float)*
						     largestDim*largestDim));
		CK(cuedMap[m].shad= (float *)malloc(sizeof(float)*
						    largestDim*largestDim));
		CK(cuedMap[m].radial= (float *)malloc(sizeof(float)*
						      largestDim*largestDim));
	      }/* Alloc memory */
	    
	    for(m=0;m<numFrames;m++)
	      {/* Try different angles */
		
		numFramesMm = numFrames - m;

		/* Now estimate time */
		time(&tloc);
		secLeft = numFrames * 1;   /* a guess */
		if(numCols > 240)
		  secLeft = numFrames * 12; /* a guess */
		else if(numCols > 190)
		  secLeft = numFrames * 8; /* a guess */
		else if(numCols > 160)
		  secLeft = numFrames * 4; /* a guess */

		if(m>0)
		  {/* Can't estimate unless we've already done one */
		    secLeft = (int)(
				    (float)(((float)((int)(tloc - oldtloc))+
					     1.0)/
					    (float)m) * 
				    (float)(numFrames - m)
				    );
		    /* +1 For truncated seconds */
		    
		    /* Suppose we're taking less than one second to do a frame
		     * ??? */
		    /* secLeft = MAX(secLeft,(numFrames-m)); */
		  }/* Can't estimate unless we've already done one */
		
		/* Frame the location of the stop button */
		XDrawLine(dsply,XtWindow(mapShadeStopWidget),image_gcRED,
			  9,9,17,17);
		XDrawLine(dsply,XtWindow(mapShadeStopWidget),image_gcRED,
			  17,9,9,17);
		XFlush(dsply);

		if(cuedMap[m].depth && cuedMap[m].shad)
		  {/* We've enough memory */
		    thisMap= cuedMap[m].depth;
		    thisShadMap= cuedMap[m].shad;
		    thisRadialMap = cuedMap[m].radial;
		    
		    if(m>0)
		      {/* OK we need to step something */
			if(whichStep==SURFDENS)
			  {/* surface density */
			    surfaceDensity += stepVal;
			    if(unblobSwitch)
			      unblob_tryit_activate((numFrames-m),0,0); 
			                                /* Need to recalc
							 * unblob if density
							 * changes */
			  }/* surface density */
			else if(whichStep==FRACT)
			  {
			    globFract += stepVal;
			  }
			else if(whichStep==PHI)
			  {/* phi */
			    if(!stereoMapShade ||
			       (stereoMapShade && (m%2)==0))
			      {/* OK to increment */
				if(memoryMiser)
				  globPhi += (double)stepVal;
				else globPhi -= (double)stepVal;
				if(globPhi >= (double)360.0)
				  globPhi -= (double)360.0;
				if(globPhi <= (double)-360.0)
				  globPhi += (double)360.0;
			      }/* OK to increment */
			  }/* phi */
			else if(whichStep==OMEGA)
			  {/* omega */
			    if(!stereoMapShade ||
			       (stereoMapShade && (m%2)==0))
			      {/* OK to increment */
				if(memoryMiser)
				  globOmega += (double)stepVal;
				else globOmega -= (double)stepVal;
				if(globOmega >= (double)360.0)
				  globOmega -= (double)360.0;
				if(globOmega <= (double)-360.0)
				  globOmega += (double)360.0;
			      }/* OK to increment */
			  }/* omega */
			else if(whichStep==THETA)
			  {/* theta */
			    if(!stereoMapShade ||
			       (stereoMapShade && (m%2)==0))
			      {/* OK to increment */
				if(memoryMiser)
				  globTheta += (double)stepVal;
				else globTheta -= (double)stepVal;
				if(globTheta >= (double)360.0)
				  globTheta -= (double)360.0;
				if(globTheta <= (double)-360.0)
				  globTheta += (double)360.0;
			      }/* OK to increment */
			  }/* theta */
			else if(whichStep==MAXRADIUS)
			  globMaxRadius += stepVal;
			else if(whichStep == UNBLOBSIZE)
			  {/* DO unblob */
			    unblobSize += (int)stepVal;
			    unblob_tryit_activate((numFrames-m),0,0);
			    unblobSwitch = TRUE;
			  }/* DO unblob */
			else if(whichStep == UNBLOBRADIUS)
			  {/* DO unblob */
			    radiusUnblob += stepVal;
			    unblob_tryit_activate((numFrames-m),0,0);
			    unblobSwitch = TRUE;
			  }/* DO unblob */
			else if(whichStep == MAXX)
			  {/* DO unblob */
			    cutMaxValues[0] += stepVal;
			      resolution_tryit_activate(0,&numFramesMm,0);
			  }/* DO unblob */
			else if(whichStep == MAXY)
			  {/* DO unblob */
			    cutMaxValues[1] += stepVal;
			      resolution_tryit_activate(0,&numFramesMm,0);
			  }/* DO unblob */
			else if(whichStep == MAXZ)
			  {/* DO unblob */
			    cutMaxValues[2] += stepVal;
			      resolution_tryit_activate(0,&numFramesMm,0);
			  }/* DO unblob */
			else if(whichStep == PIXELSIZE)
			  {/* DO highres */
			    globPixelSize += stepVal;
			    if(lowRes)
			      {/* Do low res */
				/* Do nothing */
			      }/* Do low res */
			    else
			      {/* DO hires stuff */
				resolution_tryit_activate(0,&numFramesMm,0);
				/* Need to recalc unblob if we calc
				 * a new buffer */
				if(unblobSwitch)
				  {/* We've got to unblob it */
				    if(unblobBuffer != (*float_img_buf1))
				      free(unblobBuffer);
				    unblobBuffer= NULL;
				    unblobBuffer=(float *)malloc(sizeof(float)*
						numCols*numRows*origNumSecs);
				    if(!unblobBuffer)
				      {/* Can't allocate it */
					askThem(
		"Insufficient memory for unblob buffer\n\
 - continuing anyway without unblobbing\n",FALSE,
					       XmDIALOG_INFORMATION);
				      }/* Can't allocate it */
				    else
				      unblob_tryit_activate((numFrames-m),
							    0,0); 
				  }/* We've got to unblob it */
			      }/* DO hires stuff */
			  }/* DO hires */
		      }/* OK we need to step something */

		    /* For stereo adjust the theta concept */
		    if(stereoMapShade && ((m%2)==1))
		      {/* theta */
			if(memoryMiser)
			  globTheta -= (double)6.0;
			else globTheta += (double)6.0;
			if(globTheta >= (double)360.0)
			  globTheta -= (double)360.0;
			if(globTheta <= (double)-360.0)
			  globTheta += (double)360.0;
		      }/* theta */
		    
		    /* Init this for unblob if we need it */
		    maxUnblob = 0.0 - surfaceDensity;
		    
		    if(unblobSwitch)
		      this_float_image_buffer1= unblobBuffer;
		    else this_float_image_buffer1= (*float_img_buf1);

		    /* SHow them how long they've got left */
		    minLeft = (int)(float)secLeft / 60.0;
		    secLeft = secLeft - (minLeft * 60);
		    sprintf(outStr,
		          "[%d] Estimated time left (min:sec) -    %02d:%02d",
			    (numFrames-m),minLeft,secLeft);
		    
		    XmTextSetString(timeLeftLabelWidget, outStr);
		    XFlush(dsply);
		    

		    /* No tilting in dph_tilt we do it all in 
		     * dph_nomem_tilt_fast()*/
		    globTilt = FALSE;
		    
		    for(i=0;i<(largestDim);i++)
		      for(j=0;j<(largestDim);j++)
			{/* Do all pixels */
			  thisMap[((i * largestDim)+j)]= 0.0;
			  thisShadMap[((i * largestDim)+j)]= 0.0;
			  thisRadialMap[((i * largestDim)+j)]= 0.0;
			  rwa_tmp2d_map[((i * largestDim)+j)]= (double)0.0;
			  rwa_tmp_map[j]= (double)0.0;  
			}/* Do all pixels */

		    /* See if they want to bugout or not */
		    if(XCheckTypedWindowEvent(dsply,
					      XtWindow(mapShadeStopWidget),
					      ButtonPress,
					      &event))
		      {/* STOP!!! */
			XFlush(dsply);
			goto noMoreCueing;
		      }/* STOP!!! */

		    /* Historical from FORTRAN - get rid of this later. */
		    ikey = 'M';
		    
		    /* Always compute the dpeth-cue even if you're doing
		     * radial cueing as you need depth to shade the radial. */
		    saveRadialMapShade= radialMapShade;
		    radialMapShade = FALSE;
		    if(negUniverseCut)
		      {/* Stuff appropriately */
			col1 = 0;
			col2 = numCols;
			row1 = 0;
			row2 = numRows;
			sec1 = 0;
			sec2 = origNumSecs;
		      }/* Stuff appropriately */
		    else
		      {/* Optimize */
			if(TRUE)
			  {/* lores */
			    col1 = cutMinValues[0];
			    col2 = cutMaxValues[0];
			    row1 = cutMinValues[1];
			    row2 = cutMaxValues[1];
			    sec1 = cutMinValues[2];
			    sec2 = cutMaxValues[2];
			  }/* lores */
			else
			  {/* hires */
			    col1 = 0;
			    col2 = numCols;
			    row1 = 0;
			    row2 = numRows;
			    sec1 = 0;
			    sec2 = origNumSecs;
			  }/* hires */
		      }/* Optimize */

		    /* The concept of xcen/ycen/zcen is independent 
		     * for rotation and for radially shading - we may 
		     * need a varaible for both??? */
		    if(!lowRes)
		      {/* Compute center for high res images */
			float ifItWereEntireImageSize,
			      perCentSize,
			      centerEntire,
			      highResZero;
			
			perCentSize = 
			  (float)( (float)cutMaxValues[0]/(float)lowResCols) -
			    (float)( (float)cutMinValues[0]/(float)lowResCols);
			if(perCentSize <= 0.0)
			  perCentSize = 1.0;
			ifItWereEntireImageSize = numCols / perCentSize;
			centerEntire = ifItWereEntireImageSize / 2.0;
			highResZero = ifItWereEntireImageSize * 
			  (float)( (float)cutMinValues[0]/(float)lowResCols);
			radialXcen = (int)(centerEntire - highResZero);

			perCentSize = 
			  (float)( (float)cutMaxValues[1]/(float)lowResRows) -
			    (float)( (float)cutMinValues[1]/(float)lowResRows);
			if(perCentSize <= 0.0)
			  perCentSize = 1.0;
			ifItWereEntireImageSize = numRows / perCentSize;
			centerEntire = ifItWereEntireImageSize / 2.0;
			highResZero = ifItWereEntireImageSize * 
			  (float)( (float)cutMinValues[1]/(float)lowResRows);
			radialYcen = (int)(centerEntire - highResZero);

			perCentSize = 
			  (float)( (float)cutMaxValues[2]/(float)lowResSecs) -
			    (float)( (float)cutMinValues[2]/(float)lowResSecs);
			if(perCentSize <= 0.0)
			  perCentSize = 1.0;
			ifItWereEntireImageSize = origNumSecs / perCentSize;
			centerEntire = ifItWereEntireImageSize / 2.0;
			highResZero = ifItWereEntireImageSize * 
			  (float)( (float)cutMinValues[2]/(float)lowResSecs);
			radialZcen = (int)(centerEntire - highResZero);

			topRadialSec = (int)(highResZero +
					     ((ifItWereEntireImageSize *
					      perCentSize)*2.0));
			                 /* times 2.0 In case we rotate it */

		      }/* Compute center for high res images */
		    else
		      {/* Just init them */
			radialXcen = radialYcen = radialZcen = 0;
			topRadialSec = (int)origNumSecs;
		      }/* Just init them */
		    
		    if(radialProjections)
		      {/* DO projections */
			float x_org,y_org,z_org;
			float minF,maxF,scFactor, thisRadius,smallestDens;
			float omega,phi,thetaY,sinX,sinY,cosX,cosY;
			int colOffset,rowOffset;

			phi    = globPhi;
			thetaY = globTheta;
			omega  = globOmega;
			/* Not used anymore 
			cosX = (float)cos((double)deg2rad((double)phi));
			sinX = (float)sin((double)deg2rad((double)phi));
			cosY = (float)cos((double)deg2rad((double)thetaY));
			sinY = (float)sin((double)deg2rad((double)thetaY));
			*/
			x_org = (float)(((int)(numCols) + 1.0)/2.0)-1.0;
			y_org = (float)(((int)(numRows) + 1.0)/2.0)-1.0;
			z_org = (float)(((int)(origNumSecs) + 1.0)/2.0)-1.0;

			thisRadius = globMaxRadius;

			if(thisRadius > (largestDim * 0.50))
			  thisRadius = largestDim  * 0.50;

			smallestDens = secMinPixel-1.0;
			
			colOffset = (largestDim-numCols)/2;
			rowOffset = (largestDim-numRows)/2;

			if(icosRadProj)
			  {/* Do icosahedral radial projections */
			    char *s1;
			    float thresh,slid;
			    s1 = XmTextGetString(dioThreshTextW);
			    thresh = (float)strtod(s1,0);
			    XtFree(s1);
			    s1 = XmTextGetString(dioSlideTextW);
			    slid = (float)strtod(s1,0);
			    XtFree(s1);
			    icosRadialProject ((float *)
					       this_float_image_buffer1,
					       numCols,numRows,origNumSecs,
					       (float *)thisShadMap,
					       x_org, y_org, z_org, 
					       thisRadius, smallestDens,
					       thetaY,phi,omega,
					       colOffset, rowOffset,
					       largestDim, irev, whichIcos,
					       thresh,slid, destroyIcos,
					       globMinRadius);
			  }/* Do icosahedral radial projections */
			else
			  {/* Do spherical */
			    rad_dens_fast_rob_C ((float *)
						 this_float_image_buffer1,
						 numCols,numRows,origNumSecs,
						 (float *)thisShadMap,
						 x_org, y_org, z_org, 
						 thisRadius, smallestDens,
						 thetaY,phi,omega,
						 colOffset, rowOffset,
						 largestDim, irev);
			  }/* Do spherical */

			minF= maxF= thisShadMap[0];
			for(i=0;i<(largestDim);i++)
			  for(j=0;j<(largestDim);j++)
			    {/* Do all pixels */
			      if(thisShadMap[((i * largestDim)+j)]!=0.0)
				{/* OK to use */
				  minF= MIN(minF,thisShadMap[((i * 
							    largestDim)+j)]);
				  maxF= MAX(thisShadMap[((i * 
							  largestDim)+j)],
					    maxF);
				}/* OK to use */
			    }/* Do all pixels */

			if(minF!=maxF)
			  scFactor = 255.0 / (maxF - minF);
			else scFactor = 1.0;
			
			for(i=0;i<(largestDim);i++)
			  for(j=0;j<(largestDim);j++)
			    {/* Do all pixels */
			      if(thisShadMap[((i * largestDim)+j)]!=0.0)
				{/* OK to use */
				  thisShadMap[((i * largestDim)+j)] =
				    (thisShadMap[((i * largestDim)+j)] - 
				     minF) *
				       scFactor;
				  if(revRadProj)
				    {/* Reverse */
				      /* Now reverse the contrast - they do it
				       * backward for some reason */
				      thisShadMap[((i * largestDim)+j)] = 
					(float)
					  (255.0 -
					   thisShadMap[((i * largestDim)+j)] );
				    }/* Reverse */
				}/* OK to use */
			      
			    }/* Do all pixels */
			
		      }/* DO projections */
		    
		    else
		      {/* Do radial/depth-cued shading */

			if(!truePhi)
			  {/* Use older routine */
			    emmap_rob_dphshd_C(
					    (float *)this_float_image_buffer1,
					       (float *)thisMap,
					       (float *)thisRadialMap,
					       (float *)thisShadMap,
					       (float *)(*float_img_buf1),
					       (saveRadialMapShade) ? 
					              FALSE : TRUE,
					       surfaceDensity,
					       numCols, numRows,
					       sec1,sec2,
					       col1, col2, 
					       row1, row2,
					       ikey,
					       globPhi, 
					       globTheta,
					       globMinRadius, globMaxRadius,
					       xCen, yCen, zCen,
					       radialXcen, radialYcen, 
					       radialZcen,
					       lowRes, topRadialSec);
			  }/* Use older routine */
			else
			  {/* Try to do a TRUE phi rotation */

			    thetaPhi_dphshd_C(
					    (float *)this_float_image_buffer1,
					      (float *)(*float_img_buf1),
					      numCols, numRows,
					      (float *)thisMap, /* final */
					      FALSE, FALSE,
					      sec1,sec2,
					      col1, col2,
					      row1, row2,
					      1, irev, surfaceDensity, 
					      (saveRadialMapShade) ? 
					          FALSE : TRUE,
					      (float *)thisShadMap,
					      globPhi, globTheta, globOmega,
					      globMinRadius, globMaxRadius,
					      xCen, yCen, zCen,
					      radialXcen, radialYcen, 
					      radialZcen,
					      lowRes, topRadialSec, 
					    samRenStep);
			  }/* Try to do a TRUE phi rotation */
			
			
		      }/* Do radial/depth-cued shading */
		    
		    /* Now compute the radial if you need it */
		    radialMapShade= saveRadialMapShade;
		    if(radialMapShade)
		      {/* radial */
			if(negUniverseCut)
			  {/* Stuff appropriately */
			    col1 = 0;
			    col2 = numCols;
			    row1 = 0;
			    row2 = numRows;
			    sec1 = 0;
			    sec2 = origNumSecs;
			  }/* Stuff appropriately */
			else
			  {/* Optimize */
			      if(TRUE)
			      {/* lores */
				col1 = cutMinValues[0];
				col2 = cutMaxValues[0];
				row1 = cutMinValues[1];
				row2 = cutMaxValues[1];
				sec1 = cutMinValues[2];
				sec2 = cutMaxValues[2];
			      }/* lores */
			    else
			      {/* hires */
				col1 = 0;
				col2 = numCols;
				row1 = 0;
				row2 = numRows;
				sec1 = 0;
				sec2 = origNumSecs;
			      }/* hires */
			  }/* Optimize */

			if(!truePhi)
			  {/* DO older rtn */
			    emmap_rob_dphshd_C(
					     (float *)this_float_image_buffer1,
					       (float *)thisMap,
					       (float *)thisRadialMap,
					       (float *)thisShadMap,
					       (float *)(*float_img_buf1),
					       TRUE,
					       surfaceDensity,
					       numCols, numRows,
					       sec1,sec2,
					       col1, col2,
					       row1, row2,
					       ikey,
					       globPhi, globTheta,
					       globMinRadius, globMaxRadius,
					       xCen, yCen, zCen,
					       radialXcen, radialYcen, 
					       radialZcen,
					       lowRes, topRadialSec);
			  }/* DO older rtn */
			else
			  {/* Try a TRUE phi rotate */
			    thetaPhi_dphshd_C(
					    (float *)this_float_image_buffer1,
					      (float *)(*float_img_buf1),
					      numCols, numRows,
					      (float *)thisRadialMap,
					      FALSE, FALSE,
					      sec1,sec2,
					      col1, col2,
					      row1, row2,
					      1, irev, surfaceDensity, TRUE,
					      (float *)thisShadMap,
					      globPhi, globTheta, globOmega,
					      globMinRadius, globMaxRadius,
					      xCen, yCen, zCen,
					      radialXcen, radialYcen, 
					      radialZcen,
					      lowRes, topRadialSec,samRenStep);
			  }/* Try a TRUE phi rotate */
		      }/* radial */
		    
		    /* Remember how many */
		    numSecs= nsecs= m+1;
		    
		    /* Now adjust back for next pass if stereo mode */
		    if(stereoMapShade && ((m%2)==1))
		      {/* theta */
			if(memoryMiser)
			  globTheta += (double)6.0;
			else globTheta -= (double)6.0;
			if(globTheta >= (double)360.0)
			  globTheta -= (double)360.0;
			if(globTheta <= (double)-360.0)
			  globTheta += (double)360.0;
		      }/* theta */
		    
		  }/* We've enough memory */	
		else
		  {/* no more */
		    askThem("Trouble allocating memory.",
			    FALSE,XmDIALOG_INFORMATION);
		  }/* no more */
	      }/* Try different angles */
	    
	    /* For some reason emmap_dphshd_C() makes this too large
	     * sometimes */
	    if(lowRes)
	      {
		map_max = MIN(map_max,(float)(origNumSecs/2)+1.0); 
	      }

	    /* We've got to trap here in case we've set conditions such
	     * that we don't have any data in the map to shade. */
	    if((map_max<(float)((float)r4_min*0.9))
	       && (map_min>(float)((float)r4_max*0.9)))
	      {/* We've got no data */
		map_min= -20.0;
		map_max= 120.0;
	      }/* We've got no data */
	    
	    /* Go in here if you want depth-cued shading */
	    if(secondMapShade && !radialMapShade 
	       )
	      {/* Do shading? */
		sinx           = sin(deg2rad(globAngx));
		cosx           = cos(deg2rad(globAngx));
		siny           = sin(deg2rad(globAngy));
		cosy           = cos(deg2rad(globAngy));
		
		/* Get this from interface in future */
		cenx= ceny= cenz= (origNumSecs/2)+1;
		
		emmap_rob_shd1((cue_t *)cuedMap,
			       rwa_tmp2d_map, rwa_tmp_map,
			       largestDim, largestDim,
			       1, numFrames,
			       saveGlobFract, TRUE,
			       saveGlobAngx,stepVal,
			       saveGlobAngy,whichStep,
			       (0.0), ANGX, ANGY, FRACT);
	      }/* Do shading? */
	    
	    /* Make sure we see the lowest value in the map as something other
	     * than background */
	    if(radialMapShade)
	      map_min -= fabs(((map_max - map_min) * 0.05));
	    else 
	      {
		map_min -= fabs(((map_max - map_min) * 0.01));
		map_max += fabs(((map_max - map_min) * 0.01)); 
	      }
	    
	    if(map_max==map_min)
	      map_max += 10.0;

	    if(lowRes && (radialMapShade && !radialCS))
	      {/* DOn't scale them */
		/* Reset these to the whole range */
		map_max = (float)(origNumSecs/2)+1.0;
		map_min = 0.0 - (map_max * 0.025); /* SO we get the zeroth 
						    * one displayed */
	      }/* DOn't scale them */

	    /* Now scale them all the same */
	    dph_rescale_rob (largestDim, largestDim, numSecs, 
			     (cue_t *)cuedMap, tryColors, 
			     &radialStepSize, hiColor);
	    
	    /* Go in here if you want radial-cued shading */
	    if(secondMapShade && radialMapShade)
	      {/* Do shading for radial */
		
		sinx           = sin(deg2rad(globAngx));
		cosx           = cos(deg2rad(globAngx));
		siny           = sin(deg2rad(globAngy));
		cosy           = cos(deg2rad(globAngy));
		
		/* Get this from interface in future */
		cenx= ceny= cenz= (origNumSecs/2)+1;
		
		emmap_rob_shd1((cue_t *)cuedMap,
			       rwa_tmp2d_map, rwa_tmp_map,
			       largestDim, largestDim,
			       1, numFrames,
			       saveGlobFract, FALSE,
			       saveGlobAngx,stepVal,
			       saveGlobAngy,whichStep,
			       0.0, ANGX, ANGY, FRACT);
	      }/* Do shading for radial */
	    
	    thisMinMapPixel = 9999.0;
	    thisMaxMapPixel = -9999.0;
	    thisMinMapPixel = HUGE;   /* dgm HUGE instead of 9999.0 */
	    thisMaxMapPixel = -HUGE;  /* dgm -HUGE instead of -9999.0 */
	    
	    /* Stuff the main window display buffer -
	     * Only stuff it on the last pass */
	    for(m=0;m<numFrames;m++)
	      {/* Do all frames */
		float mapPixel;
		
		thisMap= cuedMap[m].depth;
		thisShadMap= cuedMap[m].shad;
		thisRadialMap= cuedMap[m].radial;
		
		for(i=0;i<(largestDim);i++)
		  {/* rows */
		    for(j=0;j<(largestDim);j++)
		      {/* Do all pixels */
			
 			if(!radialProjections &&
			   !secondMapShade && 
			   !radialMapShade
			   )
			  mapPixel = (float)thisMap[(((int)(i)*
						      largestDim)+
						     (int)(j))];
			else
			  mapPixel = (float)thisShadMap[(((int)(i)*
							  largestDim)+
							 (int)(j))];
			
			(*map_img_buf)[((m*(largestDim*largestDim))+
					((i*(largestDim))+j))]= 
					  (float)mapPixel;
			
			if(mapPixel != 0)
			  {/* DOn't count background in this */
			    if(mapPixel > thisMaxMapPixel)
			      thisMaxMapPixel= (float)mapPixel;
			    if(mapPixel < thisMinMapPixel)
			      thisMinMapPixel= (float)mapPixel;
			  }/* DOn't count background in this */
			
		      }/* Do all pixels */
		  }/* rows */
	      }/* Do all frames */
	    
	    /* Trap for empty display */
	    if(map_max==0.0 && map_min==0.0)
	      {map_max=20.0; map_min= -10.0;}
	    
	    if(thisMaxMapPixel == thisMinMapPixel)
	      thisMinMapPixel -= 0.001;  /* Just get it so that we've
				      * some kind of range */
	    if(calcMapPixel)
	      {/* You can modify the concept of max and min pixel */
		minMapPixel = MIN(thisMinMapPixel,minMapPixel);
		maxMapPixel = MAX(thisMaxMapPixel,maxMapPixel);
	      }/* You can modify the concept of max and min pixel */
	    else
	      {/* Just stuff it */
		minMapPixel = thisMinMapPixel;
		maxMapPixel = thisMaxMapPixel;
	      }/* Just stuff it */

	    cuedMaxPixel = maxMapPixel;
	    cuedMinPixel = minMapPixel;

	    contrastMax = 0;
	    contrastMin = 999;
	    dispSections= FALSE;
	    stuffMapImage((float *)(*map_img_buf), 
			  (float *)(*map_img_buf), 
			  file_width,file_height,
			  nsecs,1, 0);
	    

	    /* Reset the concept of curBox in the movie/editBox screen
	     * so we don't try and reference a part of the buffer we
	     * don't have.  Oh, and BTW we have no concept of lastNumBox
	     * with MAPs - that's for scanned image boxes and should 
	     * be zero here.
	     */
	    curBox = MIN(curBox,(nsecs-1));
	    sprintf(outStr,"%d",curBox+lastNumBox+1);
	    XmTextSetString(editBoxTextWidget, outStr);
	    XFlush(dsply);

	    /* Display the image in the window created above */
	    if(XtIsManaged(main_window_widget))
	      XPutImage (dsply, image_window1, image_gc, image_image1, 0, 0,
			 0, 0, image_width, image_height);
	    
	    XFlush (dsply);
	    
	    DrawWindow();
	    /* Redraw the histogram */
	    histogram_activate(&minP, &maxP);
	    
	    
	    surfaceDensity = saveSurfaceDensity; 
	    globFract = saveGlobFract; 
	    globAngx = saveGlobAngx; 
	    globAngy = saveGlobAngy; 
	    globTheta = saveGlobTheta; 
	    globPhi = saveGlobPhi; 
	    globOmega = saveGlobOmega; 
	    globMaxRadius = saveGlobMaxRadius; 
	    globMinRadius = saveGlobMinRadius; 
	    globPixelSize = saveGlobPixelSize; 
	    unblobSize = saveUnblobSize;
	    mapNumSecs = nsecs;
	    
	    /* Free this every time */
	    /* NOOOOOO - done in groundZero() rtn!!!
	       free(rwa_tmp2d_map);
	       free(rwa_tmp_map);
	       */

	  }/* Go ahead */
	
	/* Show them what they're seeing */
	XtSetArg(arglst[0], XmNlabelPixmap, unsectionPixmap);
	XtSetArg(arglst[1], XmNselectPixmap, sectionPixmap);
	XtSetArg(arglst[2], XmNrecomputeSize, FALSE);
	XtSetValues(sectionWidget, arglst, 3);
	sprintf(outStr,"1");
	XmTextSetString(imgNumTextWidget,outStr);
	
	goto allDone;
      noMoreCueing:;
	surfaceDensity = saveSurfaceDensity; 
	globFract = saveGlobFract; 
	globAngx = saveGlobAngx; 
	globAngy = saveGlobAngy; 
	globTheta = saveGlobTheta; 
	globPhi = saveGlobPhi; 
	globOmega = saveGlobOmega; 
	globMaxRadius = saveGlobMaxRadius; 
	globMinRadius = saveGlobMinRadius; 
	globPixelSize = saveGlobPixelSize; 
	unblobSize = saveUnblobSize;
	mapNumSecs = nsecs;
	
      allDone:;

	sprintf(outStr,"Select parameters.");
	XmTextSetString(timeLeftLabelWidget, outStr);
	XFlush(dsply);
	
	/* Make them remanage the map movie screen to get it setup properly
	 * for this movie they've just calculated. */
	XtUnmanageChild(editBoxScreenWidget);
	XFlush(dsply);

	/* Don't trash this just yet -= they may want to zoom the
	 * display in virtual_read_activate() -->> stuffMapImage()
	 * so they can use this pointer before someone reallocates it...
	 *
	 * free((*map_img_buf)); 
	 * (*map_img_buf)= NULL; 
	 */
	
	/* Show them the volume stuff if they want it */
	updateNewVolume();

	unblobSwitch = FALSE;
	if(unblobBuffer != (*float_img_buf1))
	  free(unblobBuffer);
	unblobBuffer= NULL;

      }/* OK we've got map data */
  
  if(dispIcosa)
    {/* Draw cisoahedron */

      calcSizeFactor(zoomFactor);
      DrawIcosahedron(deg2rad(0.0-globTheta),
		      deg2rad(0.0-globPhi),
		      deg2rad(globOmega),
		      0, 0,
		      (float)((float)numCols*sizeFactor*2.0), 
		      image_window1); 

      DrawIcosahedron(deg2rad(0.0-globTheta),
		      deg2rad(0.0-globPhi),
		      deg2rad(globOmega),
		      500, 255,
		      numCols,
		      XtWindow(mapShadeScreenWindow));

    }/* Draw cisoahedron */

  /* Reset these in case we've trashed them etc... */
  cutMaxValues[0] = saveCutMax0Value;
  cutMaxValues[1] = saveCutMax1Value;
  cutMaxValues[2] = saveCutMax2Value;


}/* MAPSHADE_ACTIVATE_WORK */




/* ************************************************************
 * rotate_xyz_activate() - We were going to implement an xyz
 * rotation for the 3D rendering screen based on the values chosen
 * in the theta/phi/omega widgets.  We were planning on just doing
 * an XYZ to TPO conversion and then letting the 3D rendering rtns
 * just go on their merry way using the TPO like normal.  We
 * have yet to devise the XYZ to TPO conversion scheme yet. 
*/
void rotate_xyz_activate(Widget widget, int *tag, 
			 XmScrollBarCallbackStruct *callback_data)
{/* ROTATE_XYZ_ACTIVATE */
  XmString simpleString;

  rotateXYZ = !rotateXYZ;

  /* We've got to decide here if what we've got is XYZ or TPO
   * check the rotateXYZ logical and then call XDY's rtns as appropriate
   * XYZ-TPO */
  /* e.g. We might want to convert globPhi back and forth and
   * update the text widgets as appropriate here */

  if(rotateXYZ)
    {/* Feedback xyz */
      sprintf(outStr,"X");
      createSimpleString(&simpleString,outStr);
      XtSetArg(arglst[0], XmNlabelString, simpleString);
      XtSetArg(arglst[1], XmNrecomputeSize, FALSE );
      XtSetValues(thetaButtonW, arglst, 2); 
      XFlush(dsply);
      XmStringFree(simpleString); 
      XFlush(dsply);
      sprintf(outStr,"Y");
      createSimpleString(&simpleString,outStr);
      XtSetArg(arglst[0], XmNlabelString, simpleString);
      XtSetArg(arglst[1], XmNrecomputeSize, FALSE );
      XtSetValues(phiButtonW, arglst, 2); 
      XFlush(dsply);
      XmStringFree(simpleString); 
      XFlush(dsply);
      sprintf(outStr,"Z");
      createSimpleString(&simpleString,outStr);
      XtSetArg(arglst[0], XmNlabelString, simpleString);
      XtSetArg(arglst[1], XmNrecomputeSize, FALSE );
      XtSetValues(omegaButtonW, arglst, 2); 
      XFlush(dsply);
      XmStringFree(simpleString); 
      XFlush(dsply);
    }/* Feedback xyz */
  else
    {/* Feedback tpo */
      sprintf(outStr,"Theta");
      createSimpleString(&simpleString,outStr);
      XtSetArg(arglst[0], XmNlabelString, simpleString);
      XtSetArg(arglst[1], XmNrecomputeSize, FALSE );
      XtSetValues(thetaButtonW, arglst, 2); 
      XFlush(dsply);
      XmStringFree(simpleString); 
      XFlush(dsply);
      sprintf(outStr,"Phi");
      createSimpleString(&simpleString,outStr);
      XtSetArg(arglst[0], XmNlabelString, simpleString);
      XtSetArg(arglst[1], XmNrecomputeSize, FALSE );
      XtSetValues(phiButtonW, arglst, 2); 
      XFlush(dsply);
      XmStringFree(simpleString); 
      XFlush(dsply);
      sprintf(outStr,"Omega");
      createSimpleString(&simpleString,outStr);
      XtSetArg(arglst[0], XmNlabelString, simpleString);
      XtSetArg(arglst[1], XmNrecomputeSize, FALSE );
      XtSetValues(omegaButtonW, arglst, 2); 
      XFlush(dsply);
      XmStringFree(simpleString); 
      XFlush(dsply);
    }/* Feedback tpo */


}/* ROTATE_XYZ_ACTIVATE */





/* ************************************************************
 * foldAxis_activate() - They want to change the axis they
 * want to view.
*/
void foldAxis_activate(Widget widget, int *tag, 
		       XmAnyCallbackStruct *callback_data)
{/* foldAxis_ACTIVATE */
  XmString simpleString;

  /* Which axis? */
  if(*tag==5)
    {/* five fold */
      globTheta= (double)(360.0+32.0);
      globPhi = 360.0;
      globOmega = 360.0;
    }/* five fold */
  else if(*tag==3)
    {/* 3 fold */
      globTheta = (double)(360.0+69.0);
      globPhi = 360.0;
      globOmega = 360.0;
    }/* 3 fold */
  else if(*tag==2)
    {/* 2 fold */
      globTheta = (double)(360.0+0.0);
      globPhi = 360.0;
      globOmega = 360.0;
    }/* 2 fold */

  /* We've got to decide here if what we've got is XYZ or TPO
   * check the rotateXYZ logical and then call XDY's rtns as appropriate
   * XYZ-TPO */

  /* Now set the scrollbar to the new value */
  XtSetArg(arglst[0], XmNvalue, (int)globTheta);
  XtSetValues(thetaScrollWidget, arglst, 1); 
  XtSetArg(arglst[0], XmNvalue, (int)globPhi);
  XtSetValues(phiScrollWidget, arglst, 1); 
  XtSetArg(arglst[0], XmNvalue, (int)globOmega);
  XtSetValues(omegaScrollWidget, arglst, 1); 

  globTheta -= 360.0;
  if(globTheta != 0.0)
    globTilt = TRUE;
  else globTilt = FALSE;
  
  globPhi -= 360.0;
  if(globPhi != 0.0)
    globTilt = TRUE;
  else globTilt = FALSE;
  
  globOmega -= 360.0;
  if(globOmega != 0.0)
    globTilt = TRUE;
  else globTilt = FALSE;
  
  /* We've got to decide here if what we've got is XYZ or TPO
   * check the rotateXYZ logical and then call XDY's rtns as appropriate
   * XYZ-TPO */

  /* Give feedback */
  sprintf(outStr,"%.0f",globTheta);
  createSimpleString(&simpleString, outStr);
  XtSetArg(arglst[0], XmNvalue, outStr );
  XtSetArg(arglst[1], XmNrecomputeSize, FALSE );
  XtSetValues(thetaValueWidget, arglst, 2); 
  XFlush(dsply);
  XmStringFree(simpleString); 
  
  sprintf(outStr,"%.0f",globOmega);
  createSimpleString(&simpleString, outStr);
  XtSetArg(arglst[0], XmNvalue, outStr );
  XtSetArg(arglst[1], XmNrecomputeSize, FALSE );
  XtSetValues(omegaValueWidget, arglst, 2); 
  XFlush(dsply);
  XmStringFree(simpleString); 
  
  sprintf(outStr,"%.0f",globPhi);
  createSimpleString(&simpleString, outStr);
  XtSetArg(arglst[0], XmNvalue, outStr );
  XtSetArg(arglst[1], XmNrecomputeSize, FALSE );
  XtSetValues(phiValueWidget, arglst, 2); 
  XFlush(dsply);
  XmStringFree(simpleString); 
  
  if(immediateMapShade)
    mapShade_activate(0,tag,0);

}/* foldAxis_ACTIVATE */



/* ************************************************************
 * flickArrow_create() - Remember widget numbers!
 */
void flickArrow_create(Widget widget, int *tag, 
		       XmAnyCallbackStruct *callback_data)
{/* flickArrow_create */

  if(*tag==XmARROW_UP)
    flickArrowUpWidget = widget;
  else if(*tag==XmARROW_DOWN)
    flickArrowDownWidget = widget;

}/* flickArrow_create */




/* ************************************************************
 * movieArrow_create() - Remember widget numbers!
 */
void movieArrow_create(Widget widget, int *tag, 
		       XmAnyCallbackStruct *callback_data)
{/* movieArrow_create */

  if(*tag==XmARROW_UP)
    movieArrowUpWidget = widget;
  else if(*tag==XmARROW_DOWN)
    movieArrowDownWidget = widget;

}/* movieArrow_create */




/*************************************************************
 * DrawThermometer() - Draw the "thermometer" to indicate 
 * movie speed on editBox widnwo.
*/
void DrawThermometer()
{/*DrawThermometer*/

  XDrawLine(dsply, XtWindow(playFormWidget), 
	    image_gcRED,
	    1,1,(50),1);
  XDrawLine(dsply, XtWindow(playFormWidget), 
	    image_gcRED,
	    1,2,(50),2);
  XDrawLine(dsply, XtWindow(playFormWidget), 
	    image_gcGREEN,
	    1,1,(50-movieSpeed),1);
  XDrawLine(dsply, XtWindow(playFormWidget), 
	    image_gcGREEN,
	    1,2,(50-movieSpeed),2);
  XFlush(dsply);

}/*DrawThermometer*/



/*************************************************************
 * DrawFlickThermometer() - Draw the "thermometer" to indicate 
 * movie speed on editBox window.
*/
void DrawFlickThermometer()
{/*DrawFlickThermometer*/

  XDrawLine(dsply, XtWindow(subtractMapScreen), 
	    image_gcRED,
	    401,95,(450),95);
  XDrawLine(dsply, XtWindow(subtractMapScreen), 
	    image_gcRED,
	    401,96,(450),96);
  XDrawLine(dsply, XtWindow(subtractMapScreen), 
	    image_gcGREEN,
	    401,95,(450-flickSpeed),95);
  XDrawLine(dsply, XtWindow(subtractMapScreen), 
	    image_gcGREEN,
	    401,96,(450-flickSpeed),96);
  XFlush(dsply);


}/*DrawFlickThermometer*/



/* ************************************************************
 * flickArrowActivate() - They want to change the speed
 * of the flicker.
*/
void flickArrowActivate(Widget widget, int *tag, 
			XmPushButtonCallbackStruct *callback_data)
{/* FLICKARROWACTIVATE */
  int numClicks;
  XEvent event;

  numClicks = callback_data->click_count;

  /* Which arrow? */
  if(*tag==XmARROW_UP)
    {/* Up */
      flickSpeed -= numClicks;
    }/* Up */
  else if(*tag==XmARROW_DOWN)
    {/* Dopwn */
      flickSpeed += numClicks;
    }/* Down */
  else printf("flickSpeed unchanged.\n");
  
  flickSpeed = MAX(0,MIN(flickSpeed, 50));
  
  DrawFlickThermometer();
  
}/* FLICKARROWACTIVATE */





/* ************************************************************
 * mapMovieArrowActivate() - They want to change the speed
 * at which the frames of the movie are displayed.
*/
void mapMovieArrowActivate(Widget widget, int *tag, 
			   XmPushButtonCallbackStruct *callback_data)
{/* MAPMOVIEARROWACTIVATE */
  int numClicks;
  XEvent event;

  numClicks = callback_data->click_count;

  /* Which arrow? */
  if(*tag==XmARROW_UP)
    {/* Up */
      movieSpeed -= numClicks;
    }/* Up */
  else if(*tag==XmARROW_DOWN)
    {/* Dopwn */
      movieSpeed += numClicks;
    }/* Down */
  else printf("movieSpeed unchanged.\n");
  
  movieSpeed = MAX(0,MIN(movieSpeed, 50));
  
  DrawThermometer();
  
}/* MAPMOVIEARROWACTIVATE */






/* ************************************************************
 * mapMovie_show_activate() - They want to show their map sections
 * displayed as a "movie".
 */
void mapMovie_show_activate(Widget widget, int *tag, 
			    XmAnyCallbackStruct *callback_data)
{/* mapMovie_show_activate */
  int i,j,k,numEvents,speed;
  float saveContourLevel;
  XEvent event;
  char curStr[12];

  if(makingMovie)
    {/* DOn't do otherwise */
      if(*tag==MAPSTOP)
	{/* Do nothing */
	  XtSetArg(arglst[0], XmNlabelPixmap, playPixmap);
	  XtSetArg(arglst[1], XmNselectPixmap, playingPixmap);
	  XtSetArg(arglst[2], XmNrecomputeSize, FALSE);
	  XtSetValues(playWidget, arglst, 3); 
	  XtSetArg(arglst[0], XmNlabelPixmap, stopPixmap);
	  XtSetArg(arglst[1], XmNselectPixmap, stoppedPixmap);
	  XtSetArg(arglst[2], XmNrecomputeSize, FALSE);
	  XtSetValues(stopWidget, arglst, 3); 
	}/* Do nothing */
      else if(*tag==FLASHLIGHT)
	{/* flashLight the planet */
	  flashLightIt = ! flashLightIt;
	  if(flashLightIt)
	    {/* Put up flashLight icon */
	      XtSetArg(arglst[0], XmNlabelPixmap, flashLightPixmap);
	      XtSetArg(arglst[1], XmNselectPixmap, noflashLightPixmap);
	      XtSetArg(arglst[2], XmNrecomputeSize, FALSE);
	      XtSetArg(arglst[3], XmNfillOnArm, FALSE);
	      XtSetValues(flashLightWidget, arglst, 4); 
	    }/* Put up flashLight icon */
	  else 
	    {/* Put up noflashLight icon */
	      XtSetArg(arglst[0], XmNlabelPixmap, noflashLightPixmap);
	      XtSetArg(arglst[1], XmNselectPixmap, flashLightPixmap);
	      XtSetArg(arglst[2], XmNrecomputeSize, FALSE);
	      XtSetArg(arglst[3], XmNfillOnArm, FALSE);
	      XtSetValues(flashLightWidget, arglst, 4); 
	    }/* Put up noflashLight icon */
	  if(flashLightIt)
	    attrs.cursor= lanternCursor;
	  else attrs.cursor= None;
	  XChangeWindowAttributes(dsply, XtWindow(editBoxWidget), 
				  CWCursor, &attrs);
	  XFlush(dsply);
	}/* flashLight the planet */
      else if(*tag==ROCKER)
	{/* rock the planet */
	  rockIt = ! rockIt;
	  if(rockIt)
	    {/* Put up rock icon */
	      XtSetArg(arglst[0], XmNlabelPixmap, rockPixmap);
	      XtSetArg(arglst[1], XmNselectPixmap, norockPixmap);
	      XtSetArg(arglst[2], XmNrecomputeSize, FALSE);
	      XtSetArg(arglst[3], XmNfillOnArm, FALSE);
	      XtSetValues(rockWidget, arglst, 4); 
	    }/* Put up rock icon */
	  else 
	    {/* Put up norock icon */
	      XtSetArg(arglst[0], XmNlabelPixmap, norockPixmap);
	      XtSetArg(arglst[1], XmNselectPixmap, rockPixmap);
	      XtSetArg(arglst[2], XmNrecomputeSize, FALSE);
	      XtSetArg(arglst[3], XmNfillOnArm, FALSE);
	      XtSetValues(rockWidget, arglst, 4); 
	    }/* Put up norock icon */
	}/* rock the planet */
      else if (*tag==MAP || *tag==INFINIMAP)
	{/* Do the movie */
	  int goOn,switchThem,thisEndMovie;
	  XtSetArg(arglst[0], XmNlabelPixmap, playingPixmap);
	  XtSetArg(arglst[1], XmNselectPixmap, playPixmap);
	  XtSetArg(arglst[2], XmNrecomputeSize, FALSE);
	  XtSetArg(arglst[3], XmNfillOnArm, TRUE);
	  XtSetValues(playWidget, arglst, 4); 
	  XtSetArg(arglst[0], XmNlabelPixmap, noinfiniPixmap);
	  XtSetArg(arglst[1], XmNselectPixmap, infiniPixmap);
	  XtSetArg(arglst[2], XmNrecomputeSize, FALSE);
	  XtSetArg(arglst[3], XmNfillOnArm, TRUE);
	  XtSetValues(infiniWidget, arglst, 4); 
	  goOn = TRUE;
	  switchThem = FALSE;
	  saveContourLevel= contourLevel;
	  thisEndMovie = endMovie;
	  if(shadedMaps && stereoMapShade)
	    thisEndMovie = endMovie - 1;
	  while(goOn)
	    {/* Do until done */
	      /* Frame the location of the stop button */
	      XDrawLine(dsply,XtWindow(stopWidget),image_gcRED,
			9,9,17,17);
	      XDrawLine(dsply,XtWindow(stopWidget),image_gcRED,
			17,9,9,17);
	      XFlush(dsply);
	      if(switchThem)
		{/* for rockIt effect */
		  for(i=thisEndMovie;i>=startMovie;
		      (shadedMaps && stereoMapShade)?
		      i-=2:i--)
		    {/* SHow them in order */
		      /* Human numbers from one not zero */
		      sprintf(curStr,"%d",(i+1));
		      XmTextSetString(editBoxTextWidget, curStr);
		      stuffDisplayMapBuffer(i,editBoxZoom,0);
		      if(shadedMaps && stereoMapShade)
			stuffDisplayMapBuffer(i+1,editBoxZoom,numCols);
		      DrawEditBoxWindow(TRUE);
		      if(useCont)
			{/* Use on the movie */
			  for(k=0;k<totContours;k++)
			    {/* Stuff all */
			      /* Now Non-destructively draw the contour in 
			       * the window buffer */
			      drawContour(conTours[k],FALSE,i,
					  editBoxWidget,
					  float_image_buffer1,0,0,
					  editBox_display_buffer);
			    }/* Stuff all */
			  XFlush(dsply);
			}/* Use on the movie */
		      if(XCheckTypedWindowEvent(dsply,
						XtWindow(stopWidget),
						ButtonPress,
						&event))
			{/* STOP!!! */
			  XFlush(dsply);
			  goto noMore;
			}/* STOP!!! */
		      else if(XCheckTypedWindowEvent(dsply,
						XtWindow(movieArrowUpWidget),
						     ButtonPress,
						     &event))
			{/* STOP!!! */
			  if((event.xbutton.button == Button1))
			    movieSpeed--;
			  if((event.xbutton.button == Button2))
			    movieSpeed -= 2;
			  if((event.xbutton.button == Button3))
			    movieSpeed -= 5;
			  movieSpeed = MAX(0,MIN(movieSpeed, 50));
			  
			  /* Now draw the "thermometer" to indicate 
			   * movie speed */
			  DrawThermometer();
			}/* STOP!!! */
		      else if(XCheckTypedWindowEvent(dsply,
						XtWindow(movieArrowDownWidget),
						     ButtonPress,
						     &event))
			{/* STOP!!! */
			  if((event.xbutton.button == Button1))
			    movieSpeed++;
			  if((event.xbutton.button == Button2))
			    movieSpeed += 2;
			  if((event.xbutton.button == Button3))
			    movieSpeed += 5;
			  movieSpeed = MAX(0,MIN(movieSpeed, 50));
			  
			  /* Now draw the "thermometer" to indicate 
			   * movie speed */
			  DrawThermometer();
			}/* STOP!!! */
		      else if(XCheckTypedWindowEvent(dsply,
						     XtWindow(movieMinWidget),
						     ButtonPress,
						     &event))
			{/* STOP!!! */
			  XPutBackEvent(dsply,&event);
			  XFlush(dsply);
			  XSync(dsply,FALSE);
			  XFlush(dsply);
			}/* STOP!!! */
		      else if(XCheckTypedWindowEvent(dsply,
						     XtWindow(movieMaxWidget),
						     ButtonPress,
						     &event))
			{/* STOP!!! */
			  XPutBackEvent(dsply,&event);
			  XFlush(dsply);
			  XSync(dsply,FALSE);
			  XFlush(dsply);
			}/* STOP!!! */
		      else if(XCheckTypedWindowEvent(dsply,
						     XtWindow(movieDoneWidget),
						     ButtonPress,
						     &event))
			{/* STOP!!! */
			  XPutBackEvent(dsply,&event);
			  XFlush(dsply);
			  XSync(dsply,FALSE);
			  goto noMore; 
			}/* STOP!!! */
		      XFlush(dsply);
		      for(speed=0;speed<movieSpeed;speed++)
			em_sleep(1);
		    }/* SHow them in order */
		}/* for rockIt effect */
	      else
		{/* normal */
		  for(i=startMovie;i<=thisEndMovie;(shadedMaps && 
						    stereoMapShade)?
		      i+=2:i++)
		    {/* SHow them in order */
		      /* Human numbers from one not zero */
		      sprintf(curStr,"%d",(i+1));
		      XmTextSetString(editBoxTextWidget, curStr);
		      stuffDisplayMapBuffer(i,editBoxZoom,0);
		      if(shadedMaps && stereoMapShade)
			stuffDisplayMapBuffer(i+1,editBoxZoom,numCols);
		      DrawEditBoxWindow(TRUE);
		      if(useCont)
			{/* Use on the movie */
			  for(k=0;k<totContours;k++)
			    {/* Stuff all */
			      /* Now Non-destructively draw the contour in 
			       * the window buffer */
			      drawContour(conTours[k],FALSE,i,
					  editBoxWidget,
					  float_image_buffer1,0,0,
					  editBox_display_buffer);
			    }/* Stuff all */
			  XFlush(dsply);
			}/* Use on the movie */
		      if(XCheckTypedWindowEvent(dsply,
						XtWindow(stopWidget),
						ButtonPress,
						&event))
			{/* STOP!!! */
			  XFlush(dsply);
			  goto noMore;
			}/* STOP!!! */
		      else if(XCheckTypedWindowEvent(dsply,
						XtWindow(movieArrowUpWidget),
						     ButtonPress,
						     &event))
			{/* STOP!!! */
			  if((event.xbutton.button == Button1))
			    movieSpeed--;
			  if((event.xbutton.button == Button2))
			    movieSpeed -= 2;
			  if((event.xbutton.button == Button3))
			    movieSpeed -= 5;
			  movieSpeed = MAX(0,MIN(movieSpeed, 50));
			  
			  /* Now draw the "thermometer" to indicate 
			   * movie speed */
			  DrawThermometer();
			}/* STOP!!! */
		      else if(XCheckTypedWindowEvent(dsply,
						XtWindow(movieArrowDownWidget),
						     ButtonPress,
						     &event))
			{/* STOP!!! */
			  if((event.xbutton.button == Button1))
			    movieSpeed++;
			  if((event.xbutton.button == Button2))
			    movieSpeed += 2;
			  if((event.xbutton.button == Button3))
			    movieSpeed += 5;
			  movieSpeed = MAX(0,MIN(movieSpeed, 50));
			  
			  /* Now draw the "thermometer" to indicate 
			   * movie speed */
			  DrawThermometer();
			}/* STOP!!! */
		      else if(XCheckTypedWindowEvent(dsply,
						     XtWindow(movieMinWidget),
						     ButtonPress,
						     &event))
			{/* STOP!!! */
			  XPutBackEvent(dsply,&event);
			  XFlush(dsply); 
			  goto noMore;
			}/* STOP!!! */
		      else if(XCheckTypedWindowEvent(dsply,
						     XtWindow(movieMaxWidget),
						     ButtonPress,
						     &event))
			{/* STOP!!! */
			  XPutBackEvent(dsply,&event);
			  XFlush(dsply); 
			  goto noMore;
			}/* STOP!!! */
		      else if(XCheckTypedWindowEvent(dsply,
						     XtWindow(movieDoneWidget),
						     ButtonPress,
						     &event))
			{/* STOP!!! */
			  XPutBackEvent(dsply,&event);
			  XFlush(dsply);
			  XSync(dsply,FALSE);
			  goto noMore; 
			}/* STOP!!! */
		      XFlush(dsply);
		      for(speed=0;speed<movieSpeed;speed++)
			em_sleep(1);
		    }/* SHow them in order */
		}/* normal */
	      if(rockIt)
		{/* Rock it back the other way */
		  switchThem = !switchThem;
		}/* Rock it back the other way */
	      if(*tag==INFINIMAP)
		goOn = TRUE;
	      else goOn = FALSE;
	    }/* Do until done */
	noMore:;
	  contourLevel= saveContourLevel;
	  XtSetArg(arglst[0], XmNlabelPixmap, playPixmap);
	  XtSetArg(arglst[1], XmNselectPixmap, playingPixmap);
	  XtSetArg(arglst[2], XmNrecomputeSize, FALSE);
	  XtSetArg(arglst[3], XmNfillOnArm, FALSE);
	  XtSetValues(playWidget, arglst, 4); 
	  XtSetArg(arglst[0], XmNlabelPixmap, infiniPixmap);
	  XtSetArg(arglst[1], XmNselectPixmap, noinfiniPixmap);
	  XtSetArg(arglst[2], XmNrecomputeSize, FALSE);
	  XtSetArg(arglst[3], XmNfillOnArm, FALSE);
	  XtSetValues(infiniWidget, arglst, 4); 
	}/* Do the movie */
      XFlush(dsply);
    }/* DOn't do otherwise */
  
  /* Find a better place to put this ... */
  if(*tag==USESECTION)
    {/* They want to disp sections or cued map  */
      dispSections = !dispSections;
      if(dispSections)
	{/* They want sections */
	  XtSetArg(arglst[0], XmNsensitive, TRUE);
	  XtSetValues(paintMapButtonW, arglst, 1); 
	  XtSetValues(mapMovieButtonW, arglst, 1); 
	  XFlush(dsply);
	  sprintf(outStr,"1");
	  XmTextSetString(imgNumTextWidget,outStr);
	  numSecs= nsecs= lastSec;
	  maxMapPixel= secMaxPixel;
	  minMapPixel= secMinPixel;
	  XtSetArg(arglst[0], XmNlabelPixmap, sectionPixmap);
	  XtSetArg(arglst[1], XmNselectPixmap, unsectionPixmap);
	  XtSetArg(arglst[2], XmNrecomputeSize, FALSE);
	  XtSetValues(sectionWidget, arglst, 3); 
	  /* Now stuff the display buffer from what was read off the
	   * disk - strt at zero'th(1) section */
	  stuffMapImage((float *)map_image_buffer, 
			(float *)float_image_buffer1, 
			file_width,file_height,
			nsecs,1, 0);

	}/* They want sections */
      else
	{/* They want cued map */
	  if(shadedMaps)
	    {/* Only do if they've got any to disp */
	      sprintf(outStr,"1");
	      XmTextSetString(imgNumTextWidget,outStr);
	      numSecs= nsecs= mapNumSecs;
	      maxMapPixel= cuedMaxPixel;
	      minMapPixel= cuedMinPixel;

	      stuffMapImage((float *)map_image_buffer, 
			    (float *)map_image_buffer, 
			    file_width,file_height,
			    nsecs,1, 0);
	  
	      XtSetArg(arglst[0], XmNlabelPixmap, unsectionPixmap);
	      XtSetArg(arglst[1], XmNselectPixmap, sectionPixmap);
	      XtSetArg(arglst[2], XmNrecomputeSize, FALSE);
	      XtSetValues(sectionWidget, arglst, 3);
	      dispSections= FALSE;
	    }/* Only do if they've got any to disp */
	  else
	    dispSections= TRUE;
	}/* They want cued map */

      /* Display the image in the window created above */
      XPutImage (dsply, image_window1, image_gc, image_image1, 0, 0,
		 0, 0, image_width, image_height);
      
      XFlush (dsply);
      
      DrawWindow();
      /* Redraw the histogram */
      histogram_activate(&minP, &maxP);
    }/* They want to disp sections or cued map  */
  
}/* mapMovie_show_activate */



/* ************************************************************
 * mapMovie_deactivate() - They are done with the map movie.
 */
void mapMovie_deactivate(Widget widget, int *tag, 
			 XmAnyCallbackStruct *callback_data)
{/* mapMovie_deactivate */

  XtSetArg(arglst[0], XmNsensitive, TRUE);
  XtSetValues(playWidget, arglst, 1); 
  XtSetValues(stopWidget, arglst, 1); 
  XtSetValues(infiniWidget, arglst, 1); 
  XtSetValues(rockWidget, arglst, 1); 
  XtSetValues(flashLightWidget, arglst, 1); 

  XtManageChild(undoWidget);
  XtManageChild(savePaintWidget);
  XtManageChild(paintFormWidget);
  XtUnmanageChild(playFormWidget);
  XtUnmanageChild(movieSaveWidget); 
  XtUnmanageChild(movieMinWidget);
  XtUnmanageChild(movieMaxWidget);
  XtUnmanageChild(maxLabelWidget);
  XtUnmanageChild(minLabelWidget);
  XtUnmanageChild(editBoxScreenWidget);
  makingMovie= FALSE;

}/* mapMovie_deactivate */





/*-----------------------------------*/
/*************************************
 * EDITBOX callbacks                 *
 *************************************/
/*-----------------------------------*/
/* ************************************************************
 * editBox_activate() - They want to edit their boxes.  Stuff the
 * first one up there for them and show them the screen.
 * The editBox screen is also used for painting and showing 3D
 * rendered movies.
 */
void editBox_activate(Widget widget, int *tag, 
		      XmAnyCallbackStruct *callback_data)
{/* editBox_activate */
  XmString simpleString;
  
  sprintf(outStr,"%d",curBox+lastNumBox+1);
  XmTextSetString(editBoxTextWidget, outStr);
  XFlush(dsply);

  makingMovie= FALSE;
  createSimpleString(&simpleString,"Box Editor");
  XtSetArg(arglst[0], XmNdialogTitle, simpleString);
  XtSetValues(editBoxScreenWidget, arglst, 1); 
  XFlush(dsply);
  raiseIt(editBoxScreenWidget);
  XFlush(dsply);
  XmStringFree(simpleString); 
/* malloc_verify(); */

  XtManageChild(undoWidget);
  XtManageChild(savePaintWidget);
  XtManageChild(paintFormWidget);
  XtUnmanageChild(playFormWidget);
  XtUnmanageChild(movieSaveWidget); 
  XtUnmanageChild(movieMinWidget);
  XtUnmanageChild(movieMaxWidget);
  XtUnmanageChild(maxLabelWidget);
  XtUnmanageChild(minLabelWidget);

  if(fileType==PACKEDPORTABLE)
    stuffDisplayPackBoxBuffer(curBox, editBoxZoom, FALSE);
  else stuffDisplayBoxBuffer(curBox,image_fileName.addr,editBoxZoom);
      
  DrawEditBoxWindow(FALSE);

}/* editBox_activate */



/* ************************************************************
 * editBoxForm_create() - Remember the form widget numbers so
 * we can manage/unmanage them appropriately.
 */
void editBoxForm_create(Widget widget, int *tag, 
			XmAnyCallbackStruct *callback_data)
{/* editBoxForm_create */
  
  if(*tag==MAP)
    playWidget= widget;
  else if(*tag==MAPSTOP)
    stopWidget= widget;
  else if(*tag==INFINIMAP)
    infiniWidget= widget;
  else if(*tag==PLAYFORM)
    XtUnmanageChild(playFormWidget= widget);
  else if(*tag==PAINTFORM)
    XtUnmanageChild(paintFormWidget= widget);
  else if(*tag==MOVIEMIN)
    XtUnmanageChild(movieMinWidget= widget);
  else if(*tag==MOVIEMAX)
    XtUnmanageChild(movieMaxWidget= widget);
  else if(*tag==ROCKER)
    rockWidget= widget;
  else if(*tag==FLASHLIGHT)
    flashLightWidget= widget;
  else if(*tag==UNDO)
    undoWidget= widget;
  else if(*tag==SAVEPAINT)
    savePaintWidget= widget;
  else if(*tag==FFTMIN)
    minFFTwidget= widget;
  else if(*tag==FFTMAX)
    maxFFTwidget= widget;
  else if(*tag==FFTTFAC)
    tfacFFTwidget= widget;
  else if(*tag==THEOMIN)
    theominFFTwidget= widget;
  else if(*tag==THEOMAX)
    theomaxFFTwidget= widget;
  else if(*tag==MOVIESAVE)
    movieSaveWidget= widget;
  else if(*tag==MOVIEDONE)
    movieDoneWidget= widget;
  else if(*tag==CONTOUR)
    movieContourScrollWidget= widget;
  else if(*tag==DELTA)
    movieDeltaScrollWidget= widget;
  else if(*tag==NUMCONT)
    movieNumContScrollWidget= widget;

}/* editBoxForm_create */



/* ************************************************************
 * editBox_save_activate() - They want to save all of their
 * edits on this box.  Save in memory the buffer on the screen.
 */
void editBox_save_activate(Widget widget, int *tag, 
			   XmAnyCallbackStruct *callback_data)
{/* EDITBOX_SAVE_ACTIVATE */
  int i,j,y,x,k,thisZoomFactor,thisRadius;
  int yOut,yIn,outHeight,outWidth;
  float *thisBuf;

  if(*tag==SAVEPAINT)
    {/* They are painting */
      if(fileType!=MAP)
	{/* Save the display buffer info */
	  
	  /* Calc the zoom */
	  box[curBox].boxZoom= thisZoomFactor = editBoxZoom;
	  calcBoxSizeFactor(thisZoomFactor);
	  
	  box[curBox].inMemory = TRUE;  /* Make them paint it again */
	  
	  if(fileType==PACKEDPORTABLE)
	    {/* packed file input */
	      thisBuf = packImg[curBox];
	      thisRadius = packRadius;
	      outHeight = purdueHeader.numRows;
	      outWidth  = purdueHeader.numCols;
	    }/* packed file input */
	  else
	    {/* boxing from image file */
	      thisBuf = box[curBox].edit_mirror_buffer;
	      thisRadius = boxRadius;
	      outHeight = (thisRadius*2)+1;
	      outWidth  = (thisRadius*2)+1;
	    }/* boxing from image file */
	  
	  /* Reallocate the memory in case we had a larger one last time */
	  if(thisBuf)
	    free(thisBuf);
	  CK(thisBuf = (float *)malloc(sizeof(float)*(outWidth*outHeight)));
	  /* malloc_verify(); */
	  
	  for(i=0;i<outHeight;i++)
	    {/* rows */
	      yOut = i*outWidth;
	      yIn  = (i*sizeFactor) * BOX_WIDTH;
	      for(j=0;j<outWidth;j++)
		{/* cols */
		  thisBuf[(yOut+j)] = 
		    editBox_mirror_buffer[(yIn+(int)(j*sizeFactor))];
		}/* cols */
	    }/* rows */
	  
	  if(fileType==PACKEDPORTABLE)
	    packImg[curBox]= thisBuf;
	  else box[curBox].edit_mirror_buffer= thisBuf;
	  
	}/* Save the display buffer info */

      DrawEditBoxWindow(FALSE);

    }/* They are painting */
  else if(*tag==MOVIESAVE)
    {/* Save the movie */

      system("rm robem*.pcx"); 

      //writePCX("robem",endMovie-startMovie+1);
      writePCX("robem",endMovie, startMovie);

      /* Not working ...
      system("rm \\~/robem\\*.pcx"); 
      sprintf(outStr,"mv robem\\*.pcx ~/");
      printf("[%s]\n",outStr);
      system(outStr);
      */

      /* Not working ....
      system("convert robem*.pcx movie.gif ");
      system("rm robem*.pcx");  
      */
      system("convert robem*.pcx movie.gif ");

      askThem(
	"Movie saved in PCX format in robem*.pcx files.",
	      FALSE,
	      XmDIALOG_INFORMATION);

    }/* Save the movie */
  

}/* EDITBOX_SAVE_ACTIVATE */



/* ************************************************************
 * editBox_undo_activate() - They want to undo all of their
 * edits on this box.  Restuff the box from disk and
 * show them the screen.
 */
void editBox_undo_activate(Widget widget, int *tag, 
			   XmAnyCallbackStruct *callback_data)
{/* EDITBOX_UNDO_ACTIVATE */
  int i,j,k,l;

  if(fileType!=PACKEDPORTABLE)
    {/* Make them paint it again */
      box[curBox].inMemory = FALSE;  
      free(box[curBox].editBox_buffer);
    }/* Make them paint it again */
  else
    {/* redraw from the mirror_buffer */
      for(i=0;i<BOX_HEIGHT;i++)
	{/* i  */
	  for(j=0;j<BOX_WIDTH;j++)
	    {/* j */
	      if(hiColor)
		{/* hi */
		  XPutPixel(editBox_image1, x, y, 
			    pixels_return[(int)
				  ((editBox_undo_buffer[((i*BOX_WIDTH)+j)]
					    *scalePackFactor)+6.0)]);
		  editBox_display_buffer[((y*BOX_WIDTH)+x)]=
		    (BYTE)(int)((editBox_undo_buffer[((i*BOX_WIDTH)+j)]
				 *scalePackFactor)+6.0);
		}/* hi */
	      else
		{/* lo */
		  editBox_display_buffer[((i*BOX_WIDTH)+j)]= 
		    editBox_mirror_buffer[((i*BOX_WIDTH)+j)]= 
		      (float)pixels_return[(int)
				    ((editBox_undo_buffer[((i*BOX_WIDTH)+j)]
				      *scalePackFactor)+6.0)];
		}/* lo */
	    }/* j */
	}/* i */
    }/* redraw from the mirror_buffer */
  if(fileType==PACKEDPORTABLE)
    stuffDisplayPackBoxBuffer(curBox, editBoxZoom, FALSE);
  else stuffDisplayBoxBuffer(curBox,image_fileName.addr,editBoxZoom);
      
/* malloc_verify(); */
  DrawEditBoxWindow(FALSE);

}/* EDITBOX_UNDO_ACTIVATE */



/* ************************************************************
 * DrawEditBoxWindow() - Redraw window e.g. on expose events
 * or whenever it's prudent for the editBox drawing area.
 */
void DrawEditBoxWindow(int mapMovie)
{/* DrawEditBoxWindow */
  int draw_width, draw_height;

  if((numBoxes > 0 || makingMovie) ||
     (fileType==PACKEDPORTABLE && purdueHeader.nimg > 0))
    {/* OK there's one to draw */
      draw_width = BOX_WIDTH;
      draw_height = BOX_HEIGHT;
      
      if(liteSwitch)
	{
	  draw_width  = numCols;
	  draw_height = numRows;
	}

      if(XtIsManaged(editBoxScreenWidget))
	XPutImage(dsply, editBox_window1, editBox_gc, editBox_image1, 0, 0,
		  0, 0,  draw_width,  draw_height);
      XFlush (dsply);

      DrawThermometer();
      
    }/* OK there's one to draw */
  
}/* DrawEditBoxWindow */



/* ***************************************************************
 * editBoxClick() - Every time the mouse button is clicked call
 * here.  This routine handles the drawing of background
 * into the box image or restoring same in the Paint Boxes 
 * functionality.
*/
void editBoxClick(Widget widget, int *tag, 
		  XmDrawingAreaCallbackStruct *draw_data)
{/* EDITBOXCLICK */
  XEvent event;
  int x, y, i, j, k, yPaint,xPaint,xMax,yMax,thisRadius;
  int yPaintMap,xPaintMap;
  int thisOffset;
  float thisPixel,offset;
  char ans;

  /* Do appropriate depending on event type */
  if(draw_data != 0)
    {/* Only look at events meant for us */

      if(fileType==PACKEDPORTABLE)
	thisRadius = packRadius;
      else thisRadius = boxRadius;

      event= *(draw_data->event);
      
      switch (event.type) 
	{/* Find out which event we have */
	case ButtonRelease:
	  
	      
	  x = (int)(event.xbutton.x);  /* Get x,y copords of point pos */
	  y = (int)(event.xbutton.y);
	  
	  if((event.xbutton.button == Button1))
	    {/* Left most button */
	      flashLightIt = FALSE;
	      /* Put up noflashLight icon */
	      XtSetArg(arglst[0], XmNlabelPixmap, noflashLightPixmap);
	      XtSetArg(arglst[1], XmNselectPixmap, flashLightPixmap);
	      XtSetArg(arglst[2], XmNrecomputeSize, FALSE);
	      XtSetArg(arglst[3], XmNfillOnArm, FALSE);
	      XtSetValues(flashLightWidget, arglst, 4); 
	      attrs.cursor= None;
	      XChangeWindowAttributes(dsply, XtWindow(editBoxWidget), 
				      CWCursor, &attrs);
	      XFlush(dsply);
	    }/* Left most button */
	  
	  break;
	  
	case ButtonPress:
	      
	  x = (int)(event.xbutton.x);  /* Get x,y coords of point pos */
	  y = (int)(event.xbutton.y);
	  
	  if(flashLightIt)
	    flashLightIt=99;  /* Turn it really ON */

	  calcBoxSizeFactor(editBoxZoom);
	  
	  /* Button1 means draw in background value */
	  if((event.xbutton.button == Button1))
	    {/* Button 1 */

	      if(paintMapSwitch)
		{/* We are painting a MAP */
		  /* Get brush size */
		  yPaintMap = (y * recipSizeFactor);
		  xPaintMap = (x * recipSizeFactor);
		  yMax= xMax = numCols;
		  
		  xPaintMap = MAX(0,MIN(xPaintMap,xMax));
		  yPaintMap = MAX(0,MIN(yPaintMap,yMax));
		  
		  /* Repaint it */
		  if(normalExcise)
		    paintMap[((curBox*numCols*numRows)+
			      (yPaintMap*numCols)+xPaintMap)]= 
				float_image_buffer1[((curBox*numCols*numRows)+
					 (yPaintMap*numCols)+xPaintMap)];
		  else
		    paintMap[((curBox*numCols*numRows)+
			      (yPaintMap*numCols)+xPaintMap)]= 0.0;
		    
		  
		  yPaint = y - (y % (int)sizeFactor);
		  xPaint = x - (x % (int)sizeFactor);
		  yMax= xMax = ((numCols*sizeFactor));
		  
		  xPaint = MAX(0,MIN(xPaint,(xMax)));
		  yPaint = MAX(0,MIN(yPaint,(yMax)));
		  for(i=yPaint;i<(yPaint+sizeFactor);i++)
		    for(j=xPaint;j<(sizeFactor+xPaint);j++)
		      {
			if(hiColor)
			  editBox_display_buffer[((i*BOX_WIDTH)+j)]= 
			    RED;
			else
			  editBox_display_buffer[((i*BOX_WIDTH)+j)]= 
			    pixels_return[RED];
		      }

		  stuffDisplayMapBuffer(curBox,editBoxZoom,0);
		  DrawEditBoxWindow(TRUE);
		}/* We are painting a MAP */
	      else
		{/* Normal */
		  /* How big is the brush? */
		  yPaint = y - (paintSize/2);
		  xPaint = x - (paintSize/2);
		  yMax= xMax = ((thisRadius*2)+1) * sizeFactor;
		  
		  xPaint = MAX(0,MIN(xPaint,(xMax-paintSize)));
		  yPaint = MAX(0,MIN(yPaint,(yMax-paintSize)));
		  
		  /* Paint it in brush size */
		  for(i=yPaint;i<(yPaint+paintSize);i++)
		    for(j=xPaint;j<(paintSize+xPaint);j++)
		      if(bkgrndMethod==USEAVGPIXEL)
			{/* avg */
			  if(hiColor)
			    {/* hi */
			      XPutPixel(editBox_image1, j, i, 
					pixels_return[(int)backDispChar]);
			    }/* hi */
			  editBox_display_buffer[((i*BOX_WIDTH)+j)]= 
			    backDispChar;
			  editBox_mirror_buffer[((i*BOX_WIDTH)+j)]= 
			    (float)backChar;
			}/* avg */
		      else
			{/* random */
			  float thisVal;
			  thisVal = calcFloatBackground(editBox_int_buffer, 
							    xMax, 
							    yMax,
							    BOX_WIDTH,
							    fileType,
							    bkgrndMethod,
							    boxMethod,
							    packRadius,
							    boxRadius,
							    featherRadius,
							    recipSizeFactor,
							    tryColors
							    );
			  if(hiColor)
			    {/* hi */
			      XPutPixel(editBox_image1, j, i, 
					pixels_return[(6+(int)thisVal)]);
			    }/* hi */
			  else
			    editBox_display_buffer[((i*BOX_WIDTH)+j)]= 
			      pixels_return[(6+(int)thisVal)];

			  editBox_mirror_buffer[((i*BOX_WIDTH)+j)]= 
			    (float)pixels_return[(6+(int)calcFloatBackground(
							editBox_int_buffer, 
							    xMax, 
							    yMax,
							    BOX_WIDTH,
							    fileType,
							    bkgrndMethod,
							    boxMethod,
							    packRadius,
							    boxRadius,
							    featherRadius,
							    recipSizeFactor,
							    tryColors
							    ))];
			}/* random */
		}/* Normal */
	      
	    }/* Button 1 */
	  
	  /* Button2 is for replacing with image info kinda like
	   * an undo feature.  */
	  else if((event.xbutton.button == Button2))
	    {/* button 2 */

	      if(paintMapSwitch)
		{/* We are painting a MAP */
		  /* Get brush size */
		  yPaintMap = (y * recipSizeFactor);
		  xPaintMap = (x * recipSizeFactor);
		  yMax= xMax = numCols;
		  
		  xPaintMap = MAX(0,MIN(xPaintMap,xMax));
		  yPaintMap = MAX(0,MIN(yPaintMap,yMax));
		  
		  /* Repaint it */
		  if(normalExcise)
		    paintMap[((curBox*numCols*numRows)+
			      (yPaintMap*numCols)+xPaintMap)]= 0.0;
		  else
		    paintMap[((curBox*numCols*numRows)+
			      (yPaintMap*numCols)+xPaintMap)]= 
				float_image_buffer1[((curBox*numCols*numRows)+
					 (yPaintMap*numCols)+xPaintMap)];
		  
		  yPaint = y - (y % (int)sizeFactor);
		  xPaint = x - (x % (int)sizeFactor);
		  yMax= xMax = ((numCols*sizeFactor));
		  
		  xPaint = MAX(0,MIN(xPaint,(xMax)));
		  yPaint = MAX(0,MIN(yPaint,(yMax)));


		  if(minMapPixel < 0.0)
		    offset= (0.0 - minMapPixel);
		  else offset= 0.0; 
		  thisPixel=  float_image_buffer1[((curBox*numCols*numRows)+
						   (yPaintMap*numCols)+
						   xPaintMap)] +
						     offset;
	      
		  thisOffset = ((int) ( (int) ( ((float)thisPixel) * 
					       scaleMapFactor )   ) + 6);
	      
		  for(i=yPaint;i<(yPaint+sizeFactor);i++)
		    for(j=xPaint;j<(sizeFactor+xPaint);j++)
		      {
			if(hiColor)
			  {
			    XPutPixel(editBox_image1, j, i, 
				      pixels_return[thisOffset]);
			    editBox_display_buffer[((i*BOX_WIDTH)+j)]= 
			      thisOffset;
			  }
			else
			  editBox_display_buffer[((i*BOX_WIDTH)+j)]= 
			    (BYTE)pixels_return[thisOffset];
		      }
		  stuffDisplayMapBuffer(curBox,editBoxZoom,0);
		  DrawEditBoxWindow(TRUE);
		}/* We are painting a MAP */
	      else
		{/* Normal */

		  /* Get brush size */
		  yPaint = y - (paintSize/2);
		  xPaint = x - (paintSize/2);
		  yMax= xMax = ((thisRadius*2)+1) * sizeFactor;
		  
		  xPaint = MAX(0,MIN(xPaint,(xMax-paintSize)));
		  yPaint = MAX(0,MIN(yPaint,(yMax-paintSize)));
		  
		  /* Repaint it */
		  for(i=yPaint;i<(yPaint+paintSize);i++)
		    {/* i  */
		      int yIn,xIn;
		      /* Topmost point of box */
		      yIn = ((box[curBox].y*sizeFactor) - 
			     (thisRadius*sizeFactor));   
		      /* Leftmost point in box */
		      xIn = ((box[curBox].x*sizeFactor) - 
			     (thisRadius*sizeFactor));   
		      for(j=xPaint;j<(paintSize+xPaint);j++)
			{/* j */
			  editBox_mirror_buffer[((i*BOX_WIDTH)+j)]=
			    (float)editBox_undo_buffer[((i*BOX_WIDTH)+j)];
			  if(fileType==PACKEDPORTABLE)
			    {
			      if(hiColor)
				{/* hi */
				  XPutPixel(editBox_image1, j, i, 
					    pixels_return[(int)
				  ((editBox_undo_buffer[((i*BOX_WIDTH)+j)]
				    *scalePackFactor)+6.0)]);
				  editBox_display_buffer[((i*BOX_WIDTH)+j)]= 
				    (BYTE)(int)
				      ((editBox_undo_buffer[((i*BOX_WIDTH)+j)]
					*scalePackFactor)+6.0);
				}/* hi */
			      else
				editBox_display_buffer[((i*BOX_WIDTH)+j)]= 
				  pixels_return[(int)
			       ((editBox_undo_buffer[((i*BOX_WIDTH)+j)]
						  *scalePackFactor)+6.0)];
			    }
			  else
			    editBox_display_buffer[((i*BOX_WIDTH)+j)]= 
			      pixels_return[(int)
				   ((editBox_undo_buffer[((i*BOX_WIDTH)+j)]
					      *scaleBoxFactor)+6.0)];
			}/* j */
		    }/* i  */
		}/* Normal */

	    }/* button 2 */
	      
	  /* Button3 is for filling in for paintMap.  */
	  else if((event.xbutton.button == Button3))
	    {/* button 3 */

	      if(paintMapSwitch)
		{/* We are painting a MAP */
		  int i,j,k,xNew,yNew;
		  int xLo,xHi,yLo,yHi;

		  yLo = y - (2 *sizeFactor);
		  yHi = y + (2 *sizeFactor);
		  xLo = x - (2 *sizeFactor);
		  xHi = x + (2 *sizeFactor);
		  for(yNew=yLo;yNew < yHi;yNew++)
		    {/* Do all Y */
		      for(xNew=xLo;xNew < xHi;xNew++)
			{/* Do all X */
			  /* Get brush size */
			  yPaintMap = (yNew * recipSizeFactor);
			  xPaintMap = (xNew * recipSizeFactor);
			  yMax= xMax = numCols;
			  
			  xPaintMap = MAX(0,MIN(xPaintMap,xMax));
			  yPaintMap = MAX(0,MIN(yPaintMap,yMax));
			  
			  /* Repaint it */
			  if(normalExcise)
			    paintMap[((curBox*numCols*numRows)+
				      (yPaintMap*numCols)+xPaintMap)]= 
					float_image_buffer1[((curBox*numCols*
							    numRows)+
							   (yPaintMap*numCols)+
							   xPaintMap)];
			  else
			    paintMap[((curBox*numCols*numRows)+
				      (yPaintMap*numCols)+xPaintMap)]= 0.0;
			  
			  yPaint = yNew - (yNew % (int)sizeFactor);
			  xPaint = xNew - (xNew % (int)sizeFactor);
			  yMax= xMax = ((numCols*sizeFactor));
			  
			  xPaint = MAX(0,MIN(xPaint,(xMax)));
			  yPaint = MAX(0,MIN(yPaint,(yMax)));
			  for(i=yPaint;i<(yPaint+sizeFactor);i++)
			    for(j=xPaint;j<(sizeFactor+xPaint);j++)
			      {
				if(hiColor)
				  {/* hi */
				    XPutPixel(editBox_image1, j, i, 
					      pixels_return[RED]);
				    editBox_display_buffer[((i*BOX_WIDTH)+j)]=
				      (BYTE)RED;
				  }/* hi */
				else
				  editBox_display_buffer[((i*BOX_WIDTH)+j)]= 
				    pixels_return[RED];
			      }
			}/* Do all X */
		    }/* Do all Y */

		  stuffDisplayMapBuffer(curBox,editBoxZoom,0);
		  DrawEditBoxWindow(TRUE);

		}/* We are painting a MAP */
	    }/* button 3 */
	  
	  DrawEditBoxWindow(FALSE);

	  break;
	      
	case KeyPress:
	  ans= XLookupKeysym((XKeyEvent *)&event, 0);
	  printf ("Keyed[%c]\n",ans);
	  
	  break;
	  
	default:
	  break;
	}/* Find out which event we have */
    }/* Only look at events meant for us */
  
  calcSizeFactor(zoomFactor);
  XFlush(dsply);            /* Make changes current */
  
}/* EDITBOXCLICK */



/*************************************************************
 * editBoxMotion() - Whenever the pointer moves across our 
 * editBox area call here. This is used to do the functions
 * of editBoxclick above except do them repeatedly while
 * the pointer moves across the box.  Used in painting
 * boxes.
*/
void editBoxMotion(Widget widget, caddr_t unused, 
		   register XMotionEvent *pEvent)
{/* editBoxMotion */
  int x, y, k, xPaint, yPaint, xMax, yMax, thisRadius;
  int yPaintMap,xPaintMap;
  static int lastX=0;
  int thisOffset;
  float thisPixel,offset;
  XmString simpleString;
  char ourStr[32];
  float xBoxCen, yBoxCen;
  float thisGlobAngx,thisGlobAngy;

  if(fileType==PACKEDPORTABLE)
    thisRadius = packRadius;
  else thisRadius = boxRadius;

  switch (pEvent->type) 
    {/* Find out which event we have */
	  
    case MotionNotify:
	      
      x = (int)(pEvent->x);         /* Get coords of pointer */
      y = (int)(pEvent->y);
      
      calcBoxSizeFactor(editBoxZoom);
      
      /* This is the Button1 mouse down state */
      if(pEvent->state == 256)
	{/* Button one = 256 find the name for this!!! */

	  if(shadedMaps && !paintMapSwitch)
	    {/* We've got depth cued shaded data */
	      int xDiff,yDiff;
	      double absXdiff,absYdiff;
	      if(flashLightIt >=99)
		{/* OK to flashLight it */

		  xBoxCen = (float)(((float)largestDim * sizeFactor) * 0.5);
		  yBoxCen = (float)(((float)largestDim * sizeFactor) * 0.5);

		  xDiff = x - (int)xBoxCen;
		  yDiff = y - (int)yBoxCen;

		  if((xDiff) != 0.0)
		    thisGlobAngx = (atan((double)((float)xDiff/
						  (float)xBoxCen)));
		  else thisGlobAngx = 0.0;

		  if((xBoxCen) != 0.0)
		    thisGlobAngy = (atan((double)((float)yDiff/
						  (float)xBoxCen)));
		  else thisGlobAngy = 0.0;

		  sinx           = sin(thisGlobAngx);
		  cosx           = cos(thisGlobAngx);

		  siny           = sin(thisGlobAngy);
		  cosy           = cos(thisGlobAngy);
		  
		  sprintf(outStr,"%dx,%dy",
			  (int)rad2deg(thisGlobAngx),
			  (0-(int)rad2deg(thisGlobAngy))
			  );
		  createSimpleString(&simpleString,outStr);
		  XtSetArg(arglst[0], XmNlabelString, simpleString);
		  XtSetValues(lightSourceLabelWidget, arglst, 1); 
		  XFlush(dsply);
		  XmStringFree(simpleString); 
		  
		  /* Get this from interface in future */
		  cenx= ceny= cenz= (origNumSecs/2)+1;
		  
		  free(rwa_tmp2d_map);
		  free(rwa_tmp_map);
		  CK(rwa_tmp2d_map = (double *)malloc(((largestDim * 
							largestDim) + 
						      22) *
						     sizeof(double)));
		  CK(rwa_tmp_map   = (double *)malloc((largestDim + 22) * 
						     sizeof(double)));
		  emmap_rob_shd1((cue_t *)cuedMap,
				 rwa_tmp2d_map, rwa_tmp_map,
				 largestDim, largestDim,
				 (curBox+1), (curBox+1), 
				 globFract, FALSE,  /* Don't use for min/max */
				 rad2deg(thisGlobAngx),0.0,
				 rad2deg(thisGlobAngy),whichStep,
				 0.0, ANGX, ANGY, FRACT);
		  stuffDisplayMapBuffer(curBox,editBoxZoom,0);
		  DrawEditBoxWindow(TRUE);
		}/* OK to flashLight it */
	      else
		{/* Move movie frames based on motion */
		  char curStr[256];
		  if(x > lastX)
		    {/* Move left */
		      curBox--;
		    }/* Move left */
		  else
		    {/* move right */
		      curBox++;
		    }/* move right */
		  /* Set the text widget to agree */
		  if(shadedMaps && stereoMapShade)
		    curBox = MAX(0,MIN(curBox,(numSecs-2)));
		  else curBox = MAX(0,MIN(curBox,(numSecs-1)));
		  /* Human numbers from one not zero */
		  sprintf(curStr,"%d",(curBox+lastNumBox+1));
		  XmTextSetString(editBoxTextWidget, curStr);
		  XFlush(dsply);
		  /* Display the new box */
		  stuffDisplayMapBuffer(curBox,editBoxZoom,0);
		  if(shadedMaps && stereoMapShade)
		    stuffDisplayMapBuffer(curBox+1,editBoxZoom,numCols);
		  DrawEditBoxWindow(TRUE);
		  lastX = x;
		}/* Move movie frames based on motion */


	    }/* We've got depth cued shaded data */
	  else
	    {/* Regular boixing painting */

	      if(paintMapSwitch)
		{/* We are painting a MAP */
		  /* Get brush size */
		  yPaintMap = (y * recipSizeFactor);
		  xPaintMap = (x * recipSizeFactor);
		  yMax= xMax = numCols;
		  
		  xPaintMap = MAX(0,MIN(xPaintMap,xMax));
		  yPaintMap = MAX(0,MIN(yPaintMap,yMax));
		  
		  /* Repaint it */
		  if(normalExcise)
		    paintMap[((curBox*numCols*numRows)+
			      (yPaintMap*numCols)+xPaintMap)]= 
				float_image_buffer1[((curBox*numCols*numRows)+
				       (yPaintMap*numCols)+xPaintMap)];
		  else
		    paintMap[((curBox*numCols*numRows)+
			      (yPaintMap*numCols)+xPaintMap)]= 
				0.0;
		    
		  
		  yPaint = y - (y % (int)sizeFactor);
		  xPaint = x - (x % (int)sizeFactor);
		  yMax= xMax = ((numCols*sizeFactor));
		  
		  xPaint = MAX(0,MIN(xPaint,(xMax)));
		  yPaint = MAX(0,MIN(yPaint,(yMax)));
		  for(i=yPaint;i<(yPaint+sizeFactor);i++)
		    for(j=xPaint;j<(sizeFactor+xPaint);j++)
		      {
			if(hiColor)
			  {/* hi */
			    XPutPixel(editBox_image1, j, i, 
				      pixels_return[RED]);
			    editBox_display_buffer[((i*BOX_WIDTH)+j)]=
			      (BYTE)RED;
			  }/* hi */
			else
			  editBox_display_buffer[((i*BOX_WIDTH)+j)]= 
			    pixels_return[RED];
		      }
		  
		  stuffDisplayMapBuffer(curBox,editBoxZoom,0);
		  DrawEditBoxWindow(TRUE);

		}/* We are painting a MAP */
	      else
		{/* Normal */
		  /* get brush size */
		  yPaint = y - (paintSize/2);
		  xPaint = x - (paintSize/2);
		  
		  yMax= xMax = ((thisRadius*2)+1) * sizeFactor;
		  
		  xPaint = MAX(0,MIN(xPaint,(xMax-paintSize)));
		  yPaint = MAX(0,MIN(yPaint,(yMax-paintSize)));
		  
		  /* Paint appropriate */
		  for(i=yPaint;i<(yPaint+paintSize);i++)
		    for(j=xPaint;j<(paintSize+xPaint);j++)
		      if(bkgrndMethod==USEAVGPIXEL)
			{/* avg */
			  if(hiColor)
			    {/* hi */
			      XPutPixel(editBox_image1, j, i, 
					pixels_return[(int)backDispChar]);
			      editBox_display_buffer[((i*BOX_WIDTH)+j)]=
				(BYTE)backDispChar;
			    }/* hi */
			  else
			    editBox_display_buffer[((i*BOX_WIDTH)+j)]= 
			      backDispChar;
			  editBox_mirror_buffer[((i*BOX_WIDTH)+j)]= (float)backChar;
			}/* avg */
		      else
			{/* random */
			  float thisVal;
			  thisVal = calcFloatBackground(editBox_int_buffer, 
						      xMax, 
						      yMax,
						      BOX_WIDTH,
						      fileType,
						      bkgrndMethod,
						      boxMethod,
						      packRadius,
						      boxRadius,
						      featherRadius,
						      recipSizeFactor,
						      tryColors
						      );
			  if(hiColor)
			    {/* hi */
			      XPutPixel(editBox_image1, j, i, 
					pixels_return[6+(int)thisVal]);
			      editBox_display_buffer[((y*BOX_WIDTH)+x)]=
				(BYTE)6+thisVal;
			      editBox_mirror_buffer[((i*BOX_WIDTH)+j)]= 
				(float)(6.0+thisVal);
			    }/* hi */
			  else
			    {/* lo */
			      editBox_display_buffer[((i*BOX_WIDTH)+j)]= 
				pixels_return[(6+(int)thisVal)];
			      editBox_mirror_buffer[((i*BOX_WIDTH)+j)]= 
				(float)pixels_return[(6+(int)thisVal)];
			    }/* lo */
			}/* random */
		}/* Normal */
	      
	      XPutImage(dsply, editBox_window1, editBox_gc, editBox_image1, 
			xPaint, yPaint,
			xPaint, yPaint,  paintSize, paintSize);
	      XFlush (dsply);
	    }/* Regular boixing painting */
	  
	}/* Button one */

      /* This is the button2 mouse down state */
      else if(pEvent->state == 512)
	{/* Button two = 512 find the name for this!!! */
	  
	  if(paintMapSwitch)
	    {/* We are painting a MAP */
	      /* Get brush size */
	      yPaintMap = (y * recipSizeFactor);
	      xPaintMap = (x * recipSizeFactor);
	      yMax= xMax = numCols;
	      
	      xPaintMap = MAX(0,MIN(xPaintMap,xMax));
	      yPaintMap = MAX(0,MIN(yPaintMap,yMax));
	      
	      /* Repaint it */
	      if(normalExcise)
		paintMap[((curBox*numCols*numRows)+
			  (yPaintMap*numCols)+xPaintMap)]= 0.0;
	      else
		paintMap[((curBox*numCols*numRows)+
			  (yPaintMap*numCols)+xPaintMap)]= 
			    float_image_buffer1[((curBox*numCols*numRows)+
				       (yPaintMap*numCols)+xPaintMap)];
	      
	      yPaint = y - (y % (int)sizeFactor);
	      xPaint = x - (x % (int)sizeFactor);
	      yMax= xMax = ((numCols*sizeFactor));
	      
	      xPaint = MAX(0,MIN(xPaint,(xMax)));
	      yPaint = MAX(0,MIN(yPaint,(yMax)));
	      
	      
	      if(minMapPixel < 0.0)
		offset= (0.0 - minMapPixel);
	      else offset= 0.0; 
	      thisPixel=  float_image_buffer1[((curBox*numCols*numRows)+
					       (yPaintMap*numCols)+
					       xPaintMap)] +
						 offset;
	      
	      thisOffset = ((int) ( (int) ( ((float)thisPixel) * 
					   scaleMapFactor )   ) + 6);
	      
	      for(i=yPaint;i<(yPaint+sizeFactor);i++)
		for(j=xPaint;j<(sizeFactor+xPaint);j++)
		    {
			if(hiColor)
			  {/* hi */
			    XPutPixel(editBox_image1, j, i, 
				      pixels_return[thisOffset]);
			    editBox_display_buffer[((i*BOX_WIDTH)+j)]=
			      (BYTE)thisOffset;
			    }/* hi */
			  else
			    editBox_display_buffer[((i*BOX_WIDTH)+j)]= 
			      (BYTE)pixels_return[thisOffset];
		    }
	      stuffDisplayMapBuffer(curBox,editBoxZoom,0);
	      DrawEditBoxWindow(TRUE);
	    }/* We are painting a MAP */
	  else
	    {/* Normal */
	      /* get brush size */
	      yPaint = y - (paintSize/2);
	      xPaint = x - (paintSize/2);
	      yMax= xMax = ((thisRadius*2)+1) * sizeFactor;
	      
	      xPaint = MAX(0,MIN(xPaint,(xMax-paintSize)));
	      yPaint = MAX(0,MIN(yPaint,(yMax-paintSize)));
	      
	      /* Paint appropriate */
	      for(i=yPaint;i<(yPaint+paintSize);i++)
		{/* i  */
		  for(j=xPaint;j<(paintSize+xPaint);j++)
		    {/* j */
		      int thisOffset;
		      editBox_mirror_buffer[((i*BOX_WIDTH)+j)]=
			(float)editBox_undo_buffer[((i*BOX_WIDTH)+j)];
		      if(fileType==PACKEDPORTABLE)
			{
			  thisOffset = (int)
			    ((editBox_undo_buffer[((i*BOX_WIDTH)+j)]
			      *scalePackFactor)+6.0);
			  if(hiColor)
			    {/* hi */
			      XPutPixel(editBox_image1, j, i, 
					pixels_return[thisOffset]);
			      editBox_display_buffer[((i*BOX_WIDTH)+j)]=
				(BYTE)thisOffset;
			    }/* hi */
			  else
			    editBox_display_buffer[((i*BOX_WIDTH)+j)]= 
			    pixels_return[thisOffset];
			}
		      else
			{
			  thisOffset = (int)
			    ((editBox_undo_buffer[((i*BOX_WIDTH)+j)]
			      *scaleBoxFactor)+6.0);
			  if(hiColor)
			    {/* hi */
			      XPutPixel(editBox_image1, j, i, 
					pixels_return[thisOffset]);
			      editBox_display_buffer[((i*BOX_WIDTH)+j)]=
				(BYTE)thisOffset;
			    }/* hi */
			  else
			    editBox_display_buffer[((i*BOX_WIDTH)+j)]= 
			      pixels_return[thisOffset];
			}
		    }/* j */
		}/* i  */
	    }/* Normal */
	  
	  XPutImage(dsply, editBox_window1, editBox_gc, editBox_image1, 
		    xPaint, yPaint,
		    xPaint, yPaint,  paintSize, paintSize);
	  XFlush (dsply);
	  
	}/* Button two=512 */

      /* Button3 is for filling in for paintMap.  */
      else if(pEvent->state == 1024)
	{/* button 3 */
	  int i,j,k,xNew,yNew;
	  int xLo,xHi,yLo,yHi;
	  
	  if(paintMapSwitch)
	    {/* We are painting a MAP */
	      
	      yLo = y - (2 *sizeFactor);
	      yHi = y + (2 *sizeFactor);
	      xLo = x - (2 *sizeFactor);
	      xHi = x + (2 *sizeFactor);
	      for(yNew=yLo;yNew < yHi;yNew++)
		{/* Do all Y */
		  for(xNew=xLo;xNew < xHi;xNew++)
		    {/* Do all X */
		      /* Get brush size */
		      yPaintMap = (yNew * recipSizeFactor);
		      xPaintMap = (xNew * recipSizeFactor);
		      yMax= xMax = numCols;
		      
		      xPaintMap = MAX(0,MIN(xPaintMap,xMax));
		      yPaintMap = MAX(0,MIN(yPaintMap,yMax));
		      
		      /* Repaint it */
		      if(normalExcise)
			paintMap[((curBox*numCols*numRows)+
				  (yPaintMap*numCols)+xPaintMap)]= 
				    float_image_buffer1[((curBox*numCols*
							  numRows)+
							 (yPaintMap*numCols)+
							 xPaintMap)];
		      else
			paintMap[((curBox*numCols*numRows)+
				  (yPaintMap*numCols)+xPaintMap)]= 0.0;
			
		      
		      yPaint = yNew - (yNew % (int)sizeFactor);
		      xPaint = xNew - (xNew % (int)sizeFactor);
		      yMax= xMax = (int)((numCols*sizeFactor));
		      
		      xPaint = MAX(0,MIN(xPaint,(xMax)));
		      yPaint = MAX(0,MIN(yPaint,(yMax)));
		      for(i=yPaint;i<(yPaint+(int)sizeFactor);i++)
			for(j=xPaint;j<((int)sizeFactor+xPaint);j++)
			  {
			    if(hiColor)
			      {/* hi */
				XPutPixel(editBox_image1, j, i, 
					  pixels_return[RED]);
				editBox_display_buffer[((i*BOX_WIDTH)+j)]=
				  (BYTE)RED;
			      }/* hi */
			    else
			      editBox_display_buffer[((i*BOX_WIDTH)+j)]= 
				pixels_return[RED];
			  }
		    }/* Do all X */
		}/* Do all Y */
	      
	      stuffDisplayMapBuffer(curBox,editBoxZoom,0);
	      DrawEditBoxWindow(TRUE);

	    }/* We are painting a MAP */

	  xLo -= sizeFactor;
	  yLo -= sizeFactor;
	  XPutImage(dsply, editBox_window1, editBox_gc, editBox_image1, 
		    xLo, yLo,
		    xLo, yLo,  
		    6*(int)sizeFactor, 
		    6*(int)sizeFactor
		    );
	  XFlush (dsply);
	  
	}/* button 3 */

      break;
	  
    default:
      break;
    }/* Find out which event we have */
  
  calcSizeFactor(zoomFactor);
  XFlush(dsply);              /* Make changes current */

}/* editBoxMotion */



/* ************************************************************
 * editBox_screen_deactivate() - Get rid of the editBox screen.
 */
void editBox_screen_deactivate(Widget widget, int *tag, 
			       XmAnyCallbackStruct *callback_data)
{/* editBox_screen_deactIVATE */
  
  XtManageChild(undoWidget);
  XtManageChild(savePaintWidget);
  XtManageChild(paintFormWidget);
  XtUnmanageChild(playFormWidget);
  XtUnmanageChild(movieSaveWidget); 
  XtUnmanageChild(movieMinWidget);
  XtUnmanageChild(movieMaxWidget);
  XtUnmanageChild(maxLabelWidget);
  XtUnmanageChild(minLabelWidget);
  XtUnmanageChild(editBoxScreenWidget);
  /* paintMapSwitch = FALSE;  Where do I reset this????? */

}/* editBox_screen_deactIVATE */



/* ************************************************************
 * editBoxText_changed() - Whenever they change the value of
 * box number text widget call here and make the interface
 * display the appropriate box.
 */
void editBoxText_changed(Widget widget, int *tag, 
			 XmAnyCallbackStruct *callback_data)
{/* EDITBOXTEXT_CHANGED */
  char *editBoxText;

  if(inited)
    {/* Don't do until we've got everything setup */
      editBoxText = XmTextGetString(editBoxTextWidget); 
      
      if(!EQU(editBoxText,""))
	{/* Don't do in the interim state */
	  if(makingMovie)
	    {/* Display the movie frame */
	      curBox = (int)strtod(editBoxText, 0);
	      if(shadedMaps && stereoMapShade)
		curBox = MAX(1,MIN(curBox,(numSecs-1)));
	      else curBox = MAX(1,MIN(curBox,(numSecs)));
	      /* Computer numbers from zero not one */
	      curBox--;
	      
	      /* Draw new box */
	      stuffDisplayMapBuffer(curBox,editBoxZoom,0);
	      if(shadedMaps && stereoMapShade)
		stuffDisplayMapBuffer(curBox+1,editBoxZoom,numCols);

	      DrawEditBoxWindow(TRUE);
	    }/* Display the movie frame */
	  else
	    {/* Display box */
	      curBox = (int)strtod(editBoxText, 0);
	      if(fileType==PACKEDPORTABLE)
		curBox = MAX(1,MIN(curBox,(purdueHeader.nimg)));
	      else curBox = MAX((1),
				MIN(curBox,(numBoxes)));
	      /* Computer numbers from zero not one */
	      curBox--;
	      
	      /* Draw new box */
	      if(fileType==PACKEDPORTABLE)
		stuffDisplayPackBoxBuffer(curBox, editBoxZoom, FALSE);
	      else stuffDisplayBoxBuffer(curBox,image_fileName.addr,
					 editBoxZoom);
	      
	      DrawEditBoxWindow(FALSE);
	    }/* Display box */
	}/* Don't do in the interim state */
      XtFree(editBoxText);

      sprintf(outStr,"%d",curBox+lastNumBox+1);
      XmTextSetString(editBoxTextWidget, outStr);
      XFlush(dsply);
    }/* Don't do until we've got everything setup */
  
}/* EDITBOXTEXT_CHANGED */



/* ************************************************************
 * editBoxZoomText_changed() - Whenever they change the value
 * of the editBox zoom text widget call here.  Redraw the
 * box based on the zoom.
*/
void editBoxZoomText_changed(Widget widget, int *tag, 
			     XmAnyCallbackStruct *callback_data)
{/* EDITBOXZOOMTEXT_CHANGED */
  char *editBoxText;

  if(inited)
    {/* Don't do until we've got everything setup */
      XFlush(dsply);
      editBoxText = XmTextGetString(editBoxZoomTextWidget); 
      
      editBoxZoom = (int)strtod(editBoxText, 0);
      
      /* Redraw for zoom */
      if(fileType==PACKEDPORTABLE)
	stuffDisplayPackBoxBuffer(curBox, editBoxZoom, FALSE);
      else if(makingMovie)
	{/* Movies */
	  stuffDisplayMapBuffer(curBox,editBoxZoom,0);
	  if(shadedMaps && stereoMapShade)
	    stuffDisplayMapBuffer(curBox+1,editBoxZoom,numCols);
	}/* Movies */
      else stuffDisplayBoxBuffer(curBox,image_fileName.addr,editBoxZoom);
      
      DrawEditBoxWindow(FALSE);
      XtFree(editBoxText);
/* malloc_verify(); */
    }/* Don't do until we've got everything setup */
  
}/* EDITBOXZOOMTEXT_CHANGED */



/* ************************************************************
 * editBoxPaintArrowActivate() - They want to change the current
 * displayed box by using the arrows.  
*/
void editBoxPaintArrowActivate(Widget widget, int *tag, 
			       XmAnyCallbackStruct *callback_data)
{/* EDITBOXPAINTARROWACTIVATE */

  /* Which arrow? */
  if(*tag==XmARROW_UP)
    {/* Up */
      paintSize++;
    }/* Up */
  else if(*tag==XmARROW_DOWN)
    {/* Dopwn */
      paintSize--;
    }/* Down */
  else printf("editBoxPaint unchanged.\n");
  paintSize = MAX(1,MIN(paintSize, 50));

  sprintf(outStr,"%d",paintSize);
  XmTextSetString(editBoxPaintTextWidget,outStr); 
  
}/* EDITBOXPAINTARROWACTIVATE */



/* ************************************************************
 * editBoxZoomArrowActivate() - They want to change the current
 * concept of zoom.
*/
void editBoxZoomArrowActivate(Widget widget, int *tag, 
			      XmAnyCallbackStruct *callback_data)
{/* EDITBOXZOOMARROWACTIVATE */

  /* Which arrow? */
  if(*tag==XmARROW_UP)
    {/* Up */
      editBoxZoom++;
    }/* Up */
  else if(*tag==XmARROW_DOWN)
    {/* Dopwn */
      editBoxZoom--;
    }/* Down */
  else printf("editBoxZoom unchanged.\n");

  sprintf(outStr,"%d",editBoxZoom);
  XmTextSetString(editBoxZoomTextWidget,outStr); 
  
  /* Redraw for zoom */
  if(fileType==PACKEDPORTABLE)
    stuffDisplayPackBoxBuffer(curBox, editBoxZoom, FALSE);
  else if(makingMovie)
    {/* Movies */
      stuffDisplayMapBuffer(curBox,editBoxZoom,0);
      if(shadedMaps && stereoMapShade)
	stuffDisplayMapBuffer(curBox+1,editBoxZoom,numCols);
    }/* Movies */
  else stuffDisplayBoxBuffer(curBox,image_fileName.addr,editBoxZoom);
  
  DrawEditBoxWindow(FALSE);
  XFlush(dsply);

}/* EDITBOXZOOMARROWACTIVATE */



/* ************************************************************
 * editBoxArrowActivate() - They want to change the current
 * displayed box by using the arrows. 
*/
void editBoxArrowActivate(Widget widget, int *tag, 
			  XmAnyCallbackStruct *callback_data)
{/* EDITBOXArrowActIVATE */
  char curStr[12];

  /* Which arrow? */
  if(*tag==XmARROW_UP)
    {/* Up */
     curBox++;
    }/* Up */
  else if(*tag==XmARROW_DOWN)
    {/* Dopwn */
      curBox--;
    }/* Down */
  else printf("EditBox unchanged.\n");

  if(makingMovie)
    {/* Put the proper movie frame up there */
      /* Set the text widget to agree */
      if(shadedMaps && stereoMapShade)
	curBox = MAX(0,MIN(curBox,(numSecs-2)));
      else curBox = MAX(0,MIN(curBox,(numSecs-1)));
      /* Human numbers from one not zero */
      sprintf(curStr,"%d",(curBox+lastNumBox+1));
      XmTextSetString(editBoxTextWidget, curStr);
      XFlush(dsply);
      /* Display the new box */
      stuffDisplayMapBuffer(curBox,editBoxZoom,0);
      if(shadedMaps && stereoMapShade)
	stuffDisplayMapBuffer(curBox+1,editBoxZoom,numCols);
      DrawEditBoxWindow(TRUE);
    }/* Put the proper movie frame up there */
  else
    {/* Put box up */
      /* Set the text widget to agree */
      if(fileType==PACKEDPORTABLE)
	curBox = MAX(0,MIN(curBox,(purdueHeader.nimg-1)));
      else curBox = MAX(0,MIN(curBox,((numBoxes)-1)));
      /* Human numbers from one not zero */
      sprintf(curStr,"%d",(curBox+lastNumBox+1));
      XmTextSetString(editBoxTextWidget, curStr);
      XFlush(dsply);
      /* Display the new box */
      if(fileType==PACKEDPORTABLE)
	stuffDisplayPackBoxBuffer(curBox, editBoxZoom, FALSE);
      else stuffDisplayBoxBuffer(curBox,image_fileName.addr,editBoxZoom);
      DrawEditBoxWindow(FALSE);
    }/* Put box up */

}/* EDITBOXArrowActIVATE */



/* ************************************************************
 * editBoxPaintText_changed() - They want to change the brush
 * size for painting.
 */
void editBoxPaintText_changed(Widget widget, int *tag, 
			      XmAnyCallbackStruct *callback_data)
{/* EDITBOXPAINTTEXT_CHANGED */
  char *editBoxPaintText;

  editBoxPaintText = XmTextGetString(editBoxPaintTextWidget); 

  paintSize = (int)strtod(editBoxPaintText, 0);
  paintSize = MAX(1,MIN(paintSize, 50));
  XtFree(editBoxPaintText);
/* malloc_verify(); */

}/* EDITBOXPAINTTEXT_CHANGED */



/* ************************************************************
 * editBoxText_create() - Remember the text widgets on the
 * editBox screen and init them.
 */
void editBoxText_create(Widget widget, int *tag, 
			XmAnyCallbackStruct *callback_data)
{/* EDITBOXTEXT_CREATE */
  
  if(*tag==PAINT)
    {/* Remember the paint widget number for later */
      editBoxPaintTextWidget= widget;
      
      XmTextSetString(editBoxPaintTextWidget, "10"); 
      paintSize= 10;
    }/* Remember the paint widget number for later */
  else if(*tag==ZOOM)
    {/* Remember the zoom widget number for later */
      editBoxZoomTextWidget= widget;
      
      XmTextSetString(editBoxZoomTextWidget, "1"); 
      editBoxZoom= 1;
    }/* Remember the zoom widget number for later */
  else if(*tag==BOXNUM)
    {/* Setup for first box */
      /* Remember the widget number for later */
      editBoxTextWidget= widget;
      
      XmTextSetString(editBoxTextWidget, "1"); 
      curBox= 0;
      /*  handle_cr(editBoxTextWidget); may not be needed */
      handle_cr(editBoxTextWidget);
    }/* Setup for first box */
  else if(*tag==CONTOUR)
    {/* Setup for contour */
      /* Remember the widget number for later */
      contourTextWidget= widget;
    }/* Setup for contour */
  else if(*tag==BOXUP)
    {/* Setup for boxup */
      boxUpWidget= widget;
      handle_cr(widget);
    }/* Setup for boxup */
  else if(*tag==BOXDOWN)
    {/* Setup for boxdown */
      boxDownWidget= widget;
      handle_cr(widget);
    }/* Setup for boxdown */
  else if(*tag==PAINTUP)
    {/* Setup for paintup */
      paintUpWidget= widget;
      handle_cr(widget);
    }/* Setup for paintup */
  else if(*tag==PAINTDOWN)
    {/* Setup for paintdown */
      paintDownWidget= widget;
      handle_cr(widget);
    }/* Setup for paintdown */
  else if(*tag==ICOSAUP)
    {/* Setup for icosaup */
      icosaUpWidget= widget;
      handle_cr(widget);
    }/* Setup for icosaup */
  else if(*tag==ICOSADOWN)
    {/* Setup for icosadown */
      icosaDownWidget= widget;
      handle_cr(widget);
    }/* Setup for icosadown */

  if(robemFontList!=0)
    {/* They've specified an override font */
      XtSetArg(arglst[0], XmNfontList,  robemFontList);
      XtSetArg(arglst[1], XmNlabelFontList,  robemFontList);
      XtSetArg(arglst[2], XmNbuttonFontList,  robemFontList);
      XtSetValues(widget, arglst, 3);
      XFlush(dsply);
    }/* They've specified an override font */
  
}/* EDITBOXTEXT_CREATE */





/*-----------------------------------*/
/*************************************
 * RULER callbacks                   *
 *************************************/
/*-----------------------------------*/
/* ************************************************************
 * rulerText_changed() - They have new values for magnification
 * and pixel length.
*/
void rulerText_changed(Widget widget, int *tag, 
		       XmAnyCallbackStruct *callback_data)
{/* RULERTEXT_CHANGED */
  char *rulerText;

  if(*tag==MAG)
    {/* Magnification changed */
      rulerText = XmTextGetString(rulerMagTextWidget); 
      rulerMag = (float)strtod(rulerText, 0);
      XtFree(rulerText);
/* malloc_verify(); */

    }/* Magnification changed */
  else if(*tag==PIXEL)
    {/* Pixel width changed */
      rulerText = XmTextGetString(rulerPixelTextWidget); 
      rulerPixelWidth = (float)strtod(rulerText, 0);
      XtFree(rulerText);
/* malloc_verify(); */

    }/* Pixel width changed */


}/* RULERTEXT_CHANGED */



/* ************************************************************
 * rulerText_create() - Remeber these text widgets to get
 * values from them later.
*/
void rulerText_create(Widget widget, int *tag, 
		      XmAnyCallbackStruct *callback_data)
{/* RULERTEXT_CREATE */

  if(*tag==MAG)
    {/* Magnification create */
      rulerMagTextWidget = widget;
    }/* Magnification create */
  else if(*tag==PIXEL)
    {/* Pixel width create */
      rulerPixelTextWidget = widget;
    }/* Pixel width create */
  else if (*tag==MINMAPPIXEL)
    minMapTextW = widget;
  else if (*tag==MAXMAPPIXEL)
    maxMapTextW = widget;

}/* RULERTEXT_CREATE */






/*-----------------------------------*/
/*************************************
 * HELP Screen -                     *
 *************************************/
/*-----------------------------------*/

/* ************************************************************
 * help_save_activate() - They have some new changes for the
 * help document - save them.  Turn this function off later
 * after the help has been fully built/debugged.
*/
void help_save_activate(Widget widget, int *tag, 
			XmAnyCallbackStruct *callback_data)
{/* help_save_actIVATE */
  FILE *fOut;
  char *helpBuf;

  /* Make sure we've got a backup */
  /*  system("cp help.doc help.doc_back"); */
  fOut = fopen("help.doc","w");

  if(fOut)
    {/* Write current buffer */
      helpBuf= XmTextGetString(helpTextWidget);
      fwrite(helpBuf,sizeof(char),strlen(helpBuf),fOut);
      XtFree(helpBuf); 
    }/* Write current buffer */

  fclose(fOut);

  system("mail sinkovit@sdsc.edu < help.doc");


}/* help_save_actIVATE */




/* ************************************************************
 * help_screen_deactivate() - Get rid of the help screen.
 */
void help_screen_deactivate(Widget widget, int *tag, 
			    XmAnyCallbackStruct *callback_data)
{/* help_screen_deactIVATE */

  XtUnmanageChild(helpScreenWidget);
  
}/* help_screen_deactIVATE */




/* ************************************************************
 * help_screen_activate() - Put the helpscreen on the display.
 * Read the document from disk and stuff the text buffer with it.
 */
void help_screen_activate(Widget widget, int *tag, 
			  XmAnyCallbackStruct *callback_data)
{/* help_screen_actIVATE */
  fString_t passCommand;
  int status;
  long nowaitflag;

  /* Just run it on UNIX */
  /***************** OBSOLETE
  sprintf(outStr,"/bio/3dem2/rwa/devel/help_robem/main_robem.do &");
  system(outStr);
  **********************************************/
  askThem("Use the question mark '?' button for help.",
	  FALSE,
	  XmDIALOG_INFORMATION);


}/* help_screen_actIVATE */



/*************************************************************
 * context_help_screen_activate() - We are trying a new way of
 * providing help by giving them a question mark "?" button
 * on each screen that will eventually contain some useful
 * tips on how to use that screen.
 */
void context_help_screen_activate(Widget widget, int *tag, 
				  XmAnyCallbackStruct *callback_data)
{/* context_help_screen_actIVATE */
  fString_t passCommand;
  int status,done;
  long nowaitflag;
  FILE *fIn;
  char inChar;
  char *helpBuf,*s1,*s2;
  char srchStr[256],inStr[1024];

  /* Make a backup */
/*  system("cp help.doc help.doc_orig"); */

  if(*tag==UNITCELL)
    sprintf(srchStr,"UNITCELL");
  else if(*tag==XYPLOT)
    sprintf(srchStr,"XYPLOT");
  else if(*tag==HISTOGRAM)
    sprintf(srchStr,"HISTOGRAM");
  else if(*tag==FOURIERFILT)
    sprintf(srchStr,"FOURIERFILT");
  else if(*tag==USEMAPSHADE)
    sprintf(srchStr,"USEMAPSHADE");
  else if(*tag==ROTFLTHELP)
    sprintf(srchStr,"ROTFLTHELP");
  else if(*tag==HELPTOOLW)
    sprintf(srchStr,"HELPTOOLW");
  else if(*tag==SUBTRACTMAP)
    sprintf(srchStr,"SUBTRACTMAP");
  else if(*tag==TABEGX)
    sprintf(srchStr,"TRUNCADD");
  else if(*tag==THETAT)
    sprintf(srchStr,"THETAT");
  else if(*tag==FIXADD)
    sprintf(srchStr,"FIXADD");
  else if(*tag==FIX2D)
    sprintf(srchStr,"FIX2D");
  else if(*tag==BOXHELP)
    sprintf(srchStr,"BOXHELP");
  else if(*tag==APPODIZEHELP)
    sprintf(srchStr,"APPODIZEHELP");
  else if(*tag==FOURAVERHELP)
    sprintf(srchStr,"FOURAVERHELP");
  else if(*tag==NORMALHELP)
    sprintf(srchStr,"NORMALHELP");
  else if(*tag==MANIPPPHELP)
    sprintf(srchStr,"MANIPPPHELP");
  else if(*tag==CAVGPPHELP)
    sprintf(srchStr,"CAVGPPHELP");
  else if(*tag==DELTAOMEGAPPHELP)
    sprintf(srchStr,"DELTAOMEGAPPHELP");
  else if(*tag==DELTAXYPARTPROCHELP)
    sprintf(srchStr,"DELTAXYPARTPROCHELP");
  else if(*tag==AUTODRHELP)
    sprintf(srchStr,"AUTODRHELP");
  else if(*tag==AUTO2DDRHELP)
    sprintf(srchStr,"AUTO2DDRHELP");
  else if(*tag==MODELHELP)
    sprintf(srchStr,"MODELHELP");
  else if(*tag==BINNINGHELP)
    sprintf(srchStr,"BINNINGHELP");
  else if(*tag==INTERPHELP)
    sprintf(srchStr,"INTERPHELP");
  else if(*tag==DIOHELP)
    sprintf(srchStr,"DIOHELP");
  else if(*tag==IMGCONVHELP)
    sprintf(srchStr,"IMGCONVHELP");
  else if(*tag==FFTHELP)
    {/* WHich kind of fft help */

      if(measureFFT)
	sprintf(srchStr,"FFTMEASURE");
      else if(drawCTF)
	sprintf(srchStr,"CTF");
      else if(rotfltSwitch)
	sprintf(srchStr,"ROTFLT");
      else
	sprintf(srchStr,"LATREFINE");

    }/* WHich kind of fft help */
  else
    {/* ??? */
      sprintf(outStr,"[CONTEXT_HELP] I don't understand [%d]\n",*tag);
      askThem(outStr,
	      FALSE,
	      XmDIALOG_INFORMATION);
      goto noHelp;      
    }/* ??? */

  CK(s2= helpBuf= (char *)malloc((sizeof(char)*64000)));

  /* Read the doc */
  sprintf(outStr,"%s/context_help.mdoc",getenv("EMDIR"));
  fIn = fopen(outStr,"r");


  if(fIn)
    {/* Write current buffer */
      fread(helpBuf,sizeof(char),63000,fIn);
      fclose(fIn);
    }/* Write current buffer */
  else
    {/* error */
      char tmpStr[512];
      sprintf(tmpStr,"Can't find context_help.mdoc file in EMDIR area. [%s]",
	      outStr);
      askThem(tmpStr,
	      FALSE,
	      XmDIALOG_INFORMATION);
      goto noHelp;
    }/* error */

  done = FALSE;
  s2= strstr(helpBuf,srchStr);
  
  if(s2)
    {/* Found it */
      s2 += strlen(srchStr);
      s1= helpBuf;
      sprintf(s1,"%s\n",srchStr);
      s1 += strlen(srchStr)+1;
      *s1++ = '\n';
      while(!done)
	{/* Stuff and read */
	  *s1 = *s2;
	  s1 ++;
	  s2 ++;
	  if((*s2 == 'D') &&
	     (*(s1-1)=='N') && (*(s1-2)=='E'))
	    done = TRUE;
	}/* Stuff and read */
      *s1++= 'D';
      *s1 = '\0';
      goto AllDone;
    }/* Found it */
  else
    sprintf(helpBuf,"No Help found.");

 AllDone:;
  
  /* Stuff the text buffer */
  XmTextSetString(helpTextWidget,helpBuf);
  free(helpBuf);

  raiseIt(helpScreenWidget);

 noHelp:;

}/* context_help_screen_actIVATE */



/* ************************************************************
 * help_text_create() - Rememebr the help text widget number
 * for stuffing later.
*/
void help_text_create(Widget widget, int *tag, 
		      XmAnyCallbackStruct *callback_data)
{/* help_text_CREATE */

  helpTextWidget = widget;

}/* help_text_CREATE */





/*********************************************************
 * setFont_create() - Setup these widgets with the font
 * we get from the resource database.
*/
void setFont_create(Widget widget, int *tag, 
		    XmAnyCallbackStruct *callback_data)
{/* SETFONT_CREATE */

  if(robemFontList!=0)
    {/* They've specified an override font */
      XtSetArg(arglst[0], XmNfontList,  robemFontList);
      XtSetArg(arglst[1], XmNlabelFontList,  robemFontList);
      XtSetArg(arglst[2], XmNbuttonFontList,  robemFontList);
      XtSetValues(widget, arglst, 3);
      XFlush(dsply);
    }/* They've specified an override font */

  if(*tag==EDITBOX)
    editBoxButtonW= widget;
  else if(*tag==PAINTMAP)
    paintMapButtonW= widget;
  else if(*tag==REMOVEBLEMISH)
    removeBlemishButtonW= widget;
  else if(*tag==FIX)
    fixButtonW= widget;
  else if(*tag==ROTATE)
    rotateButtonW= widget;
  else if(*tag==ROTATEEXIT)
    rotateExitW= widget;
  else if(*tag==RESCALE)
    rescaleButtonW= widget;
  else if(*tag==SUBTRACTMAP)
    subtractMapButtonW= widget;
/*
Why is this here ???
  else if(*tag==SUBTRACTMAPEXIT)
    subtractExitW= widget;
*/
  else if(*tag==CALCSFS)
    calcSFButtonW= widget;
  else if(*tag==RCALCSFS)
    RcalcSFButtonW= widget;
  else if(*tag==HARDROTATE)
    hardRotateButtonW= widget;
  else if(*tag==INVERT)
    invertButtonW= widget;
  else if(*tag==TRUNCADD)
    truncaddButtonW= widget;
  else if(*tag==TRUNCADDEXIT)
    truncaddExitW= widget;
  else if(*tag==FIXEXIT)
    fixExitW= widget;
  else if(*tag==FIX2DEXIT)
    fix2DExitW= widget;
  else if(*tag==MAPMOVIE)
    mapMovieButtonW= widget;
  else if(*tag==CIRCAVG)
    circAvgButtonW= widget;
  else if(*tag==MAPSHADE)
    mapShadeButtonW= widget;
  else if(*tag==CORRELATE)
    correlateButtonW= widget;
  else if(*tag==CORRELATE2)
    correlateButton2W= widget;
  else if(*tag==TIFF)
    rawButtonW= widget;
  else if(*tag==COLOR)
    colorButtonW= widget;
  else if(*tag==CIRCAVGLIST)
    circAvgListW= widget;
  else if(*tag==FOURIERFILT)
    fourierFiltWidget= widget;
  else if(*tag==ADDAVERAGE)
    addAverageWidget= widget;
  else if(*tag==BINNING)
    binningWidget= widget;
  else if(*tag==INTERPOLATE)
    interpolateWidget= widget;
  else if(*tag==MODEL)
    modelWidget= widget;
  else if(*tag==CHECKBKG)
    checkBkgWidget= widget;
  else if(*tag==NORMAL)
    normalizeWidget= widget;
  else if(*tag==NORMALBKG)
    normalizeBkgWidget= widget;
  else if(*tag==LINGRAD)
    linGradWidget= widget;
  else if(*tag==AUTOBOX)
    autoBox_widget= widget;
  else if(*tag==AUTOBOX2)
    autoBox2_widget= widget;
  else if(*tag==USEMAPSHADE)
    {/* map shade 3d rendering screen */
      mapShadeScreenWindow = widget;
      handle_cr(widget);
    }/* map shade 3d rendering screen */
  else if(*tag==CUTMINTEXT)
    {/* map shade 3d rendering screen */
      cutFormW = widget;
      handle_cr(widget);
    }/* map shade 3d rendering screen */
  else if(*tag==UNBLOBSIZE)
    {/* map shade 3d rendering screen */
      unblobFormW = widget;
      handle_cr(widget);
    }/* map shade 3d rendering screen */
  else if(*tag==PARMTOGS)
    parmTogsW = widget;
  else if(*tag==PARMTOGSLABEL)
    parmTogsLabelW = widget;
  else if(*tag==PARMHEAD)
    parmheadW =	widget;
  else if(*tag==BATCHPARAMFILES)
    batchParamFilesW = widget;
  else if(*tag==HISTCONTROLS)
    histControlsW = widget;
  else if(*tag==PCX)
    pcxButtonW = widget;
  else if(*tag==STEPTOGGLE)
    {
      movieFormW = widget;
    }

}/* SETFONT_CREATE */




/* ************************************************************
 * createLock() - Remeber the lock widget number for
 * setting the pixmaps later.
 */
void createLock(Widget widget, int *tag, XmAnyCallbackStruct *callback_data)
{/* CREATELOCK */

  if(*tag==USELOCK)
    lockWidget= widget;
  else if(*tag==USESECTION)
    {/* Section/cued map */
      sectionWidget= widget;
    }/* Section/cued map */
  else if(*tag==COLOR)
    {/* color editor tool */
      colorToolWidget= widget;
    }/* color editor tool */
  else if(*tag==VERSIONW)
    {/* version tool */
      versionToolWidget= widget;
    }/* version tool */

}/* CREATELOCK */




/*************************************************************
 * tryThis_activate() - They're happy with the scrollbar 
 * selections they've made so compute it.
 */
void tryThis_activate(Widget widget, int *tag, 
		      XmAnyCallbackStruct *callback_data)
{/* TRYTHIS_ACTIVATE */
  int firstTime;

  mapShade_activate(0,&firstTime,0);

}/* TRYTHIS_ACTIVATE */



/****************************************************************
 * em_sleep() - I can't find a VMS sleep() :-(
*/
void em_sleep(int i)
{/* em_sleep */
  int j,k;
  float a,b,c,d;

  for(j=0;j<10000;j++)
    {
      a = j / (i + 2.2);
      for(k=0;k<1000;k++)
	{
	  d = k / (i+5.5);
	  b = a + d;
	  c = b / 7.7;
	}
    }

}/* em_sleep */




/************************************************************
 * boxgrid2_activate() - Try putting a default grid of boxes
 * out there and then do a reference refine on them.  This
 * is an experiment for autoboxing empty particles.
 */
void boxgrid2_activate(Widget widget, int *tag, 
		       XmAnyCallbackStruct *callback_data)
{/* BOXGRID2_ACTIVATE */
  int i,j,m,n,aveX,aveY;
  int saveRefAuto,saveDCR;
  int save_file_width,save_file_height;
  int save_boxRadius,save_emmaRadius,save_emptyShellWidth;
  float *save_inCore_buffer, accumVal;
  int shrinkFactor, numAccum;
  int compressed;

  //  Save what we've got now 
  save_inCore_buffer = inCore_buffer;
  save_file_width   = file_width;
  save_file_height  =  file_height;
  save_boxRadius    = boxRadius;
  save_emmaRadius   = emmaRadius; 
  save_emptyShellWidth   = emptyShellWidth; 
 
  // Compute how much we need to shrink
  compressed   = FALSE;
  shrinkFactor = 1;
  while(!compressed)
    {/* Do until done */
      if((boxRadius/shrinkFactor) < 20)
	{/* done */
	  compressed = TRUE;
	}/* done */
      else
	{/* SHrink it */
	  shrinkFactor *= 2;
	}/* SHrink it */
    }/* Do until done */
  printf("Shrinking by [%d]\n",shrinkFactor);

  // Now shrink it all 
  if(shrinkFactor > 1)
    {/* shrink */
      file_height /= shrinkFactor;
      file_width  /= shrinkFactor;
      boxRadius   /= shrinkFactor;
      emmaRadius  /= shrinkFactor;
      emptyShellWidth  /= shrinkFactor;
      CK(inCore_buffer = (float *)malloc(sizeof(float)*
					 file_height*file_width));
      numAccum = shrinkFactor * shrinkFactor;
      for(i=0;i<file_height;i++)
	{/* row */
	  for(j=0;j<file_width;j++)
	    {/* col */
	      accumVal = 0.0;
	      for(m=0;m<shrinkFactor;m++)
		{
		  for(n=0;n<shrinkFactor;n++)
		    {accumVal += save_inCore_buffer[(((i*shrinkFactor)+m)*
						     save_file_width)+
						   ((j*shrinkFactor)+n)];}
		}
	      inCore_buffer[(i*file_width)+j] = (float)(accumVal / 
							(float)numAccum);
	    }/* col */
	}/* row */
    }/* shrink */
  fullScreen = FALSE;
  virtual_read_activate(-99,-99);

  //  This works
  aveX = emmaRadius * 2;
  aveY = emmaRadius * 2;
  //  This is too many particles - takes forever to eliminate dups
  aveX = (int)((float)emmaRadius * 1.5) ;
  aveY = (int)((float)emmaRadius * 1.5) ;
  if(debugSwitch && debugLevel==3)
    {/* Let's try this experiment */
      aveX = emmaRadius * 2.15;
      aveY = emmaRadius * 2.15;
    }/* Let's try this experiment */

  for(i = emmaRadius; i < file_height; i+=aveY)
    {/* do y */
      for(j = emmaRadius; j < file_width; j += aveX)
	{/* do x */
	  if(numBoxes < MAXBOX)
	    {
	      box[numBoxes].x = j;
	      box[numBoxes].y = i;
	      box[numBoxes].reference = FALSE;
	      box[numBoxes].correlation = 1.0;
	      box[numBoxes].avgdensity = 0.0;
	      box[numBoxes].stdev = 0.0;
	      numBoxes ++;
	    }
	}/* do x */
    }/* do y */
  DrawWindow();

  firstAMMSTDEVtime = TRUE;
  //  Let them decide XmTextSetString(refLastIterTextW,"25");
  saveRefAuto = refAutomate;
  refAutomate = TRUE;
  saveDCR = darkCircleRef;
  darkCircleRef =  99;
  ref_correlate_activate();
  refAutomate = saveRefAuto;
  darkCircleRef = saveDCR;

  //  Restore  what we had
  if(shrinkFactor > 1)
    free(inCore_buffer);
  inCore_buffer =   save_inCore_buffer ;
  file_width    =   save_file_width   ;
  file_height   =   save_file_height ;
  boxRadius     =   save_boxRadius   ;
  emmaRadius    =   save_emmaRadius  ; 
  emptyShellWidth =  save_emptyShellWidth ;
  
  //  Rejigger all of the box coords
  for(i=0;i<numBoxes;i++)
    {
      box[i ].x *= shrinkFactor;
      box[i ].y *= shrinkFactor;
    }
  fullScreen = FALSE;
  virtual_read_activate(-99,-99);


}/* BOXGRID2_ACTIVATE */





/* ************************************************************
 * boxgrid_activate() - Try putting a default grid of boxes
 * out there and then do a reference refine on them.
 */
void boxgrid_activate(Widget widget, int *tag, 
		      XmAnyCallbackStruct *callback_data)
{/* BOXGRID_ACTIVATE */
  int i,j,aveX,aveY;
  int saveRefAuto,saveDCR;
  int passTag;

  if(*tag==-99 || firstGridding)
    {/* OK to make a new grid */
      //  This works
      aveX = emmaRadius * 2;
      aveY = emmaRadius * 2;
      //  This is too many particles - takes forever to eliminate dups
      aveX = (int)((float)emmaRadius * 1.5) ;
      aveY = (int)((float)emmaRadius * 1.5) ;
      if(debugSwitch && debugLevel==3)
	{/* Let's try this experiment */
	  aveX = emmaRadius * 2.15;
	  aveY = emmaRadius * 2.15;
	}/* Let's try this experiment */
      
      
      for(i = emmaRadius; i < file_height; i+=aveY)
	{/* do y */
	  for(j = emmaRadius; j < file_width; j += aveX)
	    {/* do x */
	      if(numBoxes < MAXBOX)
		{
		  box[numBoxes].x = j;
		  box[numBoxes].y = i;
		  box[numBoxes].reference = FALSE;
		  box[numBoxes].correlation = 1.0;
		  box[numBoxes].avgdensity = 0.0;
		  box[numBoxes].stdev = 0.0;
		  numBoxes ++;
		}
	    }/* do x */
	}/* do y */
      DrawWindow();
    }/* OK to make a new grid */
  firstGridding = FALSE;

  if(*tag != -99)
    {/* do frist time */
      firstAMMSTDEVtime = TRUE;
      // Let them decide      XmTextSetString(refLastIterTextW,"25");
      //  saveRefAuto = refAutomate;
      //refAutomate = TRUE;
      //saveDCR = darkCircleRef;
      //darkCircleRef =  99;
      
      passTag= CORRELATE3;
      correlate_activate(widget, &passTag,
			 callback_data);
      
      //refAutomate = saveRefAuto;
      //darkCircleRef = saveDCR;
    }/* do frist time */

}/* BOXGRID_ACTIVATE */





/***************************************************************
 * autoBox_activate() - Do the autoboxing type that they have
 * selected.
 */
void autoBox_activate(Widget widget, int *tag, 
		      XmAnyCallbackStruct *callback_data)
{/* AUTOBOX_ACTIVATE */
  FILE *fIn;
  float  *autoBox_buffer_raw;
  float  *temp_byte_buffer, *Accum, *auxAccum;
  unsigned short level;
  int totSum1;
  int total_bkg, dataRange;
  int autoboxHeight, autoboxWidth, autoboxArea, thisRadius;
  int i, j, k, dx, dy, yOut, beginY, beginX;
  int kj, jn, j1n, i1, j1, avgx, avgy, radi; 
  int ii, jj, ti, tj, clusterSize, sp;
  int startJ, endJ, startI, endI;
  int averageBorder = 5;
  int *hist_p;
  float *autoBox_buffer, *prob_p, *sum_p;
  float averageArea, sum, avgEnhanceFactor, sum1, sum2;
  float avgVal, minAvg, maxAvg, corr;
  float clusterSizeUplimit, clusterSizeDownlimit; 
  XPoint *stack, *cluster;
  XmString simpleString;
  float img_bkg, minF, maxF, img_avg, img_std, scaleF;
  int   *empty_ref_buffer,*empty_full_buffer,tbSize;
  
  firstAMMSTDEVtime = TRUE;
  /*************** Autobox [1] (EMMA method) *************/
  if(*tag==AUTOBOX || *tag==CORRELATE)
    {

      DrawWindow();

      printf("autoBox_activate() called [1].\n");
      XmTextSetString(timeLeftBoxW,"Autoboxing.");
      XFlush(dsply);

      /***** Load in current radius *****/
      if(fileType==PACKEDPORTABLE)
	{
	  askThem("Sorry, Only scanned images can be autoBoxed.",
		 FALSE,
		 XmDIALOG_INFORMATION);
	  return;
	}
      else
	thisRadius = emmaRadius;
      
      /***** Image compressed to particle  10< radius < 20 *****/ 
      compress_factor = (int) (thisRadius/20 + 1 );
      if(compress_factor < 1) 
	compress_factor = 1;
      thisRadius = (int) (thisRadius/compress_factor + 0.5);
      radi = thisRadius;
      tbSize = (radi*2)+1;
      autoboxHeight = (int)file_height/compress_factor;
      autoboxWidth = (int)file_width/compress_factor;
      autoboxArea = autoboxHeight*autoboxWidth;

      if(emptyBox && FALSE)
	{/* empties */
	  float totPixVal, totAccum,dist,xSq,ySq;
	  int maxDist,minDist,tbSizeHalf;
	  int xPos,yPos,thisValue,m,n;

	  /* Build ref buffer */
	  empty_full_buffer = (int *)malloc(sizeof(int)*tbSize*tbSize);
	  empty_ref_buffer  = (int *)malloc(sizeof(int)*tbSize*tbSize);
	  totPixVal = 0.0;
	  tbSizeHalf = (int)( tbSize * 0.5);
	  maxDist = radi*radi;
	  minDist = (int)(maxDist * 0.60);
	  for(i=0;i<tbSize;i++)
	    {
	      ySq = (i-tbSizeHalf)*(i-tbSizeHalf);
	      for(j=0;j<tbSize;j++)
		{/* Get the buffer */
		  xSq = (j-tbSizeHalf)*(j-tbSizeHalf);
		  dist = xSq + ySq;
		  
		  /* Just give me a 1 or 0 */
		  if(dist > minDist && dist <= maxDist)
		    empty_ref_buffer[((i*tbSize)+j)] = 1;
		  else
		    empty_ref_buffer[((i*tbSize)+j)] = 0;
		  if(dist <= maxDist)
		    empty_full_buffer[((i*tbSize)+j)] = 1;
		  else
		    empty_full_buffer[((i*tbSize)+j)] = 0;
		  /* DEBUG 
		  thisValue = empty_full_buffer[((i*tbSize)+j)];
		  XPutPixel(image_image1,
			    j,i,
			    pixels_return[(int)(thisValue*100)+10]
			    );
		  */
		}/* Get the buffer */
	    }
	  /* DEBUG
	  XPutImage(dsply, image_window1, image_gc, image_image1, 
		    0, 0,
		    0, 0,  
		    tbSize, 
		    tbSize);
	  XFlush(dsply);
	  */
	}/* empties */

      /***** Open the disk file for load in raw image *****/
      fIn= NULL;
      fIn= fopen(image_fileName.addr,"r");
      if(fIn==NULL)
	{/* Error */
	  printf(" Error open image file for refine center.\n");
	  XmTextSetString(timeLeftBoxW,
			  "Error open image file for refine center.");
	  XFlush(dsply);
	  return;
	}/* Error */
      
      /***** allocate space for image and temp byte buffer *****/
      if((autoBox_buffer_raw=
	  (float *)malloc(autoboxArea*sizeof(float)))==(float *)NULL)
	{
	  printf("\nError: Allocate particle image for autoboxing");
	  XmTextSetString(timeLeftBoxW,
			  "Error allocating particle image.");
	  XFlush(dsply);
	  return;
	}
      if((autoBox_buffer = (float *)calloc(autoboxArea,
					   sizeof(float))) == (float *)NULL)
	{
	  printf("\nError: Allocate particle image for autoboxing");
	  free(autoBox_buffer_raw);
	  XmTextSetString(timeLeftBoxW,
			  "Error allocating particle image.");
	  XFlush(dsply);
	  return;
	}
      if((temp_byte_buffer=
	  (float *)malloc(file_width*sizeof(float)))==(float *)NULL)
	{
	  printf("\nError: Allocate particle image for autoboxing");
	  free(autoBox_buffer_raw);
	  free(autoBox_buffer);
	  XmTextSetString(timeLeftBoxW,
			  "Error allocating particle image.");
	  XFlush(dsply);
	  return;
	}

      /* Let's scale this */
      calcStats(inCore_buffer, &img_bkg, &minF,
		&maxF, &img_avg, &img_std,
		file_width, file_height, file_width);
      scaleF = 255.0/(maxF-minF);
      
      XmTextSetString(timeLeftBoxW,"Load raw image");
      XFlush(dsply);
      /***** Load in autobox raw image *****/
      for(j=0; j<autoboxHeight; j++)
	{
	  if(!imgInCore)
	    {/* Get from disk */
	      yOff = headOff + (j*compress_factor)*file_width;
	      sysError = fseek(fIn,yOff,0);
	      if(sysError)
		{
		  printf("Error: fseek in auto box.\n");
		  free(autoBox_buffer_raw);
		  free(autoBox_buffer);
		  free(temp_byte_buffer);
		  XmTextSetString(timeLeftBoxW,
				  "fseek() error.");
		  XFlush(dsply);
		  return;
		}
	      fread(temp_byte_buffer, sizeof(BYTE), (file_width), fIn);
	    }/* Get from disk */
	  else
	    {/* Get from core */
	      yOff = (j*compress_factor)*file_width;
	      memcpy(temp_byte_buffer,&inCore_buffer[(yOff)],
		     (file_width*sizeof(float)));
	    }/* Get from core */

	  /* Stuff the autoBox_buffer_raw */
	  yOut = (int) j*autoboxWidth;
	  for(k=0; k<autoboxWidth; k++)
	    autoBox_buffer_raw[yOut+k]= 
	      (float)(0.5 + 
		      ((temp_byte_buffer[k*compress_factor]-minF)*scaleF)
		      /2.0);
	}
      
      fclose(fIn);
      free(temp_byte_buffer);
      

      /*********** Histogram equalization ***********/
      XmTextSetString(timeLeftBoxW,"Histogram Equalize");
      XFlush(dsply);
      dataRange = 128;
      if((prob_p=(float*)malloc(dataRange*sizeof(float)))==(float*)NULL ||
	 (sum_p=(float*)malloc(dataRange*sizeof(float)))==(float*)NULL ||
	 (hist_p = (int *)malloc(dataRange*sizeof(int))) == (int *)NULL)
	{
	  printf("\nError: Allocate buffers for Histogram equalization\n");
	  XmTextSetString(timeLeftBoxW,
			  "Error allocating buffers for Hist Equal.");
	  XFlush(dsply);
	  return;
	}
      
      for(i=0;i<dataRange;i++)
	{
	  hist_p[i] = 0.;
	  prob_p[i] = 0.;
	}
      
      for(k=0;k<autoboxHeight;k++)
	for(j=0;j<autoboxWidth;j++)
	  hist_p[(int)autoBox_buffer_raw[k*autoboxWidth+j]]++;
      
      for(i=0;i<dataRange;i++)
	prob_p[i] = (float)hist_p[i]/(float)autoboxArea;
      
      for(i=0;i<dataRange;i++)
	{
	  sum_p[i] = 0.;
	  for(j=0;j<=i;j++)
	    sum_p[i] += prob_p[j];
	}
      
      for(k=0;k<autoboxHeight;k++)
	{
	  kj = k*autoboxWidth;
	  for(j=0;j<autoboxWidth;j++)
	    {
	      level = (unsigned short)(0.5+(dataRange - 1)*
				       sum_p[(int)autoBox_buffer_raw[kj+j]]);
	      if(level >= dataRange)
		level = dataRange - 1;
	      autoBox_buffer_raw[kj+j] = level;
	    }
	}
      
      free(hist_p);
      free(prob_p); 
      free(sum_p);
      

      /*********** Averaging regard to 10< radius <20 *************/
      if(radi < 12)
	averageBorder = 2;
      else
	averageBorder = 3;
      averageArea = 4 * averageBorder * averageBorder;
      minAvg = maxAvg = 0;
      
      for(j=averageBorder;j<autoboxHeight-averageBorder;j++)
	{
	  jn = j*autoboxWidth; sum = 0.;
	  for(j1=j-averageBorder;j1<j+averageBorder;j1++)
	    {
	      j1n = j1*autoboxWidth;
	      for(i1=0;i1<2*averageBorder;i1++)
		sum += autoBox_buffer_raw[j1n+i1];
	    }
	  avgVal = autoBox_buffer[jn+averageBorder] = sum/averageArea;
	  if(maxAvg < avgVal) maxAvg = avgVal;
	  
	  for(i=averageBorder+1;i<autoboxWidth-averageBorder;i++)
	    {
	      for(j1=j-averageBorder;j1<j+averageBorder;j1++)
		{
		  j1n = j1*autoboxWidth;
		  sum += autoBox_buffer_raw[j1n+i+averageBorder-1];
		  sum -= autoBox_buffer_raw[j1n+i-averageBorder-1];
		}
	      
	      avgVal = autoBox_buffer[jn+i] = sum/averageArea;
	      if(maxAvg < avgVal) maxAvg = avgVal;
	    }
	}
      
      free(autoBox_buffer_raw);
      

      /***** Enhence the contrast of averaged image *****/
      avgEnhanceFactor = (dataRange-1.)/(maxAvg - minAvg);  
      
      for(j=0; j<autoboxHeight; j++)
	{
	  yOut = j*autoboxWidth;
	  for(k=0; k<autoboxWidth; k++)
	    autoBox_buffer[yOut+k] =  (int) (0.5 +
					     avgEnhanceFactor*
					     autoBox_buffer[yOut+k]);
	}
      
      
      /***** Local background substraction by cross-point mathod *****/
      XmTextSetString(timeLeftBoxW,"Cross-Point Method");
      XFlush(dsply);
      if(globAccum)
	free(globAccum);
      globAccum = (int *)NULL;
      if((globAccum=(int *)malloc(autoboxArea*sizeof(int)))==(int *)NULL)
	{
	  printf("\ncannot allocate globAccum");
	  free(autoBox_buffer);
	  XmTextSetString(timeLeftBoxW,
			  "Error allocating buffers for globAccum.");
	  XFlush(dsply);
	  return;
	}
      if((Accum=(float *)malloc(autoboxArea*sizeof(float)))==(float *)NULL)
	{
	  printf("\ncannot allocate Accum");
	  free(autoBox_buffer);
	  XmTextSetString(timeLeftBoxW,
			  "Error allocating buffers for Accum.");
	  XFlush(dsply);
	  return;
	}
      
      if((auxAccum=(float *)malloc(autoboxArea*sizeof(float)))==(float *)NULL)
	{
	  printf("\ncannot allocate auxAccum");
	  free(autoBox_buffer);
	  free(Accum);
	  XmTextSetString(timeLeftBoxW,
			  "Error allocating buffers for Accum.");
	  XFlush(dsply);
	  return;
	}
      
      numBoxes = 0;
      for(j=0;j<autoboxHeight;j++)
	for(i=0;i<autoboxWidth;i++)
	  Accum[j*autoboxWidth+i] = 0;
      
      /* scan left to right, top to bottom */
      XmTextSetString(timeLeftBoxW,"Cross-Point Method - Pass 1");
      XFlush(dsply);
      for(j=radi+1;j<autoboxHeight-radi-1;j++)
	{
	  for(i=radi+1;i<autoboxWidth-radi-1;i++)
	    {
	      if(autoBox_buffer[j*autoboxWidth+i] -
		 autoBox_buffer[j*autoboxWidth+i+radi+1] >= prag)
		{
		  if(autoBox_buffer[j*autoboxWidth+i+radi+1] -
		     autoBox_buffer[(j+radi+1)*autoboxWidth+i+
				    radi+1] >= prag)
		    {
		      Accum[(j+radi+1)*autoboxWidth+i+radi+1] = 1;
		    }
		  else
		    if(autoBox_buffer[j*autoboxWidth+i+radi+1] -
		       autoBox_buffer[(j+radi+1)*autoboxWidth+i+
				      radi+1] <= -prag)
		      {
			Accum[j*autoboxWidth+i+radi+1] = 1;
		      }
		  
		    else
		      if(autoBox_buffer[j*autoboxWidth+i+radi+1] -
			 autoBox_buffer[(j-radi-1)*autoboxWidth+i+
					radi+1] >= prag)
			{
			  Accum[(j-radi-1)*autoboxWidth+i+radi+1] = 1;
			}
		      else
			if(autoBox_buffer[j*autoboxWidth+i+radi+1] -
			   autoBox_buffer[(j-radi-1)*autoboxWidth+i+
					  radi+1] <= -prag)
			  {
			    Accum[j*autoboxWidth+i+radi+1] = 1;
			  }
		}
	      else
		if(autoBox_buffer[j*autoboxWidth+i] -
		   autoBox_buffer[j*autoboxWidth+i+radi+1]
		   <= -prag)
		  {
		    if(autoBox_buffer[j*autoboxWidth+i] -
		       autoBox_buffer[(j+radi+1)*autoboxWidth+i]
		       >= prag)
		      {
			Accum[(j+radi+1)*autoboxWidth+i] = 1;
		      }
		    else
		      if(autoBox_buffer[j*autoboxWidth+i] -
			 autoBox_buffer[(j+radi+1)*autoboxWidth+i]
			 <= -prag)
			{
			  Accum[j*autoboxWidth+i] = 1;
			}
		    
		      else
			if(autoBox_buffer[j*autoboxWidth+i] -
			   autoBox_buffer[(j-radi-1)*autoboxWidth+i]
			   >= prag)
			  {
			    Accum[(j-radi-1)*autoboxWidth+i] = 1;
			  }
			else
			  if(autoBox_buffer[j*autoboxWidth+i] -
			     autoBox_buffer[(j-radi-1)*autoboxWidth+i]
			     <= -prag)
			    {
			      Accum[j*autoboxWidth+i] = 1;
			    }
		  }
	    }
	}
      
      
      /* scan top to bottom, left to right */
      XmTextSetString(timeLeftBoxW,"Cross-Point Method - Pass 2");
      XFlush(dsply);
      for(i=radi+1;i<autoboxWidth-radi-1;i++)
	{
	  for(j=radi+1;j<autoboxHeight-radi-1;j++)
	    {
	      if(autoBox_buffer[j*autoboxWidth+i] -
		 autoBox_buffer[(j+radi+1)*autoboxWidth+i]
		 >= prag)
		{
		  if(autoBox_buffer[(j+radi+1)*autoboxWidth+i] -
		     autoBox_buffer[(j+radi+1)*autoboxWidth+i+
                                    radi+1] >= prag)
		    {
		      Accum[(j+radi+1)*autoboxWidth+i+radi+1] = 1;
		    }
		  else
		    if(autoBox_buffer[(j+radi+1)*autoboxWidth+i] -
		       autoBox_buffer[(j+radi+1)*autoboxWidth+i+
				      radi+1] <= -prag)
		      {
			Accum[(j+radi+1)*autoboxWidth+i] = 1;
		      }
		  
		    else
		      if(autoBox_buffer[(j+radi+1)*autoboxWidth+i] -
			 autoBox_buffer[(j+radi+1)*autoboxWidth+i-
					radi-1] >= prag)
			{
			  Accum[(j+radi+1)*autoboxWidth+i-radi-1] = 1;
			}
		      else
			if(autoBox_buffer[(j+radi+1)*autoboxWidth+i] -
			   autoBox_buffer[(j+radi+1)*autoboxWidth+i-
					  radi-1] <= -prag)
			  {
			    Accum[(j+radi+1)*autoboxWidth+i] = 1;
			  }
		}
	      else
		if(autoBox_buffer[j*autoboxWidth+i] -
		   autoBox_buffer[(j+radi+1)*autoboxWidth+i]
		   <= -prag)
		  {
		    if(autoBox_buffer[j*autoboxWidth+i] -
		       autoBox_buffer[j*autoboxWidth+i+radi+1]
		       >= prag)
		      {
			Accum[j*autoboxWidth+i+radi+1] = 1;
		      }
		    else
		      if(autoBox_buffer[j*autoboxWidth+i] -
			 autoBox_buffer[j*autoboxWidth+i+radi+1]
			 <= -prag)
			{
			  Accum[j*autoboxWidth+i] = 1;
			}
		    
		      else
			if(autoBox_buffer[j*autoboxWidth+i] -
			   autoBox_buffer[j*autoboxWidth+i-radi-1]
			   >= prag)
			  {
			    Accum[j*autoboxWidth+i-radi-1] = 1;
			  }
			else
			  if(autoBox_buffer[j*autoboxWidth+i] -
			     autoBox_buffer[j*autoboxWidth+i-radi-1]
			     <= -prag)
			    {
			      Accum[j*autoboxWidth+i] = 1;
			    }
		  }
	    }
	}
      
      /* Do dilation first */
      if(autoBoxSwitch)
	nDilates = autoBoxDilates;
      printf("nDilates[%d]\n",nDilates);
      for(k=0;k<nDilates;k++)
	{/* Dilation layers */
	  for(j=1;j<autoboxHeight-1;j++)
	    {
	      for(i=1;i<autoboxWidth-1;i++)
		{
		  auxAccum[j*autoboxWidth+i] = Accum[j*autoboxWidth+i];
		  if(!Accum[j*autoboxWidth+i])
		    {
		      if(Accum[j*autoboxWidth+i-1]+
			 Accum[j*autoboxWidth+i+1]+
			 Accum[(j-1)*autoboxWidth+i]+
			 Accum[(j+1)*autoboxWidth+i]+
			 Accum[(j-1)*autoboxWidth+i-1]+
			 Accum[(j-1)*autoboxWidth+i+1]+
			 Accum[(j+1)*autoboxWidth+i-1]+
			 Accum[(j+1)*autoboxWidth+i+1] > 0)
			auxAccum[j*autoboxWidth+i] = 1;
		    }
		}
	    }
      
	  /* Now take the dilated data as our gold standard */
	  for(j=1;j<autoboxHeight-1;j++)
	    {
	      for(i=1;i<autoboxWidth-1;i++)
		Accum[j*autoboxWidth+i] = auxAccum[j*autoboxWidth+i];
	    }
	  
	}/* Dilation layers */

      /* Now do thinning */
      if(autoBoxSwitch)
	nThins = autoBoxThin;
      printf("nThins[%d]\n",nThins);
      for(k=0;k<nThins;k++)
	{/* Thinning layers */
	  for(j=1;j<autoboxHeight-1;j++)
	    {
	      for(i=1;i<autoboxWidth-1;i++)
		{
		  auxAccum[j*autoboxWidth+i] = Accum[j*autoboxWidth+i];
		  if(Accum[j*autoboxWidth+i])
		    {
		      if(Accum[j*autoboxWidth+i-1]+
			 Accum[j*autoboxWidth+i+1]+
			 Accum[(j-1)*autoboxWidth+i]+
			 Accum[(j+1)*autoboxWidth+i]+
			 Accum[(j-1)*autoboxWidth+i-1]+
			 Accum[(j-1)*autoboxWidth+i+1]+
			 Accum[(j+1)*autoboxWidth+i-1]+
			 Accum[(j+1)*autoboxWidth+i+1] < 8)
			auxAccum[j*autoboxWidth+i] = 0;
		    }
		}
	    }
      
	  /* Now take the eroded data as our gold standard */
	  for(j=1;j<autoboxHeight-1;j++)
	    {
	      for(i=1;i<autoboxWidth-1;i++)
		Accum[j*autoboxWidth+i] = auxAccum[j*autoboxWidth+i];
	    }
	  
	}/* Thinning layers */


      /* *******************************************************
       * Now do anything appropriate for empty particles       *
       * *******************************************************/
      if(emptyBox && FALSE)
	{/* Do for empties */
	  int m,n,totHits;
	  sprintf(outStr,"Looking for empty particles\n");
	  XmTextSetString(timeLeftBoxW,outStr);
	  XFlush(dsply);
	  for(j=1;j<autoboxHeight-tbSize;j++)
	    {/* y */
	      for(i=1;i<autoboxWidth-tbSize;i++)
		{/* x */
		  totHits = 0;
		  for(m=0;m<tbSize;m++)
		    {/* height */
		      for(n=0;n<tbSize;n++)
			{/* width */
			  if(empty_ref_buffer[((m*tbSize)+n)] == 
			     Accum[(((j+m)*autoboxWidth)+i+n)])
			    totHits++;
			}/* width */
		    }/* height */
		  if(totHits > ((tbSize*tbSize)*0.75))
		    {/* Fill it up */
		      sprintf(outStr,"Found empty particle at x[%d]y[%d]\n",
			      i,j);
		      XmTextSetString(timeLeftBoxW,outStr);
		      XFlush(dsply);
		      for(m=0;m<tbSize;m++)
			{/* height */
			  for(n=0;n<tbSize;n++)
			    {/* width */
			      if(empty_full_buffer[((m*tbSize)+n)]==1) 
				auxAccum[(((j+m)*autoboxWidth)+i+n)] = 2;
			    }/* width */
			}/* height */
		      /* Advance beyond */
		      i += tbSize;
		    }/* Fill it up */
		  else
		    {/* Just do normal */
		      if(auxAccum[j*autoboxWidth+i] < 2)
			auxAccum[j*autoboxWidth+i] = Accum[j*autoboxWidth+i];
		    }/* Just do normal */
		}/* x */
	    }/* y */
      
	  /* Now take the model enhanced data as our gold standard */
	  for(j=1;j<autoboxHeight-1;j++)
	    {
	      for(i=1;i<autoboxWidth-1;i++)
		if(auxAccum[j*autoboxWidth+i] == 2)
		  Accum[j*autoboxWidth+i] = 1;
		else
		  Accum[j*autoboxWidth+i] = auxAccum[j*autoboxWidth+i];
	    }
	  
	  free(empty_ref_buffer);
	  free(empty_full_buffer);
	}/* Do for empties */

      free(auxAccum);
      
      /* Show the clusters */
      if(TRUE)
	{/* Setup for Display the clustered image */
	  int cl;
	  for(cl=0;cl<autoboxArea;cl++)
	    globAccum[cl] = Accum[cl];
	}/* Setup for Display the clustered image */


      /******** Start finding particle by Stack algorithm ********/
      XmTextSetString(timeLeftBoxW,"Clustering");
      XFlush(dsply);
      if((cluster = (XPoint *)malloc(autoboxArea*sizeof(XPoint))) ==
	 (XPoint *)NULL)
	{
	  printf("\n cannot allocate cluster");
	  XmTextSetString(timeLeftBoxW,
			  "Error allocating buffers for Clustering.");
	  XFlush(dsply);
	  return;
	}
      
      if((stack = (XPoint *)malloc(autoboxArea*sizeof(XPoint))) ==
	 (XPoint *)NULL)
	{
	  printf("\n cannot allocate stack");
	  XmTextSetString(timeLeftBoxW,
			  "Error allocating buffers for Clustering.");
	  XFlush(dsply);
	  return;
	}
      
      clusterSizeUplimit = 2.0*3.14159*radi*radi - 4*3.14159*radi;
      clusterSizeDownlimit = 0.6*3.14159*radi*radi - 4*3.14159*radi;
      
      for(j=0;j<autoboxHeight;j++)
	{/* height */
	  for(i=0;i<autoboxWidth;i++)
	    {/* width */
	      if(Accum[j*autoboxWidth+i])
		{/* accum test passed */
		  sp = clusterSize = 0;
		  
		  /*---------- addPoint ------------------*/
		  cluster[clusterSize].x = i;
		  cluster[clusterSize].y = j;
		  clusterSize++;
		  if(i>0 && j>0)
		    {
		      if(Accum[(j-1)*autoboxWidth+i-1] == 1)
			{
			  stack[sp].x = i-1; stack[sp].y = j-1; sp++;
			  Accum[(j-1)*autoboxWidth+i-1] = 2;
			}
		    }
		  if(i>0)
		    {
		      if(Accum[j*autoboxWidth+i-1] == 1)
			{
			  stack[sp].x = i-1; stack[sp].y = j; sp++;
			  Accum[j*autoboxWidth+i-1] = 2;
			}
		    }
		  if(i>0 && j<autoboxHeight-1)
		    {
		      if(Accum[(j+1)*autoboxWidth+i-1] == 1)
			{
			  stack[sp].x = i-1; stack[sp].y = j+1; sp++;
			  Accum[(j+1)*autoboxWidth+i-1] = 2;
			}
		    }
		  if(j>0)
		    {
		      if(Accum[(j-1)*autoboxWidth+i] == 1)
			{
			  stack[sp].x = i; stack[sp].y = j-1; sp++;
			  Accum[(j-1)*autoboxWidth+i] = 2;
			}
		    }
		  if(j<autoboxHeight-1)
		    {
		      if(Accum[(j+1)*autoboxWidth+i] == 1)
			{
			  stack[sp].x = i; stack[sp].y = j+1; sp++;
			  Accum[(j+1)*autoboxWidth+i] = 2;
			}
		    }
		  if(i<autoboxWidth-1 && j>0)
		    {
		      if(Accum[(j-1)*autoboxWidth+i+1] == 1)
			{
			  stack[sp].x = i+1; stack[sp].y = j-1; sp++;
			  Accum[(j-1)*autoboxWidth+i+1] = 2;
			}
		    }
		  if(i<autoboxWidth-1)
		    {
		      if(Accum[j*autoboxWidth+i+1] == 1)
			{
			  stack[sp].x = i+1; stack[sp].y = j; sp++;
			  Accum[j*autoboxWidth+i+1] = 2;
			}
		    }
		  if(i<autoboxWidth-1 && j<autoboxHeight-1)
		    {
		      if(Accum[(j+1)*autoboxWidth+i+1] == 1)
			{
			  stack[sp].x = i+1; stack[sp].y = j+1; sp++;
			  Accum[(j+1)*autoboxWidth+i+1] = 2;
			}
		    }
		  Accum[j*autoboxWidth+i] = 0;
		  
		  while(sp>0)
		    {/* sp > 0 */
		      
		      /*--------- removeStackTop ----------------*/
		      ti = stack[sp-1].x; tj = stack[sp-1].y; sp --;
		      
		      /*---------- addPoint ------------------*/
		      cluster[clusterSize].x = ti;
		      cluster[clusterSize].y = tj;
		      clusterSize++;
		      if(clusterSize >= autoboxArea)
			{
			  printf("\ncluster too big");
			  free(stack);
			  free(cluster);
			  free(Accum);
			  free(autoBox_buffer);
			  XmTextSetString(timeLeftBoxW,
					  "Error cluster too big.");
			  XFlush(dsply);
			  return;
			}
		      if(sp >= autoboxArea)
			{
			  printf("\nstack too big");
			  free(stack);
			  free(cluster);
			  free(Accum);
			  free(autoBox_buffer);
			  XmTextSetString(timeLeftBoxW,
					  "Error cluster too big.");
			  XFlush(dsply);
			  return;
			}
		      if(ti>0 && tj>0)
			{
			  if(Accum[(tj-1)*autoboxWidth+ti-1] == 1)
			    {
			      stack[sp].x = ti-1; stack[sp].y = tj-1; sp++;
			      Accum[(tj-1)*autoboxWidth+ti-1] = 2;
			    }
			}
		      if(ti>0)
			{
			  if(Accum[tj*autoboxWidth+ti-1] == 1)
			    {
			      stack[sp].x = ti-1; stack[sp].y = tj; sp++;
			      Accum[tj*autoboxWidth+ti-1] = 2;
			    }
			}
		      if(ti>0 && tj<autoboxHeight-1)
			{
			  if(Accum[(tj+1)*autoboxWidth+ti-1] == 1)
			    {
			      stack[sp].x = ti-1; stack[sp].y = tj+1; sp++;
			      Accum[(tj+1)*autoboxWidth+ti-1] = 2;
			    }
			}
		      if(tj>0)
			{
			  if(Accum[(tj-1)*autoboxWidth+ti] == 1)
			    {
			      stack[sp].x = ti; stack[sp].y = tj-1; sp++;
			      Accum[(tj-1)*autoboxWidth+ti] = 2;
			    }
			}
		      if(tj<autoboxHeight-1)
			{
			  if(Accum[(tj+1)*autoboxWidth+ti] == 1)
			    {
			      stack[sp].x = ti; stack[sp].y = tj+1; sp++;
			      Accum[(tj+1)*autoboxWidth+ti] = 2;
			    }
			}
		      if(ti<autoboxWidth-1 && tj>0)
			{
			  if(Accum[(tj-1)*autoboxWidth+ti+1] == 1)
			    {
			      stack[sp].x = ti+1; stack[sp].y = tj-1; sp++;
			      Accum[(tj-1)*autoboxWidth+ti+1] = 2;
			    }
			}
		      if(ti<autoboxWidth-1)
			{
			  if(Accum[tj*autoboxWidth+ti+1] == 1)
			    {
			      stack[sp].x = ti+1; stack[sp].y = tj; sp++;
			      Accum[tj*autoboxWidth+ti+1] = 2;
			    }
			}
		      if(ti<autoboxWidth-1 && tj<autoboxHeight-1)
			{
			  if(Accum[(tj+1)*autoboxWidth+ti+1] == 1)
			    {
			      stack[sp].x = ti+1; stack[sp].y = tj+1; sp++;
			      Accum[(tj+1)*autoboxWidth+ti+1] = 2;
			    }
			}
		      Accum[tj*autoboxWidth+ti] = 0;
		    }/* sp > 0 */
		  

		  if(clusterSize > clusterSizeDownlimit &&
		     clusterSize < clusterSizeUplimit )
		    {/* clusterSize in limits */
		      avgx = avgy = 0;
		      for(k=0;k<clusterSize;k++)
			{
			  avgx += cluster[k].x;
			  avgy += cluster[k].y;
			}
		      avgx /= clusterSize; avgy /= clusterSize;

		      startJ = avgy - radi;
		      endJ   = avgy + radi;
		      startI = avgx - radi;
		      endI   = avgx + radi;
		      if(startJ > radi && endJ <(autoboxHeight-radi) &&
			 startI > radi && endI <(autoboxWidth-radi) )
			{/* OK to do this one */
			  /*---------- correlParticles -------------*/
			  sum2 = (autoBox_buffer[(avgy-radi-1)*
						 autoboxWidth+avgx]+
				  autoBox_buffer[(avgy+radi+1)*
						 autoboxWidth+avgx]+
				  autoBox_buffer[(avgy-radi-1)*
						 autoboxWidth+avgx-
						 radi-1]+
				  autoBox_buffer[(avgy+radi+1)*
						 autoboxWidth+avgx-
						 radi-1]+
				  autoBox_buffer[(avgy-radi-1)*
						 autoboxWidth+avgx+
						 radi+1]+
				  autoBox_buffer[(avgy+radi+1)*
						 autoboxWidth+avgx+
						 radi+1]+
				  autoBox_buffer[avgy*autoboxWidth+
						 avgx-radi-1]+
				  autoBox_buffer[avgy*autoboxWidth+
						 avgx+radi+1])/8.;
			  
			  sum1 = 0.;
			  totSum1 = 0;
			  
			  startJ = MAX(startJ,radi);
			  endJ   = MIN(endJ,(autoboxHeight-radi));
			  startI = MAX(startI,radi);
			  endI   = MIN(endI,(autoboxWidth-radi));
			  for(jj=startJ;jj<endJ;jj++)
			    {
			      for(ii=startI;ii<endI;ii++)
				{
				  if((ii-avgx)*(ii-avgx)+
				     (jj-avgy)*(jj-avgy) 
				     < radi*radi)
				    {
				      sum1 += autoBox_buffer[jj*
							    autoboxWidth+ii];
				      totSum1++;
				    }
				}
			    }
			}/* OK to do this one */
		      else 
			{/* Make it a lousy correlation */
			  sum1 = 0.0;  
			  sum2 = 1.0;
			}/* Make it a lousy correlation */
		      
		      corr = fabs(sum1/(3.14*radi*radi)-sum2);
		      
		      if(corr >= 2.5 && avgy < autoboxHeight-radi &&
			 avgx < autoboxWidth-radi)
			{
			  if(numBoxes < MAXBOX)
			    {
			      box[numBoxes].x = avgx * compress_factor;
			      box[numBoxes].y = avgy * compress_factor;
			      box[numBoxes].reference = FALSE;
			      box[numBoxes].correlation = corr;
			      box[numBoxes].avgdensity = 
				    ((sum1/(float)totSum1)/256.0);
			      box[numBoxes].avgdensity = 
				    (float)(sum1/(float)totSum1);
			      numBoxes ++;
			    }
			  else
			    {/* Overflow */
			      printf(
	       "\n[AUTOBOX_ACTIVATE] Too many boxes [%d][%d]; box ignored.",
				     numBoxes,MAXBOX);
			    }/* Overflow */
			}
		      
		    }/* clusterSize in limits */

		}/* accum test passed */
	    }/* width */
	}/* height */
      
      free(stack);
      free(cluster);
      free(Accum);
      free(autoBox_buffer);
      
      if(*tag==CORRELATE)
	{/* DO the reference correlation automatically */

	  for(i=0; i<numBoxes; i++)
	    box[i].reference = TRUE;
	  
	  ref_correlate_activate();

	  /* Remember for next time */
	  for(i=0; i<numBoxes; i++)
	    boxRefLast[i] = box[i].reference ;
	  
	  sprintf(outStr,"Done.");
	  createSimpleString(&simpleString,outStr);
	  XtSetArg(arglst[0], XmNlabelString, simpleString );
	  XtSetArg(arglst[1], XmNrecomputeSize, FALSE );
	  XtSetValues(autoCorrLabelW, arglst, 2); 
	  XFlush(dsply);
	  XmStringFree(simpleString); 
	  
	}/* DO the reference correlation automatically */

      DrawWindow();
      
      printf("\nFinish automatic boxing\n");
      sprintf(outStr,"Done.  [%d] boxes.",numBoxes);
      XmTextSetString(timeLeftBoxW,outStr);
      XFlush(dsply);

      sprintf(outStr,"Done.");
      createSimpleString(&simpleString,outStr);
      XtSetArg(arglst[0], XmNlabelString, simpleString );
      XtSetArg(arglst[1], XmNrecomputeSize, FALSE );
      XtSetValues(autoBoxLabelW, arglst, 2); 
      XFlush(dsply);
      XmStringFree(simpleString); 
  
    }/* Autobox [1] (EMMA method) */
  
  

  /************************************************************
   ******************** Autobox [2] (New method) **************
   ************************************************************/
  else if(*tag==AUTOBOX2)
  { /* Autobox [2] (New method) */
    printf("\nautoBox_activate called [2].\n");

    /* Load in current radius */
     thisRadius = emmaRadius;

    radi = thisRadius;
    autoboxHeight = file_height;
    autoboxWidth = file_width;
    autoboxArea = autoboxHeight*autoboxWidth;

    /* Open the disk file for load in raw image */
    fIn= NULL;
    fIn= fopen(image_fileName.addr,"r");
    if(fIn==NULL)
    {/* Error */
      printf(" Error open image file for refine center.\n");
      XmTextSetString(timeLeftBoxW,
		      "Error opening image.");
      XFlush(dsply);
      return;
    }/* Error */

    /***** allocate space for image and temp byte buffer *****/
    if((autoBox_buffer_raw=
	(float *)malloc(autoboxArea*sizeof(float)))==(float *)NULL)
    {
      printf("\nError: Allocate particle image for autoboxing");
      XmTextSetString(timeLeftBoxW,
		      "Error allocating particle buffers.");
      XFlush(dsply);
      return;
    }
    if((autoBox_buffer = (float *)calloc(autoboxArea,
                sizeof(float))) == (float *)NULL)
    {
      printf("\nError: Allocate particle image for autoboxing");
      free(autoBox_buffer_raw);
      XmTextSetString(timeLeftBoxW,
		      "Error allocating particle buffers.");
      XFlush(dsply);
      return;
    }
    if((temp_byte_buffer=
	(float *)malloc(autoboxWidth*sizeof(float)))==(float *)NULL)
    {
      printf("\nError: Allocate particle image for autoboxing");
      free(autoBox_buffer_raw);
      free(autoBox_buffer);
      XmTextSetString(timeLeftBoxW,
		      "Error allocating particle buffers.");
      XFlush(dsply);
      return;
    }

    /****** Image compression regarding to radius ***/


    /* Let's scale this */
    calcStats(inCore_buffer, &img_bkg, &minF,
	      &maxF, &img_avg, &img_std,
	      file_width, file_height, file_width);
    scaleF = 255.0/(maxF-minF);
    
    /* Load in raw image */
    for(j=0; j<autoboxHeight; j++)
    {
	  if(!imgInCore)
	    {/* Get from disk */
	      yOff = headOff + j*autoboxWidth;
	      sysError = fseek(fIn,yOff,0);
	      if(sysError)
		{
		  printf("Error: fseek in auto box.\n");
		  free(autoBox_buffer_raw);
		  free(autoBox_buffer);
		  free(temp_byte_buffer);
		  XmTextSetString(timeLeftBoxW,
				  "fseek() error in autobox.");
		  XFlush(dsply);
		  return;
		}
	      fread(temp_byte_buffer, sizeof(BYTE), (autoboxWidth), fIn);
	    }/* Get from disk */
	  else
	    {/* Get from core */
	      yOff = j*autoboxWidth;
	      memcpy(temp_byte_buffer,&inCore_buffer[(yOff)],
		     (autoboxWidth*sizeof(float)));
	    }/* Get from core */

      /* Stuff the autoBox_buffer_raw */
      yOut = j*autoboxWidth;
      for(k=0; k<autoboxWidth; k++)
        autoBox_buffer_raw[yOut+k]= ((temp_byte_buffer[k]-minF)*scaleF)/2.0;
    }

    fclose(fIn);
    free(temp_byte_buffer);

    /*********** Histogram equalization ***********/
    dataRange = 128;
    if((prob_p=(float*)malloc(dataRange*sizeof(float)))==(float*)NULL ||
        (sum_p=(float*)malloc(dataRange*sizeof(float)))==(float*)NULL)
    {
      printf("\nError: Allocate buffers for Histogram equalization\n");
      XmTextSetString(timeLeftBoxW,
		      "Error allocating prob_p buffers.");
      XFlush(dsply);
      return;
    }

    for(i=0;i<dataRange;i++)
      prob_p[i] = 0.;

    for(k=0;k<autoboxHeight;k++)
      for(j=0;j<autoboxWidth;j++)
         prob_p[(int)autoBox_buffer_raw[k*autoboxWidth+j]]++;

    for(i=0;i<dataRange;i++)
      prob_p[i] = (float)prob_p[i]/(float)(autoboxArea);

    for(i=0;i<dataRange;i++)
    {
      sum_p[i] = 0.;
      for(j=0;j<=i;j++)
        sum_p[i] += prob_p[j];
    }
    for(k=0;k<autoboxHeight;k++)
    {
      kj = k*autoboxWidth;
      for(j=0;j<autoboxWidth;j++)
      {
        level = (unsigned short)(0.5+(dataRange - 1)*
                sum_p[(int)autoBox_buffer_raw[kj+j]]);
        if(level >= dataRange)
          level = dataRange - 1;
        autoBox_buffer_raw[kj+j] = level;
      }
    }

    free(autoBox_buffer_raw);
    free(prob_p);
    free(sum_p);

    /*********** local background mapping *************/



    /*********** frequency density mapping ************/



    /*********** circling, skip blank area ************/


    free(autoBox_buffer);

    DrawWindow();

    printf("\nFinish automatic boxing\n");
  } /* Autobox [2] (New method) */

}/* AUTOBOX_ACTIVATE */



/****************************************************************
 * correlate_activate() - Auto-refine particle centers .
 */
void correlate_activate(Widget widget, int *tag, 
			XmAnyCallbackStruct *callback_data)
{/* CORRELATE_ACTIVATE */
  float *refineBox_buffer;
  float *refBox_byte_buffer;
  int img_bkg, total_bkg;
  int i, j, k, cx, cy, dx, dy;
  int arraySize, Radius_old, outHeight, outWidth;
  int yOut, beginY, beginX;
  int outsideTotal, outsideNo, insideTotal, insideNo;
  float outsideAvg, insideAvg, dividedAvg;
  FILE *fIn;
  XmString simpleString;

  firstAMMSTDEVtime = TRUE;
  if(*tag==CLEARBOXES)
    {/* CLear all boxes */
      for(i=0;i<numBoxes;i++)
	if(box[i].inMemory==TRUE)
	  {/* Free it */
	    free(box[i].editBox_buffer);
	    box[i].editBox_buffer= NULL;
	    box[i].inMemory= FALSE;
	  }/* Free it */
      numBoxes= numRects= numPoly= 0;

      /* Change cursor back to what it was */
      attrs.cursor= old_cursor;
      XChangeWindowAttributes(dsply, XtWindow(drawingWidget), 
			      CWCursor, &attrs);
      XFlush(dsply);
      
      sprintf(outStr,"     ");
      createSimpleString(&simpleString,outStr);
      XtSetArg(arglst[0], XmNlabelString, simpleString );
      XtSetArg(arglst[1], XmNrecomputeSize, FALSE );
      XtSetValues(autoBoxLabelW, arglst, 2); 
      XtSetValues(refCorrLabelW, arglst, 2); 
      XtSetValues(autoCorrLabelW, arglst, 2); 
      XFlush(dsply);
      XmStringFree(simpleString); 
  
      DrawWindow();
    }/* CLear all boxes */
  /******************* Auto_correlation method *******/
  else if(*tag==CORRELATE)
    {/* correlate [1] */

      /* This may take awhile so give them the watch cursor */
      attrs.cursor= watch_cursor;
      XChangeWindowAttributes(dsply, XtWindow(drawingWidget), 
			      CWCursor, &attrs);
      XFlush(dsply);
      
      printf("correlate_activate()[1] called.\n");
      XmTextSetString(timeLeftBoxW,"Auto-correlation refinement.");
      XFlush(dsply);
      
      /* Reset reference particles Marker*/
      for(i=0; i<numBoxes; i++)
	box[i].reference = FALSE;
      
      /* Load in radius and enlarge it 25% */
      if(fileType==PACKEDPORTABLE || fileType==PORTABLE )
	{
	  askThem("Sorry, Only box images can be refined.",
		 FALSE,
		 XmDIALOG_INFORMATION);
	  goto outCorr1;
	  /* return; */
	}
      else
	{
	  Radius_old = boxRadius;
	  boxRadius = (int) (Radius_old * 1.25);
	}
      
      outHeight = outWidth = (boxRadius*2)+1;
      arraySize = outHeight * outWidth;
      
      /* allocate space for the boxed particle image */
      if((refineBox_buffer = (float *)calloc(arraySize,
					     sizeof(float))) == (float *)NULL)
	{
	  printf("\nError: Allocate particle image for refinement");
	  XmTextSetString(timeLeftBoxW,
			  "Error allocating buffers for refine centers.");
	  XFlush(dsply);
	  goto outCorr1;
	  /* return; */
	}
      if((refBox_byte_buffer = (float *)calloc((BOX_WIDTH*2),
					     sizeof(float))) == (float *)NULL)
	{
	  printf("\nError: Allocate particle image for refinement[refBox]");
	  XmTextSetString(timeLeftBoxW,
			  "Error allocating buffers for refine centers.");
	  XFlush(dsply);
	  free(refineBox_buffer);
	  goto outCorr1;
	  /* return; */
	}
      
      /* Open the disk file for load in boxed image */
      fIn= NULL;
      fIn= fopen(image_fileName.addr,"r");
      if(fIn==NULL)
	{/* Error */
	  printf(" Error reading image file for refine center.\n");
	  free(refineBox_buffer);
	  free(refBox_byte_buffer);
	  boxRadius = Radius_old; 
	  XmTextSetString(timeLeftBoxW,
			  "Error opening image for refine centers.");
	  XFlush(dsply);
	  goto outCorr1;
	  /* return; */
	}/* Error */
      
      /*****  Refine the center for each box *****/
      for(i=0;i<numBoxes;i++)
	{/* Do all boxes */
	  /* Give feedback */
	  sprintf(outStr,"Boxes left - [%d]",(numBoxes-i));
	  XmTextSetString(timeLeftBoxW,outStr);
	  XFlush(dsply);
	  /* CHeck if they want to abort or not etc... */
	  if(XCheckTypedWindowEvent(dsply,
				    XtWindow(boxStopWidget),
				    ButtonPress,
				    &event))
	    {/* STOP!!! */
	      sprintf(outStr,"Operation aborted by user.");
	      XmTextSetString(timeLeftBoxW,outStr);
	      XFlush(dsply);
	      free(refineBox_buffer);
	      free(refBox_byte_buffer);
	      goto outCorr1;
	    }/* STOP!!! */

	  /* Load in particle image for refinement */
	  beginY = box[i].y - boxRadius;
	  beginX = box[i].x - boxRadius;
	  for(j=0; j<outHeight; j++)
	    {
	      if(!imgInCore)
		{/* Get from disk */
		  yOff = headOff + (beginY + j)*file_width + beginX;
		  sysError = fseek(fIn,yOff,0);
		  if(sysError)
		    {
		      printf("Error: fseek in refine center.\n");
		      free(refineBox_buffer);
		      boxRadius = Radius_old; 
		      XmTextSetString(timeLeftBoxW,
				      "Error fseek() in refine centers.");
		      XFlush(dsply);
		      free(refBox_byte_buffer);
		      goto outCorr1;
		      /* return; */
		    }
		  fread(refBox_byte_buffer, sizeof(BYTE),
			(outWidth),fIn);
		}/* Get from disk */
	      else
		{/* Get from core */
		  yOff = (beginY + j)*file_width + beginX;
		  memcpy(refBox_byte_buffer,&inCore_buffer[(yOff)],
			 (outWidth*sizeof(float)));
		}/* Get from core */
	      
	      /* Stuff the refineBox_buffer */
	      yOut = j*outWidth;
	      for(k=0; k<outWidth; k++)
		refineBox_buffer[yOut+k]= (short)refBox_byte_buffer[k];
	    }
	  
	  /* compute image background for each particle*/
	  total_bkg = 0;
	  
	  for(j=0; j<outWidth; j++)
	    total_bkg += (int)refineBox_buffer[j];       /*top row*/
	  for(j=arraySize-outWidth; j<arraySize; j++)
	    total_bkg += (int)refineBox_buffer[j];       /*bottom row */
	  for(j=outWidth; j<arraySize-outWidth; j+=outWidth)
	    total_bkg += (int)refineBox_buffer[j];       /*left column*/
	  for(j=2*outWidth-1; j<arraySize-outWidth; j+=outWidth)
	    total_bkg += (int)refineBox_buffer[j];       /*right column*/
	  
	  img_bkg = (int) total_bkg/(2*outHeight+ 2*outWidth - 4);
	  
	  ccf_refine_(refineBox_buffer, &img_bkg, &outWidth,
		      &outHeight, &dx, &dy);
	  
	  /* reset boxing center to new value */
	  box[i].x +=  dx;
	  box[i].y +=  dy;
	  
	  /***********  Remove bad particle ***********/
	  /* Load in refined particle image */
	  beginY = box[i].y - boxRadius;
	  beginX = box[i].x - boxRadius;
	  for(j=0; j<outHeight; j++)
	    {
	      if(!imgInCore)
		{/* Get from disk */
		  yOff = headOff + (beginY + j)*file_width + beginX;
		  sysError = fseek(fIn,yOff,0);
		  if(sysError)
		    {
		      printf("Error: fseek in refine center.\n");
		      free(refineBox_buffer);
		      boxRadius = Radius_old;
		      XmTextSetString(timeLeftBoxW,
				      "Error fseek() in refine centers.");
		      XFlush(dsply);
		      free(refBox_byte_buffer);
		      goto outCorr1;
		      /* return; */
		    }
		  fread(tempBox_byte_buffer, sizeof(float),
			(outWidth),fIn);
		}/* Get from disk */
	      else
		{/* Get from core */
		  yOff = (beginY + j)*file_width + beginX;
		  memcpy(tempBox_byte_buffer,&inCore_buffer[(yOff)],
			 (outWidth*sizeof(float)));
		}/* Get from core */
	      
	      /* Stuff the refineBox_buffer */
	      yOut = j*outWidth;
	      for(k=0; k<outWidth; k++)
		refineBox_buffer[yOut+k]= (float)tempBox_byte_buffer[k];
	    }
	  
	  /* compute image background for each particle*/
	  total_bkg = 0;
	  
	  for(j=0; j<outWidth; j++)
	    total_bkg += (int)refineBox_buffer[j];       /*top row*/
	  for(j=arraySize-outWidth; j<arraySize; j++)
	    total_bkg += (int)refineBox_buffer[j];       /*bottom row */
	  for(j=outWidth; j<arraySize-outWidth; j+=outWidth)
	    total_bkg += (int)refineBox_buffer[j];       /*left column*/
	  for(j=2*outWidth-1; j<arraySize-outWidth; j+=outWidth)
	    total_bkg += (int)refineBox_buffer[j];       /*right column*/
	  
	  img_bkg = (int) total_bkg/(2*outHeight+ 2*outWidth - 4);
	  
	  /* Mark bad particles by comparing local background */
	  outsideTotal = 0;
	  outsideNo = 0;
	  insideTotal = 0;
	  insideNo = 0;
	  cx = cy = boxRadius; 
	  for(j=0;j<outHeight;j++)
	    {
	      for(k=0;k<outWidth;k++)
		{
		  if(((j-cy)*(j-cy)+(k-cx)*(k-cx))>Radius_old*Radius_old*1.1)
		    {
		      outsideNo ++;
		      if((refineBox_buffer[j*outWidth+k]) < img_bkg)
			outsideTotal ++;
		    }
		  if(((j-cy)*(j-cy)+(k-cx)*(k-cx))<Radius_old*Radius_old)
		    {
		      insideNo ++;
		      if((refineBox_buffer[j*outWidth+k]) < img_bkg)
			insideTotal ++;
		    }
		}
	    }
	  outsideAvg = (float)outsideTotal/(float)outsideNo;
	  insideAvg = (float)insideTotal/(float)insideNo;
	  dividedAvg = insideAvg / outsideAvg;
	  if( dividedAvg <=1.15 )
            box[i].reference = -1;
      
	  box[i].correlation = dividedAvg;
	  box[i].avgdensity = insideAvg/256.0;
	  box[i].avgdensity = (float)insideAvg;

	  if(refineDisp)
	    {/* Give feedback as to which circle we're working on */
	      int arcRadius;
	      /* Size of the circle */
	      calcSizeFactor(zoomFactor);
	      /* Which is best??? */
	      arcRadius = (int)(sizeFactor * Radius_old);
	      arcRadius = (int)(sizeFactor * boxRadius);
	      rwaXDrawArc(dsply, XtWindow(drawingWidget), image_gcBLUE,
			  (((box[i].x*sizeFactor)-(virXoffset*sizeFactor))-
			   arcRadius), 
			  (((box[i].y*sizeFactor)-(virYoffset*sizeFactor))-
			   arcRadius),
			  (arcRadius*2),(arcRadius*2),
			  0, (360*64));
	    }/* Give feedback as to which circle we're working on */
	  
	}/* Do all boxes */


      /* Remove bad particle */
      i = 0; j = 0;
      do
	{
	  if(box[i].reference == -1)
	    {
	      /* Reset reference particles Marker*/
	      box[i].reference = FALSE;
	      i ++;
	      numBoxes --;
	    }
	  else
	    {
	      box[j].x = box[i].x;
	      box[j].y = box[i].y;
	      /* Reset reference particles Marker*/
	      box[j].reference = FALSE;
	      i ++; j ++;
	    }
	}
      while (j<numBoxes);
      
      /* Recover the radius */
      boxRadius = Radius_old;
      
      fclose(fIn);
      free(refBox_byte_buffer);
      free(refineBox_buffer);
      
    outCorr1:;
      
      sprintf(outStr,"Done.  [%d] boxes.",numBoxes);
      XmTextSetString(timeLeftBoxW,outStr);
      XFlush(dsply);

      sprintf(outStr,"Done.");
      createSimpleString(&simpleString,outStr);
      XtSetArg(arglst[0], XmNlabelString, simpleString );
      XtSetArg(arglst[1], XmNrecomputeSize, FALSE );
      XtSetValues(autoCorrLabelW, arglst, 2); 
      XFlush(dsply);
      XmStringFree(simpleString); 

      /* Change cursor back to what it was */
      attrs.cursor= old_cursor;
      XChangeWindowAttributes(dsply, XtWindow(drawingWidget), 
			      CWCursor, &attrs);
      XFlush(dsply);
      
      for(i=0; i<numBoxes; i++)
	box[i].reference = FALSE;
      DrawWindow();
      
    }/* correlate [1] */
  
  
  /**************  Refence based correlation ***********/
  else if(*tag==CORRELATE2 || *tag==CORRELATE3)
    {/* corr [2] */
      printf("correlate_activate()[2] called.\n");
      XmTextSetString(timeLeftBoxW,"Reference based refinement.");
      XFlush(dsply);
      
      if(*tag==CORRELATE2)
	{/* Make sure get some reference particles */
	  ref_box_flag = TRUE;
	  for(i=0;i<numBoxes;i++)
	    box[i].reference = FALSE;  /* Reset reference particles Marker*/
	  XtManageChild(refscreen);
	  sprintf(outStr,"\n\
Mouse modes:\n\
   Left - Select this particle for use in reference\n\
          OR Make a new particle at this lcoation and use it\n\
             for the reference.\n\
   Middle - Remove this particle altogether\n\
\n\
Use the Middle button to remove any particle that doesn't have an\n\
acceptable center and replace it using the Left button.\n\
\n\
When finished selecting press the 'DoIt' button.\n");
	  XmTextSetString(refTextWidget,outStr);
	  /* When "doit" button is pushed, the routine 
	     ref_correlate_activate() will be called */ 

	  /* Manage proper children */
	  XtManageChild(refCircAvgW);
	  XtManageChild(refAutomateW);
	  XtManageChild(refReadfileW);
	  XtManageChild(refReadFileLabelW);
	  XtManageChild(darkRefToggleW);
	  XtManageChild(refAllW);
	  XtManageChild(refResetW);

	  XtUnmanageChild(refRegridW);

	  DrawWindow();
	}/* Make sure get some reference particles */
      else if(*tag==CORRELATE3)
	{/* empty particle boxing */
	  ref_box_flag = TRUE;
	  for(i=0;i<numBoxes;i++)
	    box[i].reference = FALSE;  /* Reset reference particles Marker*/
	  XtManageChild(refscreen);
	  sprintf(outStr,"\n\
Empty Particle boxing:\n\
This routine tries to refine a grid of boxes using reference matching.\n\
The reference is built based on the Emma Radius and Empty Shell Width parameters.\n\
\n\
Choose how many iterations you'd like to try and press the DoIt button.\n\
Use the Stop Sign button on the 3D Rendering display to cancel this\n\
algorithm at any time.\n\
\n\
REGRID button - use this to start over with a new grid of boxes.\n\
\n");
	  XmTextSetString(refTextWidget,outStr);
	  /* When "doit" button is pushed, the routine 
	     ref_correlate_activate() will be called */ 

	  refAutomate = TRUE;
	  darkCircleRef =  99;

	  XtSetArg(arglst[0], XmNset, TRUE);
	  XtSetValues(darkRefToggleW, arglst, 1);
	  XtSetArg(arglst[0], XmNset, TRUE);
	  XtSetValues(refAutomateW, arglst, 1);


	  /* Manage proper children */
	  XtUnmanageChild(refCircAvgW);
	  XtUnmanageChild(refAutomateW);
	  XtUnmanageChild(refReadfileW);
	  XtUnmanageChild(refReadFileLabelW);
	  XtUnmanageChild(darkRefToggleW);
	  XtUnmanageChild(refAllW);
	  XtUnmanageChild(refResetW);

	  XtManageChild(refRegridW);


	  DrawWindow();
	}/* empty particle boxing */
      else
	{/* Just go for it */
	  ref_correlate_activate();
	}/* Just go for it */
    }/* corr [2] */
  
}/* CORRELATE_ACTIVATE */




/**************************************************************
 * ref_correlate_activate() - Refine particle centers by
 * reference.
 */
void ref_correlate_activate()
{/* REF_CORRELATE_ACTIVATE */
  float *refineBox_buffer;
  float *ref_refineBox_buffer;
  float *ref_refineBox_buffer_short;
  float *refBox_byte_buffer;
  float *ref_refineBox_fft;
  float *tmp_refineBox_buffer;
  float  maxP,minP;
  int totDx,totDy,numChanged,totalBugOut;
  float varDx,varDy,sumsqDx,sumsqDy,sumDx,sumDy;
  float thisScaleFac,meanDx,meanDy,stdDx,stdDy;
  float img_bkg, total_bkg, ref_img_bkg;
  int cx, cy;
  int i, j, k, dx, dy, fft_dim1, fft_dim2, fft_nfac, iter, lastIter;
  int arraySize, Radius_old, outHeight, outWidth;
  int yOut, beginY, beginX, saveNumBoxes;
  int numBoxes_ref, maxborder;
  int outsideTotal, outsideNo, insideTotal, insideNo;
  float outsideAvg, insideAvg, dividedAvg;
  FILE *fIn;
  XmString simpleString;


  printf ("DBUG ra[%d]  dcr[%d]\n",
	  refAutomate ,
	  darkCircleRef);



  /* This may take awhile so give them the watch cursor */
  attrs.cursor= watch_cursor;
  XChangeWindowAttributes(dsply, XtWindow(drawingWidget), 
			  CWCursor, &attrs);
  XFlush(dsply);

  /* Load in radius and enlarge it 25% */
  if(fileType==PACKEDPORTABLE)
    {
      askThem("Sorry, Only BYTE box scanned images can be refined.",
	      FALSE,
	      XmDIALOG_INFORMATION);
      return;
    }

  /* Show them what's happening */
  /* DONT DO IT !!! raiseIt(statusScreenWidget);
   * XFlush(dsply); */

  totalBugOut = FALSE;
  numChanged = 99999999;
  numChanged = INT_MAX;  /* dgm INT_MAX instead of 9999999 */
  lastIter = 1;  /* Default */
  if(refAutomate)
    {/* Try and do this a number of times */
      char *thisStr;
      thisStr = XmTextGetString(refLastIterTextW);
      lastIter = (int)strtod(thisStr,0);
      XtFree(thisStr);
      sprintf(outStr,"%d",lastIter);
      XmTextSetString(refLastIterTextW,outStr);
      XFlush(dsply);
    }/* Try and do this a number of times */

  if(autoBoxSwitch)
    lastIter = autoBoxIter;

  /* Init these so we can look at changes per cycle */
  for(i=0; i<numBoxes; i++)
    box[i].nochange = 0;

  for(iter=0;iter<lastIter;iter++)
    {/* Do iter times or until they opt out */
printf("Iter = [%d][%d][%d][%d][%d]\n",iter,lastIter, emmaRadius, boxRadius,emptyShellWidth);
      
      Radius_old = boxRadius;
      boxRadius = (int) (Radius_old * 1.25);
      
      outHeight = outWidth = (boxRadius*2)+1;
      arraySize = outHeight * outWidth;
      
      /* Remove particles outside file limits */
      i = 0; j = 0;
      do
	{
	  if(
	     box[i].x+(boxRadius+1) > file_width ||
	     box[i].y+(boxRadius+1) > file_height ||
	     box[i].x-(boxRadius+1) < 0 ||
	     box[i].y-(boxRadius+1) < 0 
	     )
	    {
	      /* Reset reference particles Marker*/
	      /* box[i].reference = FALSE; Do this elsewhere */
	      i ++;
	      numBoxes --;
	    }
	  else
	    {
	      box[j].x = box[i].x;
	      box[j].y = box[i].y;
	      box[j].reference = box[i].reference;
	      box[j].correlation = box[i].correlation;
	      box[j].avgdensity = box[i].avgdensity;
	      box[j].stdev = box[i].stdev;
	      /* Reset reference particles Marker*/
	      /* box[j].reference = FALSE; Do this elsewhere */
	      i ++; j ++;
	    }
	}
      while (j<numBoxes);

      if(iter>0)
	sprintf(outStr,"Trying pass [%d]   Variations last time [%d]\n\
MeanDx[%.1f]  MeanDy[%.1f]\n\
StdDx [%.1f]  StdDy [%.1f]\n",
		iter+1,
		numChanged,
		meanDx,meanDy,
		stdDx,stdDy
		);
      else
	sprintf(outStr,"Trying pass [%d]",iter+1);
      printf("%s\n",outStr);
      if(checkStatus(outStr))
	{/* done */
	  iter= lastIter+11;
	  goto bugOut1;
	}/* done */
      
      totDx = totDy = 0;
      sumDx = sumDy = 0.0;
      varDx = sumsqDx = 0.0;
      varDy = sumsqDy = 0.0;
      
      /* allocate space for the boxed particle image */
      if((refineBox_buffer = (float *)calloc(arraySize,
					     sizeof(float))) == (float *)NULL)
	{
	  printf("\nError: Allocate particle image for refinement");
	  return;
	}
      
      if((tmp_refineBox_buffer = (float *)calloc(arraySize,
					    sizeof(float))) == (float *)NULL)
	{
	  printf("\nError: Allocate particle image for refinement");
	  free(refineBox_buffer);
	  return;
	}
      if((ref_refineBox_buffer = (float *)calloc(arraySize,
					   sizeof(float))) == (float *)NULL)
	{
	  printf("\nError: Allocate particle image for refinement");
	  free(refineBox_buffer);
	  return;
	}
      if((ref_refineBox_buffer_short = (float *)calloc(arraySize,
				    sizeof(float))) == (float *)NULL)
	{
	  printf("\nError: Allocate particle image for refinement");
	  free(refineBox_buffer);
	  return;
	}
      if((refBox_byte_buffer = (float *)calloc(arraySize,
					      sizeof(float))) == (float *)NULL)
	{
	  printf("\nError: Allocate particle image for refinement[refBox]"
		 );
	  free(refineBox_buffer);
	  free(tmp_refineBox_buffer);
	  free(ref_refineBox_buffer);
	  free(ref_refineBox_buffer_short);
	  return; 
	}
      
      /* Open the disk file for load in boxed image */
      fIn= NULL;
      fIn= fopen(image_fileName.addr,"r");
      if(fIn==NULL)
	{/* Error */
	  printf(" Error reading image file for refine center.\n");
	  boxRadius = Radius_old;
	  free(refineBox_buffer);
	  free(tmp_refineBox_buffer);
	  free(ref_refineBox_buffer);
	  free(ref_refineBox_buffer_short);
	  return;
	}/* Error */
      
      if(refReadFile)
	{/* Just use a disk file as the reference */
	  fString_t rWrite,inFile;
	  int nc,nr,ns,halfHt,halfOut;
	  float thisValue;
	  float *tmpFltArray;
	  forPifHead_t thisHead;
	  
	  rWrite.addr = malloc(sizeof(char)*4);
	  rWrite.length = 1;
	  sprintf(rWrite.addr,"r");
	  inFile.addr = malloc(sizeof(char)*512);
	  sprintf(inFile.addr,refReadFileName);
	  inFile.length = strlen(refReadFileName);

	  thisHead.luNum = 17;

	  pif_open_(inFile.addr, &(thisHead.luNum), rWrite.addr);

	  pif_read_gh(&thisHead);
	  thisHead.imgNum = 1;
	  pif_read_dh(&thisHead);

	  if(thisHead.mode != 46)
	    {/* error */
	      sprintf(outStr,
		  "Invalid PIF mode for reference image [%d].  Use mode==46",
		      thisHead.mode);
	      askThem(outStr,
		      FALSE,
		      XmDIALOG_INFORMATION);
	    }/* error */
	  else
	    {/* OK to continue */
	      int yIn,xIn;
	      ns = 1;
	      nc = thisHead.numCols;
	      nr = thisHead.numRows;
	      tmpFltArray = (float *)malloc(sizeof(float)*nc*nr);
	      pif_read_boxi4(&thisHead, tmpFltArray);

	      printf("thisHead.bkg = [%f]\n",(float)thisHead.img_bkg);
	      /* First pad with background */
	      for(i=0; i<(outHeight*outWidth); i++)
		ref_refineBox_buffer_short[i] = (float)thisHead.img_bkg;
	  
	      halfHt = nc / 2;
	      halfOut = outHeight / 2;

	      for(i=0; i<outHeight; i++)
		{
		  yOut = i*outWidth;
		  yIn  = halfHt - (halfOut - i);
		  for(j=0; j<outWidth; j++)
		    {/* Do all */
		      xIn = halfHt - (halfOut -j);
		      
		      if((yIn < 0) ||
			 (yIn > (nr-1)) ||
			 (xIn < 0) ||
			 (xIn > (nc-1)))
			{/* do nothing its already background */
			}/* do nothing its already background */
		      else
			ref_refineBox_buffer_short[yOut+j] = (float)
			  tmpFltArray[(yIn*nc)+xIn] ;
		      
		      if(referenceDisp)
			{/* Now stuff the display buffer so we can see it */
			  if(hiColor)
			    {
			      thisValue = ref_refineBox_buffer_short[yOut+j];
			      XPutPixel(image_image1,
					j, i,
					pixels_return[(int)thisValue]
					);
			    }
			  else
			    image_buffer1[((i*buffer_width)+j)]=
			      (BYTE)pixels_return[
					(int)ref_refineBox_buffer[yOut+j]] ;
			}/* Now stuff the display buffer so we can see it */
		      
		    }/* Do all */
		}
	      
	      if(referenceDisp)
		{/* Just put this stuff up  */
		  XPutImage(dsply, image_window1, image_gc, image_image1, 
			    0, 0,
			    0, 0,  outWidth, outHeight);
		  XFlush(dsply);
		}/* Just put this stuff up  */

	      free(tmpFltArray);
	      
	    }/* OK to continue */

	  free(rWrite.addr);
	  free(inFile.addr);

	}/* Just use a disk file as the reference */
      else if(darkCircleRef)
	{/* Just use a dark circle as the reference */
	  int thisDist,boxDist,ySq,halfHt,halfWid,emptyDist;
	  float thisValue;
	  boxDist   = (emmaRadius * emmaRadius);
	  emptyDist = (int)(((float)(emmaRadius - emptyShellWidth)) * 
			    ((float)(emmaRadius - emptyShellWidth)));
	  halfHt  = outHeight/2;
	  halfWid = outWidth/2;
	  for(i=0; i<outHeight; i++)
	    {
	      yOut = i*outWidth;
	      ySq = (i-halfHt)*(i-halfHt);
	      for(j=0; j<outWidth; j++)
		{/* Do all */
		  thisDist = ySq + ((j-halfWid)*(j-halfWid));
		  if(thisDist < boxDist)
		    ref_refineBox_buffer[yOut+j] = 50.0;  /* Dark */
		  else
		    ref_refineBox_buffer[yOut+j] = 150.0;  /* Background */
		  /* Check if we want an empty model */
		  if(darkCircleRef==99 && thisDist < emptyDist)
		    ref_refineBox_buffer[yOut+j] = 175.0;  /* Empty */
		  /* WORKS ??? 
		   * ref_refineBox_buffer[yOut+j] = 150.0;  /+ Empty */
		    
		  ref_refineBox_buffer_short[yOut+j] = (float)
		    ref_refineBox_buffer[yOut+j] ;
		  
		  if(referenceDisp)
		    {/* Now stuff the display buffer so we can see it */
		      if(hiColor)
			{
			  /* OLDWAYimage_buffer1[((i*buffer_width)+j)]=
			   *pixels_return[(int)ref_refineBox_buffer[yOut+j]];*/
			  thisValue = ref_refineBox_buffer[yOut+j];
			    XPutPixel(image_image1,
				      j, i,
				      pixels_return[(int)thisValue]
				      );
			}
		      else
			image_buffer1[((i*buffer_width)+j)]=
			  (BYTE)pixels_return[
				       (int)ref_refineBox_buffer[yOut+j]] ;
		    }/* Now stuff the display buffer so we can see it */
		  
		}/* Do all */
	    }
	  
	  if(referenceDisp)
	    {/* Just put this stuff up  */
	      XPutImage(dsply, image_window1, image_gc, image_image1, 
			0, 0,
			0, 0,  outWidth, outHeight);
	      XFlush(dsply);
	    }/* Just put this stuff up  */
	  
	}/* Just use a dark circle as the reference */
      else
	{/* Averaging reference particles */
	  
	  
	  numBoxes_ref = 0;
	  for(i=0; i<numBoxes; i++)
	    {/* Do all boxes */
	      if(box[i].reference==1)
		{
		  numBoxes_ref ++;
		  beginY = box[i].y - boxRadius;
		  beginX = box[i].x - boxRadius;
		  
		  for(j=0; j<outHeight; j++)
		    {
		      if(!imgInCore)
			{/* Get from disk */
			  yOff = headOff + (beginY + j)*file_width + 
			    beginX;
			  sysError = fseek(fIn,yOff,0);
			  if(sysError)
			    {
			      printf("Error: fseek in refine center.\n");
			      boxRadius = Radius_old;
			      free(refineBox_buffer);
			      free(ref_refineBox_buffer);
			      free(ref_refineBox_buffer_short);
			      return;
			    }
			  fread(refBox_byte_buffer, sizeof(BYTE),
				(outWidth),fIn);
			}/* Get from disk */
		      else
			{/* Get from core */
			  yOff = (beginY + j)*file_width + beginX;
			  memcpy(refBox_byte_buffer,&inCore_buffer[(yOff)],
				 (outWidth*sizeof(float)));
			}/* Get from core */
		      
		      /* Stuff the ref_refineBox_buffer */
		      yOut = j*outWidth;
		      for(k=0; k<outWidth; k++)
			{/* Do all cols */
			  ref_refineBox_buffer[yOut+k] +=
			    (float)refBox_byte_buffer[k];
			}/* Do all cols */
		    }
		}
	    }/* Do all boxes */
	  
	  
	  if(numBoxes_ref == 0)
	    {
	      sprintf(outStr,
		      "You haven't chosen any reference particles yet.");
	      askThem(outStr,
		      FALSE,
		      XmDIALOG_INFORMATION);
	      ref_refineBox_fft = malloc(1);
	      iter = lastIter;
	      goto bugOut1;
	    }
	  
	  for(i=0; i<outHeight; i++)
	    {
	      yOut = i*outWidth;
	      for(j=0; j<outWidth; j++)
		{ref_refineBox_buffer[yOut+j] = 
		   (float)(0.5 + ref_refineBox_buffer[yOut+j]/
			   (float)numBoxes_ref);
		 ref_refineBox_buffer_short[yOut+j] = (float)
		   ref_refineBox_buffer[yOut+j] ;}
	    }
	  


	  if(refCircAvg && referenceDisp)
	    {/* Just put this stuff up  */
	      maxP=minP=ref_refineBox_buffer[0];
	      for(j=0; j<outHeight; j++)
		{/* hieght */
		  yOut = j*outWidth;
		  for(k=0; k<outWidth; k++)
		    {/* Do all cols */
		      minP = MIN(minP,ref_refineBox_buffer[yOut+k]);
		      maxP = MAX(maxP,ref_refineBox_buffer[yOut+k]);
		    }/* Do all cols */
		}/* hieght */
	      thisScaleFac = (float)(tryColors-11.0)/(float)(maxP-minP);
	      for(j=0; j<outHeight; j++)
		{/* hieght */
		  float thisValue;
		  yOut = j*outWidth;
		  for(k=0; k<outWidth; k++)
		    {/* Do all cols */
		      if(hiColor)
			{
			thisValue=
		   ((float)(ref_refineBox_buffer[yOut+k]-minP)*thisScaleFac)
						    + 9.0
						     ;
		      XPutPixel(image_image1,
				k+400, j,
				pixels_return[(int)thisValue]
				);
			}
		      else
			image_buffer1[((j*buffer_width)+k)]=
			  (BYTE)pixels_return[(int)(
		((float)(ref_refineBox_buffer[yOut+k]-minP)*thisScaleFac)
						    + 9
						    )] ;
		      /* Put histograph of middle line here */
		    }/* Do all cols */
		}/* hieght */
	      XPutImage(dsply, image_window1, image_gc, image_image1, 
			400, 0,
			400, 0,  outWidth, outHeight);
	      XFlush(dsply);
	    }/* Just put this stuff up  */

	  if(refCircAvg)
	    {/* Circular average the reference */
	      float *tmpBuffer,thisBkg;
	      int ncol,nrow,interp,yOff,xOff,nterp;
	      float xc,yc,dist;
	      float *rint;
	      int *irnum;
	      
	      ncol = outWidth;
	      nrow = outHeight;
	      xc = (float)((float)ncol/2.0);
	      yc = (float)((float)nrow/2.0);
	      interp= 2;  /* Original way */
	      interp= 1;  /* Check circAvgMotion() to make sure 
			   * they agree */
	      nterp = interp*(int)(
				   (int)((MAX(ncol,nrow)))
				   / (int)2 
				   ) 
		+ 1;
	      
	      CK(rint= (float *)malloc(sizeof(float)*nterp*3));
	      
	      CK(irnum= (int *)malloc(sizeof(int)*nterp*3));
	      CK(tmpBuffer= (float *)malloc(sizeof(float)*
					   outWidth*outHeight));
	      
	      /* Stuff buffer */
	      for(i=0;i<outHeight;i++)
		for(j=0;j<outWidth;j++)
		  {tmpBuffer[(i*outWidth)+j]= 
		     (float)ref_refineBox_buffer[(i*outWidth)+j];}
	      
	      thisBkg = 0.0;
	      img_1d_rob(tmpBuffer,
			 ncol,nrow,
			 xc,yc,
			 interp,nterp,
			 rint,irnum,
			 thisBkg);
	      
	      for(i=0;i<outHeight;i++)
		{
		  yOff = ((float)i-yc)*((float)i-yc);
		  for(j=0;j<outWidth;j++)
		    {
		      xOff = ((float)j-xc)*((float)j-xc);
		      dist = (float)sqrt((float)(xOff+yOff));
		      tmp_refineBox_buffer[(i*outWidth)+j]=
			(float)rint[((int)dist)];
		    }
		}
	      
	      free(irnum);
	      free(rint);
	      free(tmpBuffer);
	      
	      if(iter >= 0)
		{/* debugging - try this */
		  for(j=0; j<outHeight; j++)
		    {/* hieght */
		      yOut = j*outWidth;
		      for(k=0;k<outWidth;k++)
			{
			  ref_refineBox_buffer[yOut+k]=
			    tmp_refineBox_buffer[yOut+k];
			}
		    }

		}/* debugging - try this */
	      
	    }/* Circular average the reference */
	  
	  
	  if(referenceDisp)
	    {/* Just put this stuff up  */
	      maxP=minP=tmp_refineBox_buffer[0];
	      for(j=0; j<outHeight; j++)
		{/* hieght */
		  yOut = j*outWidth;
		  for(k=0; k<outWidth; k++)
		    {/* Do all cols */
		      minP = MIN(minP,tmp_refineBox_buffer[yOut+k]);
		      maxP = MAX(maxP,tmp_refineBox_buffer[yOut+k]);
		    }/* Do all cols */
		}/* hieght */
	      thisScaleFac = (float)(tryColors-11.0)/(float)(maxP-minP);
	      for(j=0; j<outHeight; j++)
		{/* hieght */
		  float thisValue;
		  yOut = j*outWidth;
		  for(k=0; k<outWidth; k++)
		    {/* Do all cols */
		      if(hiColor)
			{
			thisValue=
		   ((float)(tmp_refineBox_buffer[yOut+k]-minP)*thisScaleFac)
						    + 9.0
						     ;
		      XPutPixel(image_image1,
				k, j,
				pixels_return[(int)thisValue]
				);
			}
		      else
			image_buffer1[((j*buffer_width)+k)]=
			  (BYTE)pixels_return[(int)(
		((float)(tmp_refineBox_buffer[yOut+k]-minP)*thisScaleFac)
						    + 9
						    )] ;
		      /* Put histograph of middle line here */
		    }/* Do all cols */
		}/* hieght */
	      
	      XPutImage(dsply, image_window1, image_gc, image_image1, 
			0, 0,
			0, 0,  outWidth, outHeight);
	      XFlush(dsply);
	    }/* Just put this stuff up  */
	  
	}/* Averaging reference particles */



      /* compute image background for averaged particle*/
      total_bkg = 0;
      for(j=0; j<outWidth; j++)
	total_bkg += (float)ref_refineBox_buffer[j];       /*top row*/
      for(j=arraySize-outWidth; j<arraySize; j++)
	total_bkg += (float)ref_refineBox_buffer[j];       /*bottom row */
      for(j=outWidth; j<arraySize-outWidth; j+=outWidth)
	total_bkg += (float)ref_refineBox_buffer[j];       /*left column*/
      for(j=2*outWidth-1; j<arraySize-outWidth; j+=outWidth)
	total_bkg += (float)ref_refineBox_buffer[j];       /*right column*/
      
      ref_img_bkg = (float) total_bkg/(float)(2.0*outHeight+ 2*outWidth - 4);
      
      /***** What is the best fft size? ****/
      maxborder = MAX(outWidth, outHeight);
      fft_dim1 = 1;
      while(fft_dim1 < maxborder)
	fft_dim1 *= 2;
      fft_dim2 = fft_dim1; 
      
      
      /***** Apply fft on averaged reference partical once *****/
      if((ref_refineBox_fft = (float *)calloc( (2*fft_dim1*fft_dim2),
				      sizeof(float) )) == (float *)NULL)
	{
	  printf("\nError: Allocate particle image for refinement");
	  boxRadius = Radius_old;
	  free(refineBox_buffer);
	  free(tmp_refineBox_buffer);
	  free(ref_refineBox_buffer);
	  free(ref_refineBox_buffer_short);
	  return;
	}
      
      /* Init */
      for(j=0;j<(2*fft_dim1*fft_dim2);j++)
	ref_refineBox_fft[j]= 0.0;
      
      fft_avg_ref_(ref_refineBox_buffer_short, ref_refineBox_fft, 
		   &ref_img_bkg, &outWidth, &outHeight, &fft_dim1, 
		   &fft_dim2);
      

      /*****  Refine the center for each box *****/
      if(showCC)
	{/* Just do the zero'th one */
	  float min,max;
	  float sFac;
	  saveNumBoxes = numBoxes;
	  numBoxes = 1;
	  
	  /* Now show the reference on the screen */
	  min = max = ref_refineBox_buffer[0];
	  for(i=0; i<outHeight; i++)
	    {
	      yOut = i*outWidth;
	      for(j=0; j<outWidth; j++)
		{/* Do all */
		  max = MAX(ref_refineBox_buffer[yOut+j],max);
		  min = MIN(ref_refineBox_buffer[yOut+j],min);
		}/* Do all */
	    }
	  sFac = (tryColors) / (max-min);

	  for(i=0; i<outHeight; i++)
	    {
	      yOut = i*outWidth;
	      for(j=0; j<outWidth; j++)
		{/* Do all */
		  if(hiColor)
		    {/* hi */
		      int thisValue;
		      thisValue = (ref_refineBox_buffer[yOut+j]-min)*sFac ;
		      XPutPixel(image_image1,
				j, i,
				pixels_return[thisValue]
				);
		    }/* hi */
		  else
		    image_buffer1[(y*image_width)+j]=
		      (BYTE)((ref_refineBox_buffer[yOut+j]-min)*sFac) ;
		}/* Do all */
	    }
	  /* Display the image in the window created above */
	  XPutImage (dsply, image_window1, image_gc, image_image1, 0, 0,
		     0, 0, outWidth, outHeight);
	  XFlush (dsply);
	}/* Just do the zero'th one */
      

      for(i=0;i<numBoxes;i++)
	{/* Do all these boxes */
	  
	  /* Give feedback to user */
	  if(lastIter > 1)
	    sprintf(outStr,"Iteration[%d]  Boxes left - [%d]",
		    iter+1,(numBoxes-i));
	  else
	    sprintf(outStr,"Boxes left - [%d]",(numBoxes-i));
	  XmTextSetString(timeLeftBoxW,outStr);
	  XFlush(dsply);
	  /* CHeck if they want to abort or not etc... */
	  if(XCheckTypedWindowEvent(dsply,
				    XtWindow(boxStopWidget),
				    ButtonPress,
				    &event))
	    {/* STOP!!! */
	      sprintf(outStr,"Operation aborted by user.");
	      XmTextSetString(timeLeftBoxW,outStr);
	      XFlush(dsply);
	      totalBugOut = TRUE;
	      goto bugOut1;
	    }/* STOP!!! */
	  
	  if(box[i].nochange < 2)
	    {/* Only do those who've changed a lot */

	      /* Load in particle image for refinement */
	      beginY = box[i].y - boxRadius;
	      beginX = box[i].x - boxRadius;
	      
	      for(j=0; j<outHeight; j++)
		{
		  if(!imgInCore)
		    {/* Get from disk */
		      yOff = headOff + (beginY + j)*file_width + beginX;
		      sysError = fseek(fIn,yOff,0);
		      if(sysError)
			{
			  printf("Error: fseek in refine center.\n");
			  goto bugOut1;
			}
		  fread(refBox_byte_buffer, sizeof(BYTE),
			(outWidth),fIn);
		    }/* Get from disk */
		  else
		    {/* Get from core */
		      yOff = (beginY + j)*file_width + beginX;
		      memcpy(refBox_byte_buffer,&inCore_buffer[(yOff)],
			     (outWidth*sizeof(float)));
		    }/* Get from core */
		  
		  /* Stuff the refineBox_buffer */
		  yOut = j*outWidth;
		  for(k=0; k<outWidth; k++)
		    refineBox_buffer[yOut+k]= (float)refBox_byte_buffer[k];
		}
	      
	      /* compute image background for each particle*/
	      total_bkg = 0;
	      
	      for(j=0; j<outWidth; j++)
		total_bkg += (float)refineBox_buffer[j];       /*top row*/
	      for(j=arraySize-outWidth; j<arraySize; j++)
		total_bkg += (float)refineBox_buffer[j];       /*bottom row */
	      for(j=outWidth; j<arraySize-outWidth; j+=outWidth)
		total_bkg += (float)refineBox_buffer[j];       /*left column*/
	      for(j=2*outWidth-1; j<arraySize-outWidth; j+=outWidth)
		total_bkg += (float)refineBox_buffer[j];       /*right column*/
	      
	      img_bkg = (float) total_bkg/(float)(2*outHeight+ 2*outWidth - 4);
	      
	      
	      ref_ccf_refine_(ref_refineBox_fft, refineBox_buffer, &img_bkg,
			      &outWidth, &outHeight, &fft_dim1, &fft_dim2, 
			      &dx, &dy);
	      
	      if(dx==0 && dy==0)
		{/* No change */
		  box[i].nochange++;
		}/* No change */
	      else
		{/* Reset */
		  box[i].nochange= 0;
		}/* Reset */

	      /* Set boxing center to new value */
	      box[i].x +=  dx;  /* x MAY be biased 1 to left?????? */
	      box[i].y +=  dy;
	      sumDx += (int)fabs((float)dx);
	      sumDy += (int)fabs((float)dy);
	      totDx ++;
	      totDy ++;
	      sumsqDx += ((float)dx * (float)dx);
	      sumsqDy += ((float)dy * (float)dy);
	      
	      
	      
	      /***********  remove bad particle ************/
	      /* Load in refined particle image */
	      beginY = box[i].y - boxRadius;
	      beginX = box[i].x - boxRadius;
	      if((beginX < 0 || beginY < 0) ||
		 (((box[i].y+outHeight)-file_height)>=0) ||
		 (((box[i].x+outWidth)-file_width)>=0) 
		 )

		{/* Throw it out - it's off the edge */
		  box[i].reference = -1;
		  goto byPassThrowOut;
		}/* Throw it out - it's off the edge */
	      for(j=0; j<outHeight; j++)
		{
		  if(!imgInCore)
		    {/* Get from disk */
		      yOff = headOff + (beginY + j)*file_width + beginX;
		      sysError = fseek(fIn,yOff,0);
		      if(sysError)
			{
			  printf("Error: fseek in refine center.\n");
			  goto bugOut1;
			}
		      fread(refBox_byte_buffer, sizeof(BYTE),
			    (outWidth),fIn);
		    }/* Get from disk */
		  else
		    {/* Get from core */
		      yOff = (beginY + j)*file_width + beginX;
		      memcpy(refBox_byte_buffer,&inCore_buffer[(yOff)],
			     (outWidth*sizeof(float)));
		    }/* Get from core */
		  
		  /* Stuff the refineBox_buffer */
		  yOut = j*outWidth;
		  for(k=0; k<outWidth; k++)
		    refineBox_buffer[yOut+k]= (float)refBox_byte_buffer[k];
		}
	      
	      /* compute image background for each particle*/
	      total_bkg = 0;
	      
	      for(j=0; j<outWidth; j++)
		total_bkg += (float)refineBox_buffer[j];       /*top row*/
	      for(j=arraySize-outWidth; j<arraySize; j++)
		total_bkg += (float)refineBox_buffer[j];       /*bottom row */
	      for(j=outWidth; j<arraySize-outWidth; j+=outWidth)
		total_bkg += (float)refineBox_buffer[j];       /*left column*/
	      for(j=2*outWidth-1; j<arraySize-outWidth; j+=outWidth)
		total_bkg += (float)refineBox_buffer[j];       /*right column*/
	      
	      img_bkg = (float) total_bkg/(float)(2*outHeight+ 2*outWidth - 4);
	      
	      /* Mark bad particles by comparing local background */
	      outsideTotal = 0;
	      outsideNo = 0;
	      insideTotal = 0;
	      insideNo = 0;
	      cx = cy = boxRadius;
	      for(j=0;j<outHeight;j++)
		{
		  for(k=0;k<outWidth;k++)
		    {
		      if(((j-cy)*(j-cy)+(k-cx)*(k-cx))>Radius_old*Radius_old*
			 1.05)
			{
			  outsideNo ++;
			  if((refineBox_buffer[j*outWidth+k]) < img_bkg)
			    outsideTotal ++;
			}
		      if(((j-cy)*(j-cy)+(k-cx)*(k-cx))<Radius_old*Radius_old)
			{
			  insideNo ++;
			  if((refineBox_buffer[j*outWidth+k]) < img_bkg)
			    insideTotal ++;
			}
		    }
		}
	      outsideAvg = (float)outsideTotal/(float)outsideNo;
	      insideAvg = (float)insideTotal/(float)insideNo;
	      if(outsideAvg != 0.0)
		dividedAvg = insideAvg / outsideAvg;
	      else dividedAvg = refCutoff+1.00;  /* Make sure it's included */

	    byPassThrowOut:;
	      
	      box[i].correlation = dividedAvg;
	      box[i].avgdensity = insideAvg/256.0;
	      box[i].avgdensity = (float)insideAvg;
	      
	      if(refineDisp)
		{/* Give feedback as to which circle we're working on */
		  int arcRadius;
		  /* Size of the circle */
		  calcSizeFactor(zoomFactor);
		  arcRadius = (int)(sizeFactor * Radius_old);
		  rwaXDrawArc(dsply, XtWindow(drawingWidget), image_gcBLUE,
			      (((box[i].x*sizeFactor)-(virXoffset*sizeFactor))-
			       arcRadius), 
			      (((box[i].y*sizeFactor)-(virYoffset*sizeFactor))-
			       arcRadius),
			      (arcRadius*2),(arcRadius*2),
			      0, (360*64));
		}/* Give feedback as to which circle we're working on */
	      
	    }/* Only do those who've changed a lot */
	  else
	    {/* SHow them we're done with this one */
	      if(refineDisp)
		{/* Give feedback as to which circle we're working on */
		  int arcRadius;
		  /* Size of the circle */
		  calcSizeFactor(zoomFactor);
		  arcRadius = (int)(sizeFactor * Radius_old);
		  rwaXDrawArc(dsply, XtWindow(drawingWidget), image_gcWHITE,
			      (((box[i].x*sizeFactor)-(virXoffset*sizeFactor))-
			       arcRadius), 
			      (((box[i].y*sizeFactor)-(virYoffset*sizeFactor))-
			       arcRadius),
			      (arcRadius*2),(arcRadius*2),
			      0, (360*64));
		}/* Give feedback as to which circle we're working on */
	    }/* SHow them we're done with this one */
	  
	}/* Do all these boxes */
      
      
      /* Let's try and throw out duplicate boxes */
      for(j=0;j<numBoxes;j++)
	{/* Look through all boxes */
	  int dist;
	  for(i=0;i<numBoxes;i++)
	    {/* second */
	      if((i != j) && (box[j].reference!=-1))
		{/* Dont compare this to itself */
		  dist = (box[i].x - box[j].x)*(box[i].x - box[j].x)
		    +
		    (box[i].y - box[j].y)*(box[i].y - box[j].y);
		  if(dist < 80)
		    {/* Throw it out - it's duplicate */
		      box[i].reference = -1;
		      /* Now average the coords of the 2 particles */
		      box[j].y = (int)(((float)box[j].y +(float)box[i].y)*0.5);
		      box[j].x = (int)(((float)box[j].x +(float)box[i].x)*0.5);
		    }/* Throw it out - it's duplicate */
		}/* Dont compare this to itself */
	    }/* second */
	}/* Look through all boxes */

      /* Remove bad particle */
      i = 0; j = 0;
      do
	{
	  if(box[i].reference == -1)
	    {
	      /* Reset reference particles Marker*/
	      /* box[i].reference = FALSE; Do this elsewhere */
	      i ++;
	      numBoxes --;
	    }
	  else
	    {
	      box[j].x = box[i].x;
	      box[j].y = box[i].y;
	      box[j].reference = box[i].reference;
	      box[j].correlation = box[i].correlation;
	      box[j].avgdensity = box[i].avgdensity;
	      box[j].stdev = box[i].stdev;
	      /* Reset reference particles Marker*/
	      /* box[j].reference = FALSE; Do this elsewhere */
	      i ++; j ++;
	    }
	}
      while (j<numBoxes);
      

      /* Let's see if we've changed much from last time... */
      /* Now let's do stats */
      meanDy = (float)((float)sumDy / (float)(totDy));
      stdDy  = 0.0;
      varDy  = (sumsqDy / (float)(totDy)) - (meanDy * meanDy);
      if(varDy > 0.0)
	stdDy = (float)sqrt(varDy);
      else stdDy = (float)sqrt(fabs(varDy));
      meanDx = (float)((float)sumDx / (float)(totDx));
      stdDx  = 0.0;
      varDx  = (sumsqDx / (float)(totDx)) - (meanDx * meanDx);
      if(varDx > 0.0)
	stdDx = (float)sqrt(varDx);
      else stdDx = (float)sqrt(fabs(varDx));
      numChanged = sumDx + sumDy;
      if(numChanged < (int)(numBoxes*0.02))
	{/* Let's stop */
	  sprintf(outStr,"Stopping - not much changed. [Only %d boxes]",
		  numChanged);
	  printf("%s\n",outStr);
	  checkStatus(outStr);
	  iter= lastIter+11;
	}/* Let's stop */
      
      if(showCC)
	{/* Just did the zero'th one */
	  numBoxes = saveNumBoxes;
	}/* Just did the zero'th one */
      
      /* Make sure they can toggle this now... */
      XtSetArg(arglst[0], XmNsensitive, TRUE);
      /*  XtSetValues(showccToggleW, arglst, 1);  */
      
    bugOut1:
      /* Recover the radius */
      boxRadius = Radius_old;
      
      free(ref_refineBox_fft);
      
      fclose(fIn);
      free(refBox_byte_buffer);
      free(refineBox_buffer);
      free(tmp_refineBox_buffer);
      free(ref_refineBox_buffer);
      free(ref_refineBox_buffer_short);
      if(totalBugOut)
	goto totalBugOut2;

      DrawWindow();
      
      
    }/* Do iter times or until they opt out */
 totalBugOut2:;
  
  sprintf(outStr,"Done.  [%d] boxes.",numBoxes);
  XmTextSetString(timeLeftBoxW,outStr);
  XFlush(dsply);
  
  sprintf(outStr,"Done.");
  createSimpleString(&simpleString,outStr);
  XtSetArg(arglst[0], XmNlabelString, simpleString );
  XtSetArg(arglst[1], XmNrecomputeSize, FALSE );
  XtSetValues(refCorrLabelW, arglst, 2); 
  XFlush(dsply);
  XmStringFree(simpleString); 
  
  /* Setup for next time */
  for(i=0; i<numBoxes; i++)
    box[i].reference = FALSE;
  
  DrawWindow();

  /* Change cursor back to what it was */
  attrs.cursor= old_cursor;
  XChangeWindowAttributes(dsply, XtWindow(drawingWidget), 
			  CWCursor, &attrs);
  XFlush(dsply);

}/* REF_CORRELATE_ACTIVATE */





/*************************************************************
 * ref_select_activate() - Mark all boxes as candidates for
 * reference correlation standard.
 */
void ref_select_activate(Widget widget, int *tag, 
			 XmAnyCallbackStruct *callback_data)
{/* REF_SELECT_ACTIVATE */
  int i,passTag;

  if(*tag==ALL)
    {/* Mark all reference particles Marker*/
      for(i=0; i<numBoxes; i++)
	box[i].reference = TRUE;
    }/* Mark all reference particles Marker*/
  else if(*tag==LAST)
    {/* Last time reference particles Marker*/
      /* OBSOLETE */
      for(i=0; i<numBoxes; i++)
	{
	  box[i].reference = boxRefLast[i];
	}
    }/* Last time reference particles Marker*/
  else if(*tag==RESET)
    {/* Reset reference particles Marker*/
      for(i=0; i<numBoxes; i++)
	box[i].reference = FALSE;
    }/* Reset reference particles Marker*/
  else if(*tag==REGRID)
    {/* Reset grid*/
      for(i=0; i<numBoxes; i++)
	box[i].reference = FALSE;

      /* CLear all boxes */
      for(i=0;i<numBoxes;i++)
	if(box[i].inMemory==TRUE)
	  {/* Free it */
	    free(box[i].editBox_buffer);
	    box[i].editBox_buffer= NULL;
	    box[i].inMemory= FALSE;
	  }/* Free it */
      numBoxes= numRects= numPoly= 0;
      
      passTag = -99;
      boxgrid_activate(widget,&passTag,callback_data);
    }/* Reset grid*/

  XFlush(dsply);

  DrawWindow();

}/* REF_SELECT_ACTIVATE */



/*************************************************************
 * ref_screen_deactivate() - Make ref_screen go away.
 */
void ref_screen_deactivate(Widget widget, int *tag, 
			   XmAnyCallbackStruct *callback_data)
{/* REF_SCREEN_DEACTIVATE */

  /* OBSOLETE for select last functionality */
  for(i=0; i<numBoxes; i++)
    {
      boxRefLast[i] = box[i].reference ;
    }
  ref_box_flag = FALSE;
  XtUnmanageChild(refscreen);
  ref_correlate_activate();

}/* REF_SCREEN_DEACTIVATE */


/*************************************************************
 * ref_text_create() - Remember widget number!
 */
void ref_text_create(Widget widget, int *tag, 
		     XmAnyCallbackStruct *callback_data)
{/* REF_TEXT_CREATE */

  refTextWidget = widget;

}/* REF_TEXT_CREATE */


/*************************************************************
 * stats_activate() - Show them the image statistics.
 */
void stats_activate(Widget widget, int *tag, 
		    XmAnyCallbackStruct *callback_data)
{/* STATS_ACTIVATE */

  raiseIt(statscreen);

}/* STATS_ACTIVATE */



/*************************************************************
 * displayPifHead() - Show them exactly what we think are the
 * contents of the header(s) of the PIF file.
*/
void displayPifHead(forPifHead_t thisHead)
{/* DISPLAYPIFHEAD */
  char pifHeadStr[4028];
  char pixSizUnitsStr[32];

  if(thisHead.pixSizUnits==1)
    sprintf(pixSizUnitsStr,"Angs");
  else if(thisHead.pixSizUnits==2)
    sprintf(pixSizUnitsStr,"nm");
  else
    sprintf(pixSizUnitsStr,"Unk");


      sprintf(pifHeadStr,"\nPIF Header for file [%s]\n\
\n\
    nimg                    =   [%d]\n\
    imgNum                  =   [%d]\n\
    floatScaleFactor        =   [%f]\n\
    recipFloatScaleFactor   =   [%f]\n\
    int4scaleF              =   [%f]\n\
    htype                   =   [%d]\n\
    numCols                 =   [%d]\n\
    numRows                 =   [%d]\n\
    numSecs                 =   [%d]\n\
    mode                    =   [%d]\n\
    endianNess              =   [%d]\n\
    versionStr              =   [%s]\n\
    img_bkg                 =   [%f]\n\
    datai4_min              =   [%d]\n\
    datai4_max              =   [%d]\n\
    datar4_min              =   [%f]\n\
    datar4_max              =   [%f]\n\
    datar4_avg              =   [%f]\n\
    datar4_std              =   [%f]\n\
    descriptionStr          =   [%s]\n\
    timeStr                 =   [%s]\n\
    micrographStr           =   [%s]\n\
    scanNumberStr           =   [%s]\n\
    aoverb                  =   [%f]\n\
    map_abang               =   [%f]\n\
    a_cell                  =   [%f]\n\
    b_cell                  =   [%f]\n\
    c_cell                  =   [%f]\n\
    alp                     =   [%f]\n\
    bet                     =   [%f]\n\
    gam                     =   [%f]\n\
    dela                    =   [%f]\n\
    delb                    =   [%f]\n\
    delc                    =   [%f]\n\
    pixelSize               =   [%f]\n\
    pixSizUnits             =   [%d][%s]\n\
    t_matrix                =   [%d][%d][%d]\n\
                                [%d][%d][%d]\n\
    packRadius              =   [%d]\n\
    binFactor               =   [%d]\n\
    dthe                    =   [%f]\n\
    dphi_90                 =   [%f]\n\
    symmetry                =   [%d]\n\
    res_out                 =   [%f]\n\
    ctfMode                 =   [%d]\n\
    tempFactor              =   [%f]\n\
    wiener                  =   [%f]\n\
    borderType              =   [%d]\n\
    borderWidth             =   [%d]\n\
    rad_lo                  =   [%d]\n\
    rad_hi                  =   [%d]\n\
    pifVersion              =   [%f]\n\
    even                    =   [%d]\n\
    mrcX                    =   [%d]\n\
    mrcY                    =   [%d]\n\
    mrcZ                    =   [%d]\n\
    transX                  =   [%d]\n\
    transY                  =   [%d]\n\
    transZ                  =   [%d]\n\
    isTrunc                 =   [%d]\n\
    origX                   =   [%d]\n\
    origY                   =   [%d]\n\
    origZ                   =   [%d]\n\
\n\
\n",
	      fileName.addr,
	      thisHead.nimg,
	      thisHead.imgNum,
	      thisHead.floatScaleFactor,
	      thisHead.recipFloatScaleFactor,
	      thisHead.int4scaleF,
	      thisHead.htype,
	      thisHead.numCols,
	      thisHead.numRows,
	      thisHead.numSecs,
	      thisHead.mode,
	      thisHead.endianNess,
	      thisHead.versionStr,
	      thisHead.img_bkg,
	      thisHead.datai4_min,
	      thisHead.datai4_max,
	      thisHead.datar4_min,
	      thisHead.datar4_max,
	      thisHead.datar4_avg,
	      thisHead.datar4_std,
	      thisHead.descriptionStr,
	      thisHead.timeStr,
	      thisHead.micrographStr,
	      thisHead.scanNumberStr,
	      thisHead.aoverb,
	      thisHead.map_abang,
	      thisHead.a_cell,
	      thisHead.b_cell,
	      thisHead.c_cell,
	      thisHead.alp,
	      thisHead.bet,
	      thisHead.gam,
	      thisHead.dela,
	      thisHead.delb,
	      thisHead.delc,
	      thisHead.pixelSize,
	      thisHead.pixSizUnits,
	      pixSizUnitsStr,
	      thisHead.t_matrix[0],thisHead.t_matrix[1],
	      thisHead.t_matrix[2],thisHead.t_matrix[3],
	      thisHead.t_matrix[4],thisHead.t_matrix[5],
	      thisHead.packRadius,
	      thisHead.binFactor,
	      thisHead.dthe,
	      thisHead.dphi_90,
	      thisHead.symmetry,
	      thisHead.res_out,
	      thisHead.ctfMode,
	      thisHead.tempFactor,
	      thisHead.wiener,
	      thisHead.borderType,
	      thisHead.borderWidth,
	      thisHead.rad_lo,
	      thisHead.rad_hi,
	      thisHead.pifVersion,
	      thisHead.even,
	      thisHead.mrcX,
	      thisHead.mrcY,
	      thisHead.mrcZ,
	      thisHead.transX,
	      thisHead.transY,
	      thisHead.transZ,
	      thisHead.isTrunc,
	      thisHead.origX,
	      thisHead.origY,
	      thisHead.origZ
	      );

  
  XmTextSetString(statTextWidget,pifHeadStr);
  XFlush(dsply);
  
  XmTextSetString(titleTextW,thisHead.descriptionStr);
  XFlush(dsply);

}/* DISPLAYPIFHEAD */

  

/*************************************************************
 * pifhead_activate() - List the PIF header.
 */
void pifhead_activate(Widget widget, int *tag, 
		      XmAnyCallbackStruct *callback_data)
{/* PIFHEAD_ACTIVATE */
  fString_t rWrite,inFile;
  XmString simpleString;

  displayPifHead(pifHeader);


}/* PIFHEAD_ACTIVATE */




/*************************************************************
 * save_title_activate() - Update PIF file with this new 
 * title.
 */
void save_title_activate(Widget widget, int *tag, 
			 XmAnyCallbackStruct *callback_data)
{/* SAVE_TITLE_ACTIVATE */
  char *thisStr;
  int imgOffset,thisImgNum;
  FILE *fOut;

  thisStr = XmTextGetString(titleTextW); 
  if(strlen(thisStr) > 79)
    thisStr[79]= '\0';
  sprintf(pifHeader.descriptionStr,"%s",thisStr);
  XtFree(thisStr);
  XFlush(dsply);

  fOut= NULL;
  fOut = fopen(fileName.addr,"r+");
	  
  if(fileType==PACKEDPORTABLE)
    thisImgNum = imgNum;
  else thisImgNum = 1;
  
  imgOffset = 512 +  /* Global header */
              ((thisImgNum-1) *       /* Number of images preceding this */
	       (512 +                 /* Their data headers */
		(purdueHeader.numRows*
		 purdueHeader.numCols*
		 purdueHeader.numSecs)  /* Their data */
		)
	       ) + 116                  /* Offset of descriptionStr */
		 ;

  /* Now update the header to show new data at the end */
  sysError= fseek(fOut,imgOffset,0);
  if(sysError)
    {/* Can't go on */
      askThem("[SAVE_TITLE_ACTIVATE] Error seeking file header.",
	      TRUE,XmDIALOG_INFORMATION);
    }/* Can't go on */
  else
    {/* Update title */
      fwrite(pifHeader.descriptionStr,sizeof(char),80,fOut);
    }/* Update title */

  fclose(fOut);
  fOut= NULL;

  displayPifHead(pifHeader);

}/* SAVE_TITLE_ACTIVATE */




/*************************************************************
 * stat_calc_activate() - Calc the stats.
 */
void stat_calc_activate(Widget widget, int *tag, 
			XmAnyCallbackStruct *callback_data)
{/* STAT_CALC_ACTIVATE */

  if(fileType==MAP || fileType==FLOATMAP)
    {/* MAP */

      calcMapStats(float_image_buffer1,
		   &purdueHeader.img_bkg, 
		   &purdueHeader.map_min,
		   &purdueHeader.map_max, 
		   &purdueHeader.img_avg, 
		   &purdueHeader.img_std,
		   &numCols, &numRows, &origNumSecs);

      /* Write stats */
      sprintf(outStr,"\nStatistics for file [%s]\n\
\n\
Title [%s]\n\
\n\
File size   [%d]x[%d]x[%d]\n\
Background  [%f]\n\
MIN,MAX     [%f],[%f]\n\
AVG,STD     [%f],[%f]\n\
\n",
	      fileName.addr,
	      title,
	      numCols,	  numRows, origNumSecs,
	      purdueHeader.img_bkg, 
	      minMapPixel, maxMapPixel,
	      purdueHeader.img_avg, purdueHeader.img_std
	      );
      XmTextSetString(statTextWidget,outStr);
      XFlush(dsply);
    }/* MAP */
  else if(fileType==EMPORTABLE || fileType==BYTEPORTABLE
	  || fileType==PORTABLE)
    {/* Scanned iamge */
      int useRectangle,thisNcol,thisNrow,median;

      useRectangle = rectBox.drawable;
      
      if(useRectangle)
	{/* USe rectangle */
	  int saveFileType;
	  float *char_buffer;

	  thisNrow = rectBox.lowy - rectBox.y;
	  thisNcol = rectBox.lowx - rectBox.x;

	  CK(char_buffer= (float *)malloc(
				   sizeof(float)*thisNcol*thisNrow));

	  stuffRectangleBuffer(char_buffer, 
			       image_fileName.addr);
	  
	  saveFileType = fileType;
	  fileType = RECTPACKED;
	  calcStats(char_buffer,
		    &purdueHeader.img_bkg, 
		    &purdueHeader.img_min,
		    &purdueHeader.img_max, 
		    &purdueHeader.img_avg, 
		    &purdueHeader.img_std,
		    thisNcol, thisNrow, thisNcol);
	  fileType = saveFileType;

	  calcRectMedian(char_buffer,&median,
			thisNcol, thisNrow);

	  free(char_buffer);

	  /* Write stats */
	  sprintf(outStr,"\nStatistics for file [%s] RECTANGLE\n\
\n\
Rect size   [%d]x[%d]\n\
Rect offset x[%d]y[%d]\n\
Background  [%d]\n\
MIN,MAX     [%d],[%d]\n\
Median      [%d]\n\
AVG,STD     [%f],[%f]\n\
",
		  fileName.addr,
		  thisNcol, thisNrow,
		  rectBox.x,rectBox.y,
		  purdueHeader.img_bkg,
		  purdueHeader.img_min,purdueHeader.img_max,
		  median,
		  purdueHeader.img_avg,purdueHeader.img_std
		  );
	}/* USe rectangle */
      else
	{/* No can do */
	  calcScanStats(inCore_buffer,
			&purdueHeader.img_bkg, 
			&purdueHeader.img_min,
			&purdueHeader.img_max, 
			&purdueHeader.img_avg, 
			&purdueHeader.img_std,
			file_width, file_height, 
			file_width);

	  /* Write stats */
	  sprintf(outStr,"\nStatistics for file [%s] Image[%d]\n\
\n\
File size   [%d]x[%d]\n\
Background  [%f]\n\
MIN,MAX     [%f],[%f]\n\
AVG,STD     [%f],[%f]\n\
",
		  fileName.addr,1,
		  purdueHeader.numCols,	  purdueHeader.numRows,
		  purdueHeader.img_bkg,
		  purdueHeader.img_min,purdueHeader.img_max,
		  purdueHeader.img_avg,purdueHeader.img_std
		  );
	}/* No can do */

      XmTextSetString(statTextWidget,outStr);
    }/* Scanned iamge */
  else if(fileType==PACKEDPORTABLE)
    {/* packed boxes */
      
      calcCircleStats(packImg[(imgNum-1)],
		      &purdueHeader.img_bkg, 
		      &purdueHeader.img_min,
		      &purdueHeader.img_max, 
		      &purdueHeader.img_avg, 
		      &purdueHeader.img_std,
		      purdueHeader.numCols, purdueHeader.numRows, 
		      purdueHeader.numCols);
      
      /* Write stats */
      sprintf(outStr,"\nStatistics for file [%s] Image [%d]\n\
\n\
File size   [%d]x[%d]x[%d]\n\
Background  [%f]\n\
MIN,MAX     [%f],[%f]\n\
AVG,STD     [%f],[%f]\n\
\n\
\n",
	      fileName.addr, imgNum,
	      purdueHeader.numCols,	  purdueHeader.numRows, 
	      purdueHeader.nimg,
	      purdueHeader.img_bkg,
	      purdueHeader.img_min, purdueHeader.img_max, 
	      purdueHeader.img_avg, purdueHeader.img_std
	      );
      XmTextSetString(statTextWidget,outStr);
    }/* packed boxes */

}/* STAT_CALC_ACTIVATE */



/*************************************************************
 * stat_screen_deactivate() - Make stats go away.
 */
void stat_screen_deactivate(Widget widget, int *tag, 
			    XmAnyCallbackStruct *callback_data)
{/* STAT_SCREEN_DEACTIVATE */

  XtUnmanageChild(statscreen);

}/* STAT_SCREEN_DEACTIVATE */



/*************************************************************
 * stat_text_create() - Remember widget number!
 */
void stat_text_create(Widget widget, int *tag, 
		      XmAnyCallbackStruct *callback_data)
{/* STAT_TEXT_CREATE */

  statTextWidget = widget;

}/* STAT_TEXT_CREATE */



/***************************************************************
 * stepMode_button_change() - Remember which parameter they want 
 * stepped in the movie module.
*/
void stepMode_button_change(Widget widget, int *tag, 
			    XmAnyCallbackStruct *callback_data)
{/* STEPMODE_BUTTON_CHANGE */
  XmString simpleString;
  int i;

  sprintf(outStr,"Theta Step:");
  XtSetArg(arglst[0], XmNsensitive, TRUE);
  XtSetValues(movieFormW, arglst, 1); 
  XtSetValues(stepToggleWidget, arglst, 1); 
  XtSetValues(stepToggle2Widget, arglst, 1); 
  XtSetValues(frameTextWidget, arglst, 1); 
  XtSetValues(stepTextWidget, arglst, 1); 
  XtSetValues(stepValueLabelWidget, arglst, 1); 
  XtSetValues(numFramesLabelW, arglst, 1); 
  XFlush(dsply);
  if(whichStep==*tag)
    {/* They are toggling the same one */
      if(whichStep != NONE)
	{/* They turned it off */
	  whichStep = NONE;
	  sprintf(outStr,"NO Step:");
	  XtSetArg(arglst[0], XmNsensitive, FALSE);
	  XtSetValues(stepToggleWidget, arglst, 1); 
	  XtSetValues(stepToggle2Widget, arglst, 1); 
	  XtSetValues(frameTextWidget, arglst, 1); 
	  XtSetValues(stepTextWidget, arglst, 1); 
	  XtSetValues(stepValueLabelWidget, arglst, 1); 
	  XtSetValues(numFramesLabelW, arglst, 1); 
	  XtSetValues(movieFormW, arglst, 1); 
	}/* They turned it off */
    }/* They are toggling the same one */
  else if(*tag==SURFDENS)
    {/* surf changed */
      sprintf(outStr,"SurfDens Step:");
      whichStep = SURFDENS;
    }/* surf changed */
  else if(*tag==FRACT)
    {/* fract changed */
      sprintf(outStr,"Fract Step:");
      whichStep = FRACT;
    }/* fract changed */
  else if(*tag==ANGX)
    {/* angx changed */
      sprintf(outStr,"Angx Step:");
      whichStep = ANGX;
    }/* angx changed */
  else if(*tag==ANGY)
    {/* angy changed */
      sprintf(outStr,"Angy Step:");
      whichStep = ANGY;
    }/* angy changed */
  else if(*tag==PHI)
    {/* phi changed */
      sprintf(outStr,"Phi Step:");
      whichStep = PHI;
    }/* phi changed */
  else if(*tag==OMEGA)
    {/* omega changed */
      sprintf(outStr,"Omega Step:");
      whichStep = OMEGA;
    }/* omega changed */
  else if(*tag==THETA)
    {/* theta changed */
      sprintf(outStr,"Theta Step:");
      whichStep = THETA;
    }/* theta changed */
  else if(*tag==UNBLOBSIZE)
    {/* unblobSize changed */
      sprintf(outStr,"UnblobSize Step:");
      whichStep = UNBLOBSIZE;
    }/* unblobSize changed */
  else if(*tag==UNBLOBRADIUS)
    {/* unblobRadius changed */
      sprintf(outStr,"UnblobRadius Step:");
      whichStep = UNBLOBRADIUS;
    }/* unblobRadius changed */
  else if(*tag==MAXX)
    {/* maxx changed */
      sprintf(outStr,"Cut X Step:");
      whichStep = MAXX;
    }/* maxx changed */
  else if(*tag==MAXY)
    {/* maxy changed */
      sprintf(outStr,"Cut Y Step:");
      whichStep = MAXY;
    }/* maxy changed */
  else if(*tag==MAXZ)
    {/* maxz changed */
      sprintf(outStr,"Cut Z Step:");
      whichStep = MAXZ;
    }/* maxz changed */
  else if(*tag==MAXRADIUS)
    {/* maxradius changed */
      sprintf(outStr,"MaxRad Step:");
      whichStep = MAXRADIUS;
    }/* maxradius changed */
  else if(*tag==PIXELSIZE)
    {/* maxradius changed */
      sprintf(outStr,"PixelSize Step:");
      whichStep = PIXELSIZE;
    }/* maxradius changed */
  
  if((*tag==SURFDENS) ||
     (*tag==FRACT) ||
     (*tag==ANGX) ||
     (*tag==ANGY) ||
     (*tag==THETA) ||
     (*tag==PHI) ||
     (*tag==OMEGA) ||
     (*tag==UNBLOBSIZE) ||
     (*tag==UNBLOBRADIUS) ||
     (*tag==MAXX) ||
     (*tag==MAXY) ||
     (*tag==MAXZ) ||
     (*tag==MAXRADIUS) ||
     (*tag==PIXELSIZE)
     )
    {/* maxradius changed */

      XtSetArg(arglst[0], XmNlabelPixmap, moviePixmap);
      XtSetArg(arglst[1], XmNselectPixmap, unmoviePixmap);
      XtSetArg(arglst[2], XmNrecomputeSize, FALSE);
      XtSetValues(surfDensToggleWidget, arglst, 3); 
      XtSetValues(fractToggleWidget, arglst, 3);
      XtSetValues(angxToggleWidget, arglst, 3);
      XtSetValues(angyToggleWidget, arglst, 3);
      XtSetValues(thetaToggleWidget, arglst, 3);
      XtSetValues(phiToggleWidget, arglst, 3);
      XtSetValues(omegaToggleWidget, arglst, 3);
      XtSetValues(maxRadiusToggleWidget, arglst, 3);
      XtSetValues(pixelSizeToggleWidget, arglst, 3);
      XtSetValues(unblobsizeToggleW, arglst, 3);
      XtSetValues(unblobradiusToggleW, arglst, 3);
      XtSetValues(maxxToggleW, arglst, 3);
      XtSetValues(maxyToggleW, arglst, 3);
      XtSetValues(maxzToggleW, arglst, 3);

      /* Now turn this one on */
      if(whichStep != NONE)
	{/* Set this one ON */
	  XtSetArg(arglst[0], XmNlabelPixmap, unmoviePixmap);
	  XtSetArg(arglst[1], XmNselectPixmap, moviePixmap);
	  XtSetArg(arglst[2], XmNrecomputeSize, FALSE);
	  XtSetValues(widget, arglst, 3); 
	}/* Set this one ON */

      /* Now make them decide on a new numFrames and step value */
      XmTextSetString(frameTextWidget, ""); 
      /* DON'T DO THIS ONE!!! XmTextSetString(stepTextWidget, "");  */

    }/* maxradius changed */

  createSimpleString(&simpleString,outStr);
  XtSetArg(arglst[0], XmNlabelString, simpleString );
  XtSetArg(arglst[1], XmNrecomputeSize, FALSE );
  XtSetValues(stepValueLabelWidget, arglst, 2); 
  XFlush(dsply);
  XmStringFree(simpleString); 
  
  if(immediateMapShade)
    mapShade_activate(0,&i,0);

}/* STEPMODE_BUTTON_CHANGE */




/****************************************************************
 * stepMode_change() - Remember which parameter they want stepped
 * in the movie module.
*/
void stepMode_change(Widget widget, int *tag, 
		     XmAnyCallbackStruct *callback_data)
{/* STEPMODE_CHANGE */
  XmString simpleString;
  int i;

  sprintf(outStr,"Theta Step:");
  if(whichStep==*tag)
    {/* They are toggling the same one */
      if(whichStep != NONE)
	{/* They turned it off */
	  whichStep = NONE;
	  sprintf(outStr,"NO Step:");
	}/* They turned it off */
    }/* They are toggling the same one */
  else if(*tag==SURFDENS)
    {/* surf changed */
      sprintf(outStr,"SurfDens Step:");
      whichStep = SURFDENS;
    }/* surf changed */
  else if(*tag==FRACT)
    {/* fract changed */
      sprintf(outStr,"Fract Step:");
      whichStep = FRACT;
    }/* fract changed */
  else if(*tag==ANGX)
    {/* angx changed */
      sprintf(outStr,"Angx Step:");
      whichStep = ANGX;
    }/* angx changed */
  else if(*tag==ANGY)
    {/* angy changed */
      sprintf(outStr,"Angy Step:");
      whichStep = ANGY;
    }/* angy changed */
  else if(*tag==THETA)
    {/* theta changed */
      sprintf(outStr,"Theta Step:");
      whichStep = THETA;
    }/* theta changed */
  else if(*tag==UNBLOBSIZE)
    {/* unblobSize changed */
      sprintf(outStr,"UnblobSize Step:");
      whichStep = UNBLOBSIZE;
    }/* unblobSize changed */
  else if(*tag==UNBLOBRADIUS)
    {/* unblobRadius changed */
      sprintf(outStr,"UnblobRadius Step:");
      whichStep = UNBLOBRADIUS;
    }/* unblobRadius changed */
  else if(*tag==MAXX)
    {/* maxx changed */
      sprintf(outStr,"Cut X Step:");
      whichStep = MAXX;
    }/* maxx changed */
  else if(*tag==MAXY)
    {/* maxy changed */
      sprintf(outStr,"Cut Y Step:");
      whichStep = MAXY;
    }/* maxy changed */
  else if(*tag==MAXZ)
    {/* maxz changed */
      sprintf(outStr,"Cut Z Step:");
      whichStep = MAXZ;
    }/* maxz changed */
  else if(*tag==MEMORYMISER)
    {/* Memory Miser */
      memoryMiser = !memoryMiser;
    }/* Memory Miser */
  else if(*tag==MAXRADIUS)
    {/* maxradius changed */
      sprintf(outStr,"MaxRad Step:");
      whichStep = MAXRADIUS;
    }/* maxradius changed */
  else if(*tag==PIXELSIZE)
    {/* maxradius changed */
      sprintf(outStr,"PixelSize Step:");
      whichStep = PIXELSIZE;
    }/* maxradius changed */
  
  /* These don't have to do with the whichStep ones */
  if(*tag==MAXRADIUSTOGGLE)
    {/* maxradius label toggle changed */
      maxRadiusSwitch = !maxRadiusSwitch;
      if(maxRadiusSwitch)
	sprintf(outStr,"MaxRadius:");
      else sprintf(outStr,"MinRadius:");
    }/* maxradius Lable togglechanged */
  else if(*tag==IREV)
    {/* irev changed */
      if(irev==-1)
	{/* reverse */
	  irev= 1;
	}/* reverse */
      else if(irev==1)
	{/* reverse */
	  irev= -1;
	}/* reverse */
      /* Give feedback on what the surface density *really* is... */
      /* Make sure the user sees it with the right sign */
      sprintf(outStr,"%.2f",irev*surfaceDensity);
      XmTextSetString(surfDensValueWidget,outStr); 
    }/* irev changed */
  else if(*tag==Y)
    {/* rotateAxis changed */
      if(rotateAxis==Y)
	{/* reverse */
	  sprintf(outStr,"X axis degrees");
	  rotateAxis= X;
	}/* reverse */
      else if(rotateAxis==X)
	{/* reverse */
	  sprintf(outStr,"Y axis degrees");
	  rotateAxis= Y;
	}/* reverse */
    }/* rotateAxis changed */

  
  if((*tag==SURFDENS) ||
     (*tag==FRACT) ||
     (*tag==ANGX) ||
     (*tag==ANGY) ||
     (*tag==THETA) ||
     (*tag==UNBLOBSIZE) ||
     (*tag==UNBLOBRADIUS) ||
     (*tag==MAXX) ||
     (*tag==MAXY) ||
     (*tag==MAXZ) ||
     (*tag==MAXRADIUS) ||
     (*tag==PIXELSIZE)
     )
    {/* maxradius changed */
      XtSetArg(arglst[0], XmNset, FALSE);
      XtSetValues(surfDensToggleWidget, arglst, 1);
      XtSetValues(fractToggleWidget, arglst, 1);
      XtSetValues(angxToggleWidget, arglst, 1);
      XtSetValues(angyToggleWidget, arglst, 1);
      XtSetValues(thetaToggleWidget, arglst, 1);
      XtSetValues(phiToggleWidget, arglst, 1);
      XtSetValues(omegaToggleWidget, arglst, 1);
      XtSetValues(maxRadiusToggleWidget, arglst, 1);
      XtSetValues(pixelSizeToggleWidget, arglst, 1);
      XtSetValues(unblobsizeToggleW, arglst, 1);
      XtSetValues(unblobradiusToggleW, arglst, 1);
      XtSetValues(maxxToggleW, arglst, 1);
      XtSetValues(maxyToggleW, arglst, 1);
      XtSetValues(maxzToggleW, arglst, 1);

      /* Now turn this one on */
      if(whichStep != NONE)
	{/* Set this one ON */
	  XtSetArg(arglst[0], XmNset, TRUE);
	  XtSetValues(widget, arglst, 1); 
	}/* Set this one ON */

      /* Now make them decide on a new numFrames and step value */
      XmTextSetString(frameTextWidget, ""); 
      /* DON'T DO THIS ONE!!! XmTextSetString(stepTextWidget, "");  */

    }/* maxradius changed */

  createSimpleString(&simpleString,outStr);
  XtSetArg(arglst[0], XmNlabelString, simpleString );
  XtSetArg(arglst[1], XmNrecomputeSize, FALSE );
  if(*tag == Y)
    XtSetValues(axisToggleWidget, arglst, 2); 
  else if(*tag==MAXRADIUSTOGGLE)
    XtSetValues(maxRadiusLabelToggleWidget, arglst, 2); 
  else if(*tag==IREV)
    {/* Do nothing */}
  else XtSetValues(stepValueLabelWidget, arglst, 2); 
  XFlush(dsply);
  XmStringFree(simpleString); 
  
  if(immediateMapShade)
    mapShade_activate(0,&i,0);

}/* STEPMODE_CHANGE */




/*************************************************************
 * stepMode_create() - Remember widget numbers!
*/
void stepMode_create(Widget widget, int *tag, 
		     XmAnyCallbackStruct *callback_data)
{/* STEPMODE_CREATE */

  if(*tag==SURFDENS)
    {/* surf created */
      surfDensToggleWidget= widget;
    }/* surf created */
  else if(*tag==FRACT)
    {/* fract created */
      fractToggleWidget= widget;
    }/* fract created */
  else if(*tag==ANGX)
    {/* angx created */
      angxToggleWidget= widget;
    }/* angx created */
  else if(*tag==ANGY)
    {/* angy created */
      angyToggleWidget= widget;
    }/* angy created */
  else if(*tag==THETA)
    {/* theta created */
      thetaToggleWidget= widget;
    }/* theta created */
  else if(*tag==PHI)
    phiToggleWidget= widget;
  else if(*tag==OMEGA)
    omegaToggleWidget= widget;
  else if(*tag==MAXRADIUS)
    {/* maxrad created */
      maxRadiusToggleWidget= widget;
    }/* maxrad created */
  else if(*tag==PIXELSIZE)
    {/* maxrad created */
      pixelSizeToggleWidget= widget;
    }/* maxrad created */
  else if(*tag==UNBLOBSIZE)
    unblobsizeToggleW= widget;
  else if(*tag==UNBLOBRADIUS)
    unblobradiusToggleW= widget;
  else if(*tag==MAXX)
    maxxToggleW= widget;
  else if(*tag==MAXY)
    maxyToggleW= widget;
  else if(*tag==MAXZ)
    maxzToggleW= widget;
  else if(*tag==MAXRADIUSTOGGLE)
    {/* maxrad created */
      maxRadiusLabelToggleWidget= widget;
    }/* maxrad created */
  else if(*tag==IREV)
    {/* irev created */
      irevToggleWidget= widget;
    }/* irev created */
  else if(*tag==Y)
    {/* irev created */
      axisToggleWidget= widget;
    }/* irev created */
  else if(*tag==MAPSHADEEXIT)
    {/* mapShadeExit created */
      mapShadeExitW= widget;
    }/* mapShadeExit created */


}/* STEPMODE_CREATE */




/***************************************************************
 * movie_contour_state() - They want to do something with their
 * section and contour plot.
*/
void movie_contour_state(Widget widget, int *tag, 
			 XmScrollBarCallbackStruct *callback_data)
{/* movie_contour_state */
  XmString simpleString;
  int i;
  char *s1,*s2;

  if(*tag==CLEAR)
    {/* clear */
      stuffDisplayMapBuffer(curBox,editBoxZoom,0);
      DrawEditBoxWindow(TRUE);
      XFlush(dsply);
    }/* clear */
  if(*tag==RESET)
    {/* clear */
      stuffDisplayMapBuffer(curBox,editBoxZoom,0);
      DrawEditBoxWindow(TRUE);
      XFlush(dsply);
      totContours = 0;
      sprintf(contourStr," ");
      XmTextSetString(contourTextWidget, contourStr); 
      XFlush(dsply);
    }/* clear */
  else if(*tag==USE)
    {/* use */
      int i;

      if(numCont > 0)
	{/* Do the range */
	  for(i=0;i<numCont;i++)
	    {/* Do all range */
	      conTours[totContours]= (contourLevel+((float)i*deltaCont));
	      if(totContours < 50)
		totContours++;
	      else askThem("MAX contour levels exceeded",FALSE,
			   XmDIALOG_INFORMATION);
	    }/* Do all range */
	}/* Do the range */
      else
	{/* Do individual */
	  conTours[totContours]= contourLevel;
	  if(totContours < 50)
	    totContours++;
	  else askThem("MAX contour levels exceeded",FALSE,
		       XmDIALOG_INFORMATION);
	}/* Do individual */	  
      
      s1= contourStr;
      for(i=0;i<totContours;i++)
	{/* do all */
	  sprintf(s1,"%d  ",conTours[i]);
	  s1= (char *)(contourStr+(int)strlen(contourStr));
	}/* do all */
      XmTextSetString(contourTextWidget, contourStr); 
      XFlush(dsply);
    }/* use */
  else if(*tag==WRITE)
    {/* contour write */
      char *thisStr;
      char numStr[25];
      int saveEditBoxZoom;
      float thisContour;

      /* Get a fresh buffer */
      saveEditBoxZoom= editBoxZoom;
      /* editBoxZoom = 1;  ONLY if you want to force it to 1 */
      stuffDisplayMapBuffer(curBox,editBoxZoom,0);

      /*thisStr = XmTextGetString(contourTextWidget); */
      for(i=0;i<totContours;i++)
	{/* Stuff all */

	  /*sscanf(thisStr,"%d",numStr);
	  thisContour = (float)strtod(numStr,0);*/
	  thisContour = conTours[i];
	  printf("%f\n",thisContour);

	  /* Now destructively draw the contour in the window buffer */
	  drawContour(thisContour,TRUE,curBox,editBoxWidget,
		      float_image_buffer1,0,0,
		      editBox_display_buffer);


	}/* Stuff all */
      /*      XtFree(thisStr); */

      calcBoxSizeFactor(editBoxZoom);
      writeRAWcontour("contourSection.raw",editBox_display_buffer,
		      (numCols*sizeFactor),(numRows*sizeFactor));

      editBoxZoom= saveEditBoxZoom;
    }/* contour write */

}/* movie_contour_state */



/**********************************************************
 * drawContour() - draw the contour over the image already 
 * displayed.
*/
void drawContour(float contourLvl, 
		 int destructive, 
		 int thisCurBox,
		 Widget thisWidget,
		 float *thisFloatBuf,
		 int deltaX, int deltaY,
		 BYTE *this_display_buffer)
{/* DRAWCONTOUR */
  int i,j,k,numPts,numColsTsizeFactor,nCols,nRows;
  int box_width;
  short *short_display_image;
  float *xp,*yp;
  float sclx,scly,tour;
  short ix[40000],iy[40000],ix1[40000],iy1[40000];
  float thisAng;
  float *thisBuffer;

  if(this_display_buffer == editBox_display_buffer)
    box_width = BOX_WIDTH;
  else box_width = buffer_width;

  cosang = 0.0;          /* 90.0 degrees */
  sinang = 1.0;          /* 90.0 degrees */
  icx    = 0.0;          /* Center of image */
  icy    = 0.0;
  
  /* ??? deg */
  cosang = (float)1.0;
  sinang = (float)1.0;
  
  /* These are for scalex and scale Y - get from zoomfactor..*/
  calcBoxSizeFactor(editBoxZoom);
  sclx = sizeFactor;
  scly = sizeFactor;
  
  tour = contourLvl;
  numPts = 1;

  for(i=0;i<40000;i++)
    {/* Init */
      ix[i]=iy[i]=ix1[i]=iy1[i]= 0;
    }/* Init */
  
  if(shadedMaps && !dispSections)
    {/* SHow the shaded map */
      int colOffset,rowOffset;

      CK(xp = (float *)malloc(sizeof(float)*((int)(largestDim)+1)));
      CK(yp = (float *)malloc(sizeof(float)*((int)(largestDim)+1)));
      CK(short_display_image = (short *)malloc(sizeof(short)*((int)(largestDim
								 )+1)*
					    ((int)(largestDim)+1)));
      
      /* Need numCols and numRows here because we want to use the section
       * contour info */
      thisBuffer= (float *)(thisFloatBuf+(cutMaxValues[2]*
					  numCols*numRows));
      nCols = numCols;
      nRows = numRows;

      contur_a_rob_(thisBuffer,
		    short_display_image,
		    xp, yp,
		    &nCols,&nRows,
		    &sclx,&scly,&tour, ix,iy,ix1,iy1, &numPts,
		    &cosang,&sinang,&icx,&icy);

      
      numColsTsizeFactor = numCols * sizeFactor;
      colOffset = (largestDim-numCols)/2;
      rowOffset = (largestDim-numRows)/2;
      colOffset *= sizeFactor;
      rowOffset *= sizeFactor;

      for(i=0;i<(numPts-1);i++)
	{/* Draw all points */
	  
	  if(ix[i] >= (numColsTsizeFactor*3) || 
	     ix1[i] >= (numColsTsizeFactor*3))
	    {/* Into next cell */
	      ix[i]  -= (numColsTsizeFactor*3);
	      ix1[i] -= (numColsTsizeFactor*3);
	    }/* Into next cell */
	  else if(ix[i] >= (numColsTsizeFactor*2) || 
		  ix1[i] >= (numColsTsizeFactor*2))
	    {/* Into next cell */
	      ix[i]  -= (numColsTsizeFactor*2);
	      ix1[i] -= (numColsTsizeFactor*2);
	    }/* Into next cell */
	  else if(ix[i] >= (numColsTsizeFactor) || 
		  ix1[i] >= (numColsTsizeFactor))
	    {/* Into next cell */
	      ix[i]  -= numColsTsizeFactor;
	      ix1[i] -= numColsTsizeFactor;
	    }/* Into next cell */

	  ix[i]  += colOffset;
	  ix1[i] += colOffset;
	  iy[i]  += rowOffset;
	  iy1[i] += rowOffset;

	  if(!destructive)
	    {/* Just draw with X */
	      if(contourLvl > breakCont)
		XDrawLine(dsply, XtWindow(thisWidget), 
			  image_gcRED,
			  (int)ix[i],(int)iy[i],
			  (int)ix1[i],(int)iy1[i]);
	      else if(contourLvl == breakCont)
		XDrawLine(dsply, XtWindow(thisWidget), 
			  image_gcBLUE,
			  (int)ix[i],(int)iy[i],
			  (int)ix1[i],(int)iy1[i]);
	      else
		XDrawLine(dsply, XtWindow(thisWidget), 
			  image_gcGREEN,
			  (int)ix[i],(int)iy[i],
			  (int)ix1[i],(int)iy1[i]);
	    }/* Just draw with X */
	  
	  
	  if(destructive)
	    {/* Now draw this in the buffer itself */
	      float slope,intercept;
	      int j,k,y,startX,endX;
	      
	      /* Corrupt it */
	      /* Need to draw a line with value==pixels_return[RED] or
	       * [GREEN] in the buffer below based on ix,iy,ix1,iy1 */
	      if(ix1[i]==ix[i])
		slope = 9999.9;
	      else slope = (float)(iy1[i] - iy[i]) / (float)(ix1[i] - ix[i]);
	      intercept = (float)iy[i] - (float)(slope * (float)ix[i]);
	      startX = MIN(ix[i],ix1[i]);
	      endX   = MAX(ix[i],ix1[i]);
	      
	      startX = MAX(startX,0);
	      endX   = MIN(endX,numColsTsizeFactor);

	      for(j=startX;j<=endX;j++)
		{/* Do all X */
		  y= (int)((slope * (float)j) + intercept);

		  y = MAX(y,0);
		  y = MIN(y,numColsTsizeFactor);
		  
		  if(contourLvl > 0.0)
		    this_display_buffer[((y*box_width)+j)]=
		      pixels_return[RED];
		  else if(contourLvl == 0.0)
		    this_display_buffer[((y*box_width)+j)]=
		      pixels_return[BLUE];
		  else
		    this_display_buffer[((y*box_width)+j)]=
		      pixels_return[GREEN];
		  
		}/* Do all X */
	      
	    }/* Now draw this in the buffer itself */
	  
	}/* Draw all points */
      
      if(destructive)
	{/* Draw it */
	  DrawEditBoxWindow(TRUE);
	  XFlush(dsply);
	}/* Draw it */
      
      free(xp);
      free(yp);
      free(short_display_image);

    }/* SHow the shaded map */
  else
    {/* Do sections */

      nCols = numCols;
      nRows = numRows;
  
      numColsTsizeFactor = numCols * sizeFactor;
      
      CK(xp = (float *)malloc(sizeof(float)*((int)(numCols)+1)));
      CK(yp = (float *)malloc(sizeof(float)*((int)(numRows)+1)));
      CK(short_display_image = (short *)malloc(sizeof(short)*
					       ((int)(numCols)+1)*
					    ((int)(numRows)+1)));
      thisBuffer= (float *)(thisFloatBuf+(thisCurBox*
					  numCols*numRows));

      contur_a_rob_(thisBuffer,
		    short_display_image,
		    xp, yp,
		    &nCols,&nRows,
		    &sclx,&scly,&tour, ix,iy,ix1,iy1, &numPts,
		    &cosang,&sinang,&icx,&icy);


      for(i=0;i<(numPts-1);i++)
	{/* Draw all points */
      
	  if(ix[i] >= (numColsTsizeFactor*3) || 
	     ix1[i] >= (numColsTsizeFactor*3))
	    {/* Into next cell */
	      ix[i]  -= (numColsTsizeFactor*3);
	      ix1[i] -= (numColsTsizeFactor*3);
	    }/* Into next cell */
	  else if(ix[i] >= (numColsTsizeFactor*2) || 
		  ix1[i] >= (numColsTsizeFactor*2))
	    {/* Into next cell */
	      ix[i]  -= (numColsTsizeFactor*2);
	      ix1[i] -= (numColsTsizeFactor*2);
	    }/* Into next cell */
	  else if(ix[i] >= (numColsTsizeFactor) || 
		  ix1[i] >= (numColsTsizeFactor))
	    {/* Into next cell */
	      ix[i]  -= numColsTsizeFactor;
	      ix1[i] -= numColsTsizeFactor;
	    }/* Into next cell */

	  ix[i]  += deltaX;
	  ix1[i] += deltaX;
	  iy[i]  += deltaY;
	  iy1[i] += deltaY;
	  
	  if(!destructive)
	    {/* Just draw with X */
	      if(contourLvl > breakCont)
		XDrawLine(dsply, XtWindow(thisWidget), 
			  image_gcRED,
			  (int)ix[i],(int)iy[i],
			  (int)ix1[i],(int)iy1[i]);
	      else if(contourLvl == breakCont)
		XDrawLine(dsply, XtWindow(thisWidget), 
			  image_gcBLUE,
			  (int)ix[i],(int)iy[i],
			  (int)ix1[i],(int)iy1[i]);
	      else
		XDrawLine(dsply, XtWindow(thisWidget), 
			  image_gcGREEN,
			  (int)ix[i],(int)iy[i],
			  (int)ix1[i],(int)iy1[i]);
	    }/* Just draw with X */
	  
	  
	  if(destructive)
	    {/* Now draw this in the buffer itself */
	      float slope,intercept;
	      int j,k,y,startX,endX;
	      
	      /* Corrupt it */
	      /* Need to draw a line with value==pixels_return[RED] or
	       * [GREEN] in the buffer below based on ix,iy,ix1,iy1 */
	      if(ix1[i]==ix[i])
		slope = 9999.9;
	      else slope = (float)(iy1[i] - iy[i]) / (float)(ix1[i] - ix[i]);
	      intercept = (float)iy[i] - (float)(slope * (float)ix[i]);
	      startX = MIN(ix[i],ix1[i]);
	      endX   = MAX(ix[i],ix1[i]);
	      
	      startX = MAX(startX,0);
	      endX   = MIN(endX,numColsTsizeFactor);

	      for(j=startX;j<=endX;j++)
		{/* Do all X */
		  y= (int)((slope * (float)j) + intercept);
		  
		  y = MAX(y,0);
		  y = MIN(y,numColsTsizeFactor);

		  if(contourLvl > 0.0)
		    this_display_buffer[((y*box_width)+j)]=
		      pixels_return[RED];
		  else if(contourLvl == 0.0)
		    this_display_buffer[((y*box_width)+j)]=
		      pixels_return[BLUE];
		  else
		    this_display_buffer[((y*box_width)+j)]=
		      pixels_return[GREEN];
		  
		}/* Do all X */
	      
	    }/* Now draw this in the buffer itself */
	  
	}/* Draw all points */
      
      if(destructive)
	{/* Draw it */
	  DrawEditBoxWindow(TRUE);
	  XFlush(dsply);
	}/* Draw it */
      
      free(xp);
      free(yp);
      free(short_display_image);
    }/* Do sections */

}/* DRAWCONTOUR */





/*******************************************************
 * drawFFTcontour() - Draw contour lines in fft contour
 * window.
*/
void drawFFTcontour()
{/* DRAWFFTCONTOUR */
  int numPts,numColsTsizeFactor,nCols,nRows;
  short *short_display_image;
  float *xp,*yp;
  float sclx,scly,tour;
  short ix[40000],iy[40000],ix1[40000],iy1[40000];
  float thisAng,minF,maxF;
  float *tempFFTcontour_buffer;

  if(contourToggle)
    {/* Only do when appropriate */
      cosang = 0.0;          /* 90.0 degrees */
      sinang = 1.0;          /* 90.0 degrees */
      icx    = 0.0;          /* Center of image */
      icy    = 0.0;
      
      /* ??? deg */
      cosang = (float)1.0;
      sinang = (float)1.0;
      
      /* These are for scalex and scale Y - get from zoomfactor..*/
      sclx = 1.0;
      scly = 1.0;

      nCols = 128;
      nRows = 128;
      
      numColsTsizeFactor = 128;
      
      CK(xp = (float *)malloc(sizeof(float)*(128+1)));
      CK(yp = (float *)malloc(sizeof(float)*(128+1)));
      CK(short_display_image = (short *)malloc(sizeof(short)*(128+1)*
					    (128+1)));
      CK(tempFFTcontour_buffer = (float *)malloc(sizeof(float)*(128+1)*
					      (128+1)));
      
      minF = 999999999.9;
      maxF = -999999999.9;
      minF = HUGE;   /* dgm HUGE instead of 9999999.9 */
      maxF = -HUGE;  /* dgm -HUGE instead of -9999999.9 */
      for(i=0;i<128;i++)
	for(j=0;j<128;j++)
	  {/* DO all */
	    tempFFTcontour_buffer[((i*128)+j)]= 
	      (float)fft_contour_buffer[(((i*1)*128)+(j*1))];
	    minF= MIN(tempFFTcontour_buffer[((i*128)+j)],minF);
	    maxF= MAX(tempFFTcontour_buffer[((i*128)+j)],maxF);
	  }/* DO all */

      for(tour = (maxF-1.0);tour > (minF);tour -= ((maxF-minF)/4.0))
	{/* DO alot of contours */
	  
	  numPts = 1;

	  for(i=0;i<40000;i++)
	    {/* Init */
	      ix[i]=iy[i]=ix1[i]=iy1[i]= 0;
	    }/* Init */
	  
	  contur_a_rob_(tempFFTcontour_buffer,
			short_display_image,
			xp, yp,
			&nCols,&nRows,
			&sclx,&scly,&tour, ix,iy,ix1,iy1, &numPts,
			&cosang,&sinang,&icx,&icy);
	  
	  for(i=0;i<(numPts-1);i++)
	    {/* Draw all points */
	      
	      if(ix[i] >= (numColsTsizeFactor*3) || 
		 ix1[i] >= (numColsTsizeFactor*3))
		{/* Into next cell */
		  ix[i]  -= (numColsTsizeFactor*3);
		  ix1[i] -= (numColsTsizeFactor*3);
		}/* Into next cell */
	      else if(ix[i] >= (numColsTsizeFactor*2) || 
		      ix1[i] >= (numColsTsizeFactor*2))
		{/* Into next cell */
		  ix[i]  -= (numColsTsizeFactor*2);
		  ix1[i] -= (numColsTsizeFactor*2);
		}/* Into next cell */
	      else if(ix[i] >= (numColsTsizeFactor) || 
		      ix1[i] >= (numColsTsizeFactor))
		{/* Into next cell */
		  ix[i]  -= numColsTsizeFactor;
		  ix1[i] -= numColsTsizeFactor;
		}/* Into next cell */
	      

	      ix[i]  *= 1;    /* Put it back into 128 space */
	      ix1[i] *= 1;
	      iy[i]  *= 1;
	      iy1[i] *= 1;

	      /* Just draw with X */
	      if(tour < (maxF*0.5))
		XDrawLine(dsply, fftContour_window1, 
			  image_gcRED,
			  (int)ix[i],(int)iy[i],
			  (int)ix1[i],(int)iy1[i]);
	      else if((tour < (maxF-1.0)) && (tour >= (maxF*0.5)))
		XDrawLine(dsply, fftContour_window1, 
			  image_gcBLUE,
			  (int)ix[i],(int)iy[i],
			  (int)ix1[i],(int)iy1[i]);
	      else
		XDrawLine(dsply, fftContour_window1, 
			  image_gcGREEN,
			  (int)ix[i],(int)iy[i],
			  (int)ix1[i],(int)iy1[i]);
	      
	    }/* Draw all points */
	  
	}/* DO alot of contours */
      
      free(xp);
      free(yp);
      free(short_display_image);
    }/* Only do when appropriate */


}/* DRAWFFTCONTOUR */



/**************************************************************
 * RwaActivateWidget() - Activate the widget passed in.
 * This routine makes the interface behave just as if you
 * had pressed a push button to activate it.
*/
void RwaActivateWidget(Widget w)
{/* RWAACTIVATEWIDGET */
  Arg arglst[2];

  if(XmIsGadget(w))
    {/* IS a gadget */
      (*((XmGadgetClass)XtClass(w))->gadget_class.arm_and_activate)(w,0,0,0);
      return;
    }/* IS a gadget */
  else if(XmIsPrimitive(w))
    {/* IS a gadget */
      (*((XmPrimitiveWidgetClass)
	 XtClass(w))->primitive_class.arm_and_activate)(w,0,0,0);
      return;
    }/* IS a gadget */
  
}/* RWAACTIVATEWIDGET */




/*************************************************************
 * movieMinMax_change() - They want to limit the movie frames
 * from min to max.
*/
void movieMinMax_change(Widget widget, int *tag, 
			XmScrollBarCallbackStruct *callback_data)
{/* MOVIEMINMAX_CHANGE */
  XmString simpleString;

  if(*tag==MOVIEMIN)
    {/* surf changed */
      startMovie = callback_data->value;
      startMovie = MIN(startMovie,endMovie);
      XtVaSetValues(movieMinWidget, XmNvalue, startMovie, NULL);
    }/* surf changed */
  else if(*tag==MOVIEMAX)
    {/* surf changed */
      endMovie = callback_data->value;
      endMovie = MAX(endMovie,startMovie);
      XtVaSetValues(movieMaxWidget, XmNvalue, endMovie, NULL);
    }/* surf changed */
  else if(*tag==BREAKPOINT)
    {/* delta changed */

      breakCont = (float)contourLevel;

      sprintf(outStr,"%.0f",breakCont);
      createSimpleString(&simpleString,outStr);
      XtSetArg(arglst[0], XmNlabelString, simpleString);
      XtSetValues(movieBreakLabelWidget, arglst, 1); 
      XFlush(dsply);
      XmStringFree(simpleString); 
      XFlush(dsply);

    }/* delta changed */
  else if(*tag==DELTA)
    {/* delta changed */
      deltaCont = (float)callback_data->value;

      sprintf(outStr,"%.0f",deltaCont);
      createSimpleString(&simpleString,outStr);
      XtSetArg(arglst[0], XmNlabelString, simpleString);
      XtSetValues(movieDeltaLabelWidget, arglst, 1); 
      XFlush(dsply);
      XmStringFree(simpleString); 
      XFlush(dsply);

    }/* delta changed */
  else if(*tag==NUMCONT)
    {/* numCOnt changed */
      numCont = (int)callback_data->value;

      sprintf(outStr,"%d",numCont);
      createSimpleString(&simpleString,outStr);
      XtSetArg(arglst[0], XmNlabelString, simpleString);
      XtSetValues(movieNumContLabelWidget, arglst, 1); 
      XFlush(dsply);
      XmStringFree(simpleString); 
      XFlush(dsply);

    }/* numCOnt changed */
  else if(*tag==CONTOUR || *tag==TRYIT)
    {/* contour changed */
      
      if(*tag==CONTOUR)
	{/* New contour */

	  stuffDisplayMapBuffer(curBox,editBoxZoom,0);
	  DrawEditBoxWindow(TRUE);
	  XFlush(dsply);

	  contourLevel = (float)callback_data->value;
	  
	  sprintf(outStr,"  %.0f  ",contourLevel);
	  createSimpleString(&simpleString,outStr);
	  XtSetArg(arglst[0], XmNlabelString, simpleString);
	  XtSetValues(movieContourLabelWidget, arglst, 1); 
	  XFlush(dsply);
	  XmStringFree(simpleString); 
	  XFlush(dsply);

	  /* Non-destructively draw it */
	  XFlush(dsply);
	  drawContour(contourLevel, FALSE, curBox, editBoxWidget,
		      float_image_buffer1,0,0,
		      editBox_display_buffer);

	}/* New contour */

      if(*tag==TRYIT)
	{/* Now let's draw the contour */
	  int i;

	  if(crysDisping)
	    {/* Special for crysDisp */
	      for(i=0;i<totContours;i++)
		{/* Non-destructively draw it */
		  drawContour(conTours[i],
			      TRUE, 0, drawingWidget,
			      crys_float_image_buffer,
			      0,0,image_buffer1);
		}/* Non-destructively draw it */
	    }/* Special for crysDisp */
	  else
	    {/* Normal */
	      for(i=0;i<numCont;i++)
		{/* DO all contours */
		  
		  /* Non-destructively draw it */
		  drawContour((contourLevel+((float)i*deltaCont)), 
			      FALSE, curBox, editBoxWidget,float_image_buffer1,
			      0,0,editBox_display_buffer);
		  
		}/* DO all contours */
	      
	    }/* Normal */
	  
	}/* Now let's draw the contour */


    }/* contour changed */


  if(*tag == MOVIEMAX || *tag == MOVIEMIN)
    {/* OK to do this */
      int tag;
      sprintf(outStr,"%d",startMovie+1);  /* +1 for humans */
      createSimpleString(&simpleString,outStr);
      XtSetArg(arglst[0], XmNlabelString, simpleString);
      XtSetValues(minLabelWidget, arglst, 1); 
      XFlush(dsply);
      XmStringFree(simpleString); 
      
      sprintf(outStr,"%d",endMovie+1);    /* +1 for humans */
      createSimpleString(&simpleString,outStr);
      XtSetArg(arglst[0], XmNlabelString, simpleString);
      XtSetValues(maxLabelWidget, arglst, 1); 
      XFlush(dsply);
      XmStringFree(simpleString); 
      tag= INFINIMAP;
      
      /* How do we get it to start the infiniLoop again??? */
      /* mapMovie_show_activate(0, &tag, 0); */
      /* RwaActivateWidget(infiniWidget); This messes up the mwm  */
    }/* OK to do this */


}/* MOVIEMINMAX_CHANGE */







/* ************************************************************
 * circAvg_screen_deactivate() - Get rid of the circAvg screen.
 */
void circAvg_screen_deactivate(Widget widget, int *tag, 
			       XmAnyCallbackStruct *callback_data)
{/* circAvg_screen_deactIVATE */

  XtUnmanageChild(circAvgScreenWidget);
  
}/* circAvg_screen_deactIVATE */




/* ************************************************************
 * circAvg_pifOut_activate() - PifOut the circAvg
 */
void circAvg_pifOut_activate(Widget widget, int *tag, 
			     XmAnyCallbackStruct *callback_data)
{/* CIRCAVG_PIFOUT_ACTIVATE */
  int i,j,thisNterp, xcen,ycen, dist, luNum, imgOffset;
  float *thisRint,*thisStDev;
  fString_t fName,readWrite;
  char outFileName[256];
  BYTE *outBuf;
  float *outBufF;
  forPifHead_t thisHead;
  float  scaleFactor, img_bkg, img_min,
    img_max, img_avg, img_std, maxF,minF;

  fName.addr = (char *)malloc(sizeof(char)*64);
  readWrite.addr = (char *)malloc(sizeof(char)*64);

  pif_init_head_(&thisHead);

  thisHead.numCols =  numCols;
  thisHead.numRows =  numSecs;
  CK(outBuf = (BYTE *)malloc(sizeof(BYTE)*
				  thisHead.numRows*
				  thisHead.numCols));
  CK(outBufF = (float *)malloc(sizeof(float)*
				  thisHead.numRows*
				  thisHead.numCols));

  xcen = numCols/2;
  ycen = numRows/2;
  thisRint = circAvg[selectedCircAvg].rint;
  for(j=0;j<numRows;j++)
    {/* Rows */
      
      for(i=0;i<numCols;i++)
	{/* Do entire image */
	  dist = (int)sqrt((float)(((j-ycen)*(j-ycen))+((i-xcen)*(i-xcen))));
	  outBufF[((j*numCols)+i)]= (float)thisRint[dist];

	}/* Do entire image */

    }/* Rows */
  
  calcStats(outBufF,
		&img_bkg, &img_min,
		&img_max, &img_avg, &img_std,
		thisHead.numCols,thisHead.numRows,
		thisHead.numCols);
  
  minF = (float)img_min;
  maxF = (float)img_max;
  scaleFactor = 255.0 / (maxF-minF);

  /* Now scale it */
  for(j=0;j<(thisHead.numRows);j++)
    {/* Rows */
      for(i=0;i<(thisHead.numCols);i++)
	{/* Do entire image */
	  outBuf[((j*thisHead.numCols)+i)]= (BYTE)
	    (((float)outBufF[((j*thisHead.numCols)+i)]-minF) * 
	     scaleFactor);
	}/* Do entire image */
    }/* Rows */
  calcByteStats(outBuf,
		&img_bkg, &img_min,
		&img_max, &img_avg, &img_std,
		thisHead.numCols,thisHead.numRows,
		thisHead.numCols);


  sprintf(outFileName,"spherAver.pif");
  sprintf(fName.addr,outFileName);
  sprintf(readWrite.addr,"w");

  luNum = 7;
  thisHead.luNum = luNum;

  pif_open_(fName.addr, &luNum, readWrite.addr);


  thisHead.nimg = 1;
  thisHead.htype= 1;
  thisHead.numSecs =  1;  /* No maps here */

  thisHead.mode = 6;

  pif_write_gh_(&thisHead);

      
  thisHead.datai4_min = (int)img_min;
  thisHead.datar4_min = (int)img_min;
  thisHead.datai4_max = (float)img_max;
  thisHead.datar4_max = (float)img_max;
  thisHead.datar4_avg = img_avg;
  thisHead.datar4_std = img_std;
  thisHead.img_bkg = img_bkg;
  thisHead.imgNum = 1;

  pif_write_dh_(&thisHead);
      
  /* [3] Now write the image info */
  imgOffset = 1024;  

  pif_write_byte_image_(&(thisHead.luNum),
			outBuf,
			&imgOffset,
			&(thisHead.numCols),
			&(thisHead.numRows),
			&(thisHead.numSecs));
  
  free(fName.addr);
  free(readWrite.addr);
  free(outBuf);
  free(outBufF);
  
  pif_close(&thisHead.luNum);
  
  askThem("spherAver.pif file writtten.",FALSE,
	  XmDIALOG_INFORMATION);
  

  
}/* CIRCAVG_PIFOUT_ACTIVATE */



/* ************************************************************
 * circAvg_print_activate() - Print the circAvg
 */
void circAvg_print_activate(Widget widget, int *tag, 
			    XmAnyCallbackStruct *callback_data)
{/* CIRCAVG_PRINT_ACTIVATE */
  int i,j,thisNterp;
  float *thisRint,*thisStDev;
  FILE *fOut;

  fOut = fopen("spherical_average.dat","w");

  if(fOut)
    {/* Do it */
      for(j=0;j<currentCircAvg;j++)
	{/* DO all curves */
	  thisRint   = circAvg[j].rint;
	  thisStDev  = circAvg[j].stDev;
	  thisNterp  = circAvg[j].nterp;
	  if(fOut)
	    fprintf(fOut,"Map [%s] - number of points:[%d]\n",
		    circAvg[j].mapName,
		    circAvg[j].nterp
		    );
	  for(i=0;i<thisNterp;i++)
	    {/* DO all */
	      fprintf(fOut,"%f  %f\n",thisRint[i],thisStDev[i]);
	    }/* DO all */
	}/* DO all curves */
      fclose(fOut);
      askThem("spherical_average.dat file writtten.",FALSE,
	      XmDIALOG_INFORMATION);

    }/* Do it */
  else
    askThem("Unable to open spherical_average.dat file.",FALSE,
	    XmDIALOG_INFORMATION);
  
}/* CIRCAVG_PRINT_ACTIVATE */



/* ******************************************************************
 * circAvg_screen_activate() - Put the circAvgscreen on the display.
 */
void circAvg_screen_activate(Widget widget, int *tag, 
			     XmAnyCallbackStruct *callback_data)
{/* CIRCAVG_SCREEN_ACTIVATE */
  int i,j,thisNterp;
  float *thisRint;

  /* Now draw the circular average already calculated when the MAP 
   * was read in. */
  minCA= r4_max;
  maxCA= r4_min;
  for(j=0;j<currentCircAvg;j++)
    {/* DO all curves */
      thisRint  = circAvg[j].rint;
      thisNterp = circAvg[j].nterp;
      for(i=0;i<thisNterp;i++)
	{/* DO all */
	  if(thisRint[i]!=0.0)
	    {/* OK to use */
	      minCA = MIN(minCA,thisRint[i]);
	      maxCA = MAX(maxCA,thisRint[i]);
	    }/* OK to use */
	}/* DO all */
    }/* DO all curves */

  scaleCircFactor = 245.0 / ((float)(maxCA-minCA)); 

  raiseIt(circAvgScreenWidget);

}/* CIRCAVG_SCREEN_ACTIVATE */



/* ************************************************************
 * DrawCircAvgWindow() - Redraw window e.g. on expose events
 * or whenever it's prudent.
 */
void DrawCircAvgWindow(Widget widget, int *tag, 
		       XmAnyCallbackStruct *callback_data)
{/* DrawCircAvgWindow */
  int draw_width, draw_height;
  int i,j,scaledY,scaledYp1,thisX,thisXp1,thisNterp;
  float *thisRint;
  XmString simpleString;

  if(*tag==CLEAR)
    {/* Clear it out */
      for(j=0;j<3;j++)
	{/* i */
	  sprintf(outStr," ");
	  createSimpleString(&simpleString,outStr);
	  XtSetArg(arglst[0], XmNlabelString, simpleString);
	  XtSetArg(arglst[1], XmNrecomputeSize, FALSE );
	  XtSetValues(circAvgLabelW[j], arglst, 2); 
	  XFlush(dsply);
	  XmStringFree(simpleString); 
	  sprintf(outStr,"---");
	  createSimpleString(&simpleString,outStr);
	  XtSetArg(arglst[0], XmNlabelString, simpleString);
	  XtSetArg(arglst[1], XmNrecomputeSize, FALSE );
	  XtSetArg(arglst[2], XmNforeground, 
		   drawCircs[j].pixelColor );
	  XtSetValues(circAvgLineW[j], arglst, 3); 
	  XFlush(dsply);
	  XmStringFree(simpleString); 
	}/* i */
      numCircAvgDraw= 0; 
    }/* Clear it out */

  draw_width = 256;
  draw_height = 256;
  if(XtIsManaged(circAvgScreenWidget))
    XPutImage(dsply, circAvg_window1, circAvg_gc, circAvg_image1, 0, 0,
	      0, 0,  draw_width,  draw_height);

  /* Now draw axes */
  XDrawLine(dsply, XtWindow(circAvgWidget), image_gcWHITE,
	    5,0,5,256);
  XDrawLine(dsply, XtWindow(circAvgWidget), image_gcWHITE,
	    0,251,256,251);

  /* Draw zero line */
  scaledY = 251.0 - ((0.0-minCA)*scaleCircFactor);
  XDrawLine(dsply, XtWindow(circAvgWidget), image_gcWHITE,
	    6,(int)scaledY,1024,(int)scaledY);

  XFlush (dsply);
 
  for(j=0;j<numCircAvgDraw;j++)
    {/* Do all */
      
      thisRint  = circAvg[drawCircs[j].index].rint;
      thisNterp = circAvg[drawCircs[j].index].nterp;
      
      for(i=caOffsetX;i<(thisNterp-1);i++)
	{/* DO all */
	  scaledY = (thisRint[i]-minCA)*scaleCircFactor;
	  scaledYp1 = (thisRint[i+1]-minCA)*scaleCircFactor;
	  scaledY = (245.0 - scaledY)+6.0;
	  scaledYp1 = (245.0 - scaledYp1)+6.0;
	  thisX= ((i-caOffsetX)*circAvgSize)+6;
	  thisXp1= ((i-caOffsetX)*circAvgSize)+7;
	  XDrawLine(dsply, XtWindow(circAvgWidget), 
		    drawCircs[j].colorGC,
		    thisX,(int)scaledY,(thisXp1),(int)scaledYp1);
	}/* DO all */
      
    }/* Do all */
  
  XFlush (dsply);

}/* DrawCircAvgWindow */





/* ************************************************************
 * circAvg_list_proc() - Call here everytime they make a new
 * choice form the circAvg MASTER list.  Plot the appropriate
 * circAvg already calculated when the MAP was read in.
*/
static void circAvg_list_proc(Widget widget, int *tag, 
			      XmListCallbackStruct *list)
{/* CIRCAVG_LIST_PROC */
  Arg arglst[5];
  int i, j, foundOne, foundIt, thisColor, thisEntry,passTag;
  float scaledY,scaledYp1;
  float *thisRint;
  GC thisGC;
  char *thisStr;
  XmString list_txt;
  XmString simpleString;
  char tempStr[256];

  foundOne= (int) (list->item_position);   /* Numbers from 1 not 0 */
  foundOne--;  /* COmputer numbers from 0 not 1 */
  selectedCircAvg= foundOne;

  /* Now point to the one we want */
  thisRint = circAvg[foundOne].rint;

  /* We'll always fill up the last place regardless */
  if(numCircAvgDraw==3)
    numCircAvgDraw= 2;

  if(numCircAvgDraw==0)
    {
      thisGC = image_gcRED;
      drawCircs[numCircAvgDraw].colorGC = image_gcRED;
    }
  else if(numCircAvgDraw==1)
    {
      thisGC = image_gcGREEN;
      drawCircs[numCircAvgDraw].colorGC = image_gcGREEN;
    }
  else if(numCircAvgDraw==2)
    {
      thisGC = image_gcBLUE;
      drawCircs[numCircAvgDraw].colorGC = image_gcBLUE;
    }

  /* Remember these */
  if(!XmStringGetLtoR(list->item, 
		      XmSTRING_DEFAULT_CHARSET, 
		      &(thisStr)))
    {printf("Error extracting filename.\n");}
  sprintf(drawCircs[numCircAvgDraw].fName,thisStr);
  XtFree(thisStr);
  drawCircs[numCircAvgDraw].index   = foundOne;
  drawCircs[numCircAvgDraw].pixelColor = pixels_return[(RED+numCircAvgDraw)];

  sprintf(outStr,"[%s] ",drawCircs[numCircAvgDraw].fName);
  createSimpleString(&simpleString,outStr);
  XtSetArg(arglst[0], XmNlabelString, simpleString);
  XtSetArg(arglst[1], XmNrecomputeSize, FALSE );
  XtSetValues(circAvgLabelW[numCircAvgDraw], arglst, 2); 
  XFlush(dsply);
  XmStringFree(simpleString); 
  
  sprintf(outStr,"---");
  createSimpleString(&simpleString,outStr);
  XtSetArg(arglst[0], XmNlabelString, simpleString);
  XtSetArg(arglst[1], XmNrecomputeSize, FALSE );
  XtSetArg(arglst[2], XmNforeground, 
	   drawCircs[numCircAvgDraw].pixelColor );
  XtSetValues(circAvgLineW[numCircAvgDraw], arglst, 3); 
  XFlush(dsply);
  XmStringFree(simpleString); 

  numCircAvgDraw++;
  if(numCircAvgDraw > 2)
    numCircAvgDraw= 3;

  passTag= 0;
  DrawCircAvgWindow(0,&passTag,0);

}/* CIRCAVG_LIST_PROC */






/* @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
 * circAvgMotion() - Whenever the pointer moves across our 
 * CircAvg drawing area(s) call here.  This rotuine
 * basically takes care of drawing that little white line 
 * you see when the cursor is dragged across the circAvg
 * areas.
*/
void circAvgMotion(Widget widget, caddr_t unused, 
		   register XMotionEvent *pEvent)
{/* circAvgMotion */
  XmString simpleString;
  int x, y,lastDraw,foundIt,thisEntry;
  static int lastX= -999;
  static int lastI= -999;
  static short lastRed,lastGreen,lastBlue;
  int i, j, thisI;
  float *thisRint;

  /* Make all changes current */
  XFlush(dsply);               

  switch (pEvent->type) 
    {/* Find out which event we have */
	  
    case LeaveNotify:

      if(lastX > -99)
	{/* redraw last one */
	  XDrawLine(dsply, XtWindow(circAvgWidget), image_gcBLACK,
		    lastX,252,lastX,256);
	  XDrawLine(dsply, XtWindow(circAvgWidget), image_gcBLACK,
		    lastX,1,lastX,6);
	}/* redraw last one */
      if(lastI > -99)
	{/* Redraw the last one */
	  color_defs[lastI].red    = lastRed;
	  color_defs[lastI].green  = lastGreen;
	  color_defs[lastI].blue   = lastBlue;
	  if(masterProcess)
	    {
	      if(hiColor)
		{XAllocColor (dsply, color_map, &color_defs[lastI]);
		 pixels_return[lastI] = color_defs[lastI].pixel;}
	      else
		XStoreColor (dsply, color_map, &color_defs[lastI]);
	    }
	}/* Redraw the last one */
      
      break;

    case MotionNotify:
	      
      /* Get coords of pointer */
      x = (int)(pEvent->x);     
      y = (int)(pEvent->y);
      
      if(lastX > -99)
	{/* redraw last one */
	  XDrawLine(dsply, XtWindow(circAvgWidget), image_gcBLACK,
		    lastX,252,lastX,256);
	  XDrawLine(dsply, XtWindow(circAvgWidget), image_gcBLACK,
		    lastX,1,lastX,6);
	}/* redraw last one */
      XDrawLine(dsply, XtWindow(circAvgWidget), image_gcWHITE,
		x,252,x,256);
      XDrawLine(dsply, XtWindow(circAvgWidget), image_gcWHITE,
		x,1,x,6);
      
      for(i=0;i<numCircAvgDraw;i++)
	{/* Do all */
	  int thisX,thisNterp;
	  float *thisRint;
	  thisX= ((x-6)*(1.0/circAvgSize))+caOffsetX;
	  thisRint  = circAvg[drawCircs[i].index].rint;
	  thisNterp = circAvg[drawCircs[i].index].nterp;
	  if(thisX < thisNterp)
	    sprintf(outStr,"[%s] %d,%.2f",drawCircs[i].fName,thisX,
		    thisRint[thisX]);
	  else
	    sprintf(outStr,"[%s] %d,UNDEF",drawCircs[i].fName,thisX);
	  createSimpleString(&simpleString,outStr);
	  XtSetArg(arglst[0], XmNlabelString, simpleString);
	  XtSetArg(arglst[1], XmNrecomputeSize, FALSE );
	  XtSetValues(circAvgLabelW[i], arglst, 2); 
	  XFlush(dsply);
	  XmStringFree(simpleString); 

	  sprintf(outStr,"---");
	  createSimpleString(&simpleString,outStr);
	  XtSetArg(arglst[0], XmNlabelString, simpleString);
	  XtSetArg(arglst[1], XmNrecomputeSize, FALSE );
	  XtSetArg(arglst[2], XmNforeground, 
		   drawCircs[i].pixelColor );
	  XtSetValues(circAvgLineW[i], arglst, 3); 
	  XFlush(dsply);
	  XmStringFree(simpleString); 

	}/* Do all */
      
      if(radialMapShade && !dispSections && selectedCircAvg>=0)
	{/* Change the colormap at this radius */
	  if(lastI > -99)
	    {/* Redraw the last one */
	      color_defs[lastI].red    = lastRed;
	      color_defs[lastI].green  = lastGreen;
	      color_defs[lastI].blue   = lastBlue;
	      if(masterProcess)
		{
		  if(hiColor)
		    {XAllocColor (dsply, color_map, &color_defs[lastI]);
		     pixels_return[lastI] = color_defs[lastI].pixel;}
		  else
		    XStoreColor (dsply, color_map, &color_defs[lastI]);
		}
	    }/* Redraw the last one */
	  /* (/ circAvgSize) because we use (* circAvgSize) for display 
	   * in circAvg_list_proc()   */
	  y = (x - 6) / circAvgSize;
	  y = MAX(y,0);
	  i= MIN(tryColors-1,MAX(y,6));
	  /*i += pixels_return[0];  */
	  thisI= i;

	  thisRint = circAvg[selectedCircAvg].rint;
	  foundIt = FALSE;	
	  for(thisEntry=0;thisEntry<256;thisEntry++)
	    if(!foundIt)
	      {/* see if this is it */
		if((radialValue[thisEntry]<= (i)) &&  
		   (radialValue[thisEntry+1]>= (i)) ) 
		  {
		    thisI= thisEntry+pixels_return[0];
		    foundIt= TRUE;
		  }
	      }/* see if this is it */

	  /* Make sure we're in range */
	  thisI= MIN(pixels_return[tryColors-1],MAX(thisI,pixels_return[6]));

	  /* Tell them what position in the colormap we are */
	  sprintf(outStr,"%3d",thisI);
	  createSimpleString(&simpleString,outStr);
	  XtSetArg(arglst[0], XmNlabelString, simpleString);
	  XtSetArg(arglst[1], XmNrecomputeSize, FALSE );
	  XtSetValues(colormapValueWidget[ENTRY], arglst, 2); 
	  XFlush(dsply);
	  XmStringFree(simpleString); 

	  if(thisI==pixels_return[0]+6)
	    sprintf(outStr,"%3d",0);
	  else sprintf(outStr,"%.1f",radialValue[thisI]);

	  createSimpleString(&simpleString,outStr);
	  XtSetArg(arglst[0], XmNlabelString, simpleString);
	  XtSetArg(arglst[1], XmNrecomputeSize, FALSE );
	  XtSetValues(radialValueWidget, arglst, 2); 
	  XFlush(dsply);
	  XmStringFree(simpleString); 

	  if(foundIt)
	    {/* Then we can change a color */
	      lastRed  = color_defs[thisI].red;
	      lastGreen= color_defs[thisI].green;
	      lastBlue = color_defs[thisI].blue;
	      color_defs[thisI].red    = 65000;
	      color_defs[thisI].green  = 0;
	      color_defs[thisI].blue   = 0;
	      if(masterProcess)
		{
		  if(hiColor)
		    {XAllocColor (dsply, color_map, &color_defs[thisI]);
		     pixels_return[thisI] = color_defs[thisI].pixel;}
		  else
		    XStoreColor (dsply, color_map, &color_defs[thisI]);
		}
	      
	      lastI = thisI;
	    }/* Then we can change a color */

	}/* Change the colormap at this radius */

      lastX = x;
      
      break;
	  
    default:
      break;
    }/* Find out which event we have */
  
  XFlush(dsply);              /* Make changes current */
  
}/* circAvgMotion */




/* ************************************************************
 * circAvgArrowActivate() - They want to change the current
 * circavg displayed by using the arrows.
*/
void circAvgArrowActivate(Widget widget, int *tag, 
			  XmAnyCallbackStruct *callback_data)
{/* circAvgARROWACTIVATE */
  int passTag;

  /* Which arrow? */
  if(*tag==XmARROW_UP)
    {/* Up */
      circAvgSize++;
    }/* Up */
  else if(*tag==XmARROW_DOWN)
    {/* Down */
      circAvgSize--;
    }/* Down */
  else printf("circAvg unchanged.\n");
  circAvgSize = MAX(1,MIN(circAvgSize, 4));

  sprintf(outStr,"%d",circAvgSize);
  XmTextSetString(circAvgTextWidget,outStr); 
  
  passTag= 0;
  DrawCircAvgWindow(0,&passTag,0);

}/* circAvgARROWACTIVATE */



/*************************************************************
 * circAvg_change() - They want to change the offset in the X
 * direction for the circAvg screen.
*/
void circAvg_change(Widget widget, int *tag, 
		    XmScrollBarCallbackStruct *callback_data)
{/* CIRCAVG_CHANGE */
  XmString simpleString;
  int passTag;
  float *thisRint;
  int thisNterp;

  if(*tag==CAINVERT)
    {/* toggle invert contrast */
      caInvertContrast = !caInvertContrast;

      for(j=0;j<numCircAvgDraw;j++)
	{/* Do all */
	  
	  thisRint  = circAvg[drawCircs[j].index].rint;
	  thisNterp = circAvg[drawCircs[j].index].nterp;
	  
	  for(i=caOffsetX;i<(thisNterp);i++)
	    {/* DO all */
	      thisRint[i] *= -1.0;
	    }/* DO all */
	  
	}/* Do all */
      
    }/* toggle invert contrast */
  else
    {/* DO scrollbar */
      caOffsetX = (int)callback_data->value;
      
    }/* DO scrollbar */
  
  /* Recalc the Y scaling factors based on this new offset.
   * We may be able to show more of the curves out here if
   * we've got a huge noise spike at the center of the map
   * e.g. */
  minCA= r4_max;
  maxCA= r4_min;
  for(j=0;j<currentCircAvg;j++)
    {/* DO all curves */
      thisRint  = circAvg[j].rint;
      thisNterp = circAvg[j].nterp;
      for(i=caOffsetX;i<(thisNterp);i++)
	{/* DO all */
	  if(thisRint[i]!=0.0)
	    {/* OK to use */
	      minCA = MIN(minCA,thisRint[i]);
	      maxCA = MAX(maxCA,thisRint[i]);
	    }/* OK to use */
	}/* DO all */
    }/* DO all curves */
  
  if(maxCA==minCA)
    minCA = maxCA - 0.01;
  scaleCircFactor = 245.0 / ((float)(maxCA-minCA)); 
  
  passTag= 0;
  DrawCircAvgWindow(0,&passTag,0);
      
}/* CIRCAVG_CHANGE */






/* ************************************************************
 * circAvgText_changed() - They want to change circavg size.
 */
void circAvgText_changed(Widget widget, int *tag, 
			 XmAnyCallbackStruct *callback_data)
{/* circAvgTEXT_CHANGED */
  char *circAvgText;
  int i;

  circAvgText = XmTextGetString(circAvgTextWidget); 

  circAvgSize = (int)strtod(circAvgText, 0);
  circAvgSize = MAX(1,MIN(circAvgSize, 4));
  XtFree(circAvgText);

}/* circAvgTEXT_CHANGED */



/* ************************************************************
 * circAvgText_create() - Remember the text widgets on the
 * circAvg screen and init them.
 */
void circAvgText_create(Widget widget, int *tag, 
			XmAnyCallbackStruct *callback_data)
{/* circAvgTEXT_CREATE */
  
  if(*tag==0)
    {/* This */
      circAvgTextWidget= widget;
      XmTextSetString(circAvgTextWidget, "1"); 
    }/* This */
  else if(*tag==11)
    circAvgLineW[0]= widget;
  else if(*tag==12)
    circAvgLineW[1]= widget;
  else if(*tag==13)
    circAvgLineW[2]= widget;
  else if(*tag==1)
    circAvgLabelW[0]= widget;
  else if(*tag==2)
    circAvgLabelW[1]= widget;
  else if(*tag==3)
    circAvgLabelW[2]= widget;
  else if(*tag==CAX)
    circAvgScrollW= widget;
  
  circAvgSize= 1;
  
}/* circAvgTEXT_CREATE */



/***************************************************************
 * findMaxMin()- Generic rtn for finding the min/max in the
 * given fft buffer.
*/
void findMaxMin(float *fftImg_buffer,
		float *maxF, float *minF, float *offset,
		int logSwitch, int fftWidth, int fftHeight,
		int bkgVal)
{/* FINDMAXMIN */
  int i,j,jindx,jindxp1;
  float thisFloat;

  *maxF = -HUGE;  /* dgm -HUGE instead of -99999.9 */  
  *minF = HUGE;   /* dgm HUGE instead of 9999999.9 */
  for(j=0;j<fftHeight;j++)
    for(i=0;i<(fftWidth*2);i+= 2)
      {/* Do entire image */
	
	jindx   = ((j*(int)(fftWidth*2))+(i));

	thisFloat= fftImg_buffer[jindx];
	*maxF = MAX(*maxF,thisFloat);
	*minF = MIN(*minF,thisFloat);

      }/* Do entire image */
  *offset = *minF;

}/* FINDMAXMIN */





/*****************************************************************
 * calcNewLat() - Calculate new lattice ax,ay,bx,by based on the
 * values they've chosen from the scrollbars for astar,bstar,gamstr,
 * and omega(OLD gammaLat).
 * See latref_calc() in latref.for for code from which this was taken.
*/
void calcNewLat(int numScroll)
{/*  CALCNEWLAT */
  
  if(!layerLines)
    {/* 2D lattice */
      if(numScroll==2)
	{/* gammaStr == latScroll 3*/
	  gamstr = latScrollValue[2];
	  map_abang = 180.0 - gamstr;
	  /* Don't forget to add omega also ... */
	  anga = angaSave + omega;
	  angb = latScrollValue[2] + anga;

	  byL  = bstar * sin(deg2rad(angb));
	  bxL  = bstar * cos(deg2rad(angb));
	  ayL  = astar * (float)sin((double)deg2rad(anga));
	  axL  = astar * (float)cos((double)deg2rad(anga));

	  /*  omega= 180.0 - gamstr; */
	}/* gammaStr == latScroll 3*/
      else if(numScroll==0)
	{/* astar */
	  astar = latScrollValue[0];
	  if(bstar != 0.0 && astar != 0.0)
	    aoverb = 1.0 / (astar / bstar);

	  ayL = astar * (float)sin((double)deg2rad(anga));
	  axL = astar * (float)cos((double)deg2rad(anga));

	  /* adim = idim1*raster/(astar*sind(gamstr)); */
	}/* astar */
      else if(numScroll==1)
	{/* bstar */
	  bstar = (float)latScrollValue[1];
	  if(bstar != 0.0)
	    aoverb = 1.0 / (astar / bstar);

	  byL = (float)(bstar * (float)sin((double)deg2rad(angb)));
	  bxL = (float)(bstar * (float)cos((double)deg2rad(angb)));

	  /* bdim = idim1*raster/(bstar*sind(gamstr)); */
	}/* bstar */
      else if(numScroll==3)
	{/* omega */
	  omega = latScrollValue[3];

	  /* Shouldn't we do this from the anga and angb ??
	   * what if the gamstr has changed ???  */
	  /* Don't forget gamstr */
	  anga = angaSave + omega;
	  angb = latScrollValue[2] + anga;

	  ayL  = astar * (float)sin((double)deg2rad(anga));
	  axL  = astar * (float)cos((double)deg2rad(anga));
	  byL  = bstar * (float)sin((double)deg2rad(angb));
	  bxL  = bstar * (float)cos((double)deg2rad(angb));
	  
	}/* omega */
    }/* 2D lattice */
  else
    {/* layerlines */
      if(numScroll==3)
	omega = latScrollValue[3];
      else if(numScroll==1)
	dstar = latScrollValue[1];
    }/* layerlines */

}/*  CALCNEWLAT */






/****************************************************************************
 * fft_dsp_c() - Derived from the TSB fortran FFT_DSP().  We're trying to 
 * do 1024x1024 FFT's now and display them in 512x512.  Also be aware of the
 * problesm involving 256x256 and 128x128 fft's...
 *
C  **************************************************************************
C  *                                                                        *
C  *                 *****     SUBROUTINE FFT_DSP_rob *****                 *
C  *                                                                        *
C  *    THE INTENSITIES FROM EMIMGFFT ARE DISPLAYED USING A 64 GREY-LEVEL   *
C  *      SCALE.  THE TOP TWO BITS ARE USED BY THE MASK PROGRAM FOR         *
C  *      SUBSEQUENT PSEUDO-OPTICAL FILTERING (USING EMFFT).                *
C  *                                                                        *
C  *    ONLY VALUES BETWEEN THRESH_LO AND THRESH_HI ARE MAPPED INTO THE     *
C  *     COLOR TABLE, WITH OUTLIERS SET TO THE OUTSIDE LIMITS OF THE CT.    *
C  *                                                                        *
C  *    LAST MODIFIED  3-MAR-1992 BY T. BAKER.                              *
C  **************************************************************************
C
*/
void fft_dsp_c(float *fft, BYTE *img, 
	       int fftSize1, int fftSize2,
	       int ncol, int nrow, 
	       float thresh_hi, float thresh_lo,
	       int hiPixel, int loPixel, 
	       int linear, int amps)
{/* FFT_DSP_C */
  int i, j, k, l, fftSize2D2, fftSize2T2, jT2;
  int lastRow,lastCol,ii,jj, thisOutRow;
  float    xmult, ymult;
  float    testlo, testhi;
  float    slopeb, fract_min, fract_max;
  float    maxvalue, nonlin_k, nonlin_s, value;
  
  printf("fftSizes [%d,%d]\n",fftSize1,fftSize2);
  printf("ncol,nrow[%d,%d]\n",ncol,nrow);

  testhi = hiPixel;
  testlo = loPixel;
  
  xmult = (float)ncol / (float)fftSize1;
  ymult = (float)nrow / (float)fftSize2;

  lastCol = MAX((int)xmult,1);
  lastRow = MAX((int)ymult,1);

  
  fftSize2D2 = fftSize2 / 2;
  fftSize2T2 = fftSize2 * 2;

  slopeb = (testhi-testlo)/(thresh_hi-thresh_lo);
  
  /*
   *  C  *****************************************
   *  C  *    FILL UP IMG ARRAY AFTER SCALING    *
   *  C  *****************************************
   */
  
  if(xmult >= 1.0 && ymult >= 1.0)
    {/* The FFT is smaller or equal to the display size */

      for(i=0;i<(fftSize1);i++)
	{/* do i = 1,((fftSize2/2)+1) */
	  
	  if(i==0)
	    thisOutRow = fftSize2D2 - 1;
	  else if(i <= fftSize2D2)
	    thisOutRow = fftSize2D2 - 1 + i;
	  else
	    thisOutRow = i - fftSize2D2 - 1;
	  thisOutRow = MAX(thisOutRow,0);
	  thisOutRow = MIN(thisOutRow,(fftSize1-1));
	  
	  for(j=0;j<(fftSize2D2);j++)
	    {/* do j = 1,jtest ! right side of fft */
	      
	      jT2 = 2 * j;
	      fft[(jT2+1+(i*fftSize2T2))] *= 
		tempFactors[((int)(i*ymult))][2*(int)(j*xmult)];
	      fft[(jT2+(i*fftSize2T2))] *= 
		tempFactors[((int)(i*ymult))][2*(int)(j*xmult)];
	      value = ((fft[(jT2+1+(i*fftSize2T2))] * 
			fft[(jT2+1+(i*fftSize2T2))])
		       + 
		       (fft[(jT2+(i*fftSize2T2))] * fft[(jT2+(i*fftSize2T2))])
		       );
	      if(amps != 0) value = sqrt(value);
	      if(linear==0 && 
		 value != 0.0)value = log10(value);
	      
	      value = ((value - thresh_lo)*slopeb) + testlo;
	      
	      if(value < testlo) 
		value = testlo;
	      if(value > testhi) 
		value = testhi;
	      
	      for(ii=0;ii<lastRow;ii++)
		{/* Expand it out */
		  for(jj=0;jj<lastCol;jj++)
		    {/* Expand cols */
		      img[((int)(((thisOutRow*ymult)+ii)*ncol)+
			   (int)((j+fftSize2D2)*xmult)+jj)] = 
			     (BYTE)(value);
		    }/* Expand cols */
		}/* Expand it out */
	      
	    }/* do j = 1,jtest ! right side of fft */
	  
	  
	  for(j=(fftSize2/2);j<(fftSize2);j++)
	    {/* do j = (jtest+1),fftSize1 ! left side of fft */
	      jT2 = 2 * j;
	      fft[(jT2+1+(i*fftSize2T2))] *= 
		tempFactors[((int)(i*ymult))][2*(int)(j*xmult)];
	      fft[(jT2+(i*fftSize2T2))] *= 
		tempFactors[((int)(i*ymult))][2*(int)(j*xmult)];
	      value = ((fft[(jT2+1+(i*fftSize2T2))] * 
			fft[(jT2+1+(i*fftSize2T2))])
		       + 
		       (fft[(jT2+(i*fftSize2T2))] * fft[(jT2+(i*fftSize2T2))])
		       );
	      
	      if(amps != 0) value = sqrt(value);
	      if(linear==0 &&
		 value != 0.0)value = log10(value);
	      
	      value = ((value - thresh_lo)*slopeb) + testlo;
	      
	      if(value < testlo) value = testlo;
	      if(value > testhi) value = testhi;
	      
	      for(ii=0;ii<lastRow;ii++)
		{/* Expand it out */
		  for(jj=0;jj<lastCol;jj++)
		    {/* Expand cols */
		      img[((int)(((thisOutRow*ymult)+ii)*ncol)+
			   (int)((j-fftSize2D2)*xmult)+jj)] = 
			     (BYTE)(value);
		    }/* Expand cols */
		}/* Expand it out */
	      
	    }/* do j = (jtest+1),fftSize1 ! left side of fft */
	  
	}/* do i = 1,((fftSize2/2)+1) */
    }/* The FFT is smaller or equal to the display size */
  else if(xmult < 1.0 && ymult < 1.0)
    {/* The FFT is larger than the display size */
      int step,fftSize1D2mult,fftSize2D2mult;
      float totValue,numPnts;

      step = (int)(1.0 / xmult);
      fftSize2D2mult = (int)(fftSize2D2 * xmult);
      for(i=0;i<(fftSize1);i+=step)
	{/* do all rows */
	  
	  if(i==0)
	    thisOutRow = fftSize2D2mult - 1;
	  else if(i <= fftSize2D2)
	    thisOutRow = fftSize2D2mult - 1 + (i*xmult);
	  else
	    thisOutRow = (i*xmult) - fftSize2D2mult - 1;
	  thisOutRow = MAX(thisOutRow,0);
	  thisOutRow = MIN(thisOutRow,((int)(fftSize1*xmult)-1)); 
	  
	  for(j=0;j<(fftSize2D2);j+=step)
	    {/* do right side of fft */
	      
	      numPnts  = 0.0;
	      totValue = 0.0;
	      for(ii=0;ii<step;ii++)
		{/* Expand it out */
		  
		  for(jj=0;jj<step;jj++)
		    {/* Expand cols */
		      jT2 = 2 * (j+jj);
		      fft[(jT2+1+((i+ii)*fftSize2T2))] *= 
			tempFactors[((int)(i*ymult))][2*(int)(j*xmult)];
		      fft[(jT2+((i+ii)*fftSize2T2))] *= 
			tempFactors[((int)(i*ymult))][2*(int)(j*xmult)];
		      value = ((fft[(jT2+1+((i+ii)*fftSize2T2))] * 
				fft[(jT2+1+((i+ii)*fftSize2T2))])
			       + 
			       (fft[(jT2+((i+ii)*fftSize2T2))] * 
				fft[(jT2+((i+ii)*fftSize2T2))])
			       );
		      
		      if(amps != 0) value = sqrt(value);
		      if(linear==0 && 
			 value != 0.0)value = log10(value);
		      
		      value = ((value - thresh_lo)*slopeb) + testlo;
		      
		      if(value < testlo) 
			value = testlo;
		      if(value > testhi) 
			value = testhi;
		      
		      totValue += value;
		      numPnts  += 1.0;
		    }/* Expand cols */
		}/* Expand it out */
	      
	      img[((int)(thisOutRow*ncol)+
		   (int)((j+fftSize2D2)*xmult))] = 
		     (BYTE)(totValue/numPnts);
	      
	    }/* do right side of fft */
	  
	  
	  for(j=(fftSize2D2);j<(fftSize2);j+=step)
	    {/* do left side of fft */
	      
	      numPnts  = 0.0;
	      totValue = 0.0;
	      for(ii=0;ii<step;ii++)
		{/* Expand it out */
		  
		  for(jj=0;jj<step;jj++)
		    {/* Expand cols */
		      jT2 = 2 * (j+jj);
		      fft[(jT2+1+((i+ii)*fftSize2T2))] *= 
			tempFactors[((int)(i*ymult))][2*(int)(j*xmult)];
		      fft[(jT2+((i+ii)*fftSize2T2))] *= 
			tempFactors[((int)(i*ymult))][2*(int)(j*xmult)];
		      value = ((fft[(jT2+1+((i+ii)*fftSize2T2))] * 
				fft[(jT2+1+((i+ii)*fftSize2T2))])
			       + 
			       (fft[(jT2+((i+ii)*fftSize2T2))] * 
				fft[(jT2+((i+ii)*fftSize2T2))])
			       );
		      
		      if(amps != 0) value = sqrt(value);
		      if(linear==0 && 
			 value != 0.0)value = log10(value);
		      
		      value = ((value - thresh_lo)*slopeb) + testlo;
		      
		      if(value < testlo) 
			value = testlo;
		      if(value > testhi) 
			value = testhi;
		      
		      totValue += value;
		      numPnts  += 1.0;
		    }/* Expand cols */
		}/* Expand it out */
	      
	      
	      img[((int)(thisOutRow*ncol)+
		   (int)((j-fftSize2D2)*xmult))] = 
		     (BYTE)(totValue/numPnts);
	      
	    }/* do left side of fft */
	  
	}/* do all rows) */

    }/* The FFT is larger than the display size */
  else
    {/* Don't know */
      askThem("[FFT_DSP_C] Don't know how to do this...",
	      FALSE,
	      XmDIALOG_QUESTION);
    }/* Don't know */
  

}/* FFT_DSP_C */





/*******************************************************************
 * calcFFT() - Stuff the buffer and calc the FFT.  Display based on
 * scroll values.
*/
void calcFFT(int doFFT, int thisBox, int tryBackground)
{/* CALCFFT */
  int i,m,j,k,thisPixel,fftRow,fftCol,
        thisNcols,thisNrows,arg1,arg2,arg3,arg4,jindx,jindxp1;
  float distThis;
  int insideLimits,strtAvg,endAvg,thisImg,thisNimg,useRectangle;
  int thisPixelValue;
  int ksign,iform,linear;
  float thisMin,thisMax,maxF,minF,offset,recipThisNimg;
  float theoreticalMax,theoreticalMin;
  float scaleFactor, thisFFTvalue, thisFloat, thisSign;
  float recipSize;
  int iTrecip,iTidim2, iTfftSize2;
  float thisBackground;
  float *temp_buffer;
  float  *char_buffer, *thisBuffer;
  XmString simpleString;
  int thisMaxInt,thisMinInt,sliderSize, array_size, hiPixel,loPixel;
  double aveIntensity,maxAve,minAve;
  int minAveCoordX,minAveCoordY;
  int maxAveCoordX,maxAveCoordY;
  float thisVal,thisValP1;

  numCols = purdueHeader.numCols;
  numRows = purdueHeader.numRows;

  useRectangle = rectBox.drawable;

  if(useRectangle)
    {/* USe rectangle */
      numRows = rectBox.lowy - rectBox.y;
      numCols = rectBox.lowx - rectBox.x;
    }/* USe rectangle */
  
  if(numCols > 1024 || numRows > 1024)
    {/* Not yet */
      askThem("Image sizes greater than 1024x1024 not yet implemented.",
	      FALSE,
	      XmDIALOG_QUESTION);
      goto noFFT;
    }/* Not yet */

  CK(temp_buffer= (float *)malloc(
			 sizeof(float)*numCols*numRows));

  if(doFFT)
    {/* First time - calc the fft */

      /* We've got a fresh FFT here so no shifting of the origin
       * has occurred yet. */
      apOrigXtot = apOrigYtot = 0.0;

      /* Find out what fftSize they want and get the reciprical for
       * use later. */
      recipSize = (float)((float)fftSize / (float)idim1 );

      /* Init out the FFT buffers */
      for(i=0;i<idim1;i++)
	for(j=0;j<(idim2*2);j++)
	  {/* Init all */
	    fftImage_gold[(i*(idim2*2))+j]= 
	      fftImage_buffer2[(i*(idim2*2))+j]= 0.0;
	  }/* Init all */

      CK(char_buffer= (float *)malloc(
			 sizeof(float)*numCols*numRows));


      if(averageFFT)
	{/* Setup for doing all of the images */
	  strtAvg = 0;
	  endAvg  = purdueHeader.nimg;
	  thisNimg = 0;
	  /* Find out how many we've got */
	  for(i=strtAvg;i<endAvg;i++)
	    if(!deletePackedImage[i])
	      thisNimg++;
	  if(thisNimg<=0)
	    goto noFFT;
	}/* Setup for doing all of the images */
      else
	{/* Just do this one */
	  strtAvg = imgNum-1;
	  endAvg  = imgNum;
	  thisNimg = 1;
	}/* Just do this one */
      recipThisNimg = (float)(1.0 / (float)thisNimg);

      for(thisImg=strtAvg;thisImg < endAvg;thisImg++)
	{/* Do all images wanted */

	  if(!averageFFT || (averageFFT && !deletePackedImage[thisImg]))
	    {/* OK to do */

	      if(fileType==PACKEDPORTABLE || fileType==MAP)
		sprintf(outStr,"%d",thisImg+1);
	      else
		sprintf(outStr,"%d",1);
	      XmTextSetString(fftImgNumW,outStr);
	      XmTextSetString(rotfltImgNumW,outStr);
	      XFlush(dsply);
	      
	      thisBuffer= (float *)packImg[(thisImg)]; 
	      
	      if(useRectangle)
		{/* USe rectangle box */
		  stuffRectangleBuffer(char_buffer, 
				       image_fileName.addr);
		}/* USe rectangle box */

	      for(i=0;i<numRows;i++)
		for(j=0;j<numCols;j++)
		  {/* Do all */
		    if(useRectangle)
		      temp_buffer[((i*numCols)+j)]= (float)
			char_buffer[((i*numCols)+j)];
		    else if(fileType==PACKEDPORTABLE)
		      temp_buffer[((i*numCols)+j)]= (float) 
			thisBuffer[((i*numCols)+j)];
		    else if(fileType==MAP || fileType==FLOATMAP)
		      temp_buffer[((i*numCols)+j)]= (float)
			float_image_buffer1[((thisImg*numRows*numCols)+
					     (i*numCols)+j)];
		    else
		      {/* Just get it from here */
			temp_buffer[((i*numCols)+j)]= (float)
			  int_image_buffer1[((i*image_width)+j)];
			char_buffer[((i*numCols)+j)]= (float)
			  temp_buffer[((i*numCols)+j)];
		      }/* Just get it from here */
		  }/* Do all */
	      
	      if(tryBackground)
		{/* Try this one */
		  int xcen,ycen,xSq,ySq,iMycen,jMxcen,thisRad,distnce;
		  
		  thisBackground = fftBkgValue;
		  
printf("Trying bkg[%f]\n",thisBackground);
		}/* Try this one */
	      else
		{/* Get global */
		  if(fileType==PACKEDPORTABLE && !useRectangle)
		    {
		      /* Don't do this it messes up gap.pif
		      printf("ibkg[%f] pI[%f]\n",
			     purdueHeader.img_bkg,(*packImg[thisImg]));
		      purdueHeader.img_bkg = (float)(*packImg[thisImg]);
		      */
		      if(packRadius > 0)
			purdueHeader.img_bkg = (int)(*packImg[thisImg]);
		    }
		  else if((fileType==EMPORTABLE || fileType==BYTEPORTABLE ||
			   fileType==PORTABLE ) &&
			  !useRectangle)
		    {/* Do nothing */
		      /*  just use what we read from the file */
		      purdueHeader.img_bkg = pifHeader.img_bkg;
		      /* purdueHeader.img_bkg= (short)purdueHeader.img_avg; */
		    }
		  else if(fileType==MAP || fileType==FLOATMAP )
		    {/* purdueHeader.img_bkg= 0; Don't do this
		                                EMPORTABLEI4 images have
						a valid background */
		    }
		  else
		    {/* Unknown source or userectangle*/
		      if(useRectangle)
			{
			  int saveFileType;
			  saveFileType = fileType;
			  fileType = RECTPACKED;
			  calcStats(char_buffer,
				    &purdueHeader.img_bkg, 
				    &purdueHeader.img_min,
				    &purdueHeader.img_max, 
				    &purdueHeader.img_avg, 
				    &purdueHeader.img_std,
				    numCols, numRows, numCols);
			  fileType = saveFileType;
			}
		      else 
			{/* Else just use what we read from the file */
			  purdueHeader.img_bkg = pifHeader.img_bkg;
			}/* Else just use what we read from the file */
		    }/* Unknown source or userectangle */
		  thisBackground = purdueHeader.img_bkg;
		
printf("Using bkg[%f]\n",thisBackground);

		  /* Setup the FFT Background widget */
		  fftBkgValue = purdueHeader.img_bkg;
		  sliderSize = 5;
		  XtSetArg(arglst[0], XmNmaximum, (int)(purdueHeader.img_max+
							(float)sliderSize));
		  XtSetArg(arglst[1], XmNminimum, (int)
			   (purdueHeader.img_min));
		  XtSetArg(arglst[2], XmNsliderSize, (int)(sliderSize));
		  XtSetArg(arglst[3], XmNvalue, (int)fftBkgValue);
		  XtSetArg(arglst[4], XmNpageIncrement, (int)(sliderSize));
/*		  XtSetValues(fftBkgScrollW, arglst, 5);  */
		  XFlush(dsply);
		  
		  sprintf(outStr,"%d",fftBkgValue);  
		  createSimpleString(&simpleString,outStr);
		  XtSetArg(arglst[0], XmNlabelString, simpleString);
/*		  XtSetValues(fftBkgValueW, arglst, 1);  */
		  XFlush(dsply);
		  XmStringFree(simpleString); 
		  
		}/* Get global */
	      
	      ksign = 1;       /* Do a "forward" FFT */
	      iform = 1;       /* Data is imaginary */

	      
	  if(fileType == MAP)
	    {/* DO for MAP */
	      int nsec,em_wrt,fft_full,thisSec,fft_iend;
	      float fft_fzero, fft_origx, fft_origy;
	      thisBackground = purdueHeader.img_bkg;

	      nsec = 1;
	      em_wrt = 0;
	      thisSec = 1;
	      fft_full = 1;
	      fft_origx = fft_origy = 0.0;
	      fft_iend = 1;
	      fft_fzero = 0.0;

	      map_fft_fill_robemorien_((float_image_buffer1+
					(thisImg*numCols*numRows)), 
				       &numCols, &numRows,
				       fftImage_buffer2, 
				       &fftSize, &fftSize, 
				       &thisBackground);


	    }/* DO for MAP */
	      else
		{/* Do for images */
		  map_fft_fill_robemorien_(temp_buffer, &numCols, &numRows, 
					   fftImage_buffer2, 
					   &fftSize, &fftSize, 
					   &thisBackground);
		}/* Do for images */


              /* Check to see if they want to stop */
	      /* Frame the location of the stop button */
	      XDrawLine(dsply,XtWindow(fftStopW),image_gcRED,
			12,12,20,20);
	      XDrawLine(dsply,XtWindow(fftStopW),image_gcRED,
			20,12,12,20);
	      XFlush(dsply);
	      
	      /* See if they want to bugout or not */
	      if(XCheckTypedWindowEvent(dsply,
					XtWindow(fftStopW),
					ButtonPress,
					&event))
		{/* STOP!!! */
		  XFlush(dsply);
		  free(char_buffer);
		  free(temp_buffer);
		  goto noFFT;
		}/* STOP!!! */
	      else
		{/* OK to continue ... */
		}/* OK to continue ... */

	      /* Now do the FFT */
	      fft_2d_rob_ (fftImage_buffer2, &fftSize, &fftSize,
			   &ksign, &iform);


	      /* This is curious - We need to have the FFT buffer 
	       * built by stuffing this way for the Fourier Averaging
	       * rtn but we need it the other way for the rtn that
	       * calculates ax,ay,bx,by  (latcen2). */
	      if(!averageFFT)
		{/* DOn't do for averaged FFT */
		  /* Let's try putting this into TSB space and 
		   * see if it helps */
		  for(i=0;i<numRows/2;i++)
		    {
		      float tmpB;
		      for(j=0;j<numCols;j++)
			{/* Do all */
			  tmpB = temp_buffer[((i*numCols)+j)];
			  temp_buffer[((i*numCols)+j)]= 
			    temp_buffer[(((numRows-i-1)*numCols)+j)] ;
			  temp_buffer[(((numRows-i-1)*numCols)+j)] = tmpB ; 
			}/* Do all */
		    }
		  map_fft_fill_robemorien_(temp_buffer, &numCols, &numRows, 
					   fftImage_fourAver, 
					   &fftSize, &fftSize, 
					   &thisBackground);
		  
		  fft_2d_rob_ (fftImage_fourAver, &fftSize, &fftSize,
			       &ksign, &iform);
		}/* DOn't do for averaged FFT */
	      

	      /* This is only valid when averaging now ... */
	      /* ctfEnhance = FALSE; */

	      /* Now accumulate this one into the GOLD  */
	      for(i=0;i<fftSize;i++)
		{/* i */

		  iTfftSize2 = i * (fftSize*2);  
		  for(j=0;j<(fftSize*2);j+=2)
		    {/* Accumulate all */

		      if(averageFFT)
			{/* average */
			  /* This is only valid when averaging now ... 
			   * what follows is an "incoherent" average of
			   * the data - makes the CTF rings look pretty. */
			  /* ctfEnhance = TRUE; */
			  if(ctfEnhance)
			    {/* DO this kludge - why does it enhance??? */
			      thisVal   = fftImage_buffer2[(iTfftSize2)+
							   (int)(j)];
			      thisValP1   = fftImage_buffer2[(iTfftSize2)+
							     (int)(j)+1];
			      fftImage_gold[(iTfftSize2+j)] += 
				(((thisVal*thisVal)+(thisValP1*thisValP1)) * 
				 recipThisNimg);
			      fftImage_gold[(iTfftSize2+j+1)] = 0.0; 
			    }/* DO this kludge - why does it enhance??? */
			  else
			    {/* "normal" */
			      thisVal   = fftImage_buffer2[(iTfftSize2)+
							   (int)(j)];
			      fftImage_gold[(iTfftSize2+j)] += (thisVal * 
								recipThisNimg);
			      thisValP1   = fftImage_buffer2[(iTfftSize2)+
							     (int)(j)+1];
			      fftImage_gold[(iTfftSize2+j+1)] += (thisValP1 * 
							      recipThisNimg);
			    }/* "normal" */
			}/* average */
		      else
			{/* Individual */
			  if(FALSE) /* Never do ctfEnhance for individuals*/
			    {/* DO this kludge - why does it enhance??? */
			      thisVal   = fftImage_buffer2[(iTfftSize2)+
							   (int)(j)];
			      thisValP1   = fftImage_buffer2[(iTfftSize2)+
							     (int)(j)+1];
			      fftImage_gold[(iTfftSize2+j)] = 
				(((thisVal*thisVal)+(thisValP1*thisValP1)));
			    }/* DO this kludge - why does it enhance??? */
			  else
			    {/* "Normal" */
			      double tmpVal;
			      fftImage_gold[(iTfftSize2+j)] = 
				fftImage_buffer2[(iTfftSize2+j)];
			      fftImage_gold[(iTfftSize2+j+1)] = 
				fftImage_buffer2[(iTfftSize2+j+1)];

			    }/* "Normal" */
			}/* Individual */


	thisVal = (
		     (fftImage_gold[(iTfftSize2+j)]*
		      fftImage_gold[(iTfftSize2+j)]) +
		     (fftImage_gold[(iTfftSize2+j+1)]*
		      fftImage_gold[(iTfftSize2+j+1)]) );
	
	if(thisVal < 0.0)
	  {
	    printf("Found one [%d][%d]-[%f]\n",i,j,thisVal);
	  }


		    }/* Accumulate all */
		}/* i */
	      
	    }/* OK to do */
	  
	}/* Do all images wanted */

      free(char_buffer);
      
    }/* First time - calc the fft */

  thisMin = HUGE;   /* dgm HUGE instead of 99999999.0 */
  thisMax = -HUGE;  /* dgm HUGE instead of -99999999.9 */
  thisMin = r4_max;
  thisMax = r4_min;
  recipSize = (float)((float)512.0 / (float)idim1 );
  for(i=0;i<(int)(fftSize);i++)
    for(j=0;j<(int)(fftSize*2);j+=2)
      {/* Do all image */
	float thisValue, val2,val3;
	thisPixel = (int)(i*(fftSize*2))+j;
	thisValue = (
		     (fftImage_gold[(thisPixel)]*
		      fftImage_gold[(thisPixel)]) +
		     (fftImage_gold[(thisPixel+1)]*
		      fftImage_gold[(thisPixel+1)]) );
	
	/* NOTE - SHouldn't we just apply the tempFactor here ??? */
	thisValue *= tempFactors[(int)(i*recipSize)][(int)(j*recipSize)];
	if(tempFactors[(int)(i*recipSize)][(int)(j*recipSize)] < 0.0)
	  {
	    printf("Found one [%d][%d]-[%f][%f]\n",
		   i,j,thisValue,
		   tempFactors[(int)(i*recipSize)][(int)(j*recipSize)]);
	    printf("temp[i,j] = [%d,%d]\n",
		   (int)(i*recipSize),
		   (int)(j*recipSize));
	  }
	if(fftDispType==AMPLITUDE)
	  thisValue = sqrt(thisValue);
	if(fftType == LOGFFT && thisValue > 0.0)
	  thisValue = log10(thisValue);
	thisMax = MAX(thisValue,thisMax);
	thisMin = MIN(thisValue,thisMin);
      }/* Do all image */
  
  if(averageFFT)
    {/* Do special */
      sprintf(outStr,"Aver");
      XmTextSetString(fftImgNumW,outStr);
    }/* Do special */
  
  /* Now setup the scroll bars to hold the new max/min */
  /* We can't exceed what a 32 bit signed integer will hold!! 
   * - Really 2.17 Gig but let's be safe... 
   */
  theoreticalMax = thisMax;  /* Capture these so we can stuff the label */
  theoreticalMin = thisMin;
  if (thisMin == 0.0)
    thisMin = 1.0;
  if(thisMax <= thisMin)
    thisMax = thisMin + 1.0;
  if(doFFT && !fftMinMaxLock)
    {/* OK to change fft min max value */
      globMinFFTvalue = minFFTvalue = thisMin;
      globMaxFFTvalue = maxFFTvalue = thisMax;
    }/* OK to change fft min max value */
  
  /* OK so suppose they've changed to Amplitudes from Intensities
   * we need to adjust the min/max values in case the range is less
   * now */
  if(!fftMinMaxLock)
    {/* keep it in range */
      minFFTvalue = MAX(minFFTvalue,thisMin);
      maxFFTvalue = MIN(maxFFTvalue,thisMax);
    }/* keep it in range */

  if(fftType==LINEARFFT)
    thisMax *= 0.001;  /* Do this in 1/1000 ths */
  else thisMax *= 100.0;  /* Do this in 100 ths */
  
  if(fftType==LINEARFFT)
    thisMin *= 0.001;  /* Do this in 1/1000 ths */
  else thisMin *= 100.0;  /* Do this in 100 ths */
  
  thisMaxInt = (int)MIN(thisMax,1800000000.0);   /* Why????? */
  imageMaxFFTvalue = (float)thisMax;
  thisMinInt = (int)thisMin;
  
  if(thisMaxInt <= thisMinInt)
    thisMaxInt = thisMinInt + 1;
  
  sliderSize = (thisMaxInt - thisMinInt) / 100;
  sliderSize = MAX(sliderSize,1);
  
  XtSetArg(arglst[0], XmNmaximum, (int)(thisMaxInt+(float)sliderSize));
  XtSetArg(arglst[1], XmNminimum, (int)(thisMinInt));
  XtSetArg(arglst[2], XmNsliderSize, (int)(sliderSize));
  
  if(!fftMinMaxLock)
    {/* OK to change fft min max value */
      if(fftType==LINEARFFT)/* 1/1000th */
	XtSetArg(arglst[3], XmNvalue, (int)(minFFTvalue/1000.0));
      else XtSetArg(arglst[3], XmNvalue, (int)(minFFTvalue*100.0));
      XtSetArg(arglst[4], XmNpageIncrement, (int)(sliderSize));
      if(sliderSize > 11.0)
	XtSetArg(arglst[5], XmNincrement, (int)(sliderSize/10.0));
      else XtSetArg(arglst[5], XmNincrement, (int)(1.0));
      XtSetValues(minFFTscroll, arglst, 6);
    }/* OK to change fft min max value */
  else
    {/* Not OK */
      XtSetArg(arglst[3], XmNpageIncrement, (int)(sliderSize));
      if(sliderSize > 11.0)
	XtSetArg(arglst[4], XmNincrement, (int)(sliderSize/10.0));
      else XtSetArg(arglst[4], XmNincrement, (int)(1.0));
      XtSetValues(minFFTscroll, arglst, 5);
    }/* Not OK */
  
  XtSetArg(arglst[0], XmNmaximum, (int)(thisMaxInt+(float)sliderSize));
  XtSetArg(arglst[1], XmNminimum, (int)(thisMinInt));
  XtSetArg(arglst[2], XmNsliderSize, (int)(sliderSize));
  if(!fftMinMaxLock)
    {/* OK to change fft min max value */
      if(fftType==LINEARFFT)/* 1/1000th */
	XtSetArg(arglst[3], XmNvalue, (int)(maxFFTvalue/1000.0));
      else XtSetArg(arglst[3], XmNvalue, (int)(maxFFTvalue*100.0));
      XtSetArg(arglst[4], XmNpageIncrement, (int)(sliderSize));
      if(sliderSize > 11.0)
	XtSetArg(arglst[5], XmNincrement, (int)(sliderSize/10.0));
      else XtSetArg(arglst[5], XmNincrement, (int)(1.0));
      XtSetValues(maxFFTscroll, arglst, 6);
    }/* OK to change fft min max value */
  else
    {/* Not OK */
      XtSetArg(arglst[3], XmNpageIncrement, (int)(sliderSize));
      if(sliderSize > 11.0)
	XtSetArg(arglst[4], XmNincrement, (int)(sliderSize/10.0));
      else XtSetArg(arglst[4], XmNincrement, (int)(1.0));
      XtSetValues(maxFFTscroll, arglst, 5);
    }/* Not OK */
  
  XSync(dsply,FALSE);
  XFlush(dsply);
  
  /* Stuff the current values of these */
  sprintf(outStr,"%.2f",minFFTvalue);   
  createSimpleString(&simpleString,outStr);
  XtSetArg(arglst[0], XmNlabelString, simpleString);
  XtSetValues(minFFTwidget, arglst, 1); 
  XFlush(dsply);
  XmStringFree(simpleString); 
  sprintf(outStr,"%.2f",maxFFTvalue);     
  createSimpleString(&simpleString,outStr);
  XtSetArg(arglst[0], XmNlabelString, simpleString);
  XtSetValues(maxFFTwidget, arglst, 1); 
  XFlush(dsply);
  XmStringFree(simpleString); 
  
  /* Now stuff the theoretical MAX/MIN labels */
  sprintf(outStr,"Range Min:  %.2f",theoreticalMin);  
  createSimpleString(&simpleString,outStr);
  XtSetArg(arglst[0], XmNlabelString, simpleString);
  XtSetValues(theominFFTwidget, arglst, 1); 
  XFlush(dsply);
  XmStringFree(simpleString); 
  sprintf(outStr,"Max:  %.2f",theoreticalMax);    
  createSimpleString(&simpleString,outStr);
  XtSetArg(arglst[0], XmNlabelString, simpleString);
  XtSetValues(theomaxFFTwidget, arglst, 1); 
  XFlush(dsply);
  XmStringFree(simpleString); 
  
  XSync(dsply,FALSE);
  XFlush(dsply);

  
  /* STuff the mirror buffer - we don't have to sew it together here
   * let's just do it for i=0 to < idim1 and idim2 etc... */
  for(i=0;i<idim1;i++)
    for(j=0;j<(idim2*2);j++)
      {/* Init all */
	fftImage_buffer2[(i*(idim2*2))+j]= 
	  fftImage_gold[(i*(idim2*2))+j] ;
      }/* Init all */

  if(lopHipSwitch)
    {/* Just do a lop hip thing */
      int thisTrue;
      float res_max,res_min,sizeF, cutmaxLo,cutmaxHi;
      char *thisStr;

      thisTrue =   1;
      thisStr = XmTextGetString(lopTextW); 
      cutmaxLo  = (float)strtod(thisStr,0);
      XtFree(thisStr);
      thisStr = XmTextGetString(hipTextW); 
      cutmaxHi  = (float)strtod(thisStr,0);
      XtFree(thisStr);
      sizeF = (float)((float)fftSize/(float)512.0);
      res_max =  (float)sqrt((double)lowResoFilter);
      res_max *= sizeF;
      res_min =  (float)sqrt((double)hiResoFilter);
      res_min *= sizeF;

      fft_lop_(fftImage_buffer2, 
	       &fftSize, &fftSize, 
	       &res_max, &cutmaxLo, &thisTrue);
      fft_hip_(fftImage_buffer2, 
	       &fftSize, &fftSize, 
	       &res_min, &cutmaxHi, &thisTrue);

    }/* Just do a lop hip thing */
  else
    {/* Normal */

      if(filterToggle && pickedLatPts) 
	{/* filter it */
	  float axPass,ayPass,bxPass,byPass;
	  
	  if(!layerLines)
	    {/* Do a 2d lattice */
	      short *thisMask;
	      
	      CK(thisMask = (short *)malloc(sizeof(short)*
					    (fftSize*fftSize*4))); 
	      
	      arg1= arg2= fftSize;
	      
	      axPass = axL;
	      ayPass = ayL;
	      bxPass = bxL;
	      byPass = byL;
	      
	      fft_fltmsk_rob_(fftImage_buffer2,
			      thisMask,
			      &arg1,
			      &arg2,
			      &axPass,
			      &ayPass,
			      &bxPass,
			      &byPass,
			      &filtSizeValue,
			      &icut,           /* icut ==1 for soft cutoff */
			      &lowResoFilter,
			      &hiResoFilter
			      );
	      
	      free(thisMask);
	      
	      
	    }/* Do a 2d lattice */
	  else if(pickedLatPts)
	    {/* DO layer lines */
	      float space,fall,slope,intercept;
	      
	      if(latLineOrder > 0)
		{/* OK to do */
		  fall  = (float)filtSizeValue;
		  /* printf("fall = %f\n",fall); */
		  slope = ((float)(
				   (256-latPt[0].measuredY)-
				   (256-latPt[1].measuredY)
				   ))/
				     ((float)(latPt[0].measuredX-
					      latPt[1].measuredX));
		  intercept = (float)((float)((float)(256-
						      latPt[0].measuredY))) - 
		    (float)(slope * (float)((float)(latPt[0].measuredX-256)));
		  if(pickLattice != -99 && !latValChanged)
		    {/* calc it */
		      latScrollValue[1] = 
			dstar = dstarSave = (fabs((float)(intercept)*
						  cos(atan2(slope,1.0))) / 
					     (float)latLineOrder);    
		      latScrollValue[3]= omega= omegaSave = 
			rad2deg(atan2((0.0-slope),1.0));
		      latReset_activate(0,0,0); 
		    }/* Calc it */
		  else
		    { /* Get from scroll */
		      dstar = latScrollValue[1]; 
		      omega = latScrollValue[3]; 
		    } /* Get from scroll */
		  space = dstar;
		  /* fltllmsk needs these in degrees not radians */
		  slope = omega;
		  slope = 0.0 - omega;  /* I think this is correct ??? */
		  /* Try this */
		  slope = 0.0 - deg2rad(omega);

		  fft_fltllmsk_rob_ (fftImage_buffer2, 
				     &fftSize, &fftSize, 
				     &slope, 
				     &lowResoFilter,
				     &hiResoFilter,
				     &space, &fall, &icut);

		}/* OK to do */
	      else askThem("Must have a positive Lattice Line Order.",
			   FALSE,XmDIALOG_INFORMATION);
	      
	    }/* DO layer lines */
	  
	}/* filter it */
    }/* Normal */
  
  /* Stuff (filtered) buffer back for display */
  arg1 = fftSize;
  arg2 = fftSize;
  arg3 = 1024;
  arg4 = 1024;

  /* Try this with fftImage_buffer2 after we clean up how it's stuff
   * in the sew-together stuff above... */
  if(hiColor)
    {
      loPixel = (int)8;
      hiPixel = (int)(tryColors-2);
    }
  else
    {/* 8 bit */
      loPixel = (int)pixels_return[(8)];
      hiPixel = (int)pixels_return[(tryColors-2)];
    }/* 8 bit */

  if(fftType==LINEARFFT)
    linear = -1;
  else linear = 0;     /* Log */

  if(disp1024fft && fftSize == 1024)
    {/* display it */
      int i;
      BYTE *byte_inCore_buffer;
      byte_inCore_buffer = (BYTE *)malloc(sizeof(BYTE)*1024*1024);
      if(int_image_buffer1 != NULL)
	free(int_image_buffer1);
      CK(int_image_buffer1 = (unsigned short *)malloc(
				   ((defDisplayBufferSize*
				     defDisplayBufferSize)*
				    sizeof(unsigned short))+1));
      if(imgInCore)
	free(inCore_buffer);
      CK(inCore_buffer = (float *)malloc(sizeof(float)*1024*1024));
      imgInCore = TRUE;
      file_width = 1024;
      file_height = 1024;
      fullScreen= TRUE;
      
      fft_dsp_c(fftImage_buffer2, 
		byte_inCore_buffer,
		arg1, arg2, arg3, arg4,
		maxFFTvalue,
		minFFTvalue,
		hiPixel,
		loPixel,
		linear,
		(fftDispType==AMPLITUDE)?TRUE:FALSE);

      for(i=0;i<(1024*1024);i++)
	inCore_buffer[i] = byte_inCore_buffer[i];
      free(byte_inCore_buffer);

      virtual_read_activate(-99,-99);

    }/* display it */

  /* Stuff (filtered) buffer back for display */
  arg1 = fftSize;
  arg2 = fftSize;
  arg3 = 512;
  arg4 = 512;

  /* Try this with fftImage_buffer2 after we clean up how it's stuff
   * in the sew-together stuff above... */
  if(hiColor)
    {
      loPixel = (int)8;
      hiPixel = (int)(tryColors-2);
    }
  else
    {/* 8 bit */
      loPixel = (int)pixels_return[(8)];
      hiPixel = (int)pixels_return[(tryColors-2)];
    }/* 8 bit */

  if(fftType==LINEARFFT)
    linear = -1;
  else linear = 0;     /* Log */

  printf("maxFFT[%f]  min[%f]\n",maxFFTvalue,minFFTvalue);
  fft_dsp_c(fftImage_buffer2, 
	    fft_buffer1,
	    arg1, arg2, arg3, arg4,
	    maxFFTvalue,
	    minFFTvalue,
	    hiPixel,
	    loPixel,
	    linear,
	    (fftDispType==AMPLITUDE)?TRUE:FALSE);


  for(i=0;i<512;i++)
    for(j=0;j<512;j++)
      {/* All cols */
	fft_gold_buffer1[((i*512)+j)]= fft_buffer1[((i*512)+j)];
      }/* All cols */

  if(backTransform)
    {/* Show them what the back transform is */
      float *fftImage_copy;
      /* Now put it back in "normal" space */
      ksign = 0-1;     /* Do a "reverse" FFT */
      iform = 1;       /* Data is imaginary */

      /* Back transform into a copy of the filtered FFT just in case
       * they want to do Fourier Averaging or etc... */ 
      CK(fftImage_copy= (float *)malloc(sizeof(float)*idim1*idim2*2));

      for(i=0;i<idim1;i++)
	for(j=0;j<(idim2*2);j++)
	  {/* Init all */
	    fftImage_copy[(i*(idim2*2))+j]= 
	      fftImage_buffer2[(i*(idim2*2))+j] ;
	  }/* Init all */

      fft_2d_rob_(fftImage_copy, &fftSize, &fftSize, &ksign, &iform);

      /* Let's look at what scaling needs done */
      scaleFactor = 0.0;
      findMaxMin(fftImage_copy, 
		 &maxF, &minF, &offset, FALSE, fftSize, fftSize,
		 purdueHeader.img_bkg);

      /* Defloat it */
      maxF += (float)purdueHeader.img_bkg;
      minF += (float)purdueHeader.img_bkg;


      if(applyCTF)
	{/* Do for packed boxes */
	  if((pifMode==6))
	    {/* byte */
	      if(maxF==minF)
		scaleFactor= 1.0;
	      else scaleFactor = (float)(255.0) / (maxF-minF);
	    }/* byte */
	  else
	    {/* short */
	      if(maxF==minF)
		scaleFactor= 1.0;
	      else scaleFactor = (float)(65500.0) / (maxF-minF);
	    }/* short */
	}/* Do for packed boxes */
      else
	{/* Normal */
	  if((pifMode==6))
	    {/* byte */
	      if(maxF==minF)
		scaleFactor= 1.0;
	      else scaleFactor = (float)(tryColors - 10.0) / (maxF-minF);
	    }/* byte */
	  else
	    {/* short */
	      if(maxF==minF)
		scaleFactor= 1.0;
	      else scaleFactor = (float)(tryColors - 10.0) / (maxF-minF);
	    }/* short */
	}/* Normal */

      offset = minF;

      /* Now put the back-transformed image into the display buffer */
      if(applyCTF)
	thisBuffer = packImg[thisBox];

      calcSizeFactor(zoomFactor);
      for(j=0;j<(numRows);j++)
	{/* Rows */
	  int yOut,xOut,l,m;
	  int thisChar;
	  yOut = (j*sizeFactor);
	  for(i=0;i<(numCols);i++)
	    {/* Do entire image */
	      /* Offset this one beside the original */
	      xOut = (i*sizeFactor)+(numCols*sizeFactor);

	      /* If we need to adjust for TSB orientation 
	      jindx   = (((numRows-j-1)*(int)(fftSize*2))+(i*2));
	      */
	      jindx   = ((j*(int)(fftSize*2))+(i*2));

	      /* jindxp1 = ((j*(int)(fftSize*2))+(i*2)+1); */
	      
	      /* Definitely need to add img_bkg back in!!! (To defloat)*/
	      thisFloat= fftImage_copy[jindx]+(float)purdueHeader.img_bkg;
	      if(hiColor)
		thisChar = (int)(
				 (float)(((float)((thisFloat)-
						offset) 
					* scaleFactor)));
	      else
		thisChar = (int)(
				 pixels_return[9] +
				 (int)(((float)((thisFloat)-
						offset) 
					* scaleFactor)));
	      if(applyCTF)
		{/* Stuff this into the packed box */
		  thisBuffer[((j*numCols)+i)]=  (float)(thisFloat);
		}/* Stuff this into the packed box */
	      else
		{/* Normal */
		  for(l=0;l<=(int)sizeFactor;l++)
		    for(m=0;m<=(int)sizeFactor;m++)
		      {
			if(hiColor)
			  {
			    XPutPixel(image_image1,
				      (xOut+m), (yOut+l),
				      pixels_return[9+thisChar]);
			    image_buffer1_disp[(((yOut+l)*buffer_width)+
						(xOut+m))]= 
			      (unsigned char)(thisChar); 
			  }
			else
			  image_buffer1[(((yOut+l)*buffer_width)+(xOut+m))]= 
			    (unsigned char)(thisChar); 
		      }
		}/* Normal */
	    }/* Do entire image */
	}/* Rows */
      XFlush(dsply);
      free(fftImage_copy);

      if(applyCTF && stuffBackground)
	{/* Stuff outside of circle with background */
	  int oldFileType,thisRad,k,l,outHeight,outWidth,
	      xCent,yCent,dist,xSq,ySq;
	  int yOut;
	  float backChar;

	  oldFileType = fileType;
	  fileType = REBOX;
	  thisBuffer = packImg[thisBox];
	  calcCircleStats(thisBuffer,
			  &purdueHeader.img_bkg, 
			  &purdueHeader.img_min,
			  &purdueHeader.img_max, 
			  &purdueHeader.img_avg, 
			  &purdueHeader.img_std,
			  numCols, numRows, numCols);
	  fileType = oldFileType;

	  /* Calc the background */
	  backChar= (float)purdueHeader.img_bkg;
      
	  /* Draw the background  for all > radius in box */
	  /* Where is the center? */
	  yCent = numRows / 2;
	  xCent = numCols / 2;
	  thisRad = xCent * xCent;
	  outHeight = numRows;
	  outWidth  = numCols;
	  for(k=0;k<outHeight;k++)
	    {/* Do the y direction */
	      yOut = (k * numCols);
	      ySq = (k - yCent) * (k - yCent);
	      
	      /* Stuff all > radius in background character */
	      for(l=0;l<outWidth;l++)
		{/* Stuff the outbuffer */
		  xSq = (l - xCent) * (l - xCent);
		  dist = (ySq+xSq);
		  if(dist > thisRad)
		    thisBuffer[(yOut+l)]=  backChar;
		}/* Stuff the outbuffer */

	    }/* Do the y direction */
	  
	}/* Stuff outside of circle with background */

      if(fileType==MAP)  /* Show more */
	XPutImage (dsply, image_window1, image_gc, image_image1, 0, 0,
		   0, 0, (image_width*3), (image_height*2));
      else
	XPutImage (dsply, image_window1, image_gc, image_image1, 0, 0,
		   0, 0, (image_width), image_height);
	
      XFlush (dsply);
      
    }/* Show them what the back transform is */
  
  free(temp_buffer);
/*   DrawFFTWindow(); */

 noFFT:;

}/* CALCFFT */




/*************************************************************
 * read_fftbuffer_activate() - Read an fft buffer from disk.
 * Used for debuggin other programs mainly.
 */
void read_fftbuffer_activate(Widget widget, int *tag, 
			     XmAnyCallbackStruct *callback_data)
{/* READ_FFTBUFFER_ACTIVATE */
  int saveBackTransform;
  FILE *fIn;
  
  fIn = fopen("fft_buffer.dat", "r");
  if(fIn)
    {/* OK to do */

      fread(fftImage_gold,sizeof(float),((fftSize*2)*fftSize),fIn);

      applyCTF = FALSE;
      saveBackTransform = backTransform;
      backTransform = TRUE;  /* Let's back transform it */
      
      calcFFT(FALSE, 0, FALSE);
      
      XSync(dsply,FALSE);
      XFlush(dsply);
      DrawFFTWindow();
      raiseIt(main_window_widget); 
      raiseIt(fftscreen);
      XFlush(dsply);
      
      /* Now stuff it back they way it used to be */
      backTransform = saveBackTransform;

      fclose(fIn);
    }/* OK to do */
  else
    {/* Can't find it */
      sprintf(outStr,
   "Can't find fft_buffer.dat in the directory you started robem from.");
      askThem(outStr,FALSE,XmDIALOG_INFORMATION);
    }/* Can't find it */

}/* READ_FFTBUFFER_ACTIVATE */




/*************************************************************
 * fft_writeFB_activate() - Write various FFT buffers to disk.
 */
void fft_writeFB_activate(Widget widget, int *tag, 
			  XmAnyCallbackStruct *callback_data)
{/* fft_writeFB_actIVATE */
  float img_bkg;
  float img_min, img_max;
  float img_avg, img_std;
  int ncol, nrow, yOut;
  char outFileName[256],rawOutFileName[256];
  int i,j,k,l,m,imgOffset;
  forPifHead_t thisHead;
  fString_t fName,readWrite;
  float minF,maxF,scaleFactor;
  float floatScaleFactor,recipFloatScaleFactor;
  int luNum;
  BYTE *temp_buffer;
  float *tempF_buffer;
  FILE *fOut;

  fName.addr = (char *)malloc(sizeof(char)*64);
  readWrite.addr = (char *)malloc(sizeof(char)*64);

  pif_init_head_(&thisHead);

  if(*tag==WRITEFFT)
    {/* init */
      sprintf(outFileName,"fftImage.pif");
      sprintf(rawOutFileName,"fftImage.raw");
      thisHead.numCols =  512;
      thisHead.numRows =  512;
      CK(temp_buffer = (BYTE *)malloc(sizeof(BYTE)*
				      thisHead.numRows*
				      thisHead.numCols));
      for(j=0;j<512;j++)
	{/* Rows */
	  yOut = 511 - j;
	  for(i=0;i<512;i++)
	    {/* Do entire image */
	      /* Remember to put it into TSB space */
	      temp_buffer[((j*512)+i)]= (BYTE)fft_buffer1[((yOut*512)+i)];
	    }/* Do entire image */
	}/* Rows */

    }/* init */
  else 
    {/* init back */
      sprintf(outFileName,"backImage.pif");
      sprintf(rawOutFileName,"backImage.raw");
      thisHead.numCols =  numCols;
      thisHead.numRows =  numRows;

      CK(temp_buffer = (BYTE *)malloc(sizeof(BYTE)*numRows*numCols));
      CK(tempF_buffer = (float *)malloc(sizeof(float)*numRows*numCols));

      calcSizeFactor(zoomFactor);
      for(j=0;j<(numRows);j++)
	{/* Rows */
	  int yOut,xOut,l,m;
	  BYTE thisChar;
	  /* Remember to put it into TSB space */
	  yOut = ((numRows-1-j)*sizeFactor);
	  for(i=0;i<(numCols);i++)
	    {/* Do entire image */
	      /* Offset this one beside the original */
	      xOut = (i*sizeFactor)+(numCols*sizeFactor);
	      temp_buffer[((j*numCols)+i)]= (BYTE)
		image_buffer1_disp[((yOut*buffer_width)+xOut)];
	      tempF_buffer[((j*numCols)+i)]= (float)
		image_buffer1_disp[((yOut*buffer_width)+xOut)];
	    }/* Do entire image */
	}/* Rows */

    }/* init back */

  calcByteStats(temp_buffer,
		&img_bkg, &img_min,
		&img_max, &img_avg, &img_std,
		thisHead.numCols,thisHead.numRows,
		thisHead.numCols);
  
  minF = (float)img_min;
  maxF = (float)img_max;
  scaleFactor = 255.0 / (maxF-minF);

  /* Now scale it */
  for(j=0;j<(thisHead.numRows);j++)
    {/* Rows */
      for(i=0;i<(thisHead.numCols);i++)
	{/* Do entire image */
	  temp_buffer[((j*thisHead.numCols)+i)]= (BYTE)
	    (((float)temp_buffer[((j*thisHead.numCols)+i)]-minF) * 
	     scaleFactor);
	}/* Do entire image */
    }/* Rows */

  fOut = fopen(rawOutFileName,"w");

  sprintf(fName.addr,outFileName);
  sprintf(readWrite.addr,"w");

  luNum = 7;
  thisHead.luNum = luNum;
  pif_open_(fName.addr, &luNum, readWrite.addr);

  calcMapStats(tempF_buffer,
	       &img_bkg, &img_min, &img_max, 
	       &img_avg, &img_std,
	       &thisHead.numCols, &thisHead.numRows, &thisHead.numSecs);
  calcScaleFactor(&floatScaleFactor,		     
		  &recipFloatScaleFactor, 
		  &int4scaleF, 
		  img_min,img_max,
		  img_avg,img_std,
		  PACKEDPORTABLE);

  thisHead.floatScaleFactor = floatScaleFactor;
  thisHead.recipFloatScaleFactor = recipFloatScaleFactor;
  thisHead.int4scaleF = int4scaleF;
  thisHead.nimg = 1;
  thisHead.htype= 1;
  thisHead.numSecs =  1;  /* No maps here */

  if(*tag==WRITEBACK)
    thisHead.mode = 46;
  else
    thisHead.mode = 6;

  pif_write_gh_(&thisHead);

  m=0;  /* Only 1 image */
      
  thisHead.datai4_min = (int)img_min;
  thisHead.datar4_min = (int)img_min;
  thisHead.datai4_max = (float)img_max;
  thisHead.datar4_max = (float)img_max;
  thisHead.datar4_avg = img_avg;
  thisHead.datar4_std = img_std;
  thisHead.img_bkg = img_bkg;
  thisHead.packRadius = packRadius;
  if(featherMode==LINEAR)
    thisHead.borderType = 1;
  else if(featherMode==GAUSSIAN)
    thisHead.borderType = 2;
  else if(featherMode==EXPONENTIAL)
    thisHead.borderType = 3;
  else 
    thisHead.borderType = 0;
  thisHead.borderWidth = featherRadius-boxRadius;
  
  thisHead.imgNum = m+1;
  pif_write_dh_(&thisHead);
      
  /* [3] Now write the image info */
  imgOffset = 
    512+                   /* Gloal header */
      ((m+1)*512)+         /* Image header */
	((m)*              /* The image itself */
	 thisHead.numCols*
	 thisHead.numRows)*
	   sizeof(BYTE);
  
  if(*tag==WRITEFFT)
    {/* Write the FFT */
      pif_write_byte_image_(&(thisHead.luNum),
			    (temp_buffer),
			    &imgOffset,
			    &(thisHead.numCols),
			    &(thisHead.numRows),
			    &(thisHead.numSecs));
    }/* Write the FFT */
  else if(*tag==WRITEBACK)
    {/* Write the Back transform */
      pif_write_wmapi4_(&(thisHead),
			(tempF_buffer),
			&(thisHead.numCols),
			&(thisHead.numRows),
			&(thisHead.numSecs));
    }/* Write the Back transform */

  /* Flip it back from TSB space for RAW file */
  for(j=0;j<(thisHead.numRows/2);j++)
    {/* Rows */
      int yOut;
      BYTE thisChar;
      yOut = thisHead.numRows - j - 1;
      for(i=0;i<(thisHead.numCols);i++)
	{/* Do entire image */
	  thisChar= (BYTE)temp_buffer[((j*thisHead.numCols)+i)];
	  temp_buffer[((j*thisHead.numCols)+i)]= (BYTE)
	    temp_buffer[((yOut*thisHead.numCols)+i)];
	  temp_buffer[((yOut*thisHead.numCols)+i)]= (BYTE)thisChar;
	}/* Do entire image */
    }/* Rows */
  
  fwrite(temp_buffer,sizeof(BYTE),(thisHead.numCols*
				   thisHead.numRows*
				   thisHead.numSecs),fOut);
      
  fclose(fOut);
  pif_close(&thisHead.luNum);

  free(temp_buffer);

  sprintf(outStr,"Successfully wrote [%s] and [%s]\n\
Dimensions [%d][%d]\n",
	  outFileName,
	  rawOutFileName,
	  thisHead.numCols,
	  thisHead.numRows
	  );

  askThem(outStr,FALSE,XmDIALOG_INFORMATION);
  
}/* fft_writeFB_actIVATE */




/* ************************************************************
 * fft_activate() - Put the fftscreen on the display.
 */
void fft_activate(Widget widget, int *tag, 
		  XmAnyCallbackStruct *callback_data)
{/* fft_actIVATE */
  int i;
  XmString simpleString;

  if(purdueHeader.numCols > 0 && purdueHeader.numRows > 0)
    {/* OK to go */

      averageFFT= FALSE;    /* Don't want average anymore */

      raiseIt(fftscreen);

      XtSetArg(arglst[0], XmNsensitive, TRUE);
      XtSetValues(a128FFTWidget, arglst, 1); 
      XtSetValues(a256FFTWidget, arglst, 1); 
      XtSetValues(a512FFTWidget, arglst, 1);
      XtSetValues(a1024FFTWidget, arglst, 1); 

      /* Start as small as we can get */
      fftSize = 128;
      if(rectBox.drawable)
	{/* They may just want to look at their rectangle */
	  askThem("I'm using the rectangle box you defined\n\
earlier - NOT the entire image.",
		  FALSE,XmDIALOG_INFORMATION);
	  if(rectBox.width > 128 ||
	     rectBox.height > 128)
	    {/* CAn't do 128 x 128 */
	      XtSetArg(arglst[0], XmNsensitive, FALSE);
	      XtSetValues(a128FFTWidget, arglst, 1); 
	      fftSize = 256;
	    }/* CAn't do 128 x 128 */
	  if(rectBox.width > 256 ||
	     rectBox.height > 256)
	    {/* CAn't do 256 x 256 */
	      XtSetArg(arglst[0], XmNsensitive, FALSE);
	      XtSetValues(a256FFTWidget, arglst, 1); 
	      fftSize = 512;
	    }/* CAn't do 256 x 256 */
	}/* They may just want to look at their rectangle */
      else
	{/* Use whole image */
	  if(purdueHeader.numCols > 128 ||
	     purdueHeader.numRows > 128)
	    {/* CAn't do 128 x 128 */
	      XtSetArg(arglst[0], XmNsensitive, FALSE);
	      XtSetValues(a128FFTWidget, arglst, 1); 
	      fftSize = 256;
	    }/* CAn't do 128 x 128 */
	  if(purdueHeader.numCols > 256 ||
	     purdueHeader.numRows > 256 )
	    {/* CAn't do 256 x 256 */
	      XtSetArg(arglst[0], XmNsensitive, FALSE);
	      XtSetValues(a256FFTWidget, arglst, 1); 
	      fftSize = 512;
	    }/* CAn't do 256 x 256 */
	}/* Use whole image */


      /* Let's change the lopHip scrolls accordingly */
      XtSetArg(arglst[0], XmNmaximum, (int)(fftSize*1.222));
      XtSetArg(arglst[1], XmNminimum, (int)0);
      XtSetArg(arglst[2], XmNpageIncrement, (int)(fftSize/32));
      XtSetValues(lopScrollW, arglst, 3);
      XtSetValues(hipScrollW, arglst, 3);

      /* This may have been reset by rotflt e.g. */
      idim1= idim2= 1024;   /* MAX fft size */
      CK(fftImage_gold= (float *)malloc(sizeof(float)*(idim1*(idim2*2))));
      CK(fftImage_fourAver= (float *)malloc(sizeof(float)*(idim1*(idim2*2))));
      CK(fftImage_buffer2= (float *)malloc(sizeof(float)*(idim1*(idim2*2))));
      
      fftMode= TRUE;
      
      if(measureFFT)
	{/* Unmanage these from the FFT screen */

	  lopHipSwitch = FALSE;
	  drawCTF = FALSE;
	  measureFFT = TRUE;
	  pickLattice = -99;  /* Make them draw it all over again */
	  backTransform = FALSE;
	  currentLatPt=0;
	  /* Turn interface off */
	  XtSetArg(arglst[0], XmNset, FALSE);
	  /* XtSetValues(pickLatticeToggleW, arglst, 1);  */
	  XtSetValues(backFFTToggleW, arglst, 1);  
	  XtSetArg(arglst[0], XmNset, FALSE);
	  XtSetValues(pickCTFtoggleW, arglst, 1); 
	  
	  XtUnmanageChild(lophipForm);
	  XtUnmanageChild(ctfTogglesForm);
	  XtUnmanageChild(ctfCurveForm);
	  XtUnmanageChild(ctfScrollsForm);
	  XtUnmanageChild(ctfParamsW);
	  
	  XtUnmanageChild(latPointsForm);
	  XtUnmanageChild(latTogglesForm);
	  /* XtUnmanageChild(latResetW); */
	  XtUnmanageChild(latResetButW);
	  XtUnmanageChild(latResetLabelW);
	  /* XtUnmanageChild(filterControlsForm); */
	  
	  XtManageChild(fftMeasureForm);
	  XtManageChild(fftFeedbackForm);
	  
	  for(i=0;i<4;i++)
	    {/* Do all */
	      XtUnmanageChild(latticeScrollW[i]);
	      XtUnmanageChild(latticeLabel1W[i]);
	      XtUnmanageChild(latticeLabel2W[i]);
	    }/* Do all */
	  
	  /* Give feedback on FFT window title bar */
	  sprintf(outStr,"FFT - Measure FFT");
	  createSimpleString(&simpleString,outStr);
	  XtSetArg(arglst[0], XmNdialogTitle, simpleString );
	  XtSetArg(arglst[1], XmNrecomputeSize, FALSE );
	  XtSetValues(fftscreen, arglst, 2); 
	  XFlush(dsply);
	  XmStringFree(simpleString); 
	  XFlush(dsply);

	}/* Unmanage these from the FFT screen */
      
      /* Force the colormap to normal grayscale */
      for(i=7;i<(tryColors);i++)
	{/* Do the colors we have been assigned */
	  if(hiColor)
	    {/* hi */
	      color_defs[i].pixel = 
		gen_color_defs[i].pixel;
	      color_defs[i].red   = 
		gen_color_defs[i].red;
	      color_defs[i].green = 
		gen_color_defs[i].green;
	      color_defs[i].blue  = 
		gen_color_defs[i].blue;
	      color_defs[i].flags = 
		gen_color_defs[i].flags;
	    }/* hi */
	  else
	    {/* lo */
	      color_defs[pixels_return[i]].pixel = 
		gen_color_defs[pixels_return[i]].pixel;
	      color_defs[pixels_return[i]].red   = 
		gen_color_defs[pixels_return[i]].red;
	      color_defs[pixels_return[i]].green = 
		gen_color_defs[pixels_return[i]].green;
	      color_defs[pixels_return[i]].blue  = 
		gen_color_defs[pixels_return[i]].blue;
	      color_defs[pixels_return[i]].flags = 
		gen_color_defs[pixels_return[i]].flags;
	    }/* lo */
	  if(masterProcess)
	    {
	      if(hiColor)
		{XAllocColor (dsply, color_map, &color_defs[i]);
		 pixels_return[i] = color_defs[i].pixel;}
	      else
		XStoreColor (dsply, color_map, &color_defs[pixels_return[i]]);
	    }
	}/* Do the colors we have been assigned */

      if(fftSize==128)
	RwaActivateWidget(a128FFTWidget); 
      else if(fftSize==256)
	RwaActivateWidget(a256FFTWidget); 
      else if(fftSize==512)
	RwaActivateWidget(a512FFTWidget); 
      else if(fftSize==1024)
	RwaActivateWidget(a1024FFTWidget); 
      
      calcFFT(TRUE, 0, 0);
      DrawFFTWindow();

    }/* OK to go */
  
}/* fft_actIVATE */




/* ************************************************************
 * fftForm_create() - Remember form widget numbers.
 */
void fftForm_create(Widget widget, int *tag, 
		    XmAnyCallbackStruct *callback_data)
{/* fftForm_create */
  
  if(*tag==LATREFINE)
    {/* Manage all of the lattice refinement stuff*/
      latPointsForm = widget;
    }/* Manage all of the lattice refinement stuff*/
  else if(*tag==LATTOGGLES)
    {/* Manage all of the lattice refinement stuff*/
      latTogglesForm = widget;
    }/* Manage all of the lattice refinement stuff*/
  else if(*tag==LATRESETBUT)
    {/* reset and save */
      latResetButW = widget;
    }/* reset and save */
  else if(*tag==LATRESETLABEL)
    {/* reset and save */
      latResetLabelW = widget;
    }/* reset and save */
  else if(*tag==LATRESET)
    {/* reset and save */
      latResetW = widget;
    }/* reset and save */
  else if(*tag==FILTCONTROLS)
    {/* filt controsl */
      filterControlsForm= widget;
    }/* filt controsl */
  else if(*tag==FFTIMAGES)
    fftImagesForm= widget;
  else if(*tag==CTFTOGGLES)
    ctfTogglesForm= widget;
  else if(*tag==CTFPARAMS)
    ctfParamsW= widget;
  else if(*tag==DISPTOGGLES)
    displayFFTForm= widget;
  else if(*tag==CTFCURVE)
    ctfCurveForm= widget;
  else if(*tag==CTFSCROLLS)
    ctfScrollsForm= widget;
  else if(*tag==FFTMEASURE)
    fftMeasureForm= widget;
  else if(*tag==FFTMEASPIXL)
    fftFeedbackForm= widget;
  else if(*tag==LOPHIP)
    lophipForm= widget;

}/* fftForm_create */





/* ************************************************************
 * ctf_reset_activate() - Init CTF params based on what 
 * microscope they've chosen for the default.
 */
void ctf_reset_activate(widget, tag, callback_data)
Widget  widget;
int     *tag;
XmAnyCallbackStruct *callback_data;
{/* CTF_RESET_ACTIVATE */
  int i,j,thisReso;
  XmString simpleString;
  int spCoExp;
  float 
    afractThisValue, 
    pfractThisValue ,
    resolutionThisValue, 
    avoltageThisValue, 
    spaberrationThisValue, 
    chaberrationThisValue, 
    spCoherenceThisValue, 
    spCoFactor ,
    tmCoherenceThisValue, 
    majDefocusThisValue, 
    minDefocusThisValue, 
    majAngThisValue, 
    wienerThisValue ,
    defocusThisValue ;
  char inStr[256];


  if(*tag==ALL)
    {/* normal */
      afractThisValue = 0.0;
      pfractThisValue = 1.0;
      /*      resolutionThisValue = 1.0;  */
      avoltageThisValue = 100000.0;
      avoltageThisValue = 200000.0;   /* Default to 200 kV for FEG */
      spaberrationThisValue = 2.0;
      chaberrationThisValue = 2.0;
      spCoherenceThisValue = 5.0;
      spCoExp = -3;
      spCoFactor = 0.001;
      tmCoherenceThisValue = 0.0;
      majDefocusThisValue = 1.0;
      minDefocusThisValue = 1.0;
      majAngThisValue = 0.0;
      wienerThisValue = 0.0;
      defocusThisValue = 1.000;
    }/* normal */
  else if(*tag==FEG)
    {/* CM200 feg */
      afractThisValue = 0.05;
      pfractThisValue = 0.95;
      /*      resolutionThisValue = 1.0; */
      avoltageThisValue = 200000.0;
      spaberrationThisValue = 2.0;
      chaberrationThisValue = 2.0;
      spCoherenceThisValue = 5.0;
      spCoExp = -2;
      spCoFactor = 0.01;
      tmCoherenceThisValue = 0.5;
      majDefocusThisValue = 1.0;
      minDefocusThisValue = 1.0;
      majAngThisValue = 0.0;
      wienerThisValue = 0.0;
      defocusThisValue = 1.000;
    }/* feg */
  else if(*tag==FEG300)
    {/* feg 300 */
      afractThisValue = 0.04;
      pfractThisValue = 0.96;
      /*      resolutionThisValue = 1.0;  */
      avoltageThisValue = 300000.0;
      spaberrationThisValue = 2.0;
      chaberrationThisValue = 2.0;
      spCoherenceThisValue = 5.0;
      spCoExp = -3;
      spCoFactor = 0.001;
      tmCoherenceThisValue = 0.5;
      majDefocusThisValue = 1.0;
      minDefocusThisValue = 1.0;
      majAngThisValue = 0.0;
      wienerThisValue = 0.0;
      defocusThisValue = 1.000;
    }/* feg 300 */
  else if(*tag==EM300)
    {/* nermag */
      afractThisValue = 0.07;
      pfractThisValue = 0.93;
      /*      resolutionThisValue = 1.0; */
      avoltageThisValue = 100000.0;
      spaberrationThisValue = 1.6;
      chaberrationThisValue = 1.3;
      spCoherenceThisValue = 5.0;
      spCoExp = -3;
      spCoFactor = 0.001;
      tmCoherenceThisValue = 1.6;
      majDefocusThisValue = 1.0;
      minDefocusThisValue = 1.0;
      majAngThisValue = 0.0;
      wienerThisValue = 0.0;
      defocusThisValue = 1.000;
    }/* nermag */
  else if(*tag==EM420W)
    {/* em420 */
      afractThisValue = 0.07;
      pfractThisValue = 0.93;
      /*      resolutionThisValue = 1.0; */
      avoltageThisValue = 120000.0;
      spaberrationThisValue = 2.0;
      chaberrationThisValue = 2.0;
      spCoherenceThisValue = 1.0;
      /* Now set the exponent to 00 */
      spCoFactor = 1.0;
      spCoExp = 0;
      tmCoherenceThisValue = 1.6;
      majDefocusThisValue = 1.0;
      minDefocusThisValue = 1.0;
      majAngThisValue = 0.0;
      wienerThisValue = 0.0;
      defocusThisValue = 1.000;
    }/* em420 */
  else if(*tag==EM420L)
    {/* em420 LaB6 */
      afractThisValue = 0.07;
      pfractThisValue = 0.93;
      /*      resolutionThisValue = 1.0; */
      avoltageThisValue = 120000.0;
      spaberrationThisValue = 2.0;
      chaberrationThisValue = 2.0;
      spCoherenceThisValue = 0.5;
      spCoExp = 0;
      spCoFactor = 1.0;
      tmCoherenceThisValue = 0.8;
      majDefocusThisValue = 1.0;
      minDefocusThisValue = 1.0;
      majAngThisValue = 0.0;
      wienerThisValue = 0.0;
      defocusThisValue = 1.000;
    }/* em420 LaB6 */
  else if(*tag==JEOL_EM_2010F_HT_FEG)
    {/* JEOL_EM-2010F-HT_FEG */
      afractThisValue = 0.05;
      pfractThisValue = 0.95;
      /*      resolutionThisValue = 1.0; */
      avoltageThisValue = 200000.0;
      spaberrationThisValue = 1.4;
      chaberrationThisValue = 1.8;
      spCoherenceThisValue = 5.0;
      spCoExp = -3;
      spCoFactor = 0.001;
      tmCoherenceThisValue = 0.5;
      majDefocusThisValue = 1.0;
      minDefocusThisValue = 1.0;
      majAngThisValue = 0.0;
      wienerThisValue = 0.0;
      defocusThisValue = 1.000;
    }/* JEOL_EM-2010F-HT_FEG */
  else if(*tag==JEOL_EM_2010F_UHR_FEG)
    {/* JEOL_EM-2010F-UHR_FEG */
      afractThisValue = 0.05;
      pfractThisValue = 0.95;
      /*      resolutionThisValue = 1.0; */
      avoltageThisValue = 200000.0;
      spaberrationThisValue = 0.5;
      chaberrationThisValue = 1.1;
      spCoherenceThisValue = 5.0;
      spCoExp = -3;
      spCoFactor = 0.001;
      tmCoherenceThisValue = 0.5;
      majDefocusThisValue = 1.0;
      minDefocusThisValue = 1.0;
      majAngThisValue = 0.0;
      wienerThisValue = 0.0;
      defocusThisValue = 1.000;
    }/* JEOL_EM-2010F-UHR_FEG */
  else if(*tag==JEOL1010)
    {/* JEOL1010 */
      afractThisValue = 0.05;
      pfractThisValue = 0.95;
      /*      resolutionThisValue = 1.0; */
      avoltageThisValue = 100000.0;
      spaberrationThisValue = 1.9;
      chaberrationThisValue = 1.9;
      spCoherenceThisValue = 5.0;
      spCoExp = -3;
      spCoFactor = 0.001;
      tmCoherenceThisValue = 0.5;
      majDefocusThisValue = 1.0;
      minDefocusThisValue = 1.0;
      majAngThisValue = 0.0;
      wienerThisValue = 0.0;
      defocusThisValue = 1.000;
    }/* JEOL1010 */
  else if(*tag==Hitachi_H9000)
    {/* Hitachi_H9000 */
      afractThisValue = 0.05;
      pfractThisValue = 0.95;
      /*      resolutionThisValue = 1.0; */
      avoltageThisValue = 120000.0;
      spaberrationThisValue = 0.9;
      chaberrationThisValue = 2.0;
      spCoherenceThisValue = 0.5;
      spCoExp = -3;
      spCoFactor = 0.001;
      tmCoherenceThisValue = 0.5;
      majDefocusThisValue = 1.0;
      minDefocusThisValue = 1.0;
      majAngThisValue = 0.0;
      wienerThisValue = 0.0;
      defocusThisValue = 1.000;
    }/* Hitachi_H9000 */
  else if(*tag==Hitachi_HF2000)
    {/* Hitachi_HF2000 */
      afractThisValue = 0.05;
      pfractThisValue = 0.95;
      /*      resolutionThisValue = 1.0; */
      avoltageThisValue = 200000.0;
      spaberrationThisValue = 1.2;
      chaberrationThisValue = 1.5;
      spCoherenceThisValue = 0.2;
      spCoExp = -3;
      spCoFactor = 0.001;
      tmCoherenceThisValue = 0.3;
      majDefocusThisValue = 1.0;
      minDefocusThisValue = 1.0;
      majAngThisValue = 0.0;
      wienerThisValue = 0.0;
      defocusThisValue = 1.000;
    }/* Hitachi_HF2000 */
  else if(*tag==FEI_Tecnai_G2_Polara)
    {/* FEI_Tecnai_G2_Polara */
      afractThisValue = 0.04;
      pfractThisValue = 0.96;
      /*      resolutionThisValue = 1.0;  */
      avoltageThisValue = 300000.0;
      spaberrationThisValue = 2.0;
      chaberrationThisValue = 2.0;
      spCoherenceThisValue = 0.1;
      spCoExp = -3;
      spCoFactor = 0.001;
      tmCoherenceThisValue = 0.7;
      majDefocusThisValue = 1.0;
      minDefocusThisValue = 1.0;
      majAngThisValue = 0.0;
      wienerThisValue = 0.0;
      defocusThisValue = 1.000;
    }/* FEI_Tecnai_G2_Polara */
  else if(*tag==FEI_Tecnai_G2_Sphera)
    {/* FEI_Tecnai_G2_Sphera */
      afractThisValue = 0.05;
      pfractThisValue = 0.95;
      /*      resolutionThisValue = 1.0; */
      avoltageThisValue = 200000.0;
      spaberrationThisValue = 2.0;
      chaberrationThisValue = 2.0;
      spCoherenceThisValue = 1.0;
      spCoExp = -3;
      spCoFactor = 0.001;
      tmCoherenceThisValue = 1.0;
      majDefocusThisValue = 1.0;
      minDefocusThisValue = 1.0;
      majAngThisValue = 0.0;
      wienerThisValue = 0.0;
      defocusThisValue = 1.000;
    }/* FEI_Tecnai_G2_Sphera */

  sprintf(inStr,"%.2f",afractThisValue);
  XmTextSetString(ctfAmpFactorW,inStr); 
  sprintf(inStr,"%.0f",avoltageThisValue*0.001);  /* In Kv */
  XmTextSetString(ctfKvW,inStr); 
  sprintf(inStr,"%.1f",spaberrationThisValue);
  XmTextSetString(ctfSpAberW,inStr); 
  sprintf(inStr,"%.1f",wienerThisValue);
  XmTextSetString(ctfWienerW,inStr); 


}/* CTF_RESET_ACTIVATE */






/*************************************************************
 * ctf_focus() - Give them feedback as to units and what 
 * exactly the values seen on the FFT screen are for.
 */
void ctf_focus(Widget widget, int *tag, XmAnyCallbackStruct *callback_data)
{/* CTF_FOCUS */
  XmString simpleString;

  if(widget==ctfModeW)
    sprintf(outStr,"MODE= 0,1,2,3");
  else if(widget==ctfScanStepW)
    sprintf(outStr,"ScanStep in");
  else if(widget==ctfPixSizeW)
    sprintf(outStr,"PixelSize in");
  else if(widget==ctfAmpFactorW)
    sprintf(outStr,"Amplitude Contrast"); 
 else if(widget==ctfKvW)
    sprintf(outStr,"Voltage in");
  else if(widget==ctfWaveW)
    sprintf(outStr,"Wavelength is");
  else if(widget==ctfSpAberW)
    sprintf(outStr,"Spherical Aberration");
  else if(widget==ctfMajorDefocusW)
    sprintf(outStr,"Major Defocus");
  else if(widget==ctfMinorDefocusW)
    sprintf(outStr,"Minor Defocus");
  else if(widget==ctfMagW)
    sprintf(outStr,"Magnification");
  else if(widget==ctfWienerW)
    sprintf(outStr,"Wiener is ");
  else if(widget==ctfAngMajorW)
    sprintf(outStr,"Major Angle");
  else if(widget==ctfTFactorW)
    sprintf(outStr,"Temp Factor");

  createSimpleString(&simpleString,outStr);
  XtSetArg(arglst[0], XmNlabelString, simpleString);
  XtSetValues(ctfHelpLabelW, arglst, 1); 
  XFlush(dsply);
  XmStringFree(simpleString); 
  XFlush(dsply);

  if(widget==ctfModeW)
    sprintf(outStr,"");
  else if(widget==ctfScanStepW)
    sprintf(outStr,"microns");
  else if(widget==ctfPixSizeW)
    sprintf(outStr,"nanometers");
  else if(widget==ctfAmpFactorW)
    sprintf(outStr,"in percent"); 
 else if(widget==ctfKvW)
    sprintf(outStr,"kiloVolts");
  else if(widget==ctfWaveW)
    sprintf(outStr,"auto-calculated");
  else if(widget==ctfSpAberW)
    sprintf(outStr,"in millimeters");
  else if(widget==ctfMajorDefocusW)
    sprintf(outStr,"in microns");
  else if(widget==ctfMinorDefocusW)
    sprintf(outStr,"in microns");
  else if(widget==ctfMagW)
    sprintf(outStr,"in mag");
  else if(widget==ctfWienerW)
    sprintf(outStr,"unitless");
  else if(widget==ctfAngMajorW)
    sprintf(outStr,"in degrees");
  else if(widget==ctfTFactorW)
    sprintf(outStr,"in Ang^2");

  createSimpleString(&simpleString,outStr);
  XtSetArg(arglst[0], XmNlabelString, simpleString);
  XtSetValues(ctfHelpLabel2W, arglst, 1); 
  XFlush(dsply);
  XmStringFree(simpleString); 
  XFlush(dsply);

}/* CTF_FOCUS */



/* ************************************************************
 * fft_mode_activate() - Select which mode to run in.
 */
void fft_mode_activate(Widget widget, int *tag, 
		       XmAnyCallbackStruct *callback_data)
{/* fft_mode_actIVATE */
  int i,saveBackTransform;
  XmString simpleString;
  int ctfMode, view, units, fft_full, filter;
  float volts,pixelSize,ampContrast_in,defocus_in,
        defocus_minor_in,input_phi,sphericalAbbr;
  char *ctfText;
  float angMajor,tFactor,wiener,d_star,pi,calcWavelength;
  float radian,degree;
  int mode, thisTag;
  float filt_fac[10];
  float cc,beta,dE;

  thisTag = *tag;
  if(*tag == APPLYINVERSE1CTF)
    {ctfFilter = 1;
    thisTag = APPLYINVERSECTF;}
  else if(*tag == APPLYINVERSE2CTF)
    {ctfFilter = 2;
    thisTag = APPLYINVERSECTF;}
  else if(*tag == APPLYINVERSE3CTF)
    {ctfFilter = 3;
    thisTag = APPLYINVERSECTF;}
  else if(*tag == APPLYINVERSE4CTF)
    {ctfFilter = 4;
    thisTag = APPLYINVERSECTF;}
  else if(*tag == APPLYINVERSE5CTF)
    {ctfFilter = 5;
    thisTag = APPLYINVERSECTF;}


  if(thisTag==APPLYCTF || thisTag==APPLYINVERSECTF || 
     thisTag==APPLYINVERSECTF2 || thisTag==APPLYINVERSECTF3 || 
     thisTag==BACKTRANS)
    {/* Apply CTF to current FFT and back transform */

      ctfMode = 1;
      view    = 1;

      ctfText = XmTextGetString(ctfKvW); 
      volts   = (int)strtod(ctfText, 0);
      volts *= 1000.0;    /* Convert from kiloVolts to volts */
      XtFree(ctfText);

      ctfText = XmTextGetString(ctfPixSizeW); 
      pixelSize   = (float)strtod(ctfText, 0);
      XtFree(ctfText);


      /* These should be in microns ??? */
      ctfText = XmTextGetString(ctfMajorDefocusW); 
      ctfMajorDefocusValue   = (float)strtod(ctfText, 0);
      XtFree(ctfText);
      ctfText = XmTextGetString(ctfMinorDefocusW); 
      ctfMinorDefocusValue   = (float)strtod(ctfText, 0);
      XtFree(ctfText);
      defocus_in        = ctfMajorDefocusValue;
      defocus_minor_in  = ctfMinorDefocusValue;

      input_phi   =  0.0;

      ctfText = XmTextGetString(ctfSpAberW); 
      sphericalAbbr   = (float)strtod(ctfText, 0);
      XtFree(ctfText);

/*********************************************
 *
 *   Never EVER get wavelength from interface 

   ALWAYS calculate it from volts !!!!
      ctfText = XmTextGetString(ctfWaveW); 
      wavelength   = (float)strtod(ctfText, 0);
      XtFree(ctfText);
**********************************************/
      wavelength = wavel_(&volts);

      ctfText = XmTextGetString(ctfAmpFactorW); 
      ampContrast_in = (float)strtod(ctfText, 0);
      XtFree(ctfText);

      if(pixelSize != 0.0)
	{/* OK to do */

	  units = 2;  /* e.g. we're using nanometers for pixelSize */
	  fft_full = 1;
	  d_star = 0.025;
	  pi = PI;
	  radian = 0.0;
	  piraddeg_(&pi, &radian, &degree);

	  ctfText = XmTextGetString(ctfModeW); 
	  mode = (int)strtod(ctfText, 0);
	  XtFree(ctfText);
	  ctfText = XmTextGetString(ctfWienerW); 
	  wiener = (float)strtod(ctfText, 0);
	  XtFree(ctfText);
	  if(wiener==0.0)
	    wiener= 0.001;
	  sprintf(outStr,"%f",wiener);
	  XmTextSetString(ctfWienerW,outStr);
	  XFlush(dsply);
	  ctfText = XmTextGetString(ctfAngMajorW); 
	  angMajor = (float)strtod(ctfText, 0);
	  /* We need to change the sign of angMajor here because our
	   * XWindows based buffer is upside down.  When we write
	   * to disk we need to use the sign just as given from
	   * the interface */
	  angMajor = 0.0 - angMajor ;
	  XtFree(ctfText);
	  ctfText = XmTextGetString(ctfTFactorW); 
	  tFactor = (float)strtod(ctfText, 0);
	  XtFree(ctfText);


	  /* Init these for now - get from interface eventually. */
	  cc = 0.0;
	  dE= 0.0;
	  beta = 0.0;

	  filt_fac[0] = wiener;
	  filt_fac[1] = 0.1;
	  if(thisTag==APPLYCTF)
	    {ctf_multiply_(&mode,
			  fftImage_gold,
			  &fftSize, &fftSize, 
			  &units, 
			  &pixelSize,
			  &pi, &radian,
			  &wavelength,
			  &defocus_in, &defocus_minor_in,
			  &angMajor,
			  &ampContrast_in,
			  &sphericalAbbr, 
			  &cc,&beta,&dE,&volts,
			  &tFactor, &fft_full
			  );}
	  else if(thisTag==APPLYINVERSECTF)
	    {
	      filter = ctfFilter;
	      ctf_divide_(&mode,
			  &filter,
			  fftImage_gold,
			  &fftSize, &fftSize, 
			  &units, 
			  &pixelSize,
			  &pi, &radian,
			  &wavelength,
			  &defocus_in, &defocus_minor_in,
			  &angMajor,
			  &ampContrast_in,
			  &sphericalAbbr, 
			  filt_fac,
			  &tFactor,
			  &fft_full,
			  &cc,&beta,&dE,&volts
			  );
	    }
	  
	  applyCTF = FALSE;
	  saveBackTransform = backTransform;
	  backTransform = TRUE;  /* Let's back transform it */

	  calcFFT(FALSE, 0, FALSE);

	  XSync(dsply,FALSE);
	  XFlush(dsply);
	  DrawFFTWindow();
	  /*raiseIt(main_window_widget); 
	  raiseIt(fftscreen); */
	  XFlush(dsply);

	  /* Now stuff it back they way it used to be */
	  backTransform = saveBackTransform;

	}/* OK to do */
      else
	{/* Not yet */
	  askThem("You've got to enter a valid PixelSize.",
		  FALSE,XmDIALOG_INFORMATION);
	}/* Not yet */
      
    bugOut:;
      
    }/* Apply CTF to current FFT and back transform */
  else if(thisTag==LOPHIP)
    {/* Manage all of the LOP/HIP stuff */

      lopHipSwitch = TRUE;
      drawCTF = FALSE;
      pickLattice = -99;  /* Make them draw it all over again */
      /* We will force them to pick lattic epoints here */
      pickedLatPts= FALSE;
      backTransform = FALSE;
      XtSetArg(arglst[0], XmNsensitive, TRUE);
      XtSetValues(drawHiLoToggleW,arglst,1);

      currentLatPt=0;
      /* Turn interface off */
      XtSetArg(arglst[0], XmNset, FALSE);
      /* XtSetValues(pickLatticeToggleW, arglst, 1);  */
      XtSetValues(backFFTToggleW, arglst, 1);  
      XtSetArg(arglst[0], XmNset, FALSE);
      XtSetValues(pickCTFtoggleW, arglst, 1); 

      XtManageChild(lophipForm);
      XtUnmanageChild(ctfTogglesForm);
      XtUnmanageChild(ctfCurveForm);
      XtUnmanageChild(ctfScrollsForm);
      XtUnmanageChild(ctfParamsW);

      XtUnmanageChild(latPointsForm);
      XtUnmanageChild(latTogglesForm);
      /* XtUnmanageChild(latResetW); */
      XtUnmanageChild(latResetButW);
      XtUnmanageChild(latResetLabelW);
      /* XtUnmanageChild(filterControlsForm); */

      XtUnmanageChild(fftMeasureForm);
      XtUnmanageChild(fftFeedbackForm);

      for(i=0;i<4;i++)
	{/* Do all */
	  XtUnmanageChild(latticeScrollW[i]);
	  XtUnmanageChild(latticeLabel1W[i]);
	  XtUnmanageChild(latticeLabel2W[i]);
	}/* Do all */

      measureFFT = FALSE;

      /* Give feedback on FFT window title bar */
      sprintf(outStr,"FFT - LOP/HIP FIltering");
      createSimpleString(&simpleString,outStr);
      XtSetArg(arglst[0], XmNdialogTitle, simpleString );
      XtSetArg(arglst[1], XmNrecomputeSize, FALSE );
      XtSetValues(fftscreen, arglst, 2); 
      XFlush(dsply);
      XmStringFree(simpleString); 
      XFlush(dsply);

    }/* Manage all of the LOP/HIP stuff */
  else if(thisTag==CTF)
    {/* Manage all of the CTF stuff */

      lopHipSwitch = FALSE;
      drawCTF = TRUE;
      pickLattice = -99;  /* Make them draw it all over again */
      /* We will force them to pick lattic epoints here */
      pickedLatPts= FALSE;
      backTransform = FALSE;
      currentLatPt=0;
      /* Turn interface off */
      XtSetArg(arglst[0], XmNset, FALSE);
      /* XtSetValues(pickLatticeToggleW, arglst, 1);  */
      XtSetValues(backFFTToggleW, arglst, 1);  
      XtSetArg(arglst[0], XmNset, FALSE);
      XtSetValues(pickCTFtoggleW, arglst, 1); 

      XtManageChild(ctfTogglesForm);
      XtManageChild(ctfCurveForm);
      XtManageChild(ctfScrollsForm);
      XtManageChild(ctfParamsW);

      XtUnmanageChild(lophipForm);
      XtUnmanageChild(latPointsForm);
      XtUnmanageChild(latTogglesForm);
      /* XtUnmanageChild(latResetW); */
      XtUnmanageChild(latResetButW);
      XtUnmanageChild(latResetLabelW);
      /* XtUnmanageChild(filterControlsForm); */

      XtUnmanageChild(fftMeasureForm);
      XtUnmanageChild(fftFeedbackForm);

      for(i=0;i<4;i++)
	{/* Do all */
	  XtUnmanageChild(latticeScrollW[i]);
	  XtUnmanageChild(latticeLabel1W[i]);
	  XtUnmanageChild(latticeLabel2W[i]);
	}/* Do all */

      measureFFT = FALSE;

      /* Give feedback on FFT window title bar */
      sprintf(outStr,"FFT - CTF Estimation");
      createSimpleString(&simpleString,outStr);
      XtSetArg(arglst[0], XmNdialogTitle, simpleString );
      XtSetArg(arglst[1], XmNrecomputeSize, FALSE );
      XtSetValues(fftscreen, arglst, 2); 
      XFlush(dsply);
      XmStringFree(simpleString); 
      XFlush(dsply);

    }/* Manage all of the CTF stuff */
  else if(thisTag==LATREFINE)
    {/* Manage all of the lattice refinement stuff*/


      /* Force it into 512 size mode */
      if(fftSize != 512)
	{/* Do it in 512 */
	  fftSize = 512;
	  RwaActivateWidget(a512FFTWidget); 
	  calcFFT(TRUE, 0, 0);
	  DrawFFTWindow();
	  askThem(
           "Lattice Refinement and Filtering only in 512x512 FFTs for now.",
		  FALSE,XmDIALOG_INFORMATION);
	}/* Do it in 512 */

      lopHipSwitch = FALSE;
      drawCTF = FALSE;
      pickLattice = 1;  /* Make them draw it all over again */
      backTransform = FALSE;
      currentLatPt=0;
      /* Turn interface off */
      XtSetArg(arglst[0], XmNset, FALSE);
      /* XtSetValues(pickLatticeToggleW, arglst, 1);  */
      XtSetValues(backFFTToggleW, arglst, 1);  
      XtSetArg(arglst[0], XmNset, FALSE);
      XtSetValues(pickCTFtoggleW, arglst, 1); 

      XtUnmanageChild(lophipForm);
      XtUnmanageChild(ctfTogglesForm);
      XtUnmanageChild(ctfCurveForm);
      XtUnmanageChild(ctfScrollsForm);
      XtUnmanageChild(ctfParamsW);

      XtManageChild(latPointsForm);
      XtManageChild(latTogglesForm);
      /*      XtManageChild(latResetW); */
      XtManageChild(latResetButW);
      XtManageChild(latResetLabelW);
      /* XtManageChild(filterControlsForm); */

      XtUnmanageChild(fftMeasureForm);
      XtUnmanageChild(fftFeedbackForm);


      for(i=0;i<4;i++)
	{/* Do all */
	  XtManageChild(latticeScrollW[i]);
#define  USETHIS_IN_DEBUGGING 1
#ifdef USETHIS_IN_DEBUGGING
	  XtManageChild(latticeLabel1W[i]);
#endif /* #ifdef USETHIS_IN_DEBUGGING */
	  XtManageChild(latticeLabel2W[i]);
	}/* Do all */


      measureFFT = FALSE;

      /* Give feedback on FFT window title bar */
      sprintf(outStr,"FFT - Lattice Refinement");
      createSimpleString(&simpleString,outStr);
      XtSetArg(arglst[0], XmNdialogTitle, simpleString );
      XtSetArg(arglst[1], XmNrecomputeSize, FALSE );
      XtSetValues(fftscreen, arglst, 2); 
      XFlush(dsply);
      XmStringFree(simpleString); 
      XFlush(dsply);


    }/* Manage all of the lattice refinement stuff*/
  else if(thisTag==FFTMEASURE)
    {/* UnManage all */

      lopHipSwitch = FALSE;
      drawCTF = FALSE;
      measureFFT = TRUE;
      pickLattice = -99;  /* Make them draw it all over again */
      backTransform = FALSE;
      currentLatPt=0;
      /* Turn interface off */
      XtSetArg(arglst[0], XmNset, FALSE);
      /* XtSetValues(pickLatticeToggleW, arglst, 1);  */
      XtSetValues(backFFTToggleW, arglst, 1);  
      XtSetArg(arglst[0], XmNset, FALSE);
      XtSetValues(pickCTFtoggleW, arglst, 1); 

      XtUnmanageChild(lophipForm);
      XtUnmanageChild(ctfTogglesForm);
      XtUnmanageChild(ctfCurveForm);
      XtUnmanageChild(ctfScrollsForm);
      XtUnmanageChild(ctfParamsW);

      XtUnmanageChild(latPointsForm);
      XtUnmanageChild(latTogglesForm);
      /* XtUnmanageChild(latResetW); */
      XtUnmanageChild(latResetButW);
      XtUnmanageChild(latResetLabelW);
      /* XtUnmanageChild(filterControlsForm); */

      XtManageChild(fftMeasureForm);
      XtManageChild(fftFeedbackForm);

      sprintf(outStr,"[] Pixels.");
      createSimpleString(&simpleString,outStr);
      XtSetArg(arglst[0], XmNlabelString, simpleString);
      XtSetValues(fftMeasPixW, arglst, 1); 
      XFlush(dsply);
      XmStringFree(simpleString); 

      sprintf(outStr,"[] Angstroms.");
      createSimpleString(&simpleString,outStr);
      XtSetArg(arglst[0], XmNlabelString, simpleString);
      XtSetValues(fftMeasAngW, arglst, 1); 
      XFlush(dsply);
      XmStringFree(simpleString); 

      for(i=0;i<4;i++)
	{/* Do all */
	  XtUnmanageChild(latticeScrollW[i]);
	  XtUnmanageChild(latticeLabel1W[i]);
	  XtUnmanageChild(latticeLabel2W[i]);
	}/* Do all */

      /* Give feedback on FFT window title bar */
      sprintf(outStr,"FFT - Measure FFT");
      createSimpleString(&simpleString,outStr);
      XtSetArg(arglst[0], XmNdialogTitle, simpleString );
      XtSetArg(arglst[1], XmNrecomputeSize, FALSE );
      XtSetValues(fftscreen, arglst, 2); 
      XFlush(dsply);
      XmStringFree(simpleString); 
      XFlush(dsply);

    }/* UnManage all */
  else
    {/* Done */
      fft_screen_deactivate(0,0,0);
    }/* Done */
	
  /* Make them do a new circavg */
  calcFFT1DcircAvg= FALSE;
  

}/* fft_mode_actIVATE */




/* ************************************************************
 * ctf_appodize_activate() - Get appodizing info for writing
 * This from FFT.
 */
void ctf_appodize_activate(Widget widget, int *tag, 
			   XmAnyCallbackStruct *callback_data)
{/* CTF_APPODIZE_ACTIVATE */

  appodizing = WRITEFFT;
  writeFFTType = *tag;
  /* This is border width now */
  sprintf(outStr,"%d",featherRadius-boxRadius);
  XmTextSetString(appodBoxRadW,outStr);
  sprintf(outStr,"%d",featherRadius);
  XmTextSetString(appodFeaRadW,outStr);
  raiseIt(appodizeScreenWidget);

}/* CTF_APPODIZE_ACTIVATE */





/* ************************************************************
 * ctf_all_activate() - Apply a ctf multiplication or division.
 */
void ctf_all_activate()
{/* CTF_ALL_ACTIVATE */
  int i,saveBackTransform,bufNum;
  int ctfMode, view, units, fft_full;
  float volts,pixelSize,ampContrast_in,defocus_in,
        defocus_minor_in,input_phi,sphericalAbbr;
  char *ctfText;
  float angMajor,tFactor,wiener,d_star,pi,calcWavelength;
  float radian,degree;
  int mode,filter;
  int j,k,fftRow,fftCol,
        thisNcols,thisNrows,arg1,arg2,arg3,arg4,jindx,jindxp1;
  float distThis;
  int insideLimits,strtAvg,endAvg,thisImg,thisNimg,useRectangle;
  BYTE thisPixelValue;
  int ksign,iform,linear;
  float thisMin,thisMax,maxF,minF,offset,recipThisNimg;
  float theoreticalMax,theoreticalMin;
  float scaleFactor, thisFFTvalue, thisFloat, thisSign;
  float recipSize;
  int iTrecip,iTidim2, iTfftSize2, icount;
  float thisBackground;
  float *temp_buffer;
  float  thisPixel,*char_buffer, *thisBuffer, *tmp2;
  float *real_buffer;
  XmString simpleString;
  int thisMaxInt,thisMinInt,sliderSize, array_size, hiPixel,loPixel;
  float *fftImage_copy;
  forPifHead_t thisHead;
  int luNum,imgOffset;
  float img_bkg,img_min,img_max;
  float img_minF,img_maxF;
  float img_avg,img_std;
  fString_t fName,readWrite;
  float cc, dE, beta, filt_fac[10];
  int thisBoxRadius;
  BYTE *byte_buffer;
  int m,n;
  float min,max,scaleF;

  if(!doApplyCtf)
    {/* ok to check */
      sprintf(outStr,"I'm using CTF filter==[%d]  Do you want me to continue?",
	      ctfFilter);
      if(!askThem(outStr,
		  FALSE,XmDIALOG_INFORMATION))
	goto noDoit;
    }/* ok to check */


  numCols = purdueHeader.numCols;
  numRows = purdueHeader.numRows;

  CK(temp_buffer= (float *)malloc(
			 sizeof(float)*numCols*numRows));
  CK(real_buffer= (float *)malloc(
			 sizeof(float)*numCols*numRows*purdueHeader.nimg));

  /* Find out what fftSize they want and get the reciprical for
   * use later. */
  recipSize = (float)((float)fftSize / (float)idim1 );

  CK(char_buffer= (float *)malloc(
			 sizeof(float)*numCols*numRows));
  
  CK(fftImage_copy= (float *)malloc(sizeof(float)*idim1*idim2*2));
  
  /* Open up the output file */
  fName.addr = (char *)malloc(sizeof(char)*264);
  readWrite.addr = (char *)malloc(sizeof(char)*64);
  pif_init_head(&thisHead);
  sprintf(fName.addr,"ctf_corrected_imgs.pif");
  sprintf(readWrite.addr,"w");
  
  luNum = 7;
  thisHead.luNum = luNum;

  pif_open_(fName.addr, &luNum, readWrite.addr);

	      
  strtAvg = 0;
  endAvg  = purdueHeader.nimg;
  thisNimg = 0;
  /* Find out how many we've got */
  for(i=strtAvg;i<endAvg;i++)
    if(!deletePackedImage[i])
      thisNimg++;
  recipThisNimg = (float)(1.0 / (float)thisNimg);
  
  icount = 0;  /* Count the images */
  for(thisImg=strtAvg;thisImg < endAvg;thisImg++)
    {/* Do all images wanted */
      
      if(!deletePackedImage[thisImg])
	{/* OK to do */
	  
	  /* Tell them what we're doing */
	  sprintf(outStr,"%d",thisImg+1);
	  XmTextSetString(fftImgNumW,outStr);
	  XFlush(dsply);

	  
	  if(fileType==PACKEDPORTABLE || fileType==MAP)
	    sprintf(outStr,"%d",thisImg+1);
	  else
	    sprintf(outStr,"%d",1);
	  XmTextSetString(fftImgNumW,outStr);
	  XFlush(dsply);

	  /* Init out the FFT buffers */
	  for(i=0;i<idim1;i++)
	    for(j=0;j<(idim2*2);j++)
	      {/* Init all */
		fftImage_gold[(i*(idim2*2))+j]= 0.0;
	      }/* Init all */
	  
	  if(fileType == MAP)
	    {/* DO for MAP */
	      int nsec,em_wrt,fft_full,thisSec,fft_iend;
	      float fft_fzero, fft_origx, fft_origy;
	      thisBackground = purdueHeader.img_bkg;

	      nsec = 1;
	      em_wrt = 0;
	      thisSec = 1;
	      fft_full = 1;
	      fft_origx = fft_origy = 0.0;
	      fft_iend = 1;
	      fft_fzero = 0.0;
	      /* NOTE:  Use tsb orientation when writing to disk */
	      map_fft_fill_tsborien_((float_image_buffer1+
				 (thisImg*numCols*numRows)), 
				&numCols, &numRows,
				fftImage_gold, 
				&fftSize, &fftSize, 
				&thisBackground);

	    }/* DO for MAP */
	  else
	    {/* Do for packed images */
	      thisBuffer= (float *)packImg[(thisImg)]; 
	      
	      for(i=0;i<numRows;i++)
		for(j=0;j<numCols;j++)
		  {/* Do all */
/* WHy oh why do we have to do this here ??? - when to shift to TSB orientation
 * and when not to ????? 
We need to do this because we will write this in TSB orientation on
disk.
		    temp_buffer[(((numRows-i-1)*numCols)+j)]= (unsigned short)
		      thisBuffer[((i*numCols)+j)];
		    temp_buffer[((i*numCols)+j)]= (unsigned short)
		      thisBuffer[((i*numCols)+j)];
*/
		    temp_buffer[(((numRows-i-1)*numCols)+j)]= (float)
		      thisBuffer[((i*numCols)+j)];
		  }/* Do all */
	      
	      /* Get global */
	      if(packRadius > 0)
		purdueHeader.img_bkg = (float)(*packImg[thisImg]);
	      thisBackground = purdueHeader.img_bkg;


	      map_fft_fill_robemorien_(temp_buffer, &numCols, &numRows, 
				       fftImage_gold, 
				       &fftSize, &fftSize, &thisBackground);
	    }/* Do for packed images */
	  
	  ksign = 1;       /* Do a "forward" FFT */
	  iform = 1;       /* Data is imaginary */
	      
	  fft_2d_rob_ (fftImage_gold, &fftSize, &fftSize,
		       &ksign, &iform);

	  /* Apply CTF to current FFT and back transform */
	  
	  ctfMode = 1;
	  view    = 1;
	  
	  ctfText = XmTextGetString(ctfKvW); 
	  volts   = (int)strtod(ctfText, 0);
	  volts *= 1000.0;    /* Convert from kiloVolts to volts */
	  XtFree(ctfText);
	  
	  ctfText = XmTextGetString(ctfPixSizeW); 
	  pixelSize   = (float)strtod(ctfText, 0);
	  XtFree(ctfText);
	  
	  
	  /* These should be in microns ??? */
	  ctfText = XmTextGetString(ctfMajorDefocusW); 
	  ctfMajorDefocusValue   = (float)strtod(ctfText, 0);
	  XtFree(ctfText);
	  ctfText = XmTextGetString(ctfMinorDefocusW); 
	  ctfMinorDefocusValue   = (float)strtod(ctfText, 0);
	  XtFree(ctfText);
	  defocus_in        = ctfMajorDefocusValue;
	  defocus_minor_in  = ctfMinorDefocusValue;
	  
	  input_phi   =  0.0;
	  
	  ctfText = XmTextGetString(ctfSpAberW); 
	  sphericalAbbr   = (float)strtod(ctfText, 0);
	  XtFree(ctfText);
	  
/*********************************************
 *
 *   Never EVER get wavelength from interface 

   ALWAYS calculate it from volts !!!!
      ctfText = XmTextGetString(ctfWaveW); 
      wavelength   = (float)strtod(ctfText, 0);
      XtFree(ctfText);
**********************************************/
          wavelength = wavel_(&volts);
	  
	  ctfText = XmTextGetString(ctfAmpFactorW); 
	  ampContrast_in = (float)strtod(ctfText, 0);
	  XtFree(ctfText);
	  
	  if(doApplyCtf)
	    pixelSize  = applyCtfPixelSize;

	  if(pixelSize != 0.0)
	    {/* OK to do */
	      
	      units = 2;  /* e.g. we're using nanometers for pixelSize */
	      fft_full = 1;
	      d_star = 0.025;
	      pi = PI;
	      radian = 0.0;
	      piraddeg_(&pi, &radian, &degree);

	      
	      ctfText = XmTextGetString(ctfModeW); 
	      mode = (int)strtod(ctfText, 0);
	      XtFree(ctfText);
	      ctfText = XmTextGetString(ctfWienerW); 
	      wiener = (float)strtod(ctfText, 0);
	      XtFree(ctfText);
	      if(wiener==0.0)
		wiener= 0.001;
	      sprintf(outStr,"%f",wiener);
	      XmTextSetString(ctfWienerW,outStr);
	      XFlush(dsply);
	      ctfText = XmTextGetString(ctfAngMajorW); 
	      angMajor = (float)strtod(ctfText, 0);
	      /* We DO NOT need to change the sign of angMajor here. 
	       * we are writing
	       * to disk and we need to use the sign just as given from
	       * the interface.  e.g. The buffer was stuffed in
	       * TSB orientation above so no flipping needed.
	      angMajor = 0.0 - angMajor;   */
	      XtFree(ctfText);
	      ctfText = XmTextGetString(ctfTFactorW); 
	      tFactor = (float)strtod(ctfText, 0);
	      XtFree(ctfText);
	      
	      
	      if(doApplyCtf)
		{/* batch mode */
		  wiener     = applyCtfWiener;
		  tFactor    = applyCtfTFactor;
		  mode       = applyCtfMode;  /*(int) */
		  angMajor   = applyCtfAngMajor; /* in nm */
		  pixelSize  = applyCtfPixelSize;
		  defocus_in = applyCtfDefocus_in;
		  defocus_minor_in = applyCtfDefocus_minor_in;
		  ampContrast_in   = applyCtfAmpContrast_in;
		  sphericalAbbr    = applyCtfSphericalAbbr;
		  volts            = applyCtfVolts;  /* int in volts NOT 
						      * kiloVolts*/
		}/* batch mode */

	      /* Init these for now - get from interface eventually. */
	      cc = 0.0;
	      dE= 0.0;
	      beta = 0.0;

	      filt_fac[0] = wiener;
	      filt_fac[1] = 0.1;
	      filter = ctfFilter;
	      if(writeFFTType==APPLYCTF || writeFFTType==APPLYCTFR4)
		ctf_multiply_(&mode,
			      fftImage_gold,
			      &fftSize, &fftSize, 
			      &units, 
			      &pixelSize,
			      &pi, &radian,
			      &wavelength,
			      &defocus_in, &defocus_minor_in,
			      &angMajor,
			      &ampContrast_in,
			      &sphericalAbbr, 
			      &cc,&beta,&dE,&volts,
			      &tFactor, &fft_full
			      );
	      else if(writeFFTType==APPLYINVERSECTF || 
		      writeFFTType==APPLYINVERSECTFR4)
		{
		  ctf_divide_(&mode,
			      &filter,
			      fftImage_gold,
			      &fftSize, &fftSize, 
			      &units, 
			      &pixelSize,
			      &pi, &radian,
			      &wavelength,
			      &defocus_in, &defocus_minor_in,
			      &angMajor,
			      &ampContrast_in,
			      &sphericalAbbr, 
			      filt_fac,
			      &tFactor,
			      &fft_full,
			      &cc,&beta,&dE,&volts
			      );
		}

	      
	      /****** debugging stuff 
		      printf("fftSize=[%d]\n",fftSize);
		      bufNum = 0;
		      debug_fft_buffer_(fftImage_gold, &fftSize, &fftSize, 
		      &bufNum);
	      *************************/
	      applyCTF = FALSE;

	      /* Show them what the back transform is */
	      /* Now put it back in "normal" space */
	      ksign = 0-1;     /* Do a "reverse" FFT */
	      iform = 1;       /* Data is imaginary */
	      
	      /* Back transform into a copy of the filtered FFT just in case
	       * they want to do Fourier Averaging or etc... */ 
	      
	      for(i=0;i<idim1;i++)
		for(j=0;j<(idim2*2);j++)
		  {/* Init all */
		    fftImage_copy[(i*(idim2*2))+j]= 
		      fftImage_gold[(i*(idim2*2))+j] ;
		  }/* Init all */
	      

	      if(writeFFTType==APPLYCTF || writeFFTType==APPLYINVERSECTF)
		{/* DO it in BYTE */

		  fft_2d_rob_(fftImage_copy, &fftSize, &fftSize, 
			      &ksign, &iform);
	      
		  /* Let's look at what scaling needs done */
		  scaleFactor = 0.0;
		  findMaxMin(fftImage_copy, 
			     &maxF, &minF, &offset, FALSE, fftSize,fftSize,
			     purdueHeader.img_bkg);
		  
		  /* Defloat it */
		  maxF += (float)purdueHeader.img_bkg;
		  minF += (float)purdueHeader.img_bkg;
		  
		  if(maxF==minF)
		    scaleFactor= 1.0;
		  else scaleFactor = (float)(255.0) / (maxF-minF);
		  offset = minF;
		  
		  /* Now put the back-transformed image into the 
		   * display buffer */
		  thisBuffer = char_buffer;
		  
		  calcSizeFactor(zoomFactor);
#define  SCALEIT_here_742  1
#ifdef SCALEIT_here_742
		  for(j=0;j<(numRows);j++)
		    {/* Rows */
		      int yOut,xOut,l,m;
		      BYTE thisChar;
		      yOut = (j*sizeFactor);
		      for(i=0;i<(numCols);i++)
			{/* Do entire image */
			  /* Offset this one beside the original */
			  xOut = (i*sizeFactor)+(numCols*sizeFactor);
			  
			  jindx   = ((j*(int)(fftSize*2))+(i*2));
			  /* jindxp1 = ((j*(int)(fftSize*2))+(i*2)+1); */
			  
			  /* Definitely need to add img_bkg back in!!! 
			   * (To defloat)*/
			  thisFloat= fftImage_copy[jindx]+
			    (float)purdueHeader.img_bkg;
			  thisBuffer[((j*numCols)+i)]=  
			    (unsigned char)((thisFloat-minF)*scaleFactor);
			  
			}/* Do entire image */
		    }/* Rows */
#else
		  for(j=0;j<(numRows);j++)
		    {/* Rows */
		      int yOut,xOut,l,m;
		      BYTE thisChar;
		      yOut = (j*sizeFactor);
		      for(i=0;i<(numCols);i++)
			{/* Do entire image */
			  /* Offset this one beside the original */
			  xOut = (i*sizeFactor)+(numCols*sizeFactor);
			  
			  jindx   = ((j*(int)(fftSize*2))+(i*2));
			  /* jindxp1 = ((j*(int)(fftSize*2))+(i*2)+1); */
			  
			  /* Definitely need to add img_bkg back in!!! 
			   * (To defloat)*/
			  thisFloat= fftImage_copy[jindx]+
			    (float)purdueHeader.img_bkg;
			  thisBuffer[((j*numCols)+i)]=  
			    (unsigned char)((thisFloat));
			  
			}/* Do entire image */
		    }/* Rows */

#endif	/* ifdef SCALEIT */	  
		  
		  calcCircleStats(char_buffer,
				  &img_bkg, 
				  &img_min,
				  &img_max,
				  &img_avg, 
				  &img_std,
				  numCols, numRows, numCols);
		  
		  calcScaleFactor(&floatScaleFactor,		     
				  &recipFloatScaleFactor, 
				  &int4scaleF, 
				  img_min,img_max,
				  img_avg,img_std,
				  PACKEDPORTABLE);
		  
		  thisBuffer = char_buffer;

		  thisBoxRadius = featherRadius;
		  backChar = (float)img_bkg;
		  appodize_it(thisBuffer, 
			      numCols, 
			      numRows, numCols,
			      fileType,
			      bkgrndMethod,
			      boxMethod,
			      packRadius,
			      boxRadius,
			      featherRadius,
			      recipSizeFactor,
			      sizeFactor,
			      tryColors,
			      featherMode,
			      &backChar,
			      thisBoxRadius,
			      forceAvgBkg,
			      img_avg
			     );

		  thisHead.numCols = numCols;
		  thisHead.numRows = numRows;
		  thisHead.numSecs = 1;
		  /* Is this OBsolete???  Don't we ALWAYS write float data 
		   * from now on???  */
		  thisHead.mode = 6;
		  thisHead.nimg = thisNimg;
		  thisHead.floatScaleFactor = floatScaleFactor;
		  thisHead.recipFloatScaleFactor = recipFloatScaleFactor;
		  thisHead.int4scaleF = int4scaleF;
		  thisHead.htype= 1;
		  
		  pif_write_gh_(&thisHead);
		  
		  m=icount;
		  icount++;
		  
		  thisHead.datai4_min = (int)img_min;
		  thisHead.datar4_min = (int)img_min;
		  thisHead.datai4_max = (float)img_max;
		  thisHead.datar4_max = (float)img_max;
		  thisHead.datar4_avg = img_avg;
		  thisHead.datar4_std = img_std;
		  thisHead.img_bkg = img_bkg;
		  thisHead.packRadius = packRadius;
		  if(featherMode==LINEAR)
		    thisHead.borderType = 1;
		  else if(featherMode==GAUSSIAN)
		    thisHead.borderType = 2;
		  else if(featherMode==EXPONENTIAL)
		    thisHead.borderType = 3;
		  else 
		    thisHead.borderType = 0;
		  thisHead.borderWidth = featherRadius-boxRadius;
		  
		  thisHead.imgNum = m+1;
		  pif_write_dh_(&thisHead);
		  
		  /* [3] Now write the image info */
		  imgOffset = 
		    512+                   /* Gloal header */
		      ((m+1)*512)+         /* Image header */
			((m)*              /* The image itself */
			 thisHead.numCols*
			 thisHead.numRows)*
			   sizeof(BYTE);
		  
		  byte_buffer = (BYTE *)malloc(sizeof(BYTE)*
					 thisHead.numCols*thisHead.numRows);
		  for(m=0;m<thisHead.numRows;m++)
		    for(n=0;n<thisHead.numCols;n++)
		      {
			min = MIN(min,char_buffer[((m*thisHead.numCols)+n)]);
			max = MAX(max,char_buffer[((m*thisHead.numCols)+n)]);
		      }
		  scaleF = 255.0 / max;
		  for(m=0;m<thisHead.numRows;m++)
		    for(n=0;n<thisHead.numCols;n++)
		      {
			byte_buffer[((m*thisHead.numCols)+n)] = (BYTE)
			  (char_buffer[((m*thisHead.numCols)+n)] * scaleF);
		      }
		  pif_write_byte_image_(&(thisHead.luNum),
					(byte_buffer),
					&imgOffset,
					&(thisHead.numCols),
					&(thisHead.numRows),
					&(thisHead.numSecs));
		  
		  free(byte_buffer);
		}/* DO it in BYTE */
	      else
		{/* DO it R4 */
		  float *thisRealBuffer;

		  thisBackground = purdueHeader.img_bkg;

		  fft_2dbt_r4_(fftImage_copy, &fftSize, &fftSize, 
			       (real_buffer+(thisImg*numCols*numRows)),
			       &numCols, &numRows,
			       &fft_full,
			       &thisBackground);


		}/* DO it R4 */
	      
	      
	    }/* OK to do */
	  
	}/* OK to do */
    }/* Do all images wanted */
  
  if(writeFFTType==APPLYCTFR4 || writeFFTType==APPLYINVERSECTFR4)
    {/* Write R4 images */

      if(doApplyCtf)
	nimg = thisNimg;
      img_bkg = img_minF = img_maxF = img_avg = img_std = 0.0;
      calcRealStats(real_buffer,
		    &img_bkg, 
		    &img_minF,
		    &img_maxF,
		    &img_avg, 
		    &img_std,
		    numCols, numRows, numCols, nimg);
      printf("img_max/min[%f,%f]\n",img_maxF,img_minF);
      printf("nc,nr,nimg=[%d,%d,%d]\n",numCols,numRows,nimg);

      thisHead.numCols = numCols;
      thisHead.numRows = numRows;
      thisHead.numSecs = 1;
      thisHead.mode = 46;
      thisHead.nimg = thisNimg;
      thisHead.floatScaleFactor = (float)MAXINT / (2.0 + 
						   (float)(MAX((img_maxF+2.0),
					 fabs((float)(img_minF-2.0)))));
      thisHead.recipFloatScaleFactor = 1.0/(thisHead.floatScaleFactor);
      thisHead.int4scaleF = thisHead.floatScaleFactor;
      thisHead.htype= 1;
      thisHead.datai4_min = (int)img_minF;
      thisHead.datar4_min = (float)img_minF;
      thisHead.datai4_max = (int)img_maxF;
      thisHead.datar4_max = (float)img_maxF;
      thisHead.datar4_avg = img_avg;
      thisHead.datar4_std = img_std;
      
      pif_write_gh_(&thisHead);
      
      icount = 0;
      for(thisImg=strtAvg;thisImg < endAvg;thisImg++)
	{/* Do all images wanted */
	  int *thisPackImg;
	  float *thisFloatBuffer;
	  int nr,nc;
	  float minF,maxF,scaleF;

	  if(!deletePackedImage[thisImg])
	    {/* OK to do */
	      m=icount;
	      icount++;
	      
	      /* Now get stats on just this one */
	      calcRealStats((real_buffer+(thisImg*numCols*numRows)),
			    &img_bkg, 
			    &img_minF,
			    &img_maxF,
			    &img_avg, 
			    &img_std,
			    numCols, numRows, numCols, 1);
	      
	      thisHead.datai4_min = (int)img_minF;
	      thisHead.datar4_min = (float)img_minF;
	      thisHead.datai4_max = (int)img_maxF;
	      thisHead.datar4_max = (float)img_maxF;
	      thisHead.datar4_avg = img_avg;
	      thisHead.datar4_std = img_std;
	      thisHead.floatScaleFactor = (float)MAXINT / (2.0 + 
						   (float)(MAX((img_maxF+2.0),
					 fabs((float)(img_minF-2.0)))));
	      thisHead.recipFloatScaleFactor = 1.0/(thisHead.floatScaleFactor);
	      thisHead.int4scaleF = thisHead.floatScaleFactor;
	      thisHead.img_bkg = img_bkg;
	      thisHead.packRadius = packRadius;
	      if(featherMode==LINEAR)
		thisHead.borderType = 1;
	      else if(featherMode==GAUSSIAN)
		thisHead.borderType = 2;
	      else if(featherMode==EXPONENTIAL)
		thisHead.borderType = 3;
	      else 
		thisHead.borderType = 0;
	      thisHead.borderWidth = featherRadius-boxRadius;
	      
	      thisHead.imgNum = m+1;

	      pif_write_dh_(&thisHead);
	      
	      /* WHY ???? */
	      thisHead.imgOffset= 512 +
		(512 * (thisImg+1)) +
		((thisImg) * numCols * numRows *4);
	      pif_write_boxi4_(&thisHead, (real_buffer+
					   (thisImg*numCols*numRows)));
	      

	    }/* OK to do */
	}/* Do all images wanted */
    }/* Write R4 images */
  
  
  pif_close(&(thisHead.luNum));
  
  if(!doApplyCtf)
    askThem("Images written to ctf_corrected_imgs.pif file.",
	    FALSE,XmDIALOG_INFORMATION);

  free(fName.addr);
  free(readWrite.addr);
  free(char_buffer);
  free(temp_buffer);
  free(real_buffer);
  free(fftImage_copy);
  
  noDoit:;

}/* CTF_ALL_ACTIVATE */





/*************************************************************
 * fft_1DCA_activate() - Do a 1D circ avg
 */
void fft_1DCA_activate(Widget widget, int *tag, 
		       XmAnyCallbackStruct *callback_data)
{/* fft_1DCA_actIVATE */
  int   i,j,k,l;
  float *rint;
  int   *irnum;
  float xc,yc;
  int   ncol,nrow;
  float minCA, maxCA, scaleCircFactor;
  float tmpBuffer[(512*512)];
  int   scaledY, scaledYp1, thisX, thisXp1;
  int   interp,nterp;
  float thisBkg;
  FILE  *fOut;

  ncol = 512;
  nrow = 512;
  xc = (float)((float)ncol/2.0);
  yc = (float)((float)nrow/2.0);
  interp= 2;  /* Original way */
  interp= 1;  /* Check circAvgMotion() to make sure they agree */
  nterp = interp*(int)(
		       (int)((MAX(ncol,nrow)))
		       / (int)2 
		       ) 
    + 1;
  
  CK(rint= (float*)malloc(sizeof(float)*nterp*3));
	      
  CK(irnum= (int*)malloc(sizeof(int)*nterp*3));

  /* Stuff buffer */
  for(i=0;i<512;i++)
    for(j=0;j<512;j++)
      tmpBuffer[(i*512)+j]= (float)fft_gold_buffer1[(i*512)+j];

  thisBkg = 0.0;
  img_1d_rob(tmpBuffer,
	     ncol,nrow,
	     xc,yc,
	     interp,nterp,
	     rint,irnum,
	     thisBkg);

  free(irnum);
  
  /* Recalc the Y scaling factors based on this new offset.
   * We may be able to show more of the curves out here if
   * we've got a huge noise spike at the center of the map
   * e.g. */
  minCA= r4_max;
  maxCA= r4_min;

  /* Open up out file */
  fOut = fopen("fft1dca.txt","w");

  rint[0]= 0.0;     /* Hard code for now ... */
  for(i=0;i<256;i++)
    {/* DO all */
      if(rint[i]!=0.0)
	{/* OK to use */
	  minCA = MIN(minCA,rint[i]);
	  maxCA = MAX(maxCA,rint[i]);
	}/* OK to use */
      if(fOut && i > 0)
	{/* Print it */
	  float numAngs;
	  numAngs = ((512.0 / (float)i)*fftAngsPerPixel);
	  if(fftAngsPerPixel != 0.0)
	    fprintf(fOut,"%f,%f\n",numAngs,rint[i]);
	  else fprintf(fOut,"%d,%f\n",i,rint[i]);
	}/* Print it */
    }/* DO all */
  
  if(fOut)
    {/* Went OK */
      fclose(fOut);
      /* askThem("1D CircAvg written to sys$login:fft1dca.txt file.",
	      FALSE,XmDIALOG_INFORMATION); */
    }/* Went OK */
  else
    askThem("Cannot open sys$login:fft1dca.txt file.",
	    FALSE,XmDIALOG_INFORMATION);
  
  if(disp1Dca)
    {/* Display the 1D circavg on the fft window */
      int xSq,ySq;
      float dist;

      scaleCircFactor = (float)(tryColors-10.0) / ((float)(maxCA-minCA)); 

      /* Init */
      for(i=0;i<512;i++)
	for(j=0;j<512;j++)
	  {
	    if(hiColor)
	      {
		fft_buffer1[(i*512)+j]= 8;
	      }
	    else
	      fft_buffer1[(i*512)+j]= (BYTE)pixels_return[8];
	  }

      for(i=0;i<512;i++)
	{/* rows */
	  ySq = (i-256)*(i-256);
	  for(j=0;j<512;j++)
	    {/* Circ avg the buffer */
	      xSq = (j-256)*(j-256);
	      dist = sqrt((double)(xSq+ySq));
	      if(dist <= 256.0)
		{/* OK to do */
		  if(hiColor)
		    {
		  fft_buffer1[(i*512)+j]=    ((float)
					      (rint[(int)dist]-minCA) *
					      scaleCircFactor
					      );
		    }
		  else
		  fft_buffer1[(i*512)+j]= (BYTE)pixels_return[7+
						    (int)
						     ((float)
						      (rint[(int)dist]-minCA) *
						      scaleCircFactor
						      )];
		}/* OK to do */
	    }/* Circ avg the buffer */
	}/* rows */
      

      if(XtIsManaged(fftscreen))
	XPutImage(dsply, fft_window1, fft_gc, fft_image1, 0, 0,
		  0, 0,  512, 512);
      XFlush (dsply);

    }/* Display the 1D circavg on the fft window */
  else
    {/* Just put back the fft buffer */
      DrawFFTWindow();
    }/* Just put back the fft buffer */

  scaleCircFactor = 505.0 / ((float)(maxCA-minCA)); 
  
  for(i=256;i<512;i++)
    {/* DO all */
      scaledY = (rint[(int)((float)(i-256)/(float)fftZoomFactor)]-
		 minCA)*scaleCircFactor;
      scaledYp1 = (rint[(int)((float)((i-256)+1)/(float)fftZoomFactor)]-
		   minCA)*scaleCircFactor;
      scaledY = (510.0 - scaledY);
      scaledYp1 = (510.0 - scaledYp1);
      thisX= i;
      thisXp1= i+1;
      if(i==256)
	scaledY = 256;
      XDrawLine(dsply, XtWindow(fftDrawingWidget), 
		image_gcGREEN,
		thisX,(int)scaledY,(thisXp1),(int)scaledYp1);
    }/* DO all */
  
  free(rint);

  calcFFT1DcircAvg= TRUE;

}/* fft_1DCA_actIVATE */




/* ************************************************************
 * fft_screen_deactivate() - Get the fft screen
 * off the display.
 */
void fft_screen_deactivate(Widget widget, int *tag, 
			   XmAnyCallbackStruct *callback_data)
{/* fft_screen_deactIVATE */
  
  if(fftMode)
    {/* Free from last time */
      if(fftImage_gold)
	free(fftImage_gold);
      fftImage_gold = NULL;
      if(fftImage_fourAver)
	free(fftImage_fourAver);
      fftImage_fourAver = NULL;
      if(fftImage_buffer2)
	free(fftImage_buffer2);
      fftImage_buffer2 = NULL;
    }/* Free from last time */
  fftMode= FALSE;

  XtUnmanageChild(fftscreen);
  
}/* fft_screen_deactIVATE */



/*************************************************************
 * calcMeanTFpixel() - Calculate the mean transform pixel
 * length so we can draw this new defocus.
*/
void calcMeanTFpixel(float focusMajValue)
{/* CALCMeanTFpixel */
  int i,j;
  float a,b,c,alphaSq,alpha,spatialFreq,distance;

}/* CALCMeanTFpixel */




/*************************************************************
 * calcNodes() - Calculate the ctf nodes.
*/
void calcNodes()
{/* CALCNODES */
  int i,j;
  float invPixels,distA,distB,distAsq,distBsq;
  float spatialFreq,alpha,pixelSize,ampFactor;
  float numAngA,numAngB;
  char *ctfText;

  ctfMajorDefocusValue = ctffocusMajValue ;
  ctfMinorDefocusValue = ctfFocusMinvalue ;

  ctfText = XmTextGetString(ctfPixSizeW); 
  pixelSize   = (float)strtod(ctfText, 0);
  XtFree(ctfText);
  ctfText = XmTextGetString(ctfAmpFactorW); 
  ampFactor   = (float)strtod(ctfText, 0);
  XtFree(ctfText);

  /* Now get all distances for all nodes */
  for(i=0;i<MAXCTFS;i++)
    {/* DO 6 of them */
      float pi;
      int node;
      pi = PI;
      node = i + 1;
      ctf_nodes_misha_(&pi,&wavelength,&ctfMajorDefocusValue,&spAberration,
		     &ampFactor, &node, &distA);
      ctf_nodes_misha_(&pi,&wavelength,&ctfMinorDefocusValue,&spAberration,
		     &ampFactor, &node, &distB);

      /* DEBUG
            printf("Returned distA,distB[%f,%f]\n",distA,distB);
      */

      distA *= (512.0*pixelSize);
      distB *= (512.0*pixelSize);
      distAsq = (distA * distA);
      distBsq = (distB * distB);
      ctfPts[i].distAsq= distAsq;
      ctfPts[i].distA  = distA;
      ctfPts[i].distBsq= distBsq;
      ctfPts[i].distB  = distB;
      ctfPts[i].omega  = ctfAngMajorValue;
    }/* DO 6 of them */

}/* CALCNODES */






/*************************************************************
 * calcTFpixel() - Calculate the transform pixel
 * length so we can draw these new defocusi.
*/
void calcTFpixel(float focusMajValue, float focusMinValue)
{/* CALCTFpixel */
  int i,j;
  float a,b,c,alphaSq,alpha,spatialFreq,distance;

  a= (spAberration * 1000000.0) / 2.0;
  b= focusMajValue;
  c= (wavelength / PI);
  alphaSq = (
	       (b) + sqrt(
			  (b * b) -
			  (4.0 * a * c)
			  )
	       ) /
		 (2.0 * a);
  alpha = sqrt(alphaSq);
  spatialFreq = alpha / wavelength;
  distance = (512.0 * (scanStep * 1000.0) * spatialFreq) / 
                      magnification;
  ctfPts[ctfCurveNum].distA =   distance;


  a= (spAberration * 1000000.0) / 2.0;
  b= focusMinValue;
  c= (wavelength / PI);
  alphaSq = (
	       (b) + sqrt(
			  (b * b) -
			  (4.0 * a * c)
			  )
	       ) /
		 (2.0 * a);
  alpha = sqrt(alphaSq);
  spatialFreq = alpha / wavelength;
  distance = (512.0 * (scanStep * 1000.0) * spatialFreq) / 
                      magnification;
  ctfPts[ctfCurveNum].distB =   distance;

}/* CALCTFpixel */





/***************************************************************
 * calcDefocus() - Calculate the major and minor defocus values
*/
void calcDefocus(float *majorValue, float *minorValue)
{/* CALCDEFOCUS */
  int i,j,nodes;
  float invPixels,distA,distB,distAsq,distBsq;
  float spatialFreq,alpha,pixelSize,ampFactor,volts;
  float numAngA,numAngB,pi,radian,degree,focus;
  char *ctfText;

  piraddeg_(&pi, &radian, &degree);

  distA   = ctfPts[ctfCurveNum].distA;
  distB   = ctfPts[ctfCurveNum].distB;
  distAsq = ctfPts[ctfCurveNum].distAsq;
  distBsq = ctfPts[ctfCurveNum].distBsq;

  /* Do we want to use pixelSize here instead of scanStep/mag ??? */

  ctfText = XmTextGetString(ctfPixSizeW); 
  pixelSize   = (float)strtod(ctfText, 0);
  XtFree(ctfText);
  ctfText = XmTextGetString(ctfAmpFactorW); 
  ampFactor   = (float)strtod(ctfText, 0);
  XtFree(ctfText);
  ctfText = XmTextGetString(ctfKvW); 
  volts   = (float)strtod(ctfText, 0);
  volts  *= 1000.0;
  XtFree(ctfText);

  wavelength = wavel_(&volts);


  /* Use variable ctfCurveNum as the ctf node number */

  nodes = ctfCurveNum + 1;
  
  printf("\n\n*********************************\n");
  printf("distA,B[%f,%f]\n",distA,distB);

  if(pickLattice != -99)
    {/* Only do if points have been picked */
      spatialFreq =  ((float)distA / (float)512.0) / pixelSize;
      printf("spatialFreq = [%f] pixelSIze=[%f] distA[%f]\n",
	     spatialFreq,pixelSize,distA);
      printf ("wave[%f] ampFactor[%f] nodes[%d] cs[%f]\n",
	      wavelength, ampFactor, nodes, spAberration);
      ctf_focus_misha_(&pi, &wavelength, &focus, &spAberration, 
		     &ampFactor, &nodes, &spatialFreq);
      ctfMajorDefocusValue = focus;
      
      spatialFreq =  ((float)distB / (float)512.0) / pixelSize;
      printf("spatialFreq = [%f] pixelSIze=[%f] distB[%f]\n",
	     spatialFreq,pixelSize,distB);
      ctf_focus_misha_(&pi, &wavelength, &focus, &spAberration, 
		     &ampFactor, &nodes, &spatialFreq);
      ctfMinorDefocusValue = focus;
    }/* Only do if points have been picked */
  
  /* Now set up the text widgets */
  *majorValue = ctfMajorDefocusValue;
  *minorValue = ctfMinorDefocusValue;
  sprintf(outStr,"%f",ctfMajorDefocusValue);
  XmTextSetString(ctfMajorDefocusW,outStr);
  sprintf(outStr,"%f",ctfMinorDefocusValue);
  XmTextSetString(ctfMinorDefocusW,outStr);

  calcNodes();
  printf("calcDefocus [%f,%f]\n",
	 *majorValue, *minorValue);
  
}/* CALCDEFOCUS */




/***********************************************************
 * redo_ctfFocMajMin() -
 */
void redo_ctfFocMajMin()
{/* REDO_CTFFOCMAJMIN */

  calcDefocus(&ctffocusMajValue, &ctfFocusMinvalue);
  printf("redo [%f,%f]\n",
	 ctffocusMajValue, ctfFocusMinvalue);


  sprintf(outStr,"%f",ctfFocusMinvalue);  
  XmTextSetString(ctfFocusMinvalueW, outStr);
  XmTextSetString(ctfMinorDefocusW,outStr);
  sprintf(outStr,"%.2f",rad2deg(ctfAngMajorValue));  
  XmTextSetString(ctfAngMajorValueW, outStr);
  XmTextSetString(ctfAngMajorW, outStr); 
  sprintf(outStr,"%f",ctffocusMajValue);  
  XmTextSetString(ctffocusMajValueW, outStr);
  XmTextSetString(ctfMajorDefocusW,outStr);
  calcNodes();
  calcFFT(FALSE, 0, FALSE);
  DrawFFTWindow();
  XSync(dsply,FALSE);

}/* REDO_CTFFOCMAJMIN */




/***************************************************************
 * calcCTF() - Calc CTF curve.
*/
void calcCTF()
{/* CALCCTF */
  int i,j,k,thisJ,thisRowSq,thisColSq;
  float res_step,defocus,alpha,chi_alpha,thisCs,tmp,
        pfract, afract, dist, wave, thisDefocus;
  float distA,distB,invPixels;
  BYTE * thisBuf;

  res_step = 1.0 / 32.0;
  defocus  = ctffocusMajValue;
  pfract   = 1.0;
  afract   = 0.0;

  /* From ctfDisp 
  wavelength = 1230.0/sqrt(voltage + (1.0e-6)*(voltage*voltage));
  wave = wavelength/1000.0;		/+! convert to nm units +/
  */

  /* Convert ALL to nm units */
  wave = wavelength/1000.0;		 
  thisDefocus = defocus * 1000.0;
  thisCs   = spAberration * 1000000.0;

  for(i=0;i<512;i++)
    {/* Do rows */
      thisRowSq = (i-256)*(i-256);
      for(j=0;j<1024;j+=2)
	{/* Do all */
	  /* Calc CTF factor but remember to do it from distance
	   * from the middle */
	  thisJ = j;
	  thisColSq = (thisJ/2 - 256)*(thisJ/2 - 256);
	  dist = sqrt((float)(thisRowSq+thisColSq));
	  
	  alpha     = (dist)*wave*res_step;
	  chi_alpha = (THISTWOPI/wave) *
	    (((float)(thisDefocus)*(alpha*alpha)/2.0) - 
	     (thisCs*(alpha*alpha*alpha*alpha)/4.0));
	  
	  tmp       = pfract*sin(chi_alpha) + afract*cos(chi_alpha);
	  
	  /* Orig */
	  ctfValues[i][j] = tmp;
	  ctfValues[i][j+1] = dist;

	  /* Why this???
	  ctfValues[i][j] = 1.0 - fabs(tmp);
	  ctfValues[i][j+1] = 1.0 - fabs(tmp);
	  */

	}/* Do all */
    }/* Do rows */
  
  if(debugSwitch)
    {/* For debugging lets just display this on screen */
      for(i=0;i<512;i++)
	{/* Do rows */
	  for(j=0;j<1024;j+=2)
	    {/* Do all */
	      if(ctfValues[i][j] > 0.0)
		fft_gold_buffer1[((i*512)+(j/2))]= pixels_return[RED];
	      else if(ctfValues[i][j] < 0.0)
		fft_gold_buffer1[((i*512)+(j/2))]= pixels_return[GREEN];
	      else
		fft_gold_buffer1[((i*512)+(j/2))]= pixels_return[BLUE];
	    }/* Do all */
	}/* Do rows */
      
      DrawFFTWindow();
      XFlush(dsply);
      for(i=0;i<200;i++)
	{
	  em_sleep(5);
	}
    }/* For debugging lets just display this on screen */
  
printf("leaving calcCTF\n");

}/* CALCCTF */





/***************************************************************
 * drawSCURVEonFFT() - Draw CTF curve on the FFT screen.
*/
void drawSCURVEonFFT()
{/* DRAWSCURVEONFFT */
  int i,j,k,thisJ,thisRowSq,thisColSq,window,thisX,thisY;
  int windowLeft,thisPixel;
  float res_step,defocus,alpha,chi_alpha,thisCs,tmp,
        pfract, afract, dist, wave, thisDefocus;
  float distA,distB,invPixels,lastVal,totPixels;
  float slope,minPixel,maxPixel,scaleF;
  BYTE * thisBuf;
  char *thisText;
#define CNVPTX(x) ((fftZoomFactor==2) ? ((x-fftZoomX) * 2) : (x))

  /* Draw background */
  for(i=0;i<256;i++)
    {/* Do cols */
      XDrawLine(dsply, XtWindow(fftDrawingWidget), image_gcGREY[100],
		CNVPTX(256+i),256,
		CNVPTX(256+i),512);      
    }/* Do cols */

  if(scurveClick.x != 256.0)
    dist = sqrt((float)
		((float)(256.0 - (float)scurveClick.y) *
		(float)(256.0 - (float)scurveClick.y) 
		+
		(float)((float)scurveClick.x - 256.0) *
		(float)((float)scurveClick.x - 256.0)));
  else
    dist = abs((float)((float)scurveClick.x - 256.0));
  // Do we really want this ???  dist *=  (float)((float)fftSize / 512.0);


  if(scurveClick.x != 256.0)
    slope = ((float)(256.0 - (float)scurveClick.y) / 
	     (float)((float)scurveClick.x - 256.0));
  else
    slope = 0.0;
  if(slope==0.0)
    slope = 0.00000001;

  res_step = 1.0 / 32.0;
  defocus  = ctffocusMajValue;
  pfract   = 1.0;
  afract   = 0.0;

  /* From ctfDisp 
  wavelength = 1230.0/sqrt(voltage + (1.0e-6)*(voltage*voltage));
  wave = wavelength/1000.0;		/+! convert to nm units +/
  */

  /* Convert ALL to nm units */
  wave = wavelength/1000.0;		 
  thisDefocus = defocus * 1000.0;
  thisCs   = spAberration * 1000000.0;
  
  /* Get stats for data */
  minPixel = maxPixel = auto2DdataBuf[0];
  for(i=0;i<512;i++)
    for(j=0;j<256;j++)
      {/* Zoom it normal */
	thisPixel = ((i*256)+j);
	minPixel= MIN(auto2DdataBuf[thisPixel],minPixel);
	maxPixel= MAX(auto2DdataBuf[thisPixel],maxPixel);
      }/* Zoom it normal */
  scaleF = (float)((float)((float)255.0) / 
		   (float)(maxPixel-minPixel));

  /* First draw all data on screen */
  for(i=0;i<(int)dist;i++)
    {/* Do cols */
      int x,y;
      float xStep, spatialI;
      
      spatialI = i * ((float)fftSize / 512.0);
      xStep = sqrt((1.0/((slope*slope)+1.0)));
      x = (int)((float)spatialI*xStep);

      y = (int)(((float)spatialI*xStep)*slope);

      /* printf("i[%d] xStep[%f] x[%d] y[%d]\n",i,xStep,x,y); */


      /* Now draw scurve */
      if((y > 255) || (y<-254))
	tmp = 511;
      else
	tmp = 511 - (int)((auto2DdataBuf[(((256-y)*256)+x)] -minPixel)*scaleF);

      XDrawLine(dsply, XtWindow(fftDrawingWidget), image_gcBLUE,
		CNVPTX(256+i),(int)511,
		CNVPTX(256+i),(int)tmp);

    }/* Do rows */
  
  /* Get stats for curve */
  minPixel = maxPixel = auto2DdrawBuf[0];
  for(i=0;i<512;i++)
    for(j=0;j<256;j++)
      {/* Zoom it normal */
	thisPixel = ((i*256)+j);
	minPixel= MIN(auto2DdrawBuf[thisPixel],minPixel);
	maxPixel= MAX(auto2DdrawBuf[thisPixel],maxPixel);
      }/* Zoom it normal */
  scaleF = (float)((float)((float)255.0) / 
		   (float)(maxPixel-minPixel));

  lastVal = (128.0+256.0);
  for(i=0;i<(int)dist;i++)
    {/* Do cols */
      int x,y;
      float xStep, spatialI;

      spatialI = i * ((float)fftSize / 512.0);
      xStep = sqrt((1.0/((slope*slope)+1.0)));
      x = (int)((float)spatialI*xStep);

      y = (int)(((float)spatialI*xStep)*slope);

      /* Now draw scurve */
      if((y > 255) || (y<-254))
	tmp = 511;
      else
	tmp = 511 - (int)((auto2DdrawBuf[(((256-y)*256)+x)] -minPixel)*scaleF);

      XDrawLine(dsply, XtWindow(fftDrawingWidget), image_gcGREEN,
		CNVPTX(256+i),(int)lastVal,
		CNVPTX(256+i+1),(int)tmp);

      lastVal = tmp;
      
    }/* Do rows */
  

}/* DRAWSCURVEONFFT */





/* ************************************************************
 * ctf_ioparms_activate() - Used to write the CTF params
 * to a disk file for sharing with the ctfDisp prg.
 */
void ctf_ioparms_activate(Widget widget, int *tag, 
			  XmAnyCallbackStruct *callback_data)
{/* CTF_IOPARMS_ACTIVATE */
  FILE *fIO;
  char *ctfText;
  char inStr[256];
  char fileStr[512];

  if(*tag==0)
    {/* Output */

      sprintf(fileStr,"%s/ctf_ioparms.dat",getenv("HOME"));
      fIO = fopen(fileStr,"w");

      if(fIO)
	{/* Do it */

	  ctfText = XmTextGetString(ctfAmpFactorW); 
	  fprintf(fIO,"%s\n",ctfText);
	  XtFree(ctfText);
	  ctfText = XmTextGetString(ctfSpAberW); 
	  fprintf(fIO,"%s\n",ctfText);
	  XtFree(ctfText);
	  ctfText = XmTextGetString(ctfMajorDefocusW); 
	  fprintf(fIO,"%s\n",ctfText);
	  XtFree(ctfText);
	  ctfText = XmTextGetString(ctfMinorDefocusW); 
	  fprintf(fIO,"%s\n",ctfText);
	  XtFree(ctfText);
	  ctfText = XmTextGetString(ctfKvW); 
	  fprintf(fIO,"%s\n",ctfText);
	  XtFree(ctfText);
	  ctfText = XmTextGetString(ctfWienerW); 
	  fprintf(fIO,"%s\n",ctfText);
	  XtFree(ctfText);
	  ctfText = XmTextGetString(ctfAngMajorW); 
	  fprintf(fIO,"%s\n",ctfText);
	  XtFree(ctfText);

	  fclose(fIO);

	}/* Do it */
      else askThem("Can't open ctf_ioparms.dat file.",
		   FALSE,XmDIALOG_INFORMATION);
    }/* Output */
  else
    {/* Input */

      sprintf(fileStr,"%s/ctf_ioparms.dat",getenv("HOME"));
      fIO = fopen(fileStr,"r");

      if(fIO)
	{/* Do it */

	  get_str(fIO, '\n', inStr);
	  XmTextSetString(ctfAmpFactorW,inStr); 
	  get_str(fIO, '\n', inStr);
	  XmTextSetString(ctfSpAberW,inStr); 
	  get_str(fIO, '\n', inStr);
	  XmTextSetString(ctfMajorDefocusW,inStr); 
	  get_str(fIO, '\n', inStr);
	  XmTextSetString(ctfMinorDefocusW,inStr); 
	  get_str(fIO, '\n', inStr);
	  XmTextSetString(ctfKvW,inStr); 
	  get_str(fIO, '\n', inStr);
	  XmTextSetString(ctfWienerW,inStr); 
	  get_str(fIO, '\n', inStr);
	  XmTextSetString(ctfAngMajorW,inStr); 

	  fclose(fIO);
	}/* Do it */
      else askThem("Can't open ctf_ioparms.dat file.",
		   FALSE,XmDIALOG_INFORMATION);
    }/* Input */

}/* CTF_IOPARMS_ACTIVATE */





/* ************************************************************
 * ctfCurve_cr_activate() - They have hit the CReturn key while
 * in the ctfCurve Text widget.  
 */
void ctfCurve_cr_activate(Widget widget, int *tag, 
			  XmAnyCallbackStruct *callback_data)
{/* ctfCurve_CR_ACTIVATE */
  char *ctfText;

  ctfText = XmTextGetString(ctfCurveTextW); 
  ctfCurveNum = (int)strtod(ctfText, 0);
  ctfCurveNum--;     /* Human numbers from 1 */
  if(ctfCurveNum >= MAXCTFS)
    ctfCurveNum = MAXCTFS-1;
  if(ctfCurveNum < 0)
    ctfCurveNum = 0;
  XtFree(ctfText);

  maxCtfCurve = MAX((ctfCurveNum+1),maxCtfCurve);

  /* Restuff the old points */
  if(!ctfPts[ctfCurveNum].pickedPts)
    {/* Guess some new ones */
      ctfPts[ctfCurveNum].latX0= latPt[0].measuredX;
      ctfPts[ctfCurveNum].latY0= latPt[0].measuredY;
      ctfPts[ctfCurveNum].latX1= latPt[1].measuredX;
      ctfPts[ctfCurveNum].latY1= latPt[1].measuredY;
/*      ctfPts[ctfCurveNum].distA = ctfPts[0].distA;
      ctfPts[ctfCurveNum].distB = ctfPts[0].distB;*/
    }/* Guess some new ones */
  else
    {/* restuff */
      latPt[0].measuredX = ctfPts[ctfCurveNum].latX0;
      latPt[0].measuredY = ctfPts[ctfCurveNum].latY0;
      latPt[1].measuredX = ctfPts[ctfCurveNum].latX1;
      latPt[1].measuredY = ctfPts[ctfCurveNum].latY1;
    }/* restuff */
  ctfPts[ctfCurveNum].pickedPts= TRUE;

  sprintf(outStr,"%d",ctfCurveNum+1);
  XmTextSetString(ctfCurveTextW,outStr);

  DrawFFTWindow();

}/* ctfCurve_CR_ACTIVATE */





/* ************************************************************
 * fftImgNum_cr_activate() - They have hit the CReturn key while
 * in the fftImgNum Text widget.  Work on the image they've entered.
 */
void fftImgNum_cr_activate(Widget widget, int *tag, 
			   XmAnyCallbackStruct *callback_data)
{/* fftImgNum_CR_ACTIVATE */
  char *imgNumText;

  averageFFT= FALSE;

  imgNumText = XmTextGetString(widget); /* Could be fftImgNumW or 
					 * rotfltImgNumW */ 
  imgNum = (int)strtod(imgNumText, 0);
  if(imgNum > purdueHeader.nimg)
    imgNum = purdueHeader.nimg;
  if(imgNum < 1)
    imgNum = 1;
  XtFree(imgNumText);

  sprintf(outStr,"%d",imgNum);
  XmTextSetString(imgNumTextWidget,outStr);
  XmTextSetString(fftImgNumW,outStr);
  XmTextSetString(rotfltImgNumW,outStr);
  if(fileType==PACKEDPORTABLE)
    stuffPackedImage(purdueHeader.numCols,
		     purdueHeader.numRows,
		     packImg,purdueHeader.nimg,
		     imgNum);
  DrawWindow(); 


  calcFFT(TRUE, 0, 0);
  DrawFFTWindow();

  /* Make them do a new circavg */
  calcFFT1DcircAvg= FALSE;

  /* DO if rotflt'ing */
  newOrigin = TRUE;
  if(XtIsManaged(rotfltscreen))
    {
      iopt[0]= 1;      /* 1 == Refine symmetry axis */
      iopt[1]= 0;
      iopt[2]= 0;      /* ==1  means plot power spectrum */
      iopt[3]= 0;
      
      rotflt_tryit_activate(0,0,0);
    }


}/* fftImgNum_CR_ACTIVATE */





/*************************************************************
 * fft_change() - They want to limit the display of the FFT
 * from min to max.  Also other FFT parameters are changed here.
*/
void fft_change(Widget widget, int *tag, 
		XmScrollBarCallbackStruct *callback_data)
{/* FFT_CHANGE */
  int i,j;
  XmString simpleString;

  if(!inited)
    goto notInited;

  if(*tag==FFTMIN)
    {/* surf changed */
      if(fftType==LINEARFFT)
	minFFTvalue = (float)(callback_data->value*1000.0);/* 1/1000th */
      else minFFTvalue = (float)(callback_data->value/100.0);/* 1/100th */

      if(fftType==LINEARFFT)
	minFFTvalue = MIN(minFFTvalue,(maxFFTvalue-1000.0));
      else minFFTvalue = MIN(minFFTvalue,(maxFFTvalue-0.01));

      /* Give feedmack */
      sprintf(outStr,"%.2f",minFFTvalue);  
      createSimpleString(&simpleString,outStr);
      XtSetArg(arglst[0], XmNlabelString, simpleString);
      XtSetValues(minFFTwidget, arglst, 1); 
      XFlush(dsply);
      XmStringFree(simpleString); 

    }/* surf changed */
  else if(*tag==FFTMAX)
    {/* surf changed */
      if(fftType==LINEARFFT)
	maxFFTvalue = (float)(callback_data->value*1000.0);/* 1/1000th */
      else maxFFTvalue = (float)(callback_data->value/100.0);/* 1/100th */

      if(fftType==LINEARFFT)
	maxFFTvalue = MAX(maxFFTvalue,(minFFTvalue+1000.0));
      else maxFFTvalue = MAX(maxFFTvalue,(minFFTvalue+0.01));

    }/* surf changed */
  else if(*tag==FFTTFAC)
    {/* New Temp factor */
      tempFactor = (int)callback_data->value;

      sprintf(outStr,"%d",tempFactor);  
      createSimpleString(&simpleString,outStr);
      XtSetArg(arglst[0], XmNlabelString, simpleString);
      XtSetValues(tfacFFTwidget, arglst, 1); 
      XFlush(dsply);
      XmStringFree(simpleString); 

      /* Calc all temp factors */
      if(tempFactor == 0)
	{/* No temp factor */
	  for(i=0;i<1024;i++)
	    for(j=0;j<2048;j++)
	      tempFactors[i][j] = 1.0;
	}/* No temp factor */
      else
	{/* Calc it */
	  float B,dist,numTemp,floatVal;
	  int thisRowSq,thisColSq,thisJ;
	  
	  floatVal = (float)callback_data->value;
	  
	  numTemp = (float)(floatVal*0.1);
	  if(numTemp > 10.0)
	    numTemp= 10.0;
	  
	  numTemp = 4.0/(pow(2.0,numTemp));
	  B = 4.0 / (float)(numTemp * numTemp);
	  for(i=0;i<512;i++)
	    {/* Do rows */
	      if(i < 256)
		thisRowSq = (i)*(i);
	      else thisRowSq = (512-i)*(512-i);
	      for(j=0;j<1024;j+=2)
		{/* Do all */
		  /* Calc TEMP factor but remmeber to do it from distance
		   * from the middle */
		  thisJ = j;
		  thisJ = (j - 512); /* Try this */
		  if(j > 512)
		    thisJ -= 512;
		  else thisJ += 512;
		  thisColSq = (thisJ/2)*(thisJ/2);
		  dist = (float)(thisRowSq+thisColSq);
		  if(dist > 0.0)
		    tempFactors[i][j] = exp(-B / (4.0 * (65536.0 / dist)*
					       (65536.0 / dist)));
		  else tempFactors[i][j] = 1.0;
		  if(tempFactors[i][j] == 0.0)
		    tempFactors[i][j]= 0.00000001;
		  tempFactors[i][j] = 1.0 / tempFactors[i][j];
		  tempFactors[i][j+1] = tempFactors[i][j];
		}/* Do all */
	    }/* Do rows */
	}/* Calc it */
      
    }/* New Temp factor */
  else if(*tag==FFTBKG)
    {/* FFTBKG scroll */
      fftBkgValue = (int)callback_data->value;
      sprintf(outStr,"%d",fftBkgValue);  /* +1 for humans */
      createSimpleString(&simpleString,outStr);
      XtSetArg(arglst[0], XmNlabelString, simpleString);
      XtSetValues(fftBkgValueW, arglst, 1); 
      XFlush(dsply);
      XmStringFree(simpleString); 
    }/* FFTBKG scroll */
  else if(*tag==CTF_ANGMAJOR)
    {/* CTF omega scroll */
      ctfAngMajorValue = (float)callback_data->value;
      ctfAngMajorValue *= 0.01;
      ctfAngMajorValue = deg2rad(ctfAngMajorValue);  
      changeCTFtext = TRUE;
      sprintf(outStr,"%.2f",rad2deg(ctfAngMajorValue));  /* +1 for humans */
      XmTextSetString(ctfAngMajorValueW, outStr);
      XmTextSetString(ctfAngMajorW, outStr); 
    }/* CTF omega scroll */
  else if(*tag==CTF_FOCUSMAJ)
    {/* CTF mean Defocus scroll */
      ctffocusMajValue = (float)callback_data->value;
      ctffocusMajValue *= 0.01;
      /* OBSOLETE???      calcTFpixel(ctffocusMajValue, ctfFocusMinvalue); */
      calcNodes();
      /* calcMeanTFpixel(ctffocusMajValue); */
      changeCTFtext = TRUE;
      sprintf(outStr,"%f",ctffocusMajValue);
      XmTextSetString(ctffocusMajValueW, outStr);
      XmTextSetString(ctfMajorDefocusW,outStr);
    }/* CTF mean Defocus scroll */
  else if(*tag==CTF_FOCUSMIN)
    {/* CTF a/b scroll */
      ctfFocusMinvalue = (float)callback_data->value;
      ctfFocusMinvalue *= 0.01;
      /* OBSOLETE???      calcTFpixel(ctffocusMajValue, ctfFocusMinvalue); */
      calcNodes();
      changeCTFtext = TRUE;
      sprintf(outStr,"%f",ctfFocusMinvalue);  /* +1 for humans */
      XmTextSetString(ctfFocusMinvalueW, outStr);
      XmTextSetString(ctfMinorDefocusW,outStr);
    }/* CTF a/b scroll */
  else if(*tag==FILTSIZE)
    {/* FILTER Size scroll */
      filtSizeValue = (float)callback_data->value;
      filtSizeValue *= 0.5;
      sprintf(outStr,"%.0f",filtSizeValue*2.0);  /* +1 for humans */
      createSimpleString(&simpleString,outStr);
      XtSetArg(arglst[0], XmNlabelString, simpleString);
      XtSetValues(filtLabelWidget, arglst, 1); 
      XFlush(dsply);
      XmStringFree(simpleString); 
    }/* FILTER Size scroll */
  else if (*tag==LATCOLOR)
    {/* lat color */
      int thisVal;
      thisVal = (int)callback_data->value;
      image_gcLAT = image_gcGREY[thisVal];
    }/* lat color */
  else if (*tag==HIP)
    {/* hip */
      int thisVal;
      thisVal = (int)callback_data->value;
      sprintf(outStr,"%f",(float)thisVal);
      XmTextSetString(hipTextW,outStr);
    }/* hip */
  else if (*tag==LOP)
    {/* lop */
      int thisVal;
      thisVal = (int)callback_data->value;
      sprintf(outStr,"%f",(float)thisVal);
      XmTextSetString(lopTextW,outStr);
    }/* lop */

  if(*tag==FFTMIN || *tag==FFTMAX)
    {/* MIN/MAX scrolls */
      if(minFFTvalue >= maxFFTvalue)
	maxFFTvalue = minFFTvalue + 1.0;
      
      
      sprintf(outStr,"%.2f",minFFTvalue);  
      createSimpleString(&simpleString,outStr);
      XtSetArg(arglst[0], XmNlabelString, simpleString);
      XtSetValues(minFFTwidget, arglst, 1); 
      XFlush(dsply);
      XmStringFree(simpleString); 
      
      sprintf(outStr,"%.2f",maxFFTvalue);    
      createSimpleString(&simpleString,outStr);
      XtSetArg(arglst[0], XmNlabelString, simpleString);
      XtSetValues(maxFFTwidget, arglst, 1); 
      XFlush(dsply);
      XmStringFree(simpleString); 
    }/* MIN/MAX scrolls */
  
  if(callback_data->event != 0)
    {/* OK to go */

      if(*tag==FFTBKG)
	{/* Here */
	  calcFFT(TRUE, 0, TRUE);
	}/* Here */
      else calcFFT(FALSE, 0, FALSE);
      
      XSync(dsply,FALSE);
      XFlush(dsply);
      DrawFFTWindow();
      XSync(dsply,FALSE);
      XFlush(dsply);
    }/* OK to go */

 notInited:;

}/* FFT_CHANGE */



/*************************************************************
 * fft_change_toggle() - They want to change fft Size.
 * 
*/
void fft_change_toggle(Widget widget, int *tag, 
		       XmAnyCallbackStruct *callback_data)
{/* FFT_CHANGE_TOGGLE */
  int i,j;
  XmString simpleString;

  if(!inited || callback_data->event == 0)
    goto notInited;

  if(*tag==A1024FFT || *tag==A512FFT || *tag==A256FFT || *tag==A128FFT)
    rotflt_screen_deactivate(0,0,0);

  if(*tag==A1024FFT)
    { fftSize = 1024; }
  else if(*tag==A512FFT)
    { fftSize = 512;       }
  else if(*tag==A256FFT)
    { fftSize = 256;       }
  else if(*tag==A128FFT)
    { fftSize = 128;       }
  else if(*tag==AMPLITUDE)
    { 
      if(fftDispType==AMPLITUDE)
	{fftDispType = INTENSITY;
	 createSimpleString(&simpleString, "Intensity");}
      else 
	{fftDispType = AMPLITUDE;
	 createSimpleString(&simpleString, "Amplitude");}

      /* XtSetArg(arglst[0], XmNset, FALSE); */
      XtSetArg(arglst[0], XmNlabelString, simpleString);
      XtSetValues(amplitudeToggleW, arglst, 1); 
      XFlush(dsply);
      XmStringFree(simpleString);
    }
  else if (*tag==FFTZOOM)
    {/* Zoom the FFT */
      fftZoom = !fftZoom;
      if(fftZoom)
	{/* Turn the others off */

	  fftZoomFactor = 2;

	  fftZoomX = 128;  /* Upper left hand corner */
	  fftZoomY = 128;
	  fftZoomX = MIN(MAX(fftZoomX,0),(512-128));
	  fftZoomY = MIN(MAX(fftZoomY,0),(512-128));

	}/* Turn the others off */
      else
	{/* Go back to normal */
	  fftZoomFactor = 1;
	  fftZoomX= fftZoomY= 0;
	}/* Go back to normal */
    }/* Zoom the FFT */
  else if(*tag==FFTCONTOUR)
    { contourToggle = !contourToggle; }
  else if (*tag==DISP1DCA)
    {/* Display 1D circ avg */
      disp1Dca = !disp1Dca;
    }/* Display 1D circ avg */
  else if (*tag==CTFENHANCE)
    {/* Display CTF enhancing */
      ctfEnhance = !ctfEnhance;
    }/* Display ctf enhancing */
  else if(*tag==FFTLOCK)
    { fftMinMaxLock = !fftMinMaxLock; }
  else if(*tag==DISP1024FFT)
    { 
      disp1024fft = !disp1024fft; 
      printf("disp1024[%d]\n",disp1024fft);
    }
  else if(*tag==LINEARFFT)
    {/* LINEAR */
      fftType = LINEARFFT; 
    }/* LINEAR */
  else if(*tag==LOGFFT)
    {/* LOG */ 
      fftType = LOGFFT; 
    }/* LOG */ 
  else if(*tag==NONLINFFT)
    { fftType = NONLINFFT; }
  else if(*tag==BACKFFT)
    { 
      backTransform = !backTransform; 
    }
  else if(*tag==FILTER)
    {/* FILTER fft */
      filterToggle = !filterToggle;
      if(filterToggle)
	{/* Put up filtering tools */
	  XtSetArg(arglst[0], XmNsensitive, TRUE);
	  /*	  XtSetValues(filtSizeWidget, arglst, 1);  */
	  XtSetValues(drawHiLoToggleW,arglst,1);
	  XtSetArg(arglst[0], XmNsensitive, FALSE);
	  XtSetValues(pick2pointsLabelW,arglst,1);
	  /* Turn off pick lattice mode */
	  pickLattice = -99;
	}/* Put up filtering tools */
      else
	{/* Put up filtering tools */
	  XtSetArg(arglst[0], XmNsensitive, FALSE);
	  /* XtSetValues(filtSizeWidget, arglst, 1);  */
	  XtSetValues(drawHiLoToggleW,arglst,1);
	  XtSetArg(arglst[0], XmNsensitive, TRUE);
	  XtSetValues(pick2pointsLabelW,arglst,1);
	  /* Turn on pick lattice mode */
	  pickLattice = 1;
	  currentLatPt=0; 
	  contour= TRUE;
	}/* Put up filtering tools */

    }/* FILTER fft */
  else if(*tag==DRAWLAT)
    {/* DRAWLAT fft */
      drawLatToggle = !drawLatToggle;
    }/* DRAWLAT fft */
  else if(*tag==DRAWHILO)
    {/* DRAW hi and low pas filters */
      drawHiLoToggle = !drawHiLoToggle;
    }/* DRAW hi and low pas filters */
  else if(*tag==PICKLATTICE || *tag==PICKCTF)
    {/* Pick Lattice/CTF points */

      if(*tag==PICKCTF)
	{/* Turn it WAY off for pick ctf points */
	  if(pickLattice != -99)
	    {pickLattice = -99;  
	     drawCTF = FALSE; }
	  else 
	    {pickLattice = 1;
	     drawCTF = TRUE; }
	}/* Turn it WAY off for pick ctf points */
      else
	{/* All else */
	  /* Turn it on */
	  pickLattice = 1;
	}/* All else */

      contour= TRUE;
      /*	  pickedLatPts= FALSE; They may have some from before */
      currentLatPt=0; 
      stuffContourBuffer();
      /* Turn off Filtering */
      filterToggle = FALSE;
      XtSetArg(arglst[0], XmNsensitive, FALSE);
      /* XtSetValues(filtSizeWidget, arglst, 1);  */
      XtSetValues(drawHiLoToggleW,arglst,1);
    }/* Pick Lattice/CTF points */
  else if(*tag==LAYERLINES || *tag==TWODLINES)
    {/* line Lattice points */
      if(*tag==LAYERLINES)
	layerLines= TRUE;
      else layerLines = FALSE;

      filterToggle = FALSE;
      XtSetArg(arglst[0], XmNsensitive, FALSE);
      /* XtSetValues(filtSizeWidget, arglst, 1);  */
      XtSetValues(drawHiLoToggleW,arglst,1);

      /* Turn it on */
      pickLattice = 1;
      contour= TRUE;
      /*	  pickedLatPts= FALSE; They may have some from before */
      currentLatPt=0; 
      stuffContourBuffer();

      XtSetArg(arglst[0], XmNset, TRUE);
      /* XtSetValues(pickLatticeToggleW, arglst, 1);  */
      XtSetArg(arglst[0], XmNset, FALSE);
      XtSetValues(filterFFTtoggleW, arglst, 1); 

      if(layerLines)
	{/* Turn it on */

	  /* Reinit these */
	  XmTextSetString(lat1XtextW,"");
	  XmTextSetString(lat1YtextW,"");
	  XmTextSetString(lat2XtextW,"");
	  XmTextSetString(lat2YtextW,"");

	  /* No resolution filter with layer lines yet */
	  resoX= 0;
	  resoY= 0;
	  lowResoFilter = 320000;  /* Ridiculously high */
	  hiResoFilter  = 0;       /* Ridiculously low */

	  sprintf(outStr,"Draw Layer Lines"); 
	  createSimpleString(&simpleString,outStr);
	  XtSetArg(arglst[0], XmNlabelString, simpleString);
	  XtSetArg(arglst[1], XmNset, TRUE);
	  XtSetValues(drawLatToggleW, arglst, 2); 
	  /* We have to force the scrolls to be inited */
	  drawLatToggle = TRUE;
	  XFlush(dsply);
	  XmStringFree(simpleString); 
	  sprintf(outStr,"Filter Mask Slit Diameter"); 
	  createSimpleString(&simpleString,outStr);
	  XtSetArg(arglst[0], XmNlabelString, simpleString);
	  XtSetValues(filtDescWidget, arglst, 1); 
	  XFlush(dsply);
	  XmStringFree(simpleString); 
	  XtUnmanageChild(lat1YtextW);
	  XtUnmanageChild(lat2XtextW);
	  XtUnmanageChild(lat2YtextW);
	  XtUnmanageChild(lat2LabelW);
	  XtUnmanageChild(millerLabelW);
	  sprintf(outStr,"LayrLine#"); 
	  createSimpleString(&simpleString,outStr);
	  XtSetArg(arglst[0], XmNlabelString, simpleString);
	  XtSetValues(lat1LabelW, arglst, 1); 
	  XFlush(dsply);
	  XmStringFree(simpleString); 
	  XFlush(dsply);
	  if(latLineOrder <= 1)
	    latLineOrder = 1;
	  sprintf(outStr,"%d",latLineOrder);
	  XmTextSetString(lat1XtextW, outStr);
	  XFlush(dsply);
	  XtSetArg(arglst[0], XmNsensitive, FALSE);
	  XtSetValues(latticeScrollW[0], arglst, 1); 
	  XtSetValues(latticeScrollW[2], arglst, 1); 
	  sprintf(outStr,"n/a");
	  XmTextSetString(latticeLabel2W[0], outStr); 
	  XmTextSetString(latticeLabel2W[2], outStr); 

/* Why do we need to do this ???
	  latPt[2].x = latPt[0].x;
	  latPt[2].y = latPt[0].y;
	  latPt[2].measuredX = latPt[0].measuredX;
	  latPt[2].measuredY = latPt[0].measuredY;
	  latPt[3].x = latPt[1].x;
	  latPt[3].y = latPt[1].y;
	  latPt[3].measuredX = latPt[1].measuredX;
	  latPt[3].measuredY = latPt[1].measuredY;
*/
	  latReset_activate(0,0,0);
	}/* Turn it on */
      else
	{/* Tuirn it off */

	  /* Reinit these */
	  XmTextSetString(lat1XtextW,"");
	  XmTextSetString(lat1YtextW,"");
	  XmTextSetString(lat2XtextW,"");
	  XmTextSetString(lat2YtextW,"");

	  sprintf(outStr,"Draw 2D Lattice"); 
	  createSimpleString(&simpleString,outStr);
	  XtSetArg(arglst[0], XmNlabelString, simpleString);
	  XtSetValues(drawLatToggleW, arglst, 1); 
	  XFlush(dsply);
	  XmStringFree(simpleString); 
	  sprintf(outStr,"Filter Mask Hole Diameter:"); 
	  createSimpleString(&simpleString,outStr);
	  XtSetArg(arglst[0], XmNlabelString, simpleString);
	  XtSetValues(filtDescWidget, arglst, 1); 
	  XFlush(dsply);
	  XtManageChild(lat1YtextW);
	  XtManageChild(lat2YtextW);
	  XtManageChild(lat2XtextW);
	  XtManageChild(lat2LabelW);
	  XtManageChild(millerLabelW);
	  sprintf(outStr,"h1, k1"); 
	  createSimpleString(&simpleString,outStr);
	  XtSetArg(arglst[0], XmNlabelString, simpleString);
	  XtSetValues(lat1LabelW, arglst, 1); 
	  XFlush(dsply);
	  XmStringFree(simpleString); 
	  XFlush(dsply);
	  XtSetArg(arglst[0], XmNsensitive, TRUE);
	  XtSetValues(latticeScrollW[0], arglst, 1); 
	  XtSetValues(latticeScrollW[2], arglst, 1); 
	  XFlush(dsply);
/* Why do we ned to do this ???
	  latPt[0].x = latPt[2].x;
	  latPt[0].y = latPt[2].y;
	  latPt[0].measuredX = latPt[2].measuredX;
	  latPt[0].measuredY = latPt[2].measuredY;
	  latPt[1].x = latPt[3].x;
	  latPt[1].y = latPt[3].y;
	  latPt[1].measuredX = latPt[3].measuredX;
	  latPt[1].measuredY = latPt[3].measuredY;
*/
	  latReset_activate(0,0,0);
	}/* Tuirn it off */
    }/* Line Lattice points */
  else if (*tag==SOFT)
    {/* Soft cutoff */
      icut= 1;
    }/* Soft cutoff */
  else if (*tag==HARD)
    {/* Hard cutoff */
      icut= 0;
    }/* Hard cutoff */


  if(*tag==LINEARFFT || *tag==LOGFFT)
    {/* LOG or LINEAR fft */
      float thisMin,thisMax;
      int thisMaxInt, thisMinInt, fftHeight,fftWidth, thisPixel,
          sliderSize;

      fftHeight = fftWidth = 512;
      thisMin = r4_max;
      thisMax = r4_min;
      for(i=0;i<(int)(fftHeight);i++)
	for(j=0;j<(int)(fftWidth*2);j+=2)
	  {/* Do all image */
	    float thisValue;
	    thisPixel = (int)(i*(fftWidth*2))+j;
	    thisValue = (
			 (fftImage_gold[(thisPixel)]*
			  fftImage_gold[(thisPixel)]) +
			 (fftImage_gold[(thisPixel+1)]*
			  fftImage_gold[(thisPixel+1)]) );
	    thisValue *= tempFactors[(int)(i*0.5)][(int)(j*0.5)];
	    if(fftType == LOGFFT && thisValue > 0.0)
	      thisValue = log10(thisValue);
	    thisMax = MAX(thisValue,thisMax);
	    thisMin = MIN(thisValue,thisMin);
	  }/* Do all image */

      if(averageFFT)
	{/* Do special */
	  sprintf(outStr,"Aver");
	  XmTextSetString(fftImgNumW,outStr);
	}/* Do special */

      /* Now setup the scroll bars to hold the new max/min */
      /* We can't exceed what a 32 bit signed integer will hold!! 
       * - Really 2.17 Gig but let's be safe... 
       */
      /* Suppose thisMin was 65 for linear but 1.8 for log ??? */
      minFFTvalue = thisMin;
      maxFFTvalue = thisMax;

      if(fftType==LINEARFFT)
	thisMax *= 0.001;  /* Let's get rid of a lot of the top stuff */

      if(fftType==LINEARFFT)
	thisMax *= 0.001;  /* Do this in 1/1000 ths */
      else thisMax *= 100.0;  /* Do this in 100 ths */
      if(fftType==LINEARFFT)
	thisMin *= 0.001;  /* Do this in 1/1000 ths */
      else thisMin *= 100.0;  /* Do this in 100 ths */

      thisMaxInt = (int)MIN(thisMax,1800000000.0);
      imageMaxFFTvalue = (float)thisMax;
      thisMinInt = (int)thisMin;

      if(thisMaxInt <= thisMinInt)
	thisMaxInt = thisMinInt + 1;

      sliderSize = (thisMaxInt - thisMinInt) / 100;
      sliderSize = MAX(sliderSize,1);

      XtSetArg(arglst[0], XmNmaximum, (int)(thisMaxInt+(float)sliderSize));
      XtSetArg(arglst[1], XmNminimum, (int)(thisMinInt));
      XtSetArg(arglst[2], XmNsliderSize, (int)(sliderSize));
      if(!fftMinMaxLock)
	{/* OK to change fft min max value */
	  if(fftType==LINEARFFT)/* 1/1000th */
	    XtSetArg(arglst[3], XmNvalue, (int)(minFFTvalue/1000.0));
	  else XtSetArg(arglst[3], XmNvalue, (int)(minFFTvalue*100.0));
	  XtSetArg(arglst[4], XmNpageIncrement, (int)(sliderSize));
	  if(sliderSize > 11.0)
	    XtSetArg(arglst[5], XmNincrement, (int)(sliderSize/10.0));
	  else XtSetArg(arglst[5], XmNincrement, (int)(1.0));
	  XtSetValues(minFFTscroll, arglst, 6);
	}/* OK to change fft min max value */
      else
	{/* Not OK */
	  XtSetArg(arglst[3], XmNpageIncrement, (int)(sliderSize));
	  if(sliderSize > 11.0)
	    XtSetArg(arglst[4], XmNincrement, (int)(sliderSize/10.0));
	  else XtSetArg(arglst[4], XmNincrement, (int)(1.0));
	  XtSetValues(minFFTscroll, arglst, 5);
	}/* Not OK */

      XtSetArg(arglst[0], XmNmaximum, (int)(thisMaxInt+(float)sliderSize));
      XtSetArg(arglst[1], XmNminimum, (int)(thisMinInt));
      XtSetArg(arglst[2], XmNsliderSize, (int)(sliderSize));
      if(!fftMinMaxLock)
	{/* OK to change fft min max value */
	  if(fftType==LINEARFFT)/* 1/1000th */
	    XtSetArg(arglst[3], XmNvalue, (int)(maxFFTvalue/1000.0));
	  else XtSetArg(arglst[3], XmNvalue, (int)(maxFFTvalue*100.0));
	  XtSetArg(arglst[4], XmNpageIncrement, (int)(sliderSize));
	  if(sliderSize > 11.0)
	    XtSetArg(arglst[5], XmNincrement, (int)(sliderSize/10.0));
	  else XtSetArg(arglst[5], XmNincrement, (int)(1.0));
	  XtSetValues(maxFFTscroll, arglst, 6);
	}/* OK to change fft min max value */
      else
	{/* Not OK */
	  XtSetArg(arglst[3], XmNpageIncrement, (int)(sliderSize));
	  if(sliderSize > 11.0)
	    XtSetArg(arglst[4], XmNincrement, (int)(sliderSize/10.0));
	  else XtSetArg(arglst[4], XmNincrement, (int)(1.0));
	  XtSetValues(maxFFTscroll, arglst, 5);
	}/* Not OK */
      XSync(dsply,FALSE);
      XFlush(dsply);

      sprintf(outStr,"%.2f",minFFTvalue);  
      createSimpleString(&simpleString,outStr);
      XtSetArg(arglst[0], XmNlabelString, simpleString);
      XtSetValues(minFFTwidget, arglst, 1); 
      XFlush(dsply);
      XmStringFree(simpleString); 
      
      sprintf(outStr,"%.2f",maxFFTvalue);    
      createSimpleString(&simpleString,outStr);
      XtSetArg(arglst[0], XmNlabelString, simpleString);
      XtSetValues(maxFFTwidget, arglst, 1); 
      XSync(dsply,FALSE);
      XFlush(dsply);
      XmStringFree(simpleString); 
    }/* LOG or LINEAR fft */


  if((*tag==A1024FFT) ||
     (*tag==A512FFT) ||
     (*tag==A256FFT) ||
     (*tag==A128FFT) 
     )
    { /* Let's change the lopHip scrolls accordingly */
      XtSetArg(arglst[0], XmNmaximum, (int)(fftSize*1.222));
      XtSetArg(arglst[1], XmNminimum, (int)0);
      XtSetArg(arglst[2], XmNpageIncrement, (int)(fftSize/32));
      XtSetValues(lopScrollW, arglst, 3);
      XtSetValues(hipScrollW, arglst, 3);
    } /* Let's change the lopHip scrolls accordingly */

  /* Recalc FFT */
  
  if(*tag != FFTLOCK)
    {/* Don't do anything for FFTLOCK */
      if(*tag==A1024FFT ||
	 *tag==A512FFT  ||
	 *tag==A256FFT  ||
	 *tag==A128FFT  ||
	 *tag==RECALCFFT)
	{/* recalc it */
	  averageFFT = FALSE;
	  calcFFT(TRUE, 0, 0); 
	}/* recalc it */
      else
	{/* DOn't need to recalc it */
	  calcFFT(FALSE, 0, 0); 
	}/* DOn't need to recalc it */
      
      XFlush(dsply);
      DrawFFTWindow();
      XFlush(dsply);
    }/* Don't do anything for FFTLOCK */
  

  /* Make them do a new circavg */
  calcFFT1DcircAvg= FALSE;

 notInited:;

}/* FFT_CHANGE_TOGGLE */



/* ************************************************************
 * fft_create() - Remember the widgets on the
 * fft screen.
 */
void fft_create(Widget widget, int *tag, 
		XmAnyCallbackStruct *callback_data)
{/* fft_CREATE */
  int i,j;
  XmString simpleString;
  XtCallbackRec callBck[5];    /* Our callbacks */
  XFontStruct *font;           /* Holds font information */
  XmFontList font_list;
  XmStringCharSet cs= "ISOLatin1";

  if(*tag==FFTMIN)
    minFFTscroll= widget;
  else if(*tag==FFTMAX)
    maxFFTscroll= widget;
  else if(*tag==FFTTFAC)
    tfacFFTscroll= widget;
  else if(*tag==LINEARFFT)
    linearFFTtoggleW= widget;
  else if(*tag==LOGFFT)
    logFFTtoggleW= widget;
  else if(*tag==NONLINFFT)
    nonlinFFTtoggleW= widget;
  else if(*tag==PICKLATTICE)
    pickLatticeToggleW= widget;
  else if(*tag==BACKFFT)
    backFFTToggleW= widget;
  else if(*tag==PICKCTF)
    pickCTFtoggleW= widget;
  else if(*tag==CTF_FOCUSMIN_VALUE)
    {ctfFocusMinvalueW= widget;
     handle_cr(widget);}
  else if(*tag==CTF_ANGMAJOR_VALUE)
    {ctfAngMajorValueW= widget;
     handle_cr(widget);}
  else if(*tag==CTF_DEFOCUS_VALUE)
    {ctffocusMajValueW= widget;
     handle_cr(widget);}
  else if(*tag==CTF_FOCUSMIN)
    ctfFocusMinscrollW= widget;
  else if(*tag==CTF_ANGMAJOR)
    ctfAngMajorScrollW= widget;
  else if(*tag==CTF_FOCUSMAJ)
    ctffocusMajScrollW= widget;
  else if(*tag==FFTZOOM)
    fftZoomToggleW = widget;
  else if(*tag==FILTER)
    filterFFTtoggleW = widget;
  else if(*tag==FILTSIZE)
    filtSizeWidget = widget;
  else if(*tag==FILTLABEL)
    filtLabelWidget = widget;
  else if(*tag==FILTDESC)
    filtDescWidget = widget;
  else if(*tag==DRAWLAT)
    drawLatToggleW = widget;
  else if(*tag==DRAWHILO)
    drawHiLoToggleW = widget;
  else if(*tag==A1024FFT)
    a1024FFTWidget= widget;
  else if(*tag==A512FFT)
    a512FFTWidget= widget;
  else if(*tag==A256FFT)
    a256FFTWidget= widget;
  else if(*tag==A128FFT)
    a128FFTWidget= widget;
  else if(*tag==HARD)
    hardToggleW = widget;
  else if(*tag==SOFT)
    softToggleW = widget;
  else if(*tag==CONTLABEL)
    contLabelWidget = widget;
  else if(*tag==FFTFEEDBACK)
    fftFeedbackLabelW = widget;
  else if(*tag==FFTBKG)
    fftBkgScrollW = widget;
  else if(*tag==FFTBKGVALUE)
    fftBkgValueW = widget;
  else if(*tag==TWODLINES)
    twodLinesToggleW = widget;
  else if(*tag==LAYERLINES)
    layerLinesToggleW = widget;
  else if(*tag==AMPLITUDE)
    amplitudeToggleW = widget;
  else if(*tag==LOP)
    lopScrollW = widget;
  else if(*tag==HIP)
    hipScrollW = widget;
  else if(*tag==FFTFEEDBACK1)
    fftFeedbackLabel1W = widget;
  else if(*tag==FFTFEEDBACK2)
    fftFeedbackLabel2W = widget;
  else if(*tag==FFTFEEDBACK3)
    fftFeedbackLabel3W = widget;
  else if(*tag==APORIG1)
    apOrigLabel1W = widget;
  else if(*tag==APORIG2)
    apOrigLabel2W = widget;
  else if(*tag==XmARROW_UP)
    {fftImgnumArrowUpW = widget;
    handle_cr(widget);}
  else if(*tag==XmARROW_DOWN)
    {fftImgnumArrowDownW = widget;
    handle_cr(widget);}
  else if (*tag==ALL)
    {/* Create all the dynamic widgets */

      printf("Creating lattice scrolls\n");

      for(i=0;i<5;i++)
	{/* Do 4 of them */
	  
	  XtSetArg(arglst[0], XmNx, 145);
	  XtSetArg(arglst[1], XmNy, (530+(i*40)) );
	  XtSetArg(arglst[2], XmNorientation, XmHORIZONTAL);
	  XtSetArg(arglst[3], XmNwidth, 150); 
	  XtSetArg(arglst[4], XmNmaximum, 122); 
	  XtSetArg(arglst[5], XmNminimum, -100);
 	  XtSetArg(arglst[6], XmNborderWidth, 2); 
	  XtSetArg(arglst[7], XmNvalue, 0);

	  latticeScrollW[i]= (Widget)XmCreateScrollBar(widget, "", arglst, 8);
	  
	  if(i==0)
	    sprintf(outStr,"A*");
	  else if(i==1)
	    sprintf(outStr,"B*");
	  else if(i==2)
	    sprintf(outStr,"Gamma*");
	  else if(i==3)
	    sprintf(outStr,"Omega");

	  createSimpleString(&simpleString,outStr);
	  XtSetArg(arglst[0], XmNx, 15);
	  XtSetArg(arglst[1], XmNy, (530+(i*40)) );
	  XtSetArg(arglst[2], XmNlabelString, simpleString); 
	  XtSetArg(arglst[3], XmNwidth, 93); 
	  latticeLabel1W[i]= (Widget)XmCreateLabel(widget, "", arglst, 4);
	  XFlush(dsply);
	  XmStringFree(simpleString);
	  
	  XtSetArg(arglst[0], XmNx, 315);
	  XtSetArg(arglst[1], XmNy, (525+(i*40)) );
	  XtSetArg(arglst[2], XmNcolumns, 5);
	  XtSetArg(arglst[3], XmNeditable, TRUE);
	  latticeLabel2W[i]= (Widget)XmCreateText(widget, "", arglst, 4);
	  XFlush(dsply);


	  /* Now put them on the screen */
	  if(FALSE)
	    {/* Default is not to manage thses */
	      XtManageChild(latticeScrollW[i]);
	      XtManageChild(latticeLabel1W[i]);
	      XtManageChild(latticeLabel2W[i]);
	    }/* Default is not to manage thses */

	  /* Add callbacks to our widgets */
	  callBck[0].callback= latScrollChanged;
	  callBck[0].closure= 0;
	  callBck[1].callback= 0;
	  callBck[1].closure= 0;
	  XtAddCallbacks(latticeScrollW[i], XmNvalueChangedCallback, 
			 callBck);  
	  XtAddCallbacks(latticeScrollW[i], XmNdragCallback, callBck);  

 	  handle_cr(latticeLabel2W[i]); 

	}/* Do 4 of them */

      XmStringFree(simpleString);

    }/* Create all the dynamic widgets */


  
}/* fft_CREATE */




/* ************************************************************
 * latScrollChanged() - When the lat scroll[i] on the screen
 * changes value call here.
 */
void latScrollChanged(Widget widget, int *tag, 
		      XmScrollBarCallbackStruct *callback_data)
{/* latScrollChanged */
  int thisVal,i,thisOne;
  float floatVal;
  XmString simpleString;

  thisOne= -99;
  for(i=0;i<12;i++)
    if(widget==latticeScrollW[i])
      thisOne= i;

  if(thisOne >=0)
    {/* Only do if found */
      latValChanged = TRUE;

      floatVal = (float)callback_data->value;
      if(thisOne == 3) /* omega */
	latScrollValue[thisOne] = floatVal / 10.0;
      else latScrollValue[thisOne] = floatVal / 100.0;
      sprintf(outStr,"%.2f",latScrollValue[thisOne]);

      XmTextSetString(latticeLabel2W[thisOne], outStr); 
      
      /* Reasons this is called are drag and value_changed - 
       * see /usr/include/Xm/Xm.h for defn's */
      calcNewLat(thisOne);

      calcFFT(FALSE, 0, 0);
      DrawFFTWindow();

    }/* Only do if found */

}/* latScrollChanged */




/* ************************************************************
 * latTextChanged() - When the lat text[i] on the screen
 * changes value call here.
 */
void latTextChanged(Widget widget, int *tag, 
		    XmAnyCallbackStruct *callback_data)
{/* LATTEXTCHANGED */
  int thisVal,i,thisOne;
  float floatVal;
  XmString simpleString;
  char *thisStr;

  thisOne= -99;
  for(i=0;i<12;i++)
    if(widget==latticeLabel2W[i])
      thisOne= i;

  if(thisOne >=0)
    {/* Only do if found */

      latValChanged = TRUE;

      thisStr = XmTextGetString(latticeLabel2W[thisOne]);
      floatVal = (float)strtod(thisStr,0);
      XtFree(thisStr);

      if(thisOne==3) /* Omega */
	latScrollValue[thisOne] = floatVal;
      else latScrollValue[thisOne] = floatVal;

/* Don't change you'll get an infinite loop !!
      XtSetArg(arglst[0], XmNvalue, (int)(floatVal*100.0));
      XtSetValues(latticeScrollW[thisOne], arglst, 1);
*/      

      /* Reasons this is called is value_changed - 
       * see /usr/include/Xm/Xm.h for defn's */
      calcNewLat(thisOne);

      calcFFT(FALSE, 0, 0);
      DrawFFTWindow();

    }/* Only do if found */

}/* LATTEXTCHANGED */




/*************************************************************
 * hipTextChanged() - When the HIP text widgets on the screen
 * change value call here.  Only called when the CR is pressed
 * in the text widget.
 */
void hipTextChanged(Widget widget, int *tag, 
		    XmAnyCallbackStruct *callback_data)
{/* HIPTEXTCHANGED */
  int thisVal,i,thisOne;
  float floatVal;
  XmString simpleString;
  char *thisStr;
  char *latText;

  latText = XmTextGetString(widget); 
  scanStep = (int)strtod(latText,0);
  XtFree(latText);

  XtSetArg(arglst[0], XmNvalue, (int)scanStep);
  XtSetValues(hipScrollW, arglst, 1);  

  calcFFT(FALSE, 0, 0);
  DrawFFTWindow();
      
}/* HIPTEXTCHANGED */





/***************************************************************
 * lopTextChanged() - When the LOP text widgets on the screen
 * change value call here.  Only called when the CR is pressed
 * in the text widget.
 */
void lopTextChanged(Widget widget, int *tag, 
		    XmAnyCallbackStruct *callback_data)
{/* LOPTEXTCHANGED */
  int thisVal,i,thisOne;
  float floatVal;
  XmString simpleString;
  char *thisStr;
  char *latText;

  latText = XmTextGetString(widget); 
  scanStep = (int)strtod(latText,0);
  XtFree(latText);

  XtSetArg(arglst[0], XmNvalue, (int)scanStep);
  XtSetValues(lopScrollW, arglst, 1);  

  calcFFT(FALSE, 0, 0);
  DrawFFTWindow();
      
}/* LOPTEXTCHANGED */




/* ************************************************************
 * lohiRadiusChanged() - When the LOP or HIP radius text 
 * widgets on the screen change value call here.  Only called 
 * when the CR is pressed in the text widget.
 */
void lohiRadiusChanged(Widget widget, int *tag, 
		       XmAnyCallbackStruct *callback_data)
{/* lohiRadiusCHANGED */
  int thisVal,i,thisOne;
  float floatVal;
  XmString simpleString;
  char *thisStr;
  char *latText;

  latText = XmTextGetString(lopRadiusTextW); 
  scanStep = (int)strtod(latText,0);
  XtFree(latText);
  lowResoFilter = scanStep * scanStep;

  latText = XmTextGetString(hipRadiusTextW); 
  scanStep = (int)strtod(latText,0);
  XtFree(latText);
  hiResoFilter = scanStep * scanStep;

  sprintf(outStr,"%d",(int)sqrt((float)lowResoFilter));
  XmTextSetString(lopRadiusTextW,outStr);
  sprintf(outStr,"%d",(int)sqrt((float)hiResoFilter));
  XmTextSetString(hipRadiusTextW,outStr);


  calcFFT(FALSE, 0, 0);
  DrawFFTWindow();
      
}/* lohiRadiusCHANGED */




/***************************************************************
 * ctfTextChanged() - When the CTF text widgets on the screen
 * change value call here.  Only called when the CR is pressed
 * in the text widget.
 */
void ctfTextChanged(Widget widget, int *tag, 
		    XmAnyCallbackStruct *callback_data)
{/* CTFTEXTCHANGED */
  int thisVal,i,thisOne;
  float floatVal;
  XmString simpleString;
  char *thisStr;
  char *latText;

  latText = XmTextGetString(widget); 

  if(widget==ctfScanStepW)
    {/* Get scan step */
      scanStep = (float)strtod(latText,0);
      if(scanStep > 0.0)
	{/* ok to calc */
	  pixSize = ((scanStep*1000.0)/magnification);
	  fftAngsPerPixel = ((float)pixSize*10.0);
	  sprintf(outStr,"%.4f",pixSize);
	  XmTextSetString(ctfPixSizeW,outStr);	
	  sprintf(outStr,"%.3f",fftAngsPerPixel);
	  XmTextSetString(fftMeasurePSTextW,outStr);
	  XFlush(dsply);

	  redo_ctfFocMajMin();

	}/* ok to calc */
    }/* Get scan step */
  else if(widget==ctfPixSizeW)
    {/* Get pixsize */
      pixSize = (float)strtod(latText,0);
      fftAngsPerPixel = ((float)pixSize*10.0);
      if(pixSize > 0.0)
	{/* ok to calc */
	  sprintf(outStr,"%.3f",fftAngsPerPixel);
	  XmTextSetString(fftMeasurePSTextW,outStr);
	  magnification = ((scanStep*1000.0)/pixSize);
	  sprintf(outStr,"%.0f",magnification);
	  XmTextSetString(ctfMagW,outStr);	
	  XFlush(dsply);

	  redo_ctfFocMajMin();

	}/* ok to calc */
    }/* Get pixsize */
  else if(widget==ctfMagW)
    {/* Get mag */
      magnification = (float)strtod(latText,0);
      if(magnification > 0.0)
	{/* ok to calc */
	  pixSize = ((scanStep*1000.0)/magnification);
	  fftAngsPerPixel = ((float)pixSize*10.0);
	  sprintf(outStr,"%.4f",pixSize);
	  XmTextSetString(ctfPixSizeW,outStr);	
	  sprintf(outStr,"%.3f",fftAngsPerPixel);
	  XmTextSetString(fftMeasurePSTextW,outStr);
	  XFlush(dsply);

	  redo_ctfFocMajMin();

	}/* ok to calc */
    }/* Get mag */

  XtFree(latText);
      
}/* CTFTEXTCHANGED */




/**************************************************************
 * drawCross() - Draw Cross to indicate x,y point slected.
*/
void drawCross(int drawX,int drawY,int drawXfin,int drawYfin,
	       GC colorGC)
{/* Draw cross */
  int arcRadius;
  
  arcRadius = 5;
  if(pickLattice != -99)
    XDrawLine(dsply, XtWindow(fftDrawingWidget), colorGC,
	      drawX,drawY,
	      drawXfin,drawYfin);
  else
    {
      /* Tim doesn't like these in CTF eSTIMATION SCREEN 
      XDrawArc(dsply, XtWindow(fftDrawingWidget), colorGC,
	       (drawX-arcRadius), 
	       (drawY-arcRadius),
	       (arcRadius*2),(arcRadius*2),
	       0, (360*64));
      **********************************/
    }
  
}/* Draw cross */



/****************************************************************
 * fft_flicker_activate() - They want to flicker the drawing of
 * the CTF with the raw FFT.
 */
void fft_flicker_activate(Widget widget, int *tag, 
			  XmAnyCallbackStruct *callback_data)
{/* FFT_FLICKER_ACTIVATE */
  int saveDrawCtf;

  saveDrawCtf = drawCTF;
  if(*tag==1)
    {/* No drawings */
      drawCTF = FALSE;
    }/* No drawings */

  DrawFFTWindow();

  drawCTF = saveDrawCtf;

}/* FFT_FLICKER_ACTIVATE */







/**************************************************************
 * DrawFFTWindow() - Redraw window e.g. on expose events
 * or whenever it's prudent.
 */
void DrawFFTWindow()
{/* DrawFFTWindow */
  int draw_width, draw_height, drawX,drawY,drawXfin,drawYfin;
  float xcen,ycen;
  int nx,ny,ik,ix,iy,ih,ix1,ix2,iy1,iy2;
  int h1,k1,h2,k2,i,j,thisY,loPixel,hiPixel,thisPixel;
  int minPixel,maxPixel;
  float x1,x2,y1,y2,thisNode,tempScaleF, calcValue;
  float scaleCircFactor,maxCA,minCA,scaledY,scaledYp1;
  int thisX, thisXp1, is1D, is2D;
#define CNVPTXC(x) ((fftZoomFactor==2) ? ((x-fftZoomX) * 2 +1) : (x))
#define CNVPTYC(y) ((fftZoomFactor==2) ? ((y-fftZoomY) * 2 - fftZoomFactor) : (y-fftZoomFactor))
  
#define CNVPTX(x) ((fftZoomFactor==2) ? ((x-fftZoomX) * 2) : (x))
#define CNVPTY(y) ((fftZoomFactor==2) ? ((y-fftZoomY) * 2) : (y))
  
#define DRAWBLANK(x) ((doingAutoDefocusRefine && adrBlankDraw) ? ((x-256) * 2) : (x))
#define DRAWBLANKY(y) ((doingAutoDefocusRefine && adrBlankDraw) ? (256-((256-y) * 2)) : (y))

  if(!drawLite)
    goto bugOutFFT;


  /* STuff the display buffer based on how zoomed we are */
  if(fftZoomFactor==1)
    {/* Normal */
      for(i=0;i<512;i++)
	for(j=0;j<512;j++)
	  {/* Zoom it normal */
	    thisPixel = ((i*512)+j);
	    if(hiColor)
	      {/* hi */
		XPutPixel(fft_image1,
			  j, i,
			  /* pixels_return[fft_gold_buffer1[thisPixel]] */
			  pixels_return[fft_gold_buffer1[thisPixel]]
			  );
	      }/* hi */
	    /* Always do this??? */
	    fft_buffer1[thisPixel]= fft_gold_buffer1[thisPixel];
	  }/* Zoom it normal */
    }/* Normal */
  else
    {/* Zoomed */
      for(i=0;i<512;i++)
	for(j=0;j<512;j++)
	  {/* Zoom it up */
	    thisX= (fftZoomX+(j/2));
	    thisY= (fftZoomY+(i/2));
	    thisPixel = ((i*512)+j);
	    if(hiColor)
	      {/* hi */
		XPutPixel(fft_image1,
			  j, i,
			  pixels_return[fft_gold_buffer1[((thisY*512)+thisX)]]
			  );
	      }/* hi */
	    /* Always do this ??? */
	    fft_buffer1[(thisPixel)]= 
	      fft_gold_buffer1[((thisY*512)+thisX)];
	  }/* Zoom it up */
    }/* Zoomed */
  
  if(doingAutoDefocusRefine && adrBlankDraw)
    {/* Black it out */
      for(i=0;i<512;i++)
	for(j=0;j<512;j++)
	  {/* Zoom it up */
	    thisPixel = ((i*512)+j);
	    fft_buffer1[(thisPixel)]= fft_gold_buffer1[(((128+(i/2))*512)+
							(256-(j/2)))];
	}/* Zoom it up */
    }/* Black it out */
	
  is1D = XtIsManaged(autoDRScreenWidget); 
  is2D = XtIsManaged(auto2DDRScreenWidget); 

  if(is2D && drawAuto2Dctf)
    {/* Draw the auto2Dctf feedback */
      int x,y;
      for(y=0;y<512;y++)
	{/* y */
	  for(x=0;x<256;x++)
	    {/* x */	      
	      XPutPixel(fft_image1,
			256+x, y,
			pixels_return[(int)auto2DdrawBuf[((y*256)+x)]]
			);
	    }/* x */
	}/* y */
      XFlush(dsply);
    }/* Draw the auto2Dctf feedback */

  draw_width = 512;
  draw_height = 512;
  if(XtIsManaged(fftscreen))
    XPutImage(dsply, fft_window1, fft_gc, fft_image1, 0, 0,
	      0, 0,  draw_width,  draw_height);
  XFlush (dsply);


  if(is1D || drawCTF)
    {/* Draw the CTF elipse */
      int x1,x2,y1,y2,minX,minY,maxX,maxY,beginEl,endEl;
      int *dX1P,*dY1P,*drawX1,*drawY1;
      float distAsq,distBsq,distA,distB,cosAng1,sinAng1;
      float ang1,angThis;
      char numStr[64];

      /* We calculate drawX1 and drawY1 below 0 and above
       * 1024 so make thee arrays to have entries there or
       * we get memory violations/trashing */
      dX1P = (int *)malloc(sizeof(int)*2048);
      dY1P = (int *)malloc(sizeof(int)*2048);
      drawX1 = dX1P + 512;
      drawY1 = dY1P + 512;

      for(j= 0;j<MAXCTFS;j++)
	{/* Draw all CTF curves */

	  if(pickedLatPts)
	    {/* Only do if there's two */
	      
	      thisNode = (float)((float)j+1.0);

	      distB = ctfPts[j].distB;
	      distBsq = distB * distB;
	      distA = ctfPts[j].distA;
	      distAsq = distA * distA;
	      ang1 = ctfAngMajorValue;

	      beginEl= 512-((int)distA); /* We offset this by 256 pixels*/
	      endEl= 512+((int)distA);
	      /* DOn't threshold it 
	      beginEl = MAX(0,beginEl);
	      endEl = MIN(endEl,1023);
	      */

	      for(i=(beginEl);i<=(endEl);i++)
		{/* Do all Y */
		  int thisX, thisY, thisI;
		  double tempF, thisR;
		  
		  thisI = i-256;  /* We're offsetting this by 256 pixels */
		  
		  tempF= (double)fabs(distBsq *
				      (1.0 - (
					      (float)((thisI-256)*
						      (thisI-256))/ 
					      distAsq)
				       )
				      );
		  tempF= (float)sqrt((double)tempF); 

		  if(thisI==(beginEl-256) || (thisI==endEl-256))
		    tempF= 0.0;

		  if(thisI==256)
		    angThis = (float)atan((double)
					  (
					   ((float)(tempF))/
					   ((float)(0.0000001))
					   )
					  );
		  else
		    angThis = (float)atan((double)
					  (
					   ((float)(tempF))/
					   ((float)(256-thisI))
					   )
					  );
		  
		  thisX = 256 - thisI;
		  thisY = (int)tempF;
		  thisR = (double)sqrt((double)((thisX*thisX)+(thisY*thisY)));
		  
		  /* Now rotate in ang1 */
		  drawX1[i] = 256 + 
		    (int)(thisR * cos((double)(angThis-ang1)));
		  drawY1[i] = 256 + 
		    (int)(thisR * sin((double)(angThis-ang1)));
		  
		  drawX1[i] = DRAWBLANK(drawX1[i]);
		  drawY1[i] = DRAWBLANKY(drawY1[i]);

		}/* Do all Y */
	      

	      /* Now draw them */
	      for(i=(beginEl);i<(endEl);i++)
		{/* Do all Y */
		  /* i==256 draws a line down the middle */
		  if(i!=512)
		    {/* Then draw them */
		      if(j==ctfCurveNum)
			{/* Special */
			  XDrawLine(dsply, XtWindow(fftDrawingWidget), 
				    image_gcRED,
				    CNVPTXC(drawX1[i]),
				    CNVPTYC(drawY1[i]),
				    CNVPTXC(drawX1[i+1]),
				    CNVPTYC(drawY1[i+1]));
			  if(!doingAutoDefocusRefine ||
			     (doingAutoDefocusRefine && !adrBlankDraw))
			  XDrawLine(dsply, XtWindow(fftDrawingWidget), 
				    image_gcRED,
				    CNVPTXC((256) + 
					   (256-drawX1[i])),
				    CNVPTYC((256) - 
					   (drawY1[i]-256)),
				    CNVPTXC((256) + 
					   (256-drawX1[i+1])),
				    CNVPTYC((256) - 
					   (drawY1[i+1]-256)));  
			}/* Special */
		      else
			{/* Normal */
			  XDrawLine(dsply, XtWindow(fftDrawingWidget), 
				    image_gcLAT,
				    CNVPTXC(drawX1[i]),
				    CNVPTYC(drawY1[i]),
				    CNVPTXC(drawX1[i+1]),
				    CNVPTYC(drawY1[i+1]));

			  if(!doingAutoDefocusRefine ||
			     (doingAutoDefocusRefine && !adrBlankDraw))
			  XDrawLine(dsply, XtWindow(fftDrawingWidget), 
				    image_gcLAT,
				    CNVPTXC((256) + 
					   (256-drawX1[i])),
				    CNVPTYC((256) - 
					   (drawY1[i]-256)),
				    CNVPTXC((256) + 
					   (256-drawX1[i+1])),
				    CNVPTYC((256) - 
					   (drawY1[i+1]-256)));  
			}/* Normal */
		      XFlush(dsply);  
		    }/* Then draw them */
		}/* Do all Y */
	      
	      
	      XFlush (dsply);
	    }/* Only do if there's two */
      
	}/* Draw all CTF curves */
      free(dX1P);
      free(dY1P);
      
      if(pickedLatPts)
	{/* Only do if there's two */
	  /* Reset these just in case */
	  latPt[0].measuredX= ctfPts[ctfCurveNum].latX0;
	  latPt[0].measuredY= ctfPts[ctfCurveNum].latY0;
	  latPt[1].measuredX= ctfPts[ctfCurveNum].latX1;
	  latPt[1].measuredY= ctfPts[ctfCurveNum].latY1;
	}/* Only do if there's two */

      /* Now see if there is any autoDR (Automatic Defocus Refinement
       * stuff to draw */
      if(is1D && doingAutoDefocusRefine)
	{/* Only do if appropriate */
	  int i,j;

	  if(ca1d)
	    {/* Don't do if nothing there */
	      maxCA = minCA = ca1d[3];
	      minCA = 0.0;
	      for(i=3;i<256;i++)
		{/* Do all */
		  minCA = MIN(ca1d[i],minCA);
		  maxCA = MAX(ca1d[i],maxCA);
		}/* Do all */
	      if(maxCA != minCA)
		{/* OK to do */
		  scaleCircFactor = 505.0 / ((float)(maxCA-minCA)); 
		}/* OK to do */
	    }/* Don't do if nothing there */

	  if(adrCircAvgDraw)
	    {/* Draw circavg */
	      
	      if(ca1d)
		{/* Don't do if nothing there */
		  for(i=256;i<511;i++)
		    {/* DO all */
		      scaledY = (ca1d[(i-256)]-minCA)*scaleCircFactor;
		      scaledYp1 = (ca1d[(i-256)+1]-minCA)*scaleCircFactor;
		      scaledY = (510.0 - scaledY);
		      scaledYp1 = (510.0 - scaledYp1);
		      thisX= i;
		      thisXp1= i+1;
		      if(i==256)
			scaledY = 256;
		      thisX = DRAWBLANK(thisX);
		      thisXp1= DRAWBLANK(thisXp1);
		      XDrawLine(dsply, XtWindow(fftDrawingWidget), 
				image_gcGREEN,
				CNVPTX(thisX),(int)scaledY,
				CNVPTX(thisXp1),(int)scaledYp1);
		    }/* DO all */
		}/* Don't do if nothing there */
	      
	    }/* Draw circavg */
	  if(adrBDraw)
	    {/* Draw B */
	      if(adr_b)
		{
		      for(i=256;i<511;i++)
			{/* DO all */
			  scaledY = (adr_b[(i-256)]-minCA)*scaleCircFactor;
			  scaledYp1 = (adr_b[(i-256)+1]-minCA)*scaleCircFactor;
			  scaledY = (510.0 - scaledY);
			  scaledYp1 = (510.0 - scaledYp1);
			  thisX= i;
			  thisXp1= i+1;
			  if(i==256)
			    scaledY = 256;
			  thisX = DRAWBLANK(thisX);
			  thisXp1= DRAWBLANK(thisXp1);
			  XDrawLine(dsply, XtWindow(fftDrawingWidget), 
				    image_gcBLUE,
				    CNVPTX(thisX),(int)scaledY,
				    CNVPTX(thisXp1),(int)scaledYp1);
			}/* DO all */
		}
	    }/* Draw B */
	  if(adrSDraw)
	    {/* Draw S */
	      if(adr_s)
		{
		      for(i=256;i<511;i++)
			{/* DO all */
			  scaledY = (adr_s[(i-256)]-minCA)*scaleCircFactor;
			  scaledYp1 = (adr_s[(i-256)+1]-minCA)*scaleCircFactor;
			  scaledY = (510.0 - scaledY);
			  scaledYp1 = (510.0 - scaledYp1);
			  thisX= i;
			  thisXp1= i+1;
			  if(i==256)
			    scaledY = 256;
			  thisX = DRAWBLANK(thisX);
			  thisXp1= DRAWBLANK(thisXp1);
			  XDrawLine(dsply, XtWindow(fftDrawingWidget), 
				    image_gcRED,
				    CNVPTX(thisX),(int)scaledY,
				    CNVPTX(thisXp1),(int)scaledYp1);
			}/* DO all */
		}
	    }/* Draw S */
	  if(adrSminusBDraw)
	    {/* Draw SminusB */
	      if(adr_sminusb)
		{
		  maxCA = minCA = adr_sminusb[0];
		  for(i=0;i<256;i++)
		    {/* Do all */
		      minCA = MIN(adr_sminusb[i],minCA);
		      maxCA = MAX(adr_sminusb[i],maxCA);
		    }/* Do all */
		  if(maxCA != minCA)
		    {/* OK to do */
		      scaleCircFactor = 250.0 / ((float)(maxCA-minCA)); 
		      
		      for(i=256;i<511;i++)
			{/* DO all */
			  scaledY = (adr_sminusb[(i-256)]-minCA)*
			    scaleCircFactor;
			  scaledYp1 = (adr_sminusb[(i-256)+1]-minCA)*
			    scaleCircFactor;
			  scaledY = (255.0 - scaledY);
			  scaledYp1 = (255.0 - scaledYp1);
			  thisX= i;
			  thisXp1= i+1;
			  if(i==256)
			    scaledY = 256;
			  thisX = DRAWBLANK(thisX);
			  thisXp1= DRAWBLANK(thisXp1);
			  XDrawLine(dsply, XtWindow(fftDrawingWidget), 
				    image_gcBLUE,
				    CNVPTX(thisX),(int)scaledY,
				    CNVPTX(thisXp1),(int)scaledYp1);
			}/* DO all */
		    }/* OK to do */
		}
	    }/* Draw SminusB */
	  if(adrDDraw)
	    {/* Draw D */
	      if(adr_d)
		{
		  /* Use sminusb to scale here */
		  maxCA = minCA = adr_sminusb[0];
		  for(i=0;i<256;i++)
		    {/* Do all */
		      minCA = MIN(adr_sminusb[i],minCA);
		      maxCA = MAX(adr_sminusb[i],maxCA);
		    }/* Do all */
		  if(maxCA != minCA)
		    {/* OK to do */
		      scaleCircFactor = 250.0 / ((float)(maxCA-minCA)); 
		      for(i=256;i<511;i++)
			{/* DO all */
			  scaledY = (adr_d[(i-256)]-minCA)*scaleCircFactor;
			  scaledYp1 = (adr_d[(i-256)+1]-minCA)*scaleCircFactor;
			  scaledY = (255.0 - scaledY);
			  scaledYp1 = (255.0 - scaledYp1);
			  thisX= i;
			  thisXp1= i+1;
			  if(i==256)
			    scaledY = 256;
			  thisX = DRAWBLANK(thisX);
			  thisXp1= DRAWBLANK(thisXp1);
			  XDrawLine(dsply, XtWindow(fftDrawingWidget), 
				    image_gcWHITE,
				    CNVPTX(thisX),(int)scaledY,
				    CNVPTX(thisXp1),(int)scaledYp1);
			}/* DO all */
		    }/* OK to do */
		}
	    }/* Draw D */
 	  if(adrCtfsqDraw)
	    {/* Draw ctfSq */
	      if(adr_ctfsq)
		{
		  maxCA = minCA = adr_ctfsq[0];
		  for(i=0;i<256;i++)
		    {/* Do all */
		      minCA = MIN(adr_ctfsq[i],minCA);
		      maxCA = MAX(adr_ctfsq[i],maxCA);
		    }/* Do all */
		  if(maxCA != minCA)
		    {/* OK to do */
		      scaleCircFactor = 250.0 / ((float)(maxCA-minCA)); 
		      
		      for(i=256;i<511;i++)
			{/* DO all */
			  scaledY = (adr_ctfsq[(i-256)]-minCA)*scaleCircFactor;
			  scaledYp1 = (adr_ctfsq[(i-256)+1]-minCA)*
			    scaleCircFactor;
			  scaledY = (255.0 - scaledY);
			  scaledYp1 = (255.0 - scaledYp1);
			  thisX= i;
			  thisXp1= i+1;
			  if(i==256)
			    scaledY = 256;
			  thisX = DRAWBLANK(thisX);
			  thisXp1= DRAWBLANK(thisXp1);
			  XDrawLine(dsply, XtWindow(fftDrawingWidget), 
				    image_gcGREEN,
				    CNVPTX(thisX),(int)scaledY,
				    CNVPTX(thisXp1),(int)scaledYp1);
			}/* DO all */
		    }/* OK to do */
		}
	    }/* Draw ctfSq */

	}/* Only do if appropriate */

      if(is2D &&  doingAuto2DDefocusRefine )
	{/* Draw the axes 2D refine */
	  float tmpAng, curAngle;
	  char *ctfText;
	  ix1 = 256;
	  iy1 = 256;

	  ctfText = XmTextGetString(ctfAngMajorW); 
	  curAngle = (float)strtod(ctfText, 0);
	  XtFree(ctfText);
	  if(curAngle > 180.0)
	    curAngle -= 180.0;

	  /* First do major */
	  if(curAngle < 90.0)
	    {
	      ix2 = 512;
	      iy2 = (int)(256.0 - (256.0 * tan(deg2rad(curAngle))));
	    }
	  else if(curAngle == 90.0)
	    {
	      ix2 = ix1;
	      iy2 = 0;
	    }
	  else
	    {
	      tmpAng = 90.0 - (curAngle-90.0);
	      ix2 = 0;
	      iy2 = (int)(256.0 - (256.0 * tan(deg2rad(tmpAng))));
	    }
	  XDrawLine(dsply, XtWindow(fftDrawingWidget), 
		    image_gcLAT,
		    ix1,iy1,
		    ix2,iy2);
	  /* Now minor */
	  curAngle -= 90.0;
	  if(curAngle < 0.0)
	    {
	      tmpAng = 0.0 - curAngle;
	      ix2 = 512;
	      iy2 = (int)(256.0 + (256.0 * tan(deg2rad(tmpAng))));
	    }
	  else
	    {
	      ix2 = 512;
	      iy2 = (int)(256.0 - (256.0 * tan(deg2rad(curAngle))));
	    }
	  XDrawLine(dsply, XtWindow(fftDrawingWidget), 
		    image_gcLAT,
		    ix1,iy1,
		    ix2,iy2);


	}/* Draw the axes 2D refine */

    }/* Draw the CTF elipse */
  else
    {/* Lattice refinement */
      if(!layerLines)
	{/* 2D draw */
	  if(drawLatToggle && pickedLatPts && !measureFFT)
	    {/* Draw 2D lattice lines */
	      
	      xcen= ycen = 256.0;
	      
	      if(axL != 0.0)
		nx = abs((int)(512.0/(axL*fftZoomFactor))) + 11;
	      if(byL != 0.0)
		ny = abs((int)(512.0/(byL*fftZoomFactor))) + 11;
	      if(ny < 50000 && nx < 50000)
		{/* Don't do ridiculous ones */
	      
	      /*
		 c  **************************
		 c  *    astar grid lines    *
		 c  **************************
		 */
	      ik = -(ny/2) - 2;
	      for(iy=0;iy<ny;iy++)
		{/* do iy = 1,ny */
		  ik = ik + 1;
		  ih = -(nx/2) - 1;
		  x1 = (float)(xcen) + ih*axL*fftZoomFactor + 
		    ik*bxL*fftZoomFactor;

		  y1 = (float)(ycen) - ih*ayL*fftZoomFactor - 
		    ik*byL*fftZoomFactor;
		  y1--;

		  for(ix=0;ix<nx;ix++)
		    {/* do ix = 1,nx */
		      x2 = x1 + axL*fftZoomFactor;
		      y2 = y1 - ayL*fftZoomFactor;
		      ix1 = (int)(x1+0.5);
		      iy1 = (int)(y1+0.5);
		      ix2 = (int)(x2+0.5);
		      iy2 = (int)(y2+0.5);
		      XDrawLine(dsply, XtWindow(fftDrawingWidget), 
				image_gcLAT,
				ix1,iy1,ix2,iy2);
		      
		      /* call lx_vecclip (ix1,iy1,ix2,iy2,128,ixlim-128,0,
			 iylim,icol)*/
		      x1 = x2;
		      y1 = y2;
		      ih = ih + 1;
		      
		    }/* do ix = 1,nx */
		}/* do iy = 1,ny */
	      
	      /*
		 c  **************************
		 c  *    bstar grid lines    *
		 c  **************************
		 */
	      ih = -(nx/2) - 2;
	      for(ix=0;ix<nx;ix++)
		{/* do ix = 1,nx */
		  ih = ih + 1;
		  ik = -(ny/2) - 1;
		  x1 = (float)(xcen) + ih*axL*fftZoomFactor + 
		    ik*bxL*fftZoomFactor;

		  y1 = (float)(ycen) - ih*ayL*fftZoomFactor - 
		    ik*byL*fftZoomFactor;
		  y1--;

		  for(iy=0;iy<ny;iy++)
		    {/* do iy = 1,ny */
		      x2 = x1+bxL*fftZoomFactor;
		      y2 = y1-byL*fftZoomFactor;
		      ix1 = (int)(x1+0.5);
		      iy1 = (int)(y1+0.5);
		      ix2 = (int)(x2+0.5);
		      iy2 = (int)(y2+0.5);
		      XDrawLine(dsply, XtWindow(fftDrawingWidget), 
				image_gcLAT,
				ix1,iy1,ix2,iy2);
		      /*call lx_vecclip (ix1,iy1,ix2,iy2,128,ixlim-128,0,
			iylim,icol)*/
		      x1 = x2;
		      y1 = y2;
		      ik = ik+1;
		    }/* do iy = 1,ny */
		  
		}/* do ix = 1,nx */
	      
		}/* Don't do ridiculous ones */

	    }/* Draw 2D lattice lines */
	}/* 2D draw */
      else if(drawLatToggle && pickedLatPts && !measureFFT)
	{/* Just draw layerlines */
	  float space,fall,slope,slopeRad,intercept,thisI;
	  float cosang,tanang;
	  int ix1,ix2,iy1,iy2;
	  if(latLineOrder > 0)
	    {/* OK to do */
	      fall  = (float)filtSizeValue;
	      slope = ((float)(
			       (256-latPt[0].measuredY)-
			       (256-latPt[1].measuredY)
			       ))/
				 ((float)(latPt[0].measuredX-
					  latPt[1].measuredX));
	      intercept = (float)((float)((float)(256-
					 latPt[0].measuredY))) - 
					   (float)(slope * (float)((float)(
					latPt[0].measuredX-256)));
	      if(pickLattice != -99 && !latValChanged)
		{/* Calc it */
		  latScrollValue[1] = 
		    dstar = dstarSave= (fabs((float)(intercept)*
					     cos(atan2(slope,1.0))) / 
					(float)latLineOrder);    
		  latScrollValue[3]= omega= omegaSave = 
		    rad2deg(atan2((slope),1.0));
		  /* Do we really need to do this here??
		   * May cause infinite looping  SIgnal no redraw with
		   * -99 tag */
		  latReset_activate(0,-99,0);  
		   
		}/* Calc it */
	      else
		{ /* Get from scroll */
		  dstar = latScrollValue[1]; 
		  omega = latScrollValue[3]; 

		  slope     = tan(deg2rad(omega)); 
		  intercept = cos(deg2rad(omega)); 
		} /* Get from scroll */

	      space = dstar;  /* ORIG way */
	      cosang = cos(deg2rad(omega));
	      tanang = tan(deg2rad(omega));

	      if(cos(omega) != 0.0)
		intercept = dstar / cos(deg2rad(omega));

	      thisI = 256.0;
	      while(thisI < 2750.0)
		{/* Draw lines */
		  
		  /* Draw in positive direction. */
		  ix1= 0;
		  iy1= (int)((float)thisI-(float)(slope*
					      (float)((float)ix1-256.0)));
		  ix2= 511;
		  iy2= (int)((float)thisI-(float)(slope*
					      (float)((float)ix2-256.0)));
		  /* Why do we have to do this???  */
		  	      iy1--;
			      iy2--;
		  XDrawLine(dsply, XtWindow(fftDrawingWidget), image_gcLAT,
			    ix1,iy1,ix2,iy2);
		  
		  if(thisI!=256.0)
		    {/* Draw in negative direction. */
		      ix1= 0;
		      iy1= (int)((float)(256.0-(thisI-256.0))-
				 (float)(slope*(float)((float)ix1-256.0)));
		      ix2= 511;
		      iy2= (int)((float)(256.0-(thisI-256.0))-
				 (float)(slope*(float)((float)ix2-256.0)));
		      /* Why do we have to do this??? */
		      iy1--;
		      iy2--;
		      XDrawLine(dsply, XtWindow(fftDrawingWidget), 
				image_gcLAT,
				ix1,iy1,ix2,iy2);
		    }/* Draw in negative direction. */

		  /* ORIGWAY thisI += fabs(space); */
		  thisI += (float)(fabs(intercept)*(float)fftZoomFactor);

		}/* Draw lines */
	      
	    }/* OK to do */
	  else askThem("Must have a positive Lattice Line Order.",
		       FALSE,XmDIALOG_INFORMATION);
	  
	}/* Just draw layerlines */
      
    }/* Lattice refinement */
  
  
  if(!doingAutoDefocusRefine)
    {/* Only draw these if not auto defocus refining */
      if(((pickLattice > 0) || pickedLatPts) &&
	 !filterToggle && !measureFFT)
	{/* Draw what they've selected */
	  drawX= CNVPTX(latPt[0].measuredX) - 5;
	  drawY= CNVPTY(latPt[0].measuredY)-1;
	  drawXfin= CNVPTX(latPt[0].measuredX) + 5;
	  drawYfin= CNVPTY(latPt[0].measuredY)-1;
	  if(pickLattice != -99)
	    drawCross(drawX,drawY,drawXfin,drawYfin, image_gcGREEN);
	  drawX= CNVPTX(latPt[0].measuredX);
	  drawY= CNVPTY(latPt[0].measuredY) - 5;
	  drawXfin= CNVPTX(latPt[0].measuredX);
	  drawYfin= CNVPTY(latPt[0].measuredY) + 5; 
	  if(pickLattice != -99)
	    drawCross(drawX,drawY,drawXfin,drawYfin, image_gcGREEN);
	  else drawCross(CNVPTX(latPt[0].measuredX),CNVPTY(latPt[0].measuredY),
			 0,0,image_gcGREEN);
	  
	  drawX= CNVPTX(latPt[1].measuredX) - 5;
	  drawY= CNVPTY(latPt[1].measuredY)-1;
	  drawXfin= CNVPTX(latPt[1].measuredX) + 5;
	  drawYfin= CNVPTY(latPt[1].measuredY)-1;
	  if(pickLattice != -99)
	    drawCross(drawX,drawY,drawXfin,drawYfin, image_gcRED);
	  drawX= CNVPTX(latPt[1].measuredX);
	  drawY= CNVPTY(latPt[1].measuredY) - 5;
	  drawXfin= CNVPTX(latPt[1].measuredX);
	  drawYfin= CNVPTY(latPt[1].measuredY) + 5;
	  if(pickLattice != -99)
	    drawCross(drawX,drawY,drawXfin,drawYfin, image_gcRED);
	  else drawCross(CNVPTX(latPt[1].measuredX),CNVPTY(latPt[1].measuredY),
			 0,0,image_gcRED);
	}/* Draw what they've selected */
    }/* Only draw these if not auto defocus refining */
  
  if( (filterToggle && drawHiLoToggle) ||
      (lopHipSwitch && drawHiLoToggle) )
    {/* Draw the low/hi pass filters */
      int drawX,drawY,arcRadius;
      if(hiResoFilter > 0)
	{/* Draw it */
	  drawX = 256;
	  drawY = 256;
	  if(TRUE || debugLevel==7)
	    {/* DEBUG try these */
	      drawY= 255;
	      if(fftSize==128)
		{/* DO this */
		  drawX = 257;
		  drawY = 254;
		}/* DO this */
	      else if(fftSize==256)
		{/* DO this */
		  drawX = 256;
		  drawY = 255;
		}/* DO this */
	    }/* DEBUG try these */
	  arcRadius = (int)(sqrt((double)hiResoFilter) * 
			    (float)fftZoomFactor);
	  XDrawArc(dsply, XtWindow(fftDrawingWidget), image_gcRED,
		   (drawX-arcRadius), 
		   (drawY-arcRadius),
		   (arcRadius*2),(arcRadius*2),
		   0, (360*64));
	}/* Draw it */
      if(lowResoFilter < 320000)
	{/* Draw it */
	  drawX = 256;
	  drawY = 256;
	  if(TRUE || debugLevel==7)
	    {/* DEBUG try these */
	      drawY= 255;
	      if(fftSize==128)
		{/* DO this */
		  drawX = 257;
		  drawY = 254;
		}/* DO this */
	      else if(fftSize==256)
		{/* DO this */
		  drawX = 256;
		  drawY = 255;
		}/* DO this */
	    }/* DEBUG try these */
	  arcRadius = (int)(sqrt((double)lowResoFilter) * 
			    (float)fftZoomFactor);
	  XDrawArc(dsply, XtWindow(fftDrawingWidget), image_gcGREEN,
		   (drawX-arcRadius), 
		   (drawY-arcRadius),
		   (arcRadius*2),(arcRadius*2),
		   0, (360*64));
	}/* Draw it */

      if(lopHipSwitch && drawHiLoToggle)
	{/* Draw dotted circles where the drop off is */
	  float hiCutmax,loCutmax;
	  char *inStr;

	  inStr     = XmTextGetString(lopTextW); 
	  loCutmax  = (float)strtod(inStr,0);
	  XtFree(inStr);
	  inStr     = XmTextGetString(hipTextW); 
	  hiCutmax  = (float)strtod(inStr,0);
	  XtFree(inStr);
	  if(hiResoFilter > 0)
	    {/* Draw it */
	      drawX = 256;
	      drawY = 256;
	      if(TRUE || debugLevel==7)
		{/* DEBUG try these */
		  drawY= 255;
		  if(fftSize==128)
		    {/* DO this */
		      drawX = 257;
		      drawY = 254;
		    }/* DO this */
		  else if(fftSize==256)
		    {/* DO this */
		      drawX = 256;
		      drawY = 255;
		    }/* DO this */
		}/* DEBUG try these */
	      arcRadius = (int)((sqrt((double)hiResoFilter)-hiCutmax) * 
				(float)fftZoomFactor);
	      XDrawArc(dsply, XtWindow(fftDrawingWidget), image_gcDOTred,
		       (drawX-arcRadius), 
		       (drawY-arcRadius),
		       (arcRadius*2),(arcRadius*2),
		       0, (360*64));
	    }/* Draw it */
	  if(lowResoFilter < 320000)
	    {/* Draw it */
	      drawX = 256;
	      drawY = 256;
	      if(TRUE || debugLevel==7)
		{/* DEBUG try these */
		  drawY= 255;
		  if(fftSize==128)
		    {/* DO this */
		      drawX = 257;
		      drawY = 254;
		    }/* DO this */
		  else if(fftSize==256)
		    {/* DO this */
		      drawX = 256;
		      drawY = 255;
		    }/* DO this */
		}/* DEBUG try these */
	      arcRadius = (int)((sqrt((double)lowResoFilter)+loCutmax) * 
				(float)fftZoomFactor);
	      XDrawArc(dsply, XtWindow(fftDrawingWidget), image_gcDOTgreen,
		       (drawX-arcRadius), 
		       (drawY-arcRadius),
		       (arcRadius*2),(arcRadius*2),
		       0, (360*64));
	    }/* Draw it */
	  
	}/* Draw dotted circles where the drop off is */


    }/* Draw the low/hi pass filters */

  if(contour)
    {/* Put up contour image */
      if(XtIsManaged(fftscreen))
	XPutImage(dsply, fftContour_window1, fftContour_gc, 
		  fftContour_image1, 
		  0, 0, 0, 0,  128,  128);
      drawFFTcontour();
      XFlush (dsply);
    }/* Put up contour image */

  if(is2D &&  doingAuto2DDefocusRefine  && 
      (drawAuto2Dscurve ||
       a2drSDraw ||
       a2drSminusBDraw ||
       a2drCtfsqDraw ||
       a2drBDraw ||
       a2drDDraw
       ))
    {/* Draw the auto2Dscurve feedback */
      float psi, phi, pi, radian, degree, defocus;
      float thisAngle,ratio,xDiff,yDiff,saveDef;

      /* Now let's draw the appropriate s curve */
      piraddeg_(&pi,&radian,&degree);

      phi = auto2Dangle * radian;
      
      /* Now calc the angle on the screen */
      psi = ratio = 0.0;
      if(
	 (scurveClick.x < 512) &&
	 (scurveClick.y < 256) &&
	 (scurveClick.y > 0)
	 )
	{/* OK to do */
	  yDiff = (256.0 - (float)scurveClick.y);
	  xDiff = ((float)scurveClick.x-256.0);
	  if(xDiff < 0.0)
	    xDiff = 0.0 - xDiff;
	  if(xDiff != 0.0)
	    ratio = yDiff / xDiff;
	  else
	    ratio = 999999999.9;
	}/* OK to do */
      thisAngle = atan((float)ratio);
      if(scurveClick.x < 256.0)
	thisAngle = (90.0*radian) + 
	  ((90.0*radian) - (float)atan((float)ratio));
      /* If degrees needed 
	 psi = thisAngle * degree;
      */
      psi = thisAngle ;

      if(ratio > 0.0)
	{/* OK to go */
	  ctf_astig_(&defocus ,
		     &auto2DdefMajor,
		     &auto2DdefMinor,
		     &psi,
		     &phi);	

	  sprintf(outStr,"[%f] defocus",defocus);
	  XDrawString(dsply, XtWindow(fftDrawingWidget), 
		      image_gcRED,
		      15, 15, outStr, strlen(outStr));

	  XDrawLine(dsply, XtWindow(fftDrawingWidget), 
		    image_gcRED,
		    257,255,
		    scurveClick.x,scurveClick.y);
	  XFlush(dsply);
	  
	  if(drawAuto2Dscurve)
	    {
	      saveDef = ctffocusMajValue;
	      ctffocusMajValue = defocus;
	      drawSCURVEonFFT();
	      ctffocusMajValue = saveDef;
	    }
	}/* OK to go */

    }/* Draw the auto2Dscurve feedback */

  if(is2D && doingAuto2DDefocusRefine  && 
      (a2drSDraw ||
       a2drSminusBDraw ||
       a2drCtfsqDraw ||
       a2drBDraw ||
       a2drDDraw))
    {/* Get curves */
      float defMajor, defMinor, angle, slope, angdraw;
      float   volts, ampContrast, Cs, pixelSize;
      float minS,maxS,scaleS;
      int thisPixel, xfac;
      char *ctfText;
      
      ctfText = XmTextGetString(a2ddrfjW);
      defMajor = (float)strtod(ctfText,0);
      XtFree(ctfText);
      ctfText = XmTextGetString(a2ddrfnW);
      defMinor = (float)strtod(ctfText,0);
      XtFree(ctfText);
      ctfText = XmTextGetString(a2ddragW);
      angle = (float)strtod(ctfText,0);
      XtFree(ctfText);
      ctfText = XmTextGetString(ctfKvW); 
      volts   = (float)strtod(ctfText, 0);
      volts *= 1000.0;    /* Convert from kiloVolts to volts */
      XtFree(ctfText);
      ctfText = XmTextGetString(ctfSpAberW); 
      Cs   = (float)strtod(ctfText, 0);
      XtFree(ctfText);
      /* DONT DO THIS??? Cs *= 10000000.0;  /+ Put into Angstroms */
      ctfText = XmTextGetString(ctfAmpFactorW); 
      ampContrast = (float)strtod(ctfText, 0);
      XtFree(ctfText);
      ctfText = XmTextGetString(ctfPixSizeW); 
      pixelSize   = (float)strtod(ctfText, 0);
      XtFree(ctfText);
      pixelSize *= 10.0;     /* Convert to Angs */
      
      if(scurveClick.x != 256.0)
	slope = ((float)(256.0 - (float)scurveClick.y) / 
		 (float)((float)scurveClick.x - 256.0));
      else
	slope = 0.0;
      if(slope==0.0)
	slope = 0.00000001;
      angdraw = atan(slope);
      angdraw = rad2deg(angdraw);

      robangle_curves_(&angdraw, 
		       &angle, 
		       poly_paras,
		       &defMajor,&defMinor,
		       &hkl_max,
		       &order_b,&order_d,
		       fftasum, 
		       &volts,
		       &ampContrast,
		       &Cs,
		       &pixelSize,
		       a2dr_s,
		       a2dr_sminusb,
		       a2dr_ctfsq,
		       a2dr_d,
		       a2dr_b
		       );
      
      xfac = (int)(512.0 / (float)fftSize);

      maxS = minS = a2dr_s[2];
      for(j=2;j<256;j++)
	{/* Zoom it normal */
	  thisPixel = j;
	  minS= MIN(a2dr_s[thisPixel],minS);
	  maxS= MAX(a2dr_s[thisPixel],maxS);
	}/* Zoom it normal */
      if(maxS != minS)
	scaleS = (float)((float)((float)505.0) / 
				  (float)(maxS-minS));
      
      if(a2drSDraw)
	{/* s */
	  if(a2dr_s)
	    {
	      maxCA = maxS;
	      minCA = minS;
	      scaleCircFactor = scaleS;
	      // printf("a2dr_s min[%f]max[%f]\n",minCA,maxCA);

	      for(i=256;i<511;i++)
		{/* DO all */
		  scaledY = (a2dr_s[(i-256)]-minCA)*scaleCircFactor;
		  scaledYp1 = (a2dr_s[(i-256)+1]-minCA)*scaleCircFactor;
		  scaledY = (510.0 - scaledY);
		  scaledYp1 = (510.0 - scaledYp1);
		  thisX= (i-256)*xfac;
		  thisXp1= ((i-256)+1)*xfac;
		  thisX += 256;
		  thisXp1 += 256;
		  if(i==256)
		    scaledY = 256;
		  thisX = DRAWBLANK(thisX);
		  thisXp1= DRAWBLANK(thisXp1);
		  XDrawLine(dsply, XtWindow(fftDrawingWidget), 
			    image_gcRED,
			    CNVPTX(thisX),(int)scaledY,
			    CNVPTX(thisXp1),(int)scaledYp1);
		}/* DO all */
	    }
	}/* s */
      if(a2drSminusBDraw)
	{/* sminusb */
	  if(a2dr_sminusb)
	    {
	      maxCA = minCA = a2dr_sminusb[2];
	      for(i=2;i<256;i++)
		{/* Do all */
		  minCA = MIN(a2dr_sminusb[i],minCA);
		  maxCA = MAX(a2dr_sminusb[i],maxCA);
		}/* Do all */
	      //  printf("a2dr_sminusb min[%f]max[%f]\n",minCA,maxCA);

	      if(maxCA != minCA)
		{/* OK to do */
		  scaleCircFactor = 250.0 / ((float)(maxCA-minCA)); 
		  
		  for(i=256;i<511;i++)
		    {/* DO all */
		      scaledY = (a2dr_sminusb[(i-256)]-minCA)*
			scaleCircFactor;
		      scaledYp1 = (a2dr_sminusb[(i-256)+1]-minCA)*
			scaleCircFactor;
		      scaledY = (255.0 - scaledY);
		      scaledYp1 = (255.0 - scaledYp1);
		      thisX= (i-256)*xfac;
		      thisXp1= ((i-256)+1)*xfac;
		      thisX += 256;
		      thisXp1 += 256;
		      if(i==256)
			scaledY = 256;
		      thisX = DRAWBLANK(thisX);
		      thisXp1= DRAWBLANK(thisXp1);
		      XDrawLine(dsply, XtWindow(fftDrawingWidget), 
				image_gcBLUE,
				CNVPTX(thisX),(int)scaledY,
				CNVPTX(thisXp1),(int)scaledYp1);
		    }/* DO all */
		}/* OK to do */
	    }
	}/* sminusb */
      if(a2drCtfsqDraw)
	{/* ctfsq */
	  if(a2dr_ctfsq)
	    {
	      maxCA = minCA = a2dr_ctfsq[2];
	      for(i=2;i<256;i++)
		{/* Do all */
		  minCA = MIN(a2dr_ctfsq[i],minCA);
		  maxCA = MAX(a2dr_ctfsq[i],maxCA);
		}/* Do all */
	      //   printf("a2dr_ctfsq min[%f]max[%f]\n",minCA,maxCA);

	      if(maxCA != minCA)
		{/* OK to do */
		  scaleCircFactor = 250.0 / ((float)(maxCA-minCA)); 
		  
		  for(i=256;i<511;i++)
		    {/* DO all */
		      scaledY = (a2dr_ctfsq[(i-256)]-minCA)*
			scaleCircFactor;
		      scaledYp1 = (a2dr_ctfsq[(i-256)+1]-minCA)*
			scaleCircFactor;
		      scaledY = (255.0 - scaledY);
		      scaledYp1 = (255.0 - scaledYp1);
		      thisX= (i-256)*xfac;
		      thisXp1= ((i-256)+1)*xfac;
		      thisX += 256;
		      thisXp1 += 256;
		      if(i==256)
			scaledY = 256;
		      thisX = DRAWBLANK(thisX);
		      thisXp1= DRAWBLANK(thisXp1);
		      XDrawLine(dsply, XtWindow(fftDrawingWidget), 
				image_gcGREEN,
				CNVPTX(thisX),(int)scaledY,
				CNVPTX(thisXp1),(int)scaledYp1);
		    }/* DO all */
		}/* OK to do */
	    }
	}/* ctfsq */
      if(a2drDDraw)
	{/* d */
	  if(a2dr_d)
	    {
	      maxCA = minCA = a2dr_d[2];
	      for(i=2;i<256;i++)
		{/* Do all */
		  minCA = MIN(a2dr_d[i],minCA);
		  maxCA = MAX(a2dr_d[i],maxCA);
		}/* Do all */
	      // printf("a2dr_d min[%f]max[%f]\n",minCA,maxCA);

	      if(maxCA != minCA)
		{/* OK to do */
		  scaleCircFactor = 250.0 / ((float)(maxCA-minCA)); 
		  for(i=256;i<511;i++)
		    {/* DO all */
		      scaledY = (a2dr_d[(i-256)]-minCA)*scaleCircFactor;
		      scaledYp1 = (a2dr_d[(i-256)+1]-minCA)*scaleCircFactor;
		      scaledY = (255.0 - scaledY);
		      scaledYp1 = (255.0 - scaledYp1);
		      thisX= (i-256)*xfac;
		      thisXp1= ((i-256)+1)*xfac;
		      thisX += 256;
		      thisXp1 += 256;
		      if(i==256)
			scaledY = 256;
		      thisX = DRAWBLANK(thisX);
		      thisXp1= DRAWBLANK(thisXp1);
		      XDrawLine(dsply, XtWindow(fftDrawingWidget), 
				image_gcWHITE,
				CNVPTX(thisX),(int)scaledY,
				CNVPTX(thisXp1),(int)scaledYp1);
		    }/* DO all */
		}/* OK to do */
	    }
	}/* d */
      if(a2drBDraw)
	{/* b */
	  if(a2dr_b)
	    {
	      maxCA = minCA = a2dr_b[2];
	      for(i=2;i<256;i++)
		{/* Do all */
		  minCA = MIN(a2dr_b[i],minCA);
		  maxCA = MAX(a2dr_b[i],maxCA);
		}/* Do all */
	      // printf("a2dr_b min[%f]max[%f]\n",minCA,maxCA);
	      /* Just use S scaling */
	      maxCA = maxS;
	      minCA = minS;
	      scaleCircFactor = scaleS;

	      if(maxCA != minCA)
		{/* OK to do */
		  scaleCircFactor = 250.0 / ((float)(maxCA-minCA)); 
		  for(i=256;i<511;i++)
		    {/* DO all */
		      scaledY = (a2dr_b[(i-256)]-minCA)*scaleCircFactor;
		      scaledYp1 = (a2dr_b[(i-256)+1]-minCA)*scaleCircFactor;
		      scaledY = (510.0 - scaledY);
		      scaledYp1 = (510.0 - scaledYp1);
		      thisX= (i-256)*xfac;
		      thisXp1= ((i-256)+1)*xfac;
		      thisX += 256;
		      thisXp1 += 256;
		      if(i==256)
			scaledY = 256;
		      thisX = DRAWBLANK(thisX);
		      thisXp1= DRAWBLANK(thisXp1);
		      XDrawLine(dsply, XtWindow(fftDrawingWidget), 
				image_gcBLUE,
				CNVPTX(thisX),(int)scaledY,
				CNVPTX(thisXp1),(int)scaledYp1);
		    }/* DO all */
		}/* OK to do */
	    }
	}/* b */
      
    }/* Get curves */
  
 bugOutFFT:;

}/* DrawFftWindow */



/* ************************************************************
 * latText_create() - Remember the latText text widgets on the
 * FFT screen.
 */
void latText_create(Widget widget, int *tag, 
		    XmAnyCallbackStruct *callback_data)
{/* latText_CREATE */
  
  if(*tag==LAT1X)
    lat1XtextW= widget;
  else if(*tag==LAT1Y)
    lat1YtextW= widget;
  else if(*tag==LAT2X)
    lat2XtextW= widget;
  else if(*tag==LAT2Y)
    lat2YtextW= widget;
  else if(*tag==PICK2POINTS)
    pick2pointsLabelW= widget;
  else if(*tag==MILLERIND)
    millerLabelW= widget;
  else if(*tag==LATLAB1)
    lat1LabelW= widget;
  else if(*tag==LATLAB2)
    lat2LabelW= widget;
  else if(*tag==RASTERSTEP)
    rasterStepW= widget;
  else if(*tag==FFTIMGNUM)
    {/* Img num */
      fftImgNumW= widget;
      handle_cr(fftImgNumW);
    }/* Img num */
  else if(*tag==ROTFLTIMGNUM)
    {/* Img num */
      rotfltImgNumW= widget;
      handle_cr(rotfltImgNumW);
    }/* Img num */
  else if(*tag==CTFCURVENUM)
    {/* Img num */
      ctfCurveTextW= widget;
      handle_cr(ctfCurveTextW);
    }/* Img num */
  else if(*tag==SCANSTEP)
    {ctfScanStepW= widget;
     handle_cr(widget);}
  else if(*tag==PIXSIZE)
    {ctfPixSizeW = widget;
     handle_cr(widget);}
  else if(*tag==MAG)
    {ctfMagW= widget;
     handle_cr(widget);}
  else if(*tag==KV)
    ctfKvW = widget;
  else if(*tag==AMPFACTOR)
    ctfAmpFactorW= widget;
  else if(*tag==WAVE)
    ctfWaveW= widget;
  else if(*tag==SPABER)
    ctfSpAberW= widget;
  else if(*tag==MAJORDEFOCUS)
    {
      ctfMajorDefocusW= widget;
      XtUnmanageChild(ctfMajorDefocusW);  /* Not used except to pass values */
    }
  else if(*tag==MINORDEFOCUS)
    {
      ctfMinorDefocusW= widget;
      XtUnmanageChild(ctfMinorDefocusW);  /* Not used except to pass values */
    }
  else if(*tag==CTFHELPLABEL)
    ctfHelpLabelW= widget;
  else if(*tag==CTFHELPLABEL2)
    ctfHelpLabel2W= widget;
  else if(*tag==WIENER)
    ctfWienerW= widget;
  else if(*tag==ANGMAJOR)
    {
      ctfAngMajorW= widget;
      XtUnmanageChild(ctfAngMajorW);  /* Not used except to pass values */
    }
  else if(*tag==TFACTOR)
    ctfTFactorW= widget;
  else if(*tag==MODE)
    ctfModeW= widget;
  else if(*tag==FFTMEASPIXL)
    fftMeasPixW= widget;
  else if(*tag==FFTMEASANGL)
    fftMeasAngW= widget;
  else if(*tag==SHIFTORIGX)
    shiftOrigXW= widget;
  else if(*tag==SHIFTORIGY)
    shiftOrigYW= widget;
  else if(*tag==MOVEORIGX)
    moveOrigXW= widget;
  else if(*tag==MOVEORIGY)
    moveOrigYW= widget;
  else if(*tag==CTFMEASURE)
    ctfMeasureLabelW= widget;
  else if(*tag==CTFMEASURE2)
    ctfMeasureLabel2W= widget;
  else if(*tag==LOP)
    {lopTextW= widget;
     handle_cr(lopTextW);}
  else if(*tag==HIP)
    {hipTextW= widget;
     handle_cr(hipTextW);}
  else if(*tag==LOPRADIUS)
    {lopRadiusTextW= widget;
     handle_cr(lopRadiusTextW);}
  else if(*tag==HIPRADIUS)
    {hipRadiusTextW= widget;
     handle_cr(hipRadiusTextW);}
  else if(*tag==FFTMEASURE)
    fftMeasurePSTextW= widget;

  
}/* latText_CREATE */




/* ************************************************************
 * latText_changed() - Whenever lattice text widgets changed
 * call here.
 */
void latText_changed(Widget widget, int *tag, 
		     XmAnyCallbackStruct *callback_data)
{/* latText_CHANGED */
  char *latText;
  int i;
  XmString simpleString; 

  latText = XmTextGetString(widget); 

  if(*tag==RASTERSTEP)
    {/* Get new raster step size */
      rasterStep = (float)strtod(latText,0);
    }/* Get new raster step size */
  else if(*tag==FARADIUS)
    {/* Get new fourAver Radius */
      fourAverRadius = (float)strtod(latText,0);
    }/* Get new fourAver Radius */
  else if(*tag==FACUT)
    {/* Get new fourAver Cut */
      fourAverCut = (float)strtod(latText,0);
    }/* Get new fourAver Cut */
  else 
    if(*tag==SCANSTEP)
    {/* Get scan step */
      scanStep = (float)strtod(latText,0);
      if(scanStep > 0.0)
	{/* ok to calc */
	  sprintf(outStr,"%.4f",((scanStep*1000.0)/magnification));
	  createSimpleString(&simpleString,outStr);
	  XtSetArg(arglst[0], XmNvalue, outStr );
	  XtSetArg(arglst[1], XmNrecomputeSize, FALSE );
	  XtSetValues(ctfPixSizeW, arglst, 2); 
	  XFlush(dsply);
	  XmStringFree(simpleString); 
	  XFlush(dsply);
	}/* ok to calc */
    }/* Get scan step */
  else if(*tag==PIXSIZE)
    {/* Get pixsize */
      pixSize = (float)strtod(latText,0);
      fftAngsPerPixel = ((float)strtod(latText,0)*10.0);
      if(pixSize > 0.0)
	{/* ok to calc */
	  sprintf(outStr,"%.3f",fftAngsPerPixel);
	  XmTextSetString(fftMeasurePSTextW,outStr);
	  sprintf(outStr,"%.3f",((scanStep*1000.0)/pixSize));
	  createSimpleString(&simpleString,outStr);
	  XtSetArg(arglst[0], XmNvalue, outStr );
	  XtSetArg(arglst[1], XmNrecomputeSize, FALSE );
	  XtSetValues(ctfMagW, arglst, 2); 
	  XFlush(dsply);
	  XmStringFree(simpleString); 
	  XFlush(dsply);
	}/* ok to calc */
    }/* Get pixsize */
  else if(*tag==MAG)
    {/* Get mag */
      magnification = (float)strtod(latText,0);
      if(magnification > 0.0)
	{/* ok to calc */
	  sprintf(outStr,"%.4f",((scanStep*1000.0)/magnification));
	  createSimpleString(&simpleString,outStr);
	  XtSetArg(arglst[0], XmNvalue, outStr );
	  XtSetArg(arglst[1], XmNrecomputeSize, FALSE );
	  XtSetValues(ctfPixSizeW, arglst, 2); 
	  XFlush(dsply);
	  XmStringFree(simpleString); 
	  XFlush(dsply);
	}/* ok to calc */
    }/* Get mag */
  else if(*tag==KV)
    {/* Get kv */
      float volts,calcWavelength;
      kv = (float)strtod(latText,0);
      volts = kv * 1000.0;
      if(volts > 0.0)
	{/* ok to calc */
	  calcWavelength = 1.23 / sqrt ((volts + (1.0e-6 * (volts*volts))));
	  sprintf(outStr,"%f",(calcWavelength));
	  XmTextSetString(ctfWaveW,outStr);	
	  XFlush(dsply);
	}/* ok to calc */
    }/* Get kv */
  else if(*tag==WIENER)
    {/* Get  */
      wiener = (float)strtod(latText,0);
    }/* Get  */
  else if(*tag==ANGMAJOR)
    {/* Get  */
      angMajor = (float)strtod(latText,0);
    }/* Get  */
  else if(*tag==TFACTOR)
    {/* Get  */
      ctfTFactor = (float)strtod(latText,0);
    }/* Get  */
  else if(*tag==MODE) 
   {/* Get  */
      ctfDivideMode = (int)strtod(latText,0);
    }/* Get  */
  else if(*tag==WAVE)
    {/* Get wave */
      wavelength = (float)strtod(latText,0);
    }/* Get wave */
  else if(*tag==SPABER)
    {/* Get spAber */
      spAberration = (float)strtod(latText,0);
    }/* Get spAber */
  else if(*tag==FFTMEASURE)
    {/* Get fftAngs */
      fftAngsPerPixel = (float)strtod(latText,0);
    }/* Get fftAngs */
  else if(*tag==FFTMEASLN)
    {/* Get fftAngs */
      fftMeasLineNum = (int)strtod(latText,0);
    }/* Get fftAngs */
  else if(*tag==CTFCURVENUM)
    {/* Get ctf curve num */
      ctfCurveNum = (int)strtod(latText,0);
      ctfCurveNum--;   /* Human numbers from 1 not 0 */
      if(ctfCurveNum >= MAXCTFS)
	ctfCurveNum = MAXCTFS-1;
      if(ctfCurveNum < 0)
	ctfCurveNum = 0;
    }/* Get ctf curve num */

  if(*tag != FFTMEASURE && *tag != FFTMEASLN && *tag != CTFCURVENUM)
    {/* DOn't do for measurement */
      if(layerLines)
	{/* Get lattice line order */
	  latLineOrder = (int)strtod(latText, 0);
	}/* Get lattice line order */
      else 
	{/* 2D lattice */
	  if(*tag==LAT1X)
	    {/* */ 
	      latPt[0].x = (int)strtod(latText, 0);
	      /* Remeber for when we need it */
	      latLineOrder = (int)strtod(latText, 0); 
	    }/* */
	  else if(*tag==LAT1Y)
	    {/* */
	      latPt[0].y = (int)strtod(latText, 0);
	    }/* */
	  else if(*tag==LAT2X)
	    {/* */
	      latPt[1].x = (int)strtod(latText, 0);
	    }/* */
	  else if(*tag==LAT2Y)
	    {/* */
	      latPt[1].y = (int)strtod(latText, 0);
	    }/* */
	}/* 2D lattice */
    }/* DOn't do for measurement */
  
  XtFree(latText);
  
}/* latText_CHANGED */



/* *****************************************************************
 * ctfText_changed() - Whenever any of the ctf value widgets change
 * call here.
 */
void ctfText_changed(Widget widget, int *tag, 
		     XmAnyCallbackStruct *callback_data)
{/* CTFTEXT_CHANGED */
  char *ctfText;
  int i;

  ctfText = XmTextGetString(widget); 

  if(*tag==CTF_FOCUSMIN_VALUE)
    {/* Get new ab ratio */
      ctfFocusMinvalue = (float)strtod(ctfText,0);
      calcNodes();
/*      calcTFpixel(ctffocusMajValue, ctfFocusMinvalue); Also funky!!! */
/* Don't set the scrollbar it gives very funky results!!!  */
      XtSetArg(arglst[0], XmNvalue, (int)(ctfFocusMinvalue*100.0));
      if(!changeCTFtext)
	XtSetValues(ctfFocusMinscrollW, arglst, 1);  
      sprintf(outStr,"%f",ctfFocusMinvalue);
      XmTextSetString(ctfMinorDefocusW,outStr);
    }/* Get new ab ratio */
  else if(*tag==CTF_ANGMAJOR_VALUE)
    {/* Get omega */
      ctfAngMajorValue = (float)strtod(ctfText,0);
/* Don't set the scrollbar it gives very funky results!!! */
      XtSetArg(arglst[0], XmNvalue, (int)(ctfAngMajorValue*100.0));
      if(!changeCTFtext)
	XtSetValues(ctfAngMajorScrollW, arglst, 1); 
      ctfAngMajorValue = (float)deg2rad(ctfAngMajorValue);
      sprintf(outStr,"%.2f",rad2deg(ctfAngMajorValue));
      XmTextSetString(ctfAngMajorW, outStr); 
    }/* Get omega */
  else if(*tag==CTF_DEFOCUS_VALUE)
    {/* Get defocus */
      ctffocusMajValue = (float)strtod(ctfText,0);
      calcNodes();
/* Also funky!!!      calcTFpixel(ctffocusMajValue, ctfFocusMinvalue); */
/*      calcMeanTFpixel(ctffocusMajValue); */
/* Don't set the scrollbar it gives very funky results!!! */
      XtSetArg(arglst[0], XmNvalue, (int)(ctffocusMajValue *
					  100.0));
      if(!changeCTFtext)
	XtSetValues(ctffocusMajScrollW, arglst, 1); 
      sprintf(outStr,"%f",ctffocusMajValue);
      XmTextSetString(ctfMajorDefocusW,outStr);
    }/* Get defocus */

  changeCTFtext = FALSE; /* We just did */

  XtFree(ctfText);
  
  XFlush(dsply);


}/* CTFTEXT_CHANGED */



/********************************************************
 * ctfValue_change() - When the ctf is changed call here
 * to recalc the CTF and then redisplay.
 */
void ctfValue_change(Widget widget, int *tag, 
		     XmAnyCallbackStruct *callback_data)
{/* CTFVALUE_CHANGE */

  calcFFT(FALSE, 0, FALSE);
  
  XSync(dsply,FALSE);
  XFlush(dsply);
  DrawFFTWindow();
  XSync(dsply,FALSE);
  XFlush(dsply);

}/* CTFVALUE_CHANGE */






/**********************************************************************
 * stuffContourBuffer() - Stuff the current concepts of x,y for lattice
 * points in zoomed up contour buffer
*/
void stuffContourBuffer()
{/* STUFFCONTOURBUFFER */
  int i,j;
  int thisX, thisY;
  char thisColor[16];
  XmString simpleString;

  if(currentLatPt < 2)
    {/* fiorst */
      /* contour = TRUE; */
      for(i=0;i<128;i++)
	{/* rows */
	  for(j=0;j<128;j++)
	    {/* Stuff all */

	      thisX= (latPt[currentLatPt].measuredX-8+(j/8));
	      thisY= (latPt[currentLatPt].measuredY-8+(i/8));
	      if(hiColor)
		{
		  fft_contour_buffer[((i*128)+j)]= 
		    fft_gold_buffer1[(((latPt[currentLatPt].measuredY-8+
					(i/8))*
				       512)+
				      (latPt[currentLatPt].measuredX-8+
				       (j/8)))];
		  XPutPixel(fftContour_image1,
			    j, i,
			    pixels_return[fft_contour_buffer[((i*128)+j)]]);
		}
	      else
		{
		  fft_contour_buffer[((i*128)+j)]= 
		    fft_gold_buffer1[(((latPt[currentLatPt].measuredY-8+
					(i/8))*
				       512)+
				      (latPt[currentLatPt].measuredX-8+
				       (j/8)))];
		}
	    }/* Stuff all */
	}/* rows */
      
      /* Now draw the contour plot in the window */
      /* USe contur_a_rob() fortran routine */
      /* drawFFTcontour(); Done in DrawFFTWindow() */

      if(currentLatPt==0)
	{/* green */
	  XtSetArg(arglst[0], XmNborderColor, pixels_return[GREEN]);
	}/* green */
      else 
	{/* RED */
	  XtSetArg(arglst[0], XmNborderColor, pixels_return[RED]);
	}/* RED */

      XtSetValues(contourWidget, arglst, 1); 
      XFlush(dsply);
      
      if(layerLines)
	sprintf(outStr,"[%1d] xy[%d,%d]",
		currentLatPt+1,
		latPt[currentLatPt].measuredX,
		latPt[currentLatPt].measuredY
		);    
      else if(drawCTF)
	sprintf(outStr,"[%1d][%d,%d] xy[%d,%d]",
		currentLatPt+1,
		latPt[currentLatPt].x,
		latPt[currentLatPt].y,
		latPt[currentLatPt].measuredX-256,
		256-latPt[currentLatPt].measuredY
	      );    
      else sprintf(outStr,"[%1d][%d,%d] xy[%d,%d]",
	      currentLatPt+1,
	      latPt[currentLatPt].x,
	      latPt[currentLatPt].y,
	      latPt[currentLatPt].measuredX,
	      latPt[currentLatPt].measuredY
	      );    
      createSimpleString(&simpleString,outStr);
      XtSetArg(arglst[0], XmNlabelString, simpleString);
      /* OLD XtSetValues(contLabelWidget, arglst, 1);  */
      XtSetValues(fftFeedbackLabel2W, arglst, 1); 
      XFlush(dsply);
      XmStringFree(simpleString); 
      sprintf(outStr,"");
      createSimpleString(&simpleString,outStr);
      XtSetArg(arglst[0], XmNlabelString, simpleString);
      /* OLD       XtSetValues(fftFeedbackLabelW, arglst, 1);  */
      XtSetValues(fftFeedbackLabel1W, arglst, 1); 
      XFlush(dsply);
      XmStringFree(simpleString); 
      sprintf(outStr,"");
      createSimpleString(&simpleString,outStr);
      XtSetArg(arglst[0], XmNlabelString, simpleString);
      XtSetValues(fftFeedbackLabel3W, arglst, 1); 
      XFlush(dsply);
      XmStringFree(simpleString); 
    }/* fiorst */
  else
    {/* none */
      /* contour = FALSE; */
      for(i=0;i<128;i++)
	{/* rows */
	  for(j=0;j<128;j++)
	    {/* Stuff all */
	      if(hiColor)
		{
		  XPutPixel(fftContour_image1,
			    j, i,
			    pixels_return[BLACK]);
		  fft_contour_buffer[((i*128)+j)]= BLACK;
		}
	      else
		fft_contour_buffer[((i*128)+j)]= pixels_return[BLACK];
	    }/* Stuff all */
	}/* rows */
    }/* none */


}/* STUFFCONTOURBUFFER */




/****************************************************************
 * findLatPt(x,y) - Which lattice point are they wanting?
*/
void findLatPt(int x, int y)
{/* FINDLATPT */
  int i,j,dist1, dist2;

  dist1= ((latPt[0].measuredX - x)*(latPt[0].measuredX - x)) +
         ((latPt[0].measuredY - y)*(latPt[0].measuredY - y)) ;

  dist2= ((latPt[1].measuredX - x)*(latPt[1].measuredX - x)) +
         ((latPt[1].measuredY - y)*(latPt[1].measuredY - y)) ;

  if(dist2 > dist1)
    currentLatPt = 0;
  else 
    currentLatPt = 1;


}/* FINDLATPT */



/************************************************************
 * convertFFTcoords() - Converts from x,y positions from the
 * X Windows window (e.g. gotten from fftMotion()) into
 * indexes in the fft array.
*/
void convertFFTcoords(int x, int y, int *thisX, int *thisY)
{/* CONVERTFFTCOORDS */

  if(y<=256 && x <256)
    {/* upper left */
      *thisY = (256-y);  /* Goes to lower right */
      *thisX = (256+x);
    }/* upper left */
  else if(y<=256 && x >=256)
    {/* upper right */
      *thisY = (256-y);  /* Goes to lower left */
      *thisX = (x-256);
    }/* upper right */
  else if(y>256 && x < 256)
    {/* lower left */
      *thisY = (512 - (y-256));   /* Goes to upper right */
      *thisX = (256 + x);
    }/* lower left */
  else
    {/* lower right */
      *thisY = (512 - (y-256));    /* Goes to upper left */
      *thisX = (x-256);
    }/* lower right */

}/* CONVERTFFTCOORDS */




/*************************************************************
 * fftMotion() - Whenever the pointer moves across our FFT
 * window call here.  This is where we draw the zoomed up
 * display in the contour window.
*/
void fftMotion(Widget widget, caddr_t unused, 
	       register XMotionEvent *pEvent)
{/* fftMotion */
  float thisRadius,sizeF,apart,bpart,amp,phase,thisAngs;
  XmString simpleString;
  int iTfftSize2;
  int x, y,lastDraw;
  int i, j, k, thisX,thisY;
  static int lastX=-99;
  static int lastY=-99;
#define CNVPTX(x) ((fftZoomFactor==2) ? ((x-fftZoomX) * 2) : (x))

  /* Make all changes current */
  XFlush(dsply);               

  switch (pEvent->type) 
    {/* Find out which event we have */
	  
    case MotionNotify:
	      
      /* Get coords of pointer */
      x = (int)(pEvent->x);     
      y = (int)(pEvent->y);

      if(pickLattice == -99)
	{/* DOn't do if they're trying to pick points */
	  
	  sizeF = fftSize / 512.0;
	  /* ORIG way   thisX = (int)(((float)x-256.0)); */
	  thisX = ((x-256)/fftZoomFactor);
	  if(thisX < 0 && fftSize < 512)
	    thisX -= (int)(512.0/fftSize)-1;
	  /* ORIG way   thisY = (int)(((float)255.0-(float)y)); */
	  thisY = (int)((float)(255-y)/fftZoomFactor);
	  if(thisY < 0 && fftSize < 512)
	    thisY -= (int)(512/fftSize)-2;
	  thisX = MIN(MAX(thisX,-256),255);
	  thisY = MIN(MAX(thisY,-256),255);
	  thisX = (int)((float)thisX * sizeF);
	  thisY = (int)((float)thisY * sizeF);

	  if(fftImage_buffer2)
	    {/* OK to do */
	      fft_point_(fftImage_buffer2,
			 &fftSize,&fftSize,
			 &thisX,&thisY,&thisRadius,&apart,&bpart);
	    }/* OK to do */
	  if(x != 256)
	    sprintf(outStr,"x[%d]y[%d] radius[%.0f] Angs[%.2f]",
		    thisX,thisY,thisRadius,
		    ((512.0/((float)(x-256.0)/(float)fftZoomFactor))*
		     fftAngsPerPixel));
	      else
		sprintf(outStr,"x[%d]y[%d] radius[%.0f] Angs[undef]",
			thisX,thisY,thisRadius);
	  createSimpleString(&simpleString,outStr);
	  XtSetArg(arglst[0], XmNlabelString, simpleString);
	  /* OLD XtSetValues(fftFeedbackLabelW, arglst, 1);  */
	  XtSetValues(fftFeedbackLabel1W, arglst, 1); 
	  XFlush(dsply);
	  XmStringFree(simpleString); 
	  sprintf(outStr,"a[%.2E] b[%.2E]",
		  apart,bpart);
	  createSimpleString(&simpleString,outStr);
	  XtSetArg(arglst[0], XmNlabelString, simpleString);
	  /* OLD XtSetValues(contLabelWidget, arglst, 1);  */
	  XtSetValues(fftFeedbackLabel2W, arglst, 1); 
	  XFlush(dsply);
	  XmStringFree(simpleString); 
	  convert_abpart_(&apart,&bpart,&amp,&phase);
	  sprintf(outStr,"amp[%.2E] phase[%.2E]",
		  amp,phase);
	  createSimpleString(&simpleString,outStr);
	  XtSetArg(arglst[0], XmNlabelString, simpleString);
	  XtSetValues(fftFeedbackLabel3W, arglst, 1); 
	  XFlush(dsply);
	  XmStringFree(simpleString); 

	  if(fftZoomFactor==2)
	    {/* Adjust for zoom */
	      x= fftZoomX + (x / 2);
	      y= fftZoomY + (y / 2);
	    }/* Adjust for zoom */
	  for(i=0;i<128;i++)
	    {/* rows */
	      int thisX,thisY;
	      for(j=0;j<128;j++)
		{/* Stuff all */
		  thisX= (x-8+(j/8));
		  thisY= (y-8+(i/8));
		  if(thisX < 512 && thisY < 512 &&
		     thisX >= 0 && thisY >= 0)
		    {
		      fft_contour_buffer[((i*128)+j)]= 
			fft_gold_buffer1[((thisY*512)+thisX)];
		      if(hiColor)
			{
			  XPutPixel(fftContour_image1,
				    j, i,
				    pixels_return[fft_contour_buffer[((i*128)
								      +j)]]);
			}
		    }
		  else
		    {
		      if(hiColor)
			{
			  fft_contour_buffer[((i*128)+j)]= BLACK;
			  XPutPixel(fftContour_image1,
				    j, i,
				    pixels_return[fft_contour_buffer[BLACK]]);
			}
		      else
			fft_contour_buffer[((i*128)+j)]= pixels_return[BLACK];
		    }
		}/* Stuff all */
	    }/* rows */
	  

	  if(XtIsManaged(fftscreen))
	    {/* Draw the contour window */
	      XPutImage(dsply, fftContour_window1, fftContour_gc, 
			fftContour_image1, 
			0, 0, 0, 0,  128,  128);
	      XDrawLine(dsply, fftContour_window1, 
			image_gcBLACK,
			64,60,64,68);
	      XDrawLine(dsply, fftContour_window1, 
			image_gcBLACK,
			60,64,68,64);
	    }/* Draw the contour window */
	  XFlush (dsply);
	}/* DOn't do if they're trying to pick points */
      
      if(doingAutoDefocusRefine || calcFFT1DcircAvg)
	{/* Give feedback */
	  XmString simpleString;
	  int distX;

	  if(doingAutoDefocusRefine && adrBlankDraw)
	    sprintf(outStr,"Radius = [%d] Pixels.",(x/2));
	  else
	    sprintf(outStr,"Radius = [%d] Pixels.",(x-256));
	  createSimpleString(&simpleString,outStr);
	  XtSetArg(arglst[0], XmNlabelString, simpleString);
	  XtSetValues(fftMeasPixW, arglst, 1); 
	  XFlush(dsply);
	  XmStringFree(simpleString); 

	  /* FIrst calc dist on X axis only */
	  if(doingAutoDefocusRefine && adrBlankDraw)
	    {/* special */
	      if(x > 1)
		sprintf(outStr,"[%.2f] Angs",
			((512.0/((float)(x/2)))*
			 fftAngsPerPixel));
	      else
		sprintf(outStr,"[undef] Angs");
	    }/* special */
	  else
	    {/* normal */
	      if(x != 256)
		sprintf(outStr,"[%.2f] Angs",
			((512.0/((float)(x-256.0)))*
			 fftAngsPerPixel));
	      else
		sprintf(outStr,"[undef] Angs");
	    }/* normal */
	    
	  createSimpleString(&simpleString,outStr);
	  XtSetArg(arglst[0], XmNlabelString, simpleString);
	  XtSetArg(arglst[1], XmNforeground, pixels_return[RED] );
	  XtSetValues(fftMeasAngW, arglst, 2); 
	  XtSetValues(ctfMeasureLabelW, arglst, 2); 
	  XFlush(dsply);
	  XmStringFree(simpleString); 

	  /* Now calc dist based on pointer */
	  distX = (int)sqrt((float)(
				    (x-256.0)*(x-256.0) +
				    (y-256.0)*(y-256.0)
				    ));
	  if(doingAutoDefocusRefine && adrBlankDraw)
	    {/* special */
	      if(distX > 1)
		sprintf(outStr,"[%.2f] Angs",
			((512.0/((float)(distX/2)))*
			 fftAngsPerPixel));
	      else
		sprintf(outStr,"[undef] Angs");
	    }/* special */
	  else
	    {/* normal */
	      if(distX != 256)
		sprintf(outStr,"[%.2f] Angs",
			((512.0/((float)(distX)))*
			 fftAngsPerPixel));
	      else
		sprintf(outStr,"[undef] Angs");
	    }/* normal */
	    
	  createSimpleString(&simpleString,outStr);
	  XtSetArg(arglst[0], XmNlabelString, simpleString);
	  XtSetArg(arglst[1], XmNforeground, pixels_return[BLUE] );
	  XtSetValues(ctfMeasureLabel2W, arglst, 2); 
	  XFlush(dsply);
	  XmStringFree(simpleString); 

	  /* Don't do it 	  x = CNVPTX(x);  */
	  if(lastX >=0 && lastY >= 0)
	    {/* redraw last one */
	      if(fftMotionFeedback)
		{
		  XDrawLine(dsply, XtWindow(fftDrawingWidget), 
			    image_gcBLACK,
			    lastX,512,lastX,505);
		  XDrawLine(dsply, XtWindow(fftDrawingWidget), 
			    image_gcBLACK,
			    lastX,255,lastX,256);
		}
	    }/* redraw last one */

	  /* Now put thisX in drawable coords */
	  x = (int)(pEvent->x);     
	  x = MAX(256,x);
	  sizeF = 1.0;
	  thisX = 256 + (int)((float)(x-256.0) * sizeF);
	  if(fftMotionFeedback)
	    {
	      XDrawLine(dsply, XtWindow(fftDrawingWidget), 
			image_gcRED,
			thisX,512,thisX,505);
	      XDrawLine(dsply, XtWindow(fftDrawingWidget), 
			image_gcRED,
			thisX,255,thisX,256);
	    }
	  lastX = thisX;
	  lastY = y;

	  /* Now draw where the pointer has been */
	  x = (int)(pEvent->x);     
	  y = (int)(pEvent->y);     
	  if(fftMotionFeedback)
	    XDrawLine(dsply, XtWindow(fftDrawingWidget), 
		      image_gcBLUE,
		      x,y,x,y-1);
	  
	}/* Give feedback */

      break;
	  
    default:
      break;
    }/* Find out which event we have */
  
  XFlush(dsply);              /* Make changes current */


}/* fftMotion */



/******************************************************************
 * fftMeasureDraw()- Draw the things needed for the FFT Measurement
 * stuff
*/
void fftMeasureDraw(int button)
{/* FFTMEASUREDRAW */
  int thisX,thisY;
  XmString simpleString;
#define CNVPTX(x) ((fftZoomFactor==2) ? ((x-fftZoomX) * 2) : (x))
#define CNVPTY(y) ((fftZoomFactor==2) ? ((y-fftZoomY) * 2) : (y))

  /* Refresh the FFT in window */
  DrawFFTWindow();

  /* Now draw all points */
  thisX = fftMeasures[0].x;
  thisY = fftMeasures[0].y;
  XDrawPoint(dsply,XtWindow(fftDrawingWidget),image_gcRED,
    thisX,thisY);
  drawCross(CNVPTX(thisX),CNVPTY(thisY),
	    CNVPTX(thisX+2),CNVPTY(thisY+2),
	    image_gcRED);
  thisX = fftMeasures[1].x;
  thisY = fftMeasures[1].y;
  XDrawPoint(dsply,XtWindow(fftDrawingWidget),image_gcBLUE,
	     thisX,thisY);
  drawCross(CNVPTX(thisX),CNVPTY(thisY),
	    CNVPTX(thisX+2),CNVPTY(thisY+2),
	    image_gcBLUE);

  if(TRUE)
    {/* Now draw Line */
      int ix1,ix2,iy1,iy2;
      float lngth;

      /* Zoom it for drawing */
      ix1= CNVPTX(fftMeasures[0].x);
      ix2= CNVPTX(fftMeasures[1].x);
      iy1= CNVPTY(fftMeasures[0].y);
      iy2= CNVPTY(fftMeasures[1].y);
      XDrawLine(dsply, XtWindow(fftDrawingWidget), 
		image_gcGREEN,
		ix1,iy1,ix2,iy2);
      /* DON'T Zoom it for measuring */
      ix1= fftMeasures[0].x;
      ix2= fftMeasures[1].x;
      iy1= fftMeasures[0].y;
      iy2= fftMeasures[1].y;
      lngth = (float)sqrt((double)
			  ((ix2-ix1)*
			   (ix2-ix1)) + 
			  ((iy2-iy1)*
			   (iy2-iy1)));
      
      sprintf(outStr,"[%.2f] Pixels.   xy[%d,%d] to xy[%d,%d]",
	      lngth*(float)((float)fftSize/512.0),
	      ix1,iy1,ix2,iy2
	      );
      createSimpleString(&simpleString,outStr);
      XtSetArg(arglst[0], XmNlabelString, simpleString);
      XtSetValues(fftMeasPixW, arglst, 1); 
      XFlush(dsply);
      XmStringFree(simpleString); 

      if(lngth==0.0)
	sprintf(outStr,"[n/a] Angstroms.");
      else
	sprintf(outStr,"[%f] Angstroms.",((512.0/lngth)*fftAngsPerPixel));
      createSimpleString(&simpleString,outStr);
      XtSetArg(arglst[0], XmNlabelString, simpleString);
      XtSetValues(fftMeasAngW, arglst, 1); 
      XFlush(dsply);
      XmStringFree(simpleString); 
      
    }/* Now draw Line */

}/* FFTMEASUREDRAW */




/*********************************************************
 * calcLattice()- Calc the lattice based on prg settings
 * determined by other callbacks.
*/
void calcLattice(int recalcAxBxAyBy)
{/* CALCLATTICE */
  float xcen,ycen;
  int nx,ny,ik,ix,iy,ih,ix1,ix2,iy1,iy2;
  int h1,k1,h2,k2,arg1;
  float x1,x2,y1,y2,farg1;
  
  xcen= 256.0;
  ycen= 256.0;
  
  /* Now compute distance from origin */
  h1= latPt[0].x;
  k1= latPt[0].y;
  h2= latPt[1].x;
  k2= latPt[1].y;
  
  x1= (float)latPt[0].measuredX;
  y1= (float)latPt[0].measuredY;
  x2= (float)latPt[1].measuredX;
  y2= (float)latPt[1].measuredY;
  
  /* Put it in TSB space */
  y1= 512.0 -(float)latPt[0].measuredY;
  y2= 512.0 -(float)latPt[1].measuredY;
  
  if(recalcAxBxAyBy)
    {/* Not recalced ablove */

      latcen2_(&h1,&k1,&x1,&y1,&h2,&k2,&x2,&y2,
	       &axL,&ayL,&bxL,&byL,&xcen,&ycen);

      /*
      printf("Values gotten from latcen2:\n\
	  ax[%f]ay[%f]bx[%f]by[%f]\n",
	     axL,ayL,bxL,byL);
	     */
      
      /* Let's always store anga angb gamstr 
       * and omega in degrees */
      anga= rad2deg(atan2((double)ayL,(double)axL));
      angb= rad2deg(atan2((double)byL,(double)bxL));
      
      /* Let's try this ... ??? 
	 anga = 0.0 - anga;
	 angb = 0.0 - angb;
	 */
      
      ang    = angb-anga;
      astar  = sqrt(axL*axL+ayL*ayL);
      bstar  = sqrt(bxL*bxL+byL*byL);
      gamstr = angb-anga;
      
    }/* Not recalced ablove */
  
  /* Stuff this for things like the crysDisp screen */
  map_abang = 180.0 - gamstr;
  if(bstar != 0.0 && astar != 0.0)
    aoverb = 1.0 / (astar / bstar);
  
  omega  = 180.0-gamstr;
  
  angaSave      = anga;
  angbSave      = angb;
  angSave       = ang;
  
  axLSave     = axL;
  ayLSave     = ayL;
  bxLSave     = bxL;	
  byLSave     = byL;
  
  astarSave = astar;
  bstarSave = bstar;
  gamstrSave = gamstr;
  omegaSave = omega;
  
  latScrollValue[0]= astar;
  latScrollValue[1]= bstar;
  latScrollValue[2]= gamstr;
  latScrollValue[3]= omega;
  
  latReset_activate(0,0,0);
  
}/* CALCLATTICE */




/* @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
 * fftClick() - Every time the mouse button is clicked call
 * here.  This routine handles the selecting of
 * lattice points for the FFT.
 *            Button1 is the Left mouse button
 *            Button2 is the Middle mouse button
 *            Button3 is the Right mouse button
 */
void fftClick(Widget widget, int *tag, 
	      XmDrawingAreaCallbackStruct *draw_data)
{/* FFTCLICK */
  XEvent event;
  int x, y, k, i,j,reDraw, thisOne, recalcAxBxAyBy;
  char ans;
  XmString simpleString;
  char outStr[16];

  /* Do appropriate depending on event type */
  if(draw_data != 0)
    {/* Only look at events meant for us */

      /* Turn this off */
      calcFFT1DcircAvg= FALSE;

      event= *(draw_data->event);

      switch (event.type) 
	{/* Find out which event we have */

	case ButtonRelease:

	  x = (int)(event.xbutton.x);      /* Get x,y copords of point pos */
	  y = (int)(event.xbutton.y);

	  break;

	  
	case ButtonPress:

	  /* They are trying to find points */
	  x = (int)(event.xbutton.x);      /* Get x,y copords of point pos */
	  y = (int)(event.xbutton.y);

	  recalcAxBxAyBy= TRUE;

	  /* Put coords back into "normal" space */
	  if(doingAutoDefocusRefine && adrBlankDraw)
	    {/* special */
	      x = ((x/2)+256);
	    }/* special */
	  else
	    if((widget!=contourWidget) && fftZoomFactor==2)
	      {/* Adjust for zoom */
		x= fftZoomX + (x / 2);
		y= fftZoomY + (y / 2);
	      }/* Adjust for zoom */

	  if(drawAuto2Dscurve)
	    {/* remember where they clicked */
	      scurveClick.x = x;
	      scurveClick.y = y;
	      DrawFFTWindow();
	    }/* remember where they clicked */

	  if(measureFFT)
	    {/* They are measuring the FFT here. */
	      if((event.xbutton.button == Button1))
		{/* Pick first point */
		  fftMeasures[0].x = x;
		  fftMeasures[0].y = y;
		  fftMeasureDraw(event.xbutton.button);
		}/* Pick first point */
	      if((event.xbutton.button == Button2))
		{/* Pick second point */
		  fftMeasures[1].x = x;
		  fftMeasures[1].y = y;
		  fftMeasureDraw(event.xbutton.button);
		}/* Pick second point */
	      if((event.xbutton.button == Button3))
		{/* measure */
		  fftMeasures[0].x = 256;
		  fftMeasures[0].y = 256;
		  fftMeasures[1].x = x;
		  fftMeasures[1].y = y;
		  fftMeasureDraw(event.xbutton.button);
		}/* measure */
	      XFlush(dsply);
	    }/* They are measuring the FFT here. */
	  else if(pickLattice != -99)
	    {/* OK to keep on trying */
	      latValChanged = FALSE;

	      if(widget==contourWidget)
		{/* They've selected from CONTOUR drawing area */
		  if(contour)
		    {/* Don't do if not intended */
		      if((event.xbutton.button == Button1))
			{/* Pick appropriate Lattice point */
			  if(pickLattice==1)
			    {/* Get first points */
			      latPt[0].measuredX += ((x/8) - 8);
			      latPt[0].measuredY += ((y/8) - 8)+1;
			      stuffContourBuffer();
			    }/* Get first points */
			  else if(pickLattice==2)
			    {/* Get second points */
			      latPt[1].measuredX += ((x/8) - 8);
			      latPt[1].measuredY += ((y/8) - 8)+1;
			      stuffContourBuffer();
			    }/* Get second points */
			  else if(pickedLatPts)
			    {/* Just find out which one we're dealing with */
			      /* findLatPt(x,y); */
			      latPt[currentLatPt].measuredX += ((x/8) - 8);
			      latPt[currentLatPt].measuredY += ((y/8) - 8)+1;
			      stuffContourBuffer();		   
			    }/* Just find out which one we're dealing with */
			}/* Pick appropriate Lattice point */
		    }/* Don't do if not intended */
		}/* They've selected from CONTOUR drawing area */
	      else
		{/* Main fft screen */
		  if((event.xbutton.button == Button1))
		    {/* Pick appropriate Lattice point */
		    }/* Pick appropriate Lattice point */
		  else if((event.xbutton.button == Button2))
		    {/* yongning */
		      int ncol,nrow;
		      if(fileType==PACKEDPORTABLE)
			{/* packed box */
			  ncol = purdueHeader.numCols;
			  nrow = purdueHeader.numRows;
			}/* packed box */
		      else
			{/* other */
			  ncol = numCols;
			  nrow = numRows;
			}/* other */
		      printf("x,y[%d,%d]\n",x,y);
		      recalcAxBxAyBy= FALSE;
		    }/* yongning */
		  else if((event.xbutton.button == Button3))
		    {/* Give feedback */
		      short ix,iy,jh,jk;
		      float xtmp,ytmp,fmax,degree;
		      int lx_lim,ly_lim,ncol,nrow;
		      float axold,bxold,ayold,byold;
		      float *apart,*bpart;
		      
		      /* Now give some info on what's there */
		      ix = (x-255);
		      iy = (512 - y)-255;
		      lx_lim = 512;
		      ly_lim = 512;
		      axold = axL;
		      ayold = ayL;
		      bxold = bxL;
		      byold = byL;
		      fmax = (float)sqrt(pow(10.0,(double)maxFFTvalue))
			* 0.1; /* Why times 0.1 ??? */
		      if(fileType==PACKEDPORTABLE)
			{/* packed box */
			  ncol = purdueHeader.numCols;
			  nrow = purdueHeader.numRows;
			}/* packed box */
		      else
			{/* other */
			  ncol = numCols;
			  nrow = numRows;
			}/* other */
		      CK(apart = (float *)malloc(sizeof(float)*fftSize));
		      CK(bpart = (float *)malloc(sizeof(float)*fftSize));
		      
		      fft_latref2_index_ (&ix, &iy, 
					  &fftSize, &fftSize, 
					  &jh, &jk,
					  &xtmp, &ytmp, 
					  &lx_lim, &ly_lim, 
					  &axold, &ayold, &bxold, &byold);
		      
		      fft_latref2_spot_ (fftImage_buffer2, 
					 &fftSize, &fftSize, 
					 apart, bpart,
					 &ix, &iy, 
					 &fmax, &degree,
					 &ncol, &nrow);

		      free(apart);
		      free(bpart);
		    }/* Give feedback */

		  if((event.xbutton.button != Button3))
		    {/* if it wasn't the third button they clicked above try
		      * and find a lattice point for it ... */
		      if(pickLattice==1)
			{/* Get first points */
			  latPt[0].measuredX = x;
			  latPt[0].measuredY = y;
			  currentLatPt=0;
			  stuffContourBuffer();		   
			  pickLattice= 2;
			}/* Get first points */
		      else if(pickLattice==2)
			{/* Get second points */
			  latPt[1].measuredX = x;
			  latPt[1].measuredY = y;
			  currentLatPt=1;
			  stuffContourBuffer();		   
			  pickLattice= 0;
			  pickedLatPts= TRUE;
			}/* Get second points */
		      else if(pickedLatPts)
			{/* Just find out which one we're dealing with */
			  findLatPt(x,y);
			  latPt[currentLatPt].measuredX = x;
			  latPt[currentLatPt].measuredY = y;
			  stuffContourBuffer();		   
			}/* Just find out which one we're dealing with */
		    }/* if it wasn't the third button they clicked above try
		      * and find a lattice point for it ... */
		}/* Main fft screen */
	      
	      
	      /* Calc lattice - they may have moved where they want it */
	      if(!layerLines && pickedLatPts && !drawCTF)
		{/* OK to calc */

		  calcLattice(recalcAxBxAyBy);

		}/* OK to calc */
	      
	      if(drawCTF && disp1Dca )
		{/* Just default it to where they clicked */
		  latPt[0].measuredX = x;
		  latPt[0].measuredY = 256;
		  latPt[1].measuredX = 256;
		  latPt[1].measuredY = x;
		  currentLatPt=1;
		  stuffContourBuffer();		   
		  pickLattice= 0;
		  pickedLatPts= TRUE;
		}/* Just default it to where they clicked */

	      /******************************************/
	      /* DO the CTF stuff here                  */
	      /******************************************/
	      if(pickedLatPts && drawCTF)
		{/* DO CTF  */
		  int thisMax, thisMin,x1,y1,x2,y2,sliderSize;
		  float ang1,distAsq,distBsq,distA,distB, meanValue;

		  ctfPts[ctfCurveNum].latX0= latPt[0].measuredX;
		  ctfPts[ctfCurveNum].latY0= latPt[0].measuredY;
		  ctfPts[ctfCurveNum].latX1= latPt[1].measuredX;
		  ctfPts[ctfCurveNum].latY1= latPt[1].measuredY;
		  ctfPts[ctfCurveNum].pickedPts= TRUE;
		  
		  if(latPt[0].measuredX==256)
		    latPt[0].measuredX=257;
		  
		  if(latPt[0].measuredX != 256)
		    ang1 = (float)atan((double)
				       (
					fabs((float)(latPt[0].measuredY-256))/
					fabs((float)(latPt[0].measuredX-256))
					)
				       );
		  else
		    ang1 = deg2rad(90.0);
		  
		  /* If we're greater than 90 degrees add that in */
		  if(latPt[0].measuredX < 256)
		    ang1 = deg2rad(90.0) + (deg2rad(90.0) - ang1);
		  /* If we're greater than 180 degrees add that in */
		  if(latPt[0].measuredY > 256)
		    ang1 = deg2rad(180.0) + (deg2rad(180.0) - ang1);
	
		  x1= latPt[0].measuredX;
		  x2= latPt[1].measuredX;
		  y1= latPt[0].measuredY;
		  y2= latPt[1].measuredY;

		  distAsq =    ((x1-256)*(x1-256))+
		    ((y1-256)*(y1-256))
		      ;
		  distBsq =    ((x2-256)*(x2-256))+
		    ((y2-256)*(y2-256))
		      ;
		  distA = sqrt(distAsq);
		  distB = sqrt(distBsq);

		  ctfPts[ctfCurveNum].distAsq= distAsq;
		  ctfPts[ctfCurveNum].distA  = distA;
		  ctfPts[ctfCurveNum].distBsq= distBsq;
		  ctfPts[ctfCurveNum].distB  = distB;
		  ctfPts[ctfCurveNum].omega  = ang1;

		  ctfFocusMinvalue = distA / distB;
		  ctfAngMajorValue = ang1;

		  calcDefocus(&ctffocusMajValue, &ctfFocusMinvalue);
		  printf("fftClick [%f,%f]\n",
			 ctffocusMajValue, ctfFocusMinvalue);
#define SETUP_SCROLLS 1
#ifdef SETUP_SCROLLS
		  /* Now setup the scroll bars */
		  XtSetArg(arglst[0], XmNincrement, (int)(10));
		  XtSetArg(arglst[1], XmNpageIncrement, (int)(100));
		  XtSetArg(arglst[2], XmNvalue, (int)(rad2deg(ctfAngMajorValue)
						      *100.0));
		  XtSetArg(arglst[3], XmNmaximum, (int)18888);
		  XtSetArg(arglst[4], XmNminimum, (int)-18000);
		  XtSetValues(ctfAngMajorScrollW, arglst, 5);

		  XtSetArg(arglst[0], XmNvalue, (int)(ctfFocusMinvalue*100.0));
		  XtSetArg(arglst[1], XmNincrement, (int)(1));
		  XtSetArg(arglst[2], XmNpageIncrement, (int)(10));
		  XtSetArg(arglst[3], XmNmaximum, (int)11111);
		  XtSetArg(arglst[4], XmNminimum, (int)-10000);
		  XtSetValues(ctfFocusMinscrollW, arglst, 5);

		  XtSetArg(arglst[0], XmNvalue, (int)(ctffocusMajValue*100.0));
		  XtSetArg(arglst[1], XmNincrement, (int)(1));
		  XtSetArg(arglst[2], XmNpageIncrement, (int)(10));
		  XtSetArg(arglst[3], XmNmaximum, (int)11111);
		  XtSetArg(arglst[4], XmNminimum, (int)-10000);
		  XtSetValues(ctffocusMajScrollW, arglst, 5);

#else

		  /* Now setup the scroll bars */
		  XtSetArg(arglst[0], XmNvalue, (int)(rad2deg(ctfAngMajorValue)
						      *100.0));
		  XtSetValues(ctfAngMajorScrollW, arglst, 1);
		  XtSetArg(arglst[0], XmNvalue, (int)(ctffocusMajValue*100.0));
		  XtSetValues(ctffocusMajScrollW, arglst, 1);
		  XtSetArg(arglst[0], XmNvalue, (int)(ctfFocusMinvalue*100.0));
		  XtSetValues(ctfFocusMinscrollW, arglst, 1);
#endif


		  sprintf(outStr,"%f",ctfFocusMinvalue);  
		  XmTextSetString(ctfFocusMinvalueW, outStr);
		  XmTextSetString(ctfMinorDefocusW,outStr);
		  sprintf(outStr,"%.2f",rad2deg(ctfAngMajorValue));  
		  XmTextSetString(ctfAngMajorValueW, outStr);
		  XmTextSetString(ctfAngMajorW, outStr); 
		  sprintf(outStr,"%f",ctffocusMajValue);  
		  XmTextSetString(ctffocusMajValueW, outStr);
		  XmTextSetString(ctfMajorDefocusW,outStr);

		}/* Do CTF  /*


	      /* Draw new Curve */
	      DrawFFTWindow();

	      
	    }/* OK to keep on trying */
	  else if(filterToggle || lopHipSwitch)
	    {/* Get resolution filter point */

	      if(TRUE)
		{/* No reso filter with layer lines yet */
		  if((event.xbutton.button == Button1))
		    {/* Select new resolution */
		      resoX= x;
		      resoY= y;
		      lowResoFilter = (((x-256)*(x-256))+((y-256)*(y-256)));
		    }/* Select new resolution */
		  if((event.xbutton.button == Button2))
		    {/* Select new resolution */
		      resoX= x;
		      resoY= y;
		      hiResoFilter = (((x-256)*(x-256))+((y-256)*(y-256)));
		    }/* Select new resolution */
		  else if((event.xbutton.button == Button3))
		    {/* Reset resolution */
		      resoX= 0;
		      resoY= 0;
		      lowResoFilter = 320000;  /* Ridiculously high */
		      hiResoFilter  = 0;  /* Ridiculously lo */
		    }/* Reset resolution */

		  sprintf(outStr,"%d",(int)sqrt((float)lowResoFilter));
		  XmTextSetString(lopRadiusTextW,outStr);
		  sprintf(outStr,"%d",(int)sqrt((float)hiResoFilter));
		  XmTextSetString(hipRadiusTextW,outStr);

		  calcFFT(FALSE, 0, 0);
		  DrawFFTWindow();
		}/* No reso filter with layer lines yet */

	    }/* Get resolution filter point */

	  break;
	  
	case KeyPress:
	  ans= XLookupKeysym((XKeyEvent *)&event, 0);
	  printf ("Keyed[%c]\n",ans);

	  break;
	  
	default:
	  break;
	}/* Find out which event we have */
    }/* Only look at events meant for us */
  
  XFlush(dsply);            /* Make changes current */

  
}/* FFTCLICK */



/*************************************************************
 * contourWindow_create() - Remember the contour area window
 * pointer for use later.  
 */
void contourWindow_create(Widget widget, int *tag, 
			  XmAnyCallbackStruct *callback_data)
{/* CONTOURWINDOW_CREATE */
  
  /* Remember the widget number for later */
  contourWidget= widget;
  
}/* CONTOURWINDOW_CREATE */




/*************************************************************
 * ctfArrowActivate() - They want to change the current
 * CTF by using the arrows.
*/
void ctfArrowActivate(Widget widget, int *tag, 
		      XmPushButtonCallbackStruct *callback_data)
{/* CTFARROWACTIVATE */
  char outStr[16];

  if(*tag==XmARROW_UP)
    {/* Up */
      ctfCurveNum++;
    }/* Up */
  else if(*tag==XmARROW_DOWN)
    {/* Dopwn */
      ctfCurveNum--;
    }/* Down */
  else printf("CTf curve num unchanged.\n");
  
  if(ctfCurveNum >= MAXCTFS)
    ctfCurveNum = MAXCTFS-1;
  if(ctfCurveNum < 0)
    ctfCurveNum = 0;

  maxCtfCurve = MAX((ctfCurveNum+1),maxCtfCurve);

  /* Restuff the old points */
  if(!ctfPts[ctfCurveNum].pickedPts)
    {/* Guess some new ones */
      ctfPts[ctfCurveNum].latX0= latPt[0].measuredX;
      ctfPts[ctfCurveNum].latY0= latPt[0].measuredY;
      ctfPts[ctfCurveNum].latX1= latPt[1].measuredX;
      ctfPts[ctfCurveNum].latY1= latPt[1].measuredY;
      /*      ctfPts[ctfCurveNum].distA = ctfPts[0].distA;
      ctfPts[ctfCurveNum].distB = ctfPts[0].distB;  */
    }/* Guess some new ones */
  else
    {/* restuff */
      latPt[0].measuredX = ctfPts[ctfCurveNum].latX0;
      latPt[0].measuredY = ctfPts[ctfCurveNum].latY0;
      latPt[1].measuredX = ctfPts[ctfCurveNum].latX1;
      latPt[1].measuredY = ctfPts[ctfCurveNum].latY1;
    }/* restuff */
  ctfPts[ctfCurveNum].pickedPts= TRUE;

  sprintf(outStr,"%d",ctfCurveNum+1);
  XmTextSetString(ctfCurveTextW,outStr);

  DrawFFTWindow();

}/* CTFARROWACTIVATE */




/**************************************************************
 * doTheIcosaArrow() - Draw the Icosahedron
*/
void doTheIcosaArrow(float speedFactor)
{/* DOTHEICOSAARROW */

  icosaScaleF *= speedFactor;

  if(dispIcosa)
    {/* Draw cisoahedron */

      DrawWindow();
      
      calcSizeFactor(zoomFactor);
      DrawIcosahedron(deg2rad(0.0-globTheta),
		      deg2rad(0.0-globPhi),
		      deg2rad(globOmega),
		      0, 0,
		      (float)((float)numCols*sizeFactor*2.0), 
		      image_window1); 

      DrawIcosahedron(deg2rad(0.0-globTheta),
		      deg2rad(0.0-globPhi),
		      deg2rad(globOmega),
		      500, 255,
		      numCols,
		      XtWindow(mapShadeScreenWindow));

    }/* Draw cisoahedron */

}/* DOTHEICOSAARROW */



/* ************************************************************
 * icosaArrowActivateButton() - Arrows change the speed at
 * which the icosahedron is expanded or contracted.
 */
void icosaArrowActivateButton(Widget widget, XEvent *event)
{/* ICOSAARROWACTIVATEBUTTON */
  XEvent thisEvent;
  float speedFactor;

  thisEvent= *(event);
  
  /* Get which button */
  if(widget==icosaDownWidget)
    {/* DOnw */
      if((thisEvent.xbutton.button == Button1))
	speedFactor= 0.99;
      if((thisEvent.xbutton.button == Button2))
	speedFactor= 0.95;
      if((thisEvent.xbutton.button == Button3))
	speedFactor= 0.9;
    }/* DOnw */
  else
    {/* UP */
      if((thisEvent.xbutton.button == Button1))
	speedFactor= 1.01;
      if((thisEvent.xbutton.button == Button2))
	speedFactor= 1.05;
      if((thisEvent.xbutton.button == Button3))
	speedFactor= 1.1;
    }/* UP */
    
  doTheIcosaArrow(speedFactor);
  
}/* ICOSAARROWACTIVATEBUTTON */





/*************************************************************
 * icosaArrowActivate() - They want to change the current
 * size of the icosahedron by using the arrows.
*/
void icosaArrowActivate(Widget widget, int *tag, 
			XmPushButtonCallbackStruct *callback_data)
{/* ICOSAARROWACTIVATE */

  if(*tag==XmARROW_UP)
    {/* Up */
      doTheIcosaArrow(1.01);
    }/* Up */
  else if(*tag==XmARROW_DOWN)
    {/* Dopwn */
      doTheIcosaArrow(0.99);
    }/* Down */
  else printf("Unknown ICOSA_ARROW.\n");
  
}/* ICOSAARROWACTIVATE */





/*************************************************************
 * samRenArrowActivate() - They want to change the sampling
 * for 3D render.
*/
void samRenArrowActivate(Widget widget, int *tag, 
			 XmPushButtonCallbackStruct *callback_data)
{/* SAMRENARROWACTIVATE */
  char *thisStr;
  int thisVal;

  thisStr = XmTextGetString(sampleRenderWidget);
  thisVal = (int)strtod(thisStr,0);
  XtFree(thisStr);

  if(*tag==XmARROW_UP)
    {/* Up */
      thisVal++; 
    }/* Up */
  else if(*tag==XmARROW_DOWN)
    {/* Dopwn */
      thisVal--; 
    }/* Down */
  else printf("Unknown SAMREN_ARROW.\n");
  
  thisVal = MAX(MIN(thisVal,10),1);

  sprintf(outStr,"%d",thisVal);
  XmTextSetString(sampleRenderWidget,outStr);
    
  if(thisVal > 1 && globFract >= 0.75)
    {/* set shading fraction */
      globFract = 0.25;
      sprintf(outStr,"%.2f",globFract);
      XmTextSetString(fractValueWidget,outStr); 
      XtSetArg(arglst[0], XmNvalue, (int)(globFract*1000.0));
      XtSetValues(fractScrollWidget, arglst, 1);
      XFlush(dsply);
    }/* set shading fraction */


}/* SAMRENARROWACTIVATE */





/*************************************************************
 * fftArrowActivate() - They want to change the current
 * image to FFT by using the arrows.
*/
void fftArrowActivate(Widget widget, int *tag, 
		      XmPushButtonCallbackStruct *callback_data)
{/* FFTARROWACTIVATE */

  /* Just call the radiusArrow call back rtn. */
  if(*tag==XmARROW_UP)
    imgNumArrowActivate(imgnumArrowUpWidget, tag, callback_data);
  else imgNumArrowActivate(imgnumArrowDownWidget, tag, callback_data);

  averageFFT= FALSE;    /* Don't want average anymore */

  calcFFT(TRUE, 0, 0);
  DrawFFTWindow();

  /* Make them do a new circavg */
  calcFFT1DcircAvg= FALSE;

  /* DO if rotflt'ing */
  newOrigin = TRUE;
  if(XtIsManaged(rotfltscreen))
    {
      iopt[0]= 1;      /* 1 == Refine symmetry axis */
      iopt[1]= 0;
      iopt[2]= 0;      /* ==1  means plot power spectrum */
      iopt[3]= 0;
      rotflt_tryit_activate(0,0,0);
    }

}/* FFTARROWACTIVATE */



/*************************************************************
 * fftImageNumArrowButton() - They want to change the current
 * image to FFT by using the arrows.
*/
void fftImgNumArrowButton(Widget widget, XEvent *event)
{/* FFTIMGNUMARROWBUTTON */
  XEvent thisEvent;
  float speedFactor;

  if(!locked)
    {/* Can't do if we're locked out */

      thisEvent= *(event);

      /* Get which button */
      if((thisEvent.xbutton.button == Button1))
	speedFactor = 1.0;;
      if((thisEvent.xbutton.button == Button2))
	speedFactor = 2.0;
      if((thisEvent.xbutton.button == Button3))
	speedFactor = 5.0;

      if((widget==rotfltImgnumArrowUpW) ||
	 (widget==fftImgnumArrowUpW))
	doTheImgNum(imgnumArrowUpWidget, speedFactor);
      else 
	doTheImgNum(imgnumArrowDownWidget, speedFactor);

    }/* Can't do if we're locked out */

  averageFFT= FALSE;    /* Don't want average anymore */

  calcFFT(TRUE, 0, 0);
  DrawFFTWindow();

  /* Make them do a new circavg */
  calcFFT1DcircAvg= FALSE;

  /* DO if rotflt'ing */
  newOrigin = TRUE;
  if(XtIsManaged(rotfltscreen))
    {
      iopt[0]= 1;      /* 1 == Refine symmetry axis */
      iopt[1]= 0;
      iopt[2]= 0;      /* ==1  means plot power spectrum */
      iopt[3]= 0;
      
      rotflt_tryit_activate(0,0,0);
    }

}/* FFTIMGNUMARROWBUTTON */








/********************************************************************
 * latReset_activate() - Reset axL, ayL, bxL and byL
 * to the original values they were calculated to by when the cursor
 * selected the points.
 */
void latReset_activate(Widget widget, int *tag, 
		       XmAnyCallbackStruct *callback_data)
{/* LATRESET_ACTIVATE */
  int thisOne,thisMin,thisMax,i,sliderSize;
  XmString simpleString;

  axL = axLSave;
  ayL = ayLSave;
  bxL = bxLSave;
  byL = byLSave;

  anga = angaSave;
  angb = angbSave;

  astar  = astarSave;
  bstar  = bstarSave;
  gamstr = gamstrSave;
  omega  = omegaSave;
  dstar  = dstarSave;

  /* Stuff this for things like the crysDisp screen */
  map_abang = 180.0 - gamstr;
  if(bstar != 0.0 && astar != 0.0)
    aoverb = 1.0 / (astar / bstar);

  latScrollValue[0]= astarSave;
  if(layerLines)
    latScrollValue[1]= dstarSave;
  else latScrollValue[1]= bstarSave;
  latScrollValue[2]= gamstrSave;
  latScrollValue[3]= omegaSave;
  if(!layerLines)
    latScrollValue[3]= 0.0;         /* Omega starts at 0 now ... */

  if(inited)
    {/* Now setup the scrolls and labels */
      for(i=0;i<4;i++)
	{/* Do all */
	  
	  if(!layerLines ||
	     (layerLines && ((i==3)||(i==1))))
	    {/* OK to do */
	      sprintf(outStr,"%.2f",(latScrollValue[i]));
	      XmTextSetString(latticeLabel2W[i], outStr); 
	      
	      /* Now setup the scroll bars to hold the new max/min */
	      if(latScrollValue[i] < 0.0)
		{/* Negative number */
		  thisMax = (int)((latScrollValue[i]-
				   (latScrollValue[i] * 0.5)) * 100.0);
		  thisMin = (int)((latScrollValue[i]+
				   (latScrollValue[i] * 0.5)) * 100.0);
		}/* Negative number */
	      else
		{/* Positive number */
		  thisMax = (int)((latScrollValue[i]+
				   (latScrollValue[i] * 0.5)) * 100.0);
		  thisMin = (int)((latScrollValue[i]-
				   (latScrollValue[i] * 0.5)) * 100.0);
		}/* Positive number */
	    }/* OK to do */
	  
	  if(i==3 && !layerLines)
	    {/* omega */
	      thisMax= 6000;  /* 60 degrees times 100 */
	      thisMin= -6000;
	      latScrollValue[i]= 0.0;
	      omega = 0.0;
	      sprintf(outStr,"%.2f",(latScrollValue[i]));
	      XmTextSetString(latticeLabel2W[i], outStr); 
	    }/* omega */
	  
	  if(!layerLines ||
	     (layerLines && (i==3 || i==1)))
	    {/* OK to do */
	      if(thisMax == thisMin)
		{/* Make them separate */
		  thisMax += 500;
		  thisMin -= 500;
		}/* Make them separate */
	      if(layerLines && (i==3 || i==1))
		{
		  if(i==3)
		    {
		      thisMax += 1800;   /* +/- 180 degrees */
		      thisMin -= 1800; 
		    }
		  else
		    {
		      thisMax += 500; 
		      thisMin -= 500; 
		    }
		}
	      sliderSize = (thisMax - thisMin) / 100;
	      sliderSize = MAX(sliderSize,1);
	      
	      XtSetArg(arglst[0], XmNmaximum, (int)(thisMax+
						    (float)sliderSize));
	      XtSetArg(arglst[1], XmNminimum, (int)(thisMin));
	      XtSetArg(arglst[2], XmNsliderSize, (int)(sliderSize));
	      if(i==3)  /* omega */
		XtSetArg(arglst[3], XmNvalue, (int)(latScrollValue[i]*10.0));
	      else XtSetArg(arglst[3], XmNvalue, (int)(latScrollValue[i]*
						       100.0));
	      XtSetValues(latticeScrollW[i], arglst, 4);
	    }/* OK to do */
	  
	  XSync(dsply,FALSE);
	  XFlush(dsply);
	  
	}/* Do all */
      
      /*  if(!layerLines) */
      if(!(*tag==-99))
	{/* OK to draw */
	  calcFFT(FALSE, 0, 0);
	  DrawFFTWindow();
	}/* OK to draw */
    }/* Now setup the scrolls and labels */
  
}/* LATRESET_ACTIVATE */




/**********************************************************
 * fft_average_activate() - They want to look at the
 * average FFT of all images .
*/
void fft_average_activate(Widget widget, int *tag, 
			  XmAnyCallbackStruct *callback_data)
{/* fft_average_actIVATE */

  averageFFT = TRUE;
  calcFFT(TRUE, 0, 0);
  DrawFFTWindow();

}/* fft_average_actIVATE */




/**********************************************************
 * fourAver_work() - Generate the Structure factors.
*/
void fourAver_work(Widget widget, int *tag, 
		   XmAnyCallbackStruct *callback_data)
{/* FOURAVER_WORK */
  float axPass,ayPass,bxPass,byPass;
  int i,j,k,lastSF;
  float x1,x2,y1,y2,thisMin,thisMax,scaleFac,tmpF;
  char  *thisStr;
  int   ncol,nrow,nsec;
  int   sf_type, em_ndim;
  float alp_star,bet_star,gam_star;
  float a_star,b_star,c_star;
  float alp,bet,gam;
  fString_t outfileName;
  char outFilename[256];
  float x_shift, y_shift, thisMap_abang;
  int   fft_full;
  int   ikey;
  FILE *fOut;

  if(axL==0.0 && bxL==0.0 && ayL==0.0 && byL==0.0)
    {/* Can't goon */
      askThem("Must create a lattice first.",
	      FALSE,XmDIALOG_INFORMATION);
    }/* Can't goon */
  else
    {/* OK */

      pif_init_head(&faHeader);

      /* Let's try these */
      axPass = axL;
      ayPass = ayL;
      bxPass = bxL;
      byPass = byL; 

      /* Use this for debugging with gap.img  
      axPass = 41.452003;
      ayPass = 0.814141;
      bxPass = 19.514853;
      byPass = 36.546074;
      */

      x1 = axPass;
      y1 = ayPass;
      x2 = bxPass;
      y2 = byPass;

      printf("ax[%f]ay[%f]bx[%f]by[%f]\n",x1,y1,x2,y2);

      if(fileType==PACKEDPORTABLE)
	{/* packed box */
	  ncol = purdueHeader.numCols;
	  nrow = purdueHeader.numRows;
	}/* packed box */
      else
	{/* other */
	  ncol = numCols;
	  nrow = numRows;
	}/* other */

      emfft_spt_rob1_ (fftImage_fourAver, &fftSize, &fftSize,
		       &axPass,&ayPass,&bxPass,&byPass, &nsf,
		       &x1,&x2,&y1,&y2, &ncol, &nrow,
		       &rasterStep,&fourAverRadius,&fourAverCut);



      if(sFacs != NULL)
	free(sFacs);
      sFacs = (strucFact_t *)malloc(sizeof(strucFact_t)*nsf*2);
      if(fa_sFacs != NULL)
	free(fa_sFacs);
      fa_sFacs = (strucFact_t *)malloc(sizeof(strucFact_t)*nsf*2);

      
      emfft_spt_rob2_ (fftImage_fourAver, &fftSize, &fftSize,  
		       &axPass,&ayPass,&bxPass,&byPass, 
		       &nsf,
		       fa_sFacs,
		       &x1,&x2,&y1,&y2, 
		       &ncol, &nrow,
		       &faHeader.a_star,   
		       &faHeader.b_star,   
		       &faHeader.c_star, 
		       &faHeader.alp_star,   
		       &faHeader.bet_star, 
		       &faHeader.gam_star,
		       &rasterStep,&fourAverRadius,&fourAverCut,
		       &shiftFFT);
      
      rec_to_real_(&faHeader.a_cell, &faHeader.b_cell, 
		   &faHeader.c_cell, 
		   &faHeader.alp, &faHeader.bet, 
		   &faHeader.gam, 
		   &vol,
		   &faHeader.a_star, &faHeader.b_star, 
		   &faHeader.c_star, 
		   &vol_star, 
		   &faHeader.alp_star, &faHeader.bet_star, 
		   &faHeader.gam_star);
      


      thisMin = thisMax = fa_sFacs[0].a_amp;
      origNsf = nsf;  /* We may expand this in the impose rtns so remember */
      for(i=0;i<nsf;i++)
	{/* Check all SFs */

	  /* Copy gold std into temp area */
	  sFacs[i].h = fa_sFacs[i].h;
	  sFacs[i].k = fa_sFacs[i].k;
	  sFacs[i].l = fa_sFacs[i].l;
	  sFacs[i].a_amp = fa_sFacs[i].a_amp;
	  sFacs[i].b_phi = fa_sFacs[i].b_phi;
	  sFacs[i].dstr = fa_sFacs[i].dstr;

	  thisMax = MAX(fa_sFacs[i].a_amp,thisMax);
	  thisMax = MAX(fa_sFacs[i].b_phi,thisMax);
	  thisMax = MAX(fa_sFacs[i].sig,thisMax);
	  thisMax = MAX(fa_sFacs[i].dstr,thisMax);

	  thisMin = MIN(fa_sFacs[i].a_amp,thisMin);
	  thisMin = MIN(fa_sFacs[i].b_phi,thisMin);
	  thisMin = MIN(fa_sFacs[i].sig,thisMin);
	  thisMin = MIN(fa_sFacs[i].dstr,thisMin);

	  if(fa_sFacs[i].a_amp == 0.0 &&
	     fa_sFacs[i].b_phi == 0.0 )
	    {
	      lastSF = i;
	    }

	}/* Check all SFs */


    }/* OK */

}/* FOURAVER_WORK */




/**********************************************************
 * fourAver_backit() - They want a back transform based
 * on the SFs.
*/
void fourAver_backit(int firstTime)
{/* FOURAVER_BACKIT */
  float axPass,ayPass,bxPass,byPass;
  int i,j,k,lastSF;
  int backSizeA,backSizeB;
  float x1,x2,y1,y2,thisMin,thisMax,scaleFac,tmpF;
  char  *thisStr;
  int   ncol,nrow,nsec;
  int   sf_type, em_ndim;
  float alp_star,bet_star,gam_star;
  float a_star,b_star,c_star;
  float alp,bet,gam;
  fString_t outfileName;
  char outFilename[256];
  float x_shift, y_shift, thisMap_abang;
  int   fft_full;
  int   ikey;

  sf_type = 1;
  em_ndim = 2;  /* 2d data here */
  
  /* Now convert Structure factors back into a 2D MAP */
  if((float_image_buffer1) &&
     (float_image_buffer1 != lowRes_buffer1))
    {free(float_image_buffer1);  }
  
  /* These are already calculated 
     map_abang = 90.0;
     aoverb = 1.0;
     thisMap_abang = 180.0 - map_abang; 
     */
  thisMap_abang = map_abang;
  
  adim = bdim = 256; 
  backSizeA = adim;
  backSizeB = bdim;

  float_image_buffer1 = (float *) 
    malloc(sizeof(float)*2*backSizeA*backSizeB);

  emsf_backt_rob_(&sf_type,
		  &em_ndim,
		  &nsf,
		  sFacs,
		  &map_min,&map_max,
		  &aoverb, 
		  &faHeader.alp_star, &faHeader.bet_star, 
		  &faHeader.gam_star,
		  &faHeader.a_star, &faHeader.b_star, 
		  &faHeader.c_star,
		  &vol, &vol_star, 
		  &faHeader.a_cell, &faHeader.b_cell, 
		  &faHeader.c_cell, 
		  &thisMap_abang,
		  &faHeader.alp,&faHeader.bet,&faHeader.gam,
		  float_image_buffer1,
		  &backSizeA,&backSizeB);

  fortranBuffer = TRUE;     /* Next time use fortran to free() this */
  
  cdNumCols = ncol = backSizeA;
  cdNumRows = nrow = backSizeB;
  cdNumSecs = 1;
  
  printf("\n\nNew dimensions ->  %d,%d,%d\n",ncol,nrow,1);
  
  /* Now scale the buffer to fit a BYTE output */
  if(map_max == map_min)
    scaleFac = 1.0;
  else scaleFac = 254.0 / (map_max - map_min);

  for(i=0;i<cdNumRows;i++)
    {
      for(j=0;j<cdNumCols;j++)
	{/* Scale them all */
	  /* j*2 because the fourt() rtn stuffs it in the first 
	   * position */
	  float_image_buffer1[((i*(cdNumCols*2))+
			       (j*2))] = 
				 (float_image_buffer1[((i*(cdNumCols*2))+
						       (j*2))] - 
				  map_min) * scaleFac;
	  
	}/* Scale them all */
    }
  
  fourierAveragedCrysDisp = TRUE;
  
  crysDisp_screen_activate(0,&firstTime,0); 


}/* FOURAVER_BACKIT */






/**********************************************************
 * cDisp_newAdimBdim_changed() - They changed the adim
 * and/or bdim concept.  redraw
*/
void cDisp_newAdimBdim_changed(Widget widget, int *tag, 
			       XmAnyCallbackStruct *callback_data)
{/* CDISP_NEWADIMBDIM_CHANGED */

  fourAver_backit(FALSE);

}/* CDISP_NEWADIMBDIM_CHANGED */





/**********************************************************
 * fourAver_write_sf() - Write to disk.
*/
void fourAver_write_sf(int tag)
{/* FOURAVER_WRITE_SF */

  /* This is done in fourAver_work() above... */
  sprintf(outStr,"Wrote structure factors to file [four_aver_sf.psf]");
  askThem(outStr,FALSE,XmDIALOG_INFORMATION);


}/* FOURAVER_WRITE_SF */





/**********************************************************
 * fourAver_refine_impose() - Impose symmetry.
*/
void fourAver_refine_impose(int tag)
{/* FOURAVER_REFINE_IMPOSE */
  float axPass,ayPass,bxPass,byPass;
  int i,j,k,lastSF;
  float x1,x2,y1,y2,thisMin,thisMax,scaleFac,tmpF;
  char  *thisStr;
  int   ncol,nrow,nsec;
  int   sf_type, em_ndim;
  float alp_star,bet_star,gam_star;
  float a_star,b_star,c_star;
  float alp,bet,gam;
  fString_t outfileName;
  char outFilename[256];
  float x_shift, y_shift, thisMap_abang;
  int   fft_full;
  int   ikey;
  int nlim;

  sf_type = 1;
  em_ndim = 2;  /* 2d data here */
  
  /* Now check to see if we want to do anything to these
   * structure factors or not */
  if(tag==1)
    {/* Refine phase origins */
      if(refinePhaseOrigin)
	{/* Refine the SF's for 2,3,4,or 6 symmetry */
	  float sf_sumfsq;
	  ikey = refinePhaseOrigin;

	  emsf_origins_rob_(&em_ndim,&sf_type,&nsf,
			    &sf_sumfsq, 
			    sFacs, &ikey);

	}/* Refine the SF's for 2,3,4,or 6 symmetry */
      
    }/* Refine phase origins */
  else if(tag==2)
    {/* Impose symmetries */
      float thisAoverb;

      if(imposeSymmetry)
	{/* Impose the symmetry */
	  nlim = nsf * 2 ;
	  if(imposeSymmetry == P2)
	    ikey = 1;
	  else if(imposeSymmetry == PGG)
	    ikey = 8;
	  else if(imposeSymmetry == CMM)
	    ikey = 9;
	  else if(imposeSymmetry == P4)
	    ikey = 10;
	  else if(imposeSymmetry == P4M)
	    ikey = 11;
	  else if(imposeSymmetry == P4G)
	    ikey = 12;
	  else if(imposeSymmetry == P6)
	    ikey = 16;
	  else if(imposeSymmetry == P6M)
	    ikey = 17;
	  else 
	    {/* Error */
	      sprintf(outStr,"Invalid Symmetry [%d]",imposeSymmetry);
	      askThem(outStr,FALSE,XmDIALOG_INFORMATION);
	      goto noSymm;
	    }/* Error */
	  thisAoverb = aoverbTemp;

	  emsf_sym_rob_(&em_ndim, &sf_type, &nlim, &nsf, 
			sFacs,
			&gamstr, 
			&ikey, &thisAoverb);

	  if(thisAoverb != aoverbTemp)
	    {/* Set it */
	      aoverbTemp = thisAoverb;
	      sprintf(outStr,"%f",aoverbTemp);
	      XmTextSetString(aoverbTextW,outStr);
	    }/* Set it */

	  if(map_abangTemp != (180.0 - gamstr))
	    {/* Set it */
	      map_abangTemp = 180.0 - gamstr;
	      sprintf(outStr,"%f",map_abangTemp);
	      XmTextSetString(map_abangTextW,outStr);
	      /* Now setup map and ab scrolls and tryit */
	      crysDisp_cr_activate(0,0,0);
	    }/* Set it */

	noSymm:;
	  
	}/* Impose the symmetry */
      
    }/* Impose symmetries */
  
  
  if(writeFourAverSF)
    {/* Write the sfs */
      /* Do we really need to clip the last zero ones ??? */
      /* nsf = lastSF;   */
      
      /* Let's write out a temporary PIF file to see what's
       * going on */
      sprintf(outFilename,"four_aver_sf.psf");

      pif_open_(outFilename,&faHeader.luNum,"w");
      
      faHeader.nimg  = 1;
      thisMax = MAX(thisMax,fabs(thisMin));
      thisMax = MAX(thisMax,360.0);
      printf("thisMax[%f]\n",thisMax);
      faHeader.floatScaleFactor        = (double)((double) MAXINT / 
						  (double) thisMax);
      faHeader.recipFloatScaleFactor   = (1.0/faHeader.floatScaleFactor);
      faHeader.int4scaleF              = faHeader.floatScaleFactor;
      faHeader.numCols   = 2;  /* We're 2D here */
      faHeader.numRows   = nsf;
      faHeader.mode      = 31;
      
      printf("map_abang here - [%f]\n",map_abang);
      faHeader.map_abang = map_abang;
      faHeader.aoverb = aoverb;
      if(featherMode==LINEAR)
	faHeader.borderType = 1;
      else if(featherMode==GAUSSIAN)
	faHeader.borderType = 2;
      else if(featherMode==EXPONENTIAL)
	faHeader.borderType = 3;
      else 
	faHeader.borderType = 0;
      faHeader.borderWidth = featherRadius-boxRadius;
      
      
      pif_write_gh(&faHeader);
      faHeader.imgNum    = 1;
      pif_write_dh(&faHeader);
      pif_write_sf(&faHeader, sFacs);
      pif_close(&faHeader.luNum);
      
      sprintf(outStr,"Wrote structure factors to file [four_aver_sf.psf]");
      askThem(outStr,FALSE,XmDIALOG_INFORMATION);
    }/* Write the sfs */
  
  /* Now convert Structure factors back into a 2D MAP */
  fourAver_backit(FALSE);
  
}/* FOURAVER_REFINE_IMPOSE */




/*******************************************************************
 * fourAver_manage() - Get the fourAver screen up.
 */
void fourAver_manage(Widget widget, int *tag, 
		     XmAnyCallbackStruct *callback_data)
{/* FOURAVER_MANAGE */
  XmString simpleString;
  int tag2;

  if(!averageFFT)
    {/* DOn't do for averaged FFT */

      sprintf(outStr,"No symmetry imposed yet");
      sprintf(outStr,"%s",outStr);
      createSimpleString(&simpleString,outStr);
      XtSetArg(arglst[0], XmNlabelString, simpleString);
      XtSetArg(arglst[1], XmNrecomputeSize, FALSE );
      XtSetValues(imposeLabelW, arglst, 2); 
      XFlush(dsply);
      XmStringFree(simpleString); 
      
      sprintf(outStr,"No refinement yet");
      createSimpleString(&simpleString,outStr);
      XtSetArg(arglst[0], XmNlabelString, simpleString);
      XtSetArg(arglst[1], XmNrecomputeSize, FALSE );
      XtSetValues(refineLabelW, arglst, 2); 
      XFlush(dsply);
      XmStringFree(simpleString); 
      
      imposeSymmetry = refinePhaseOrigin = 0;
      
      /* Generate gold standard structure factors */
      fourAver_work(widget, tag, callback_data);

      /* Just give us a default view of it */
      adim = bdim = 256;
      fourAver_backit(TRUE);

      raiseIt(main_window_widget);
      XFlush(dsply);

      raiseIt(fourAverScreen);

      /* Let's set the zoom */
      XmTextSetString(scaleXvalueW,"1.0");  
      /*
	raiseIt(crysDispscreen); */
      tag2 = TRUE;
      crysDisp_screen_activate(0,&tag2,0);
  
    }/* DOn't do for averaged FFT */
  
}/* FOURAVER_MANAGE */




/*******************************************************************
 * fourAver_resetSfs() - Reset to last gold STD of SFs.
 */
void fourAver_resetSfs(Widget widget, int *tag, 
		       XmAnyCallbackStruct *callback_data)
{/* FOURAVER_RESETSFS */
  int i;
  XmString simpleString;

  phaseOrigX = 0.0;
  phaseOrigY = 0.0;
  sprintf(outStr,"0.0");
  XmTextSetString(moveOrigXW,outStr);
  XmTextSetString(moveOrigYW,outStr);
  sprintf(outStr,"");
  XmTextSetString(shiftOrigXW,outStr);
  XmTextSetString(shiftOrigYW,outStr);

  nsf = origNsf;  /* We may have expanded this in the impose rtns */
  for(i=0;i<nsf;i++)
    {/* Check all SFs */
      
      /* Copy gold std into temp area */
      sFacs[i].h = fa_sFacs[i].h;
      sFacs[i].k = fa_sFacs[i].k;
      sFacs[i].l = fa_sFacs[i].l;
      sFacs[i].a_amp = fa_sFacs[i].a_amp;
      sFacs[i].b_phi = fa_sFacs[i].b_phi;
      sFacs[i].dstr = fa_sFacs[i].dstr;

    }/* Check all SFs */

  sprintf(outStr,"%s","No symmetry imposed yet");
  createSimpleString(&simpleString,outStr);
  XtSetArg(arglst[0], XmNlabelString, simpleString);
  XtSetArg(arglst[1], XmNrecomputeSize, FALSE );
  XtSetValues(imposeLabelW, arglst, 2); 
  XFlush(dsply);
  XmStringFree(simpleString); 
  sprintf(outStr,"%s","No refinement yet");
  createSimpleString(&simpleString,outStr);
  XtSetArg(arglst[0], XmNlabelString, simpleString);
  XtSetArg(arglst[1], XmNrecomputeSize, FALSE );
  XtSetValues(refineLabelW, arglst, 2); 
  XFlush(dsply);
  XmStringFree(simpleString); 

  /* Now reset map_abang and aoverb */
  crysDisp_reset_activate(0,0,0);

  fourAver_backit(FALSE);
  
}/* FOURAVER_RESETSFS */



/*************************************************************
 * movePhaseOrigin_activate() -  Move the concept of phase 
 * origin.
 */
void movePhaseOrigin_activate(Widget widget, int *tag, 
			      XmAnyCallbackStruct *callback_data)
{/* MOVEPHASEORIGIN_ACTIVATE */
  int i;
  char *shiftText;
  float moveX,moveY;

  if(deltaPhaseShift)
    {/* Do relative */
      shiftText = XmTextGetString(shiftOrigYW); 
      moveY = (float)strtod(shiftText,0);
      XtFree(shiftText);
      shiftText = XmTextGetString(shiftOrigXW); 
      moveX = (float)strtod(shiftText,0);
      XtFree(shiftText);
      
      phaseOrigX += moveX;
      phaseOrigY += moveY;
      sprintf(outStr,"%f",phaseOrigX);
      XmTextSetString(moveOrigXW,outStr);
      sprintf(outStr,"%f",phaseOrigY);
      XmTextSetString(moveOrigYW,outStr);
    }/* Do relative */
  else
    {/* Do absolute */
      shiftText = XmTextGetString(moveOrigYW); 
      moveY = (float)strtod(shiftText,0);
      XtFree(shiftText);
      shiftText = XmTextGetString(moveOrigXW); 
      moveX = (float)strtod(shiftText,0);
      XtFree(shiftText);
      
      phaseOrigX = moveX;
      phaseOrigY = moveY;

      nsf = origNsf;  /* We may have expanded this in the impose rtns */
      for(i=0;i<nsf;i++)
	{/* Check all SFs */
	  
	  /* Copy gold std into temp area */
	  sFacs[i].h = fa_sFacs[i].h;
	  sFacs[i].k = fa_sFacs[i].k;
	  sFacs[i].l = fa_sFacs[i].l;
	  sFacs[i].a_amp = fa_sFacs[i].a_amp;
	  sFacs[i].b_phi = fa_sFacs[i].b_phi;
	  sFacs[i].dstr = fa_sFacs[i].dstr;
	  
	}/* Check all SFs */

    }/* Do absolute */


  /* Now shift them structure factors */
  rob_sf_pshift_(sFacs,&nsf, &moveX,&moveY);

  /* Now init for next time */
  sprintf(outStr,"");
  XmTextSetString(shiftOrigXW,outStr); 
  XmTextSetString(shiftOrigYW,outStr); 

  fourAver_backit(FALSE);


}/* MOVEPHASEORIGIN_ACTIVATE */




/*********************************************************************
 * fourAver_deactivate() - Free memory and get fourAver interface off 
 * the display.
 */
void fourAver_deactivate(Widget widget, int *tag, 
			 XmAnyCallbackStruct *callback_data)
{/* FOURAVER_DEACTIVATE */

  XtUnmanageChild(fourAverScreen);
  XtUnmanageChild(crysDispscreen);

  if(fa_sFacs != NULL)
    free(fa_sFacs);
  if(sFacs != NULL)
    free(sFacs);
  fa_sFacs = sFacs = NULL;

  if(fourierAveragedCrysDisp && XtIsManaged(fftscreen))
    {/* Redraw for fft */

      virtual_read_activate(-99,-99);

      raiseIt(fftscreen);
    }/* Redraw for fft */
  fourierAveragedCrysDisp = FALSE;

}/* FOURAVER_DEACTIVATE */




/******************************************************************
 * impose_changed() - They've selected a new symmetry.
 */
void impose_changed(Widget widget, int *tag, 
		    XmAnyCallbackStruct *callback_data)
{/* IMPOSE_CHANGED */
  XmString simpleString;
  char symm[32];

  imposeSymmetry = *tag;

  if(imposeSymmetry == P2)
    sprintf(symm,"P2 Imposed");
  else if(imposeSymmetry == PGG)
    sprintf(symm,"PGG Imposed");
  else if(imposeSymmetry == CMM)
    sprintf(symm,"CMM Imposed");
  else if(imposeSymmetry == P4)
    sprintf(symm,"P4 Imposed");
  else if(imposeSymmetry == P4M)
    sprintf(symm,"P4M Imposed");
  else if(imposeSymmetry == P4G)
    sprintf(symm,"P4G Imposed");
  else if(imposeSymmetry == P6)
    sprintf(symm,"P6 Imposed");
  else if(imposeSymmetry == P6M)
    sprintf(symm,"P6M Imposed");
  else 
    sprintf(symm,"No symmetry imposed yet");

  sprintf(outStr,"%s",symm);
  createSimpleString(&simpleString,outStr);
  XtSetArg(arglst[0], XmNlabelString, simpleString);
  XtSetArg(arglst[1], XmNrecomputeSize, FALSE );
  XtSetValues(imposeLabelW, arglst, 2); 
  XFlush(dsply);
  XmStringFree(simpleString); 

  fourAver_refine_impose(2);

}/* IMPOSE_CHANGED */






/******************************************************************
 * refine_changed() - They've changed the concept of folding.
 */
void refine_changed(Widget widget, int *tag, 
		    XmAnyCallbackStruct *callback_data)
{/* refine_CHANGED */
  XmString simpleString;

  refinePhaseOrigin = *tag;

  if(refinePhaseOrigin == 1)
    sprintf(outStr,"2 Fold Refined");
  else if(refinePhaseOrigin == 2)
    sprintf(outStr,"3 Fold Refined");
  else if(refinePhaseOrigin == 3)
    sprintf(outStr,"4 Fold Refined");
  else if(refinePhaseOrigin == 4)
    sprintf(outStr,"6 Fold Refined");
  else sprintf(outStr,"No refinement yet ");
  createSimpleString(&simpleString,outStr);
  XtSetArg(arglst[0], XmNlabelString, simpleString);
  XtSetArg(arglst[1], XmNrecomputeSize, FALSE );
  XtSetValues(refineLabelW, arglst, 2); 
  XFlush(dsply);
  XmStringFree(simpleString); 

  fourAver_refine_impose(1);

}/* refine_CHANGED */







/******************************************************************
 * backSize_changed() - They want a new size of back transform
 */
void backSize_changed(Widget widget, int *tag, 
		      XmAnyCallbackStruct *callback_data)
{/* BACKSIZE_CHANGED */
  XmString simpleString;

  if(*tag==256)
    backSize = 2;
  else if(*tag==128)
    backSize = 1;
  else
    backSize = 0;

  sprintf(outStr,"Size of Back Transform = %d",*tag);
  createSimpleString(&simpleString,outStr);
  XtSetArg(arglst[0], XmNlabelString, simpleString);
  XtSetArg(arglst[1], XmNrecomputeSize, FALSE );
  XtSetValues(backSizeW, arglst, 2); 
  XFlush(dsply);
  XmStringFree(simpleString); 

  fourAver_backit(FALSE);

}/* BACKSIZE_CHANGED */





/* ************************************************************
 * rotflt_screen_deactivate() - Get the rotflt screen
 * off the display.
 */
void rotflt_screen_deactivate(Widget widget, int *tag, 
			      XmAnyCallbackStruct *callback_data)
{/* rotflt_screen_deactIVATE */
  
  file_width = orig_file_width;
  rotfltSwitch = FALSE;
  firstRotFlt = FALSE;
  /* Why can't I free this???  free(tempRotflt_buffer);  */
  XtUnmanageChild(rotfltscreen); 

  /* Reset these for "normal" FFT'ing */
  /* Already should be OK - we're useing
   * idim1Rotflt as variable now ... idim1= idim2= 1024;*//* MAX fft size */
  
}/* rotflt_screen_deactIVATE */




/* ************************************************************
 * rotflt_screen_activate() - Put the rotfltscreen on the display.
 */
void rotflt_screen_activate(Widget widget, int *tag, 
			    XmAnyCallbackStruct *callback_data)
{/* rotflt_screen_actIVATE */
  int thisMaxInt, thisMinInt,  sliderSize, i, idim1Calc;

  orig_file_width = file_width;
  rotfltSwitch = TRUE;
  firstRotFlt = TRUE;
  boxMethod= USERECT;
  newOrigin = TRUE;
  winScale = 1.0;

  idim1Calc = 1;
  for(i=0;i<10;i++)
    {
      idim1Calc *= 2;
      if(purdueHeader.numCols > (idim1Calc))
	idim1Rotflt = idim1Calc;
    }
  idim1Rotflt *= 2;
  idim1Rotflt = fftSize;
  CK(tempRotflt_buffer= (float *)malloc(sizeof(float)*2*fftSize*fftSize));

  if(fft_step <= 0.0)
    fft_step = 1.0;

  thisMaxInt = (int)(((idim1Rotflt / 2) - 2)/fft_step);
  thisMinInt = (int)(1);
  sliderSize = 5;
  nannuli = (int)(idim1Rotflt / 4);
  XtSetArg(arglst[0], XmNmaximum, (int)(thisMaxInt+(float)sliderSize));
  XtSetArg(arglst[1], XmNminimum, (int)(thisMinInt));
  XtSetArg(arglst[2], XmNsliderSize, (int)(sliderSize));
  XtSetArg(arglst[3], XmNvalue, (int)(nannuli));
  if(thisMaxInt > 200)
    XtSetArg(arglst[4], XmNpageIncrement, (int)(thisMaxInt/100));
  else if(thisMaxInt > 20)
    XtSetArg(arglst[4], XmNpageIncrement, (int)(thisMaxInt/10));
  else 
    XtSetArg(arglst[4], XmNpageIncrement, (int)(thisMaxInt));
  XtSetValues(nannuliScrollW, arglst, 5);

  /* Calc max symmetry Limit */
  /* Don't set lg_end variable here - it's done on the interface */
  thisMaxInt = (3.1415 * purdueHeader.numCols) * 0.5;
  XtSetArg(arglst[0], XmNmaximum, (int)(thisMaxInt+(float)sliderSize));
  XtSetArg(arglst[1], XmNsliderSize, (int)(sliderSize));
  if(thisMaxInt > 200)
    XtSetArg(arglst[2], XmNpageIncrement, (int)(thisMaxInt/10));
  else if(thisMaxInt > 20)
    XtSetArg(arglst[2], XmNpageIncrement, (int)(thisMaxInt/10));
  else 
    XtSetArg(arglst[2], XmNpageIncrement, (int)(thisMaxInt/10));
  XtSetValues(lgendScrollW, arglst, 3);

  sprintf(outStr,"%.2f",fft_step);
  XmTextSetString(fftStepTextW,outStr);

  omega = 0.0;
  XtSetArg(arglst[0], XmNvalue, (int)(omega));
  XtSetValues(phi0ScrollW, arglst, 1);

  iopt[0]= 1;      /* 1 == Refine symmetry axis */
  iopt[1]= 0;
  iopt[2]= 0;      /* ==1  means plot power spectrum */
  iopt[3]= 0;

  raiseIt(rotfltscreen);
  rotflt_tryit_activate(0,0,0);

}/* rotflt_screen_actIVATE */





/* ************************************************************
 * rotflt_write_activate() - Wriet the rotflt img to disk.
 */
void rotflt_write_activate(Widget widget, int *tag, 
			   XmAnyCallbackStruct *callback_data)
{/* ROTFLT_WRITE_ACTIVATE */
  int ns,maxI,minI,radiusM1M1;
  float scaleF;
  forPifHead_t thisHead;
  fString_t rWrite,inFile,forName;
  BYTE *outBuf,*tmpByte_buffer;
  FILE *fOut;

  CK(outBuf = (BYTE *)malloc(sizeof(BYTE)*numCols*numRows));
  CK(tmpByte_buffer = (BYTE *)malloc(sizeof(BYTE)*numCols*numRows));
  pif_init_head(&thisHead);
  thisHead.luNum = 4;
  rWrite.addr = malloc(sizeof(char)*4);
  rWrite.length = 4;
  forName.addr = malloc(sizeof(char)*64);
  forName.length = 64;
  sprintf(rWrite.addr,"w");
  sprintf(forName.addr,"rotflt_fimg.pif");

  pif_open_(forName.addr, &(thisHead.luNum), rWrite.addr);
	  
  /* Pick appropriate statistics calculation based
   * on data type we have */
  thisHead.numCols = numCols;
  thisHead.numRows = numRows;
  thisHead.numSecs = 1;
  
  thisHead.mode = 6;

  ns = 1;
  radiusM1M1=  (2 * radiusM1) - 1;
  calcMapStats(rho,
	       &thisHead.img_bkg, 
	       &thisHead.datar4_min,
	       &thisHead.datar4_max, 
	       &thisHead.datar4_avg, 
	       &thisHead.datar4_std,
	       &radiusM1M1,
	       &radiusM1M1,
	       &ns);
  thisHead.datar4_max  += 1.0;
  thisHead.datar4_min  -= 1.0;
  scaleF = 225.0 / ((thisHead.datar4_max) - (thisHead.datar4_min));
  maxI = minI = 128;
  for(i=0;i<numRows;i++)
    for(j=0;j<numCols;j++)
      {
	outBuf[((i*numCols)+j)] = (BYTE)((rho[((i*radiusM1M1)+j)] - 
					  thisHead.datar4_min)*scaleF);
 	minI = MIN(minI,outBuf[((i*numCols)+j)]);
	maxI = MAX(maxI,outBuf[((i*numCols)+j)]);
      }
  calcByteStats(outBuf,
	    &thisHead.img_bkg, 
	    &thisHead.datar4_min,
	    &thisHead.datar4_max, 
	    &thisHead.datar4_avg, 
	    &thisHead.datar4_std,
	    numCols, numRows, numCols);
  thisHead.datai4_max= (int)thisHead.datar4_max;
  thisHead.datai4_min= (int)thisHead.datar4_min;
    
  pif_write_gh(&thisHead);
  
  thisHead.imgNum = 1;
  
  pif_write_dh(&thisHead);
  thisHead.imgOffset += 512;  /* Just wrote the header */
  
  tsbworld_byte(outBuf,
		tmpByte_buffer,
		numCols,
		numCols,numRows);
  pif_write_byte_image_(&(thisHead.luNum),
			tmpByte_buffer,
			&(thisHead.imgOffset),
			&(thisHead.numCols),
			&(thisHead.numRows),
			&(thisHead.numSecs));

  free(forName.addr);
  free(rWrite.addr);
  pif_close(&(thisHead.luNum));


  fOut = fopen("rotflt_fimg.raw","w");
  fwrite(outBuf, sizeof(BYTE),(numRows*numCols),fOut);
  fclose(fOut);

  sprintf(outStr,"RAW file stats:\n\
\n\
     Filename= [rotflt_fimg.raw]\n\
\n\
     Frames= [%d]\n\
     Width = [%d]\n\
     Height= [%d]\n\
\n\
     8 bit byte pixels\n\
     NO header\n\
\n\
     Min[%d]\n\
     Max[%d]\n",1,(numCols),numRows,
	  minI,maxI);
  askThem(outStr,FALSE,XmDIALOG_INFORMATION);
  
  askThem("File rotflt_fimg.pif written to disk.",
	  FALSE,XmDIALOG_INFORMATION);

  free(outBuf);
  free(tmpByte_buffer);

  
}/* ROTFLT_WRITE_ACTIVATE */




/* ************************************************************
 * rotflt_create() - Remember the rotflt screen for use later.  
 */
void rotflt_create(Widget widget, int *tag, 
		   XmAnyCallbackStruct *callback_data)
{/* ROTFLT_CREATE */
  
  /* Remember the widget number for later */
  if(*tag==ROTFLTDRAW)
    rotfltDrawWidget= widget;
  else if(*tag==ALL)
    rotfltWidget= widget;
  else if(*tag==YSCALE)
    yscaleScrollW= widget;
  else if(*tag==NANNULISCROLL)
    nannuliScrollW= widget;
  else if(*tag==NANNULI)
    nannuliLabelW= widget;
  else if(*tag==YORIG)
    {/* yorig */
      yorigTextW= widget;
      handle_cr(yorigTextW);
    }/* yorig */
  else if(*tag==XORIG)
    {/* xorig */
      xorigTextW= widget;
      handle_cr(xorigTextW);
    }/* xorig */
  else if(*tag==FFTSTEP)
    {/* fft step */
      fftStepTextW= widget;
      handle_cr(fftStepTextW);
    }/* fft step */
  else if(*tag==PHI0SCROLL)
    phi0ScrollW= widget;
  else if(*tag==PHI0)
    phi0LabelW= widget;
  else if(*tag==JZERO)
    jZerotoggleW= widget;
  else if(*tag==BACKT)
    backTtoggleW= widget;
  else if(*tag==NROT)
    nrotLabelW= widget;
  else if(*tag==LGEND)
    lgendLabelW= widget;
  else if(*tag==LGENDSCROLL)
    lgendScrollW= widget;
  else if(*tag==LOGROTFLT)
    logROTFLTtoggleW= widget;
  else if(*tag==LINROTFLT)
    linROTFLTtoggleW= widget;
  else if(*tag==TRYIT)
    tryitButtonW= widget;
  else if(*tag==REFINE)
    refineButtonW= widget;
  else if(*tag==REFINERESET)
    refineResetButtonW= widget;
  else if(*tag==XmARROW_UP)
    {rotfltImgnumArrowUpW = widget;
    handle_cr(widget);}
  else if(*tag==XmARROW_DOWN)
    {rotfltImgnumArrowDownW = widget;
    handle_cr(widget);}

}/* ROTFLT_CREATE */



/*************************************************************
 * DrawROTFLTWindow() - Redraw window e.g. on expose events
 * or whenever it's prudent.
 */
void DrawROTFLTWindow()
{/* DrawROTFLTWindow */
  int draw_width, draw_height, drawX,drawY,drawXfin,drawYfin;
  float xcen,ycen,thisMin,thisMax,scaleFactor;
  int nx,ny,ik,ix,iy,ih,ix1,ix2,iy1,iy2;
  int h1,k1,h2,k2,lastVal,lastX,lastXaccum;
  float x1,x2,y1,y2;
  char numStr[128];

  draw_width = 512;
  draw_height = 512;
  if(XtIsManaged(rotfltscreen))
    XPutImage(dsply, rotflt_window1, rotflt_gc, rotflt_image1, 0, 0,
	      0, 0,  draw_width,  draw_height);
  XFlush (dsply);

  /* Now draw axes */
  XDrawLine(dsply, XtWindow(rotfltDrawWidget), image_gcWHITE,
	    25,0,25,512);
  XDrawLine(dsply, XtWindow(rotfltDrawWidget), image_gcWHITE,
	    0,485,512,485);

  if(  logROTFLTplot )
    {/* Draw markers for log plot */
      drawX = 1;
      drawY = 10;
      sprintf(numStr,"100");
      XDrawString(dsply, XtWindow(rotfltDrawWidget), image_gcRED,
		  drawX, drawY, numStr, strlen(numStr));
      XFlush (dsply);
      for(i=0;i<(yFactor);i++)
	{/* Horizontal grid */
	  int offset;
	  offset = 475 / (yFactor);
	  XDrawLine(dsply, XtWindow(rotfltDrawWidget), image_gcGREY[150],
		    25,((i+1)*offset),512,((i+1)*offset));
	  drawX = 1;
	  drawY = ((i+1)*offset);
	  sprintf(numStr,"%.2f",(float)(100.0 / (pow(10.0,(i+1)))));
	  XDrawString(dsply, XtWindow(rotfltDrawWidget), image_gcRED,
		      drawX, drawY, numStr, strlen(numStr));
	  XFlush (dsply);
	}/* Horizontal grid */
    }/* Draw markers for log plot */
  else
    {/* Draw markers for linear */
      for(i=0;i<4;i++)
	{/* Horizontal grid */
	  int offset;
	  float maxVal, tickVal;
	  maxVal = 100.0/(float)yFactor;
	  tickVal = maxVal / 5.0;
	  drawX = 1;
	  drawY = 10;
	  sprintf(numStr,"%.1f",maxVal);
	  XDrawString(dsply, XtWindow(rotfltDrawWidget), image_gcRED,
		      drawX, drawY, numStr, strlen(numStr));
	  XFlush (dsply);
	  offset = 475 / (5);
	  XDrawLine(dsply, XtWindow(rotfltDrawWidget), image_gcGREY[150],
		    25,((i+1)*offset),512,((i+1)*offset));
	  drawX = 1;
	  drawY = ((i+1)*offset);
	  sprintf(numStr,"%.1f",(float)((maxVal-tickVal)-((float)i*tickVal)));
	  XDrawString(dsply, XtWindow(rotfltDrawWidget), image_gcRED,
		      drawX, drawY, numStr, strlen(numStr));
	  XFlush (dsply);
	}/* Horizontal grid */
    }/* Draw markers for linear */

  /* Vertical grid */
  for(i=0;i<(radiusM1*2*5);i+=5)
    {/* Do all */
      int y,yp1;
      y   = 0;
      if((i%25)==0)
	yp1 = 490;
      else yp1 = 485;
      if(i%50==0)
	XDrawLine(dsply, XtWindow(rotfltDrawWidget), image_gcGREY[150],
		  ((i)+26),(int)y,((i)+26),(int)yp1);
      else if((i%25)==0)
	XDrawLine(dsply, XtWindow(rotfltDrawWidget), image_gcGREY[100],
		  ((i)+26),(int)y,((i)+26),(int)yp1);
    }/* Do all */
  
  /* Now draw tick mark numbers */
  for(i=0;i<(radiusM1*2*5);i+=5)
    {/* Do all */
      int drawX, drawY;
      if((i%50)==0)
	{/* Do every tenth */
	  drawX = 20+i;
	  drawY = 500;
	  sprintf(numStr,"%2d",(int)(xFactor*(float)((float)i/5.0)));
	  XDrawString(dsply, XtWindow(rotfltDrawWidget), image_gcRED,
		      drawX, drawY, numStr, strlen(numStr));
	  XFlush (dsply);
	}/* Do every tenth */
    }/* Do all */


  XFlush (dsply);

  /* Now draw power spectrum */
  thisMin = 99999.9;
  thisMax = -99999.9;
  thisMin = HUGE;  /* dgm HUGE instead of 99999.9 */
  thisMax = -HUGE; /* dgm -HUGE instead of -99999.9 */
  for(i=0;i<(radiusM1*2)+1;i++)
    {/* Do all */
      if(logROTFLTplot)
	{/* log */
	  if(powSpec[i] < 0.001)
	    powSpec[i]= 0.001;
	  thisMin = MIN(log(powSpec[i]),thisMin);
	  thisMax = MAX(log(powSpec[i]),thisMax);
	}/* log */
      else
	{/* linear */
	  thisMin = MIN(powSpec[i],thisMin);
	  thisMax = MAX(powSpec[i],thisMax);
	}/* linear */
    }/* Do all */
  if(logROTFLTplot && (logCutoff > 0.0))
    thisMin = log(logCutoff);

  if(thisMax > thisMin)
    scaleFactor = (400.0/(thisMax-thisMin));
  else scaleFactor = 1.0;

  if(sump1 > 0.0)
    {/* Don't do unless we've got a power spectrum */
      drawX = 100;
      drawY = 10;
      sprintf(numStr,"%.2f%% of azimuthal power is %3d-fold.",
	      sump3,nrot);
      XDrawString(dsply, XtWindow(rotfltDrawWidget), image_gcGREEN,
		  drawX, drawY, numStr, strlen(numStr));
      drawX = 100;
      drawY = 25;
      sprintf(numStr,"This plus G-zero is %6.2f%% of total power.", 
	      sump2);
      XDrawString(dsply, XtWindow(rotfltDrawWidget), image_gcGREEN,
		  drawX, drawY, numStr, strlen(numStr));
      drawX = 100;
      drawY = 40;
      sprintf(numStr,"Threshold of noise = %.2f%%", ((1.0/nrot)*100.0));
      XDrawString(dsply, XtWindow(rotfltDrawWidget), image_gcGREEN,
		  drawX, drawY, numStr, strlen(numStr));
    }/* Don't do unless we've got a power spectrum */

  lastVal = 99;
  lastX = -5;
  lastXaccum = 10;
  for(i=0;i<(radiusM1*2*5);i+=5)
    {/* Do all */
      int y,yp1,thisPlotVal,nextVal;
      float x,xp1;
      thisPlotVal = (int)(xFactor * (float)((float)i/5.0));
      nextVal = (int)xFactor;
      if(xFactor < 1.0)
	{/* DO special */
	  if((int)(xFactor * (float)((float)(i+5)/5.0)) > lastVal)
	    nextVal = 1;
	}/* DO special */
      lastVal = thisPlotVal + nextVal;
      if(nextVal > 0)
	{/* For xfactors < 1 this will prevent multiple draws */
	  if(logROTFLTplot)
	    {/* log */
	      x   = MAX(log(powSpec[thisPlotVal]),thisMin);
	      xp1 = MAX(log(powSpec[thisPlotVal+((int)nextVal)]),thisMin);
	      y   = (int)475-(((x - thisMin)*scaleFactor));
	      yp1 = (int)475-(((xp1 - thisMin)*scaleFactor));
	    }/* log */
	  else
	    {/* linear */
	      x   = powSpec[thisPlotVal];
	      xp1 = powSpec[thisPlotVal+((int)nextVal)];
	      y   = (int)475-(yFactor*((x - thisMin)*scaleFactor));
	      yp1 = (int)475-(yFactor*((xp1 - thisMin)*scaleFactor));
	    }/* linear */
	  XDrawLine(dsply, XtWindow(rotfltDrawWidget), image_rotfltgcRED,
		    ((lastX)+26+5),(int)y,((lastX)+26+lastXaccum),(int)yp1);
	  XDrawLine(dsply, XtWindow(rotfltDrawWidget), image_gcBLUE,
		    (0),(int)0,(0),(int)1);
	  lastX = i;
	  lastXaccum = 5;
	}/* For xfactors < 1 this will prevent multiple draws */
      lastXaccum +=5;
    }/* Do all */

}/* DrawROTFLTWindow */




/*******************************************************************
 * rotfltClick() - Every time the mouse button is clicked call
 * here.  
 *            Button1 is the Left mouse button
 *            Button2 is the Middle mouse button
 *            Button3 is the Right mouse button
 */
void rotfltClick(Widget widget, int *tag, 
		 XmDrawingAreaCallbackStruct *draw_data)
{/* ROTFLTCLICK */
  XEvent event;
  int x, y, k, i,j,reDraw, thisOne;
  char ans;

  /* Do appropriate depending on event type */
  if(draw_data != 0)
    {/* Only look at events meant for us */
      event= *(draw_data->event);

      switch (event.type) 
	{/* Find out which event we have */

	case ButtonRelease:
	  x = (int)(event.xbutton.x);      /* Get x,y copords of point pos */
	  y = (int)(event.xbutton.y);

	  break;
	  
	case ButtonPress:
	  /* They are trying to find points */
	  x = (int)(event.xbutton.x);      /* Get x,y copords of point pos */
	  y = (int)(event.xbutton.y);

	  printf("[%d,%d] clicked\n",x,y);
	  
	  break;
	  
	case KeyPress:
	  ans= XLookupKeysym((XKeyEvent *)&event, 0);
	  printf ("Keyed[%c]\n",ans);

	  break;
	  
	default:
	  break;
	}/* Find out which event we have */
    }/* Only look at events meant for us */
  
  XFlush(dsply);            /* Make changes current */

  
}/* ROTFLTCLICK */




/*****************************************************************
 * rotfltMotion() - Whenever the pointer moves across our ROTFLT
 * window call here.  
*/
void rotfltMotion(Widget widget, caddr_t unused, 
		  register XMotionEvent *pEvent)
{/* rotfltMotion */
  int x, y,lastDraw;
  int i, j, k, thisX,thisY;

  /* Make all changes current */
  XFlush(dsply);               

  switch (pEvent->type) 
    {/* Find out which event we have */
	  
    case MotionNotify:
	      
      /* Get coords of pointer */
      x = (int)(pEvent->x);     
      y = (int)(pEvent->y);
      
      break;
      
    default:
      break;
    }/* Find out which event we have */
  
  XFlush(dsply);              /* Make changes current */
  

}/* rotfltMotion */





/* ************************************************************
 * scaleText_changed() - Whenever text widgets change
 * call here.
 */
void scaleText_changed(Widget widget, int *tag, 
		       XmAnyCallbackStruct *callback_data)
{/* scaleText_CHANGED */
  char *scaleText;
  int i;

  scaleText = XmTextGetString(widget); 
  if(*tag==XSCALE)
    {/* x */
      /*       xFactor = (float)strtod(scaleText,0); */
    }/* x */
  else if (*tag==YSCALE)
    {/* y */
    }/* y */
  else if (*tag==FFTSTEP)
    {/* fft_step */
      fft_step = (float)strtod(scaleText,0);
      if(fft_step < 0.0)
	{/* Invalid */
	  askThem("Cannot have an FFT step less than or equal to zero",
		  FALSE,XmDIALOG_INFORMATION);
	  fft_step = fabs(fft_step);
	  if(fft_step == 0.0)
	    fft_step = 1.0;
	  sprintf(outStr,"%.2f",fft_step);
	  XmTextSetString(fftStepTextW, outStr);
	  XFlush(dsply);
	}/* Invalid */
    }/* fft_step */

  XtFree(scaleText);
  
}/* scaleText_CHANGED */





/* ************************************************************
 * fft_step_changed() - Whenever fft_step text widget gets
 * a carriage return event call here.
 */
void fft_step_changed(Widget widget, int *tag, 
		      XmAnyCallbackStruct *callback_data)
{/* FFT_STEP_CHANGED */
  XmString simpleString;
  char *scaleText;
  int thisMaxInt,thisMinInt,sliderSize;

  scaleText = XmTextGetString(widget); 

  if(widget==fftStepTextW)
    {/* FFTSTEP widget */
      
      fft_step = (float)strtod(scaleText,0);
      if(fft_step < 0.0)
	{/* Invalid */
	  askThem("Cannot have an FFT step less than or equal to zero",
		  FALSE,XmDIALOG_INFORMATION);
	  fft_step = fabs(fft_step);
	  sprintf(outStr,"%.2f",fft_step);
	  XmTextSetString(fftStepTextW, outStr);
	  XFlush(dsply);
	}/* Invalid */
      else if (fft_step == 0.0)
	{/* Invalid */
	  askThem("Cannot have an FFT step equal to zero\n\
I'm resetting this to 1.0.",
		  FALSE,XmDIALOG_INFORMATION);
	  fft_step = 1.0;
	  sprintf(outStr,"%.2f",fft_step);
	  XmTextSetString(fftStepTextW, outStr);
	  XFlush(dsply);
	}/* Invalid */
      XtFree(scaleText);
      
      /**********************************************
       * We should be able to bump up the nannuli 
       * if we've got a fft_step less than 1.0 (or
       * also we should have to reduce nannuli if
       * we've got an fft_step > 1.0.  The fortran
       * code doesn't handle this very well yet so let's
       * just not use this until we can get emrotflt_main()
       * to do the right thing.
       * nannuli = MIN(nannuli,thisMaxInt);
       **********************************************/
      thisMaxInt = (int)(idim1Rotflt / 2) - 1;
      thisMaxInt = (int)((float)thisMaxInt / fft_step)-1;
      thisMinInt = (int)(1);
      sliderSize = 5;
      nannuli = MIN(nannuli,thisMaxInt);
      XtSetArg(arglst[0], XmNmaximum, (int)(thisMaxInt+(float)sliderSize));
      XtSetArg(arglst[1], XmNminimum, (int)(thisMinInt));
      XtSetArg(arglst[2], XmNsliderSize, (int)(sliderSize));
      XtSetArg(arglst[3], XmNvalue, (int)(nannuli));
      if(thisMaxInt > 200)
	XtSetArg(arglst[4], XmNpageIncrement, (int)(thisMaxInt/100));
      else if(thisMaxInt > 20)
	XtSetArg(arglst[4], XmNpageIncrement, (int)(thisMaxInt/10));
      else 
	XtSetArg(arglst[4], XmNpageIncrement, (int)(thisMaxInt));
      XtSetValues(nannuliScrollW, arglst, 5);
      
      sprintf(outStr,"%d",(nannuli));
      createSimpleString(&simpleString,outStr);
      XtSetArg(arglst[0], XmNlabelString, simpleString);
      XtSetArg(arglst[1], XmNrecomputeSize, FALSE );
      XtSetValues(nannuliLabelW, arglst, 2); 
      XFlush(dsply);
      XmStringFree(simpleString); 
      XFlush(dsply);
    }/* FFTSTEP widget */
  else if(widget==xorigTextW)
    {/* xorig text */
    }/* xorig text */
  else if(widget==yorigTextW)
    {/* yorig text */
    }/* yorig text */
  
  /*   if(backTransROTFLT) */
    rotflt_tryit_activate(0,0,0);

}/* FFT_STEP_CHANGED */





/* ************************************************************
 * doTheRotfltArrow() - Called from rotflt_arrow_activate and 
 * rotflt_arrow23_activate().  Do the work required to change
 * the xorig/yorig concept.
 */
void doTheRotfltArrow(Widget widget, float speedFactor)
{/* DOTHEROTFLTARROW */
  char newVal[22];
  char *thisStr;
  float thisFloat;

  if(widget==xorigUpArrowW)
    {/* Up */
      thisStr = XmTextGetString(xorigTextW);
      thisFloat = (float)strtod(thisStr,0);
      thisFloat += speedFactor;
      sprintf(outStr,"%.2f",thisFloat);
      XmTextSetString(xorigTextW,outStr);
      XFlush(dsply);
      XtFree(thisStr);
    }/* Up */
  else if(widget==xorigDownArrowW)
    {/* Dopwn */
      thisStr = XmTextGetString(xorigTextW);
      thisFloat = (float)strtod(thisStr,0);
      thisFloat -= speedFactor;
      sprintf(outStr,"%.2f",thisFloat);
      XmTextSetString(xorigTextW,outStr);
      XFlush(dsply);
      XtFree(thisStr);
    }/* Down */
  else if(widget==yorigUpArrowW)
    {/* Up */
      thisStr = XmTextGetString(yorigTextW);
      thisFloat = (float)strtod(thisStr,0);
      thisFloat += speedFactor;
      sprintf(outStr,"%.2f",thisFloat);
      XmTextSetString(yorigTextW,outStr);
      XFlush(dsply);
      XtFree(thisStr);
    }/* Up */
  else if(widget==yorigDownArrowW)
    {/* Dopwn */
      thisStr = XmTextGetString(yorigTextW);
      thisFloat = (float)strtod(thisStr,0);
      thisFloat -= speedFactor;
      sprintf(outStr,"%.2f",thisFloat);
      XmTextSetString(yorigTextW,outStr);
      XFlush(dsply);
      XtFree(thisStr);
    }/* Down */
  else printf("Unknown widget value [%d].\n",widget);

  /*   if(backTransROTFLT) */
    rotflt_tryit_activate(0,0,0);

}/* DOTHEROTFLTARROW */



/* ************************************************************
 * rotflt_arrow_activate() - This rtn is for the first mouse 
 * button only on the arrow widget.
 */
void rotflt_arrow_activate(Widget widget, int *tag, 
			   XmPushButtonCallbackStruct *callback_data)
{/* rotflt_arrow_activate */
  float speedFactor;

  speedFactor = 1.0;
  
  doTheRotfltArrow(widget, speedFactor);

}/* rotflt_arrow_activate */




/* ************************************************************
 * rotflt_arrow23_activate() - This rtn is for the second and 
 * third mouse buttons on the arrow widget.
 */
void rotflt_arrow23_activate(Widget widget, XEvent *event)
{/* rotflt_arrow23_activate */
  XEvent thisEvent;
  float speedFactor;

  thisEvent= *(event);

  /* Get which button */
  if((thisEvent.xbutton.button == Button1))
    speedFactor = 1.0;
  if((thisEvent.xbutton.button == Button2))
    speedFactor = 0.1;
  if((thisEvent.xbutton.button == Button3))
    speedFactor = 0.01;
  
  doTheRotfltArrow(widget, speedFactor);

}/* rotflt_arrow23_activate */






/* ************************************************************
 * rotflt_forceTryit_activate() - Try to Rotationally Filter this FFT.
 */
void rotflt_forceTryit_activate(Widget widget, int *tag, 
				XmAnyCallbackStruct *callback_data)
{/* ROTFLT_FORCETRYIT_ACTIVATE */
  int saveVal;

  saveVal = backTransROTFLT;
  backTransROTFLT = TRUE;  /* FOrce it */
  rotflt_tryit_activate(0,0,0);
  backTransROTFLT = saveVal;

}/* ROTFLT_FORCETRYIT_ACTIVATE */




/* ************************************************************
 * rotflt_tryit_activate() - Try to Rotationally Filter this FFT.
 */
void rotflt_tryit_activate(Widget widget, int *tag, 
			   XmAnyCallbackStruct *callback_data)
{/* ROTFLT_TRYIT_ACTIVATE */
  float thisMin, thisMax, scaleFactor;
  float elipX, elipY, 
        ascal, bscal, 
        del_br, del_lr, 
        twodd, fft_origx, fft_origy,
        pow_scale, xorigSave,yorigSave;
  int irad, max_bo, 
      radiusM1M1,
      lg_start;
  int iTnumCols,imgTnumCols;
  int numPixels;
  int x,y,m,l,n,thisInt,yOut,yOff;
  float *thisBuffer;
  unsigned short *tempS_buffer;
  XmString simpleString;
  char *thisStr;

/*  Get these from interface now 
  nrot = 5;                    
  nannuli = 30;        
  fft_step = 1.0;
  phi0= 0.0;
  jzero= 0;
*/

  /* Give them feed_back on what it actually is */
  sprintf(outStr,"%.2f",fft_step);
  XmTextSetString(fftStepTextW, outStr);
  XFlush(dsply);

  phi0 = omega;   /* What was gotten from the interface? */

  radiusM1= (purdueHeader.numCols/2);

  /* Get del_lr from scroll eventually */
  del_lr = 1.0;    /* Don't Try and calc in rotflt_info() */

  del_br = 1.0;    /* Already calculated in rotflt_info() */
  ascal = 1.0;
  bscal = 0.0;
  elipX = 1.0;
  elipY = 1.0;

  /* Only init once!!!!  Use last xorig and origY each successive time!!! */
  /* oldway:  if(iopt[0]==1)  */
  if(newOrigin)
    {/* Recalc the origin */
      xorig = (purdueHeader.numCols/2);       /* FFT origx and origy */
      yorig = (purdueHeader.numCols/2);      
      sprintf(outStr,"%.2f",xorig);
      XmTextSetString(xorigTextW,outStr);
      sprintf(outStr,"%.2f",yorig);
      XmTextSetString(yorigTextW,outStr);
      newOrigin = FALSE;
      newStart = TRUE;
    }/* Recalc the origin */
  else
    {/* Get from last time */
      newStart = FALSE;
      thisStr = XmTextGetString(xorigTextW);
      xorig = (float)strtod(thisStr,0);
      xorig -= 1.0;  /* Xorig is one off */
      XtFree(thisStr);
      thisStr = XmTextGetString(yorigTextW);
      yorig = (float)strtod(thisStr,0);
      XtFree(thisStr);
      /* Let's put it in tsb space */
      yorig = purdueHeader.numRows - yorig;
    }/* Get from last time */

  pow_scale = 100.0;

  lg_start = 0;         /* Calculated in rotflt_info() */
  /*  lg_end   = 0;  Symmetry Limit */

  max_bo = 0;                   /* Doesn't matter what it's set to - recalced
				 * in rotflt_info_rob() */
  
  if(fileType==PACKEDPORTABLE)
    {
      thisBuffer= (float *)packImg[(imgNum-1)]; 
      numCols = purdueHeader.numCols;
      numRows = purdueHeader.numRows;
    }  
  else
    {/* Scanned */
      thisBuffer= (float *)inCore_buffer;
      numCols = file_width;
      numRows = file_height;
    }/* Scanned */


  /*******************************************************
   * RHO is defined as: 
   *
   *     REAL*4  RHO(2*NLR-1,2*NLR-1)
   *
   * in robemF.f so make sure we've got enough allocated here to
   * allow for any discrepancies 
   **********************************************************/
  if(rho)
    free(rho);
  CK(rho= (float *)malloc(sizeof(float)*(numCols)*(numRows)*2));
    
  /* Calc zoom */
  calcSizeFactor(zoomFactor);

  /* Stuff our fft buffer */
  for(i=0;i<fftSize;i++)
    for(j=0;j<(fftSize*2);j++)
      {/* Init all */
	tempRotflt_buffer[(i*(fftSize*2))+j]= 
	  fftImage_buffer2[(i*(fftSize*2))+j];
      }/* Init all */

  for(i=0;i<1000;i++)
    powSpec[i]= 0.0;

  /* These may get corrupted */
  xorigSave= xorig;
  yorigSave= yorig;

  emrotflt_main_(powSpec,
		 &elipX,&elipY,
		 &irad,
		 &idim1Rotflt,
		 &phi0,&ascal,&bscal,
		 &twodd,&jzero,&radiusM1,&max_bo,
		 &fft_step,&nannuli,
		 &del_br,
		 rho,
		 &del_lr,
		 iopt,
		 &lg_start,&lg_end,
		 &xorig,&yorig,
		 &pow_scale, &nrot,
		 &numCols, &numRows,
		 &sump1, &sump2, &sump3,
		 &fft_origx, &fft_origy, &winScale, sigma,
		   tempRotflt_buffer, 
		 /* fftImage_buffer2, */
		 &newStart, &sigMax, &sigMaxTot);
  
  sprintf(outStr,"%.3f ",winScale);
  createSimpleString(&simpleString,outStr);
  XtSetArg(arglst[0], XmNlabelString, simpleString);
  XtSetArg(arglst[1], XmNrecomputeSize, FALSE );
  XtSetValues(refineResetLabelW, arglst, 2); 
  XFlush(dsply);
  XmStringFree(simpleString); 
  XFlush(dsply);

  /* Give feedback on nannuli */
  XtSetArg(arglst[0], XmNvalue, (int)(nannuli));
  XtSetValues(nannuliScrollW, arglst, 1);
  sprintf(outStr,"%d",(nannuli));
  createSimpleString(&simpleString,outStr);
  XtSetArg(arglst[0], XmNlabelString, simpleString);
  XtSetArg(arglst[1], XmNrecomputeSize, FALSE );
  XtSetValues(nannuliLabelW, arglst, 2); 
  XFlush(dsply);
  XmStringFree(simpleString); 
  XFlush(dsply);


  if(iopt[0]==1)
    {/* origin refined */
      /* Put it back in XWindows space (in TSB space above) */
      yorig = purdueHeader.numRows - yorig;
      /* Why do this here ??? */
      xorig = xorig + (1.0+fft_origx);
      /* It's in TSB space */
      yorig = yorig - fft_origy;
      sprintf(outStr,"%.2f",xorig);
      XmTextSetString(xorigTextW,outStr);
      sprintf(outStr,"%.2f",yorig);
      XmTextSetString(yorigTextW,outStr);
    }/* origin refined */
  else
    {/* Just use our originals */
      xorig= xorigSave;
      yorig= yorigSave;
      /* Put it back in tsb space */
      yorig = purdueHeader.numRows - yorig;
    }/* Just use our originals */

  firstRotFlt = FALSE;

  thisMin = 99999.9;
  thisMax = -99999.9;
  thisMin = HUGE;   /* dgm HUGE instead of 99999.9 */
  thisMax = -HUGE;  /* dgm -HUGE instead of -99999.9 */
  radiusM1M1=  (2 * radiusM1) - 1;
  file_width = orig_file_width + radiusM1M1;
  for(i=0;i<radiusM1M1;i++)
    for(j=0;j<radiusM1M1;j++)
      {/* Do all */
	thisMin = MIN(rho[((i*radiusM1M1)+j)],thisMin);
	thisMax = MAX(rho[((i*radiusM1M1)+j)],thisMax);
      }/* Do all */
  if(thisMax > thisMin)
    scaleFactor = ((tryColors-14.0)/(thisMax-thisMin));
  else scaleFactor = 1.0;
  if(scaleFactor < minRotfltScaleF)
    minRotfltScaleF = scaleFactor;
  if(!stretchRotflt)  
    scaleFactor = minRotfltScaleF;

  /* Calc zoom */
  calcSizeFactor(zoomFactor);

  if(backTransROTFLT) 
    {/* Put rotflt image beside original */

      /* Redraw original in case it's got a crosshair on it or etc... */
      virtual_read_activate(-99,-99);
      
      /* 1 Put rotflt image beside original here */
      for(y=0;y<radiusM1M1;y++)
	{/* Do lines */
	  
	  if(sizeFactor > 1.0)
	    {/* We are zooming out */
	      for(m=0;m<(int)sizeFactor;m++)
		{/* Do all zoomed lines */
		  yOff = (buffer_width * (((int)(y*sizeFactor)) + m)) ;
		  for(l=0;l<radiusM1M1;l++)
		    {/* Do all cols */
		      yOut = yOff + (radiusM1M1*sizeFactor) + 
			((int)(l*sizeFactor));
		      for(n=0;n<(int)sizeFactor;n++)
			{
			  if(hiColor)
			    {/* XPutPixel here ??? */
			      XPutPixel(image_image1,
					(radiusM1M1*sizeFactor)+
					(l*sizeFactor)+n, 
					(int)(y*sizeFactor)+m,
					pixels_return[(int)((
					     rho[((y*radiusM1M1)+l)]-
							     thisMin)*
							    scaleFactor)+6]
					);
			    }/* XPutPixel here ??? */
		      else
			*(image_buffer1+yOut+n)= 
			  (unsigned char)
			  pixels_return[(int)((rho[((y*radiusM1M1)+l)]-
					       thisMin)*
					      scaleFactor)+6];
			}
		    }/* Do all cols */
		}/* Do all zoomed lines */
	    }/* We are zooming out */
	  else
	    {/* We are zooming IN */
	      yOff = (buffer_width * ((int)(y*sizeFactor))) ;
	      for(l=0;l<radiusM1M1;l+=(int)recipSizeFactor)
		{/* Do all cols */
		  yOut = yOff + (radiusM1M1*sizeFactor) + 
		    ((int)(l*sizeFactor));
		  thisInt= 0;
		  for(n=0;n<(int)recipSizeFactor;n++)
		{
		  if(hiColor)
		    {/* XPutPixel here ??? */
		      thisInt += (int)((rho[((y*radiusM1M1)+l+n)]-
					thisMin)
				       *scaleFactor);
		    }/* XPutPixel here ??? */		    
		  else
		    thisInt += (BYTE)pixels_return[(int)((
					     rho[((y*radiusM1M1)+l+n)]-
							  thisMin)
							 *scaleFactor)+6];
		}
		  if(hiColor)
		    XPutPixel(image_image1,
			      (l*sizeFactor)+(radiusM1M1*sizeFactor), 
			      (int)(y*sizeFactor),
			      pixels_return[(int)(thisInt/
						  (int)recipSizeFactor)+6]);
		  else 
		    *(image_buffer1+yOut)= (unsigned char)(thisInt/
						 (int)recipSizeFactor);
		}/* Do all cols */
	    }/* We are zooming IN */
	}/* Do lines */
    }/* Put rotflt image beside original */
  
  
  if(rotfltResiduals)
    {/* Put up contour plot of residuals */
      thisMin = thisMax = sigma[0];
      for(y=0;y<17;y++)
	{/* Do lines */
	  for(m=0;m<17;m++)
	    {/* Do all zoomed lines */
	      thisMax = MAX(thisMax,sigma[(y*17)+m]);
	      thisMin = MIN(thisMin,sigma[(y*17)+m]);
	    }/* Do all zoomed lines */
	}/* Do lines */
      if((thisMax-thisMin)==0.0)
	scaleFactor = 1.0;
      else scaleFactor = (tryColors-9) / (thisMax -thisMin);
      numPixels = (int)((radiusM1M1 *sizeFactor) / 17);

      for(y=0;y<(17*numPixels);y+=numPixels)
	{/* Do lines */
	  for(m=0;m<(17*numPixels);m+=numPixels)
	    {/* Do all zoomed lines */
	      for(l=0;l<numPixels;l++)
		{/* y */
		  yOff = (buffer_width * (y+l)) ;
		  yOut = yOff + (radiusM1M1*sizeFactor) + (radiusM1M1*
							   sizeFactor);
		  for(n=0;n<numPixels;n++)
		    {
		      if(hiColor)
			{
			  XPutPixel(image_image1,
				    (radiusM1M1*sizeFactor) + 
				    (radiusM1M1*sizeFactor)+m+n, 
				    (int)(y+l),
				    pixels_return[
					  (int)((int)(
				          ((sigma[(((y/numPixels)*17)+
					            (m/numPixels))] - 
						   thisMin) * scaleFactor)
						      )
						+6)
					  ]);
			}
		      else
			*(image_buffer1+yOut+m+n)= 
			  (unsigned char)
			    pixels_return[
					  (int)((int)(
				             ((sigma[(((y/numPixels)*17)+
					   	       (m/numPixels))] - 
							thisMin) * scaleFactor)
						      )
						+6)
					  ];
		    }
		}/* y */
	    }/* Do all zoomed lines */
	}/* Do lines */
    }/* Put up contour plot of residuals */
  

  if(rotfltCrosshair)
    {/* Draw crosshair in image */
      float y,x;

      y = (float)numRows - (float)yorig;  
      x = (float)xorig;

      if(sizeFactor > 1.0)
	{/* We are zooming out */
	  for(m=0;m<(int)sizeFactor;m++)
	    {/* Do all zoomed lines */
	      yOff = (buffer_width * (((int)(y*sizeFactor)) + m)) ;
	      for(l=0;l<radiusM1M1;l++)
		{/* Do all cols */
		  yOut = yOff + ((int)(l*sizeFactor));
		  for(n=0;n<(int)sizeFactor;n++)
		    {
		      if(hiColor)
			{
			  XPutPixel(image_image1,
				    (int)(l*sizeFactor)+n, 
				    (int)(y*sizeFactor)+m,
				    pixels_return[RED]);
			}
		      else
			*(image_buffer1+yOut+n)= 
			  (unsigned char)
			    pixels_return[RED];
		    }
		}/* Do all cols */
	      for(l=0;l<radiusM1M1;l++)
		{/* Do all rows */
		  yOff = (buffer_width * (((int)(l*sizeFactor)) + m)) ;
		  yOut = yOff + ((int)(xorig*sizeFactor));
		  for(n=0;n<(int)sizeFactor;n++)
		    {
		      if(hiColor)
			{
			  XPutPixel(image_image1,
				    (int)(x*sizeFactor)+n, 
				    (int)(l*sizeFactor)+m,
				    pixels_return[RED]);
			}
		      else
			*(image_buffer1+yOut+n)= 
			  (unsigned char)
			    pixels_return[RED];
		    }
		}/* Do all rows */
	    }/* Do all zoomed lines */
	}/* We are zooming out */
      else
	{/* We are zooming IN */
	  yOff = (buffer_width * ((int)(y*sizeFactor))) ;
	  for(l=0;l<radiusM1M1;l+=(int)recipSizeFactor)
	    {/* Do all cols */
	      yOut = yOff + ((int)(l*sizeFactor));
	      if(hiColor)
		{
		  XPutPixel(image_image1,
			    (l*sizeFactor), (((int)(y*sizeFactor))),
			    pixels_return[RED]);
		}
	      else
		*(image_buffer1+yOut)= (BYTE)pixels_return[RED];
	    }/* Do all cols */
	  for(l=0;l<radiusM1M1;l+=(int)recipSizeFactor)
	    {/* Do all rows */
	      yOff = (buffer_width * ((int)(l*sizeFactor))) ;
	      yOut = yOff + ((int)(xorig*sizeFactor));
	      if(hiColor)
		{
		  XPutPixel(image_image1,
			    (xorig*sizeFactor), (((int)(l*sizeFactor))),
			    pixels_return[RED]);
		}
	      else
		*(image_buffer1+yOut)= (BYTE)pixels_return[RED];
	    }/* Do all rows */
	}/* We are zooming IN */

    }/* Draw crosshair in image */

  /* We may not want to refine the origin next time ... */
  iopt[0] = 0;  /* Make them force it again */

  XPutImage (dsply, image_window1, image_gc, image_image1, 0, 0,
	     0, 0, (file_width*sizeFactor), (file_height*sizeFactor));
  XFlush (dsply);
  DrawWindow(); 

 noRotFlt:;

  if(widget != (Widget)-99)
    DrawROTFLTWindow();


}/* ROTFLT_TRYIT_ACTIVATE */




/* ************************************************************
 * rotflt_change() - Whenever rotflt scroll widgets changed
 * call here.
 */
void rotflt_change(Widget widget, int *tag, 
		   XmScrollBarCallbackStruct *callback_data)
{/* rotflt_CHANGE */
  XEvent event;
  char *rotflt,outStr[64];
  XmString simpleString;
  int i;

  if(!inited)
    goto notInited;

  if(*tag==NANNULI)
    {/* */ 
      nannuli = (int)callback_data->value;
      nannuli = MAX(nannuli,1);
      sprintf(outStr,"%d",(nannuli));
      createSimpleString(&simpleString,outStr);
      XtSetArg(arglst[0], XmNlabelString, simpleString);
      XtSetArg(arglst[1], XmNrecomputeSize, FALSE );
      XtSetValues(nannuliLabelW, arglst, 2); 
      XFlush(dsply);
      XmStringFree(simpleString); 
      XFlush(dsply);
      if((nannuli*fft_step) >= (idim1Rotflt/2))
	{/* Reset fft_step */
	  printf("Resetting fft_step\n");
	  fft_step = (float)(idim1Rotflt/2)/(float)(nannuli+1); /* +1 to make sure */
	  sprintf(outStr,"%.2f",fft_step);
	  XmTextSetString(fftStepTextW, outStr);
	  XFlush(dsply);
	}/* Reset fft_step */
      /*      if(backTransROTFLT) */
	rotflt_tryit_activate(0,0,0);
    }/* */
  else if(*tag==LGEND)
    {/* */ 
      lg_end = (int)callback_data->value;
      if(lg_end < (nrot))
	{/* ohoh */
	  lg_end = nrot;
	  XtSetArg(arglst[0], XmNvalue, (int)(lg_end));
	  XtSetValues(lgendScrollW, arglst, 1);
	  XFlush(dsply);
	}/* ohoh */
      sprintf(outStr,"%d",(lg_end));
      createSimpleString(&simpleString,outStr);
      XtSetArg(arglst[0], XmNlabelString, simpleString);
      XtSetArg(arglst[1], XmNrecomputeSize, FALSE );
      XtSetValues(lgendLabelW, arglst, 2); 
      XFlush(dsply);
      XmStringFree(simpleString); 
      XFlush(dsply);
      /*if(backTransROTFLT) */
	rotflt_tryit_activate(0,0,0);
    }/* */
  else if(*tag==NROT)
    {/* */ 
      nrot = (int)callback_data->value;
      sprintf(outStr,"%d",(nrot));
      createSimpleString(&simpleString,outStr);
      XtSetArg(arglst[0], XmNlabelString, simpleString);
      XtSetArg(arglst[1], XmNrecomputeSize, FALSE );
      XtSetValues(nrotLabelW, arglst, 2); 
      XFlush(dsply);
      XmStringFree(simpleString); 
      XFlush(dsply);
      if(lg_end < nrot)
	{/* ohoh */
	  lg_end = nrot;
	  XtSetArg(arglst[0], XmNvalue, (int)(lg_end));
	  XtSetValues(lgendScrollW, arglst, 1);
	  XFlush(dsply);
	  sprintf(outStr,"%d",(lg_end));
	  createSimpleString(&simpleString,outStr);
	  XtSetArg(arglst[0], XmNlabelString, simpleString);
	  XtSetArg(arglst[1], XmNrecomputeSize, FALSE );
	  XtSetValues(lgendLabelW, arglst, 2); 
	  XFlush(dsply);
	  XmStringFree(simpleString); 
	  XFlush(dsply);
	}/* ohoh */
      /*if(backTransROTFLT) */
	rotflt_tryit_activate(0,0,0);
    }/* */
  else if (*tag==XSCALE)
    {/* x scale */
      xFactor = (float)callback_data->value;
      if(xFactor < 1.0)
	{/* Do it in 0.0 - 1.0 range */
	  float tmpF;
	  tmpF = 0.9 - ((0.0 - xFactor)/10.0);
	  tmpF = MAX(tmpF,0.1);
	  xFactor = tmpF;
	}/* Do it in 0.0 - 1.0 range */
      DrawROTFLTWindow();
    }/* x scale */
  else if (*tag==YSCALE)
    {/* y scale */
      yFactor = (float)callback_data->value;
      if(logROTFLTplot)
	{/* Now setup yscale scroll and logCutoff */
	  logCutoff = (10.0/ pow(10.0,(yFactor-1)));
	  saveLogYFactor = yFactor;
	}/* Now setup yscale scroll and logCutoff */
      else 
	{/* Linear */
	  if(yFactor < 1.0)  
	    {/* Do it in 0.0 - 1.0 range */
	      float tmpF;
	      tmpF = 0.9 - ((0.0 - yFactor)/10.0);
	      tmpF = MAX(tmpF,0.1);
	      yFactor = tmpF;
	    }/* Do it in 0.0 - 1.0 range */
	  saveLinYFactor = yFactor;
	}/* Linear */
      DrawROTFLTWindow();
    }/* y scale */
  else if (*tag==JZERO)
    {/* jzero */
      if(jzero == 0)
	jzero= 1;
      else jzero = 0;
      /* OLDWAY
       *if(jzero==0)
       *sprintf(outStr,"JZERO included"); 
       *else sprintf(outStr,"JZERO NOT included"); 
       *createSimpleString(&simpleString,outStr);
       *XtSetArg(arglst[0], XmNlabelString, simpleString);
       *XtSetValues(jZerotoggleW, arglst, 1); 
       *XFlush(dsply);
       *XmStringFree(simpleString); 
       *XFlush(dsply);
       *
       *XtSetArg(arglst[0], XmNset, FALSE);
       *XtSetValues(jZerotoggleW, arglst, 1); 
       ********************************/
/*      DrawROTFLTWindow(); */
      /* if(backTransROTFLT) */
	rotflt_tryit_activate(0,0,0);
    }/* jzero */
  else if (*tag==PHI0)
    {/* phi0 */
      omega = (float)callback_data->value;
      sprintf(outStr,"%.0f",(omega));
      createSimpleString(&simpleString,outStr);
      XtSetArg(arglst[0], XmNlabelString, simpleString);
      XtSetArg(arglst[1], XmNrecomputeSize, FALSE );
      XtSetValues(phi0LabelW, arglst, 2); 
      XFlush(dsply);
      XmStringFree(simpleString); 
      XFlush(dsply);
      /* if(backTransROTFLT) */
	rotflt_tryit_activate((Widget)-99,0,0);
	/*      DrawROTFLTWindow(); */
    }/* phi0 */
  else if (*tag==LOGROTFLT  || *tag==LINROTFLT)
    {/* log scale */
      if (*tag==LOGROTFLT)
	logROTFLTplot = TRUE;
      else if (*tag==LINROTFLT)
	logROTFLTplot = FALSE;

      if(logROTFLTplot)
	{/* Now setup yscale scroll and logCutoff */
	  logCutoff = 10.0;
	  yFactor= saveLogYFactor;
	  logCutoff = (10.0/ pow(10.0,(yFactor-1)));
	  XtSetArg(arglst[0], XmNmaximum, (int)(5+1));
	  XtSetArg(arglst[1], XmNminimum, (int)(1));
	  XtSetArg(arglst[2], XmNsliderSize, (int)(1));
	  XtSetArg(arglst[3], XmNvalue, (int)(yFactor));
	  XtSetArg(arglst[4], XmNpageIncrement, (int)(1));
	  XtSetValues(yscaleScrollW, arglst, 5);
	  XFlush(dsply);
	  XtSetArg(arglst[0], XmNset, TRUE);
	  XtSetValues(logROTFLTtoggleW, arglst,1);
	  XtSetArg(arglst[0], XmNset, FALSE);
	  XtSetValues(linROTFLTtoggleW, arglst,1);
	  XFlush(dsply);
	}/* Now setup yscale scroll and logCutoff */
      else
	{/* Now setup yscale scroll */
	  yFactor= saveLinYFactor;
	  XtSetArg(arglst[0], XmNmaximum, (int)(100+5));
	  XtSetArg(arglst[1], XmNminimum, (int)(1));
	  XtSetArg(arglst[2], XmNsliderSize, (int)(5));
	  XtSetArg(arglst[3], XmNvalue, (int)(yFactor));
	  XtSetArg(arglst[4], XmNpageIncrement, (int)(5));
	  XtSetValues(yscaleScrollW, arglst, 5);
	  XFlush(dsply);
	  XtSetArg(arglst[0], XmNset, FALSE);
	  XtSetValues(logROTFLTtoggleW, arglst,1);
	  XtSetArg(arglst[0], XmNset, TRUE);
	  XtSetValues(linROTFLTtoggleW, arglst,1);
	  XFlush(dsply);
	}/* Now setup yscale scroll */

      DrawROTFLTWindow(); 
    }/* log scale */
  else if (*tag==STRETCH)
    {/* stretch */
      stretchRotflt = !stretchRotflt;
      rotflt_tryit_activate(0,0,0);
    }/* stretch */
  else if (*tag==BACKT)
    {/* back */
      backTransROTFLT= !backTransROTFLT;
      if(backTransROTFLT)
	{/* Turn button to insenitive */
	  XtSetArg(arglst[0], XmNsensitive, FALSE);
	  XtSetValues(tryitButtonW, arglst, 1); 
	}/* Turn button to insenitive */
      else
	{/* Turn button sensitive */
	  XtSetArg(arglst[0], XmNsensitive, TRUE);
	  XtSetValues(tryitButtonW, arglst, 1); 
	}/* Turn button sensitive */
      /* if(backTransROTFLT) */
	rotflt_tryit_activate(0,0,0);
    }/* back */
  else if (*tag==REFINE)
    {/* refine */
      XtSetArg(arglst[0], XmNset, FALSE);
      XtSetValues(refineButtonW, arglst, 1); 
      iopt[0]= 1;
      rotflt_tryit_activate(0,0,0);
    }/* refine */
  else if (*tag==ROTFLTCROSS)
    rotfltCrosshair = !rotfltCrosshair;
  else if (*tag==ROTFLTRESID)
    rotfltResiduals = !rotfltResiduals;
  else if (*tag==REFINERESET)
    {
      winScale = 1.0;
      sprintf(outStr,"%.3f ",winScale);
      createSimpleString(&simpleString,outStr);
      XtSetArg(arglst[0], XmNlabelString, simpleString);
      XtSetArg(arglst[1], XmNrecomputeSize, FALSE );
      XtSetValues(refineResetLabelW, arglst, 2); 
      XFlush(dsply);
      XmStringFree(simpleString); 
      XFlush(dsply);
    }


 notInited:;

}/* rotflt_CHANGE */






/*************************************************************
 * doTheCdispZoom() - Do what's necessary to accommodate their 
 * zooming on the cdisp screen.
*/
void doTheCdispZoom(Widget widget, float speedFactor)
{/* DOTHECDISPZOOM */
  float thisSizeFactor,thisF;
  int newX,newY,reason;
  char *zoomStr;

  zoomStr = XmTextGetString(scaleXvalueW);
  thisF = (float)strtod(zoomStr,0);
  XtFree(zoomStr);

  if(speedFactor <= 0.0)
    {/* Button 1 */

      if(widget==cDispZoomArrowUpW)
	thisF *= 2.0;
      else if(widget==cDispZoomArrowDownW)
	thisF *= 0.5;

      /* Now give us a clean power of two */
      thisF = pow(2.0,(double)((int)rwalog2(thisF)));

    }/* Button 1 */
  else
    {/* Buttons 2 or 3 */
      if(widget==cDispZoomArrowUpW)
	thisF += (0.01*speedFactor);
      else if(widget==cDispZoomArrowDownW)
	thisF -= (0.01*speedFactor);
    }/* Buttons 2 or 3 */

  if(thisF != 0.0)
    {/* OK to do */
      x_size = (int)((float)thisF * (float)cdNumCols);

      sprintf(outStr,"%.2f",thisF);
      XmTextSetString(scaleXvalueW,outStr);
      XFlush(dsply);

    }/* OK to do */

  /* Now redraw with new zoominess */
  crysDisp_cr_activate(0,0,0);
  
}/* DOTHECDISPZOOM */



/*************************************************************
 * mainCdisp_zoom_activate() -  This routine is called
 * when the first mouse button is used and causes the XmNactivate
 * callback.
*/
void mainCdisp_zoom_activate(Widget widget, int *tag, 
			     XmPushButtonCallbackStruct *callback_data)
{/* MAINCDISP_ZOOM_ACTIVATE */
  float speedFactor;

  if(inited)
    {/* Only do if it exists */

      doTheCdispZoom(widget, -1.0);
      
    }/* Only do if it exists */

}/* MAINCDISP_ZOOM_ACTIVATE */







/*************************************************************
 * cDisp_zoom_activate() - They want to change the current
 * zoom factor.
*/
void cDisp_zoom_activate(Widget widget, XEvent *event)
{/* CDISP_ZOOM_ACTIVATE */
  float speedFactor;
  XEvent thisEvent;

  if(inited)
    {/* Only do if it exists */

      thisEvent = *event;

      /* Get which cDisp */
      if((thisEvent.xbutton.button == Button1))
	speedFactor = -1.0;;
      if((thisEvent.xbutton.button == Button2))
	speedFactor = 5.0;
      if((thisEvent.xbutton.button == Button3))
	speedFactor = 1.0;

      doTheCdispZoom(widget, speedFactor);

    }/* Only do if it exists */


}/* CDISP_ZOOM_ACTIVATE */





/******************************************************************
 * cDispSymm_changed() -  Ricardo, this is the callback that
 * happens whenever the user makes a choice from the "Plane
 * Group Symmetry" menu on the Crystal Display screen.  All this
 * does is call your routine with the appropriate args and then
 * tries to redraw the crystal.
 *
 *    phaseGroupSymmetry == 1    ->   p1
 *    phaseGroupSymmetry == 2    ->   p2
 *    phaseGroupSymmetry == 3    ->   pm
 *    phaseGroupSymmetry == 4    ->   pg
 *
 *    float *crys_float_image_buffer   - Is the buffer of image data
 *                                       for the unit cell.
 *    int  cdNumCols                   - number of columns in 
 *                                       crys_float_image_buffer buffer
 *    int  cdNumRows                   - number of rows in 
 *                                       crys_float_image_buffer buffer
 */
void cDispSymm_changed(Widget widget, int *tag, 
		       XmAnyCallbackStruct *callback_data)
{/* CDISPSYMM_CHANGED */
  int phaseGroupSymmetry;
  char *origStr;
  float origX,origY;

  phaseGroupSymmetry = *tag;

  origStr = XmTextGetString(cDispSymmOrigYW);
  origY = (float)strtod(origStr,0);
  XtFree(origStr);
    
  origStr = XmTextGetString(cDispSymmOrigXW);
  origX = (float)strtod(origStr,0);
  XtFree(origStr);
    
  ricardo_routine(phaseGroupSymmetry, &crys_float_image_buffer,
		  &cdNumCols, &cdNumRows, origX, origY);

  /* Now tryit */
  crysDisp_tryit_activate(0,0,0);

}/* CDISPSYMM_CHANGED */



/*************************************************************
 * crysDisp_contour_activate() - They want to draw a contour
 * from the crysDisp
 */
void crysDisp_contour_activate(Widget widget, int *tag, 
			       XmAnyCallbackStruct *callback_data)
{/* CRYSDISP_CONTOUR_ACTIVATE */
  XmString simpleString;

  /* Now setup the scrolls accordingly */
  /* Always use the section min and max because we cue off of the sections
   * to contour shaded views */
  XtSetArg(arglst[0], XmNmaximum, (int)(255.0));
  XtSetArg(arglst[1], XmNminimum, (int)(-255.0));
  XtSetArg(arglst[2], XmNsliderSize, (int)(5));
  XtSetArg(arglst[3], XmNvalue, (int)(0));
  XtSetArg(arglst[4], XmNpageIncrement, (int)(5));
  XtSetArg(arglst[5], XmNincrement, (int)(1));
  XtSetValues(movieContourScrollWidget, arglst, 6);
  XFlush(dsply);

  /* And the value labels */
  sprintf(outStr,"%d",(0));
  createSimpleString(&simpleString,outStr);
  XtSetArg(arglst[0], XmNlabelString, simpleString);
  XtSetValues(movieContourLabelWidget, arglst, 1); 
  XFlush(dsply);
  XmStringFree(simpleString); 
  XFlush(dsply);

  raiseIt(movie_contour_screen);

}/* CRYSDISP_CONTOUR_ACTIVATE */





/* ***************************************************************
 * crysDisp_create() - Remember the crysDisp screen for use later.  
 */
void crysDisp_create(Widget widget, int *tag, 
		     XmAnyCallbackStruct *callback_data)
{/* crysDisp_CREATE */
  
  if(*tag==ADSCALE)
    aDscaleTextW = widget;
  else if(*tag==SCALEX)
    {scaleXvalueW= widget;
     handle_cr(widget);}
  else if(*tag==NPIXY)
    npixyTextW= widget;
  else if(*tag==NCELLX)
    ncellxvalueW= widget;
  else if(*tag==NCELLY)
    ncellyvalueW= widget;
  else if(*tag==MAP_ABANGSCROLL)
    mapScrollW= widget;
  else if(*tag==MAP_ABANGLABEL)
    mapValueW= widget;
  else if(*tag==MAP_ABANG)
    {/* map_abang */
      map_abangTextW= widget;
      handle_cr(map_abangTextW);
    }/* map_abang */
  else if(*tag==CRYSDISPRESET)
    crysDispResetW= widget;
  else if(*tag==AOVERBSCROLL)
    abScrollW= widget;
  else if(*tag==AOVERBLABEL)
    abValueW= widget;
  else if(*tag==AOVERB)
    {/* aoverb */
      aoverbTextW= widget;
      handle_cr(aoverbTextW);
    }/* aoverb */
  else if(*tag==SCALEXSCROLL)
    scaleXScrollW= widget;
  else if(*tag==NCELLXSCROLL)
    ncellxScrollW= widget;
  else if(*tag==NCELLYSCROLL)
    ncellyScrollW= widget;
  else if(*tag==ADIMSCROLL)
    adimScrollW= widget;
  else if(*tag==BDIMSCROLL)
    bdimScrollW= widget;
  else if(*tag==GAMMAXTALSCROLL)
    gammaXtalScrollW= widget;
  else if(*tag==ADIM)
    {adimTextW= widget;
     handle_cr(widget);}
  else if(*tag==BDIM)
    {bdimTextW= widget;
     handle_cr(widget);}
  else if(*tag==GAMMAXTAL)
    {gammaXtalTextW= widget;
     handle_cr(widget);}
  else if(*tag==SCALEXZOOMUP)
    {cDispZoomArrowUpW=widget;
     handle_cr(widget);}
  else if(*tag==SCALEXZOOMDOWN)
    {cDispZoomArrowDownW=widget;
     handle_cr(widget);}
  else if(*tag==SYMMORIGX)
    cDispSymmOrigXW = widget;
  else if(*tag==SYMMORIGY)
    cDispSymmOrigYW = widget;
  
}/* crysDisp_CREATE */


/* ************************************************************
 * crysDisp_screen_deactivate() - Get the crysDisp screen
 * off the display.
 */
void crysDisp_screen_deactivate(Widget widget, int *tag, 
				XmAnyCallbackStruct *callback_data)
{/* crysDisp_screen_deactIVATE */
  
  crysDisping = FALSE;
  XtUnmanageChild(crysDispscreen);
  free(crys_float_image_buffer);
  crys_float_image_buffer = NULL;

}/* crysDisp_screen_deactIVATE */




/* ************************************************************
 * crysDisp_screen_activate() - Put the crysDispscreen on the display.
 */
void crysDisp_screen_activate(Widget widget, int *tag, 
			      XmAnyCallbackStruct *callback_data)
{/* crysDisp_screen_actIVATE */
  int i,j;
  XmString simpleString;
  static int oldXsize=-99;
  float *thisPackImg;

  crysDisping = TRUE;

  if(fileType==PACKEDPORTABLE && !fourierAveragedCrysDisp)
    {/* Do a little farbling here */
      numCols = purdueHeader.numCols;
      numRows = purdueHeader.numRows;
      cdNumCols = purdueHeader.numCols;
      cdNumRows = purdueHeader.numRows;
      cdNumSecs = 1;
      /* We've got a bug whereby these are bogus ... */
      /* aoverb = 1.0;
	 map_abang = 90.0; */
    }/* Do a little farbling here */
  
  if(!fourierAveragedCrysDisp)
    {/* Init */
      cdNumCols = purdueHeader.numCols;
      cdNumRows = purdueHeader.numRows;
      cdNumSecs = 1;
    }/* Init */

  if(cdNumCols > 0)
    {/* Don't do if no x_size */
      CK(crys_float_image_buffer= (float *)malloc(sizeof(float)*
						  cdNumCols*cdNumRows));
      thisPackImg = packImg[0];
      for(i=0;i<cdNumRows;i++)
	{
	  for(j=0;j<cdNumCols;j++)
	    {/* Redo in TSB format */
	      if(fileType==PACKEDPORTABLE && !fourierAveragedCrysDisp)
		{
		  crys_float_image_buffer[((i*cdNumCols)+j)]= 
		    thisPackImg[((((numRows-1)-i)*numCols)+j)]; 
		  /* We've got to trap for zero values for B&W images */
		  if(crys_float_image_buffer[((i*cdNumCols)+j)]<=2.0)
		    crys_float_image_buffer[((i*cdNumCols)+j)]= 2.0;
		}
	      else if(fourierAveragedCrysDisp)
		{
		  /* j*2 because the fourt() rtn stuffs it in the first 
		   * position */
		  /* Also - why do I have to rotate this 90 degrees ??? */
		  /* OLDWAY works
		   * crys_float_image_buffer[((i*cdNumCols)+j)]= 
		   * float_image_buffer1[((((cdNumRows-1)-i)*(cdNumCols*2))+
		   * (j*2))];
		   */
		  /* Works 3/31/99 
		     crys_float_image_buffer[(((cdNumRows-1-j)*
		     cdNumCols)+i)]= 
		     float_image_buffer1[((i*(cdNumCols*2))+
		     (j*2))];  
		     Works 3/31/99 */
		  /* Works 4/1/99
		    crys_float_image_buffer[((j*cdNumCols)+i)]= 
		      float_image_buffer1[((i*(cdNumCols*2))+
					   (j*2))]; */

		  /*  Original orientation gotten from FOURT rtn 
		  crys_float_image_buffer[((i*cdNumCols)+j)]= 
		    float_image_buffer1[((i*(cdNumCols*2))+
					 (j*2))];
					 */
		  /* Rotate 90 degrees - The fourt() rtn stuffs this
		   * buffer one column at a time - so we've got to
		   * compensate for that here.  */
		  crys_float_image_buffer[((j*cdNumCols)+i)]= 
		    float_image_buffer1[((i*(cdNumCols*2))+
					 (j*2))];

		}
	      else
		{
		  crys_float_image_buffer[((i*cdNumCols)+j)]= 
		    float_image_buffer1[((((numRows-1)-i)*numCols)+j)];
		}
	    }/* Redo in TSB format */
	}

      
      /* DOn't display those annoying numnbers */
      dispMethod = FALSE;       /* Toggle this guy */

      if(cdNumCols != oldXsize)
	{/* They've changed it */
	  x_size= cdNumCols;
	  npixy= cdNumRows;

	  /* Now setup the scroll accordingly */
	  XtSetArg(arglst[0], XmNmaximum, (int)((x_size*10)+5));
	  XtSetArg(arglst[1], XmNminimum, (int)(2));
	  XtSetArg(arglst[2], XmNsliderSize, (int)(5));
	  XtSetArg(arglst[3], XmNvalue, (int)(x_size));
	  XtSetArg(arglst[4], XmNpageIncrement, (int)(x_size/5));
	  XtSetArg(arglst[5], XmNincrement, (int)MAX((x_size/50),1));
	  /* XtSetValues(scaleXScrollW, arglst, 6); */
	  XtSetArg(arglst[4], XmNpageIncrement, (int)(x_size/10));
	  XtSetValues(adimScrollW, arglst, 5);
	  XFlush(dsply);
	  adim = (int)x_size;
	  bdim = cdNumRows;
	  XtSetArg(arglst[4], XmNpageIncrement, (int)(bdim/10));
	  XtSetValues(bdimScrollW, arglst, 5);
	  gammaXtal = 90.0;
	  XtSetArg(arglst[0], XmNmaximum, (int)(365.0));
	  XtSetArg(arglst[1], XmNminimum, (int)(0));
	  XtSetArg(arglst[2], XmNsliderSize, (int)(10));
	  XtSetArg(arglst[3], XmNvalue, (int)(gammaXtal));
	  XtSetArg(arglst[4], XmNpageIncrement, (int)(10));
	  XtSetValues(gammaXtalScrollW, arglst, 5);
	  XFlush(dsply);

	  /* And the value label */
	  sprintf(outStr,"%.2f",(float)((float)x_size/(float)cdNumCols)); 
	  XmTextSetString(scaleXvalueW,outStr);
	  sprintf(outStr,"%d",(adim)); 
	  XmTextSetString(adimTextW,outStr);
	  sprintf(outStr,"%d",(bdim)); 
	  XmTextSetString(bdimTextW,outStr);
	  sprintf(outStr,"%.2f",(float)(gammaXtal)); 
	  XmTextSetString(gammaXtalTextW,outStr);
	  XFlush(dsply);
	}/* They've changed it */
      oldXsize = cdNumCols;

      if(*tag==1)
	{/* Init these */
	  if(map_abang == 0.0)
	    map_abang = 90.0;
	  map_abangTemp = map_abang;
	  sprintf(outStr,"%f",map_abangTemp);
	  XmTextSetString(map_abangTextW, outStr);
	  aoverbTemp = aoverb;
	  sprintf(outStr,"%f",aoverbTemp);
	  XmTextSetString(aoverbTextW, outStr);
	}/* Init these */
      
      /* Put screen on display */
      raiseIt(crysDispscreen);

      /* Now setup map and ab scrolls and tryit */
      crysDisp_cr_activate(0,0,0);

    }/* Don't do if no x_size */
  else
    crysDisping= FALSE;

}/* crysDisp_screen_actIVATE */





/* *****************************************************************
 * crysDisp_reset_activate() - They have hit the RESET button
 * Reset delta scrolls to zero, reinit map_abangTemp and aoverbTemp
 * and call tryit().
 */
void crysDisp_reset_activate(Widget widget, int *tag, 
			     XmAnyCallbackStruct *callback_data)
{/* crysDisp_RESET_ACTIVATE */
  XmString simpleString;

  /* Reinit gloabls */
  if(map_abang == 0.0)
    map_abang = 90.0;
  map_abangTemp = map_abang;
  sprintf(outStr,"%.2f",map_abangTemp);
  XmTextSetString(map_abangTextW, outStr);
  aoverbTemp = aoverb;
  sprintf(outStr,"%.2f",aoverbTemp);
  XmTextSetString(aoverbTextW, outStr);
      
  /* Now setup the scrolls accordingly */
  deltaMapABang= 0.0;
/* OLDWAY   XtSetArg(arglst[0], XmNvalue, (int)deltaMapABang); */
  XtSetArg(arglst[0], XmNvalue, (int)map_abangTemp);
  XtSetValues(mapScrollW, arglst, 1);
  XFlush(dsply);
  /* And the value label */
  sprintf(outStr,"%.2f",(float)(map_abangTemp)); 
  createSimpleString(&simpleString,outStr);
  XtSetArg(arglst[0], XmNlabelString, simpleString);
  XtSetValues(mapValueW, arglst, 1); 
  XFlush(dsply);
  XmStringFree(simpleString); 
  XFlush(dsply);

  /* Now setup the scrolls accordingly */
  deltaAoverB= 0.0;
  /* OLDWAY  XtSetArg(arglst[0], XmNvalue, (int)deltaAoverB); */
  XtSetArg(arglst[0], XmNvalue, (int)aoverbTemp*200);
  XtSetValues(abScrollW, arglst, 1);
  XFlush(dsply);
  /* And the value label */
  sprintf(outStr,"%.2f",(float)(aoverbTemp)); 
  createSimpleString(&simpleString,outStr);
  XtSetArg(arglst[0], XmNlabelString, simpleString);
  XtSetValues(abValueW, arglst, 1); 
  XFlush(dsply);
  XmStringFree(simpleString); 
  XFlush(dsply);

  crysDisp_tryit_activate(0,0,0);

}/* crysDisp_RESET_ACTIVATE */





/* ************************************************************
 * crysDisp_cr_activate() - They have hit the CReturn key while
 * in the map_abang or aoverb Text widgets.  Reset delta scrolls 
 * to zero and call tryit().
 */
void crysDisp_cr_activate(Widget widget, int *tag, 
			  XmAnyCallbackStruct *callback_data)
{/* crysDisp_CR_ACTIVATE */
  XmString simpleString;

  /* Now setup the scrolls accordingly */
  deltaMapABang= 0.0;
  /* OLDWAY XtSetArg(arglst[0], XmNvalue, (int)deltaMapABang); */
  XtSetArg(arglst[0], XmNvalue, (int)map_abangTemp);
  XtSetValues(mapScrollW, arglst, 1);
  XFlush(dsply);
  /* And the value label */
  sprintf(outStr,"%.2f",(float)(map_abangTemp)); 
  createSimpleString(&simpleString,outStr);
  XtSetArg(arglst[0], XmNlabelString, simpleString);
  XtSetValues(mapValueW, arglst, 1); 
  XFlush(dsply);
  XmStringFree(simpleString); 
  XFlush(dsply);

  /* Now setup the scrolls accordingly */
  deltaAoverB= 0.0;
/* OLDWAY  XtSetArg(arglst[0], XmNvalue, (int)deltaAoverB); */
  XtSetArg(arglst[0], XmNvalue, (int)aoverbTemp*200);
  XtSetValues(abScrollW, arglst, 1);
  XFlush(dsply);
  /* And the value label */
  sprintf(outStr,"%.2f",(float)(aoverbTemp)); 
  createSimpleString(&simpleString,outStr);
  XtSetArg(arglst[0], XmNlabelString, simpleString);
  XtSetValues(abValueW, arglst, 1); 
  XFlush(dsply);
  XmStringFree(simpleString); 
  XFlush(dsply);

  crysDisp_tryit_activate(0,0,0);

}/* crysDisp_CR_ACTIVATE */





/* ************************************************************
 * crysDisp_tryit_activate() - Try to Anglularly display this 
 * float map image.
 */
void crysDisp_tryit_activate(Widget widget, int *tag, 
			     XmAnyCallbackStruct *callback_data)
{/* crysDisp_TRYIT_ACTIVATE */
  unsigned short *disp_buffer;
  int ny, section, plotX, plotY; 
  float xstrt,ystrt,new_min,angle,aoverbPass,sinang,cosang;
  float scaleFactor,thisMin,thisMax,scale,passGamma;
  int y_size;
  int thisPixel,icenx,iceny;
  int k,l,ycellOffset,xcellOffset;

  /* Try this */
  angle = map_abangTemp;
  aoverbPass= aoverbTemp;

  if(angle == 0.0)
    angle = 90.0;     /* Just give it something */

  if(aoverbPass == 0.0)
    aoverbPass = 1.0;

  sinang = sin(deg2rad(angle));
  cosang = cos(deg2rad(angle));

  /* Now convert angle to radians */
  angle = deg2rad(angle);

  aDscale= scale  = (float)(x_size)/(float)(cdNumCols);
  ny     = (int)(scale*(float)(cdNumRows)/aoverbPass);
  y_size = (int)(sinang*ny);

  if(buffer_width < display_width)
    icenx = (buffer_width / 2);
  else icenx = (display_width / 2);
  if(buffer_height < display_height)
    iceny = (buffer_height / 2);
  else iceny = (display_height / 2);

  xstrt = icenx - ((float)(x_size*ncellx) + 
		   ncelly*cosang*ny)/2.0;
  ystrt = (float)((float)iceny - 
		  ((float)((float)y_size*(float)ncelly)/2.0));

  /* Center it better */
  xstrt -= (float)((float)x_size/1.0);
  ystrt -= (float)((float)y_size/1.0);

  npixx = x_size;
  npixy = (int)y_size;  

  section = 1;
  new_min = minMapPixel;

  if(buffer_width < display_width)
    plotX = buffer_width;
  else plotX = display_width;
  if(buffer_height < display_height)
    plotY = buffer_height;
  else plotY = display_height;
  CK(disp_buffer= (unsigned short *)malloc(sizeof(unsigned short)*
					   (plotX*plotY)));
  for(i=0;i<plotY;i++)
    for(j=0;j<plotX;j++)
      disp_buffer[((i*plotX)+j)]= 0;

  /* Put gamma into radians */
  passGamma = deg2rad(gammaXtal);

  mapdsp_gplot_rob_ (crys_float_image_buffer, 
		     disp_buffer, 
		     &aDscale, 
		     &cdNumCols, &cdNumRows, &cdNumSecs, 
		     &npixx, &npixy, 
		     &xstrt, &ystrt, 
		     &ny, 
		     &ncellx, &ncelly, 
		     &section,
		     &new_min,
		     &angle,
		     &aoverbPass,
		     &plotX,&plotY,
		     &adim, &bdim, &passGamma);

  thisMin = 99999.0;
  thisMax = -99999.0;
  thisMin = HUGE;   /* dgm HUGE instead of 99999.9 */
  thisMax = -HUGE;  /* dgm -HUGE instead of -99999.9 */
  for(i=0;i<(plotY);i++)
    for(j=0;j<(plotX);j++)
      {/* Put it right beside the orig */
	thisPixel= (float)disp_buffer[((i*plotX)+j)];
	thisMin = MIN(thisPixel,thisMin);
	thisMax = MAX(thisPixel,thisMax);
      }/* Put it right beside the orig */

  if(thisMax > thisMin)
    scaleFactor = (((float)tryColors-7.0)/(thisMax-thisMin));
  else scaleFactor = 1.0;

  /* Init it to black ... */
  for(i=0;i<(buffer_height);i++)
    for(j=0;j<(buffer_width);j++)
      {
	if(hiColor)
	  XPutPixel(image_image1, j, i, pixels_return[BLACK]);
	else
	  image_buffer1[((i*buffer_width)+j+cdNumCols)]= pixels_return[BLACK];
      }

  for(i=0;i<(plotY);i++)
    for(j=0;j<(plotX);j++)
      {/* Put it right beside the orig */
	if(hiColor)
	  {
	    XPutPixel(image_image1, j, ((plotY-1)-i), 
		      pixels_return[(int)((disp_buffer[((i*plotX)+j)]-
					   thisMin)*scaleFactor)+6]);
	  }
	else
	  {
	    image_buffer1[((((plotY-1)-i)*buffer_width)+j)]= 
	      (unsigned char)
	      pixels_return[(int)((disp_buffer[((i*plotX)+j)]-
				   thisMin)*scaleFactor)+6];
	  }
      }/* Put it right beside the orig */

  if(!hiColor)
    XPutImage (dsply, image_window1, image_gc, image_image1, 0, 0,
	       0, 0, (display_width), display_height);
  XFlush (dsply);
  DrawWindow();

  free(disp_buffer);


}/* crysDisp_TRYIT_ACTIVATE */





/* ************************************************************
 * crystalDisp_changed() - Whenever crysDisp text widgets changed
 * call here.
 */
void crystalDisp_changed(Widget widget, int *tag, 
			 XmAnyCallbackStruct *callback_data)
{/* crystalDisp_CHANGED */
  char *rotflt,outStr[64];
  XmString simpleString;
  int i, thisI;
  float thisF;
  char *crystalDispText;

  if(*tag==UPDATE)
    {/* Update toggle */
      updateCrys = !updateCrys;
    }/* Update toggle */
  else
    {/* Text widgets */
      crystalDispText = XmTextGetString(widget); 
      
      if(*tag==ADSCALE)
	{/* */ 
	  aDscale = (float)strtod(crystalDispText,0);
	}/* */
      else if (*tag==SCALEX)
	{/* scaleX */
	  thisF = (float)strtod(crystalDispText,0);
	  if(thisF != 0.0)
	    {/* OK to do */
	      x_size = (int)((float)thisF * (float)cdNumCols);
	      /*
	      XtSetArg(arglst[0], XmNvalue, (int)x_size);
	      XtSetValues(scaleXScrollW, arglst, 1);
	      XFlush(dsply);
	      */
	      if(aoverbTemp != 0.0)
		{/* OK */
		  thisI = (int)((float)((float)cdNumCols) * thisF);
		  if(thisI != 0)
		    {/* OK to do */
		      adim = (int)thisI;
		      XtSetArg(arglst[0], XmNvalue, (int)adim);
		      XtSetValues(adimScrollW, arglst, 1);
		      sprintf(outStr,"%d",adim);
		      XmTextSetString(adimTextW,outStr);
		      XFlush(dsply);
		    }/* OK to do */
		  thisI = (int)((float)((float)cdNumRows / aoverbTemp) * 
				thisF);
		  if(thisI != 0)
		    {/* OK to do */
		      bdim = (int)thisI;
		      XtSetArg(arglst[0], XmNvalue, (int)bdim);
		      XtSetValues(bdimScrollW, arglst, 1);
		      sprintf(outStr,"%d",bdim);
		      XmTextSetString(bdimTextW,outStr);
		      XFlush(dsply);
		    }/* OK to do */
		}/* OK */
	      if(aoverbTemp != 0.0) /* Wait until calculated */
		crysDisp_tryit_activate(0,0,0);
	    }/* OK to do */
	}/* scaleX */
      else if (*tag==AOVERB)
	{/* aoverb */
	  aoverbTemp = (float)strtod(crystalDispText,0);
	  if(aoverbTemp != 0.0)
	    {/* OK */
	      thisI = (int)((float)cdNumRows / aoverbTemp)*
		(float)((float)x_size/(float)cdNumCols);
	      if(thisI != 0)
		{/* OK to do */
		  bdim = (int)thisI;
		  XtSetArg(arglst[0], XmNvalue, (int)bdim);
		  XtSetValues(bdimScrollW, arglst, 1);
		  sprintf(outStr,"%d",bdim);
		  XmTextSetString(bdimTextW,outStr);
		  XFlush(dsply);
		}/* OK to do */
	    }/* OK */
	}/* aoverb */
      else if (*tag==MAP_ABANG)
	{/* map_abang */
	  map_abangTemp = (float)strtod(crystalDispText,0);
	}/* map_abang */
      else if (*tag==NPIXY)
	{/* npixy */
	  npixy = (int)strtod(crystalDispText,0);
	}/* npixy */
      else if (*tag==NCELLX)
	{/* ncellx */
	  ncellx = (int)strtod(crystalDispText,0);
	}/* ncellx */
      else if (*tag==NCELLY)
	{/* ncelly */
	  ncelly = (int)strtod(crystalDispText,0);
	}/* ncelly */
      else if (*tag==GSCALE)
	{/* g scale */
	  gscale = (float)strtod(crystalDispText,0);
	}/* g scale */
      else if (*tag==ADIM)
	{/* adim */
	  thisI = (int)strtod(crystalDispText,0);
	  if(thisI != 0)
	    {/* OK to do */
	      adim = (int)thisI;
	      XtSetArg(arglst[0], XmNvalue, (int)adim);
	      XtSetValues(adimScrollW, arglst, 1);
	      XFlush(dsply);
	      if(crys_float_image_buffer != 0 &&
		 aoverbTemp != 0.0) /* Wait until calculated */
		crysDisp_tryit_activate(0,0,0);
	    }/* OK to do */
	}/* adim */
      else if (*tag==BDIM)
	{/* bdim */
	  thisI = (int)strtod(crystalDispText,0);
	  if(thisI != 0)
	    {/* OK to do */
	      bdim = thisI;
	      XtSetArg(arglst[0], XmNvalue, (int)bdim);
	      XtSetValues(bdimScrollW, arglst, 1);
	      XFlush(dsply);
	      if(crys_float_image_buffer != 0 &&
		 aoverbTemp != 0.0) /* Wait until calculated */
		crysDisp_tryit_activate(0,0,0);
	    }/* OK to do */
	}/* bdim */
      else if (*tag==GAMMAXTAL)
	{/* gammaXtal */
	  thisF = (float)strtod(crystalDispText,0);
	  if(thisF != 0.0)
	    {/* OK to do */
	      gammaXtal = thisF;
	      XtSetArg(arglst[0], XmNvalue, (int)gammaXtal);
	      XtSetValues(gammaXtalScrollW, arglst, 1);
	      XFlush(dsply);
	      if(aoverbTemp != 0.0) /* Wait until calculated */
		crysDisp_tryit_activate(0,0,0);
	    }/* OK to do */
	}/* gammaXtal */
      
      XtFree(crystalDispText);
    }/* Text widgets */
  
}/* crystalDisp_CHANGED */





/* ************************************************************
 * crysDisp_scroll_change() - Whenever crysDisp_scroll scroll 
 * widgets changed call here.
 */
void crysDisp_scroll_change(Widget widget, int *tag, 
			    XmScrollBarCallbackStruct *callback_data)
{/* CRYSDISP_SCROLL_CHANGE */
  char *crysDisp_scroll,outStr[64];
  XmString simpleString;
  int i,newMaxCell;
  
  if(!inited)
    goto notInited;

  if(*tag==SCALEX)
    {/* */ 
      /* Get zoom factor */
      x_size = (int)callback_data->value;
      sprintf(outStr,"%.2f",(float)((float)x_size/(float)cdNumCols)); 
      XmTextSetString(scaleXvalueW,outStr);
      XFlush(dsply);

      /* Now reset ncell* scrolls so they can use all of the
       * display */
      if(buffer_width < display_width)
	newMaxCell = buffer_width / x_size;
      else newMaxCell = display_width / x_size;
      newMaxCell *= 4;
      if(newMaxCell == 0 )
	newMaxCell = 1;

      if(newMaxCell < ncelly)
	{/* Revalue it */
	  ncelly = (int)newMaxCell;
	  sprintf(outStr,"%d",ncelly); 
	  createSimpleString(&simpleString,outStr);
	  XtSetArg(arglst[0], XmNlabelString, simpleString);
	  XtSetValues(ncellyvalueW, arglst, 1); 
	  XFlush(dsply);
	  XmStringFree(simpleString); 
	  XFlush(dsply);
	}/* Revalue it */

      if(newMaxCell < ncellx)
	{/* Revalue it */
	  ncellx = (int)newMaxCell;
	  sprintf(outStr,"%d",ncellx); 
	  createSimpleString(&simpleString,outStr);
	  XtSetArg(arglst[0], XmNlabelString, simpleString);
	  XtSetValues(ncellxvalueW, arglst, 1); 
	  XFlush(dsply);
	  XmStringFree(simpleString); 
	  XFlush(dsply);
	}/* Revalue it */

      XtSetArg(arglst[0], XmNmaximum, (int)(newMaxCell+5));
      XtSetArg(arglst[1], XmNsliderSize, (int)(5));
      XtSetArg(arglst[2], XmNvalue, (int)(ncellx));
      XtSetValues(ncellxScrollW, arglst, 3);
      XtSetArg(arglst[0], XmNmaximum, (int)(newMaxCell+5));
      XtSetArg(arglst[1], XmNsliderSize, (int)(5));
      XtSetArg(arglst[2], XmNvalue, (int)(ncelly));
      XtSetValues(ncellyScrollW, arglst, 3);

    }/* */
  else if (*tag==NCELLX)
    {/* ncellx */
      ncellx = (int)callback_data->value;
      sprintf(outStr,"%d",ncellx); 
      createSimpleString(&simpleString,outStr);
      XtSetArg(arglst[0], XmNlabelString, simpleString);
      XtSetValues(ncellxvalueW, arglst, 1); 
      XFlush(dsply);
      XmStringFree(simpleString); 
      XFlush(dsply);
    }/* ncellx */
  else if (*tag==NCELLY)
    {/* ncelly */
      ncelly = (int)callback_data->value;
      sprintf(outStr,"%d",ncelly); 
      createSimpleString(&simpleString,outStr);
      XtSetArg(arglst[0], XmNlabelString, simpleString);
      XtSetValues(ncellyvalueW, arglst, 1); 
      XFlush(dsply);
      XmStringFree(simpleString); 
      XFlush(dsply);
    }/* ncelly */
  else if (*tag==MAP_ABANG)
    {/* map_abang */
      map_abangTemp = (float)callback_data->value;
      sprintf(outStr,"%.2f",(map_abangTemp));
      createSimpleString(&simpleString,outStr);
      XtSetArg(arglst[0], XmNlabelString, simpleString);
      XtSetValues(mapValueW, arglst, 1); 
      XFlush(dsply);
      XmStringFree(simpleString); 
      sprintf(outStr,"%.2f",(map_abangTemp));
      XmTextSetString(map_abangTextW, outStr);
      XFlush(dsply);
    }/* map_abang */
  else if (*tag==AOVERB)
    {/* aoverb */
      aoverbTemp = (float)callback_data->value;
      aoverbTemp *= 0.005;    /* Do it in 5/1000's */
      sprintf(outStr,"%.2f",(aoverbTemp));
      createSimpleString(&simpleString,outStr);
      XtSetArg(arglst[0], XmNlabelString, simpleString);
      XtSetValues(abValueW, arglst, 1); 
      XFlush(dsply);
      XmStringFree(simpleString); 
      sprintf(outStr,"%.2f",(aoverbTemp));
      XmTextSetString(aoverbTextW, outStr);
      XFlush(dsply);
    }/* aoverb */
  else if (*tag==ADIM)
    {/* adim */
      adim = (int)callback_data->value;
      sprintf(outStr,"%d",(adim));
      XmTextSetString(adimTextW,outStr);
      XFlush(dsply);
    }/* adim */
  else if (*tag==BDIM)
    {/* bdim */
      bdim = (int)callback_data->value;
      sprintf(outStr,"%d",(bdim));
      XmTextSetString(bdimTextW,outStr);
      XFlush(dsply);
    }/* bdim */
  else if (*tag==GAMMAXTAL)
    {/* gammaXtal */
      gammaXtal = (float)callback_data->value;
      sprintf(outStr,"%.2f",(gammaXtal));
      XmTextSetString(gammaXtalTextW,outStr);
      XFlush(dsply);
    }/* gammaXtal */

  if(updateCrys || (callback_data->reason != XmCR_DRAG))
    crysDisp_tryit_activate(0,0,0);

 notInited:;

}/* CRYSDISP_SCROLL_CHANGE */




/******************************************************************
 * linGrad_activate() - They want to linearly gradient their
 * image(s).
 */
void linGrad_activate(Widget widget, int *tag, 
		      XmAnyCallbackStruct *callback_data)
{/* linGrad_ACTIVATE */
  int   gradRadius,i,j,k,thisPixel;
  float img_bkg, offset;
  float img_min,img_max;
  float img_avg,img_std;
  float *tmpBuf, img_bkg_int;
  float *thisPackBuf;
  int   minP,maxP,saveMaxPackPixel;
  float scaleFactor,expFac;
  int   xcen,ycen,xSq,ySq,jMycen,kMxcen,thisRad,distnce,origBoxNum;
  char  *imgNumText;
	      
  if(fileType==PACKEDPORTABLE)
    {/* OK */

      imgNumText = XmTextGetString(boxInitWidget); 
      origBoxNum =  (int)strtod(imgNumText, 0);
      XtFree(imgNumText);
      
      xcen = ycen = packRadius;
      thisRad = packRadius * packRadius;

      /* This may take awhile so give them the watch cursor */
      attrs.cursor= watch_cursor;
      XChangeWindowAttributes(dsply, XtWindow(drawingWidget), 
			      CWCursor, &attrs);
      XFlush(dsply);
      
      numCols = purdueHeader.numCols;
      numRows = purdueHeader.numRows;

      CK(tmpBuf= (float *)malloc(sizeof(float)*numCols*numRows));

      saveMaxPackPixel = maxPackPixel;
      maxPackPixel= -99999999.0;
      minPackPixel= 99999999.0;
      maxPackPixel= INT_MIN;  /* dgm INT_MIN instead of -9999999 */
      minPackPixel= INT_MAX;  /* dgm INT_MAX instead of 9999999 */
      for(i=0;i<purdueHeader.nimg;i++)
	{/* Accumlate all images */

	  sprintf(outStr,"%d",i+1);
	  XmTextSetString(boxInitWidget,outStr);

	  if(!deletePackedImage[i])
	    {/* USe this one */
	      calcCircleStats(packImg[i],
			      &img_bkg, 
			      &img_min,
			      &img_max,
			      &img_avg, 
			      &img_std,
			      numCols, numRows, numCols);
	      /* printf("1) bkg=%d min=%d  max=%d  avg=%f  std=%f\n",
		     img_bkg,
		     img_min,
		     img_max,
		     img_avg,
		     img_std);        */

	      /* Stuff packed byte image into short image buffer */
	      thisPackBuf = packImg[i];
	      for(j=0;j<numRows;j++)
		for(k=0;k<numCols;k++)
		  {/* Do all */
		    thisPixel = ((j*numCols)+k);
		    tmpBuf[(int)thisPixel]= 
		      (float)(thisPackBuf[(int)thisPixel]);
		  }/* Do all */

	      /* img_lingrad depends on the background being right */
	      if(packRadius > 0.0)
		gradRadius = numCols / 2 ;
	      else gradRadius = 0; /* prevent img_lingrad_rob() from 
				    * doing circles */
	      /* img_bkg_short = (short)img_bkg; */


  /* In img_lingrad_rob():
   *    March 1998 RWA - I've had to multiply the IMG() pixel values by
   *    100.0 to increase the dynamic range so that the TRUS*() rtns 
   *    work properly.  Trouble is we're passing in BYTE info in this
   *    short buffer and only using 0-255 when TRUS expects 32000 short
   *    values.  Changed it to pass in the expFac (expanding Factor).
   *    Seems with packed boxes we can't use this large of an expanding
   *    factor (100.0).  Let's try 10.0 * numCols/500.0 so a 5000x5000
   *    scanned image woudl have our 100.0 factor and a 50x50 packed box
   *    would have an expFac==1.0.
   */
	      expFac = 10.0 * ((float)numCols /100.0);

	      if(expFac < 1.0)
		expFac = 1.0;

	      /* Let's just force this to be 1.0 for now - I think
	       * expFac is more important for large scanned images???
	      expFac = 1.0;   */

	      img_lingrad_rob_(tmpBuf, &numCols, &numRows, 
			      &gradRadius, &img_bkg, &expFac);

	      /* Find min/max */
	      minP = maxP = tmpBuf[0];
	      for(j=0;j<numRows;j++)
		for(k=0;k<numCols;k++)
		  {/* Do all */
		    thisPixel = ((j*numCols)+k);
		    minP = MIN(minP,tmpBuf[(thisPixel)]);
		    maxP = MAX(maxP,tmpBuf[(thisPixel)]);
		  }/* Do all */
	      /* Why tryColors-50.0 here ???  Why not 255.0 ??? */
	      if(maxP == minP)
		{/* no contrast */
		  scaleFactor = 1.0;
		  minP = 1.0;
		}/* no contrast */
	      else
		{/* normal */
		  scaleFactor = ((float)tryColors-50.0) / (float)(maxP-minP);
		  scaleFactor = ((float)saveMaxPackPixel) / (float)(maxP-minP);
		}/* normal */

	      /* Stuff short image buffer back into Pack array*/
	      thisPackBuf = packImg[i];
	      for(j=0;j<numRows;j++)
		for(k=0;k<numCols;k++)
		  {/* Do all */
		    thisPixel = ((j*numCols)+k);
		    thisPackBuf[(thisPixel)]= (float)
		      (float)((float)(tmpBuf[(thisPixel)]-minP)*scaleFactor)
		      +1.0;
		  }/* Do all */

	      if(packRadius > 0.0)
		{/* Now recalc the backGround */
		  calcCircleStats(packImg[i],
				  &img_bkg, 
				  &img_min,
				  &img_max,
				  &img_avg, 
				  &img_std,
				  numCols, numRows, numCols);
		  /* And apply new background */
		  for(j=0;j<numRows;j++)
		    {/* rows */
		      jMycen = j - ycen;
		      ySq    = (jMycen*jMycen);
		      for(k=0;k<numCols;k++)
			{/* Do all */
			  kMxcen = k - xcen;
			  xSq    = (kMxcen*kMxcen);
			  distnce = (xSq+ySq);
			  thisPixel = ((j*numCols)+k);
			  if(distnce > thisRad)
			    {/* Stuff it */
			      thisPackBuf[(thisPixel)]= (float)img_bkg;
			    }/* Stuff it */
			  
			  /* Now find min/max */
			  maxPackPixel= (MAX(maxPackPixel,
					     thisPackBuf[thisPixel]));
			  minPackPixel= (MIN(minPackPixel,
					     thisPackBuf[thisPixel]));
			}/* Do all */
		    }/* rows */
		}/* Now recalc the backGround */
	      else
		{/* Just get what min max we've got */
		  for(j=0;j<numRows;j++)
		    {/* rows */
		      for(k=0;k<numCols;k++)
			{/* Do all */
			  thisPixel = ((j*numCols)+k);
			  /* Now find min/max */
			  maxPackPixel= (MAX(maxPackPixel,
					     thisPackBuf[thisPixel]));
			  minPackPixel= (MIN(minPackPixel,
					     thisPackBuf[thisPixel]));
			}/* Do all */
		    }/* rows */
		}/* Just get what min max we've got */
	      
	    }/* USe this one */
	}/* Accumlate all images */


#define TRY_SCALING_PACKED 1
#ifdef TRY_SCALING_PACKED
      /* Now clip this guy to our colormap */ 
      if(minPackPixel < 0.0)
	scalePackFactor = (((float)tryColors - 7.0) / 
			   (maxPackPixel-minPackPixel));
      else scalePackFactor = (((float)tryColors - 7.0) / 
			      maxPackPixel);
#else
      /* Now clip this guy to our colormap */ 
      scalePackFactor = (((float)tryColors - 7.0) / 256.0);
#endif

      free(tmpBuf);

      sprintf(outStr,"%d",origBoxNum);
      XmTextSetString(boxInitWidget,outStr);

      stuffPackedImage(purdueHeader.numCols,
		       purdueHeader.numRows,
		       packImg,purdueHeader.nimg,
		       imgNum);  /* Start at radius'th image */
      
      DrawWindow(); 

      /* Change cursor back to what it was */
      attrs.cursor= old_cursor;
      XChangeWindowAttributes(dsply, XtWindow(drawingWidget), 
			      CWCursor, &attrs);
      XFlush(dsply);

      if(normLinSwitch || doLinAlso)
	printf("Linear Gradient finished OK.\n");
      else
	askThem("Linear Gradient finished OK.",
		FALSE,XmDIALOG_INFORMATION);
    }/* OK */
  else if((fileType==EMPORTABLE ||
	   fileType==BYTEPORTABLE ||
	   fileType==PORTABLE) && imgInCore)
    {/* OK */

      /* This may take awhile so give them the watch cursor */
      attrs.cursor= watch_cursor;
      XChangeWindowAttributes(dsply, XtWindow(drawingWidget), 
			      CWCursor, &attrs);
      XFlush(dsply);
      
      numCols = file_width;
      numRows = file_height;

      CK(tmpBuf= (float *)malloc(sizeof(float)*numCols*numRows));

      calcStats(inCore_buffer,
		&img_bkg, 
		&img_min,
		&img_max,
		&img_avg, 
		&img_std,
		numCols, numRows, numCols);

      /* Stuff packed byte image into short image buffer */
      for(j=0;j<numRows;j++)
	for(k=0;k<numCols;k++)
	  {/* Do all */
	    thisPixel = ((j*numCols)+k);
	    tmpBuf[thisPixel]= (float)(inCore_buffer[thisPixel]);
	  }/* Do all */
      
      /* img_lingrad depends on the background being right */
      if(numRows > numCols)
	gradRadius = numRows * 1.5 ;
      else gradRadius = numCols * 1.5 ;
      /* img_bkg_short = (short)img_bkg; */


  /* In img_lingrad_rob():
   *    March 1998 RWA - I've had to multiply the IMG() pixel values by
   *    100.0 to increase the dynamic range so that the TRUS*() rtns 
   *    work properly.  Trouble is we're passing in BYTE info in this
   *    short buffer and only using 0-255 when TRUS expects 32000 short
   *    values.  Changed it to pass in the expFac (expanding Factor).
   *    Seems with packed boxes we can't use this large of an expanding
   *    factor (100.0).  Let's try 10.0 * numCols/500.0 so a 5000x5000
   *    scanned image woudl have our 100.0 factor and a 50x50 packed box
   *    would have an expFac==1.0.
   */
      expFac = 10.0 * ((float)numCols /500.0);
      expFac = 100.0;
      img_lingrad_rob_(tmpBuf, &numCols, &numRows, 
		       &gradRadius, &img_bkg, &expFac);
      
      /* Stuff short image buffer back into Pack array*/
      minP = maxP = (float)tmpBuf[0];
      for(j=0;j<numRows;j++)
	for(k=0;k<numCols;k++)
	  {/* Do all */
	    thisPixel = ((j*numCols)+k);
	    minP = MIN(minP,(float)tmpBuf[(thisPixel)]);
	    maxP = MAX(maxP,(float)tmpBuf[(thisPixel)]);
	  }/* Do all */
      scaleFactor = ((float)tryColors-9.0) / (maxP-minP);

      for(j=0;j<numRows;j++)
	for(k=0;k<numCols;k++)
	  {/* Do all */
	    thisPixel = ((j*numCols)+k);
	    inCore_buffer[(thisPixel)]= (float)
	      (((float)tmpBuf[(thisPixel)]-minP)*scaleFactor);
	  }/* Do all */

      free(tmpBuf);
      
      /* Make them regenerate a full-screen image */
      /*      fullScreen = TRUE; */
      fullScreen_buffer.dispx = fullScreen_buffer.dispy = 0;

      virtual_read_activate(0,0);
      
      /* Change cursor back to what it was */
      attrs.cursor= old_cursor;
      XChangeWindowAttributes(dsply, XtWindow(drawingWidget), 
			      CWCursor, &attrs);
      XFlush(dsply);
      
      askThem("Linear Gradient finished OK.",
	      FALSE,XmDIALOG_INFORMATION);
    }/* OK */
  else
    {/* Tell them why we're not doing this */
      askThem("Cannot perform Linear gradient for this data.",
	      FALSE,XmDIALOG_INFORMATION);
    }/* Tell them why we're not doing this */

 noCalc:;

}/* linGrad_ACTIVATE */






/*******************************************************************
 * normalize_activate() - They want to normalize their images
 * across the entire data set. *tag==2 means they want normalization
 * to happen for Backgrounds instead of the usual way using means.
 */
void normalize_activate(Widget widget, int *tag, 
			XmAnyCallbackStruct *callback_data)
{/* normalize_ACTIVATE */
  int i,j,k,tot;
  float *img_bkg;
  float *img_min,*img_max;
  float totMean,totStd,*img_avg,*img_std;
  float globMean, globStd, stdAdjustFactor;
  float globMin, globMax;
  int   yBuf;
  float newScaleFactor, new_minF, new_maxF;
  float   new_min, new_max;
  int   constant;
  float *thisBuf;
  int xcen,ycen,xSq,ySq,jMycen,kMxcen,thisRad,distnce,thisPixel;
  float *thisPackBuf;
  char *imgNumText;
  int origBoxNum;

  kindOfNormalize = *tag;

  if(fileType==PACKEDPORTABLE)
    {/* OK */

      imgNumText = XmTextGetString(boxInitWidget); 
      origBoxNum =  (int)strtod(imgNumText, 0);
      XtFree(imgNumText);

      xcen = ycen = packRadius;
      thisRad = packRadius * packRadius;

      if(*tag==2)
	{/* They want background normalization */
	  if(!askThem("Normalization of Backgrounds should only be used\n\
for cosmetic purposes.  For programs like PFT\n\
you should use Normalize Means&Vars.\n\
Continue?",
		  FALSE,XmDIALOG_QUESTION))
	    goto noCalc;

	}/* They want background normalization */

      /* This may take awhile so give them the watch cursor */
      attrs.cursor= watch_cursor;
      XChangeWindowAttributes(dsply, XtWindow(drawingWidget), 
			      CWCursor, &attrs);
      XFlush(dsply);

      CK(img_bkg= (float *)malloc(sizeof(float)*purdueHeader.nimg));
      CK(img_min= (float *)malloc(sizeof(float)*purdueHeader.nimg));
      CK(img_max= (float *)malloc(sizeof(float)*purdueHeader.nimg));
      CK(img_avg= (float *)malloc(sizeof(float)*purdueHeader.nimg));
      CK(img_std= (float *)malloc(sizeof(float)*purdueHeader.nimg));

      numCols = purdueHeader.numCols;
      numRows = purdueHeader.numRows;

      tot = 0;
      totMean = totStd= 0.0;
      globMin = 99999;
      globMax = -99999;
      globMin = HUGE;  /* dgm HUGE instead of 99999 */
      globMax = -HUGE; /* dgm HUGE instead of -99999 */

      for(i=0;i<purdueHeader.nimg;i++)
	{/* Accumlate all images */

	  sprintf(outStr,"A[%d]",i+1);
	  XmTextSetString(boxInitWidget,outStr);

	  if(!deletePackedImage[i])
	    {/* USe this one */
	      calcCircleStats(packImg[i],
			      &img_bkg[i], 
			      &img_min[i],
			      &img_max[i],
			      &img_avg[i], 
			      &img_std[i],
			      numCols, numRows, numCols);
	       /* printf("Norm min[i]=%d  max[i]=%d  avg=%f  std=%f\n",
		     img_min[i],
		     img_max[i],
		     img_avg[i],
		     img_std[i]);    */
	      if(*tag==2)
		totMean += img_bkg[i];
	      else 
		totMean += img_avg[i];

	      totStd  += img_std[i];

	      globMin = MIN(globMin,img_min[i]);
	      globMax = MAX(globMax,img_max[i]);

	      tot++;

	    }/* USe this one */
	}/* Accumlate all images */

      if(tot > 0)
	{/* OK to calc */
	  globMean = totMean / (float)tot;
	  globStd  = totStd  / (float)tot;
	}/* OK to calc */
      else
	{/* Error */
	  if(normLinSwitch)
	    printf("Nothing to calc.\n");
	  else
	    askThem("Nothing to calc.",FALSE,XmDIALOG_INFORMATION);
	  goto noCalc;
	}/* Error */

       /* printf("\n\nglobMin=%d  max=%d\n",globMin,globMax); 
	 printf("GlobMean=%f  globStd=%f\n",globMean,globStd);    */


      printf("Done globSTD changed to [%f]\n",globStd);    
      printf("     globMean           [%f]\n",globMean);    

      free(img_bkg);
      free(img_min);
      free(img_max);
      free(img_avg);
      free(img_std);

      if(dumpNormStats)
	{/* dump to file */
	  FILE *fDump;
	  fDump = fopen(dumpNormFile,"a");
	  if(fDump)
	    {/* ok */
	      fprintf(fDump,"[%s] mean[%f] stDev[%f]\n",
		      fileName.addr,
		      globMean,
		      globStd);
	      fclose(fDump);
	    }/* ok */
	}/* dump to file */

      /* Change cursor back to what it was */
      attrs.cursor= old_cursor;
      XChangeWindowAttributes(dsply, XtWindow(drawingWidget), 
			      CWCursor, &attrs);
      XFlush(dsply);


      /* Suppose you get the gloabl mean and StDev from the interface
       * and then just apply that here... ??? */
      sprintf(outStr,"%f",globMean);
      XmTextSetString(normalMeanW,outStr);
      sprintf(outStr,"%f",globStd);
      XmTextSetString(normalStdW,outStr);
      if(!normLinSwitch)
	raiseIt(normalScreen);

    }/* OK */
  else
    {/* Tell them why we're not doing this */
      if(normLinSwitch)
	printf("Normalize Means&Vars only for Packed boxes.\n");
      else
	askThem("Normalize Means&Vars only for Packed boxes.",
		FALSE,XmDIALOG_INFORMATION);
    }/* Tell them why we're not doing this */

  
 noCalc:;

}/* normalize_ACTIVATE */






/*******************************************************************
 * normalize_activate_2() - They want to normalize their images
 * across the entire data set. *tag==2 means they want normalization
 * to happen for Backgrounds instead of the usual way using means.
 */
void normalize_activate_2(Widget widget, int *tag, 
			  XmAnyCallbackStruct *callback_data)
{/* normalize_ACTIVATE_2 */
  int   i,j,k,tot;
  float   *img_bkg;
  float   *img_min,*img_max;
  float totMean,totStd,*img_avg,*img_std;
  float globMean, globStd, stdAdjustFactor;
  float   globMin, globMax;
  int   yBuf;
  float newScaleFactor, new_minF, new_maxF;
  int   new_min, new_max;
  int   constant;
  float   *thisBuf;
  int   xcen,ycen,xSq,ySq,jMycen,kMxcen,thisRad,distnce,thisPixel;
  float   *thisPackBuf;
  char  *imgNumText,*thisText;
  int   origBoxNum;

  thisText= XmTextGetString(normalMeanW);
  globMean = (float)strtod(thisText,0);
  XtFree(thisText);
  thisText= XmTextGetString(normalStdW);
  globStd = (float)strtod(thisText,0);
  XtFree(thisText);
  
  if(normLinSwitch)
    {/* Use command line values */
      if(normLinMean==0.0 && normLinStd==0.0)
	{/* Just use what we've got */
	  printf("Using calculated values Mean[%f] Std[%f]\n",
		 globMean,globStd);
	}/* Just use what we've got */
      else
	{/* Use command line vals */
	  globMean = normLinMean;
	  globStd  = normLinStd;
	  printf("Using command line values Mean[%f] Std[%f]\n",
		 globMean,globStd);
	}/* Use command line vals */
    }/* Use command line values */
  *tag = kindOfNormalize;
  
  imgNumText = XmTextGetString(boxInitWidget); 
  origBoxNum =  (int)strtod(imgNumText, 0);
  XtFree(imgNumText);
  
  xcen = ycen = packRadius;
  thisRad = packRadius * packRadius;
  
  /* This may take awhile so give them the watch cursor */
  attrs.cursor= watch_cursor;
  XChangeWindowAttributes(dsply, XtWindow(drawingWidget), 
			  CWCursor, &attrs);
  XFlush(dsply);
  
  CK(img_bkg= (float *)malloc(sizeof(float)*purdueHeader.nimg));
  CK(img_min= (float *)malloc(sizeof(float)*purdueHeader.nimg));
  CK(img_max= (float *)malloc(sizeof(float)*purdueHeader.nimg));
  CK(img_avg= (float *)malloc(sizeof(float)*purdueHeader.nimg));
  CK(img_std= (float *)malloc(sizeof(float)*purdueHeader.nimg));
  
  numCols = purdueHeader.numCols;
  numRows = purdueHeader.numRows;
  
  for(i=0;i<purdueHeader.nimg;i++)
    {/* Accumlate all images */
      
      sprintf(outStr,"A[%d]",i+1);
      XmTextSetString(boxInitWidget,outStr);
      
      if(!deletePackedImage[i])
	{/* USe this one */
	  calcCircleStats(packImg[i],
			  &img_bkg[i], 
			  &img_min[i],
			  &img_max[i],
			  &img_avg[i], 
			  &img_std[i],
			  numCols, numRows, numCols);
	}/* USe this one */
    }/* Accumlate all images */
  
  
  /* Suppose you get the gloabl mean and StDev from the interface
   * and then just apply that here... ??? */
  
  printf("Using globSTD  == [%f]\n",globStd);    
  printf("      globMean == [%f]\n",globMean);    
  
  
  /***************************************************************
   * Now scale each image to the new min/max and add the constant
   * to bring it in line with the global Mean.
   ***************************************************************/
  minPackPixel= maxPackPixel = globMean;
  for(i=0;i<purdueHeader.nimg;i++)
    {/* Adjust all images */
      
      sprintf(outStr,"C[%d]",i+1);
      XmTextSetString(boxInitWidget,outStr);
      
      if(!deletePackedImage[i])
	{/* USe this one */
	  
	  if(img_std[i] != 0.0)
	    stdAdjustFactor = globStd / img_std[i];
	  else stdAdjustFactor = 1.0;
	  
	  if(*tag==2)
	    {/* BKG */
	      new_min = (float)(((img_min[i] - img_bkg[i])*stdAdjustFactor)
			      + img_avg[i] + 0.49999);
	      new_max = (float)(((img_max[i] - img_bkg[i])*stdAdjustFactor)
			      + img_avg[i] + 0.49999);
	    }/* BKG */
	  else
	    {/* AVG */
	      new_min = (float)(((img_min[i] - img_avg[i])*stdAdjustFactor)
			      + img_avg[i] + 0.49999);
	      new_max = (float)(((img_max[i] - img_avg[i])*stdAdjustFactor)
			      + img_avg[i] + 0.49999);
	    }/* AVG */
	  
	  thisBuf = packImg[i];
	  for(j=0;j<numRows;j++)
	    {/* rows */
	      yBuf = j * numCols;
	      for(k=0;k<numCols;k++)
		{/* Scale all image */
		  /* Get this to expand/contract around the 
		   * new global Mean */
		  if(*tag==2)
		    thisBuf[(yBuf+k)]= 
		      (float)(globMean +
			     ((thisBuf[(yBuf+k)]-img_bkg[i]) *
			      stdAdjustFactor));
		  else
		    thisBuf[(yBuf+k)]= 
		      (float)(globMean +
			     ((thisBuf[(yBuf+k)]-img_avg[i]) *
			      stdAdjustFactor));
		  maxPackPixel= MAX(maxPackPixel,thisBuf[(yBuf+k)]);
		  minPackPixel= MIN(minPackPixel,thisBuf[(yBuf+k)]);
		}/* Scale all image */
	    }/* rows */
	  
	  
	  /* Now recalc the backGround */
	  calcCircleStats(packImg[i],
			  &img_bkg[i], 
			  &img_min[i],
			  &img_max[i],
			  &img_avg[i], 
			  &img_std[i],
			  numCols, numRows, numCols);
	  /* And apply new background */
	  thisPackBuf = packImg[i];
	  for(j=0;j<numRows;j++)
	    {/* rows */
	      jMycen = j - ycen;
	      ySq    = (jMycen*jMycen);
	      for(k=0;k<numCols;k++)
		{/* Do all */
		  kMxcen = k - xcen;
		  xSq    = (kMxcen*kMxcen);
		  distnce = (xSq+ySq);
		  if(distnce > thisRad)
		    {/* Stuff it */
		      thisPixel = ((j*numCols)+k);
		      if(forceAvgBkg)
			thisPackBuf[(thisPixel)]= (float)img_avg[i];
		      else
			thisPackBuf[(thisPixel)]= (float)img_bkg[i];
		    }/* Stuff it */
		  
		}/* Do all */
	    }/* rows */
	  
	}/* USe this one */
      
    }/* Adjust all images */
  minPackPixel -= 1.5;
  maxPackPixel += 1.5;
  if(minPackPixel < 0.0)
    scalePackFactor = (((float)tryColors - 7.0) / 
		       (maxPackPixel-minPackPixel));
  else scalePackFactor = (((float)tryColors - 7.0) / 
			  maxPackPixel);

  free(img_bkg);
  free(img_min);
  free(img_max);
  free(img_avg);
  free(img_std);
  
  sprintf(outStr,"%d",origBoxNum);
  XmTextSetString(boxInitWidget,outStr);
  
  stuffPackedImage(purdueHeader.numCols,
		   purdueHeader.numRows,
		   packImg,purdueHeader.nimg,
		   imgNum);  /* Start at radius'th image */
  
  DrawWindow(); 
  
  /* Change cursor back to what it was */
  attrs.cursor= old_cursor;
  XChangeWindowAttributes(dsply, XtWindow(drawingWidget), 
			  CWCursor, &attrs);
  XFlush(dsply);
  
  if(*tag==2)
    askThem("Normalize Bkgs&Vars finished OK.",
	    FALSE,XmDIALOG_INFORMATION);
  else
    {
      if(normLinSwitch)
	printf("Normalize Means&Vars finished OK.\n");
      else
	askThem("Normalize Means&Vars finished OK.",
		FALSE,XmDIALOG_INFORMATION);
    }
  XtUnmanageChild(normalScreen);
  
}/* normalize_ACTIVATE_2 */




/*******************************************************************
 * normal_deactivate() - Get normalization screen off the display.
 */
void normal_deactivate(Widget widget, int *tag, 
		       XmAnyCallbackStruct *callback_data)
{/* NORMAL_DEACTIVATE */

  XtUnmanageChild(normalScreen);
  
}/* NORMAL_DEACTIVATE */





/*******************************************************************
 * checkBkg_activate() - Check boxes for bad bkgs.
 */
void checkBkg_activate(Widget widget, int *tag, 
		       XmAnyCallbackStruct *callback_data)
{/* CHECKBKG_ACTIVATE */
  int i,j,k,tot;
  float *img_bkg;
  float *img_min,*img_max;
  float *img_avg,*img_std,totMean,meanVar;
  float  *thisBuf;
  int xcen,ycen,xSq,ySq,jMycen,kMxcen,thisRad,distnce,thisPixel;
  float  *thisPackBuf;
	      
  /* This may take awhile so give them the watch cursor */
  attrs.cursor= watch_cursor;
  XChangeWindowAttributes(dsply, XtWindow(drawingWidget), 
			  CWCursor, &attrs);
  XFlush(dsply);
  
  CK(img_bkg= (float *)malloc(sizeof(float)*purdueHeader.nimg));
  CK(img_min= (float *)malloc(sizeof(float)*purdueHeader.nimg));
  CK(img_max= (float *)malloc(sizeof(float)*purdueHeader.nimg));
  CK(img_avg= (float *)malloc(sizeof(float)*purdueHeader.nimg));
  CK(img_std= (float *)malloc(sizeof(float)*purdueHeader.nimg));
  
  numCols = purdueHeader.numCols;
  numRows = purdueHeader.numRows;
  
  tot = 0; totMean= 0.0;
  for(i=0;i<purdueHeader.nimg;i++)
    {/* Accumlate all images */
      calcCircleStats(packImg[i],
		      &img_bkg[i], 
		      &img_min[i],
		      &img_max[i],
		      &img_avg[i], 
		      &img_std[i],
		      numCols, numRows, numCols);
      if(img_bkg[i] != (float)*(packImg[i]))
	{/* accum */
	  tot++;
	  totMean += (float)((float)*(packImg[i])-img_bkg[i]);
	}/* accum */
    }/* Accumlate all images */
  
  if(tot > 0)
    {/* There are some to change */
  
      sprintf(outStr,"I've found [%d] boxes with bogus background values.\n\
The mean difference from the theoretical background is [%f].\n\n\
Do you want me to change them?\n",tot,(float)(totMean/(float)tot));

      if(askThem(outStr, FALSE,XmDIALOG_QUESTION))
	{/* OK let's change them */

	  for(i=0;i<purdueHeader.nimg;i++)
	    {/* Accumlate all images */
	      
	      xcen = ycen = packRadius;
	      thisRad = packRadius * packRadius;

	      /* Now recalc the backGround */
	      calcCircleStats(packImg[i],
			      &img_bkg[i], 
			      &img_min[i],
			      &img_max[i],
			      &img_avg[i], 
			      &img_std[i],
			      numCols, numRows, numCols);
	      if(img_bkg[i] != (float)*(packImg[i]))
		{/* And apply new background */
		  thisPackBuf = packImg[i];
		  for(j=0;j<numRows;j++)
		    {/* rows */
		      jMycen = j - ycen;
		      ySq    = (jMycen*jMycen);
		      for(k=0;k<numCols;k++)
			{/* Do all */
			  kMxcen = k - xcen;
			  xSq    = (kMxcen*kMxcen);
			  distnce = (xSq+ySq);
			  if(distnce > thisRad)
			    {/* Stuff it */
			      thisPixel = ((j*numCols)+k);
			      thisPackBuf[(thisPixel)]= (float)img_bkg[i];
			    }/* Stuff it */
			  
			}/* Do all */
		    }/* rows */
		}/* And apply new background */
	      
	    }/* Accumlate all images */
	}/* OK let's change them */
    }/* There are some to change */
  else
    {/* Nothing to change */

      askThem("All backgrounds checked out OK.",
	      FALSE,XmDIALOG_INFORMATION);

    }/* Nothing to change */

  free(img_bkg);
  free(img_min);
  free(img_max);
  free(img_avg);
  free(img_std);
  
  
  stuffPackedImage(purdueHeader.numCols,
		   purdueHeader.numRows,
		   packImg,purdueHeader.nimg,
		   imgNum);  /* Start at radius'th image */
  
  DrawWindow(); 
  
  /* Change cursor back to what it was */
  attrs.cursor= old_cursor;
  XChangeWindowAttributes(dsply, XtWindow(drawingWidget), 
			  CWCursor, &attrs);
  XFlush(dsply);
  
  
}/* CHECKBKG_ACTIVATE */




/*************************************************************
 * mb_scroll_change() - Text widgets form the model building
 * screen have changed.
 */
void mb_scroll_change(Widget widget, int *tag, 
		      XmScrollBarCallbackStruct *callback_data)
{/* MB_SCROLL_CHANGE */
  int thisVal,i,j,thisOne,secOut,rowOut,colOut;
  int intVal;
  XmString simpleString;

  intVal = (int)callback_data->value;
  sprintf(outStr,"%d",intVal);

  if(*tag==ROTATETSCROLL)
    {
      XmTextSetString(mbRotatetTextW, outStr); 
    }
  else if(*tag==ROTATEPSCROLL)
    {
      XmTextSetString(mbRotatepTextW, outStr); 
    }
  else if(*tag==ROTATEOSCROLL)
    {
      XmTextSetString(mbRotateoTextW, outStr); 
    }
  else if(*tag==DIMENSIONXSCROLL)
    {
      XmTextSetString(mbDimensionxTextW, outStr); 
    }
  else if(*tag==DIMENSIONYSCROLL)
    {
      XmTextSetString(mbDimensionyTextW, outStr); 
    }
  else if(*tag==DIMENSIONZSCROLL)
    {
      XmTextSetString(mbDimensionzTextW, outStr); 
    }
  else if(*tag==CENTERXSCROLL)
    {
      XmTextSetString(mbCenterxTextW, outStr); 
    }
  else if(*tag==CENTERYSCROLL)
    {
      XmTextSetString(mbCenteryTextW, outStr); 
    }
  else if(*tag==CENTERZSCROLL)
    {
      XmTextSetString(mbCenterzTextW, outStr); 
    }
  else if(*tag==DIFFSEC2SCROLL)
    {
      //      mbCenter2ScrollValue = floatVal / 100.0;
      // sprintf(outStr,"%.2f",mbCenter2ScrollValue);
      // XmTextSetString(mbCenter2TextWidget, outStr); 
    }

  if(TRUE)
    {/* OK to display the goods */

      if(mbImmediate)
	model_activate(widget, tag, callback_data);
      
    }/* OK to display the goods */

}/* MB_SCROLL_CHANGE */




/*************************************************************
 * modelObject_change() - They want to change modelObject.
 * 
*/
void modelObject_change(Widget widget, int *tag, 
			XmScrollBarCallbackStruct *callback_data)
{/* MODELOBJECT_CHANGE */
  XmString simpleString;
  
  modelType = *tag;
  if(*tag==SQUARE)
    {/* square */
      sprintf(outStr,"Object to build:  Cube");
      XtManageChild(mbDimensionyScrollW);
      XtManageChild(mbDimensionyTextW);
      XtManageChild(mbDimensionzScrollW);
      XtManageChild(mbDimensionzTextW);
      XtManageChild(mbRotatetScrollW);
      XtManageChild(mbRotatetTextW);
      XtManageChild(mbRotatepScrollW);
      XtManageChild(mbRotatepTextW);
      XtManageChild(mbRotateoScrollW);
      XtManageChild(mbRotateoTextW);
    }/* square */
  else if(*tag==CIRCLE)
    {/* circle */
      sprintf(outStr,"Object to build:  Sphere");
      XtUnmanageChild(mbDimensionyScrollW);
      XtUnmanageChild(mbDimensionyTextW); 
      XtUnmanageChild(mbDimensionzScrollW);
      XtUnmanageChild(mbDimensionzTextW);
      XtUnmanageChild(mbRotatetScrollW);
      XtUnmanageChild(mbRotatetTextW);
      XtUnmanageChild(mbRotatepScrollW);
      XtUnmanageChild(mbRotatepTextW);
      XtUnmanageChild(mbRotateoScrollW);
      XtUnmanageChild(mbRotateoTextW);
    }/* circle */
  else if(*tag==ICOS)
    {/* icos */
      sprintf(outStr,"Object to build:  Icosahedron");
      XtUnmanageChild(mbDimensionyScrollW);
      XtUnmanageChild(mbDimensionyTextW);
      XtManageChild(mbDimensionzScrollW);
      XtManageChild(mbDimensionzTextW);
      XtManageChild(mbRotatetScrollW);
      XtManageChild(mbRotatetTextW);
      XtManageChild(mbRotatepScrollW);
      XtManageChild(mbRotatepTextW);
      XtManageChild(mbRotateoScrollW);
      XtManageChild(mbRotateoTextW);
    }/* icos */
  else if(*tag==CYLINDER)
    {/* cylinder */
      sprintf(outStr,"Object to build:  Cylinder");
      XtUnmanageChild(mbDimensionyScrollW);
      XtUnmanageChild(mbDimensionyTextW);
      XtManageChild(mbDimensionzScrollW);
      XtManageChild(mbDimensionzTextW);
      XtManageChild(mbRotatetScrollW);
      XtManageChild(mbRotatetTextW);
      XtManageChild(mbRotatepScrollW);
      XtManageChild(mbRotatepTextW);
      XtManageChild(mbRotateoScrollW);
      XtManageChild(mbRotateoTextW);
    }/* cylinder */
  else if(*tag==TRIANGLE)
    {/* triangle */
      sprintf(outStr,"Object to build:  Triangle");
      XtManageChild(mbDimensionyScrollW);
      XtManageChild(mbDimensionyTextW);
      XtManageChild(mbDimensionzScrollW);
      XtManageChild(mbDimensionzTextW);
      XtManageChild(mbRotatetScrollW);
      XtManageChild(mbRotatetTextW);
      XtManageChild(mbRotatepScrollW);
      XtManageChild(mbRotatepTextW);
      XtManageChild(mbRotateoScrollW);
      XtManageChild(mbRotateoTextW);
    }/* triangle */
  else
    {/* ?? */
      printf ("MODELOBJECT_CHANGE - Unknown object\n");
      modelType = SQUARE;  /* Default it here */
      sprintf(outStr,"Bug in object code.");
    }/* ?? */


  createSimpleString(&simpleString,outStr);
  XtSetArg(arglst[0], XmNlabelString, simpleString);
  XtSetValues(modelLabelW, arglst, 1); 
  XFlush(dsply);
  XmStringFree(simpleString); 


}/* MODELOBJECT_CHANGE */




/*******************************************************************
 * model_screen_activate() - Get the Model building screen up
 * on the display.
 */
void model_screen_activate(Widget widget, int *tag, 
			   XmAnyCallbackStruct *callback_data)
{/* MODEL_SCREEN_ACTIVATE */
  int xCen,yCen,zCen;

  if(model_undo_buffer)
    {free(model_undo_buffer);
    model_undo_buffer = 0;}
  CK(model_undo_buffer = (float *)malloc(sizeof(float)*
					 numCols*numRows*numSecs));
  for(i=0;i<(numSecs*numRows*numCols);i++)
    model_undo_buffer[i] = float_image_buffer1[i];

  raiseIt(statusScreenWidget);
  XFlush(dsply);
  XSync(dsply,0);
  checkStatus("Configuring widgets ...");

  xCen = (numCols / 2) + 1 ;  /* +1 for human user */
  yCen = (numRows / 2) + 1 ;  /* +1 for human user */
  zCen = (origNumSecs / 2) + 1 ;  /* +1 for human user */
  XtVaSetValues(mbDimensionxScrollW,XmNmaximum, (int)((numCols)+10),NULL);
  XtVaSetValues(mbDimensionxScrollW,XmNsliderSize, (int)(10),NULL);
		
  XtVaSetValues(mbDimensionyScrollW,XmNmaximum, (int)((numRows)+10),NULL);
  XtVaSetValues(mbDimensionyScrollW,XmNsliderSize, (int)(10),NULL);
		
  XtVaSetValues(mbDimensionzScrollW,XmNmaximum, (int)((origNumSecs)+10),NULL);
  XtVaSetValues(mbDimensionzScrollW,XmNsliderSize, (int)(10),NULL);
		
  XtVaSetValues(mbCenterxScrollW,XmNmaximum, (int)((numCols)+10),NULL);
  XtVaSetValues(mbCenterxScrollW,XmNsliderSize, (int)(10),NULL);
  XtVaSetValues(mbCenterxScrollW,XmNvalue, (int)(xCen),NULL);
		
  XtVaSetValues(mbCenteryScrollW,XmNmaximum, (int)((numRows)+10),NULL);
  XtVaSetValues(mbCenteryScrollW,XmNsliderSize, (int)(10),NULL);
  XtVaSetValues(mbCenteryScrollW,XmNvalue, (int)(yCen),NULL);
		
  XtVaSetValues(mbCenterzScrollW,XmNmaximum, (int)((origNumSecs)+10),NULL);
  XtVaSetValues(mbCenterzScrollW,XmNsliderSize, (int)(10),NULL);
  XtVaSetValues(mbCenterzScrollW,XmNvalue, (int)(zCen),NULL);
		
  sprintf(outStr,"%d",xCen);
  XmTextSetString(mbCenterxTextW, outStr); 
  sprintf(outStr,"%d",yCen);
  XmTextSetString(mbCenteryTextW, outStr); 
  sprintf(outStr,"%d",zCen);
  XmTextSetString(mbCenterzTextW, outStr); 

  raiseIt(mapShadeScreenWidget);
  raiseIt(statusScreenWidget);
  XFlush(dsply);
  XSync(dsply,0);

  mapShadeScreen_activate(0,0,0);
  XFlush(dsply);
  raiseIt(modelScreenWidget);

  mapShade_activate(0,tag,0);      

  XtUnmanageChild(statusScreenWidget);


}/* MODEL_SCREEN_ACTIVATE */



/* ************************************************************
 * model_deactivate() - Get rid of Model Bulding interface
 * from the display.
 */
void model_deactivate(Widget widget, int *tag, 
		      XmAnyCallbackStruct *callback_data)
{/* MODEL_DEACTIVATE */
  
  if(model_undo_buffer)
    {free(model_undo_buffer);
    model_undo_buffer = 0;}
  XtUnmanageChild(modelScreenWidget); 
  XtUnmanageChild(mapShadeScreenWidget); 
  mapShadeScreen_deactivate(0,0,0);

}/* MODEL_DEACTIVATE */





/*******************************************************************
 * model_activate() - This is the "Tryit" button.
 */
void model_activate(Widget widget, int *tag, 
		    XmAnyCallbackStruct *callback_data)
{/* MODEL_ACTIVATE */
  int i,j,k,newI,newJ,newK;
  int xCen,yCen,zCen,dimX,dimY,dimZ;
  int theta,phi,omega;
  int begX,begY,begZ,endX,endY,endZ;
  int zSq,ySq,xSq,dimXsq;
  int iter,numIter;
  double thisCol,thisRow,thisSecn;
  double t1,t2,t3,t4,t5,t6,t7,t8,t9;
  double ctheta,cphi,comega,stheta,sphi,somega;
  float modelFillValuePlus;
  float thetas[25], phis[25];
  char  *mFillText;

  mFillText = XmTextGetString(modelFillW); 
  modelFillValue =  (float)strtod(mFillText, 0);
  XtFree(mFillText);
  modelFillValuePlus = modelFillValue*1.01;

  mFillText = XmTextGetString(mbCenterxTextW); 
  xCen =  (int)strtod(mFillText, 0);
  XtFree(mFillText);
  mFillText = XmTextGetString(mbCenteryTextW); 
  yCen =  (int)strtod(mFillText, 0);
  XtFree(mFillText);
  mFillText = XmTextGetString(mbCenterzTextW); 
  zCen =  (int)strtod(mFillText, 0);
  XtFree(mFillText);
  mFillText = XmTextGetString(mbDimensionxTextW); 
  dimX =  (int)strtod(mFillText, 0);
  XtFree(mFillText);
  mFillText = XmTextGetString(mbDimensionyTextW); 
  dimY =  (int)strtod(mFillText, 0);
  XtFree(mFillText);
  mFillText = XmTextGetString(mbDimensionzTextW); 
  dimZ =  (int)strtod(mFillText, 0);
  XtFree(mFillText);
  mFillText = XmTextGetString(mbRotatetTextW); 
  theta =  (int)strtod(mFillText, 0);
  XtFree(mFillText);
  mFillText = XmTextGetString(mbRotatepTextW); 
  phi   =  (int)strtod(mFillText, 0);
  XtFree(mFillText);
  mFillText = XmTextGetString(mbRotateoTextW); 
  omega =  (int)strtod(mFillText, 0);
  XtFree(mFillText);

  dimXsq = dimX*dimX;

  xCen--;  /* Now in computer space */
  yCen--;
  zCen--;
  begX = xCen - (dimX/2);
  begY = yCen - (dimY/2);
  begZ = zCen - (dimZ/2);
  endX = begX + dimX;
  endY = begY + dimY;
  endZ = begZ + dimZ;

  /* We've got a bug whereby we've got a lot of problems with
   * the original map_prj_all() rtn that does the rotation. */
  /* WHY does phi not have to have the sign reversed here like
   * it does in emUnix.c in the thetaPhi_dphshd_C() rtn??? */
  comega = (double)cos((double)deg2rad((double)phi));
  somega = (double)sin((double)deg2rad((double)phi));
  cphi = (double)cos((double)deg2rad((double)omega));
  sphi = (double)sin((double)deg2rad((double)omega));
  ctheta = (double)cos((double)deg2rad((double)0.0-theta));
  stheta = (double)sin((double)deg2rad((double)0.0-theta));

  /*********************************************
   *    COMPUTE ELEMENTS OF ROTATION MATRIX    *
   *********************************************/
  t1 = ctheta*cphi*comega + sphi*somega;
  t2 = ctheta*sphi*comega - cphi*somega;
  t3 = -stheta*comega;
  t4 = ctheta*cphi*somega - sphi*comega;
  t5 = ctheta*sphi*somega + cphi*comega;
  t6 = -stheta*somega;
  t7 = stheta*cphi;
  t8 = stheta*sphi;
  t9 = ctheta;

  raiseIt(statusScreenWidget);
  XFlush(dsply);

  /* Always refresh from last saved */
  for(i=0;i<(origNumSecs*numRows*numCols);i++)
    float_image_buffer1[i] = model_undo_buffer[i] ;

  numIter = 1;
  if(isIcosRepl)
    {/* Let's find the vertices we need */
      float thisRad;
      int thisXcen,thisYcen,thisZcen;

      printf("Doing icos repl \n");
      thisRad = (float)xCen;
      thisXcen = numCols / 2;
      thisYcen = numRows / 2;
      thisZcen = origNumSecs / 2;

      /* Stuff the icosahedron values */
      get_verticies();

      /* OK let's change the coords now */
      for (i=0;i<icosabox.num_points;i++) 
	{/* DO all */
	  icosabox.vertex[i].lx *= thisRad;
	  icosabox.vertex[i].ly *= thisRad;
	  icosabox.vertex[i].lz *= thisRad;
	  /* Now put it in our box */
	  icosabox.vertex[i].lx += thisXcen;
	  icosabox.vertex[i].ly += thisYcen;
	  icosabox.vertex[i].lz += thisZcen;

	  thetas[i] = (float)(icosabox.vertex[i].lx - thisXcen) /
	    (float)(icosabox.vertex[i].lz - thisZcen) ;
	  thetas[i] = atan(thetas[i]);
	  thetas[i] = rad2deg(thetas[i]);
	  phis[i] = (float)(icosabox.vertex[i].ly - thisYcen) /
	    (float)(icosabox.vertex[i].lx - thisXcen) ;
	  phis[i] = atan(phis[i]);
	  phis[i] = rad2deg(phis[i]);

	}/* DO all */
      
      numIter = icosabox.num_points;

    }/* Let's find the vertices we need */

  /* OK - create a new object for every replication we may need */
  for(iter=0;iter<numIter;iter++)
    {/* DO all iter */
      if(isIcosRepl)
	{/* Let's find the centers */
	  sprintf(outStr,"Placing Icos position[%d].",iter+1);
	  checkStatus(outStr);
	  xCen = icosabox.vertex[iter].lx;
	  yCen = icosabox.vertex[iter].ly;
	  zCen = icosabox.vertex[iter].lz;
	  begX = xCen - (dimX/2);
	  begY = yCen - (dimY/2);
	  begZ = zCen - (dimZ/2);
	  endX = begX + dimX;
	  endY = begY + dimY;
	  endZ = begZ + dimZ;

	  /* Now do the rotation */
	  omega = 0.0;
	  /* Why do we have to do this for this case ??? */
	  if(thetas[iter] > -0.0001 && thetas[iter] < 0.0001)
	    {/* */
	      if(iter ==  5 || iter == 7)
		thetas[iter] = 0.0 - (90.0 - 31.717468);
	      else
		thetas[iter] = 90.0 - 31.717468;
	    }/* */
	  comega = (double)cos((double)deg2rad((double)phis[iter]));
	  somega = (double)sin((double)deg2rad((double)phis[iter]));
	  cphi = (double)cos((double)deg2rad((double)omega));
	  sphi = (double)sin((double)deg2rad((double)omega));
	  ctheta = (double)cos((double)deg2rad((double)0.0-thetas[iter]));
	  stheta = (double)sin((double)deg2rad((double)0.0-thetas[iter]));
	  
	  /*********************************************
	   *    COMPUTE ELEMENTS OF ROTATION MATRIX    *
	   *********************************************/
	  t1 = ctheta*cphi*comega + sphi*somega;
	  t2 = ctheta*sphi*comega - cphi*somega;
	  t3 = -stheta*comega;
	  t4 = ctheta*cphi*somega - sphi*comega;
	  t5 = ctheta*sphi*somega + cphi*comega;
	  t6 = -stheta*somega;
	  t7 = stheta*cphi;
	  t8 = stheta*sphi;
	  t9 = ctheta;
	  
	}/* Let's find the centers */

      if(modelType==SQUARE)
	{/* do cube */
	  for(i=begZ;i<endZ;i++)
	    {/* secs */
	      for(j=begY;j<endY;j++)
		{/* rows */
		  for(k=begX;k<endX;k++)
		    {/* cols */
		      newTranslateThetaPhiOmega(&thisCol,&thisRow,&thisSecn,
						xCen,yCen,zCen,
						k,j,i,
						t1,t2,t3,t4,t5,t6,t7,t8,t9);
		      float_image_buffer1[(((int)thisSecn*numCols*numRows)+
					   ((int)thisRow*numCols)+ 
					   (int)thisCol)] = modelFillValue;
		    }/* cols */
		}/* rows */
	    }/* secs */
	  /* We have to corrupt a corner pixel to make sure our
	   * shading algorithm can display it when not rotated */
	  //      newTranslateThetaPhiOmega(&thisCol,&thisRow,&thisSecn,
	  //			xCen,yCen,zCen,
	  //			begX,begY,(endZ-1),
	  //			t1,t2,t3,t4,t5,t6,t7,t8,t9);
	  //      float_image_buffer1[(((int)thisSecn*numCols*numRows)+
	  //		   ((int)thisRow*numCols)+ 
	  //		   (int)thisCol)] = 0.0;
	}/* do cube */
      else if(modelType==CIRCLE)
	{/* do sphere */
	  for(i=0;i<origNumSecs;i++)
	    {/* secs */
	      zSq = (i-zCen)*(i-zCen);
	      for(j=0;j<numRows;j++)
		{/* rows */
		  ySq = (j-yCen)*(j-yCen);
		  for(k=0;k<numCols;k++)
		    {/* cols */
		      xSq = (k-xCen)*(k-xCen) + ySq + zSq;
		      if(xSq < dimXsq)
			float_image_buffer1[((i*numCols*numRows)+
					     (j*numCols)+ 
					     k)] = modelFillValue;
		    }/* cols */
		}/* rows */
	    }/* secs */
	  
	}/* do sphere */
      else if(modelType==TRIANGLE)
	{/* do triangle */
	  int *triMask;
	  float slopeLeft,interLeft,slopeRight,interRight;
	  float thisY;
	  
	  triMask = (int *)malloc(numCols*numRows*sizeof(int));
	  
	  for(j=0;j<numRows;j++)
	    for(k=0;k<numCols;k++)
	      triMask[((j*numCols)+k)] = TRUE;
	  
	  if(xCen != begX && xCen != endX)
	    {/* OK to do */
	      slopeLeft = (float)(endY - begY) / (float)(xCen - begX);
	      interLeft = (float)begY - (float)(slopeLeft * (float)begX);
	      slopeRight = (float)(endY - begY) / (float)(xCen - endX);
	      interRight = (float)((float)begY-
				   (float)(slopeRight*(float)endX));
	      
	      for(j=begY;j<endY;j++)
		for(k=begX;k<endX;k++)
		  {/* fill in mask */
		    if(k <= xCen)
		      {/* left */
			thisY = k*slopeLeft+interLeft;
			if(j >= (int)thisY)
			  triMask[((j*numCols)+k)] = FALSE;
		      }/* left */
		    else if(k > xCen)
		      {/* right */
			thisY = k*slopeRight+interRight;
			if(j >= (int)thisY)
			  triMask[((j*numCols)+k)] = FALSE;
		      }/* right */
		    else
		      triMask[((j*numCols)+k)] = FALSE;
		  }/* fill in mask */
	      
	      /* Now that we have the mask just rotate it and fill buffer */
	      for(i=begZ;i<endZ;i++)
		for(j=begY;j<endY;j++)
		  for(k=begX;k<endX;k++)
		    {/* fill in buffer */
		      if(triMask[((j*numCols)+k)])
			{/* OK to fill */
			  newTranslateThetaPhiOmega(&thisCol,&thisRow,
						    &thisSecn,
						    xCen,yCen,zCen,
						    k,j,i,
						    t1,t2,t3,t4,t5,
						    t6,t7,t8,t9);
			  float_image_buffer1[(((int)thisSecn*numCols*numRows)+
					       ((int)thisRow*numCols)+ 
					       (int)thisCol)] = modelFillValue;
			}/* OK to fill */
		    }/* fill in buffer */
	      /* We have to corrupt a corner pixel to make sure our
	       * shading algorithm can display it when not rotated */
	      //  newTranslateThetaPhiOmega(&thisCol,&thisRow,&thisSecn,
	      //		    xCen,yCen,zCen,
	      //		    begX+1,begY+1,(endZ-1),
	      //		    t1,t2,t3,t4,t5,t6,t7,t8,t9);
	      //float_image_buffer1[(((int)thisSecn*numCols*numRows)+
	      //	       ((int)thisRow*numCols)+ 
	      //	       (int)thisCol)] = 0.0;
	    }/* OK to do */
	  
	  free(triMask);
	}/* do triangle */
      else if(modelType==CYLINDER)
	{/* do cylinder */
	  
	  for(i=begZ;i<endZ;i++)
	    {/* secs */
	      for(j=0;j<numRows;j++)
		{/* rows */
		  ySq = (j-yCen)*(j-yCen);
		  for(k=0;k<numCols;k++)
		    {/* cols */
		      xSq = (k-xCen)*(k-xCen) + ySq;
		      if(xSq < dimXsq)
			{/* OK to do */
			  newTranslateThetaPhiOmega(&thisCol,&thisRow,
						    &thisSecn,
						    xCen,yCen,zCen,
						    k,j,i,
						    t1,t2,t3,t4,t5,
						    t6,t7,t8,t9);
			  float_image_buffer1[(((int)thisSecn*numCols*numRows)+
					       ((int)thisRow*numCols)+ 
					       (int)thisCol)] = modelFillValue;
			}/* OK to do */
		    }/* cols */
		}/* rows */
	    }/* secs */
	  /* We have to corrupt a corner pixel to make sure our
	   * shading algorithm can display it when not rotated */
	  // newTranslateThetaPhiOmega(&thisCol,&thisRow,&thisSecn,
	  //			xCen,yCen,zCen,
	  //			begX+1,yCen,(endZ-1),
	  //		    	t1,t2,t3,t4,t5,t6,t7,t8,t9);
	  // float_image_buffer1[(((int)thisSecn*numCols*numRows)+
	  //	   ((int)thisRow*numCols)+ 
	  //		   (int)thisCol)] = 0.0;
	  
	}/* do cylinder */
      else if(modelType==ICOS)
	{/* do icos */
	  double thisX,thisY,thisZ,thisRad,thisDepth;
	  int iter;
	  
	  sprintf(outStr,"Calculating Icos shell.");
	  checkStatus(outStr);
	  
	  for(iter = 0 ; iter < dimZ; iter++)
	    {/* Do for this thickness */
	      dimX -= 1;
	      if(dimX >= 0)
		{/* OK to do */
		  for(i=0;i<origNumSecs;i++)
		    {/* secs */
		      for(j=0;j<numRows;j++)
			{/* rows */
			  ySq = (j-yCen)*(j-yCen);
			  for(k=0;k<numCols;k++)
			    {/* cols */
			      xSq = (k-xCen)*(k-xCen) + ySq;
			      thisX = (double)k - (double)xCen;
			      thisY = (double)j - (double)yCen;
			      thisZ = (double)i - (double)zCen;
			      thisRad = (double)dimX;
			      if(doIcosshell(thisX,thisY,thisZ,
					     thisRad,thisDepth))
				{/* OK to do */
				  newTranslateThetaPhiOmega(&thisCol,&thisRow,
							    &thisSecn,
							    xCen,yCen,zCen,
							    k,j,i,
							    t1,t2,t3,t4,t5,
							    t6,t7,t8,t9);
				  float_image_buffer1[(((int)thisSecn*numCols*
							numRows)+
						       ((int)thisRow*numCols)+ 
						       (int)thisCol)] = 
				    modelFillValue;
				}/* OK to do */
			    }/* cols */
			}/* rows */
		    }/* secs */
		}/* OK to do */
	    }/* Do for this thickness */
	  
	}/* do icos */

    }/* DO all iter */
  
  /* We've got "holes" in the objects that are rotated to
   * the icosahedral vertices.  Do a nearest neighbor search
   * to fill in the holes. */
  if(isIcosRepl)
    {/* Fill in the blanks */
      int numGood,thisPixel,curPixel,nrnc;

      sprintf(outStr,"Cleaning up.");
      checkStatus(outStr);
	  
      nrnc = numRows*numCols;
      for(i=1;i<origNumSecs-1;i++)
	{/* secs */
	  for(j=1;j<numRows-1;j++)
	    {/* rows */
	      for(k=1;k<numCols-1;k++)
		{/* cols */
		  numGood = 0;
		  thisPixel = i*nrnc+j*numCols + k;
		  if(float_image_buffer1[thisPixel] != modelFillValue)
		    {/* Now accum the nearest neighbors */
		      /* (i-1) */
		      curPixel = (i-1)*nrnc+(j-1)*numCols + (k-1);
		      if(float_image_buffer1[curPixel] == modelFillValue)
			numGood++;
		      curPixel = (i-1)*nrnc+(j-1)*numCols + (k);
		      if(float_image_buffer1[curPixel] == modelFillValue)
			numGood++;
		      curPixel = (i-1)*nrnc+(j-1)*numCols + (k+1);
		      if(float_image_buffer1[curPixel] == modelFillValue)
			numGood++;
		      curPixel = (i-1)*nrnc+(j)*numCols + (k-1);
		      if(float_image_buffer1[curPixel] == modelFillValue)
			numGood++;
		      curPixel = (i-1)*nrnc+(j)*numCols + (k);
		      if(float_image_buffer1[curPixel] == modelFillValue)
			numGood++;
		      curPixel = (i-1)*nrnc+(j)*numCols + (k+1);
		      if(float_image_buffer1[curPixel] == modelFillValue)
			numGood++;
		      curPixel = (i-1)*nrnc+(j+1)*numCols + (k-1);
		      if(float_image_buffer1[curPixel] == modelFillValue)
			numGood++;
		      curPixel = (i-1)*nrnc+(j+1)*numCols + (k);
		      if(float_image_buffer1[curPixel] == modelFillValue)
			numGood++;
		      if(numGood > 1)
			{/* We can't pass unless we've gotten one by now */
			  curPixel = (i-1)*nrnc+(j+1)*numCols + (k+1);
			  if(float_image_buffer1[curPixel] == modelFillValue)
			    numGood++;

			  /* (i) */
			  curPixel = (i)*nrnc+(j-1)*numCols + (k-1);
			  if(float_image_buffer1[curPixel] == modelFillValue)
			    numGood++;
			  curPixel = (i)*nrnc+(j-1)*numCols + (k);
			  if(float_image_buffer1[curPixel] == modelFillValue)
			    numGood++;
			  curPixel = (i)*nrnc+(j-1)*numCols + (k+1);
			  if(float_image_buffer1[curPixel] == modelFillValue)
			    numGood++;
			  curPixel = (i)*nrnc+(j)*numCols + (k-1);
			  if(float_image_buffer1[curPixel] == modelFillValue)
			    numGood++;
			  curPixel = (i)*nrnc+(j)*numCols + (k);
			  if(float_image_buffer1[curPixel] == modelFillValue)
			    numGood++;
			  curPixel = (i)*nrnc+(j)*numCols + (k+1);
			  if(float_image_buffer1[curPixel] == modelFillValue)
			    numGood++;
			  curPixel = (i)*nrnc+(j+1)*numCols + (k-1);
			  if(float_image_buffer1[curPixel] == modelFillValue)
			    numGood++;
			  curPixel = (i)*nrnc+(j+1)*numCols + (k);
			  if(float_image_buffer1[curPixel] == modelFillValue)
			    numGood++;
			  curPixel = (i)*nrnc+(j+1)*numCols + (k+1);
			  if(float_image_buffer1[curPixel] == modelFillValue)
			    numGood++;

			  /* (i+1) */
			  curPixel = (i+1)*nrnc+(j-1)*numCols + (k-1);
			  if(float_image_buffer1[curPixel] == modelFillValue)
			    numGood++;
			  curPixel = (i+1)*nrnc+(j-1)*numCols + (k);
			  if(float_image_buffer1[curPixel] == modelFillValue)
			    numGood++;
			  curPixel = (i+1)*nrnc+(j-1)*numCols + (k+1);
			  if(float_image_buffer1[curPixel] == modelFillValue)
			    numGood++;
			  curPixel = (i+1)*nrnc+(j)*numCols + (k-1);
			  if(float_image_buffer1[curPixel] == modelFillValue)
			    numGood++;
			  curPixel = (i+1)*nrnc+(j)*numCols + (k);
			  if(float_image_buffer1[curPixel] == modelFillValue)
			    numGood++;
			  curPixel = (i+1)*nrnc+(j)*numCols + (k+1);
			  if(float_image_buffer1[curPixel] == modelFillValue)
			    numGood++;
			  curPixel = (i+1)*nrnc+(j+1)*numCols + (k-1);
			  if(float_image_buffer1[curPixel] == modelFillValue)
			    numGood++;
			  curPixel = (i+1)*nrnc+(j+1)*numCols + (k);
			  if(float_image_buffer1[curPixel] == modelFillValue)
			    numGood++;
			  curPixel = (i+1)*nrnc+(j+1)*numCols + (k+1);
			  if(float_image_buffer1[curPixel] == modelFillValue)
			    numGood++;
			  
			  if(numGood > 18)
			    float_image_buffer1[thisPixel] = modelFillValue;
			}/* We can't pass unless we've gotten one by now */

		    }/* Now accum the nearest neighbors */
		}/* cols */
	    }/* rows */
	}/* secs */

      
    }/* Fill in the blanks */

  XtUnmanageChild(statusScreenWidget);
  XFlush(dsply);
  mapShade_activate(0,tag,0);      

}/* MODEL_ACTIVATE */





/*******************************************************************
 * model_save_activate() - This is the "Save" button.
 */
void model_save_activate(Widget widget, int *tag, 
			 XmAnyCallbackStruct *callback_data)
{/* MODEL_SAVE_ACTIVATE */
  int i,j,k;

  /* Save what we have */
  for(i=0;i<(origNumSecs*numRows*numCols);i++)
    model_undo_buffer[i] = float_image_buffer1[i];

}/* MODEL_SAVE_ACTIVATE */




/*******************************************************************
 * model_undo_activate() - Undo last model operation.
 */
void model_undo_activate(Widget widget, int *tag, 
			 XmAnyCallbackStruct *callback_data)
{/* MODEL_UNDO_ACTIVATE */

  for(i=0;i<(origNumSecs*numRows*numCols);i++)
    float_image_buffer1[i] = model_undo_buffer[i] ;
  mapShade_activate(0,tag,0);      


}/* MODEL_UNDO_ACTIVATE */




/*******************************************************************
 * model_clear_activate() - CLear all of the MAP volume.
 */
void model_clear_activate(Widget widget, int *tag, 
			  XmAnyCallbackStruct *callback_data)
{/* MODEL_CLEAR_ACTIVATE */

  for(i=0;i<(origNumSecs*numRows*numCols);i++)
    float_image_buffer1[i] = 0.0;

  mapShade_activate(0,tag,0);      

}/* MODEL_CLEAR_ACTIVATE */





/*******************************************************************
 * binning_screen_activate() - Get the binning screen up on
 * the display.
 */
void binning_screen_activate(Widget widget, int *tag, 
			     XmAnyCallbackStruct *callback_data)
{/* BINNING_SCREEN_ACTIVATE */

  raiseIt(binningScreenWidget);

}/* BINNING_SCREEN_ACTIVATE */



/*******************************************************************
 * binning_activate() - Do the binning operation.
 */
void binning_activate(Widget widget, int *tag, 
		      XmAnyCallbackStruct *callback_data)
{/* BINNING_ACTIVATE */
  float *tmpBuf;
  int i,j,k,newHeight,newWidth,binStep;
  int m,n,o,numGrid;
  float minF,maxF,totPixels;

  if(doBinning==BIN2X2)
    {
      printf("Do 2x2\n");
      newHeight = file_height / 2;
      newWidth  = file_width / 2;
      binStep = 2;
    }
  else if(doBinning==BIN3X3)
    {
      printf("Do 3x3\n");
      newHeight = file_height / 3;
      newWidth  = file_width / 3;
      binStep = 3;
    }
  else if(doBinning==BIN4X4)
    {
      printf("Do 4x4\n");
      newHeight = file_height / 4;
      newWidth  = file_width / 4;
      binStep = 4;
    }
  else
    {/* error */
      askThem("You must choose the Binning Size first.",
	      FALSE,XmDIALOG_INFORMATION);
      goto noBin;
    }/* error */

  XtUnmanageChild(binningScreenWidget); 
  XFlush(dsply);

  numGrid = binStep * binStep;

  for(i=0;i<file_height;i+=binStep)
    {/* rows */
      for(j=0;j<file_width;j+=binStep)
	{/* cols */
	  if(i+binStep <file_height && j+binStep <file_width)
	    {/* OK to do */
	      totPixels = 0.0;
	      for(m=0;m<binStep;m++)
		for(n=0;n<binStep;n++)
		  {/* tot them up */
		    totPixels += (float)inCore_buffer[(((i+m)*file_width)+
						       (j+n))] ;
		  }/* tot them up */
	      inCore_buffer[(((i/binStep)*newWidth)+(j/binStep))] = 
		(float)totPixels / numGrid;
	    }/* OK to do */
	}/* cols */
    }/* rows */
  
  /* Now adjust concept of image size */
  purdueHeader.numCols = file_width  = newWidth;
  purdueHeader.numRows = file_height = newHeight;
  
  /* Make it generate a new fullscreen buffer */
  fullScreen_buffer.dispx = fullScreen_buffer.dispy = 0;
  virtual_read_activate(0,0);
  
  DrawWindow(); 
	  
 noBin:;


}/* BINNING_ACTIVATE */




/*******************************************************************
 * interp_screen_activate() - Get interpolate screen up.
 */
void interp_screen_activate(Widget widget, int *tag, 
			    XmAnyCallbackStruct *callback_data)
{/* INTERP_SCREEN_ACTIVATE */

  raiseIt(interpScreenWidget);

}/* INTERP_SCREEN_ACTIVATE */



/*******************************************************************
 * interpolate_activate() - Do the actual interpolation.
 */
void interpolate_activate(Widget widget, int *tag, 
			  XmAnyCallbackStruct *callback_data)
{/* INTERPOLATE_ACTIVATE */
  int i,j,k,offset,new_width,new_height,success;
  float maxF,minF,scaleF,range,origSize,outputSize;
  float rx,ry,ix,iy,
	  rotation,
	  scale,
	  filler;
  float *tmpBuf;
  char fileStr[512];
  char *inStr, *s1;

  XtUnmanageChild(interpScreenWidget); 
  XFlush(dsply);

  /* Get pixel sizes from interface */
  inStr    = XmTextGetString(origTextW); 
  origSize =  (float)strtod(inStr, 0);
  XtFree(inStr);
  inStr    = XmTextGetString(outputTextW); 
  outputSize =  (float)strtod(inStr, 0);
  XtFree(inStr);

  if(doInterp)
    {/* Do in batch mode */
      origSize   = doInterpOrigSize;
      outputSize = doInterpOutputSize;
    }/* Do in batch mode */

  if(outputSize == 0.0 || origSize == 0.0 ||
     (outputSize < origSize))
    {/* Error */
      if(outputSize == 0.0 || origSize == 0.0)
	 askThem("Error - Pixel size of 0.0 not allowed.",
		 FALSE,XmDIALOG_INFORMATION);
	 else
	 askThem("Error - Output pixSize is greater than original pixSize.",
		 FALSE,XmDIALOG_INFORMATION);
    }/* Error */
  else
    {/* OK to go on */

      tmpBuf = (float *)malloc(sizeof(float)*file_height*file_width);
      
      if(tmpBuf)
	{/* OK to go on */
	  
	  minF = maxF = (float)inCore_buffer[0];
	  for(i=0;i<file_height;i++)
	    {/* rows */
	      for(j=0;j<file_width;j++)
		{/* cols */
		  tmpBuf[((i*file_width)+j)] = 
		    (float)inCore_buffer[((i*file_width)+j)] ;
		  minF = MIN(minF,tmpBuf[((i*file_width)+j)]);
		  maxF = MAX(maxF,tmpBuf[((i*file_width)+j)]);
		}/* cols */
	    }/* rows */
	  range  = maxF - minF;
	  offset = (int)minF;
	  
	  
	  rx= ry= ix =iy= 0.0;
	  rotation = 0.0;
	  scale = origSize / outputSize;
	  filler = (float)tmpBuf[0];
	  
	  /* OK Let's Get rid of some memory to help inter2d out */
	  free(inCore_buffer);

	  success= interp2d(tmpBuf, &file_width,  &file_height,
			    &rx,&ry,&ix,&iy,
			    &rotation,
			    &scale,
			    &filler
			    );
	  printf("Success=[%d]\n",success);
	  
	  /* Now stuff interpolated buffer back into BYTE */
	  minF = maxF = tmpBuf[0];
	  for(i=0;i<file_height;i++)
	    {/* rows */
	      for(j=0;j<file_width;j++)
		{/* cols */
		  minF = MIN(minF,tmpBuf[((i*file_width)+j)]);
		  maxF = MAX(maxF,tmpBuf[((i*file_width)+j)]);
		}/* cols */
	    }/* rows */
	  scaleF = range / (maxF - minF);
	  
	  if(!success)
	    {/* Can't do it */
	      askThem("Error - Can't get enough memory.",
		 FALSE,XmDIALOG_INFORMATION);
	      /* Pretend we didn't do anything */
	      scale = 1.0;
	    }/* Can't do it */

	  /* Now adjust concept of image size */
	  new_width  = file_width  * scale;
	  new_height = file_height * scale;
	  
	  /* OK Let's Get back the memory we freed to help inter2d out */
	  inCore_buffer = (float *)malloc(sizeof(float)*new_width*new_height);

	  for(i=0;i<new_height;i++)
	    {/* rows */
	      for(j=0;j<new_width;j++)
		{/* cols */
		  inCore_buffer[((i*new_width)+j)] = (float)(
							    offset + 
		        (float)((tmpBuf[((i*file_width)+j)]-minF) *scaleF)
							    );
		}/* cols */
	    }/* rows */
	  /* Now adjust concept of image size */
	  purdueHeader.numCols = file_width  = new_width;
	  purdueHeader.numRows = file_height = new_height;
	  
	  free(tmpBuf);
	  
	  if(success)
	    {/* OK to ask them about box coords */
	    }/* OK to ask them about box coords */
	  
	  /* Make it generate a new fullscreen buffer */
	  fullScreen_buffer.dispx = fullScreen_buffer.dispy = 0;
	  virtual_read_activate(0,0);
	  
	  DrawWindow(); 
	  
	  if(!doInterp)
	    askThem("Finished.  \n\
    Now you MUST write out this new file before you can box.",
		    FALSE,XmDIALOG_INFORMATION);

	}/* OK to go on */
      else
	{/* no memory */
	  askThem("Insufficient memory for interpolation",
		  FALSE,XmDIALOG_INFORMATION);
	}/* no memory */
      
    }/* OK to go on */


}/* INTERPOLATE_ACTIVATE */





/*******************************************************************
 * addAverage_activate() - Add an averaged image to dataset.
 */
void addAverage_activate(Widget widget, int *tag, 
			 XmAnyCallbackStruct *callback_data)
{/* ADDAVERAGE_ACTIVATE */
  int i,j,k,m,ncol,nrow;
  float *thisImg,*avgImg;
  float *accumImg,divisor;

  if(fileType==PACKEDPORTABLE)
    {/* Can only do for packed images now */

      m = purdueHeader.nimg;

      if(packImg[m]!=NULL)
	{free(packImg[m]);
	 packImg[m]= NULL;}
      CK(packImg[m]= (float *)malloc(sizeof(float) *
				   purdueHeader.numCols*
				   purdueHeader.numRows));
      
      CK(accumImg= (float *)malloc(sizeof(float) *
				   purdueHeader.numCols*
				   purdueHeader.numRows));
      
      /* Init */
      nrow = purdueHeader.numRows;
      ncol = purdueHeader.numCols;
      for(j=0;j<nrow;j++)
	{/* rows */
	  for(k=0;k<ncol;k++)
	    {/* cols */
	      accumImg[((j*ncol)+k)] = 0.0;
	    }/* cols */
	}/* rows */
      
      /* Now accum the lot */
      for(i=0;i<purdueHeader.nimg;i++)
	{/* Do all */
	  thisImg = packImg[i];
	  for(j=0;j<nrow;j++)
	    {/* rows */
	      for(k=0;k<ncol;k++)
		{/* cols */
		  accumImg[((j*ncol)+k)] += 
		    (float)((float)thisImg[((j*ncol)+k)]);
		}/* cols */
	    }/* rows */
	}/* Do all */

      /* Now average them */
      if(purdueHeader.nimg > 0)
	divisor = 1.0 / (float)purdueHeader.nimg;
      avgImg = packImg[m];
      for(j=0;j<nrow;j++)
	{/* rows */
	  for(k=0;k<ncol;k++)
	    {/* cols */
	      accumImg[((j*ncol)+k)] *= (float)divisor; 
	      avgImg[((j*ncol)+k)] = (float)accumImg[((j*ncol)+k)];
	    }/* cols */
	}/* rows */
      
      free(accumImg);

      /* We've got a new one now */
      purdueHeader.nimg++;

      stuffPackedImage(purdueHeader.numCols,
		       purdueHeader.numRows,
		       packImg,purdueHeader.nimg,
		       imgNum);  /* Start at radius'th image */
      
      DrawWindow(); 
  
      sprintf(outStr,"Averaged Image is now image number [%d]",
	      purdueHeader.nimg);
      askThem(outStr,FALSE,XmDIALOG_INFORMATION);

    }/* Can only do for packed images now */

}/* ADDAVERAGE_ACTIVATE */





/*******************************************************************
 * fourierFilt_activate() - Setup the fourFilter screen.
 */
void fourierFilt_activate(Widget widget, int *tag, 
			  XmAnyCallbackStruct *callback_data)
{/* FOURIERFILT_ACTIVATE */
  int idim1;
  XmString simpleString;

  sprintf(outStr,"0.0");
  XmTextSetString(resMinTextW,outStr);
  XmTextSetString(cutMinTextW,outStr);
  XmTextSetString(cutMaxTextW,outStr);
  
  if(purdueHeader.numCols > 512)
    idim1 = 1024;
  else if(purdueHeader.numCols > 256)
    idim1 = 512;
  else if(purdueHeader.numCols > 128)
    idim1 = 256;
  else if(purdueHeader.numCols > 64)
    idim1 = 128;
  else 
    idim1 = 64;
  
  /* And the value labels */
  sprintf(outStr,"FFT size = %d x %d",idim1,idim1);
  createSimpleString(&simpleString,outStr);
  XtSetArg(arglst[0], XmNlabelString, simpleString);
  XtSetValues(fourierFiltLabelW, arglst, 1); 
  XFlush(dsply);
  XmStringFree(simpleString); 
  XFlush(dsply);
  
  sprintf(outStr,"%f",1.41 * idim1);
  XmTextSetString(resMaxTextW,outStr);
  
  raiseIt(fourierFiltScreen);
  

}/* FOURIERFILT_ACTIVATE */





/*******************************************************************
 * fourierFilt_deactivate() - Get rid of fourFilt screen.
 */
void fourierFilt_deactivate(Widget widget, int *tag, 
			    XmAnyCallbackStruct *callback_data)
{/* FOURIERFILT_DEACTIVATE */

  XtUnmanageChild(fourierFiltScreen);
  
}/* FOURIERFILT_DEACTIVATE */





/*******************************************************************
 * fourierFilt_work() - Do the fourier filtering.
 */
void fourierFilt_work(Widget widget, int *tag, 
		      XmAnyCallbackStruct *callback_data)
{/* FOURIERFILT_WORK */
  int i,j,k,tot;
  float *img;
  float *thisBuf;
  float *fft;
  int ncol,nrow,idim1,idim2,implemented,imgBkgP;
  char *thisStr;
  float resmin,resmax,cutmin,cutmax;

  /* This may take awhile so give them the watch cursor */
  attrs.cursor= watch_cursor;
  XChangeWindowAttributes(dsply, XtWindow(drawingWidget), 
			  CWCursor, &attrs);
  XFlush(dsply);
  
  ncol = purdueHeader.numCols;
  nrow = purdueHeader.numRows;

  if(purdueHeader.numCols > 512)
    {idim2= idim1 = 1024;}
  else if(purdueHeader.numCols > 256)
    {idim2= idim1 = 512;}
  else if(purdueHeader.numCols > 128)
    {idim2= idim1 = 256;}
  else if(purdueHeader.numCols > 64)
    {idim2= idim1 = 128;}
  else 
    {idim2= idim1 = 64;}
  
  thisStr = XmTextGetString(resMinTextW);
  resmin  = (float)strtod(thisStr,0);
  XtFree(thisStr);
  thisStr = XmTextGetString(resMaxTextW);
  resmax  = (float)strtod(thisStr,0);
  XtFree(thisStr);
  thisStr = XmTextGetString(cutMinTextW);
  cutmin  = (float)strtod(thisStr,0);
  XtFree(thisStr);
  thisStr = XmTextGetString(cutMaxTextW);
  cutmax  = (float)strtod(thisStr,0);
  XtFree(thisStr);

  CK(img= (float *)malloc(sizeof(float)*ncol*nrow));
  CK(fft= (float *)malloc(sizeof(float)*idim1*idim2*2));
      
  for(i=0;i<purdueHeader.nimg;i++)
    {/* DO all images */

      thisBuf = packImg[i];

      imgBkgP = *(thisBuf);
      
      for(j=0;j<nrow;j++)
	for(k=0;k<ncol;k++)
	  {/* Restuff */
	    img[((j*ncol)+k)] = (float)thisBuf[((j*ncol)+k)];
	  }/* Restuff */

      img_filt_ (img, &ncol, &nrow, fft, &idim1, &idim2,
		 &imgBkgP, &cutmin, &resmin, &cutmax, &resmax);

      for(j=0;j<nrow;j++)
	for(k=0;k<ncol;k++)
	  {/* Restuff */
	    thisBuf[((j*ncol)+k)] = (float)img[((j*ncol)+k)];
	  }/* Restuff */

    }/* DO all images */
  
  free(img);
  free(fft);

  /* Always force mode==46 output after we do this!! */
  fileType = 46;

  stuffPackedImage(purdueHeader.numCols,
		   purdueHeader.numRows,
		   packImg,purdueHeader.nimg,
		   imgNum);  /* Start at radius'th image */
  
  DrawWindow(); 
  
  XtUnmanageChild(fourierFiltScreen);

  /* Change cursor back to what it was */
  attrs.cursor= old_cursor;
  XChangeWindowAttributes(dsply, XtWindow(drawingWidget), 
			  CWCursor, &attrs);
  XFlush(dsply);

  askThem("Fourier Filtering finished OK.",FALSE,XmDIALOG_INFORMATION);

}/* FOURIERFILT_WORK */




/***********************
 *    UNBLOB           *
 ***********************/

/* ************************************************************
 * unblob_create_label() - Remember the unblob screen for use later.  
 */
void unblob_create_label(Widget widget, int *tag, 
			 XmAnyCallbackStruct *callback_data)
{/* unblob_CREATE_LABEL */
  
  if(*tag==MINX)
    cutMinValueW[0]= widget;
  else if(*tag==MINY)
    cutMinValueW[1]= widget;
  else if(*tag==MINZ)
    cutMinValueW[2]= widget;
  else if(*tag==MAXX)
    cutMaxValueW[0]= widget;
  else if(*tag==MAXY)
    cutMaxValueW[1]= widget;
  else if(*tag==MAXZ)
    cutMaxValueW[2]= widget;
  else if(*tag==PERX)
    perValueW[0]= widget;
  else if(*tag==PERY)
    perValueW[1]= widget;
  else if(*tag==PERZ)
    perValueW[2]= widget;
  else if(*tag==NEGUNIV)
    negUnivW= widget;
  else if(*tag==USERADPROJ)
    {/* radporj */
      radProjToggleW= widget;
      XtUnmanageChild(radProjToggleW);
    }/* radporj */
  else if(*tag==USERADPROJCS)
    {/* radporj */
      radProjCSToggleW= widget;
      XtUnmanageChild(radProjCSToggleW);
    }/* radporj */
  else if(*tag==USERADIALCS)
    {/* radial */
      radialCSToggleW= widget;
    }/* radial */
  else if(*tag==ICOSRADPROJ)
    {
      icosRadProjToggleW = widget;
      XtUnmanageChild(icosRadProjToggleW);
    }

}/* unblob_CREATE_LABEL */




/* ************************************************************
 * unblob_create() - Remember the unblob screen for use later.  
 */
void unblob_create(Widget widget, int *tag, 
		   XmAnyCallbackStruct *callback_data)
{/* unblob_CREATE */
  
  if(*tag==UNBLOBMIN)
    unblobMinScrollW= widget;
  else if(*tag==UNBLOBMAX)
    unblobMaxScrollW= widget;
  else if(*tag==UNBLOBRADIUS)
    unblobRadiusScrollW= widget;
  else if(*tag==UNBLOBSIZE)
    unblobSizeScrollW= widget;
  else if(*tag==UBRADIUSV)
    unblobRadiusValueW= widget;
  else if(*tag==UBMINV)
    unblobMinValueW= widget;
  else if(*tag==UBMAXV)
    unblobMaxValueW= widget;
  else if(*tag==UBSIZEV)
    unblobSizeValueW= widget;
  else if(*tag==MINX)
    cutMinScrollW[0]= widget;
  else if(*tag==MINY)
    cutMinScrollW[1]= widget;
  else if(*tag==MINZ)
    cutMinScrollW[2]= widget;
  else if(*tag==MAXX)
    cutMaxScrollW[0]= widget;
  else if(*tag==MAXY)
    cutMaxScrollW[1]= widget;
  else if(*tag==MAXZ)
    cutMaxScrollW[2]= widget;
  else if(*tag==UBFILL)
    unblobFillValueW= widget;
  else if(*tag==UNBLOBRESET)
    {unblobResetW= widget;
    }

  if(*tag!=UBFILL && *tag!= UNBLOBRESET)
    handle_cr(widget);

}/* unblob_CREATE */




/* ************************************************************
 * unblob_screen_deactivate() - Get the unblob screen
 * off the display.
 */
void unblob_screen_deactivate(Widget widget, int *tag, 
			      XmAnyCallbackStruct *callback_data)
{/* unblob_screen_deactIVATE */
  
/* OBSOLETE */
  XtUnmanageChild(unblobscreen);
  unblobSwitch = FALSE;
  if(unblobBuffer != float_image_buffer1)
    free(unblobBuffer);

}/* unblob_screen_deactIVATE */




/* ************************************************************
 * unblob_screen_activate() - Put the unblobscreen on the display.
 */
void unblob_screen_activate(Widget widget, int *tag, 
			    XmAnyCallbackStruct *callback_data)
{/* unblob_screen_actIVATE */
  int i,j;
  XmString simpleString;

  minUnblob= (int)secMinPixel;
  maxUnblob= (int)0.0 - surfaceDensity;
  printf("secMinPixel[%f]max[%f]\n",secMinPixel,secMaxPixel);
  /* Divide by 100 for percentage */
  unblobSize= (int)((numCols*numRows*origNumSecs) * 
		    (float)((float)0.001/100.0));

  /* Now setup the scrolls accordingly */
  XtSetArg(arglst[0], XmNmaximum, (int)((numCols/2)+5));
  XtSetArg(arglst[1], XmNminimum, (int)(0));
  XtSetArg(arglst[2], XmNsliderSize, (int)(5));
  XtSetArg(arglst[3], XmNvalue, (int)(0));
  XtSetArg(arglst[4], XmNpageIncrement, (int)(10));
  XtSetValues(unblobRadiusScrollW, arglst, 5);
  XFlush(dsply);
  XtSetArg(arglst[0], XmNvalue, (int)(0));
  XtSetValues(unblobSizeScrollW, arglst, 1);
  XFlush(dsply);
  /* And the value labels */
  sprintf(outStr,"%d",(0));
  createSimpleString(&simpleString,outStr);
  XtSetArg(arglst[0], XmNlabelString, simpleString);
  XtSetValues(unblobRadiusValueW, arglst, 1); 
  XFlush(dsply);
  XmStringFree(simpleString); 
  XFlush(dsply);
  sprintf(outStr,"%.3f",(0.001));
  XmTextSetString(unblobSizeValueW,outStr); 
  XFlush(dsply);
  
  /* Now setup cut scrolls and values widgets */
  reset_xyz_scrolls();

  CK(unblobBuffer=(float *)malloc(sizeof(float)*numCols*numRows*origNumSecs));
  
  unblobSwitch = TRUE;
  raiseIt(unblobscreen);  

}/* unblob_screen_actIVATE */




/* ************************************************************
 * unblob_text_change() - Whenever unblob text widgets changed
 * call here.
 */
void unblob_text_change(Widget widget, int *tag, 
			XmAnyCallbackStruct *callback_data)
{/* UNBLOB_TEXT_CHANGE */
  char *unblobText;
  float thisFloat;

  if(inited)
    {/* Only do if it exists */
      unblobText = XmTextGetString(unblobSizeValueW); 
      thisFloat  = (float)strtod(unblobText,0);
      unblobSize = (int)((numCols*numRows*origNumSecs) * 
			  (float)((float)thisFloat/100.0));
      XtFree(unblobText);
    }/* Only do if it exists */

}/* UNBLOB_TEXT_CHANGE */





/* ************************************************************
 * unblob_scroll_change() - Whenever unblob scroll widgets changed
 * call here.
 */
void unblob_scroll_change(Widget widget, int *tag, 
			  XmScrollBarCallbackStruct *callback_data)
{/* UNBLOB_SCROLL_CHANGE */
  XmString simpleString;
  int i,thisVal;
  float thisFloat;

  if(*tag==UNBLOBMIN)
    {/* */ 
      thisVal = (int)callback_data->value;
      sprintf(outStr,"%d",thisVal);
      minUnblob= thisVal;
      createSimpleString(&simpleString,outStr);
      XtSetArg(arglst[0], XmNlabelString, simpleString);
      XtSetValues(unblobMinValueW, arglst, 1); 
      XFlush(dsply);
      XmStringFree(simpleString); 
      XFlush(dsply);
    }/* */
  else if (*tag==UNBLOBMAX)
    {/*  */
      thisVal = (int)callback_data->value;
      sprintf(outStr,"%d",thisVal); 
      maxUnblob= thisVal;
      createSimpleString(&simpleString,outStr);
      XtSetArg(arglst[0], XmNlabelString, simpleString);
      XtSetValues(unblobMaxValueW, arglst, 1); 
      XFlush(dsply);
      XmStringFree(simpleString); 
      XFlush(dsply);
    }/*  */
  else if (*tag==UNBLOBRADIUS)
    {/*  */
      thisVal = (int)callback_data->value;
      sprintf(outStr,"%d",thisVal); 
      radiusUnblob= thisVal;
      createSimpleString(&simpleString,outStr);
      XtSetArg(arglst[0], XmNlabelString, simpleString);
      XtSetValues(unblobRadiusValueW, arglst, 1); 
      XFlush(dsply);
      XmStringFree(simpleString); 
      XFlush(dsply);
    }/*  */
  else if (*tag==UNBLOBSIZE)
    {/*  */
      thisFloat = (float)callback_data->value;
      thisFloat *= 0.001;
      sprintf(outStr,"%.3f%%",thisFloat); 
      /* Divide by 100 for percentage */
      unblobSize= (int)((numCols*numRows*origNumSecs) * 
			  (float)((float)thisFloat/100.0));
      XmTextSetString(unblobSizeValueW,outStr); 
      XFlush(dsply);
    }/*  */
  else if (*tag==MINX || *tag==MINY || *tag==MINZ)
    {/*  */
      thisVal = (int)callback_data->value;
      sprintf(outStr,"%d",thisVal+1); 
      if(*tag==MINY)
	cutMaxValues[(*tag-MINX)]= numRows-thisVal;
      else
	cutMinValues[(*tag-MINX)]= thisVal;
      createSimpleString(&simpleString,outStr);
      XtSetArg(arglst[0], XmNlabelString, simpleString);
      XtSetValues(cutMinValueW[(*tag-MINX)], arglst, 1); 
      XFlush(dsply);
      XmStringFree(simpleString); 
      XFlush(dsply);
      /* Now do percent label */
      if(*tag==MINX)
	perMinValues[(*tag-MINX)]= (float)thisVal/(float)numCols;
      else if(*tag==MINY)
	perMinValues[(*tag-MINX)]= (float)thisVal/(float)numRows;
      else if(*tag==MINZ)
	perMinValues[(*tag-MINX)]= (float)thisVal/(float)origNumSecs;
      sprintf(outStr,"%.1f%% to %.1f%%",perMinValues[(*tag-MINX)]*100.0,
	      perMaxValues[(*tag-MINX)]*100.0); 
      createSimpleString(&simpleString,outStr);
      XtSetArg(arglst[0], XmNlabelString, simpleString);
      XtSetValues(perValueW[(*tag-MINX)], arglst, 1); 
      XFlush(dsply);
      XmStringFree(simpleString); 
      XFlush(dsply);
    }/*  */
  else if (*tag==MAXX || *tag==MAXY || *tag==MAXZ)
    {/*  */
      thisVal = (int)callback_data->value;
      if(thisVal > 2)
	sprintf(outStr,"%d",thisVal-1); 
      else sprintf(outStr,"%d",1); 
      if(*tag==MAXY)
	cutMinValues[(*tag-MAXX)]= numRows-thisVal;
      else
	cutMaxValues[(*tag-MAXX)]= thisVal;
      createSimpleString(&simpleString,outStr);
      XtSetArg(arglst[0], XmNlabelString, simpleString);
      XtSetValues(cutMaxValueW[(*tag-MAXX)], arglst, 1); 
      XFlush(dsply);
      XmStringFree(simpleString); 
      XFlush(dsply);
      /* Now do percent label */
      if(*tag==MAXX)
	perMaxValues[(*tag-MAXX)]= (float)thisVal/(float)numCols;
      else if(*tag==MAXY)
	perMaxValues[(*tag-MAXX)]= (float)thisVal/(float)numRows;
      else if(*tag==MAXZ)
	perMaxValues[(*tag-MAXX)]= (float)thisVal/(float)origNumSecs;
      sprintf(outStr,"%.1f%% to %.1f%%",perMinValues[(*tag-MAXX)]*100.0,
	      perMaxValues[(*tag-MAXX)]*100.0); 
      createSimpleString(&simpleString,outStr);
      XtSetArg(arglst[0], XmNlabelString, simpleString);
      XtSetValues(perValueW[(*tag-MAXX)], arglst, 1); 
      XFlush(dsply);
      XmStringFree(simpleString); 
      XFlush(dsply);
    }/*  */

  if((callback_data->reason != XmCR_DRAG))
    {/* Do something */;}

  if(immediateMapShade)
    mapShade_activate(0,&i,0);   /* Don't do this if we're just dragging */

}/* UNBLOB_SCROLL_CHANGE */




/*************************************************************
 * unblob_tryit_activate() - UNBLOB the map.
 */
void unblob_tryit_activate(Widget widget, int *tag, 
			   XmAnyCallbackStruct *callback_data)
{/* UNBLOB_TRYIT_ACTIVATE */
  int bCnt,uniqueBlobs,status,killFlag;
  float bgMin,bgMax,minIntensity,maxIntensity;
  float killValue,contMin,contMax;
  int i,j,k,thisSecRow;
  int thisPixel;
  int maxVolume;
  blobStructure_t blobList[BMAX];
  int cutSecs,cutRows,cutCols,clipIt;
  float *unblobitBuf;
  int   thisUnblobSize;
  char *unblobText;
  float thisFloat;

  /* Give feedback */
  sprintf(outStr,"[%d] Calculating Unblob.",widget);
  XmTextSetString(timeLeftLabelWidget, outStr);
  XFlush(dsply);

  if(lowRes)
    {/* Don't do this for High Resolution images */
      /* Clear our temp buffer */
      if(unblobBuffer != float_image_buffer1)
	{/* Don't zero out our gold standard!!! */
	  if(unblobBuffer != float_image_buffer1)
	    {/* Don't zero out our gold standard!!! */
	      for(i=0;i<origNumSecs;i++)
		for(j=0;j<numRows;j++)
		  {/* rowsec */
		    thisSecRow= (i*numRows*numCols)+(j*numCols);
		    for(k=0;k<numCols;k++)
		      unblobBuffer[(thisSecRow+k)]= 0.0;
		  }/* rowsec */
	    }/* Don't zero out our gold standard!!! */
	  
	  if(negUniverseCut)
	    {/* Cut out the negative space */
	      /* First stuff the buffer with everything */
	      for(i=0;i<origNumSecs;i++)
		{/* Secs */
		  for(j=0;j<numRows;j++)
		    {/* rows */
		      for(k=0;k<numCols;k++)
			{/* COLS  */
			  if(reverseSections)
			    {/* Reverse the sections */
			      unblobBuffer[(
					    ((origNumSecs-i)*numRows*numCols)+
					    (j*numCols)+
					    k
					    )] =
			      (float)float_image_buffer1[(
							  (i*numRows*numCols)+
							  (j*numCols)+
							  k
							  )] ;
			    }/* Reverse the sections */
			  else
			    {/* Normal */
			      unblobBuffer[(
					    (i*numRows*numCols)+
					    (j*numCols)+
					    k
					    )] =
			      (float)float_image_buffer1[(
							  (i*numRows*numCols)+
							  (j*numCols)+
							  k
							  )] ;
			    }/* Normal */
			}/* COLS */
		    }/* rows */
		}/* Secs */
	      /* Now cut out the negative space */
	      /* Now cut it if appropriate else just stuff it */
	      for(i=cutMinValues[2];i<cutMaxValues[2];i++)
		{/* Secs */
		  for(j=cutMinValues[1];j<cutMaxValues[1];j++)
		    {/* rows */
		      for(k=cutMinValues[0];k<cutMaxValues[0];k++)
			{/* COLS  */
			  if(reverseSections)
			    {/* Reverse the sections */
			      unblobBuffer[(
					    ((origNumSecs-i)*numRows*numCols)+
					    (j*numCols)+
					    k
					    )] = (float)0.0;
			    }/* Reverse the sections */
			  else
			    {/* Normal */
			      unblobBuffer[(
					    (i*numRows*numCols)+
					    (j*numCols)+
					    k
					    )] = (float)0.0;
			    }/* Normal */
			}/* COLS */
		    }/* rows */
		}/* Secs */
	      
	    }/* Cut out the negative space */
	  else
	    {/* Normal */
	      /* Now cut it if appropriate else just stuff it */
	      for(i=cutMinValues[2];i<cutMaxValues[2];i++)
		{/* Secs */
		  for(j=cutMinValues[1];j<cutMaxValues[1];j++)
		    {/* rows */
		      for(k=cutMinValues[0];k<cutMaxValues[0];k++)
			{/* COLS  */
			  if(reverseSections)
			    {/* Reverse the sections */
			      unblobBuffer[(
					    ((origNumSecs-i)*numRows*numCols)+
					    (j*numCols)+
					    k
					    )] =
			      (float)float_image_buffer1[(
							  (i*numRows*numCols)+
							  (j*numCols)+
							  k
							  )] ;
			    }/* Reverse the sections */
			  else
			    {/* Normal */
			      unblobBuffer[(
					    (i*numRows*numCols)+
					    (j*numCols)+
					    k
					    )] =
		    (float)float_image_buffer1[(
						(i*numRows*numCols)+
						(j*numCols)+
						k
						)] ;
			    }/* Normal */
			}/* COLS */
		    }/* rows */
		}/* Secs */
	    }/* Normal */
	}/* Don't zero out our gold standard!!! */
      else
	{/* unblobBuffer is the gold standard float_image_buffer1 */
	  if(negUniverseCut)
	    {/* Cut out the negative space */
	      /* First stuff the buffer with everything */
	      /* Already done - it *IS* the gold std!! */
	      
	      /* Now cut out the negative space */
	      /* Now cut it if appropriate else just stuff it */
	      for(i=cutMinValues[2];i<cutMaxValues[2];i++)
		{/* Secs */
		  for(j=cutMinValues[1];j<cutMaxValues[1];j++)
		    {/* rows */
		      for(k=cutMinValues[0];k<cutMaxValues[0];k++)
			{/* COLS  */
			  if(reverseSections)
			    {/* Reverse the sections */
			      unblobBuffer[(
					    ((origNumSecs-i)*numRows*numCols)+
					    (j*numCols)+
					    k
					    )] = (float)0.0;
			    }/* Reverse the sections */
			  else
			    {/* Normal */
			      unblobBuffer[(
					    (i*numRows*numCols)+
					    (j*numCols)+
					    k
					    )] = (float)0.0;
			    }/* Normal */
			}/* COLS */
		    }/* rows */
		}/* Secs */
	      
	    }/* Cut out the negative space */
	  else
	    {/* Normal */
	      /* Now zero out the cut stuff if appropriate */
	      for(i=0;i<origNumSecs;i++)
		{/* Secs */
		  for(j=0;j<numRows;j++)
		    {/* rows */
		      for(k=0;k<numCols;k++)
			{/* COLS  */
			  if((i < cutMinValues[2] || i>= cutMaxValues[2]) ||
			     (j < cutMinValues[1] || j>= cutMaxValues[1]) ||
			     (k < cutMinValues[0] || k>= cutMaxValues[0]))
			    {/* OK to CUT */
			      if(reverseSections)
				{/* Reverse the sections */
				  unblobBuffer[(
						((origNumSecs-i)*
						 numRows*numCols)+
						(j*numCols)+
						k
						)] = 0.0;
				}/* Reverse the sections */
			      else
				{/* Normal */
				  unblobBuffer[(
						(i*numRows*numCols)+
						(j*numCols)+
						k
						)] = 0.0;
				}/* OK to CUT */
			    }/* Normal */
			}/* COLS */
		    }/* rows */
		}/* Secs */
	    }/* Normal */
	}/* unblobBuffer is the gold standard float_image_buffer1 */
    }/* Don't do this for High Resolution images */
  else
    {/* We've got to stuff the buffer for high res images */
      if(unblobBuffer != float_image_buffer1)
	{/* STuff it */
	  for(i=0;i<origNumSecs;i++)
	    for(j=0;j<numRows;j++)
	      {/* rowsec */
		thisSecRow= (i*numRows*numCols)+(j*numCols);
		for(k=0;k<numCols;k++)
		  unblobBuffer[(thisSecRow+k)]= 
		    float_image_buffer1[(thisSecRow+k)];
	      }/* rowsec */
	}/* STuff it */
    }/* We've got to stuff the buffer for high res images */


  if(cleanUnblob)
    {/* Just cut out anything of lesser density */

      for(i=0;i<origNumSecs;i++)
	{/* Secs */
	  for(j=0;j<numRows;j++)
	    {/* rows */
	      for(k=0;k<numCols;k++)
		{/* COLS  */
		  thisPixel = (
				((i)*numRows*numCols)+
				(j*numCols)+
				k
				);
		  /* Use greater than here because of sign difference in
		   * TSB density values. */
		  if((unblobBuffer[thisPixel] > maxUnblob))
		    unblobBuffer[thisPixel] = (float)0.0;
		}/* COLS  */
	    }/* rows */
	}/* Secs */

    }/* Just cut out anything of lesser density */


  if(radiusUnblob > 0 || unblobSize > 0)
    {/* They want to unblob */

      unblobText = XmTextGetString(unblobFillValueW); 
      killValue  = (float)strtod(unblobText,0);
      XtFree(unblobText);
      printf("UB fill [%f]\n",killValue);

      contMin = (float)minUnblob*(irev * -1);
      contMax = (float)maxUnblob*(irev * -1);
      if(contMin > contMax)
	{/* switch them */
	  float tmpF;
	  tmpF = contMin;
	  contMin = contMax;
	  contMax = tmpF;
	}/* switch them */
      
      unblobText = XmTextGetString(unblobSizeValueW); 
      thisFloat  = (float)strtod(unblobText,0);
      unblobSize = (int)((float)(numCols*numRows*origNumSecs) * 
			  (float)((float)thisFloat/100.0));
      XtFree(unblobText);

      clipIt = (
		cutMinValues[0] != 0.0 ||
		cutMinValues[1] != 0.0 ||
		cutMinValues[2] != 0.0 ||
		cutMaxValues[0] != numCols ||
		cutMaxValues[1] != numRows ||
		cutMaxValues[2] != origNumSecs 
		);
      if(negUniverseCut)
	clipIt= FALSE;
      if(!lowRes)
	clipIt = FALSE;

      if(clipIt)
	{/* Pass smaller cube to unblob rtns */
	  cutSecs = (cutMaxValues[2] - cutMinValues[2]);
	  cutRows = (cutMaxValues[1] - cutMinValues[1]);
	  cutCols = (cutMaxValues[0] - cutMinValues[0]);
	  unblobitBuf = NULL;
	  CK(unblobitBuf = (float *)malloc(sizeof(float)*
					   cutSecs*cutRows*cutCols));
	  if(!unblobitBuf)
	    {/* Error */
	      askThem("[UNBLOB_TRYIT_ACTIVATE]  Error allocating unblobitBuf.",
		      TRUE,XmDIALOG_INFORMATION);
	      goto destructor;
	    }/* Error */
	  
	  for(i=0;i<cutSecs;i++)
	    {/* Secs */
	      int thatPixel;
	      for(j=0;j<cutRows;j++)
		{/* rows */
		  for(k=0;k<cutCols;k++)
		    {/* COLS  */
		      thisPixel = (
				   ((i)*cutRows*cutCols)+
				   (j*cutCols)+
				   k
				   );
		      thatPixel = (
				   ((i+cutMinValues[2])*numRows*numCols)+
				   ((j+cutMinValues[1])*numCols)+
				   (k+cutMinValues[0])
				   );
		      unblobitBuf[thisPixel]= unblobBuffer[thatPixel];
		    }/* COLS  */
		}/* rows */
	    }/* Secs */
	}/* Pass smaller cube to unblob rtns */
      else
	{/* STuff with what we've got */
	  cutSecs = origNumSecs;
	  cutRows = numRows;
	  cutCols = numCols;
	  unblobitBuf = unblobBuffer;
	}/* STuff with what we've got */

      bCnt = BMAX;
      for(i=0;i<BMAX;i++)
	{/* Init it out */
	  blobList[i].id  = 0;
	  blobList[i].size= 0;
	  blobList[i].minInt= 0.0;
	  blobList[i].maxInt= 0.0;
	}/* Init it out */
      
      killFlag = FALSE;

      thisUnblobSize = unblobSize;

      count_kill_blobs_rob_(&bCnt, &uniqueBlobs, 
			    &contMin, &contMax,
			    &killFlag,
			    &thisUnblobSize,
			    &killValue,
			    &bgMin, &bgMax,
			    &status, unblobitBuf,
			    &cutCols, &cutRows, &cutSecs, 
			    blobList,
			    &radiusUnblob, &cleanUnblob);

      /* This is done for us from the interface - see thisUnblobSize
       * variable 
       maxVolume = (numCols*numRows*origNumSecs);
       GetBlobSizeLimit_rob(&thisUnblobSize, &killValue, blobList, &bCnt, 
       &uniqueBlobs, &maxVolume);
       */
      
 
      /*
	 c Set up header for new file
	 write(*, '(a,2f8.1)') ' >> Old min/max intensities : '
	 &					, minIntensity, maxIntensity
	 */
      
      getblobminmaxint_rob_(blobList, &bCnt, &thisUnblobSize, &killValue, 
			    &minIntensity, &maxIntensity);

      minIntensity = MIN(bgMin,minIntensity);
      maxIntensity = MAX(bgMax,maxIntensity);

      /*	write(*, '(a,2f8.1)') ' >> New min/max intensities : '
       * &			       , minIntensity, maxIntensity
       *
       * Count_Kill_Blobs_rob(INFILEUNIT, OUTFILEUNIT, bCnt, uniqueBlobs, 
       * blobList, contour1, contour2, .true., blobSizeLim, 
       * killValue, bgMin, bgMax, status);
       *
       */

      killFlag = TRUE;
      count_kill_blobs_rob_(&bCnt, &uniqueBlobs, 
			    &contMin, &contMax,
			    &killFlag,
			    &thisUnblobSize,
			    &killValue,
			    &bgMin, &bgMax,
			    &status, unblobitBuf,
			    &cutCols, &cutRows, &cutSecs, 
			    blobList,
			    &radiusUnblob, &cleanUnblob);


      if(clipIt)
	{/* Now stuff it back again */
	  for(i=0;i<cutSecs;i++)
	    {/* Secs */
	      int thatPixel;
	      for(j=0;j<cutRows;j++)
		{/* rows */
		  for(k=0;k<cutCols;k++)
		    {/* COLS  */
		      thisPixel = (
				   ((i)*cutRows*cutCols)+
				   (j*cutCols)+
				   k
				   );
		      thatPixel = (
				   ((i+cutMinValues[2])*numRows*numCols)+
				   ((j+cutMinValues[1])*numCols)+
				   (k+cutMinValues[0])
				   );
		      unblobBuffer[thatPixel]= unblobitBuf[thisPixel];
		    }/* COLS  */
		}/* rows */
	    }/* Secs */
	  
	  /* Now give back memory */
	  free(unblobitBuf);
	  unblobitBuf= NULL;
	}/* Now stuff it back again */
      
    }/* They want to unblob */
  
 destructor:;

}/* UNBLOB_TRYIT_ACTIVATE */




/*************************************************************
 * unblob_reset_activate() - Go back to the last gold STD
 * map to undo any unblobbing.
 */
void unblob_reset_activate(Widget widget, int *tag, 
			   XmAnyCallbackStruct *callback_data)
{/* UNBLOB_RESET_ACTIVATE */
  int i;

  XtSetArg(arglst[0], XmNvalue, (int)(0));
  XtSetValues(unblobSizeScrollW, arglst, 1);
  XFlush(dsply);
  sprintf(outStr,"0.000 %%");
  XmTextSetString(unblobSizeValueW,outStr); 
  unblobSize = 0;
  XFlush(dsply);

  if(immediateMapShade)
    mapShade_activate(0,&i,0);

}/* UNBLOB_RESET_ACTIVATE */




/*************************************************************
 * unblob_save_activate() - Save this unblobbed MAP as the
 * new gold STD MAP.
 */
void unblob_save_activate(Widget widget, int *tag, 
			  XmAnyCallbackStruct *callback_data)
{/* UNBLOB_SAVE_ACTIVATE */
  int i,j,k;
  char outStr[64];

  if(askThem("This operation destructively overwrites\n\
the Map in core with the UNBLOB Map.  You must\n\
reread the disk Map file to undo this.",FALSE,XmDIALOG_QUESTION))
    {/* OK to go on */

      /* Calculate the unblob buffer */
      CK(unblobBuffer=(float *)malloc(sizeof(float)*
				      numCols*numRows*origNumSecs));

      unblobSwitch = TRUE;
      if(!unblobBuffer)
	{/* Can't allocate it */
	  if(askThem("Insufficient memory - do you want to unblob the\n\
Gold Standard memory buffer?  (This is a \n\
destructive operation - you'll have to re-read\n\
from disk to undo this.)",FALSE,XmDIALOG_QUESTION))
	    unblobBuffer = float_image_buffer1;
	  else
	    unblobSwitch = FALSE;
	}/* Can't allocate it */
	  
      if(unblobSwitch)
	{/* unblob it */
	  unblob_tryit_activate(0,0,0);
	  
	  if(unblobBuffer != float_image_buffer1)
	    {/* Save our temp buffer  */
	      for(i=0;i<origNumSecs;i++)
		{/* secs */
		  for(j=0;j<numRows;j++)
		    {/* rows */
		      for(k=0;k<numCols;k++)
			{/* cols */
			  float_image_buffer1[(
					       (i*numRows*numCols)+
					       (j*numCols)+
					       k
					       )] =
						 (float)unblobBuffer[(
							 (i*numRows*numCols)+
							      (j*numCols)+
								      k
								      )] ;
			}/* cols */
		    }/* rows */
		}/* secs */
	      free(unblobBuffer);
	    }/* Save our temp buffer  */
	  
	}/* unblob it */
      unblobBuffer = NULL;
      unblobSwitch = FALSE;

      sprintf(outStr,"Select parameters.");
      XmTextSetString(timeLeftLabelWidget, outStr);
      XFlush(dsply);

    }/* OK to go on */

}/* UNBLOB_SAVE_ACTIVATE */





/****************************************************************
 * reverse_sections() - Set switches to do 3D Rendering properly.
 */
void reverse_sections(Widget widget, int *tag, 
		      XmAnyCallbackStruct *callback_data)
{/* REVERSE_SECTIONS */
  XmString simpleString;

  if(*tag==NEGUNIV)
    {/* Neg univ */
      negUniverseCut = !negUniverseCut;
      if(negUniverseCut)
	sprintf(outStr,"NOT Cut:");
      else sprintf(outStr,"Cut:");
      createSimpleString(&simpleString,outStr);
      XtSetArg(arglst[0], XmNlabelString, simpleString );
      XtSetArg(arglst[1], XmNrecomputeSize, FALSE );
      XtSetValues(negUnivW, arglst, 2); 
      XFlush(dsply);
      XmStringFree(simpleString); 
    }/* Neg univ */
  if(*tag==CLEANUNBLOB)
    cleanUnblob = !cleanUnblob;
  else if(*tag==DISPICOSA)
    {/* display the Icosahedron?? */
      dispIcosa = !dispIcosa;
      if(!dispIcosa)
	{/* Init the display */
	  int sx,sy,ex,ey,xoffset,yoffset;
	  xoffset = 500;
	  yoffset = 290;
	  sx = (int)xoffset - 35;
	  ex = (int)xoffset + 35;
	  for(i=0;i<70;i++)
	    {/* Init */
	      sy = (int)(yoffset - 35)+i;
	      ey = (int)sy;
	      XDrawLine(dsply,XtWindow(mapShadeScreenWindow),
			image_gcGREY[100],
			sx,sy,ex,ey);
	    }/* Init */
	}/* Init the display */
    }/* display the Icosahedron?? */
  else if(*tag==USERADPROJ)
    revRadProj = !revRadProj;
  else if(*tag==ICOSRADPROJ)
    {
      icosRadProj = !icosRadProj;
      raiseIt(dioScreen);
    }
  else if(*tag==USERADIALCS)
    radialCS = !radialCS;
  else if(*tag==USERADPROJCS)
    radProjCS = !radProjCS;
  else if(*tag==ALL)
    {/* Reverse sections */
      reverseSections = !reverseSections; 
    }/* Reverse sections */
  
}/* REVERSE_SECTIONS */





/**************************************************************
 * reverse_sections() - Physically reverse sections in the MAP.
 */
void destructive_reverse_sections(Widget widget, int *tag, 
				  XmAnyCallbackStruct *callback_data)
{/* DESTRUCTIVE_REVERSE_SECTIONS */
  int i,j,k,secBegin,secEnd,rowBegin,rowEnd,colBegin,colEnd;
  int saveDispSections;
  float tempFloat;

  if(*tag==X)
    {/* Reverse Columns */
      /* We are doing this destructively now. */
      for(k=0;k<(origNumSecs);k++)
	{/* Secs */
	  secBegin = k*numRows*numCols;
	  for(j=0;j<numRows;j++)
	    {/* Rows */
	      rowBegin = (j*numCols);
	      for(i=0;i<numCols/2;i++)
		{/* Reverse all */
		  colBegin = secBegin + rowBegin + i;
		  colEnd   = secBegin + rowBegin + (numCols - 1 - i);
		  tempFloat = float_image_buffer1[colEnd];
		  float_image_buffer1[colEnd] = 
		    float_image_buffer1[colBegin];
		  float_image_buffer1[colBegin] = tempFloat;
		}/* Reverse all */
	    }/* Rows */
	}/* Secs */
    }/* Reverse Columns */
  else if(*tag==Y)
    {/* Reverse rows */
      /* We are doing this destructively now. */
      for(k=0;k<(origNumSecs);k++)
	{/* Secs */
	  secBegin = k*numRows*numCols;
	  for(j=0;j<(numRows/2);j++)
	    {/* Rows */
	      rowBegin = (j*numCols);
	      rowEnd   = ((numRows - 1 - j)*numCols);
	      for(i=0;i<numCols;i++)
		{/* Reverse all */
		  colBegin = secBegin + rowBegin + i;
		  colEnd   = secBegin + rowEnd + i;
		  tempFloat = float_image_buffer1[colEnd];
		  float_image_buffer1[colEnd] = 
		    float_image_buffer1[colBegin];
		  float_image_buffer1[colBegin] = tempFloat;
		}/* Reverse all */
	    }/* Rows */
	}/* Secs */
    }/* Reverse rows */
  else if(*tag==Z)
    {/* Reverse sections */
      /* We are doing this destructively now. */
      for(k=0;k<(origNumSecs/2);k++)
	{/* Secs */
	  secBegin = k*numRows*numCols;
	  secEnd   = (origNumSecs - 1 - k) *numRows*numCols;
	  for(j=0;j<numRows;j++)
	    {/* Rows */
	      rowBegin = (j*numCols);
	      for(i=0;i<numCols;i++)
		{/* Reverse all */
		  colBegin = secBegin + rowBegin + i;
		  colEnd   = secEnd + rowBegin + i;
		  tempFloat = float_image_buffer1[colEnd];
		  float_image_buffer1[colEnd] = 
		    float_image_buffer1[colBegin];
		  float_image_buffer1[colBegin] = tempFloat;
		}/* Reverse all */
	    }/* Rows */
	}/* Secs */
    }/* Reverse sections */
  
  /* Their shaded views are no longer valid - show them 
   * their new sections */
  shadedMaps= FALSE;
  dispSections = TRUE;
  XtSetArg(arglst[0], XmNlabelPixmap, sectionPixmap);
  XtSetArg(arglst[1], XmNselectPixmap, unsectionPixmap);
  XtSetArg(arglst[2], XmNrecomputeSize, FALSE);
  XtSetValues(sectionWidget, arglst, 3); 
  if(dispSections)
    {/* Update the sections - if shaded views then they have to recalc */
      /* Now stuff the display buffer from our new FIXed map */
      stuffMapImage((float *)map_image_buffer, 
		    (float *)float_image_buffer1, 
		    file_width,file_height,
		    nsecs,imgNum, 0);
      DrawWindow();
    }/* Update the sections - if shaded views then they have to recalc */
  
  XtUnmanageChild(invertscreen);
  
}/* DESTRUCTIVE_REVERSE_SECTIONS */





/* ************************************************************
 * fixit_create() - Remember the fixit screen for use later.  
 */
void fixit_create(Widget widget, int *tag, 
		  XmAnyCallbackStruct *callback_data)
{/* fixit_CREATE */
  
  if(*tag==MINFIX)
    minFixTextW= widget;
  else if(*tag==MAXFIX)
    maxFixTextW= widget;
  else if(*tag==PAGEFIX)
    pageFixTextW= widget;
  else if(*tag==INCFIX)
    incFixTextW= widget;

}/* fixit_CREATE */



/* ************************************************************
 * fixit_changed() - Whenever crysDisp text widgets changed
 * call here.
 */
void fixit_changed(Widget widget, int *tag, 
		   XmAnyCallbackStruct *callback_data)
{/* fixit_CHANGED */
  char *rotflt,outStr[64];
  XmString simpleString;
  int i;
  char *fixitText;

  fixitText = XmTextGetString(widget); 
  
  if(*tag==MINFIX)
    {/* */ 
      minFixValue = (float)strtod(fixitText,0);
    }/* */
  else if (*tag==MAXFIX)
    {/*  */
      maxFixValue = (int)strtod(fixitText,0);
    }/*  */
  else if (*tag==PAGEFIX)
    {/*  */
      pageFixValue = (int)strtod(fixitText,0);
    }/*  */
  else if (*tag==INCFIX)
    {/*  */
      incFixValue = (int)strtod(fixitText,0);
    }/*  */
  
  XtFree(fixitText);

}/* fixit_CHANGED */





/*************************************************************
 * fixit_activate() - Setup the fixit screen on display.
 */
void fixit_activate(Widget widget, int *tag, 
		    XmAnyCallbackStruct *callback_data)
{/* FIXIT_ACTIVATE */

  if(*tag==FFTMIN)
    fixitWidget= minFFTscroll;
  else if(*tag==FFTMAX)
    fixitWidget= maxFFTscroll;

  XtSetArg(arglst[0], XmNminimum, &minFixValue);
  XtSetArg(arglst[1], XmNmaximum, &maxFixValue);
  XtSetArg(arglst[2], XmNincrement, &incFixValue);
  XtSetArg(arglst[3], XmNpageIncrement, &pageFixValue);
  XtGetValues(fixitWidget, arglst, 4);
  XFlush(dsply);

  sprintf(outStr,"%d",minFixValue);
  XmTextSetString(minFixTextW,outStr);
  sprintf(outStr,"%d",maxFixValue);
  XmTextSetString(maxFixTextW,outStr);
  sprintf(outStr,"%d",pageFixValue);
  XmTextSetString(pageFixTextW,outStr);
  sprintf(outStr,"%d",incFixValue);
  XmTextSetString(incFixTextW,outStr);
  XFlush(dsply);

  raiseIt(fixitscreen);

}/* FIXIT_ACTIVATE */





/*************************************************************
 * fixit_deactivate() - Get rid of fixit screen.
 */
void fixit_deactivate(Widget widget, int *tag, 
		      XmAnyCallbackStruct *callback_data)
{/* FIXIT_DEACTIVATE */
  int sliderSize;

  /* Now instantiate the changes they've made */
  if(*tag!=CANCEL)
    {/* MAke changes current */
      sliderSize = (int)pageFixValue;
      XtSetArg(arglst[0], XmNmaximum, (int)((maxFixValue)+sliderSize));
      XtSetArg(arglst[1], XmNminimum, (int)(minFixValue));
      XtSetArg(arglst[2], XmNsliderSize, (int)(sliderSize));
      XtSetArg(arglst[3], XmNincrement, (int)(incFixValue));
      XtSetArg(arglst[4], XmNpageIncrement, (int)(pageFixValue));
      XtSetValues(fixitWidget, arglst, 5);
      XFlush(dsply);
    }/* MAke changes current */

  XtUnmanageChild(fixitscreen);

}/* FIXIT_DEACTIVATE */





/*************************************************************
 * movie_contour_activate() - Get movie Contour screen up on 
 * display.
 */
void movie_contour_activate(Widget widget, int *tag, 
			    XmAnyCallbackStruct *callback_data)
{/* MOVIE_CONTOUR_ACTIVATE */
  XmString simpleString;

  /* Now setup the scrolls accordingly */
  /* Always use the section min and max because we cue off of the sections
   * to contour shaded views */
  XtSetArg(arglst[0], XmNmaximum, (int)(secMaxPixel+5.0));
  XtSetArg(arglst[1], XmNminimum, (int)(secMinPixel));
  XtSetArg(arglst[2], XmNsliderSize, (int)(5));
  XtSetArg(arglst[3], XmNvalue, (int)(0));
  XtSetArg(arglst[4], XmNpageIncrement, (int)(5));
  XtSetArg(arglst[5], XmNincrement, (int)(1));
  XtSetValues(movieContourScrollWidget, arglst, 6);
  XFlush(dsply);

  /* And the value labels */
  sprintf(outStr,"%d",(0));
  createSimpleString(&simpleString,outStr);
  XtSetArg(arglst[0], XmNlabelString, simpleString);
  XtSetValues(movieContourLabelWidget, arglst, 1); 
  XFlush(dsply);
  XmStringFree(simpleString); 
  XFlush(dsply);

  raiseIt(movie_contour_screen);

}/* MOVIE_CONTOUR_ACTIVATE */





/*************************************************************
 * movie_contour_deactivate() - Get rid of movieContour.
 */
void movie_contour_deactivate(Widget widget, int *tag, 
			      XmAnyCallbackStruct *callback_data)
{/* MOVIE_CONTOUR_DEACTIVATE */

  XtUnmanageChild(movie_contour_screen);

}/* MOVIE_CONTOUR_DEACTIVATE */




/******************************************************************
 *------------------------------------------------------------
 * histEqualize - Gotten from EMMA in hist.c module. Sept 1997
 *------------------------------------------------------------
 */
void histEqualize(int dim1, int dim2, 
		  unsigned char *image, 
		  int firstPixel, int ncolors, 
		  unsigned char *outBuffer,
		  int min, int max, 
		  int buffer_width)
{/* HISTEQUALIZE */
   int i, j, k;
   int *hist;
   float *prob, *sum;
   float scaleFactor;
   unsigned short level;
   unsigned short *short_image;

   CK(hist = (int *)malloc(ncolors*sizeof(int)));
   CK(prob = (float *)malloc(ncolors*sizeof(float)));
   CK(sum = (float *)malloc(ncolors*sizeof(float)));

   for(i=0;i<ncolors;i++)
   {
     hist[i] = 0;
     prob[i] = 0.;
   }

   /* dcm  This code is replaced by the following
      for(j=0;j<dim1;j++)
      for(k=0;k<dim2;k++)
      hist[image[k*dim1+j]]++;
      */ 
   
   if(hiColor)
     {
       short_image = (unsigned short *)image;
       max = 65500;
       scaleFactor= (float)(ncolors-1)/(float)(max-min);
       for(k=0;k<dim2;k++)
	 for(j=0;j<dim1;j++)
	   {/* Do all */
	     hist[(int)((short_image[k*buffer_width+j]-min)*scaleFactor)]++;
	   }/* Do all */
     }
   else
     {/* lo */
       scaleFactor= (float)(ncolors-1)/(float)(max-min);
       for(k=0;k<dim2;k++)
	 for(j=0;j<dim1;j++)
	   {/* Do all */
	     hist[(int)((image[k*buffer_width+j]-min)*scaleFactor)]++;
	   }/* Do all */
     }/* lo */
   
   for(i=0;i<ncolors;i++)
   {
     prob[i] = (float)hist[i]/(float)(dim1*dim2);
   }
     
   for(i=0;i<ncolors;i++)
   {
     sum[i] = 0.;
     for(j=0;j<=i;j++)
       sum[i] += prob[j];
   }

   for(i=0;i<ncolors;i++)
   {
     level = (unsigned short)(0.5+(ncolors-1)*sum[i]);
   }

   /* dcm  j and k loops have been reversed */
   for(k=0;k<dim2;k++)
     {/* K */
       int kTdim1,kTdim1Pj,thisVal;
       kTdim1 = k*buffer_width;
       for(j=0;j<dim1;j++)
	 {
	   kTdim1Pj = kTdim1 + j;
	   if(hiColor)
	     level = (unsigned short)(0.5+(ncolors-1)*sum[(int)
					   ((short_image[kTdim1Pj]-min)*
							   scaleFactor)]);
	   else
	     level = (unsigned short)(0.5+(ncolors-1)*sum[(int)
						((image[kTdim1Pj]-min)*
							   scaleFactor)]);
	   if(level >= ncolors)
	     level = ncolors-1;
	   /* transfImage[currImage][kTdim1+j] = (float)level; */
	   if(hiColor)
	     {
	       thisVal = (int)((float)firstPixel + (float)level);
	       XPutPixel(image_image1, j, k, pixels_return[thisVal]);
	       histogram[thisVal]++;
	     }
	   else
	     {
	       outBuffer[kTdim1Pj] = (float)firstPixel + (float)level;
	       histogram[outBuffer[kTdim1Pj]]++;
	     }
	 }
     }/* K */

   free(hist);
   free(prob); 
   free(sum);

}/* HISTEQUALIZE */




/*************************************************************
 * invert_activate() - Get invert screen on display.
 */
void invert_activate(Widget widget, int *tag, 
		     XmAnyCallbackStruct *callback_data)
{/* INVERT_ACTIVATE */

  raiseIt(invertscreen);

}/* INVERT_ACTIVATE */



/*************************************************************
 * invert_deactivate() - Get rid of invert screen.
 */
void invert_deactivate(Widget widget, int *tag, 
		       XmAnyCallbackStruct *callback_data)
{/* INVERT_DEACTIVATE */

  XtUnmanageChild(invertscreen);

}/* INVERT_DEACTIVATE */



/*************************************************************
 * updateDimensions() - They want to change dims on truncate
 * or addto
 */
void updateDimensions()
{/* UPDATEDIMENSIONS */
  int endVal,begVal,diffVal;
  char *thisText;

  thisText = XmTextGetString(taBegXW); 
  begVal  = (int)strtod(thisText,0);
  XtFree(thisText);
  thisText = XmTextGetString(taEndXW); 
  endVal  = (int)strtod(thisText,0);
  XtFree(thisText);
  diffVal = endVal - (begVal-1);
  if((diffVal % 2)==0)
    sprintf(outStr,"%d Columns EVEN",diffVal);
  else sprintf(outStr,"%d Columns",diffVal);
  doLabel(outStr, taDimXW);

  thisText = XmTextGetString(taBegYW); 
  begVal  = (int)strtod(thisText,0);
  XtFree(thisText);
  thisText = XmTextGetString(taEndYW); 
  endVal  = (int)strtod(thisText,0);
  XtFree(thisText);
  diffVal = endVal - (begVal-1);
  if((diffVal % 2)==0)
    sprintf(outStr,"%d Rows EVEN",diffVal);
  else sprintf(outStr,"%d Rows",diffVal);
  doLabel(outStr, taDimYW);

  thisText = XmTextGetString(taBegZW); 
  begVal  = (int)strtod(thisText,0);
  XtFree(thisText);
  thisText = XmTextGetString(taEndZW); 
  endVal  = (int)strtod(thisText,0);
  XtFree(thisText);
  diffVal = endVal - (begVal-1);
  if((diffVal % 2)==0)
    sprintf(outStr,"%d Sections EVEN",diffVal);
  else sprintf(outStr,"%d Sections",diffVal);
  if(fileType!=PACKEDPORTABLE)
    doLabel(outStr, taDimZW);

  XFlush(dsply);

}/* UPDATEDIMENSIONS */




/*************************************************************
 * doTheTa2Arrow() - Called from ta2ArrowActivate and 
 * ta2ArrowActivateButton().  
 */
void doTheTa2Arrow(Widget widget, float speedFactor)
{/* DOTHETA2ARROW */
  int endVal,begVal;
  char *thisText;

  if(widget==ta2DownArrowW)
    speedFactor *= -1.0;
  
  if(taArrowX)
    {/* OK to do */
      thisText = XmTextGetString(taEndXW); 
      endVal  = (int)strtod(thisText,0);
      endVal += (int)speedFactor;
      sprintf(outStr,"%d",endVal);
      XmTextSetString(taEndXW,outStr); 
      XtFree(thisText);
    }/* OK to do */

  if(taArrowY)
    {/* OK to do */
      thisText = XmTextGetString(taEndYW); 
      endVal  = (int)strtod(thisText,0);
      endVal += (int)speedFactor;
      sprintf(outStr,"%d",endVal);
      XmTextSetString(taEndYW,outStr); 
      XtFree(thisText);
    }/* OK to do */
  
  if((taArrowZ) && (fileType!=PACKEDPORTABLE))
    {/* OK to do */
      thisText = XmTextGetString(taEndZW); 
      endVal  = (int)strtod(thisText,0);
      endVal += (int)speedFactor;
      sprintf(outStr,"%d",endVal);
      XmTextSetString(taEndZW,outStr); 
      XtFree(thisText);
    }/* OK to do */

  updateDimensions();

}/* DOTHETA2ARROW */




/*************************************************************
 * ta2ArrowActivateButton() - CHaneg the spped on truncate/
 * addto arrows.
 */
void ta2ArrowActivateButton(Widget widget, XEvent *event)
{/* TA2ARROWACTIVATEBUTTON */
  XEvent thisEvent;
  float speedFactor;

  if(!locked)
    {/* Can't do if we're locked out */
      
      thisEvent= *(event);

      /* Get which button */
      if((thisEvent.xbutton.button == Button1))
	speedFactor = 1.0;;
      if((thisEvent.xbutton.button == Button2))
	speedFactor = 5.0;
      if((thisEvent.xbutton.button == Button3))
	speedFactor = 10.0;

      doTheTa2Arrow(widget, speedFactor);

    }/* Can't do if we're locked out */

}/* TA2ARROWACTIVATEBUTTON */





/*************************************************************
 * ta2ArrowActivate() - truncate/addto arrow
 */
void ta2ArrowActivate(Widget widget, XEvent *event)
{/* TA2ARROWACTIVATE */
  float speedFactor;

  speedFactor = 1.0;
  doTheTa2Arrow(widget, speedFactor);

}/* TA2ARROWACTIVATE */





/*************************************************************
 * doTheGbsArrow() - Called from gbsArrowActivate and 
 * gbsArrowActivateButton().  
 */
void doTheGbsArrow(Widget widget, float speedFactor)
{/* DOTHEGBSARROW */
  int begVal,endVal;
  char *thisText;

  if(widget==gbsDownArrowW)
    speedFactor *= -1.0;
  
  if(taArrowX)
    {/* OK to do */
      thisText = XmTextGetString(taBegXW); 
      begVal  = (int)strtod(thisText,0);
      begVal -= (int)speedFactor;
      sprintf(outStr,"%d",begVal);
      XmTextSetString(taBegXW,outStr); 
      XtFree(thisText);
      thisText = XmTextGetString(taEndXW); 
      endVal  = (int)strtod(thisText,0);
      endVal += (int)speedFactor;
      sprintf(outStr,"%d",endVal);
      XmTextSetString(taEndXW,outStr); 
      XtFree(thisText);
    }/* OK to do */

  if(taArrowY)
    {/* OK to do */
      thisText = XmTextGetString(taBegYW); 
      begVal  = (int)strtod(thisText,0);
      begVal -= (int)speedFactor;
      sprintf(outStr,"%d",begVal);
      XmTextSetString(taBegYW,outStr); 
      XtFree(thisText);
      thisText = XmTextGetString(taEndYW); 
      endVal  = (int)strtod(thisText,0);
      endVal += (int)speedFactor;
      sprintf(outStr,"%d",endVal);
      XmTextSetString(taEndYW,outStr); 
      XtFree(thisText);
    }/* OK to do */
  
  if((taArrowZ) && (fileType!=PACKEDPORTABLE))
    {/* OK to do */
      thisText = XmTextGetString(taBegZW); 
      begVal  = (int)strtod(thisText,0);
      begVal -= (int)speedFactor;
      sprintf(outStr,"%d",begVal);
      XmTextSetString(taBegZW,outStr); 
      XtFree(thisText);
      thisText = XmTextGetString(taEndZW); 
      endVal  = (int)strtod(thisText,0);
      endVal += (int)speedFactor;
      sprintf(outStr,"%d",endVal);
      XmTextSetString(taEndZW,outStr); 
      XtFree(thisText);
    }/* OK to do */

  updateDimensions();

}/* DOTHEGBSARROW */




/*************************************************************
 * gbsArrowActivateButton() - Global box size arrows
 */
void gbsArrowActivateButton(Widget widget, XEvent *event)
{/* GBSARROWACTIVATEBUTTON */
  XEvent thisEvent;
  float speedFactor;

  if(!locked)
    {/* Can't do if we're locked out */
      
      thisEvent= *(event);

      /* Get which button */
      if((thisEvent.xbutton.button == Button1))
	speedFactor = 1.0;;
      if((thisEvent.xbutton.button == Button2))
	speedFactor = 2.0;
      if((thisEvent.xbutton.button == Button3))
	speedFactor = 5.0;

      doTheGbsArrow(widget, speedFactor);

    }/* Can't do if we're locked out */

}/* GBSARROWACTIVATEBUTTON */





/*************************************************************
 * gbsArrowActivate() - Global box size arrows
 */
void gbsArrowActivate(Widget widget, XEvent *event)
{/* GBSARROWACTIVATE */
  float speedFactor;

  speedFactor = 1.0;
  doTheGbsArrow(widget, speedFactor);

}/* GBSARROWACTIVATE */





/*************************************************************
 * doTheTaArrow() - Called from taArrowActivate and 
 * taArrowActivateButton().  
 */
void doTheTaArrow(Widget widget, float speedFactor)
{/* DOTHETAARROW */
  int begVal;
  char *thisText;

  if(widget==taDownArrowW)
    speedFactor *= -1.0;
  
  if(taArrowX)
    {/* OK to do */
      thisText = XmTextGetString(taBegXW); 
      begVal  = (int)strtod(thisText,0);
      begVal += (int)speedFactor;
      sprintf(outStr,"%d",begVal);
      XmTextSetString(taBegXW,outStr); 
      XtFree(thisText);
    }/* OK to do */

  if(taArrowY)
    {/* OK to do */
      thisText = XmTextGetString(taBegYW); 
      begVal  = (int)strtod(thisText,0);
      begVal += (int)speedFactor;
      sprintf(outStr,"%d",begVal);
      XmTextSetString(taBegYW,outStr); 
      XtFree(thisText);
    }/* OK to do */
  
  if((taArrowZ) && (fileType!=PACKEDPORTABLE))
    {/* OK to do */
      thisText = XmTextGetString(taBegZW); 
      begVal  = (int)strtod(thisText,0);
      begVal += (int)speedFactor;
      sprintf(outStr,"%d",begVal);
      XmTextSetString(taBegZW,outStr); 
      XtFree(thisText);
    }/* OK to do */

  updateDimensions();

}/* DOTHETAARROW */




/*************************************************************
 * taArrowActivateButton() - Change speed on truncate/addto arrows.
 */
void taArrowActivateButton(Widget widget, XEvent *event)
{/* TAARROWACTIVATEBUTTON */
  XEvent thisEvent;
  float speedFactor;

  if(!locked)
    {/* Can't do if we're locked out */
      
      thisEvent= *(event);

      /* Get which button */
      if((thisEvent.xbutton.button == Button1))
	speedFactor = 1.0;;
      if((thisEvent.xbutton.button == Button2))
	speedFactor = 5.0;
      if((thisEvent.xbutton.button == Button3))
	speedFactor = 10.0;

      doTheTaArrow(widget, speedFactor);

    }/* Can't do if we're locked out */

}/* TAARROWACTIVATEBUTTON */





/*************************************************************
 * taArrowActivate() - truncate/adddto arrow
 */
void taArrowActivate(Widget widget, XEvent *event)
{/* TAARROWACTIVATE */
  float speedFactor;

  speedFactor = 1.0;
  doTheTaArrow(widget, speedFactor);

}/* TAARROWACTIVATE */





/*************************************************************
 * truncadd_half_activate() - Just do what's needed to
 * truncate or addto half of the dataset.
 */
void truncadd_half_activate(Widget widget, int *tag, 
			    XmAnyCallbackStruct *callback_data)
{/* TRUNCADD_HALF_ACTIVATE */

  if((*tag==FRONT)  && (fileType!=PACKEDPORTABLE))
    {/* */
      sprintf(outStr,"%d",(origNumSecs/2)+1);
      XmTextSetString(taBegZW,outStr); 
      sprintf(outStr,"%d",largestDim);
      XmTextSetString(taEndZW,outStr); 
    }/* */
  else if((*tag==BACK)  && (fileType!=PACKEDPORTABLE))
    {/* */
      sprintf(outStr,"%d",1);
      XmTextSetString(taBegZW,outStr); 
      sprintf(outStr,"%d",(origNumSecs/2)+1);
      XmTextSetString(taEndZW,outStr); 
    }/* */
  else if(*tag==LEFT)
    {/* */
      sprintf(outStr,"%d",1);
      XmTextSetString(taBegXW,outStr); 
      sprintf(outStr,"%d",(numCols/2)+1);
      XmTextSetString(taEndXW,outStr); 
    }/* */
  else if(*tag==RIGHT)
    {/* */
      sprintf(outStr,"%d",(numCols/2)+1);
      XmTextSetString(taBegXW,outStr); 
      sprintf(outStr,"%d",largestDim);
      XmTextSetString(taEndXW,outStr); 
    }/* */
  else if(*tag==TOP)
    {/* */
      sprintf(outStr,"%d",1);
      XmTextSetString(taBegYW,outStr); 
      sprintf(outStr,"%d",(numRows/2)+1);
      XmTextSetString(taEndYW,outStr); 
    }/* */
  else if(*tag==BOTTOM)
    {/* */
      sprintf(outStr,"%d",(numRows/2)+1);
      XmTextSetString(taBegYW,outStr); 
      sprintf(outStr,"%d",largestDim);
      XmTextSetString(taEndYW,outStr); 
    }/* */
  else if(*tag==WHOLE)
    {/* */
      sprintf(outStr,"%d",1);
      XmTextSetString(taBegXW,outStr); 
      XmTextSetString(taBegYW,outStr); 
      if (fileType!=PACKEDPORTABLE)
	XmTextSetString(taBegZW,outStr); 
      sprintf(outStr,"%d",largestDim);
      XmTextSetString(taEndXW,outStr); 
      XmTextSetString(taEndYW,outStr); 
      if (fileType!=PACKEDPORTABLE)
	XmTextSetString(taEndZW,outStr); 
    }/* */

  updateDimensions();
  XFlush(dsply);

  sprintf(outStr,"%d",(largestDim/2)+1);
/*  XmTextSetString(taCenterXW,outStr); 
  XmTextSetString(taCenterYW,outStr); 
  XmTextSetString(taCenterZW,outStr); 
*/
  XFlush(dsply);

}/* TRUNCADD_HALF_ACTIVATE */




/*************************************************************
 * truncadd_activate() - get the truncate/addto screen up.
 */
void truncadd_activate(Widget widget, int *tag, 
		       XmAnyCallbackStruct *callback_data)
{/* TRUNCADD_ACTIVATE */

  sprintf(outStr,"%d",1);
  XmTextSetString(taBegXW,outStr); 
  XmTextSetString(taBegYW,outStr); 
  XmTextSetString(taBegZW,outStr); 
  if(fileType==PACKEDPORTABLE)
    sprintf(outStr,"%d",purdueHeader.numCols);
  else
    sprintf(outStr,"%d",largestDim);
  XmTextSetString(taEndXW,outStr); 
  if(fileType==PACKEDPORTABLE)
    sprintf(outStr,"%d",purdueHeader.numRows);
  XmTextSetString(taEndYW,outStr); 
  if(fileType==PACKEDPORTABLE)
    sprintf(outStr,"%d",purdueHeader.nimg);
  XmTextSetString(taEndZW,outStr); 
  sprintf(outStr,"%d",(largestDim/2)+1);
/*  XmTextSetString(taCenterXW,outStr); 
  XmTextSetString(taCenterYW,outStr); 
  XmTextSetString(taCenterZW,outStr); 
*/
  if(fileType==PACKEDPORTABLE)
    {/* Turn off z plane */
      XtSetArg(arglst[0], XmNsensitive, FALSE);
    }/* Turn off z plane */
  else
    {/* Turn ON z plane */
      XtSetArg(arglst[0], XmNsensitive, TRUE);
    }/* Turn ON z plane */
  XtSetValues(taEndZW, arglst, 1);  
  XtSetValues(taBegZW, arglst, 1);  

  sprintf(outStr,"%.2f",(float)pifHeader.img_bkg);
  if(EQU(outStr,"nan"))
    sprintf(outStr,"0.0");
  XmTextSetString(taFillW,outStr); 
  updateDimensions();
  XFlush(dsply);

  raiseIt(truncaddscreen);
  mapShadeScreen_deactivate(0,0,0);

}/* TRUNCADD_ACTIVATE */




/*************************************************************
 * truncadd_deactivate() - get rid of truncate/addto screen.
 */
void truncadd_deactivate(Widget widget, int *tag, 
			 XmAnyCallbackStruct *callback_data)
{/* TRUNCADD_DEACTIVATE */

  XtUnmanageChild(truncaddscreen);

}/* TRUNCADD_DEACTIVATE */




/*************************************************************
 * truncadd_do_activate() - Actually do the truncation or
 * the adding to.
 */
void truncadd_do_activate(Widget widget, int *tag, 
			  XmAnyCallbackStruct *callback_data)
{/* TRUNCADD_DO_ACTIVATE */
  int i,j,k;
  char *truncaddText;
  int begX,begY,begZ,endX,endY,endZ,thisX,thisY,thisZ,tmpY;
  int centerX,centerY,centerZ,origX,origY,origZ;
  int residX,residY,residZ;
  int newNumCols,newNumRows,newNumSecs;
  float *temp_float_buffer;
  float thisPixel,fillValue;
  XmString simpleString;

  truncaddText = XmTextGetString(taFillW); 
  fillValue = (float)strtod(truncaddText,0);
  XtFree(truncaddText);
printf("Fill value = [%f]\n",fillValue);

  truncaddText = XmTextGetString(taBegXW); 
  begX = (int)strtod(truncaddText,0);
  XtFree(truncaddText);
  truncaddText = XmTextGetString(taBegYW); 
  begY = (int)strtod(truncaddText,0);
  XtFree(truncaddText);
  truncaddText = XmTextGetString(taBegZW); 
  begZ = (int)strtod(truncaddText,0);
  XtFree(truncaddText);

  truncaddText = XmTextGetString(taEndXW); 
  endX = (int)strtod(truncaddText,0);
  XtFree(truncaddText);
  truncaddText = XmTextGetString(taEndYW); 
  endY = (int)strtod(truncaddText,0);
  XtFree(truncaddText);
  truncaddText = XmTextGetString(taEndZW); 
  endZ = (int)strtod(truncaddText,0);
  XtFree(truncaddText);

/*
  truncaddText = XmTextGetString(taCenterXW); 
  centerX = (int)strtod(truncaddText,0);
  XtFree(truncaddText);
  truncaddText = XmTextGetString(taCenterYW); 
  centerY = (int)strtod(truncaddText,0);
  XtFree(truncaddText);
  truncaddText = XmTextGetString(taCenterZW); 
  centerZ = (int)strtod(truncaddText,0);
  XtFree(truncaddText);
*/

  if(doTrunc)
    {/* Fill it up */
      begX = truncBegX;
      endX = truncEndX;
      begY = truncBegY;
      endY = truncEndY;
      begZ = truncBegZ;
      endZ = truncEndZ;
      fillValue = truncVal;
    }/* Fill it up */


  /* Humans number from 1 not zero */
  begX--;
  begY--;
  begZ--;
  centerX--;
  centerY--;
  centerZ--;



  newNumCols = endX - begX;
  newNumRows = endY - begY;
  newNumSecs = endZ - begZ;
  numCols = purdueHeader.numCols;
  featherRadius += (newNumCols - numCols)/2;
  boxRadius     += (newNumCols - numCols)/2;
  
  /* Y is a special case here - Tim numbers row starting
   * from the bottom.
   */

  if(fileType == PACKEDPORTABLE)
    newNumSecs = 1;  /* We don't have secs here */

  if(newNumCols <= 0 || newNumRows <=0 || newNumSecs <=0)
    {/* Can't do */
      sprintf(outStr,"Can't do this:  [cols=%d,rows=%d,secs=%d]",
	      newNumCols,
	      newNumRows,
	      newNumSecs);
      if(doTrunc)
	printf("%s\n",outStr);
      else
	askThem(outStr,FALSE,XmDIALOG_INFORMATION);
      goto noTruncAdd;
    }/* Can't do */


  if(fileType == PACKEDPORTABLE)
    {/* DO for packed boxes */
      int m,n,o,nc,nr;
      float *thisPackImg;
      float *tmpInt;

      /* Now calc how far we are from original */
      residY = centerY - (purdueHeader.numRows/2);
      residX = centerX - (purdueHeader.numCols/2);

      nc = purdueHeader.numCols;
      nr = purdueHeader.numRows;
      for(m=0;m<purdueHeader.nimg;m++)
	{/* DO all */

	  CK(tmpInt = (float *)malloc(sizeof(float)*newNumCols*newNumRows));
	  for(j=0;j<newNumRows;j++)
	    for(k=0;k<newNumCols;k++)
	      tmpInt[((j*newNumCols)+k)] = fillValue;

	  thisPackImg = packImg[m];
	  fillValue = (float)thisPackImg[0];
	  printf("Fill value is bkg value = [%f]\n",fillValue);

	  for(j=begY;j<endY;j++)
	    {/* Do rows */

	      /***************************
               * TSB orientation :
		 thisY = (newNumRows-(j-begY)-1)*newNumCols;
	         origY = (nr - j) - 1 ;
		 We've changed it to be XWindows coords as the code
		 in the 3D Rendering screen will be VERY challenging
		 to make that consistent with this convention.
		 RWa March 2004 
		 * RWA April 2004 Here's the code to make it XWindows:
	           thisY = (begY-1)*newNumCols;
	           origY = j;
		 *  I've changedf back to TSB orientation for all
                    in robem
	      ************************************/
	      thisY = (newNumRows-(j-begY)-1)*newNumCols;
	      origY = (nr - j) - 1 ;
	      
	      for(k=begX;k<endX;k++)
		{/* Do cols */
		  
		  thisX = (k-begX);
		  origX = k;
		  
		  if(origY < nr && 
		     origX < nc &&
		     origY >= 0 &&
		     origX >= 0 
		     )
		    thisPixel= tmpInt[(thisY+thisX)] =
		      thisPackImg[((origY*nc)+origX)];
		  else
		    thisPixel= 
		      tmpInt[(thisY+thisX)] = 
		      fillValue;
		  
		}/* Do cols */
	      
	    }/* Do rows */

	  packImg[m] = tmpInt;
	  free(thisPackImg);

	}/* DO all */	  

      /* Now clip this guy to our colormap */ 
      if(minPackPixel < 0.0)
	scalePackFactor = (((float)tryColors - 7.0) / 
			   (maxPackPixel-minPackPixel));
      else scalePackFactor = (((float)tryColors - 7.0) / 
			      maxPackPixel);
      printf("minPackPixel [%f]  max[%f]     scale[%f]\n",
	     minPackPixel,
	     maxPackPixel,
	     scalePackFactor);
      
      purdueHeader.numCols = newNumCols;
      purdueHeader.numRows = newNumRows;
      packRadius = purdueHeader.numCols/2;
      stuffPackedImage(purdueHeader.numCols,
		       purdueHeader.numRows,
		       packImg,purdueHeader.nimg,
		       imgNum);  /* Start at radius'th image */
      DrawWindow(); 

      XtUnmanageChild(truncaddscreen);
      
      sprintf(outStr,
	      "Boxes successfully truncated/added to");
      if(doTrunc)
	printf("%s\n",outStr);
      else
	askThem(outStr,FALSE,XmDIALOG_INFORMATION);
      
      /* Give feedback on main window title bar */
      if(!masterProcess)
	sprintf(outStr,"Robem: [%s] - [%d,%d,%d] [%s]",fileName.addr,
		purdueHeader.numCols,purdueHeader.numRows,
		purdueHeader.nimg, "CHILD");
      else
	sprintf(outStr,"Robem: [%s] - [%d,%d,%d]",fileName.addr,
		purdueHeader.numCols,purdueHeader.numRows,
		purdueHeader.nimg);
      createSimpleString(&simpleString,outStr);
      XtSetArg(arglst[0], XmNdialogTitle, simpleString );
      XtSetArg(arglst[1], XmNrecomputeSize, FALSE );
      XtSetValues(main_window_widget, arglst, 2); 
      XFlush(dsply);
      XmStringFree(simpleString); 
      XFlush(dsply);
      
    }/* DO for packed boxes */
  else
    {/* MAP */
	  
      /* Now calc how far we are from original */
      residZ = centerZ - (origNumSecs/2);
      residY = centerY - (numRows/2);
      residX = centerX - (numCols/2);

      temp_float_buffer = float_image_buffer1;
      CK(float_image_buffer1 = (float *)malloc(sizeof(float)*
					       newNumCols*
					       newNumRows*
					       newNumSecs));
      
      maxMapPixel= -99999999.9;
      minMapPixel= 99999999.9;
      maxMapPixel = -HUGE;  /* dgm -HUGE instead of -99999.9 */
      minMapPixel = HUGE;   /* dgm HUGE instead of 9999999.9 */
       for(i=begZ;i<endZ;i++)
	{/* Do sections */
	  
	  thisZ = (i-begZ)*newNumCols*newNumRows;
	  origZ = i;
	  
	  for(j=begY;j<endY;j++)
	    {/* Do rows */
	      
	      /* Original way :  
		 thisY = (j-begY)*newNumCols;
		 origY = j; 
	      */
	      /***************************
               * TSB orientation :
		 thisY = (newNumRows-(j-begY)-1)*newNumCols;
	         origY = (numRows - j) - 1 ;
		 We've changed it to be XWindows coords as the code
		 in the 3D Rendering screen will be VERY challenging
		 to make that consistent with this convention.
		 RWa March 2004 
		 * RWA April 2004 Here's the code to make it XWindows:
	           thisY = (j-begY)*newNumCols;
	           origY = j;
		 *  I've changedf back to TSB orientation for all
                    in robem
	      ************************************/
	      thisY = (newNumRows-(j-begY)-1)*newNumCols;
	      origY = (numRows - j) - 1 ;
	      
	      for(k=begX;k<endX;k++)
		{/* Do cols */
		  
		  thisX = (k-begX);
		  origX = k;
		  
		  if(origZ < origNumSecs && 
		     origY < numRows && 
		     origX < numCols &&
		     origZ >= 0 &&
		     origY >= 0 &&
		     origX >= 0 
		     )
		    thisPixel= float_image_buffer1[(thisZ+thisY+thisX)] =
		      temp_float_buffer[(
					 (origZ*numCols*numRows)+
					 (origY*numCols)+origX
					 )];
		  else
		    thisPixel= 
		      float_image_buffer1[(thisZ+thisY+thisX)] = 
		      fillValue;
		  
		  if(thisPixel > maxMapPixel)
		    maxMapPixel= (float)thisPixel;
		  if(thisPixel < minMapPixel)
		    minMapPixel= (float)thisPixel;
		  
		}/* Do cols */
	      
	    }/* Do rows */
	  
	}/* Do sections */
      
      /* Now get rid of the old */
      free(temp_float_buffer);
      
      /* Keep track for traslation stuff */
      pifHeader.transX += begX;
      pifHeader.transY += begY;
      pifHeader.transZ += begZ;
      if(!pifHeader.isTrunc)
	{/* First truncation */
	  pifHeader.origX= numCols;
	  pifHeader.origY= numRows;
	  pifHeader.origZ= numSecs;
	}/* First truncation */
      pifHeader.isTrunc = TRUE;
      /* printf("here 111 [%d][%d][%d][%d][%d][%d][%d]\n",
	     pifHeader.transX,
	     pifHeader.transY,
	     pifHeader.transZ,
	     pifHeader.origX,
	     pifHeader.origY,
	     pifHeader.origZ,
	     pifHeader.isTrunc 
	     ); */


      numCols = newNumCols;
      numRows = newNumRows;
      numSecs = newNumSecs;
      origNumSecs= numSecs;
      
      /* Now remember these forever */
      lowRes_buffer1 = float_image_buffer1;
      lowResCols = numCols;
      lowResRows = numRows;
      lowResSecs = numSecs;
      
      /* Remember how big we are on disk */
      file_width  = numCols;
      file_height = numRows;
      
      purdueHeader.numCols = numCols;
      purdueHeader.numRows = numRows;
      purdueHeader.numSecs = origNumSecs;
      
      /* For writing new file */
      fileBinSecs  = numSecs;
      fileBinRows  = numRows;
      fileBinCols  = numCols;
      
      xCen = (double)(numCols/2);
      yCen = (double)(numRows/2);
      zCen = (double)(origNumSecs/2);
      
      largestDim = numCols;
      largestDim = MAX(largestDim,numRows);
      largestDim = MAX(largestDim,origNumSecs);
      
      nsecs= nimg= numSecs;
      
      /* We've got to make this larger for asymmetric maps 
       * and all maps that have information "outside" of the
       * normal sphere of our normal maps,
       * or else we just keep cliping the box into a sphere */
      if(numCols==largestDim && numRows==largestDim
	 && (origNumSecs)==largestDim)
	{/* Symmetric */
	  aSymmetric = FALSE;
	  origGlobMaxRadius= globMaxRadius = (double)((double)origNumSecs/2)
	    +5.0;
	  globMinRadius = (double)0.0;
	}/* Symmetric */
      else
	{/* asymmetric */
	  aSymmetric = TRUE;
	  origGlobMaxRadius=  globMaxRadius = (double)((double)origNumSecs);
	  globMinRadius = (double)0.0;
	}/* asymmetric */
      
      /* Set scrolls on depthcue screen */
      /* Give feedback */
      sprintf(outStr,"%.0f",globMinRadius);
      createSimpleString(&simpleString,outStr);
      XtSetArg(arglst[0], XmNvalue, outStr );
      XtSetArg(arglst[1], XmNrecomputeSize, FALSE );
      XtSetValues(minRadiusValueWidget, arglst, 2); 
      XFlush(dsply);
      XmStringFree(simpleString); 
      
      sprintf(outStr,"%.0f",globMaxRadius);
      createSimpleString(&simpleString,outStr);
      XtSetArg(arglst[0], XmNvalue, outStr );
      XtSetArg(arglst[1], XmNrecomputeSize, FALSE );
      XtSetValues(maxRadiusValueWidget, arglst, 2); 
      XFlush(dsply);
      XmStringFree(simpleString); 
      
      
      /* Radius is origNumSecs/2 */
      XtVaSetValues(maxRadiusScrollWidget,XmNmaximum, 
		    (int)((globMaxRadius*2)+5),NULL);
      XtVaSetValues(maxRadiusScrollWidget,XmNsliderSize, (int)(5),NULL);
      XtVaSetValues(maxRadiusScrollWidget,XmNvalue, (int)(globMaxRadius),NULL);
		    
      XtVaSetValues(minRadiusScrollWidget,XmNmaximum, 
		    (int)((globMaxRadius*2)+5),NULL);
      XtVaSetValues(minRadiusScrollWidget,XmNsliderSize, (int)(5),NULL);
      XtVaSetValues(minRadiusScrollWidget,XmNvalue, (int)(globMinRadius),NULL);
      
      printf("MAX[%f]  MIN[%f]\n",maxMapPixel,minMapPixel);
      secMaxPixel = maxMapPixel;
      secMinPixel = minMapPixel;
      
      minSurf = (int)MAX(minMapPixel,0.0);
      maxSurf = MAX(maxMapPixel,(0.0-minMapPixel));
      /* We're doing the surfaceDensity in tenths now */
      if(maxSurf > 5000.0)
	{
	  XtVaSetValues(surfDensScrollWidget,XmNmaximum, 
			(int)((maxSurf*10.0)+50.0),NULL);
	  XtVaSetValues(surfDensScrollWidget,XmNminimum, 
			(int)(minSurf*10.0),NULL);
	  XtVaSetValues(surfDensScrollWidget,XmNsliderSize, 
			(int)(maxSurf*0.01),NULL);
	  XtVaSetValues(surfDensScrollWidget,XmNpageIncrement, 
			(int)(maxSurf*0.01),NULL);
	  XtVaSetValues(surfDensScrollWidget,XmNincrement, 
			(int)(maxSurf*0.002),NULL);
	  XtVaSetValues(surfDensScrollWidget,XmNvalue, (int)(10),NULL);
	}
      else if(maxSurf > 500.0)
	{
	  XtVaSetValues(surfDensScrollWidget,XmNmaximum, 
			(int)((maxSurf*10.0)+50.0),NULL);
	  XtVaSetValues(surfDensScrollWidget,XmNminimum, 
			(int)(minSurf*10.0),NULL);
	  XtVaSetValues(surfDensScrollWidget,XmNsliderSize, (int)(500),NULL);
	  XtVaSetValues(surfDensScrollWidget,XmNpageIncrement, (int)500,NULL);
	  XtVaSetValues(surfDensScrollWidget,XmNincrement, (int)100,NULL);
	  XtVaSetValues(surfDensScrollWidget,XmNvalue, (int)(10),NULL);
	}
      else
	{
	  /* We're doing the surfaceDensity in tenths now */
	  XtVaSetValues(surfDensScrollWidget,XmNmaximum, 
			(int)((maxSurf*10.0)+50.0),NULL);
	  XtVaSetValues(surfDensScrollWidget,XmNminimum, 
			(int)(minSurf*10.0),NULL);
	  XtVaSetValues(surfDensScrollWidget,XmNsliderSize, (int)(50),NULL);
	  XtVaSetValues(surfDensScrollWidget,XmNpageIncrement, (int)50,NULL);
	  XtVaSetValues(surfDensScrollWidget,XmNincrement, (int)10,	NULL);
	  XtVaSetValues(surfDensScrollWidget,XmNvalue, (int)(10),NULL);
	}
      surfaceDensity= 1.0;  /* 1 tenth of 10 */
      
      /* Give feedback */
      sprintf(outStr,"%d",-1);
      createSimpleString(&simpleString,outStr);
      XtSetArg(arglst[0], XmNvalue, outStr );
      XtSetArg(arglst[1], XmNrecomputeSize, FALSE );
      XtSetValues(surfDensValueWidget, arglst, 2); 
      XFlush(dsply);
      XmStringFree(simpleString); 
      
      createSimpleString(&simpleString,"SecNum");
      XtSetArg(arglst[0], XmNlabelString, simpleString);
      XtSetArg(arglst[1], XmNrecomputeSize, FALSE );
      XtSetValues(imgNumButtonW, arglst, 2); 
      XFlush(dsply);
      XmStringFree(simpleString); 
      
      sprintf(outStr,"1");
      XmTextSetString(imgNumTextWidget,outStr);
      
      
      /* To make this work for now let's stuff the nimg field with the
       * number of sections.  Really we should change any and all references
       * to nimg to be origNumSecs. We really only work on one image at a 
       * time so this may never have any affect.
       */
      nsecs= purdueHeader.nimg = origNumSecs;
      
      /* Write stats */
      calcMapStats(float_image_buffer1,
		   &purdueHeader.img_bkg, 
		   &purdueHeader.map_min,
		   &purdueHeader.map_max, 
		   &purdueHeader.img_avg, 
		   &purdueHeader.img_std,
		   &numCols, &numRows, &origNumSecs);
      
      /* Write stats */
      sprintf(outStr,"\nStatistics for file [%s]\n\
\n\
Title [%s]\n\
\n\
File size   [%d]x[%d]x[%d]\n\
Background  [%f]\n\
MIN,MAX     [%f],[%f]\n\
AVG,STD     [%f],[%f]\n\
\n",
	      fileName.addr,
	      title,
	      numCols,	  numRows, origNumSecs,
	      purdueHeader.img_bkg, 
	      minMapPixel, maxMapPixel,
	      purdueHeader.img_avg, purdueHeader.img_std
	      );
      XmTextSetString(statTextWidget,outStr);
      XFlush(dsply);
      
      
      XtUnmanageChild(truncaddscreen);
      
      sprintf(outStr,
	      "Map successfully truncated/added to");
      if(doTrunc)
	printf("%s\n",outStr);
      else
	askThem(outStr,FALSE,XmDIALOG_INFORMATION);
      
      /* Their shaded views are no longer valid - show them 
       * their new sections */
      shadedMaps= FALSE;
      dispSections = TRUE;
      XtSetArg(arglst[0], XmNlabelPixmap, sectionPixmap);
      XtSetArg(arglst[1], XmNselectPixmap, unsectionPixmap);
      XtSetArg(arglst[2], XmNrecomputeSize, FALSE);
      XtSetValues(sectionWidget, arglst, 3); 
      if(dispSections)
	{stuffMapImage((float *)map_image_buffer, 
		       (float *)float_image_buffer1, 
		       file_width,file_height,
		       nsecs,
		       imgNum, 0);  /* Start at zero'th image */
	DrawWindow(); }
      

      /* Give feedback on main window title bar */
      if(!masterProcess)
	sprintf(outStr,"Robem: [%s] - [%d,%d,%d] [%s]",fileName.addr,
		numCols,numRows,numSecs, "CHILD");
      else
	sprintf(outStr,"Robem: [%s] - [%d,%d,%d]",fileName.addr,
		numCols,numRows,numSecs);
      createSimpleString(&simpleString,outStr);
      XtSetArg(arglst[0], XmNdialogTitle, simpleString );
      XtSetArg(arglst[1], XmNrecomputeSize, FALSE );
      XtSetValues(main_window_widget, arglst, 2); 
      XFlush(dsply);
      XmStringFree(simpleString); 
      XFlush(dsply);
      
    }/* MAP */
  
 noTruncAdd:;
  
}/* TRUNCADD_DO_ACTIVATE */




/* ************************************************************
 * truncadd_changed() - Whenever text widgets changed
 * call here.
 */
void truncadd_changed(Widget widget, int *tag, 
		      XmAnyCallbackStruct *callback_data)
{/* truncadd_CHANGED */
  char *rotflt,outStr[64];
  XmString simpleString;
  int i,thisVal;
  char *truncaddText;

}/* TRUNCADD_CHANGED */





/*************************************************************
 * rotate_activate() - Get physical rotate MAP screen up.
 */
void rotate_activate(Widget widget, int *tag, 
		     XmAnyCallbackStruct *callback_data)
{/* ROTATE_ACTIVATE */

  raiseIt(rotatescreen);

}/* ROTATE_ACTIVATE */




/*************************************************************
 * rotate_deactivate() - Get rid of physical rotate MAP screen.
 */
void rotate_deactivate(Widget widget, int *tag, 
		       XmAnyCallbackStruct *callback_data)
{/* ROTATE_DEACTIVATE */

  XtUnmanageChild(rotatescreen);

}/* ROTATE_DEACTIVATE */




/**********************************************************
 * phys_map_rotate() - Actually rotate the MAP in memory.
*/
void phys_map_rotate(float *gold_buffer,
		     float *temp_buffer,
		     float theta, float phi, float omega)
{/* PHYS_MAP_ROTATE */
  char axis;
  int  i,j,k;
  fString_t fAxis;

  if(gold_buffer != temp_buffer)
    {/* Stuff it */
      for(i=0;i<origNumSecs;i++)
	{
	  for(j=0;j<numRows;j++)
	    {
	      yOff = (i*numCols*numRows)+(j*numCols);
	      for(k=0;k<numCols;k++)
		{
		  temp_buffer[yOff+k]= gold_buffer[yOff+k];
		}
	    }
	}
    }/* Stuff it */

  /* Do Phi */
  axis = 'Z';
  if(phi != 0.0)
    map_rot_(temp_buffer,
	     &numCols, &numRows, &origNumSecs,
	     &phi, &axis);
  /* Do Theta */
  axis = 'Y';
  if(theta != 0.0)
    map_rot_(temp_buffer,
	     &numCols, &numRows, &origNumSecs,
	     &theta, &axis);
  /* Do Omega */
  axis = 'Z';
  if(omega != 0.0)
    map_rot_(temp_buffer,
	     &numCols, &numRows, &origNumSecs,
	     &omega, &axis);
  

}/* PHYS_MAP_ROTATE */




/*******************************************************
 * new_phys_map_rotate() -  Better way to rotate???
*/
void new_phys_map_rotate(float *gold_buffer,
			 float *temp_buffer,
			 float theta, float phi, float omega)
{/* NEW_PHYS_MAP_ROTATE */
  int  i,j,k,yOff,zOff,xOff;
  float tempF;
  double t1,t2,t3,t4,t5,t6,t7,t8,t9;
  double cphi,sphi,ctheta,stheta,comega,somega;
  double thisCol,thisRow,thisSecn;
  double xCen,yCen,zCen;
  int ncol,nrow;

  /* We've got a bug whereby we've got a lot of problems with
   * the original map_prj_all() rtn that does the rotation. */
  /* WHY does phi not have to have the sign reversed here like
   * it does in emUnix.c in the thetaPhi_dphshd_C() rtn??? */
  comega = (double)cos((double)deg2rad((double)phi));
  somega = (double)sin((double)deg2rad((double)phi));
  cphi = (double)cos((double)deg2rad((double)omega));
  sphi = (double)sin((double)deg2rad((double)omega));
  ctheta = (double)cos((double)deg2rad((double)0.0-theta));
  stheta = (double)sin((double)deg2rad((double)0.0-theta));

  /*********************************************
   *    COMPUTE ELEMENTS OF ROTATION MATRIX    *
   *********************************************/
  t1 = ctheta*cphi*comega + sphi*somega;
  t2 = ctheta*sphi*comega - cphi*somega;
  t3 = -stheta*comega;
  t4 = ctheta*cphi*somega - sphi*comega;
  t5 = ctheta*sphi*somega + cphi*comega;
  t6 = -stheta*somega;
  t7 = stheta*cphi;
  t8 = stheta*sphi;
  t9 = ctheta;

  ncol = numCols;
  nrow = numRows;
  xCen = (double)(numCols/2);
  yCen = (double)(numRows/2);
  zCen = (double)(origNumSecs/2);

  for(i=0;i<origNumSecs;i++)
    {
      zOff = (i*numCols*numRows);
      for(j=0;j<numRows;j++)
	{
	  yOff = zOff+(j*numCols);
	  for(k=0;k<numCols;k++)
	    {
	      xOff = yOff + k;
	      newTranslateThetaPhiOmega(&thisCol,&thisRow,&thisSecn,
					xCen,yCen,zCen,
					k,j,i,
					t1,t2,t3,t4,t5,t6,t7,t8,t9);
	      if(thisCol >= 1.0 &&
		 thisCol <= (numCols-2) &&
		 thisRow >= 1.0 &&
		 thisRow <= (numRows-2) &&
		 thisSecn >= 1.0 &&
		 thisSecn <= (origNumSecs-2))
		temp_buffer[xOff]=   getVoxel(gold_buffer,
					      thisCol,thisRow,thisSecn,
					      theta, phi, omega,
					      ncol, nrow);
	      else temp_buffer[xOff] = 0.0;

	    }
	}
    }

}/* NEW_PHYS_MAP_ROTATE */




/*************************************************************
 * rotate_do_activate() - Do the rotate and then display it.
 */
void rotate_do_activate(Widget widget, int *tag, 
			XmAnyCallbackStruct *callback_data)
{/* ROTATE_DO_ACTIVATE */
  char *rotateText;
  float rotateTheta,rotatePhi,rotateOmega;
  char axis;
  int  i;
  fString_t fAxis;
  float *rotate_buffer;

  rotateText = XmTextGetString(rotateThetaW); 
  rotateTheta = (float)strtod(rotateText,0);
  XtFree(rotateText);
  rotateText = XmTextGetString(rotatePhiW); 
  rotatePhi = (float)strtod(rotateText,0);
  XtFree(rotateText);
  rotateText = XmTextGetString(rotateOmegaW); 
  rotateOmega = (float)strtod(rotateText,0);
  XtFree(rotateText);

  rotateTheta = 0.0 - rotateTheta;

  phys_map_rotate(float_image_buffer1,
		  float_image_buffer1,
		  rotateTheta,
		  rotatePhi,
		  rotateOmega
		  );

  XtUnmanageChild(rotatescreen);
  
  /* DO we really need to do this??? ********
  sprintf(outStr,
	  "Map successfully rotated \n\
theta %.2f  phi %.2f  omega %.2f \n",
	  (0.0-rotateTheta),
	  rotatePhi,
	  rotateOmega);
  askThem(outStr,FALSE,XmDIALOG_INFORMATION);
  **********************/

  /* Their shaded views are no longer valid - show them 
   * their new sections 
   * shadedMaps= FALSE;
   * dispSections = TRUE;
   * XtSetArg(arglst[0], XmNlabelPixmap, sectionPixmap);
   * XtSetArg(arglst[1], XmNselectPixmap, unsectionPixmap);
   * XtSetArg(arglst[2], XmNrecomputeSize, FALSE);
   * XtSetValues(sectionWidget, arglst, 3); 
   */

  if(dispSections)
    {stuffMapImage((float *)map_image_buffer, 
		   (float *)float_image_buffer1, 
		  file_width,file_height,
		  nsecs,
		  imgNum, 0);  /* Start at zero'th image */
     DrawWindow(); }
  else
    {/* Try and depth cue it etc.. */
      /* Let's force a "2-fold" display */
      i= 2;
      foldAxis_activate(widget, &i, callback_data);
    }/* Try and depth cue it etc.. */

 noRotate:;

}/* ROTATE_DO_ACTIVATE */




/* ************************************************************
 * rotate_create() - Remember the rotate screen for use later.  
 */
void rotate_create(Widget widget, int *tag, 
		   XmAnyCallbackStruct *callback_data)
{/* ROTATE_CREATE */

  if(*tag==THETAT)
    rotateThetaW = widget;
  else if(*tag==PHIT)
    rotatePhiW = widget;
  else if(*tag==OMEGAT)
    rotateOmegaW = widget;
  
}/* ROTATE_CREATE */




/* ************************************************************
 * fix2D_flicker_activate() - Flicker the new fixed 2D image
 * with what was on disk.
 */
void fix2D_flicker_activate(Widget widget, int *tag, 
			    XmAnyCallbackStruct *callback_data)
{/* FIX2D_FLICKER_ACTIVATE */
  int i,j,k,x,y,offset;
  int numRows,numCols;
  int *thisImg,*dispImg;

  flickerFix2D = *tag;

  if(fileType==PACKEDPORTABLE)
    stuffPackedImage(purdueHeader.numCols,
		     purdueHeader.numRows,
		     packImg,purdueHeader.nimg,
		     imgNum);  /* Start at radius'th image */
  else
    {/* do it */
      /* Force a recalc of full screen */
      fullScreen_buffer.dispx = fullScreen_buffer.dispy = 0;
      virtual_read_activate(-99,-99);
    }/* do it */

  DrawWindow();

}/* FIX2D_FLICKER_ACTIVATE */




/* ************************************************************
 * fix2D_create() - Remember the fix2D screen widgets for use 
 * later.  
 */
void fix2D_create(Widget widget, int *tag, 
		  XmAnyCallbackStruct *callback_data)
{/* FIX2D_CREATE */

  if(*tag==FIXREGIONS)
    fix2DRegionsTextW = widget;
  else if(*tag==FIXSTD)
    fix2DStdTextW = widget;
  
}/* FIX2D_CREATE */




/*************************************************************
 * fix2D_do_activate() - Do the fixing of this image
 */
void fix2D_do_activate(Widget widget, int *tag, 
		       XmAnyCallbackStruct *callback_data)
{/* FIX2D_DO_ACTIVATE */
  int  i,j,l,m;
  char *fixText;
  int fixRegions;
  float fixStd;
  float thisBkg, thisMin, thisMax;
  int   cutHigh, cutLow; 
  float thisAvg, thisStd;
  float *thisImg,*thisImg2;

  raiseIt(statusScreenWidget);
  XFlush(dsply);
  sprintf(outStr,"Setting up to fix values.");
  checkStatus(outStr);
  
  fixText = XmTextGetString(fix2DRegionsTextW); 
  fixRegions = (int)strtod(fixText,0);
  XtFree(fixText);
  fixText = XmTextGetString(fix2DStdTextW); 
  fixStd = (float)strtod(fixText,0);
  XtFree(fixText);

  if(doFix && batchFixStd != 0.0)
    {/* We're in batch mode and they've asked for a specific stDev */
      fixStd = batchFixStd;
    }/* We're in batch mode and they've asked for a specific stDev */

  fixRegions = MAX(MIN(fixRegions,MAXREGIONS),1);
  fixStd     = MAX(MIN(fixStd,100.0),0.1);
  printf("Using stDev==[%f]\n",fixStd);

  if(fileType==PACKEDPORTABLE)
    fixRegions = 1;

  sprintf(outStr,"%d",fixRegions);
  XmTextSetString(fix2DRegionsTextW,outStr);
  sprintf(outStr,"%.2f",fixStd);
  XmTextSetString(fix2DStdTextW,outStr);


  /* Let's reload all data in case it's corrupted from last time */
  if(fileType==PACKEDPORTABLE)
    {/* packed */
      for(i=0;i<purdueHeader.nimg;i++)
	{
	  memcpy(packImg[i],fix2DpackImg_buffer[i],
		 (purdueHeader.numCols*
		  purdueHeader.numRows*sizeof(int)));
	  thisImg2 = fix2DpackImg_diff[i];
	  for(l=0;l<purdueHeader.numRows;l++)
	    for(m=0;m<purdueHeader.numCols;m++)
	      {
		thisImg2[((l*purdueHeader.numCols)+m)] = 0;
	      }
	}
    }/* packed */
  else
    {/* scanned */
      memcpy(inCore_buffer,fix2DinCore_buffer,
	     (file_width*file_height*sizeof(float)));
      for(l=0;l<file_height;l++)
	for(m=0;m<file_width;m++)
	  {
	    fix2DinCore_diff[((l*file_width)+m)] = 0;
	  }
    }/* scanned */

  if(*tag==FIXRESTORE)
    {/* Just get out we've restored it above */
      goto restoreDone;
    }/* Just get out we've restored it above */

  if(*tag==FIXPREVIEW || *tag==FIX2D)
    {/* DO it non-destructively */
      int i,j,k;

      if(fileType==PACKEDPORTABLE)
	{/* packed */
	  for(k=0;k<purdueHeader.nimg;k++)
	    {/* img */
	      sprintf(outStr,"Processing Image [%d].",k);
	      checkStatus(outStr);

	      thisImg  = packImg[k];
	      thisImg2 = fix2DpackImg_diff[k];
	      calcSubScanStats(thisImg,
			       &thisBkg, 
			       &thisMin,
			       &thisMax,
			       &thisAvg, 
			       &thisStd,
			       purdueHeader.numCols, 
			       purdueHeader.numRows, 
			       purdueHeader.numCols,
			       0, 0);
	      cutHigh = (int)(thisAvg + (fixStd * thisStd));
	      cutLow  = (int)(thisAvg - (fixStd * thisStd));
	      for(i=0;i<purdueHeader.numRows;i++)
		{
		  for(j=0;j<purdueHeader.numCols;j++)
		    {
		      if(thisImg[((i*purdueHeader.numCols)+j)] > cutHigh ||
			 thisImg[((i*purdueHeader.numCols)+j)] < cutLow)
			{/* Cut it off */
			  thisImg[((i*purdueHeader.numCols)+j)] = 
			    (int)(thisAvg);
			  thisImg2[((i*purdueHeader.numCols)+j)] = 
			    (BYTE)TRUE;
			}/* Cut it off */
		    }
		}
	    }/* img */

	}/* packed */
      else
	{/* scanned */
	  int thisX, thisY, xStep,yStep,thisOne;
	  int drawX,drawY;

	  xStep = (int)(file_width  / (float)fixRegions);
	  yStep = (int)(file_height / (float)fixRegions);
	  xStep++; yStep++;  /* Make sure we get the last niblet in last 
			      * region */
	  
	  for(thisY=0,thisOne=1;thisY<file_height;thisY+= yStep,thisOne++)
	    {/* DO y */
	      sprintf(outStr,"Processing region [%d].",thisOne);
	      checkStatus(outStr);

	      for(thisX=0;thisX<file_width;thisX+=xStep)
		{/* X region */

		  calcSubScanStats(inCore_buffer,
				   &thisBkg, 
				   &thisMin,
				   &thisMax,
				   &thisAvg, 
				   &thisStd,
				   ((thisX+xStep) > file_width)?
				         (file_width-thisX) :xStep, 
				   ((thisY+yStep) > file_height)?
				         (file_height-thisY) :yStep, 
				   file_width,
				   thisX,thisY);

		    drawX = (int)(thisX/xStep);
		    drawY = (int)(thisY/yStep);

		  fix2Davg[(drawX)][(drawY)] = thisAvg;
		  fix2Dstd[(drawX)][(drawY)] = thisStd;

		  cutHigh = (int)(thisAvg + (fixStd * thisStd));
		  cutLow  = (int)(thisAvg - (fixStd * thisStd));

		  for(i=thisY;i<(thisY+yStep);i++)
		    {
		      for(j=thisX;j<(thisX+xStep);j++)
			{
			  if(((int)inCore_buffer[((i*file_width)+j)] > 
			      cutHigh) ||
			     ((int)inCore_buffer[((i*file_width)+j)] < 
			      cutLow)
			     )
			    {/* Cut it off */
			      if((i<file_height) && (j<file_width))
				{/* Do it */
				  inCore_buffer[((i*file_width)+j)] =	
				    (float)(thisAvg);
				  fix2DinCore_diff[((i*file_width)+j)] = 
				    (BYTE)TRUE;
				}/* Do it */
			    }/* Cut it off */
			}
		    }
		}/* X region */
	    }/* DO y */

	}/* scanned */

    }/* DO it non-destructively */

 restoreDone:;

  sprintf(outStr,"Writing to screen.");
  checkStatus(outStr);
  
  if(fileType==PACKEDPORTABLE)
    stuffPackedImage(purdueHeader.numCols,
		     purdueHeader.numRows,
		     packImg,purdueHeader.nimg,
		     imgNum);  /* Start at radius'th image */
  else
    {/* do it */
      /* Force a recalc of full screen */
      fullScreen_buffer.dispx = fullScreen_buffer.dispy = 0;
      virtual_read_activate(-99,-99);
    }/* do it */
  
  DrawWindow(); 
  

  if(*tag==FIX2D)
    {/* Destroy the data */

      if(fileType==MAP)
	{/* do nothing */}
      else
	{/* 2d */
	  int i;
	  if(fileType==PACKEDPORTABLE)
	    {for(i=0;i<purdueHeader.nimg;i++)
	      {
		if(fix2DpackImg_diff[i])
		  free(fix2DpackImg_diff[i]);
		fix2DpackImg_diff[i]= NULL;
		if(fix2DpackImg_buffer[i])
		  free(fix2DpackImg_buffer[i]);
		fix2DpackImg_buffer[i]= NULL;
	      }
	    }
	  else
	    {
	      if(fix2DinCore_diff)
		free(fix2DinCore_diff);
	      fix2DinCore_diff= NULL;
	      if(fix2DinCore_buffer)
		free(fix2DinCore_buffer);
	      fix2DinCore_buffer= NULL;
	    }
	}/* 2d */
      XtUnmanageChild(fix2Dscreen);

    }/* Destroy the data */
  
  XtUnmanageChild(statusScreenWidget);


}/* FIX2D_DO_ACTIVATE */




/*************************************************************
 * removeBlemish_activate() - Remove statistically outlying
 * data from this scanned image - maybe there's a speck of
 * dust on the micrograph when it's scanned?
 */
void removeBlemish_activate(Widget widget, int *tag, 
			    XmAnyCallbackStruct *callback_data)
{/* REMOVEBLEMISH_ACTIVATE */
  float largHalf;
  int i;

  /* Get rid of 3D rendering if it's there */
  if(XtIsManaged(mapShadeScreenWidget))
    mapShadeScreen_deactivate(0,0,0);

  sprintf(outStr,"%d",1);
  XmTextSetString(fixBegXW,outStr); 
  XmTextSetString(fixBegYW,outStr); 
  XmTextSetString(fixBegZW,outStr); 

  sprintf(outStr,"%d",largestDim);
  if(fileType==PACKEDPORTABLE)
    sprintf(outStr,"%d",purdueHeader.numCols);
  XmTextSetString(fixEndXW,outStr); 
  if(fileType==PACKEDPORTABLE)
    sprintf(outStr,"%d",purdueHeader.numRows);
  XmTextSetString(fixEndYW,outStr); 
  if(fileType==PACKEDPORTABLE)
    sprintf(outStr,"%d",purdueHeader.nimg);  /* No sections - boxes */
  XmTextSetString(fixEndZW,outStr); 

  /* Fill the corners */
  sprintf(outStr,"%d",(int)((float)largestDim*1.5));
  if(fileType==PACKEDPORTABLE)
    sprintf(outStr,"%d",(int)(purdueHeader.numRows*1.5));
  XmTextSetString(fixMaxWidget,outStr);

  XFlush(dsply);

  /* 2d */
  for(i=0;i<purdueHeader.nimg;i++)
    {
      if(fix2DpackImg_diff[i])
	free(fix2DpackImg_diff[i]);
      fix2DpackImg_diff[i] = NULL;
      if(fix2DpackImg_buffer[i])
	free(fix2DpackImg_buffer[i]);
      fix2DpackImg_buffer[i] = NULL;
    }
  if(fix2DinCore_diff)
    {free(fix2DinCore_diff);
    fix2DinCore_diff= NULL;}
  if(fix2DinCore_buffer)
    {free(fix2DinCore_buffer);
    fix2DinCore_buffer= NULL;}
  if(fileType==PACKEDPORTABLE)
    {/* packed */
      for(i=0;i<purdueHeader.nimg;i++)
	{/* nimg */
	  fix2DpackImg_diff[i] = (float *)malloc(sizeof(float)*
					       (purdueHeader.numCols*
						purdueHeader.numRows));
	  fix2DpackImg_buffer[i] = (float *)malloc(sizeof(float)*
						 (purdueHeader.numCols*
						  purdueHeader.numRows));
	  memcpy(fix2DpackImg_buffer[i],packImg[i],
		 (purdueHeader.numCols*
		  purdueHeader.numRows*sizeof(int)));
	}/* nimg */
    }/* packed */
  else
    {/* scanned */
      CK(fix2DinCore_diff = (float *)malloc(sizeof(float)*
					  file_width *
					  file_height));
      fix2DinCore_buffer = (float *)malloc(sizeof(float)*
					 file_width *
					 file_height);
      memcpy(fix2DinCore_buffer,inCore_buffer,
	     (file_width*file_height*sizeof(float)));
    }/* scanned */
  
  raiseIt(fix2Dscreen);
  
}/* REMOVEBLEMISH_ACTIVATE */





/*************************************************************
 * fix_activate() - Get fix screen up.
 */
void fix_activate(Widget widget, int *tag, 
		  XmAnyCallbackStruct *callback_data)
{/* FIX_ACTIVATE */
  float largHalf;

  /* Get rid of 3D rendering if it's there */
  if(XtIsManaged(mapShadeScreenWidget))
    mapShadeScreen_deactivate(0,0,0);

  sprintf(outStr,"%d",1);
  XmTextSetString(fixBegXW,outStr); 
  XmTextSetString(fixBegYW,outStr); 
  XmTextSetString(fixBegZW,outStr); 

  sprintf(outStr,"%d",largestDim);
  if(fileType==PACKEDPORTABLE)
    sprintf(outStr,"%d",purdueHeader.numCols);
  XmTextSetString(fixEndXW,outStr); 
  if(fileType==PACKEDPORTABLE)
    sprintf(outStr,"%d",purdueHeader.numRows);
  XmTextSetString(fixEndYW,outStr); 
  if(fileType==PACKEDPORTABLE)
    sprintf(outStr,"%d",purdueHeader.nimg);  /* No sections - boxes */
  XmTextSetString(fixEndZW,outStr); 

  /* Fill the corners */
  sprintf(outStr,"%d",(int)((float)largestDim*1.5));
  if(fileType==PACKEDPORTABLE)
    sprintf(outStr,"%d",(int)(purdueHeader.numRows*1.5));
  XmTextSetString(fixMaxWidget,outStr);

  XFlush(dsply);

  raiseIt(fixscreen);

}/* FIX_ACTIVATE */





/*************************************************************
 * fix_deactivate() - Get rid of fix screen.
 */
void fix_deactivate(Widget widget, int *tag, 
		    XmAnyCallbackStruct *callback_data)
{/* FIX_DEACTIVATE */
  int i;

  XtUnmanageChild(fixscreen);
  XtUnmanageChild(fix2Dscreen);

}/* FIX_DEACTIVATE */



/*************************************************************
 * fix2D_deactivate() - Get rid of fix2D screen.
 */
void fix2D_deactivate(Widget widget, int *tag, 
		      XmAnyCallbackStruct *callback_data)
{/* FIX2D_DEACTIVATE */
  int i;

  /* 2d */

  /* Let's reload all data in case it's corrupted from last time */
  if(fileType==PACKEDPORTABLE)
    {/* packed */
      int i;
      BYTE *thisImg;
      for(i=0;i<purdueHeader.nimg;i++)
	{
	  memcpy(packImg[i],fix2DpackImg_buffer[i],
		 (purdueHeader.numCols*
		  purdueHeader.numRows*sizeof(int)));
	}
    }/* packed */
  else
    {/* scanned */
      memcpy(inCore_buffer,fix2DinCore_buffer,
	     (file_width*file_height*sizeof(float)));
    }/* scanned */
  
  if(fileType==PACKEDPORTABLE)
    {
      for(i=0;i<purdueHeader.nimg;i++)
	{
	  if(fix2DpackImg_diff[i])
	    free(fix2DpackImg_diff[i]);
	  fix2DpackImg_diff[i] = NULL;
	  if(fix2DpackImg_buffer[i])
	    free(fix2DpackImg_buffer[i]);
	  fix2DpackImg_buffer[i] = NULL;
	}
    }
  else
    {
      if(fix2DinCore_diff)
	{free(fix2DinCore_diff);
	fix2DinCore_diff= NULL;}
      if(fix2DinCore_buffer)
	{free(fix2DinCore_buffer);
	fix2DinCore_buffer= NULL;}
    }
  
  if(fileType==PACKEDPORTABLE)
    stuffPackedImage(purdueHeader.numCols,
		     purdueHeader.numRows,
		     packImg,purdueHeader.nimg,
		     imgNum);  /* Start at radius'th image */
  else
    {/* do it */
      fullScreen = FALSE;  /* Force a recalc of full screen */
      virtual_read_activate(-99,-99);
    }/* do it */
  
  DrawWindow(); 
  
  XtUnmanageChild(fixscreen);
  XtUnmanageChild(fix2Dscreen);

}/* FIX2D_DEACTIVATE */



/* ************************************************************
 * fix_create() - Remember the fix screen for use later.  
 */
void fix_create(Widget widget, int *tag, 
		XmAnyCallbackStruct *callback_data)
{/* fix_CREATE */

  if(*tag==FIXVALUE)
    fixTextWidget = widget;
  else if(*tag==FIXMIN)
    fixMinWidget = widget;
  else if(*tag==FIXMAX)
    fixMaxWidget = widget;
  else if(*tag==TAFILL)
    taFillW = widget;
  else if(*tag==TABEGX)
    taBegXW = widget;
  else if(*tag==TABEGY)
    taBegYW = widget;
  else if(*tag==TABEGZ)
    taBegZW = widget;
  else if(*tag==TAENDX)
    taEndXW = widget;
  else if(*tag==TAENDY)
    taEndYW = widget;
  else if(*tag==TAENDZ)
    taEndZW = widget;
  else if(*tag==TACENTERX)
    taCenterXW = widget;
  else if(*tag==TACENTERY)
    taCenterYW = widget;
  else if(*tag==TACENTERZ)
    taCenterZW = widget;
  else if(*tag==FIXBEGX)
    fixBegXW = widget;
  else if(*tag==FIXBEGY)
    fixBegYW = widget;
  else if(*tag==FIXBEGZ)
    fixBegZW = widget;
  else if(*tag==FIXENDX)
    fixEndXW = widget;
  else if(*tag==FIXENDY)
    fixEndYW = widget;
  else if(*tag==FIXENDZ)
    fixEndZW = widget;
  else if(*tag==FIXGAUSSSTD)
    {
      fixGaussStdTextW = widget;
      XtUnmanageChild(widget);
    }


}/* fix_CREATE */



/* ************************************************************
 * fix_change() - Whenever text widgets changed
 * call here.
 */
void fix_change(Widget widget, int *tag, 
		XmAnyCallbackStruct *callback_data)
{/* fix_change */

  fixMode = *tag;

  if(fixMode==NOISE || fixMode==GAUSSNOISE)
    {/* Init fixValue field */
      sprintf(outStr,"1.0 STD");
      XmTextSetString(fixTextWidget,outStr); 
      XFlush(dsply);
    }/* Init fixValue field */

  if(fixMode == GAUSSNOISE)
    {
      XtManageChild(fixGaussStdLabelW);
      XtManageChild(fixGaussStdTextW);
      sprintf(outStr,"%f",purdueHeader.img_std);
      XmTextSetString(fixGaussStdTextW,outStr);
    }
  else
    {
      XtUnmanageChild(fixGaussStdLabelW);
      XtUnmanageChild(fixGaussStdTextW);
    }

}/* fix_change */





/*************************************************************
 * fix_do_activate() - Do the fix.
 */
void fix_do_activate(Widget widget, int *tag, 
		     XmAnyCallbackStruct *callback_data)
{/* FIX_DO_ACTIVATE */
  int i,j,k,xcen,ycen,zcen,lastZ;
  float fixValue;
  float minRad,maxRad,thisRad,thisPixel,saveMinRad,saveMaxRad;
  float xSq,ySq,zSq,qSq;
  char *fixText;
  float noise,minSurf,maxSurf;
  double factor;
  int begX,begY,begZ,endX,endY,endZ,tmpY,thisNumSecs;
  float *thisPackImg;
  float *tmpS;
  int icount,ncol,nrow;
  float maxF,minF,scaleF,stDev;
  float *gauss;
  
  fixText = XmTextGetString(fixTextWidget); 
  fixValue = (float)strtod(fixText,0);
  XtFree(fixText);

  fixText = XmTextGetString(fixMinWidget); 
  saveMinRad= minRad = (float)strtod(fixText,0);
  minRad *= minRad;  /* Do this squared */
  XtFree(fixText);

  fixText = XmTextGetString(fixMaxWidget); 
  saveMaxRad = maxRad = (float)strtod(fixText,0);
  maxRad = (maxRad*maxRad);  /* Do this squared */
  XtFree(fixText);

  fixText = XmTextGetString(fixBegXW); 
  begX = (int)strtod(fixText,0);
  XtFree(fixText);

  fixText = XmTextGetString(fixBegYW); 
  begY = (int)strtod(fixText,0);
  XtFree(fixText);

  fixText = XmTextGetString(fixBegZW); 
  begZ = (int)strtod(fixText,0);
  XtFree(fixText);

  fixText = XmTextGetString(fixEndXW); 
  endX = (int)strtod(fixText,0);
  XtFree(fixText);

  fixText = XmTextGetString(fixEndYW); 
  endY = (int)strtod(fixText,0);
  XtFree(fixText);

  fixText = XmTextGetString(fixEndZW); 
  endZ = (int)strtod(fixText,0);
  XtFree(fixText);

  /* Humans number from 1 not zero */
  begX--;
  begY--;
  begZ--;

  /* Always do fixValue STD */
  noise  = (float)(purdueHeader.img_std) * fixValue;   
  if(fixMode==GAUSSNOISE)
    {/* Check to see what's entered */
      fixText = XmTextGetString(fixGaussStdTextW); 
      noise = (float)strtod(fixText,0) * fixValue;
      XtFree(fixText);
    }/* Check to see what's entered */

  if(doNoise)
    {/* We're in batch mode and they've asked for a specific stDev */
      fixValue = noiseStd;
      noise    = (float)(purdueHeader.img_std) * fixValue ;
      fixMode  = noiseType;
      begX = begY = begZ = 0;
      endX = purdueHeader.numCols;
      endY = purdueHeader.numRows;
      endZ = purdueHeader.nimg;
    }/* We're in batch mode and they've asked for a specific stDev */
  printf("begEnd [%d,%d,%d,%d,%d,%d]\n",
	 begX , begY , begZ ,
	 endX ,
	 endY ,
	 endZ 
	 );
  /* Get our fraction of the random number */
  factor= 1.0 / DENOMINATOR;

  if(fileType==PACKEDPORTABLE)
    {/* DO packed boxes */
      /* Y is a special case here - Tim numbers row starting
       * from the bottom.
       */
      tmpY = endY;
      endY = purdueHeader.numRows - begY;
      begY = purdueHeader.numRows - tmpY;
      
      begX = MAX(begX,0);
      begY = MAX(begY,0);
      begZ = MAX(begZ,0);
      endX = MIN(endX,purdueHeader.numCols);
      endY = MIN(endY,purdueHeader.numRows);
      endZ = MIN(endZ,purdueHeader.nimg);
    }/* DO packed boxes */
  else if(fileType==MAP)
    {/* MAPs */
      /* Y is a special case here - Tim numbers row starting
       * from the bottom.
       */
      tmpY = endY;
      endY = numRows - begY;
      begY = numRows - tmpY;
      
      begX = MAX(begX,0);
      begY = MAX(begY,0);
      begZ = MAX(begZ,0);
      endX = MIN(endX,largestDim);
      endY = MIN(endY,largestDim);
      endZ = MIN(endZ,largestDim);
      /* We may have shaded map display where numSecs is less than 
       * largestDim so account for that here */
      thisNumSecs = numSecs;
      if(shadedMaps)
	thisNumSecs = largestDim;

      endZ = MIN(endZ,thisNumSecs);  /* We may have a PRJ file here that has
				      * reduced number of sections */
    }/* MAPs */
  else if(fileType==BYTEPORTABLE || fileType==PORTABLE)
    {/* Add noise to scan */
      begZ = 0;  /* Only one scan */
      endZ = 1;
      begX = 0;
      endX = file_width;
      begY = 0;
      endY = file_height;
      thisNumSecs = 1;
    }/* Add noise to scan */

  if(fixMode==GAUSSNOISE && 
     (fileType==PACKEDPORTABLE || fileType==BYTEPORTABLE || fileType==MAP))
    {/* Add gaussian noise to packed boxes/scans */

      if(fileType==BYTEPORTABLE)
	{/* Setup for scans */
	  calcScanStats(inCore_buffer,
			&purdueHeader.img_bkg, 
			&purdueHeader.img_min,
			&purdueHeader.img_max, 
			&purdueHeader.img_avg, 
			&purdueHeader.img_std,
			file_width, file_height, 
			file_width);
	  noise  = (float)(purdueHeader.img_std) * fixValue;   
	  stDev = (float)purdueHeader.img_std;
	  ncol  = file_width;
	  nrow  = file_height;
	  tmpS = (float *)malloc(sizeof(float) *
				 file_width*file_height);
	}/* Setup for scans */
      else if(fileType==MAP)
	{/* MAP */
	  stDev = (float)purdueHeader.img_std;
	  ncol  = numCols;
	  nrow  = numRows;
	  tmpS = (float *)malloc(sizeof(float) *
				 origNumSecs *
				 numCols *
				 numRows);
	}/* MAP */
      else
	{/* Setup for packed boxes */
	  stDev = (float)purdueHeader.img_std;
	  ncol  = purdueHeader.numCols;
	  nrow  = purdueHeader.numRows;
	  tmpS = (float *)malloc(sizeof(float) *
				 purdueHeader.nimg *
				 purdueHeader.numCols *
				 purdueHeader.numRows);
	}/* Setup for packed boxes */


      icount = 1000;
      gauss = (float *)malloc(sizeof(float) * icount);

      for(i=begZ;i<endZ;i++)
	{/* imgs */
	  /* Init */
	  for(j=0;j<nrow;j++)
	    {/* rows */
	      for(k=0;k<ncol;k++)
		{/* cols */
		  tmpS[(i*ncol*nrow)+(j*ncol)+k] = 0.0;
		}/* cols */
	    }/* rows */

	  if(fileType==PACKEDPORTABLE)
	    {/* packed images */
	      thisPackImg = packImg[i];
	      for(j=begY;j<endY;j++)
		{/* rows */
		  for(k=begX;k<endX;k++)
		    {/* cols */
		      tmpS[(i*ncol*nrow)+(j*ncol)+k] = 
			(float)thisPackImg[(j*ncol)+k] ;
		    }/* cols */
		}/* rows */
	    }/* packed images */
	  else if(fileType==MAP)
	    {/* MAP image */
	      for(j=begY;j<endY;j++)
		{/* rows */
		  for(k=begX;k<endX;k++)
		    {/* cols */
		      tmpS[(i*ncol*nrow)+(j*ncol)+k] = 
			(float)float_image_buffer1[(i*ncol*nrow)+(j*ncol)+k] ;
		    }/* cols */
		}/* rows */	      
	    }/* MAP image */
	  else
	    {/* scanned image */
	      for(j=begY;j<endY;j++)
		{/* rows */
		  for(k=begX;k<endX;k++)
		    {/* cols */
		      tmpS[(i*ncol*nrow)+(j*ncol)+k] = 
			(float)inCore_buffer[(j*ncol)+k] ;
		    }/* cols */
		}/* rows */
	    }/* scanned image */

	  gauss_load_(gauss, &icount);
	  /* for(j=0;j<icount;j++)
	     printf("gauss[%d] = %f\n",j,gauss[j]);
	  printf("noise(std) = %f\n",noise);
	  printf("gauss[%d] = %f\n",1,gauss[1]);  
	  printf("gauss[%d] = %f\n",11,gauss[11]);  
	  printf("gauss[%d] = %f\n",111,gauss[111]);  
	  printf("gauss[%d] = %f\n",555,gauss[555]);  
	  printf("gauss[%d] = %f\n",988,gauss[988]);     */
	  img_gnoise_rob_((tmpS+(i*ncol*nrow)), 
		      &ncol,
		      &nrow,
		      &noise,
		      gauss,
		      &icount
		      );
	}/* imgs */

      minF = maxF = (float)tmpS[(begZ*nrow*ncol)+(begY*ncol)+begX];
      for(i=begZ;i<endZ;i++)
	{/* imgs */
	  for(j=begY;j<endY;j++)
	    {/* rows */
	      for(k=begX;k<endX;k++)
		{/* cols */
		  minF = MIN((float)tmpS[(i*nrow*ncol)+(j*ncol)+k],minF);
		  maxF = MAX((float)tmpS[(i*nrow*ncol)+(j*ncol)+k],maxF);
		}/* cols */
	    }/* rows */
	}/* imgs */

      scaleF = ((float)tryColors - 7.0) / (maxF - minF);

      /* Now put it back */
      for(i=begZ;i<endZ;i++)
	{/* imgs */
	  
	  if(fileType==MAP)
	    {/* MAP */
	      maxMapPixel = maxF;
	      minMapPixel = minF;
	      for(j=begY;j<endY;j++)
		{/* rows */
		  for(k=begX;k<endX;k++)
		    {/* cols */
		      float_image_buffer1[(i*ncol*nrow)+(j*ncol)+k] =
			(((float)tmpS[(i*ncol*nrow)+
				      (j*ncol)+k])) ;
		    }/* cols */
		}/* rows */
	    }/* MAP */
	  else
	    {/* normal */
	      if(fileType==PACKEDPORTABLE)
		thisPackImg = packImg[i];
	      else thisPackImg = inCore_buffer;
	      for(j=begY;j<endY;j++)
		{/* rows */
		  for(k=begX;k<endX;k++)
		    {/* cols */
		      /************ OLDWAY in BYTE format
		      thisPackImg[(j*ncol)+k] =
			(((float)tmpS[(i*ncol*nrow)+
				      (j*ncol)+k] - minF) * scaleF) + 1;
		      *************/
		      thisPackImg[(j*ncol)+k] =
			tmpS[(i*ncol*nrow)+
			     (j*ncol)+k];
		    }/* cols */
		}/* rows */
	    }/* normal */
	}/* imgs */

      /* Now get min/max */
      minF = maxF = (float)tmpS[0];
      if(fileType!=MAP)
	{/* OK  */
	  for(i=0;i<purdueHeader.nimg;i++)
	    {/* imgs */
	      if(fileType==PACKEDPORTABLE)
		thisPackImg = packImg[i];
	      else thisPackImg = inCore_buffer;
	      for(j=0;j<nrow;j++)
		{/* rows */
		  for(k=0;k<ncol;k++)
		    {/* cols */
		      minF = MIN((float)thisPackImg[(j*ncol)+k],minF);
		      maxF = MAX((float)thisPackImg[(j*ncol)+k],maxF);
		    }/* cols */
		}/* rows */
	    }/* imgs */
	}/* OK  */
      printf("min,max = [%f,%f]\n",minF,maxF);

      free(tmpS);
      free(gauss);

      if(fileType==PACKEDPORTABLE)
	{/* Setup packed stuff */
	  maxPackPixel = maxF;
	  minPackPixel = minF;
	  
#define TRY_SCALING_PACKED 1
#ifdef TRY_SCALING_PACKED
	  /* Now clip this guy to our colormap */ 
	  if(minPackPixel < 0.0)
	    scalePackFactor = (((float)tryColors - 7.0) / 
			       (maxPackPixel-minPackPixel));
	  else scalePackFactor = (((float)tryColors - 7.0) / 
				  maxPackPixel);
#else
	  /* Now clip this guy to our colormap */ 
	  scalePackFactor = (((float)tryColors - 7.0) / 256.0);
#endif
	  stuffPackedImage(purdueHeader.numCols,
			   purdueHeader.numRows,
			   packImg,purdueHeader.nimg,
			   imgNum);
	}/* Setup packed stuff */
      else if(fileType==MAP)
	{/* MAp */
	  /* Make sure we have some kind of range */
	  if(maxMapPixel==minMapPixel)
	    maxMapPixel = minMapPixel + 1.0;
	  printf("MAX[%f]  MIN[%f]\n",maxMapPixel,minMapPixel);
	  secMaxPixel = maxMapPixel;
	  secMinPixel = minMapPixel;
	  
	  calcMapStats(float_image_buffer1,
		       &purdueHeader.img_bkg, 
		       &purdueHeader.map_min,
		       &purdueHeader.map_max, 
		       &purdueHeader.img_avg, 
		       &purdueHeader.img_std,
		       &numCols, &numRows, &origNumSecs);
	  stuffMapImage((float *)map_image_buffer, 
			(float *)float_image_buffer1, 
			file_width,file_height,
			nsecs,imgNum, 0);
	  DrawWindow();
	}/* MAp */
      else
	{/* Scan */
	  /* Make it generate a new fullscreen buffer */
	  fullScreen_buffer.dispx = fullScreen_buffer.dispy = 0;
	  virtual_read_activate(0,0);
	}/* Scan */

      DrawWindow(); 

      goto allDone;

    }/* Add gaussian noise to packed boxes/scans */
  else if(fixMode==GAUSSNOISE || fileType==BYTEPORTABLE)
    {/* Not yet */
      askThem("Not yet.",FALSE,XmDIALOG_INFORMATION);
      goto byPass;
    }/* Not yet */


  if(fileType==PACKEDPORTABLE || fileType==PORTABLE)
    {/* Do for packed boxes */
      float newValue;
      float tmp1,tmp2,tmp3,maxValue;
      float *thisTmpS;
      xcen =   purdueHeader.numCols/2;
      ycen =   purdueHeader.numRows/2;

      thisTmpS = (float *)malloc(sizeof(float) *
			     purdueHeader.nimg *
			     purdueHeader.numCols *
			     purdueHeader.numRows);

      for(i=begZ;i<endZ;i++)
	{/* imgs */

	  if(fileType==PORTABLE)
	    thisPackImg = inCore_buffer;
	  else
	    thisPackImg = packImg[i];

	  ncol  = purdueHeader.numCols;
	  nrow  = purdueHeader.numRows;
	  /* Init */
	  maxValue = thisPackImg[0];
	  for(j=0;j<nrow;j++)
	    {/* rows */
	      for(k=0;k<ncol;k++)
		{/* cols */
		  thisTmpS[(i*nrow*ncol)+(j*ncol)+k] = thisPackImg[(j*ncol)+k];
		  maxValue = MAX(maxValue,thisPackImg[(j*ncol)+k]);
		}/* cols */
	    }/* rows */

	  for(j=begY;j<endY;j++)
	    {/* rows */
	      ySq = (float)((j-ycen)*(j-ycen));
	      
	      for(k=begX;k<endX;k++)
		{/* cols */
		  xSq= (float)((k-xcen)*(k-xcen));
		  thisRad = (xSq+ySq);
		  if((thisRad >= minRad && thisRad <=maxRad))
		    {/* OK to change */
		      newValue = (float)thisPackImg[(j*purdueHeader.numCols)+
						    k];
		      if(fixMode==FIXADD)
			newValue +=  fixValue;
		      else if(fixMode==FIXMULT)
			newValue =  ((float)newValue * (float)fixValue);
		      else if(fixMode==FIXOVER)
			newValue =  fixValue;
		      else if(fixMode==INVERT)
			{
			  if(fileType==PORTABLE)
			    newValue = maxValue - newValue;
			  else  /* For maps??? */
			    newValue = 0.0 - newValue;
			}
		      else if(fixMode==NOISE)
			{/* Noise */
			  tmp1 = ((((float)random()*factor)-0.5)*2.0*noise);

			  newValue +=  tmp1;
			}/* Noise */
		      
		      thisTmpS[(i*nrow*ncol)+
			   (j*purdueHeader.numCols)+k]= (float)newValue;

		    }/* OK to change */
		}/* cols */
	    }/* rows */

	}/* imgs */
      
      if(fixMode==NOISE)
	{/* Noise */
	  minF = maxF = (float)thisTmpS[(begZ*nrow*ncol)+(begY*ncol)+begX];
	  for(i=begZ;i<endZ;i++)
	    {/* imgs */
	      for(j=begY;j<endY;j++)
		{/* rows */
		  for(k=begX;k<endX;k++)
		    {/* cols */
		      minF = MIN((float)thisTmpS[(i*nrow*ncol)+(j*ncol)+k],minF);
		      maxF = MAX((float)thisTmpS[(i*nrow*ncol)+(j*ncol)+k],maxF);
		    }/* cols */
		}/* rows */
	    }/* imgs */
	  
	  scaleF = ((float)tryColors - 7.0) / (maxF - minF);
	}/* Noise */

      /* Now put it back */
      for(i=begZ;i<endZ;i++)
	{/* imgs */
	  if(fileType==PORTABLE)
	    thisPackImg = inCore_buffer;
	  else
	    thisPackImg = packImg[i];
	  for(j=begY;j<endY;j++)
	    {/* rows */
	      for(k=begX;k<endX;k++)
		{/* cols */
		  if(fixMode==NOISE)
		    thisPackImg[(j*ncol)+k] =
		      (float)thisTmpS[(i*ncol*nrow)+
				      (j*ncol)+k] ;
		  else
		    thisPackImg[(j*ncol)+k] = 
		      thisTmpS[(i*ncol*nrow)+(j*ncol)+k];
		}/* cols */
	    }/* rows */
	}/* imgs */

      /* Now get min/max */
      /* Why only for noise???       if(fixMode==NOISE) */
      if(TRUE)
	{/* Only for nosie */
	  minF = maxF = (float)thisTmpS[0];
	  for(i=0;i<purdueHeader.nimg;i++)
	    {/* imgs */
	      if(fileType==PORTABLE)
		thisPackImg = inCore_buffer;
	      else
		thisPackImg = packImg[i];
	      for(j=0;j<nrow;j++)
		{/* rows */
		  for(k=0;k<ncol;k++)
		    {/* cols */
		      minF = MIN((float)thisPackImg[(j*ncol)+k],minF);
		      maxF = MAX((float)thisPackImg[(j*ncol)+k],maxF);
		    }/* cols */
		}/* rows */
	    }/* imgs */
	  
	  maxPackPixel = maxF;
	  minPackPixel = minF;
	  
#define TRY_SCALING_PACKED 1
#ifdef TRY_SCALING_PACKED
	  /* Now clip this guy to our colormap */ 
	  if(minPackPixel < 0.0)
	    scalePackFactor = (((float)tryColors - 7.0) / 
			       (maxPackPixel-minPackPixel));
	  else scalePackFactor = (((float)tryColors - 7.0) / 
				  maxPackPixel);
#else
	  /* Now clip this guy to our colormap */ 
	  scalePackFactor = (((float)tryColors - 7.0) / 256.0);
#endif
	}/* Only for nosie */


      free(thisTmpS);
      if(fileType==PORTABLE)
	{/* Do nothing */
	  fullScreen= FALSE;
	  virtual_read_activate(-99,-99);
	}/* Do nothing */
      else
	stuffPackedImage(purdueHeader.numCols,
			 purdueHeader.numRows,
			 packImg,purdueHeader.nimg,
			 imgNum);
      DrawWindow(); 

    }/* Do for packed boxes */
  else if(fileType==MAP)
    {/* Do for MAP */

      /* OLDWAY xcen =   ycen =   zcen = largestDim/2; */
      xcen = numCols/2;
      ycen = numRows/2;
      zcen = numSecs/2;

      for(i=begZ;i<endZ;i++)
	{/* secs */
	  zSq = (float)((i-zcen)*(i-zcen));
	  
	  for(j=begY;j<endY;j++)
	    {/* rows */
	      ySq = (float)((j-ycen)*(j-ycen));
	      
	      for(k=begX;k<endX;k++)
		{/* cols */
		  xSq= (float)((k-xcen)*(k-xcen));
		  thisRad = zSq + (xSq+ySq);
		  if((fixIcosFill && callXdy((double)i,(double)j,(double)k,
					      (double)saveMinRad,
					      (double)saveMaxRad)) ||
		     (thisRad >= minRad && thisRad <=maxRad))
		    {/* OK to change */
		      if(fixMode==FIXADD)
			float_image_buffer1[(
					     (i*numRows*numCols)+
					     (j*numCols)+
					     k
					     )] +=  fixValue;
		      else if(fixMode==FIXMULT)
			float_image_buffer1[(
					     (i*numRows*numCols)+
					     (j*numCols)+
					     k
					     )] *=  fixValue;
		      else if(fixMode==FIXOVER)
			float_image_buffer1[(
					     (i*numRows*numCols)+
					     (j*numCols)+
					     k
					     )] =  (float)fixValue;
		      else if(fixMode==INVERT)
			float_image_buffer1[(
					     (i*numRows*numCols)+
					     (j*numCols)+
					     k
					     )] *=  -1.0;
		      else if(fixMode==NOISE)
			{/* Noise */
			  float_image_buffer1[(
					       (i*numRows*numCols)+
					       (j*numCols)+
					       k
					       )] +=  (((random()*factor)-0.5)*
						       2.0*noise);
			}/* Noise */
		    }/* OK to change */
		  
		}/* cols */
	    }/* rows */
	}/* secs */
      
      /* Find the new min/max - we may be only
       * doing a part of the MAP here so we need to include all of
       * the ones that aren't procesed here in the min/max. */
      maxMapPixel= -99999999.9;
      minMapPixel= 99999999.9;
      maxMapPixel = -HUGE;  /* dgm -HUGE instead of -99999.9 */
      minMapPixel = HUGE;   /* dgm HUGE instead of 9999999.9 */
      
      if(thisNumSecs < numRows)
	{/* Not symmetrical e.g. may be a PRJ file */
	  lastZ = thisNumSecs;
	}/* Not symmetrical e.g. may be a PRJ file */
      else
	lastZ = largestDim;

      for(i=0;i<lastZ;i++)
	{/* secs */
	  for(j=0;j<largestDim;j++)
	    {/* rows */
	      for(k=0;k<largestDim;k++)
		{/* cols */
		  
		  thisPixel= float_image_buffer1[(
						  (i*numRows*numCols)+
						  (j*numCols)+
						  k
						  )];
		  if(thisPixel > maxMapPixel)
		    maxMapPixel= (float)thisPixel;
		  if(thisPixel < minMapPixel)
		    minMapPixel= (float)thisPixel;
		  
		}/* cols */
	    }/* rows */
	}/* secs */
       
     
      minSurf = (int)MAX(minMapPixel,0.0);
      maxSurf = MAX(maxMapPixel,(0.0-minMapPixel));
      /* We're doing the surfaceDensity in tenths now */
      if(maxSurf > 5000.0)
	{
	  XtVaSetValues(surfDensScrollWidget,XmNmaximum, 
			(int)((maxSurf*10.0)+50.0),NULL);
	  XtVaSetValues(surfDensScrollWidget,XmNminimum, 
			(int)(minSurf*10.0),NULL);
	  XtVaSetValues(surfDensScrollWidget,XmNsliderSize, 
			(int)(maxSurf*0.01),NULL);
	  XtVaSetValues(surfDensScrollWidget,XmNpageIncrement, 
			(int)(maxSurf*0.01),NULL);
	  XtVaSetValues(surfDensScrollWidget,XmNincrement, 
			(int)(maxSurf*0.002),NULL);
	  XtVaSetValues(surfDensScrollWidget,XmNvalue, (int)(10),NULL);
	  
	}
      else if(maxSurf > 500.0)
	{
	  XtVaSetValues(surfDensScrollWidget,XmNmaximum, 
			(int)((maxSurf*10.0)+50.0),NULL);
	  XtVaSetValues(surfDensScrollWidget,XmNminimum, 
			(int)(minSurf*10.0),NULL);
	  XtVaSetValues(surfDensScrollWidget,XmNsliderSize, (int)(500),NULL);
	  XtVaSetValues(surfDensScrollWidget,XmNpageIncrement, (int)500,NULL);
	  XtVaSetValues(surfDensScrollWidget,XmNincrement, (int)100,	NULL);
	  XtVaSetValues(surfDensScrollWidget,XmNvalue, (int)(10),NULL);
	  
	}
      else
	{
	  XtVaSetValues(surfDensScrollWidget,XmNmaximum, 
			(int)((maxSurf*10.0)+50.0),NULL);
	  XtVaSetValues(surfDensScrollWidget,XmNminimum, 
			(int)(minSurf*10.0),NULL);
	  XtVaSetValues(surfDensScrollWidget,XmNsliderSize, (int)(50),NULL);
	  XtVaSetValues(surfDensScrollWidget,XmNpageIncrement, (int)50,NULL);
	  XtVaSetValues(surfDensScrollWidget,XmNincrement, (int)10,	NULL);
	  XtVaSetValues(surfDensScrollWidget,XmNvalue, (int)(10),NULL);
	  
	}
      
      /* Their shaded views are no longer valid - show them 
       * their new sections */
      shadedMaps= FALSE;
      dispSections = TRUE;
      XtSetArg(arglst[0], XmNlabelPixmap, sectionPixmap);
      XtSetArg(arglst[1], XmNselectPixmap, unsectionPixmap);
      XtSetArg(arglst[2], XmNrecomputeSize, FALSE);
      XtSetValues(sectionWidget, arglst, 3); 
      nsecs = origNumSecs;
      if(dispSections)
	{/* Update the sections - if shaded views then they have to recalc */
	  /* Now stuff the display buffer from our new FIXed map */
	  stuffMapImage((float *)map_image_buffer, 
			(float *)float_image_buffer1, 
			file_width,file_height,
			nsecs,imgNum, 0);
	  DrawWindow();
	}/* Update the sections - if shaded views then they have to recalc */
      
    }/* Do for MAP */
  else
    {/* not yet */
      askThem("FIX not yet implemented for this file type.",
	      FALSE,XmDIALOG_INFORMATION);
      goto byPass;
    }/* not yet */

 allDone:;

  if(fixMode==FIXADD)
    sprintf(outStr,
	    "Successfully %s %f",
	    "Added",
	    fixValue);
  else if(fixMode==FIXMULT)
    sprintf(outStr,
	    "Successfully %s %f",
	    "Multiplied by",
	    fixValue);
  else if(fixMode==FIXOVER)
    sprintf(outStr,
	    "Successfully %s %f",
	    "Overwrote with",
	    fixValue);
  else if(fixMode==INVERT)
    sprintf(outStr,
	    "Successfully INVERTED voxels");
  else if(fixMode==NOISE || fixMode==GAUSSNOISE)
    {sprintf(outStr,
	    "Successfully added %.1f STD of noise",fixValue);
    printf("%s\n",outStr);
    }
  
  
  /*   XtUnmanageChild(fixscreen); */
  if(!doNoise)
    askThem(outStr,FALSE,XmDIALOG_INFORMATION);
  
  byPass:;

}/* FIX_DO_ACTIVATE */




/*************************************************************
 * fftMeasWrite_activate() - Write to disk the info on
 * this line of the FFT.
 */
void fftMeasWrite_activate(Widget widget, int *tag, 
			   XmAnyCallbackStruct *callback_data)
{/* fftMeasWrite_ACTIVATE */
  int i,j,k,thisPixel,fftWidth;
  FILE *fOut;

  if(fftImage_gold != NULL)
    {/* Write file */

      fOut= fopen("fftline.dat","w");
      
      fftMeasLineNum = MAX(MIN(fftMeasLineNum,511),0);
      fprintf(fOut,"\n\nLine number [%d] of FFT obtained from\n",
	      fftMeasLineNum);
      if(averageFFT)
	fprintf(fOut,"Img [Average] of [%s]\n",fileName.addr);
      else fprintf(fOut,"Img [%d] of [%s]\n\n",imgNum,fileName.addr);
      
      fprintf(fOut,"A Part        \tB Part\n");
      fprintf(fOut,"--------------\t---------------\n");
      fftWidth= 512; 
      for(i=fftMeasLineNum;i<(int)(fftMeasLineNum+1);i++)
	for(j=0;j<(int)(fftWidth*2);j+=2)
	  {/* Do all image */
	    float thisValue;
	    thisPixel = (int)(i*(fftWidth*2))+j;
	    fprintf(fOut,"%f\t%f\n",
		    fftImage_gold[(thisPixel)],
		    fftImage_gold[(thisPixel+1)]
		    );

	  }/* Do all image */

      fclose(fOut);

      askThem("File fftline.dat written successfully.",
	      FALSE,XmDIALOG_INFORMATION);

    }/* Write file */
  else
    askThem("No FFT to write!",FALSE,XmDIALOG_INFORMATION);


}/* fftMeasWrite_ACTIVATE */




/*************************************************************
 * point_color_activate() - Change the color of the point screen
 * label. Corresponds to how drawing is done.
 */
void point_color_activate(Widget widget, int *tag, 
			  XmAnyCallbackStruct *callback_data)
{/* POINT_COLOR_ACTIVATE */

  pointColor++;
  if(pointColor > 2)
    pointColor = 0;

  if(pointColor==0)
    {/* red */
      XtSetArg(arglst[0], XmNforeground, pixels_return[RED] );
      XtSetValues(widget, arglst, 1); 
    }/* red */
  else if(pointColor==1)
    {/* green */
      XtSetArg(arglst[0], XmNforeground, pixels_return[GREEN] );
      XtSetValues(widget, arglst, 1); 
    }/* green */
  else if(pointColor==2)
    {/* blue */
      XtSetArg(arglst[0], XmNforeground, pixels_return[BLUE] );
      XtSetValues(widget, arglst, 1); 
    }/* blue */

}/* POINT_COLOR_ACTIVATE */



/*************************************************************
 * point_deactivate() - Get rid of point screen.
 */
void point_deactivate(Widget widget, int *tag, 
		      XmAnyCallbackStruct *callback_data)
{/* POINT_DEACTIVATE */

  XtUnmanageChild(pointscreen);

}/* POINT_DEACTIVATE */



/* ************************************************************
 * point_create() - Remember the point screen for use later.  
 */
void point_create(Widget widget, int *tag, 
		  XmAnyCallbackStruct *callback_data)
{/* point_CREATE */

  pointLabelWidget[*tag] = widget;
  
}/* point_CREATE */



/*******************************************************************
 * drawPointScreen() - Draw this string on the pointscreen.
*/
void drawPointScreen(char *thisStr, int pos)
{/*  DRAWPOINTSCREEN */
  XmString simpleString;

  createSimpleString(&simpleString, thisStr);
  XtSetArg(arglst[0], XmNlabelString, simpleString);
  XtSetArg(arglst[1], XmNrecomputeSize, FALSE );
  XtSetArg(arglst[2], XmNalignment, XmALIGNMENT_BEGINNING);
  XtSetValues(pointLabelWidget[pos], arglst, 3); 

  XFlush(dsply);
  XmStringFree(simpleString); 


}/*  DRAWPOINTSCREEN */





/*******************************************************************
 * findRealValue() - Find out what the "real" value was for this
 * pixel in the disk file.
*/
float findRealValue(int posX, int posY,
		    int *thisImg,
		    int *sectY,
		    int *sectX)
{/*  FINDREALVALUE */
  int unZoomedX, unZoomedY, nrows,ncols;
  float returnVal;

  returnVal = 0.0;
  calcSizeFactor(zoomFactor);
  unZoomedX = posX * recipSizeFactor;
  unZoomedY = posY * recipSizeFactor;
  
  nrows= purdueHeader.numRows;
  ncols= purdueHeader.numCols;

  if(fileType==EMPORTABLE || fileType==BYTEPORTABLE || fileType==PORTABLE)
    {/* Scanned image */
      if(posY < image_height && posX < image_width)
	returnVal = (float)int_image_buffer1[((posY*image_width)+posX)];
      *sectX = virXoffset+(unZoomedX);
      *sectY = virYoffset+(unZoomedY);
      *thisImg = 0;
    }/* Scanned image */
  else if(fileType==PACKEDPORTABLE)
    {/* Packed boxes */
      float *thisPckImg;
      int sectRow,sectCol,thisSection,sectOffsetX,sectOffsetY;
      sectRow = posY / (int)(nrows*sizeFactor);
      sectCol = posX / (int)(ncols*sizeFactor);
      thisSection = packStartImage + (sectRow*packSectionsPerRow) + sectCol;
      sectOffsetY = posY % (int)(nrows*sizeFactor);
      sectOffsetX = posX % (int)(ncols*sizeFactor);
      sectOffsetY *= recipSizeFactor;
      sectOffsetX *= recipSizeFactor;
      *thisImg = 0;
      *sectY   = 0;
      *sectX   = 0;
      if(thisSection < purdueHeader.nimg)
	{/* OK to do */
	  *thisImg = thisSection + 1;
	  *sectY   = purdueHeader.numRows - sectOffsetY;
	  *sectX   = sectOffsetX + 1;
	  thisPckImg = packImg[thisSection];
	  returnVal = (float)thisPckImg[((sectOffsetY*ncols)+
					 sectOffsetX
					 )];
	}/* OK to do */
    }/* Packed boxes */
  else if(fileType==MAP)
    {/* MAP */
      int sectRow,sectCol,thisSection,sectOffsetX,sectOffsetY;
      sectRow = posY / (int)(nrows*sizeFactor);
      sectCol = posX / (int)(ncols*sizeFactor);
      thisSection = mapStartImage + (sectRow*mapSectionsPerRow) + sectCol;
      sectOffsetY = posY % (int)(nrows*sizeFactor);
      sectOffsetX = posX % (int)(ncols*sizeFactor);
      sectOffsetY *= recipSizeFactor;
      sectOffsetX *= recipSizeFactor;
      *thisImg = 0;
      *sectY   = 0;
      *sectX   = 0;
      if(thisSection < nsecs)
	{/* OK to do */

	  /* Put coords in TSB space */
	  *thisImg = thisSection + 1;
	  /* TSB orient */
	  *sectY   = purdueHeader.numRows - sectOffsetY;
	  *sectX   = sectOffsetX + 1;

	  if(shadedMaps && !dispSections)
	    returnVal = (float)map_image_buffer[((thisSection*nrows*ncols)+
						 (sectOffsetY*ncols)+
						 sectOffsetX
						 )];
	  else
	    returnVal = (float)float_image_buffer1[((thisSection*nrows*ncols)+
						    (sectOffsetY*ncols)+
						    sectOffsetX
						    )];
	}/* OK to do */
    }/* MAP */

  return(returnVal);

}/*  FINDREALVALUE */





/*************************************************************
 * doTheZoom() - Do what's necessary to accommodate their 
 * zooming on the main screen.  This rtn is called from 
 * main_zoom_activate() and button_zoom_activate().
*/
void doTheZoom(Widget widget, float speedFactor)
{/* DOTHEZOOM */
  float thisSizeFactor;
  int newX,newY,reason;

  calcSizeFactor(zoomFactor);
  if(widget==zoomArrowUpWidget)
    thisSizeFactor = (float)sizeFactor + (0.01*speedFactor);
  else if(widget==zoomArrowDownWidget)
    thisSizeFactor = (float)sizeFactor - (0.01*speedFactor);
  else thisSizeFactor = (float)sizeFactor;
  
  /* I need a log base 2 routine "log2()" that will allow
   * me to convert the input gotten above into meaningful
   * zoomFactors. 
   * e.g.
   *
   *      zoomFactor = (float)log2(thisSizeFactor);
   *
   */
  zoomFactor = (float)rwalog2(thisSizeFactor);
  
  zoomFactor = MAX(-4.0,MIN(zoomFactor,4.0));
  
  sprintf(outStr,"%.2f",thisSizeFactor);
  XmTextSetString(zoomTextW,outStr);
  
  /* Now redraw with new zoominess */
  fullScreen= FALSE;
  newX = (display_width / 2);
  newY = (display_height / 2);
  virtual_read_activate(newX,newY);
  
}/* DOTHEZOOM */






/*************************************************************
 * main_zoom_activate() - They want to change the current
 * displayed box by using the arrows.  This routine is called
 * when the first mouse button is used and causes the XmNactivate
 * callback.
*/
void main_zoom_activate(Widget widget, int *tag, 
			XmPushButtonCallbackStruct *callback_data)
{/* MAIN_ZOOM_ACTIVATE */
  float speedFactor;

  if(inited)
    {/* Only do if it exists */

      if(widget==boxinitArrowUpWidget ||
	 widget==boxinitArrowDownWidget )
	{/* box init */
	  speedFactor = 1.0;
	  
	  if(widget==boxinitArrowDownWidget)
	    speedFactor *= -1.0;
	  
	  doTheBoxInit(widget, speedFactor);
	  
	}/* box init */
      else
	{/* Zoom */
	  
	  doTheZoom(widget, 1.0);
	  
	}/* Zoom */
      
    }/* Only do if it exists */

}/* MAIN_ZOOM_ACTIVATE */







/*************************************************************
 * button_zoom_activate() - They want to change the current
 * displayed box by using the arrows.
*/
void button_zoom_activate(Widget widget, XEvent *event)
{/* BUTTON_ZOOM_ACTIVATE */
  float speedFactor;
  XEvent thisEvent;

  if(inited)
    {/* Only do if it exists */

      thisEvent = *event;

      /* Get which button */
      if((thisEvent.xbutton.button == Button1))
	speedFactor = 1.0;;
      if((thisEvent.xbutton.button == Button2))
	speedFactor = 2.0;
      if((thisEvent.xbutton.button == Button3))
	speedFactor = 5.0;

      doTheZoom(widget, speedFactor);

    }/* Only do if it exists */


}/* BUTTON_ZOOM_ACTIVATE */



/*-----------------------------------*/
/*************************************
 * Radius Callbacks                  *
 *************************************/
/*-----------------------------------*/
/* ************************************************************
 * radiusText_create() - Remeber the radius text widget number
 * for use later.  Init the widget to a value of 1.
 */
void radiusText_create(Widget widget, int *tag, 
		       XmAnyCallbackStruct *callback_data)
{/* radiusTEXT_CREATE */
  
  if(*tag==BOXRADIUSBUTTON)
    {/* boxRadius */
      /* Remember the widget number for later */
      boxRadiusTextWidget= widget;
      XmTextSetString(boxRadiusTextWidget, "0"); 
      handle_cr(boxRadiusTextWidget);
    }/* boxRadius */
  else if (*tag==BOXRADIUSDOWN)
    {/* radius DOWN */
      /* Remember the widget number for later */
      boxRadiusArrowDownWidget= widget;
      handle_cr(boxRadiusArrowDownWidget);
    }/* radius DOWN */
  else if (*tag==BOXRADIUSUP)
    {/* radius UP */
      /* Remember the widget number for later */
      boxRadiusArrowUpWidget= widget;
      handle_cr(boxRadiusArrowUpWidget);
    }/* radius UP */
  else if(*tag==EMPTYSHELLBUTTON)
    {/* emptyshell */
      /* Remember the widget number for later */
      emptyShellTextWidget= widget;
      XmTextSetString(emptyShellTextWidget, "5"); 
      handle_cr(emptyShellTextWidget);
    }/* emptyshell */
  else if (*tag==EMPTYSHELLDOWN)
    {/* radius DOWN */
      /* Remember the widget number for later */
      emptyShellArrowDownWidget= widget;
      handle_cr(emptyShellArrowDownWidget);
    }/* radius DOWN */
  else if (*tag==EMPTYSHELLUP)
    {/* radius UP */
      /* Remember the widget number for later */
      emptyShellArrowUpWidget= widget;
      handle_cr(emptyShellArrowUpWidget);
    }/* radius UP */
  else if(*tag==EMMARADIUSBUTTON)
    {/* emmaRadius */
      /* Remember the widget number for later */
      emmaRadiusTextWidget= widget;
      XmTextSetString(emmaRadiusTextWidget, "45"); 
      handle_cr(emmaRadiusTextWidget);
    }/* emmaRadius */
  else if (*tag==EMMARADIUSDOWN)
    {/* radius DOWN */
      /* Remember the widget number for later */
      emmaRadiusArrowDownWidget= widget;
      handle_cr(emmaRadiusArrowDownWidget);
    }/* radius DOWN */
  else if (*tag==EMMARADIUSUP)
    {/* radius UP */
      /* Remember the widget number for later */
      emmaRadiusArrowUpWidget= widget;
      handle_cr(emmaRadiusArrowUpWidget);
    }/* radius UP */
  else if(*tag==FEATHERRADIUSBUTTON)
    {/* featherRadius */
      /* Remember the widget number for later */
      featherRadiusTextWidget= widget;
      XmTextSetString(featherRadiusTextWidget, "101"); 
      handle_cr(featherRadiusTextWidget);
    }/* featherRadius */
  else if (*tag==FEATHERRADIUSDOWN)
    {/* radius DOWN */
      /* Remember the widget number for later */
      featherRadiusArrowDownWidget= widget;
      handle_cr(featherRadiusArrowDownWidget);
    }/* radius DOWN */
  else if (*tag==FEATHERRADIUSUP)
    {/* radius UP */
      /* Remember the widget number for later */
      featherRadiusArrowUpWidget= widget;
      handle_cr(featherRadiusArrowUpWidget);
    }/* radius UP */
  else if (*tag==CIRCLEWIDTH)
    {/* circle width */
      /* Remember the widget number for later */
      circleWidthToggleW= widget;
    }/* circle width */
  else if (*tag==DARKREF)
    {/* darkref */
      /* Remember the widget number for later */
      darkRefToggleW= widget;
    }/* darkref */
  else if (*tag==SHOWBOXINGSTATS)
    {/* showboxingstats */
      /* Remember the widget number for later */
      showBoxingStatsToggleW= widget;
    }/* showboxingstats */
  else if (*tag==SHOWDEBUG)
    {/* showDebug */
      /* Remember the widget number for later */
      showDebugToggleW= widget;
    }/* showDebug */
  else if (*tag==SHOWCC)
    {/* showcc */
      /* Remember the widget number for later */
      showccToggleW= widget;
    }/* showcc */
  else if (*tag==BOXSTOP)
    {/* box stop */
      /* Remember the widget number for later */
      boxStopWidget= widget;
    }/* box stop */
  else if (*tag==VERSIONW)
    {/* version text */
      versionTextWidget = widget;
      handle_cr(versionTextWidget);
    }/* version text */
  else if (*tag==VERSIONDOWN)
    {/* version downarrowtext */
      versionDownWidget = widget;
    }/* version downarrowtext */
  else if (*tag==VERSIONUP)
    {/* version uparrowtext */
      versionUpWidget = widget;
    }/* version uparrowtext */
  else if (*tag==AMMAVGDENS)
    {/* version uparrowtext */
      autoMinMaxLabelW = widget;
    }/* version uparrowtext */
  else if (*tag==REFALL)
    {/* select all reference for autoboxing */
      refAllW = widget;
    }/* select all reference for autoboxing */

}/* radiusTEXT_CREATE */



/* ************************************************************
 * radiusText_changed() - When they change the radiusText widget
 * they need to press the "radius" button next to it to get
 * the change instantiated.  We change the radius here but they
 * need to call the radius_activate() callback (called when the
 * button is pressed) to get it displayed.
 */
void radiusText_changed(Widget widget, int *tag, 
			XmAnyCallbackStruct *callback_data)
{/* radiusTEXT_CHANGED */
  char *radiusText;

  if(inited && !locked)
    {/* Can't do if we're locked out */
      if(*tag==BOXRADIUSBUTTON)
	{/* radius */
	  radiusText = XmTextGetString(boxRadiusTextWidget); 
	  
	  boxRadius = featherRadius - (int)strtod(radiusText, 0);
	  if(boxRadius < 1)
	    boxRadius = 1;

	  XtFree(radiusText);
	  XFlush(dsply);

	}/* radius */
      else if(*tag==EMPTYSHELLBUTTON)
	{/* radius */
	  radiusText = XmTextGetString(emptyShellTextWidget); 
	  
	  emptyShellWidth = (int)strtod(radiusText, 0);
	  if(emptyShellWidth < 1)
	    emptyShellWidth = 1;
	  if(emptyShellWidth >= emmaRadius)
	    emptyShellWidth = emmaRadius - 1;
	  XtFree(radiusText);
	  XFlush(dsply);

	}/* radius */
      else if(*tag==EMMARADIUSBUTTON)
	{/* radius */
	  radiusText = XmTextGetString(emmaRadiusTextWidget); 
	  
	  emmaRadius = (int)strtod(radiusText, 0);
	  if(emmaRadius < 1)
	    emmaRadius = 1;
	  if(emmaRadius < 1)
	    emmaRadius = 1;
	  if(emmaRadius >= featherRadius)
	    {
	      boxRadius += (emmaRadius+1)-featherRadius;
	      featherRadius = emmaRadius+1;
	    }

	  XtFree(radiusText);
	  XFlush(dsply);

	}/* radius */
      else if(*tag==FEATHERRADIUSBUTTON)
	{/* radius */
	  int oldSize;
	  oldSize = featherRadius - boxRadius;	    
	  radiusText = XmTextGetString(featherRadiusTextWidget); 
	  
	  featherRadius = (int)strtod(radiusText, 0);
	  featherRadius = (featherRadius -1)/2;
	  if(featherRadius < 1)
	    featherRadius = 1;
	  boxRadius = featherRadius - oldSize;

	  XtFree(radiusText);
	  XFlush(dsply);

	}/* radius */
      else
	{/* radius */
	  printf("Unknown radius mode [%d]\n",*tag);
	}/* radius */

    }/* Can't do if we're locked out */

  DrawWindow(); 

}/* radiusTEXT_CHANGED */




/* ************************************************************
 * doTheRadius() - Called from radiusArrowActivate and 
 * radiusArrowActivateButton().  Do the work required to change
 * the radius concept.
 */
void doTheRadius(Widget widget, float speedFactor)
{/* DOTHERADIUS */
  char newVal[22];
  int oldSize;

  oldSize = featherRadius - boxRadius;
  if(widget==boxRadiusArrowDownWidget)
    {/* Up */
      if((boxRadius+(int)speedFactor) <= featherRadius)
	boxRadius += (int)speedFactor;
      else
	boxRadius = featherRadius;
      if(boxRadius < 1)
	boxRadius = 1;
    }/* Up */
  else if(widget==boxRadiusArrowUpWidget)
    {/* Dopwn */
      boxRadius -= (int)speedFactor;
      if(boxRadius < 1)
	boxRadius = 1;
    }/* Down */
  else if(widget==emptyShellArrowUpWidget)
    {/* Up */
      emptyShellWidth += (int)speedFactor;
      if(emptyShellWidth < 1)
	emptyShellWidth = 1;
      if(emptyShellWidth >= emmaRadius)
	emptyShellWidth = emmaRadius-1;
    }/* Up */
  else if(widget==emptyShellArrowDownWidget)
    {/* Dopwn */
      emptyShellWidth -= (int)speedFactor;
      if(emptyShellWidth < 1)
	emptyShellWidth = 1;
    }/* Down */
  else if(widget==emmaRadiusArrowUpWidget)
    {/* Up */
      emmaRadius += (int)speedFactor;
      if(emmaRadius < 1)
	emmaRadius = 1;
      if(emmaRadius >= featherRadius)
	{/* Inc it */
	  featherRadius = emmaRadius + 1;
	  boxRadius = featherRadius - oldSize;
	}/* Inc it */
    }/* Up */
  else if(widget==emmaRadiusArrowDownWidget)
    {/* Dopwn */
      emmaRadius -= (int)speedFactor;
      if(emmaRadius < 1)
	emmaRadius = 1;
    }/* Down */
  else if(widget==featherRadiusArrowUpWidget)
    {/* Up */
      featherRadius += (int)speedFactor;
      if(featherRadius < 1)
	featherRadius = 1;
      boxRadius = featherRadius - oldSize;
    }/* Up */
  else if(widget==featherRadiusArrowDownWidget)
    {/* Dopwn */
      featherRadius -= (int)speedFactor;
      if(featherRadius < 1)
	featherRadius = 1;
      if(featherRadius < emmaRadius)
	featherRadius = emmaRadius;
      boxRadius = featherRadius - oldSize;
    }/* Down */
  else printf("radius unchanged.\n");
  
  sprintf(newVal,"%d",featherRadius-boxRadius);
  XmTextSetString(boxRadiusTextWidget, newVal);
  sprintf(newVal,"%d",emptyShellWidth);
  XmTextSetString(emptyShellTextWidget, newVal);
  sprintf(newVal,"%d",emmaRadius);
  XmTextSetString(emmaRadiusTextWidget, newVal);
  sprintf(newVal,"%d",((2*featherRadius)+1));
  XmTextSetString(featherRadiusTextWidget, newVal);
  DrawWindow(); 

}/* DOTHERADIUS */




/* ************************************************************
 * radiusArrowActivate() - Instead of making them type the 
 * new radius just let them clikc the arrow.  Change the radius
 * accordingly.  This rtn is for the first mouse button only.
 */
void radiusArrowActivate(Widget widget, int *tag, 
			 XmPushButtonCallbackStruct *callback_data)
{/* radiusARROWACTIVATE */
  float speedFactor;

  if(!locked)
    {/* Can't do if we're locked out */
      
      speedFactor = 1.0;

      doTheRadius(widget, speedFactor);

    }/* Can't do if we're locked out */

}/* radiusARROWACTIVATE */




/* ************************************************************
 * radiusArrowActivateButton() - Instead of making them type the 
 * new radius just let them clikc the arrow.  Change the radius
 * accordingly.  This rtn is for the second and third mouse buttons.
 */
void radiusArrowActivateButton(Widget widget, XEvent *event)
{/* radiusARROWACTIVATEBUTTON */
  XEvent thisEvent;
  float speedFactor;

  if(!locked)
    {/* Can't do if we're locked out */
      
      thisEvent= *(event);

      /* Get which button */
      if((thisEvent.xbutton.button == Button1))
	speedFactor = 1.0;;
      if((thisEvent.xbutton.button == Button2))
	speedFactor = 2.0;
      if((thisEvent.xbutton.button == Button3))
	speedFactor = 5.0;

      doTheRadius(widget, speedFactor);

    }/* Can't do if we're locked out */

}/* radiusARROWACTIVATEBUTTON */




/* ************************************************************
 * appod_create() - remember widgets from create appodize screen.
*/
void appod_create(Widget widget, int *tag, 
		  XmAnyCallbackStruct *callback_data)
{/* APPOD_CREATE */
  
  /* Remember the widget number for later */
  if(*tag==BOXRADIUSBUTTON)
    appodBoxRadW= widget;
  else if(*tag==FEATHERRADIUSBUTTON)
    appodFeaRadW= widget;
  
}/* APPOD_CREATE */




/* ************************************************************
 * appodize_deactivate() - Get the appodizing screen
 * off the display.
 */
void appodize_deactivate(Widget widget, int *tag, 
			 XmAnyCallbackStruct *callback_data)
{/* APPODIZE_DEACTIVATE */

  XtUnmanageChild(appodizeScreenWidget);
  XtUnmanageChild(statusScreenWidget);

}/* APPODIZE_DEACTIVATE */





/* ************************************************************
 * appodize_activate() - Get appodize screen up.
 */
void appodize_activate(Widget widget, int *tag, 
		       XmAnyCallbackStruct *callback_data)
{/* APPODIZE_ACTIVATE */
  int outHeight, outWidth ;  
  char *thisStr;

  XtUnmanageChild(appodizeScreenWidget);
  raiseIt(statusScreenWidget);
  XFlush(dsply);
  XSync(dsply,0);
  
  thisStr = XmTextGetString(appodFeaRadW);
  featherRadius = (int)strtod(thisStr,0);
  free(thisStr);
  /* This is now border Width */
  thisStr = XmTextGetString(appodBoxRadW);
  boxRadius = featherRadius - (int)strtod(thisStr,0);
  if(emmaRadius >= (boxRadius-1))
    emmaRadius = boxRadius-2;  /* Make sure we don't get thresholded out with
				* the emma radius */
  free(thisStr);

  /* Now setup boxing screen for new values */
  sprintf(outStr,"%d",((2*featherRadius)+1));
  XmTextSetString(featherRadiusTextWidget, outStr); 
  sprintf(outStr,"%d",featherRadius-boxRadius);
  XmTextSetString(boxRadiusTextWidget, outStr); 

  if(appodizing == CIRCLEPACKED)
    {/* Write out circvle packed images */
      /* Calc our dimensions */
      outHeight= outWidth = (boxRadius*2)+1;
      /* Shouldn't we use the featherRadius here to determine
       * outWidth and outHeight for write_pif() ??? */
      outHeight= outWidth = (featherRadius*2)+1;
      write_pif(outWidth, outHeight, fileName.addr, CIRCLEPACKED);
    }/* Write out circvle packed images */
  else if(appodizing == PACKEDPORTABLE)
    {/* Write out already packed boxes */
      /* Calc our dimensions */
      if(packRadius <= 0)
	outHeight= outWidth = purdueHeader.numCols;
      else outHeight= outWidth = (packRadius*2)+1;	
      if(purdueHeader.numCols != purdueHeader.numRows)
	{outHeight = purdueHeader.numRows;
	 outWidth  = purdueHeader.numCols;}
      write_pif(outWidth, outHeight, fileName.addr, PACKEDPORTABLE);
    }/* Write out already packed boxes */
  else if(appodizing == WRITEFFT)
    {/* Write out from FFT screen */
      ctf_all_activate();
    }/* Write out from FFT screen */

  XtUnmanageChild(statusScreenWidget);

}/* APPODIZE_ACTIVATE */




/* ************************************************************
 * boxcolor_change() - Get the boxing color.
 */
void boxcolor_change(Widget widget, int *tag, 
		     XmAnyCallbackStruct *callback_data)
{/* BOXCOLOR_CHANGE */

  if(boxing_GC == image_gcGREEN)
    {
      boxing_GC = image_gcRED;
      XtSetArg(arglst[0], XmNforeground, pixels_return[RED] );
      XtSetValues(widget, arglst, 1); 
    }
  else if(boxing_GC == image_gcRED)
    {
      boxing_GC = image_gcBLUE;
      XtSetArg(arglst[0], XmNforeground, pixels_return[BLUE] );
      XtSetValues(widget, arglst, 1); 
    }
  else if(boxing_GC == image_gcBLUE)
    {
      boxing_GC = image_gcYELLOW;
      XtSetArg(arglst[0], XmNforeground, pixels_return[YELLOW] );
      XtSetValues(widget, arglst, 1); 
    }
  else if(boxing_GC == image_gcYELLOW)
    {
      boxing_GC = image_gcWHITE;
      XtSetArg(arglst[0], XmNforeground, pixels_return[WHITE] );
      XtSetValues(widget, arglst, 1); 
    }
  else if(boxing_GC == image_gcWHITE)
    {
      boxing_GC = image_gcBLACK;
      XtSetArg(arglst[0], XmNforeground, pixels_return[BLACK] );
      XtSetValues(widget, arglst, 1); 
    }
  else if(boxing_GC == image_gcBLACK)
    {
      boxing_GC = image_gcPLACEHOLDER;
      XtSetArg(arglst[0], XmNforeground, pixels_return[PLACEHOLDER] );
      XtSetValues(widget, arglst, 1); 
    }
  else if(boxing_GC == image_gcPLACEHOLDER)
    {
      boxing_GC = image_gcGREEN;
      XtSetArg(arglst[0], XmNforeground, pixels_return[GREEN] );
      XtSetValues(widget, arglst, 1); 
    }
  else 
    {
      boxing_GC = image_gcGREEN;
      XtSetArg(arglst[0], XmNforeground, pixels_return[GREEN] );
      XtSetValues(widget, arglst, 1); 
    }

  DrawWindow();

}/* BOXCOLOR_CHANGE */




  
/* ************************************************************
 * box_deactivate() - Get the boxing screen
 * off the display.
 */
void box_deactivate(Widget widget, int *tag, 
		    XmAnyCallbackStruct *callback_data)
{/* BOX_DEACTIVATE */
  
  XtUnmanageChild(boxScreenWidget);
  
}/* BOX_DEACTIVATE */




/*************************************************************
 * box_change() - They want to change box circle size.
 * 
*/
void box_change(Widget widget, int *tag, 
		XmScrollBarCallbackStruct *callback_data)
{/* BOX_CHANGE */
  XmString simpleString;

  if(*tag==CIRCLEWIDTH)
    {/* toggling circleWidth */
      dispThickCircles = !dispThickCircles;
      if(dispThickCircles)
	sprintf(outStr,"Draw Thin Circles"); 
      else sprintf(outStr,"Draw Thick Circles"); 
      
      createSimpleString(&simpleString,outStr);
      XtSetArg(arglst[0], XmNlabelString, simpleString);
      XtSetValues(circleWidthToggleW, arglst, 1); 
      XFlush(dsply);
      XmStringFree(simpleString); 
      XFlush(dsply);
    }/* toggling circleWidth */
  else if(*tag==CYCLEREF)
    {/* toggling cycleref */
      cycleRef = !cycleRef;
    }/* toggling cycleref */
  else if(*tag==DARKREF)
    {/* toggling darkRef */
      darkCircleRef = !darkCircleRef;
      if(darkCircleRef)
	{/* Thsi guy is OK now */
	  refCircAvg = FALSE;
	  XtSetArg(arglst[0], XmNset, FALSE);
	  XtSetValues(refCircAvgW, arglst, 1);

	  XtSetArg(arglst[0], XmNsensitive, FALSE);
	  XtSetValues(refAllW, arglst, 1);  
	  XFlush(dsply);
	}/* Thsi guy is OK now */
      else
	{/* not dark circle */
	  XtSetArg(arglst[0], XmNsensitive, TRUE);
	  XtSetValues(refAllW, arglst, 1);  
	  XFlush(dsply);
	}/* not dark circle */
    }/* toggling darkRef */
  else if(*tag==SHOWDEBUG)
    {/* toggling showdebug */
      showDebug = !showDebug;
    }/* toggling showdebug */
  else if(*tag==SHOWBOXINGSTATS)
    {/* toggling showboxingstats */
      showBoxingSwitch = !showBoxingSwitch;
    }/* toggling showboxingstats */
  else if(*tag==SHOWCC)
    {/* toggling showcc */
      showCC = !showCC;
      if(showCC)
	sprintf(outStr,"PointMode Box Finder"); 
      else sprintf(outStr,"NO PointMode Box Finder"); 
      
      createSimpleString(&simpleString,outStr);
      XtSetArg(arglst[0], XmNlabelString, simpleString);
      /* XtSetValues(showccToggleW, arglst, 1);  */
      XFlush(dsply);
      XmStringFree(simpleString); 
      XFlush(dsply);
    }/* toggling showcc */
  else if (*tag==BOXRADIUSBUTTON)
    {/* Toggling box radius display */
      boxRadiusDisp = !boxRadiusDisp;
    }/* Toggling box radius display */
  else if (*tag==EMPTYSHELLBUTTON)
    {/* Toggling empty radius display */
      emptyShellDisp = !emptyShellDisp;
    }/* Toggling empty radius display */
  else if (*tag==EMMARADIUSBUTTON)
    {/* Toggling emma radius display */
      emmaRadiusDisp = !emmaRadiusDisp;
    }/* Toggling emma radius display */
  else if (*tag==FEATHERRADIUSBUTTON)
    {/* Toggling feather radius display */
      featherRadiusDisp = !featherRadiusDisp;
    }/* Toggling feather radius display */
  else if (*tag==DRAWREFINE)
    {/* Toggling refinement display */
      refineDisp = !refineDisp;
    }/* Toggling refinement display */
  else if (*tag==DRAWCLUSTERS)
    {/* Toggling draw clusters */
      clusterDisp = !clusterDisp;
    }/* Toggling draw clusters */
  else if (*tag==DRAWREFERENCE)
    {/* Toggling draw reference */
      referenceDisp = !referenceDisp;
    }/* Toggling draw reference */
  else if (*tag==DISPCENTERS)
    {/* Toggling display of centers */
      dispCenters = !dispCenters;
    }/* Toggling display of centers */
  else if ((*tag==AMMCORRCOEF)||(*tag==AMMAVGDENS)||(*tag==AMMSTDEV))
    {/* Toggling thresholding */
      autoMinMaxThreshold = *tag;
      if ((*tag==AMMCORRCOEF))
	sprintf(outStr,"Correlation Coef Cutoffs");
      else if ((*tag==AMMAVGDENS))
	sprintf(outStr,"Average Density Cutoffs");
      else if ((*tag==AMMSTDEV))
	sprintf(outStr,"Stdev Density Cutoffs");
      createSimpleString(&simpleString,outStr);
      XtSetArg(arglst[0], XmNlabelString, simpleString);
      XtSetArg(arglst[1], XmNrecomputeSize, FALSE );
      XtSetValues(autoMinMaxLabelW, arglst, 2); 
      XFlush(dsply);
      XmStringFree(simpleString); 
      XFlush(dsply);

      showBoxingLevel = *tag;

    }/* Toggling thresholding */

  DrawWindow();

}/* BOX_CHANGE */




/* ************************************************************
 * autobox_scroll_change() - Whenever autobox scroll widgets changed
 * call here.
 */
void autobox_scroll_change(Widget widget, int *tag, 
			   XmScrollBarCallbackStruct *callback_data)
{/* AUTOBOX_SCROLL_CHANGE */
  XmString simpleString;
  int i,removeThem,x,y;
  static float minScrollVal= 0.0,maxScrollVal=1.0;
  float thisVal,minVal,maxVal;

  if(*tag==THINNING)
    {/* Change number of thinning layers */

      nThins = (int)callback_data->value;

      sprintf(outStr,"%d",nThins); 
      
      createSimpleString(&simpleString,outStr);
      XtSetArg(arglst[0], XmNlabelString, simpleString);
      XtSetValues(thinLabelW, arglst, 1); 
      XFlush(dsply);
      XmStringFree(simpleString); 
      XFlush(dsply);

    }/* Change number of thinning layers */
  else if(*tag==DILATE)
    {/* Change number of dilation layers */

      nDilates = (int)callback_data->value;

      sprintf(outStr,"%d",nDilates); 
      
      createSimpleString(&simpleString,outStr);
      XtSetArg(arglst[0], XmNlabelString, simpleString);
      XtSetValues(dilateLabelW, arglst, 1); 
      XFlush(dsply);
      XmStringFree(simpleString); 
      XFlush(dsply);

    }/* Change number of dilation layers */
  else if(*tag==PRAG)
    {/* Change prag */

      prag = ((float)callback_data->value/10.0);

      sprintf(outStr,"%.1f",prag); 
      
      createSimpleString(&simpleString,outStr);
      XtSetArg(arglst[0], XmNlabelString, simpleString);
      /* XtSetValues(pragLabelW, arglst, 1);  */
      XFlush(dsply);
      XmStringFree(simpleString); 
      XFlush(dsply);

    }/* Change prag */
  else if(*tag==REFCUTOFF)
    {/* Change refCutoff */

      /* Ratio cutoff to throw out particles */
      refCutoff = ((float)callback_data->value/100.0);

      sprintf(outStr,"%.2f",refCutoff); 
      
      createSimpleString(&simpleString,outStr);
      XtSetArg(arglst[0], XmNlabelString, simpleString);
      /* XtSetValues(refCutoffLabelW, arglst, 1);  */
      XFlush(dsply);
      XmStringFree(simpleString); 
      XFlush(dsply);

    }/* Change refCutoff */
  else
    {/* Do min/max/remove thing */
      removeThem = FALSE;
      if(*tag==AUTOBOX)
	{/* MIN */ 
	  minScrollVal = (float)callback_data->value/100.0;
	}/* MIN */
      else if (*tag==CORRELATE)
	{/* MAX */
	  maxScrollVal = (float)callback_data->value/100.0;
	}/* MAX */
      else if (*tag==REMOVEBOXES)
	{/* MAX */
	  removeThem = TRUE;
	}/* MAX */
      
      if(autoMinMaxThreshold==AMMCORRCOEF)
	{/* corr coef */

	  /* First find MIN and MAX */
	  maxVal= minVal = box[0].correlation;
	  for(i=0;i<numBoxes;i++)
	    {/* Check all */
	      box[i].reference = FALSE;
	      minVal = MIN(box[i].correlation,minVal);
	      maxVal = MAX(box[i].correlation,maxVal);
	    }/* Check all */

	  threshHi = minVal + ((maxVal-minVal)*maxScrollVal);
	  threshLo = minVal + ((maxVal-minVal)*minScrollVal);
      
	  /* Now mark those outside range */
	  i = 0;
	  do
	    {/* Check all */
	      if(box[i].correlation<threshLo ||
		 box[i].correlation>threshHi )
		{/* CAndidate for removal */
		  if(removeThem)
		    {/* Seek and destroy */
		      removeBox(box[i].x, box[i].y, TRUE);
		      /* DON'T Go to next one - we've got to check this new 
		       * one that's taken i's place... */
		    }/* Seek and destroy */
		  else
		    {/* Just mark it */
		      box[i].reference = TRUE;
		      i++;  /* Go to next one */
		    }/* Just mark it */
		}/* CAndidate for removal */
	      else i++; /* Go to next one */
	    }/* Check all */
	  while(i<numBoxes);
	  
	}/* corr coef */
      else if(autoMinMaxThreshold==AMMAVGDENS)
	{/* Avg density */
	  int beginY,beginX,outHeight,outWidth,j,k,l;
	  float tmpBuffer[20555];
	  float totVal;

	  /* The current value of avgdensity is incorrest I'm thinking
	   * so let's recalc it here */
	  outHeight = outWidth = (boxRadius*2)+1;
	  for(i=0;i<numBoxes;i++)
	    {/* do all */
	      beginY = box[i].y - boxRadius;
	      beginX = box[i].x - boxRadius;
	      totVal = 0.0;
	      for(j=0; j<outHeight; j++)
		{
		  yOff = (beginY + j)*file_width + beginX;
		  memcpy(tmpBuffer,&inCore_buffer[(yOff)],
			 (outWidth*sizeof(float)));
		  for(k=0;k<outWidth;k++)
		    totVal += tmpBuffer[k];
		}
	      box[i].avgdensity = totVal ;
	    }/* do all */

	  /* First find MIN and MAX */
	  maxVal= minVal = box[0].avgdensity;
	  for(i=0;i<numBoxes;i++)
	    {/* Check all */
	      box[i].reference = FALSE;
	      minVal = MIN(box[i].avgdensity,minVal);
	      maxVal = MAX(box[i].avgdensity,maxVal);
	    }/* Check all */
	  
	  threshHi = minVal + ((maxVal-minVal)*maxScrollVal);
	  threshLo = minVal + ((maxVal-minVal)*minScrollVal);
      
	  /* Now mark those outside range */
	  i = 0;
	  do
	    {/* Check all */
	      if(box[i].avgdensity<threshLo ||
		 box[i].avgdensity>threshHi )
		{/* CAndidate for removal */
		  if(removeThem)
		    {/* Seek and destroy */
		      removeBox(box[i].x, box[i].y, TRUE);
		      /* DON'T Go to next one - we've got to check this new 
		       * one that's taken i's place... */
		    }/* Seek and destroy */
		  else
		    {/* Just mark it */
		      box[i].reference = TRUE;
		      i++;  /* Go to next one */
		    }/* Just mark it */
		}/* CAndidate for removal */
	      else i++; /* Go to next one */
	    }/* Check all */
	  while(i<numBoxes);
	  
	}/* Avg density */
      else if(autoMinMaxThreshold==AMMSTDEV)
	{/* stdev */
	  int beginY,beginX,outHeight,outWidth,j,k,l;
	  float tmpBuffer[20555];
	  float totVal,img_bkg,img_min,img_max,img_avg,img_std;

	  outHeight = outWidth = (boxRadius*2)+1;
	  if(firstAMMSTDEVtime)
	    {/* only do once */
	      for(i=0;i<numBoxes;i++)
		{/* do all */
		  beginY = box[i].y - boxRadius;
		  beginX = box[i].x - boxRadius;
		  totVal = 0.0;
		  stuffDisplayBoxBuffer(i, image_fileName.addr, 1);
		  box[i].stdev = purdueHeader.img_std ;
		  printf("box[%d] stdev[%f]\n",i,box[i].stdev);
		}/* do all */
	    }/* only do once */
	  firstAMMSTDEVtime = FALSE;

	  /* First find MIN and MAX */
	  maxVal= minVal = box[0].stdev;
	  for(i=0;i<numBoxes;i++)
	    {/* Check all */
	      box[i].reference = FALSE;
	      minVal = MIN(box[i].stdev,minVal);
	      maxVal = MAX(box[i].stdev,maxVal);
	    }/* Check all */
	  
	  threshHi = minVal + ((maxVal-minVal)*maxScrollVal);
	  threshLo = minVal + ((maxVal-minVal)*minScrollVal);
      
	  /* Now mark those outside range */
	  i = 0;
	  do
	    {/* Check all */
	      if(box[i].stdev<threshLo ||
		 box[i].stdev>threshHi )
		{/* CAndidate for removal */
		  if(removeThem)
		    {/* Seek and destroy */
		      removeBox(box[i].x, box[i].y, TRUE);
		      /* DON'T Go to next one - we've got to check this new 
		       * one that's taken i's place... */
		    }/* Seek and destroy */
		  else
		    {/* Just mark it */
		      box[i].reference = TRUE;
		      i++;  /* Go to next one */
		    }/* Just mark it */
		}/* CAndidate for removal */
	      else i++; /* Go to next one */
	    }/* Check all */
	  while(i<numBoxes);
	  
	}/* stdev */
      

      if((callback_data->reason != XmCR_DRAG))
	{/* Do something */
	  DrawWindow();
	}/* Do something */
      
      sprintf(outStr,"Done.  [%d] boxes.",numBoxes);
      XmTextSetString(timeLeftBoxW,outStr);
      XFlush(dsply);
      
      if (*tag==REMOVEBOXES)
	{/* Reset this for next time */
	  minScrollVal = 0.0;
	  maxScrollVal = 1.0;
	  
	  /* Now setup the scrolls accordingly */
	  XtSetArg(arglst[0], XmNvalue, (int)(0));
	  XtSetValues(minBoxScrollWidget, arglst, 1);
	  XtSetArg(arglst[0], XmNvalue, (int)(100));
	  XtSetValues(maxBoxScrollWidget, arglst, 1);
	  XFlush(dsply);
	}/* Reset this for next time */
    }/* Do min/max/remove thing */
  
}/* AUTOBOX_SCROLL_CHANGE */




/*************************************************************
 * drawFFTLiteActivate() - They want to draw the window over the net.
 * drawLite causes a smaller drawing area and also forces the
 * user to use the "D" button when he wants a fresh display.
*/
void drawFFTLiteActivate(Widget widget, int *tag, 
			 XmPushButtonCallbackStruct *callback_data)
{/* DRAWFFTLITEACTIVATE */

  drawLite = TRUE;
  DrawFFTWindow();
  if(*tag==2)
    DrawWindow();
  drawLite = FALSE;

}/* DRAWFFTLITEACTIVATE */





/*************************************************************
 * drawLiteActivate() - They want to draw the window over the net.
*/
void drawLiteActivate(Widget widget, int *tag, 
		      XmPushButtonCallbackStruct *callback_data)
{/* drawLiteACTIVATE */

  drawLite = TRUE;
  DrawWindow();
  drawLite = FALSE;

}/* drawLiteACTIVATE */





/*******************************************************
 * convertDeltaXY() - convert X,Y coord to display coords.
*/
void convertDeltaXY(int *convX, int *convY, float fX, float fY) 
{/* CONVERTDELTAXY */
  float factor;

  allMax = MAX(deltaMax,(0.0-deltaMin));
  if(allMax != 0.0)
    factor = (float)(220.0 / allMax);
  *convX = (int)(256.0 + (fX * factor));
  *convY = (int)(256.0 - (fY * factor));

}/* CONVERTDELTAXY */



/******************************************************
 * convertXYDelta() - Convert display coords into
 * theta and phi
*/
void convertXYDelta(int x, int y, 
		     float *thisX, float *thisY)
{/* CONVERTXYDELTA */
  float factor;

  allMax = MAX(deltaMax,(0.0-deltaMin));
  factor = (float)(allMax / 220.0);

  *thisX = (float)(x-256)*factor;
  *thisY = (float)(256-y)*factor;

}/* CONVERTXYDELTA */






/*******************************************************
 * convertXY() - convert X,Y coord to display coords.
*/
void convertXY(int *convX, int *convY, float fX, float fY) 
{/* CONVERTXY */
  float xFactor,yFactor;

  if(symmetry==5)
    {/* Draw a diamond shape */
      float tempTheta,sinTheta;

      if(fY> 90.0)
	{tempTheta = 90.0 - (fY-90.0);}
      else tempTheta = fY;

      sinTheta = sin((double)deg2rad(tempTheta));

      xFactor = (float)((7.0*sinTheta) * (32.0 / phiMax));
      *convX = (int)(fX * xFactor);
      *convX += 256;
      
      yFactor = (float)(21.0 * ((90.0 - 69.0)/(theta2 - theta1)));
      
      fY = fY - theta1;
      *convY = (int)(fY * yFactor);
      
      *convY = 20 + *convY;

    }/* Draw a diamond shape */
  else
    {/* Everything else */

      xFactor = (float)(7.0 * (32.0 / phiMax));
      *convX = (int)(fX * xFactor);
      *convX += 256;
      
      yFactor = (float)(21.0 * ((90.0 - 69.0)/(theta2 - theta1)));
      
      fY = fY - theta1;
      *convY = (int)(fY * yFactor);
      
      *convY = 20 + *convY;

    }/* Everything else */
  

}/* CONVERTXY */





/*************************************************************
 * batchjob_allasu_activate() - Batch process particles.
*/
void batchjob_allasu_activate(Widget widget, int *tag, 
			      XmPushButtonCallbackStruct *callback_data)
{/* BATCHJOB_ALLASU_ACTIVATE */
  int i, foundOne;

  for(i=0;i<totParamFiles;i++)
    {/* Get all */

      read_param(batchListStr[i], FALSE);

    }/* Get all */

  /* Update display */
  DrawICOSWindow();

  raiseIt(partprocscreen);

}/* BATCHJOB_ALLASU_ACTIVATE */





/*************************************************************
 * batchjob_submit_activate() - Submit a batch job for all 
 * the particles at which you've been looking.
*/
void batchjob_submit_activate(Widget widget, int *tag, 
			      XmPushButtonCallbackStruct *callback_data)
{/* BATCHJOB_SUBMIT_ACTIVATE */
  char *batchStr;
  char sysCall[128],mailNode[128];
  FILE *fOut;

  batchStr = XmTextGetString(batchjobTextW);

  fOut = fopen(paramBatchJobFilename,"w");
  if(fOut)
    {/* OK to write */
      
      fprintf(fOut,"%s\n",batchStr);
      fclose(fOut);

      sprintf(mailNode,"%s@baker1",getenv("USER"));

      sprintf(sysCall,"%s < /dev/null | mail %s &",
	      paramBatchJobFilename,
	      mailNode);

      system(sysCall);

      askThem("Batch Job submitted to OS for execution.",
	      FALSE,XmDIALOG_INFORMATION);	      

    }/* OK to write */
  else
    {/* can't do it */
    }/* can't do it */

  XtFree(batchStr);

}/* BATCHJOB_SUBMIT_ACTIVATE */





/*************************************************************
 * batchjob_activate() - Get batch job screen up.
*/
void batchjob_activate(Widget widget, int *tag, 
		       XmPushButtonCallbackStruct *callback_data)
{/* BATCHJOB_aCTIVATE */

  raiseIt(batchjobscreen);

}/* BATCHJOB_aCTIVATE */





/*************************************************************
 * batchjob_deactivate() - get rid of batch job screen.
*/
void batchjob_deactivate(Widget widget, int *tag, 
			 XmPushButtonCallbackStruct *callback_data)
{/* BATCHJOB_DEACTIVATE */

  XtUnmanageChild(batchjobscreen);

}/* BATCHJOB_DEACTIVATE */




/*************************************************************
 * batch_focus() - Give helpful feedback about the meaning
 * of each widget.
 */
void batch_focus(Widget widget, int *tag, 
		 XmAnyCallbackStruct *callback_data)
{/* BATCH_FOCUS */
  XmString simpleString;

  if(widget==batchModeW)
    sprintf(outStr,"Enter MODE.");
  else if(widget==batchBinW)
    sprintf(outStr,"Enter BIN.");
  else if(widget==batchSymW)
    sprintf(outStr,"Enter SYM.");
  else if(widget==batchDtheW)
    sprintf(outStr,"Enter DTHE.");
  else if(widget==batchDphi90W)
    sprintf(outStr,"Enter DPHI90.");
  else if(widget==batchIlistW)
    sprintf(outStr,"Enter ILIST.");
  else if(widget==batchRadLoW)
    sprintf(outStr,"Enter RADLO.");
  else if(widget==batchRadHiW)
    sprintf(outStr,"Enter RADHI.");
  else if(widget==batchRadStepW)
    sprintf(outStr,"Enter RADSTEP.");
  else if(widget==batchAnnLoW)
    sprintf(outStr,"Enter ANNLO.");
  else if(widget==batchAnnHiW)
    sprintf(outStr,"Enter ANNHI.");
  else if(widget==batchResLoW)
    sprintf(outStr,"Enter RESLO.");
  else if(widget==batchResHiW)
    sprintf(outStr,"Enter RESHI.");
  else if(widget==batchJCutW)
    sprintf(outStr,"Enter JCUT.");
  else if(widget==batchSigW)
    sprintf(outStr,"Enter SIG.");
  else if(widget==batchMagCenW)
    sprintf(outStr,"Enter MAGCEN.");
  else if(widget==batchMagStepW)
    sprintf(outStr,"Enter MAGSTEP.");
  else if(widget==batchMagNumW)
    sprintf(outStr,"Enter MAGNUM.");
  else if(widget==batchMagNormW)
    sprintf(outStr,"Enter MAGNORM.");
  else if(widget==batch3DMapW)
    sprintf(outStr,"Enter 3D MAP Filename.");
  else if(widget==batchPrjsFileW)
    sprintf(outStr,"Enter PRJS Filename.");
  else if(widget==batchPftsFileW)
    sprintf(outStr,"Enter PFTS Filename.");

  createSimpleString(&simpleString,outStr);
  XtSetArg(arglst[0], XmNlabelString, simpleString);
  XtSetValues(batchHelpLabelW, arglst, 1); 
  XFlush(dsply);
  XmStringFree(simpleString); 
  XFlush(dsply);

}/* BATCH_FOCUS */



/* ************************************************************
 * batchparam_list_proc() - REad the specific file from the
 * batch job and display.
*/
void batchparam_list_proc(Widget widget, int *tag, 
			  XmListCallbackStruct *list)
{/* BATCHPARAM_LIST_PROC */
  Arg arglist[2];
  char *thisStr;
  int i, foundOne;

  foundOne= (int) (list->item_position);   /* Numbers from 1 not 0 */
  if(!XmStringGetLtoR(list->item, 
		      XmSTRING_DEFAULT_CHARSET, 
		      &thisStr))
    {printf("Error extracting List value.\n");}

  read_param(thisStr,TRUE);

  /* Update display */
  DrawWindow();
  DrawICOSWindow();

  XtFree(thisStr);  

  raiseIt(partprocscreen);

}/* BATCHPARAM_LIST_PROC */








/*************************************************************
 * partproc_write_activate() - Write a new paramfile. 
*/
void partproc_write_activate(Widget widget, int *tag, 
			     XmPushButtonCallbackStruct *callback_data)
{/* PARTPROC_WRITE_aCTIVATE */

  filterFiles("*",ALL); 

  writeType= PARAMFILE;
  readWrite= WRITING;

  raiseIt(fileSel_widget);

}/* PARTPROC_WRITE_aCTIVATE */





/*************************************************************
 * partproc_proj_activate() - Project the MAP for each
 * particle and display on screen.
 */
void partproc_proj_activate(Widget widget, int *tag, 
			    XmAnyCallbackStruct *callback_data)
{/* PARTPROC_PROJ_ACTIVATE */

  mapProjW = widget;

}/* PARTPROC_PROJ_ACTIVATE */




/*************************************************************
 * partproc_activate() - They want to do particle processing.
*/
void partproc_activate(Widget widget, int *tag, 
		       XmPushButtonCallbackStruct *callback_data)
{/* PARTPROC_ACTIVATE */

  yscale = 1.0;  /* How do we set this??? */
  raiseIt(partprocscreen);
  XFlush(dsply);
  /*  raiseIt(batchjobscreen);
      XFlush(dsply); */

}/* PARTPROC_aCTIVATE */




/*************************************************************
 * partproc_delta_activate() - Add Delta x,y to all particles.
 */
void partproc_delta_activate(Widget widget, int *tag, 
			     XmAnyCallbackStruct *callback_data)
{/* PARTPROC_DELTA_ACTIVATE */

  raiseIt(deltaXYpartprocScreen);


}/* PARTPROC_DELTA_ACTIVATE */


/*************************************************************
 * delatXYpp_deactivate() - get rid of screen.
 */
void deltaXYpp_deactivate(Widget widget, int *tag, 
			  XmAnyCallbackStruct *callback_data)
{/* DELATXYPP_DEACTIVATE */

  XtUnmanageChild(deltaXYpartprocScreen);

}/* DELTAXYPP_DEACTIVATE */



/*************************************************************
 * partproc_delta_work() - Add Delta x,y to all particles.
 * Do the work here.
 */
void partproc_delta_work(Widget widget, int *tag, 
			 XmAnyCallbackStruct *callback_data)
{/* PARTPROC_DELTA_WORK */
  P_INFO *thisInfo;
  int thisX,thisY,thisXp1,thisYp1;
  float deltaX,deltaY,x,y;
  char *inStr;

  inStr = XmTextGetString(deltaXpartprocW);
  deltaX = (float)strtod(inStr,0);
  XtFree(inStr);
  inStr = XmTextGetString(deltaYpartprocW);
  deltaY = (float)strtod(inStr,0);
  XtFree(inStr);

  thisInfo = info.info;
  while(thisInfo != NULL)
    {/* Do all of them */

      x = thisInfo->x + deltaX;
      y = thisInfo->y + deltaY;

      thisInfo->x=x; 
      thisInfo->y=y; 
      thisInfo->origX = x;
      thisInfo->origY = y;
      thisInfo->deltaX = thisInfo->x - (float)(((float)
						purdueHeader.numCols / 
						2.0)-0.5);
      thisInfo->deltaY = thisInfo->y - (float)(((float)
						purdueHeader.numRows / 
						2.0)-0.5);
      deltaMax = MAX(thisInfo->deltaX, deltaMax);
      deltaMax = MAX(thisInfo->deltaY, deltaMax);
      deltaMin = MIN(thisInfo->deltaX, deltaMin);
      deltaMin = MIN(thisInfo->deltaY, deltaMin);
      deltaXmin= MIN(thisInfo->deltaX, deltaXmin);
      deltaYmin= MIN(thisInfo->deltaY, deltaYmin);
      deltaXmax= MAX(thisInfo->deltaX, deltaXmax);
      deltaYmax= MAX(thisInfo->deltaY, deltaYmax);
      thisInfo->distance = sqrt((thisInfo->deltaX*thisInfo->deltaX)+
			    (thisInfo->deltaY*thisInfo->deltaY));
      distanceMin = MIN(distanceMin,thisInfo->distance);
      distanceMax = MAX(distanceMax,thisInfo->distance);
      
      thisInfo = thisInfo->next;

    }/* Do all of them */


  XtUnmanageChild(deltaXYpartprocScreen);
  stuffPackedImage(purdueHeader.numCols,
		   purdueHeader.numRows,
		   packImg,purdueHeader.nimg,
		   imgNum);
  DrawWindow();
  DrawICOSWindow();
  XtUnmanageChild(deltaXYpartprocScreen);

}/* PARTPROC_DELTA_WORK */





/*************************************************************
 * partproc_deltaomega_activate() - Add Delta omega to all particles.
 */
void partproc_deltaomega_activate(Widget widget, int *tag, 
				  XmAnyCallbackStruct *callback_data)
{/* PARTPROC_DELTAOMEGA_ACTIVATE */

  raiseIt(deltaOmegaPPScreen);


}/* PARTPROC_DELTAOMEGA_ACTIVATE */


/*************************************************************
 * delatOmegapp_deactivate() - get rid of screen.
 */
void deltaOmegapp_deactivate(Widget widget, int *tag, 
			     XmAnyCallbackStruct *callback_data)
{/* DELATOMEGAPP_DEACTIVATE */

  XtUnmanageChild(deltaOmegaPPScreen);

}/* DELTAOMEGAPP_DEACTIVATE */



/*************************************************************
 * partproc_deltaomega_work() - Add Delta omega to all particles.
 * Do the work here.
 */
void partproc_deltaomega_work(Widget widget, int *tag, 
			      XmAnyCallbackStruct *callback_data)
{/* PARTPROC_DELTAOMEGA_WORK */
  P_INFO *thisInfo;
  int thisX,thisY,thisXp1,thisYp1;
  float deltaOmega,omega;
  char *inStr;

  inStr = XmTextGetString(deltaOmegaPPW);
  deltaOmega = (float)strtod(inStr,0);
  XtFree(inStr);

  thisInfo = info.info;
  while(thisInfo != NULL)
    {/* Do all of them */

      omega = thisInfo->omega + deltaOmega;
      if(omega > 360.0)
	omega -= 360.0;
      if(omega < 0.0)
	omega += 360.0;
      
      thisInfo->omega = omega; 
      
      thisInfo = thisInfo->next;

    }/* Do all of them */


  XtUnmanageChild(deltaXYpartprocScreen);
  stuffPackedImage(purdueHeader.numCols,
		   purdueHeader.numRows,
		   packImg,purdueHeader.nimg,
		   imgNum);
  DrawWindow();
  DrawICOSWindow();
  XtUnmanageChild(deltaOmegaPPScreen);

}/* PARTPROC_DELTAOMEGA_WORK */



/*************************************************************
 * binning_change() - Chnage binning concept.
 */
void binning_change(Widget widget, int *tag, 
		    XmAnyCallbackStruct *callback_data)
{/* BINNING_CHANGE */

  doBinning = *tag;

  if(*tag==BIN2X2)
    {/* Radomize the centers */
      sprintf(outStr,"2x2 selected");
      XmTextSetString(binningTextW,outStr);
    }/* Radomize the centers */
  else if(*tag==BIN3X3)
    {/* Radomize the centers */
      sprintf(outStr,"3x3 selected");
      XmTextSetString(binningTextW,outStr);
    }/* Radomize the centers */
  else if(*tag==BIN4X4)
    {/* Radomize the centers */
      sprintf(outStr,"4x4 selected");
      XmTextSetString(binningTextW,outStr);
    }/* Radomize the centers */
  else 
    {/* future use */
      printf("Not yet.\n");
    }/* future use */

}/* BINNING_CHANGE */




/*************************************************************
 * manipPP_change() - Change partproc concepts.
 */
void manipPP_change(Widget widget, int *tag, 
		    XmAnyCallbackStruct *callback_data)
{/* MANIPPP_CHANGE */

  doManipPP = *tag;

  if(*tag==RANDCENTER)
    {/* Radomize the centers */
      sprintf(outStr,"1.0 STD");
      XmTextSetString(manipPPvalueW,outStr);
    }/* Radomize the centers */
  else if(*tag==UNKPP)
    {/* future use */
      printf("Not yet.\n");
    }/* future use */

}/* MANIPPP_CHANGE */



/*************************************************************
 * manipPP_activate() - get screen up.
 */
void manipPP_activate(Widget widget, int *tag, 
		      XmAnyCallbackStruct *callback_data)
{/* MANIPPP_ACTIVATE */

  raiseIt(manipPPScreen);

}/* MANIPPP_ACTIVATE */



/*************************************************************
 * manipPP_deactivate() - get rid of screen
 */
void manipPP_deactivate(Widget widget, int *tag, 
			XmAnyCallbackStruct *callback_data)
{/* MANIPPP_DEACTIVATE */

  XtUnmanageChild(manipPPScreen);

}/* MANIPPP_DEACTIVATE */



/*************************************************************
 * partproc_manip_work() - 
 * Do the work here.
 */
void partproc_manip_work(Widget widget, int *tag, 
			 XmAnyCallbackStruct *callback_data)
{/* PARTPROC_MANIP_WORK */
  P_INFO *thisInfo;
  char *inStr;
  int pMin,pMax;
  float numStd,factor,noise,mean,stDev,tmp1;

  if(doManipPP==RANDCENTER)
    {/* Randomize the centers */

      inStr = XmTextGetString(manipPPvalueW);
      numStd = (float)strtod(inStr,0);
      XtFree(inStr);

      particleCalcStats(&pMin, &pMax, &mean, &stDev);

      noise  = (float)(stDev) * numStd;   
      /* Get our fraction of the random number */
      factor= 1.0 / DENOMINATOR;

      thisInfo = info.info;
      while(thisInfo != NULL)
	{/* Do all of them */
	  
	  tmp1 = ((((float)random()*factor)-0.5)*2.0*noise);

	  thisInfo->x += tmp1;

	  tmp1 = ((((float)random()*factor)-0.5)*2.0*noise);

	  thisInfo->y += tmp1;
	  
	  thisInfo->x = MIN(MAX(thisInfo->x,0.0),(float)purdueHeader.numCols);
	  thisInfo->y = MIN(MAX(thisInfo->y,0.0),(float)purdueHeader.numRows);

	  thisInfo = thisInfo->next;
	  
	}/* Do all of them */
      
    }/* Randomize the centers */

  stuffPackedImage(purdueHeader.numCols,
		   purdueHeader.numRows,
		   packImg,purdueHeader.nimg,
		   imgNum);
  DrawWindow();
  DrawICOSWindow();
  XtUnmanageChild(manipPPScreen);

}/* PARTPROC_MANIP_WORK */





/*************************************************************
 * partproc_deltaoldNew_work() - Add Delta omega to all particles
 * to convert old parameter files into new ones (July 1999).
 * Do the work here.
 */
void partproc_deltaoldNew_work(Widget widget, int *tag, 
			       XmAnyCallbackStruct *callback_data)
{/* PARTPROC_DELTAOLDNEW_WORK */
  P_INFO *thisInfo;
  float deltaOmega,omega;

  deltaOmega = (float)90.0;

  thisInfo = info.info;
  while(thisInfo != NULL)
    {/* Do all of them */

      omega = thisInfo->omega + deltaOmega;
      if(omega > 360.0)
	omega -= 360.0;
      if(omega < 0.0)
	omega += 360.0;

      thisInfo->omega = omega; 
      
      thisInfo = thisInfo->next;

    }/* Do all of them */


  XtUnmanageChild(deltaXYpartprocScreen);
  stuffPackedImage(purdueHeader.numCols,
		   purdueHeader.numRows,
		   packImg,purdueHeader.nimg,
		   imgNum);
  DrawWindow();
  DrawICOSWindow();
  XtUnmanageChild(deltaOmegaPPScreen);

}/* PARTPROC_DELTAOLDNEW_WORK */





/*************************************************************
 * partproc_rebox_activate() - Rebox the specifed particle(s).
 */
void partproc_rebox_activate(Widget widget, int *tag, 
			     XmAnyCallbackStruct *callback_data)
{/* PARTPROC_REBOX_ACTIVATE */
  int i,j,k,l,m,yOut,startI,endI,box_width;
  int thisX, thisY;
  float deltaX, deltaY;
  float *thisBoxBuf;
  int yIn,xIn,outHeight,outWidth,ok2go;
  int firstCol,firstRow, thisBox;
  FILE *fOut,*fIn;
  static int beenWarned = FALSE;
  
  ok2go = TRUE;
  if(EQU(paramScanFilename,""))
    {/* No scan file read */
      askThem("No scan file read.",
	      FALSE,XmDIALOG_INFORMATION);
      ok2go = FALSE;
    }/* No scan file read */

  if(ok2go && beenWarned)
    {/* They've been warned - pass this test . */
    }/* They've been warned - pass this test . */
  else if(ok2go)
    {/* Warn them */
      if(ok2go && 
	 askThem("This is a destructive operation.  The particle\n\
disk file will be permanently changed as well as\n\
the box coordinates file.  Do you want to continue?",
		 FALSE,XmDIALOG_QUESTION))
	{/* destruct? */
	  ok2go = TRUE;
	}/* destruct? */
      else ok2go = FALSE;
    }/* Warn them */
  
  beenWarned = TRUE;
  
  if(ok2go && *tag==ALL)
    {/* destruct? */
      if(askThem("You've requested that ALL particles be reboxed.\n\
Is this what you *really* want?",
		 FALSE,XmDIALOG_QUESTION))
	ok2go = TRUE;
      else ok2go = FALSE;
    }/* destruct? */
  
  if(ok2go)
    {/* OK to go */
      
      if(*tag==0)
	{/* Do a single one */
	  startI = imgNum;
	  endI   = startI + 1;
	}/* Do a single one */
      else if(*tag==ALL)
	{/* DO them all */
	  startI = 1;
	  endI   = purdueHeader.nimg+1;
	}/* DO them all */
      
      for(m=startI;m<endI;m++)
	{/* Do all desired */
	  
	  thisX = 0;
	  thisY = 0;
	  
	  findNearestParticle(&thisX,&thisY,&foundPart,FALSE,i,FALSE,0);
	  
	  deltaX = (float)(foundPart->x - (float)(purdueHeader.numCols/2+1));
	  deltaY = (float)(foundPart->y - (float)(purdueHeader.numRows/2+1));
	  
	  thisBox = foundPart->particle_number-1;
	  boxRadius = purdueHeader.numCols / 2;
	  box[thisBox].x += deltaX;
	  box[thisBox].y += deltaY;
	  
	  /* Get these so that they are NINT() rounded */
	  box[thisBox].x += 0.0499999;
	  box[thisBox].y += 0.0499999;
	  
	  firstCol = box[thisBox].x - boxRadius;
	  firstRow = box[thisBox].y - boxRadius;

	  if(box[thisBox].edit_mirror_buffer)
	    free(box[thisBox].edit_mirror_buffer);
	  box[thisBox].inMemory = FALSE;

	  thisBoxBuf = packImg[thisBox];
	  
	  /* Get dimensions */
	  outHeight= outWidth = ((boxRadius*2)+1);
	  box_width = purdueHeader.numCols;

	  fIn  = fopen(paramScanFilename, "r");
	  for(i=0;i<outHeight;i++)
	    {/* Rows */
	      yIn = 1024 + ((firstRow+i)*scanCols) + firstCol; 
	      sysError = fseek(fIn,yIn,0);
	      if(sysError)
		{
		  askThem("Error: fseek in reading new particle.",
			  FALSE, XmDIALOG_INFORMATION);
		  goto noWriteThis;
		}
	      yOut = (i * box_width);
	      /* We need to read the floatInt*4 format here for mode==46!! */
	      fread(&(thisBoxBuf[(yOut)]), sizeof(BYTE), box_width, fIn);

	    }/* Rows */
      
	  /* Calc the background */
	  backChar= (float)calcFloatBackground(thisBoxBuf, 
					   outWidth, 
					   outHeight, box_width,
					   fileType,
					   bkgrndMethod,
					   boxMethod,
					   packRadius,
					   boxRadius,
					   featherRadius,
					   recipSizeFactor,
					   tryColors
					   );
	  
	  /* Draw the background  for all > radius in box */
	  for(k=0;k<outHeight;k++)
	    {/* Do the y direction */
	      int yOut,yCent,xCent;
	      float ySq,xSq,dist;
	      
	      /* Where is the center? */
	      yOut = (k * box_width);
	      yCent = ((boxRadius ));
	      xCent = ((boxRadius ));
	      
	      /* Stuff all > radius in background character */
	      for(l=0;l<outWidth;l++)
		{/* Stuff the outbuffer */
		  
		  ySq = (k - yCent) * (k - yCent);
		  xSq = (l - xCent) * (l - xCent);
		  
		  dist = sqrt((ySq+xSq));
		  
		  if(dist > ((boxRadius)+1.0))
		    thisBoxBuf[(yOut+(l))]=  backChar;

		}/* Stuff the outbuffer */
	    }/* Do the y direction */
	  
	  fOut = fopen(paramBoxesFilename,"r+");
	  /* Now write this new box to the disk file */
	  yOut = 512 + (512 * (m)) + ((m-1) * (outWidth*outHeight));
	  sysError = fseek(fOut,yOut,0);
	  if(sysError)
	    {
	      askThem("Error: fseek in writing new particle.",
		      FALSE, XmDIALOG_INFORMATION);
	      goto noWriteThis;
	    }
	  fwrite(thisBoxBuf, sizeof(BYTE),
		(outWidth*outHeight),fOut);

	  /* Now write new Box Coords file */
	  writeBoxCoords(paramBoxCoordsFilename, PARAMSCAN);

	  foundPart->x = (float)boxRadius + 1.0;
	  foundPart->y = (float)boxRadius + 1.0;

	noWriteThis:;
	  fclose(fOut);
	  fclose(fIn);

	}/* Do all desired */

      findNearestParticle(&thisX,&thisY,&foundPart,FALSE,imgNum,TRUE,0);

    }/* OK to go */


}/* PARTPROC_REBOX_ACTIVATE */




/*************************************************************
 * cavgPP_change() - Change circavg concepts in partproc
 * screen.
 */
void cavgPP_change(Widget widget, int *tag, 
		   XmAnyCallbackStruct *callback_data)
{/* CAVGPP_CHANGE */


  if(*tag==CAVGDO)
    {/* Circ Avg */
      doCavgPP = *tag;
      XtManageChild(writeCavgToggleW);
      XtUnmanageChild(nFoldLabelW);
      XtUnmanageChild(nFoldTextW);
    }/* Circ Avg */
  else if(*tag==NSYMDO)
    {/* nsym */
      doCavgPP = *tag;
      XtUnmanageChild(writeCavgToggleW);
      XtManageChild(nFoldLabelW);
      XtManageChild(nFoldTextW);
    }/* nsym */
  else if(*tag==WRITECAVG)
    {/* write */
      writeCavg = !writeCavg;
    }/* write */
  else if((*tag==PARAMXY) || (*tag==BOXXY))
    {/* param/box xy */
      cavgPPcenter = *tag;
    }/* param/box xy */


}/* CAVGPP_CHANGE */



/*************************************************************
 * cavgPP_do_activate() - Do circavg
 */
void cavgPP_do_activate(Widget widget, int *tag, 
			XmAnyCallbackStruct *callback_data)
{/* CAVGPP_DO_ACTIVATE */
  char *nFoldText;
  int nFold,i,j,k,numAlloc,Ncol,Nrow,inum;
  float *thisRint,xcen,ycen;
  float *thisImg;
  float  *pkImg;
  FILE *fOut;

  nFoldText = XmTextGetString(nFoldTextW); 
  nFold = (int)strtod(nFoldText,0);
  XtFree(nFoldText);
  XFlush(dsply);

  if(doCavgPP == 0)
    {/* warning */
      askThem("You must choose whether you want Circular Averaging or n Symmetry applied.",
	      FALSE,XmDIALOG_INFORMATION);
      goto noCavgPP;
    }/* warning */

  if(cavgPPcenter == 0)
    {/* warning */
      askThem("You must choose a method for centers (param file or box).",
	      FALSE,XmDIALOG_INFORMATION);
      goto noCavgPP;
    }/* warning */

  if(writeCavg)
    {/* Open file */
      fOut = fopen("circular_avg.asc","w");
      if(!fOut)
	{/* error */
	  askThem("Cannot open file circular_avg.asc.",
		  FALSE,XmDIALOG_INFORMATION);
	  goto noCavgPP;
	}/* error */
    }/* Open file */


  if(nFold > 99 || nFold < 2)
    {/* warning */
      askThem("N Fold symmetry outside limits - resetting.",
	      FALSE,XmDIALOG_INFORMATION);
      nFold = MIN(MAX(nFold,2),99);
      sprintf(outStr,"%d",nFold);
      XmTextSetString(nFoldTextW,outStr);
    }/* warning */

  /***********************************/
  /* OK now call routine to compute  */
  /***********************************/
  Ncol = purdueHeader.numCols;
  Nrow = purdueHeader.numRows;

  if(totCavgPPrint > 0)
    {/* free memory */
      for(i=0;i<totCavgPPrint;i++)
	free(cavgPPrint[i]);
    }/* free memory */


  totCavgPPrint = purdueHeader.nimg;
  /* malloc() by *6 because we want to go to the edges of the box and
   * it has an interg==3 in img_sym() */
  numAlloc = (int)(purdueHeader.numCols * 6);
  for(i=0;i<totCavgPPrint;i++)
    cavgPPrint[i] = (float *)malloc(sizeof(float)*numAlloc);


  for(i=0;i<totCavgPPrint;i++)
    {
      thisRint = cavgPPrint[i];
      for(j=0;j<numAlloc;j++)
	thisRint[j] = 0.0;
    }

  thisImg = malloc(sizeof(float) *
		   purdueHeader.numCols *
		   purdueHeader.numRows);

  /* 2.0 for interp in img_sym() and 0.5 for radius not diameter 
   * 0.90 for 90% - we're plotting too much ... ??? */
  inum = (int)((2.0*0.5*0.90) * sqrt((double)
			  (purdueHeader.numCols * purdueHeader.numCols +
			   purdueHeader.numRows * purdueHeader.numRows
			   )
			  ));
  if(writeCavg)
    {/* Write cavg to disk file */
      fprintf(fOut,"Total particles                 = %d\n",
	      purdueHeader.nimg);
      fprintf(fOut,"Number data points per particle = %d\n",
	      inum);
      fprintf(fOut,"*******************************************\n");
    }/* Write cavg to disk file */

  for(i=0;i<purdueHeader.nimg;i++)
    {/* DO all */
      int nrot;
      float img_avg, img_std;
      int   img_min, img_max, img_bkg;

      if(doCavgPP = CAVGDO)
	nrot = 101;
      else nrot = nFold;
      
      pkImg = packImg[i];
      thisRint = cavgPPrint[i];

      for(j=0;j<purdueHeader.numRows;j++)
	for(k=0;k<purdueHeader.numCols;k++)
	  {
	    thisImg[((j*purdueHeader.numCols)+k)] = (float)
	      pkImg[((j*purdueHeader.numCols)+k)];
	  }

      if(cavgPPcenter==PARAMXY)
	{/* Find param center */
	  P_INFO *thisInfo;

	  thisInfo = info.info;
	  while(thisInfo != NULL)
	    {/* Do all of them */

	      if(thisInfo->particle_number == i+1)
		{/* Found it */
		  xcen = (float)(thisInfo->x);
		  ycen = (float)(thisInfo->y);

printf("Found id[%d,%d]  x[%f] y[%f]\n",
       thisInfo->particle_number, i,
       thisInfo->x,
       thisInfo->y
       );
		}/* Found it */

	      thisInfo = thisInfo->next;
	    }/* Do all of them */

	}/* Find param center */
      else
	{/* Use box center */
	  xcen = (float)((float)purdueHeader.numCols/2.0);
	  ycen = (float)((float)purdueHeader.numRows/2.0);
	}/* Use box center */

      r4_img_sym_(
	       thisImg,
	       &Ncol,
	       &Nrow,
	       &xcen,&ycen,
	       &nrot,
	       &img_min,
	       &img_max,
	       &img_avg,
	       &img_std,
	       &img_bkg,
	       thisRint,
	       &numAlloc
	       );

      if(writeCavg)
	{/* Write cavg to disk file */
	  fprintf(fOut,"Particle #%d\n",i+1);
	  /* 2.0 because that's the interp used in img_sym */
	  for(j=0;j<inum;j++)
	    fprintf(fOut,"%f  %f\n",(j / 2.0),thisRint[j]);
	  fprintf(fOut,"*******************************************\n");
	}/* Write cavg to disk file */


    }/* DO all */
  free(thisImg);

  if(writeCavg)
    {/* Close file */
      fclose(fOut);
      askThem("Circular averages written to file 'circular_avg.asc'.",
	      FALSE,XmDIALOG_INFORMATION);
    }/* Close file */



  noCavgPP:;

}/* CAVGPP_DO_ACTIVATE */




/*************************************************************
 * cavgPP_activate() - get part proc cricavg screen up.
 */
void cavgPP_activate(Widget widget, int *tag, 
		     XmAnyCallbackStruct *callback_data)
{/* CAVGPP_ACTIVATE */

  XtUnmanageChild(nFoldLabelW);
  XtUnmanageChild(nFoldTextW);
  XtUnmanageChild(writeCavgToggleW);
  raiseIt(cavgPPScreen); 

}/* CAVGPP_ACTIVATE */



/*************************************************************
 * cavgPP_deactivate() - get rid of screen.
 */
void cavgPP_deactivate(Widget widget, int *tag, 
		       XmAnyCallbackStruct *callback_data)
{/* CAVGPP_DEACTIVATE */

  XtUnmanageChild(cavgPPScreen);

}/* CAVGPP_DEACTIVATE */




/*************************************************************
 * partproc_sort_activate() - Sort particles.
 */
void partproc_sort_activate(Widget widget, int *tag, 
			    XmAnyCallbackStruct *callback_data)
{/* PARTPROC_SORT_ACTIVATE */
  XmString simpleString;

  icosaMotioning = FALSE;

  XtSetArg(arglst[0], XmNset, FALSE );
  
  XtSetValues(idtW, arglst, 1); 
  XtSetValues(thetatW, arglst, 1); 
  XtSetValues(phitW, arglst, 1); 
  XtSetValues(omegatW, arglst, 1); 
  XtSetValues(xtW, arglst, 1); 
  XtSetValues(ytW, arglst, 1); 
  XtSetValues(deltaXTW, arglst, 1); 
  XtSetValues(deltaYTW, arglst, 1); 
  XtSetValues(absDeltaXTW, arglst, 1); 
  XtSetValues(absDeltaYTW, arglst, 1); 
  XtSetValues(distanceTW, arglst, 1); 
  XtSetValues(magtW, arglst, 1); 
  XtSetValues(pftccW, arglst, 1); 
  XtSetValues(prjccW, arglst, 1); 
  XtSetValues(cmpccW, arglst, 1); 
  
  XFlush(dsply);

  idtoggle= thetatoggle= phitoggle= omegatoggle= xtoggle= ytoggle=
    magtoggle= cc1toggle= cc2toggle= cc3toggle= distancetoggle=
      deltaXtoggle = deltaYtoggle= absDeltaXtoggle= absDeltaYtoggle= FALSE;

  XtSetArg(arglst[0], XmNforeground, pixels_return[WHITE] );
  XtSetValues(magtW, arglst, 1); 
  XtSetValues(pftccW, arglst, 1); 
  XtSetValues(prjccW, arglst, 1); 
  XtSetValues(cmpccW, arglst, 1); 

  sprintf(outStr,"Choose key for SORT: ");
  createSimpleString(&simpleString,outStr);
  XtSetArg(arglst[0], XmNlabelString, simpleString);
  XtSetValues(parmTogsLabelW, arglst, 1); 
  XFlush(dsply);
  XmStringFree(simpleString); 
  XFlush(dsply);

  XtManageChild(parmTogsW);
  projASU = FALSE;
  XtUnmanageChild(projasuW);
  XtUnmanageChild(parmheadW);
  sprintf(outStr,"Choose SORT key.");
  XmTextSetString(partPointTextW,outStr);
  icoDispMode = ICOSORT;
  icoSortLevel = 0;
  DrawICOSWindow();

}/* PARTPROC_SORT_ACTIVATE */




/*************************************************************
 * partproc_screen_deactivate() - Get the partproc screen
 * off the display.
 */
void partproc_screen_deactivate(Widget widget, int *tag, 
				XmAnyCallbackStruct *callback_data)
{/* PARTPROC_SCREEN_DEACTIVATE */
  
  XtUnmanageChild(partprocscreen);
  
}/* PARTPROC_SCREEN_DEACTIVATE */





/*************************************************************
 * text_create() - Remeber the  text widget number
 * for use later.  
 */
void text_create(Widget widget, int *tag, 
		 XmAnyCallbackStruct *callback_data)
{/* TEXT_CREATE */
  
  if(*tag==PARAMFILE)
    {/* paramfile */
      /* Remember the widget number for later */
      paramFileTextW= widget;
      handle_cr(paramFileTextW);
    }/* paramfile */
  else if(*tag==PARAMMAP)
    {/* paramMapfile */
      /* Remember the widget number for later */
      paramMapFileTextW= widget;
      handle_cr(paramMapFileTextW);
    }/* paramMapfile */
  else if(*tag==PARAMSCAN)
    {/* paramScanfile */
      /* Remember the widget number for later */
      paramScanFileTextW= widget;
      handle_cr(paramScanFileTextW);
    }/* paramScanfile */
  else if(*tag==PROJCONTOUR)
    {/* projection contour text */
      /* Remember the widget number for later */
      projContourTextW= widget;
      handle_cr(projContourTextW);
    }/* projection contour text */
  else if(*tag==PROJASU)
    {/* projection ASU text */
      /* Remember the widget number for later */
      projasuTextW= widget;
      handle_cr(projasuTextW);
    }/* projection ASU text */
  else if(*tag==MOVEMENT)
    movementTextWidget = widget;
  else if(*tag==DIAM)
    diamTextWidget = widget;
  else if(*tag==INFOPROMPT)
    infoPromptTextW = widget;
  else if(*tag==PIXSIZEP)
    pixsizeW =	widget;
  else if(*tag==UNITSP)
    unitsW =	widget;
  else if(*tag==VOLTS)
    voltsW = widget;
  else if(*tag==AMPFACS)
    ampfacsW = widget;
  else if(*tag==DELF_MAJ)
    delfMajW = widget;
  else if(*tag==DELF_MIN)
    delfMinW = widget;
  else if(*tag==ANG_MAJ)
    angMajW = widget;
  else if(*tag==CS_COEF)
    csCoefW = widget;
  else if(*tag==BATCHJOB)
    batchjobFileTextW = widget;
  else if(*tag==BATCHJOBTEXT)
    batchjobTextW = widget;
  else if(*tag==THETAPROJ)
    thetaProjW = widget;
  else if(*tag==PHIPROJ)
    phiProjW = widget;
  else if(*tag==OMEGAPROJ)
    omegaProjW = widget;
  else if(*tag==TITLE)
    titleTextW = widget;
  else if(*tag==BATCHMODE)
    batchModeW = widget;
  else if(*tag==BATCHBIN)
    batchBinW = widget;
  else if(*tag==BATCHSYM)
    batchSymW = widget;
  else if(*tag==BATCHDTHE)
    batchDtheW = widget;
  else if(*tag==BATCHDPHI90)
    batchDphi90W = widget;
  else if(*tag==BATCHILIST)
    batchIlistW = widget;
  else if(*tag==BATCHRADLO)
    batchRadLoW = widget;
  else if(*tag==BATCHRADHI)
    batchRadHiW = widget;
  else if(*tag==BATCHRADSTEP)
    batchRadStepW = widget;
  else if(*tag==BATCHANNLO)
    batchAnnLoW = widget;
  else if(*tag==BATCHANNHI)
    batchAnnHiW = widget;
  else if(*tag==BATCHRESLO)
    batchResLoW = widget;
  else if(*tag==BATCHRESHI)
    batchResHiW = widget;
  else if(*tag==BATCHJCUT)
    batchJCutW = widget;
  else if(*tag==BATCHSIG)
    batchSigW = widget;
  else if(*tag==BATCHMAGCEN)
    batchMagCenW = widget;
  else if(*tag==BATCHMAGSTEP)
    batchMagStepW = widget;
  else if(*tag==BATCHMAGNUM)
    batchMagNumW = widget;
  else if(*tag==BATCHMAGNORM)
    batchMagNormW = widget;
  else if(*tag==BATCH3DMAP)
    batch3DMapW = widget;
  else if(*tag==BATCHPRJSFILE)
    batchPrjsFileW = widget;
  else if(*tag==BATCHPFTSFILE)
    batchPftsFileW = widget;
  else if(*tag==NEWMINDENS)
    newMinDensTextW= widget;
  else if(*tag==NEWMAXDENS)
    newMaxDensTextW = widget;
  else if(*tag==RESMINTEXT)
    resMinTextW = widget;
  else if(*tag==RESMAXTEXT)
    resMaxTextW = widget;
  else if(*tag==CUTMINTEXT)
    cutMinTextW = widget;
  else if(*tag==CUTMAXTEXT)
    cutMaxTextW = widget;
  else if(*tag==BINTEXT)
    {binTextW = widget;
     handle_cr(binTextW);}
  else if(*tag==STEPTEXT)
    {stepTextW = widget;
     handle_cr(stepTextW);}
  else if(*tag==PPINTEXT)
    {ppinTextW = widget;
     handle_cr(ppinTextW);}
  else if(*tag==STRTXPP)
    {strtXPPW = widget;
     handle_cr(strtXPPW);}
  else if(*tag==ENDXPP)
    {endXPPW = widget;
     handle_cr(endXPPW);}
  else if(*tag==MAXYPP)
    {maxYPPW = widget;
     handle_cr(maxYPPW);}
  else if(*tag==PPMIN)
    {ppMinTextW = widget;
     handle_cr(ppMinTextW);}
  else if(*tag==PPMAX)
    {ppMaxTextW = widget;
     handle_cr(ppMaxTextW);}
  else if(*tag==FILE1TEXT)
    subtractFile1TextW = widget;
  else if(*tag==FILE2TEXT)
    {
      subtractFile2TextW = widget;
      handle_cr(widget);
    }
  else if(*tag==DELTAX)
    deltaXpartprocW = widget;
  else if(*tag==DELTAY)
    deltaYpartprocW = widget;
  else if(*tag==DELTAOMEGA)
    deltaOmegaPPW = widget;
  else if(*tag==MANIPPPVALUE)
    manipPPvalueW = widget;
  else if(*tag==NORMALMEAN)
    normalMeanW = widget;
  else if(*tag==NORMALSTD)
    normalStdW = widget;
  else if(*tag==SUBMAGSTART)
    {subMagStartW = widget;
     handle_cr(widget);}
  else if(*tag==SUBMAGSTEP)
    subMagStepW = widget;
  else if(*tag==SUBMAGNUMSTEP)
    subMagNumStepW = widget;
  else if(*tag==SUBMAGNUMMINRAD)
    subMagMinRadW = widget;
  else if(*tag==SUBMAGNUMMAXRAD)
    subMagMaxRadW = widget;
  else if(*tag==SUBMAGNUMRADSTEP)
    subMagRadStepW = widget;
  else if(*tag==MAGSECTEXT)
    {magSecTextWidget = widget;
     handle_cr(widget);
   }
  else if(*tag==THRESHTEXT)
    {threshTextWidget = widget;
     handle_cr(widget);
   }
  else if(*tag==DIFFSECTEXT)
    {diffSecTextWidget = widget;
     handle_cr(widget);
   }
  else if(*tag==DIFFSEC2TEXT)
    {diffSec2TextWidget = widget;
     handle_cr(widget);
   }
  else if(*tag==ORDERBKG)
    {orderBkgW = widget;}
  else if(*tag==ORDERSIG)
    {orderSigW = widget;}
  else if(*tag==ADRHIRES)
    {adrHiResW = widget;}
  else if(*tag==ADRDEFOCUS)
    {adrDefocusW = widget;}
  else if(*tag==ADRCHISQ)
    {adrChisqW = widget;}
  else if(*tag==ORIGTEXT)
    {origTextW = widget;}
  else if(*tag==OUTPUTTEXT)
    {outputTextW = widget;}
  else if(*tag==REFLASTITER)
    {refLastIterTextW = widget;}
  else if(*tag==ADRA0)
    {adra0W = widget;}
  else if(*tag==ADRA1)
    {adra1W = widget;}
  else if(*tag==ADRA2)
    {adra2W = widget;}
  else if(*tag==ADRA3)
    {adra3W = widget;}
  else if(*tag==ADRA4)
    {adra4W = widget;}
  else if(*tag==ADRA5)
    {adra5W = widget;}
  else if(*tag==ADRB0)
    {adrb0W = widget;}
  else if(*tag==ADRB1)
    {adrb1W = widget;}
  else if(*tag==ADRB2)
    {adrb2W = widget;}
  else if(*tag==ADRB3)
    {adrb3W = widget;}
  else if(*tag==ADRB4)
    {adrb4W = widget;}
  else if(*tag==ADRB5)
    {adrb5W = widget;}
  else if(*tag==NFOLDTEXT)
    {nFoldTextW = widget;
     XtUnmanageChild(nFoldTextW);}
  else if(*tag==NEWXTEXT)
    {fftnewXtextW = widget;}
  else if(*tag==NEWYTEXT)
    {fftnewYtextW = widget;}
  else if(*tag==DIOTHRESHTEXT)
    {dioThreshTextW = widget;}
  else if(*tag==DIOSLIDETEXT)
    {dioSlideTextW = widget;}
  else if(*tag==REFCIRCAVG)
    {refCircAvgW = widget;}
  else if(*tag==REFAUTOMATE)
    {refAutomateW = widget;}
  else if(*tag==REFREADFILE)
    {refReadfileW = widget;}
  else if(*tag==RESET)
    {refResetW = widget;}
  else if(*tag==REGRID)
    {refRegridW = widget;}
  else if(*tag==A2DDRSCWIN)
    {a2ddrscwinW = widget;}
  else if(*tag==A2DDRFJ)
    {a2ddrfjW = widget;}
  else if(*tag==A2DDRFN)
    {a2ddrfnW = widget;}
  else if(*tag==A2DDRAG)
    {a2ddragW = widget;}
  else if(*tag==A2DDRTF)
    {a2ddrtfW = widget;}
  else if(*tag==A2DDRCSQ)
    {a2ddrcsqW = widget;}
  else if(*tag==MODELFILL)
    {modelFillW = widget;}
  else if(*tag==ROTATETTEXT)
    {mbRotatetTextW = widget;}
  else if(*tag==ROTATEPTEXT)
    {mbRotatepTextW = widget;}
  else if(*tag==ROTATEOTEXT)
    {mbRotateoTextW = widget;}
  else if(*tag==DIMENSIONXTEXT)
    {mbDimensionxTextW = widget;}
  else if(*tag==DIMENSIONYTEXT)
    {mbDimensionyTextW = widget;}
  else if(*tag==DIMENSIONZTEXT)
    {mbDimensionzTextW = widget;}
  else if(*tag==CENTERXTEXT)
    {mbCenterxTextW = widget;}
  else if(*tag==CENTERYTEXT)
    {mbCenteryTextW = widget;}
  else if(*tag==CENTERZTEXT)
    {mbCenterzTextW = widget;}
  else if(*tag==BINNINGTEXT)
    {binningTextW = widget;}



}/* TEXT_CREATE */




/*************************************************************
 * paramFile_browse_activate() - Read file from disk.
*/
void paramFile_browse_activate(Widget widget, int *tag, 
			       XmPushButtonCallbackStruct *callback_data)
{/* PARAMFILE_BROWSE_aCTIVATE */

  filterFiles("*",ALL); 
  
  /* Is it a PARAMFILE or PARAMMAP or PARAMSCAN */
  if(*tag==PARAMFILE)
    readType= PARAMFILE;
  else if(*tag==PARAMMAP)
    readType= PARAMMAP;
  else if(*tag==PARAMSCAN)
    readType= PARAMSCAN;
  else if(*tag==BATCHJOB)
    readType= BATCHJOB;

  readWrite= READING;

  raiseIt(fileSel_widget);

}/* PARAMFILE_BROWSE_aCTIVATE */






/*************************************************************
 * projContour_cr_activate() - Draw the contour on the display.
 */
void projContour_cr_activate(Widget widget, int *tag, 
			     XmAnyCallbackStruct *callback_data)
{/* projContour_cr_ACTIVATE */

  if(EQU(paramMapFilename,""))
    {/* Can't go on */
      askThem("No MAP open can't contour the projection.",
	      FALSE,XmDIALOG_INFORMATION);
    }/* Can't go on */
  else
    {/* Do it */
      stuffPackedImage(purdueHeader.numCols,
		       purdueHeader.numRows,
		       packImg,purdueHeader.nimg,
		       imgNum);
      DrawWindow(); 
    }/* Do it */
  
}/* projContour_cr_ACTIVATE */



/*************************************************************
 * paramScanFile_read_activate() - Show them the status screen
 * an dread the param file.
 */
void paramScanFile_read_activate(Widget widget, int *tag, 
				 XmAnyCallbackStruct *callback_data)
{/* PARAMSCANFILE_READ_ACTIVATE */
  char *paramText;

  raiseIt(statusScreenWidget);
  XFlush(dsply);
  XSync(dsply,0);

  paramText = XmTextGetString(paramScanFileTextW); 

  /* Now read the SCAN file */
  read_param_scan(paramText);

  XtFree(paramText);
  XFlush(dsply);

}/* PARAMSCANFILE_READ_ACTIVATE */



/*************************************************************
 * paramMapFile_read_activate() - Show them the status and read 
 * MAP listed in the param file.
 */
void paramMapFile_read_activate(Widget widget, int *tag, 
				XmAnyCallbackStruct *callback_data)
{/* PARAMMAPFILE_READ_ACTIVATE */
  char *paramText;

  raiseIt(statusScreenWidget);
  XFlush(dsply);
  XSync(dsply,0);

  paramText = XmTextGetString(paramMapFileTextW); 

  /* Now read the MAP file */
  read_param_map(paramText);

  XtFree(paramText);
  XFlush(dsply);

}/* PARAMMAPFILE_READ_ACTIVATE */



/*************************************************************
 * paramFile_read_activate() - Callback to actually read the
 * param file.
 */
void paramFile_read_activate(Widget widget, int *tag, 
			     XmAnyCallbackStruct *callback_data)
{/* PARAMFILE_READ_ACTIVATE */
  char *paramText;

  paramText = XmTextGetString(paramFileTextW); 
  read_param(paramText,TRUE);
  XtFree(paramText);
  XFlush(dsply);

}/* PARAMFILE_READ_ACTIVATE */



/******************************************************
 * write_param ()- Write the parameter file.
*/
void write_param(char *fName)
{/* WRITE_PARAM */
  int i,j,k;
  char *thisStr;
  P_INFO *thisInfo;
  FILE *fOut;

  if(haveParamFile)
    {/* We have one to do */

      fOut = fopen(fName,"r");
      if(fOut)
	{/* Exists */
	  fclose(fOut);
	  sprintf(outStr,"[%s] Already exists - Overwrite anyway?",fName);
	  if(!askThem(outStr,FALSE,XmDIALOG_QUESTION))
	    goto noWrite;
	}/* Exists */
      
      fOut = fopen(fName,"w");
      if(fOut)
	{/* OK to write */

	  /* Write packed boxes name */
	  fprintf(fOut,"%s\n",paramBoxesFilename);

	  /* Write second line */
	  thisStr= XmTextGetString(pixsizeW);
	  fprintf(fOut,"%s, ",thisStr);
	  XtFree(thisStr);
	  thisStr= XmTextGetString(unitsW);
	  fprintf(fOut,"%s, ",thisStr);
	  XtFree(thisStr);
	  thisStr= XmTextGetString(voltsW);
	  fprintf(fOut,"%s, ",thisStr);
	  XtFree(thisStr);
	  thisStr= XmTextGetString(ampfacsW);
	  fprintf(fOut,"%s, ",thisStr);
	  XtFree(thisStr);
	  thisStr= XmTextGetString(delfMajW);
	  fprintf(fOut,"%s, ",thisStr);
	  XtFree(thisStr);
	  thisStr= XmTextGetString(delfMinW);
	  fprintf(fOut,"%s, ",thisStr);
	  XtFree(thisStr);
	  thisStr= XmTextGetString(angMajW);
	  fprintf(fOut,"%s, ",thisStr);
	  XtFree(thisStr);
	  thisStr= XmTextGetString(csCoefW);
	  fprintf(fOut,"%s\n",thisStr);
	  XtFree(thisStr);
	  
	  /* Now write any untoggled particles */
	  thisInfo = info.info;
	  while(thisInfo != NULL)
	    {/* Do all */
	      if(!thisInfo->toggle)
		{/* OK to do */
		  fprintf(fOut,
	    "%7d,%8.3f,%8.3f,%8.3f,%8.3f,%7.3f,%6.3f,%6.3f,%6.3f,%6.3f\n",
			  thisInfo->particle_number,
			  thisInfo->theta,
			  thisInfo->phi,
			  thisInfo->omega,
			  thisInfo->x,
			  thisInfo->y,
			  thisInfo->mag,
			  thisInfo->cc1,
			  thisInfo->cc2,
			  thisInfo->cc3
			  );
		}/* OK to do */
	      else
		{/* NOT OK to do */
		  fprintf(fOut,
	    "%7d,%8.3f,%8.3f,%8.3f,%8.3f,%7.3f,%6.3f,%6.3f,%6.3f,%6.3f\n",
			  (0 - thisInfo->particle_number),
			  thisInfo->theta,
			  thisInfo->phi,
			  thisInfo->omega,
			  thisInfo->x,
			  thisInfo->y,
			  thisInfo->mag,
			  thisInfo->cc1,
			  thisInfo->cc2,
			  thisInfo->cc3
			  );
		}/* NOT OK to do */
	      thisInfo = thisInfo->next;
	    }/* Do all */

	  if(thisInfo==NULL)
	    {/* They called us from FFT screen - just default it */
	      int ll;
	      float factor,randNum;
	      float thisTheta, thisPhi, thisOmega;

	      /* Get our fraction of the random number */
	      factor= 1.0 / DENOMINATOR;
	      randNum= (random() * factor);

	      for(ll=0;ll<purdueHeader.nimg;ll++)
		{/* Do all */
		  randNum= (random() * factor);
		  thisTheta = 69.0 + (21.0 * randNum);
		  randNum= (random() * factor);
		  thisPhi = ((thisTheta - 69.0)*31.7/21.0) * randNum;

		  /* XDY wants some negative values for phi to
		   * fill the ASU ...  randNum is between 0.0 and 1.0 */
		  randNum= (random() * factor);
		  if(randNum > 0.5)
		    thisPhi *= -1.0;

		  randNum= (random() * factor);
		  thisOmega = 360.0 * randNum;
		  fprintf(fOut,
	    "%7d,%8.3f,%8.3f,%8.3f,%8.3f,%7.3f,%6.3f,%6.3f,%6.3f,%6.3f\n",
			  ll+1,
			  thisTheta,
			  thisPhi,
			  thisOmega,
			  (float)(purdueHeader.numCols/2),
			  (float)(purdueHeader.numRows/2),
			  1.000,
			  -99.9,
			  -99.9,
			  -99.9
			  );
		}/* Do all */
	    }/* They called us from FFT screen - just default it */

	  fclose(fOut);

	  sprintf(outStr,"[%s] file written.",fName);
	  askThem(outStr,FALSE,XmDIALOG_INFORMATION);
	}/* OK to write */
      else
	{/* error */
	  sprintf(outStr,"Can't open [%s] for writing.",fName);
	  askThem(outStr,FALSE,XmDIALOG_INFORMATION);
	}/* error */
    }/* We have one to do */
  else
    {/* error */
      sprintf(outStr,"No param file open.");
      askThem(outStr,FALSE,XmDIALOG_INFORMATION);
    }/* error */

 noWrite:;

}/* WRITE_PARAM */




/*********************************************************
 * read_batchjob ()- Read the batch file.  VMS scpecific.
*/
void read_batchjob(char *fName)
{/* READ_BATCHJOB */
  int i,beginParms;
  char junk;
  XmString simpleString;
  char inStr[5120];
  char parm1[32],parm2[32],parm3[32],parm4[32],
       parm5[32],parm6[32],parm7[32],parm8[32];
  forPifHead_t thisHead;
  fString_t rWrite,inFile;
  char *s1;
  FILE *fIn;
  
  XmTextSetString(batchjobFileTextW,fName); 
  sprintf(paramBatchJobFilename,"%s",fName);
  
  /* Open and read it */
  fIn= 0;
  fIn= fopen(fName,"r");
  if(fIn)
    {/* Read it */

      totParamFiles= 0;
      for(i=0;i<MAXPACKEDFILES;i++)
	{if(batchListStr[i])
	  free(batchListStr[i]);
	 batchListStr[i]= 0;}

      fread(inStr,sizeof(char),5100,fIn);
      fclose(fIn);      

      XmTextSetString(batchjobTextW,inStr); 

      /* Now read it one field at a time */
      fIn= fopen(fName,"r");

      XmListDeleteAllItems(batchParamFilesW);
      beginParms = 0;
      while(!feof(fIn))
	{/* DO all */
	  get_str(fIn, '\n', inStr);
	  if(inStr[0]=='e' &&
	     inStr[1]=='c' &&
	     inStr[2]=='h' &&
	     inStr[3]=='o')
	    {/* BeginParms */
	      if(beginParms == 0)
		{/* First line */
		  sscanf(inStr,"%s%s%s%s%s%s%s",
			 parm1,parm2,parm3,parm4,parm5,parm6,parm7);
		  XmTextSetString(batchModeW,parm2);                    
		  XmTextSetString(batchBinW,parm3);
		  XmTextSetString(batchSymW,parm4);
		  XmTextSetString(batchDtheW,parm5);
		  XmTextSetString(batchDphi90W,parm6);
		  XmTextSetString(batchIlistW,parm7);
		}/* First line */
	      else if(beginParms == 1)
		{/* Second line */
		  sscanf(inStr,"%s%s%s%s%s%s%s",
			 parm1,parm2,parm3,parm4,parm5,parm6,parm7);
		  XmTextSetString(batchRadLoW,parm2);
		  XmTextSetString(batchRadHiW,parm3);
		  XmTextSetString(batchRadStepW,parm4);
		  XmTextSetString(batchAnnLoW,parm5);
		  XmTextSetString(batchAnnHiW,parm6);
		}/* Second line */
	      else if(beginParms == 2)
		{/* line 3 */
		  sscanf(inStr,"%s%s%s%s%s%s%s",
			 parm1,parm2,parm3,parm4,parm5,parm6,parm7);
		  XmTextSetString(batchResLoW,parm2);
		  XmTextSetString(batchResHiW,parm3);
		  XmTextSetString(batchJCutW,parm4);
		  XmTextSetString(batchSigW,parm5);
		}/* line 3 */
	      else if(beginParms == 3)
		{/* line 4 */
		  sscanf(inStr,"%s%s%s%s%s%s%s",
			 parm1,parm2,parm3,parm4,parm5,parm6,parm7);
		  XmTextSetString(batchMagCenW,parm2);
		  XmTextSetString(batchMagStepW,parm3);
		  XmTextSetString(batchMagNumW,parm4);
		  XmTextSetString(batchMagNormW,parm5);
		}/* line 4 */
	      else if(beginParms == 4)
		{/* line 5 */
		  sscanf(inStr,"%s%s%s%s%s%s%s",
			 parm1,parm2,parm3,parm4,parm5,parm6,parm7);
		  XmTextSetString(batch3DMapW,parm2);
		  read_param_map(parm2);
		}/* line 5 */
	      else if(beginParms == 5)
		{/* line 6 */
		  sscanf(inStr,"%s%s%s%s%s%s%s",
			 parm1,parm2,parm3,parm4,parm5,parm6,parm7);
		  XmTextSetString(batchPrjsFileW,parm2);
		}/* line 6 */
	      else if(beginParms == 6)
		{/* line 7 */
		  sscanf(inStr,"%s%s%s%s%s%s%s",
			 parm1,parm2,parm3,parm4,parm5,parm6,parm7);
		  XmTextSetString(batchPftsFileW,parm2);
		}/* line 7 */
	      else if(beginParms >= 7)
		{/* line 8 thru ??? */
		  sscanf(inStr,"%s%s%s%s%s%s%s",
			 parm1,parm2,parm3,parm4,parm5,parm6,parm7);
		  if(!EQU(parm2,""))
		    {/* Do all non-nul */
		      createSimpleString(&simpleString, parm2);
		      XmListAddItem(batchParamFilesW, simpleString, 
				    ((beginParms-7)+1));  
		      XFlush(dsply);
		      XmStringFree(simpleString); 
		    }/* Do all non-nul */
		}/* line 8 thru ??? */

	      beginParms++;

	    }/* BeginParms */
	  
	}/* DO all */

      fclose(fIn);      

    }/* Read it */
  else 
    {/* error */
      XmTextSetString(batchjobFileTextW,""); 
      sprintf(paramBatchJobFilename,"");
      sprintf(outStr,"Can't open [%s]\n",fName);
      askThem(outStr,TRUE,XmDIALOG_INFORMATION);
      goto destructor;
    }/* error */
  
 destructor:;
  XtUnmanageChild(statusScreenWidget);
  XFlush(dsply);
  
}/* READ_BATCHJOB */



/******************************************************
 * read_param_scan ()- Read the scan file.
*/
void read_param_scan(char *fName)
{/* READ_PARAM_SCAN */
  int i;
  char junk;
  XmString simpleString;
  char inStr[512];
  forPifHead_t thisHead;
  fString_t rWrite,inFile;
  char *s1;
  FILE *fIn;
  
  if(!haveParamFile)
    {/* Must read a param file first */
      askThem("You must read a parameter file first.",
	      FALSE,
	      XmDIALOG_INFORMATION);
      goto destructor;
    }/* Must read a param file first */
  
  XmTextSetString(paramScanFileTextW,fName); 
  sprintf(paramScanFilename,"%s",fName);
  
  /* Open and read it */
  fIn= 0;
  fIn= fopen(fName,"r");
  if(fIn)
    {/* Read it */
      fclose(fIn);      
      i = 7;

      rWrite.addr = malloc(sizeof(char)*4);
      rWrite.length = 1;
      inFile.addr = malloc(sizeof(char)*64);
      inFile.length = 64;

      sprintf(inFile.addr,"%s",fName);
      sprintf(rWrite.addr,"r");

      pif_open_(inFile.addr, &i, rWrite.addr);
      
      free(rWrite.addr);
      free(inFile.addr);

      thisHead.luNum = i;
      thisHead.mode  = -99;
      pif_read_gh_(&thisHead);

      if(thisHead.mode != 0)
	{/* BAD mode */

	  XmTextSetString(paramScanFileTextW,""); 
	  sprintf(paramScanFilename,"");
	  sprintf(outStr,"Invalid PIF mode for scanned data [%s]\n",fName);
	  askThem(outStr,TRUE,XmDIALOG_INFORMATION);
	  
	  pif_close(&i);
	  goto destructor;

	}/* BAD mode */

      scanCols = thisHead.numCols;
      scanRows = thisHead.numRows;

      pif_close(&i);

    }/* Read it */
  else 
    {/* error */
      XmTextSetString(paramScanFileTextW,""); 
      sprintf(paramScanFilename,"");
      sprintf(outStr,"Can't open [%s]\n",fName);
      askThem(outStr,TRUE,XmDIALOG_INFORMATION);
      goto destructor;
    }/* error */
  
  s1 = strrchr(fName,'.');
  if(s1)
    *s1 = '\0';
  
  sprintf(paramBoxCoordsFilename,"%s.bcrd",fName);
  sprintf(outStr,"Reading [%s]",paramBoxCoordsFilename);
  if(checkStatus(outStr))
    goto destructor;
  readBoxCoords(paramBoxCoordsFilename);
  
  
 destructor:;
  XtUnmanageChild(statusScreenWidget);
  
}/* READ_PARAM_SCAN */



/******************************************************
 * read_param_map ()- Read the map file.
*/
void read_param_map(char *fName)
{/* READ_PARAM_MAP */
  int i;
  char junk;
  XmString simpleString;
  char inStr[512];
  char *s1;
  FILE *fIn;
  int strtCol,strtRow,strtSec;
  float a_star, b_star, c_star, alp_star, bet_star, gam_star,
        pixelSize;
  int pixSizUnits,ctfMode;
  float res_out,tempFactor,wiener;
  forPifHead_t thisHead;

  printf("Reading [%s]\n",fName);
  XmTextSetString(paramMapFileTextW,fName); 
  sprintf(paramMapFilename,"%s",fName);

  /* Open and read it */
  fIn= 0;
  fIn= fopen(fName,"r");
  if(fIn)
    {/* Read it */
      int i,j,k,l,m,n;
      int thisNimg,thisNcol,thisNrow,thisNsec;
      fString_t rWrite,inFile;
      int htype,imgOffset,convTsb;
      char versionStr[256],outStr[512], descriptionStr[100];
      XmString simpleString;
      fString_t descStr;
      float a_cell, b_cell,c_cell;
      int this_img_bkg,this_img_min,this_img_max,thisSymmetry;
      float this_img_avg,this_img_std;
      float aoverb,  map_abang;
      float map_scalef;
      int binCol,binRow,binSec;
      
      
      fclose(fIn);

      sprintf(outStr,"Reading [%s]",fName);
      if(checkStatus(outStr))
	goto destructor;

      if(projASU)
	{/* But *do* get rid of any unwanted memory */
	  for(i=0;i<purdueHeader.nimg;i++)
	    if(packImg[i])
	      {free(packImg[i]);
	       packImg[i]= NULL;}
	  projASU = FALSE;
	  purdueHeader.nimg = 0;
	  lastPackedNumImgs= 0;
	}/* But *do* get rid of any unwanted memory */

      thisHead.luNum= 4;
      rWrite.addr = malloc(sizeof(char)*4);
      rWrite.length = 1;
      descStr.addr = malloc(sizeof(char)*64);
      descStr.length = 32;
      inFile.addr = malloc(sizeof(char)*64);
      inFile.length = 64;
      sprintf(inFile.addr,"%s",fName);
      sprintf(rWrite.addr,"r");

      pif_open_(inFile.addr, &(thisHead.luNum), rWrite.addr);
      
      pif_read_gh(&thisHead);
      pifMode = thisHead.mode;
      thisNcol = thisHead.numCols;
      thisNrow = thisHead.numRows;
      thisNsec = thisHead.numSecs;
      
      if(pifMode!=20)
	{/* NOt a map!! */
	  sprintf(outStr,"Invalid MAP mode [%s]\n",fName);
	  askThem(outStr,FALSE,XmDIALOG_INFORMATION);
	  goto destructor;
	}/* NOt a map!! */
      
      for(j=0;j<1;j++)
	{/* Read all iamges */
	  if(pifMode==20)
	    {/* MAP file */
	      int thisImg;
	      thisImg = j + 1;  /* For fortran */
	      thisHead.imgNum = thisImg;
	      pif_read_dh(&thisHead);
	      imgOffset = thisHead.imgOffset;
	      recipFloatScaleFactor = thisHead.recipFloatScaleFactor;

	      mrc_alpha = (int) thisHead.alp;
	      mrc_beta = (int) thisHead.bet;
	      mrc_gamma = (int) thisHead.gam;
	      fileBinSecs= binSec  = thisNsec;
	      fileBinRows= binRow  = thisNrow;
	      fileBinCols= binCol  = thisNcol;
	      fileBinFactor = binFactor= thisHead.binFactor;
	      /* Normal */
	      if(binFactor==2)
		{/* 1/2 */
		  if(pifMode==20)
		    binSec  = thisNsec/2+1;  
		  binRow  = thisNrow/2+1;
		  binCol  = thisNcol/2+1;
		  if((binSec%2)==0)
		    binSec--;
		  if((binRow%2)==0)
		    binRow--;
		  if((binCol%2)==0)
		    binCol--;
		}/* 1/2 */
	      
	      if(float_image_buffer1 != NULL)
		free(float_image_buffer1);

	      CK(float_image_buffer1 = (float *)malloc(sizeof(float)*
						       binCol*binRow*binSec));
	      
	      maxMapPixel= -99999999.9;
	      minMapPixel= 99999999.9;
              maxMapPixel = -HUGE;  /* dgm -HUGE instead of -99999.9 */
              minMapPixel = HUGE;   /* dgm HUGE instead of 9999999.9 */
	      convTsb = TRUE;
	      pif_tsbworld(convTsb);
	      if(pifMode==20)
		{/* read it */
		  binFactor = 1;
		  strtCol = strtRow = strtSec = 1;
		  pif_read_map_short_image_(&(thisHead.luNum), 
					    float_image_buffer1,
					    &imgOffset,
					    &binCol, &binRow, &binSec,
					    &recipFloatScaleFactor,
					    &minMapPixel,&maxMapPixel, 
					    &binFactor,
					    &strtCol, &strtRow, &strtSec,
					    &binCol, &binRow, &binSec);
		}/* read it */
	      else
		{/* Error */
		  askThem("Bad mode for MAP.",FALSE,XmDIALOG_INFORMATION);
		}/* Error */
	      
	      pif_close_(&(thisHead.luNum));
	      
	      thisNcol = binCol;
	      thisNrow = binRow;
	      thisNsec = binSec;
	      origNumSecs= thisNsec;
	      
	    }/* MAP file */
	  
	}/* Read all iamges */
      
      free(rWrite.addr);
      free(descStr.addr);
      free(inFile.addr);

      xCen = (double)(thisNcol/2);
      yCen = (double)(thisNrow/2);
      zCen = (double)(origNumSecs/2);
      
      largestDim = thisNcol;
      largestDim = MAX(largestDim,thisNrow);
      largestDim = MAX(largestDim,origNumSecs);
      
      mapCols = binCol;
      mapRows = binRow;
      mapSecs = binSec;

    }/* Read it */
  else 
    {/* error */
      sprintf(outStr,"Can't open [%s]\n",fName);
      askThem(outStr,TRUE,XmDIALOG_INFORMATION);
    }/* error */

 destructor:;
  XtUnmanageChild(statusScreenWidget);

}/* READ_PARAM_MAP */



/******************************************************
 * read_param ()- Read the parameter file.
*/
void read_param(char *fName, int readPacked)
{/* READ_PARAM */
  int i,tiltFormat;
  char junk,sysCall[256];
  XmString simpleString;
  char inStr[512],tFname[512];
  char *s1;
  FILE *fIn;
  
  printf("Reading [%s]\n",fName);
  XmTextSetString(paramFileTextW,fName); 
  sprintf(inParamFilename,"%s",fName);

  /* Open and read it */
  fIn= 0;
  fIn= fopen(fName,"r");
  if(fIn)
    {/* Read it */

      /* Get packed boxes filename */
      /* What are we passed? */
      readPackedFile = readPacked;
      fscanf(fIn,"%s",inStr);
      tiltFormat = FALSE;
      if(EQU(inStr,"TILT_FORMAT"))
	{readPackedFile = FALSE;
	 tiltFormat = TRUE;}
      if(!EQU(inStr,"") && readPackedFile)
	{/* Ok to read */
	  
	  sprintf(tFname,"%s",inStr);
	  sprintf(fileName.addr,"%s",tFname);
	  sprintf(paramBoxesFilename,"%s",tFname);
	  raiseIt(statusScreenWidget);
	  XFlush(dsply);
	  XSync(dsply,0);
	  
	  sprintf(outStr,"Reading [%s]",fileName.addr);
	  if(checkStatus(outStr))
	    goto destructor;
	      
	  /* Remember this for the next time */
	  if(numBoxes > 0)
	    lastNumBox = (lastNumBox + numBoxes);

	  /* Don't do this - we need to be able to keep
	   * MAPs and such around etc... */
	  /* Init for this new image */
	  /* groundZero(); */

	  if(projASU)
	    {/* But *do* get rid of any unwanted memory */
	      for(i=0;i<purdueHeader.nimg;i++)
		if(packImg[i])
		  {free(packImg[i]);
		   packImg[i]= NULL;}
	      projASU = FALSE;
	      purdueHeader.nimg = 0;
	    }/* But *do* get rid of any unwanted memory */

	  /* Suppose they've got a previous packed file read */
	  for(i=0;i<purdueHeader.nimg;i++)
	    if(packImg[i])
	      {free(packImg[i]);
	       packImg[i]= NULL;}
	  purdueHeader.nimg = 0;

	  /* Tell them what we're doing */
	  sprintf(outStr,"%d",lastNumBox+1);
	  XmTextSetString(boxInitWidget, outStr);
	  XFlush(dsply);

	  /* Remember what our filename was - especially if we had to 

	   * change it to portable or etc...
	   */
	  sprintf(image_fileName.addr,"%s",fileName.addr);


	  /* Let's see if we can't just use our own */
	  mainRobem_sub_UNIX(&int_image_buffer1, 
			     &fileName, &i, &j);

	  /* Ooops no image get out of here */
	  if(i==0 || j==0)
	    goto destructor;

	  /* Give feedback on main window title bar */
	  if(!masterProcess)
	    sprintf(outStr,"Robem: [%s][%d][%d] - %d images total. [%s]",
		    fileName.addr,
		    purdueHeader.numCols,
		    purdueHeader.numRows,
		    purdueHeader.nimg,"CHILD");
	  else
	    sprintf(outStr,"Robem: [%s][%d][%d] - %d images total.",
		    fileName.addr,
		    purdueHeader.numCols,
		    purdueHeader.numRows,
		    purdueHeader.nimg);
	  createSimpleString(&simpleString,outStr);
	  XtSetArg(arglst[0], XmNdialogTitle, simpleString );
	  XtSetArg(arglst[1], XmNrecomputeSize, FALSE );
	  XtSetValues(main_window_widget, arglst, 2); 
	  XFlush(dsply);
	  XmStringFree(simpleString); 
	  XFlush(dsply);

	  /* Remember how big we are on disk */
	  file_width = i;
	  file_height = j;
	  lastPackedNumImgs = purdueHeader.nimg;

	  pickAutoZoom(i,j);

	  /* Now clip this and only get 1k x 1k maximum at a time */
	  i= defDisplayBufferSize;
	  j= defDisplayBufferSize;

	  /* Point to where the display data is */
	  image_image1->data = (char *)image_buffer1;
	  
	}/* Ok to read */

      sprintf(outStr,"Reading parameters");
      if(checkStatus(outStr))
	goto destructor;

      /* Get second line */
      fscanf(fIn,"%s",inStr);
      if(s1=strrchr(inStr,','))
	*s1= '\0';
      XmTextSetString(pixsizeW,inStr);
      pixSizeP= (float)strtod(inStr,0);
      fscanf(fIn,"%s",inStr);
      if(s1=strrchr(inStr,','))
	*s1= '\0';
      XmTextSetString(unitsW,inStr);
      unitsP= (int)strtod(inStr,0);
      fscanf(fIn,"%s",inStr);
      if(s1=strrchr(inStr,','))
	*s1= '\0';
      XmTextSetString(voltsW,inStr);
      volts= (float)strtod(inStr,0);
      fscanf(fIn,"%s",inStr);
      if(s1=strrchr(inStr,','))
	*s1= '\0';
      XmTextSetString(ampfacsW,inStr);
      ampFacs= (float)strtod(inStr,0);
      fscanf(fIn,"%s",inStr);
      if(s1=strrchr(inStr,','))
	*s1= '\0';
      XmTextSetString(delfMajW,inStr);
      delf_maj= (float)strtod(inStr,0);
      fscanf(fIn,"%s",inStr);
      if(s1=strrchr(inStr,','))
	*s1= '\0';
      XmTextSetString(delfMinW,inStr);
      delf_min= (float)strtod(inStr,0);
      fscanf(fIn,"%s",inStr);
      if(s1=strrchr(inStr,','))
	*s1= '\0';
      XmTextSetString(angMajW,inStr);
      ang_maj= (float)strtod(inStr,0);
      fscanf(fIn,"%s",inStr);
      if(s1=strrchr(inStr,','))
	*s1= '\0';
      XmTextSetString(csCoefW,inStr);
      cs_coeff= (float)strtod(inStr,0);
      /* printf("%f  %f  %f  %f  %f  %f  %f\n",
	     pixSizeP,volts,ampFacs,delf_maj,delf_min,ang_maj,cs_coeff); */

      rewind(fIn);
      if(readPackedFile || tiltFormat)
	{/* Init for new file */
	  Free_Particle_Info(&info);
	  deltaXmin = deltaYmin = deltaMin = 99999999.9;
	  deltaXmax = deltaYmax = deltaMax = -99999999.9;
	}/* Init for new file */
      Get_Particle_Info(&info,fIn,purdueHeader.numCols,purdueHeader.numRows);
      find_nearest_neighbors(&info);
      /* Why do we have to add missing particles ???
       * add_missing_particles(&info, purdueHeader.nimg,
       * purdueHeader.numCols,purdueHeader.numRows);
       */

      /* We've got one now */
      haveParamFile = TRUE;
	
      if(readPackedFile)
	stuffPackedImage(purdueHeader.numCols,
			 purdueHeader.numRows,
			 packImg,purdueHeader.nimg,
			 1);  /* Start at zero'th(1) image */

      DrawICOSWindow();

    }/* Read it */
  else 
    {/* error */
      sprintf(outStr,"Can't open [%s]\n",fName);
      askThem(outStr,TRUE,XmDIALOG_INFORMATION);
    }/* error */

 destructor:;
  XtUnmanageChild(statusScreenWidget);
  fclose(fIn);

}/* READ_PARAM */



/***********************************************************
 * icosWindow_create() - Rememebr the widgets use din the 
 * draw the icosahedron interface.
*/
void icosWindow_create(Widget widget, int *tag, 
		       XmAnyCallbackStruct *callback_data)
{/* ICOSWINDOW_CREATE */
  
  /* Remember the widget number for later */
  if(*tag==MAINICOS)
    icosaWidget= widget;
  else if(*tag==INCDEC)
    incDecW= widget;
  else if(*tag==IDT)
    idtW= widget;
  else if(*tag==THETAT)
    thetatW= widget;
  else if(*tag==PHIT)
    phitW= widget;
  else if(*tag==OMEGAT)
    omegatW= widget;
  else if(*tag==XT)
    xtW= widget;
  else if(*tag==YT)
    ytW= widget;
  else if(*tag==MAGT)
    magtW= widget;
  else if(*tag==PFTCC)
    pftccW= widget;
  else if(*tag==PRJCC)
    prjccW= widget;
  else if(*tag==CMPCC)
    cmpccW= widget;
  else if(*tag==DELTAXT)
    deltaXTW= widget;
  else if(*tag==DELTAYT)
    deltaYTW= widget;
  else if(*tag==ABSDELTAXT)
    absDeltaXTW= widget;
  else if(*tag==ABSDELTAYT)
    absDeltaYTW= widget;
  else if(*tag==DISTANCET)
    distanceTW= widget;
  
}/* ICOSWINDOW_CREATE */




/*************************************************************
 * unitCell_background() - Draw the basic background for
 * the unit cell drawing later.
 */
void unitCell_background()
{/* UNITCELL_BACKGROUND */
  P_INFO *thisInfo;
  int thisX,thisY,thisXp1,thisYp1;
  float thisTheta, thisPhi, negPhiMax;
  char outStr[256];
  
  thisTheta = 90.0;

  find_phimax_(&symmetry, &thisTheta, &phiMax);

  negPhiMax = 0.0 - phiMax;

  thisPhi = 0.0;

  get_t1t2p1_(&symmetry, &theta1, &theta2, &thisPhi);


  if(symmetry==5)
    {/* Draw a diamond shape */
      /* Draw dotted lines to show plotting "space" */
      thisX = (int)0.0;
      thisY = (int)0.0;
      convertXY(&thisX, &thisY, 0.0, 0.0);
      thisXp1 = (int)negPhiMax;
      thisYp1 = (int)90.0;
      convertXY(&thisXp1, &thisYp1, negPhiMax, 90.0);
      XDrawLine(dsply,icosa_window1,image_gcDOT,
		(thisX),(thisY),
		(thisXp1),(thisYp1));
      thisX = (int)negPhiMax;
      thisY = (int)90.0;
      convertXY(&thisX, &thisY, negPhiMax, 90.0);
      thisXp1 = (int)0.0;
      thisYp1 = (int)180.0;
      convertXY(&thisXp1, &thisYp1, 0.0, 180.0);
      XDrawLine(dsply,icosa_window1,image_gcDOT,
		(thisX),(thisY),
		(thisXp1),(thisYp1));
      thisX = (int)0.0;
      thisY = (int)180.0;
      convertXY(&thisX, &thisY, 0.0, 180.0);
      thisXp1 = (int)phiMax;
      thisYp1 = (int)90.0;
      convertXY(&thisXp1, &thisYp1, phiMax, 90.0);
      XDrawLine(dsply,icosa_window1,image_gcDOT,
		(thisX),(thisY),
		(thisXp1),(thisYp1));
      thisX = (int)phiMax;
      thisY = (int)90.0;
      convertXY(&thisX, &thisY, phiMax, 90.0);
      thisXp1 = (int)0.0;
      thisYp1 = (int)0.0;
      convertXY(&thisXp1, &thisYp1, 0.0, 0.0);
      XDrawLine(dsply,icosa_window1,image_gcDOT,
		(thisX),(thisY),
		(thisXp1),(thisYp1));
    }/* Draw a diamond shape */
  else
    {/* Draw a triangle */
      /* Draw dotted lines to show plotting "space" */
      thisX = (int)negPhiMax;
      thisY = (int)theta2;
      convertXY(&thisX, &thisY, negPhiMax, theta2);
      thisXp1 = 0;
      thisYp1 = (int)theta1;
      convertXY(&thisXp1, &thisYp1, 0.0, theta1);
      XDrawLine(dsply,icosa_window1,image_gcDOT,
		(thisX),(thisY),
		(thisXp1),(thisYp1));
      thisX = (int)negPhiMax;
      thisY = (int)theta2;
      convertXY(&thisX, &thisY, negPhiMax, theta2);
      thisXp1 = (int)phiMax;
      thisYp1 = (int)theta2;
      convertXY(&thisXp1, &thisYp1, phiMax, theta2);
      XDrawLine(dsply,icosa_window1,image_gcDOT,
		(thisX),(thisY),
		(thisXp1),(thisYp1));
      thisX = 0;
      thisY = (int)theta1;
      convertXY(&thisX, &thisY, 0.0, theta1);
      thisXp1 = (int)phiMax;
      thisYp1 = (int)theta2;
      convertXY(&thisXp1, &thisYp1, phiMax, theta2);
      XDrawLine(dsply,icosa_window1,image_gcDOT,
		(thisX),(thisY),
		(thisXp1),(thisYp1));
    }/* Draw a triangle */

  /* Draw tick marks */
  thisX = (int)negPhiMax;
  thisY = 490;
  sprintf(outStr,"%d",thisX);
  convertXY(&thisX, &thisYp1, negPhiMax, 490.0);
                                       /* thisY already in dsply space */
  XDrawLine(dsply,icosa_window1,image_gcGREEN,
	    (thisX),(thisY),
	    (thisX),(thisY+5));
  XDrawString(dsply, icosa_window1, image_gcWHITE,
	      thisX, (thisY+15), outStr, strlen(outStr));

  thisX = 0;
  sprintf(outStr,"%d",thisX);
  convertXY(&thisX, &thisYp1, 0.0, 490.0);
                                     /* thisY already in dsply space */
  XDrawLine(dsply,icosa_window1,image_gcGREEN,
	    (thisX),(thisY),
	    (thisX),(thisY+5));
  XDrawString(dsply, icosa_window1, image_gcWHITE,
	      thisX, (thisY+15), outStr, strlen(outStr));

  thisX = (int)phiMax;
  sprintf(outStr,"%d",thisX);
  convertXY(&thisX, &thisYp1, phiMax, 490.0);
                                    /* thisY already in dsply space */
  XDrawLine(dsply,icosa_window1,image_gcGREEN,
	    (thisX),(thisY),
	    (thisX),(thisY+5));
  XDrawString(dsply, icosa_window1, image_gcWHITE,
	      thisX, (thisY+15), outStr, strlen(outStr));

  thisX = YAXIS;
  thisY = (int)theta1;
  sprintf(outStr,"%d",thisY);
  convertXY(&thisXp1, &thisY, (float)YAXIS, theta1);
                                      /* thisX already in dsply space */
  XDrawLine(dsply,icosa_window1,image_gcGREEN,
	    (thisX-5),(thisY),
	    (thisX),(thisY));
  XDrawString(dsply, icosa_window1, image_gcWHITE,
	      (thisX-YAXIS), (thisY), outStr, strlen(outStr));
  thisY = (int)theta2;
  sprintf(outStr,"%d",thisY);
  convertXY(&thisXp1, &thisY, (float)YAXIS, theta2);  
                                 /* thisX already in dsply space */
  XDrawLine(dsply,icosa_window1,image_gcGREEN,
	    (thisX-5),(thisY),
	    (thisX),(thisY));
  XDrawString(dsply, icosa_window1, image_gcWHITE,
	      (thisX-YAXIS), (thisY), outStr, strlen(outStr));


}/* UNITCELL_BACKGROUND */






/*************************************************************
 * unitCell_work() - Do what it takes to get the unit cell
 * drawn correctly on the display.
 */
void unitCell_work(Widget widget, int *tag, 
		   XmAnyCallbackStruct *callback_data)
{/* unitCell_work */
  P_INFO *thisInfo;
  int thisX,thisY,thisXp1,thisYp1;
  float thisTheta, thisPhi, negPhiMax;
  char outStr[256];
  
  unitCell_background();

  thisInfo = info.info;
  while(thisInfo != NULL)
    {/* Do all of them */
      thisX = (int)thisInfo->phi;
      thisY = (int)thisInfo->theta;
      convertXY(&thisX, &thisY, thisInfo->phi, thisInfo->theta);
      
      if (!thisInfo->toggle) 
	{/* OK to draw normally */
	  /* Now draw an "X" */
	  XDrawLine(dsply,icosa_window1,image_gcRED,
		    (thisX - 5),(thisY-5),
		    (thisX + 5),(thisY+5));
	  XDrawLine(dsply,icosa_window1,image_gcRED,
		    (thisX + 5),(thisY-5),
		    (thisX - 5),(thisY+5));
	}/* OK to draw normally */
      else if(thisInfo->neighbors > 1)
	{/* Draw in GREEN */
	  /* Now draw an "X" */
	  XDrawLine(dsply,icosa_window1,image_gcGREEN,
		    (thisX - 5),(thisY-5),
		    (thisX + 5),(thisY+5));
	  XDrawLine(dsply,icosa_window1,image_gcGREEN,
		    (thisX + 5),(thisY-5),
		    (thisX - 5),(thisY+5));
	}/* Draw in GREEN */
      else
	{/* Draw in BLUE */
	  /* Now draw an "X" */
	  XDrawLine(dsply,icosa_window1,image_gcBLUE,
		    (thisX - 5),(thisY-5),
		    (thisX + 5),(thisY+5));
	  XDrawLine(dsply,icosa_window1,image_gcBLUE,
		    (thisX + 5),(thisY-5),
		    (thisX - 5),(thisY+5));
	}/* Draw in BLUE */
      if(thisInfo->neighbors > 1)
	{/* Feedback how many neighbors */
	  sprintf(outStr,"%d",thisInfo->neighbors);
	  XDrawString(dsply, icosa_window1, image_gcGREEN,
		      (thisX-5), (thisY+5), outStr, strlen(outStr));
	}/* Feedback how many neighbors */

      if(thisInfo==foundPart)
	{/* Indicate this is the current one found */
	  
	  /* Draw a box around the one we found */
	  XDrawLine(dsply,icosa_window1,image_gcWHITE,
		    (thisX - 5),(thisY-5),
		    (thisX + 5),(thisY-5));
	  XDrawLine(dsply,icosa_window1,image_gcWHITE,
		    (thisX - 5),(thisY+5),
		    (thisX + 5),(thisY+5));
	  XDrawLine(dsply,icosa_window1,image_gcWHITE,
		    (thisX - 5),(thisY-5),
		    (thisX - 5),(thisY+5));
	  XDrawLine(dsply,icosa_window1,image_gcWHITE,
		    (thisX + 5),(thisY-5),
		    (thisX + 5),(thisY+5));
	}/* Indicate this is the current one found */

      thisInfo = thisInfo->next;
    }/* Do all of them */

  if(drawIcos && foundPart!=NULL)
    {/* Draw this icosahedron */
      float thisTheta,thisPhi,thisOmega;

      thisTheta = foundPart->theta;
      thisPhi   = foundPart->phi;
      if(findNearest && useOmega)
	{/* Use real omega value - gotten above */
	  thisOmega = foundPart->omega;
	}/* Use real omega value - gotten above */
      else thisOmega = 0.0;

      /*DrawIcosahedron(thisTheta,thisPhi,thisOmega);  */
      DrawIcosahedron(deg2rad(thisTheta),
		      deg2rad(thisPhi),
		      deg2rad(thisOmega),
		      80, 0, SCALE, icosa_window1); 

    }/* Draw this icosahedron */

}/* unitCell_work */






/*************************************************************
 * xyPlot_background() - Draw the generic background for the
 * xy plot.
 */
void xyPlot_background()
{/* XYPLOT_BACKGROUND */
  P_INFO *thisInfo;
  int thisX,thisY,thisXp1,thisYp1,i,j;
  char outStr[256];
  
  /* Draw dotted lines to show plotting "space" */
  thisX = 0;
  thisY = 256;
  thisXp1 = 512;
  thisYp1 = 256;
  XDrawLine(dsply,icosa_window1,image_gcDOT,
	    (thisX),(thisY),
	    (thisXp1),(thisYp1));
  thisX = 256;
  thisY = 0;
  thisXp1 = 256;
  thisYp1 = 512;
  XDrawLine(dsply,icosa_window1,image_gcDOT,
	    (thisX),(thisY),
	    (thisXp1),(thisYp1));

  /* Draw tick marks */
  /* X Axis */
  allMax = MAX(deltaMax,(0.0-deltaMin));
  convertDeltaXY(&thisX,&thisY,allMax,0.0);
  sprintf(outStr,"%.1f",allMax);
  XDrawLine(dsply,icosa_window1,image_gcWHITE,
	    (thisX),(0),
	    (thisX),(512));
  XDrawString(dsply, icosa_window1, image_gcWHITE,
	      thisX+2, 256, outStr, strlen(outStr));
  convertDeltaXY(&thisX,&thisY,(0.0-allMax),0.0);
  sprintf(outStr,"%.1f",(0.0-allMax));
  XDrawLine(dsply,icosa_window1,image_gcWHITE,
	    (thisX),(0),
	    (thisX),(512));
  XDrawString(dsply, icosa_window1, image_gcWHITE,
	      thisX-30, 256, outStr, strlen(outStr));

  /* Y Axis */
  convertDeltaXY(&thisX,&thisY,0.0,allMax);
  sprintf(outStr,"%.1f",allMax);
  XDrawLine(dsply,icosa_window1,image_gcWHITE,
	    (0),(thisY),
	    (512),(thisY));
  XDrawString(dsply, icosa_window1, image_gcWHITE,
	      256, thisY-5, outStr, strlen(outStr));
  convertDeltaXY(&thisX,&thisY,0.0,(0.0-allMax));
  sprintf(outStr,"%.1f",(0.0-allMax));
  XDrawLine(dsply,icosa_window1,image_gcWHITE,
	    (0),(thisY),
	    (512),(thisY));
  XDrawString(dsply, icosa_window1, image_gcWHITE,
	      256,thisY+12, outStr, strlen(outStr));


  /* Now for Standard Deviation rings */
  sprintf(outStr,"STDev rings (X %.2f, Y %.2f)",deltaStdX,deltaStdY);
  XDrawLine(dsply,icosa_window1,image_gcGREY[100],
	    (50),(10),
	    (70),(10));
  XDrawString(dsply, icosa_window1, image_gcWHITE,
	      80,15, outStr, strlen(outStr));
  for(j=1;j<7;j++)
    {/* Draw the STD elipses */
      int x1,x2,y1,y2,minX,minY,maxX,maxY,beginEl,endEl;
      int drawX1[1024],drawY1[1024];
      int distAI,distBI;
      float distAsq,distBsq,distA,distB,cosAng1,sinAng1;
      float ang1,angThis,thisDeltaStdX,thisDeltaStdY;
      char numStr[64];

      thisDeltaStdX = deltaStdX*(float)j;
      thisDeltaStdY = deltaStdY*(float)j;
      if((thisDeltaStdX < allMax) &&
	 (thisDeltaStdY < allMax) )
	{/* OK to do */
	  convertDeltaXY(&distAI,&distBI,thisDeltaStdX,thisDeltaStdY);
	  distA = (float)fabs((float)(distAI-256));
	  distB = (float)fabs((float)(distBI-256));
	  
	  distBsq = distB * distB;
	  distAsq = distA * distA;
	  distAsq = MAX(distAsq,1.0); /* Can't divide by 0.0 */
	  ang1 = 0.0;
	  
	  beginEl= 512-((int)distA); /* We offset this by 256 pixels*/
	  endEl= 512+((int)distA);
	  beginEl = MAX(0,beginEl);
	  endEl = MIN(endEl,1023);
	  
	  
	  for(i=(beginEl);i<=(endEl);i++)
	    {/* Do all Y */
	      int thisX, thisY, thisI;
	      double tempF, thisR;
	      
	      thisI = i-256;  /* We're offsetting this by 256 pixels */
	      
	      tempF= (double)fabs(distBsq *
				  (1.0 - (
					  (float)((thisI-256)*
						  (thisI-256))/ 
					  distAsq)
				   )
				  );
	      tempF= (float)sqrt((double)tempF); 
	      
	      if(thisI==(beginEl-256) || (thisI==endEl-256))
		tempF= 0.0;
	      
	      if(thisI==256)
		angThis = (float)atan((double)
				      (
				       ((float)(tempF))/
				       ((float)(0.0000001))
				       )
				      );
	      else
		angThis = (float)atan((double)
				      (
				       ((float)(tempF))/
				       ((float)(256-thisI))
				       )
				      );
	      
	      thisX = 256 - thisI;
	      thisY = (int)tempF;
	      thisR = (double)sqrt((double)((thisX*thisX)+(thisY*thisY)));
	      
	      /* Now rotate in ang1 */
	      drawX1[i] = 256 + (int)(thisR * cos((double)(angThis-ang1)));
	      drawY1[i] = 256 + (int)(thisR * sin((double)(angThis-ang1)));
	      
	    }/* Do all Y */
	  
	  
	  /* Now draw them */
	  for(i=(beginEl);i<(endEl);i++)
	    {/* Do all Y */
	      /* i==256 draws a line down the middle */
	      if(i!=512)
		{/* Then draw them */
		  XDrawLine(dsply, icosa_window1, image_gcGREY[100],
			    (drawX1[i]),
			    (drawY1[i]),
			    (drawX1[i+1]),
			    (drawY1[i+1]));
		  XDrawLine(dsply, icosa_window1, image_gcGREY[100],
			    (256 + (256-drawX1[i])),
			    (256 - (drawY1[i]-256)),
			    (256 + (256-drawX1[i+1])),
			    (256 - (drawY1[i+1]-256)));  
		  XFlush(dsply);  
		}/* Then draw them */
	    }/* Do all Y */
	  
	  
	  XFlush (dsply);
	}/* OK to do */
      

    }/* Draw the STD elipses */



}/* XYPLOT_BACKGROUND */






/*************************************************************
 * xyPlot_work() - Do what it takes to get the x,y plot 
 * drawn correctly.
 */
void xyPlot_work(Widget widget, int *tag, 
		 XmAnyCallbackStruct *callback_data)
{/* XYPLOT_WORK */
  P_INFO *thisInfo;
  int thisX,thisY,thisXp1,thisYp1;
  float thisTheta, thisPhi, negPhiMax;
  char outStr[256];
  
  /* Recalculate the STD stats  (We may have changed one of the centers) */
  deltaXmin = deltaYmin = deltaMin = 99999999.9;
  deltaXmax = deltaYmax = deltaMax = -99999999.9;
  find_nearest_neighbors(&info);

  xyPlot_background();

  thisInfo = info.info;

  while(thisInfo != NULL)
    {/* Do all of them */

      convertDeltaXY(&thisX,&thisY,thisInfo->deltaX,thisInfo->deltaY);
      
      if (!thisInfo->toggle) 
	{/* OK to draw normally */
	  /* Now draw an "X" */
	  XDrawLine(dsply,icosa_window1,image_gcRED,
		    (thisX - 5),(thisY-5),
		    (thisX + 5),(thisY+5));
	  XDrawLine(dsply,icosa_window1,image_gcRED,
		    (thisX + 5),(thisY-5),
		    (thisX - 5),(thisY+5));
	}/* OK to draw normally */
      else
	{/* Draw in BLUE */
	  /* Now draw an "X" */
	  XDrawLine(dsply,icosa_window1,image_gcBLUE,
		    (thisX - 5),(thisY-5),
		    (thisX + 5),(thisY+5));
	  XDrawLine(dsply,icosa_window1,image_gcBLUE,
		    (thisX + 5),(thisY-5),
		    (thisX - 5),(thisY+5));
	}/* Draw in BLUE */

      if(thisInfo==foundPart)
	{/* Indicate this is the current one found */
	  
	  /* Draw a box around the one we found */
	  XDrawLine(dsply,icosa_window1,image_gcWHITE,
		    (thisX - 5),(thisY-5),
		    (thisX + 5),(thisY-5));
	  XDrawLine(dsply,icosa_window1,image_gcWHITE,
		    (thisX - 5),(thisY+5),
		    (thisX + 5),(thisY+5));
	  XDrawLine(dsply,icosa_window1,image_gcWHITE,
		    (thisX - 5),(thisY-5),
		    (thisX - 5),(thisY+5));
	  XDrawLine(dsply,icosa_window1,image_gcWHITE,
		    (thisX + 5),(thisY-5),
		    (thisX + 5),(thisY+5));
	}/* Indicate this is the current one found */

      thisInfo = thisInfo->next;
    }/* Do all of them */

}/* XYPLOT_WORK */





/*************************************************************
 * plotCavgPP_work() - DO whatever you need to get the cavg
 * plot on the display.
 */
void plotCavgPP_work(Widget widget, int *tag, 
		     XmAnyCallbackStruct *callback_data)
{/* PLOTCAVGPP_WORK */
  P_INFO *thisInfo;
  int thisX,thisY,thisXp1,thisYp1,caOffset, inum;
  float thisTheta, thisPhi, negPhiMax;
  float *thisRint,scaleCircFactor,minCA,maxCA,circAvgSize;
  float scaledY,scaledYp1;
  char outStr[256];
  
  if(imgNum > totCavgPPrint)
    {/* CAnt do it */
      sprintf(outStr,"No Circular Average computed yet.");
      XDrawString(dsply,icosa_window1,image_gcWHITE,
		  100,250,outStr,strlen(outStr));
      goto noCavgPlot;
    }/* CAnt do it */

  thisRint = (float *)cavgPPrint[imgNum-1];

  minCA = maxCA = thisRint[3];
  /* 2.0 for interp in img_sym() and 0.5 for radius not diameter 
   * 0.90 for 90% - we're plotting too much ... ??? */
  inum = (int)((2.0*0.5*0.90) * sqrt((double)
			  (purdueHeader.numCols * purdueHeader.numCols +
			   purdueHeader.numRows * purdueHeader.numRows
			   )
			  ));
  for(i=0;i<(int)(inum);i++)
    {/* get stats */
      minCA = MIN(minCA,thisRint[i]);
      maxCA = MAX(maxCA,thisRint[i]);
    }/* get stats */
  scaleCircFactor = 490.0 / (maxCA - minCA);
  circAvgSize = 1;  /* Get from interface later */
  caOffset = 0;

  for(i=0;i<(int)(inum);i++)
    {/* Plot them out */

	  scaledY = (thisRint[i]-minCA)*scaleCircFactor;
	  scaledYp1 = (thisRint[i+1]-minCA)*scaleCircFactor;
	  scaledY = (490.0 - scaledY);
	  scaledYp1 = (490.0 - scaledYp1);
	  thisX= ((i-caOffsetX)*circAvgSize)+22;
	  thisXp1= ((i-caOffsetX)*circAvgSize)+23;


	  XDrawLine(dsply,icosa_window1,image_gcGREEN,
		    (thisX),(int)(scaledY),
		    (thisXp1),(int)(scaledYp1));



    }/* Plot them out */

  noCavgPlot:;

}/* PLOTCAVGPP_WORK */





/**************************************************************
 *  Draw_Histogram() - Draws a histogram 
 */
void Draw_Histogram()
{ /* Draw_Histogram */
  char newVal[12];


  if (magtoggle) {
    Single_Line(MAGT,image_gcGREEN);
  } 
  if (cc1toggle) {
    Single_Line(PFTCC,image_gcBLUE);
  }
  if (cc2toggle) {
    Single_Line(PRJCC,image_gcGREEN);
  }
  if (cc3toggle) {
    Single_Line(CMPCC,image_gcRED);
  }

  /* Figure out how to do this ... 
  sprintf(newVal,"%d",(yscale));
  XmTextSetString(highTextWidget, newVal);
  */
  
} /* Draw_Histogram */ 




/*****************************************************
 * Single_Line() -
*/
void Single_Line(int whichParam,GC color)
{/* SINGLE_LINE */
  int startx,endx,starty,endy,i,j;
  float x;
  int badCount,lastBadCount;

  startx = YAXIS; starty = 0;
  endx = startx; endy = starty;
  lastBadCount = 0;
  for (x=startbox;x<endbox;x+=step) 
    {
      if(x==startbox)
	{/* Show them the value */
	  sprintf(outStr,"%.2f",x);
	  XDrawString(dsply,icosa_window1,image_gcWHITE,
		      20,505,outStr,strlen(outStr));
	}/* Show them the value */
      endy = find_number_in_range(info.info,x,x+step,whichParam,
				  &badCount);
      endy += badCount;
      for(i=0;i<pixelstep;i++)
	{/* Draw bars */
	  if(badCount > 0 || (endy==0 && lastBadCount != 0))
	    XDrawLine(dsply,icosa_window1,image_gcGREY[150],
		      startx+i,490-starty*yscale,
		      startx+i,490-endy*yscale);
	  else
	    XDrawLine(dsply,icosa_window1,color,
		      startx+i,490-starty*yscale,
		      startx+i,490-endy*yscale);
	  starty = endy;
	}/* Draw bars */
      if(endy > 0)
	lastBadCount = badCount;
      startx = endx;
      endx += (int)pixelstep;
      starty = endy;
    }

  /* Show them the end value */
  x=endbox;
  sprintf(outStr,"%.2f",x);
  XDrawString(dsply,icosa_window1,image_gcWHITE,
	      455,505,outStr,strlen(outStr));
  XFlush(dsply);

}/* SINGLE_LINE */
  



/****************************************************************
 * histogram_work() - The real work of drawing the histogram
 * happens here.
*/
void histogram_work()
{/* histogram_work */
  float y;
  int yStep;

  sprintf(outStr,"#");
  XDrawString(dsply, icosa_window1, image_gcWHITE,
	      500, 200, outStr, strlen(outStr));
  sprintf(outStr,"P");
  XDrawString(dsply, icosa_window1, image_gcWHITE,
	      500, 220, outStr, strlen(outStr));
  sprintf(outStr,"a");
  XDrawString(dsply, icosa_window1, image_gcWHITE,
	      500, 230, outStr, strlen(outStr));
  sprintf(outStr,"r");
  XDrawString(dsply, icosa_window1, image_gcWHITE,
	      500, 240, outStr, strlen(outStr));
  sprintf(outStr,"t");
  XDrawString(dsply, icosa_window1, image_gcWHITE,
	      500, 250, outStr, strlen(outStr));
  sprintf(outStr,"i");
  XDrawString(dsply, icosa_window1, image_gcWHITE,
	      500, 260, outStr, strlen(outStr));
  sprintf(outStr,"c");
  XDrawString(dsply, icosa_window1, image_gcWHITE,
	      500, 270, outStr, strlen(outStr));
  sprintf(outStr,"l");
  XDrawString(dsply, icosa_window1, image_gcWHITE,
	      500, 280, outStr, strlen(outStr));
  sprintf(outStr,"e");
  XDrawString(dsply, icosa_window1, image_gcWHITE,
	      500, 290, outStr, strlen(outStr));
  sprintf(outStr,"s");
  XDrawString(dsply, icosa_window1, image_gcWHITE,
	      500, 300, outStr, strlen(outStr));
  sprintf(outStr,"range");
  XDrawString(dsply, icosa_window1, image_gcWHITE,
	      250, 500, outStr, strlen(outStr));
  
  sprintf(outStr,"%1.0f",snumb);
  XDrawString(dsply, icosa_window1, image_gcWHITE,
	      YAXIS-10, 500, outStr, strlen(outStr));
  sprintf(outStr,"%1.0f",480.0/(float)yscale);
  XDrawString(dsply, icosa_window1, image_gcWHITE,
	      0, 22, outStr, strlen(outStr));
  
  yStep = (int)((480.0/yscale) / 10.0);
  yStep = MAX(yStep,1);
  /* Now draw grid */
  for(y=(480.0/yscale);y>0.0;y-=(float)yStep)
    {/* Do all */
      int stX,eX,stY,eY;
      stX = 24;
      eX = 480;
      stY = YAXIS;
      eY = YAXIS;
      eY= stY = 512 - (int)(YAXIS+(int)((float)480.0*
					((float)(y*yscale)/(float)480.0)));
      XDrawLine(dsply, icosa_window1, image_gcGREY[100],
		stX,stY,    /* startX, startY */
		eX,eY);
      if(y < (float)(480.0/yscale))
	{/* Don't do the first */
	  sprintf(outStr,"%.0f",y);
	  XDrawString(dsply, icosa_window1, image_gcWHITE,
		      1, eY, outStr, strlen(outStr));
	}/* Don't do the first */

    }/* Do all */

}/* histogram_work */





/******************************************************
 * DrawICOSWindow() - Redraw window e.g. on expose events
 * or whenever it's prudent.
 */
void DrawICOSWindow()
{/* DRAWICOSWINDOW */
  int draw_width, draw_height, i;
  Widget	widget= 0;
  int    *tag= 0;
  XmAnyCallbackStruct *callback_data= 0;
  
  /* Draw all of image */
  draw_width = 512;
  draw_height = 512;
  XPutImage(dsply, icosa_window1, image_gc, icosa_image1, 0, 0,
	    0, 0,  draw_width,  draw_height);
  XFlush (dsply);

  if(icoDispMode!=XYPLOT)
    {/* Draw axes */
      XDrawLine(dsply, icosa_window1, image_gcWHITE,
		YAXIS, YAXIS-10,               /* startX, startY */
		YAXIS, 490);            /* End X, endY */
      XDrawLine(dsply, icosa_window1, image_gcWHITE,
		YAXIS, 490,
		490, 490);
      
      for(i=0;i<512;i++)
	{/* DO the entire range */
	  if(i< (ico_toolsScrollValue*4.0))
	    XDrawLine(dsply, image_window1, image_gcRED,
		      (i+30), 70,
		      (i+30), 442);
	  
	}/* DO the entire range */
    }/* Draw axes */

  if(projASU)
    {/* Do the projASU stuff */
      XtUnmanageChild(ppMinMaxFormW);
      projASUunitCell_work(widget,tag,callback_data); 
    }/* Do the projASU stuff */
  else if(icoDispMode==XYPLOT) 
    {
      XtManageChild(ppMinMaxFormW);
      xyPlot_work(widget,tag,callback_data); 
    }
  else if(icoDispMode==PLOTCAVGPP) 
    {
      XtUnmanageChild(ppMinMaxFormW);
      plotCavgPP_work(widget,tag,callback_data); 
    }
  else if(icoDispMode==UNITCELL) 
    {
      XtUnmanageChild(ppMinMaxFormW);
      unitCell_work(widget,tag,callback_data); 
    }
  else if(icoDispMode==HISTOGRAM) 
    {

      XtUnmanageChild(ppMinMaxFormW);
      XtUnmanageChild(incDecW);
      XtUnmanageChild(idtW);
      XtUnmanageChild(thetatW);
      XtUnmanageChild(phitW);
      XtUnmanageChild(omegatW);
      XtUnmanageChild(xtW);
      XtUnmanageChild(ytW);
      XtUnmanageChild(deltaXTW);
      XtUnmanageChild(deltaYTW);
      XtUnmanageChild(absDeltaXTW);
      XtUnmanageChild(absDeltaYTW);
      XtUnmanageChild(distanceTW);
      XtManageChild(magtW);
      XtManageChild(pftccW);
      XtManageChild(prjccW);
      XtManageChild(cmpccW);

      XtManageChild(histControlsW);

      histogram_work();
      Draw_Histogram();
    }  
  else if(icoDispMode==ICOSORT) 
    {
      XtUnmanageChild(ppMinMaxFormW);
      XtManageChild(incDecW);
      XtManageChild(idtW);
      XtManageChild(thetatW);
      XtManageChild(phitW);
      XtManageChild(omegatW);
      XtManageChild(xtW);
      XtManageChild(ytW);
      XtManageChild(magtW);
      XtManageChild(pftccW);
      XtManageChild(prjccW);
      XtManageChild(cmpccW);
      XtManageChild(deltaXTW);
      XtManageChild(deltaYTW);
      XtManageChild(absDeltaXTW);
      XtManageChild(absDeltaYTW);
      XtManageChild(distanceTW);
      XtUnmanageChild(histControlsW);
      if(icoSortLevel==0)
	sprintf(outStr,"Choose Sort key.");
      else if(icoSortLevel==1)
	sprintf(outStr,"Sorting ...");
      else if(icoSortLevel==2)
	sprintf(outStr,"Sort Finished.");
      else
	sprintf(outStr,"Sort Mode");
      XDrawString(dsply, icosa_window1, image_gcWHITE,
		  200, 256, outStr, strlen(outStr));
    }
  
}/* DRAWICOSWINDOW */



/*************************************************************
 * unitCell_activate() - They want the unit cell displayed.
 */
void unitCell_activate(Widget widget, int *tag, 
		       XmAnyCallbackStruct *callback_data)
{/* UNITCELL_ACTIVATE */

  if(!findNearest)
    icosaMotioning = TRUE;
  XtUnmanageChild(parmTogsW);
  XtUnmanageChild(projasuW);
  projASU = FALSE;
  XtManageChild(parmheadW);
  icoDispMode = UNITCELL;
  DrawICOSWindow(); 

}/* UNITCELL_ACTIVATE */



/*************************************************************
 * ico_hist_activate() - They want the histogram displayed.
 */
void ico_hist_activate(Widget widget, int *tag, 
		       XmAnyCallbackStruct *callback_data)
{/* ICO_HIST_ACTIVATE */
  XmString simpleString;

  icosaMotioning = FALSE;

  sprintf(outStr,"Choose parameters for histogram: ");
  createSimpleString(&simpleString,outStr);
  XtSetArg(arglst[0], XmNlabelString, simpleString);
  XtSetValues(parmTogsLabelW, arglst, 1); 
  XFlush(dsply);
  XmStringFree(simpleString); 
  XFlush(dsply);

  XtManageChild(parmTogsW);
  XtUnmanageChild(parmheadW);
  projASU = FALSE;
  XtUnmanageChild(projasuW);
  sprintf(outStr,"Choose Histogram parameters.");
  XmTextSetString(partPointTextW,outStr);
  icoDispMode = HISTOGRAM;
  DrawICOSWindow();

}/* ICO_HIST_ACTIVATE */




/*************************************************************
 * xyPlot_activate() - They have chosen to have the x,y plot
 * displayed.
 */
void xyPlot_activate(Widget widget, int *tag, 
		     XmAnyCallbackStruct *callback_data)
{/* XYPLOT_ACTIVATE */
  float thisDistanceMax;

  if(!findNearest)
    icosaMotioning = TRUE;
  XtUnmanageChild(parmTogsW);
  XtUnmanageChild(projasuW);
  projASU = FALSE;
  XtManageChild(parmheadW);

  /* Let's do this by 1000's so we can get the granularity we
   * need for small absolute max distances */
  thisDistanceMax = distanceMax * 1000.0;
  minXYplot = 0.0;
  maxXYplot = distanceMax;
  sprintf(outStr,"%f",minXYplot);
  XmTextSetString(ppMinTextW,outStr);
  sprintf(outStr,"%f",maxXYplot);
  XmTextSetString(ppMaxTextW,outStr);
  XtVaSetValues(ppMinScrollW,XmNmaximum, (int)(thisDistanceMax),NULL);
  XtVaSetValues(ppMinScrollW,XmNminimum, (int)(0),NULL);
  XtVaSetValues(ppMinScrollW,XmNsliderSize, (int)(thisDistanceMax/20.0),NULL);
  XtVaSetValues(ppMinScrollW,XmNpageIncrement, 
		(int)(thisDistanceMax/20.0),NULL);
  XtVaSetValues(ppMinScrollW,XmNincrement, (int)(thisDistanceMax/500.0),NULL);
  XtVaSetValues(ppMinScrollW,XmNvalue, (int)(0),NULL);
		
  XtVaSetValues(ppMaxScrollW,XmNmaximum, (int)(thisDistanceMax*1.11),NULL);
  XtVaSetValues(ppMaxScrollW,XmNminimum, (int)(0),NULL);
  XtVaSetValues(ppMaxScrollW,XmNsliderSize, (int)(thisDistanceMax/20.0),NULL);
  XtVaSetValues(ppMaxScrollW,XmNpageIncrement, 
		(int)(thisDistanceMax/20.0),NULL);
  XtVaSetValues(ppMaxScrollW,XmNincrement, (int)(thisDistanceMax/500.0),NULL);
  XtVaSetValues(ppMaxScrollW,XmNvalue, (int)(thisDistanceMax+1),NULL);

  icoDispMode = XYPLOT;
  DrawICOSWindow(); 

}/* XYPLOT_ACTIVATE */





/*************************************************************
 * plotCavgPP_activate() - Setup for plotting the cavg.
 */
void plotCavgPP_activate(Widget widget, int *tag, 
			 XmAnyCallbackStruct *callback_data)
{/* PLOTCAVGPP_ACTIVATE */
  float thisDistanceMax;

  if(!findNearest)
    icosaMotioning = FALSE;

  icoDispMode = PLOTCAVGPP;
  DrawICOSWindow(); 

}/* PLOTCAVGPP_ACTIVATE */





/*************************************************************
 * DrawIcosahedron() - Actually do the icosahedron drawing.
*/
void DrawIcosahedron(float yrot,float xrot,float zrot, 
		     int xoffset, int yoffset, 
		     float scaleFactor, 
		     Window thisWin)  
{/* DRAWICOSAHEDRON */
  int i,drawIt;
  int sx,sy,ex,ey;
  GC thisGC;

  /* Let's rotate omega 90 degrees so it matches our convention */
  if(thisWin==icosa_window1)
    zrot += deg2rad(90.0);

  if(xoffset == 80)
    thisGC = image_gcRED;
  else if(thisWin==XtWindow(mapShadeScreenWindow)) 
    thisGC = image_gcRED;
  else thisGC = image_gcBLUE;

  /* Blank out last icosahedron */
  if(thisWin==icosa_window1)
    XPutImage(dsply, icosa_window1, icosa_gc, icosa_image1, 
	      300+xoffset,0,300+xoffset,0,
	      110, 110); 

  if(thisWin==XtWindow(mapShadeScreenWindow))
    {/* Init the display */
      sx = (int)xoffset - 35;
      ex = (int)xoffset + 35;
      for(i=0;i<70;i++)
	{/* Init */
	  sy = (int)(yoffset - 35)+i;
	  ey = (int)sy;
	  XDrawLine(dsply,thisWin,image_gcGREY[100],
		    sx,sy,ex,ey);
	}/* Init */
    }/* Init the display */

  /* Stuff the icosahedron values */
  get_verticies();

  for (i=0;i<icosabox.num_points;i++) {
    roty(&icosabox.vertex[i].lx,&icosabox.vertex[i].lz,yrot);
    rotx(&icosabox.vertex[i].ly,&icosabox.vertex[i].lz,xrot);
    rotz(&icosabox.vertex[i].lx,&icosabox.vertex[i].ly,zrot);  

  }
  if(thisWin==XtWindow(mapShadeScreenWindow))
    scale_verticies(120.0);
  else scale_verticies(scaleFactor*icosaScaleF);
  Convert3d_to_scrcords();
  if(thisWin==icosa_window1)
    translate_verticies(TRANSLATEX+xoffset, TRANSLATEY+yoffset);
  else if(thisWin==XtWindow(mapShadeScreenWindow))
    translate_verticies(xoffset, yoffset);
  else translate_verticies((largestDim/2)*sizeFactor,
			   (largestDim/2)*sizeFactor
			   );

  for (i=0;i<icosabox.num_faces;i++) 
    {

      /* We only want to draw the front half in main window */
      if(thisWin==icosa_window1)
	drawIt = TRUE;
      else if(thisWin==XtWindow(mapShadeScreenWindow))
	drawIt = TRUE;
      else if(icosabox.vertex[icosabox.faces[i].x].lz >= 0.0)
	drawIt = TRUE;
      else drawIt = FALSE;

      if(drawIt)
	{/* OK to draw */
	  sx = (int)icosabox.vertex[icosabox.faces[i].x].sx;
	  sy = (int)icosabox.vertex[icosabox.faces[i].x].sy;
	  ex = (int)icosabox.vertex[icosabox.faces[i].y].sx;
	  ey = (int)icosabox.vertex[icosabox.faces[i].y].sy;
	  XDrawLine(dsply,thisWin,thisGC,
		    sx,sy,ex,ey);
	  sx = (int)icosabox.vertex[icosabox.faces[i].y].sx;
	  sy = (int)icosabox.vertex[icosabox.faces[i].y].sy;
	  ex = (int)icosabox.vertex[icosabox.faces[i].z].sx;
	  ey = (int)icosabox.vertex[icosabox.faces[i].z].sy;
	  XDrawLine(dsply,thisWin,thisGC,
		    sx,sy,ex,ey);
	  sx = (int)icosabox.vertex[icosabox.faces[i].z].sx;
	  sy = (int)icosabox.vertex[icosabox.faces[i].z].sy;
	  ex = (int)icosabox.vertex[icosabox.faces[i].x].sx;
	  ey = (int)icosabox.vertex[icosabox.faces[i].x].sy;
	  XDrawLine(dsply,thisWin,thisGC,
		    sx,sy,ex,ey);
	}/* OK to draw */
    }
  
  if(thisWin==icosa_window1 && xoffset==0)
    {/* Draw theta phi coords */
      XDrawString(dsply, XtWindow(icosaWidget), image_gcBLUE,
		  310, 100, outStr, strlen(outStr));
      XFlush(dsply);
    }/* Draw theta phi coords */

}/* DRAWICOSAHEDRON */



/******************************************************
 * convert ThetaPhi() - Convert display coords into
 * theta and phi
*/
void convertThetaPhi(int x, int y, 
		     float *thisTheta, float *thisPhi)
{/* CONVERTTHETAPHI */
  float xFactor,yFactor;

  if(symmetry==5)
    {/* Do for symm==5 */
      float sinTheta,tempTheta;

      *thisPhi = (float)x;
      *thisTheta = (float)y;

      yFactor = (float)(21.0 * ((90.0 - 69.0)/(theta2 - theta1)));
      
      *thisTheta -= 20.0;
      *thisTheta /= yFactor;
      *thisTheta += theta1;

      if((*thisTheta)> 90.0)
	{tempTheta = 90.0 - ((*thisTheta)-90.0);}
      else tempTheta = *thisTheta;

      sinTheta = sin((double)deg2rad(tempTheta));

      xFactor = (float)((7.0*sinTheta) * (32.0 / phiMax));
      if(xFactor==0.0)
	xFactor = 0.000001;

      *thisPhi -= 256.0 ;
      *thisPhi /= xFactor;
      
    }/* Do for symm==5 */
  else
    {/* Everything else */
      *thisPhi = (float)x;
      *thisTheta = (float)y;
      xFactor = (float)(7.0 * (32.0 / phiMax));
      yFactor = (float)(21.0 * ((90.0 - 69.0)/(theta2 - theta1)));
      
      *thisPhi -= 256.0 ;
      *thisPhi /= xFactor;
      
      *thisTheta -= 20.0;
      *thisTheta /= yFactor;
      *thisTheta += theta1;
    }/* Everything else */

}/* CONVERTTHETAPHI */




/********************************************************************
 * find_nearest_neighbors() - Find how many particles are near this
 * one.
*/
void find_nearest_neighbors(INFO *head)
{/* FIND_NEAREST_NEIGHBORS */
  int thisX,thisY,thatX,thatY,distnce;
  P_INFO *thisInfo,*thatInfo;
  float totValuesY;
  double varY,sumsqY,avgY;
  int tot;
  float totValuesX;
  double varX,sumsqX,avgX;

  deltaStdY= deltaStdX = 0.0; 
  tot= 0; 
  totValuesX= totValuesY= 0.0;
  sumsqX = sumsqY= 0.0;
  thisInfo = head->info;
  while(thisInfo != NULL)
    {/* Do all of them */
      thisX = (int)thisInfo->phi;
      thisY = (int)thisInfo->theta;
      convertXY(&thisX,&thisY, thisInfo->phi, thisInfo->theta);
      thatInfo = head->info;
      thisInfo->neighbors = 0;
      while(thatInfo != NULL)
	{/* Find neighbors */
	  thatX = (int)thatInfo->phi;
	  thatY = (int)thatInfo->theta;
	  convertXY(&thatX,&thatY, thatInfo->phi, thatInfo->theta);
	  distnce = (thisX-thatX)*(thisX-thatX)+
	    (thisY-thatY)*(thisY-thatY);
	  if(distnce <= (1.0))
	    {/* Found a new candidate within 1 pixel */
	      thisInfo->neighbors++;
	    }/* Found a new candidate within 1 pixel */
	  thatInfo = thatInfo->next;
	}/* Find neighbors */

      /* While we're here let's calc the deltaXY stats */
      sumsqY += (thisInfo->deltaY * thisInfo->deltaY);
      totValuesY +=  (float)thisInfo->deltaY;
      sumsqX += (thisInfo->deltaX * thisInfo->deltaX);
      totValuesX +=  (float)thisInfo->deltaX;
      tot++;

      /* Recalc these just in case we've changed one */
      deltaMax = MAX(thisInfo->deltaX, deltaMax);
      deltaMax = MAX(thisInfo->deltaY, deltaMax);
      deltaMin = MIN(thisInfo->deltaX, deltaMin);
      deltaMin = MIN(thisInfo->deltaY, deltaMin);
      deltaXmin= MIN(thisInfo->deltaX, deltaXmin);
      deltaYmin= MIN(thisInfo->deltaY, deltaYmin);
      deltaXmax= MAX(thisInfo->deltaX, deltaXmax);
      deltaYmax= MAX(thisInfo->deltaY, deltaYmax);
      thisInfo->distance = sqrt((thisInfo->deltaX*thisInfo->deltaX)+
				(thisInfo->deltaY*thisInfo->deltaY));
      
      thisInfo = thisInfo->next;

    }/* Do all of them */

  /* Calc stats */
  avgY = (float)((float)totValuesY / (float)(tot));
  varY = (sumsqY / (float)(tot)) - (avgY * avgY);
  if(varY > 0.0)
    deltaStdY = (float)sqrt(varY);
  else deltaStdY = (float)sqrt(fabs(varY));

  avgX = (float)((float)totValuesX / (float)(tot));
  varX = (sumsqX / (float)(tot)) - (avgX * avgX);
  if(varX > 0.0)
    deltaStdX = (float)sqrt(varX);
  else deltaStdX = (float)sqrt(fabs(varX));


}/* FIND_NEAREST_NEIGHBORS */




/**********************************************************
 * findNearestParticle() - What's the nearest paritcle to
 * this one.  Used to find out which particle they want
 * detail printed.
*/
void findNearestParticle(int *x, int *y,
			 P_INFO **foundPart,
			 int toggleFound,
			 int makeThisParticle,
			 int showPackedBox,
			 int buttonNum)
{/* FINDNEARESTPARTICLE */
  int i,j,thisX,thisY,minX,minY;
  int distnce,minDist,minParticle;
  float minOmega;
  P_INFO *thisInfo;
  P_INFO *goodOnes[20];
  int numFound;
  static int lastParticle = -99;
  XmString simpleString;

  if(!readPackedFile)
    showPackedBox = FALSE;
  numFound = 0;
  minParticle = -99;
  minDist = 999999999;
  minDist = INT_MAX;  /* dgm INT_MAX instead of 99999999 */
  thisInfo = info.info;
  while(thisInfo != NULL)
    {/* Do all of them */

      if(icoDispMode==UNITCELL)
	{/* Unit Cell */
	  thisX = (int)thisInfo->phi;
	  thisY = (int)thisInfo->theta;
	  convertXY(&thisX, &thisY, thisInfo->phi, thisInfo->theta);
	}/* Unit Cell */
      else
	{/* XYPLOT */
	  thisX = (int)thisInfo->deltaX;
	  thisY = (int)thisInfo->deltaY;
	  convertDeltaXY(&thisX, &thisY, thisInfo->deltaX, thisInfo->deltaY);
	}/* XYPLOT */

      distnce = (thisX - *x)*(thisX - *x) +
	(thisY - *y)*(thisY - *y);
      if(makeThisParticle > 0)
	{/* They want to find it by particle number */
	  if(thisInfo->rank == makeThisParticle)
	    {/* Found it */
	      minDist = distnce;
	      minParticle = thisInfo->rank;
	      minX = thisX;
	      minY = thisY;
	      minOmega = thisInfo->omega;
	      *foundPart = thisInfo;
	    }/* Found it */
	}/* They want to find it by particle number */
      else if(distnce < minDist)
	{/* Found a new candidate */
	  minDist = distnce;
	  minParticle = thisInfo->rank;
	  minX = thisX;
	  minY = thisY;
	  minOmega = thisInfo->omega;
	  *foundPart = thisInfo;
	}/* Found a new candidate */
      
      thisInfo = thisInfo->next;
    }/* Do all of them */

  if(makeThisParticle <= 0)
    {/* Now try and find duplicates */
      thisInfo = info.info;
      while(thisInfo != NULL)
	{/* Do all of them */
	  thisX = (int)thisInfo->phi;
	  thisY = (int)thisInfo->theta;
	  convertXY(&thisX, &thisY, thisInfo->phi, thisInfo->theta);
	  
	  distnce = (thisX - minX)*(thisX - minX) +
	    (thisY - minY)*(thisY - minY);
	  if(distnce <= (1.0))
	    {/* Found a new candidate within 1 pixel */
	      goodOnes[numFound++] = thisInfo;
	    }/* Found a new candidate within 1 pixel */
	  thisInfo = thisInfo->next;
	}/* Do all of them */
    }/* Now try and find duplicates */

  if(buttonNum == Button3)
    {/* Just give them a report on what's what */
      char *s1,*s2;
      s1 = outStr;
      sprintf(s1,"Legend - \n\n\
     U  -  Particle thrown out in UnitCell display\n\
     X  -  Particle thrown out in XY Plot  display\n\
     T  -  Particle thrown out from thresholded XY Plot\n\
     M  -  Particle thrown out in Histogram display for bad Mag\n\
     F  -  Particle thrown out in Histogram display for bad PftCC\n\
     R  -  Particle thrown out in Histogram display for bad PrjCC\n\
     C  -  Particle thrown out in Histogram display for bad CmpCC\n\
     I  -  Particle thrown out in Input File\n\n");
      s2 = outStr + strlen(s1);
      s1 = s2;
      
      if(numFound > 1)
	{/* Do multiples */
	  for(i=0;i<numFound;i++)
	    {/* Do all */
	      sprintf(s1,
      "Part[%d] - Theta[%.3f]Phi[%.3f]Omega[%.3f]  [%c%c%c%c%c%c%c%c]\n",
		      goodOnes[i]->particle_number,
		      goodOnes[i]->theta,
		      goodOnes[i]->phi,
		      goodOnes[i]->omega,
		      (goodOnes[i]->toggle & UCBAD)?'U':' ',
		      (goodOnes[i]->toggle & XYBAD)?'X':' ',
		      (goodOnes[i]->toggle & XYBADTHRESH)?'T':' ',
		      (goodOnes[i]->toggle & HISTMAGBAD)?'M':' ',
		      (goodOnes[i]->toggle & HISTPFTBAD)?'F':' ',
		      (goodOnes[i]->toggle & HISTPRJBAD)?'R':' ',
		      (goodOnes[i]->toggle & HISTCMPBAD)?'C':' ',
		      (goodOnes[i]->toggle & INFILEBAD)?'I':' '
		      );
	      s2 = outStr + strlen(outStr);
	      s1 = s2;
	    }/* Do all */
	}/* Do multiples */
      else
	{/* Do single */
	  if(foundPart)
	    {/* DO it */
	      sprintf(s1,
	    "Part[%d] - Theta[%.3f]Phi[%.3f]Omega[%.3f]  [%c%c%c%c%c%c%c%c]\n",
		      (*foundPart)->particle_number,
		      (*foundPart)->theta,
		      (*foundPart)->phi,
		      (*foundPart)->omega,
		      ((*foundPart)->toggle & UCBAD)?'U':' ',
		      ((*foundPart)->toggle & XYBAD)?'X':' ',
		      ((*foundPart)->toggle & XYBADTHRESH)?'T':' ',
		      ((*foundPart)->toggle & HISTMAGBAD)?'M':' ',
		      ((*foundPart)->toggle & HISTPFTBAD)?'F':' ',
		      ((*foundPart)->toggle & HISTPRJBAD)?'R':' ',
		      ((*foundPart)->toggle & HISTCMPBAD)?'C':' ',
		      ((*foundPart)->toggle & INFILEBAD)?'I':' '
		      );
	    }/* DO it */
	  else sprintf(outStr,"No particle found.");
	}/* Do single */
      
      XmTextSetString(helpTextWidget,outStr);
      raiseIt(helpScreenWidget);
      
    }/* Just give them a report on what's what */
  else
    {/* normal */
      if(numFound > 1)
	{/* We've got to make them choose one */
	  int numWanted,saveIcosaMotioning;
	  char *s1,*s2;
	  
	  sprintf(outStr,"Multiple particles found.  Choose one:\n");
	  XmTextSetString(infoPromptTextW,outStr); 
	  
	  for(i=0;i<numFound;i++)
	    {/* Do all */
	      sprintf(outStr,
		      "Part[%d] - Theta[%.3f]Phi[%.3f]Omega[%.3f]  [%s]",
		      goodOnes[i]->particle_number,
		      goodOnes[i]->theta,
		      goodOnes[i]->phi,
		      goodOnes[i]->omega,
		      (goodOnes[i]->toggle)?"OFF":"ON");
	      createSimpleString(&simpleString, outStr);
	      XmListAddItem(infoListW, simpleString, i+1);
	      XFlush(dsply);
	      XmStringFree(simpleString); 
	    }/* Do all */
	  
	  saveIcosaMotioning = icosaMotioning;
	  icosaMotioning = FALSE;
	  sprintf(outStr,"Multiple particles found.  Choose one:");
	  numWanted= askThemAnswer(outStr,numFound);
	  /* Do we really want to turn this back on just yet??? */
	  icosaMotioning = saveIcosaMotioning;
	  *foundPart = goodOnes[numWanted];
	  thisX = (int)(*foundPart)->phi;
	  thisY = (int)(*foundPart)->theta;
	  convertXY(&thisX, &thisY, (*foundPart)->phi, (*foundPart)->theta);
	  minX = thisX;
	  minY = thisY;
	  minParticle = (*foundPart)->rank;
	  /* Let's toggle it here before we display it */
	}/* We've got to make them choose one */
      
      /* OK we know which one they want now so continue */
      *x = minX;
      *y = minY;
      if(toggleFound)
	{/* Toggle it back off */
	  if(icoDispMode==UNITCELL)
	    {/* Unit Cell display */
	      if((*foundPart)->toggle & UCBAD)
		{
		  (*foundPart)->toggle &= NOTUCBAD;
		}
	      else
		{
		  (*foundPart)->toggle |= UCBAD;
		}
	    }/* Unit Cell display */
	  else if(icoDispMode==XYPLOT)
	    {/* XY plot */
	      if((*foundPart)->toggle & XYBAD)
		{
		  (*foundPart)->toggle &= NOTXYBAD;
		}
	      else 
		{
		  (*foundPart)->toggle |= XYBAD;
		}
	    }/* XY plot */
	}/* Toggle it back off */
      
      if(showPackedBox)
	{/* Show them this one */
	  imgNum = minParticle;
	  sprintf(outStr,"%d",imgNum);
	  XmTextSetString(imgNumTextWidget, outStr);
	  stuffPackedImage(purdueHeader.numCols,
			   purdueHeader.numRows,
			   packImg,purdueHeader.nimg,
			   imgNum);
	  DrawWindow(); 
	}/* Show them this one */
      
      lastParticle = minParticle;
      
      /* First draw what we've got */
      DrawICOSWindow();
      
      sprintf(outStr,
	      "Particle[%9d]  Theta [%9.3f]      Phi[%9.3f]  Omega[%9.3f]\n\
       X[%9.3f]       Y[%9.3f]      Mag[%9.3f]\n\
  pft_cc[%9.3f]  prj_cc[%9.3f]   cmp_cc[%9.3f]\n\
  deltaX[%9.3f]  deltaY[%9.3f] distance[%9.3f]\n\
   state[%3s]         rank[%d]\n",
	      (*foundPart)->particle_number, 
	      (*foundPart)->theta,
	      (*foundPart)->phi,
	      (*foundPart)->omega,
	      (*foundPart)->x,
	      (*foundPart)->y,
	      (*foundPart)->mag,
	      (*foundPart)->cc1,
	      (*foundPart)->cc2,
	      (*foundPart)->cc3,
	      (*foundPart)->deltaX,
	      (*foundPart)->deltaY,
	      (*foundPart)->distance,
	      ((*foundPart)->toggle)?" OFF":"ON",
	      (*foundPart)->rank
	      );
      
      XmTextSetString(partPointTextW, outStr);
      sprintf(outStr,"%d",imgNum);
      XmTextSetString(ppinTextW, outStr);
      XFlush(dsply);
    }/* normal */
  
}/* FINDNEARESTPARTICLE */




/**********************************************************
 * findHistParticle() - Mark this particle as bad.  May be
 * used in future when writing a param file from this
 * interface.
*/
void findHistParticle(int *x, int buttonNum)
{/* FINDHISTPARTICLE */
  int i,j,thisX,thisY,minX,minY;
  int distnce,minDist,minParticle;
  P_INFO *thisInfo;
  int numFound;
  float thisVal;
  int binVal;
  XmString simpleString;

  /* First let's see what the cutoff is */
  thisVal = (float)((*x)-YAXIS); /* Offset pixels in X direction */
  thisVal = (float)((*x));
  thisVal = startbox + (thisVal / 480.0);
  
  /* Now find the lowest value for this band of the histogram */
  binVal = (int)((max-min)/(float)step);
  for(i=0;i<(binVal-1);i++)
    {/* Look for range */
      float tmpVal1,tmpVal2;
      tmpVal1 = ((endbox-startbox)*(step*(float)i)) ;
      tmpVal2 = ((endbox-startbox)*(step*(float)(i+1))) ;
      if(tmpVal1 <= thisVal && thisVal < tmpVal2)
	{/* Found it */
	  thisVal = tmpVal1;
	}/* Found it */
    }/* Look for range */


  numFound = 0;
  thisInfo = info.info;
  while(thisInfo != NULL)
    {/* Do all of them */

      if(buttonNum == Button1)
	{/* Take out all to the left */
	  if(magtoggle)
	    {/* Chekc mag */
	      if(thisInfo->mag < thisVal)
		thisInfo->toggle |= HISTMAGBAD;
	    }/* Chekc mag */
	  if(cc1toggle)
	    {/* Chekc mag */
	      if(thisInfo->cc1 < thisVal)
		thisInfo->toggle |= HISTPFTBAD;
	    }/* Chekc mag */
	  if(cc2toggle)
	    {/* Chekc mag */
	      if(thisInfo->cc2 < thisVal)
		thisInfo->toggle |= HISTPRJBAD;
	    }/* Chekc mag */
	  if(cc3toggle)
	    {/* Chekc mag */
	      if(thisInfo->cc3 < thisVal)
		thisInfo->toggle |= HISTCMPBAD;
	    }/* Chekc mag */
	}/* Take out all to the left */
      else if(buttonNum == Button2)
	{/* Take out all to the right */
	  if(magtoggle)
	    {/* Chekc mag */
	      if(thisInfo->mag > thisVal)
		thisInfo->toggle |= HISTMAGBAD;
	    }/* Chekc mag */
	  if(cc1toggle)
	    {/* Chekc mag */
	      if(thisInfo->cc1 > thisVal)
		thisInfo->toggle |= HISTPFTBAD;
	    }/* Chekc mag */
	  if(cc2toggle)
	    {/* Chekc mag */
	      if(thisInfo->cc2 > thisVal)
		thisInfo->toggle |= HISTPRJBAD;
	    }/* Chekc mag */
	  if(cc3toggle)
	    {/* Chekc mag */
	      if(thisInfo->cc3 > thisVal)
		thisInfo->toggle |= HISTCMPBAD;
	    }/* Chekc mag */
	}/* Take out all to the right */
      else if(buttonNum == Button3)
	{/* REstore them */
	  if(magtoggle)
	    {/* Chekc mag */
	      if(thisInfo->mag < thisVal)
		thisInfo->toggle &= NOTHISTMAGBAD;
	    }/* Chekc mag */
	  if(cc1toggle)
	    {/* Chekc mag */
	      if(thisInfo->cc1 < thisVal)
		thisInfo->toggle &= NOTHISTPFTBAD;
	    }/* Chekc mag */
	  if(cc2toggle)
	    {/* Chekc mag */
	      if(thisInfo->cc2 < thisVal)
		thisInfo->toggle &= NOTHISTPRJBAD;
	    }/* Chekc mag */
	  if(cc3toggle)
	    {/* Chekc mag */
	      if(thisInfo->cc3 < thisVal)
		thisInfo->toggle &= NOTHISTCMPBAD;
	    }/* Chekc mag */

	}/* REstore them */

      thisInfo = thisInfo->next;
    }/* Do all of them */


  /* First draw what we've got */
  DrawICOSWindow();

  XFlush(dsply);

}/* FINDHISTPARTICLE */




/*@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
 * icosaMotion() - Whenever the pointer moves across our 
 * window call here.  This is where we redraw the circles
 * if they're moving a box or print the coords of the
 * pointer as they move.
*/
void icosaMotion(Widget widget, caddr_t unused, 
		 register XMotionEvent *pEvent)
{/* ICOSAMOTION */
  int x, y,lastDraw;
  int i, j, k,thisParticle;
  float thisTheta, thisPhi, thisOmega;
  XmString simpleString;
  char infoStr[1024];
  
  /* Make all changes current */
  XFlush(dsply);               
  
  switch (pEvent->type) 
    {/* Find out which event we have */
      
    case MotionNotify:
      
      /* Get coords of pointer */
      x = (int)(pEvent->x);     
      y = (int)(pEvent->y);
      
      if(icoDispMode==UNITCELL)
	{/* Unit cell */
	  if(drawIcos)
	    {/* Draw this icosahedron */
	      
	      /* We're just travelling here */
	      thisOmega = 0.0;
	      convertThetaPhi(x, y, &thisTheta, &thisPhi);
	      sprintf(outStr,"%.1f,%.1f",thisTheta,thisPhi);
	      /*DrawIcosahedron(thisTheta,thisPhi,thisOmega);  */
	      DrawIcosahedron(deg2rad(thisTheta),
			      deg2rad(thisPhi),
			      deg2rad(thisOmega),
			      0, 0, SCALE, icosa_window1); 
	      
	    }/* Draw this icosahedron */
	}/* Unit cell */
      
      break;
      
    default:
      break;
    }/* Find out which event we have */
  
  XFlush(dsply);              /* Make changes current */
  
}/* ICOSAMOTION */



/* @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
 * icosaClick() - Every time the mouse button is clicked
 * in the partproc icosa_area window call here.  
 *
 *            Button1 is the Left mouse button
 *            Button2 is the Middle mouse button
 *            Button3 is the Right mouse button
 */
void icosaClick(Widget widget, int *tag, 
		XmDrawingAreaCallbackStruct *draw_data)
{/* ICOSACLICK */
  XEvent event;
  int x, y, k, i,j,reDraw, thisOne;
  float thisOmega,thisPhi,thisTheta;
  char ans;
  XmString simpleString;

  /* Do appropriate depending on event type */
  if(draw_data != 0)
    {/* Only look at events meant for us */

      event= *(draw_data->event);

      switch (event.type) 
	{/* Find out which event we have */

	case ButtonRelease:

	  x = (int)(event.xbutton.x);      /* Get x,y copords of point pos */
	  y = (int)(event.xbutton.y);

	  break;

	  
	case ButtonPress:

	  /* They are trying to find points */
	  x = (int)(event.xbutton.x);      /* Get x,y copords of point pos */
	  y = (int)(event.xbutton.y);

	  if(icoDispMode==ICOSORT)
	    {/* Do nothing */
	    }/* Do nothing */
	  else if(icoDispMode==HISTOGRAM)
	    {/* Do hist */
	      findHistParticle(&x,event.xbutton.button);
	    }/* Do hist */
	  else
	    {/* find */
	      if((event.xbutton.button == Button1))
		{/* Button 1 == left */
		  if(findNearest)
		    findNearestParticle(&x,&y,&foundPart,FALSE,-99,TRUE,
					event.xbutton.button);
		}/* Button 1 == left */
	      else if((event.xbutton.button == Button2))
		{/* Toggle this guy off */
		  if(findNearest)
		    findNearestParticle(&x,&y,&foundPart,TRUE,-99,TRUE,
					event.xbutton.button);
		}/* Toggle this guy off */
	      if((event.xbutton.button == Button3))
		{/* Button 3 == right */
		  if(findNearest)
		    findNearestParticle(&x,&y,&foundPart,FALSE,-99,TRUE,
					event.xbutton.button);
		}/* Button 3 == right */
	    }/* find */
	  
	  break;
	  
	case KeyPress:
	  ans= XLookupKeysym((XKeyEvent *)&event, 0);
	  printf ("Keyed[%c]\n",ans);

	  break;
	  
	default:
	  break;
	}/* Find out which event we have */
    }/* Only look at events meant for us */
  
  XFlush(dsply);            /* Make changes current */

  
}/* ICOSACLICK */





/*************************************************************
 * binPPArrowActivate() - They want to change the bin
 * for the histogram on the PartProc screen by using the 
 * arrows.
*/
void binPPArrowActivate(Widget widget, int *tag, 
			XmAnyCallbackStruct *callback_data)
{/* BINPPARROWACTIVATE */
  float speedFactor;
  char  *inStr;
  int   binVal;
  float stepVal;

  inStr = XmTextGetString(binTextW);
  binVal = (int)strtod(inStr,0);
  XtFree(inStr);

  if(*tag==XmARROW_UP)
    binVal++;
  else 
    binVal--;

  binVal= MAX(binVal,1);

  sprintf(outStr,"%d",binVal);
  XmTextSetString(binTextW,outStr);

  /* Now set the stepVal accordingly */
  stepVal = (float)(max-min)/binVal;
  stepVal= MAX(stepVal,0.0000001);
  sprintf(outStr,"%.3f",stepVal);
  XmTextSetString(stepTextW,outStr);

  /* Now make those changes current */
  paramMode_change(widget,tag,callback_data);

}/* BINPPARROWACTIVATE */





/*************************************************************
 * binPPArrowButton() - They want to change the current
 * displayed box by using the arrows.
*/
void binPPArrowButton(Widget widget, XEvent *event)
{/* BINPPARROWBUTTON */
  XEvent thisEvent;
  float speedFactor;
  char *inStr;
  int binVal;
  float stepVal;
  int     tag;
  XmAnyCallbackStruct callback_data;

  inStr = XmTextGetString(binTextW);
  binVal = (int)strtod(inStr,0);
  XtFree(inStr);

  thisEvent= *(event);
  
  /* Get which button */
  if((thisEvent.xbutton.button == Button1))
    speedFactor = 1.0;
  if((thisEvent.xbutton.button == Button2))
    speedFactor = 2.0;
  if((thisEvent.xbutton.button == Button3))
    speedFactor = 5.0;
  
  if(widget==binPPUpW)
    binVal += (int)speedFactor;
  else 
    binVal -= (int)speedFactor;

  binVal= MAX(binVal,1);

  sprintf(outStr,"%d",binVal);
  XmTextSetString(binTextW,outStr);

  /* Now set the stepVal accordingly */
  stepVal = (float)(max-min)/binVal;
  stepVal= MAX(stepVal,0.0000001);
  sprintf(outStr,"%.3f",stepVal);
  XmTextSetString(stepTextW,outStr);

  /* Now make those changes current */
  tag = BINTEXT;
  paramMode_change(widget,&tag,&callback_data);

}/* BINPPARROWBUTTON */





/*************************************************************
 * binPPtext_changed() - They hit the carriage return key
 * in the binTextW so make those changes current.
*/
void binPPtext_changed(Widget widget, XEvent *event)
{/* BINPPTEXT_CHANGED */
  int     tag;
  XmAnyCallbackStruct callback_data;
  char *binStr,*stepStr;
  int   binVal;
  float stepVal;

  binStr = XmTextGetString(binTextW);
  binVal = (int)strtod(binStr,0);
  if(binVal < 1)
    binVal = 1;
  XtFree(binStr);
  sprintf(outStr,"%d",binVal);
  XmTextSetString(binTextW,outStr);

  /* Now set the stepVal accordingly */
  stepVal = (float)(max-min)/(float)binVal;
  stepVal= MAX(stepVal,0.0000001);
  sprintf(outStr,"%.3f",stepVal);
  XmTextSetString(stepTextW,outStr);

  /* Now make those changes current */
  tag = BINTEXT;
  paramMode_change(widget,&tag,&callback_data);

}/* BINPPTEXT_CHANGED */





/*************************************************************
 * stepPPArrowActivate() - They want to change the step
 * for the histogram on the PartProc screen by using the 
 * arrows.
*/
void stepPPArrowActivate(Widget widget, int *tag, 
			 XmAnyCallbackStruct *callback_data)
{/* STEPPPARROWACTIVATE */
  float speedFactor;
  char *inStr;
  int binVal;
  float stepVal;

  inStr = XmTextGetString(stepTextW);
  stepVal = (float)strtod(inStr,0);
  XtFree(inStr);

  if(*tag==XmARROW_UP)
    stepVal+= 1.0;
  else 
    stepVal-= 1.0;

  stepVal= MAX(stepVal,0.0000001);

  sprintf(outStr,"%.3f",stepVal);
  XmTextSetString(stepTextW,outStr);

  /* Now set the binVal accordingly */
  binVal = (int)((max-min)/(float)stepVal);
  if(binVal < 1)
    binVal = 1;
  sprintf(outStr,"%d",binVal);
  XmTextSetString(binTextW,outStr);

  /* Now make those changes current */
  paramMode_change(widget,tag,callback_data);

}/* STEPPPARROWACTIVATE */





/*************************************************************
 * stepPPArrowButton() - They want to change the current
 * displayed box by using the arrows.
*/
void stepPPArrowButton(Widget widget, XEvent *event)
{/* STEPPPARROWBUTTON */
  XEvent thisEvent;
  float speedFactor;
  char *inStr;
  float stepVal;
  int binVal;
  int     tag;
  XmAnyCallbackStruct callback_data;

  inStr = XmTextGetString(stepTextW);
  stepVal = (float)strtod(inStr,0);
  XtFree(inStr);

  thisEvent= *(event);
  
  /* Get which button */
  if((thisEvent.xbutton.button == Button1))
    speedFactor = 1.0;
  if((thisEvent.xbutton.button == Button2))
    speedFactor = 2.0;
  if((thisEvent.xbutton.button == Button3))
    speedFactor = 5.0;
  
  if(widget==stepPPUpW)
    stepVal += speedFactor;
  else 
    stepVal -= speedFactor;

  stepVal= MAX(stepVal,0.0000001);

  sprintf(outStr,"%.3f",stepVal);
  XmTextSetString(stepTextW,outStr);

  /* Now set the binVal accordingly */
  binVal = (int)((max-min)/(float)stepVal);
  if(binVal < 1)
    binVal = 1;
  sprintf(outStr,"%d",binVal);
  XmTextSetString(binTextW,outStr);

  /* Now make those changes current */
  tag = STEPTEXT;
  paramMode_change(widget,&tag,&callback_data);

}/* STEPPPARROWBUTTON */





/*************************************************************
 * stepPPtext_changed() - They hit the carriage return key
 * in the stepTextW so make those changes current.
*/
void stepPPtext_changed(Widget widget, XEvent *event)
{/* STEPPTEXT_CHANGED */
  int     tag;
  XmAnyCallbackStruct callback_data;
  char *stepStr;
  int binVal;
  float stepVal;

  stepStr = XmTextGetString(stepTextW);
  stepVal = (float)strtod(stepStr,0);
  stepVal= MAX(stepVal,0.0000001);
  XtFree(stepStr);

  /* Now set the binVal accordingly */
  binVal = (int)((max-min)/(float)stepVal);
  if(binVal < 1)
    binVal = 1;
  sprintf(outStr,"%d",binVal);
  XmTextSetString(binTextW,outStr);

  /* Now make those changes current */
  tag = STEPTEXT;
  paramMode_change(widget,&tag,&callback_data);

}/* STEPPTEXT_CHANGED */





/*************************************************************
 * ppintext_changed() - They hit the carriage return key
 * in the ppinTextW so make those changes current.
*/
void ppintext_changed(Widget widget, XEvent *event)
{/* PPINTEXT_CHANGED */
  char *thisStr;

  thisStr = XmTextGetString(ppinTextW);
  imgNum = (int)strtod(thisStr,0);
  XtFree(thisStr);

  if(fileType==PACKEDPORTABLE || fileType==MAP)
    imgNum = MIN(imgNum,purdueHeader.nimg);
  if(imgNum < 1)
    imgNum = 1;
  if(shadedMaps && !dispSections)
    imgNum = MIN(imgNum,mapNumSecs);

  doTheImgNum(imgnumArrowUpWidget, 0);

}/* PPINTEXT_CHANGED */





/*************************************************************
 * ppinArrowActivate() - They want to change the current
 * displayed box by using the arrows.
*/
void ppinArrowActivate(Widget widget, int *tag, 
		       XmAnyCallbackStruct *callback_data)
{/* PPINARROWACTIVATE */
  float speedFactor;

  if(!locked)
    {/* Can't do if we're locked out */
      
      speedFactor = 1.0;;
      if(*tag==XmARROW_UP)
	doTheImgNum(imgnumArrowUpWidget, speedFactor);
      else 
	doTheImgNum(imgnumArrowDownWidget, speedFactor);

    }/* Can't do if we're locked out */

}/* PPINARROWACTIVATE */





/*************************************************************
 * ppinArrowButton() - They want to change the current
 * displayed box by using the arrows.
*/
void ppinArrowButton(Widget widget, XEvent *event)
{/* PPINARROWBUTTON */
  XEvent thisEvent;
  float speedFactor;

  if(!locked)
    {/* Can't do if we're locked out */
      
      thisEvent= *(event);

      /* Get which button */
      if((thisEvent.xbutton.button == Button1))
	speedFactor = 1.0;;
      if((thisEvent.xbutton.button == Button2))
	speedFactor = 2.0;
      if((thisEvent.xbutton.button == Button3))
	speedFactor = 5.0;

      if(widget==ppinUpW)
	doTheImgNum(imgnumArrowUpWidget, speedFactor);
      else 
	doTheImgNum(imgnumArrowDownWidget, speedFactor);

    }/* Can't do if we're locked out */

}/* PPINARROWBUTTON */





/************************************************************
 * global_minmax() - Find the min/max for this dataset.
 */
void global_minmax() 
{/* GLOBAL_MINMAX */

  /* We're only doing mag,pftcc,prjcc,cmpcc now so these will do */
  min = 1.0;
  max = -1.0;
  if (thetatoggle) {
    if (thetamin < min) min = thetamin;
    if (thetamax > max) max = thetamax;
  }
  if (phitoggle) {
    if (phimin < min) min = phimin;
    if (phi_max > max) max = phi_max;
  }
  if (omegatoggle) {
    if (omegamin < min) min = omegamin;
    if (omegamax > max) max = omegamax;
  }
  if (xtoggle) {
    if (xmin < min) min = xmin;
    if (xmax > max) max = xmax;
  }
  if (ytoggle) {
    if (ymin < min) min = ymin;
    if (ymax > max) max = ymax;
  }
  if (deltaXtoggle) {
    if (deltaXmin < min) min = deltaXmin;
    if (deltaXmax > max) max = deltaXmax;
  }
  if (deltaYtoggle) {
    if (deltaYmin < min) min = deltaYmin;
    if (deltaYmax > max) max = deltaYmax;
  }
  if (absDeltaXtoggle) {
    if (fabs(deltaXmin) < min) min = fabs(deltaXmin);
    if (fabs(deltaXmax) > max) max = fabs(deltaXmax);
  }
  if (absDeltaYtoggle) {
    if (fabs(deltaYmin) < min) min = fabs(deltaYmin);
    if (fabs(deltaYmax) > max) max = fabs(deltaYmax);
  }
  if (distancetoggle) {
    if (distanceMin < min) min = distanceMin;
    if (distanceMax > max) max = distanceMax;
  }
  if (magtoggle) {
    if (magmin < min) min = magmin;
    if (magmax > max) max = magmax;
  } 
  if (cc1toggle) {
    if (cc1min < min) min = cc1min;
    if (cc1max > max) max = cc1max;
    /* Try RWa July 30, 1998 */
    min = 0.0;
    max = 1.0;
  }
  if (cc2toggle) {
    if (cc2min < min) min = cc2min;
    if (cc2max > max) max = cc2max;
    /* Try RWa July 30, 1998 */
    min = 0.0;
    max = 1.0;
  }
  if (cc3toggle) {
    if (cc3min < min) min = cc3min;
    if (cc3max > max) max = cc3max;
    /* Try RWa July 30, 1998 */
    min = 0.0;
    max = 1.0;
  }
}/* GLOBAL_MINMAX */




/****************************************************************
 * find_min_yscale() - Find the minimum yscale for this data.
 */
void find_min_yscale()
{/* FIND_MIN_YSCALE */
  int minyscale=480;
  float heightFound;
  int thisBadCount;
  float thisAvgBad;

  yscale = minyscale;

  avgBadCount = 0.0;
  if (thetatoggle) {
    heightFound = find_max_height(info.info,startbox,endbox,step,
				  THETAT,&thisBadCount);
    heightFound += thisBadCount;
    thisAvgBad = (float)thisBadCount / (float)(heightFound);
    avgBadCount = MAX(avgBadCount,thisAvgBad);
    heightFound = MIN(heightFound,(float)maxYbox);
                                              /* They may want to override */
    yscale = 480.0/heightFound;
    if (yscale < minyscale)
      minyscale = yscale;
  }
  if (phitoggle) {
    heightFound = find_max_height(info.info,startbox,endbox,step,
				  PHIT,&thisBadCount);
    heightFound += thisBadCount;
    thisAvgBad = (float)thisBadCount / (float)(heightFound);
    avgBadCount = MAX(avgBadCount,thisAvgBad);
    heightFound = MIN(heightFound,(float)maxYbox);
                                              /* They may want to override */
    yscale = 480.0/heightFound;
    /*    yscale = 480.0/find_max_height(info.info,startbox,endbox,
	  step,PHIT); ,&thisBadCount*/
    if (yscale < minyscale)
      minyscale = yscale;
  }
  if (omegatoggle) {
    heightFound = find_max_height(info.info,startbox,endbox,step,
				  OMEGAT,&thisBadCount);
    heightFound += thisBadCount;
    thisAvgBad = (float)thisBadCount / (float)(heightFound);
    avgBadCount = MAX(avgBadCount,thisAvgBad);
    heightFound = MIN(heightFound,(float)maxYbox);
                                              /* They may want to override */
    yscale = 480.0/heightFound;
    /*    yscale = 480.0/find_max_height(info.info,startbox,endbox,
	  step,OMEGAT); ,&thisBadCount*/
    if (yscale < minyscale)
      minyscale = yscale;
  }
  if (xtoggle) {
    heightFound = find_max_height(info.info,startbox,endbox,step,
				  XT,&thisBadCount);
    heightFound += thisBadCount;
    thisAvgBad = (float)thisBadCount / (float)(heightFound);
    avgBadCount = MAX(avgBadCount,thisAvgBad);
    heightFound = MIN(heightFound,(float)maxYbox);
                                              /* They may want to override */
    yscale = 480.0/heightFound;
    /*    yscale = 480.0/find_max_height(info.info,startbox,endbox,
	  step,XT); ,&thisBadCount*/
    if (yscale < minyscale)
      minyscale = yscale;
  }
  if (deltaXtoggle) {
    heightFound = find_max_height(info.info,startbox,endbox,step,
				  DELTAXT,&thisBadCount);
    heightFound += thisBadCount;
    thisAvgBad = (float)thisBadCount / (float)(heightFound);
    avgBadCount = MAX(avgBadCount,thisAvgBad);
    heightFound = MIN(heightFound,(float)maxYbox);
                                              /* They may want to override */
    yscale = 480.0/heightFound;
    /*    yscale = 480.0/find_max_height(info.info,startbox,endbox,
	  step,DELTAXT); ,&thisBadCount*/
    if (yscale < minyscale)
      minyscale = yscale;
  }
  if (absDeltaXtoggle) {
    heightFound = find_max_height(info.info,startbox,endbox,step,
				  ABSDELTAXT,&thisBadCount);
    heightFound += thisBadCount;
    thisAvgBad = (float)thisBadCount / (float)(heightFound);
    avgBadCount = MAX(avgBadCount,thisAvgBad);
    heightFound = MIN(heightFound,(float)maxYbox);
                                              /* They may want to override */
    yscale = 480.0/heightFound;
    /*    yscale = 480.0/find_max_height(info.info,startbox,endbox,
	  step,ABSDELTAXT); ,&thisBadCount*/
    if (yscale < minyscale)
      minyscale = yscale;
  }
  if (ytoggle) {
    heightFound = find_max_height(info.info,startbox,endbox,step,
				  YT,&thisBadCount);
    heightFound += thisBadCount;
    thisAvgBad = (float)thisBadCount / (float)(heightFound);
    avgBadCount = MAX(avgBadCount,thisAvgBad);
    heightFound = MIN(heightFound,(float)maxYbox);
                                              /* They may want to override */
    yscale = 480.0/heightFound;
    /*    yscale = 480.0/find_max_height(info.info,startbox,endbox,
	  step,YT); ,&thisBadCount*/
    if (yscale < minyscale)
      minyscale = yscale;
  }
  if (deltaYtoggle) {
    heightFound = find_max_height(info.info,startbox,endbox,step,
				  DELTAYT,&thisBadCount);
    heightFound += thisBadCount;
    thisAvgBad = (float)thisBadCount / (float)(heightFound);
    avgBadCount = MAX(avgBadCount,thisAvgBad);
    heightFound = MIN(heightFound,(float)maxYbox);
                                              /* They may want to override */
    yscale = 480.0/heightFound;
    /*    yscale = 480.0/find_max_height(info.info,startbox,endbox,
	  step,DELTAYT); ,&thisBadCount*/
    if (yscale < minyscale)
      minyscale = yscale;
  }
  if (absDeltaYtoggle) {
    heightFound = find_max_height(info.info,startbox,endbox,step,
				  ABSDELTAYT,&thisBadCount);
    heightFound += thisBadCount;
    thisAvgBad = (float)thisBadCount / (float)(heightFound);
    avgBadCount = MAX(avgBadCount,thisAvgBad);
    heightFound = MIN(heightFound,(float)maxYbox);
                                              /* They may want to override */
    yscale = 480.0/heightFound;
    /*    yscale = 480.0/find_max_height(info.info,startbox,endbox,
	  step,ABSDELTAYT); ,&thisBadCount*/
    if (yscale < minyscale)
      minyscale = yscale;
  }
  if (distancetoggle) {
    heightFound = find_max_height(info.info,startbox,endbox,step,
				  DISTANCET,&thisBadCount);
    heightFound += thisBadCount;
    thisAvgBad = (float)thisBadCount / (float)(heightFound);
    avgBadCount = MAX(avgBadCount,thisAvgBad);
    heightFound = MIN(heightFound,(float)maxYbox);
                                              /* They may want to override */
    yscale = 480.0/heightFound;
    /*    yscale = 480.0/find_max_height(info.info,startbox,endbox,
	  step,DISTANCET); ,&thisBadCount*/
    if (yscale < minyscale)
      minyscale = yscale;
  }
  if (magtoggle) {
    heightFound = find_max_height(info.info,startbox,endbox,step,
				  MAGT,&thisBadCount);
    heightFound += thisBadCount;
    thisAvgBad = (float)thisBadCount / (float)(heightFound);
    avgBadCount = MAX(avgBadCount,thisAvgBad);
    heightFound = MIN(heightFound,(float)maxYbox);
                                              /* They may want to override */
    yscale = 480.0/heightFound;
    /*    yscale = 480.0/find_max_height(info.info,startbox,endbox,
	  step,MAGT); ,&thisBadCount*/
    if (yscale < minyscale)
      minyscale = yscale;
  }
  if (cc1toggle) {
    heightFound = find_max_height(info.info,startbox,endbox,step,
				  PFTCC,&thisBadCount);
    heightFound += thisBadCount;
    thisAvgBad = (float)thisBadCount / (float)(heightFound);
    avgBadCount = MAX(avgBadCount,thisAvgBad);
    heightFound = MIN(heightFound,(float)maxYbox);
                                              /* They may want to override */
    yscale = 480.0/heightFound;
    /*    yscale = 480.0/find_max_height(info.info,startbox,endbox,
	  step,PFTCC); ,&thisBadCount*/
    if (yscale < minyscale)
      minyscale = yscale;
  }
  if (cc2toggle) {
    heightFound = find_max_height(info.info,startbox,endbox,step,
				  PRJCC,&thisBadCount);
    heightFound += thisBadCount;
    thisAvgBad = (float)thisBadCount / (float)(heightFound);
    avgBadCount = MAX(avgBadCount,thisAvgBad);
    heightFound = MIN(heightFound,(float)maxYbox);
                                              /* They may want to override */
    yscale = 480.0/heightFound;
    /*    yscale = 480.0/find_max_height(info.info,startbox,endbox,
	  step,PRJCC); ,&thisBadCount*/
    if (yscale < minyscale)
      minyscale = yscale;
  }
  if (cc3toggle) {
    heightFound = find_max_height(info.info,startbox,endbox,step,
				  CMPCC,&thisBadCount);
    heightFound += thisBadCount;
    thisAvgBad = (float)thisBadCount / (float)(heightFound);
    avgBadCount = MAX(avgBadCount,thisAvgBad);
    heightFound = MIN(heightFound,(float)maxYbox);
                                              /* They may want to override */
    yscale = 480.0/heightFound;
    /*    yscale = 480.0/find_max_height(info.info,startbox,endbox,
	  step,CMPCC); ,&thisBadCount*/
    if (yscale < minyscale)
      minyscale = yscale;
  }
  yscale = minyscale;
  if(yscale == 0.0)
    yscale = 1.0;  /* Just give it something */

}/* FIND_MIN_YSCALE */



/********************************************************
 * paramMode_change() - They've toggled a different 
 * mode.
 */
void paramMode_change(Widget widget, int *tag, 
		      XmAnyCallbackStruct *callback_data)
{/* PARAMMODE_CHANGE */
  char newVal[12];
  XmString simpleString;
  char  *binStr,*minMaxStr;
  int binVal;
  int totalToggled;

  if(icoDispMode==ICOSORT)
    {/* Init everything */
      idtoggle= thetatoggle= phitoggle= omegatoggle= xtoggle= ytoggle=
	magtoggle= cc1toggle= cc2toggle= cc3toggle= distancetoggle=
	  deltaXtoggle = deltaYtoggle= absDeltaXtoggle= absDeltaYtoggle= FALSE;
      
      XtSetArg(arglst[0], XmNforeground, pixels_return[WHITE] );
      XtSetValues(magtW, arglst, 1); 
      XtSetValues(pftccW, arglst, 1); 
      XtSetValues(prjccW, arglst, 1); 
      XtSetValues(cmpccW, arglst, 1); 

      /* Just in case we need it for ICOSORT */
      XtSetArg(arglst[0], XmNset, FALSE );
      if(icoDispMode==ICOSORT)
	{/* Do the sorting thing */
	  XtSetValues(idtW, arglst, 1); 
	  XtSetValues(thetatW, arglst, 1); 
	  XtSetValues(phitW, arglst, 1); 
	  XtSetValues(omegatW, arglst, 1); 
	  XtSetValues(xtW, arglst, 1); 
	  XtSetValues(ytW, arglst, 1); 
	  XtSetValues(deltaXTW, arglst, 1); 
	  XtSetValues(deltaYTW, arglst, 1); 
	  XtSetValues(absDeltaXTW, arglst, 1); 
	  XtSetValues(absDeltaYTW, arglst, 1); 
	  XtSetValues(distanceTW, arglst, 1); 
	  XtSetValues(magtW, arglst, 1); 
	  XtSetValues(pftccW, arglst, 1); 
	  XtSetValues(prjccW, arglst, 1); 
	  XtSetValues(cmpccW, arglst, 1); 
	}/* Do the sorting thing */
      XFlush(dsply);
      /* Turn the new one ON */
      XtSetArg(arglst[0], XmNset, TRUE );
    }/* Init everything */

  if (info.info) {
    find_minmax(info.info); 
  }
  else {
    XtSetArg(arglst[0], XmNset, FALSE );
    XtSetValues(widget, arglst, 1); 
    XFlush(dsply);
    askThem("No File opened!",TRUE,XmDIALOG_QUESTION);
    return;
  }
  if(*tag==INCDEC) {
    incSort = !incSort;
    if(icoDispMode==ICOSORT)
      {/* Set label */
	if(incSort)
	  sprintf(outStr,"Increasing");
	else sprintf(outStr,"Decreasing");
	createSimpleString(&simpleString,outStr);
	XtSetArg(arglst[0], XmNlabelString, simpleString);
	XtSetValues(incDecW, arglst, 1); 
	XmStringFree(simpleString);
      }/* Set label */
    goto noSortYet;
  }
  else if(*tag==IDT) {
    if (idtoggle) idtoggle = 0;
    else idtoggle = 1;
    global_minmax(); 
    if(icoDispMode==ICOSORT)
      XtSetValues(idtW, arglst, 1); 
  }
  else if(*tag==THETAT) {
    if (thetatoggle) thetatoggle = 0;
    else thetatoggle = 1;
    global_minmax();
    if(icoDispMode==ICOSORT)
      XtSetValues(thetatW, arglst, 1); 
  }
  else if(*tag==PHIT) {
    if (phitoggle) phitoggle = 0;
    else phitoggle = 1;
    global_minmax();
    if(icoDispMode==ICOSORT)
      XtSetValues(phitW, arglst, 1); 
  }
  else if(*tag==OMEGAT) {
    if (omegatoggle) omegatoggle = 0;
    else omegatoggle = 1;
    global_minmax();
    if(icoDispMode==ICOSORT)
      XtSetValues(omegatW, arglst, 1); 
  }
  else if(*tag==XT) {
    if (xtoggle) xtoggle = 0;
    else xtoggle = 1;
    global_minmax();
    if(icoDispMode==ICOSORT)
      XtSetValues(xtW, arglst, 1); 
  }
  else if(*tag==DELTAXT) {
    deltaXtoggle = !deltaXtoggle;
    global_minmax();
    if(icoDispMode==ICOSORT)
      XtSetValues(deltaXTW, arglst, 1); 
  }
  else if(*tag==ABSDELTAXT) {
    absDeltaXtoggle = !absDeltaXtoggle;
    global_minmax();
    if(icoDispMode==ICOSORT)
      XtSetValues(absDeltaXTW, arglst, 1); 
  }
  else if(*tag==YT) {
    if (ytoggle) ytoggle = 0;
    else ytoggle = 1;
    global_minmax();
    if(icoDispMode==ICOSORT)
      XtSetValues(ytW, arglst, 1); 
  }
  else if(*tag==DELTAYT) {
    deltaYtoggle = !deltaYtoggle;
    global_minmax();
    if(icoDispMode==ICOSORT)
      XtSetValues(deltaYTW, arglst, 1); 
  }
  else if(*tag==ABSDELTAYT) {
    absDeltaYtoggle = !absDeltaYtoggle;
    global_minmax();
    if(icoDispMode==ICOSORT)
      XtSetValues(absDeltaYTW, arglst, 1); 
  }
  else if(*tag==DISTANCET) {
    distancetoggle = !distancetoggle;
    global_minmax();
    if(icoDispMode==ICOSORT)
      XtSetValues(distanceTW, arglst, 1); 
  }
  else if(*tag==MAGT) {
    if (magtoggle) magtoggle = 0;
    else magtoggle = 1;
    global_minmax();
    if(icoDispMode==ICOSORT)
      XtSetValues(magtW, arglst, 1); 
    else if(icoDispMode==HISTOGRAM) 
      {
	XtSetArg(arglst[0], XmNset, FALSE );
	XtSetArg(arglst[1], XmNforeground, pixels_return[WHITE] );
	XtSetValues(pftccW, arglst, 2); 
	XtSetValues(prjccW, arglst, 2); 
	XtSetValues(cmpccW, arglst, 2); 
	cc1toggle = cc2toggle=cc3toggle= FALSE;
	if(magtoggle)
	  XtSetArg(arglst[0], XmNforeground, pixels_return[GREEN] );
	else XtSetArg(arglst[0], XmNforeground, pixels_return[WHITE] );
	XtSetValues(magtW, arglst, 1); 
      }
    XFlush(dsply);
  }
  else if(*tag==PFTCC) {
    if (cc1toggle) cc1toggle = 0;
    else cc1toggle = 1;
    global_minmax();
    if(icoDispMode==ICOSORT)
      XtSetValues(pftccW, arglst, 1); 
    else if(icoDispMode==HISTOGRAM) 
      {
	XtSetArg(arglst[0], XmNset, FALSE );
	XtSetArg(arglst[1], XmNforeground, pixels_return[WHITE] );
	XtSetValues(magtW, arglst, 2); 
	magtoggle= FALSE;
	if(cc1toggle)
	  XtSetArg(arglst[0], XmNforeground, pixels_return[BLUE] );
	else XtSetArg(arglst[0], XmNforeground, pixels_return[WHITE] );
	XtSetValues(pftccW, arglst, 1); 
      }
    XFlush(dsply);
  }
  else if(*tag==PRJCC) {
    if (cc2toggle) cc2toggle = 0;
    else cc2toggle = 1;
    global_minmax();
    if(icoDispMode==ICOSORT)
      XtSetValues(prjccW, arglst, 1); 
    else if(icoDispMode==HISTOGRAM) 
      {
	XtSetArg(arglst[0], XmNset, FALSE );
	XtSetArg(arglst[1], XmNforeground, pixels_return[WHITE] );
	XtSetValues(magtW, arglst, 2); 
	magtoggle= FALSE;
	if(cc2toggle)
	  XtSetArg(arglst[0], XmNforeground, pixels_return[GREEN] );
	else XtSetArg(arglst[0], XmNforeground, pixels_return[WHITE] );
	XtSetValues(prjccW, arglst, 1); 
      }
    XFlush(dsply);
  }
  else if(*tag==CMPCC) {
    if (cc3toggle) cc3toggle = 0;
    else cc3toggle = 1;
    global_minmax();
    if(icoDispMode==ICOSORT)
      XtSetValues(cmpccW, arglst, 1); 
    else if(icoDispMode==HISTOGRAM) 
      {
	XtSetArg(arglst[0], XmNset, FALSE );
	XtSetArg(arglst[1], XmNforeground, pixels_return[WHITE] );
	XtSetValues(magtW, arglst, 2); 
	magtoggle= FALSE;
	if(cc3toggle)
	  XtSetArg(arglst[0], XmNforeground, pixels_return[RED] );
	else XtSetArg(arglst[0], XmNforeground, pixels_return[WHITE] );
	XtSetValues(cmpccW, arglst, 1); 
      }
    XFlush(dsply);
  }

  if (!thetatoggle && !omegatoggle && !phitoggle && !xtoggle && !ytoggle
      && !magtoggle && !cc1toggle && !cc2toggle && !cc3toggle &&
      !deltaXtoggle && !absDeltaXtoggle && !distancetoggle &&
      !deltaYtoggle && !absDeltaYtoggle)
  {
    startbox = 0.0;
    endbox = 1.0;
    step = 0.016;
  }
  else
    {
      startbox = min; endbox = max;  
    }
      
  if(min>=max)
    {/* Can't have this */
      max = min + 0.01;
    }/* Can't have this */

  /* Check to see if they've entered anything on the screen */
  minMaxStr = XmTextGetString(strtXPPW);
  if(!EQU(minMaxStr,""))
    startbox = MIN((float)strtod(minMaxStr,0),max);
  XtFree(minMaxStr);
  minMaxStr = XmTextGetString(endXPPW);
  if(!EQU(minMaxStr,""))
    endbox = MAX((float)strtod(minMaxStr,0),min);
  XtFree(minMaxStr);
  minMaxStr = XmTextGetString(maxYPPW);
  if(!EQU(minMaxStr,""))
    maxYbox = (int)strtod(minMaxStr,0);
  maxYbox = MAX(1,maxYbox);
  XtFree(minMaxStr);
  sprintf(outStr,"%f",startbox);
  XmTextSetString(strtXPPW,outStr);
  sprintf(outStr,"%f",endbox);
  XmTextSetString(endXPPW,outStr);
  sprintf(outStr,"%d",maxYbox);
  XmTextSetString(maxYPPW,outStr);

  binStr = XmTextGetString(binTextW);
  binVal = (int)strtod(binStr,0);
  if(binVal < 1)
    binVal = 1;
  XtFree(binStr);
  sprintf(outStr,"%d",binVal);
  XmTextSetString(binTextW,outStr);

  step = (float)((max-min)/480.0)*(480.0/(float)binVal);
  if(endbox==startbox)
    pixelstep = (step*linelength)/(1);
  else pixelstep = (step*linelength)/(endbox-startbox);
  find_min_yscale();
  /*
  sprintf(newVal,"%f",(step)); 
  XmTextSetString(stepTextWidget, newVal);
  sprintf(newVal,"%f",(startbox));
  XmTextSetString(startTextWidget, newVal);
  sprintf(newVal,"%f",(endbox)); 
  XmTextSetString(endTextWidget, newVal); 
  */
  XFlush(dsply);

  if(icoDispMode==ICOSORT)
    {/* They want to sort */
      int x,y;

      totalToggled = idtoggle + thetatoggle + 
	phitoggle + omegatoggle + 
	  xtoggle + ytoggle + magtoggle + distancetoggle +
	    deltaXtoggle + deltaYtoggle + absDeltaXtoggle + absDeltaYtoggle + 
	      cc1toggle + cc2toggle + cc3toggle;
      if(totalToggled != 1)
	{/* not right */
	  if(totalToggled <= 0)
	    {/* OK - they're choosing a new one */}
	  else
	    askThem("You can only sort on one key.",FALSE,
		    XmDIALOG_INFORMATION);
	}/* not right */
      else
	{/* Sort it */

	  icoSortLevel = 1;
	  DrawICOSWindow();
	  if(idtoggle)
	    sortParticles(info.info,purdueHeader.nimg, 2, incSort);
	  else if(thetatoggle)
	    sortParticles(info.info,purdueHeader.nimg, 3, incSort);
	  else if(phitoggle)
	    sortParticles(info.info,purdueHeader.nimg, 4, incSort);
	  else if(omegatoggle)
	    sortParticles(info.info,purdueHeader.nimg, 5, incSort);
	  else if(xtoggle)
	    sortParticles(info.info,purdueHeader.nimg, 6, incSort);
	  else if(ytoggle)
	    sortParticles(info.info,purdueHeader.nimg, 7, incSort);
	  else if(magtoggle)
	    sortParticles(info.info,purdueHeader.nimg, 8, incSort);
	  else if(cc1toggle)
	    sortParticles(info.info,purdueHeader.nimg, 9, incSort);
	  else if(cc2toggle)
	    sortParticles(info.info,purdueHeader.nimg, 10, incSort);
	  else if(cc3toggle)
	    sortParticles(info.info,purdueHeader.nimg, 11, incSort);
	  else if(deltaXtoggle)
	    sortParticles(info.info,purdueHeader.nimg, 13, incSort);
	  else if(deltaYtoggle)
	    sortParticles(info.info,purdueHeader.nimg, 14, incSort);
	  else if(absDeltaXtoggle)
	    sortParticles(info.info,purdueHeader.nimg, 15, incSort);
	  else if(absDeltaYtoggle)
	    sortParticles(info.info,purdueHeader.nimg, 16, incSort);
	  else if(distancetoggle)
	    sortParticles(info.info,purdueHeader.nimg, 17, incSort);

	  icoSortLevel = 2;
	  sprintf(outStr,"SORT finished.");
	  XmTextSetString(partPointTextW,outStr);

	  findNearestParticle(&x,&y,&foundPart,FALSE,1,TRUE,0);
	  
	}/* Sort it */

    }/* They want to sort */


  DrawICOSWindow();

 noSortYet:;

}/* PARAMMODE_CHANGE */



/********************************************************
 * markXYplotOutliers() - Mark those points outside the
 * given distance as outliers.  The toggle value tells
 * the display rtn to give feedback that this point is
 * an outlier. 
*/
void markXYplotOutliers()
{/* MARKXYPLOTOUTLIERS */
  int numFound;
  P_INFO *thisInfo;

  /* First unmark any previous ones */
  thisInfo = info.info;
  while(thisInfo != NULL)
    {/* Do all of them */
      thisInfo->toggle &= NOTXYBADTHRESH;
      thisInfo = thisInfo->next;
    }/* Do all of them */

  numFound = 0;
  thisInfo = info.info;
  while(thisInfo != NULL)
    {/* Do all of them */
      /* XYPLOT */
      if((thisInfo->distance < minXYplot) ||
	 (thisInfo->distance > maxXYplot) )
	{/* Found a new candidate */
	  numFound++;
	  thisInfo->toggle |= XYBADTHRESH;
	}/* Found a new candidate */
      thisInfo = thisInfo->next;
    }/* Do all of them */

  DrawICOSWindow(); 

}/* MARKXYPLOTOUTLIERS */





/********************************************************
 * ppMinMax_change() - They want to change the concept 
 * o min/max.
 */
void ppMinMax_change(Widget widget, int *tag, 
		     XmAnyCallbackStruct *callback_data)
{/* PPMINMAX_CHANGE */
  char *thisStr;
  float thisVal;

  thisStr = XmTextGetString(widget);
  thisVal = (float)strtod(thisStr,0);
  XtFree(thisStr);

  thisVal = MAX(thisVal,0.0);
  thisVal = MIN(thisVal,distanceMax);
  /* Now show them if it's changed */
  sprintf(outStr,"%f",thisVal);
  XmTextSetString(widget,outStr);

  /* Put this value into "scroll space" */
  thisVal *= 1000.0;

  if(widget==ppMinTextW)
    {
      minXYplot = thisVal / 1000.0;
      XtVaSetValues(ppMinScrollW,
		    XmNvalue, (int)(thisVal),
		    NULL);
    }
  else if(widget==ppMaxTextW)
    {
      maxXYplot = thisVal / 1000.0;
      XtVaSetValues(ppMaxScrollW,
		    XmNvalue, (int)(thisVal),
		    NULL);
    }
  else
    printf("Don't know...\n");

  markXYplotOutliers();

}/* PPMINMAX_CHANGE */



/***************************************************************
 * ppMinMax_scroll_change() - Whenever ppMinMax_scroll scroll 
 * widgets changed call here.
 */
void ppMinMax_scroll_change(Widget widget, int *tag, 
			    XmScrollBarCallbackStruct *callback_data)
{/* PPMINMAX_SCROLL_CHANGE */
  char *ppMinMax_scroll,outStr[64];
  XmString simpleString;
  int i,newMaxCell;
  float scrollVal;
  if(!inited)
    goto notInited;

  scrollVal = (float)callback_data->value;
  
  sprintf(outStr,"%f",scrollVal/1000.0); 

  if(*tag==PPMIN)
    {/* */ 
      XmTextSetString(ppMinTextW,outStr);
      minXYplot = scrollVal / 1000.0;
    }/* */
  else if (*tag==PPMAX)
    {/* ppmax */
      XmTextSetString(ppMaxTextW,outStr);
      maxXYplot = scrollVal / 1000.0;
    }/* ppmax */

  /* if((callback_data->reason != XmCR_DRAG))  */
    markXYplotOutliers();

  XFlush(dsply);

 notInited:;

}/* PPMINMAX_SCROLL_CHANGE */





/*************************************************************
 * symmetry_activate() - They've chosen a new symmetry.
 */
void symmetry_activate(Widget widget, int *tag, 
		       XmAnyCallbackStruct *callback_data)
{/* symmetry_ACTIVATE */

  symmetry = *tag;
  DrawICOSWindow(); 

}/* symmetry_ACTIVATE */





/********************************************************
 * diamText_change() - They want a different diameter.
 */
void diamText_change(Widget widget, int *tag, 
		     XmAnyCallbackStruct *callback_data)
{/* DIAMTEXT_CHANGE */
  char *diamStr;
  float diameter;
  float dThe,move1,move2,move3,move4,move5;
  char outStr[2048];

  diamStr = XmTextGetString(diamTextWidget);
  diameter = (float)strtod(diamStr,0);
  XtFree(diamStr);

  if(whichViewMove == MOVEMENT)
    {/* Movement of surface features */
      if(diameter > 0.0)
	{/* Calc it */
	  dThe = 0.10;
	  move1 = (2.0 * PI * (float)diameter * dThe)/360.0;
	  dThe = 0.25;
	  move2 = (2.0 * PI * (float)diameter * dThe)/360.0;
	  dThe = 0.50;
	  move3 = (2.0 * PI * (float)diameter * dThe)/360.0;
	  dThe = 1.0;
	  move4 = (2.0 * PI * (float)diameter * dThe)/360.0;
	  dThe = 2.0;
	  move5 = (2.0 * PI * (float)diameter * dThe)/360.0;
	  
	  sprintf(outStr,
		  "\t        \t\t\t Del_theta (degrees)\n\
\tParticle\t\t\t \n\
\tDiameter\t\t2.0\t1.0\t0.5\t0.25\t0.10\n\
\t--------\t\t---------------------------------------\n\
\t%.0f\t\t\t%.1f\t%.1f\t%.1f\t%.1f\t%.2f\n",
		  diameter, move5,move4,move3,move2,move1);
	  XmTextSetString(movementTextWidget,outStr);
	  XFlush(dsply);
	  
	}/* Calc it */
    }/* Movement of surface features */
  else
    {/* Views per ASU */
	  
      sprintf(outStr,
	      "\t        \t\t\t Symmetry\n\
\tAngle   \t\t\t \n\
\tDelta   \t\t    532\t     52\t      5\t      1\n\
\t--------\t\t----------------------------------\n\
\t%.2f\t\t\t%7d\t%7d\t%7d\t%7d\n\
\t%.2f\t\t\t%7d\t%7d\t%7d\t%7d\n\
\t%.2f\t\t\t%7d\t%7d\t%7d\t%7d\n\
\t%.2f\t\t\t%7d\t%7d\t%7d\t%7d\n\
\t%.2f\t\t\t%7d\t%7d\t%7d\t%7d\n\
\t%.2f\t\t\t%7d\t%7d\t%7d\t%7d\n",
	      3.0,52,266,519,2353,
	      2.0,96,573,1127,5253,
	      1.0,382,2173,4309,20809,
	      0.5,1448,8471,16869,82869,
	      0.25,5653,33439,66733,330751,
	      0.1,34427,207358,414353,2064448);
      XmTextSetString(movementTextWidget,outStr);
      XFlush(dsply);
      
    }/* Views per ASU */


}/* DIAMTEXT_CHANGE */




/*************************************************************
 * views_activate() - Setup for view and get movementscreen
 * up on the display.
 */
void views_activate(Widget widget, int *tag, 
		    XmAnyCallbackStruct *callback_data)
{/* views_ACTIVATE */
  XmString simpleString;

  whichViewMove = VIEWS;

  XmTextSetString(diamTextWidget, "Not yet.");

  sprintf(outStr,"Number of Views per ASU");
  createSimpleString(&simpleString,outStr);
  XtSetArg(arglst[0], XmNdialogTitle, simpleString );
  XtSetArg(arglst[1], XmNrecomputeSize, FALSE );
  XtSetValues(movementscreen, arglst, 2); 
  XFlush(dsply);
  XmStringFree(simpleString); 
  XFlush(dsply);

  sprintf(outStr,"Angle Delta");
  createSimpleString(&simpleString,outStr);
  XtSetArg(arglst[0], XmNlabelString, simpleString );
  XtSetArg(arglst[1], XmNrecomputeSize, FALSE );
  XtSetValues(movementLabelWidget, arglst, 2); 
  XFlush(dsply);
  XmStringFree(simpleString); 
  XFlush(dsply);

  /* Hard code this for now ... */
  sprintf(outStr,
	      "\t        \t\t\t Symmetry\n\
\tAngle   \t\t\t \n\
\tDelta   \t\t    532\t     52\t      5\t      1\n\
\t--------\t\t----------------------------------\n\
\t%.2f\t\t\t%7d\t%7d\t%7d\t%7d\n\
\t%.2f\t\t\t%7d\t%7d\t%7d\t%7d\n\
\t%.2f\t\t\t%7d\t%7d\t%7d\t%7d\n\
\t%.2f\t\t\t%7d\t%7d\t%7d\t%7d\n\
\t%.2f\t\t\t%7d\t%7d\t%7d\t%7d\n\
\t%.2f\t\t\t%7d\t%7d\t%7d\t%7d\n",
	      3.0,52,266,519,2353,
	      2.0,96,573,1127,5253,
	      1.0,382,2173,4309,20809,
	      0.5,1448,8471,16869,82869,
	      0.25,5653,33439,66733,330751,
	      0.1,34427,207358,414353,2064448);
  XmTextSetString(movementTextWidget,outStr);
  XFlush(dsply);

  raiseIt(movementscreen);
  
}/* views_ACTIVATE */



/*************************************************************
 * movement_activate() - Get movement screen on display.
 */
void movement_activate(Widget widget, int *tag, 
		       XmAnyCallbackStruct *callback_data)
{/* movement_ACTIVATE */
  XmString simpleString;

  whichViewMove = MOVEMENT;

  XmTextSetString(diamTextWidget, " ");

  sprintf(outStr,"Movement of Surface Features");
  createSimpleString(&simpleString,outStr);
  XtSetArg(arglst[0], XmNdialogTitle, simpleString );
  XtSetArg(arglst[1], XmNrecomputeSize, FALSE );
  XtSetValues(movementscreen, arglst, 2); 
  XFlush(dsply);
  XmStringFree(simpleString); 
  XFlush(dsply);

  sprintf(outStr,"Particle Diameter:");
  createSimpleString(&simpleString,outStr);
  XtSetArg(arglst[0], XmNlabelString, simpleString );
  XtSetArg(arglst[1], XmNrecomputeSize, FALSE );
  XtSetValues(movementLabelWidget, arglst, 2); 
  XFlush(dsply);
  XmStringFree(simpleString); 
  XFlush(dsply);

  /* Default this to something */
  sprintf(outStr,
		  "\t        \t\t\t Del_theta (degrees)\n\
\tParticle\t\t\t \n\
\tDiameter\t\t2.0\t1.0\t0.5\t0.25\t0.10\n\
\t--------\t\t---------------------------------------\n");
  XmTextSetString(movementTextWidget,outStr);
  XFlush(dsply);

  raiseIt(movementscreen);
  
}/* movement_ACTIVATE */



/*************************************************************
 * movement_deactivate() - Gte rid of movement screen.
 */
void movement_deactivate(Widget widget, int *tag, 
			 XmAnyCallbackStruct *callback_data)
{/* movement_DEACTIVATE */

  XtUnmanageChild(movementscreen);
  
}/* movement_DEACTIVATE */



/* ************************************************************
 * info_list_create() - Remember widget
*/
void info_list_create(Widget widget, int *tag, 
		      XmAnyCallbackStruct *callback_data)
{/* INFO_LIST_CREATE */

  infoListW= widget;

}/* INFO_LIST_CREATE */


/* ************************************************************
 * info_list_proc() - Call here everytime they make a new
 * choice form the info MASTER list.  Update the appropriate
 * variables and make the display appear as it should.
*/
void info_list_proc(Widget widget, int *tag, 
		    XmListCallbackStruct *list)
{/* INFO_LIST_PROC */
  Arg arglist[2];
  int i, foundOne;

  foundOne= (int) (list->item_position);   /* Numbers from 1 not 0 */
  globAskAnswer = foundOne-1;
  XtUnmanageChild(infoAnswerscreen);
  gFinished= TRUE;

}/* INFO_LIST_PROC */




/*************************************************************
 * projASU_activate() - Tell them how to project.
 */
void projASU_activate(Widget widget, int *tag, 
		      XmAnyCallbackStruct *callback_data)
{/* PROJASU_ACTIVATE */

  XtUnmanageChild(parmTogsW);
  XtUnmanageChild(parmheadW);
  XtManageChild(projasuW);
  XFlush(dsply);

  askThem("Enter the desired Delta Angle below and hit the\n\
Return key.  OR enter the desired theta, phi and\n\
omega below and click the 'Project This:' button.",
	  FALSE,
	  XmDIALOG_INFORMATION);

}/* PROJASU_ACTIVATE */





/*************************************************************
 * projasu_cr_activate() - Do the projection from the 
 * param data and the MAP.
 */
void projasu_cr_activate(Widget widget, int *tag, 
			 XmAnyCallbackStruct *callback_data)
{/* PROJASU_CR_ACTIVATE */
  char  *projStr;
  int   i,j,k,l,m;
  int   nthe,nphi,curPhi,curTheta;
  float dthe,dphi_90,lastTheta;
  float the1, the2, phi1;
  int   lun,yOut1,yOut2,thisX,thisY,zOut;
  float thisOmega,thisPhi,thisTheta,thisSF;
  float maxF,minF,z1,z2,thisFloat;
  float *projn,*map2d,*thisQ;
  float *thisImg;

  if(EQU(paramMapFilename,""))
    {/* Can't go on */
      askThem("No MAP open can't do the projections.",
	      FALSE,XmDIALOG_INFORMATION);
    }/* Can't go on */
  else
    {/* Do it */

      if(haveParamFile)
	{/* Get rid of any parameter file we've got */
	  haveParamFile = FALSE;
	  sprintf(inParamFilename,"");
	  sprintf(paramBoxesFilename,"");
	  XmTextSetString(paramFileTextW,inParamFilename); 
	  purdueHeader.nimg= 0;
	  lastPackedNumImgs= 0;
	  Free_Particle_Info(&info);
	}/* Get rid of any parameter file we've got */
      
      /* Suppose they've got a previous packed file read */
      for(i=0;i<purdueHeader.nimg;i++)
	if(packImg[i])
	  {free(packImg[i]);
	   packImg[i]= NULL;}
      purdueHeader.nimg = 0;
      
      /* We've got to fake robem out that this is a packed
       * file... */
      fileType= PACKEDPORTABLE;
      /* Now clip this guy to our colormap */ 
      scalePackFactor = (((float)tryColors - 7.0) / 256.0);
      projASU = TRUE;

      /* They can write this out as RAW */
      XtSetArg(arglst[0], XmNsensitive, TRUE);
      XtSetValues(rawButtonW, arglst, 1);   

      projStr = XmTextGetString(projasuTextW);
      projASUdeltaAng = (int)((float)strtod(projStr,0)+0.4999);
      projASUdeltaAng = MAX(projASUdeltaAng,1);
      XtFree(projStr);
      sprintf(outStr,"%d",projASUdeltaAng);
      XmTextSetString(projasuTextW,outStr);
      XFlush(dsply);      

      purdueHeader.numCols = mapCols;
      purdueHeader.numRows = mapRows;
      purdueHeader.numSecs = 1;
      packRadius = (mapCols/2);

      
      /* C  ***************************************************************
	 C  *    Determine NVIEW, the number of model projection views    *
	 C  ***************************************************************
	 */
      if(projASUdeltaAng < 1)
	projASUdeltaAng = 1;
      nthe    = 0;
      dthe    = (float)projASUdeltaAng;
      dphi_90 = (float)projASUdeltaAng;


      getnview_rwa_ (&symmetry, &dthe, &dphi_90, &nthe, &nview, &maxPhi);


      /* Allocate for every one we need to enclose
       * these in a square box.
       */
      purdueHeader.nimg = (nthe * maxPhi);

      /* Allocate all storage */
      /* Allocate it whether you'll use it or not */
      for(m=0;m<purdueHeader.nimg;m++)
	{/* Allocate all */
	  CK(packImg[m]= (float *)malloc(
				       sizeof(float) *
				       purdueHeader.numCols*
				       purdueHeader.numRows));
	  /* Init this out */
	  thisImg = packImg[m];
	  for(i=0;i<purdueHeader.numRows;i++)
	    {
	      yOut1 = i*purdueHeader.numCols;
	      for(j=0;j<purdueHeader.numCols;j++)
		thisImg[(yOut1+j)]= 0.0;
	    }
	}/* Allocate all */

      /* Calc what projections we'll need */
      the1 =  0.0;
      the2 = 90.0;
      phi1 =  0.0;
      /* C	******************************************************
	 C	*  Exceptions to the above values:                   *
	 C	*  SYM = 1    for asymmetric particle,               *
	 C	*      = 2-20 for particle with nfold about z-axis,  *
	 C	*      = 532  for icosahedral particle.              *
	 C	******************************************************
	 */
      if (symmetry >= 1 && symmetry <= 20) 
	the2 = 180.0;
      else if (symmetry == 532) 
	the1 = 69.09484255211070;

      if(thetas)
	free(thetas);
      if(phis)
	free(phis);
      CK(thetas = (float *)malloc(sizeof(float)*nview));
      CK(phis   = (float *)malloc(sizeof(float)*nview));

      calc_tp_rwa_ (thetas, phis, 
		    &the1, &the2, 
		    &phi1, 
		    &dthe, &dphi_90, 
		    &nthe, &nphi, 
		    &nview, &symmetry);

      /* Init the display */
      DrawICOSWindow();

      /* Now project all */
      CK(projn = (float *)malloc(sizeof(float)*mapCols*(mapRows+2)));
      CK(map2d = (float *)malloc(sizeof(float)*mapCols*(mapRows+2)*nview));

      lun = 7;
      for(i=0;i<nview;i++)
	{/* DO all projs */
	  
	  /*
	   **************************************************************
	   *    NOTE USE OF OMEGA = 90.0 TO GET CORRECT ORIENTATIONS    *
	   **************************************************************
	   */
	  thisOmega = 90.0;
	  thisOmega = 180.0;
	  thisTheta = thetas[i];
	  thisPhi   = phis[i];
	  z1 = 1.0;
	  z2 = (float)mapSecs;

	  sprintf(outStr,"Projecting theta[%f] phi[%f].",thetas[i],phis[i]);
	  XmTextSetString(partPointTextW,outStr);
	  XFlush(dsply);

	  map_prj_ (&lun, 
		    &mapCols, &mapRows, &mapSecs, 
		    projn, (map2d+(i*mapCols*mapRows)),  
		    float_image_buffer1,
		    &thisTheta, &thisPhi, &thisOmega, 
		    &z1, &z2,
		    &map_alp, &map_bet, &map_gam);

	  /* Give feedback */
	  thisX = (int)phis[i];
	  thisY = (int)thetas[i];
	  convertXY(&thisX, &thisY, phis[i], thetas[i]);
	  
	  XDrawLine(dsply,icosa_window1,image_gcBLUE,
		    (thisX - 5),(thisY-5),
		    (thisX + 5),(thisY+5));
	  XDrawLine(dsply,icosa_window1,image_gcBLUE,
		    (thisX + 5),(thisY-5),
		    (thisX - 5),(thisY+5));

	  XFlush(dsply);
	  
	}/* DO all projs */

      minF = maxF = map2d[0];
      for(i=0;i<nview;i++)
	{/* DO all projs */
	  zOut = i*mapCols*mapRows;
	  /* Find min/max */
	  for(y=0;y<mapRows;y++)
	    {/* Rows */
	      yOut2 = y*mapCols;
	      for(x=0;x<mapCols;x++)
		{/* Cols */
		  minF = MIN(map2d[(zOut+yOut2+x)],minF);
		  maxF = MAX(map2d[(zOut+yOut2+x)],maxF);
		}/* Cols */
	    }/* Rows */
	}/* DO all projs */

      if(maxF > minF)
	thisSF = (254.0) / (maxF-minF);
      else thisSF = 1.0;

      curTheta = curPhi = 0;
      lastTheta = thetas[0];
      for(i=0;i<nview;i++)
	{/* DO all projs */
	  if(thetas[i] != lastTheta)
	    {/* We've got a new row */
	      curTheta++;
	      curPhi = 0;
	      lastTheta = thetas[i];
	    }/* We've got a new row */

	  /* Point to the appropriate packed image */
	  thisImg = packImg[((curTheta*maxPhi)+curPhi)];
	  curPhi++;

	  /* We've got to flip the projection in the Y direction.
	   * Tim draws from the bottom up -we're opposite
	   */
	  zOut = i * mapCols*mapRows;
	  for(y=0;y<mapRows;y++)
	    {/* Rows */
	      yOut1 = (mapRows - y - 1)*mapCols;
	      yOut2 = y*mapCols;
	      for(x=0;x<mapCols;x++)
		{/* Cols */
		  thisImg[yOut2+x]= (float)((map2d[(zOut+yOut1+x)]-minF)*
					   thisSF);
		}/* Cols */
	    }/* Rows */

	  /* Give feedback */
	  thisX = (int)phis[i];
	  thisY = (int)thetas[i];
	  convertXY(&thisX, &thisY, phis[i], thetas[i]);
	  
	  XDrawLine(dsply,icosa_window1,image_gcRED,
		    (thisX - 5),(thisY-5),
		    (thisX + 5),(thisY+5));
	  XDrawLine(dsply,icosa_window1,image_gcRED,
		    (thisX + 5),(thisY-5),
		    (thisX - 5),(thisY+5));
	  XFlush(dsply);
	}/* DO all projs */

      free(projn);
      free(map2d);

      /* Global - only freed() the next time
	 free(thetas);
	 free(phis);
	 */

      stuffPackedImage(purdueHeader.numCols,
		       purdueHeader.numRows,
		       packImg,purdueHeader.nimg,
		       1);  /* Start at zero'th(1) image */

      DrawWindow();

    }/* Do it */
  
}/* PROJASU_CR_ACTIVATE */



/********************************************************
 * squareText_changed() - They've changed the sqaure 
 * params.
 */
void squareText_changed(Widget widget, int *tag, 
			XmAnyCallbackStruct *callback_data)
{/* SQUARETEXT_CHANGED */
  char *squareStr;

  squareStr = XmTextGetString(widget);
  makeSquareBox = (int)strtod(squareStr,0);
  XtFree(squareStr);

}/* SQUARETEXT_CHANGED */




/*************************************************************
 * project_one() - Project from MAP.
 */
void project_one(Widget widget, int *tag, 
		 XmAnyCallbackStruct *callback_data)
{/* PROJECT_ONE */
  char  *projStr;
  int   i,j,k,l,m;
  int   nthe,nphi,curPhi,curTheta;
  float dthe,dphi_90,lastTheta,thisTheta;
  float the1, the2, phi1;
  int   lun,yOut1,yOut2,thisX,thisY,zOut;
  float thisOmega,thisPhi,thisSF;
  float maxF,minF,z1,z2,thisFloat;
  float *projn,*map2d,*thisQ;
  float *thisImg;

  if(EQU(paramMapFilename,""))
    {/* Can't go on */
      askThem("No MAP open can't do the projections.",
	      FALSE,XmDIALOG_INFORMATION);
    }/* Can't go on */
  else
    {/* Do it */

      if(haveParamFile)
	{/* Get rid of any parameter file we've got */
	  haveParamFile = FALSE;
	  sprintf(inParamFilename,"");
	  sprintf(paramBoxesFilename,"");
	  XmTextSetString(paramFileTextW,inParamFilename); 
	  purdueHeader.nimg= 0;
	  lastPackedNumImgs= 0;
	  Free_Particle_Info(&info);
	}/* Get rid of any parameter file we've got */
      
      /* Suppose they've got a previous packed file read */
      for(i=0;i<purdueHeader.nimg;i++)
	if(packImg[i])
	  {free(packImg[i]);
	   packImg[i]= NULL;}
      purdueHeader.nimg = 0;
      
      /* We've got to fake robem out that this is a packed
       * file... */
      fileType= PACKEDPORTABLE;
      /* Now clip this guy to our colormap */ 
      scalePackFactor = (((float)tryColors - 7.0) / 256.0);
      projASU = TRUE;

      /* They can write this out as RAW */
      XtSetArg(arglst[0], XmNsensitive, TRUE);
      XtSetValues(rawButtonW, arglst, 1);   

      projStr = XmTextGetString(thetaProjW);
      thisTheta = (float)strtod(projStr,0);
      XtFree(projStr);
      projStr = XmTextGetString(phiProjW);
      thisPhi = (float)strtod(projStr,0);
      XtFree(projStr);
      projStr = XmTextGetString(omegaProjW);
      thisOmega = (float)strtod(projStr,0);
      XtFree(projStr);
      XFlush(dsply);      

      purdueHeader.numCols = mapCols;
      purdueHeader.numRows = mapRows;
      purdueHeader.numSecs = 1;
      packRadius = (mapCols/2);

      
      /* NVIEW is just one here	 */
      maxPhi = nview= purdueHeader.nimg = 1;

      /* Allocate all storage */
      /* Allocate it whether you'll use it or not */
      for(m=0;m<purdueHeader.nimg;m++)
	{/* Allocate all */
	  CK(packImg[m]= (float *)malloc(
				       sizeof(float) *
				       purdueHeader.numCols*
				       purdueHeader.numRows));
	  /* Init this out */
	  thisImg = packImg[m];
	  for(i=0;i<purdueHeader.numRows;i++)
	    {
	      yOut1 = i*purdueHeader.numCols;
	      for(j=0;j<purdueHeader.numCols;j++)
		thisImg[(yOut1+j)]= 0;
	    }
	}/* Allocate all */

      if(thetas)
	free(thetas);
      if(phis)
	free(phis);
      CK(thetas = (float *)malloc(sizeof(float)*nview));
      CK(phis   = (float *)malloc(sizeof(float)*nview));
      thetas[0] = thisTheta;
      phis[0]   = thisPhi;

      /* Init the display */
      DrawICOSWindow();

      /* Now project all */
      CK(projn = (float *)malloc(sizeof(float)*mapCols*(mapRows+2)));
      CK(map2d = (float *)malloc(sizeof(float)*mapCols*(mapRows+2)*nview));

      lun = 7;
      for(i=0;i<nview;i++)
	{/* DO all projs */
	  
	  /*
	   **************************************************************
	   *    NOTE USE OF OMEGA = 90.0 TO GET CORRECT ORIENTATIONS    *
	   **************************************************************
	  thisOmega = 90.0;
	  thisOmega = 180.0;
	   */
	  /* Leave thisOmega alone ??? */

	  thisTheta = thetas[i];
	  thisPhi   = phis[i];
	  z1 = 1.0;
	  z2 = (float)mapSecs;

	  sprintf(outStr,"Projecting theta[%f] phi[%f].",thetas[i],phis[i]);
	  XmTextSetString(partPointTextW,outStr);
	  XFlush(dsply);

	  map_prj_ (&lun, 
		    &mapCols, &mapRows, &mapSecs, 
		    projn, (map2d+(i*mapCols*mapRows)),  
		    float_image_buffer1,
		    &thisTheta, &thisPhi, &thisOmega, 
		    &z1, &z2,
		    &map_alp, &map_bet, &map_gam);

	  /* Give feedback */
	  thisX = (int)phis[i];
	  thisY = (int)thetas[i];
	  convertXY(&thisX, &thisY, phis[i], thetas[i]);
	  
	  XDrawLine(dsply,icosa_window1,image_gcBLUE,
		    (thisX - 5),(thisY-5),
		    (thisX + 5),(thisY+5));
	  XDrawLine(dsply,icosa_window1,image_gcBLUE,
		    (thisX + 5),(thisY-5),
		    (thisX - 5),(thisY+5));

	  XFlush(dsply);
	  
	}/* DO all projs */

      minF = maxF = map2d[0];
      for(i=0;i<nview;i++)
	{/* DO all projs */
	  zOut = i*mapCols*mapRows;
	  /* Find min/max */
	  for(y=0;y<mapRows;y++)
	    {/* Rows */
	      yOut2 = y*mapCols;
	      for(x=0;x<mapCols;x++)
		{/* Cols */
		  minF = MIN(map2d[(zOut+yOut2+x)],minF);
		  maxF = MAX(map2d[(zOut+yOut2+x)],maxF);
		}/* Cols */
	    }/* Rows */
	}/* DO all projs */

      if(maxF > minF)
	thisSF = (254.0) / (maxF-minF);
      else thisSF = 1.0;

      curTheta = curPhi = 0;
      lastTheta = thetas[0];
      for(i=0;i<nview;i++)
	{/* DO all projs */
	  if(thetas[i] != lastTheta)
	    {/* We've got a new row */
	      curTheta++;
	      curPhi = 0;
	      lastTheta = thetas[i];
	    }/* We've got a new row */

	  /* Point to the appropriate packed image */
	  thisImg = packImg[((curTheta*maxPhi)+curPhi)];
	  curPhi++;

	  /* We've got to flip the projection in the Y direction.
	   * Tim draws from the bottom up -we're opposite
	   */
	  zOut = i * mapCols*mapRows;
	  for(y=0;y<mapRows;y++)
	    {/* Rows */
	      yOut1 = (mapRows - y - 1)*mapCols;
	      yOut2 = y*mapCols;
	      for(x=0;x<mapCols;x++)
		{/* Cols */
		  thisImg[yOut2+x]= (float)((map2d[(zOut+yOut1+x)]-minF)*
					   thisSF);
		}/* Cols */
	    }/* Rows */

	  /* Give feedback */
	  thisX = (int)phis[i];
	  thisY = (int)thetas[i];
	  convertXY(&thisX, &thisY, phis[i], thetas[i]);
	  
	  XDrawLine(dsply,icosa_window1,image_gcRED,
		    (thisX - 5),(thisY-5),
		    (thisX + 5),(thisY+5));
	  XDrawLine(dsply,icosa_window1,image_gcRED,
		    (thisX + 5),(thisY-5),
		    (thisX - 5),(thisY+5));
	  XFlush(dsply);
	}/* DO all projs */

      free(projn);
      free(map2d);

      /* Global - only freed() the next time
	 free(thetas);
	 free(phis);
	 */

      stuffPackedImage(purdueHeader.numCols,
		       purdueHeader.numRows,
		       packImg,purdueHeader.nimg,
		       1);  /* Start at zero'th(1) image */

      DrawWindow();

    }/* Do it */
  
}/* PROJECT_ONE */





/*************************************************************
 * projASUunitCell_work() - Draw all particles in this unit 
 * cell display.
 */
void projASUunitCell_work(Widget widget, int *tag, 
			  XmAnyCallbackStruct *callback_data)
{/* projASUunitCell_work */
  P_INFO *thisInfo;
  int thisX,thisY,thisXp1,thisYp1;
  float thisTheta, thisPhi, negPhiMax;
  char outStr[256];
  
  unitCell_background();

  for(i=0;i<nview;i++)
    {/* Do all of them */
      thisX = (int)phis[i];
      thisY = (int)thetas[i];
      convertXY(&thisX, &thisY, phis[i], thetas[i]);
      
      XDrawLine(dsply,icosa_window1,image_gcRED,
		(thisX - 5),(thisY-5),
		(thisX + 5),(thisY+5));
      XDrawLine(dsply,icosa_window1,image_gcRED,
		(thisX + 5),(thisY-5),
		(thisX - 5),(thisY+5));
    }/* Do all of them */


}/* projASUunitCell_work */



/*************************************************************
 * rescale_densities() - Rescale all of the density values 
 * in this MAP.
*/
void rescale_densities(Widget widget, int *tag, 
		       XmAnyCallbackStruct *callback_data)
{/* RESCALE_DENSITIES */
  int   i,j;
  int   lun,file_id,map_incore;
  float newMinMapPixel,newMaxMapPixel,map_scalef;
  char  *thisStr;
  int   *thisPkImg;

  thisStr = XmTextGetString(newMinDensTextW);
  newMinMapPixel = (float)strtod(thisStr,0);
  XtFree(thisStr);
  thisStr = XmTextGetString(newMaxDensTextW);
  newMaxMapPixel = (float)strtod(thisStr,0);
  XtFree(thisStr);
  XFlush(dsply);      

  /*  Need to generalize this to more than just maps */
  printf("ft[%d]  pack[%d]empor[%d]map[%d]\n",
	 fileType,PACKEDPORTABLE,EMPORTABLE,MAP);
  if(fileType==EMPORTABLE || fileType==BYTEPORTABLE || fileType==PORTABLE)
    {/* scanned image */
      /* Use img_scale() in imglib.f */
    }/* scanned image */
  else  if(fileType==PACKEDPORTABLE)
    {/* Do for packed boxes */
      /* Use img_scale() in imglib.f */
    }/* Do for packed boxes */
  else if(fileType==MAP)
    {/* Do for MAP */
      lun = 1;
      map_scalef = 1.0;
      file_id = 0;
      map_incore = 1;
      
      
      map_scale_ (&lun, float_image_buffer1, 
		  &numCols, &numRows, &origNumSecs,
		  &minMapPixel, &maxMapPixel, 
		  &newMinMapPixel, &newMaxMapPixel, 
		  &purdueHeader.img_min, &purdueHeader.img_max,
		  &purdueHeader.img_avg, &purdueHeader.img_std, &map_scalef, 
		  &file_id, &map_incore);

      
      secMaxPixel = maxMapPixel = newMaxMapPixel;
      secMinPixel = minMapPixel = newMinMapPixel;
      
      calcMapStats(float_image_buffer1,
		   &purdueHeader.img_bkg, 
		   &purdueHeader.map_min,
		   &purdueHeader.map_max, 
		   &purdueHeader.img_avg, 
		   &purdueHeader.img_std,
		   &numCols, &numRows, &origNumSecs);
      
      /*
       * CALL MAP_SCALE (LUN, MAP, NCOL, NROW, NSEC, DENS1, DENS2,
       * 1 DENS3, DENS4, MAP_MIN, MAP_MAX, MAP_AVG, MAP_STD, MAP_SCALEF, 
       * 2 FILE_ID, MAP_INCORE)
       */
    }/* Do for MAP */
  
  XtUnmanageChild(rescalescreen);

}/* RESCALE_DENSITIES */



/***************************************************************
 * rescale_activate() - Get interface about how to rescale this
 * MAP's densities on the display.
 */
void rescale_activate(Widget widget, int *tag, 
		      XmAnyCallbackStruct *callback_data)
{/* RESCALE_ACTIVATE */
  XmString simpleString;

  sprintf(outStr,"Old Min Density =  %.2f   Enter NEW Min: ",minMapPixel);
  createSimpleString(&simpleString,outStr);
  XtSetArg(arglst[0], XmNlabelString, simpleString );
  XtSetArg(arglst[1], XmNrecomputeSize, FALSE );
  XtSetValues(oldMinDensLabelW, arglst, 2); 
  XmStringFree(simpleString); 
  XFlush(dsply);

  sprintf(outStr,"Old Max Density =  %.2f   Enter NEW Max: ",maxMapPixel);
  createSimpleString(&simpleString,outStr);
  XtSetArg(arglst[0], XmNlabelString, simpleString );
  XtSetArg(arglst[1], XmNrecomputeSize, FALSE );
  XtSetValues(oldMaxDensLabelW, arglst, 2); 
  XmStringFree(simpleString); 
  XFlush(dsply);

  raiseIt(rescalescreen);

}/* RESCALE_ACTIVATE */



/*************************************************************
 * rescale_deactivate() - Get rid of rescale interface.
 */
void rescale_deactivate(Widget widget, int *tag, 
			XmAnyCallbackStruct *callback_data)
{/* RESCALE_DEACTIVATE */

  XtUnmanageChild(rescalescreen);

}/* RESCALE_DEACTIVATE */






/*************************************************************
 * calcSF_work() - Do what ti takes to get structure factors
 * calculated from the given MAP.
 */
void calcSF_work(int thisCalcSFs)
{/* CALCSF_WORK */
  int numFramesMm;
  XmString simpleString;
  float resHi;

  if(thisCalcSFs)
    {/* Calc them up */
      sprintf(outStr,"Calculating Structure Factors");
      checkStatus(outStr);
      resHi = 2.01; /* Let's chop it down by a factor of 2 */
      map2sfs(resHi);
      
      sprintf(outStr,"Producing High Resolution MAP");
      checkStatus(outStr);
      numFramesMm = 1;
      resolution_tryit_activate(0,&numFramesMm,0);

      /* Make sure they can do high res if they want etc. */
      XtSetArg(arglst[0], XmNsensitive, TRUE );
      XtSetValues(loHiResWidget, arglst, 1); 
      
      /* Now prevent them from trying to manipulate these - e.g.
       * make them reread from disk to manipulate. */
      XtSetArg(arglst[0], XmNsensitive, FALSE );
/*      XtSetValues(fixButtonW, arglst, 1);  */
      XtSetValues(truncaddButtonW, arglst, 1); 
      XtSetValues(rescaleButtonW, arglst, 1); 
      XtSetValues(modelWidget, arglst, 1); 
      XtSetValues(subtractMapButtonW, arglst, 1); 
      XtSetValues(calcSFButtonW, arglst, 1); 
      /* XtSetValues(RcalcSFButtonW, arglst, 1);  */
      XtSetValues(hardRotateButtonW, arglst, 1); 
      XtSetValues(invertButtonW, arglst, 1); 
      XtSetValues(rotateButtonW, arglst, 1); 
      
      if(!immediateResource)
	{/* Turn immediate update ON */
	  XtSetArg(arglst[0], XmNset, TRUE);
	  XtSetValues(immediateUpdateW, arglst, 1); 
	  immediateMapShade = TRUE;
	  XFlush(dsply);
	}/* Turn immediate update ON */
      
      zoomFactor = 1.0;  /* Let's zoom it up */
      XmTextSetString(zoomTextW, "2.00");  

      /* Their shaded views are no longer valid - show them 
       * their new sections
       * shadedMaps= FALSE;
       * dispSections = TRUE;
       * XtSetArg(arglst[0], XmNlabelPixmap, sectionPixmap);
       * XtSetArg(arglst[1], XmNselectPixmap, unsectionPixmap);
       * XtSetArg(arglst[2], XmNrecomputeSize, FALSE);
       * XtSetValues(sectionWidget, arglst, 3); 
       */
      if(dispSections)
	{/* Update the sections - if shaded views then they have to recalc */
	  /* Now stuff the display buffer from our new FIXed map */
	  stuffMapImage((float *)map_image_buffer, 
			(float *)float_image_buffer1, 
			file_width,file_height,
			nsecs,imgNum, 0);
	  DrawWindow();
	}/* Update the sections - if shaded views then they have to recalc */
      else
	{/* If we're already doing rendering then setup the scrolls
	  * appropriately */
	  mapShadeScreen_activate(0,0,0);
	}/* If we're already doing rendering then setup the scrolls
	  * appropriately */

    }/* Calc them up */
  else
    {/* Prevent any high res stuff */
      sprintf(outStr,"Low Resolution ON");
      XtSetArg(arglst[0], XmNsensitive, FALSE );
      XtSetValues(pixelSizeScrollWidget, arglst, 1); 
      XtSetValues(pixelSizeValueWidget, arglst, 1); 
/* WHy are we doing this - it's for the movie???
      XtSetValues(pixelSizeToggleWidget, arglst, 1); 
*/
      /* XtSetValues(loHiResWidget, arglst, 1);  */
      
      XtSetArg(arglst[0], XmNset, FALSE );
/* WHy are we doing this - it's for the movie???
      XtSetValues(pixelSizeToggleWidget, arglst, 1); 
*/
      lowRes = TRUE;
    }/* Prevent any high res stuff */

}/* CALCSF_WORK */




/*************************************************************
 * calcSF_activate() - Setup the interface for calculating
 * structure factors.
 */
void calcSF_activate(Widget widget, int *tag, 
		     XmAnyCallbackStruct *callback_data)
{/* CALCSF_ACTIVATE */
  int i;
  XmString simpleString;

  /* Don't do this here!!!  We'll do it on the fly as they 
     cut their map down and then change the pixel size in the HiRes
     interface ...
     
     calcSF_work(TRUE);
   */

  /* Insert the call to inter3d() here to make them a 100 cubed lowres
   * buffer to work with */

  /* DO this instead */
  /* Make sure they can do high res if they want etc. */
  XtSetArg(arglst[0], XmNsensitive, TRUE );
  XtSetValues(loHiResWidget, arglst, 1); 
      
  /* Now remember these forever */
  lowRes_buffer1 = float_image_buffer1;
  lowResCols = numCols;
  lowResRows = numRows;
  lowResSecs = origNumSecs;
  
  purdueHeader.numCols = numCols;
  purdueHeader.numRows = numRows;
  nsecs= purdueHeader.numSecs = origNumSecs;

  xCen = (double)(numCols/2);
  yCen = (double)(numRows/2);
  zCen = (double)(origNumSecs/2);

  largestDim = numCols;
  largestDim = MAX(largestDim,numRows);
  largestDim = MAX(largestDim,origNumSecs);

  /* We've got to make this larger for asymmetric maps 
   * and all maps that have information "outside" of the
   * normal sphere of our normal maps,
   * or else we just keep cliping the box into a sphere */
  if(numCols==largestDim && numRows==largestDim
     && (origNumSecs)==largestDim)
    {/* Symmetric */
      aSymmetric = FALSE;
      origGlobMaxRadius= globMaxRadius = (double)((double)origNumSecs/2)+5.0;
      globMinRadius = (double)0.0;
    }/* Symmetric */
  else
    {/* asymmetric */
      aSymmetric = TRUE;
      origGlobMaxRadius=  globMaxRadius = (double)((double)origNumSecs);
      globMinRadius = (double)0.0;
    }/* asymmetric */
  
  /* Set scrolls on depthcue screen */
  /* Give feedback */
  sprintf(outStr,"%.0f",globMinRadius);
  createSimpleString(&simpleString,outStr);
  XtSetArg(arglst[0], XmNvalue, outStr );
  XtSetArg(arglst[1], XmNrecomputeSize, FALSE );
  XtSetValues(minRadiusValueWidget, arglst, 2); 
  XFlush(dsply);
  XmStringFree(simpleString); 
  
  sprintf(outStr,"%.0f",globMaxRadius);
  createSimpleString(&simpleString,outStr);
  XtSetArg(arglst[0], XmNvalue, outStr );
  XtSetArg(arglst[1], XmNrecomputeSize, FALSE );
  XtSetValues(maxRadiusValueWidget, arglst, 2); 
  XFlush(dsply);
  XmStringFree(simpleString); 
  
  
  /* Radius is origNumSecs/2 */
  XtVaSetValues(maxRadiusScrollWidget,XmNmaximum, 
		(int)((globMaxRadius*2)+5),NULL);
  XtVaSetValues(maxRadiusScrollWidget,XmNsliderSize, (int)(5),NULL);
  XtVaSetValues(maxRadiusScrollWidget,XmNvalue, (int)(globMaxRadius),NULL);
		
  XtVaSetValues(minRadiusScrollWidget,XmNmaximum, 
		(int)((globMaxRadius*2)+5),NULL);
  XtVaSetValues(minRadiusScrollWidget,XmNsliderSize, (int)(5),NULL);
  XtVaSetValues(minRadiusScrollWidget,XmNvalue, (int)(globMinRadius),NULL);

  XtUnmanageChild(statusScreenWidget);

  i = SURFDENS;
  /* Do we really need this ???  
     mapShade_activate(0,&i,0);      
     *******************************/


}/* CALCSF_ACTIVATE */




/*************************************************************
 * map2sfs()- Convert the current map into structure factors.
*/
void map2sfs(float resHiP)
{/* map2sfs */
  XmString simpleString;
  int i,j,k,m,n,o,origNC,origNR,origNS;
  int hmax,kmax,lmax;
  char  *thisStr;
  float voxSiz,resLo,resHi,resoScale;


#ifdef GETFROMINTERFACE  
  thisStr = XmTextGetString(voxSizTextW);
  voxSiz = (float)strtod(thisStr,0);
  XtFree(thisStr);
  thisStr = XmTextGetString(resLoTextW);
  resLo = (float)strtod(thisStr,0);
  XtFree(thisStr);
  thisStr = XmTextGetString(resHiTextW);
  resHi = (float)strtod(thisStr,0);
  XtFree(thisStr);
  thisStr = XmTextGetString(resoScaleTextW);
  resoScale = (float)strtod(thisStr,0);
  XtFree(thisStr);
  XFlush(dsply);      
#else
  voxSiz = (float)1.0;
  resLo = (float)1000.0;
  resHi = (float)resHiP;
  resoScale = (float)1.0;
#endif /* GETFROMINTERFACE   */

  origNC = numCols;
  origNR = numRows;
  origNS = origNumSecs; 

  if(calcBin)
    {/* BIN this puppy down */
      float totVals;
      int origNCtNR,iOff,jOff;

      numCols /= 2;
      numRows /= 2;
      origNumSecs /= 2;
      nsecs = origNumSecs;
      origNCtNR = origNC * origNR;
      for(i=0;i<(origNS-1);i+=2)
	{/* Sections */
	for(j=0;j<(origNR-1);j+=2)
	  {
	  for(k=0;k<(origNC-1);k+=2)
	    {/* Do it */
	      totVals = 0.0;
	      for(m=0;m<2;m++)
		{
		  iOff = ((i+m)*origNCtNR);
		  for(n=0;n<2;n++)
		    {
		      jOff = ((j+n)*origNC);
		      for(o=0;o<2;o++)
			{/* Accum */
			  totVals += float_image_buffer1[(iOff)+(jOff)+k+o];
			}/* Accum */
		    }
		}
	      float_image_buffer1[((i/2)*numCols*numRows)+
				  ((j/2)*numCols)+(k/2)]= (totVals / 8.0);
	    }/* Do it */
	}
      }/* Sections */
    }/* BIN this puppy down */

  origNC = numCols;
  origNR = numRows;
  origNS = numSecs = origNumSecs;  /* They may have some depthcued stuff */


  map3dt_info_rwa_(&numCols,&numRows,&numSecs,
                   &hmax, &kmax, &lmax, &nsf,
                   &a_cell, &b_cell, &c_cell,
		   &voxSiz, &resLo, &resHi, &resoScale);

  if(numCols != origNC ||
     numRows != origNR ||
     numSecs != origNS )
    {/* We've got to chop it up */
      for(i=0;i<numSecs;i++)
	for(j=0;j<numRows;j++)
	  for(k=0;k<numCols;k++)
	    {/* Do it */
	      float_image_buffer1[(i*numCols*numRows)+
				  (j*numCols)+k]= 
	      float_image_buffer1[(i*origNC*origNR)+
				  (j*origNC)+k];
	    }/* Do it */
    }/* We've got to chop it up */
  
  if(sFacs != NULL)
    free(sFacs);
  sFacs = (strucFact_t *)malloc(sizeof(strucFact_t)*nsf);

  main_map3dt_(&numCols,&numRows,&numSecs, float_image_buffer1, 
	       sFacs, &nsf, &hmax, &kmax, &lmax);


}/* map2sfs */





/****************************************************************
 * resolution_tryit_activate() - They want to display this MAP
 * with a different pixel resolution.
 */
void resolution_tryit_activate(Widget widget, int *tag, 
			       XmAnyCallbackStruct *callback_data)
{/* RESOLUTION_TRYIT_ACTIVATE */
  XmString simpleString;
  int i,j,k,largeCol,largeRow,largeSec;
  float maxReso,largeFactor;
  char  *thisStr;
  float xminP,xmaxP,yminP,ymaxP,zminP,zmaxP,
        dxP,dyP,dzP,resHi;
  int   norien;

  if(lowRes)
    {/* No need to recalc */
      goto noResolve;
    }/* No need to recalc */

  if(pixelSizeSave == globPixelSize)
    {/* No need to recalc */
      goto noResolve;
    }/* No need to recalc */

  /* Let's build the buffer we need from the lowRes buffer */
  numCols = cutMax0Save - cutMin0Save ;
  numRows = cutMax1Save - cutMin1Save ;
  numSecs = origNumSecs = cutMax2Save - cutMin2Save ;

  if((float_image_buffer1) &&
     (float_image_buffer1 != lowRes_buffer1))
    free(float_image_buffer1);


  float_image_buffer1 = (float *)malloc(sizeof(float)*
					numCols*numRows*numSecs);
  for(i=cutMin2Save;i<cutMax2Save;i++)
    {/* DO all sections */
      for(j=cutMin1Save;j<cutMax1Save;j++)
	{/* DO all rows */
	  for(k=cutMin0Save;k<cutMax0Save;k++)
	    {/* DO all cols */

	      float_image_buffer1[(((i-(int)cutMin2Save)*(numCols*numRows)) +
				   ((j-(int)cutMin1Save)*numCols)+
				   (k-(int)cutMin0Save))] =
				     lowRes_buffer1[((i*
						     (lowResCols*lowResRows))+
						    (j*
						     lowResCols)+
						    k)];


	    }/* DO all cols */
	}/* DO all rows */
    }/* DO all sections */

  /* Let's calc our structure factors */
  sprintf(outStr,"Calculating Structure Factors");
  checkStatus(outStr);
  resHi = 2.01; /* Let's chop it down by a factor of 2 */
  map2sfs(resHi);

  zoomFactor = 1.0;  /* Let's zoom it up */
  XmTextSetString(zoomTextW, "2.00");  

  if(sFacs == NULL)
    {/* error */
      askThem("No Structure factors could be produced.",
	      FALSE,XmDIALOG_INFORMATION); 
      goto noResolve;
    }/* error */

  if(!calcLowResMap)
    {/* Give feedback */
      sprintf(outStr,
	      "[%d] Calculating High Resolution Image",*tag);
      XmTextSetString(timeLeftLabelWidget, outStr);
      XFlush(dsply);
    }/* Give feedback */
  
  maxReso = (float)1.0;
  norien = (int)1;
  xminP  = 0.0;
  xmaxP  = 1.0;
  yminP  = 0.0;
  ymaxP  = 1.0;
  zminP  = 0.0;
  zmaxP  = 1.0;
  dxP= dyP= dzP= 1.0;

  if(calcLowResMap)
    {/* Calc the gold std low res map */
      pixelSizeFactor= dxP = dyP = dzP = numCols/100.0;
      xminP  = 0.0;
      xmaxP  = 1.0;
      yminP  = 0.0;
      ymaxP  = 1.0;
      zminP  = 0.0;
      zmaxP  = 1.0;
    }/* Calc the gold std low res map */
  else 
    {/* DO what they've dialed from the 3D rendering screen */
      if(calcBin)
	dxP = dyP = dzP = (globPixelSize / 2.0); /* It's binned */
      else
	dxP = dyP = dzP = globPixelSize;
      /* Remember to cut it like the loRes was cut */
      xminP  =   ((float)cutMin0Save / (float)lowResCols);
      xmaxP  =   ((float)cutMax0Save / (float)lowResCols);
      yminP  =   ((float)cutMin1Save / (float)lowResRows);
      ymaxP  =   ((float)cutMax1Save / (float)lowResRows);
      zminP  =   ((float)cutMin2Save / (float)lowResSecs);
      zmaxP  =   ((float)cutMax2Save / (float)lowResSecs);

      /* We're only dealing with a cut buffer now so just do the whole
       * thing. */
      xminP  = 0.0;
      xmaxP  = 1.0;
      yminP  = 0.0;
      ymaxP  = 1.0;
      zminP  = 0.0;
      zmaxP  = 1.0;

      printf("xyz min/max[%f][%f][%f][%f][%f][%f]\n",
	     xminP,xmaxP,yminP,ymaxP,zminP,zmaxP);
    }/* DO what they've dialed from the 3D rendering screen */

  cutFactor = MAX((xmaxP-xminP),(ymaxP-yminP));
  cutFactor = MAX((zmaxP-zminP),cutFactor);

  /* Now convert Structure factors back into a 3D MAP */
  if((float_image_buffer1) &&
     (float_image_buffer1 != lowRes_buffer1))
    {
      free(float_image_buffer1);
    }

  /* We have to approximate how large our new buffer is going to be 
   * so we can allocate it large enough */
  largeFactor = 1.0 / dxP;
  largeCol = (int)(numCols * largeFactor);
  largeFactor = 1.0 / dyP;
  largeRow = (int)(numRows * largeFactor);
  largeFactor = 1.0 / dzP;
  largeSec = (int)(numSecs * largeFactor);
  float_image_buffer1 = 
    (float *) malloc(sizeof(float)*largeCol*largeRow*largeSec);

  main_emsf_ (&numCols, &numRows, &numSecs, &nsf, 
              sFacs, float_image_buffer1,
              &a_cell, &b_cell, &c_cell, &maxReso,
	      &xminP,&xmaxP,&yminP,&ymaxP,&zminP,&zmaxP,
	      &dxP,&dyP,&dzP,&norien);

  fortranBuffer = TRUE;     /* Next time use fortran to free() this */

  printf("\n\nNew dimensions ->  %d,%d,%d\n",numCols,numRows,numSecs);

  file_width  =  purdueHeader.numCols = numCols;
  file_height =  purdueHeader.numRows = numRows;
  nsecs = purdueHeader.numSecs = purdueHeader.nimg = origNumSecs = numSecs;

  fileBinSecs= numSecs;
  fileBinRows= numRows;
  fileBinCols= numCols;

  xCen = (double)(numCols/2);
  yCen = (double)(numRows/2);
  zCen = (double)(origNumSecs/2);

  /* We may be doing a movie (on pixelSize) so let's decide on 
   * what the largestDim is outside of here. 
   largestDim = numCols;
   largestDim = MAX(largestDim,numRows);
   largestDim = MAX(largestDim,origNumSecs);
  */
  calcMapStats(float_image_buffer1,
	       &purdueHeader.img_bkg, 
	       &purdueHeader.map_min,
	       &purdueHeader.map_max, 
	       &purdueHeader.img_avg, 
	       &purdueHeader.img_std,
	       &numCols, &numRows, &origNumSecs);
  secMaxPixel = maxMapPixel = purdueHeader.map_max;
  secMinPixel = minMapPixel = purdueHeader.map_min;
  if(secMaxPixel != secMinPixel)
    scaleMapFactor= (((float)tryColors - 7.0) / (secMaxPixel-secMinPixel));
  else 
    {
      scaleMapFactor= 1.0;
      maxMapPixel = minMapPixel + 1.0;
      secMaxPixel = secMinPixel + 1.0;
    }

#define WENEEDTHIS 1
#ifdef WENEEDTHIS
  /* We've got to make this larger for asymmetric maps 
   * and all maps that have information "outside" of the
   * normal sphere of our normal maps,
   * or else we just keep cliping the box into a sphere */
  
  /* Radius is origNumSecs/2 */
  XtVaSetValues(maxRadiusScrollWidget,XmNmaximum, 
		(int)((origNumSecs*2)+5),NULL);
  XtVaSetValues(maxRadiusScrollWidget,XmNsliderSize, (int)(5),NULL);
		
  XtVaSetValues(minRadiusScrollWidget,XmNmaximum, 
		(int)((origNumSecs*2)+5),NULL);
  XtVaSetValues(minRadiusScrollWidget,XmNsliderSize, (int)(5),NULL);
  
  printf("Resolve MAX[%f]  MIN[%f]\n",maxMapPixel,minMapPixel);
  secMaxPixel = maxMapPixel;
  secMinPixel = minMapPixel;

#endif /* WE NEEDTHIS */

  /* Give feedback on main window title bar */
  if(!masterProcess)
    sprintf(outStr,"Robem: [%s] - [%d,%d,%d] [%s]",fileName.addr,
	    numCols,numRows,numSecs, "CHILD");
  else
    sprintf(outStr,"Robem: [%s] - [%d,%d,%d]",fileName.addr,
	    numCols,numRows,numSecs);
  createSimpleString(&simpleString,outStr);
  XtSetArg(arglst[0], XmNdialogTitle, simpleString );
  XtSetArg(arglst[1], XmNrecomputeSize, FALSE );
  XtSetValues(main_window_widget, arglst, 2); 
  XFlush(dsply);
  XmStringFree(simpleString); 
  XFlush(dsply);

  pixelSizeSave = globPixelSize;

  reset_xyz_scrolls();

  noResolve:;

}/* RESOLUTION_TRYIT_ACTIVATE */





/* *******************************************************
 * flicker_change() - Toggle the concept fo flicker.
 */
void flicker_change(Widget widget, int *tag, 
		    XmToggleButtonCallbackStruct *callback_data)
{/* FLICKER_CHANGE */

  flickerIt = *tag;

}/* FLICKER_CHANGE */




/* *******************************************************
 * radTog_change() - Toggle what values get displayed.
 */
void radTog_change(Widget widget, int *tag, 
		   XmToggleButtonCallbackStruct *callback_data)
{/* RADTOG_CHANGE */

  if(*tag==RDENS)
    rdensit = !rdensit;
  else if(*tag==WRDENS)
    wrdensit = !wrdensit;
  else if(*tag==CCOEF)
    ccoefit = !ccoefit;
  else if(*tag==WCCOEF)
    wccoefit = !wccoefit;
  else if(*tag==S_WCCOEF)
    s_wccoefit = !s_wccoefit;

  DrawWindow();

}/* RADTOG_CHANGE */




/*****************************************************************
 * smMinMax_reset_activate() - Reset to last gold std values.
*/
void smMinMax_reset_activate(Widget widget, int *tag, 
			     XmAnyCallbackStruct *callback_data)
{/* SMMINMAX_RESET_ACTIVATE */

  minMapPixel = saveMinMapPixel;
  maxMapPixel = saveMaxMapPixel;
  sprintf(outStr,"Min: %.2f",minMapPixel);
  doLabel(outStr,smMinLabelW);
  sprintf(outStr,"Max: %.2f",maxMapPixel);
  doLabel(outStr,smMaxLabelW);

}/* SMMINMAX_RESET_ACTIVATE */



/*************************************************************
 * subtractMap_read() -  Read the second map to be subtracted
 * from the first.
 */
void subtractMap_read()
{/* SUBTRACTMAP_READ */
  int i,j,k;
  fString_t rWrite;
  int htype,imgOffset,convTsb;
  char versionStr[256], descriptionStr[100];
  XmString simpleString;
  fString_t descStr;
  float alp, bet, gam, minMapP, maxMapP;
  int strtCol,strtRow,strtSec;
  int thisImg,pifMode;
  forPifHead_t thisHeader;
  float recipFSF;
  fString_t inFileName;
  char *s1;

  sprintf(outStr,"Reading file ... ");
  checkStatus(outStr);
  raiseIt(statusScreenWidget);
  XFlush(dsply);

  subtractingMap = FALSE;
  i= 4;
  thisHeader.luNum = i;
  rWrite.addr = malloc(sizeof(char)*4);
  rWrite.length = 1;
  descStr.addr = malloc(sizeof(char)*64);
  descStr.length = 32;
  inFileName.addr = malloc(sizeof(char)*64);
  inFileName.length = 64;
  sprintf(rWrite.addr,"r");

  s1 = XmTextGetString(subtractFile2TextW); 
  sprintf(subtractFilename,"%s",s1);
  XtFree(s1);

  sprintf(inFileName.addr,subtractFilename);
  pif_open_(inFileName.addr, &i, rWrite.addr);

  pif_read_gh_(&thisHeader);
  recipFSF = thisHeader.recipFloatScaleFactor;
  if((numCols < thisHeader.numCols) ||
     (numRows < thisHeader.numRows) ||
     (numSecs < thisHeader.numSecs))
    {/* Can't do it */
      askThem("Map dimensions are greater than first MAP.",
	      FALSE,XmDIALOG_INFORMATION);
      goto noRead;
    }/* Can't do it */
  pifMode               = thisHeader.mode;

  if(pifMode!=20)
    {/* NOt a map!! */
      askThem("Not a MAP file.",
	      FALSE,XmDIALOG_INFORMATION);
      goto noRead;
    }/* NOt a map!! */

  thisImg = 1;  /* For fortran */
  thisHeader.imgNum = thisImg;
  pif_read_dh_(&thisHeader);
  
  recipFSF = thisHeader.recipFloatScaleFactor;
  imgOffset  = thisHeader.imgOffset; 
  pifMode    = thisHeader.mode; 

  sub2binSec  = thisHeader.numSecs;
  sub2binRow  = thisHeader.numRows;
  sub2binCol  = thisHeader.numCols;
  binFactor = thisHeader.binFactor;
  
  if(binFactor==2)
    {/* 1/2 */
      if(pifMode==20)
	sub2binSec  = thisHeader.numSecs/2+1;  
      sub2binRow  = thisHeader.numRows/2+1;
      sub2binCol  = thisHeader.numCols/2+1;
      if((sub2binSec%2)==0)
	sub2binSec--;
      if((sub2binRow%2)==0)
	sub2binRow--;
      if((sub2binCol%2)==0)
	sub2binCol--;
    }/* 1/2 */

  if(subtract_image_buffer1)
    free(subtract_image_buffer1);
  CK(subtract_image_buffer1 = (float *)malloc(sizeof(float)*
					   sub2binCol*sub2binRow*sub2binSec));
  if(subtract_image_buffer2)
    free(subtract_image_buffer2);
  CK(subtract_image_buffer2 = (float *)malloc(sizeof(float)*
					   sub2binCol*sub2binRow*sub2binSec));
  
  /* Don't reset concept of minMapPixel and maxMapPixel here!!! */
  maxMapP= -999999999.9;
  minMapP= 999999999.9;
  maxMapP = -HUGE; /* dgm -HUGE instead of -99999.9 */
  minMapP = HUGE;  /* dgm HUGE instead of 9999999.9 */
  convTsb = TRUE;
  pif_tsbworld(convTsb);
  binFactor = 1;
  strtCol = strtRow = strtSec = 1;
  pif_read_map_short_image_(&i, subtract_image_buffer1,
			    &imgOffset,
			    &sub2binCol, &sub2binRow, &sub2binSec,
			    &recipFSF,
			    &minMapP,&maxMapP, &binFactor,
			    &strtCol, &strtRow, &strtSec,
			    &sub2binCol, &sub2binRow, &sub2binSec);
  pif_close_(&i);
  
  if((sub2binCol < numCols ) ||
     (sub2binRow < numRows ) ||
     (sub2binSec < numSecs ) )
    {/* Let's just place it in a bigger buffer */
      float *tmpBuf;
      int i,j,k,rowOut,secOut,colOff,rowOff,secOff;
      int rowOut2,secOut2;
      CK(tmpBuf = (float *)malloc(sizeof(float)*
				  numCols*numRows*numSecs));
      add_map_padded(subtract_image_buffer1,
		     &sub2binCol,&sub2binRow,&sub2binSec,
		     tmpBuf,
		     &numCols,&numRows,&numSecs);
      sub2binCol = numCols;
      sub2binRow = numRows;
      sub2binSec = numSecs;
      free(subtract_image_buffer1);
      subtract_image_buffer1= tmpBuf;
      /* Let's get this one the right size also... */
      if(subtract_image_buffer2)
	free(subtract_image_buffer2);
      CK(subtract_image_buffer2 = (float *)malloc(sizeof(float)*
					   sub2binCol*sub2binRow*sub2binSec));
    }/* Let's just place it in a bigger buffer */

  /* Let's get some global ideas about min/max map pixel here */
  saveMinMapPixel = minMapPixel;
  saveMaxMapPixel = maxMapPixel;
  if(maxMapP > maxMapPixel)
    maxMapPixel = maxMapP;
  if(minMapP < minMapPixel)
    minMapPixel = minMapP;
  sprintf(outStr,"Min: %.2f",minMapPixel);
  doLabel(outStr,smMinLabelW);
  sprintf(outStr,"Max: %.2f",maxMapPixel);
  doLabel(outStr,smMaxLabelW);

  threshScrollValue  = minMapPixel;
  sprintf(outStr,"%f",threshScrollValue); 
  XmTextSetString(threshTextWidget, outStr); 
  XtVaSetValues(threshScrollWidget,XmNminimum, (int)(minMapPixel - 5.0),NULL);
  XtVaSetValues(threshScrollWidget,XmNmaximum, (int)(maxMapPixel+20),NULL);
  XtVaSetValues(threshScrollWidget,XmNsliderSize, (int)(10),NULL);
  XtVaSetValues(threshScrollWidget,XmNvalue, (int)(threshScrollValue),NULL);

  if(diffSection)
    diffSec_really_doit();

 noRead:;
  XtUnmanageChild(statusScreenWidget);

}/* SUBTRACTMAP_READ */






/*************************************************************
 * diffSecSave_activate() - Save the subtracted buffer to disk.
 */
void diffSecSave_activate(Widget widget, int *tag, 
			  XmAnyCallbackStruct *callback_data)
{/* DIFFSECSAVE_ACTIVATE */
  XmString simpleString;

  if(*tag==1)
    {/* Save it */
      radScaleM = diffSecScrollValue;
      radScaleB = diffSec2ScrollValue;
      
      sprintf(outStr,"MAP1 = (MAP2 * %.3f) + %.3f",
	      radScaleM,
	      radScaleB
	      );
      createSimpleString(&simpleString, outStr);
      XtSetArg(arglst[0], XmNlabelString, simpleString);
      XtSetArg(arglst[1], XmNrecomputeSize, FALSE );
      XtSetValues(subEquationW, arglst, 2); 
      XFlush(dsply);
      XmStringFree(simpleString); 
    }/* Save it */
  else  if(*tag==0)
    {/* Reset it */
      diffSecScrollValue  = radScaleM;
      diffSec2ScrollValue = radScaleB;
      XtSetArg(arglst[0], XmNvalue, (int)(diffSecScrollValue*100.0));
      XtSetValues(diffSecScrollWidget, arglst, 1);
      sprintf(outStr,"%.2f",diffSecScrollValue);
      XmTextSetString(diffSecTextWidget, outStr); 
      XtSetArg(arglst[0], XmNvalue, (int)(diffSec2ScrollValue*100.0));
      XtSetValues(diffSec2ScrollWidget, arglst, 1);
      sprintf(outStr,"%.2f",diffSec2ScrollValue);
      XmTextSetString(diffSec2TextWidget, outStr); 
      
      sprintf(outStr,"MAP1 = (MAP2 * %.3f) + %.3f",
	      radScaleM,
	      radScaleB
	      );
      createSimpleString(&simpleString, outStr);
      XtSetArg(arglst[0], XmNlabelString, simpleString);
      XtSetArg(arglst[1], XmNrecomputeSize, FALSE );
      XtSetValues(subEquationW, arglst, 2); 
      XFlush(dsply);
      XmStringFree(simpleString); 

      if(diffSection)
	diffSec_really_doit();

    }/* Reset it */
  else  if(*tag==2)
    {/* MAG save */
      saveBmag = magSecScrollValue;
      sprintf(outStr,"%.3f",magSecScrollValue);
      XmTextSetString(subMagStartW,outStr); 
    }/* MAG save */
  else  if(*tag==3)
    {/* MAG reset */
      magSecScrollValue  = saveBmag;
      XtSetArg(arglst[0], XmNvalue, (int)(magSecScrollValue*1000.0));
      XtSetValues(magSecScrollWidget, arglst, 1);
      sprintf(outStr,"%.3f",magSecScrollValue);
      XmTextSetString(magSecTextWidget, outStr); 

      if(diffSection)
	diffSec_really_doit();
    }/* MAG reset */
  else  if(*tag==20)
    {/* starting MAG reset */
      saveBmag = magSecScrollValue= 1.0;
      XtSetArg(arglst[0], XmNvalue, (int)(magSecScrollValue*1000.0));
      XtSetValues(magSecScrollWidget, arglst, 1);
      sprintf(outStr,"%.3f",magSecScrollValue);
      XmTextSetString(magSecTextWidget, outStr); 

      sprintf(outStr,"1.00");
      XmTextSetString(subMagStartW,outStr); 

      if(diffSection)
	diffSec_really_doit();

    }/* starting MAG reset */
  else  if(*tag==21)
    {/* step MAG reset */

      if(subMap2d3d==2)
	sprintf(outStr,"0.01");
      else sprintf(outStr,"0.005");
      XmTextSetString(subMagStepW,outStr); 

      if(diffSection)
	diffSec_really_doit();

    }/* step MAG reset */
  else  if(*tag==22)
    {/* numStep MAG reset */

      if(subMap2d3d==2)
	sprintf(outStr,"11");
      else sprintf(outStr,"5");
      XmTextSetString(subMagNumStepW,outStr); 

      if(diffSection)
	diffSec_really_doit();

    }/* numStep MAG reset */
  else  if(*tag==11)
    {/* starting RAD reset */

      sprintf(outStr,"1");
      XmTextSetString(subMagMinRadW,outStr); 
      sprintf(outStr,"%d",numSecs/2);
      XmTextSetString(subMagMaxRadW,outStr); 

      if(diffSection)
	diffSec_really_doit();

    }/* starting RAD reset */

}/* DIFFSECSAVE_ACTIVATE */




/*************************************************************
 * phaseArrowActivate() - Phase arrow activated.
*/
void phaseArrowActivate(Widget widget, int *tag, 
			XmAnyCallbackStruct *callback_data)
{/* PHASEARROWACTIVATE */
  float speedFactor;
  char *inStr;
  int binVal;
  float stepVal;

  if(*tag==PHASXUP || *tag==PHASXDOWN)
    inStr = XmTextGetString(moveOrigXW);
  else if(*tag==PHASYDOWN ||*tag==PHASYUP)
    inStr = XmTextGetString(moveOrigYW);
  
  stepVal = (float)strtod(inStr,0);
  XtFree(inStr);

  if(*tag==PHASXUP || *tag==PHASYUP)
    stepVal+= 0.10;
  else 
    stepVal-= 0.10;

  stepVal= MAX(stepVal,0.0000000);
  stepVal= MIN(stepVal,1.0000000);

  sprintf(outStr,"%.3f",stepVal);
  if(*tag==PHASXUP || *tag==PHASXDOWN)
    XmTextSetString(moveOrigXW,outStr);
  else if(*tag==PHASYDOWN ||*tag==PHASYUP)
    XmTextSetString(moveOrigYW,outStr);

}/* PHASEARROWACTIVATE */





/*************************************************************
 * phaseArrowButton() - They want to change the current
 * phase by using the arrows.
*/
void phaseArrowButton(Widget widget, XEvent *event)
{/* PHASEARROWBUTTON */
  XEvent thisEvent;
  float speedFactor;
  char *inStr;
  float stepVal;
  int binVal;
  int     tag;
  XmAnyCallbackStruct callback_data;

  if(widget==phasXupW || widget==phasXdownW)
    inStr = XmTextGetString(moveOrigXW);
  else
    inStr = XmTextGetString(moveOrigYW);

  stepVal = (float)strtod(inStr,0);
  XtFree(inStr);

  thisEvent= *(event);
  
  /* Get which button */
  if((thisEvent.xbutton.button == Button1))
    speedFactor = 0.10;
  if((thisEvent.xbutton.button == Button2))
    speedFactor = 0.05;
  if((thisEvent.xbutton.button == Button3))
    speedFactor = 0.01;
  
  if(widget==phasXupW || widget==phasYupW)
    stepVal += speedFactor;
  else 
    stepVal -= speedFactor;

  stepVal= MAX(stepVal,0.000000);
  stepVal= MIN(stepVal,1.000000);

  sprintf(outStr,"%.3f",stepVal);

  if(widget==phasXupW || widget==phasXdownW)
    XmTextSetString(moveOrigXW,outStr);
  else
    XmTextSetString(moveOrigYW,outStr);

}/* PHASEARROWBUTTON */







/*************************************************************
 * smMMaxArrowActivate() - min/max arrows
*/
void smMMaxArrowActivate(Widget widget, int *tag, 
			 XmAnyCallbackStruct *callback_data)
{/* SMMMAXARROWACTIVATE */
  float speedFactor;
  char *inStr;
  int binVal;
  float stepVal;

  if(*tag==MINRADUP || *tag==MINRADDOWN)
    inStr = XmTextGetString(subMagMinRadW);
  else if(*tag==MAXRADDOWN ||*tag==MAXRADUP)
    inStr = XmTextGetString(subMagMaxRadW);
  
  stepVal = (float)strtod(inStr,0);
  XtFree(inStr);

  if(*tag==MINRADUP || *tag==MAXRADUP)
    stepVal+= 1.0;
  else 
    stepVal-= 1.0;

  stepVal= MAX(stepVal,0.0000000);
  stepVal= MIN(stepVal,(sub2binCol/2));

  sprintf(outStr,"%.3f",stepVal);
  if(*tag==MINRADUP || *tag==MINRADDOWN)
    XmTextSetString(subMagMinRadW,outStr);
  else if(*tag==MAXRADDOWN ||*tag==MAXRADUP)
    XmTextSetString(subMagMaxRadW,outStr);

  if(diffSection)
    diffSec_really_doit();

}/* SMMMAXARROWACTIVATE */





/*************************************************************
 * smMMaxArrowButton() - They want to change the current
 * min/max by using the arrows.
*/
void smMMaxArrowButton(Widget widget, XEvent *event)
{/* SMMMAXARROWBUTTON */
  XEvent thisEvent;
  float speedFactor;
  char *inStr;
  float stepVal;
  int binVal;
  int     tag;
  XmAnyCallbackStruct callback_data;

  if(widget==smMaxRadUpW || widget==smMaxRadDownW)
    inStr = XmTextGetString(subMagMaxRadW);
  else
    inStr = XmTextGetString(subMagMinRadW);

  stepVal = (float)strtod(inStr,0);
  XtFree(inStr);

  thisEvent= *(event);
  
  /* Get which button */
  if((thisEvent.xbutton.button == Button1))
    speedFactor = 1.0;
  if((thisEvent.xbutton.button == Button2))
    speedFactor = 2.0;
  if((thisEvent.xbutton.button == Button3))
    speedFactor = 5.0;
  
  if(widget==smMaxRadUpW || widget==smMinRadUpW)
    stepVal += speedFactor;
  else 
    stepVal -= speedFactor;

  stepVal= MAX(stepVal,0.000000);
  stepVal= MIN(stepVal,(sub2binCol/2));

  sprintf(outStr,"%.3f",stepVal);

  if(widget==smMaxRadUpW || widget==smMaxRadDownW)
    XmTextSetString(subMagMaxRadW,outStr);
  else
    XmTextSetString(subMagMinRadW,outStr);

  if(diffSection)
    diffSec_really_doit();

}/* SMMMAXARROWBUTTON */







/*************************************************************
 * subtractMap_2d3d_activate() - Toggle whther they want to
 * do 2D subtraction or 3D.
 */
void subtractMap_2d3d_activate(Widget widget, int *tag, 
			       XmAnyCallbackStruct *callback_data)
{/* SUBTRACTMAP_2D3D_ACTIVATE */

  if(*tag==SUBTRACTMAP2D)
    {/* They want 2d mode */
      subMap2d3d= 2;
      XtSetArg(arglst[0], XmNforeground, pixels_return[RED]);
      XtSetValues(subMap2dW, arglst, 1); 
      XtSetArg(arglst[0], XmNforeground, pixels_return[WHITE]);
      XtSetValues(subMap3dW, arglst, 1); 
      XFlush(dsply);
    }/* They want 2d mode */
  else if(*tag==SUBTRACTMAP3D)
    {/* They want 3d mode */
      subMap2d3d= 3;
      XtSetArg(arglst[0], XmNforeground, pixels_return[WHITE]);
      XtSetValues(subMap2dW, arglst, 1); 
      XtSetArg(arglst[0], XmNforeground, pixels_return[RED]);
      XtSetValues(subMap3dW, arglst, 1); 

      /* Reset for 3D trial */
      sprintf(outStr,"0.005");
      XmTextSetString(subMagStepW,outStr); 
      sprintf(outStr,"5");
      XmTextSetString(subMagNumStepW,outStr); 

      XFlush(dsply);
    }/* They want 3d mode */
  else
    {/* error */
      printf("[2d3d_activate] Unknown mode [%d]\n",*tag);
    }/* error */

}/* SUBTRACTMAP_2D3D_ACTIVATE */




/*************************************************************
 * recalc_ymxb_thresh() - We've got to recalc y=mx+b every time
 * they change the threshold 
*/
void recalc_ymxb_thresh()
{/* RECALC_YMXB_THRESH */
  int tag;

  tag = 1;
/* Use other rtn   subtractMap_computeScaling(0,&tag,0); */
  subtractMap_findRad(0, &tag, 0);


}/* RECALC_YMXB_THRESH */




/*************************************************************
 * subtractMap_computeScaling() - Compute the scaling needed
 * when we eventully do the MAP subtraction. 
 */
void subtractMap_computeScaling(Widget widget, int *tag, 
				XmAnyCallbackStruct *callback_data)
{/* SUBTRACTMAP_computeScaling */
  int i,j,k,saveNbins;
  int yOff,zOff,xOff,nfit;
  XmString simpleString;
  fString_t descStr;
  int binFactor,ichc,sec1,sec2;
  float recipFSF;
  char *s1;
  float xc,yc,zc,radlo,radhi;
  float tbccoef,  map_std, bmag;
  int   lpol;
  float *thisBuf1,*thisBuf2;

  subtractingMap = FALSE;
/*  submapFindRad  = TRUE; */

  if(!subtract_image_buffer1)
    {/* Must read first */
      askThem("Must Read second map file first.",
	      FALSE,XmDIALOG_INFORMATION);
      goto noRead;
    }/* Must read first */

  if(tag==0)
    {/* Interactive */
      raiseIt(statusScreenWidget);
      XFlush(dsply);
    }/* Interactive */

  ichc = 2;
  xc = ((numCols+1)/2.0);
  yc = ((numRows+1)/2.0);
  zc = ((numSecs+1)/2.0);
  radlo = 0.0;
  radhi = numSecs;
  radstep = 1.0;
  s1 = XmTextGetString(subMagMinRadW); 
  radlo = (float)strtod(s1,0);
  XtFree(s1);
  /* Is this really necessary??? radlo -= 1.0;  /+ Zero one counting problem */
  if(radlo < 0.0)
    radlo = 0.0;
  s1 = XmTextGetString(subMagMaxRadW); 
  radhi = (float)strtod(s1,0);
  XtFree(s1);
  slidingWindow = 8;

  qtr_switch = 0;
  qtr_switch = 1;  /* 1 mean try doing the calc with only 1/4 of the MAP */
  if(slidingWindow < 2 || slidingWindow > 10)
    {/* unwise */
      askThem("It is considered unwise to have a Sliding Window <2 or > 10",
	      FALSE,XmDIALOG_INFORMATION);
    }/* unwise */
  /* We used to get this from the interface but now it's
   * been replaced with the sliding window param */
  radstep = 1.0; 

  s1 = XmTextGetString(subMagStartW); 
  bmag = (float)strtod(s1,0);
  XtFree(s1);

  nfit = numCols*numRows*numSecs;
  numRad = numCols / 2;
  nbins  = numRad  / radstep;
  nbins++;  /* Zero one counting problem */

  if(cntr1)
    free(cntr1);
  if(cntr2)
    free(cntr2);
  cntr1 = (float *)malloc(sizeof(float)*nfit);
  cntr2 = (float *)malloc(sizeof(float)*nfit);

  if(ccoef)
    free(ccoef);
  ccoef   = (float *)malloc(sizeof(float)*nbins);


  tbccoef = 0.0;
  for(i=0;i<magchk;i++)
    if(coefs[i]==bmag)
      tbccoef = coefs[i];
  map_std = 1.2;
  map_std = purdueHeader.img_std; /* Get float_image_buffer1 STDev */
  lpol = 6;

  if(subMap2d3d==2)
    {/* DO 2d */
      sec1= 1;
      sec2= 1;
      thisBuf1 = (float *)(&(float_image_buffer1[(int)(
					     (zc-1)*numCols*numRows)]));
      thisBuf2 = (float *)(&(subtract_image_buffer1[(int)(
					     (zc-1)*numCols*numRows)]));
      zc = 1;
    }/* DO 2d */
  else
    {/* DO 3d */
      sec1= numSecs;
      sec2= sub2binSec;
      thisBuf1 = (float *)(float_image_buffer1);
      thisBuf2 = (float *)(subtract_image_buffer1);
    }/* DO 3d */


  saveNbins = nbins;
  compute_scaling_(thisBuf1,
		   thisBuf2,
		   &numCols,&numRows,&sec1,
		   &tbccoef,
		   &sub2binCol,&sub2binRow,&sec2,
		   &nfit,
		   cntr1,cntr2,&lpol,&map_std,
		   &xc,&yc,&zc,
		   &radlo,&radhi,&radstep,
		   &bmag,
		   ccoef,&nbins,
		   &diffSecScrollValue,
		   &diffSec2ScrollValue,
		   &slidingWindow,
		   &threshScrollValue,
		   &threshAbs
		   );

  radScaleM = diffSecScrollValue;
  radScaleB = diffSec2ScrollValue;
  XtSetArg(arglst[0], XmNvalue, (int)(diffSecScrollValue*100.0));
  XtSetValues(diffSecScrollWidget, arglst, 1);
  sprintf(outStr,"%.2f",diffSecScrollValue);
  XmTextSetString(diffSecTextWidget, outStr); 
  XtSetArg(arglst[0], XmNvalue, (int)(diffSec2ScrollValue*100.0));
  XtSetValues(diffSec2ScrollWidget, arglst, 1);
  sprintf(outStr,"%.2f",diffSec2ScrollValue);
  XmTextSetString(diffSec2TextWidget, outStr); 

  sprintf(outStr,"MAP1 = (MAP2 * %.3f) + %.3f",
	  radScaleM,
	  radScaleB
	  );
  createSimpleString(&simpleString, outStr);
  XtSetArg(arglst[0], XmNlabelString, simpleString);
  XtSetArg(arglst[1], XmNrecomputeSize, FALSE );
  XtSetValues(subEquationW, arglst, 2); 
  XFlush(dsply);
  XmStringFree(simpleString); 

  nbins = saveNbins;

  sprintf(outStr,"%f",radlo);
  XmTextSetString(subMagMinRadW,outStr); 
  sprintf(outStr,"%f",radhi);
  XmTextSetString(subMagMaxRadW,outStr); 

  /* Let's let the findMagHist toggle decide this subCanDrawWindow = FALSE;  */
  if(tag==0)
    {DrawWindow();
     if(subCanDrawWindow)
       drawMinMaxRadius();
   }

 noRead:;

  if(tag==0)
    XtUnmanageChild(statusScreenWidget);

}/* SUBTRACTMAP_COMPUTESCALING */




/*************************************************************
 * subtractMap_findRad() -  Use findRad rtns to find relative mag for
 * the maps.
 */
void subtractMap_findRad(Widget widget, int *tag, 
			 XmAnyCallbackStruct *callback_data)
{/* SUBTRACTMAP_FINDRAD */
  int i,j,k,saveNbins;
  int yOff,zOff,xOff,nfit;
  int sec1,sec2;
  float *thisBuf1,*thisBuf2;
  XmString simpleString;
  fString_t descStr;
  int binFactor,ichc;
  float recipFSF;
  char *s1;
  float xc,yc,zc,radlo,radhi;
  float tbccoef,  map_std, bmag;
  int   lpol;

  /* Get rid of 3D rendering if it's there */
  if(XtIsManaged(mapShadeScreenWidget))
    mapShadeScreen_deactivate(0,0,0);

  subtractingMap = FALSE;
/*  submapFindRad  = TRUE; */
/*
  XtSetArg(arglst[0], XmNsensitive, TRUE );
  XtSetValues(computeScalingW, arglst, 1); 
*/

  if(!subtract_image_buffer1)
    {/* Must read first */
      askThem("Must Read second map file first.",
	      FALSE,XmDIALOG_INFORMATION);
      goto noRead;
    }/* Must read first */

  if(*tag!=1)
    raiseIt(statusScreenWidget);
  XFlush(dsply);

  ichc = 2;
  xc = ((numCols+1)/2.0);
  yc = ((numRows+1)/2.0);
  zc = ((numSecs+1)/2.0);
  radlo = 0.0;
  radhi = numSecs;
  radstep = 1.0;
  s1 = XmTextGetString(subMagMinRadW); 
  radlo = (float)strtod(s1,0);
  XtFree(s1);
  radlo -= 1.0;  /* Zero one counting problem */
  if(radlo < 0.0)
    radlo = 0.0;
  s1 = XmTextGetString(subMagMaxRadW); 
  radhi = (float)strtod(s1,0);
  XtFree(s1);
  slidingWindow = 8;

  /* TRUE means try doing the calc with only 1/4 of the MAP */
  qtr_switch = icosaQuarter;
  if(slidingWindow < 2 || slidingWindow > 10)
    {/* unwise */
      askThem("It is considered unwise to have a Sliding Window <2 or > 10",
	      FALSE,XmDIALOG_INFORMATION);
    }/* unwise */
  /* We used to get this from the interface but now it's
   * been replaced with the sliding window param */
  radstep = 1.0; 

  s1 = XmTextGetString(subMagStartW); 
  bmag = (float)strtod(s1,0);
  XtFree(s1);

  nfit = numCols*numRows*numSecs;
  numRad = numCols / 2;
  nbins  = numRad  / radstep;
  nbins++;  /* Zero one counting problem */

  if(cntr1)
    free(cntr1);
  if(cntr2)
    free(cntr2);
  cntr1 = (float *)malloc(sizeof(float)*nfit);
  cntr2 = (float *)malloc(sizeof(float)*nfit);

  if(rdens)
    free(rdens);
  if(wrdens)
    free(wrdens);
  if(ccoef)
    free(ccoef);
  if(wccoef)
    free(wccoef);
  if(s_wccoef)
    free(s_wccoef);
  rdens   = (float *)malloc(sizeof(float)*nbins);
  wrdens  = (float *)malloc(sizeof(float)*nbins);
  ccoef   = (float *)malloc(sizeof(float)*nbins);
  wccoef  = (float *)malloc(sizeof(float)*nbins);
  s_wccoef  = (float *)malloc(sizeof(float)*nbins);


  tbccoef = 0.0;
  for(i=0;i<magchk;i++)
    if(coefs[i]==bmag)
      tbccoef = coefs[i];
  map_std = 1.2;
  map_std = purdueHeader.img_std; /* Get float_image_buffer1 STDev */
  lpol = 6;

  saveNbins = nbins;

  if(subMap2d3d==2)
    {/* DO 2d */
      sec1= 1;
      sec2= 1;
      thisBuf1 = (float *)(&(float_image_buffer1[(int)(
					     (zc-1)*numCols*numRows)]));
      thisBuf2 = (float *)(&(subtract_image_buffer1[(int)(
					     (zc-1)*numCols*numRows)]));
      zc = 1;
    }/* DO 2d */
  else
    {/* DO 3d */
      sec1= numSecs;
      sec2= sub2binSec;
      thisBuf1 = (float *)(float_image_buffer1);
      thisBuf2 = (float *)(subtract_image_buffer1);
    }/* DO 3d */


  ccomap3cs_(thisBuf1,
	     thisBuf2,
	     &numCols,&numRows,&sec1,
	     &tbccoef,
	     &sub2binCol,&sub2binRow,&sec2,
	     &nfit,
	     cntr1,cntr2,&lpol,&map_std,
	     &xc,&yc,&zc,
	     &radlo,&radhi,&radstep,
	     &bmag,
	     rdens,wrdens,ccoef,wccoef,s_wccoef,&nbins,
	     &diffSecScrollValue,
	     &diffSec2ScrollValue,
	     &slidingWindow,
	     &qtr_switch, &subMap2d3d,
	     &threshScrollValue,
	     &threshAbs
	     );

  radScaleM = diffSecScrollValue;
  radScaleB = diffSec2ScrollValue;
  XtSetArg(arglst[0], XmNvalue, (int)(diffSecScrollValue*100.0));
  XtSetValues(diffSecScrollWidget, arglst, 1);
  sprintf(outStr,"%.3f",diffSecScrollValue);
  XmTextSetString(diffSecTextWidget, outStr); 
  XtSetArg(arglst[0], XmNvalue, (int)(diffSec2ScrollValue*100.0));
  XtSetValues(diffSec2ScrollWidget, arglst, 1);
  sprintf(outStr,"%.3f",diffSec2ScrollValue);
  XmTextSetString(diffSec2TextWidget, outStr); 

  sprintf(outStr,"MAP1 = (MAP2 * %.3f) + %.3f",
	  radScaleM,
	  radScaleB
	  );
  createSimpleString(&simpleString, outStr);
  XtSetArg(arglst[0], XmNlabelString, simpleString);
  XtSetArg(arglst[1], XmNrecomputeSize, FALSE );
  XtSetValues(subEquationW, arglst, 2); 
  XFlush(dsply);
  XmStringFree(simpleString); 

  nbins = saveNbins;

  /* Let's let the findmagHist toggle decide this subCanDrawWindow = FALSE;*/
  diffSec_really_doit();
  DrawWindow();

 noRead:;

  XtUnmanageChild(statusScreenWidget);

}/* SUBTRACTMAP_FINDRAD */




/**********************************************************
 * checkPeakFound() - Find the peak.
*/
int checkPeakFound()
{/* CHECKPEAKFOUND */
  int i,foundIt;
  float maxFound;

  foundIt = -99;
  maxFound = -99999.9;
  maxFound = -HUGE;  /* dgm -HUGE instead of -99999.9 */
  for(i=0;i<magchk;i++)
    {/* Do all */
      if(coefs[i] > maxFound)
	{
	  foundIt = i;
	  maxFound = coefs[i];
	}
    }/* Do all */
  if((foundIt==0) ||
     (foundIt==(magchk-1)))
    return(FALSE);
  else return(TRUE);
      
}/* CHECKPEAKFOUND */




/****************************************************************
 * drawMinMaxRadius() - Actually draw the min/max radii
 * on the display.
*/
void drawMinMaxRadius()
{/* DRAWMINMAXRADIUS */
  char *s1;
  float radlo,radhi;
  int xBegin,yBegin,xEnd,yEnd,arcRadius,i,j;

  if(showRadii )
    {/* Show radii */
      s1 = XmTextGetString(subMagMinRadW); 
      radlo = (float)strtod(s1,0);
      XtFree(s1);
      s1 = XmTextGetString(subMagMaxRadW); 
      radhi = (float)strtod(s1,0);
      XtFree(s1);

      for(i=0; i <= (2*(numCols*sizeFactor)) ; i += (numCols*sizeFactor))
	{/* Do all three */

	  xBegin = i + ((numCols * sizeFactor)/2);
	  yBegin = ((numRows * sizeFactor)/2);
	  arcRadius = (int)(radlo * sizeFactor);
	  XDrawArc(dsply, XtWindow(drawingWidget), image_gcRED,
		   (xBegin - arcRadius), 
		   (yBegin - arcRadius),
		   (arcRadius*2),(arcRadius*2),
		   0, (360*64));
	  arcRadius = (int)(radhi * sizeFactor);
	  XDrawArc(dsply, XtWindow(drawingWidget), image_gcBLUE,
		   (xBegin - arcRadius), 
		   (yBegin - arcRadius),
		   (arcRadius*2),(arcRadius*2),
		   0, (360*64));


	}/* Do all three */


    }/* Show radii */


}/* DRAWMINMAXRADIUS */






/*************************************************************
 * subtractMap_cc() -  Use cc rtns to find relative mag for
 * the maps.
 */
void subtractMap_cc(Widget widget, int *tag, 
		    XmAnyCallbackStruct *callback_data)
{/* SUBTRACTMAP_CC */
  int i,j,k,iix,iiy,iiz;
  int yOff,zOff,xOff,qtr_switch;
  XmString simpleString;
  fString_t descStr;
  int binFactor,ichc;
  float recipFSF,std;
  char *s1;
  float xc,yc,zc,radlo,radhi,bmag,magini,magstep;

  subtractingMap = FALSE;
/*  submapFindRad  = FALSE; */
/*
  XtSetArg(arglst[0], XmNsensitive, FALSE );
  XtSetValues(computeScalingW, arglst, 1); 
*/

  if(!subtract_image_buffer1)
    {/* Must read first */
      askThem("Must Read second map file first.",
	      FALSE,XmDIALOG_INFORMATION);
      goto noRead;
    }/* Must read first */

  raiseIt(statusScreenWidget);
  XFlush(dsply);

  ichc = 2;
  xc = (numCols/2.0);
  yc = (numRows/2.0);
  zc = (numSecs/2.0);
  radlo = 0.0;
  radhi = numSecs;
  radstep = 1.0;
  s1 = XmTextGetString(subMagMinRadW); 
  radlo = (float)strtod(s1,0);
  XtFree(s1);
  radlo -= 1.0;   /* To make it consistent with cc_3d() */
  s1 = XmTextGetString(subMagMaxRadW); 
  radhi = (float)strtod(s1,0);
  XtFree(s1);
  slidingWindow = 8;

  qtr_switch = 0;
  qtr_switch = 1;  /* 1 mean try doing the calc with only 1/4 of the MAP */
  /* We used to get this from the interface but now it's
   * been replaced with the sliding window param */
  radstep = 1.0; 

  bmag = 1.0;

  magini = 1.00;
  magstep = 0.1;
  magchk = 10;

  s1 = XmTextGetString(subMagStartW); 
  magini = (float)strtod(s1,0);
  XtFree(s1);
  s1 = XmTextGetString(subMagStepW); 
  magstep = (float)strtod(s1,0);
  XtFree(s1);
  s1 = XmTextGetString(subMagNumStepW); 
  saveMagchk= magchk = (int)strtod(s1,0);
  XtFree(s1);

  sprintf(outStr,"Looking around mag==[%f] ... Please wait ... ",magini);
  checkStatus(outStr);

  if(mags)
    free(mags);
  if(coefs)
    free(coefs);
  mags  = (float *)malloc(sizeof(float)*magchk);
  coefs = (float *)malloc(sizeof(float)*magchk);

  qtr_switch = icosaQuarter;

  iix= iiy= iiz = 1;

  cc_3d_(float_image_buffer1,
	 subtract_image_buffer1,
	 &numCols,&numRows,&numSecs,
	 &sub2binCol,&sub2binRow,&sub2binSec,
	 &ichc,&std,
	 &xc,&yc,&zc,
	 &magini,&magstep,&magchk,
	 &radlo,&radhi,&radstep,&bmag,
	 &iix,&iiy,&iiz,
	 mags,coefs,
	 &subMap2d3d, &qtr_switch
	 );

  magchk = saveMagchk; /* It's corrupted above */
  sprintf(outStr,"%f",bmag);
  magSecScrollValue  = saveBmag = bmag;
  XmTextSetString(subMagStartW,outStr); 
  XmTextSetString(magSecTextWidget, outStr); 
  XtSetArg(arglst[0], XmNvalue, (int)(magSecScrollValue*1000.0));
  XtSetValues(magSecScrollWidget, arglst, 1);
  if(checkPeakFound())
    {/* Step 2 times finer */
      sprintf(outStr,"%f",magstep*0.5);
      XmTextSetString(subMagStepW,outStr); 
    }/* Step 2 times finer */
  else
    {/* Stay here and step sideways */
      askThem("Peak not found.  Step Size not changed.",
	      FALSE,XmDIALOG_INFORMATION);
    }/* Stay here and step sideways */

  diffSec_really_doit();
  /* Let's let the findMagHist toggle decide this subCanDrawWindow = FALSE;*/
  DrawWindow();
  

 noRead:;

  XtUnmanageChild(statusScreenWidget);

}/* SUBTRACTMAP_CC */






/*************************************************************
 * subtractMap_doit() -  Actually do the subtracting of maps
 */
void subtractMap_doit(float *thisBuffer, float *tmpSubtract_image_buffer2)
{/* SUBTRACTMAP_DOIT */
  int i,j,k;
  int yOff,zOff,xOff;
  XmString simpleString;
  fString_t descStr;
  char *s1;
  float xc,yc,zc,mag;

  s1 = XmTextGetString(subMagStartW); 
  mag = (float)strtod(s1,0);
  XtFree(s1);
  sprintf(outStr,"Adjusting for mag==[%f] ... ",mag);
  checkStatus(outStr);

  xc = (sub2binCol/2.0);
  yc = (sub2binRow/2.0);
  zc = (sub2binSec/2.0);

  mag_write_(subtract_image_buffer1,
	     &sub2binCol,&sub2binRow,&sub2binSec,
	     &xc,&yc,&zc,&mag,
	     tmpSubtract_image_buffer2);

  sprintf(outStr,"Subtracting MAPs after adjusting density... ");
  checkStatus(outStr);

  for(i=0;i<sub2binSec;i++)
    {/* Sec */
      zOff = (i * sub2binRow * sub2binCol);
      for(j=0;j<sub2binRow;j++)
	{/* Row */
	  yOff = zOff + (j * sub2binCol);
	  for(k=0;k<sub2binCol;k++)
	    {/* Col */
	      xOff = yOff + k;
	      /* Don't forget to apply y=mx+b to second map!!! */
	      tmpSubtract_image_buffer2[xOff] = 
		  (float)((float)(tmpSubtract_image_buffer2[xOff]*
				  diffSecScrollValue)+
			  (float)diffSec2ScrollValue);
	      thisBuffer[xOff] = 
		thisBuffer[xOff] - tmpSubtract_image_buffer2[xOff];
	    }/* Col */
	}/* Row */
    }/* Sec */


}/* SUBTRACTMAP_DOIT */





/*************************************************************
 * subtractMap_work() - The real work of MAP subtracting
 * happens here.
 */
void subtractMap_work(Widget widget, int *tag, 
		      XmAnyCallbackStruct *callback_data)
{/* SUBTRACTMAP_WORK */
  int i,j,k;
  int yOff,zOff,xOff;
  XmString simpleString;
  fString_t descStr;
  char *s1;
  float xc,yc,zc,mag;
  float *tmpSubtract_image_buffer2;


  if(diffSecScrollValue == 0.0 &&
     diffSec2ScrollValue== 0.0)
    {/* They haven't done step two */
      askThem("You must finish the 'Find Radius' step first.",
	      FALSE,XmDIALOG_INFORMATION);
      goto noSubtract;
    }/* They haven't done step two */

  subtractingMap = FALSE;

  raiseIt(statusScreenWidget);
  XFlush(dsply);

  /* Call mag_write here for appropriate buffer and then subtract them */
  CK(tmpSubtract_image_buffer2 = (float *)malloc(sizeof(float)*
				      sub2binCol*sub2binRow*sub2binSec));

  /* Really subtract the buffers */
  subtractMap_doit(float_image_buffer1, tmpSubtract_image_buffer2);

  if(saveSecondMap)
    {/* Write this one to disk */
      forPifHead_t thisHead;
      fString_t rWrite,forName;
      float *tmp_float_buffer;

      /* Now write PIF file of the above */
      pif_init_head(&thisHead);
      thisHead.luNum = 4;
      rWrite.addr = malloc(sizeof(char)*4);
      rWrite.length = 4;
      forName.addr = malloc(sizeof(char)*64);
      forName.length = 64;
      sprintf(rWrite.addr,"w");
      sprintf(forName.addr,"secondMap.pif");

      pif_open_(forName.addr, &(thisHead.luNum), rWrite.addr);
	  
      /* Pick appropriate statistics calculation based
       * on data type we have */
      thisHead.datai4_min = 0;
      thisHead.datai4_max = 255;
      thisHead.numCols = sub2binCol;
      thisHead.numRows = sub2binRow;
      thisHead.numSecs = sub2binSec;
      
      thisHead.mode = 20;
      
      calcMapStats(tmpSubtract_image_buffer2,
		   &thisHead.img_bkg, 
		   &thisHead.datar4_min,
		   &thisHead.datar4_max, 
		   &thisHead.datar4_avg, 
		   &thisHead.datar4_std,
		   &sub2binCol,
		   &sub2binRow,
		   &sub2binSec);
      thisHead.datai4_max= (int)thisHead.datar4_max;
      thisHead.datai4_min= (int)thisHead.datar4_min;

      calcScaleFactor(&(thisHead.floatScaleFactor),		     
		      &(thisHead.recipFloatScaleFactor), 
		      &(thisHead.int4scaleF), 
		      (thisHead.datar4_min-1.0),
		      (thisHead.datar4_max+1.0),
		      thisHead.datar4_avg,thisHead.datar4_std,
		      MAP);

      pif_write_gh(&thisHead);
      
      thisHead.imgNum = 1;
      
      pif_write_dh(&thisHead);
      thisHead.imgOffset += 512;  /* Just wrote the header */
      
      tsbworld_float(tmpSubtract_image_buffer2,
		     subtract_image_buffer1,
		     sub2binCol,
		     sub2binRow,
		     sub2binSec);
      pif_write_wmapi2(&thisHead,
		       subtract_image_buffer1,
		       &sub2binCol,
		       &sub2binRow,
		       &sub2binSec);
      free(forName.addr);
      free(rWrite.addr);
      pif_close(&(thisHead.luNum));

      sprintf(outStr,"Second Map written to file 'secondMap.pif'");
      askThem(outStr,FALSE,XmDIALOG_INFORMATION);

    }/* Write this one to disk */

  free(tmpSubtract_image_buffer2);

  /* Write stats */
  calcMapStats(float_image_buffer1,
	       &purdueHeader.img_bkg, 
	       &purdueHeader.map_min,
	       &purdueHeader.map_max, 
	       &purdueHeader.img_avg, 
	       &purdueHeader.img_std,
	       &numCols, &numRows, &origNumSecs);
  secMaxPixel = maxMapPixel = purdueHeader.map_max;
  secMinPixel = minMapPixel = purdueHeader.map_min;
  if(secMaxPixel != secMinPixel)
    scaleMapFactor= (((float)tryColors - 7.0) / (secMaxPixel-secMinPixel));
  else 
    {
      scaleMapFactor= 1.0;
      maxMapPixel = minMapPixel + 1.0;
      secMaxPixel = secMinPixel + 1.0;
    }
  
  if(dispSections)
    {/* Update the sections - if shaded views then they have to recalc */
      /* Now stuff the display buffer from our new FIXed map */
      stuffMapImage((float *)map_image_buffer, 
		    (float *)float_image_buffer1, 
		    file_width,file_height,
		    nsecs,imgNum, 0);
      DrawWindow();
    }/* Update the sections - if shaded views then they have to recalc */
  else
    {/* If we're already doing rendering then setup the scrolls
      * appropriately */
      mapShadeScreen_activate(0,0,0);
    }/* If we're already doing rendering then setup the scrolls
      * appropriately */
  
  subtractingMap = FALSE;
  subCanDrawWindow  = TRUE; 
  drawingFindRad = FALSE;
  XtUnmanageChild(subtractMapScreen);
  if(subtract_image_buffer2)
    free(subtract_image_buffer2);
  subtract_image_buffer2 = 0;
  if(subtract_image_buffer1)
    free(subtract_image_buffer1);
  subtract_image_buffer1 = 0;
  if(diffSection_buffer)
    free(diffSection_buffer);
  diffSection_buffer = 0;
  if(subMapYmxb_buffer)
    free(subMapYmxb_buffer);
  subMapYmxb_buffer = 0;
  if(mags)
    free(mags);
  if(coefs)
    free(coefs);
  if(cntr1)
    free(cntr1);
  if(cntr2)
    free(cntr2);
  if(rdens)
    free(rdens);
  if(wrdens)
    free(wrdens);
  if(ccoef)
    free(ccoef);
  if(wccoef)
    free(wccoef);
  if(s_wccoef)
    free(s_wccoef);

 noSubtract:;

  XtUnmanageChild(statusScreenWidget);

}/* SUBTRACTMAP_WORK */





/*************************************************************
 * subtractMap_diffSection() - Toggle whether too calc a 
 * difference section or not.
 */
void subtractMap_diffSection(Widget widget, int *tag, 
			     XmAnyCallbackStruct *callback_data)
{/* SUBTRACTMAP_DIFFSECTION */

  if(!subtract_image_buffer1)
    {/* Must read first */
      askThem("Must Read second map file first.",
	      FALSE,XmDIALOG_INFORMATION);
      XtSetArg(arglst[0], XmNset, FALSE );
      XtSetValues(widget, arglst, 1); 
    }/* Must read first */
  else
    {/* OK to go */
      diffSection = !diffSection;
      if(diffSection)
	XtSetArg(arglst[0], XmNsensitive, TRUE );
      else XtSetArg(arglst[0], XmNsensitive, FALSE );
      XtSetValues(flickerTypeWidget, arglst, 1); 
      XtSetValues(diffSecScrollWidget, arglst, 1); 
      XtSetValues(diffSecTextWidget, arglst, 1); 
      XtSetValues(diffSec2ScrollWidget, arglst, 1); 
      XtSetValues(diffSec2TextWidget, arglst, 1); 
      XtSetValues(magSecScrollWidget, arglst, 1); 
      XtSetValues(magSecTextWidget, arglst, 1); 
      XtSetValues(threshScrollWidget, arglst, 1); 
      XtSetValues(threshTextWidget, arglst, 1); 
      if(diffSection)
	diffSec_really_doit();
    }/* OK to go */

}/* SUBTRACTMAP_DIFFSECTION */




/*************************************************************
 * diffSec_really_doit() - Do what ti takes to get a difference
 * section.
 */
void diffSec_really_doit()
{/* DIFFSEC_REALLY_DOIT */
  int i,j,thisOne,secOut,rowOut,colOut;
  float *thisSection1, *thisSection2;
  float maxMapP,minMapP,scaleF,tmpValue,backValue;
  float *tmpSection;
  int   thisNumSecs;
  float xc,yc,zc,xcen,ycen,xSq,ySq,distnce;
  float radlo,radhi,radloSq,radhiSq;
  char *s1;

  /* Let's let them look at it on the screen */
  /* Let's let the findmaghist toggle decide this subCanDrawWindow = TRUE;  */

  if(!subMapYmxb_buffer)
    CK(subMapYmxb_buffer= (float *)malloc(sizeof(float)*numCols*numRows));
  if(!diffSection_buffer)
    CK(diffSection_buffer= (float *)malloc(sizeof(float)*numCols*numRows));
  CK(tmpSection= (float *)malloc(sizeof(float)*numCols*numRows));
  
  secOut = (imgNum-1)*numRows*numCols;
  if(shadedMaps && !dispSections)
    {/* surface shaded */
      thisSection1 = map_image_buffer;
      thisSection2 = subtract_map_image_buffer;
    }/* surface shaded */
  else
    {/* sections */
      thisSection1 = float_image_buffer1;
      thisSection2 = subtract_image_buffer1;
    }/* sections */
  if(!thisSection2)
    {/* No go */
      free(tmpSection);
      goto notReallyDoIt;
    }/* No go */

  s1 = XmTextGetString(subMagMinRadW); 
  radlo = (float)strtod(s1,0);
  XtFree(s1);
  s1 = XmTextGetString(subMagMaxRadW); 
  radhi = (float)strtod(s1,0);
  XtFree(s1);
  radhiSq = radhi * radhi;
  radloSq = radlo * radlo;

  maxMapP = saveMaxMapPixel;
  minMapP = saveMinMapPixel;
  for(i=0;i<numRows;i++)
    {
      rowOut = i * numCols;
      for(j=0;j<numCols;j++)
	{/* Do all */
	  colOut = rowOut + j;
	  if(shadedMaps && !dispSections)
	    {/* Shaded maps */
	      tmpSection[colOut] = 
		(float)(thisSection2[secOut+colOut]);
	    }/* Shaded maps */
	  else
	    {/* sections */
	      tmpSection[colOut] = 
		(float)(thisSection2[secOut+colOut] * 
			diffSecScrollValue)
		  + diffSec2ScrollValue ;
	    }/* sections */
	  if(tmpSection[colOut] > maxMapP)
	    maxMapP = tmpSection[colOut];
	  if(tmpSection[colOut] < minMapP)
	    minMapP = tmpSection[colOut];
	}/* Do all */
    }

  /* Now adjust the subMapYmxb_buffer for the MAG */
  thisNumSecs = 1;
  xc = (sub2binCol/2.0);
  yc = (sub2binRow/2.0);
  zc = 0.0;

  mag_write_(tmpSection,
	     &numCols,&numRows,&thisNumSecs,
	     &xc,&yc,&zc,&magSecScrollValue,
	     subMapYmxb_buffer);

  free(tmpSection);

  if(shadedMaps && !dispSections)
    {/* Shaded maps */
      goto jumpOverCalc;
    }/* Shaded maps */

  /* Do the diff */
  ycen = numRows/2;
  xcen = numCols/2;
  for(i=0;i<numRows;i++)
    {
      rowOut = i * numCols;
      for(j=0;j<numCols;j++)
	{/* Do all */
	  colOut = rowOut + j;

	  /* normal */
	  diffSection_buffer[colOut] = 
	    (thisSection1[secOut+colOut] -
	     subMapYmxb_buffer[colOut]);
	  if(diffSection_buffer[colOut] > maxMapP)
	    maxMapP = diffSection_buffer[colOut];
	  if(diffSection_buffer[colOut] < minMapP)
	    minMapP = diffSection_buffer[colOut];

	  /* Make sure we get min max for the corrected image 2 */
	  if(subMapYmxb_buffer[colOut] > maxMapP)
	    maxMapP = subMapYmxb_buffer[colOut];
	  if(subMapYmxb_buffer[colOut] < minMapP)
	    minMapP = subMapYmxb_buffer[colOut];

	}/* Do all */
    }

  /* Display thresholding if appropriate */
  if(threshDoit)
    {/* Ok */
      for(i=0;i<numRows;i++)
	{/* row */
	  rowOut = i * numCols;
	  xSq    = (i-ycen)*(i-ycen);
	  for(j=0;j<numCols;j++)
	    {/* Do all */
	      colOut = rowOut + j;
	      /* Now threshold if appropriate - color in Yellow */
	      if(threshAbs)
		{/* Do for ABS */
		  float negVal,posVal;
		  ySq= (j-xcen)*(j-xcen);
		  distnce = (float)(xSq+ySq); 
		  if(distnce < radhiSq &&
		     distnce > radloSq)
		    {/* we've found one */
		      if(threshScrollValue < 0.0)
			{
			  negVal = threshScrollValue;
			  posVal = 0.0 - negVal;
			}
		      else
			{
			  posVal = threshScrollValue;
			  negVal = 0.0 - posVal;
			}
		      if(diffSection_buffer[colOut] < negVal)
			diffSection_buffer[colOut] = -1234567.0; /* Green */
		      else if(diffSection_buffer[colOut] > posVal)
			diffSection_buffer[colOut] = -1234568.0; /* Red */
		    }/* we've found one */
		}/* Do for ABS */
	      else
		{/* normal */
		  if(diffSection_buffer[colOut] < threshScrollValue)
		    {/* Candidate for thresholding */
		      if(distnce < radhiSq &&
			 distnce > radloSq)
			{/* we've found one */
			  diffSection_buffer[colOut] = -1234567.0;
			}/* we've found one */
		    }/* Candidate for thresholding */
		}/* normal */
	    }/* Do all */
	}/* row */
    }/* Ok */

  if(shadedMaps && !dispSections)
    {/* Shaded maps */
    }/* Shaded maps */

  if(!shadedMaps && dispSections)
    {/* Only for sections */
      if(maxMapP > saveMaxMapPixel)
	maxMapPixel = maxMapP;
      if(minMapP < saveMinMapPixel)
	minMapPixel = minMapP;

      sprintf(outStr,"Min: %.2f",minMapPixel);
      doLabel(outStr,smMinLabelW);
      sprintf(outStr,"Max: %.2f",maxMapPixel);
      doLabel(outStr,smMaxLabelW);
    }/* Only for sections */

 jumpOverCalc:;

  stuffMapImage((float *)map_image_buffer, 
		(float *)float_image_buffer1, 
		file_width,file_height,
		1,imgNum, 0);
  stuffMapImage((float *)subMapYmxb_buffer, 
		(float *)subMapYmxb_buffer, 
		file_width,file_height,
		1,1, numCols*sizeFactor);
  if(shadedMaps && !dispSections)
    {/* surface shaded */
      stuffMapImage((float *)subtract_map_image_buffer2, 
		    (float *)subtract_map_image_buffer2, 
		    file_width,file_height,
		    1,1, 2*(numCols*sizeFactor));
    }/* surface shaded */
  else
    {/* normal */
      stuffMapImage((float *)diffSection_buffer, 
		    (float *)diffSection_buffer, 
		    file_width,file_height,
		    1,1, 2*(numCols*sizeFactor));
    }/* normal */
  DrawWindow();
  if(subCanDrawWindow)
    drawMinMaxRadius();

 notReallyDoIt:;
  
}/* DIFFSEC_REALLY_DOIT */



/*************************************************************
 * diffSec_scroll_change() - They're changing the scrolls.
 */
void diffSec_scroll_change(Widget widget, int *tag, 
			   XmScrollBarCallbackStruct *callback_data)
{/* DIFFSEC_SCROLL_CHANGE */
  int thisVal,i,j,thisOne,secOut,rowOut,colOut;
  float floatVal;
  XmString simpleString;

  floatVal = (float)callback_data->value;


  if(*tag==DIFFSECSCROLL)
    {
      diffSecScrollValue = floatVal / 100.0;
      sprintf(outStr,"%.2f",diffSecScrollValue);
      XmTextSetString(diffSecTextWidget, outStr); 
    }
  else if(*tag==DIFFSEC2SCROLL)
    {
      diffSec2ScrollValue = floatVal / 100.0;
      sprintf(outStr,"%.2f",diffSec2ScrollValue);
      XmTextSetString(diffSec2TextWidget, outStr); 
    }
  else if(*tag==MAGSECSCROLL)
    {
      magSecScrollValue = floatVal / 1000.0;
      sprintf(outStr,"%.3f",magSecScrollValue);
      XmTextSetString(magSecTextWidget, outStr); 
    }
  else if(*tag==THRESHSCROLL)
    {
      threshScrollValue = floatVal;
      sprintf(outStr,"%.1f",threshScrollValue);
      XmTextSetString(threshTextWidget, outStr); 
      recalc_ymxb_thresh();
    }


  if(TRUE)
    {/* OK to display the goods */

      diffSec_really_doit();
      
    }/* OK to display the goods */

}/* DIFFSEC_SCROLL_CHANGE */



/* ************************************************************
 * diffSec_do() - Decide which widget called us and update the
 * appropriate value.
 */
void diffSec_do(Widget widget, int *tag, 
		XmAnyCallbackStruct *callback_data)
{/* DIFFSEC_DO */
  int thisVal,i,thisOne;
  float floatVal;
  XmString simpleString;
  char *thisStr;

  if(widget==diffSecTextWidget)
    {/* do this one */
      thisStr = XmTextGetString(diffSecTextWidget);
      diffSecScrollValue = (float)strtod(thisStr,0);
      XtFree(thisStr);
      
      XtSetArg(arglst[0], XmNvalue, (int)(diffSecScrollValue*100.0));
      XtSetValues(diffSecScrollWidget, arglst, 1);
    }/* do this one */
  else if(widget==diffSec2TextWidget)
    {/* do that one */
      thisStr = XmTextGetString(diffSec2TextWidget);
      diffSec2ScrollValue = (float)strtod(thisStr,0);
      XtFree(thisStr);
      
      XtSetArg(arglst[0], XmNvalue, (int)(diffSec2ScrollValue*100.0));
      XtSetValues(diffSec2ScrollWidget, arglst, 1);
    }/* do that one */
  else if(widget==subMagStartW)
    {/* do that one */
      thisStr = XmTextGetString(subMagStartW);
      magSecScrollValue = (float)strtod(thisStr,0);
      XtFree(thisStr);

      saveBmag = magSecScrollValue;
      XtSetArg(arglst[0], XmNvalue, (int)(magSecScrollValue*1000.0));
      XtSetValues(magSecScrollWidget, arglst, 1);
      sprintf(outStr,"%.3f",magSecScrollValue);
      XmTextSetString(magSecTextWidget, outStr); 

    }/* do that one */
  else if(widget==magSecTextWidget)
    {/* do that one */
      thisStr = XmTextGetString(magSecTextWidget);
      magSecScrollValue = (float)strtod(thisStr,0);
      XtFree(thisStr);
      
      XtSetArg(arglst[0], XmNvalue, (int)(magSecScrollValue*1000.0));
      XtSetValues(magSecScrollWidget, arglst, 1);
    }/* do that one */
  else if(widget==threshTextWidget)
    {/* do that one */
      thisStr = XmTextGetString(threshTextWidget);
      threshScrollValue = (float)strtod(thisStr,0);
      XtFree(thisStr);
      
      XtSetArg(arglst[0], XmNvalue, (int)(threshScrollValue));
      XtSetValues(threshScrollWidget, arglst, 1);

      recalc_ymxb_thresh();

    }/* do that one */

  diffSec_really_doit();

}/* DIFFSEC_DO */



/********************************************************************
 * subtractMap_findMag_disp() - Find the appropriate magnification.
 */
void subtractMap_findMag_disp(Widget widget, int *tag, 
			      XmToggleButtonCallbackStruct *callback_data)
{/* SUBTRACTMAP_FINDMAG_DISP */

  /* Let's let them look at it on the screen? - toggle it here */
  findMagHist_toggle =   !findMagHist_toggle ;
  if(findMagHist_toggle)
    subCanDrawWindow = FALSE;
  else subCanDrawWindow= TRUE;
  submapFindRad  = FALSE; 

  /* Unset findRad toggle */
  findRadHist_toggle =   FALSE;
  XtSetArg(arglst[0], XmNset, FALSE);
  XtSetValues(findRadDispW, arglst, 1); 

  DrawWindow();
  if(subCanDrawWindow)
    drawMinMaxRadius();

}/* SUBTRACTMAP_FINDMAG_DISP */




/*************************************************************
 * subtractMap_findRad_disp() - Find proper radius.
 */
void subtractMap_findRad_disp(Widget widget, int *tag, 
			      XmAnyCallbackStruct *callback_data)
{/* SUBTRACTMAP_FINDRAD_DISP */

  /* Let's let them look at it on the screen */
  findRadHist_toggle =   !findRadHist_toggle ;
  if(findRadHist_toggle)
    subCanDrawWindow = FALSE;
  else subCanDrawWindow= TRUE;

  submapFindRad  = TRUE;

  /* Unset findMag toggle */
  findMagHist_toggle =   FALSE;
  XtSetArg(arglst[0], XmNset, FALSE);
  XtSetValues(findMagDispW, arglst, 1); 

  DrawWindow();
  if(subCanDrawWindow)
    drawMinMaxRadius();

}/* SUBTRACTMAP_FINDRAD_DISP */




/*************************************************************
 * subtractMap_flicker_arm() - WHen they press the flicker
 * button call here.
 */
void subtractMap_flicker_arm(Widget widget, int *tag, 
			     XmAnyCallbackStruct *callback_data)
{/* SUBTRACTMAP_FLICKER_ARM */
  int i,j;

  if(!subtract_image_buffer1)
    {/* Must read first */
      askThem("Must Read second map file first.",
	      FALSE,XmDIALOG_INFORMATION);
    }/* Must read first */
  else
    {/* OK to go */
      /* Let's let them look at it on the screen */
      /* Let's let findMagHist toggle decide this subCanDrawWindow = TRUE; */
      /* Let's see how many flickers they want */
      if(*tag==2)
	j = 9999;     /* Virtually infinite */
      else if(*tag==1)
	j = 10;
      else
	j = 1;

      for(i=0;i<j;i++)
	{/* Do all iters */

	  /* Look here for the event to stop the infinite case */
	  if(*tag==2)
	    {/* infinite */
	      if(XCheckTypedWindowEvent(dsply,
					XtWindow(infFlickerWidget),
					ButtonPress,
					&event))
		{/* STOP!!! */
		  XFlush(dsply);
		  goto noMore;
		}/* STOP!!! */

	    }/* infinite */

	  if(diffSection)
	    {/* They're doing diffsections */
	      if(flickerIt==12)
		{/* Do 1,2 */
		  stuffMapImage((float *)subMapYmxb_buffer, 
				(float *)subMapYmxb_buffer, 
				file_width,file_height,
				1,1, 0);
		  stuffMapImage((float *)subMapYmxb_buffer, 
				(float *)subMapYmxb_buffer, 
				file_width,file_height,
				1,1, numCols*sizeFactor);
		  if(shadedMaps && !dispSections)
		    {/* surface shaded */
		      stuffMapImage((float *)subtract_map_image_buffer2, 
				    (float *)subtract_map_image_buffer2, 
				    file_width,file_height,
				    1,1, 2*(numCols*sizeFactor));
		    }/* surface shaded */
		  else
		    {/* normal */
		      stuffMapImage((float *)diffSection_buffer, 
				    (float *)diffSection_buffer, 
				    file_width,file_height,
				    1,1, 2*(numCols*sizeFactor));
		    }/* normal */
		}/* Do 1,2 */
	      else if(flickerIt == 13)
		{/* Do 1,3 */
		  if(shadedMaps && !dispSections)
		    {/* surface shaded */
		      stuffMapImage((float *)subtract_map_image_buffer2, 
				    (float *)subtract_map_image_buffer2, 
				    file_width,file_height,
				    1,1, 0);
		    }/* surface shaded */
		  else
		    {/* normal */
		      stuffMapImage((float *)diffSection_buffer, 
				    (float *)diffSection_buffer, 
				    file_width,file_height,
				    1,1, 0);
		    }/* normal */
		  stuffMapImage((float *)subMapYmxb_buffer, 
				(float *)subMapYmxb_buffer, 
				file_width,file_height,
				1,1, numCols*sizeFactor);
		  if(shadedMaps && !dispSections)
		    {/* surface shaded */
		      stuffMapImage((float *)subtract_map_image_buffer2, 
				    (float *)subtract_map_image_buffer2, 
				    file_width,file_height,
				    1,1, 2*(numCols*sizeFactor));
		    }/* surface shaded */
		  else
		    {/* normal */
		      stuffMapImage((float *)diffSection_buffer, 
				    (float *)diffSection_buffer, 
				    file_width,file_height,
				    1,1, 2*(numCols*sizeFactor));
		    }/* normal */
		}/* Do 1,3 */
	      else if(flickerIt == 23)
		{/* Do 2,3 */
		  stuffMapImage((float *)map_image_buffer, 
			    (float *)float_image_buffer1, 
				file_width,file_height,
				1,imgNum, 0);
		  if(shadedMaps && !dispSections)
		    {/* surface shaded */
		      stuffMapImage((float *)subtract_map_image_buffer2, 
				    (float *)subtract_map_image_buffer2, 
				    file_width,file_height,
				    1,1, (numCols*sizeFactor));
		      stuffMapImage((float *)subtract_map_image_buffer2, 
				    (float *)subtract_map_image_buffer2, 
				    file_width,file_height,
				    1,1, 2*(numCols*sizeFactor));
		    }/* surface shaded */
		  else
		    {/* normal */
		      stuffMapImage((float *)diffSection_buffer, 
				    (float *)diffSection_buffer, 
				    file_width,file_height,
				    1,1, (numCols*sizeFactor));
		      stuffMapImage((float *)diffSection_buffer, 
				    (float *)diffSection_buffer, 
				    file_width,file_height,
				    1,1, 2*(numCols*sizeFactor));
		    }/* normal */
		}/* Do 2,3 */
	    }/* They're doing diffsections */
	  else
	    {/* sections */
	      /* subtract_map_image_buffer is used for shaded views of the
	       * second map */
	      stuffMapImage((float *)subtract_map_image_buffer, 
			    (float *)subtract_image_buffer1, 
			    file_width,file_height,
			    nsecs,imgNum, 0);
	    }/* sections */
	  
	  DrawWindow();
	  if(subCanDrawWindow)
	    drawMinMaxRadius();

	  if((*tag==1) || (*tag==2))
	    {/* Got to disarm it also */
	      int speed;
	      for(speed=0;speed<flickSpeed;speed++)
		em_sleep(1);
	      subtractMap_flicker_disarm(widget, tag, callback_data);
	      for(speed=0;speed<flickSpeed;speed++)
		em_sleep(1);
	    }/* Got to disarm it also */

	}/* Do all iters */
    }/* OK to go */
  
 noMore:;

}/* SUBTRACTMAP_FLICKER_ARM */



/*************************************************************
 * subtractMap_flicker_disarm() - When they let up on the
 * flicker button call here.
 */
void subtractMap_flicker_disarm(Widget widget, int *tag, 
				XmAnyCallbackStruct *callback_data)
{/* SUBTRACTMAP_FLICKER_DISARM */

  if(!subtract_image_buffer1)
    {/* Must read first */
    }/* Must read first */
  else
    {/* OK to go */
      if(diffSection)
	{/* They're doing diffsections */
	  stuffMapImage((float *)map_image_buffer, 
			(float *)float_image_buffer1, 
			file_width,file_height,
			1,imgNum, 0);
	  stuffMapImage((float *)subMapYmxb_buffer, 
			(float *)subMapYmxb_buffer, 
			file_width,file_height,
			1,1, numCols*sizeFactor);
	  if(shadedMaps && !dispSections)
	    {/* surface shaded */
	      stuffMapImage((float *)subtract_map_image_buffer2, 
			    (float *)subtract_map_image_buffer2, 
			    file_width,file_height,
			    1,1, 2*(numCols*sizeFactor));
	    }/* surface shaded */
	  else
	    {/* normal */
	      stuffMapImage((float *)diffSection_buffer, 
			    (float *)diffSection_buffer, 
			    file_width,file_height,
			    1,1, 2*(numCols*sizeFactor));
	    }/* normal */
	}/* They're doing diffsections */
      else
	{/* sections */
	  stuffMapImage((float *)map_image_buffer, 
			(float *)float_image_buffer1, 
			file_width,file_height,
			nsecs,imgNum, 0);
	}/* sections */
      
      DrawWindow();
      if(subCanDrawWindow)
	drawMinMaxRadius();
    }/* OK to go */

}/* SUBTRACTMAP_FLICKER_DISARM */





/*************************************************************
 * subtractMap_activate() - Get the subtract map interface 
 * up on the display.
 */
void subtractMap_activate(Widget widget, int *tag, 
			  XmAnyCallbackStruct *callback_data)
{/* SUBTRACTMAP_ACTIVATE */

  sprintf(outStr,"%s",fileName.addr);
  XmTextSetString(subtractFile1TextW,outStr);
  sprintf(outStr,"");
  XmTextSetString(subtractFile2TextW,outStr);

  XtSetArg(arglst[0], XmNcursorPosition, strlen(outStr) );
  XtSetValues(subtractFile1TextW, arglst, 1); 
  XFlush(dsply);

  sprintf(outStr,"1.00");
  XmTextSetString(subMagStartW,outStr); 
  sprintf(outStr,"0.01");
  XmTextSetString(subMagStepW,outStr); 
  sprintf(outStr,"11");
  XmTextSetString(subMagNumStepW,outStr); 

  sprintf(outStr,"1");
  XmTextSetString(subMagMinRadW,outStr); 
  sprintf(outStr,"%d",numSecs/2);
  XmTextSetString(subMagMaxRadW,outStr); 

/*  submapFindRad  = FALSE; */
/*
  XtSetArg(arglst[0], XmNsensitive, FALSE );
  XtSetValues(computeScalingW, arglst, 1); 
*/
  if(pifHeader.pixSizUnits==1 && pifHeader.pixelSize > 0.0)
    slidingWindow = (int)(30.0 / pifHeader.pixelSize);
  else if(pifHeader.pixSizUnits==2 && pifHeader.pixelSize > 0.0)
    slidingWindow = (int)(3.0 / pifHeader.pixelSize);
  else
    slidingWindow = 5;
  slidingWindow= MAX(MIN(slidingWindow,10),2);
  sprintf(outStr,"%d",slidingWindow);

  sprintf(outStr,"%d",(pifHeader.numCols/2)+1);
  XmTextSetString(imgNumTextWidget,outStr);

  qtr_switch = 0;
  qtr_switch = 1;  /* 1 mean try doing the calc with only 1/4 of the MAP */
  raiseIt(subtractMapScreen);
  XFlush(dsply);

  /* Draw the flicker speed "thermometer" */
  DrawFlickThermometer();

}/* SUBTRACTMAP_ACTIVATE */





/*************************************************************
 * subtractMap_browse_activate() - Browse which MAP file
 * they want to be subtracted from the one already in
 * memory.
 */
void subtractMap_browse_activate(Widget widget, int *tag, 
				 XmAnyCallbackStruct *callback_data)
{/* SUBTRACTMAP_BROWSE_ACTIVATE */

  readWrite= READING;
  subtractingMap = TRUE;

  raiseIt(fileSel_widget);

}/* SUBTRACTMAP_BROWSE_ACTIVATE */





/*************************************************************
 * subtractMap_deactivate() - Get rid of subtract map interface.
 */
void subtractMap_deactivate(Widget widget, int *tag, 
			    XmAnyCallbackStruct *callback_data)
{/* SUBTRACTMAP_DEACTIVATE */

  subtractingMap = FALSE;
  subCanDrawWindow  = TRUE; 
  drawingFindRad = FALSE;
  if(mags)
    free(mags);
  if(coefs)
    free(coefs);
  if(cntr1)
    free(cntr1);
  if(cntr2)
    free(cntr2);
  if(rdens)
    free(rdens);
  if(wrdens)
    free(wrdens);
  if(ccoef)
    free(ccoef);
  if(wccoef)
    free(wccoef);
  if(s_wccoef)
    free(s_wccoef);
  XtUnmanageChild(subtractMapScreen);
  if(subtract_image_buffer2)
    free(subtract_image_buffer2);
  subtract_image_buffer2 = 0;
  if(subtract_image_buffer1)
    free(subtract_image_buffer1);
  subtract_image_buffer1 = 0;
  if(subMapYmxb_buffer)
    free(subMapYmxb_buffer);
  subMapYmxb_buffer = 0;
  if(diffSection_buffer)
    free(diffSection_buffer);
  diffSection_buffer = 0;


}/* SUBTRACTMAP_DEACTIVATE */






/*************************************************************
 * coravgArrowActivate() - correlation average arrow
*/
void coravgArrowActivate(Widget widget, int *tag, 
			 XmAnyCallbackStruct *callback_data)
{/* CORAVGARROWACTIVATE */
  float speedFactor;
  char *inStr;
  int binVal;
  float stepVal;

  inStr = XmTextGetString(emcoravgRadiusTextW);
  
  stepVal = (float)strtod(inStr,0);
  XtFree(inStr);

  if(*tag==CORAVGUP)
    stepVal += 1.0;
  else 
    stepVal -= 1.0;

  stepVal= MAX(stepVal,1.0000000);

  coravgRadius = (int)stepVal;
  sprintf(outStr,"%.1f",stepVal);
  XmTextSetString(emcoravgRadiusTextW,outStr);

  DrawWindow();  /* Show new radius */

}/* CORAVGARROWACTIVATE */





/*************************************************************
 * coravgArrowButton() - correlation average arrows get 
 * different speed depending on which mouse button is pressed.
*/
void coravgArrowButton(Widget widget, XEvent *event)
{/* CORAVGARROWBUTTON */
  XEvent thisEvent;
  float speedFactor;
  char *inStr;
  float stepVal;
  int binVal;
  int     tag;
  XmAnyCallbackStruct callback_data;

  inStr = XmTextGetString(emcoravgRadiusTextW);

  stepVal = (float)strtod(inStr,0);
  XtFree(inStr);

  thisEvent= *(event);
  
  /* Get which button */
  if((thisEvent.xbutton.button == Button1))
    speedFactor = 1.0;
  if((thisEvent.xbutton.button == Button2))
    speedFactor = 5.0;
  if((thisEvent.xbutton.button == Button3))
    speedFactor = 10.0;
  
  if(widget==coravgUpArrowW)
    stepVal += speedFactor;
  else 
    stepVal -= speedFactor;

  stepVal= MAX(stepVal,1.000000);

  coravgRadius = (int)stepVal;
  sprintf(outStr,"%.1f",stepVal);
  XmTextSetString(emcoravgRadiusTextW,outStr);

  DrawWindow();  /* Show new radius */

}/* CORAVGARROWBUTTON */







/* ************************************************************
 * emcoravg_create() - Remember the emcoravg screen for use later.  
 */
void emcoravg_create(Widget widget, int *tag, 
		     XmAnyCallbackStruct *callback_data)
{/* EMCORAVG_CREATE */
  
  if(*tag==EMCORAVGZOOM)
    emcoravgZoomTextW = widget;
  else if(*tag==EMCORAVGRADIUS)
    emcoravgRadiusTextW = widget;
  else if(*tag==CORAVGUP)
    {coravgUpArrowW = widget;
     handle_cr(widget);}
  else if(*tag==CORAVGDOWN)
    {coravgDownArrowW = widget;
     handle_cr(widget);}
  else if(*tag==LOWLIM)
    coravgLowLimScrollW = widget;
  else if(*tag==LOWLIMVAL)
    coravgLowLimValueW = widget;

  
}/* EMCORAVG_CREATE */


/*************************************************************
 * coravg_scroll_change() - Scrollbar changes.
 */
void coravg_scroll_change(Widget widget, int *tag, 
			  XmScrollBarCallbackStruct *callback_data)
{/* CORAVG_SCROLL_CHANGE */
  int thisVal,i,j,thisOne,secOut,rowOut,colOut;
  float floatVal;
  XmString simpleString;

  floatVal = (float)callback_data->value;

  coravgLowLimValue = floatVal / 100.0;

  /* For future use 
     sprintf(outStr,"%.2f",coravgLowLimValue);
     XmTextSetString(coravgLowLimTextW, outStr); 
     */
  /* Put up on interface */
  sprintf(outStr,"%.0f",coravgLowLimValue);
  createSimpleString(&simpleString,outStr);
  XtSetArg(arglst[0], XmNlabelString, simpleString);
  XtSetArg(arglst[1], XmNrecomputeSize, FALSE );
  XtSetValues(coravgLowLimValueW, arglst, 2); 
  XFlush(dsply);
  XmStringFree(simpleString); 


}/* CORAVG_SCROLL_CHANGE */




/* ************************************************************
 * emcoravg_text_changed() - Text changed.
 */
void emcoravg_text_changed(Widget widget, int *tag, 
			   XmAnyCallbackStruct *callback_data)
{/* EMCORAVG_TEXT_CHANGED */
  
  if(*tag==EMCORAVGZOOM)
    {/* Zoom widget called us */
      stuffEmcoravgImage();
    }/* Zoom widget called us */

  
}/* EMCORAVG_TEXT_CHANGED */



/* ************************************************************
 * emcoravg_screen_deactivate() - Get the emcoravg screen
 * off the display.
 */
void emcoravg_screen_deactivate(Widget widget, int *tag, 
				XmAnyCallbackStruct *callback_data)
{/* EMCORAVG_SCREEN_DEACTIVATE */
  
  if(emcoravg_image_buffer1)
    free(emcoravg_image_buffer1);
  if(emcoravg_image_buffer_disk)
    free(emcoravg_image_buffer_disk);
  if(emcoravg_image_buffer_avg)
    free(emcoravg_image_buffer_avg);
  emcoravg_image_buffer1 = NULL;
  emcoravg_image_buffer_disk = 
    emcoravg_image_buffer_avg = 
    NULL;
  XtUnmanageChild(emcoravgscreen);

}/* EMCORAVG_SCREEN_DEACTIVATE */




/* ************************************************************
 * emcoravg_save_activate() - Save to disk in PIF file.
 */
void emcoravg_save_activate(Widget widget, int *tag, 
			    XmAnyCallbackStruct *callback_data)
{/* EMCORAVG_SAVE_ACTIVATE */
  int i,j;
  forPifHead_t thisHead;
  fString_t rWrite,fName;  
  BYTE *tmpByte_buffer;
  float *tmpBuffer;
  FILE *fOut;
  float minF,maxF,scaleF;

  rWrite.addr = malloc(sizeof(char)*4);
  rWrite.length = 1;
  fName.addr = malloc(sizeof(char)*244);
  fName.length = strlen(fName.addr);

  sprintf(rWrite.addr,"w");         /* Write it */
  sprintf(fName.addr,"coravg.pif"); 
  pif_init_head(&thisHead);
  thisHead.luNum = 4;

  pif_open_(fName.addr, &(thisHead.luNum), rWrite.addr);


  /* [1] Write global file header */
  calcStats(emcoravg_image_buffer_avg,
	    &purdueHeader.img_bkg, 
	    &purdueHeader.img_min,
	    &purdueHeader.img_max, 
	    &purdueHeader.img_avg, 
	    &purdueHeader.img_std,
	    emcoravgNcol, emcoravgNrow, 
	    emcoravgNcol);
	      
  thisHead.nimg = 1;
  thisHead.numCols = emcoravgNcol;
  thisHead.numRows = emcoravgNrow;
  thisHead.numSecs = 1;
  thisHead.mode = 6;

  pif_write_gh(&thisHead);
	      
  /* [2] Write header for each image */
  thisHead.img_bkg    = (int)purdueHeader.img_avg;;
  thisHead.packRadius = emcoravgNcol/2;
		  
  thisHead.datai4_max= purdueHeader.img_max;
  thisHead.datar4_max= (float)purdueHeader.img_max;
  thisHead.datai4_min= purdueHeader.img_min;
  thisHead.datar4_min= (float)purdueHeader.img_min;
  thisHead.datar4_avg= (float)purdueHeader.img_avg;
  thisHead.datar4_std= (float)purdueHeader.img_std;

  thisHead.binFactor     = 1;
  thisHead.imgNum = 1;
  
  pif_write_dh(&thisHead);

  /* [3] Now write the image info */
  thisHead.imgOffset= 1024;
  CK(tmpBuffer=(float *)malloc(sizeof(float)*
				   emcoravgNcol*emcoravgNrow));
  CK(tmpByte_buffer=(BYTE *)malloc(sizeof(BYTE)*
				   emcoravgNcol*emcoravgNrow));
  tsbworld_float(emcoravg_image_buffer_avg,
		tmpBuffer,
		emcoravgNcol,
		emcoravgNcol,1);
  minF= maxF = tmpBuffer[0];
  for(i=0;i<(thisHead.numCols*thisHead.numRows);i++)
    {
      maxF = MAX(maxF, tmpBuffer[i]);
      minF = MIN(minF, tmpBuffer[i]);
    }
  scaleF = 254.0 /(maxF-minF);
  for(i=0;i<(thisHead.numCols*thisHead.numRows);i++)
    tmpByte_buffer[i] = ((tmpBuffer[i]-minF)*scaleF);
  pif_write_byte_image_(&(thisHead.luNum),
			tmpByte_buffer,
			&(thisHead.imgOffset),
			&(thisHead.numCols),
			&(thisHead.numRows),
			&(thisHead.numSecs));
  free(tmpByte_buffer);
  free(tmpBuffer);

  pif_close(&(thisHead.luNum));
  sprintf(outStr,"File coravg.pif written successfully.");
  askThem(outStr,FALSE,XmDIALOG_INFORMATION);
  free(rWrite.addr);
  free(fName.addr);
  

  /* Now write a raw file */
  fOut = fopen("coravg.raw","w");
  if(fOut)
    {/* Do it */
      fwrite(emcoravg_image_buffer_avg, sizeof(BYTE),
	     (emcoravgNcol*emcoravgNrow),fOut);
      fclose(fOut);
      sprintf(outStr,"File coravg.raw written with dimensions: \n\
    %d cols\n\
    %d rows\n\
    in %d eight bit bytes",emcoravgNcol,emcoravgNrow,
	      emcoravgNcol*emcoravgNrow);
      askThem(outStr,FALSE,XmDIALOG_INFORMATION);

    }/* Do it */
  else
    askThem("Can't open coravg.raw file.",FALSE,XmDIALOG_INFORMATION);

}/* EMCORAVG_SAVE_ACTIVATE */





/* ************************************************************
 * emcoravg_tryit_activate() - Do this coravg.
 */
void emcoravg_tryit_activate(Widget widget, int *tag, 
			     XmAnyCallbackStruct *callback_data)
{/* EMCORAVG_TRYIT_ACTIVATE */
  float *fft1,*fft2,*fft1_half,*fft2_half;
  int idim1,idim2,imgNcols,imgNrows,i,j,k,l,m,bufNum,cor_pass;
  BYTE *cor_byte_buffer;
  float *c_buffer,*b_buffer,*array,*cor_map,*avg_map;
  float *cor_img,*avg_img;
  int img_num,outWidth,outHeight,yOut;
  float cor_avg,cor_std,radius,scaleFactor;
  float *peaks,thisBackground;
  int nlim, npeak, lowlim, npeak_unique, ncolmp, nrowmp, arg1;
  char *s1;
  int ksign,iform,thisMin,thisMax;
  purdueHead_t thisHeader;

  if(!emcoravg_image_buffer1)
    {/* Can't go on */
      askThem("Must read reference file first.",
	      FALSE,XmDIALOG_INFORMATION);
      goto noWay;
    }/* Can't go on */

  lowlim = (int)coravgLowLimValue;

  imgNcols = purdueHeader.numCols;
  imgNrows = purdueHeader.numRows;

  fft_setdim_local_ (&imgNcols, &imgNrows , 
		     &idim1, &idim2);

  printf ("idim1,2[%d,%d]\n",idim1,idim2);

  CK(fft1 = (float *)malloc(sizeof(float)*idim1*idim2*2));
  CK(fft2 = (float *)malloc(sizeof(float)*idim1*idim2*2));
  CK(fft1_half = (float *)malloc(sizeof(float)*idim1*((idim2/2)+1)*2));
  CK(fft2_half = (float *)malloc(sizeof(float)*idim1*((idim2/2)+1)*2));
  
  CK(cor_byte_buffer = (BYTE *)malloc(sizeof(BYTE)*idim1*idim2));
  CK(c_buffer = (float *)malloc(sizeof(float)*idim1*2));
  CK(b_buffer = (float *)malloc(sizeof(float)*idim2*2));
  CK(array = (float *)malloc(sizeof(float)*2*idim1*((idim2/2)+1)));
  CK(cor_map = (float *)malloc(sizeof(float)*idim1*idim2));
  CK(cor_img = (float *)malloc(sizeof(float)*idim1*idim2));

  /* Let's try 50000 for now ... */
  nlim = 50000;
  CK(peaks = (float *)malloc(sizeof(float)*3*nlim));

  ksign = 1;       /* Do a "forward" FFT */
  iform = 1;       /* Data is imaginary */

  if(useAvgCorAvg)
    {/* Use the last average */
      for(i=0;i<emcoravgNrow;i++)
	for(j=0;j<emcoravgNcol;j++)
	  cor_img[(i*emcoravgNcol)+j] = (float)emcoravg_image_buffer_avg[
						  (i*emcoravgNcol)+j];
      /* We need a new background value */
      calcStats(emcoravg_image_buffer_avg,
		&(thisHeader.img_bkg), 
		&(thisHeader.img_min),
		&(thisHeader.img_max), 
		&(thisHeader.img_avg), 
		&(thisHeader.img_std),
		emcoravgNcol, emcoravgNrow, 
		emcoravgNcol);
    }/* Use the last average */
  else
    {/* Use the one read from disk */
      for(i=0;i<emcoravgNrow;i++)
	for(j=0;j<emcoravgNcol;j++)
	  cor_img[(i*emcoravgNcol)+j] = (float)emcoravg_image_buffer_disk[
						  (i*emcoravgNcol)+j];
      
      /* We need a new background value */
      calcStats(emcoravg_image_buffer_disk,
		&(thisHeader.img_bkg), 
		&(thisHeader.img_min),
		&(thisHeader.img_max), 
		&(thisHeader.img_avg), 
		&(thisHeader.img_std),
		emcoravgNcol, emcoravgNrow, 
		emcoravgNcol);
    }/* Use the one read from disk */


  emcoravgBkg        = thisHeader.img_bkg; 
  printf("thisBackground = [%f]\n",emcoravgBkg );

  thisBackground = emcoravgBkg;
  map_fft_fill_robemorien_(cor_img, &emcoravgNcol, &emcoravgNrow, 
			   fft1, 
			   &idim1, &idim2, &thisBackground);
  
  fft_2d_rob_(fft1, &idim1, &idim2, &ksign, &iform);
  fft_copy1_ (fft1, fft1_half, &idim1, &idim2);

  if(imgInCore)
    {/* Do for scanned image */
      for(i=0;i<imgNrows;i++)
	for(j=0;j<imgNcols;j++)
	  cor_img[(i*imgNcols)+j] = (float)inCore_buffer[
							 (i*imgNcols)+j];
    }/* Do for scanned image */
  else if(fileType==PACKEDPORTABLE)
    {/* Do for packed */
      float *thisImg;
      thisImg  = packImg[0];
      for(i=0;i<imgNrows;i++)
	for(j=0;j<imgNcols;j++)
	  cor_img[(i*imgNcols)+j] = (float)thisImg[(i*imgNcols)+j];
    }/* Do for packed */
  else
    {/* Can't do */
      sprintf(outStr,"Corr Average only for scanned or packed images.");
      askThem(outStr,FALSE,XmDIALOG_INFORMATION);
      goto noEmcoravg;
    }/* Can't do */

  thisBackground = purdueHeader.img_bkg;
  map_fft_fill_robemorien_(cor_img, &imgNcols, &imgNrows, 
		    fft2, 
		    &idim1, &idim2, &thisBackground);
  
  fft_2d_rob_(fft2, &idim1, &idim2, &ksign, &iform);
  fft_copy1_ (fft2, fft2_half, &idim1, &idim2);

  cor_pass = 0;
  cor_first_(&idim1, &idim2, array, fft1_half, fft2_half, c_buffer, &cor_pass);
	 
  img_num = 1;
  cor_second_(&idim1, &idim2, array, cor_map, 
	      cor_img, b_buffer,cor_byte_buffer,&img_num, 
	      &cor_avg, &cor_std, &thisMin, &thisMax);

  /* Now put it on the display */
  outWidth  = idim1;
  outHeight = idim2;
  printf("outdims[%d,%d]\n",outWidth,outHeight);
  thisMax = thisMin = cor_img[0];
  for(j=0; j<outHeight*outWidth; j++)
    {
      thisMax = MAX(thisMax,cor_img[j]);
      thisMin = MIN(thisMin,cor_img[j]);
    }
  scaleFactor = ((float)tryColors-9.0) / (float)(thisMax-thisMin);
      
  for(j=0; j<outHeight; j++)
    {/* hieght */
      int thisPixel;
      yOut = j*outWidth;
      for(k=0; k<outWidth; k++)
	{/* Do all cols */
	  thisPixel = (int)((((float)cor_img[yOut+k] -
			      (float)thisMin)*
			     scaleFactor)+8.0);
	  if(hiColor)
	    {/* hi */
	      /****************************
	      if(showLowLim &&
		 (thisPixel < lowlim))
		image_buffer1[((j*buffer_width)+k)]= pixels_return[RED];
	      else
		image_buffer1[((j*buffer_width)+k)]= pixels_return[thisPixel];
	      *************************/
	      if(showLowLim &&
		 (thisPixel < lowlim))
		XPutPixel(image_image1, idim1+k, j, pixels_return[RED]);
	      else
		XPutPixel(image_image1, idim1+k, j, pixels_return[thisPixel]);
	    }/* hi */
	  else
	    {/* lo */
	      if(showLowLim &&
		 (thisPixel < lowlim))
		image_buffer1[((j*buffer_width)+k)]= (BYTE)pixels_return[RED];
	      else
		image_buffer1[((j*buffer_width)+k)]=
		  (BYTE)pixels_return[thisPixel];
	    }/* lo */

	}/* Do all cols */
    }/* hieght */
  
  XPutImage(dsply, image_window1, image_gc, image_image1, 
	    0, 0,
	    0, 0,  outWidth*3, outHeight);
  XFlush(dsply);

  coravg_lim_(&nlim, peaks, &npeak, cor_map, &idim1,
	      &idim2, &lowlim, &cor_avg, &cor_std);

printf("npeak[%d]\n",npeak);

  if(npeak <= 0)
    {/* THEN */
      sprintf(outStr,"No peaks found");
      askThem(outStr,FALSE,XmDIALOG_INFORMATION);
      goto allDone;
    }/* THEN */

  s1 = XmTextGetString(emcoravgRadiusTextW);
  radius = strtod(s1,0);
  XtFree(s1);
printf("radius=[%f]\n",radius);

  arg1 = 1;
  peak_chk_(peaks, &npeak, &npeak_unique, &radius, &arg1);

  peak_sort_hi_(peaks, &npeak);
  peak_cen_(peaks, &npeak_unique, cor_map,&idim1,&idim2,&arg1);

  ncolmp = emcoravgNcol;
  nrowmp = emcoravgNrow;
  CK(avg_map= (float *)malloc(sizeof(float)*ncolmp*nrowmp));
  CK(avg_img= (float *)malloc(sizeof(float)*ncolmp*nrowmp));
  map_clear_local_ (avg_map, &ncolmp, &nrowmp);
	   
  if(imgInCore)
    {/* Do for scanned image */
      for(i=0;i<imgNrows;i++)
	for(j=0;j<imgNcols;j++)
	  cor_img[(i*imgNcols)+j] = (float)inCore_buffer[
							 (i*imgNcols)+j];
    }/* Do for scanned image */
  else if(fileType==PACKEDPORTABLE)
    {/* Do for packed */
      float *thisImg;
      thisImg  = packImg[0];
      for(i=0;i<imgNrows;i++)
	for(j=0;j<imgNcols;j++)
	  cor_img[(i*imgNcols)+j] = (float)thisImg[(i*imgNcols)+j];
    }/* Do for packed */
  else
    {/* Can't do */
      sprintf(outStr,"Corr Average only for scanned or packed images.");
      askThem(outStr,FALSE,XmDIALOG_INFORMATION);
      goto noEmcoravg;
    }/* Can't do */

  coravg_sum_(cor_img,
	      &imgNcols, &imgNrows, avg_img,avg_map,&ncolmp,
	      &nrowmp, &idim1, &idim2,  peaks, &npeak_unique, &img_num);





  /* Remember average image */
  for(i=0;i<nrowmp;i++)
    for(j=0;j<ncolmp;j++)
      {
	emcoravg_image_buffer_avg[(i*ncolmp)+j] = 
	  avg_img[(i*ncolmp)+j];
      }
  
  /* Stuff display */
  if(useAvgCorAvg)
    for(i=0;i<nrowmp;i++)
      for(j=0;j<ncolmp;j++)
	emcoravg_image_buffer1[(i*ncolmp)+j] = 
	  avg_img[(i*ncolmp)+j];
  else
    for(i=0;i<nrowmp;i++)
      for(j=0;j<ncolmp;j++)
	emcoravg_image_buffer1[(i*ncolmp)+j] = 
	  emcoravg_image_buffer_disk[(i*ncolmp)+j];

  stuffEmcoravgImage();

 noEmcoravg:;

  free (avg_map);
  free (avg_img);


  allDone:;

  free(fft1);
  free(fft2);
  free(fft1_half);
  free(fft2_half);
  free(cor_byte_buffer);
  free(c_buffer);
  free(b_buffer);
  free(array);
  free(cor_map);
  free(cor_img);
  free(peaks);

 noWay:;

}/* EMCORAVG_TRYIT_ACTIVATE */



/* ************************************************************
 * emcoravg_screen_activate() - Put the emcoravgscreen on the display.
 */
void emcoravg_screen_activate(Widget widget, int *tag, 
			      XmAnyCallbackStruct *callback_data)
{/* EMCORAVG_SCREEN_ACTIVATE */

  /* Put screen on display */
  raiseIt(emcoravgscreen);

}/* EMCORAVG_SCREEN_ACTIVATE */



/*********************************************************************
 * emcoravg_file_activate() - Browse for proper file.
 */
void emcoravg_file_activate(Widget widget, int *tag, 
			    XmAnyCallbackStruct *callback_data)
{/* EMCORAVG_FILE_ACTIVATE */

  readWrite= READING;
  readEmcoravg = TRUE;

  raiseIt(fileSel_widget);

}/* EMCORAVG_FILE_ACTIVATE */



/* ************************************************************
 * DrawEmcoravgWindow() - Redraw window e.g. on expose events
 * or whenever it's prudent.
 */
void DrawEmcoravgWindow(Widget widget, int *tag, 
			XmAnyCallbackStruct *callback_data)
{/* DRAWEMCORAVGWINDOW */
  int draw_width, draw_height;
  int i,j,scaledY,scaledYp1,thisX,thisXp1;
  float *thisRint;
  XmString simpleString;

  draw_width = 256;
  draw_height = 256;
  if(XtIsManaged(emcoravgscreen))
    XPutImage(dsply, emcoravg_window1, emcoravg_gc, emcoravg_image1, 0, 0,
	      0, 0,  draw_width,  draw_height);
  
  XFlush (dsply);

}/* DRAWEMCORAVGWINDOW */






/******************************************************
 * stuffEmcoravgImage ()- Put coravg image on display.
*/
void stuffEmcoravgImage()
{/* STUFFEMCORAVGIMAGE */
  int i,j,k,l,m;
  XmString simpleString;
  char *s1;
  int zoomIt;
  float scaleFactor,minF,maxF;

  if(emcoravg_image_buffer1)
    {/* Don't do if there's nothing to do */
      s1 = XmTextGetString(emcoravgZoomTextW); 
      zoomIt = (int)strtod(s1,0);
      XtFree(s1);
      zoomIt = MAX(zoomIt,1);
      zoomIt = MIN(zoomIt,10);
      printf("zoomit[%d]\n",zoomIt);
      
      for(j=0;j<256;j++)
	for(k=0;k<256;k++)
	  XPutPixel(emcoravg_image1, 
		    k, j,
		    pixels_return[BLACK]);
      
      /* Find min/max */
      minF = maxF = (float)emcoravg_image_buffer1[0];
      for(j=0;j<emcoravgNrow;j++)
	{
	  for(k=0;k<emcoravgNcol;k++)
	    {
	      maxF = (float)MAX((float)emcoravg_image_buffer1[
					   ((j*emcoravgNcol)+k)],maxF);
	      minF = (float)MIN((float)emcoravg_image_buffer1[
					   ((j*emcoravgNcol)+k)],minF);
	    }
	}
      
      scaleFactor = ((float)tryColors-9.0) / (maxF-minF);
      
      for(j=0;j<emcoravgNrow;j++)
	{/* disk rows */
	  for(k=0;k<emcoravgNcol;k++)
	    {/* disk cols */
	      if(((j*zoomIt)+(zoomIt-1) < 256) &&
		 ((k*zoomIt)+(zoomIt-1) < 256))
		{/* OK to do */
		  for(l=0;l<zoomIt;l++)
		    {/* zoomed buffer rows */
		      for(m=0;m<zoomIt;m++)
			{/* zoomed buffer cols */
			  if(hiColor)
			    {/* hi */
			     /* emcoravg_buffer1[((((j*zoomIt)+l)*256)+
					      (k*zoomIt)+m)] = 
						pixels_return[
				(int)((((float)emcoravg_image_buffer1[(
					    (j*emcoravgNcol)+k)]-
					minF)*
				       scaleFactor)+8.0)
							      ];
			     *****/
			      XPutPixel(emcoravg_image1, 
					(k*zoomIt)+m, (j*zoomIt)+l,
		       	pixels_return[(int)((((float)emcoravg_image_buffer1[(
					 (j*emcoravgNcol)+k)]-
					 minF)*
					     scaleFactor)+8.0)]);
			    }/* hi */
			  else
			    emcoravg_buffer1[((((j*zoomIt)+l)*256)+
					      (k*zoomIt)+m)] = 
						(BYTE)pixels_return[
				 (int)((((float)emcoravg_image_buffer1[(
					    (j*emcoravgNcol)+k)]-
					 minF)*
					scaleFactor)+8.0)
								    ];
			}/* zoomed buffer cols */
		    }/* zoomed buffer rows */
		}/* OK to do */
	    }/* disk cols */
	}/* disk rows */
      
      XPutImage(dsply, emcoravg_window1, emcoravg_gc, emcoravg_image1, 
		0, 0, 0, 0, 256, 256);
      XFlush(dsply);

    }/* Don't do if there's nothing to do */

}/* STUFFEMCORAVGIMAGE */




/******************************************************
 * read_emcoravg ()- Read the emcoravg file.
*/
void read_emcoravg(char *fName)
{/* READ_EMCORAVG */
  int i,j,k,l,m,imgOffset,beginParms,convTsb;
  XmString simpleString;
  forPifHead_t thisHeader;
  BYTE *tmpBuffer;
  fString_t rWrite,inFile,descStr,inFileName;
  FILE *fIn;

  readEmcoravg = FALSE;

  i= 4;
  thisHeader.luNum = i;
  rWrite.addr = malloc(sizeof(char)*4);
  rWrite.length = 1;
  descStr.addr = malloc(sizeof(char)*64);
  descStr.length = 32;
  inFileName.addr = malloc(sizeof(char)*64);
  inFileName.length = 64;
  sprintf(rWrite.addr,"r");

  sprintf(inFileName.addr,fName);

  pif_open_(inFileName.addr, &i, rWrite.addr);


  pif_read_gh_(&thisHeader);
  pifMode               = thisHeader.mode;

  if(pifMode!=0 && pifMode!=6)
    {/* NOt !! */
      sprintf(outStr,"Invalid PIF mode [%d]\n",pifMode);
      askThem(outStr,FALSE,XmDIALOG_INFORMATION);
      goto noRead;
    }/* NOt !! */

  /* For fortran */
  thisHeader.imgNum = 1;
  pif_read_dh_(&thisHeader);
  
  imgOffset  = thisHeader.imgOffset; 
  pifMode    = thisHeader.mode; 
  emcoravgNcol       = thisHeader.numCols; 
  emcoravgNrow       = thisHeader.numRows; 
  emcoravgBkg        = thisHeader.img_bkg; 

  if(emcoravg_image_buffer1)
    free(emcoravg_image_buffer1);
  if(emcoravg_image_buffer_disk)
    free(emcoravg_image_buffer_disk);
  if(emcoravg_image_buffer_avg)
    free(emcoravg_image_buffer_avg);
  CK(tmpBuffer = (BYTE *)malloc(sizeof(BYTE)*
					      emcoravgNcol*emcoravgNrow));
  CK(emcoravg_image_buffer1 = (float *)malloc(sizeof(float)*
					      emcoravgNcol*emcoravgNrow));
  CK(emcoravg_image_buffer_disk = (float *)malloc(sizeof(float)*
					      emcoravgNcol*emcoravgNrow));
  CK(emcoravg_image_buffer_avg = (float *)malloc(sizeof(float)*
					      emcoravgNcol*emcoravgNrow));
  
  
  pif_read_byte_image(&(thisHeader.luNum),
		      tmpBuffer,
		      &imgOffset,
		      &(thisHeader.numCols),
		      &(thisHeader.numRows),
		      &(thisHeader.numSecs));

  pif_close_(&i);
  for(i=0;i<thisHeader.numRows;i++)
    for(j=0;j<thisHeader.numCols;j++)
      emcoravg_image_buffer1[((i*thisHeader.numCols)+j)] = (float)
	tmpBuffer[((i*thisHeader.numCols)+j)];
  free(tmpBuffer);
  
  if(pifMode==6)
    {/* DO a tmp buffer */
      float *tmpByte_buffer;
      CK(tmpByte_buffer=(float *)malloc(sizeof(float)*
				       thisHeader.numCols*
				       thisHeader.numRows));

      tsbworld_float(emcoravg_image_buffer1,
		    tmpByte_buffer,
		    thisHeader.numCols, 
		    thisHeader.numCols,
		    1);
      for(i=0;i<thisHeader.numRows;i++)
	for(j=0;j<thisHeader.numCols;j++)
	  {
	    /* Just stuff the avg for now */
	    emcoravg_image_buffer_avg[((i*thisHeader.numCols)+j)]= 
	      emcoravg_image_buffer_disk[((i*thisHeader.numCols)+j)]= 
		emcoravg_image_buffer1[((i*thisHeader.numCols)+j)]= 
		  tmpByte_buffer[((i*thisHeader.numCols)+j)]; 
	  }
      free(tmpByte_buffer);

    }/* DO a tmp buffer */


  if(emcoravgNcol > 256 || emcoravgNrow > 256)
    {/* We don't do */
      sprintf(outStr,"Dimensions too large [%d,%d]\n",
	      emcoravgNcol,emcoravgNrow);
      askThem(outStr,FALSE,XmDIALOG_INFORMATION);
      goto noRead;
    }/* We don't do */


  stuffEmcoravgImage();


 noRead:;

  
  
}/* READ_EMCORAVG */





/*********************************************************************
 * fft_feedback_activate() - Give feedback on FFT values.
 */
void fft_feedback_activate(Widget widget, int *tag, 
			   XmAnyCallbackStruct *callback_data)
{/* FFT_FEEDBACK_ACTIVATE */

  raiseIt(fftFeedbackScreen);

}/* FFT_FEEDBACK_ACTIVATE */




/*************************************************************
 * apNewOrig_do_activate() - They've set a new origin.
 */
void apNewOrig_do_activate(Widget widget, int *tag, 
			   XmAnyCallbackStruct *callback_data)
{/* APNEWORIG_DO_ACTIVATE */
  float x_shift,y_shift;
  int fft_full;
  char *s1;
  XmString simpleString;

  if(*tag==APRESET)
    {/* reset orig */
      x_shift = 0.0 - apOrigXtot;
      y_shift = 0.0 - apOrigYtot;
      apOrigXtot =apOrigYtot = 0.0;  /* We're resetting it */
    }/* reset orig */
  else
    {/* normal */

      s1 = XmTextGetString(fftnewXtextW);
      x_shift = (float)strtod(s1,0);
      XtFree(s1);
      s1 = XmTextGetString(fftnewYtextW);
      y_shift = (float)strtod(s1,0);
      XtFree(s1);

      apOrigXtot += x_shift;
      apOrigYtot += y_shift;

    }/* normal */

  sprintf(outStr,"Tot X shift[%f]",apOrigXtot);
  createSimpleString(&simpleString,outStr);
  XtSetArg(arglst[0], XmNlabelString, simpleString);
  XtSetValues(apOrigLabel1W, arglst, 1); 
  XFlush(dsply);
  XmStringFree(simpleString); 
  sprintf(outStr,"Tot Y shift[%f]",apOrigYtot);
  createSimpleString(&simpleString,outStr);
  XtSetArg(arglst[0], XmNlabelString, simpleString);
  XtSetValues(apOrigLabel2W, arglst, 1); 
  XFlush(dsply);
  XmStringFree(simpleString); 
  
  fft_full = 1;
  fft_pshift_(fftImage_gold, &fftSize, &fftSize, 
	      &x_shift, &y_shift,
	      &fft_full);

  calcFFT(FALSE, 0, FALSE);

}/* APNEWORIG_DO_ACTIVATE */





/* ************************************************************
 * fftFeedback_deactivate() - Get fft feedabck interface off the
 * display.
 */
void fftFeedback_deactivate(Widget widget, int *tag, 
			    XmAnyCallbackStruct *callback_data)
{/* FFTFEEDBACK_DEACTIVATE */
  
  XtUnmanageChild(fftFeedbackScreen);

}/* FFTFEEDBACK_DEACTIVATE */





/* ************************************************************
 * dio_deactivate() - 
 */
void dio_deactivate(Widget widget, int *tag, 
		    XmAnyCallbackStruct *callback_data)
{/* DIO_DEACTIVATE */
  
  XtUnmanageChild(dioScreen);

}/* DIO_DEACTIVATE */




/* *******************************************************
 * dio_change() - 
 */
void dio_change(Widget widget, int *tag, 
		XmAnyCallbackStruct *callback_data)
{/* DIO_CHANGE */
      
  if(*tag==DIODESTROY)
    {
      askThem("Destructive not yet.",FALSE,XmDIALOG_INFORMATION);
      /* destroyIcos = !destroyIcos; */
    }
  else whichIcos = *tag;

}/* DIO_CHANGE */





/* ************************************************************
 * imgconv_screen_deactivate() - Get the imgconv screen
 * off the display.
 */
void imgconv_screen_deactivate(Widget widget, int *tag, 
			       XmAnyCallbackStruct *callback_data)
{/* IMGCONV_SCREEN_DEACTIVATE */
  
  if(imgconv_image_buffer1)
    free(imgconv_image_buffer1);
  if(imgconv_image_buffer_disk)
    free(imgconv_image_buffer_disk);
  if(imgconv_image_buffer_avg)
    free(imgconv_image_buffer_avg);
  imgconv_image_buffer1 = NULL;
  imgconv_image_buffer_disk = 
    imgconv_image_buffer_avg = 
    NULL;
  XtUnmanageChild(imgconvscreen);

}/* IMGCONV_SCREEN_DEACTIVATE */






/* ************************************************************
 * imgconv_tryit_activate() - 
 */
void imgconv_tryit_activate(Widget widget, int *tag, 
			    XmAnyCallbackStruct *callback_data)
{/* IMGCONV_TRYIT_ACTIVATE */
  float *fft1,*fft2,*fft1_half,*fft2_half;
  int idim1,idim2,imgNcols,imgNrows,i,j,k,l,m,cor_pass;
  BYTE *cor_byte_buffer;
  float *c_buffer,*b_buffer,*array,*cor_map,*avg_map;
  float *cor_img,*avg_img;
  int img_num,outWidth,outHeight,yOut;
  float cor_avg,cor_std,radius,scaleFactor;
  float *peaks;
  int nlim, npeak, lowlim, npeak_unique, ncolmp, nrowmp, arg1;
  char *s1;
  int ksign,iform,nfac;
  float thisBackground,thisMin,thisMax;
  purdueHead_t thisHeader;


  if(!imgconv_image_buffer1)
    {/* Can't go on */
      askThem("Must read reference file first.",
	      FALSE,XmDIALOG_INFORMATION);
      goto noWay;
    }/* Can't go on */

  lowlim = (int)coravgLowLimValue;

  imgNcols = purdueHeader.numCols;
  imgNrows = purdueHeader.numRows;
  fft_setdim_local_ (&imgNcols, &imgNrows , 
		     &idim1, &idim2);
  nfac = 2;
  //  fft_setdim_def_ (&imgNcols, &imgNrows , 
  //	     &idim1, &idim2, &nfac);
  printf("Chosen idim1[%d]  idim1[%d]\n",idim1,idim2);

  CK(fft1 = (float *)malloc(sizeof(float)*idim1*idim2*2));
  CK(fft2 = (float *)malloc(sizeof(float)*idim1*idim2*2));
  CK(fft1_half = (float *)malloc(sizeof(float)*idim1*((idim2/2)+1)*2));
  CK(fft2_half = (float *)malloc(sizeof(float)*idim1*((idim2/2)+1)*2));
  
  CK(cor_byte_buffer = (BYTE *)malloc(sizeof(BYTE)*idim1*idim2));
  CK(c_buffer = (float *)malloc(sizeof(float)*idim1*2));
  CK(b_buffer = (float *)malloc(sizeof(float)*idim2*2));
  CK(array = (float *)malloc(sizeof(float)*2*idim1*((idim2/2)+1)));
  CK(cor_map = (float *)malloc(sizeof(float)*idim1*idim2));
  CK(cor_img = (float *)malloc(sizeof(float)*idim1*idim2));

  /* Let's try 50000 for now ... */
  nlim = 50000;
  CK(peaks = (float *)malloc(sizeof(float)*3*nlim));

  ksign = 1;       /* Do a "forward" FFT */
  iform = 1;       /* Data is imaginary */

  for(i=0;i<imgconvNrow;i++)
    for(j=0;j<imgconvNcol;j++)
      cor_img[(i*imgconvNcol)+j] = (float)imgconv_image_buffer_disk[
						  (i*imgconvNcol)+j];
      
  /* We need a new background value */
  calcStats(imgconv_image_buffer_disk,
	    &(thisHeader.img_bkg), 
	    &(thisHeader.img_min),
	    &(thisHeader.img_max), 
	    &(thisHeader.img_avg), 
	    &(thisHeader.img_std),
	    imgconvNcol, imgconvNrow, 
	    imgconvNcol);

  imgconvBkg        = thisHeader.img_bkg; 
  printf("thisBackground = [%f]\n",imgconvBkg );

  thisBackground = imgconvBkg;
  map_fft_fill_robemorien_(cor_img, &imgconvNcol, &imgconvNrow, 
			   fft1, 
			   &idim1, &idim2, &thisBackground);
  fft_2d_rob_(fft1, &idim1, &idim2, &ksign, &iform);
  fft_copy1_ (fft1, fft1_half, &idim1, &idim2);

  if(imgInCore)
    {/* Do for scanned image */
      for(i=0;i<imgNrows;i++)
	for(j=0;j<imgNcols;j++)
	  cor_img[(i*imgNcols)+j] = (float)inCore_buffer[
							 (i*imgNcols)+j];
    }/* Do for scanned image */
  else if(fileType==PACKEDPORTABLE)
    {/* Do for packed */
      float *thisImg;
      thisImg  = packImg[0];
      for(i=0;i<imgNrows;i++)
	for(j=0;j<imgNcols;j++)
	  cor_img[(i*imgNcols)+j] = (float)thisImg[(i*imgNcols)+j];
    }/* Do for packed */
  else
    {/* Can't do */
      sprintf(outStr,"Corr Average only for scanned or packed images.");
      askThem(outStr,FALSE,XmDIALOG_INFORMATION);
      goto noImgconv;
    }/* Can't do */

  thisBackground = purdueHeader.img_bkg;
  map_fft_fill_robemorien_(cor_img, &imgNcols, &imgNrows, 
		    fft2, 
		    &idim1, &idim2, &thisBackground);
  
  fft_2d_rob_(fft2, &idim1, &idim2, &ksign, &iform);
  fft_copy1_ (fft2, fft2_half, &idim1, &idim2);

  cor_pass = 0;
  cor_first_(&idim1, &idim2, array, fft1_half, fft2_half, c_buffer, &cor_pass);
	 

  img_num = 1;
  cor_second_(&idim1, &idim2, array, cor_map, 
	      cor_img, b_buffer,cor_byte_buffer,&img_num, 
	      &cor_avg, &cor_std, &thisMin, &thisMax);

  rearrange_fft_style_(&idim1, &idim2, cor_map);

  /* Now put it on the display */
  outWidth  = idim1;
  outHeight = idim2;
  thisMax = thisMin = cor_map[0];
  for(j=0; j<(outHeight*outWidth); j++)
    {/* Do all cols */
      thisMin = MIN(thisMin,cor_map[j]);
      thisMax = MAX(thisMax,cor_map[j]);
    }/* Do all cols */
  scaleFactor = ((float)tryColors-9.0) / (float)(thisMax-thisMin);
      
#define DUMP_AND_DEBUG 1
#ifdef DUMP_AND_DEBUG
  if(TRUE)
    {/* debug */
      forPifHead_t outHead;
      float fsf,rfsf,ibkg,imin,imax,iavg,istd;
      int i;
      fString_t rWrite,inFileName;

      pif_init_head(&outHead);
      i= 9;
      outHead.luNum = i;
      rWrite.addr = malloc(sizeof(char)*4);
      rWrite.length = 1;
      inFileName.addr = malloc(sizeof(char)*64);
      inFileName.length = 64;

      outHead.mode = 46;
      outHead.numCols = outWidth;
      outHead.numRows = outHeight;
      outHead.nimg = 1;

      calcMapStats(cor_map, &ibkg, &imin, &imax, &iavg, &istd,
		   &outWidth, &outHeight, &(outHead.nimg));
      calcScaleFactor(&fsf, &rfsf, 
		      &int4scaleF, 
		      (imin-1.0), (imax+1.0), iavg, istd,
		      PACKEDPORTABLE);
      outHead.floatScaleFactor = fsf;
      outHead.recipFloatScaleFactor = rfsf;
      outHead.int4scaleF = int4scaleF;
      
      sprintf(rWrite.addr,"w");

      sprintf(inFileName.addr,"cor_map.pif");
      pif_open_(inFileName.addr, &i, rWrite.addr);

      pif_write_gh_(&outHead);
      pif_write_dh_(&outHead);
      pif_write_boxi4_(&outHead,cor_map);
      
      sprintf(outStr,"Debugging file cor_map.pif written.");
      askThem(outStr,FALSE,XmDIALOG_INFORMATION);
    }/* debug */
#endif /*DUMP_AND_DEBUG */

  for(j=0; j<outHeight; j++)
    {/* hieght */
      int thisPixel;
      yOut = j*outWidth;
      for(k=0; k<outWidth; k++)
	{/* Do all cols */
	  thisPixel = (int)((((float)cor_map[yOut+k] -
			      (float)thisMin)*
			     scaleFactor)+9.0);
	  if(hiColor)
	    {/* hi */
	      XPutPixel(image_image1, 1+k+(imgNcols), j,
			pixels_return[thisPixel]);
	    }/* hi */

	}/* Do all cols */
    }/* hieght */
  
  XPutImage(dsply, image_window1, image_gc, image_image1, 
	    0, 0,
	    0, 0,  outWidth*2, outHeight*2);
  XFlush(dsply);

  stuffImgconvImage();

 noImgconv:;


  allDone:;

  free(fft1);
  free(fft2);
  free(fft1_half);
  free(fft2_half);
  free(cor_byte_buffer);
  free(c_buffer);
  free(b_buffer);
  free(array);
  free(cor_map);
  free(cor_img);
  free(peaks);

 noWay:;

}/* IMGCONV_TRYIT_ACTIVATE */



/* ************************************************************
 * imgconv_screen_activate() - Put the imgconvscreen on the display.
 */
void imgconv_screen_activate(Widget widget, int *tag, 
			     XmAnyCallbackStruct *callback_data)
{/* IMGCONV_SCREEN_ACTIVATE */

  /* Put screen on display */
  raiseIt(imgconvscreen);


}/* IMGCONV_SCREEN_ACTIVATE */



/*********************************************************************
 * imgconv_file_activate() - Browse for appropriate file for img 
 * convolution.
 */
void imgconv_file_activate(Widget widget, int *tag, 
			   XmAnyCallbackStruct *callback_data)
{/* IMGCONV_FILE_ACTIVATE */

  readWrite= READING;
  readImgconv = TRUE;

  raiseIt(fileSel_widget);

}/* IMGCONV_FILE_ACTIVATE */




/* ************************************************************
 * imgconv_create() - Remember the imgconv screen for use later.  
 */
void imgconv_create(Widget widget, int *tag, 
		    XmAnyCallbackStruct *callback_data)
{/* IMGCONV_CREATE */
  
  
}/* IMGCONV_CREATE */




/******************************************************
 * read_imgconv ()- Read the imgconv file.
*/
void read_imgconv(char *fName)
{/* READ_IMGCONV */
  int i,j,k,l,m,imgOffset,beginParms,convTsb;
  XmString simpleString;
  BYTE *tmpBuffer;
  forPifHead_t thisHeader;
  fString_t rWrite,inFile,descStr,inFileName;
  FILE *fIn;

  readImgconv = FALSE;

  i= 4;
  thisHeader.luNum = i;
  rWrite.addr = malloc(sizeof(char)*4);
  rWrite.length = 1;
  descStr.addr = malloc(sizeof(char)*64);
  descStr.length = 32;
  inFileName.addr = malloc(sizeof(char)*64);
  inFileName.length = 64;
  sprintf(rWrite.addr,"r");

  sprintf(inFileName.addr,fName);
  pif_open_(inFileName.addr, &i, rWrite.addr);

  pif_read_gh_(&thisHeader);
  pifMode               = thisHeader.mode;

  if(pifMode!=0 && pifMode!=6 && pifMode!=46)
    {/* NOt !! */
      sprintf(outStr,"Invalid PIF mode [%d]\n",pifMode);
      askThem(outStr,FALSE,XmDIALOG_INFORMATION);
      goto noRead;
    }/* NOt !! */

  /* For fortran */
  thisHeader.imgNum = 1;
  pif_read_dh_(&thisHeader);
  
  imgOffset  = thisHeader.imgOffset; 
  pifMode    = thisHeader.mode; 
  imgconvNcol       = thisHeader.numCols; 
  imgconvNrow       = thisHeader.numRows; 
  imgconvBkg        = thisHeader.img_bkg; 

  if(imgconv_image_buffer1)
    free(imgconv_image_buffer1);
  if(imgconv_image_buffer_disk)
    free(imgconv_image_buffer_disk);
  if(imgconv_image_buffer_avg)
    free(imgconv_image_buffer_avg);
  CK(tmpBuffer = (BYTE *)malloc(sizeof(BYTE)*imgconvNcol*imgconvNrow));
  CK(imgconv_image_buffer1 = (float *)malloc(sizeof(float)*
					      imgconvNcol*imgconvNrow));
  CK(imgconv_image_buffer_disk = (float *)malloc(sizeof(float)*
					      imgconvNcol*imgconvNrow));
  CK(imgconv_image_buffer_avg = (float *)malloc(sizeof(float)*
					      imgconvNcol*imgconvNrow));
  
  
  if(pifMode==6)
    {/* byte */
      pif_read_byte_image(&(thisHeader.luNum),
			  tmpBuffer,
			  &imgOffset,
			  &(thisHeader.numCols),
			  &(thisHeader.numRows),
			  &(thisHeader.numSecs));
      for(i=0;i<thisHeader.numRows;i++)
	for(j=0;j<thisHeader.numCols;j++)
	  imgconv_image_buffer1[((i*thisHeader.numCols)+j)] = (float)
	    tmpBuffer[((i*thisHeader.numCols)+j)];
      free(tmpBuffer);
    }/* byte */
  else if(pifMode==46)
    pif_read_boxi4_(&thisHeader,imgconv_image_buffer1);

  pif_close_(&i);


  if(pifMode==6 || pifMode==46)
    {/* DO a tmp buffer */
      float *tmpByte_buffer;
      CK(tmpByte_buffer=(float *)malloc(sizeof(float)*
				       thisHeader.numCols*
				       thisHeader.numRows));

      tsbworld_float(imgconv_image_buffer1,
		    tmpByte_buffer,
		    thisHeader.numCols, 
		    thisHeader.numRows,
		    1);
      for(i=0;i<thisHeader.numRows;i++)
	for(j=0;j<thisHeader.numCols;j++)
	  {
	    /* Just stuff the avg for now */
	    imgconv_image_buffer_avg[((i*thisHeader.numCols)+j)]= 
	      imgconv_image_buffer_disk[((i*thisHeader.numCols)+j)]= 
		imgconv_image_buffer1[((i*thisHeader.numCols)+j)]= 
		  tmpByte_buffer[((i*thisHeader.numCols)+j)]; 
	  }
      free(tmpByte_buffer);

    }/* DO a tmp buffer */


  if(imgconvNcol > 512 || imgconvNrow > 512)
    {/* We don't do */
      sprintf(outStr,"Dimensions too large [%d,%d]\n",
	      imgconvNcol,imgconvNrow);
      askThem(outStr,FALSE,XmDIALOG_INFORMATION);
      goto noRead;
    }/* We don't do */

  sprintf(outStr,"Image #2:  %s",
	  fName
	  );
  createSimpleString(&simpleString,outStr);
  XtSetArg(arglst[0], XmNlabelString, simpleString);
  XtSetValues(imgconvLabelW, arglst, 1); 
  XFlush(dsply);
  XmStringFree(simpleString); 
  
  stuffImgconvImage();


 noRead:;

  
  
}/* READ_IMGCONV */






/******************************************************
 * stuffImgconvImage ()- Put img convolution img on the
 * display.
*/
void stuffImgconvImage()
{/* STUFFIMGCONVIMAGE */
  int i,j,k,l,m;
  XmString simpleString;
  char *s1;
  float scaleFactor,minF,maxF;

  if(imgconv_image_buffer1)
    {/* Don't do if there's nothing to do */
      /* Find min/max */
      minF = maxF = (float)imgconv_image_buffer1[0];
      for(j=0;j<imgconvNrow;j++)
	{
	  for(k=0;k<imgconvNcol;k++)
	    {
	      maxF = (float)MAX((float)imgconv_image_buffer1[
					   ((j*imgconvNcol)+k)],maxF);
	      minF = (float)MIN((float)imgconv_image_buffer1[
					   ((j*imgconvNcol)+k)],minF);
	    }
	}
      
      scaleFactor = ((float)tryColors-9.0) / (maxF-minF);
      
      for(j=0;j<imgconvNrow;j++)
	{/* disk rows */
	  for(k=0;k<imgconvNcol;k++)
	    {/* disk cols */
	      XPutPixel(image_image1, k, 1+j+purdueHeader.numRows,
			pixels_return[
			  (int)((((float)imgconv_image_buffer1[(
					(j*imgconvNcol)+k)]-
					      minF)*
					     scaleFactor)+8.0)
			]);
	    }/* disk cols */
	}/* disk rows */
      
    }/* Don't do if there's nothing to do */

}/* STUFFIMGCONVIMAGE */








/*  End of robem.c */



